(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 10.4' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[    115899,       3257]
NotebookOptionsPosition[    111092,       3086]
NotebookOutlinePosition[    111777,       3112]
CellTagsIndexPosition[    111644,       3106]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[BoxData[
 RowBox[{
  RowBox[{"SetDirectory", "[", 
   RowBox[{"NotebookDirectory", "[", "]"}], "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell["IBP and the masters", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "LiteRed`"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"**************** \"\>", "\[InvisibleSpace]", 
   StyleBox["\<\"LiteRed v1.82\"\>",
    StripOnInput->False,
    LineColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    FrontFaceColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    BackFaceColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    GraphicsColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    FontWeight->Bold,
    FontColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]]], 
   "\[InvisibleSpace]", "\<\" ********************\\nAuthor: Roman N. Lee, \
Budker Institute of Nuclear Physics, Novosibirsk.\\nRelease Date: \
01.06.2015\\nLiteRed stands for Loop InTEgrals REDuction.\\nThe package is \
designed for the search and application of the Integration-By-Part reduction \
rules. It also contains some other useful tools.\\nSee ?LiteRed`* for a list \
of functions. \"\>"}],
  SequenceForm["**************** ", 
   Style["LiteRed v1.82", {Bold, 
     RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]]}], 
   " ********************\nAuthor: Roman N. Lee, Budker Institute of Nuclear \
Physics, Novosibirsk.\nRelease Date: 01.06.2015\nLiteRed stands for Loop \
InTEgrals REDuction.\nThe package is designed for the search and application \
of the Integration-By-Part reduction rules. It also contains some other \
useful tools.\nSee ?LiteRed`* for a list of functions. "],
  Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"SetDim", "[", "d", "]"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Declare", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"q", ",", "l"}], "}"}], ",", "Vector", ",", 
    RowBox[{"{", 
     RowBox[{"m2", ",", "q2"}], "}"}], ",", "Number"}], "]"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"sp", "[", 
    RowBox[{"q", ",", "q"}], "]"}], "=", "q2"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "NewBasis"}]], "Input"],

Cell[BoxData[
 StyleBox["\<\"NewBasis[\!\(\*\\nStyleBox[\\\"basis\\\", \
\\\"TI\\\"]\),denominators_List,loopMomenta_List] determines a basis.\\nFirst \
argument is a basis basis.\\nSecond argument is a complete set of independent \
denominators. Shortcuts are allowed here: for the massless propagator sp[l,l]\
\\njust put l in the list. sp[a,b] stands for the scalar product of two \
vectors, sp[a]is a shortcut for sp[a,a].\\nThird argument is the list of loop \
momenta.\\nOptions:\\n    GenerateIBP\[Rule]True|False \[LongDash] determines \
whether to generate IBPs afterward,\\n    AnalyzeSectors\[Rule]True|False \
\[LongDash] determines whether to find zero sectors,\\n    FindSymmetries\
\[Rule]True|False \[LongDash] determines whether to find symmetries,\\n    \
DiskSave\[Rule]False|\!\(\*\\nStyleBox[\\\"dirname\\\", \\\"TI\\\"]\) \
\[LongDash] determines the directory in which the basis is saved.\\n    \
Append\[Rule]False|True \[LongDash] determines whether to append irreducible \
numerators if needed.\"\>", "MSG"]], "Print", "PrintUsage",
 CellTags->"Info63762687448-3325587"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NewBasis", "[", 
  RowBox[{"b1", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"sp", "[", 
      RowBox[{"q", "-", "l"}], "]"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", "l", "]"}], "-", "m2"}]}], "}"}], ",", 
   RowBox[{"{", "l", "}"}]}], "]"}]], "Input"],

Cell[BoxData["\<\"Valid basis.\\n    Ds[b1] \[LongDash] denominators,\\n    \
SPs[b1] \[LongDash] scalar products involving loop momenta,\\n    LMs[b1] \
\[LongDash] loop momenta,\\n    EMs[b1] \[LongDash] external momenta,\\n    \
Toj[b1] \[LongDash] rules to transform scalar products to \
denominators.\"\>"], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"GenerateIBP", "[", "b1", "]"}], ";"}]], "Input"],

Cell[BoxData["\<\"Integration-By-Part&Lorentz-Invariance identities are \
generated.\\n    IBP[b1] \[LongDash] integration-by-part identities,\\n    \
LI[b1] \[LongDash] Lorentz invariance identities.\"\>"], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeSectors", "[", "b1", "]"}], ";"}]], "Input"],

Cell[BoxData["\<\"Found 2 zero sectors out of 4.\\n    ZeroSectors[b1] \
\[LongDash] zero sectors,\\n    NonZeroSectors[b1] \[LongDash] nonzero \
sectors,\\n    SimpleSectors[b1] \[LongDash] simple sectors (no nonzero \
subsectors),\\n    BasisSectors[b1] \[LongDash] basis sectors (at least one \
immediate subsector is zero),\\n    ZerojRule[b1] \[LongDash] a rule to \
nullify all zero j[b1\[Ellipsis]],\\n    CutDs[b1] \[LongDash] a flag list of \
cut denominatorsj (1=cut).\"\>"], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FindSymmetries", "[", "b1", "]"}], ";"}]], "Input"],

Cell[BoxData["\<\"Found 0 mapped sectors and 2 unique sectors.\\n    \
UniqueSectors[b1] \[LongDash] unique sectors.\\n    MappedSectors[b1] \
\[LongDash] mapped sectors.\\n    SR[b1][\[Ellipsis]] \[LongDash] symmetry \
relations for j[b1,\[Ellipsis]] from UniqueSectors[b1].\\n    jSymmetries[b1,\
\[Ellipsis]] \[LongDash] symmetry rules for the sector js[b1,\[Ellipsis]] in \
UniqueSectors[b1].\\n    jRules[b1,\[Ellipsis]] \[LongDash] reduction rules \
for j[b1,\[Ellipsis]] from MappedSectors[b1].\"\>"], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"UniqueSectors", "[", "b1", "]"}], "//", 
   RowBox[{"Map", "[", "SolvejSector", "]"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
   RowBox[{"js", "[", 
    RowBox[{"b1", ",", "0", ",", "1"}], "]"}]}],
  SequenceForm["Sector ", 
   LiteRed`js[$CellContext`b1, 0, 1]],
  Editable->False]], "Print"],

Cell[BoxData["\<\"    1 master integrals found:\\nj[b1, 0, 1].\\n    \
jRules[b1, 0, 1] \[LongDash] reduction rules for the sector.\\n    MIs[b1] \
\[LongDash] updated list of the masters.\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
   RowBox[{"js", "[", 
    RowBox[{"b1", ",", "1", ",", "1"}], "]"}]}],
  SequenceForm["Sector ", 
   LiteRed`js[$CellContext`b1, 1, 1]],
  Editable->False]], "Print"],

Cell[BoxData["\<\"    1 master integrals found:\\nj[b1, 1, 1].\\n    \
jRules[b1, 1, 1] \[LongDash] reduction rules for the sector.\\n    MIs[b1] \
\[LongDash] updated list of the masters.\"\>"], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"masters", "=", 
  RowBox[{"MIs", "[", "b1", "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"j", "[", 
    RowBox[{"b1", ",", "0", ",", "1"}], "]"}], ",", 
   RowBox[{"j", "[", 
    RowBox[{"b1", ",", "1", ",", "1"}], "]"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", " ", 
   RowBox[{
    RowBox[{"j", "[", 
     RowBox[{"b1", ",", "0", ",", "2"}], "]"}], ",", " ", 
    RowBox[{"j", "[", 
     RowBox[{"b1", ",", "1", ",", "2"}], "]"}]}], "}"}], "//", 
  "IBPReduce"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "2"}], "+", "d"}], ")"}], " ", 
     RowBox[{"j", "[", 
      RowBox[{"b1", ",", "0", ",", "1"}], "]"}]}], 
    RowBox[{"2", " ", "m2"}]], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "d"}], ")"}], " ", 
       RowBox[{"j", "[", 
        RowBox[{"b1", ",", "0", ",", "1"}], "]"}]}], 
      RowBox[{"2", " ", "m2", " ", 
       RowBox[{"(", 
        RowBox[{"m2", "-", "q2"}], ")"}]}]]}], "+", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "3"}], "+", "d"}], ")"}], " ", 
      RowBox[{"j", "[", 
       RowBox[{"b1", ",", "1", ",", "1"}], "]"}]}], 
     RowBox[{"m2", "-", "q2"}]]}]}], "}"}]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Differential Equations", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Dinv"}]], "Input"],

Cell[BoxData[
 StyleBox["\<\"Dinv[\!\(\*\\nStyleBox[\\\"expr\\\", \\\"TI\\\"]\),p\
\[CenterDot]q] is the differentiation of \!\(\*\\nStyleBox[\\\"expr\\\", \
\\\"TI\\\"]\) with respect to p\[CenterDot]q. It assumes, that other scalar \
products of the external momenta are constants. Pay attention that p\
\[CenterDot]q and q\[CenterDot]p is not exactly the same in the second \
argument. If p\[CenterDot]q is used, the derivative of j[...] is expressed in \
terms of \[PartialD]/\[PartialD]p.\\nExplicit scalar functions of p\
\[CenterDot]q are differentiated correctly, unless p\[CenterDot]q is \
evaluated to something.\\nDinv[\!\(\*\\nStyleBox[\\\"expr\\\", \\\"TI\\\"]\),\
\!\(\*\\nStyleBox[\\\"x\\\", \\\"TI\\\"]\)] gives the derivative of \
\!\(\*\\nStyleBox[\\\"expr\\\", \\\"TI\\\"]\) with respect to the variable \!\
\(\*\\nStyleBox[\\\"x\\\", \\\"TI\\\"]\).\"\>", "MSG"]], "Print", "PrintUsage",
 CellTags->"Info143762687456-3325587"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"dmasters", "=", 
  RowBox[{
   RowBox[{"masters", "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"Dinv", "[", 
       RowBox[{"#", ",", "m2"}], "]"}], "&"}], "]"}]}], "//", 
   "IBPReduce"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "2"}], "+", "d"}], ")"}], " ", 
     RowBox[{"j", "[", 
      RowBox[{"b1", ",", "0", ",", "1"}], "]"}]}], 
    RowBox[{"2", " ", "m2"}]], ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "2"}], "+", "d"}], ")"}], " ", 
       RowBox[{"j", "[", 
        RowBox[{"b1", ",", "0", ",", "1"}], "]"}]}], 
      RowBox[{"2", " ", "m2", " ", 
       RowBox[{"(", 
        RowBox[{"m2", "-", "q2"}], ")"}]}]]}], "+", 
    FractionBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "3"}], "+", "d"}], ")"}], " ", 
      RowBox[{"j", "[", 
       RowBox[{"b1", ",", "1", ",", "1"}], "]"}]}], 
     RowBox[{"m2", "-", "q2"}]]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"M", "=", 
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"Coefficient", "[", 
      RowBox[{"dm", ",", "m"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"dm", ",", "dmasters"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"m", ",", "masters"}], "}"}]}], "]"}], "//", 
   "Factor"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{"-", "2"}], "+", "d"}], 
      RowBox[{"2", " ", "m2"}]], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"-", "2"}], "+", "d"}], 
       RowBox[{"2", " ", "m2", " ", 
        RowBox[{"(", 
         RowBox[{"m2", "-", "q2"}], ")"}]}]]}], ",", 
     FractionBox[
      RowBox[{
       RowBox[{"-", "3"}], "+", "d"}], 
      RowBox[{"m2", "-", "q2"}]]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Put", "[", 
  RowBox[{
   RowBox[{"M", "/.", 
    RowBox[{"d", "\[Rule]", 
     RowBox[{"4", "-", 
      RowBox[{"2", "eps"}]}]}]}], ",", "\"\<de-matrix.m\>\""}], 
  "]"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The epsilon form", "Subsection"],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{
   RowBox[{"Manual", " ", "step"}], ",", " ", 
   RowBox[{"run", ":", " ", 
    RowBox[{
     RowBox[{
      RowBox[{".", 
       RowBox[{"/", "fuchsia"}]}], " ", "reduce", " ", "de"}], "-", 
     RowBox[{"matrix", ".", "m"}], " ", "-", 
     RowBox[{"x", " ", "m2"}], " ", "-", 
     RowBox[{"m", " ", "de"}], "-", 
     RowBox[{"matrix", ".", "eps"}], " ", "-", 
     RowBox[{"t", " ", "de"}], "-", 
     RowBox[{"matrix", ".", "t"}]}]}]}], " ", "*)"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Meps", "=", 
   RowBox[{
    RowBox[{"Get", "[", "\"\<de-matrix.eps\>\"", "]"}], "//", 
    RowBox[{
     RowBox[{"Apart", "[", 
      RowBox[{"#", ",", "m2"}], "]"}], "&"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Meps", "//", "MatrixForm"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"S", "=", 
   RowBox[{"Meps", "/", "eps"}]}], ";"}]}], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", 
       FractionBox["eps", "m2"]}], "0"},
     {
      RowBox[{
       RowBox[{"-", 
        FractionBox["eps", 
         RowBox[{"m2", " ", "q2"}]]}], "+", 
       FractionBox["eps", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", "q2"}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", "eps"}], 
        RowBox[{"m2", "-", "q2"}]]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"T", "=", 
   RowBox[{"Get", "[", "\"\<de-matrix.t\>\"", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"T", "//", "MatrixForm"}]}], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         RowBox[{"2", " ", "eps"}]}], ")"}], " ", "m2"}], "0"},
     {
      RowBox[{
       RowBox[{"-", "1"}], "+", "eps"}], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "eps"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"m2", "-", "q2"}], ")"}]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell[BoxData[""], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The solution as a series in epsilon, via Mathematica", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", "J", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"J", "[", "0", "]"}], "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"0", ",", "1"}], "]"}], ",", 
    RowBox[{"C", "[", 
     RowBox[{"0", ",", "2"}], "]"}]}], "}"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"J", "[", "k_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"S", ".", 
     RowBox[{"J", "[", 
      RowBox[{"k", "-", "1"}], "]"}]}], "//", 
    RowBox[{
     RowBox[{"Integrate", "[", 
      RowBox[{"#", ",", "m2"}], "]"}], "&"}]}], "//", 
   RowBox[{
    RowBox[{"#", "+", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"k", ",", "1"}], "]"}], ",", 
       RowBox[{"C", "[", 
        RowBox[{"k", ",", "2"}], "]"}]}], "}"}]}], "&"}]}]}]}], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"C", "[", 
    RowBox[{"0", ",", "1"}], "]"}], ",", 
   RowBox[{"C", "[", 
    RowBox[{"0", ",", "2"}], "]"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "1", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"1", ",", "1"}], "]"}], "-", 
     RowBox[{
      RowBox[{"log", "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}]}]}], ",", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"log", "(", "m2", ")"}], " ", 
        RowBox[{"C", "[", 
         RowBox[{"0", ",", "1"}], "]"}]}], "q2"]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"log", "(", 
        RowBox[{"m2", "-", "q2"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}], "-", 
         RowBox[{"2", " ", "q2", " ", 
          RowBox[{"C", "[", 
           RowBox[{"0", ",", "2"}], "]"}]}]}], ")"}]}], "q2"], "+", 
     RowBox[{"C", "[", 
      RowBox[{"1", ",", "2"}], "]"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "2", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{
       SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}]}], "-", 
     RowBox[{
      RowBox[{"log", "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "1"}], "]"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"2", ",", "1"}], "]"}]}], ",", 
    RowBox[{
     FractionBox[
      RowBox[{
       RowBox[{
        TemplateBox[{"2",FractionBox["m2", "q2"]},
         "PolyLog"], " ", 
        RowBox[{"C", "[", 
         RowBox[{"0", ",", "1"}], "]"}]}], "+", 
       RowBox[{
        RowBox[{"log", "(", "m2", ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"log", "(", 
            RowBox[{"1", "-", 
             FractionBox["m2", "q2"]}], ")"}], " ", 
           RowBox[{"C", "[", 
            RowBox[{"0", ",", "1"}], "]"}]}], "-", 
          RowBox[{"C", "[", 
           RowBox[{"1", ",", "1"}], "]"}]}], ")"}]}], "+", 
       RowBox[{
        RowBox[{"log", "(", 
         RowBox[{"m2", "-", "q2"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"log", "(", 
             RowBox[{"m2", "-", "q2"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], "-", 
              RowBox[{"2", " ", "q2", " ", 
               RowBox[{"C", "[", 
                RowBox[{"0", ",", "2"}], "]"}]}]}], ")"}]}]}], "-", 
          RowBox[{"2", " ", "q2", " ", 
           RowBox[{"C", "[", 
            RowBox[{"1", ",", "2"}], "]"}]}], "+", 
          RowBox[{"C", "[", 
           RowBox[{"1", ",", "1"}], "]"}]}], ")"}]}], "+", 
       RowBox[{
        FractionBox["1", "2"], " ", 
        RowBox[{
         SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
        RowBox[{"C", "[", 
         RowBox[{"0", ",", "1"}], "]"}]}]}], "q2"], "+", 
     RowBox[{"C", "[", 
      RowBox[{"2", ",", "2"}], "]"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "3", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       FractionBox["1", "6"], " ", 
       RowBox[{
        SuperscriptBox["log", "3"], "(", "m2", ")"}], " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}]}]}], "+", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{
       SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "1"}], "]"}]}], "-", 
     RowBox[{
      RowBox[{"log", "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "1"}], "]"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"3", ",", "1"}], "]"}]}], ",", 
    RowBox[{
     RowBox[{
      FractionBox["1", 
       RowBox[{"2", " ", "q2"}]], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         TemplateBox[{"3",RowBox[{"1", "-", 
             FractionBox["m2", "q2"]}]},
          "PolyLog"], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "+", 
        RowBox[{"2", " ", 
         TemplateBox[{"3",FractionBox["m2", "q2"]},
          "PolyLog"], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "-", 
        RowBox[{"4", " ", 
         TemplateBox[{"2",RowBox[{"1", "-", 
             FractionBox["m2", "q2"]}]},
          "PolyLog"], " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "+", 
        RowBox[{"2", " ", 
         TemplateBox[{"2",FractionBox["m2", "q2"]},
          "PolyLog"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"2", " ", 
             RowBox[{"log", "(", 
              RowBox[{"1", "-", 
               FractionBox["m2", "q2"]}], ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}]}]}], "-", 
           RowBox[{
            RowBox[{"log", "(", "m2", ")"}], " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}]}], "+", 
           RowBox[{"C", "[", 
            RowBox[{"1", ",", "1"}], "]"}]}], ")"}]}], "-", 
        RowBox[{
         FractionBox["8", "3"], " ", 
         RowBox[{
          SuperscriptBox["log", "3"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "+", 
        RowBox[{
         FractionBox["16", "3"], " ", "q2", " ", 
         RowBox[{
          SuperscriptBox["log", "3"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "2"}], "]"}]}], "+", 
        RowBox[{"4", " ", 
         RowBox[{
          SuperscriptBox["log", "3"], "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "-", 
        RowBox[{"8", " ", "q2", " ", 
         RowBox[{
          SuperscriptBox["log", "3"], "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "2"}], "]"}]}], "-", 
        RowBox[{
         RowBox[{
          SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "-", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", "m2", ")"}], " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "-", 
        RowBox[{"2", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "1"}], "]"}]}], "+", 
        RowBox[{"4", " ", "q2", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "2"}], "]"}]}], "-", 
        RowBox[{"4", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "1"}], "]"}]}], "+", 
        RowBox[{"8", " ", "q2", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "2"}], "]"}]}], "-", 
        RowBox[{"8", " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "+", 
        RowBox[{"16", " ", "q2", " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "2"}], "]"}]}], "+", 
        RowBox[{"8", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"log", "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "-", 
        RowBox[{"16", " ", "q2", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"log", "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "2"}], "]"}]}], "-", 
        RowBox[{"2", " ", 
         RowBox[{
          SuperscriptBox["log", "2"], "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"log", "(", 
          FractionBox["m2", "q2"], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", "m2", ")"}], " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "1"}], "]"}]}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"2", ",", "1"}], "]"}]}], "-", 
        RowBox[{"4", " ", "q2", " ", 
         RowBox[{"log", "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"2", ",", "2"}], "]"}]}], "+", 
        RowBox[{"4", " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"log", "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "1"}], "]"}]}], "-", 
        RowBox[{"8", " ", "q2", " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", 
           FractionBox["m2", "q2"]}], ")"}], " ", 
         RowBox[{"log", "(", 
          RowBox[{"m2", "-", "q2"}], ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "2"}], "]"}]}], "-", 
        RowBox[{
         FractionBox["1", "3"], " ", 
         RowBox[{
          SuperscriptBox["log", "3"], "(", "m2", ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"0", ",", "1"}], "]"}]}], "+", 
        RowBox[{
         RowBox[{
          SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"1", ",", "1"}], "]"}]}], "-", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", "m2", ")"}], " ", 
         RowBox[{"C", "[", 
          RowBox[{"2", ",", "1"}], "]"}]}]}], ")"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"3", ",", "2"}], "]"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "4", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{
      FractionBox["1", "24"], " ", 
      RowBox[{
       SuperscriptBox["log", "4"], "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}]}], "-", 
     RowBox[{
      FractionBox["1", "6"], " ", 
      RowBox[{
       SuperscriptBox["log", "3"], "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "1"}], "]"}]}], "+", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{
       SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "1"}], "]"}]}], "-", 
     RowBox[{
      RowBox[{"log", "(", "m2", ")"}], " ", 
      RowBox[{"C", "[", 
       RowBox[{"3", ",", "1"}], "]"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"4", ",", "1"}], "]"}]}], ",", 
    RowBox[{
     RowBox[{"\[Integral]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox["eps", 
              RowBox[{"q2", " ", 
               RowBox[{"(", 
                RowBox[{"m2", "-", "q2"}], ")"}]}]], "-", 
             FractionBox["eps", 
              RowBox[{"m2", " ", "q2"}]]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              RowBox[{
               FractionBox["1", "6"], " ", 
               RowBox[{
                SuperscriptBox["log", "3"], "(", "m2", ")"}], " ", 
               RowBox[{"C", "[", 
                RowBox[{"0", ",", "1"}], "]"}]}]}], "+", 
             RowBox[{
              FractionBox["1", "2"], " ", 
              RowBox[{
               SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}]}], "-", 
             RowBox[{
              RowBox[{"log", "(", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "1"}], "]"}]}], "+", 
             RowBox[{"C", "[", 
              RowBox[{"3", ",", "1"}], "]"}]}], ")"}]}], "eps"], "-", 
         RowBox[{
          FractionBox["1", 
           RowBox[{"m2", "-", "q2"}]], "2", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             FractionBox["1", 
              RowBox[{"2", " ", "q2"}]], 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"4", " ", 
                TemplateBox[{"3",RowBox[{"1", "-", 
                    FractionBox["m2", "q2"]}]},
                 "PolyLog"], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                TemplateBox[{"3",FractionBox["m2", "q2"]},
                 "PolyLog"], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "-", 
               RowBox[{"4", " ", 
                TemplateBox[{"2",RowBox[{"1", "-", 
                    FractionBox["m2", "q2"]}]},
                 "PolyLog"], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                TemplateBox[{"2",FractionBox["m2", "q2"]},
                 "PolyLog"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"-", 
                   RowBox[{"2", " ", 
                    RowBox[{"log", "(", 
                    RowBox[{"1", "-", 
                    FractionBox["m2", "q2"]}], ")"}], " ", 
                    RowBox[{"C", "[", 
                    RowBox[{"0", ",", "1"}], "]"}]}]}], "-", 
                  RowBox[{
                   RowBox[{"log", "(", "m2", ")"}], " ", 
                   RowBox[{"C", "[", 
                    RowBox[{"0", ",", "1"}], "]"}]}], "+", 
                  RowBox[{"C", "[", 
                   RowBox[{"1", ",", "1"}], "]"}]}], ")"}]}], "-", 
               RowBox[{
                FractionBox["8", "3"], " ", 
                RowBox[{
                 SuperscriptBox["log", "3"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "+", 
               RowBox[{
                FractionBox["16", "3"], " ", "q2", " ", 
                RowBox[{
                 SuperscriptBox["log", "3"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "2"}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                RowBox[{
                 SuperscriptBox["log", "3"], "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "-", 
               RowBox[{"8", " ", "q2", " ", 
                RowBox[{
                 SuperscriptBox["log", "3"], "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "2"}], "]"}]}], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "-", 
               RowBox[{"2", " ", 
                RowBox[{"log", "(", "m2", ")"}], " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "-", 
               RowBox[{"2", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "1"}], "]"}]}], "+", 
               RowBox[{"4", " ", "q2", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "2"}], "]"}]}], "-", 
               RowBox[{"4", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "1"}], "]"}]}], "+", 
               RowBox[{"8", " ", "q2", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "2"}], "]"}]}], "-", 
               RowBox[{"8", " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "+", 
               RowBox[{"16", " ", "q2", " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "2"}], "]"}]}], "+", 
               RowBox[{"8", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "-", 
               RowBox[{"16", " ", "q2", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "2"}], "]"}]}], "-", 
               RowBox[{"2", " ", 
                RowBox[{
                 SuperscriptBox["log", "2"], "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"log", "(", 
                 FractionBox["m2", "q2"], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                RowBox[{"log", "(", "m2", ")"}], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "1"}], "]"}]}], "+", 
               RowBox[{"2", " ", 
                RowBox[{"log", "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"2", ",", "1"}], "]"}]}], "-", 
               RowBox[{"4", " ", "q2", " ", 
                RowBox[{"log", "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"2", ",", "2"}], "]"}]}], "+", 
               RowBox[{"4", " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "1"}], "]"}]}], "-", 
               RowBox[{"8", " ", "q2", " ", 
                RowBox[{"log", "(", 
                 RowBox[{"1", "-", 
                  FractionBox["m2", "q2"]}], ")"}], " ", 
                RowBox[{"log", "(", 
                 RowBox[{"m2", "-", "q2"}], ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "2"}], "]"}]}], "-", 
               RowBox[{
                FractionBox["1", "3"], " ", 
                RowBox[{
                 SuperscriptBox["log", "3"], "(", "m2", ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"0", ",", "1"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox["log", "2"], "(", "m2", ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"1", ",", "1"}], "]"}]}], "-", 
               RowBox[{"2", " ", 
                RowBox[{"log", "(", "m2", ")"}], " ", 
                RowBox[{"C", "[", 
                 RowBox[{"2", ",", "1"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{"C", "[", 
             RowBox[{"3", ",", "2"}], "]"}]}], ")"}]}]}], ")"}], 
       RowBox[{"\[DifferentialD]", "m2"}]}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"4", ",", "2"}], "]"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The solution as a series in epsilon, via GPLs", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"MakeBoxes", "[", 
   RowBox[{
    RowBox[{"Hlog", "[", 
     RowBox[{"x_", ",", "w_List"}], "]"}], ",", 
    RowBox[{"form", ":", 
     RowBox[{"(", "TraditionalForm", ")"}]}]}], "]"}], ":=", 
  RowBox[{"RowBox", "[", 
   RowBox[{"{", 
    RowBox[{"\"\<G\>\"", ",", "\"\<(\>\"", ",", 
     RowBox[{"RowBox", "[", 
      RowBox[{"Riffle", "[", 
       RowBox[{
        RowBox[{"Map", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"MakeBoxes", "[", 
            RowBox[{"#", ",", "form"}], "]"}], "&"}], ",", "w"}], "]"}], ",", 
        "\"\<,\>\""}], "]"}], "]"}], ",", "\"\<;\>\"", ",", 
     RowBox[{"MakeBoxes", "[", 
      RowBox[{"x", ",", "form"}], "]"}], ",", "\"\<)\>\""}], "}"}], 
   "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Hlog", "[", 
  RowBox[{"x", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"G", "(", 
  RowBox[{"a", ",", "b", ",", "c"}], ";", "x", ")"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Hlog", "[", 
   RowBox[{"x", ",", 
    RowBox[{"{", 
     RowBox[{"a", ",", "b", ",", "c"}], "}"}]}], "]"}], "//", 
  "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"G", "(", 
   RowBox[{"a", ",", "b", ",", "c"}], ";", "x", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"HlogSimpleIntegrate", "[", 
   RowBox[{"ex_", ",", "x_"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"symbol", ",", "a", ",", "b", ",", "w"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"ex", "//", 
         RowBox[{
          RowBox[{"Apart", "[", 
           RowBox[{"#", ",", "x"}], "]"}], "&"}]}], "//", 
        "\[IndentingNewLine]", 
        RowBox[{"ReplaceAll", "[", 
         RowBox[{
          RowBox[{"1", "/", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"a_.", "*", "x"}], "+", "b_."}], ")"}]}], 
          "\[RuleDelayed]", 
          RowBox[{
           RowBox[{"1", "/", "a"}], " ", 
           RowBox[{"symbol", "[", 
            RowBox[{
             RowBox[{"-", "b"}], "/", "a"}], "]"}]}]}], "]"}]}], "//", 
       "\[IndentingNewLine]", "Expand"}], "//", "\[IndentingNewLine]", 
      RowBox[{"ReplaceAll", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"symbol", "[", "a_", "]"}], "*", 
         RowBox[{"Hlog", "[", 
          RowBox[{"x", ",", 
           RowBox[{"{", "w___", "}"}]}], "]"}]}], "\[RuleDelayed]", 
        RowBox[{"Hlog", "[", 
         RowBox[{"x", ",", 
          RowBox[{"{", 
           RowBox[{"a", ",", "w"}], "}"}]}], "]"}]}], "]"}]}], "//", 
     "\[IndentingNewLine]", 
     RowBox[{"ReplaceAll", "[", 
      RowBox[{
       RowBox[{"symbol", "[", "a_", "]"}], "\[RuleDelayed]", 
       RowBox[{"Hlog", "[", 
        RowBox[{"x", ",", 
         RowBox[{"{", "a", "}"}]}], "]"}]}], "]"}]}]}], "\[IndentingNewLine]",
    "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"HlogSimpleIntegrate", "[", "x_", "]"}], ":=", 
  RowBox[{
   RowBox[{"HlogSimpleIntegrate", "[", 
    RowBox[{"#", ",", "x"}], "]"}], "&"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"HlogSimpleIntegrate", "[", 
   RowBox[{"ex_List", ",", "x_"}], "]"}], ":=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"HlogSimpleIntegrate", "[", "x", "]"}], ",", "ex"}], 
   "]"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"1", "/", "x"}], " ", 
   RowBox[{"Hlog", "[", 
    RowBox[{"x", ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "2"}], "}"}]}], "]"}]}], "//", 
  RowBox[{"HlogSimpleIntegrate", "[", "x", "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"G", "(", 
  RowBox[{"0", ",", "1", ",", "2"}], ";", "x", ")"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", "J", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"J", "[", "0", "]"}], "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"0", ",", "1"}], "]"}], ",", 
    RowBox[{"C", "[", 
     RowBox[{"0", ",", "2"}], "]"}]}], "}"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"J", "[", "k_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"S", ".", 
     RowBox[{"J", "[", 
      RowBox[{"k", "-", "1"}], "]"}]}], "//", 
    RowBox[{"HlogSimpleIntegrate", "[", "m2", "]"}]}], "//", 
   RowBox[{
    RowBox[{"#", "+", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"k", ",", "1"}], "]"}], ",", 
       RowBox[{"C", "[", 
        RowBox[{"k", ",", "2"}], "]"}]}], "}"}]}], "&"}]}]}]}], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"C", "[", 
    RowBox[{"0", ",", "1"}], "]"}], ",", 
   RowBox[{"C", "[", 
    RowBox[{"0", ",", "2"}], "]"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"J", "[", "1", "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"1", ",", "1"}], "]"}], "-", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"0", ",", "1"}], "]"}], " ", 
     RowBox[{"G", "(", "0", ";", "m2", ")"}]}]}], ",", 
   RowBox[{
    RowBox[{"C", "[", 
     RowBox[{"1", ",", "2"}], "]"}], "-", 
    FractionBox[
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", "0", ";", "m2", ")"}]}], "q2"], "+", 
    FractionBox[
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
    RowBox[{"2", " ", 
     RowBox[{"C", "[", 
      RowBox[{"0", ",", "2"}], "]"}], " ", 
     RowBox[{"G", "(", "q2", ";", "m2", ")"}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "2", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}]}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"2", ",", "1"}], "]"}]}], ",", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
        RowBox[{"C", "[", 
         RowBox[{"1", ",", "1"}], "]"}]}], "q2"]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
     RowBox[{"2", " ", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "+", 
     RowBox[{"4", " ", 
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"2", ",", "2"}], "]"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "3", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{
       RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}]}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "-", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"3", ",", "1"}], "]"}]}], ",", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
        RowBox[{"C", "[", 
         RowBox[{"2", ",", "1"}], "]"}]}], "q2"]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
     RowBox[{"2", " ", 
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "+", 
     RowBox[{"4", " ", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "-", 
     
     RowBox[{"8", " ", 
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
     RowBox[{"C", "[", 
      RowBox[{"3", ",", "2"}], "]"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"J", "[", "10", "]"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"10", ",", "1"}], "]"}], "-", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"9", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", "0", ";", "m2", ")"}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"8", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "-", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"7", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"6", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "-", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"5", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], ";", "m2", 
       ")"}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"4", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], ";", 
       "m2", ")"}]}], "-", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"3", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], ";",
        "m2", ")"}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
        "0"}], ";", "m2", ")"}]}], "-", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
        "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
     RowBox[{
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "1"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
        "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}]}], ",", 
    RowBox[{
     RowBox[{"C", "[", 
      RowBox[{"10", ",", "2"}], "]"}], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"9", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", "0", ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"9", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
     RowBox[{"2", " ", 
      RowBox[{"C", "[", 
       RowBox[{"9", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"8", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"8", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"8", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "+", 
     RowBox[{"4", " ", 
      RowBox[{"C", "[", 
       RowBox[{"8", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"7", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"7", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"7", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"7", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "-", 
     
     RowBox[{"8", " ", 
      RowBox[{"C", "[", 
       RowBox[{"7", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"6", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], 
      "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"6", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], 
      "q2"], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"6", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], 
      "q2"], "+", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"6", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], 
      "q2"], "-", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"6", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], 
      "q2"], "+", 
     RowBox[{"16", " ", 
      RowBox[{"C", "[", 
       RowBox[{"6", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], 
     "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"5", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], ";", "m2", 
        ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"5", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0", ",", "0", ",", "0", ",", "0"}], ";", "m2", 
        ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"5", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "0", ",", "0", ",", "0"}], ";", "m2", 
        ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"5", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "0", ",", "0"}], ";", "m2", 
        ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"5", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "0"}], ";", "m2", 
        ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"16", " ", 
       RowBox[{"C", "[", 
        RowBox[{"5", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2"}], ";", "m2",
         ")"}]}], "q2"], "-", 
     RowBox[{"32", " ", 
      RowBox[{"C", "[", 
       RowBox[{"5", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2"}], ";", "m2", 
       ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], ";", 
        "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], ";",
         "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"16", " ", 
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"32", " ", 
       RowBox[{"C", "[", 
        RowBox[{"4", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2"}],
         ";", "m2", ")"}]}], "q2"], "+", 
     RowBox[{"64", " ", 
      RowBox[{"C", "[", 
       RowBox[{"4", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{"q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2"}], 
       ";", "m2", ")"}]}], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0"}], 
        ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0"}],
         ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"16", " ", 
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", 
         "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"32", " ", 
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"64", " ", 
       RowBox[{"C", "[", 
        RowBox[{"3", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2"}], ";", "m2", ")"}]}], "q2"], "-", 
     RowBox[{"128", " ", 
      RowBox[{"C", "[", 
       RowBox[{"3", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
        "q2"}], ";", "m2", ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
          "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", 
         ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", 
         ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", 
         ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"16", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", 
         ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"32", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0",
          ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"64", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"128", " ", 
       RowBox[{"C", "[", 
        RowBox[{"2", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "+", 
     RowBox[{"256", " ", 
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2",
         ",", "q2"}], ";", "m2", ")"}]}], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
          "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", 
         ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", 
         ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", 
         ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"16", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", 
         ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"32", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0",
          ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"64", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
     FractionBox[
      RowBox[{"128", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"256", " ", 
       RowBox[{"C", "[", 
        RowBox[{"1", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "-", 
     RowBox[{"512", " ", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2",
         ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", ",",
          "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0", 
         ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", ",", "0", 
         ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"8", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", ",", "0", 
         ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"16", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0", ",", "0", 
         ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"32", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "0",
          ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"64", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
     FractionBox[
      RowBox[{"128", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], 
     "+", 
     FractionBox[
      RowBox[{"256", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], 
     "-", 
     FractionBox[
      RowBox[{"512", " ", 
       RowBox[{"C", "[", 
        RowBox[{"0", ",", "1"}], "]"}], " ", 
       RowBox[{"G", "(", 
        RowBox[{
        "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", 
         "q2", ",", "q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], 
     "+", 
     RowBox[{"1024", " ", 
      RowBox[{"C", "[", 
       RowBox[{"0", ",", "2"}], "]"}], " ", 
      RowBox[{"G", "(", 
       RowBox[{
       "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2", ",", "q2",
         ",", "q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}]}]}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Jseries", "=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"J", "[", "0", "]"}], "+", 
      RowBox[{
       RowBox[{"J", "[", "1", "]"}], "eps"}], "+", 
      RowBox[{
       RowBox[{"J", "[", "2", "]"}], 
       RowBox[{"eps", "^", "2"}]}], "+", 
      RowBox[{
       RowBox[{"J", "[", "3", "]"}], 
       RowBox[{"eps", "^", "3"}]}], "+", 
      RowBox[{
       RowBox[{"O", "[", "eps", "]"}], "^", "4"}]}], ")"}], "*", 
    RowBox[{"eps", "^", 
     RowBox[{"-", "2"}]}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Jseries", "//", "TableForm"}], "//", "TraditionalForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       InterpretationBox[
        RowBox[{
         FractionBox[
          RowBox[{"C", "[", 
           RowBox[{"0", ",", "1"}], "]"}], 
          SuperscriptBox["eps", "2"]], "+", 
         FractionBox[
          RowBox[{
           RowBox[{"C", "[", 
            RowBox[{"1", ",", "1"}], "]"}], "-", 
           RowBox[{
            RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}]}]}], "eps"], "+", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{
             RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}]}]}], "+", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "+", 
           RowBox[{"C", "[", 
            RowBox[{"2", ",", "1"}], "]"}]}], ")"}], "+", 
         RowBox[{"eps", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{
              RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "1"}], "]"}]}]}], "+", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "-", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"C", "[", 
             RowBox[{"3", ",", "1"}], "]"}]}], ")"}]}], "+", 
         InterpretationBox[
          RowBox[{"O", "(", 
           SuperscriptBox["eps", "2"], ")"}],
          SeriesData[$CellContext`eps, 0, {}, -2, 2, 1],
          Editable->False]}],
        SeriesData[$CellContext`eps, 0, {
          C[0, 1], C[1, 1] - C[0, 1] $CellContext`Hlog[$CellContext`m2, {0}], 
          C[2, 1] - C[1, 1] $CellContext`Hlog[$CellContext`m2, {0}] + 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}], C[3, 1] - 
          C[2, 1] $CellContext`Hlog[$CellContext`m2, {0}] + 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}] - 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0, 0}]}, -2, 2, 1],
        Editable->False]},
      {
       InterpretationBox[
        RowBox[{
         FractionBox[
          RowBox[{"C", "[", 
           RowBox[{"0", ",", "2"}], "]"}], 
          SuperscriptBox["eps", "2"]], "+", 
         FractionBox[
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{
              RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}]}], "q2"]}], "+", 
           FractionBox[
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
           RowBox[{"2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
           RowBox[{"C", "[", 
            RowBox[{"1", ",", "2"}], "]"}]}], "eps"], "+", 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
           RowBox[{"2", " ", "q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "-", 
           RowBox[{"2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
           RowBox[{"4", " ", "q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
           RowBox[{
            RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "1"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "+", 
           RowBox[{"q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"2", ",", "2"}], "]"}]}]}], "q2"], "+", 
         FractionBox[
          RowBox[{"eps", " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
             RowBox[{"2", " ", "q2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "2"}], "]"}], " ", 
              RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
             RowBox[{"4", " ", "q2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "2"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "+", 
             RowBox[{"4", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
             
             RowBox[{"8", " ", "q2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "2"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
             
             RowBox[{
              RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "1"}], "]"}]}], "+", 
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "-", 
             RowBox[{
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
             RowBox[{"q2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"3", ",", "2"}], "]"}]}]}], ")"}]}], "q2"], "+", 
         InterpretationBox[
          RowBox[{"O", "(", 
           SuperscriptBox["eps", "2"], ")"}],
          SeriesData[$CellContext`eps, 0, {}, -2, 2, 1],
          Editable->False]}],
        SeriesData[$CellContext`eps, 0, {
          C[0, 2], C[1, 2] - $CellContext`q2^(-1) 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {
            0}] + $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 2 
          C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}], \
$CellContext`q2^(-1) ($CellContext`q2 C[2, 2] - 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {0}] + 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
           2 $CellContext`q2 
           C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}] + 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0}] - 
           2 C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
           4 $CellContext`q2 
            C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}]), $CellContext`q2^(-1) ($CellContext`q2 C[3, 2] - 
           C[2, 1] $CellContext`Hlog[$CellContext`m2, {0}] + 
           C[2, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
           2 $CellContext`q2 
           C[2, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}] + 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0}] - 
           2 C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
           4 $CellContext`q2 
            C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] - C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0, 0}] - 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0, 
              0}] - 2 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, 0}] + 
           4 C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, $CellContext`q2}] - 8 $CellContext`q2 
           C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, $CellContext`q2}])}, -2, 2, 1],
        Editable->False]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Iseries", "=", 
   RowBox[{"T", ".", "Jseries"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Iseries", "//", "ExpandAll"}], "//", "TableForm"}], "//", 
  "TraditionalForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       InterpretationBox[
        RowBox[{
         RowBox[{"-", 
          FractionBox[
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}]}], 
           SuperscriptBox["eps", "2"]]}], "+", 
         FractionBox[
          RowBox[{
           RowBox[{"m2", " ", 
            RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}]}], "+", 
           RowBox[{"2", " ", "m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}]}], "-", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "1"}], "]"}]}]}], "eps"], "+", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"m2", " ", 
            RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "1"}], "]"}]}], "-", 
           RowBox[{"2", " ", "m2", " ", 
            RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}]}], "-", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "+", 
           RowBox[{"2", " ", "m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "1"}], "]"}]}], "-", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"2", ",", "1"}], "]"}]}]}], ")"}], "+", 
         RowBox[{"eps", " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"m2", " ", 
             RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "1"}], "]"}]}], "-", 
            RowBox[{"2", " ", "m2", " ", 
             RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}]}], "+", 
            RowBox[{"2", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "-", 
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"2", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "1"}], "]"}]}], "-", 
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"3", ",", "1"}], "]"}]}]}], ")"}]}], "+", 
         InterpretationBox[
          RowBox[{"O", "(", 
           SuperscriptBox["eps", "2"], ")"}],
          SeriesData[$CellContext`eps, 0, {}, -2, 2, 1],
          Editable->False]}],
        SeriesData[$CellContext`eps, 
         0, {-$CellContext`m2 C[0, 1], 
          2 $CellContext`m2 C[0, 1] - $CellContext`m2 
          C[1, 1] + $CellContext`m2 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0}], 
          2 $CellContext`m2 C[1, 1] - $CellContext`m2 C[2, 1] - 
          2 $CellContext`m2 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {0}] + $CellContext`m2 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {0}] - $CellContext`m2 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}], 
          2 $CellContext`m2 C[2, 1] - $CellContext`m2 C[3, 1] - 
          2 $CellContext`m2 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {0}] + $CellContext`m2 
           C[2, 1] $CellContext`Hlog[$CellContext`m2, {0}] + 
          2 $CellContext`m2 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 
             0}] - $CellContext`m2 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {0, 
            0}] + $CellContext`m2 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0, 0}]}, -2, 2, 1],
        Editable->False]},
      {
       InterpretationBox[
        RowBox[{
         FractionBox[
          RowBox[{
           RowBox[{"-", 
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "2"}], "]"}]}]}], "+", 
           RowBox[{"q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}]}], "-", 
           RowBox[{"C", "[", 
            RowBox[{"0", ",", "1"}], "]"}]}], 
          SuperscriptBox["eps", "2"]], "+", 
         FractionBox[
          RowBox[{
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}]}], "q2"], "-", 
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "+", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
           RowBox[{"2", " ", "m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
           RowBox[{"2", " ", "q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}]}], "-", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}]}], "-", 
           RowBox[{"q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}]}], "+", 
           RowBox[{"q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}]}], "+", 
           RowBox[{"C", "[", 
            RowBox[{"0", ",", "1"}], "]"}], "-", 
           RowBox[{"C", "[", 
            RowBox[{"1", ",", "1"}], "]"}]}], "eps"], "+", 
         RowBox[{"(", 
          RowBox[{
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}]}], "q2"], "-", 
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "+", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}]}], "q2"], "+", 
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
           RowBox[{"2", " ", "m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
           RowBox[{"2", " ", "q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
           RowBox[{"2", " ", "m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
           RowBox[{"2", " ", "q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
           FractionBox[
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
           RowBox[{
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "+", 
           FractionBox[
            RowBox[{"2", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "-", 
           RowBox[{"2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "1"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
           RowBox[{"4", " ", "m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
           RowBox[{"4", " ", "q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"0", ",", "2"}], "]"}], " ", 
            RowBox[{"G", "(", 
             RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}]}], "-", 
           RowBox[{"m2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"2", ",", "2"}], "]"}]}], "-", 
           RowBox[{"q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"1", ",", "2"}], "]"}]}], "+", 
           RowBox[{"q2", " ", 
            RowBox[{"C", "[", 
             RowBox[{"2", ",", "2"}], "]"}]}], "+", 
           RowBox[{"C", "[", 
            RowBox[{"1", ",", "1"}], "]"}], "-", 
           RowBox[{"C", "[", 
            RowBox[{"2", ",", "1"}], "]"}]}], ")"}], "+", 
         RowBox[{"eps", " ", 
          RowBox[{"(", 
           RowBox[{
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "1"}], "]"}]}], "q2"], "-", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"2", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "+", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"G", "(", "0", ";", "m2", ")"}], " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}]}], "q2"], "+", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "q2"], "-", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
            RowBox[{"2", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
            RowBox[{"2", " ", "q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
            RowBox[{"2", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "-", 
            RowBox[{"2", " ", "q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", "q2", ";", "m2", ")"}]}], "+", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"0", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "-", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "-", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "q2"], "+", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "0"}], ";", "m2", ")"}]}], "-", 
            FractionBox[
             RowBox[{"2", " ", "m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "+", 
            RowBox[{"2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"4", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
            RowBox[{"4", " ", "q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
            FractionBox[
             RowBox[{"2", " ", "m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"1", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "q2"], "-", 
            RowBox[{"2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
            RowBox[{"4", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"4", " ", "q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"1", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"0", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], 
            "+", 
            FractionBox[
             RowBox[{"m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "q2"], 
            "-", 
            RowBox[{
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "0", ",", "0"}], ";", "m2", ")"}]}], "+", 
            FractionBox[
             RowBox[{"2", " ", "m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "q2"],
             "-", 
            RowBox[{"2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2", ",", "0"}], ";", "m2", ")"}]}], "-", 
            FractionBox[
             RowBox[{"4", " ", "m2", " ", 
              RowBox[{"C", "[", 
               RowBox[{"0", ",", "1"}], "]"}], " ", 
              RowBox[{"G", "(", 
               RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], 
             "q2"], "+", 
            RowBox[{"4", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "1"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"8", " ", "m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "-", 
            RowBox[{"8", " ", "q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"0", ",", "2"}], "]"}], " ", 
             RowBox[{"G", "(", 
              RowBox[{"q2", ",", "q2", ",", "q2"}], ";", "m2", ")"}]}], "+", 
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "2"}], "]"}]}], "-", 
            RowBox[{"m2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"3", ",", "2"}], "]"}]}], "-", 
            RowBox[{"q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"2", ",", "2"}], "]"}]}], "+", 
            RowBox[{"q2", " ", 
             RowBox[{"C", "[", 
              RowBox[{"3", ",", "2"}], "]"}]}], "+", 
            RowBox[{"C", "[", 
             RowBox[{"2", ",", "1"}], "]"}], "-", 
            RowBox[{"C", "[", 
             RowBox[{"3", ",", "1"}], "]"}]}], ")"}]}], "+", 
         InterpretationBox[
          RowBox[{"O", "(", 
           SuperscriptBox["eps", "2"], ")"}],
          SeriesData[$CellContext`eps, 0, {}, -2, 2, 1],
          Editable->False]}],
        SeriesData[$CellContext`eps, 
         0, {-C[0, 1] - $CellContext`m2 C[0, 2] + $CellContext`q2 C[0, 2], 
          C[0, 1] + $CellContext`m2 C[0, 2] - $CellContext`q2 C[0, 2] - C[
          1, 1] - $CellContext`m2 
          C[1, 2] + $CellContext`q2 
           C[1, 2] + $CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0}] + 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - \
$CellContext`m2 $CellContext`q2^(-1) 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          2 $CellContext`m2 
           C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
          2 $CellContext`q2 
          C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}], 
          C[1, 1] + $CellContext`m2 C[1, 2] - $CellContext`q2 C[1, 2] - C[
          2, 1] - $CellContext`m2 
          C[2, 2] + $CellContext`q2 
           C[2, 2] - $CellContext`m2 $CellContext`q2^(-1) 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {
            0}] + $CellContext`m2 $CellContext`q2^(-1) 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {0}] - 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + \
$CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
          2 $CellContext`m2 
          C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          2 $CellContext`q2 
           C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - \
$CellContext`m2 $CellContext`q2^(-1) 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          2 $CellContext`m2 
           C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
          2 $CellContext`q2 
          C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - \
$CellContext`m2 $CellContext`q2^(-1) 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}] + 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 
              0}] - $CellContext`m2 $CellContext`q2^(-1) 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0}] - 
          2 C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
          2 $CellContext`m2 $CellContext`q2^(-1) 
           C[0, 
             1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] - 4 $CellContext`m2 
          C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
          4 $CellContext`q2 
           C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}], C[2, 1] + $CellContext`m2 C[2, 2] - $CellContext`q2 
          C[2, 2] - C[3, 1] - $CellContext`m2 
          C[3, 2] + $CellContext`q2 
           C[3, 2] - $CellContext`m2 $CellContext`q2^(-1) 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {
            0}] + $CellContext`m2 $CellContext`q2^(-1) 
           C[2, 1] $CellContext`Hlog[$CellContext`m2, {0}] - 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + \
$CellContext`m2 $CellContext`q2^(-1) 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
          2 $CellContext`m2 
          C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          2 $CellContext`q2 
           C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          C[2, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - \
$CellContext`m2 $CellContext`q2^(-1) 
          C[2, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + 
          2 $CellContext`m2 
           C[2, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] - 
          2 $CellContext`q2 
          C[2, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2}] + \
$CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 
             0}] - $CellContext`m2 $CellContext`q2^(-1) 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {0, 0}] - 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 
             0}] + $CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0}] + 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 
              0}] - $CellContext`m2 $CellContext`q2^(-1) 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0}] + 
          2 C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] - 2 $CellContext`m2 $CellContext`q2^(-1) 
          C[0, 
            1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
          4 $CellContext`m2 
           C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] - 4 $CellContext`q2 
          C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] - 2 
          C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
          2 $CellContext`m2 $CellContext`q2^(-1) 
           C[1, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] - 4 $CellContext`m2 
          C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + 
          4 $CellContext`q2 
           C[1, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2}] + $CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {0, 0, 0}] - 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0, 
             0}] + $CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, 0, 
              0}] - 2 C[
           0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, 0}] + 
          2 $CellContext`m2 $CellContext`q2^(-1) 
           C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, 0}] + 
          4 C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, $CellContext`q2}] - 4 $CellContext`m2 $CellContext`q2^(-1) 
          C[0, 1] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, $CellContext`q2}] + 
          8 $CellContext`m2 
           C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, $CellContext`q2}] - 8 $CellContext`q2 
          C[0, 2] $CellContext`Hlog[$CellContext`m2, {$CellContext`q2, \
$CellContext`q2, $CellContext`q2}]}, -2, 2, 1],
        Editable->False]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]]
},
WindowSize->{1920, 1046},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Magnification:>1.25 Inherited,
FrontEndVersion->"10.4 for Linux x86 (64-bit) (April 11, 2016)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info143762687456-3325587"->{
  Cell[9010, 285, 945, 13, 120, "Print",
   CellTags->"Info143762687456-3325587"]},
 "Info63762687448-3325587"->{
  Cell[2863, 98, 1096, 15, 274, "Print",
   CellTags->"Info63762687448-3325587"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info143762687456-3325587", 111424, 3097},
 {"Info63762687448-3325587", 111538, 3100}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[558, 20, 123, 3, 39, "Input"],
Cell[CellGroupData[{
Cell[706, 27, 41, 0, 57, "Subsection"],
Cell[CellGroupData[{
Cell[772, 31, 52, 1, 39, "Input"],
Cell[827, 34, 1500, 37, 150, "Print"]
}, Open  ]],
Cell[2342, 74, 77, 2, 39, "Input"],
Cell[2422, 78, 236, 7, 39, "Input"],
Cell[2661, 87, 123, 4, 39, "Input"],
Cell[CellGroupData[{
Cell[2809, 95, 51, 1, 39, "Input"],
Cell[2863, 98, 1096, 15, 274, "Print",
 CellTags->"Info63762687448-3325587"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3996, 118, 287, 9, 39, "Input"],
Cell[4286, 129, 321, 4, 150, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4644, 138, 83, 2, 39, "Input"],
Cell[4730, 142, 216, 2, 78, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4983, 149, 86, 2, 39, "Input"],
Cell[5072, 153, 494, 6, 174, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5603, 164, 86, 2, 39, "Input"],
Cell[5692, 168, 517, 6, 150, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6246, 179, 152, 4, 39, "Input"],
Cell[CellGroupData[{
Cell[6423, 187, 252, 7, 29, "Print"],
Cell[6678, 196, 204, 2, 102, "Print"],
Cell[6885, 200, 252, 7, 29, "Print"],
Cell[7140, 209, 204, 2, 102, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7393, 217, 86, 2, 39, "Input"],
Cell[7482, 221, 203, 6, 39, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[7722, 232, 252, 8, 39, "Input"],
Cell[7977, 242, 865, 30, 60, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[8891, 278, 44, 0, 57, "Subsection"],
Cell[CellGroupData[{
Cell[8960, 282, 47, 1, 39, "Input"],
Cell[9010, 285, 945, 13, 120, "Print",
 CellTags->"Info143762687456-3325587"]
}, Open  ]],
Cell[CellGroupData[{
Cell[9992, 303, 240, 8, 39, "Input"],
Cell[10235, 313, 865, 30, 60, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[11137, 348, 339, 11, 39, "Input"],
Cell[11479, 361, 574, 21, 60, "Output"]
}, Open  ]],
Cell[12068, 385, 206, 7, 39, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[12311, 397, 38, 0, 57, "Subsection"],
Cell[12352, 399, 511, 14, 39, "Input"],
Cell[CellGroupData[{
Cell[12888, 417, 395, 12, 94, "Input"],
Cell[13286, 431, 978, 31, 99, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[14301, 467, 181, 5, 67, "Input"],
Cell[14485, 474, 949, 30, 76, "Output"]
}, Open  ]],
Cell[15449, 507, 26, 0, 39, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[15512, 512, 74, 0, 57, "Subsection"],
Cell[CellGroupData[{
Cell[15611, 516, 838, 28, 94, "Input"],
Cell[16452, 546, 181, 6, 39, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[16670, 557, 92, 2, 39, "Input"],
Cell[16765, 561, 926, 30, 81, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[17728, 596, 92, 2, 39, "Input"],
Cell[17823, 600, 2159, 65, 177, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[20019, 670, 92, 2, 39, "Input"],
Cell[20114, 674, 8291, 232, 443, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[28442, 911, 92, 2, 39, "Input"],
Cell[28537, 915, 11435, 280, 615, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[40021, 1201, 67, 0, 57, "Subsection"],
Cell[40091, 1203, 761, 22, 39, "Input"],
Cell[CellGroupData[{
Cell[40877, 1229, 139, 4, 39, "Input"],
Cell[41019, 1235, 98, 2, 39, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[41154, 1242, 183, 6, 39, "Input"],
Cell[41340, 1250, 131, 4, 57, "Output"]
}, Open  ]],
Cell[41486, 1257, 2092, 61, 256, "Input"],
Cell[CellGroupData[{
Cell[43603, 1322, 253, 8, 39, "Input"],
Cell[43859, 1332, 98, 2, 39, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[43994, 1339, 800, 26, 94, "Input"],
Cell[44797, 1367, 181, 6, 39, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[45015, 1378, 54, 1, 39, "Input"],
Cell[45072, 1381, 821, 26, 65, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[45930, 1412, 92, 2, 39, "Input"],
Cell[46025, 1416, 1876, 57, 122, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[47938, 1478, 92, 2, 39, "Input"],
Cell[48033, 1482, 3103, 92, 156, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[51173, 1579, 93, 2, 39, "Input"],
Cell[51269, 1583, 22055, 615, 1348, "Output"]
}, Open  ]],
Cell[73339, 2201, 547, 18, 39, "Input"],
Cell[CellGroupData[{
Cell[73911, 2223, 102, 2, 39, "Input"],
Cell[74016, 2227, 10661, 251, 148, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[84714, 2483, 244, 7, 67, "Input"],
Cell[84961, 2492, 26103, 590, 148, "Output"]
}, Open  ]]
}, Open  ]]
}
]
*)

