(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 10.4' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[   2713211,      64045]
NotebookOptionsPosition[   2675165,      62951]
NotebookOutlinePosition[   2675531,      62967]
CellTagsIndexPosition[   2675488,      62964]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["\<\
0. Initialization: FeynArts (for diagram generation) and FeynCalc (for tensor \
algebra)\
\>", "Section"],

Cell[TextData[{
 StyleBox["See these links for documentation:\n- ", "Text"],
 StyleBox["http://www.feynarts.de/FA3Guide.pdf\n- ", "Item"],
 "https://feyncalc.github.io/FeynCalcBook/guide/FeynCalc.html"
}], "Text"],

Cell["\<\
To install the latest FeynCalc with a compatible FeynArts, evaluate this (in \
a separate instance):\
\>", "Text"],

Cell[BoxData[{
 RowBox[{"Import", "[", 
  "\"\<https://raw.githubusercontent.com/FeynCalc/feyncalc/master/install.m\>\
\"", "]"}], "\n", 
 RowBox[{"InstallFeynCalc", "[", "]"}]}], "Code",
 Evaluatable->False,
 InitializationCell->False],

Cell["\<\
FeynArts need to be patched by FeynCalc for the two to work together.
For this reason we'll let FeynCalc import FeynArts by itself.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"$LoadFeynArts", "=", "True"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"<<", "FeynCalc`"}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    StyleBox["\<\"FeynCalc \"\>", "Text",
     StripOnInput->False,
     FontWeight->Bold], "\[InvisibleSpace]", 
    StyleBox["\<\"9.2.0. For help, use the \"\>", "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox[
     TagBox[
      ButtonBox[
       RowBox[{"documentation", " ", "center"}],
       BaseStyle->"Link",
       ButtonData:>"paclet:FeynCalc/",
       ButtonNote->"paclet:FeynCalc/"],
      DisplayForm], "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox["\<\", check out the \"\>", "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox[
     TagBox[
      ButtonBox["wiki",
       BaseStyle->"Hyperlink",
       ButtonData:>{
         URL["https://github.com/FeynCalc/feyncalc/wiki"], None},
       ButtonNote->"https://github.com/FeynCalc/feyncalc/wiki"],
      DisplayForm], "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox["\<\" or write to the \"\>", "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox[
     TagBox[
      ButtonBox[
       RowBox[{"mailing", " ", 
        RowBox[{"list", "."}]}],
       BaseStyle->"Hyperlink",
       ButtonData:>{
         URL["http://www.feyncalc.org/forum/"], None},
       ButtonNote->"http://www.feyncalc.org/forum/"],
      DisplayForm], "Text",
     StripOnInput->False]}],
   SequenceForm[
    Style["FeynCalc ", "Text", Bold], 
    Style["9.2.0. For help, use the ", "Text"], 
    Style[
     DisplayForm[
      ButtonBox[
      "documentation center", BaseStyle -> "Link", ButtonData :> 
       "paclet:FeynCalc/", ButtonNote -> "paclet:FeynCalc/"]], "Text"], 
    Style[", check out the ", "Text"], 
    Style[
     DisplayForm[
      ButtonBox["wiki", ButtonData :> {
         URL["https://github.com/FeynCalc/feyncalc/wiki"], None}, BaseStyle -> 
       "Hyperlink", ButtonNote -> 
       "https://github.com/FeynCalc/feyncalc/wiki"]], "Text"], 
    Style[" or write to the ", "Text"], 
    Style[
     DisplayForm[
      ButtonBox["mailing list.", ButtonData :> {
         URL["http://www.feyncalc.org/forum/"], None}, BaseStyle -> 
       "Hyperlink", ButtonNote -> "http://www.feyncalc.org/forum/"]], "Text"]],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    StyleBox["\<\"See also the supplied \"\>", "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox[
     TagBox[
      ButtonBox[
       RowBox[{"examples", "."}],
       BaseStyle->"Hyperlink",
       ButtonFunction:>SystemOpen[
         FileNameJoin[{FeynCalc`$FeynCalcDirectory, "Examples"}]],
       Evaluator->Automatic,
       Method->"Preemptive"],
      DisplayForm], "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox["\<\" If you use FeynCalc in your research, please cite\"\>", 
     "Text",
     StripOnInput->False]}],
   SequenceForm[
    Style["See also the supplied ", "Text"], 
    Style[
     DisplayForm[
      ButtonBox[
      "examples.", BaseStyle -> "Hyperlink", ButtonFunction :> SystemOpen[
         FileNameJoin[{FeynCalc`$FeynCalcDirectory, "Examples"}]], Evaluator -> 
       Automatic, Method -> "Preemptive"]], "Text"], 
    Style[" If you use FeynCalc in your research, please cite", "Text"]],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox["\<\" \[Bullet] V. Shtabovenko, R. Mertig and F. Orellana, Comput. \
Phys. Commun., 207C, 432-444, 2016, arXiv:1601.01167\"\>", "Text",
   StripOnInput->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox["\<\" \[Bullet] R. Mertig, M. B\[ODoubleDot]hm, and A. Denner, \
Comput. Phys. Commun., 64, 345-359, 1991.\"\>", "Text",
   StripOnInput->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    StyleBox["\<\"FeynArts \"\>", "Text",
     StripOnInput->False,
     FontWeight->Bold], "\[InvisibleSpace]", 
    StyleBox["\<\"3.11 (14 Mar 2019) patched for use with FeynCalc, for \
documentation use the \"\>", "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox[
     TagBox[
      ButtonBox["manual",
       BaseStyle->"Hyperlink",
       ButtonFunction:>SystemOpen[
         First[
          FileNames[{"*.pdf", "*.PDF"}, 
           FileNameJoin[{FeynCalc`$FeynArtsDirectory, "manual"}]]]],
       Evaluator->Automatic,
       Method->"Preemptive"],
      DisplayForm], "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox["\<\" or visit \"\>", "Text",
     StripOnInput->False], "\[InvisibleSpace]", 
    StyleBox[
     TagBox[
      ButtonBox[
       RowBox[{"www", ".", "feynarts", ".", "de", "."}],
       BaseStyle->"Hyperlink",
       ButtonData:>{
         URL["http://www.feynarts.de/"], None},
       ButtonNote->"www.feynarts.de/"],
      DisplayForm], "Text",
     StripOnInput->False]}],
   SequenceForm[
    Style["FeynArts ", "Text", Bold], 
    Style[
    "3.11 (14 Mar 2019) patched for use with FeynCalc, for documentation use \
the ", "Text"], 
    Style[
     DisplayForm[
      ButtonBox[
      "manual", BaseStyle -> "Hyperlink", ButtonFunction :> SystemOpen[
         First[
          FileNames[{"*.pdf", "*.PDF"}, 
           FileNameJoin[{FeynCalc`$FeynArtsDirectory, "manual"}]]]], 
       Evaluator -> Automatic, Method -> "Preemptive"]], "Text"], 
    Style[" or visit ", "Text"], 
    Style[
     DisplayForm[
      ButtonBox["www.feynarts.de.", ButtonData :> {
         URL["http://www.feynarts.de/"], None}, BaseStyle -> "Hyperlink", 
       ButtonNote -> "www.feynarts.de/"]], "Text"]],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$FAVerbose", "=", "0"}], ";"}]], "Input"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SetDirectory", "[", 
   RowBox[{"NotebookDirectory", "[", "]"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Get", "[", "\"\<utils.m\>\"", "]"}], ";"}]}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "1. Illustration: tree level ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["\[Gamma]", "*"], "\[RightArrow]", 
    RowBox[{"u", 
     OverscriptBox["u", "_"]}]}], TraditionalForm]]],
 " matrix element"
}], "Section"],

Cell[CellGroupData[{

Cell["Generate the Feynman diagrams", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"CreateTopologies", "[", 
    RowBox[{"0", ",", 
     RowBox[{"1", "\[Rule]", "2"}], ",", 
     RowBox[{"ExcludeTopologies", "\[Rule]", 
      RowBox[{"{", "Tadpoles", "}"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"topologies", "//", 
   RowBox[{
    RowBox[{"Paint", "[", 
     RowBox[{"#", ",", 
      RowBox[{"AutoEdit", "\[Rule]", "False"}]}], "]"}], "&"}]}], 
  ";"}]}], "Input"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], 
         LineBox[{{-3.6859404417555197`*^-13, 10.}, {11.000000000008335`, 
          10.}}], LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         {PointSize[0.04], PointBox[{11., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["T1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->7.997355371900827],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            7.997355371900827]& ], {10., -0.5}, {0, -1}]},
        "\"ad/bdcd/0.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/bdcd/0.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 44}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["1",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {28.5, 69.96}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {33., 69.96}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["2",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {37.5, 69.96}, {0, 0}]},
   AspectRatio->1.0999999999999999`,
   ImageSize->{288, 288},
   PlotRange->{{0, 66}, {0, 72.6}}], TraditionalForm]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"diagrams", "=", 
   RowBox[{"InsertFields", "[", 
    RowBox[{"topologies", ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"V", "[", "1", "]"}], " ", 
       RowBox[{"(*", " ", "photon", " ", "*)"}], "}"}], "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"F", "[", "3", "]"}], 
        RowBox[{"(*", " ", 
         RowBox[{"up", " ", "quark"}], " ", "*)"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "[", "3", "]"}]}]}], 
       RowBox[{"(*", " ", 
        RowBox[{
         RowBox[{"anti", " ", "up"}], "-", 
         RowBox[{"type", " ", "quarks"}]}], " ", "*)"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"InsertionLevel", "\[Rule]", 
      RowBox[{"{", "Classes", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"GenericModel", "\[Rule]", "\"\<Lorentz\>\""}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Model", "\[Rule]", "\"\<SMQCD-modified\>\""}], 
     RowBox[{"(*", " ", 
      RowBox[{
      "Try", " ", "this", " ", "to", " ", "have", " ", "symbolic", " ", 
       "charges", " ", "for", " ", "the", " ", 
       RowBox[{"quarks", ":", " ", "\"\<SMQCD-modified\>\""}]}], " ", "*)"}], 
     ",", "\[IndentingNewLine]", 
     RowBox[{"ExcludeParticles", "\[Rule]", 
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"F", "[", "1", "]"}], 
        RowBox[{"(*", " ", "neutrinos", " ", "*)"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"F", "[", "2", "]"}], 
        RowBox[{"(*", " ", "leptons", " ", "*)"}], ",", "\[IndentingNewLine]", 
        RowBox[{"(*", " ", 
         RowBox[{
          RowBox[{
           RowBox[{"F", "[", "3", "]"}], " ", "up"}], "-", 
          RowBox[{"type", " ", "quarks"}]}], " ", "*)"}], 
        "\[IndentingNewLine]", 
        RowBox[{"(*", " ", 
         RowBox[{
          RowBox[{
           RowBox[{"F", "[", "4", "]"}], " ", "down"}], "-", 
          RowBox[{"type", " ", "quarks"}]}], " ", "*)"}], 
        "\[IndentingNewLine]", 
        RowBox[{"V", "[", "1", "]"}], 
        RowBox[{"(*", " ", "photons", " ", "*)"}], ",", "\[IndentingNewLine]", 
        RowBox[{"V", "[", "2", "]"}], 
        RowBox[{"(*", " ", 
         RowBox[{"Z", " ", "bosons"}], " ", "*)"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"V", "[", "3", "]"}], 
        RowBox[{"(*", " ", 
         RowBox[{"W", " ", "bosons"}], " ", "*)"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"(*", " ", 
         RowBox[{
          RowBox[{"V", "[", "5", "]"}], " ", "gluons"}], " ", "*)"}], 
        "\[IndentingNewLine]", "S", 
        RowBox[{"(*", " ", 
         RowBox[{"all", " ", "scalars"}], " ", "*)"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"U", "[", "1", "]"}], ",", 
        RowBox[{"U", "[", "2", "]"}], ",", 
        RowBox[{"U", "[", "3", "]"}], ",", 
        RowBox[{"U", "[", "4", "]"}], 
        RowBox[{"(*", " ", "ghosts", " ", "*)"}], ","}], 
       "\[IndentingNewLine]", 
       RowBox[{"(*", " ", 
        RowBox[{
         RowBox[{"U", "[", "5", "]"}], " ", "gluon", " ", "ghost"}], " ", 
        "*)"}], "\[IndentingNewLine]", "}"}]}]}], "\[IndentingNewLine]", 
    "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"diagrams", "//", 
   RowBox[{
    RowBox[{"Paint", "[", 
     RowBox[{"#", ",", 
      RowBox[{"AutoEdit", "\[Rule]", "False"}]}], "]"}], "&"}]}], 
  ";"}]}], "Input"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->9.996694214876033],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            9.996694214876033]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "j"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->9.996694214876033],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            9.996694214876033]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->9.996694214876033],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            9.996694214876033]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         {PointSize[0.04], PointBox[{11., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             RowBox[{"T1", " ", "C1", " ", "N1"}],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->7.997355371900827],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            7.997355371900827]& ], {10., -0.5}, {0, -1}]},
        "\"ad/bdcd/0.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/bdcd/0.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 44}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {26.25, 69.96}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {30.75, 69.96}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "j"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {35.25, 69.96}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->11.996033057851239`],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       11.996033057851239`]& ], {39.75, 69.96}, {0, 0}]},
   AspectRatio->1.0999999999999999`,
   ImageSize->{288, 288},
   PlotRange->{{0, 66}, {0, 72.6}}], TraditionalForm]], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Convert the diagrams to the amplitudes", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"ScalarProduct", "[", 
    RowBox[{"k1", ",", "k1", ",", 
     RowBox[{"SetDimensions", "\[Rule]", 
      RowBox[{"{", "d", "}"}]}]}], "]"}], "=", "0"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ScalarProduct", "[", 
    RowBox[{"k2", ",", "k2", ",", 
     RowBox[{"SetDimensions", "\[Rule]", 
      RowBox[{"{", "d", "}"}]}]}], "]"}], "=", "0"}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"famp", "=", 
  RowBox[{"CreateFeynAmp", "[", "diagrams", 
   RowBox[{"(*", 
    RowBox[{",", 
     RowBox[{"GaugeRules", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"_GaugeXi", "\[Rule]", "1"}], "}"}]}]}], "*)"}], 
   "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"FAFeynAmpList", "(", 
    RowBox[{
     RowBox[{"Process", "\[Rule]", 
      RowBox[{
       RowBox[{"(", "\[NoBreak]", GridBox[{
          {
           RowBox[{"V", "(", "1", ")"}], 
           FormBox[
            RowBox[{"\<\"p\"\>", "\[InvisibleSpace]", "1"}],
            TraditionalForm], "0", 
           RowBox[{"{", "}"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.7]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
       "\[Rule]", 
       RowBox[{"(", "\[NoBreak]", GridBox[{
          {
           RowBox[{"F", "(", 
            RowBox[{"3", ",", 
             RowBox[{"{", 
              RowBox[{
               FormBox[
                RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
                TraditionalForm], ",", 
               FormBox[
                RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "2"}],
                TraditionalForm]}], "}"}]}], ")"}], 
           FormBox[
            RowBox[{"\<\"k\"\>", "\[InvisibleSpace]", "1"}],
            TraditionalForm], 
           RowBox[{"MQU", "(", 
            FormBox[
             RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
             TraditionalForm], ")"}], 
           RowBox[{"{", 
            RowBox[{
             FractionBox[
              RowBox[{"2", " ", "Charge"}], "3"], ",", 
             FractionBox[
              RowBox[{"2", " ", "ColorCharge"}], 
              SqrtBox["3"]]}], "}"}]},
          {
           RowBox[{"-", 
            RowBox[{"F", "(", 
             RowBox[{"3", ",", 
              RowBox[{"{", 
               RowBox[{
                FormBox[
                 RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
                 TraditionalForm], ",", 
                FormBox[
                 RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "3"}],
                 TraditionalForm]}], "}"}]}], ")"}]}], 
           FormBox[
            RowBox[{"\<\"k\"\>", "\[InvisibleSpace]", "2"}],
            TraditionalForm], 
           RowBox[{"MQU", "(", 
            FormBox[
             RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
             TraditionalForm], ")"}], 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"2", " ", "Charge"}], "3"]}], ",", 
             RowBox[{"-", 
              FractionBox[
               RowBox[{"2", " ", "ColorCharge"}], 
               SqrtBox["3"]]}]}], "}"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.7]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}]}], 
     ",", 
     RowBox[{"Model", "\[Rule]", 
      RowBox[{"{", "\<\"SMQCD-modified\"\>", "}"}]}], ",", 
     RowBox[{"GenericModel", "\[Rule]", 
      RowBox[{"{", "\<\"Lorentz\"\>", "}"}]}], ",", 
     RowBox[{"AmplitudeLevel", "\[Rule]", 
      RowBox[{"{", "Classes", "}"}]}], ",", 
     RowBox[{"ExcludeParticles", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"S", ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", "1", ")"}]}], ",", 
        RowBox[{"F", "(", "1", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", "2", ")"}]}], ",", 
        RowBox[{"F", "(", "2", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", 
          RowBox[{"1", ",", 
           RowBox[{"{", "1", "}"}]}], ")"}]}], ",", 
        RowBox[{"F", "(", 
         RowBox[{"1", ",", 
          RowBox[{"{", "1", "}"}]}], ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", 
          RowBox[{"1", ",", 
           RowBox[{"{", "2", "}"}]}], ")"}]}], ",", 
        RowBox[{"F", "(", 
         RowBox[{"1", ",", 
          RowBox[{"{", "2", "}"}]}], ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", 
          RowBox[{"1", ",", 
           RowBox[{"{", "3", "}"}]}], ")"}]}], ",", 
        RowBox[{"F", "(", 
         RowBox[{"1", ",", 
          RowBox[{"{", "3", "}"}]}], ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", 
          RowBox[{"2", ",", 
           RowBox[{"{", "1", "}"}]}], ")"}]}], ",", 
        RowBox[{"F", "(", 
         RowBox[{"2", ",", 
          RowBox[{"{", "1", "}"}]}], ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", 
          RowBox[{"2", ",", 
           RowBox[{"{", "2", "}"}]}], ")"}]}], ",", 
        RowBox[{"F", "(", 
         RowBox[{"2", ",", 
          RowBox[{"{", "2", "}"}]}], ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"F", "(", 
          RowBox[{"2", ",", 
           RowBox[{"{", "3", "}"}]}], ")"}]}], ",", 
        RowBox[{"F", "(", 
         RowBox[{"2", ",", 
          RowBox[{"{", "3", "}"}]}], ")"}], ",", 
        RowBox[{"S", "(", "1", ")"}], ",", 
        RowBox[{"S", "(", "2", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"S", "(", "3", ")"}]}], ",", 
        RowBox[{"S", "(", "3", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"U", "(", "1", ")"}]}], ",", 
        RowBox[{"U", "(", "1", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"U", "(", "2", ")"}]}], ",", 
        RowBox[{"U", "(", "2", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"U", "(", "3", ")"}]}], ",", 
        RowBox[{"U", "(", "3", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"U", "(", "4", ")"}]}], ",", 
        RowBox[{"U", "(", "4", ")"}], ",", 
        RowBox[{"V", "(", "1", ")"}], ",", 
        RowBox[{"V", "(", "2", ")"}], ",", 
        RowBox[{"-", 
         RowBox[{"V", "(", "3", ")"}]}], ",", 
        RowBox[{"V", "(", "3", ")"}]}], "}"}]}], ",", 
     RowBox[{"ExcludeFieldPoints", "\[Rule]", 
      RowBox[{"{", "}"}]}], ",", 
     RowBox[{"LastSelections", "\[Rule]", 
      RowBox[{"{", "}"}]}]}], ")"}], "\[InvisibleApplication]", 
   RowBox[{"(", 
    RowBox[{"FAFeynAmp", "(", 
     RowBox[{
      RowBox[{"GraphID", "(", 
       RowBox[{
        RowBox[{"Topology", "\[LongEqual]", "1"}], ",", 
        RowBox[{"Generic", "\[LongEqual]", "1"}], ",", 
        RowBox[{"Classes", "\[LongEqual]", "1"}], ",", 
        RowBox[{"Number", "\[LongEqual]", "1"}]}], ")"}], ",", 
      RowBox[{"Integral", "[", "]"}], ",", 
      RowBox[{"\[ImaginaryI]", " ", 
       RowBox[{"SumOver", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "2"}],
           SequenceForm["Col", 2],
           Editable->False],
          TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
       RowBox[{"SumOver", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Col", 3],
           Editable->False],
          TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
       RowBox[{"SumOver", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
           SequenceForm["Gen", 2],
           Editable->False],
          TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
       RowBox[{"SumOver", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
       RowBox[{
        FormBox["\<\"ep\"\>",
         TraditionalForm], "(", 
        RowBox[{
         RowBox[{"V", "(", "1", ")"}], ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"p\"\>", "\[InvisibleSpace]", "1"}],
           SequenceForm["p", 1],
           Editable->False],
          TraditionalForm], ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Lor\"\>", "\[InvisibleSpace]", "1"}],
           SequenceForm["Lor", 1],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       FormBox[
        RowBox[{
         RowBox[{
          OverscriptBox["\<\"u\"\>", "\<\"_\"\>"], "(", 
          RowBox[{
           FormBox[
            InterpretationBox[
             RowBox[{"\<\"k\"\>", "\[InvisibleSpace]", "1"}],
             SequenceForm["k", 1],
             Editable->False],
            TraditionalForm], ",", 
           RowBox[{"MQU", "(", 
            FormBox[
             InterpretationBox[
              RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
              SequenceForm["Gen", 2],
              Editable->False],
             TraditionalForm], ")"}]}], ")"}], ".", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"\[ImaginaryI]", " ", 
             RowBox[{"FCGV", "(", "\<\"EL\"\>", ")"}], " ", 
             SubscriptBox["Q", "u"], " ", 
             RowBox[{"IndexDelta", "(", 
              RowBox[{
               FormBox[
                InterpretationBox[
                 RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "2"}],
                 SequenceForm["Col", 2],
                 Editable->False],
                TraditionalForm], ",", 
               FormBox[
                InterpretationBox[
                 RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "3"}],
                 SequenceForm["Col", 3],
                 Editable->False],
                TraditionalForm]}], ")"}], " ", 
             RowBox[{"IndexDelta", "(", 
              RowBox[{
               FormBox[
                InterpretationBox[
                 RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
                 SequenceForm["Gen", 2],
                 Editable->False],
                TraditionalForm], ",", 
               FormBox[
                InterpretationBox[
                 RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
                 SequenceForm["Gen", 3],
                 Editable->False],
                TraditionalForm]}], ")"}], " ", 
             FormBox[
              RowBox[{
               RowBox[{
                FormBox["\<\"ga\"\>",
                 TraditionalForm], "(", 
                FormBox[
                 InterpretationBox[
                  RowBox[{"\<\"Lor\"\>", "\[InvisibleSpace]", "1"}],
                  SequenceForm["Lor", 1],
                  Editable->False],
                 TraditionalForm], ")"}], ".", 
               FormBox[
                SubscriptBox["\<\"om\"\>", "\<\"-\"\>"],
                TraditionalForm]}],
              TraditionalForm]}]}], "-", 
           RowBox[{"\[ImaginaryI]", " ", 
            RowBox[{"FCGV", "(", "\<\"EL\"\>", ")"}], " ", 
            SubscriptBox["Q", "u"], " ", 
            RowBox[{"IndexDelta", "(", 
             RowBox[{
              FormBox[
               InterpretationBox[
                RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "2"}],
                SequenceForm["Col", 2],
                Editable->False],
               TraditionalForm], ",", 
              FormBox[
               InterpretationBox[
                RowBox[{"\<\"Col\"\>", "\[InvisibleSpace]", "3"}],
                SequenceForm["Col", 3],
                Editable->False],
               TraditionalForm]}], ")"}], " ", 
            RowBox[{"IndexDelta", "(", 
             RowBox[{
              FormBox[
               InterpretationBox[
                RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
                SequenceForm["Gen", 2],
                Editable->False],
               TraditionalForm], ",", 
              FormBox[
               InterpretationBox[
                RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
                SequenceForm["Gen", 3],
                Editable->False],
               TraditionalForm]}], ")"}], " ", 
            FormBox[
             RowBox[{
              RowBox[{
               FormBox["\<\"ga\"\>",
                TraditionalForm], "(", 
               FormBox[
                InterpretationBox[
                 RowBox[{"\<\"Lor\"\>", "\[InvisibleSpace]", "1"}],
                 SequenceForm["Lor", 1],
                 Editable->False],
                TraditionalForm], ")"}], ".", 
              FormBox[
               SubscriptBox["\<\"om\"\>", "\<\"+\"\>"],
               TraditionalForm]}],
             TraditionalForm]}]}], ")"}], ".", 
         RowBox[{"\<\"v\"\>", "(", 
          RowBox[{
           FormBox[
            InterpretationBox[
             RowBox[{"\<\"k\"\>", "\[InvisibleSpace]", "2"}],
             SequenceForm["k", 2],
             Editable->False],
            TraditionalForm], ",", 
           RowBox[{"MQU", "(", 
            FormBox[
             InterpretationBox[
              RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
              SequenceForm["Gen", 3],
              Editable->False],
             TraditionalForm], ")"}]}], ")"}]}],
        TraditionalForm]}]}], ")"}], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"amplitude", "=", 
  RowBox[{"FCFAConvert", "[", 
   RowBox[{"famp", ",", "\[IndentingNewLine]", 
    RowBox[{"IncomingMomenta", "\[Rule]", 
     RowBox[{"{", "q", "}"}]}], ",", "\[IndentingNewLine]", 
    RowBox[{"OutgoingMomenta", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4", ",", "k5"}], "}"}]}], 
    ",", " ", 
    RowBox[{"(*", " ", 
     RowBox[{
     "We", " ", "can", " ", "specify", " ", "more", " ", "names", " ", "than",
       " ", "we", " ", "have"}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{"LoopMomenta", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"l1", ",", "l2", ",", "l3", ",", "l4", ",", "k5"}], "}"}]}], 
    ",", "\[IndentingNewLine]", 
    RowBox[{"UndoChiralSplittings", "\[Rule]", "True"}], ",", " ", 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{
       RowBox[{"Don", "'"}], "t", " ", "split", " ", "couplings", " ", "into",
        " ", "left"}], "-", " ", 
      RowBox[{"and", " ", "right"}], "-", 
      RowBox[{"handed", " ", "ones"}]}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{"DropSumOver", "\[Rule]", "False"}], ",", " ", 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{"We", "'"}], "ll", " ", "handle", " ", "SumOver", " ", 
      "ourselves"}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{"ChangeDimension", "\[Rule]", "d"}], ",", " ", 
    RowBox[{"(*", " ", 
     RowBox[{
     "Use", " ", "this", " ", "symbol", " ", "for", " ", "the", " ", 
      "dimentionality"}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{"LorentzIndexNames", "\[Rule]", 
     RowBox[{"{", "}"}]}], ",", "\[IndentingNewLine]", 
    RowBox[{"List", "\[Rule]", "False"}], ",", "\[IndentingNewLine]", 
    RowBox[{"SMP", "\[Rule]", "True"}]}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{"\"\<SMP\>\"", "=", 
      RowBox[{"\"\<Stadard Model Parameters\>\"", ":", " ", 
       RowBox[{"use", " ", 
        RowBox[{"SMP", "[", "\"\<e\>\"", "]"}], " ", "instead", " ", "of", 
        " ", 
        RowBox[{"FCGV", "[", "\"\<EL\>\"", "]"}]}]}]}], ",", " ", 
     RowBox[{
      RowBox[{"etc", " ", "--"}], " ", "for", " ", "prettier", " ", 
      "formatting"}]}], " ", "*)"}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"\[ImaginaryI]", " ", 
   FormBox[
    RowBox[{
     SuperscriptBox[
      FormBox["\<\"\[CurlyEpsilon]\"\>",
       TraditionalForm], 
      FormBox[
       FormBox[
        FormBox["Lor1",
         TraditionalForm],
        TraditionalForm],
       TraditionalForm]], "(", 
     FormBox["q",
      TraditionalForm], ")"}],
    TraditionalForm], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox["Col2",
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox["Col3",
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
       SequenceForm["Gen", 2],
       Editable->False],
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
       SequenceForm["Gen", 3],
       Editable->False],
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\",\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"MQU", "(", 
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
          SequenceForm["Gen", 2],
          Editable->False],
         TraditionalForm], ")"}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", "\<\"e\"\>", " ", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox["Lor1",
         TraditionalForm],
        TraditionalForm]], " ", 
      SubscriptBox["Q", "u"], " ", 
      SubscriptBox["\[Delta]", 
       RowBox[{
        FormBox[
         FormBox["Col2",
          TraditionalForm],
         TraditionalForm], 
        FormBox[
         FormBox["Col3",
          TraditionalForm],
         TraditionalForm]}]], " ", 
      RowBox[{"IndexDelta", "(", 
       RowBox[{
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
          SequenceForm["Gen", 2],
          Editable->False],
         TraditionalForm], ",", 
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
          SequenceForm["Gen", 3],
          Editable->False],
         TraditionalForm]}], ")"}]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\",\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"MQU", "(", 
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
          SequenceForm["Gen", 3],
          Editable->False],
         TraditionalForm], ")"}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"amplitude", "=", 
  RowBox[{
   RowBox[{"amplitude", "//", "DotExpand"}], "//", "Contract"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"\<\"e\"\>", " ", 
   SubscriptBox["Q", "u"], " ", 
   SubscriptBox["\[Delta]", 
    RowBox[{
     FormBox[
      FormBox["Col2",
       TraditionalForm],
      TraditionalForm], 
     FormBox[
      FormBox["Col3",
       TraditionalForm],
      TraditionalForm]}]], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox["Col2",
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox["Col3",
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"IndexDelta", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
       SequenceForm["Gen", 2],
       Editable->False],
      TraditionalForm], ",", 
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
       SequenceForm["Gen", 3],
       Editable->False],
      TraditionalForm]}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
       SequenceForm["Gen", 2],
       Editable->False],
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
       SequenceForm["Gen", 3],
       Editable->False],
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\",\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"MQU", "(", 
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
          SequenceForm["Gen", 2],
          Editable->False],
         TraditionalForm], ")"}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        FormBox["\<\"\[CurlyEpsilon]\"\>",
         TraditionalForm], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\",\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"MQU", "(", 
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
          SequenceForm["Gen", 3],
          Editable->False],
         TraditionalForm], ")"}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"amplitude", "=", 
  RowBox[{"amplitude", "/.", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"MQU", "[", "_", "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{"(*", " ", 
      RowBox[{"Up", "-", 
       RowBox[{"type", " ", "quark", " ", "mass", " ", "is", " ", "zero"}]}], 
      " ", "*)"}], "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"MQD", "[", "_", "]"}], "\[Rule]", "0"}], ",", " ", 
     RowBox[{"(*", " ", 
      RowBox[{"Down", "-", 
       RowBox[{"type", " ", "quark", " ", "mass", " ", "is", " ", "zero"}]}], 
      " ", "*)"}], "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"SMP", "[", "\"\<m_u\>\"", "]"}], "\[Rule]", "0"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"SMP", "[", "\"\<m_d\>\"", "]"}], "\[Rule]", "0"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"SumOver", "[", 
       RowBox[{
        RowBox[{"_SUNIndex", "|", "_SUNFIndex"}], ",", "___"}], "]"}], 
      "\[Rule]", "1"}]}], 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{
      "These", " ", "sums", " ", "obey", " ", "the", " ", "Einstein", " ", 
       "notation"}], ",", " ", 
      RowBox[{
      "no", " ", "need", " ", "to", " ", "keep", " ", "them", " ", 
       "explicitly"}]}], " ", "*)"}], "\[IndentingNewLine]", 
    "}"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"\<\"e\"\>", " ", 
   SubscriptBox["Q", "u"], " ", 
   SubscriptBox["\[Delta]", 
    RowBox[{
     FormBox[
      FormBox["Col2",
       TraditionalForm],
      TraditionalForm], 
     FormBox[
      FormBox["Col3",
       TraditionalForm],
      TraditionalForm]}]], " ", 
   RowBox[{"IndexDelta", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
       SequenceForm["Gen", 2],
       Editable->False],
      TraditionalForm], ",", 
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
       SequenceForm["Gen", 3],
       Editable->False],
      TraditionalForm]}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
       SequenceForm["Gen", 2],
       Editable->False],
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{"SumOver", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
       SequenceForm["Gen", 3],
       Editable->False],
      TraditionalForm], ",", "3", ",", "External"}], ")"}], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        FormBox["\<\"\[CurlyEpsilon]\"\>",
         TraditionalForm], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}]}], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The matrix element from the amplitudes", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{"amplitude", "*", 
   RowBox[{"ComplexConjugate", "[", 
    RowBox[{"FCRenameDummyIndices", "[", "amplitude", "]"}], 
    "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   SuperscriptBox["\<\"e\"\>", "2"], " ", 
   SubsuperscriptBox["Q", "u", "2"], " ", 
   SubsuperscriptBox["\[Delta]", 
    RowBox[{
     FormBox[
      FormBox["Col2",
       TraditionalForm],
      TraditionalForm], 
     FormBox[
      FormBox["Col3",
       TraditionalForm],
      TraditionalForm]}], "2"], " ", 
   RowBox[{"IndexDelta", "(", 
    RowBox[{
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
       SequenceForm["Gen", 2],
       Editable->False],
      TraditionalForm], ",", 
     FormBox[
      InterpretationBox[
       RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
       SequenceForm["Gen", 3],
       Editable->False],
      TraditionalForm]}], ")"}], " ", 
   SuperscriptBox[
    RowBox[{"SumOver", "(", 
     RowBox[{
      FormBox[
       InterpretationBox[
        RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "2"}],
        SequenceForm["Gen", 2],
        Editable->False],
       TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
   SuperscriptBox[
    RowBox[{"SumOver", "(", 
     RowBox[{
      FormBox[
       InterpretationBox[
        RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
        SequenceForm["Gen", 3],
        Editable->False],
       TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        FormBox["\<\"\[CurlyEpsilon]\"\>",
         TraditionalForm], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        SuperscriptBox[
         FormBox["\<\"\[CurlyEpsilon]\"\>",
          TraditionalForm], "*"], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "//", "InputForm"}]], "Input"],

Cell["\<\
Spinor[Momentum[k1, d], 0, 1] . DiracGamma[Momentum[Polarization[q, I], d], \
d] . Spinor[-Momentum[k2, d], 0, 1]*
 Spinor[-Momentum[k2, d], 0, 1] . DiracGamma[Momentum[Polarization[q, -I], \
d], d] . Spinor[Momentum[k1, d], 0, 1]*
 IndexDelta[Index[Generation, 2], Index[Generation, \
3]]*SMP[\"e\"]^2*SMP[\"Q_u\"]^2*SumOver[Index[Generation, 2], 3, External]^2*
 SumOver[Index[Generation, 3], 3, External]^2*SUNFDelta[SUNFIndex[Col2], \
SUNFIndex[Col3]]^2\
\>", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{"Muu", "/.", 
   RowBox[{
    RowBox[{"Times", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"SumOver", "[", 
        RowBox[{
         RowBox[{"i1", ":", 
          RowBox[{"Index", "[", 
           RowBox[{"Generation", ",", "2"}], "]"}]}], ",", "3", ",", 
         "External"}], "]"}], "^", "2"}], ",", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"SumOver", "[", 
        RowBox[{
         RowBox[{"i2", ":", 
          RowBox[{"Index", "[", 
           RowBox[{"Generation", ",", "3"}], "]"}]}], ",", "3", ",", 
         "External"}], "]"}], "^", "2"}], ",", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"IndexDelta", "[", 
        RowBox[{"i1_", ",", "i2_"}], "]"}], "^", "2"}]}], 
     "\[IndentingNewLine]", "]"}], "\[Rule]", "Ngen"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   SuperscriptBox["\<\"e\"\>", "2"], " ", "Ngen", " ", 
   SubsuperscriptBox["Q", "u", "2"], " ", 
   SubsuperscriptBox["\[Delta]", 
    RowBox[{
     FormBox[
      FormBox["Col2",
       TraditionalForm],
      TraditionalForm], 
     FormBox[
      FormBox["Col3",
       TraditionalForm],
      TraditionalForm]}], "2"], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        FormBox["\<\"\[CurlyEpsilon]\"\>",
         TraditionalForm], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        SuperscriptBox[
         FormBox["\<\"\[CurlyEpsilon]\"\>",
          TraditionalForm], "*"], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{"Muu", "//", "SUNSimplify"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   SuperscriptBox["\<\"e\"\>", "2"], " ", "Ngen", " ", 
   SubscriptBox["C", "A"], " ", 
   SubsuperscriptBox["Q", "u", "2"], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        FormBox["\<\"\[CurlyEpsilon]\"\>",
         TraditionalForm], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}], " ", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{"-", 
        FormBox["k2",
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{"\[Gamma]", "\[CenterDot]", 
      FormBox[
       RowBox[{
        SuperscriptBox[
         FormBox["\<\"\[CurlyEpsilon]\"\>",
          TraditionalForm], "*"], "(", 
        FormBox["q",
         TraditionalForm], ")"}],
       TraditionalForm]}], ")"}], ".", 
    RowBox[{"(", 
     RowBox[{
      FormBox["\<\"\[CurlyPhi]\"\>",
       TraditionalForm], 
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], ")"}]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{"Muu", "//", "FermionSpinSum"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   SuperscriptBox["\<\"e\"\>", "2"], " ", "Ngen", " ", 
   SubscriptBox["C", "A"], " ", 
   SubsuperscriptBox["Q", "u", "2"], " ", 
   RowBox[{"tr", "(", 
    FormBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", 
        FormBox[
         FormBox["k1",
          TraditionalForm],
         TraditionalForm]}], ")"}], ".", 
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", 
        FormBox[
         RowBox[{
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}], ".", 
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", 
        FormBox[
         FormBox["k2",
          TraditionalForm],
         TraditionalForm]}], ")"}], ".", 
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", 
        FormBox[
         RowBox[{
          SuperscriptBox[
           FormBox["\<\"\[CurlyEpsilon]\"\>",
            TraditionalForm], "*"], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}]}],
     TraditionalForm], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{"Muu", "//", 
   RowBox[{"ReplaceAll", "[", 
    RowBox[{"DiracTrace", "\[Rule]", "Tr"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"4", " ", 
   SuperscriptBox["\<\"e\"\>", "2"], " ", "Ngen", " ", 
   SubscriptBox["C", "A"], " ", 
   SubsuperscriptBox["Q", "u", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox[
         FormBox["k1",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        FormBox[
         FormBox["k2",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{
          SuperscriptBox[
           FormBox["\<\"\[CurlyEpsilon]\"\>",
            TraditionalForm], "*"], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox[
         FormBox["k1",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{
          SuperscriptBox[
           FormBox["\<\"\[CurlyEpsilon]\"\>",
            TraditionalForm], "*"], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        FormBox[
         FormBox["k2",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox[
         FormBox["k1",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         FormBox["k2",
          TraditionalForm],
         TraditionalForm]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        FormBox[
         RowBox[{
          SuperscriptBox[
           FormBox["\<\"\[CurlyEpsilon]\"\>",
            TraditionalForm], "*"], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], "(", 
          FormBox["q",
           TraditionalForm], ")"}],
         TraditionalForm]}], ")"}]}]}], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{
   RowBox[{"Muu", "//", 
    RowBox[{
     RowBox[{"DoPolarizationSums", "[", 
      RowBox[{"#", ",", "q", ",", "0", ",", 
       RowBox[{"VirtualBoson", "\[Rule]", "True"}]}], "]"}], "&"}]}], "//", 
   "Factor"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"4", " ", 
   RowBox[{"(", 
    RowBox[{"d", "-", "2"}], ")"}], " ", 
   SuperscriptBox["\<\"e\"\>", "2"], " ", "Ngen", " ", 
   SubscriptBox["C", "A"], " ", 
   SubsuperscriptBox["Q", "u", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     FormBox[
      FormBox["k1",
       TraditionalForm],
      TraditionalForm], 
     FormBox["\<\"\[CenterDot]\"\>",
      TraditionalForm], 
     FormBox[
      FormBox["k2",
       TraditionalForm],
      TraditionalForm]}], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Muu", "=", 
  RowBox[{"Muu", "/.", 
   RowBox[{
    RowBox[{"Pair", "[", 
     RowBox[{
      RowBox[{"Momentum", "[", 
       RowBox[{"k1", ",", "d"}], "]"}], ",", 
      RowBox[{"Momentum", "[", 
       RowBox[{"k2", ",", "d"}], "]"}]}], "]"}], "\[Rule]", 
    RowBox[{
     RowBox[{"Q", "^", "2"}], "/", "2"}]}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{"d", "-", "2"}], ")"}], " ", 
   SuperscriptBox["\<\"e\"\>", "2"], " ", "Ngen", " ", 
   SuperscriptBox["Q", "2"], " ", 
   SubscriptBox["C", "A"], " ", 
   SubsuperscriptBox["Q", "u", "2"]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Aside: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["\[Gamma]", "*"], "\[RightArrow]", 
    RowBox[{"q", 
     RowBox[{
      OverscriptBox["q", "_"], " ", "/", 
      SuperscriptBox["\[Gamma]", "*"]}]}], "\[RightArrow]", 
    RowBox[{
     SuperscriptBox["\[Mu]", "+"], 
     SuperscriptBox["\[Mu]", "-"]}]}], TraditionalForm]]]
}], "Subsubsection"],

Cell[TextData[{
 "The total ",
 Cell[BoxData[
  FormBox[
   RowBox[{"M", "[", 
    RowBox[{
     SuperscriptBox["\[Gamma]", "*"], "\[RightArrow]", 
     RowBox[{"q", " ", 
      OverscriptBox["q", "_"]}]}], "]"}], TraditionalForm]]],
 " is ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    RowBox[{"(", 
     RowBox[{"d", "-", "2"}], ")"}], 
    SuperscriptBox["e", "2"], 
    SuperscriptBox["Q", "2"], 
    SubscriptBox["C", "A"], "Ngen", " ", 
    RowBox[{"(", 
     RowBox[{
      SubsuperscriptBox["Q", "u", "2"], "+", 
      SubsuperscriptBox["Q", "d", "2"]}], ")"}]}], TraditionalForm]]],
 ".\nCompare with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"M", "[", 
     RowBox[{
      SuperscriptBox["\[Gamma]", "*"], "\[RightArrow]", 
      RowBox[{
       SuperscriptBox["\[Mu]", "+"], 
       SuperscriptBox["\[Mu]", "-"]}]}], "]"}], "=", 
    RowBox[{"2", 
     RowBox[{"(", 
      RowBox[{"d", "-", "2"}], ")"}], 
     SuperscriptBox["e", "2"], 
     SuperscriptBox["Q", "2"]}]}], TraditionalForm]]],
 ".\nRatio ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"M", "[", 
      RowBox[{
       SuperscriptBox["\[Gamma]", "*"], "\[RightArrow]", 
       RowBox[{"q", 
        OverscriptBox["q", "_"]}]}], "]"}], "/", 
     RowBox[{"M", "[", 
      RowBox[{
       SuperscriptBox["\[Gamma]", "*"], "\[RightArrow]", 
       RowBox[{
        SuperscriptBox["\[Mu]", "+"], 
        SuperscriptBox["\[Mu]", "-"]}]}], "]"}]}], "=", 
    RowBox[{
     SubscriptBox["C", "A"], "Ngen", " ", 
     RowBox[{"(", 
      RowBox[{
       SubsuperscriptBox["Q", "u", "2"], "+", 
       SubsuperscriptBox["Q", "d", "2"]}], ")"}]}]}], TraditionalForm]]],
 "."
}], "Text"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "2. Hadronic tensor for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["\[Gamma]", "*"], "\[Rule]", 
    RowBox[{"u", " ", "X"}]}], TraditionalForm]]],
 " up to ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["\[Alpha]", "s", "2"], TraditionalForm]]]
}], "Section"],

Cell[CellGroupData[{

Cell["Amplitude generation", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", "MkAmplitude", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Options", "[", "MkAmplitude", "]"}], "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"AmplitudeTag", "\[Rule]", 
      RowBox[{"(", 
       RowBox[{"1", "&"}], ")"}]}], ",", 
     RowBox[{"DiagramSubset", "\[Rule]", "All"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"MkAmplitude", "[", 
    RowBox[{"nloops_Integer", ",", "particles_List", ",", 
     RowBox[{"OptionsPattern", "[", "]"}]}], "]"}], ":=", 
   "\[IndentingNewLine]", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
      "topologies", ",", "diagrams", ",", "faamplitude", ",", "amplitude"}], 
      "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"topologies", "=", 
       RowBox[{"CreateTopologies", "[", 
        RowBox[{"nloops", ",", 
         RowBox[{"1", "\[Rule]", 
          RowBox[{"Length", "[", "particles", "]"}]}], ",", 
         RowBox[{"ExcludeTopologies", "\[Rule]", "Tadpoles"}]}], "]"}]}], ";",
       "\[IndentingNewLine]", 
      RowBox[{"topologies", "=", 
       RowBox[{"topologies", "//", 
        RowBox[{
         RowBox[{"DiagramSelect", "[", 
          RowBox[{"#", ",", 
           RowBox[{"FreeQ", "[", 
            RowBox[{
             RowBox[{"Propagator", "[", "_", "]"}], "[", 
             RowBox[{"a_", ",", "a_"}], "]"}], "]"}]}], "]"}], "&"}]}]}], ";",
       "\[IndentingNewLine]", 
      RowBox[{"diagrams", "=", 
       RowBox[{"InsertFields", "[", 
        RowBox[{"topologies", ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"V", "[", "1", "]"}], " ", 
           RowBox[{"(*", " ", "photon", " ", "*)"}], "}"}], "\[Rule]", 
          "particles"}], ",", "\[IndentingNewLine]", 
         RowBox[{"InsertionLevel", "\[Rule]", 
          RowBox[{"{", "Classes", "}"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"GenericModel", "\[Rule]", "\"\<Lorentz\>\""}], ",", 
         "\[IndentingNewLine]", 
         RowBox[{"Model", "\[Rule]", "\"\<SMQCD-modified\>\""}], ",", 
         "\[IndentingNewLine]", 
         RowBox[{"ExcludeParticles", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"F", "[", "1", "]"}], ",", 
            RowBox[{"F", "[", "2", "]"}], ",", 
            RowBox[{"V", "[", "1", "]"}], ",", 
            RowBox[{"V", "[", "2", "]"}], ",", 
            RowBox[{"V", "[", "3", "]"}], ",", "S", ",", 
            RowBox[{"U", "[", "1", "]"}], ",", 
            RowBox[{"U", "[", "2", "]"}], ",", 
            RowBox[{"U", "[", "3", "]"}], ",", 
            RowBox[{"U", "[", "4", "]"}]}], "}"}]}]}], "\[IndentingNewLine]", 
        "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"OptionValue", "[", "DiagramSubset", "]"}], "=!=", "All"}], 
        ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"diagrams", "=", 
          RowBox[{"DiagramExtract", "[", 
           RowBox[{"diagrams", ",", 
            RowBox[{"OptionValue", "[", "DiagramSubset", "]"}]}], "]"}]}], 
         ";"}]}], "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"diagrams", "//", 
       RowBox[{
        RowBox[{"Paint", "[", 
         RowBox[{"#", ",", 
          RowBox[{"AutoEdit", "\[Rule]", "False"}], ",", 
          RowBox[{"Numbering", "\[Rule]", "Simple"}], ",", 
          RowBox[{"ColumnsXRows", "\[Rule]", 
           RowBox[{"{", 
            RowBox[{"6", ",", "2"}], "}"}]}], ",", 
          RowBox[{"ImageSize", "\[Rule]", 
           RowBox[{"{", 
            RowBox[{"800", ",", "300"}], "}"}]}]}], "]"}], "&"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"MatchQ", "[", 
         RowBox[{"diagrams", ",", 
          RowBox[{
           RowBox[{"TopologyList", "[", "_", "]"}], "[", "]"}]}], "]"}], ",", 
        
        RowBox[{"Return", "[", "0", "]"}]}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"(*", 
       RowBox[{
        RowBox[{"Print", "[", 
         RowBox[{"\"\<FA:\>\"", ",", 
          RowBox[{"CreateFeynAmp", "[", "diagrams", "]"}]}], "]"}], ";"}], 
       "*)"}], "\[IndentingNewLine]", 
      RowBox[{"faamplitude", "=", 
       RowBox[{"CreateFeynAmp", "[", 
        RowBox[{"diagrams", ",", 
         RowBox[{"PreFactor", "\[Rule]", "1"}]}], 
        RowBox[{"(*", 
         RowBox[{
          RowBox[{"-", "I"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"2", "Pi"}], ")"}], "^", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "d"}], " ", "LoopNumber"}], ")"}]}]}], "*)"}], 
        "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"amplitude", "=", 
       RowBox[{
        RowBox[{
         RowBox[{"FCFAConvert", "[", 
          RowBox[{"faamplitude", ",", "\[IndentingNewLine]", 
           RowBox[{"IncomingMomenta", "\[Rule]", 
            RowBox[{"{", "q", "}"}]}], ",", "\[IndentingNewLine]", 
           RowBox[{"OutgoingMomenta", "\[Rule]", 
            RowBox[{"{", 
             RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4", ",", "k5"}], 
             "}"}]}], ",", "\[IndentingNewLine]", 
           RowBox[{"LoopMomenta", "\[Rule]", 
            RowBox[{"{", 
             RowBox[{"l1", ",", "l2", ",", "l3", ",", "l4", ",", "l5"}], 
             "}"}]}], ",", "\[IndentingNewLine]", 
           RowBox[{"UndoChiralSplittings", "\[Rule]", "True"}], ",", 
           RowBox[{"(*", " ", 
            RowBox[{
             RowBox[{"Don", "'"}], "t", " ", "split", " ", "1", " ", "into", 
             " ", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"1", "\[PlusMinus]", "\[Gamma]5"}], ")"}], "/", 
              "2"}]}], " ", "*)"}], "\[IndentingNewLine]", 
           RowBox[{"DropSumOver", "\[Rule]", "False"}], ",", 
           "\[IndentingNewLine]", 
           RowBox[{"ChangeDimension", "\[Rule]", "d"}], ",", 
           "\[IndentingNewLine]", 
           RowBox[{"LorentzIndexNames", "\[Rule]", 
            RowBox[{"{", "mu", "}"}]}], ",", 
           RowBox[{"(*", " ", 
            RowBox[{
             RowBox[{"Make", " ", "'"}], 
             RowBox[{"mu", "'"}], " ", "to", " ", "the", " ", "incoming", " ",
              "photon", " ", "Lorentz", " ", "index"}], " ", "*)"}], 
           "\[IndentingNewLine]", 
           RowBox[{"List", "\[Rule]", "True"}], ",", "\[IndentingNewLine]", 
           RowBox[{"SMP", "\[Rule]", "True"}]}], "\[IndentingNewLine]", "]"}],
          "//", 
         RowBox[{
          RowBox[{"MapIndexed", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"#1", "*", 
              RowBox[{
               RowBox[{"OptionValue", "[", "AmplitudeTag", "]"}], "@@", 
               "#2"}]}], "&"}], ",", "#"}], "]"}], "&"}]}], "//", 
        RowBox[{"Apply", "[", "Plus", "]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"amplitude", "=", 
       RowBox[{"amplitude", "/.", 
        RowBox[{"{", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"_MQU", "\[Rule]", "0"}], ",", 
          RowBox[{"(*", " ", 
           RowBox[{"Up", "-", 
            RowBox[{
            "type", " ", "quark", " ", "mass", " ", "is", " ", "zero"}]}], 
           " ", "*)"}], "\[IndentingNewLine]", 
          RowBox[{"_MQD", "\[Rule]", "0"}], ",", " ", 
          RowBox[{"(*", " ", 
           RowBox[{"Down", "-", 
            RowBox[{
            "type", " ", "quark", " ", "mass", " ", "is", " ", "zero"}]}], 
           " ", "*)"}], "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SMP", "[", 
            RowBox[{"\"\<m_u\>\"", "|", "\"\<m_d\>\""}], "]"}], "\[Rule]", 
           "0"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"SumOver", "[", 
            RowBox[{
             RowBox[{"_SUNIndex", "|", "_SUNFIndex"}], ",", "___"}], "]"}], 
           "\[Rule]", "1"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"Pair", "[", 
            RowBox[{
             RowBox[{"LorentzIndex", "[", 
              RowBox[{"mu", ",", "d"}], "]"}], ",", 
             RowBox[{"Momentum", "[", 
              RowBox[{
               RowBox[{"Polarization", "[", 
                RowBox[{"q", ",", "I"}], "]"}], ",", "d"}], "]"}]}], "]"}], 
           "\[Rule]", "1"}]}], "\[IndentingNewLine]", "}"}]}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"amplitude", "//", "DotExpand"}], "//", "DiracTraceExpand"}],
         "//", "CalcColorFactor"}], "//", "Expand"}]}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"DiracTraceExpand", "[", "expr_", "]"}], ":=", 
   RowBox[{"expr", "//.", 
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"DiracTrace", "[", 
        RowBox[{"a___", ",", 
         RowBox[{"b_", "+", "c_"}], ",", "d___"}], "]"}], "\[RuleDelayed]", 
       RowBox[{"Distribute", "[", 
        RowBox[{"DiracTrace", "[", 
         RowBox[{"a", ",", 
          RowBox[{"b", "+", "c"}], ",", "d"}], "]"}], "]"}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"DiracTrace", "[", 
        RowBox[{"a___", ",", 
         RowBox[{"b_", "*", "c_"}], ",", "d___"}], "]"}], "\[RuleDelayed]", 
       RowBox[{
        RowBox[{"b", "*", 
         RowBox[{"DiracTrace", "[", 
          RowBox[{"a", ",", "c", ",", "d"}], "]"}]}], "/;", 
        RowBox[{"NonCommFreeQ", "[", "b", "]"}]}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"DiracTrace", "[", 
        RowBox[{"a___", ",", "b_", ",", "d___"}], "]"}], "\[RuleDelayed]", 
       RowBox[{
        RowBox[{"b", "*", 
         RowBox[{"DiracTrace", "[", 
          RowBox[{"a", ",", "d"}], "]"}]}], "/;", 
        RowBox[{"NonCommFreeQ", "[", "b", "]"}]}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"DiracTrace", "[", 
        RowBox[{"a___", ",", 
         RowBox[{"b_", "*", "c__"}], ",", "d___"}], "]"}], "\[RuleDelayed]", 
       RowBox[{
        RowBox[{"DiracTrace", "[", 
         RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], "/;", 
        RowBox[{"NonCommQ", "[", "b", "]"}]}]}]}], "\[IndentingNewLine]", 
     "}"}]}]}], ";"}]}], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
    "In", " ", "addition", " ", "to", " ", "what", " ", "it", " ", "looks", 
     " ", "like"}], ",", " ", 
    RowBox[{"this", " ", "sets", " ", 
     RowBox[{"Pair", "[", 
      RowBox[{
       RowBox[{"Momentum", "[", 
        RowBox[{"k", ",", "d"}], "]"}], ",", 
       RowBox[{"Momentum", "[", 
        RowBox[{"k", ",", "d"}], "]"}]}], "]"}], "\[IndentingNewLine]", "*", 
     " ", "to", " ", "zero"}], ",", " ", 
    RowBox[{
    "thus", " ", "automatically", " ", "simplifying", " ", "intermediate", 
     " ", "results", " ", "during", " ", "the", " ", "subsequent", 
     "\[IndentingNewLine]", "*", " ", "calculation"}], ",", " ", 
    RowBox[{
    "and", " ", "also", " ", "helps", " ", "DoPolarizationSums", " ", "to", 
     " ", "recognize", " ", "massless", " ", 
     RowBox[{"bosons", "."}]}]}], "\[IndentingNewLine]", " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"ScalarProduct", "[", 
      RowBox[{"k1", ",", "k1", ",", 
       RowBox[{"SetDimensions", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"4", ",", "d"}], "}"}]}]}], "]"}], "=", "0"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"ScalarProduct", "[", 
      RowBox[{"k2", ",", "k2", ",", 
       RowBox[{"SetDimensions", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"4", ",", "d"}], "}"}]}]}], "]"}], "=", "0"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"ScalarProduct", "[", 
      RowBox[{"k3", ",", "k3", ",", 
       RowBox[{"SetDimensions", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"4", ",", "d"}], "}"}]}]}], "]"}], "=", "0"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"ScalarProduct", "[", 
      RowBox[{"k4", ",", "k4", ",", 
       RowBox[{"SetDimensions", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"4", ",", "d"}], "}"}]}]}], "]"}], "=", "0"}], 
    ";"}]}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", 
   RowBox[{"amp", ",", "camp"}], "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"u0", "=", 
       RowBox[{"F", "[", 
        RowBox[{"3", ",", 
         RowBox[{"{", "1", "}"}]}], "]"}]}], ",", 
      RowBox[{"u", "=", 
       RowBox[{"F", "[", "3", "]"}]}], ",", 
      RowBox[{"d", "=", 
       RowBox[{"F", "[", "4", "]"}]}], ",", 
      RowBox[{"g", "=", 
       RowBox[{"V", "[", "5", "]"}]}], ",", 
      RowBox[{"h", "=", 
       RowBox[{"U", "[", "5", "]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{
      RowBox[{
      "Amplitudes", " ", "for", " ", "each", " ", "set", " ", "of", " ", 
       "outgoing", " ", "particles"}], ",", " ", 
      RowBox[{
      "grouped", " ", "by", " ", "the", " ", "strong", " ", "coupling", " ", 
       "power"}]}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{"(*", " ", 
     RowBox[{"Quark", " ", "amplitudes"}], " ", "*)"}], "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uU\>\"", ",", "0"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}]}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uU\>\"", ",", "0", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uU\>\"", ",", "2"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"1", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}]}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uU\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uU\>\"", ",", "4"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"2", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}]}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uU\>\"", ",", "4", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uUg\>\"", ",", "1"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}], ",", "g"}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uUg\>\"", ",", "1", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uUg\>\"", ",", "3"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"1", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}], ",", "g"}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uUg\>\"", ",", "3", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uUgg\>\"", ",", "2"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}], ",", "g", ",", "g"}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uUgg\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uUhH\>\"", ",", "2"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}], ",", "h", ",", 
          RowBox[{"-", "h"}]}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uUhH\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uUuU\>\"", ",", "2"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}], ",", "u", ",", 
          RowBox[{"-", "u"}]}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uUuU\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"\"\<uUdD\>\"", ",", "2"}], "]"}], "=", 
      RowBox[{"MkAmplitude", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", 
         RowBox[{"u0", ",", 
          RowBox[{"-", "u"}], ",", "d", ",", 
          RowBox[{"-", "d"}]}], "}"}], ",", 
        RowBox[{"AmplitudeTag", "\[Rule]", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"A", "[", 
            RowBox[{"\"\<uUdD\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
          ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"Anti", "-", 
       RowBox[{"quark", " ", "amplitudes"}]}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"(*", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<Uu\>\"", ",", "0"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u"}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<Uu\>\"", ",", "0", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<Uu\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"1", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u"}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<Uu\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<Uu\>\"", ",", "4"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"2", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u"}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<Uu\>\"", ",", "4", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<Uug\>\"", ",", "1"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u", ",", "g"}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<Uug\>\"", ",", "1", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<Uug\>\"", ",", "3"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"1", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u", ",", "g"}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<Uug\>\"", ",", "3", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<Uugg\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u", ",", "g", ",", "g"}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<Uugg\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<UuhH\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u", ",", "h", ",", 
            RowBox[{"-", "h"}]}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<UuhH\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<UuuU\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u", ",", "u", ",", 
            RowBox[{"-", "u"}]}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<UuuU\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<UudD\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "u0"}], ",", "u", ",", "d", ",", 
            RowBox[{"-", "d"}]}], "}"}], ",", 
          RowBox[{"AmplitudeTag", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"A", "[", 
              RowBox[{"\"\<UudD\>\"", ",", "2", ",", "#"}], "]"}], "&"}], 
            ")"}]}]}], "]"}]}], ";"}], "\[IndentingNewLine]", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"Gluon", " ", "amplitudes"}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"(*", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<guU\>\"", ",", "1"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{"g", ",", "u", ",", 
            RowBox[{"-", "u"}]}], "}"}]}], "]"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<gdD\>\"", ",", "1"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{"g", ",", "d", ",", 
            RowBox[{"-", "d"}]}], "}"}]}], "]"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<gguU\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{"g", ",", "g", ",", "u", ",", 
            RowBox[{"-", "u"}]}], "}"}]}], "]"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"amp", "[", 
         RowBox[{"\"\<ggdD\>\"", ",", "2"}], "]"}], "=", 
        RowBox[{"MkAmplitude", "[", 
         RowBox[{"0", ",", 
          RowBox[{"{", 
           RowBox[{"g", ",", "g", ",", "d", ",", 
            RowBox[{"-", "d"}]}], "}"}]}], "]"}]}], ";"}], 
      "\[IndentingNewLine]", "*)"}], "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"The", " ", "rest", " ", "is", " ", "zero"}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"amp", "[", 
       RowBox[{"_", ",", "_"}], "]"}], "=", "0"}], ";", "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"Conjugated", " ", "amplitudes"}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"camp", "[", 
       RowBox[{"particles_", ",", "gspower_"}], "]"}], ":=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"amp", "[", 
             RowBox[{"particles", ",", "gspower"}], "]"}], "/.", 
            RowBox[{"A", "\[Rule]", "B"}]}], "/.", 
           RowBox[{"mu", "\[Rule]", "nu"}]}], "/.", 
          RowBox[{"l1", "->", "r1"}]}], "/.", 
         RowBox[{"l2", "\[Rule]", "r2"}]}], "//", "FCRenameDummyIndices"}], "//",
        "ComplexConjugate"}]}], ";"}]}], "\[IndentingNewLine]", "]"}], 
  ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         {PointSize[0.04], PointBox[{11., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/bdcd/0.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/bdcd/0.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lH1MjVEcx0+6vaMk6s5L8hzclYinEv1xfiE1L83L2qKykvuMScVCVpgt
5MpLmThXhZmVvKdbt3lp3IuaJ6Vm5bXymoQukWRdh+08v+3s2Wfn2XPO7/v7
fh+fNanLtUMQQnPZ+vdsb7OyUgP6XxgitcVTCyjjtL7n230xoNkzLVNFxrU+
holLMbRkNheYZS9ACZpwQzKGVY6+kfESY6POvjcLw49b04b3WT0BeVnAYReG
Xm/vx3mUccyB7PLNGKTsqHh/kTEEZseuxGBOdNlQJ48G1JqxryMIw/20jFKt
xLj22LCFQzEUdjiZbRDjtAGPxR0CpExZsbWIjgLk5jnvQbUAmcd/7Q8RGXuJ
acF6tm963d4kewC66nTGf68AEBa0NVli3Nrt922XALq3/XF2iLGmpSEkR4Bf
l17eKKIjAZ1YEzyiWADjDOP9IJFxTLJoUyPAweYnlfWyOyDHvRcWfxCg5JHa
pJUYN678neiO4YVGpR60jgB0xHRodSiGWZbMlgLKuMdwvTAew7OdP10CRMaN
hs1ztmCwJaHldbIboMjf83VMn66Uc5VJEuOYJDkhE8PVhS97B6yugEo3moPX
Y6hIOVoeIDI2qpaPWYThfM3z1VppOOt3J8nAGIJp9G1KhzF9pE+X+wXItU0t
q5eHAmrPmkQbBGhIiBhtgxi7fU76UybAsug5roGiC6Cl/hHWwwKUPRxXLUnO
rD+L97XdAnwI61hAqZPCYUZzz0PZUXnfdLHq46DVQfmevfOR6Jmig3JeXPr5
MEmyV+7TtUQ4o6d2yn3bijcsaZBVSj+v7/q4qJBK6bdmekV+iGir6JF7OOdk
sjRE0euVq3vnKWqj6DnJz2FZk4wUvdUxffW2CCnzcFaf+L5OGiR8Xql2pbhO
/kP4PE25499pxAHC533vZI7bftpPuB+y7njmdln7CPfLuSjfKVHST8L99CU8
PatC7iXcb8YfY1vHid8J92O3oXyyjloI9+ta/ZU2i/Ur4X728gm92SR3E+53
DT1bUkE/Ep4Hj2zdqGPSe8LzUlpit22T+IbwPJ2uyjkUgdoJz1tPbdwOP/1T
wvM4Ie9efqXYTHheJ+y4/jRRX0d4nqEztjVdX02UvP+rL3tu8f/BX/9FoHw=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.25, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          13.000000000005457`, 14.000000000003638`}}], 
         PolygonBox[{{17.0939696961967, 14.584852813742385`}, {
          15.849461761308376`, 14.811126983722081`}, {15.962598846298224`, 
          14.019167388793148`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.280796897832168`, 15.429621715174807`}, {0, -1}], 
         LineBox[{{19.99999999999909, 5.000000000003638}, {13.00000000000091, 
          6.}}], PolygonBox[{{15.9060303038033, 5.584852813742386}, {
          17.150538238691624`, 5.8111269837220805`}, {17.03740115370178, 
          5.019167388793147}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.280796897832168`, 4.570378284825192}, {0, 1}], 
         LineBox[{{6.500000000001819, 10.}, {13., 13.999999999996362`}}], 
         PolygonBox[{{10.260994990022727`, 12.3144584553986}, {
          9.448643980243007, 11.344878217919582`}, {9.029366039711539, 
          12.026204871283216`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.332600380984282, 13.056274380900543`}, {1, -1}], 
         LineBox[{{6.499999999989086, 10.000000000003638`}, {
          12.999999999987267`, 6.000000000007276}}], 
         PolygonBox[{{9.239005009977273, 8.3144584553986}, {
          10.051356019756993`, 7.3448782179195815`}, {10.470633960288461`, 
          8.026204871283216}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.332600380984282, 6.943725619099458}, \
{1, 1}], LineBox[CompressedData["
1:eJxFlglUVEcWhluQsUVUVDiyCza87npPB3A0gCBeIkwI4IIs4hI2QRZPD4uD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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14.8493, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{6.5, 10.}], PointBox[{13., 14.}], 
          PointBox[{13., 6.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/dedfef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/dedfef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         CircleBox[{15.725, 7.904999999999999}, 2.1106041789023346`, {
          2.8558085915452476`, 5.5547723604394585`}], 
         PolygonBox[{{14.175505634327077`, 6.351385758707179}, {
          15.030237193201469`, 5.418951330844207}, {15.418751538144374`, 
          6.118277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673118`, 5.309856222011617}, {1, 1}], LineBox[CompressedData["
1:eJwVUgs0VWkbPrk2mKQkl86x9zZdxD5nb7NyCXOe8ZtETVfnhIo687uTOL+j
g5J+uWWiJlF0o8ZlNDWlXKYoMkdCLZeUUSk10a9EIYX+3bPWu971rPd9vu95
1/uSsogN/mo8Hi+Kiy85MOALrDCD9wUkDoesi1ULs4KCn7ta6Uni2IBua0Ka
FVqyT3+XUUTCu9N60+R1K+hWZjxdOkLisYdtxO+a1jDTf1MbSlPoaJjOlPhY
Y+aSbR8bN1KQPRq96lRujQjfnma+jEJldpBvli6N1gqbCFMvCtbracVv3jTu
ToYVGdtT+Pae4orGcRoT+p4uci0Kx6w0a9KbaBxIlNSeaySRVXhHZ8f/aHg8
H7wdm0SCsM9+umeaxtxbmaHRYhJBiVNvvLSEqM1xT42fJvCq/POzo5pC6N7V
M7/SQOCS87N+J67/d9MYvZ25BHydbVoVwzSkkr8FpkoCB7Xyxi4+oZFq/+Gu
MJiAWltjzGQjjbAaM4flHHelwjryztOIs807N7qLQLjSpjvjIA35mgppfA6B
eRl5TH84jR1Dd6qT6glIeraKptbQeNz70C3xEwFxyls70paGg57O2cPOJGYy
aVW2S2jkehhVuqeQGDhek/zBksbweU/hxk4S9wyMqgViGlmNF3uWW1CwkHhu
TQimsS149X1lKIW/qpJdthRzfu3OJQ0VUdjT//qt0Qcas2IPPKxto5B9b8M7
xQYh3uh0/vzTSwrVKzZ+TLsghDLVVzu3j8K88OjJES0RCjIOuTxSUQhea7bP
c70ICSem50tyKNyUfXoSkyaCadcfBgabKBSeVH8wWCbCzK72lfpfU6DuZE7r
XRNhD2/lkGMtie38yYL4KhHgU9dyPYrEAl5nztYSEQK+9TUqsyKhrlIEh2aJ
EDjro5R4TUB3/yIzWi5CXlmfVmwVAc3/iI19JJze7P2mgsME/NoN25IcRFhT
2FNkHUtgxrqdFoYk58ewYq57JAH5yu2JqXoi7GRvzMnh9pGqP+vWxCchgr96
0TOSSWDqUp+4dEiInlMpJpKrBFoblCH/7RdCLns63DxAYKQh2tWT4xl//jB8
YDGJTNXcpVXDQtz615m170NJdBcXjZZoizA+ns8vuERC91bp2A1rEd5NjpQq
x0i4rYp33yYToURjTksnS0H3wg7NV0Ui/KBMHB7xpfBjW1nxio8i7M4/f/rM
Lgrn51UdHrNk4HOof++SOG5fujV/3pQyiJmducXUn8Kz6pgTexMYKJNt9nTY
UZDUiC+knGYw+8RUW8g4iZfhD6ItKhiY3FPdGyshwZtWfzh1k0FgSrVXjJTE
Fi/H75dxPNmn0DFHncS+xruLrK8yqFk2cfcyNz8tzBNbFTJojqkt95UTOKVm
dmw0ncGKX/quBTgS+Hz26OaAnQz2q55ouesT+NR9RWbB+euycpWdfWQO2wef
xQHfMaAWvb0UeNkcsoMWLt1LGUR5L3t+45Q5DpV787IWMLAbOBrHFJqj96RH
80xDBrud9OZvqTGH16iRhY8Bg2z/E9rfD5vjTkUNZFy9/bmfVMgnsDFpQec0
p2+rFFpqrSIQkr82y8aKwd5vVB0WCQQCFPuNLMUMNGJajbdeIaDtsbpsjjeD
dcZG47n9BJI0Hise72Lwyag7WTSXhLlb2oaXJxiQvYR/tIiEwr9q5sW/GDx3
e3pabk9i5L4gt/ktg+Xj2pbXl5LI2bGkS8+EhcZEWLqDJomykoBtA44snN4V
xxffJrA5IrPWTMqCaVqKwngCcWqFs+39WXixmoa8hQTWp4pzTYNYNORPqfyy
zRFbfbZTx49FtUZYVM0rAep/wpOsVSxK+yz8oqQCnG3P3m7IsGj6cKbXrpcP
Y/9emxB9Fo2Ua4sqnY/WLqO/IwcZtCoXGiR78nG08te8jtsMnCVm897Y88Hu
8nB7UcKAeerescKGj9J/6oKOHGSglj06JHTk43V0d06IkkFPgzzrt7V8RC58
mJAYyqAuaV1BWjgfdd/I3Xb7M5iKc+vWPcKHfCIp6edABh62kkndej6my1MQ
F8ndo7fB9IMJPhJVQ+p1+xgES517LG0FqC6Lk5ccZ7DsD9cDkXIBrnU5pK+v
ZDBpfvxkXakATjom40XdDDIDkwbXtnPzfwjavJLHYjj3YO3uPgHeN83xlC5m
oX11U/O/2wRo+kcNwT+yWFEcs/jIMQEapU8UZyJY2C3Y3FLP/We3ZvDdSDoL
51+EnV+f58PaJa4/P5/FwjcvZog1OX/Xq8YvFLB4q3jdMbp6AS6n3T++i6u3
dNZr5B0xw+TOetdiTn9yfv6mgkFTKF1OqQm59/dK26ZVMlO0NHNYzWLq8xeY
4P8d3g0h
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.44794654332688, 9.690143777988382}, \
{-1, -1}], 
         {PointSize[0.04], PointBox[{11., 10.}], PointBox[{17.3, 6.5}], 
          PointBox[{13.700000000000001`, 8.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/bdce/dfefef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/bdce/dfefef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.50205345667312, 15.000143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         CircleBox[{15.275, 12.905}, 
          2.1106041789023346`, {-2.4131797068496654`, 0.28578406204454554`}], 
         PolygonBox[{{16.824494365672923`, 11.35138575870718}, {
          15.969762806798531`, 10.418951330844207`}, {15.581248461855626`, 
          11.118277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.44794654332688, 10.309856222011618`}, {-1, 1}], LineBox[CompressedData["
1:eJwtVQlQlEcaHRVQQUEOFUGIRiMzf/f/zz8YFMSEh7gqAuE+1ooHIijHBvBC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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673118`, 14.690143777988382`}, {1, -1}], 
         {PointSize[0.04], PointBox[{11., 10.}], PointBox[{17.3, 13.5}], 
          PointBox[{13.700000000000001`, 11.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["3",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becd/dfefef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becd/dfefef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAWIQHaEDJBh4HBjAQMXhRa5yk54xtwODQ2Jyh7mKw3mv
kNK4NC4HBgXNJwIpKg5ntv4MnTCT04Gh4KDX8RoVB4asgzpHznA4MCx4EH8M
yL+cUzPj5392oP42ZT6g+g239d/pGwP5Gr/PNAPNy3nWnJaWxubAYHF18s//
yg76QY6+c2ayOjDsKM3OP6HsYKRyS/3iGRagfjv3p7OVHWycbp5iYQDyDY7G
6TYrOzybVhllbszswMAR8lIHyGd7+so8M40JyH/97DFQ/ccOu4ezZjIC+Y4K
uUDz2vJunzlzhsGB4cT+2G9A+4rfy0759/+/PcMG1jX1QPfEM366k5L2z56B
YaM4J9C9ixMcNM+c+WPPEKGdeQDonybtA9eMjX/D+ce6+JRnz/wJV795zqNs
RoYfcPOk1nsC/ffNHmafS1Wz7akzX+xh7pE+pDVB1/izPcy90XG83hNnfrSH
+SeOf86VN//f28P829I+I2H3mTf2sPDoP//7VN/Ml/aw8JpT9zE1Le2ZPSw8
e1Y6ujsYP7aHhTdDKv88SYYH9rD4mPHg28x7M2/aw+LLpMkkoNb4sj0sPi22
7F8lMeukPSy+E3rX7daftdMelh5AqaRhV6ctLL0AAP+Y434=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000005457`, 15.000000000003638`}, {
          12.500000000003638`, 10.}}], 
         PolygonBox[{{16.749230176602705`, 12.832820117735137`}, {
          15.528889744907202`, 12.5}, {15.972649901887385`, 
          11.834359764529726`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.995014695850895`, 13.184877956223659`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {12.499999999998181`, 10.}}], 
         PolygonBox[{{15.750769823397293`, 7.832820117735137}, {
          16.971110255092796`, 7.5}, {16.527350098112613`, 
          6.834359764529726}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.504985304149102`, 8.184877956223659}, {-1, -1}], LineBox[CompressedData["
1:eJxFlAtMU1cYx68ortRMmdNFlIrcezm3UAHRKTVm8kEU6HQEsYjiI1AFZVBg
PKImtrZINiJL6Hj5mAiKM8ymQDViJeGh8dGhIkqpuDDXCZiUsYAJlUc0eLol
305ycvPLybnne/z/n78qJyHNg2GYGLrdX9OHWbr8gfl38VBw1vfyHxtYYOJV
n6WG8pB3OmXGr5Cy7qLEL5GHLrmk+52FsqgjqDaHh0cT+1YkMhwwuzt+7dTw
YPHzzvlEQdkQ1+bU8lARfuDVIwNlx4X7n+bz8HWkpebbPsrjvztNyTzUDJYG
di/hgZEazvVu4OHpytBpn28odxojPb0or6u+136Csu5g8/peDmRS18nSy5R7
/hpoPMNBqPigsrWNcsrM9NRuDs63ZIqmn1A+UjjY7sOBNVChyeqlXBK72DzA
wnbzqhVN3ZQNC5VHrrBQo3z6VVcn5amqYkU+C7eDQ0aCjO7znZrMbSyUJc5E
l5dRZjJ26UJYmJhTUuSRS9lyf/yuhIXv9oa6GuIoOzaXO31YsGkTsh8EU47d
FrCSYyFN3WqqXkR5/GbuezkLh0XvP5e/o/mLjMnVySw450WVVw5Snroj0n7P
wqXlSbbtLykfe5Xd18qCvfHqD1f7KcfW16tcLCwalem0ryl7g6RhHQer5DsS
IiYpL7N1z+ZxUFyRbB34wh1f1OytZg4qi4Q0G7jz19euGeHgWtzysbUFlOWj
jklfHuzDfg8vNrnrvcU2FM3DyBbtz+lvKeeqxkLSeVgQc9ReER5A84s9rj/O
/6cTnZs9vMQ6HlyWksAs6//nO836Q+cWEryf7u2Qp8YT/H+Sedj7SSnB99Wa
/vOyDoLxabJ+CeecBOOXtDeaOkUC5hcitp3USwTMX7Zjz7E9goD1SRlukN0g
AtZP+bZNnekrYH3t/8wdfeglYP37S7YmZIwT7E+w3vTTxucE+2e40iNOMhPs
r2LiqOZDGcH+N9Z53lbnEtRHsfXZ0j4lQf0cMq65uz6CoL5uxv05FB9GUH+F
j1uKKlYT1Gd9ROXk/LUE9fvcdcG4NYqgvu0vxLqC/QT1H/Y4Y1PkKYL+yH7R
M//NdYL+qVtcYA/7m6C/WtpjVFVBAvpPbb3k/2WOgP5UOMYMUy0C+jdvTv3Q
tVkB/Z3f9jrMK0aK/i8rrPV0nJbifPjx8OB+9jcpzg+le8gwgThfPgLdtL2a

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {10.75, 8.354800000000001}, {0, 1}], 
         CircleBox[{6.25, 10.618749999999999`}, 2.8187499999999996`, {
          3.3629070959375835`, 6.0618708648317945`}], 
         PolygonBox[{{5.65, 7.8}, {6.85, 7.3999999999999995`}, {6.85, 8.2}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.25, 6.8548}, {0, 1}], 
         CircleBox[{6.25, 9.381250000000001}, 2.8187499999999996`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{6.85, 12.2}, {5.65, 12.6}, {5.65, 
          11.799999999999999`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.25, 13.145199999999999`}, {0, -1}], 
         {PointSize[0.04], PointBox[{3.5, 10.}], PointBox[{12.5, 10.}], 
          PointBox[{9., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["4",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/bece/efdfdf.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/bece/efdfdf.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAWIQHaEDJBh4HBjAQMXhRa5yk54xtwODQ2Jyh7mKw3mv
kNK4NC4HBgXNJwIpKg5ntv4MnTCT04Gh4KDX8RoVB4asgzpHznA4MCx4EH8M
yL+cUzPj5392oP42ZT6g+g239d/pGwP5Gr/PNAPNy3nWnJaWxubAYHF18s//
yg76QY6+c2ayOjDsKM3OP6HsYKRyS/3iGRagfjv3p7OVHWycbp5iYQDyDY7G
6TYrOzybVhllbszswMAR8lIHyGd7+so8M40JyH/97DFQ/ccOu4ezZjIC+Y4K
uUDz2vJunzlzhsGB4cT+2G9A+4rfy0759/+/PcMG1jX1QPfEM366k5L2z56B
YaM4J9C9ixMcNM+c+WPPEKGdeQDonybtA9eMjX/D+ce6+JRnz/wJV795zqNs
RoYfcPOk1nsC/ffNHmafS1Wz7akzX+xh7pE+pDVB1/izPcy90XG83hNnfrSH
+SeOf86VN//f28P829I+I2H3mTf2sPDoP//7VN/Ml/aw8JpT9zE1Le2ZPSw8
e1Y6ujsYP7aHhTdDKv88SYYH9rD4mPHg28x7M2/aw+LLpMkkoNb4sj0sPi22
7F8lMeukPSy+E3rX7daftdMelh5AqaRhV6ctLL0AAP+Y434=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000005457`, 15.000000000003638`}, {
          12.500000000003638`, 10.}}], 
         PolygonBox[{{16.749230176602705`, 12.832820117735137`}, {
          15.528889744907202`, 12.5}, {15.972649901887385`, 
          11.834359764529726`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.995014695850895`, 13.184877956223659`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {12.499999999998181`, 10.}}], 
         PolygonBox[{{15.750769823397293`, 7.832820117735137}, {
          16.971110255092796`, 7.5}, {16.527350098112613`, 
          6.834359764529726}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.504985304149102`, 8.184877956223659}, {-1, -1}], LineBox[CompressedData["
1:eJxFlAtMU1cYx68ortRMmdNFlIrcezm3UAHRKTVm8kEU6HQEsYjiI1AFZVBg
PKImtrZINiJL6Hj5mAiKM8ymQDViJeGh8dGhIkqpuDDXCZiUsYAJlUc0eLol
305ycvPLybnne/z/n78qJyHNg2GYGLrdX9OHWbr8gfl38VBw1vfyHxtYYOJV
n6WG8pB3OmXGr5Cy7qLEL5GHLrmk+52FsqgjqDaHh0cT+1YkMhwwuzt+7dTw
YPHzzvlEQdkQ1+bU8lARfuDVIwNlx4X7n+bz8HWkpebbPsrjvztNyTzUDJYG
di/hgZEazvVu4OHpytBpn28odxojPb0or6u+136Csu5g8/peDmRS18nSy5R7
/hpoPMNBqPigsrWNcsrM9NRuDs63ZIqmn1A+UjjY7sOBNVChyeqlXBK72DzA
wnbzqhVN3ZQNC5VHrrBQo3z6VVcn5amqYkU+C7eDQ0aCjO7znZrMbSyUJc5E
l5dRZjJ26UJYmJhTUuSRS9lyf/yuhIXv9oa6GuIoOzaXO31YsGkTsh8EU47d
FrCSYyFN3WqqXkR5/GbuezkLh0XvP5e/o/mLjMnVySw450WVVw5Snroj0n7P
wqXlSbbtLykfe5Xd18qCvfHqD1f7KcfW16tcLCwalem0ryl7g6RhHQer5DsS
IiYpL7N1z+ZxUFyRbB34wh1f1OytZg4qi4Q0G7jz19euGeHgWtzysbUFlOWj
jklfHuzDfg8vNrnrvcU2FM3DyBbtz+lvKeeqxkLSeVgQc9ReER5A84s9rj/O
/6cTnZs9vMQ6HlyWksAs6//nO836Q+cWEryf7u2Qp8YT/H+Sedj7SSnB99Wa
/vOyDoLxabJ+CeecBOOXtDeaOkUC5hcitp3USwTMX7Zjz7E9goD1SRlukN0g
AtZP+bZNnekrYH3t/8wdfeglYP37S7YmZIwT7E+w3vTTxucE+2e40iNOMhPs
r2LiqOZDGcH+N9Z53lbnEtRHsfXZ0j4lQf0cMq65uz6CoL5uxv05FB9GUH+F
j1uKKlYT1Gd9ROXk/LUE9fvcdcG4NYqgvu0vxLqC/QT1H/Y4Y1PkKYL+yH7R
M//NdYL+qVtcYA/7m6C/WtpjVFVBAvpPbb3k/2WOgP5UOMYMUy0C+jdvTv3Q
tVkB/Z3f9jrMK0aK/i8rrPV0nJbifPjx8OB+9jcpzg+le8gwgThfPgLdtL2a

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {10.75, 8.354800000000001}, {0, 1}], 
         CircleBox[{6.25, 10.618749999999999`}, 2.8187499999999996`, {
          3.3629070959375835`, 6.0618708648317945`}], 
         PolygonBox[{{5.65, 7.8}, {6.85, 7.3999999999999995`}, {6.85, 8.2}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.25, 6.8548}, {0, 1}], 
         CircleBox[{6.25, 9.381250000000001}, 2.8187499999999996`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{6.85, 12.2}, {5.65, 12.6}, {5.65, 
          11.799999999999999`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.25, 13.145199999999999`}, {0, -1}], 
         {PointSize[0.04], PointBox[{3.5, 10.}], PointBox[{12.5, 10.}], 
          PointBox[{9., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["5",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/bece/efdfdf.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/bece/efdfdf.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{11.000000000003638`, 7.099999999998545}, {
          5.500000000001819, 10.}}], 
         PolygonBox[{{8.780741460291178, 8.270154502755561}, {
          7.905822204538448, 9.183673137438559}, {7.532694874879195, 
          8.476017857050321}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.1318657987852, 7.783717894247791}, \
{1, 1}], LineBox[CompressedData["
1:eJwV1AlMk2cYB/AX1iEig2Kh3LPIUUCuVRBkiJ+IrMNkK4pIEAke3A5qIFoU
tOmcVjmGytimaEoAOWxYQWWD4azAMGjFckgIyCwIyqRKnQ7ZxLH/+yVfml+e
93ief77UZW/OthRjQsguvPSX6Jbw+DIkgMKaIYnX2jTNsCFctuwRhyF6id/u
PbD6B1P2MCxvDZy3gqUJ2blz8PW4hr4eH4awlTOTfOyPlE0/kcHSGEFWIbxw
O7xfCIslb0v1cH1zq9IB1oY6HhLbMETCb3v+zxqGiMZ2+H/AZUiZIS36GawI
aCGX4UCP4/cmaT2lrnO9LfoYTWiahbVdi8wDuDvJ+Rih50XsrNxpx5B4swNu
q2D128/qNLBW3vrhFjhgfG69lz1DQjin3XLoet79wxmwcMws+gKt1xiximCV
54lPumBexZcuJXDs/OXrT2FG5tghhiMjV8wbYf6yHP3uENgQFyVfCStMOCtm
cJ88R1LKhUWfCs2/gTtOKR+aw2LiWmcDW78wn35Dz5vx6VWgf+apkPWA3qdf
2egL62Wa4EpYxMxkdGH+yE3vBUl0ffcVVTpcHBq0yZbmGWwa6wpLdNc87tJ8
bhaaLiBP8+2e7/Jh1WSey3O4WKhd8oLZ3Oj97+D4xr25495Yn7bVsAb7RwId
U7+H1X3sDUdgz2ehZQm0nsT4P4Hlf/n7esG8n2oi9qI/1d34cRasHUy48wqe
il/89qUX8ksZ9SvCfIq5qkNPYUVF134+8qkX5RnrYe1Har9bMEszMPselt7Z
V7rVAX2lNGx2oPcrdat6YJJWUxIO66bPiL0cGZI+OzaRRvtxvig8COdN+Fmf
h1U9wqkLsCEr5HEH3f9fsH0tXPz3j+xJ6k4mtQwuqI64vgQzVu36JFjTlBnG
QR7S76q3W8OsWfcaB9igXVtxA/ePGNrOWNP60TS1EB7yiIoitH5k9aUh9H/A
+7ae5qeOXpeQDHc09G9rghXD3vzXmF81eJbkwskOLwfPw+o/hpz8YHFtWZ6Q
+oqJaBLz8zivLLj0exH4K87RPLa0xBH4FysjYQTNr5l7dDksXHAqeuOJ80rT
lwvgcu7YaSWs85ROS2Dzw6+4X8FM7f34MZp/SdDZEJgIHIJ3oL/0iuNdbFgU
bmE7AWuD9mnm+ajHHbsowXyKfBPLWVhnc6mX5iEvmfJ9AfOCZGlNcM29TstF
WOFikhTuxBCfgvKXNjhPvNBt9xvMfi7LofcF2Ac95jszJKx5sDMZVlc6rj4M
B3Ys9hbBhhtRbfUwa2liUyvtdyHa4ld4/2sy9whWZRkJmuDEZYG9izCvrXrs
azj2xU03Dv1+Ms/FbIRHWg4OO8MkailsGvcXVFnEONJ64mDmUbh4NKLdDDZk
eBSawtLEkPlZnKf9OPvsecwjnbNsuAUrlrnrXOn3JQvyOwVLTxjtUiOPqd+N
T0bC7OA/v8iB5Ub1of/SvNgXLdbBnjt7N1yFxbnljXawpKFKnUDzGjBut4cN
fT2m5nAyXxkcCrN+Htnc6YHzeY15+bCSk5cohUUDV10ewvq1qgEhHLCHyf4c
/ZhqNl5whhXhsrB+WE//n92xPxWPE/M/mCEAHQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.884307456565622`, 5.282433250601188}, {1, 1}], 
         LineBox[{{10.999999999996362`, 12.950000000000793`}, {
          10.999999999996362`, 7.100000000000123}}], 
         PolygonBox[{{11., 10.625}, {10.6, 9.425}, {11.4, 9.425}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.9452, 10.025}, {-1, 0}], 
         LineBox[{{11., 12.950000000000728`}, {5.500000000005457, 10.}}], 
         PolygonBox[{{7.721255002194882, 11.191400410268166`}, {
          8.589678604650562, 12.11109625493525}, {8.967811390959675, 
          11.406102924528424`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.12216772652609, 12.236124577663222`}, \
{1, -1}], LineBox[CompressedData["
1:eJw91H0w23ccB/Cs3ISqxDRluBb1EEaX6xBPrbgefVIsVDWiy83DWM7JlCsL
Gl27KqW2sXXqccRT6VHP0m22WhYuKzZbccxT5yiptJRwHvb+7o/lLve+V36/
7ye/zzefbyw/TOBG76FQKDy8SVLUu3g5cf7PIF4/fxrJUvVoWpGdUp9pkhKH
+Fci5P71UrUYWT7DZbCRSd4VRW5IExFtWxdpfWc3Q+WI9V2igkWk/2N6XQmS
ujO3OY5k3i6P9kdOcdaT/kYKxB6M7Xc4lFjNjdhleLguavsh3JTuFG2Aeko6
1+wTeErq6OYOC6iSTXeY4zqqFw/zZeIvaHCQlfpWNaw/3m674oDvkX1nNAOP
GM4e/wcebJ9ZMDsCS38tJFaHOccEwRqtVUtyf4qNu2Ma3Jlsdlgf9ViJ7LBi
cv+xYI8jMFNtWvMAjm2azQ6FRT29hc2wdercdCZ5vu2N5CpY5JY30wCHnR64
cpNcZ7Q4D8MpB2k14XBIoEXLKixZ/mHcinjd8pu96JtzJkc6ied91LheegDu
UaQdLIRTKo2d9sPluxN9fnDvhfxLWvDg4+pLGqR+JNf4Galntq+wEdZO6NDq
gGtntFbiYH4L6046cYzoHgue/7lk8L/9+6vbSpvUpfHjX6B/kdBIOIfPldzy
pFK4PCc8ewTO1zvgcg6WfGUuGyN93KPTduyxTwP8i0uwBb3PtxXOL8pYJH2I
Ilm/J8Lz0geubHhLS+HuBd99Tx4hhIudu4VvwRSb11QprNFnVK8xcb+tQcEU
PL6vKGoBjs1Y4ZiiX+YtGu85zNqKEwbAYbKEJA3cmXhdlkp+/3OyRUPUG3mk
Cv4W5qxW8Jxh5o/5G/VkfmgNk3xY0Jt6tBG+/tDUNwvWmPrxyuCtulCnFpgT
rnM/k8y/XpbPGLkuPuRxnsxP1dAHm7AF02fBDF693SSnYz/uyiP9R0g/2nae
ZrBJotHnubCjqsLOBB7J0TP3Jv17LzW/SSyIl6uxXz1ju/I51GM9OeNVDSed
pZl3wSlDDpRIeLW9y/oqrPj48rsOsEmwescDzrL1vbCDOrWayvPL6F9hZPP+
NFwQMORZBlNfRdgPwwrV4hsBZL+UYvpTmDO5prNrh3WB98MW4N5f5HltsIJn
eVWXzOnF19lJ8Hwk1ccF1jZdEx+HqcVD3XFwyJfXTjPgkWZ2fiXcG+Vlv2mL
eW9SmE7Cw9de5qrgU46JxSZkblPjLJdhxc3gy+Scz+817NuGLZRU/ytkv/RO
tBmj3tSfbYFfkznu533qAbNaQ87WwGERLkcjYbrI7kktrGZzq/JgN4faY0Ww
uSg2sQuWTMxOiOEQlc7JKbL+pPK3AJiS/pMWBf0LZlrWyblxXApdYsCUnP61
QXIODEuUh8j8pOmu3ID9B/YYm8NN32cms8nvU8/Y1CWWxLHJfj3zGxUuoL5A
lfuCnIvejWAHGazWXXmbT/zRqWYJPOgTN3QYznKtt/GCszw7ijRk7j57zlaj
/6yiOr9xuGG0Y6cMVr9cejoA6zu5KgNgTaFB8h/wFvkftsG6mJgYzMm/C7Aa
LA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.578595720451602`, 15.372549813752972`}, {1, -1}], 
         LineBox[{{14.999999999996362`, 15.000000000001576`}, {
          14.999999999996362`, 4.99999999999934}}], 
         PolygonBox[{{15., 10.6}, {14.6, 9.4}, {15.4, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{11., 12.950000000000001`}], 
          PointBox[{11., 7.1000000000000005`}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/1d1f212d2eef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/1d1f212d2eef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{11.000000000003638`, 7.099999999998545}, {
          5.500000000001819, 10.}}], 
         PolygonBox[{{7.719258539708822, 8.82984549724444}, {
          8.967305125120804, 8.62398214294968}, {8.594177795461553, 
          7.9163268625614425`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.1318657987852, 7.783717894247791}, \
{1, 1}], LineBox[CompressedData["
1:eJwV1AlMk2cYB/AX1iEig2Kh3LPIUUCuVRBkiJ+IrMNkK4pIEAke3A5qIFoU
tOmcVjmGytimaEoAOWxYQWWD4azAMGjFckgIyCwIyqRKnQ7ZxLH/+yVfml+e
93ief77UZW/OthRjQsguvPSX6Jbw+DIkgMKaIYnX2jTNsCFctuwRhyF6id/u
PbD6B1P2MCxvDZy3gqUJ2blz8PW4hr4eH4awlTOTfOyPlE0/kcHSGEFWIbxw
O7xfCIslb0v1cH1zq9IB1oY6HhLbMETCb3v+zxqGiMZ2+H/AZUiZIS36GawI
aCGX4UCP4/cmaT2lrnO9LfoYTWiahbVdi8wDuDvJ+Rih50XsrNxpx5B4swNu
q2D128/qNLBW3vrhFjhgfG69lz1DQjin3XLoet79wxmwcMws+gKt1xiximCV
54lPumBexZcuJXDs/OXrT2FG5tghhiMjV8wbYf6yHP3uENgQFyVfCStMOCtm
cJ88R1LKhUWfCs2/gTtOKR+aw2LiWmcDW78wn35Dz5vx6VWgf+apkPWA3qdf
2egL62Wa4EpYxMxkdGH+yE3vBUl0ffcVVTpcHBq0yZbmGWwa6wpLdNc87tJ8
bhaaLiBP8+2e7/Jh1WSey3O4WKhd8oLZ3Oj97+D4xr25495Yn7bVsAb7RwId
U7+H1X3sDUdgz2ehZQm0nsT4P4Hlf/n7esG8n2oi9qI/1d34cRasHUy48wqe
il/89qUX8ksZ9SvCfIq5qkNPYUVF134+8qkX5RnrYe1Har9bMEszMPselt7Z
V7rVAX2lNGx2oPcrdat6YJJWUxIO66bPiL0cGZI+OzaRRvtxvig8COdN+Fmf
h1U9wqkLsCEr5HEH3f9fsH0tXPz3j+xJ6k4mtQwuqI64vgQzVu36JFjTlBnG
QR7S76q3W8OsWfcaB9igXVtxA/ePGNrOWNP60TS1EB7yiIoitH5k9aUh9H/A
+7ae5qeOXpeQDHc09G9rghXD3vzXmF81eJbkwskOLwfPw+o/hpz8YHFtWZ6Q
+oqJaBLz8zivLLj0exH4K87RPLa0xBH4FysjYQTNr5l7dDksXHAqeuOJ80rT
lwvgcu7YaSWs85ROS2Dzw6+4X8FM7f34MZp/SdDZEJgIHIJ3oL/0iuNdbFgU
bmE7AWuD9mnm+ajHHbsowXyKfBPLWVhnc6mX5iEvmfJ9AfOCZGlNcM29TstF
WOFikhTuxBCfgvKXNjhPvNBt9xvMfi7LofcF2Ac95jszJKx5sDMZVlc6rj4M
B3Ys9hbBhhtRbfUwa2liUyvtdyHa4ld4/2sy9whWZRkJmuDEZYG9izCvrXrs
azj2xU03Dv1+Ms/FbIRHWg4OO8MkailsGvcXVFnEONJ64mDmUbh4NKLdDDZk
eBSawtLEkPlZnKf9OPvsecwjnbNsuAUrlrnrXOn3JQvyOwVLTxjtUiOPqd+N
T0bC7OA/v8iB5Ub1of/SvNgXLdbBnjt7N1yFxbnljXawpKFKnUDzGjBut4cN
fT2m5nAyXxkcCrN+Htnc6YHzeY15+bCSk5cohUUDV10ewvq1qgEhHLCHyf4c
/ZhqNl5whhXhsrB+WE//n92xPxWPE/M/mCEAHQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.884307456565622`, 5.282433250601188}, {1, 1}], 
         LineBox[{{10.999999999996362`, 12.950000000000793`}, {
          10.999999999996362`, 7.100000000000123}}], 
         PolygonBox[{{11., 9.425}, {10.6, 10.625}, {11.4, 10.625}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.9452, 10.025}, {-1, 0}], 
         LineBox[{{11., 12.950000000000728`}, {5.500000000005457, 10.}}], 
         PolygonBox[{{8.778744997805118, 11.758599589731837`}, {
          7.5321886090403245`, 11.543897075471579`}, {7.910321395349439, 
          10.838903745064753`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.12216772652609, 12.236124577663222`}, \
{1, -1}], LineBox[CompressedData["
1:eJw91H0w23ccB/Cs3ISqxDRluBb1EEaX6xBPrbgefVIsVDWiy83DWM7JlCsL
Gl27KqW2sXXqccRT6VHP0m22WhYuKzZbccxT5yiptJRwHvb+7o/lLve+V36/
7ye/zzefbyw/TOBG76FQKDy8SVLUu3g5cf7PIF4/fxrJUvVoWpGdUp9pkhKH
+Fci5P71UrUYWT7DZbCRSd4VRW5IExFtWxdpfWc3Q+WI9V2igkWk/2N6XQmS
ujO3OY5k3i6P9kdOcdaT/kYKxB6M7Xc4lFjNjdhleLguavsh3JTuFG2Aeko6
1+wTeErq6OYOC6iSTXeY4zqqFw/zZeIvaHCQlfpWNaw/3m674oDvkX1nNAOP
GM4e/wcebJ9ZMDsCS38tJFaHOccEwRqtVUtyf4qNu2Ma3Jlsdlgf9ViJ7LBi
cv+xYI8jMFNtWvMAjm2azQ6FRT29hc2wdercdCZ5vu2N5CpY5JY30wCHnR64
cpNcZ7Q4D8MpB2k14XBIoEXLKixZ/mHcinjd8pu96JtzJkc6ied91LheegDu
UaQdLIRTKo2d9sPluxN9fnDvhfxLWvDg4+pLGqR+JNf4Galntq+wEdZO6NDq
gGtntFbiYH4L6046cYzoHgue/7lk8L/9+6vbSpvUpfHjX6B/kdBIOIfPldzy
pFK4PCc8ewTO1zvgcg6WfGUuGyN93KPTduyxTwP8i0uwBb3PtxXOL8pYJH2I
Ilm/J8Lz0geubHhLS+HuBd99Tx4hhIudu4VvwRSb11QprNFnVK8xcb+tQcEU
PL6vKGoBjs1Y4ZiiX+YtGu85zNqKEwbAYbKEJA3cmXhdlkp+/3OyRUPUG3mk
Cv4W5qxW8Jxh5o/5G/VkfmgNk3xY0Jt6tBG+/tDUNwvWmPrxyuCtulCnFpgT
rnM/k8y/XpbPGLkuPuRxnsxP1dAHm7AF02fBDF693SSnYz/uyiP9R0g/2nae
ZrBJotHnubCjqsLOBB7J0TP3Jv17LzW/SSyIl6uxXz1ju/I51GM9OeNVDSed
pZl3wSlDDpRIeLW9y/oqrPj48rsOsEmwescDzrL1vbCDOrWayvPL6F9hZPP+
NFwQMORZBlNfRdgPwwrV4hsBZL+UYvpTmDO5prNrh3WB98MW4N5f5HltsIJn
eVWXzOnF19lJ8Hwk1ccF1jZdEx+HqcVD3XFwyJfXTjPgkWZ2fiXcG+Vlv2mL
eW9SmE7Cw9de5qrgU46JxSZkblPjLJdhxc3gy+Scz+817NuGLZRU/ytkv/RO
tBmj3tSfbYFfkznu533qAbNaQ87WwGERLkcjYbrI7kktrGZzq/JgN4faY0Ww
uSg2sQuWTMxOiOEQlc7JKbL+pPK3AJiS/pMWBf0LZlrWyblxXApdYsCUnP61
QXIODEuUh8j8pOmu3ID9B/YYm8NN32cms8nvU8/Y1CWWxLHJfj3zGxUuoL5A
lfuCnIvejWAHGazWXXmbT/zRqWYJPOgTN3QYznKtt/GCszw7ijRk7j57zlaj
/6yiOr9xuGG0Y6cMVr9cejoA6zu5KgNgTaFB8h/wFvkftsG6mJgYzMm/C7Aa
LA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.578595720451602`, 15.372549813752972`}, {1, -1}], 
         LineBox[{{14.999999999996362`, 15.000000000001576`}, {
          14.999999999996362`, 4.99999999999934}}], 
         PolygonBox[{{15., 10.6}, {14.6, 9.4}, {15.4, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{11., 12.950000000000001`}], 
          PointBox[{11., 7.1000000000000005`}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/1d1f212d2eef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/1d1f212d2eef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{11.000000000003638`, 7.099999999998545}, {
          5.500000000001819, 10.}}], 
         PolygonBox[{{8.780741460291178, 8.270154502755561}, {
          7.905822204538448, 9.183673137438559}, {7.532694874879195, 
          8.476017857050321}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.1318657987852, 7.783717894247791}, \
{1, 1}], LineBox[CompressedData["
1:eJwV1AlMk2cYB/AX1iEig2Kh3LPIUUCuVRBkiJ+IrMNkK4pIEAke3A5qIFoU
tOmcVjmGytimaEoAOWxYQWWD4azAMGjFckgIyCwIyqRKnQ7ZxLH/+yVfml+e
93ief77UZW/OthRjQsguvPSX6Jbw+DIkgMKaIYnX2jTNsCFctuwRhyF6id/u
PbD6B1P2MCxvDZy3gqUJ2blz8PW4hr4eH4awlTOTfOyPlE0/kcHSGEFWIbxw
O7xfCIslb0v1cH1zq9IB1oY6HhLbMETCb3v+zxqGiMZ2+H/AZUiZIS36GawI
aCGX4UCP4/cmaT2lrnO9LfoYTWiahbVdi8wDuDvJ+Rih50XsrNxpx5B4swNu
q2D128/qNLBW3vrhFjhgfG69lz1DQjin3XLoet79wxmwcMws+gKt1xiximCV
54lPumBexZcuJXDs/OXrT2FG5tghhiMjV8wbYf6yHP3uENgQFyVfCStMOCtm
cJ88R1LKhUWfCs2/gTtOKR+aw2LiWmcDW78wn35Dz5vx6VWgf+apkPWA3qdf
2egL62Wa4EpYxMxkdGH+yE3vBUl0ffcVVTpcHBq0yZbmGWwa6wpLdNc87tJ8
bhaaLiBP8+2e7/Jh1WSey3O4WKhd8oLZ3Oj97+D4xr25495Yn7bVsAb7RwId
U7+H1X3sDUdgz2ehZQm0nsT4P4Hlf/n7esG8n2oi9qI/1d34cRasHUy48wqe
il/89qUX8ksZ9SvCfIq5qkNPYUVF134+8qkX5RnrYe1Har9bMEszMPselt7Z
V7rVAX2lNGx2oPcrdat6YJJWUxIO66bPiL0cGZI+OzaRRvtxvig8COdN+Fmf
h1U9wqkLsCEr5HEH3f9fsH0tXPz3j+xJ6k4mtQwuqI64vgQzVu36JFjTlBnG
QR7S76q3W8OsWfcaB9igXVtxA/ePGNrOWNP60TS1EB7yiIoitH5k9aUh9H/A
+7ae5qeOXpeQDHc09G9rghXD3vzXmF81eJbkwskOLwfPw+o/hpz8YHFtWZ6Q
+oqJaBLz8zivLLj0exH4K87RPLa0xBH4FysjYQTNr5l7dDksXHAqeuOJ80rT
lwvgcu7YaSWs85ROS2Dzw6+4X8FM7f34MZp/SdDZEJgIHIJ3oL/0iuNdbFgU
bmE7AWuD9mnm+ajHHbsowXyKfBPLWVhnc6mX5iEvmfJ9AfOCZGlNcM29TstF
WOFikhTuxBCfgvKXNjhPvNBt9xvMfi7LofcF2Ac95jszJKx5sDMZVlc6rj4M
B3Ys9hbBhhtRbfUwa2liUyvtdyHa4ld4/2sy9whWZRkJmuDEZYG9izCvrXrs
azj2xU03Dv1+Ms/FbIRHWg4OO8MkailsGvcXVFnEONJ64mDmUbh4NKLdDDZk
eBSawtLEkPlZnKf9OPvsecwjnbNsuAUrlrnrXOn3JQvyOwVLTxjtUiOPqd+N
T0bC7OA/v8iB5Ub1of/SvNgXLdbBnjt7N1yFxbnljXawpKFKnUDzGjBut4cN
fT2m5nAyXxkcCrN+Htnc6YHzeY15+bCSk5cohUUDV10ewvq1qgEhHLCHyf4c
/ZhqNl5whhXhsrB+WE//n92xPxWPE/M/mCEAHQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.884307456565622`, 5.282433250601188}, {1, 1}], 
         LineBox[{{10.999999999996362`, 12.950000000000793`}, {
          10.999999999996362`, 7.100000000000123}}], 
         PolygonBox[{{11., 10.625}, {10.6, 9.425}, {11.4, 9.425}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.9452, 10.025}, {-1, 0}], 
         LineBox[{{11., 12.950000000000728`}, {5.500000000005457, 10.}}], 
         PolygonBox[{{7.721255002194882, 11.191400410268166`}, {
          8.589678604650562, 12.11109625493525}, {8.967811390959675, 
          11.406102924528424`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.12216772652609, 12.236124577663222`}, \
{1, -1}], LineBox[CompressedData["
1:eJw91H0w23ccB/Cs3ISqxDRluBb1EEaX6xBPrbgefVIsVDWiy83DWM7JlCsL
Gl27KqW2sXXqccRT6VHP0m22WhYuKzZbccxT5yiptJRwHvb+7o/lLve+V36/
7ye/zzefbyw/TOBG76FQKDy8SVLUu3g5cf7PIF4/fxrJUvVoWpGdUp9pkhKH
+Fci5P71UrUYWT7DZbCRSd4VRW5IExFtWxdpfWc3Q+WI9V2igkWk/2N6XQmS
ujO3OY5k3i6P9kdOcdaT/kYKxB6M7Xc4lFjNjdhleLguavsh3JTuFG2Aeko6
1+wTeErq6OYOC6iSTXeY4zqqFw/zZeIvaHCQlfpWNaw/3m674oDvkX1nNAOP
GM4e/wcebJ9ZMDsCS38tJFaHOccEwRqtVUtyf4qNu2Ma3Jlsdlgf9ViJ7LBi
cv+xYI8jMFNtWvMAjm2azQ6FRT29hc2wdercdCZ5vu2N5CpY5JY30wCHnR64
cpNcZ7Q4D8MpB2k14XBIoEXLKixZ/mHcinjd8pu96JtzJkc6ied91LheegDu
UaQdLIRTKo2d9sPluxN9fnDvhfxLWvDg4+pLGqR+JNf4Galntq+wEdZO6NDq
gGtntFbiYH4L6046cYzoHgue/7lk8L/9+6vbSpvUpfHjX6B/kdBIOIfPldzy
pFK4PCc8ewTO1zvgcg6WfGUuGyN93KPTduyxTwP8i0uwBb3PtxXOL8pYJH2I
Ilm/J8Lz0geubHhLS+HuBd99Tx4hhIudu4VvwRSb11QprNFnVK8xcb+tQcEU
PL6vKGoBjs1Y4ZiiX+YtGu85zNqKEwbAYbKEJA3cmXhdlkp+/3OyRUPUG3mk
Cv4W5qxW8Jxh5o/5G/VkfmgNk3xY0Jt6tBG+/tDUNwvWmPrxyuCtulCnFpgT
rnM/k8y/XpbPGLkuPuRxnsxP1dAHm7AF02fBDF693SSnYz/uyiP9R0g/2nae
ZrBJotHnubCjqsLOBB7J0TP3Jv17LzW/SSyIl6uxXz1ju/I51GM9OeNVDSed
pZl3wSlDDpRIeLW9y/oqrPj48rsOsEmwescDzrL1vbCDOrWayvPL6F9hZPP+
NFwQMORZBlNfRdgPwwrV4hsBZL+UYvpTmDO5prNrh3WB98MW4N5f5HltsIJn
eVWXzOnF19lJ8Hwk1ccF1jZdEx+HqcVD3XFwyJfXTjPgkWZ2fiXcG+Vlv2mL
eW9SmE7Cw9de5qrgU46JxSZkblPjLJdhxc3gy+Scz+817NuGLZRU/ytkv/RO
tBmj3tSfbYFfkznu533qAbNaQ87WwGERLkcjYbrI7kktrGZzq/JgN4faY0Ww
uSg2sQuWTMxOiOEQlc7JKbL+pPK3AJiS/pMWBf0LZlrWyblxXApdYsCUnP61
QXIODEuUh8j8pOmu3ID9B/YYm8NN32cms8nvU8/Y1CWWxLHJfj3zGxUuoL5A
lfuCnIvejWAHGazWXXmbT/zRqWYJPOgTN3QYznKtt/GCszw7ijRk7j57zlaj
/6yiOr9xuGG0Y6cMVr9cejoA6zu5KgNgTaFB8h/wFvkftsG6mJgYzMm/C7Aa
LA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.578595720451602`, 15.372549813752972`}, {1, -1}], 
         LineBox[{{14.999999999996362`, 15.000000000001576`}, {
          14.999999999996362`, 4.99999999999934}}], 
         PolygonBox[{{15., 10.6}, {14.6, 9.4}, {15.4, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{11., 12.950000000000001`}], 
          PointBox[{11., 7.1000000000000005`}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["3",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/1d1f212d2eef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/1d1f212d2eef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{11.000000000003638`, 7.099999999998545}, {
          5.500000000001819, 10.}}], 
         PolygonBox[{{7.719258539708822, 8.82984549724444}, {
          8.967305125120804, 8.62398214294968}, {8.594177795461553, 
          7.9163268625614425`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.1318657987852, 7.783717894247791}, \
{1, 1}], LineBox[CompressedData["
1:eJwV1AlMk2cYB/AX1iEig2Kh3LPIUUCuVRBkiJ+IrMNkK4pIEAke3A5qIFoU
tOmcVjmGytimaEoAOWxYQWWD4azAMGjFckgIyCwIyqRKnQ7ZxLH/+yVfml+e
93ief77UZW/OthRjQsguvPSX6Jbw+DIkgMKaIYnX2jTNsCFctuwRhyF6id/u
PbD6B1P2MCxvDZy3gqUJ2blz8PW4hr4eH4awlTOTfOyPlE0/kcHSGEFWIbxw
O7xfCIslb0v1cH1zq9IB1oY6HhLbMETCb3v+zxqGiMZ2+H/AZUiZIS36GawI
aCGX4UCP4/cmaT2lrnO9LfoYTWiahbVdi8wDuDvJ+Rih50XsrNxpx5B4swNu
q2D128/qNLBW3vrhFjhgfG69lz1DQjin3XLoet79wxmwcMws+gKt1xiximCV
54lPumBexZcuJXDs/OXrT2FG5tghhiMjV8wbYf6yHP3uENgQFyVfCStMOCtm
cJ88R1LKhUWfCs2/gTtOKR+aw2LiWmcDW78wn35Dz5vx6VWgf+apkPWA3qdf
2egL62Wa4EpYxMxkdGH+yE3vBUl0ffcVVTpcHBq0yZbmGWwa6wpLdNc87tJ8
bhaaLiBP8+2e7/Jh1WSey3O4WKhd8oLZ3Oj97+D4xr25495Yn7bVsAb7RwId
U7+H1X3sDUdgz2ehZQm0nsT4P4Hlf/n7esG8n2oi9qI/1d34cRasHUy48wqe
il/89qUX8ksZ9SvCfIq5qkNPYUVF134+8qkX5RnrYe1Har9bMEszMPselt7Z
V7rVAX2lNGx2oPcrdat6YJJWUxIO66bPiL0cGZI+OzaRRvtxvig8COdN+Fmf
h1U9wqkLsCEr5HEH3f9fsH0tXPz3j+xJ6k4mtQwuqI64vgQzVu36JFjTlBnG
QR7S76q3W8OsWfcaB9igXVtxA/ePGNrOWNP60TS1EB7yiIoitH5k9aUh9H/A
+7ae5qeOXpeQDHc09G9rghXD3vzXmF81eJbkwskOLwfPw+o/hpz8YHFtWZ6Q
+oqJaBLz8zivLLj0exH4K87RPLa0xBH4FysjYQTNr5l7dDksXHAqeuOJ80rT
lwvgcu7YaSWs85ROS2Dzw6+4X8FM7f34MZp/SdDZEJgIHIJ3oL/0iuNdbFgU
bmE7AWuD9mnm+ajHHbsowXyKfBPLWVhnc6mX5iEvmfJ9AfOCZGlNcM29TstF
WOFikhTuxBCfgvKXNjhPvNBt9xvMfi7LofcF2Ac95jszJKx5sDMZVlc6rj4M
B3Ys9hbBhhtRbfUwa2liUyvtdyHa4ld4/2sy9whWZRkJmuDEZYG9izCvrXrs
azj2xU03Dv1+Ms/FbIRHWg4OO8MkailsGvcXVFnEONJ64mDmUbh4NKLdDDZk
eBSawtLEkPlZnKf9OPvsecwjnbNsuAUrlrnrXOn3JQvyOwVLTxjtUiOPqd+N
T0bC7OA/v8iB5Ub1of/SvNgXLdbBnjt7N1yFxbnljXawpKFKnUDzGjBut4cN
fT2m5nAyXxkcCrN+Htnc6YHzeY15+bCSk5cohUUDV10ewvq1qgEhHLCHyf4c
/ZhqNl5whhXhsrB+WE//n92xPxWPE/M/mCEAHQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.884307456565622`, 5.282433250601188}, {1, 1}], 
         LineBox[{{10.999999999996362`, 12.950000000000793`}, {
          10.999999999996362`, 7.100000000000123}}], 
         PolygonBox[{{11., 9.425}, {10.6, 10.625}, {11.4, 10.625}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.9452, 10.025}, {-1, 0}], 
         LineBox[{{11., 12.950000000000728`}, {5.500000000005457, 10.}}], 
         PolygonBox[{{8.778744997805118, 11.758599589731837`}, {
          7.5321886090403245`, 11.543897075471579`}, {7.910321395349439, 
          10.838903745064753`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.12216772652609, 12.236124577663222`}, \
{1, -1}], LineBox[CompressedData["
1:eJw91H0w23ccB/Cs3ISqxDRluBb1EEaX6xBPrbgefVIsVDWiy83DWM7JlCsL
Gl27KqW2sXXqccRT6VHP0m22WhYuKzZbccxT5yiptJRwHvb+7o/lLve+V36/
7ye/zzefbyw/TOBG76FQKDy8SVLUu3g5cf7PIF4/fxrJUvVoWpGdUp9pkhKH
+Fci5P71UrUYWT7DZbCRSd4VRW5IExFtWxdpfWc3Q+WI9V2igkWk/2N6XQmS
ujO3OY5k3i6P9kdOcdaT/kYKxB6M7Xc4lFjNjdhleLguavsh3JTuFG2Aeko6
1+wTeErq6OYOC6iSTXeY4zqqFw/zZeIvaHCQlfpWNaw/3m674oDvkX1nNAOP
GM4e/wcebJ9ZMDsCS38tJFaHOccEwRqtVUtyf4qNu2Ma3Jlsdlgf9ViJ7LBi
cv+xYI8jMFNtWvMAjm2azQ6FRT29hc2wdercdCZ5vu2N5CpY5JY30wCHnR64
cpNcZ7Q4D8MpB2k14XBIoEXLKixZ/mHcinjd8pu96JtzJkc6ied91LheegDu
UaQdLIRTKo2d9sPluxN9fnDvhfxLWvDg4+pLGqR+JNf4Galntq+wEdZO6NDq
gGtntFbiYH4L6046cYzoHgue/7lk8L/9+6vbSpvUpfHjX6B/kdBIOIfPldzy
pFK4PCc8ewTO1zvgcg6WfGUuGyN93KPTduyxTwP8i0uwBb3PtxXOL8pYJH2I
Ilm/J8Lz0geubHhLS+HuBd99Tx4hhIudu4VvwRSb11QprNFnVK8xcb+tQcEU
PL6vKGoBjs1Y4ZiiX+YtGu85zNqKEwbAYbKEJA3cmXhdlkp+/3OyRUPUG3mk
Cv4W5qxW8Jxh5o/5G/VkfmgNk3xY0Jt6tBG+/tDUNwvWmPrxyuCtulCnFpgT
rnM/k8y/XpbPGLkuPuRxnsxP1dAHm7AF02fBDF693SSnYz/uyiP9R0g/2nae
ZrBJotHnubCjqsLOBB7J0TP3Jv17LzW/SSyIl6uxXz1ju/I51GM9OeNVDSed
pZl3wSlDDpRIeLW9y/oqrPj48rsOsEmwescDzrL1vbCDOrWayvPL6F9hZPP+
NFwQMORZBlNfRdgPwwrV4hsBZL+UYvpTmDO5prNrh3WB98MW4N5f5HltsIJn
eVWXzOnF19lJ8Hwk1ccF1jZdEx+HqcVD3XFwyJfXTjPgkWZ2fiXcG+Vlv2mL
eW9SmE7Cw9de5qrgU46JxSZkblPjLJdhxc3gy+Scz+817NuGLZRU/ytkv/RO
tBmj3tSfbYFfkznu533qAbNaQ87WwGERLkcjYbrI7kktrGZzq/JgN4faY0Ww
uSg2sQuWTMxOiOEQlc7JKbL+pPK3AJiS/pMWBf0LZlrWyblxXApdYsCUnP61
QXIODEuUh8j8pOmu3ID9B/YYm8NN32cms8nvU8/Y1CWWxLHJfj3zGxUuoL5A
lfuCnIvejWAHGazWXXmbT/zRqWYJPOgTN3QYznKtt/GCszw7ijRk7j57zlaj
/6yiOr9xuGG0Y6cMVr9cejoA6zu5KgNgTaFB8h/wFvkftsG6mJgYzMm/C7Aa
LA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.578595720451602`, 15.372549813752972`}, {1, -1}], 
         LineBox[{{14.999999999996362`, 15.000000000001576`}, {
          14.999999999996362`, 4.99999999999934}}], 
         PolygonBox[{{15., 10.6}, {14.6, 9.4}, {15.4, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{11., 12.950000000000001`}], 
          PointBox[{11., 7.1000000000000005`}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["4",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/1d1f212d2eef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/1d1f212d2eef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{11.000000000003638`, 7.099999999998545}, {
          5.500000000001819, 10.}}], 
         PolygonBox[{{7.719258539708822, 8.82984549724444}, {
          8.967305125120804, 8.62398214294968}, {8.594177795461553, 
          7.9163268625614425`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.1318657987852, 7.783717894247791}, \
{1, 1}], LineBox[{{10.999999999990905`, 7.100000000005821}, {
          14.999999999985448`, 5.000000000010914}}], 
         PolygonBox[{{12.468761258297233`, 6.328900339393953}, {
          13.345305182106799`, 5.416940499470871}, {13.717172301298735`, 
          6.125258821741226}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.884307456565622`, 5.282433250601188}, {1, 1}], LineBox[CompressedData["
1:eJxFlQs0VVkYx09mRCpTHl1T4+Zy99mXhilhKtGXNOUSLSmi0KhGUimPzKBJ
lDtk0FIaj9Qor6hIRYwK1+B65TWIkjxSUgxlmWT2ncWes9ZZZ/3W2nufb3+P
/5/3/RHbfTIMw/iRV/plxqfIg2B4v/T5ChiNkIFLEgS7UlNahN2EL8raRKQg
8NqYUCzwUgfGdZaw7xcE8pnHk2sUuMBEJ12bdxJBquQ6b2cW4VuHtoeeQVDF
2Sgfab2UrFewH0tE4OvK0dJ8Tbj1sIteCYKj3T90KppqAONwSMVwFMHKxiLf
hScIy+8cc13OwsG2+urdtwk/jAv92ZuFtjfB0cIOwqv6xVMFLKQP5z5q+UCY
8X2kyWBoHut0uCfLA0at6LieGQaJbJep02zC7xRjLgdi6BYX7+2dIOvdAzjO
WRg8W89YqvcSrlic21mLIXFkfd2OPwm75mgJuzGIYmJO2F4l3FX6bUIvhtiD
frpqP0v375GktGJ4dXfgQY0jYbVBfZUiDHYe3F1n1xBuFXZ4xWBIam2wMOIR
FlVnTTpieJz64b7NQul+zZXJSzAMToWf65hHWLB06+0WFpZcGZowV5Xed8rz
RRQL+y3ug9ky6Xni7BEhC+71b983WUvP6ykLn8vC9vu7X5tL49Gw1FBoRnAx
LX5Qp5DwxRTVHZkIyja7dYYw5P7+u4NuRyLYW7BLqdqK8OWspjRSLyc5wRle
EmHXtPCeEAQ+dXZKB4cIV1idM4lDwJ27LO6UiSbJX2PB9QIEltnaTS5hhDX2
pTkOkHp5pmesryTM1Bc90mLhinF5TCOjBcwCz6jyfSx8Fhlmu+VrwqIJbv91
FiSBf6+OEBJOV4g8MMKC8tyWeHUnwq6XJb8aYgjo5+RbOxP2z4gRH8UgHnhV
y91OmDHtak/BoPLwmJ/ResLvBkJRJQadoF7Fej5hjWCLzGcYqqw8vX+TIeyV
3GvUh8FgYCI8qpPEB+kbjnVgCDq/La05n/C47trnJRg0Kw3aw+MI3/IOexaP
wW8QQpMDCde/MY/cj2GDHHLO9iCcX6KcKyB84ePiJjfCD3kb43pY2FaXE5p4
gPBLS50nSSzcjXFOeCHdHz3Suc6B5CNoYVNtEuFWTdafwwLKE3GqJIS3TqKx
pwjMzw4nRsmSeN3jIu7lIND+cMFotYWU1e+x5xBUqxcgx9j/WO5dMAKDlIBF
Gj3/84Fx9ZAfDfl0fUNQl/e603x6nu9VGTfXej793+N/0j7OUUE0Hv/+Bafq
tiAa70d1kbgqANH7vJxj/OkOmduZ+1pW96tM3kQ0H06ySSfEdxDNl1Osj/Hg
DUTzKbjpoDoh3T+d74CGvQ8UTyFaD97s70ZKXRGtl0fuJUWFtYjWs9Fal2U4
iNbbh5sXXjnKp/0Q/oTVcmvl035py7lR9VMZn/aTVl7wc2Ehn/ZbRHTpp4Bi
Pu3HRm6qi2Idn/ZrgqLSJqs3fNrPky6NcuVqiPa7xzfDb+23IjoPtkNHxKtJ
3mfm5WT0YRlOB6LzpFNtHqa3jKXzZhZ7VNk/kKXzmFO7NqymmqXz2m5rVKT6
JabzPHv0U9ZhZ0znffNz4xt2JK8zemAoyfi9rApTvVhl+8zBfghTPQmZyLxk
97mA6o2OSFHuj3kCqkc1o1eeWsgLqF7NKi23cRvHVM+EKwJM5pM6zOhdXs/N
NkkxpnqosFzLM1gaz7ReLlKQX7LJF1M9lcmwrzKzxVRv9UuVrAMNMNXjwuZ2
gzwupnrdcL7A0VQZUz2vuFYI3VKe1vv3GX1/lfAw9YOybM4cGxNM/UK/pKLa
mPThjJ94VHRoiRIw9RvdghVF4U8w9SO31/NF8nwB9av8L46rbjsmoH5WdadU
/lCpgPpdfHlnpb6KNvXD3nHLNaf3aFO/9KrYpOeTTnjaT5E7sdM+bfgXDNX5
Dw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.645199999999999`, 10.025}, {-1, 0}], 
         LineBox[{{11., 12.950000000000728`}, {5.500000000005457, 10.}}], 
         PolygonBox[{{8.778744997805118, 11.758599589731837`}, {
          7.5321886090403245`, 11.543897075471579`}, {7.910321395349439, 
          10.838903745064753`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.12216772652609, 12.236124577663222`}, \
{1, -1}], 
         LineBox[{{10.999999999998181`, 12.94999999999709}, {
          14.999999999992724`, 14.999999999992724`}}], 
         PolygonBox[{{13.533959950334125`, 14.24865447454624}, {
          12.648476366030033`, 13.345372225231012`}, {12.283603733301716`, 
          14.057318825676512`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.89785927408888, 14.74959653836316}, \
{1, -1}], LineBox[CompressedData["
1:eJxFlws0VNsfxycVQ5FG8sxrOObMzDlHj3+I+OneolJXbl091FXoKY8krzxK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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.6452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{11., 12.950000000000001`}], 
          PointBox[{11., 7.1000000000000005`}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["5",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/1d1f212d2eef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/1d1f212d2eef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{5.5, 10.}, {15., 15.}}], 
         PolygonBox[{{10.780950933429189`, 12.779447859699573`}, {
          9.90534763970386, 11.866584851347634`}, {9.532750493437762, 
          12.57451942925322}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.132893029109436`, 13.266823244692073`}, {1, -1}], 
         LineBox[{{9.700000000000728, 7.75}, {5.500000000001819, 10.}}], 
         PolygonBox[{{7.071112017982731, 9.158332847509252}, {
          8.317776547023437, 8.944259140502261}, {7.939999417011102, 
          8.239075164479235}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.472856810601101, 8.113506046455388}, \
{1, 1}], LineBox[CompressedData["
1:eJwt1Qs0VHkcB/ArVg824xl57CSPiHba2g1L3e3Iak1YTSUkhFRTpjhjlmjY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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.240929562151845`, 5.603976974329009}, {1, 1}], 
         LineBox[{{14.999999999998181`, 10.499999999996362`}, {
          9.700000000002547, 7.75}}], 
         PolygonBox[{{11.817423344901552`, 8.848663056316843}, {
          12.698352025976346`, 9.756388047082122}, {13.066801284220555`, 
          9.046285840284192}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.240929562151845`, 9.896023025670992}, {1, -1}], LineBox[CompressedData["
1:eJxFlAtMU1cYx1vcHMoiIiDSyaOT3lvYuRcQJi0ifjAsEJgyZlfe4SERgW2d
m2Sg8piyASJsUIWwiQzmqC4zDIyAIiBv52IQItXgGKK8EUvGq1uxO83M6Ulu
bn75zuP7zvl/f378p6GJRhwOJxJ/+v//gwXBsg4PBjix33tI3Fkwu2TyurgT
c5J68/4oForcbISZJzGrJUc1x1gY6AxfTnDFXHetdS6Lhcu/HDvw4AkCzpTC
88d0Fo4H1Y7Ofoc5LB+lJrAg0akdevwwV8VHZHizUN0mcYheewc4YG3X/iYL
Abe+YaNa9Vy4MNLPgPOOfde0+Zg5gx3cIgbCpf70VBzmJu3Wen8GLlh1ZvD9
MQfk2yi5DNiNKT5fEWNezQstb0PgpfHO4XhiHi0J5eYicPpsu8xeglk0k6c9
iMDvNbHmfBTm7GSRxU4EhdrD8m3HMdt7WPFtEQQ2Bj+brcScVLbflYfAli9K
29ytP0/imkEhqLtMzSTOYP721BGpL14fNLsyZILrk5t2B6Yi4MZ7Gp/C8zgu
JkK7agQJYzzHcQ/M6qFy7SiCiJBWp+U9mPPC7xUJGEjrWmcVoI9XPdv1bioD
fFFpsfUOzCGirW6/MYCU901iuZiFtfwxNQMtNwJS+gb05+vmntIs+Ji/HPy1
XF9vV8GBD1lwz1bK3pDq8y1beihnYSpGJXt/I+aq20idycKFZpFZ2E0n4LQH
xgydYMGua6+5OAWzXKmrTWbh0HDmVTMbzKPBwZGBLLyIjj47NeiI9SH1buCx
0JLkfD27BPOqZT39BOcTk/62eyTmqsGVzosMKEpSJgadMdvfLl8LZ6D0qrqs
2Ew/f/xkrjkDNSZC33EO5m3n/rbsR5BBm7dWrAkxZ7BQimBjdP+d9vU4Pnp3
bXcsAvfujoY2nn6/MlGLGEHBqvx5qgdm4bTb73wEOzeVxgkjMIukLlesEXgu
YTlnGeLvPU6tVF0yrJ9dKNpS2mfY30ewJxjGHcn5yWedd+3VGPL7t2POJ9HI
ieQ/MKK6xdMZ6hvbbZoknzfUPypTWKN+w/00n/n6z7s/Ge6vOac/MOtjR3K/
5xuqj1ohR3L/ZqY5H8yPCcn7fES1z/UohOT9VC6987V+QvK+hWeEyvolmrw/
JMuUq1doog93FReFHqaJfpqcGxUeDE30FejLLl58SRH9HaxPlU88pog+dQLt
VGMfRfTLq/B5sKGDIvreYjw+xNPHX+n/i7zulROPKNIf05UWxo6LFOmfbnhk
WWFJk/6S3enNf+FJk/5Dbg1vWcTRpD+nn5cVT56mSf+6TvSsP1JDk/4OSn+q
/KuZJv0vXmjP1HXTxB+kximFxj008Y9JjVfuSBNN/GX7uaaHVpU08R9J742g
dWk08SeG+fJ+jQ9N/GvQqXm42ogm/sbObhoW36SI/2kqw+wsPqGIP/7hXVDn
ZEsR/6yhft7wwz0B8dfJ62OqnK8ExH8P7Vv8p8cL8yt/DlnU+7MD/AcYhzfj

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 12.75}, {-1, 0}], 
         LineBox[{{15., 10.500000000001746`}, {15., 4.9999999999991696`}}], 
         PolygonBox[{{15., 8.35}, {14.6, 7.15}, {15.4, 7.15}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 7.75}, {-1, 0}], 
         {PointSize[0.04], PointBox[{15., 10.5}], 
          PointBox[{9.700000000000001, 7.75}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["6",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/de1d1f212e2f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/de1d1f212e2f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{5.5, 10.}, {15., 15.}}], 
         PolygonBox[{{10.780950933429189`, 12.779447859699573`}, {
          9.90534763970386, 11.866584851347634`}, {9.532750493437762, 
          12.57451942925322}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.132893029109436`, 13.266823244692073`}, {1, -1}], 
         LineBox[{{9.700000000000728, 7.75}, {5.500000000001819, 10.}}], 
         PolygonBox[{{7.071112017982731, 9.158332847509252}, {
          8.317776547023437, 8.944259140502261}, {7.939999417011102, 
          8.239075164479235}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.472856810601101, 8.113506046455388}, \
{1, 1}], LineBox[{{9.700000000000728, 7.749999999996362}, {
          15.000000000005457`, 4.999999999996362}}], 
         PolygonBox[{{11.817423344901552`, 6.651336943683157}, {
          12.698352025976346`, 5.7436119529178775`}, {13.066801284220555`, 
          6.453714159715808}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.240929562151845`, 5.603976974329009}, {1, 1}], LineBox[CompressedData["
1:eJwd1QtUjGkYB/AvBuO2Nd3ZypTSRTGd4zLRZeymoigZu1ExEcI4Iqq1lYlp
jVu3TWaJpkwabTcZGq1qlJJzitF1JLYtlyQ1IqbU2v+73zlzvvM77+15nnnO
+1luPxC4cxJFUbH4kTfF34VnPof6/5nHoeKi7xknwbxt5nMyYEnx8NdSWDYo
FM9mcijXyekdfbCkZ+iGN9w61t3kYM2hBBdNeo/A0hLXhMMwfV5nYCa8JDTc
qhZWr/zalQ/75GjLDGw4FOuNnWUBrFQH6u6A2UmDJy/DWQzZgmJY1LTJ7QTM
ZG1hDsMCepxlKMzuD090XMChxFdelzjBESljfsEwK6m8cgjxOlpKLeJhPf1O
TTwsqz7nmgZHrtrzYhYsPaCbcwGmisbDrlpwKNr+kbFUODUq7KM3zK4WShPI
/F+SN1CwfENZ2naYGWe6/Ik5h7IOiF7mQcZLpppVwXETZ3eZkPP+YNMfwuxE
P9Yg4uWZyGOH4U/VMfJ6uNtsxWYX7BdJWUxchZVbovUvwbJHV9tEZH6SbqEB
4ms8OelaDFxap6auwCF0q42HYMWOlTpzkW/XnPc2R8n6tmPC3XD4YERFCtm/
1sNIDlvnF38pI+tFH9aOk/pRzoxeWJa4/rOrJYdqqHVLs0C8fWlLfaPgw6Ms
/g7YJyx6ymW4a37o9yWwck7ObwqY7elV9S/Jf3Hgu/tw6+yjHH9b1L3hH+Ma
WPxZ//hl2LS3S14K+zxTCV4T31/CSocb+QXHbO04VFC1WBJBzv/2yGErTPUM
FS+D+yI9botgTXvIoW+Il+7CH8mDFYli73rYrjx37y1YtrBCmkz6JfrYunIy
PlL2Lhh+yWaICmHmohY3Z1ihaC7IhEt9cwMZsGHIHHUMHDniJxtHPT/NmCYL
hHnMHuFHUt9rB6Y6wLFXXm2jYb5Ye3CWDhxR5G5rAXcn3O3oRD4y6xZfL1jP
9nxYBdytMvX8FeZGV/2cCwdpnz6vgD/5u45mwKnJhsGTSL79D7S/w6p70zLX
wxk6sbsksMY50yEL5hfV5pD9NJqu3Dew5wkxtwcWW96SOlkhH0ZdjDHi0TY8
NN8HS7TxvE2kXv0T+y7D3U2+7VkwPURXUAPT7Fex+sn8ncVaNVm/WBuzwh7n
p+Yt6YYLX5xOPQ1TcVbBZDxWK9zWAfNkZZvJep5MPmLugHr9+Z06GzbtUt4K
hoMEXKso2PXtqehkWHXp8QwPOO5OarYc5iV33Z8Kmx0bcmuCNdmqiibkQ3us
0lPDTB/dsAuWpG73zrbBfRuz6LvhoE3ytnqYbSa46QEzn38rKYIbdFe+nw/L
bi/YcxaWrA24bgTf3Zp6Zyfso7pib0z6sbhe4grTuZ3vbeCBuJnuhuT8S+OF
nvDLML/8IeRnmj5l80FS7/EU5RO4tErtWED6jzZsVAkrZh+vHYDlTlYVN0l9
3EuXL0M+XUqLheWw8tCDNSfgRpt09waYk3vhtAqW2rYbv4HVCUmiubgXtYxF
lgY4X2HROh4Ki3f7m/s6kP/vr88XYLuAdqdzpL7TTSbVw8IfmjTtpB6G171f
w7NGadk2C5GvlJv3hdyzH7NZ0bAqllY5Coev0wjrYIVj4ugAHDvz0hk9R9Qp
/5OomdzT8b18LswO37GuEA7KmVWSDAd8eVMbB7NNXv1UCSv7P/d6wVkzmnX+
JvObNmtnw372k7XDxGM9UW3ITzjYvXQUZlYa3JfAghYXPhkX+E23Jv3gGe/4
hawv/VHnkD+cJfRqrIVFOtLpbDiko4YngfsO0PIWwWfHXU7FwkFRA6uXkP3S
+Fx/sn6hxG4N6R+ZXq8trGirU/JJf+qXuU+BOW8vKkm/C3VqnPuRP6fexKGT
+Lteq6cwu4W5Zx7izxBwR1pgjc7+I3tJPqfMNM+Ig8LHykn9peyiD3DE+f7j
NHyX7j65wTAi8Xi1X/ODRbsn1q8m8aZWLT8HU2Xi9wJYHNC8l3zHWDRudx2J
1yh9YgBuXeufzXBCPaWnNtJxz8qbPdeEwZEFiWcM4cj2fQ9vwLwU/TMMePwb
HlhL3tac/wDrU8dB
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.240929562151845`, 9.896023025670992}, {1, -1}], LineBox[CompressedData["
1:eJxFlAtMU1cYx1vcHMoiIiDSyaOT3lvYuRcQJi0ifjAsEJgyZlfe4SERgW2d
m2Sg8piyASJsUIWwiQzmqC4zDIyAIiBv52IQItXgGKK8EUvGq1uxO83M6Ulu
bn75zuP7zvl/f378p6GJRhwOJxJ/+v//gwXBsg4PBjix33tI3Fkwu2TyurgT
c5J68/4oForcbISZJzGrJUc1x1gY6AxfTnDFXHetdS6Lhcu/HDvw4AkCzpTC
88d0Fo4H1Y7Ofoc5LB+lJrAg0akdevwwV8VHZHizUN0mcYheewc4YG3X/iYL
Abe+YaNa9Vy4MNLPgPOOfde0+Zg5gx3cIgbCpf70VBzmJu3Wen8GLlh1ZvD9
MQfk2yi5DNiNKT5fEWNezQstb0PgpfHO4XhiHi0J5eYicPpsu8xeglk0k6c9
iMDvNbHmfBTm7GSRxU4EhdrD8m3HMdt7WPFtEQQ2Bj+brcScVLbflYfAli9K
29ytP0/imkEhqLtMzSTOYP721BGpL14fNLsyZILrk5t2B6Yi4MZ7Gp/C8zgu
JkK7agQJYzzHcQ/M6qFy7SiCiJBWp+U9mPPC7xUJGEjrWmcVoI9XPdv1bioD
fFFpsfUOzCGirW6/MYCU901iuZiFtfwxNQMtNwJS+gb05+vmntIs+Ji/HPy1
XF9vV8GBD1lwz1bK3pDq8y1beihnYSpGJXt/I+aq20idycKFZpFZ2E0n4LQH
xgydYMGua6+5OAWzXKmrTWbh0HDmVTMbzKPBwZGBLLyIjj47NeiI9SH1buCx
0JLkfD27BPOqZT39BOcTk/62eyTmqsGVzosMKEpSJgadMdvfLl8LZ6D0qrqs
2Ew/f/xkrjkDNSZC33EO5m3n/rbsR5BBm7dWrAkxZ7BQimBjdP+d9vU4Pnp3
bXcsAvfujoY2nn6/MlGLGEHBqvx5qgdm4bTb73wEOzeVxgkjMIukLlesEXgu
YTlnGeLvPU6tVF0yrJ9dKNpS2mfY30ewJxjGHcn5yWedd+3VGPL7t2POJ9HI
ieQ/MKK6xdMZ6hvbbZoknzfUPypTWKN+w/00n/n6z7s/Ge6vOac/MOtjR3K/
5xuqj1ohR3L/ZqY5H8yPCcn7fES1z/UohOT9VC6987V+QvK+hWeEyvolmrw/
JMuUq1doog93FReFHqaJfpqcGxUeDE30FejLLl58SRH9HaxPlU88pog+dQLt
VGMfRfTLq/B5sKGDIvreYjw+xNPHX+n/i7zulROPKNIf05UWxo6LFOmfbnhk
WWFJk/6S3enNf+FJk/5Dbg1vWcTRpD+nn5cVT56mSf+6TvSsP1JDk/4OSn+q
/KuZJv0vXmjP1HXTxB+kximFxj008Y9JjVfuSBNN/GX7uaaHVpU08R9J742g
dWk08SeG+fJ+jQ9N/GvQqXm42ogm/sbObhoW36SI/2kqw+wsPqGIP/7hXVDn
ZEsR/6yhft7wwz0B8dfJ62OqnK8ExH8P7Vv8p8cL8yt/DlnU+7MD/AcYhzfj

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 12.75}, {-1, 0}], 
         LineBox[CompressedData["
1:eJw9lQlQE2cUx6NYBhFtuaMhutkENdrdRUBIkcJT1JZDwUxV5FA8QIRiqgFU
pBACApJaOZSCSkEr5RAYagkDiiU6JKJiIOBRECxSLaIjoq2KYKXfVubbmczO
b/bL+975f4JtMmn4VA6HI0M/9v3hoWHTuwn0WAKnI66+upWCKGezPCrFCvG6
X/cUU/DdT0uV622sgRNQxdWlU8CpTAohGhF/s8t7voKC2Cf7hnZsswEON5g/
qqKg81mINN/UFp0vvXKogoLoXi9vSSVibkGlyV0KVt6vWiJ14wIn09DdYk5D
xBJuaE8d4hLPN44baDj5VPvYwno2uv98rksJDQZZa1nnRsTZ7VtCn9Kwpv9Y
g/ogy7kFQicGxDcUvY4piPdnRX8mZ6Av2dn3WBTiMEsfbRkDoq+7995ahphY
zCu+yUBsnKU2bIy9/0mLYoAB3XZzT10VYkUdKX7EgEDsXesXhBgerJXdZuCg
nfpGuil7Pst2VR0De2Y8vtDXguLpaNpfnsrAm8JZzLxsxBJmPG01A+MZ7Vqr
vYgJsT6dw4DTx77TF8vY+PPWDqhpCNFmbaeyEGsct1hG09BcUWPHY+0RM2IL
RDTQ14zHbe3QfQuXxlX/hfJ1Z0B+4CjiSMcSUzUFgoqis89no3hGMtady6XA
7WJh8VU1y6PzS1IoeP0owsIrZA7yf7n7cyUFBWN6DseYh+w5R6flU+BnZBUY
dQ5xWE2K40UKlD2HDzmsskPcEKQdpsA1nDuzrg0xx6tFSNNQ0G8WJHPno/gu
L0iOo+Hn96JzPtmIH1Pc7zU01FLlEz1axNw00JsysOlWl0dGL+LMyKZr/ihf
zYvCcm6z5+Nrg1UMJJ53NTWqQxwYazTSyMCpSN6tpETEmtYdF35nIPXA7rFR
Z8QmqdIHqB6RbYkdwQ+RPws1h0/0M/DDrqAYfi7LCZ+6XmNAZRL4uo31P2DO
wNzTqL47KyN/NEbcX/X5hSgGrD6yvjfeg+IdOaFeI2bg2ReXeslWxIrR8Kv9
qN+8h3u7DIi5qtPK4zSkZwXLuWOIHZJeTvOjoTFDO0W0HNnTdGc3mNDwbZki
L6AU8Se8xPJ2Cl5Fb6UkQuRv2Fa/FWcpaMyeMB1Ss/xwn+AwBRNmoWVLNs5F
/cmXDCZTkDjMzttc/H20/Uygvnwe/v+8+YqEGfYEtp8i8R/avpvA96+q8azf
WE1g/xZ48Cdi/iCw/xp3qe7vqQIcn0tfprmNrQDHP3Pz6ZxhngDnJyE5zkJp
IcD589FX3qh9R+D8/vZvV9rsPgLnX3JT6v2insD16Th52bbzCIHrt0LbqkwN
J3B969e8PK5ZTuD66xY+THjPxjfZH75nnvi7WhO4f/yn5Yc1mRO4v87eXayw
4BO4/8SymMoeVwL3p0PSW2nrVgL3b/9F+dvSQgL397Ipyy7t/N//D/0fynnd
5bNIgOejIa1gYiJRgOdn+tG2+PsGAZ6vCJGTg509iedPTv1CmchJPJ8D+bNW
ezWQeH5jihKn/vOCxPMtqi7f85IvxPO/odht7ZCbEOtDVELQwagvhVg/jD24
Ra9WC7G+mKj3R464CrH++L5TkTmsvUl90sntTzSNkVi/5DH3yro7Saxvhe77
+k9VkFj/rjjFF7UrSayPUZI8nfU2EuvnkZrwHrUPifV1MLM5J8KTxPq7Pkkt
tF9BYn3mXedV2XxFYv12MStSxceRWN8DhuSG3FIS6/+DXPHgnQES7we95eZ7
fy4S4v0xR2W4U5cgxPvlutlK1U29EO8fo2qNQSYU4f3k4fDI43asCO8v/eAB
l13NiCf3210j9m0P/wF3crCM
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.6452, 7.75}, {-1, 0}], 
         {PointSize[0.04], PointBox[{15., 10.5}], 
          PointBox[{9.700000000000001, 7.75}], PointBox[{5.5, 10.}], 
          PointBox[{15., 15.}], PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["7",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/de1d1f212e2f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/de1d1f212e2f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999994543, 10.}, {14.999999999994543`, 5.}}], 
         PolygonBox[{{9.719049066570811, 7.779447859699573}, {
          10.59465236029614, 6.866584851347634}, {10.967249506562238`, 
          7.57451942925322}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.132893029109436`, 6.733176755307927}, {1, 1}], 
         LineBox[{{9.550000000006548, 12.100000000002183`}, {
          5.500000000007276, 10.000000000003638`}}], 
         PolygonBox[{{8.057652718869358, 11.326190298673001`}, {
          6.808220415348643, 11.128911513906571`}, {7.176474146912643, 
          10.418707888747429`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.41630839509475, 11.821219523745839`}, \
{1, -1}], LineBox[CompressedData["
1:eJw91HtUjHkYB/BXqWa1mEhy1BrVaBCiJko06DKUmmXYKJptdZEuE2Iwa+uU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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.82286860747054, 14.931502099753644`}, {1, -1}], 
         LineBox[{{15., 8.999999999996362}, {9.549999999999272, 
          12.099999999998545`}}], 
         PolygonBox[{{12.7965337537301, 10.25334777310765}, {
          11.951234397531467`, 11.19434139604575}, {11.555698095008333`, 
          10.498963057738951`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.113448413861427`, 9.807504469530574}, {1, 1}], 
         LineBox[{{15., 8.999999999989992}, {15., 14.999999999982448`}}], 
         PolygonBox[{{15., 12.6}, {15.4, 11.4}, {14.6, 11.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 12.}, {-1, 0}], 
         LineBox[CompressedData["
1:eJxFlAtMU1cYxy+wIeCGslGxLW1vZTQwufc6WPeSxyeysTE3apEJQqQrw9Qx
QDrSTehYsRUx8jCIEoFGZgshjhooZUDSImODKQvR8tCVVcKiPCJMWARBjWyn
23K8yc3JLzfn3O983///F8rzpFnuBEHsRa9r/e+hweLrWv2BANktYpACxarH
U4sW8Zp2+XY9BX9a180JPBYQjdbLCh0FusGFDPGAi2vOh5dQUDXeM/ZAuQUI
RVOLspqCD3I6Ar7jBQDR1nzMrYuCsV3nO8utiIn22JwFCr6Oi6+8Hr8ViKnX
k4MYGuKii0Sj3YhbPjOYCmkw1wYe6fdhA3F1+8r6NRokZ9a/cb6DeC0351oA
A6wAW4PhfcR99fsdBxm4u3uvaVqMWLYekVvHQEPuLbHeE3FIr2/5IAMX25Ki
TH2u8w3tMVMMrLzL1qqzEWtWhfZpBhpbz11XbULc+KvPHzcZiE+ZHld3onoV
1YF5FgZsobr39hxCrPFYuaphYDKt9q17oYiPfN8hj2HAKI6ZZW9A7JV1u36Z
hgPkmftz7ohTlkoKmmiwzqsGUkgXZxjf3E+DqupK9HaZaz95gdxEw+rE/PHx
Ptf5l6rdRijwZJuP1kWjerxSnUsGCogknXFoFLHkIcejjILXciMrg4+i+y11
5G3RUPDhzpPTySIOEGVva83lFEg3KsM1I4jnepV/t1KgX8/XTRRwgTjdCmOT
FLhNbl6OfT4Q9Wuf/Xc+DYWcV+4OqRETXo+HFTSEsD9vo24ibvG2d3fR8Ibe
2/GqLw/VX/dJsifqj0putAUhvpE6/GkiA7EdmSX0VsQSqutJBQO1uqjnnt5H
+9sylQ9sDASlGB4ZTIj7IqdedjIgPzFfszMN8dxyQuoMA+Usl/6efW9S+2f/
bOTi/csj6Z7pUi4+X9Pc/8Tdj4v/PyBpKLbPcnB9Nyz6tFgnB9dvP/RQNP4X
B99v0e9c4kgoF99fv9nJFWuf9cdTtLjR9piL+1cUVlkUfCoQ97dzrWL4W4qH
+z+8qOJccvDwfGasP+VPlPLx/Da0F6WJwwR4vn4vfVya9aMAz7/M3yfqrIDE
+rB/1Ci7l0hi/ZguxOlH80isr4zjP5SdKiax/noKZS98qSaxPln7LHd8skms
XxPR63TsIbG+FYdfPLsrmMT6/6r48I6TSwLsj23TM2ptjQD7JyuptV8aI8D+
UltDhuSP+Nh/EbOXpZwhPvanu//uzPQuPvbvL7Vux4IG+NjfyV/cqYpa4WP/
jy7oHGHxApwPp6905+/oFuD8uFgQ0dQjJHG+zB4UFTD5JM6fyAADq3SQxPmU
YPaWZPGEOL8qDJrmA/lCnG+/sYtLgnsR/59/4f/m3zb4B9w++/4=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.6452, 7.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{15., 9.}], PointBox[{9.55, 12.1}], 
          PointBox[{5.5, 10.}], PointBox[{15., 15.}], PointBox[{15., 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["8",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/df1d1e212e2f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/df1d1e212e2f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999994543, 10.}, {14.999999999994543`, 5.}}], 
         PolygonBox[{{9.719049066570811, 7.779447859699573}, {
          10.59465236029614, 6.866584851347634}, {10.967249506562238`, 
          7.57451942925322}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.132893029109436`, 6.733176755307927}, {1, 1}], 
         LineBox[{{9.550000000006548, 12.100000000002183`}, {
          5.500000000007276, 10.000000000003638`}}], 
         PolygonBox[{{8.057652718869358, 11.326190298673001`}, {
          6.808220415348643, 11.128911513906571`}, {7.176474146912643, 
          10.418707888747429`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.41630839509475, 11.821219523745839`}, \
{1, -1}], 
         LineBox[{{9.549999999991996, 12.099999999994907`}, {
          14.999999999996362`, 14.999999999996362`}}], 
         PolygonBox[{{12.804680559934305`, 13.83184837134119}, {
          11.933218354293157`, 12.915031255369275`}, {11.557420525838236`, 
          13.621272001948348`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.151691707368592`, 14.313541446496956`}, {1, -1}], LineBox[CompressedData["

1:eJwd1A9YzHccB/DfjNT6s9Jm11SuOt3pL1dxaDq0dlTIlV1Ki3V1UxQ7Kuvh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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.767354149153686`, 9.199048423512124}, {1, 1}], LineBox[CompressedData["
1:eJw9lQs0VHkcx0cPpjwiTJasmOuO4d678mqJ+tk4R0vGa7VU5FFK5f3qySpS
VpKNVqw5nOgh77WivCmEaj1SyHJUHodiEyqz/znb+d9z5sz5nJl7f/f/+/2+
36+md6DTgRUsFus4+oi//78Y8OsXoYsGlnCpwNiQgaoH+pLKjYjZnf0P3RkI
Vqn9mHwacR2vqSKMAeWLIdM9mxEfOiddFs3Agly5Q+YIBazi1FzmJAN0j9fD
5iuI3zVI+PoxQFrFjptbI/55Jd/ZmgG7ZOtfOj/rAesyu6qcw4CE4pzT6vuI
5SOLRgdpaLftMFsTh1jf4HVNFg0+5Um2Z90Q26h9PutGA/+uimuCGeKFvw7c
UKIhaHpg5qIO4ptS/VnPKMh2Ui/4iYt42NC+MY2CpiLjZ7K6iDetGlDzpSA+
Xf3Ytq3ier1ZJhYUCGxTiEkXxFHJmWsJCjJ0GiZvhYnZ+ODgNxQ46HmaNqQj
VonJNNCgwJsbZOxQjVg4bN9sSEGlWiLHfwDxk3/CE10pKGp61G69IK435t0Y
h+7/Ii27TwadH7Ij1tZS8GKiSpTEQbzQvXBSRAHBKLSZKCNm34/t3UFDhDCq
+EcpxAkeKnkJNOyOTbrjOSV+/r0nQW009Bj90VLWhLhSs310NQPD+iks3RTE
Mf7U8hbU35ujvducxf2q+LjFg4EyzgG5DzKI325/nBPBwPy9VyXVdbrA0vEf
XhHDwGuNxTT7YMSPHswPHWeg3qqUlaqJWN/mTLwvA1HkYX5INx/1QzpZ1pIB
A+6p38t+RRwTMb9NgYHORjuDfjvEdafYcX009OYa1S5zEOt7e7Wl0ZC+aFRc
M6WD/n+FcnGiIVRZ1VO9C7HwxvN0WRryH72ZOFmD+FDzHrnHFKS+WOWcWoVY
fp2tVAoFI11d8lENiPcXrjnjSUFNtScrrxvxNWvTOVMK0oPHvijOIGZzPOy0
KGi3Kza1X4fqV25yq0bzU+HZT181QKxCfCjcRIGexMzQKlfEQRvsO0woOFh3
7u/SCMTCwD0L7hQ4KZxgXqYgtonz3ptIwbdESgXrhvj+oLTzzRSk7c5skS1E
7LCu/IIkDR9bFXcE3kF8s1JC1ZaGLrWxBZ1MxNf2P6xPpkEytG7fSDTiBJ7q
mqc0HEkPk5kU17dhV+ZKM5CnXdRgoiXuX0jRdXMG1l90VnQcFfdHdDvSiwHz
0uUjOzPF/al47x6FmNsReE2A2Kb2tDqan26nWL88/DtfaHDZr5iH7+crSPxW
4cPDz/9hfcFU4UYerl+T5WDROkji91vSS5xTukXi9xfwt+eox5D4fGn5qlJ9
PiQ+f8Di5P3LLiTuT62Vj+msI4n7Z5MdsFPkTuL+KpbGPSUCSNz/dy9PULsu
kHg+Zg3ymp/E9b/Ob1dmiYJ/B4nnmygQtJS8I/H8Ox1VV0zJ8/B+5EsJRW56
PLw/oY+/mzlvwcP79eztUsFRax7ev5qlDcrnLHl4P0fjx4Qj+jy8v52tbQJL
RR7e70sa7SAxTuL9P/+8tIMsI7E+DIVnD9qFklg/g+FpvUl6JNZX6u5doqND
2lh/0bmSh0suaWN9eoxo2Ftt18b6zYnLeL92jsD6VlHbyK8vILD+w48EiHoC
COwPMX5epnZbCewfcpYhLkJFAvvLq0/xgj8XuNh/NgdMG7dOcLE/JT63YoWM
c7F/SWbPzebMcbG/5fWZGFVJEdj/7uaHjV/lEtgf/41KbiuyIrB/SgYLWywO
Edhf2zMslFwTCey/Uo4BL5XuENifX+mNXKlvIrB/5yi20bM9BPb3ip0jzrcH
COz/891a8uF9BM4Hc86bW5HNBM6P2TfB5afyCJwvGZEGzfMnCJw/vsUyd5d2
EDif2LJVZzavJHB+bYs+zWms4uJ82zKd9b3ZMS7Ov71JtkU+6lycjwLTJAXd
Ti2cn3O3gR0Uq4Xz9fpQ5+KEOeKv+bs8LtavJvwHL+bs9Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.9452, 12.}, {-1, 0}], 
         LineBox[CompressedData["
1:eJxFlAtMU1cYxy+wIeCGslGxLW1vZTQwufc6WPeSxyeysTE3apEJQqQrw9Qx
QDrSTehYsRUx8jCIEoFGZgshjhooZUDSImODKQvR8tCVVcKiPCJMWARBjWyn
23K8yc3JLzfn3O983///F8rzpFnuBEHsRa9r/e+hweLrWv2BANktYpACxarH
U4sW8Zp2+XY9BX9a180JPBYQjdbLCh0FusGFDPGAi2vOh5dQUDXeM/ZAuQUI
RVOLspqCD3I6Ar7jBQDR1nzMrYuCsV3nO8utiIn22JwFCr6Oi6+8Hr8ViKnX
k4MYGuKii0Sj3YhbPjOYCmkw1wYe6fdhA3F1+8r6NRokZ9a/cb6DeC0351oA
A6wAW4PhfcR99fsdBxm4u3uvaVqMWLYekVvHQEPuLbHeE3FIr2/5IAMX25Ki
TH2u8w3tMVMMrLzL1qqzEWtWhfZpBhpbz11XbULc+KvPHzcZiE+ZHld3onoV
1YF5FgZsobr39hxCrPFYuaphYDKt9q17oYiPfN8hj2HAKI6ZZW9A7JV1u36Z
hgPkmftz7ohTlkoKmmiwzqsGUkgXZxjf3E+DqupK9HaZaz95gdxEw+rE/PHx
Ptf5l6rdRijwZJuP1kWjerxSnUsGCogknXFoFLHkIcejjILXciMrg4+i+y11
5G3RUPDhzpPTySIOEGVva83lFEg3KsM1I4jnepV/t1KgX8/XTRRwgTjdCmOT
FLhNbl6OfT4Q9Wuf/Xc+DYWcV+4OqRETXo+HFTSEsD9vo24ibvG2d3fR8Ibe
2/GqLw/VX/dJsifqj0putAUhvpE6/GkiA7EdmSX0VsQSqutJBQO1uqjnnt5H
+9sylQ9sDASlGB4ZTIj7IqdedjIgPzFfszMN8dxyQuoMA+Usl/6efW9S+2f/
bOTi/csj6Z7pUi4+X9Pc/8Tdj4v/PyBpKLbPcnB9Nyz6tFgnB9dvP/RQNP4X
B99v0e9c4kgoF99fv9nJFWuf9cdTtLjR9piL+1cUVlkUfCoQ97dzrWL4W4qH
+z+8qOJccvDwfGasP+VPlPLx/Da0F6WJwwR4vn4vfVya9aMAz7/M3yfqrIDE
+rB/1Ci7l0hi/ZguxOlH80isr4zjP5SdKiax/noKZS98qSaxPln7LHd8skms
XxPR63TsIbG+FYdfPLsrmMT6/6r48I6TSwLsj23TM2ptjQD7JyuptV8aI8D+
UltDhuSP+Nh/EbOXpZwhPvanu//uzPQuPvbvL7Vux4IG+NjfyV/cqYpa4WP/
jy7oHGHxApwPp6905+/oFuD8uFgQ0dQjJHG+zB4UFTD5JM6fyAADq3SQxPmU
YPaWZPGEOL8qDJrmA/lCnG+/sYtLgnsR/59/4f/m3zb4B9w++/4=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.6452, 7.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{15., 9.}], PointBox[{9.55, 12.1}], 
          PointBox[{5.5, 10.}], PointBox[{15., 15.}], PointBox[{15., 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["9",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/df1d1e212e2f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/df1d1e212e2f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         LineBox[{{16.5, 9.300000000006548}, {17.299999999999272`, 
          6.500000000010914}}], 
         PolygonBox[{{16.73516732326157, 8.476914368584493}, {
          16.680223097682095`, 7.2131971802565555`}, {17.449442255794757`, 
          7.432974082574457}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.01443788115672, 7.474182251759064}, \
{1, 0}], LineBox[CompressedData["
1:eJwl0w8slGEcB/B3kvMn7mqUWDmn7DjpsisU9qp2GuHGKUocqnGI0OVf3FRC
q7jN/C1WiWmju2KIYyld+bOyrnKtzb+EJITslvR969mevfvs2X7f5/fb81pH
JgSc0SEI4gA29SUkZ7E4JKFPwZoknPYdr5bBdZV+03y4gLZd0QenaoW1ObCA
93bbJgeSSLQpUjfC77VKeTisEutpe2ETyfqncrhrMrdWDdsW2ZG6u0iCPeZT
3AfPLBiRgTAxLrknh03lQfPlsEiH158Ls6UJfz7AhZtDDPxhfpRvG82RJJj2
yZYmcKFRrzf7n+2DVUySGJzWvnCBVfEPRTlwD6OMdIWlJSFDJHxaqOyzhx+t
eDP14d8Xp6Lp8FzLYUaLFUlEX4sdnkJe9fONztlwWBKnqoNyvLEmBua2ZWbc
hBPXx8ykwTs6I9JE1P21pYxGuN1JW+AMMzl0/3Wof8XCKGALdb7U/8UdNnyd
OaEDTz7wW70M8/inan5jPm+SuYVv4QavBCM9nJtLilks9CdsLbltBaucrQ3O
U/Nzqx4+AqdmTypaYd24QZGUyuvonP8Fz4Yvqrqp+j/OubFZ6O/VfhoD/bm4
Kya84ImYr5ciYO7ALZ8gWLO1yO0J/GY26qs/rHTK99DdjbxMPZELzP0prfeF
J+Py0hmwHy2LfgNeWeKsaZAXIA6u6IRb7gbeqICfqSO9xmCynhV5DBbxM1aW
YVHLSTMGbLiaV6iFGQ01gn70a3FUODcLM0Obr8rgLE7llBqeC1msjYKb6/gs
BZU/PfX9EJyXnzKdS3n07J298OhFun4wnBoqmXGFy20ZGXawICRyXACnf3Hw
WKPex/V5cRp80FhD+wynOn7zk8NhP6X0Hthc0xu7RHn+lqYdZjvmfvLEfcXi
AtNu6n3w3Hgy2MNSljIEryiYAyPwhlVjBVW/VOho7ID5LDbZ5POQL72/sz4O
DujqGLgARw/5LlfBy49HZF1Uf3temnXD9uKKWn0uSXxUDjq8g4eb4ulOsHTB
89IgHKs2yDoBk4IDSUrW//9SCteVYbHIv7cvUN8=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.674182251759065`, 9.785562118843277}, {0, -1}], 
         LineBox[{{14.499999999996362`, 5.69999999999709}, {
          16.499999999996362`, 9.299999999997453}}], 
         PolygonBox[{{15.208614241292821`, 6.975505634327077}, {
          16.141048669155794`, 7.830237193201469}, {15.441722848258564`, 
          8.218751538144375}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.25014377798838, 7.35205345667312}, \
{-1, 1}], LineBox[CompressedData["
1:eJwd0wtIU1EYB/Cb4aI0tdSpuWSZzpqmcw2bD9y1MEW3NEuNEjXdagTlkjIF
MzVNM0GpnOGroZHVCpfZQ3vNJB9luGoreohaKUTp1oxaW1T/04HL5cfhnO//
fdy7KicvRWZHUVQkHvKmxkZHR8WhNPV/raCpD1J9VxisHcwJzPeiqYTfpote
MNsh7brVk6bGZR7ffvBoitc7vl0FN5WzA3RwKXfeQQ4zYo/6dsL1wWbNdrhs
w5eJo/BkuNywBz7DsGalwS7C5b0tcFFAahUfTrY1ysywVTRMucGmF5LZ3aj/
qqU71hKC82qvkY9wzXfBzdewTq1OykPeYHf7iCG41O9ytA3mGPc8HISp52JJ
iTf6UMesN8Bap30PjbBjNbPRBLvE3TXGsWhqZ9QTZwfUU0TODZfCbzfv460i
Pm32V8K29phHQlhjVbqegAvZSQNJJD9jrzEZPjdTHiGFtZendv/C/UP51rRD
ZJ8pFVTAf5v6a0vIvLpz7C3I9ztj67UyMi/P8xMZMD8j3lpM+nU19N1Ffymr
BaEKUr/+/uxK+NtUV90umL0gpKkW87F7ohwTwbrKUpYLzNJVevuQ+bF9q7o8
0F/YY4sR/akqm/kKWKD0LG+DNbmT8lRYOHNgIBFW+ApDs2Hxs4J+WzDyMOZF
dbDZ8VhqN5zMMS6ZgO9Zykz5MNvt0oIE1LvV0CmlYVXfjulhOJnbeocF16v7
Irch75XeikWLyfm6eMMEzFp4JM8epi70ZMrQ79wNtfMyWBt35vtbuE2fy+TC
tNhDI8S8UtJ7xyXE3CC/QlhVcMq9iDiCP3AW5lWbfnbCvDjXgFo4842x4hWc
LYnckAn3rHjZT5Hv5TOLuxR+b/a3+MK8jrU5rag36HjMFA6bEuVr3OF2J4X/
JjKfjvn4YuTtazj8JQqmlVWN79Cf/tP+UxyY3fQgNwr+Y2eWkPt1aR3TnZgP
J+F9wVPUd7ndo+bAV0UbQ2tIPonUVcukqY1in/0ikr/loL4I1gaOaObWIW/5
zovpcPRZv6lmmPIutt8Fb0n9ujARVmWd9D4OW/5iBeEc/tNRJv0Pn1pNyg==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.325817748240937`, 5.214437881156724}, {0, 1}], 
         LineBox[{{14.5, 5.699999999998909}, {13.700000000000728`, 
          8.49999999999909}}], 
         PolygonBox[{{13.935167323261576`, 7.676914368584494}, {
          14.649442255794757`, 6.6329740825744565`}, {13.880223097682098`, 
          6.413197180256555}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.985562118843276`, 7.525817748240936}, {-1, 0}], 
         {PointSize[0.04], PointBox[{14.5, 5.7}], PointBox[{16.5, 9.3}], 
          PointBox[{11., 10.}], PointBox[{17.3, 6.5}], 
          PointBox[{13.700000000000001`, 8.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["10",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         LineBox[{{16.5, 9.300000000006548}, {17.299999999999272`, 
          6.500000000010914}}], 
         PolygonBox[{{16.73516732326157, 8.476914368584493}, {
          16.680223097682095`, 7.2131971802565555`}, {17.449442255794757`, 
          7.432974082574457}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.01443788115672, 7.474182251759064}, \
{1, 0}], LineBox[{{16.500000000007276`, 9.300000000006548}, {
          13.700000000007094`, 8.500000000003638}}], 
         PolygonBox[{{14.523085631415508`, 8.735167323261573}, {
          15.567025917425543`, 9.449442255794757}, {15.786802819743446`, 
          8.680223097682099}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.674182251759065`, 9.785562118843277}, {0, -1}], LineBox[CompressedData["
1:eJwV1AtMU1cYB/ALQnlDZ3iUh1AqUaIgDUMtAV1jxqgDSqUI9TlAqpSW2blG
CI6IiFJiXMr7JQ4maMck64SNboF45Y3DUZiZFZxjk9eGQ2BsKwh2/3OTm+aX
c893vu+fe+ufdjZBaklRlBg3+aXWzWbzRS6fKq/BxeFTxouWzGJYmz98NAqe
e/jCkni9yyrilT+fUgVPF5Hn15UCTR3Mc1ubVcBzi1VqIcxScywSYKv0oWkH
OOd6NI8Ls569R//M5lOK+Pw0O1jZUZn4Naxrlwrvh/CpyNnp5zdgU+Wr7zJh
0crM1RpYY/263xNevx934jasP6xbNezCvkSm/AHM88w8VA6/fHNudBYWHSmx
OQPfsH9scsP5tjzXYzFwzjsRP0TDk0d/lB+Aj68u/ZtD1gf2BMbCAWkzYc2w
UpEnJPungj74dBDWHmBcKoFtf9HU/0rm1a4oBmA6frV3FjbViilr9Od4ef0I
Wc+xkN2NggtnHek+eG7o9VeFsMTQdI7kldI/lt4Fv9zI2kgl50fkXZuHe33d
X3jDzD7BoDXyUYXJVMOYhy/+5JQTnB4QU5sNz90aGHKAm5yoSn/i3JYeCpZ4
3Mqu90N/N/NOP0U9eud/EgZsSF/2boQbxPTWK76YT+dUewzO6avq94OZXZoW
R1itC8sc34J8K37X6TEff/FmcAfMGytKS4cFD+o3fQsby1JzXeHFKxVtj2Fq
qVL+KBj+rXfDGfW0G0zPMpjJ1kyeJOe1b6vMgKnRD9c64UD7/JV4WJ+0z9If
/XGdUzjvwwJvgbAATnmSO5EEZxRkpxrhSZawVQW/W2T43gnzGlq/cW4k542y
8jjk/ZFEFhvhQE2JeTvM2qleckd/1bv03T4wPRHzZTIsmlpJpohlmSsVsGRB
OHob9ZWDhaGPYLVVf1IIrL1ssX8VHhQ4az5DvzQ10OxO8rM3Sd5gXro9eZYD
Z0R8/oUY5s90VvnAk5IaV60Pzo9ryt0EV9v4RFjBasneq0aSV6ucr/TmU+zz
WT/Vw+x//KL+9sK8sU8vJMH5e5eXS2GWnd8ZBsyVythimDoe63AP86m9Dpbu
hrkqdsBJuDqLfh5O9vdMS11gUXGw6gSsdhvgPgxCnh6MnjpYsv/SwVJ4cn6P
egEOnFiry4TprZFvC9EP06/F4TAsaTibeo+4aSQxATb5t53ejP4FAReoNNgw
2T0uhdnbtygKYNuxUE4zrMzbMaWD1VEpfw7B+WY/wx/E8vOaETjFzuuvQPI+
mEyKNphujCqTwbwFhllF8gmhLe4EkzrlK26wSfHW3WewrtDVpRb9sPSiQpLH
nFPcDAM2dUbu8yd5OYdfP4V5REqX/B3ke/Tt9tV7Yh+nYTNZZ310J9wJbpBL
561g7fDax3IWnMkMfYL63PZtI+MeyKOoI6gW5kfLdqfBRs9er0Mw3edmZwOb
8P9oJnkN43Ln/w/N4NiY
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.25014377798838, 7.35205345667312}, \
{-1, 1}], LineBox[CompressedData["
1:eJwd0wtIU1EYB/Cb4aI0tdSpuWSZzpqmcw2bD9y1MEW3NEuNEjXdagTlkjIF
MzVNM0GpnOGroZHVCpfZQ3vNJB9luGoreohaKUTp1oxaW1T/04HL5cfhnO//
fdy7KicvRWZHUVQkHvKmxkZHR8WhNPV/raCpD1J9VxisHcwJzPeiqYTfpote
MNsh7brVk6bGZR7ffvBoitc7vl0FN5WzA3RwKXfeQQ4zYo/6dsL1wWbNdrhs
w5eJo/BkuNywBz7DsGalwS7C5b0tcFFAahUfTrY1ysywVTRMucGmF5LZ3aj/
qqU71hKC82qvkY9wzXfBzdewTq1OykPeYHf7iCG41O9ytA3mGPc8HISp52JJ
iTf6UMesN8Bap30PjbBjNbPRBLvE3TXGsWhqZ9QTZwfUU0TODZfCbzfv460i
Pm32V8K29phHQlhjVbqegAvZSQNJJD9jrzEZPjdTHiGFtZendv/C/UP51rRD
ZJ8pFVTAf5v6a0vIvLpz7C3I9ztj67UyMi/P8xMZMD8j3lpM+nU19N1Ffymr
BaEKUr/+/uxK+NtUV90umL0gpKkW87F7ohwTwbrKUpYLzNJVevuQ+bF9q7o8
0F/YY4sR/akqm/kKWKD0LG+DNbmT8lRYOHNgIBFW+ApDs2Hxs4J+WzDyMOZF
dbDZ8VhqN5zMMS6ZgO9Zykz5MNvt0oIE1LvV0CmlYVXfjulhOJnbeocF16v7
Irch75XeikWLyfm6eMMEzFp4JM8epi70ZMrQ79wNtfMyWBt35vtbuE2fy+TC
tNhDI8S8UtJ7xyXE3CC/QlhVcMq9iDiCP3AW5lWbfnbCvDjXgFo4842x4hWc
LYnckAn3rHjZT5Hv5TOLuxR+b/a3+MK8jrU5rag36HjMFA6bEuVr3OF2J4X/
JjKfjvn4YuTtazj8JQqmlVWN79Cf/tP+UxyY3fQgNwr+Y2eWkPt1aR3TnZgP
J+F9wVPUd7ndo+bAV0UbQ2tIPonUVcukqY1in/0ikr/loL4I1gaOaObWIW/5
zovpcPRZv6lmmPIutt8Fb0n9ujARVmWd9D4OW/5iBeEc/tNRJv0Pn1pNyg==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.325817748240937`, 5.214437881156724}, {0, 1}], LineBox[CompressedData["
1:eJwd0QtIU2EUB/BribXUsijd0mzSMntoNkeZpt6C4qbT5iMdunSl1uylUL5o
kAXVAqtZttYqsReKuprDaU2bNsSGzDJntkprsR4makWmMzH73w5cLj8O3/nO
OZ/fvtyE7FkEQYTjo/+EZD9iLUn8nUGwSYKYrROo4cjgBR0WuFvc7P4NHkqn
Gq7DpvMT4Z7rSILDvUolweyXh8VcWHWoqsGF9q4R9y2w2fY0OWc5SZTs1E7R
eWVNMb/dlyTE0QMMJuz8tYHvDz+SqEJGUL/9oN+QYhlJ/JC25upgpXnkhQ9c
In9z6jgsduEKn/jQlhuCYLNu4yYpzO5d+GxwDUmQE31kGmybbo2qgpNqVTGp
8NyM+rNHYTZHEFIIW42yo9vgsf7JP2pYQ3138oedDb9dJ+GAZUcSvWHeYP7F
eNzv8U0lWgGXCi2vHsJiu3Y0ApbHMd66ov+isnxdDiw8YG3cQ8+XxuPfg23X
GgMqYMpdlzkEm0ItMx1w8FvF+Cb0n0cuOm2B8/qnZDJYHjGqa4MDKs7lvob5
qbfKLsDWz3EsNvYlTyp1DYdLVvgL0+HesVWuZvRDJb+6cxEm3D+FbYeVd5mx
D+BPoSmGOswj443XN9PvNTndzaDn290Z2whvuV6TmOUN2w93VMBZB/S72pei
nmXUVgBLv3QWboCFosvZW2HK0K3WsrAPhv6XE33e4a2IhSs9zpta6PfxmxPp
Arf5Gk8UwByy7917JklIFrs1hcDWHn3ta7jE6DPswD5aeJzhUVjQljq/Ex7M
L59k4zwZL0iqgfNORMkkMLWzy/kmnX9snzDA3cb+hEqYOaBu8UF/mlZTdBOs
OVnXVwwX/Ux5+QF2OB0TPIet5fE6L9yfpWU+WeRN753bk0b396VTTcKy8uGt
1fC6oTqvRJi6cUflgN0cTXd3wELWOR6FecWKajsLdujLj12m99+jkXShvm1e
qYeF3v/eM/pM2BTVIHIJxBz8+PUf0a9AGShdDbtlM6pi4NCVm6+E0XlORlAN
5rcJq+2hsOj3kkvOsLLxtpVD5wvv9+7zIon/AZvMCE/yHziYWCs=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.985562118843276`, 7.525817748240936}, {-1, 0}], 
         {PointSize[0.04], PointBox[{14.5, 5.7}], PointBox[{16.5, 9.3}], 
          PointBox[{11., 10.}], PointBox[{17.3, 6.5}], 
          PointBox[{13.700000000000001`, 8.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["11",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         LineBox[{{17.299999999999272`, 6.5}, {13.699999999998909`, 8.5}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJwd0n1MU1cUAPBnNQ4osZ1C6X19D8vqFBmyStOISONxIaKmdEXRdZYhIloj
CmVbjKBklbGMiGDjRwqKbRkjokHtQJgfgM3YJjq0VYkRcNol9SvZZtmIrVXq
zt0fLy+/3HPPufecm1RcvnargGEYDX70b/AMDw9rFwMjRjAswMPcr3ekoeUH
+d4jBORT3OhSIdotb675iMDPc8aqL/mVwDDCk+RdAiOmmuTmPrT+n36DgMD9
cbPDeRTtm9XIJRDI17uNx8rQ1rpxlY6A2H5dJ9eiA5ENFd8TKHm2V9eThi66
0DIWx0LTvitHqxNo/gNPJ5pYKGnJtK58h66LUyQLZWAWzYv5882HwLgKI//2
yaApXvRpVBjN6PyMloPYkq9S9gkwHqrPvrrJQXZczm3BLLT80bWCZTx0eBM/
3pKIdr63frSBh8ngoeikdLTY2Nbp5iG5Mt3ekIO2rHA+8/JwMWvL4B8FaG+C
2tHPg6WMNH9iRptfRge+5cHdfs+fbaH3f+sPpvOwM+/5BVU9Pa/hxHeDHMQZ
uyPBQzR/nKYYOFAqk0PuBrQy9QePSwa1ducH79fS+7ZlZ/IyGK4PHE7+gtYv
7kxtZGFI8bgwy0j7Fzr2t5CF0MDjXRkamq9DEmMnMLRj2h0dS9crRA/WEAi1
WDWTT7EflpGwGvvv6ndEe2zowJc2yUwChvzlkSs5aN96a7WUgHuge8aq12lY
/3h3QS4B/eU8040etHf88KlWAhmq+I2qKrTc1m4Rs2AwTVT51qDFYf3+IywU
NUpvVS5EmyVnpAoZODeXZI1K6P5Xu1/3ysCneTs1NRttFU/aV3LgE7WeieVp
PmX2j9c52NP22V9GFfW1ZV0ZPAzlCuc2rKPxwrIFOJ/7+Xs3nqf1A/mCJT/x
oM+TTy9tR+s9926PYHwEMlVeGt/6y5NBjE8qNVUF/99f/rmVh5DW/tuTBLyv
PGI3a3iQXn0TX5uGdnPzXR4OpKdfci8y0JDYXZ/LwXYFr65U0/f0cKh+QAbK
m+XsCgVauQiCKTJwxd8JLxeg9RO/Gk6woNx57tHqu1jPZ6pgpPh+ax3l05qo
w511HfjeRZv6rq5Fu8q6tq4jwKTvr6mYSfutCH/D43ya1b+X9i7Cfh5cGh2D
82xb9SJzG1pvUsxmcV51NQW7CLpozNajJRDV5T+56W4q2qqJcuD8WsyaGBsa
Ci+vjmXhP5auawE=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.791310365492, 5.560812358621961}, \
{0, 1}], CircleBox[{15.544262499999999`, 7.6960125}, 1.7212037630863173`, {
          1.2012177615541972`, 3.900181530448408}], 
         PolygonBox[{{14.449161702384592`, 9.153166107937269}, {
          13.448460892323894`, 8.379475026985792}, {14.112815702906918`, 
          7.933792757139669}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.43251255449539, 8.913609397826864}, \
{1, -1}], CircleBox[{14.9167375, 8.1169875}, 
          1.7212037630863173`, {-1.9403748920355957`, 0.7585888768586153}], 
         PolygonBox[{{16.011838297615405`, 6.659833892062731}, {
          17.01253910767611, 7.433524973014207}, {16.34818429709308, 
          7.879207242860331}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.02848744550461, 6.899390602173135}, \
{-1, 1}], LineBox[CompressedData["
1:eJwV0XtIU1EcB/CLlk5IsJE40Wpa2NRpQ9R0Zl1r0HQGs5J8VVJulo80dWpm
S6LHQlHRTVaUWYms0tJptAbNTew1THxhvsq1fBVk07DmH1rfc+Fy+XDP+Z3v
916f03lHJA4URYXiJk9KLsUVSFMMAh+aCv9S8a0e5qRKNWLYm8XIHIabNnRP
1sGL4pUIHy5NUYYfRhNcd1BTXgrrpmpdJuDikhTlMGzrM4QRi50clNwgmmIH
N2h7YKmsUH8ZFh9SDapg+idz3gTrNnNvH4cztK0OKzCVerbAFVYHuB/dEkxT
6fbD+tdsmooTzmi3wbzczv15sO7+pIa8f5/nouPA7S+8TtjJ/vTkHMt2mjJk
Oed+hI0xuWkquJ+f9kBNzl8/zy6B6yolVaeI3312vAUXlLBYHJhe21tmhmlb
kKsdfRY6/FRemG9/brk7QvqFlT06A8c3d/8xwe330jx0cATX7XcPbKyofshE
/rE29dgYLOzkl+XAfs/KDzhivkI2/tUAFzkKODTp77Hc5ehLU26CbOcqmPX2
1UwYrBRzA62wsOb6dALcG9CzKRp92TrzeCJxmKy3gXwfxqflGHjCU3L1O+zG
SwhhwUNlmRt5u2kqP7Tr5BTO62decpLCliaP9npYsBJZq4DVc/kSAayViaMa
iJNGBSvo88R9l6gGtoWsurXBzOxSjYzMs82vX4AbvQX6eFh4bceeWLi3YB/b
E1546p8TDo+Exx2zII9NNF0dBY9eucnXkLxz1pYk2KF4VlMMR1xcGqwk/3N0
6JcYbl86lzUIB/Pqc6NI38SphJ3IN2WeC42ENSkdyXJY/8a/X0TWO9eaxmDL
4rJHIaxwyTJx0f/G1pexbbCxjrOlCGYpo1l/4YG+2ZLHsJ/Lml2EvBHmQs8P
sDX4n7wF1jWqrQOwULGYQfHQh29oNsK8yVV5AEwzWr3v+NL/ATmLKdQ=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14.454159905913356`, 9.769881862693708},
           {0, -1}], 
         {PointSize[0.04], PointBox[{16.166, 9.301}], 
          PointBox[{14.295, 6.5120000000000005`}], PointBox[{11., 10.}], 
          PointBox[{17.3, 6.5}], PointBox[{13.700000000000001`, 8.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["12",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 0}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         LineBox[{{17.299999999999272`, 6.5}, {13.699999999998909`, 8.5}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJwd0n1MU1cUAPBnNQ4osZ1C6X19D8vqFBmyStOISONxIaKmdEXRdZYhIloj
CmVbjKBklbGMiGDjRwqKbRkjokHtQJgfgM3YJjq0VYkRcNol9SvZZtmIrVXq
zt0fLy+/3HPPufecm1RcvnargGEYDX70b/AMDw9rFwMjRjAswMPcr3ekoeUH
+d4jBORT3OhSIdotb675iMDPc8aqL/mVwDDCk+RdAiOmmuTmPrT+n36DgMD9
cbPDeRTtm9XIJRDI17uNx8rQ1rpxlY6A2H5dJ9eiA5ENFd8TKHm2V9eThi66
0DIWx0LTvitHqxNo/gNPJ5pYKGnJtK58h66LUyQLZWAWzYv5882HwLgKI//2
yaApXvRpVBjN6PyMloPYkq9S9gkwHqrPvrrJQXZczm3BLLT80bWCZTx0eBM/
3pKIdr63frSBh8ngoeikdLTY2Nbp5iG5Mt3ekIO2rHA+8/JwMWvL4B8FaG+C
2tHPg6WMNH9iRptfRge+5cHdfs+fbaH3f+sPpvOwM+/5BVU9Pa/hxHeDHMQZ
uyPBQzR/nKYYOFAqk0PuBrQy9QePSwa1ducH79fS+7ZlZ/IyGK4PHE7+gtYv
7kxtZGFI8bgwy0j7Fzr2t5CF0MDjXRkamq9DEmMnMLRj2h0dS9crRA/WEAi1
WDWTT7EflpGwGvvv6ndEe2zowJc2yUwChvzlkSs5aN96a7WUgHuge8aq12lY
/3h3QS4B/eU8040etHf88KlWAhmq+I2qKrTc1m4Rs2AwTVT51qDFYf3+IywU
NUpvVS5EmyVnpAoZODeXZI1K6P5Xu1/3ysCneTs1NRttFU/aV3LgE7WeieVp
PmX2j9c52NP22V9GFfW1ZV0ZPAzlCuc2rKPxwrIFOJ/7+Xs3nqf1A/mCJT/x
oM+TTy9tR+s9926PYHwEMlVeGt/6y5NBjE8qNVUF/99f/rmVh5DW/tuTBLyv
PGI3a3iQXn0TX5uGdnPzXR4OpKdfci8y0JDYXZ/LwXYFr65U0/f0cKh+QAbK
m+XsCgVauQiCKTJwxd8JLxeg9RO/Gk6woNx57tHqu1jPZ6pgpPh+ax3l05qo
w511HfjeRZv6rq5Fu8q6tq4jwKTvr6mYSfutCH/D43ya1b+X9i7Cfh5cGh2D
82xb9SJzG1pvUsxmcV51NQW7CLpozNajJRDV5T+56W4q2qqJcuD8WsyaGBsa
Ci+vjmXhP5auawE=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.791310365492, 5.560812358621961}, \
{0, 1}], CircleBox[{15.544262499999999`, 7.6960125}, 1.7212037630863173`, {
          1.2012177615541972`, 3.900181530448408}], 
         PolygonBox[{{14.449161702384592`, 9.153166107937269}, {
          13.448460892323894`, 8.379475026985792}, {14.112815702906918`, 
          7.933792757139669}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.43251255449539, 8.913609397826864}, \
{1, -1}], 
         CircleBox[{14.9167375, 8.1169875}, 
          1.7212037630863173`, {-1.9403748920355957`, 0.7585888768586153}], 
         PolygonBox[{{16.011838297615405`, 6.659833892062731}, {
          17.01253910767611, 7.433524973014207}, {16.34818429709308, 
          7.879207242860331}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.02848744550461, 6.899390602173135}, \
{-1, 1}], LineBox[CompressedData["
1:eJwV0XtIU1EcB/CLlk5IsJE40Wpa2NRpQ9R0Zl1r0HQGs5J8VVJulo80dWpm
S6LHQlHRTVaUWYms0tJptAbNTew1THxhvsq1fBVk07DmH1rfc+Fy+XDP+Z3v
916f03lHJA4URYXiJk9KLsUVSFMMAh+aCv9S8a0e5qRKNWLYm8XIHIabNnRP
1sGL4pUIHy5NUYYfRhNcd1BTXgrrpmpdJuDikhTlMGzrM4QRi50clNwgmmIH
N2h7YKmsUH8ZFh9SDapg+idz3gTrNnNvH4cztK0OKzCVerbAFVYHuB/dEkxT
6fbD+tdsmooTzmi3wbzczv15sO7+pIa8f5/nouPA7S+8TtjJ/vTkHMt2mjJk
Oed+hI0xuWkquJ+f9kBNzl8/zy6B6yolVaeI3312vAUXlLBYHJhe21tmhmlb
kKsdfRY6/FRemG9/brk7QvqFlT06A8c3d/8xwe330jx0cATX7XcPbKyofshE
/rE29dgYLOzkl+XAfs/KDzhivkI2/tUAFzkKODTp77Hc5ehLU26CbOcqmPX2
1UwYrBRzA62wsOb6dALcG9CzKRp92TrzeCJxmKy3gXwfxqflGHjCU3L1O+zG
SwhhwUNlmRt5u2kqP7Tr5BTO62decpLCliaP9npYsBJZq4DVc/kSAayViaMa
iJNGBSvo88R9l6gGtoWsurXBzOxSjYzMs82vX4AbvQX6eFh4bceeWLi3YB/b
E1546p8TDo+Exx2zII9NNF0dBY9eucnXkLxz1pYk2KF4VlMMR1xcGqwk/3N0
6JcYbl86lzUIB/Pqc6NI38SphJ3IN2WeC42ENSkdyXJY/8a/X0TWO9eaxmDL
4rJHIaxwyTJx0f/G1pexbbCxjrOlCGYpo1l/4YG+2ZLHsJ/Lml2EvBHmQs8P
sDX4n7wF1jWqrQOwULGYQfHQh29oNsK8yVV5AEwzWr3v+NL/ATmLKdQ=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.454159905913356`, 9.769881862693708}, {0, -1}], 
         {PointSize[0.04], PointBox[{16.166, 9.301}], 
          PointBox[{14.295, 6.5120000000000005`}], PointBox[{11., 10.}], 
          PointBox[{17.3, 6.5}], PointBox[{13.700000000000001`, 8.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["13",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], LineBox[CompressedData["
1:eJwt1A0w22ccB/C/UK2yyqVWQRAva20hqFRMNf7zsmY306xV56W7RoeyqoT2
Nlyus7JxM6Fl5FDFkW6jmph6O9vZjU0x0qGirp1qdOzcOBUc7ez77Pbc5XKf
e/L8/r/fN0/ick52MolBUdQpvMg7lZiMxaOp/5YLTdGf+jX1wH07ZlwW/OKt
7X6mJ03pv/51F7FZSacsEV6IG7Le4dLUHuuUT7Rw1htrY9PwUrW//RocwAxn
NML9bt3Br3uhvkOJXAoXilObI+HZWmqABdex7V0+hHNTh2uuOGPf0i47GWby
/7g06URThuC2hliYtlwv9oPTmpfLRbCkWfq83pGm2CbFfDYsPzzR4AGnGXm7
/8LzS9tcrUc5qNvXL+iCuXKPjgq4xXyfRQGsuVszlQ8v3LAbjoOZyosSFcwZ
r8sQwrP6rRpyftNVdo4LSz63YrihvlhjfpQNSyWL00rYoDBVk33K/xa1l/TH
fh4cSGz6e0wRrPKOLyZ56co2zlpgPjPGdUYtqf9V070cmPNkkWOA6Y+3DkzB
l4crivwwj8+X10otkY9qpj2pkOSj3vzFHg5b3Ul9TPI7XlJmCwsu88N8+bAx
qsAEtvkof1wB5wZtj7ShHvOBc/oPcN89gzAEdnds56zAVIKDoBP9td+aH2J5
43k5Lqft4M0Cn1R3WMIbSMjAfGecff8hljdaDwySPIKOqF+FdYd+XvOA15SO
n22iHnPOa6rKAfvLnSU6WLLtcvggnF85eLWe9Jfynud9e5oqj3pyRQZLbajF
m/CZ1Rm9iJjTF1YO94vCxvaTfluNHbfhlizv8FXMq/OtiVuEbfh/Xn0E18Vf
cAxBfZ0sndKT71+bxdLCUfePH3tK7ks086kX+tM4WH6zQ/LrrIy+A08UxlR6
knzGJxZ4mK/lZU1UCrxSapFUBStmNoUaWL7bVLFK9rfyxinMW2qifNcL+QTN
8XnRxKl3RsPhxlPfG5tJPiPZ3m/CAS8y5l6SPKvERnP4us38spsP6mvSrbSo
9yOjix8K+yw9VB+FnTJDQ+PhPlXbRgv661HPCy/AXJZo1ApO7r5YnUnOd6oO
JGK+OsOJwgxYukCLe0k+N9YaUsj5L4aHnWB+jsg6FmbylAnldvh9hq9+EAZr
FIwyd9hsw+eSJznfYRROsmnqmW3j+f2kH1GU7XfwSrp6ZJ30L775bSOcX1LL
+g3WiNepfthQ8Sz9NiyNDNSboZ7g0djeanh2K2owAfYY6heqYO5OcN04vJSX
946a3BeW4OH76K+rUJE7QOqFFh0chwN+urvLSD5/Ntc0EvPNPQ5U2KEf3bRr
TDdc3pv2tojMI6/f8wrykPZqOedJHtnZOSFwTNaJzWtw6awTdRrudavQ9pL5
wunXyL4+wsRtHqanhh5YwH+f9I+w8MX9nOQmtqK+T2tT5iHifbJYAVyh7Dkm
gkv9Co80oL8RrAjf//8f7el/AbXbzOU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], 
         LineBox[{{14.294999999996776`, 6.512000000002445}, {
          17.300000000002314`, 6.500000000003638}}], 
         PolygonBox[{{15.197504783982694`, 6.508395987551483}, {
          16.395897890982983`, 6.103607201770314}, {16.399092541051626`, 
          6.9036008231267205`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.791310365492, 5.560812358621961}, \
{0, 1}], CircleBox[{15.544262499999999`, 7.6960125}, 1.7212037630863173`, {
          1.2012177615541972`, 3.900181530448408}], 
         PolygonBox[{{14.449161702384592`, 9.153166107937269}, {
          13.448460892323894`, 8.379475026985792}, {14.112815702906918`, 
          7.933792757139669}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.43251255449539, 8.913609397826864}, \
{1, -1}], LineBox[CompressedData["
1:eJwVUHs01Hkc1WNLqVYRZuY3029G5DW/3xc95NXtvZKJpNNDnju0PVAeaWkr
UpM2YVOIMgahWkp7ClEoU3mcqPU4I5JCL+m1m5Cd/eOee+4599z7+VyhX/B6
6XgNDY1tavzPrYoQh6lnLYAFXmVO07jwKnUZk9+xwPQd8vK6GA46TZJ8jb5b
YH+AY6PqRw4qDlU+G3AWI3HJ5ZSGSgP0O8WKOxViFHMeKD1TDWDv9Gum1Xcx
/II4qY+yDCBbYuDYuJ7B+9Tie5qtBnA2+5ifmM7g98Uax+7P52CXl+x08CMG
a7el/Cwq5sAtdqNq2kcGH3tOVB1x4MJgc3jgzTEGYXyFpKKJC11nq56vwwwi
RmRFUn8edqU3SBpfMxh0T+2IeMOD8cUNGs/VeXfzOt7pSinsvmosfnCNwbac
cWHd9RQ+GA9/DjrNwEw5MMeG5iOh66X11UgGCe96rhR48vEpY+4kG18Gpl0r
n3Qd5EN+7OzUkxIG7mUxOT3H+FDaVEqDwcC/0/17aigfNqOpTe4LGYwMbNWc
v5yPcsnRHz6yDLJztmsc+Eohe1Xqw1Ixg96VZk/bzlCw5Oabz7BkcDLS6ovY
kEJVp06BlT0Dn/l3hpQKHso+t3T3uTHwCPuslNA8cMfumdeGMYgJjJfdkXNx
a2tzfmgeA9X42H0OVlwk2YamPetlsOpMYN72Dg50RqjqCmsWD4Jlb20LOXj6
4f4mbxmLCPlOQ18FB1VeRirzThbx3juoo3UcZIlvL7qvR0BZ1Uv1hVz4K17u
VCwi2OSR0Tclk4tViR+e1kkIAuoqzrRb8xCw1rw5awuBKnbkVVIrD23dM6/3
qvVkvQVDLr9SiGa3n89zJfjbNHdplDYfHrrive0OBA2yK1Vrk/kwvRNxcWwu
gVldYQJ/hI+gxcbd1ZoEf3jaXGx3ESD+suFk31YWO8VWrlXHBejZM3aLl8XC
uEU/49M1AbZrl92OCmbhNzq/z6hegBPNQdN9fmIxfv2nhMctAlRlMy8vWbD4
R7Q2X+uxALVb4uNW8VksvzBBp+22ANMHpDZ2XBbjYq4PzswQwHUses2oiEXS
Me91XoECzAsUmNYsZFHbqryrZShAh+eaox0bWOSYLHE738yH8IyP8Hk0C0lA
5PcvUXyIJ6QF5f3JwnE/x3T2PPV/666YPn/FgpOrE5D8lILkW6pxm3rf9u5J
SWUFFHQDrVeOsyd4McGh3yeFwqEtXqVyP4JC/cAZAfkU6ptX61jICKzbteyE
fRT0tFZEJxQQ+JDBiRPd+HDRC6vJryGQ5VzsfNvFh/SvplOdjwmIMm13vEyA
6BzXkHNtBKbHT556bTcHmTozQy80EWjayy3fvJqDxi/LxaO3CZ7IXFX/LqWR
VBzDNOYSTC/TbEo+SONbWGZFeRyBn4VE0V9II7e65eAR9X0l2Te8y2pozD09
frPckaAoLKFttI5GgPXuLg5FcOlspq2RksbN3p5lU4ZYRM4oqnMupVG6mXNj
WTWLrMN3mT15NAbotM/vUljYr1badiWp9SSP7PvhLL7d6HOrP0DDxSjd29af
RVRA+uUVO2mU23kUWnqzCDn8/t3WLTQmSYvDh39hscwhkXduDQ3F4qn9FrEs
9o7MdEu0o7FRv05YUMBi0YylcVvFNLRVfgfKVCycxzYV6Qhp3CsxMk7RIkix
W5b7yoCGrYcqXW5BUPrMxH1Un8b1TCe55zoCL7um0n61v3HzTebuHgJn3T0i
bVsa/t1Fb8KTCNzY2u4mPxqUiUnyw0sErx+8GC7PoNHgUTnIqySoDm1//PUl
jbaIuBdsLUFnc/lbI1shglsaFhVVE/TvnhVyI1mIkllFR2JK1H06Nft6+4SY
LYX5hzSCCbVR+UOWIni4NjwLiSTw1dvodnqHCIknam3IeoIc+VSnkpMiNIlU
Q7+ZEcyz1LxMpYvwH48Rgg0=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.02848744550461, 6.899390602173135}, \
{-1, 1}], 
         LineBox[{{16.16600000001199, 9.301000000006752}, {
          13.700000000010732`, 8.500000000003638}}], 
         PolygonBox[{{14.362348956376628`, 8.715142544224527}, {
          15.380079406439721`, 9.466291484857722}, {15.627222680807021`, 
          8.705423426693228}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.454159905913356`, 9.769881862693708}, {0, -1}], 
         {PointSize[0.04], PointBox[{16.166, 9.301}], 
          PointBox[{14.295, 6.5120000000000005`}], PointBox[{11., 10.}], 
          PointBox[{17.3, 6.5}], PointBox[{13.700000000000001`, 8.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["14",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         LineBox[{{17.299999999999272`, 6.5}, {13.699999999998909`, 8.5}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJwd0n1MU1cUAPBnNQ4osZ1C6X19D8vqFBmyStOISONxIaKmdEXRdZYhIloj
CmVbjKBklbGMiGDjRwqKbRkjokHtQJgfgM3YJjq0VYkRcNol9SvZZtmIrVXq
zt0fLy+/3HPPufecm1RcvnargGEYDX70b/AMDw9rFwMjRjAswMPcr3ekoeUH
+d4jBORT3OhSIdotb675iMDPc8aqL/mVwDDCk+RdAiOmmuTmPrT+n36DgMD9
cbPDeRTtm9XIJRDI17uNx8rQ1rpxlY6A2H5dJ9eiA5ENFd8TKHm2V9eThi66
0DIWx0LTvitHqxNo/gNPJ5pYKGnJtK58h66LUyQLZWAWzYv5882HwLgKI//2
yaApXvRpVBjN6PyMloPYkq9S9gkwHqrPvrrJQXZczm3BLLT80bWCZTx0eBM/
3pKIdr63frSBh8ngoeikdLTY2Nbp5iG5Mt3ekIO2rHA+8/JwMWvL4B8FaG+C
2tHPg6WMNH9iRptfRge+5cHdfs+fbaH3f+sPpvOwM+/5BVU9Pa/hxHeDHMQZ
uyPBQzR/nKYYOFAqk0PuBrQy9QePSwa1ducH79fS+7ZlZ/IyGK4PHE7+gtYv
7kxtZGFI8bgwy0j7Fzr2t5CF0MDjXRkamq9DEmMnMLRj2h0dS9crRA/WEAi1
WDWTT7EflpGwGvvv6ndEe2zowJc2yUwChvzlkSs5aN96a7WUgHuge8aq12lY
/3h3QS4B/eU8040etHf88KlWAhmq+I2qKrTc1m4Rs2AwTVT51qDFYf3+IywU
NUpvVS5EmyVnpAoZODeXZI1K6P5Xu1/3ysCneTs1NRttFU/aV3LgE7WeieVp
PmX2j9c52NP22V9GFfW1ZV0ZPAzlCuc2rKPxwrIFOJ/7+Xs3nqf1A/mCJT/x
oM+TTy9tR+s9926PYHwEMlVeGt/6y5NBjE8qNVUF/99f/rmVh5DW/tuTBLyv
PGI3a3iQXn0TX5uGdnPzXR4OpKdfci8y0JDYXZ/LwXYFr65U0/f0cKh+QAbK
m+XsCgVauQiCKTJwxd8JLxeg9RO/Gk6woNx57tHqu1jPZ6pgpPh+ax3l05qo
w511HfjeRZv6rq5Fu8q6tq4jwKTvr6mYSfutCH/D43ya1b+X9i7Cfh5cGh2D
82xb9SJzG1pvUsxmcV51NQW7CLpozNajJRDV5T+56W4q2qqJcuD8WsyaGBsa
Ci+vjmXhP5auawE=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.791310365492, 5.560812358621961}, \
{0, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{15.544262499999999`, 7.6960125}, 1.7212037630863173`, {
           1.2012177615541972`, 3.900181530448408}]}, 
         PolygonBox[{{14.449161702384592`, 9.153166107937269}, {
          13.448460892323894`, 8.379475026985792}, {14.112815702906918`, 
          7.933792757139669}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.43251255449539, 8.913609397826864}, \
{1, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{14.9167375, 8.1169875}, 
           1.7212037630863173`, {-1.9403748920355957`, 0.7585888768586153}]}, 
         PolygonBox[{{16.011838297615405`, 6.659833892062731}, {
          17.01253910767611, 7.433524973014207}, {16.34818429709308, 
          7.879207242860331}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.02848744550461, 6.899390602173135}, \
{-1, 1}], LineBox[CompressedData["
1:eJwV0XtIU1EcB/CLlk5IsJE40Wpa2NRpQ9R0Zl1r0HQGs5J8VVJulo80dWpm
S6LHQlHRTVaUWYms0tJptAbNTew1THxhvsq1fBVk07DmH1rfc+Fy+XDP+Z3v
916f03lHJA4URYXiJk9KLsUVSFMMAh+aCv9S8a0e5qRKNWLYm8XIHIabNnRP
1sGL4pUIHy5NUYYfRhNcd1BTXgrrpmpdJuDikhTlMGzrM4QRi50clNwgmmIH
N2h7YKmsUH8ZFh9SDapg+idz3gTrNnNvH4cztK0OKzCVerbAFVYHuB/dEkxT
6fbD+tdsmooTzmi3wbzczv15sO7+pIa8f5/nouPA7S+8TtjJ/vTkHMt2mjJk
Oed+hI0xuWkquJ+f9kBNzl8/zy6B6yolVaeI3312vAUXlLBYHJhe21tmhmlb
kKsdfRY6/FRemG9/brk7QvqFlT06A8c3d/8xwe330jx0cATX7XcPbKyofshE
/rE29dgYLOzkl+XAfs/KDzhivkI2/tUAFzkKODTp77Hc5ehLU26CbOcqmPX2
1UwYrBRzA62wsOb6dALcG9CzKRp92TrzeCJxmKy3gXwfxqflGHjCU3L1O+zG
SwhhwUNlmRt5u2kqP7Tr5BTO62decpLCliaP9npYsBJZq4DVc/kSAayViaMa
iJNGBSvo88R9l6gGtoWsurXBzOxSjYzMs82vX4AbvQX6eFh4bceeWLi3YB/b
E1546p8TDo+Exx2zII9NNF0dBY9eucnXkLxz1pYk2KF4VlMMR1xcGqwk/3N0
6JcYbl86lzUIB/Pqc6NI38SphJ3IN2WeC42ENSkdyXJY/8a/X0TWO9eaxmDL
4rJHIaxwyTJx0f/G1pexbbCxjrOlCGYpo1l/4YG+2ZLHsJ/Lml2EvBHmQs8P
sDX4n7wF1jWqrQOwULGYQfHQh29oNsK8yVV5AEwzWr3v+NL/ATmLKdQ=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.454159905913356`, 9.769881862693708}, {0, -1}], 
         {PointSize[0.04], PointBox[{16.166, 9.301}], 
          PointBox[{14.295, 6.5120000000000005`}], PointBox[{11., 10.}], 
          PointBox[{17.3, 6.5}], PointBox[{13.700000000000001`, 8.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["15",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{10.999999999998181`, 10.000000000003638`}, {
          13.69999999999709, 8.500000000003638}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          12.68023719320147, 8.608951330844206}, {13.068751538144376`, 
          9.308277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 8.499856222011616}, {1, 1}], 
         LineBox[{{17.299999999999272`, 6.5}, {13.699999999998909`, 8.5}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJwd0n1MU1cUAPBnNQ4osZ1C6X19D8vqFBmyStOISONxIaKmdEXRdZYhIloj
CmVbjKBklbGMiGDjRwqKbRkjokHtQJgfgM3YJjq0VYkRcNol9SvZZtmIrVXq
zt0fLy+/3HPPufecm1RcvnargGEYDX70b/AMDw9rFwMjRjAswMPcr3ekoeUH
+d4jBORT3OhSIdotb675iMDPc8aqL/mVwDDCk+RdAiOmmuTmPrT+n36DgMD9
cbPDeRTtm9XIJRDI17uNx8rQ1rpxlY6A2H5dJ9eiA5ENFd8TKHm2V9eThi66
0DIWx0LTvitHqxNo/gNPJ5pYKGnJtK58h66LUyQLZWAWzYv5882HwLgKI//2
yaApXvRpVBjN6PyMloPYkq9S9gkwHqrPvrrJQXZczm3BLLT80bWCZTx0eBM/
3pKIdr63frSBh8ngoeikdLTY2Nbp5iG5Mt3ekIO2rHA+8/JwMWvL4B8FaG+C
2tHPg6WMNH9iRptfRge+5cHdfs+fbaH3f+sPpvOwM+/5BVU9Pa/hxHeDHMQZ
uyPBQzR/nKYYOFAqk0PuBrQy9QePSwa1ducH79fS+7ZlZ/IyGK4PHE7+gtYv
7kxtZGFI8bgwy0j7Fzr2t5CF0MDjXRkamq9DEmMnMLRj2h0dS9crRA/WEAi1
WDWTT7EflpGwGvvv6ndEe2zowJc2yUwChvzlkSs5aN96a7WUgHuge8aq12lY
/3h3QS4B/eU8040etHf88KlWAhmq+I2qKrTc1m4Rs2AwTVT51qDFYf3+IywU
NUpvVS5EmyVnpAoZODeXZI1K6P5Xu1/3ysCneTs1NRttFU/aV3LgE7WeieVp
PmX2j9c52NP22V9GFfW1ZV0ZPAzlCuc2rKPxwrIFOJ/7+Xs3nqf1A/mCJT/x
oM+TTy9tR+s9926PYHwEMlVeGt/6y5NBjE8qNVUF/99f/rmVh5DW/tuTBLyv
PGI3a3iQXn0TX5uGdnPzXR4OpKdfci8y0JDYXZ/LwXYFr65U0/f0cKh+QAbK
m+XsCgVauQiCKTJwxd8JLxeg9RO/Gk6woNx57tHqu1jPZ6pgpPh+ax3l05qo
w511HfjeRZv6rq5Fu8q6tq4jwKTvr6mYSfutCH/D43ya1b+X9i7Cfh5cGh2D
82xb9SJzG1pvUsxmcV51NQW7CLpozNajJRDV5T+56W4q2qqJcuD8WsyaGBsa
Ci+vjmXhP5auawE=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.791310365492, 5.560812358621961}, \
{0, 1}], LineBox[CompressedData["
1:eJwVkHk0lQkYxuWmkK11dN17fd8nN/UtliRM8rQv5EaWI8mSjNTktpF0FJNI
N1qUJcnRqUTJkn13tNy4IktlaaRlOramMtOiMnf+eM97nvOc877P8yP9Q1x2
qKqoqHgr5//ddVVqp5nMAIu3VazX4mPLk9TaG2sZrHH1PCYw5qNlnSxEPE6j
ObghZMsuPpwMr3cbl9E4OS2sPrGXj5mND+fzYmi0Sd9RJVIDpC09VX8zkEZV
5XqLWQYCyMcNgrJ9aDz+8K/1kmYBXtjlSFpDaGjkVwafPSCEfWqX1v4kGuW8
a/xpU0UoFD/wFMhpDFsPOP15WITuxSXtR9UZfNkjU/g3i6Cbd37UaSODdqI5
o+C7CDx76XjjOQaCLsHuOD1DjET1vO7rYBBADD64o2OI33QaizbqsfiWUlXk
9VME4omXrs8KFlTE8NTmNyJ03hGs7wlikZuxQeovF2Hf8dBKrWgW04ROPY65
IlTfbjvRJWORl5Ro1XRKBLpR1qaIY5H6yKb8604RTg20JdiHskiff2XP4Cql
fnh6Kd+dxS8a/emNAhHqErUP31vAwinpzd3MT0LUqa4c/fmRwZXDehFDCiFu
Rzvu6yhiYBO06LJ1oRBm9eTRe3sYmOc5+8VmC2EaEG4+IWawwGXCRadciOMT
Q8Vn+mk8WvakqWBQCI84smNXJo2DRt5S7+Ui9P3I6NLeTeNNjMLrfJUIyxom
O8jX0Yj1UvQnuxviWGbMpjZbGhax0uWOqgTyIv+Y7rCKhmlQsMt3LwJ9Vcty
SgJoRIe/muuQT8CxPcN6KIVG8qLaE7WfCfSvTNbM76HB+RQM5TMkwi2DaxqV
+T7W8qJ1HElIDOquqx1g0CnXn33WjUTv2NgrjxoGPd1FCHBQ+mnhU8WqLH4N
Z90PmZNweHu6wc+OhUlizliJFgmPzfd1xSEsMvXHT64ZIJAu682hk1mElib7
fiwh4Bytrr66mMVrn/brlQkETAZvvtWUs5iIjfU7t5vArvG6L81tLB5o9p2x
diFQmJCeW9/CQnyk+KgbCJSu+eKwtZaFo+304korAiNzyoyyr7K4n1Vd/9SS
gN7z3XtlR1jMUaRf4GwJXAybf2DYgcVdX7rywRoCo0t5C2/MYjFjseilr5JX
emj3nuhuBpHCS8Wu4QTEO20G1LMYWLx9WW+RRaClrtnRTcrAN9bXxOApgWPx
7mMVGxj0lQS84+mTkBvz5LZWDBrr0mX87SRubWqV3Vfq5xo2SYfukhh5uPKg
1InBrMBn1a5TKBRmeXnWH2FgQO68UOVMYUlCdn2Ykm+5hXNXcSKF0vhJ7r3a
LFQFqhGRFRQuffZ+lr2dhVZE1OczLRQCRjRCNpSxeDXz04x5TRT+yitKnq/G
4doxnpV2IYWq065xSas5bHl0UEHFUtg26Ku2IozDFP2Pq/wkFGrGJBIijcPl
6XbGfD0KZZoz4zrvcEgx95B5t5BY+zhS7VsFhyjdjIScBBJdWQmBTjUcOkyy
+J2uJKrTMuk5VRxCv6pq8OaRwPBpn7vFHDa4asTc+EmgID+oKSWXw8Wt2poH
3xJ46pcojris/Df6T4mkl0CMoyrdGs/BYuusK879BIY0lv9w389BxW114fAH
AnEN9NJUDw7m7+zDjfVIaFmrTLa3UeYLt3rcakNi84onZmIDDgFf1s0o/V3p
SzwHbCZxaCGH+CdySOzIGNuo+zcLmWJfQcIoCb8zBXLJsJJf9ZuX/ksomAS2
uo1/ZaHw3GtyNIqC1qmM9BTlPeLysLvZQwpsrWVcmLMyj2np6j41I7y/qfPJ
JpWD08Q2cbWlEWwzb53vHFH2S1po91xihMQrV3kL15nifY2L8UxXI1QEBW2a
m2kKFcavphtGSKu78HhgxBTx+vpZm/WNoJ0yrF802wwNFwNze15QaKN6vkYu
NIOJufotQRqF/wA6XXMf
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], {13.43251255449539, 8.913609397826864}, \
{1, -1}], LineBox[CompressedData["
1:eJwVUHs01Hkc1WNLqVYRZuY3029G5DW/3xc95NXtvZKJpNNDnju0PVAeaWkr
UpM2YVOIMgahWkp7ClEoU3mcqPU4I5JCL+m1m5Cd/eOee+4599z7+VyhX/B6
6XgNDY1tavzPrYoQh6lnLYAFXmVO07jwKnUZk9+xwPQd8vK6GA46TZJ8jb5b
YH+AY6PqRw4qDlU+G3AWI3HJ5ZSGSgP0O8WKOxViFHMeKD1TDWDv9Gum1Xcx
/II4qY+yDCBbYuDYuJ7B+9Tie5qtBnA2+5ifmM7g98Uax+7P52CXl+x08CMG
a7el/Cwq5sAtdqNq2kcGH3tOVB1x4MJgc3jgzTEGYXyFpKKJC11nq56vwwwi
RmRFUn8edqU3SBpfMxh0T+2IeMOD8cUNGs/VeXfzOt7pSinsvmosfnCNwbac
cWHd9RQ+GA9/DjrNwEw5MMeG5iOh66X11UgGCe96rhR48vEpY+4kG18Gpl0r
n3Qd5EN+7OzUkxIG7mUxOT3H+FDaVEqDwcC/0/17aigfNqOpTe4LGYwMbNWc
v5yPcsnRHz6yDLJztmsc+Eohe1Xqw1Ixg96VZk/bzlCw5Oabz7BkcDLS6ovY
kEJVp06BlT0Dn/l3hpQKHso+t3T3uTHwCPuslNA8cMfumdeGMYgJjJfdkXNx
a2tzfmgeA9X42H0OVlwk2YamPetlsOpMYN72Dg50RqjqCmsWD4Jlb20LOXj6
4f4mbxmLCPlOQ18FB1VeRirzThbx3juoo3UcZIlvL7qvR0BZ1Uv1hVz4K17u
VCwi2OSR0Tclk4tViR+e1kkIAuoqzrRb8xCw1rw5awuBKnbkVVIrD23dM6/3
qvVkvQVDLr9SiGa3n89zJfjbNHdplDYfHrrive0OBA2yK1Vrk/kwvRNxcWwu
gVldYQJ/hI+gxcbd1ZoEf3jaXGx3ESD+suFk31YWO8VWrlXHBejZM3aLl8XC
uEU/49M1AbZrl92OCmbhNzq/z6hegBPNQdN9fmIxfv2nhMctAlRlMy8vWbD4
R7Q2X+uxALVb4uNW8VksvzBBp+22ANMHpDZ2XBbjYq4PzswQwHUses2oiEXS
Me91XoECzAsUmNYsZFHbqryrZShAh+eaox0bWOSYLHE738yH8IyP8Hk0C0lA
5PcvUXyIJ6QF5f3JwnE/x3T2PPV/666YPn/FgpOrE5D8lILkW6pxm3rf9u5J
SWUFFHQDrVeOsyd4McGh3yeFwqEtXqVyP4JC/cAZAfkU6ptX61jICKzbteyE
fRT0tFZEJxQQ+JDBiRPd+HDRC6vJryGQ5VzsfNvFh/SvplOdjwmIMm13vEyA
6BzXkHNtBKbHT556bTcHmTozQy80EWjayy3fvJqDxi/LxaO3CZ7IXFX/LqWR
VBzDNOYSTC/TbEo+SONbWGZFeRyBn4VE0V9II7e65eAR9X0l2Te8y2pozD09
frPckaAoLKFttI5GgPXuLg5FcOlspq2RksbN3p5lU4ZYRM4oqnMupVG6mXNj
WTWLrMN3mT15NAbotM/vUljYr1badiWp9SSP7PvhLL7d6HOrP0DDxSjd29af
RVRA+uUVO2mU23kUWnqzCDn8/t3WLTQmSYvDh39hscwhkXduDQ3F4qn9FrEs
9o7MdEu0o7FRv05YUMBi0YylcVvFNLRVfgfKVCycxzYV6Qhp3CsxMk7RIkix
W5b7yoCGrYcqXW5BUPrMxH1Un8b1TCe55zoCL7um0n61v3HzTebuHgJn3T0i
bVsa/t1Fb8KTCNzY2u4mPxqUiUnyw0sErx+8GC7PoNHgUTnIqySoDm1//PUl
jbaIuBdsLUFnc/lbI1shglsaFhVVE/TvnhVyI1mIkllFR2JK1H06Nft6+4SY
LYX5hzSCCbVR+UOWIni4NjwLiSTw1dvodnqHCIknam3IeoIc+VSnkpMiNIlU
Q7+ZEcyz1LxMpYvwH48Rgg0=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.02848744550461, 6.899390602173135}, \
{-1, 1}], LineBox[CompressedData["
1:eJwV0XtIU1EcB/CLlk5IsJE40Wpa2NRpQ9R0Zl1r0HQGs5J8VVJulo80dWpm
S6LHQlHRTVaUWYms0tJptAbNTew1THxhvsq1fBVk07DmH1rfc+Fy+XDP+Z3v
916f03lHJA4URYXiJk9KLsUVSFMMAh+aCv9S8a0e5qRKNWLYm8XIHIabNnRP
1sGL4pUIHy5NUYYfRhNcd1BTXgrrpmpdJuDikhTlMGzrM4QRi50clNwgmmIH
N2h7YKmsUH8ZFh9SDapg+idz3gTrNnNvH4cztK0OKzCVerbAFVYHuB/dEkxT
6fbD+tdsmooTzmi3wbzczv15sO7+pIa8f5/nouPA7S+8TtjJ/vTkHMt2mjJk
Oed+hI0xuWkquJ+f9kBNzl8/zy6B6yolVaeI3312vAUXlLBYHJhe21tmhmlb
kKsdfRY6/FRemG9/brk7QvqFlT06A8c3d/8xwe330jx0cATX7XcPbKyofshE
/rE29dgYLOzkl+XAfs/KDzhivkI2/tUAFzkKODTp77Hc5ehLU26CbOcqmPX2
1UwYrBRzA62wsOb6dALcG9CzKRp92TrzeCJxmKy3gXwfxqflGHjCU3L1O+zG
SwhhwUNlmRt5u2kqP7Tr5BTO62decpLCliaP9npYsBJZq4DVc/kSAayViaMa
iJNGBSvo88R9l6gGtoWsurXBzOxSjYzMs82vX4AbvQX6eFh4bceeWLi3YB/b
E1546p8TDo+Exx2zII9NNF0dBY9eucnXkLxz1pYk2KF4VlMMR1xcGqwk/3N0
6JcYbl86lzUIB/Pqc6NI38SphJ3IN2WeC42ENSkdyXJY/8a/X0TWO9eaxmDL
4rJHIaxwyTJx0f/G1pexbbCxjrOlCGYpo1l/4YG+2ZLHsJ/Lml2EvBHmQs8P
sDX4n7wF1jWqrQOwULGYQfHQh29oNsK8yVV5AEwzWr3v+NL/ATmLKdQ=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.454159905913356`, 9.769881862693708}, {0, -1}], 
         {PointSize[0.04], PointBox[{16.166, 9.301}], 
          PointBox[{14.295, 6.5120000000000005`}], PointBox[{11., 10.}], 
          PointBox[{17.3, 6.5}], PointBox[{13.700000000000001`, 8.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["16",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         LineBox[{{14.499999999992724`, 14.30000000000291}, {
          17.299999999997453`, 13.500000000003638`}}], 
         PolygonBox[{{16.476914368584495`, 13.735167323261573`}, {
          15.213197180256556`, 13.680223097682099`}, {15.432974082574459`, 
          14.449442255794757`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.325817748240937`, 14.785562118843277`}, {0, -1}], LineBox[CompressedData["

1:eJwd0w0slHEcB/AndisvCeU6Oi8J6bxsjkmyebxG3CXLuqQRMms7LivSMnIy
FLtCsbwtDaF5iXVa8x6nS2UOZZ3U1KSXu+zGHdr6/vtvz/777Pfs9/J//s/+
xPToC3oURR3FQ3YqNQXLhaZYBAdpStYxWhoAh8X81I+A83b257vBcRUOnx/A
zgFu0y7wa5eJzt8w643Pih/sF2kj9HemqRanboMEeME/0KMELjo7H1MOex3/
mCmHRW1Tq9MwW3+VRx2iKVPeJw8bV5paXBLecoJZ8hG5EO6pvPLDD16eynbo
hxUmIscgWO3oVWHoRlODxrnqI3DV1J4uPry86qi1J+9fNqkshNutpodIfmlT
7b4OuOyDqmAW9SWNxbvHYN8i9XozbGddff4V3Jp5yyKb2IJp/gIWne5T8oib
jAQ1JK5IYnLg5cbwa0LYuKdtlxmsPRnbwYWZ+lnpDDL/sUcrKvQr6yvYbgDL
YpS1j+E0Tq2UDQvOtWckw7OVzck0bGqooBzhee0NdQapV551U4Xz+WmcG9MN
+4h19eOwYDJzaJPYrIbVCQd/SxuJwHyLHqKeVjj6Hiu/Dn43zLOUwhzvl1oV
bGd6Z2QO5jYYHfbjoP9A2mQH6tlXuNXnwpK8vbxQWG803vYZTMsW1spIf6Gp
iYuwtviiUgnnSyUaHexz+226O+Zd7XsYq4+8dme69K7DbG+x5V/EdxhxrYbh
qq+aJ8uknkLduQUnTjKuyEhc/DTL2Z2mtui6wno4KkduEASrQ8QFl0j+cJ15
BKwoVSQEwlLrrGQaHrxaYcsk/TYq1w/ArkXM+2rMJ9je+2kd+Z+3BXjOwEWp
TU4DsECVMjAG03dP1efAkRYM33FyP2Zmmg+T76nx6p3779tLf8j9q+kO0cJX
vzetdcJUnmSbM+q1/JVzs2HzE60mF0j/7OI4PszdqRlohxfZPRc94Y202YFN
OGxvtoADW5VsBPNxPnSljuEFRzMC+A1ww8c0ThT8Xu7+5RcsYbQJcuA67Wce
yZOQ1GXRh31toiWe/Af/F+m3GsuV/gcJsVJu
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.658628882191852`, 12.2818774622309}, \
{-1, 0}], LineBox[{{16.5, 10.699999999998909`}, {14.5, 14.299999999999272`}}],
          PolygonBox[{{15.208614241292821`, 13.024494365672922`}, {
          16.141048669155794`, 12.16976280679853}, {15.441722848258564`, 
          11.781248461855625`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.25014377798838, 12.647946543326881`}, {-1, -1}], LineBox[CompressedData["
1:eJwV0w0slHEcB/DHS87IHHnr4gp5OzchcaLrqTV0I+Zl0TQvybzc8rIzyqlb
sxytcjZTaFnNSLLRMe+0G5FLIfNayevcaN7SdY76Ps/27Nlnv+33+/++z/NY
x6eF3tQkCMIHN/Ukvsjl8kA3ktCh4EwSlrVW3jFwQ90RtZRFEuTfKm4y3Js6
w0yA+THJZikw96tnsBGsfLpQFAfPvftZ0utEEtI1+71gOKS3azgLJkeNIzzh
g2bbQg84kFWmYwZr7hs0/XMkCWEUfWXDlSRYx0oEU3AOg8kYhFNSikxksIm5
bL0Kvrb5ZLoTPry+1HwHjt4WGfbDnLKsmkj4pME07Rss7stp9oV5S2wu1b+W
7ZbpCL+0p+c6YT7HxWfXEh7ONtSNhAUZpz8z4PxCgeIB5d87u9Zwfa2fTROs
ZKvG3Ki6c+XqOEx3mbl4GaYHhm/8gomJ0NpEeF4tLlbBrrNv9cRwuF+uchc+
EWHk9gZuGI/3X4CjR173f4JvpURW9MBqVem9dZik3TV8BPM9L53TRT5626K6
IDiWq+VqCTe5F3K14YQ/jdsOcP9Ria8U+1msfKhmwRvJK3lxcCUzatsW3ho8
S6PDO9leTBN4LWZnQOZAEsXffyj3ME+VOhorgjvPC6ZmYYZv1VwAvFhRnN9K
5d1W9vw4rNvBM5LA5f5p+jpU3dtiIIl6X37Xq9X2yCus5uACTHwULmuiLlQt
bzHhdIZ+qDk1r7SgRoPK211V5AUrhYHiiVPIryfudizcet9jtxN2aRfmPoYH
WnTWpTAv0/lFFxwyL67ogdMLUudW4ZxoG91ZeCt7NckQ+4UL7eja6M8N75az
4N6hev4JeJL+jPSGF1siXKnzjSpUfRyYbmrKTqD20R/iOVLfU2mbQQHseCOo
nQYHPKzPr4FdRGkHE5jH1ihPlMF6jRGb5bA6o2Nkkuq/pU+GUfttVvOWYRuJ
E6kNRwtkYwp4I+tQRyPySTqjtb9EWdXdGANLJQOKCdjPY8TKGObk1YW9h/k0
ZpPcDvtccch6BXM8r1aVwNr4LUVwfCIuO/I/oWlSwg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.341371117808144`, 12.718122537769101`}, {1, 0}], 
         LineBox[{{16.49999999999818, 10.700000000000728`}, {
          13.69999999999709, 11.5}}], 
         PolygonBox[{{15.676914368584494`, 10.935167323261576`}, {
          14.63297408257446, 11.649442255794757`}, {14.413197180256557`, 
          10.880223097682098`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.674182251759065`, 10.214437881156726`}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.5, 10.700000000000001`}], 
          PointBox[{14.5, 14.3}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["17",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         LineBox[{{14.499999999992724`, 14.30000000000291}, {
          17.299999999997453`, 13.500000000003638`}}], 
         PolygonBox[{{16.476914368584495`, 13.735167323261573`}, {
          15.213197180256556`, 13.680223097682099`}, {15.432974082574459`, 
          14.449442255794757`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.325817748240937`, 14.785562118843277`}, {0, -1}], 
         LineBox[{{14.5, 14.300000000007458`}, {13.700000000000728`, 
          11.500000000010914`}}], 
         PolygonBox[{{14.264832676738427`, 13.476914368584495`}, {
          13.550557744205246`, 12.432974082574457`}, {14.319776902317905`, 
          12.213197180256554`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.985562118843276`, 12.474182251759064`}, {-1, 0}], LineBox[CompressedData["

1:eJwl1AtMU2cYBuADDgRR6BiYFmFcROQmreigGxWOBgYORATEKm5Bbralo7jg
VqCMcgnUORQFAUUZbIANl6xjddQNMharQBGpkmhJ54ZQQOY2KoOlchl7z/Yn
JydP/vzf+d7vtMc9VRSfYU4QRAIu6v7/2kUS9Vex/ElCy1A7HYE9Xc4WSGBW
RY//NTjxZPR0LCwX0AKfwLpWYb093JSV8eK1AJJQXjV/V+NHEgqPJnt3ODfS
Py0PJnPspL6wwtho3A6bejn7qP1ecdrbD3xJwkoVV2YJu+WNfy6Bnwtf7/wZ
9Xv7zpoC4Spmv9lNmP3TLYslH/TXHFHNh1UyifQuzLN2+sMbNpaWHmyDq9Zd
tXPoy1OjDq6HxYW+BgXs/XR0UwPsvdNFWAJP1M5kd8FczwIiFZZcbLQfgWkt
o4nxsCm77f7fMN213eYorGZGjXqhH7Z+uUEAW/15aoYLc0OLD16GaZ8ZfpDB
MscBlgY+uSh5sxtOuTOdYYf+uCFzEQ9hVq6b5wdUnqEvaFOwKTnGphse/Egf
NA3TrF1PU/MQHxoWPabOx4wXJMEcQ3n5dzD94w/HbsASi5z3y2ExN7hcB9N9
NL9GUvuHWvI3MEnCf/XM5Br6Z8/01jnDTQnfLnXAg8pjsx4wZzLAL4kyMdC6
FZYvl44RsKrULPQVVV9vClZ4k4R0sCxwBG5Zu57Ig/v5gsUrsFbGrfOnrI/u
OAbLt9nI13fiPfrJXm6F4x5G7puCVVzOOR3y3RdlEzpY23XLthm2CpgteQob
6bFduXCLmBmxAKc8ydcnwerQ8NE3UJ9lm+LxHpyyoNeFUt4kXTwMX0989qkI
zlF61fJgad1gSTMsX6MxqmHTfM9FLWz1sjZrhPIFl2ITPFF9Kt8B/a1y3mpz
RF7Wo4rUdDjFdfc/nrDiyqRCBRsrWHzK9L6q9s2YT+fNaY09LFVsuZbMpP4f
SmcjVS99YVszTHvsmt0Hk42FmeOwg6BsTAIb89vvECyS2JsbEL6b8lcDQzYw
Z2Gd/wvykgmStC1wjV6ZIYPpd6MGLWDhcO35PZRDCs+/QD3nZ3POBi/kvfco
vQ8mzC+bN8LGoZWvy+AaZnJlOiw143dGwDz6X2HvwERDAmEBGyQb2txglnxR
OIB84QrLEDosP2BZfAk2eYhSqX2xsDD2NOww1nQmmDo/EOQdAxtuMIZPwLoT
D7IOwJ2WttYVMI3tkBwNl/Wr96rguOOXNlLnhUt+G3+D2QzBkRp4s1llAB15
5EcVr7Qwr2O+JhRusli5x0B/E2HdXx6HrermbwtgmQ0jLxPuV2bE/kj9fulO
7mlwzmFpqjXmI43id8TC0spINgvu3347zAd2lnmYxcO/NwQ5LeL5HMflWSFM
2PE++YbKt2u6oghevdAjoub1XDNlfg427V9R0+CJInMaZU7Rntbvd5D/fR6l
cHom1g7yX7Myzys=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.25014377798838, 12.647946543326881`}, {-1, -1}], LineBox[CompressedData["
1:eJwV0w0slHEcB/DHS87IHHnr4gp5OzchcaLrqTV0I+Zl0TQvybzc8rIzyqlb
sxytcjZTaFnNSLLRMe+0G5FLIfNayevcaN7SdY76Ps/27Nlnv+33+/++z/NY
x6eF3tQkCMIHN/Ukvsjl8kA3ktCh4EwSlrVW3jFwQ90RtZRFEuTfKm4y3Js6
w0yA+THJZikw96tnsBGsfLpQFAfPvftZ0utEEtI1+71gOKS3azgLJkeNIzzh
g2bbQg84kFWmYwZr7hs0/XMkCWEUfWXDlSRYx0oEU3AOg8kYhFNSikxksIm5
bL0Kvrb5ZLoTPry+1HwHjt4WGfbDnLKsmkj4pME07Rss7stp9oV5S2wu1b+W
7ZbpCL+0p+c6YT7HxWfXEh7ONtSNhAUZpz8z4PxCgeIB5d87u9Zwfa2fTROs
ZKvG3Ki6c+XqOEx3mbl4GaYHhm/8gomJ0NpEeF4tLlbBrrNv9cRwuF+uchc+
EWHk9gZuGI/3X4CjR173f4JvpURW9MBqVem9dZik3TV8BPM9L53TRT5626K6
IDiWq+VqCTe5F3K14YQ/jdsOcP9Ria8U+1msfKhmwRvJK3lxcCUzatsW3ho8
S6PDO9leTBN4LWZnQOZAEsXffyj3ME+VOhorgjvPC6ZmYYZv1VwAvFhRnN9K
5d1W9vw4rNvBM5LA5f5p+jpU3dtiIIl6X37Xq9X2yCus5uACTHwULmuiLlQt
bzHhdIZ+qDk1r7SgRoPK211V5AUrhYHiiVPIryfudizcet9jtxN2aRfmPoYH
WnTWpTAv0/lFFxwyL67ogdMLUudW4ZxoG91ZeCt7NckQ+4UL7eja6M8N75az
4N6hev4JeJL+jPSGF1siXKnzjSpUfRyYbmrKTqD20R/iOVLfU2mbQQHseCOo
nQYHPKzPr4FdRGkHE5jH1ihPlMF6jRGb5bA6o2Nkkuq/pU+GUfttVvOWYRuJ
E6kNRwtkYwp4I+tQRyPySTqjtb9EWdXdGANLJQOKCdjPY8TKGObk1YW9h/k0
ZpPcDvtccch6BXM8r1aVwNr4LUVwfCIuO/I/oWlSwg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.341371117808144`, 12.718122537769101`}, {1, 0}], LineBox[CompressedData["
1:eJwd03tIU1EcB/CrU8tosZaUZurCba2Hj1So1PIqUUtXPsq1pswxTV0tJ/lI
cKRR+UrEVMp0iWnlMhMdW4WrrP1Rw0g0m1q6DIailktJTMyy7/XA4fLhx7nn
+/vB2S5Txp21JwgiBJv6EtJUrN0ksbp2kITaEHeuE659em+YDbcvlR2130MS
30Qa637Ys8pvSzy8n3OgOhhWaa8NPIBjan1UO6n6Q83tWdgUpkt0ouodEZa9
Pji/rpwxwCWJFLlIlwovdtVkVcEBUZ/mymGRW3EQH3ZaHolthvn1TXWLHJJw
yRm88QguqfkRroGZWePCRjiTDPiYAPufKZeUwsM1sfot8HzhAV4anDd3un+M
TRKDYcLvoXBHtynyGVxuE0o2wH3G0bhGeHp0Ln8MefnHPjio4a6XXod1MBkb
c6oVng/uUVVS/b0Wb+iBT21l5l6CC43bfizCErnFfB7OdFn/LJDKx/p+QQm/
9jTm58JtlTSvq3Ajo9T0At68TH//EGY4d/2yQ7/spEnhZ1iUWHU2HD6UNW5w
RT5ywPYtF24NtOyQwTyr4m0DfEtg3qmlHN9z/Clctnx50sEX8wla6DTACvXu
JCFc2+x6vB2+5nTxbhPMF5qbKuBzS9WOE3ChN1ckgSN2vb3u4Yf5jZ9wY8Ge
T3pNR2BeQ7FyCHmNnr/FUjhz9E9JCRzEcn4jh/2/3FrYB6eopwqSYT5dnzyN
ediiElyiYWlCkOA+m8qvGPGF827m6OXwiRaHGUdYZNXaDsIRbVPTQ8jDmKpL
9IaZ88HvNDDP48JJd3jt2NJ1Fazh/7Tjwtk21V+qv2FjSUYEvJ6puRNCnU/q
LMqg7ov/wPGh6n+7w1rghvzzjZRZnza+m4S/iuvdD1H9V1a+8kX+PTS3Nsmq
P1/JhvkTC2kV8KyqW6mHxXpGznv4eXpd4AxclCXmbqL6i7Q4u2JeAnMoVwYX
HtP+CYBls9LHOpgVPUMPhXv1J8U0f7hfIaXqG+9EsnfBptLfIZvhbD/2FwHc
JzXQp/B/r9o1VjlM0PQxT+B/KysrBXA69U455H/r1k+a
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.481877462230901`, 9.541371117808147}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.5, 10.700000000000001`}], 
          PointBox[{14.5, 14.3}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["18",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         LineBox[{{17.30000000000291, 13.500000000003638`}, {
          13.700000000004366`, 11.500000000003638`}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 11.749856222011616`}, {-1, 1}], LineBox[CompressedData["

1:eJwd0wtMU2cUB/ArQyGKrj4HtQuFGMXvu2AZRa5T8Dh8oAGpOLCCYmFOwE0s
qAgM67O1FXxSsAoqCEuYCmrEEo0o0S1WvGADPtrVzc4HVnQKilPXJnquN7m5
+eX77vlyzv/eoMy1ST96MQwTjbfwNP7B83x8ODCfLwoDFw9IjoehRYdmrbhB
gL8VG101TPDll5E1BEx3nOvkT2TA6HfIH+sI1Kf2esa3ol2VVx5oCPQVzyYB
FWjVw8CjegLqsuThr3PRnK9ndx3uX3Q+YlE8WrGwKLCTwMr+RvepMLR1R5pj
MIV6ieRl+FdC/fm6nfMoKEMT0s8OQRc+vZqyl4Le+99a4p4CjHOmT5Cdgk0z
+mLzOzSj8tRIWchu1jm/HIT74UHkuQwWPIxZm+6H7kvMLzex4Lv9VotRgm5I
SR1zlQWjceVCTia4VtvpYCHu+UBO9RzhfG7E7Mcs6HPmxfyfila66/77E/dv
+ZhWKvTTcNee28bCpeS7A3UatLoZ5JUsMGXyypsG4TzX7+Z0Frw3aF907UGH
2KvJ1yxIHyUlPClDWxL+Vt6hEO8s9K7Yjva/vzGrlIJ1nV/TP/nCPBNuVsVS
WFZcHMOlCfWrN4z3osBPvd/yc7TwfsyLpnYC++T0WaMY7TxSYsB89nW32566
cB5Wd28A5gP7V3fcNqGVpbk9JQRcu4oOnY5DS6dLC7QElEk+EUM8YdiPgcw/
hvn9FuP+xox2TQhzWgionDMDaDE6pGNN10cC2VGTzyYtQEu9ivqBgmvkjc1L
Q9Cq4DmFeszn+oWeoLHo96vc33VRsOTV2c+JhHqps/zELIi0UWtmiNGiTepT
aSwUfrF8uFiG1ivGLiln4X3evfqBRLQ/P6m7FdebYhXOArRlWesJGwumyF8P
62vRNf35iQ9x3lzvtQwena1ZH4751Oh47q83aFlmx3XMJ3tiifzoGOxX3aYz
HGTBsmLYtpMUvWXvh3sqzHuUe1PVVOF7emd0BLKg9u/hWyLQHBOcbKMgfWU0
jwtGi+xRq/ZQaEsub2wYJOyvGPf9XAoyzahvp3UL/dC8OG9cf14w7QcTWqno
HM0T4HYaAosWC/OrEL3FfDh1iCnTB62w5eTg/OM2D9bltYQC49u+1ecX/F8m
pThkWWhVSn/WNgJnQq0TZGI0ZKyVVOH6T7nWzm7su68y3HwN6xnU0UOxL8Zl
DXZ8IPAJxmF/uw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.788515046681937`, 15.139182060064893`}, {0, -1}], 
         CircleBox[{15.544262499999999`, 12.303987499999998`}, 
          1.7212037630863173`, {2.383003776731178, 5.081967545625389}], 
         PolygonBox[{{14.449161702384592`, 10.84683389206273}, {
          13.448460892323894`, 11.620524973014206`}, {14.112815702906918`, 
          12.06620724286033}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.43251255449539, 11.086390602173136`}, {1, 1}], 
         CircleBox[{14.9167375, 11.8830125}, 1.7212037630863173`, {
          5.524596430320971, 8.223560199215182}], 
         PolygonBox[{{16.011838297615405`, 13.340166107937268`}, {
          17.01253910767611, 12.566475026985792`}, {16.34818429709308, 
          12.120792757139668`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.02848744550461, 13.100609397826865`}, {-1, -1}], LineBox[CompressedData["
1:eJwd0H0s1HEcB/BfpqFVdGSjh52nO86du1HKyvklpHWWrqEl7WqkPF5cuupc
V1vczmZkZGjmwkUrtJQ55JlWK5lCSh4P7dgxxRXp/euzfffd64/v5/P+fB0u
JgtjTAiC2IdD3YToEsqdJP4XkyQYW9OLGuFaT5qSAX+dPzttwyYJaYG+yxsu
tI6KT4ZX23rWPOBy5W5BJ5zT6RNpCXtsqO23c0hCqZKtDDNIYowTKwiGQze1
lN+HTSJ4BcmwdGOynA9/OZXKvQtfFm72nXLBO333r9vwrIC3XwWbB2VK4mDx
XHPIAbipUqINgFfFgfWLziQxWtLvZg27jnqrWuCJaUurEeRpOOqVUgavVe0K
UMOPH6WfLIHphG11Emwlu25fA6vrxrWB8MHZQxODsI2Zlx2bes8KWaZh3haN
hZsLLApluUbCMiZ9wJPyaGdwNayz+B4VRvUvjM8zwvqypfNZVD+2ub8/9psy
+Tn+Ee5jVuyUwwbdVKkzlZdw/6yBlxYfRMjhsZq9zVqY9Y7jMwxbZQ63v4Kf
DS2c43qQxFDztt5imLSzPXwLVlgGFV6Bs5O+OTTAoYmcLmdYl9SWNQGbP8mO
6keeDMPxq6uwUjpxTwrHSYsSjbA0v/b3Hrg74UaPDq7lvM94g/0VbHZEB+Wq
m9I7sKDiiCAPFqWl0E/AjrnLg5GweMQ/hgUvjK+3O8KGtKxUB5h4LlyZ51D/
ocrlwaaxfLMOuFejsAuH/8rk2ko4JzaqOBuWDFyrK4EN0WZOn2DGUsAODUw3
lD10Qr7QuKcXuuFWYVNYKizhHks3wgoLhawZDtfoqnwx38pv7sUfWGx7pi8b
HnP5ke+K/+BXbHyYofbppNmRMHu4bsafi/0HpvR+cN/baPciuLeeX86E1abC
9RlYtEZ7vYp+fqUJrdY85BM2Zr2E6SNGOQvuWzk9GeNC/gMYvjFU
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.237909540841969`, 9.564358586412357}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.166, 10.699}], 
          PointBox[{14.295, 13.488}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["19",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         LineBox[{{17.30000000000291, 13.500000000003638`}, {
          13.700000000004366`, 11.500000000003638`}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 11.749856222011616`}, {-1, 1}], LineBox[CompressedData["

1:eJwd0wtMU2cUB/ArQyGKrj4HtQuFGMXvu2AZRa5T8Dh8oAGpOLCCYmFOwE0s
qAgM67O1FXxSsAoqCEuYCmrEEo0o0S1WvGADPtrVzc4HVnQKilPXJnquN7m5
+eX77vlyzv/eoMy1ST96MQwTjbfwNP7B83x8ODCfLwoDFw9IjoehRYdmrbhB
gL8VG101TPDll5E1BEx3nOvkT2TA6HfIH+sI1Kf2esa3ol2VVx5oCPQVzyYB
FWjVw8CjegLqsuThr3PRnK9ndx3uX3Q+YlE8WrGwKLCTwMr+RvepMLR1R5pj
MIV6ieRl+FdC/fm6nfMoKEMT0s8OQRc+vZqyl4Le+99a4p4CjHOmT5Cdgk0z
+mLzOzSj8tRIWchu1jm/HIT74UHkuQwWPIxZm+6H7kvMLzex4Lv9VotRgm5I
SR1zlQWjceVCTia4VtvpYCHu+UBO9RzhfG7E7Mcs6HPmxfyfila66/77E/dv
+ZhWKvTTcNee28bCpeS7A3UatLoZ5JUsMGXyypsG4TzX7+Z0Frw3aF907UGH
2KvJ1yxIHyUlPClDWxL+Vt6hEO8s9K7Yjva/vzGrlIJ1nV/TP/nCPBNuVsVS
WFZcHMOlCfWrN4z3osBPvd/yc7TwfsyLpnYC++T0WaMY7TxSYsB89nW32566
cB5Wd28A5gP7V3fcNqGVpbk9JQRcu4oOnY5DS6dLC7QElEk+EUM8YdiPgcw/
hvn9FuP+xox2TQhzWgionDMDaDE6pGNN10cC2VGTzyYtQEu9ivqBgmvkjc1L
Q9Cq4DmFeszn+oWeoLHo96vc33VRsOTV2c+JhHqps/zELIi0UWtmiNGiTepT
aSwUfrF8uFiG1ivGLiln4X3evfqBRLQ/P6m7FdebYhXOArRlWesJGwumyF8P
62vRNf35iQ9x3lzvtQwena1ZH4751Oh47q83aFlmx3XMJ3tiifzoGOxX3aYz
HGTBsmLYtpMUvWXvh3sqzHuUe1PVVOF7emd0BLKg9u/hWyLQHBOcbKMgfWU0
jwtGi+xRq/ZQaEsub2wYJOyvGPf9XAoyzahvp3UL/dC8OG9cf14w7QcTWqno
HM0T4HYaAosWC/OrEL3FfDh1iCnTB62w5eTg/OM2D9bltYQC49u+1ecX/F8m
pThkWWhVSn/WNgJnQq0TZGI0ZKyVVOH6T7nWzm7su68y3HwN6xnU0UOxL8Zl
DXZ8IPAJxmF/uw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.788515046681937`, 15.139182060064893`}, {0, -1}], 
         CircleBox[{15.544262499999999`, 12.303987499999998`}, 
          1.7212037630863173`, {2.383003776731178, 5.081967545625389}], 
         PolygonBox[{{14.449161702384592`, 10.84683389206273}, {
          13.448460892323894`, 11.620524973014206`}, {14.112815702906918`, 
          12.06620724286033}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.43251255449539, 11.086390602173136`}, {1, 1}], 
         CircleBox[{14.9167375, 11.8830125}, 1.7212037630863173`, {
          5.524596430320971, 8.223560199215182}], 
         PolygonBox[{{16.011838297615405`, 13.340166107937268`}, {
          17.01253910767611, 12.566475026985792`}, {16.34818429709308, 
          12.120792757139668`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.02848744550461, 13.100609397826865`}, {-1, -1}], LineBox[CompressedData["
1:eJwd0H0s1HEcB/BfpqFVdGSjh52nO86du1HKyvklpHWWrqEl7WqkPF5cuupc
V1vczmZkZGjmwkUrtJQ55JlWK5lCSh4P7dgxxRXp/euzfffd64/v5/P+fB0u
JgtjTAiC2IdD3YToEsqdJP4XkyQYW9OLGuFaT5qSAX+dPzttwyYJaYG+yxsu
tI6KT4ZX23rWPOBy5W5BJ5zT6RNpCXtsqO23c0hCqZKtDDNIYowTKwiGQze1
lN+HTSJ4BcmwdGOynA9/OZXKvQtfFm72nXLBO333r9vwrIC3XwWbB2VK4mDx
XHPIAbipUqINgFfFgfWLziQxWtLvZg27jnqrWuCJaUurEeRpOOqVUgavVe0K
UMOPH6WfLIHphG11Emwlu25fA6vrxrWB8MHZQxODsI2Zlx2bes8KWaZh3haN
hZsLLApluUbCMiZ9wJPyaGdwNayz+B4VRvUvjM8zwvqypfNZVD+2ub8/9psy
+Tn+Ee5jVuyUwwbdVKkzlZdw/6yBlxYfRMjhsZq9zVqY9Y7jMwxbZQ63v4Kf
DS2c43qQxFDztt5imLSzPXwLVlgGFV6Bs5O+OTTAoYmcLmdYl9SWNQGbP8mO
6keeDMPxq6uwUjpxTwrHSYsSjbA0v/b3Hrg74UaPDq7lvM94g/0VbHZEB+Wq
m9I7sKDiiCAPFqWl0E/AjrnLg5GweMQ/hgUvjK+3O8KGtKxUB5h4LlyZ51D/
ocrlwaaxfLMOuFejsAuH/8rk2ko4JzaqOBuWDFyrK4EN0WZOn2DGUsAODUw3
lD10Qr7QuKcXuuFWYVNYKizhHks3wgoLhawZDtfoqnwx38pv7sUfWGx7pi8b
HnP5ke+K/+BXbHyYofbppNmRMHu4bsafi/0HpvR+cN/baPciuLeeX86E1abC
9RlYtEZ7vYp+fqUJrdY85BM2Zr2E6SNGOQvuWzk9GeNC/gMYvjFU
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.237909540841969`, 9.564358586412357}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.166, 10.699}], 
          PointBox[{14.295, 13.488}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["20",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], LineBox[CompressedData["

1:eJwt1AtMU1cYB/ALSnlKK3NICwxkjPegMNBGGLkyiFMqIqAWWFxhUErptAM3
3lgsAwY6iggYBAaTieOxdaRqUch0MKiMl7DMEljSdbixKQEUM8QH+59lN7m5
+eXc853v+87J2ZF8IibVmKKoWLzkS8lEeLxpyozAnab0hooSBTzJHuQcglWX
1JzvYG7ZdZ8GmJdQk7YIaySsgHuwUPSiN8gH/2WkPtjsQVONmunGUljj0mKz
A+Yzk5hzMC1jyr3g6LLGIf83aYrqD3mbjOtL3bIUsJkmuoQBh4dVmI7By9Kt
XXOIrwvdv27hS1MX/G4ZtcPR2YpTu2Fda0RNOmxXUHAtHhabcxY94PmgYG0a
rNxwmvzLDXbhPkmB5YVe8yqYl86ujoF57o7S03CJLGkxABa45lPJsFV1/Fvm
MKttIi4GDpm8zp1BPnZOHZaHYcqmmd0O82bXL0rg5Y/tw7NhQWjxvnOw2lLx
2wFY+eowdwTWDNre8IVlA/dTmchvdKdwyQHmnnR2PQaHn3W9w4HXEvmWPfD5
nkClJ1nP3CmN9KPPPjd3H4nPn8k/AhesV32ZC7M++XC6CV74JchZA8sFu0p1
8Nymzx2Nkb/dgba8TZ40Je3iW8TB9B999Q7woOPdn7+Gteqjf7rAQpMCnw1i
avgrW/iCq6n2oB9N3VIYhT5FPLHxB531sFxbEjAGh2S3i6dhbbpktRbmZq57
U1x4NrLzKBy3/b0ka9jZu3zFFnbQajdegTWCkM90qI9KPxO6FdZ1X7VuhXve
Ef/KgJftorpPkn7cvsl9iPjCe3mzR+CWLa7ut2GutdBlPyztfK1YSWwhXz0I
K8W7Co/BOWq3OjE5D6x3ZX6w6gWLXQM7fGQoMoVZK3UZY7DZwEj836h3oSYp
bxvya9OPG83C3Kmy5BTYKmeijvhKrUGlIfUIOEGLpJ/9yg4r9OehIfJ3JuKV
q7Y0JMLLnvrqMFiX8si+lbg230cB082Fohl4rvLuzXF44UlcmZEX5jOon1ik
3r1VNdth+twbY8Gw8vLaEAeOawjcLYJ5FdVKSziTWblyFrYLLqx8gHiMjWsv
VbBsaCqlHw4Ux/SOwQt3nn1bAo9kMfQGWG6U3hVB9nPPy64l2OxiLGUC69sL
DI9h7pVV6TDqO3NcO07GVWGM4mo4r85v7zycIy2MSoNZ/lkNU2T+8E4PPvzN
85z8PrJ/CeMZYfCNL5ruXyLjvG2JkXCUieR5BRwdX21K5hed6l3IIvWwJYfO
k3Gv8qtCst5h1dNJEl9/2TcWbjF5NsRGfnRe1gAfZtUv9Upg9ePME1HkfKlT
o76Hf0iM7BCQepupH23QH7fjrwszyPr2VTIhbNMz/+hTmGZFGLfBBWuR4e2k
X16rEzNwEc9V+1+/EkaYG7DZ+6eb/iH11A6b2OBeE9Gibhd/nP+Nzc7Ek6Oj
o3z//+9Hb/pfUWnJsA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.98789659515192, 11.13794612872654}, \
{-1, 1}], LineBox[{{14.294999999987894`, 13.488000000001193`}, {
          17.29999999999343, 13.5}}], 
         PolygonBox[{{16.397495216017305`, 13.496395987551484`}, {
          15.199102109017016`, 13.091607201770312`}, {15.195907458948373`, 
          13.89160082312672}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.791310365492, 14.439187641378037`}, \
{0, -1}], 
         CircleBox[{15.544262499999999`, 12.303987499999998`}, 
          1.7212037630863173`, {2.383003776731178, 5.081967545625389}], 
         PolygonBox[{{13.780638297615406`, 11.843366107937268`}, {
          14.11698429709308, 10.623992757139668`}, {14.781339107676104`, 
          11.069675026985792`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.43251255449539, 11.086390602173136`}, {1, 1}], LineBox[CompressedData["
1:eJwVVHtUTXkUDiMJrXKTSh41uuece+4955dM3R74FtUkeiHUkB5zw0QXFVZS
hESmB3mUoodXpDQVqVzcynjcUlmTklJSGIq8382ZP/baa6/9rf34vr22eZBy
gWKolpbWcsH+9/fz1s3QPSxFQE+DRdtnCfwvuw/mXJPisxtvkTGTxSM6NdDy
hxQ1mk7jsZtZVG1TdfbPk0HeyfySns3i2dwdso48GejwCZmNJSwc5kZlTfsh
w13XtyWHi1gkzDKeWb+AQ0IJlbUljYWb5M2ZlAwO9t2xVUcVLNb4J6QpGzj4
eZfV9VMsvHcsbhv9hsMIPfvjM7okMPaNXFk+yMFlfeFS88MSGM6b1v3pK4fa
7Q60u4cEazLqPOr/5aDTXJLqpCuB+PQircdCvWMFBnsi6xisLRbLbv3FoVE2
6JyczmBA/PVdWBqH0Eyd191KBkmPeqyLN3OwG6nHnPNm8DZzqrY8kMMa/fZx
+mCQs/uw7p8eHJpzVTI9RwZ/y1UKJTjc1W/eZu3CQP79SONCGw5xOk1fFf4M
Kj3ih7/hORSdvxN7fQeDXJcjty/LOIxp/yP1QCkDK9MzrJ4Vh4rqkOGr+xlc
6xDlT3Pk0NAzeZgZJ0HFu+aup95Cv9rhmS1KCUwHa9kbERxmaSoqJxdJUPlb
05nwUxwOxMx06uqVINU+PL2zl0P7dE3BHQMWom9m6ivWPMxuVu55LGHRPnBz
6YoEHs68W5gnx0Ltb9nGdvBwaLg94GjKIlt21famEcGbXhP3iFcSBOf1hObZ
Evj8YzV6S7EELikD7Xc8CNxNc6aLQiQImc82ZfsRiFRDRtkaStDSZVDaK8Tf
IhTDWtUMovlVx055ERiFylaoNzHwMZRtaJ1BkOHAHG+1YcBc23h6cCoRdDsr
HxjCIMxO3KXWIbjYGCwVtdLYW/DziMD7PNye9y6ZXEmje/1g1YRsHn2G76zi
T9FYpV9xdYuSh+d2xei+dBqJTWFjAlx5DC6Olx9Ko3E9l+s5J+Wxq7+VjTpE
44bf3l0uE3n4k5gPomwaY/oVcgdTAa86OeJjMQ2vwWi37xY8bB29Wrw0NKiV
k5hqGx4TDTMnRvXTeLjMLf7hIh5FoQH3LxgxMD8UYP44mkerplpr+GwGsmHp
YacKeQTOCQ+NXCPs53meefycR7jsg8YmlYHHlyPiFoFfUXL0JYtzDMattHYe
4kigF9OV6FnGINbP/3JOEAEbofmkEfKapl9F0gSCEFFUkHIfA6NRTtFJ+QQN
T3d/37qEgbtRRPWZaoIqWUHKSH0GirLG5I57BKvNrG9VCHxFn/Bad7SFINTB
rlw3iEaWyCD8eKOgTyC/UncUjfr3c2TfrxJINxVN1VFRSL0Qx9WfJHjr/JFb
u43Cl4isK5W7CI7dK1423ofCSXVz7E5hPvflrtnOjhSmpg31zZlJ4J24nho6
nUKI9dpHJmbCvawKqn0gp1De2z175GceCQNe00tdKFz2Nbk0W80jIMahMcmX
Qt+U9Hd9B3nUlMtrp4RR6Nf2yb0ZyUPb1bjQOo6Cu2XGCvtgQb90xcLKNAqV
Dj5nrVbwSInRH3viJAVtxYXIr6t5qNTKJ7+XUciz030m3cEjaVh/obKGwuLx
d8zz83ncGlBFnWiioN8WtLWijUeZ1mmvlx0UakssxQdHCfzUXPEzekrB3qct
I0dK4DL5fsHQZxRKs+bmLPMkyK3hXMYL+Hrfcq5mPUHpi6T2V7UUgruKXkSm
ElxosJvEZVEwo+n9t88RjLOZ8JNTMIU6H9XrCSqCWfvE0hGmFFo27nrC3yCw
kDmJHtSKoWyusy1SE5js70t2XStGydiinXElQr+XjgkmxmKMU4AdSCf4Ybdz
iXa9JXy86jrXbSY4/jy/MPSgJVISb8jJAoJl/u8vzt9giUbha8dICFrqPy7s
VljiPwB5daY=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.02848744550461, 13.100609397826865`}, {-1, -1}], 
         LineBox[{{16.16599999999744, 10.699000000000524`}, {
          13.69999999999709, 11.5}}], 
         PolygonBox[{{15.503651043623371`, 10.914142544224525`}, {
          14.485920593560278`, 11.66529148485772}, {14.238777319192979`, 
          10.904423426693226`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.454159905913356`, 10.230118137306292`}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.166, 10.699}], 
          PointBox[{14.295, 13.488}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["21",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         LineBox[{{17.30000000000291, 13.500000000003638`}, {
          13.700000000004366`, 11.500000000003638`}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 11.749856222011616`}, {-1, 1}], LineBox[CompressedData["

1:eJwd0wtMU2cUB/ArQyGKrj4HtQuFGMXvu2AZRa5T8Dh8oAGpOLCCYmFOwE0s
qAgM67O1FXxSsAoqCEuYCmrEEo0o0S1WvGADPtrVzc4HVnQKilPXJnquN7m5
+eX77vlyzv/eoMy1ST96MQwTjbfwNP7B83x8ODCfLwoDFw9IjoehRYdmrbhB
gL8VG101TPDll5E1BEx3nOvkT2TA6HfIH+sI1Kf2esa3ol2VVx5oCPQVzyYB
FWjVw8CjegLqsuThr3PRnK9ndx3uX3Q+YlE8WrGwKLCTwMr+RvepMLR1R5pj
MIV6ieRl+FdC/fm6nfMoKEMT0s8OQRc+vZqyl4Le+99a4p4CjHOmT5Cdgk0z
+mLzOzSj8tRIWchu1jm/HIT74UHkuQwWPIxZm+6H7kvMLzex4Lv9VotRgm5I
SR1zlQWjceVCTia4VtvpYCHu+UBO9RzhfG7E7Mcs6HPmxfyfila66/77E/dv
+ZhWKvTTcNee28bCpeS7A3UatLoZ5JUsMGXyypsG4TzX7+Z0Frw3aF907UGH
2KvJ1yxIHyUlPClDWxL+Vt6hEO8s9K7Yjva/vzGrlIJ1nV/TP/nCPBNuVsVS
WFZcHMOlCfWrN4z3osBPvd/yc7TwfsyLpnYC++T0WaMY7TxSYsB89nW32566
cB5Wd28A5gP7V3fcNqGVpbk9JQRcu4oOnY5DS6dLC7QElEk+EUM8YdiPgcw/
hvn9FuP+xox2TQhzWgionDMDaDE6pGNN10cC2VGTzyYtQEu9ivqBgmvkjc1L
Q9Cq4DmFeszn+oWeoLHo96vc33VRsOTV2c+JhHqps/zELIi0UWtmiNGiTepT
aSwUfrF8uFiG1ivGLiln4X3evfqBRLQ/P6m7FdebYhXOArRlWesJGwumyF8P
62vRNf35iQ9x3lzvtQwena1ZH4751Oh47q83aFlmx3XMJ3tiifzoGOxX3aYz
HGTBsmLYtpMUvWXvh3sqzHuUe1PVVOF7emd0BLKg9u/hWyLQHBOcbKMgfWU0
jwtGi+xRq/ZQaEsub2wYJOyvGPf9XAoyzahvp3UL/dC8OG9cf14w7QcTWqno
HM0T4HYaAosWC/OrEL3FfDh1iCnTB62w5eTg/OM2D9bltYQC49u+1ecX/F8m
pThkWWhVSn/WNgJnQq0TZGI0ZKyVVOH6T7nWzm7su68y3HwN6xnU0UOxL8Zl
DXZ8IPAJxmF/uw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.788515046681937`, 15.139182060064893`}, {0, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{15.544262499999999`, 12.303987499999998`}, 
           1.7212037630863173`, {2.383003776731178, 5.081967545625389}]}, 
         PolygonBox[{{14.449161702384592`, 10.84683389206273}, {
          13.448460892323894`, 11.620524973014206`}, {14.112815702906918`, 
          12.06620724286033}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.43251255449539, 11.086390602173136`}, {1, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{14.9167375, 11.8830125}, 1.7212037630863173`, {
           5.524596430320971, 8.223560199215182}]}, 
         PolygonBox[{{16.011838297615405`, 13.340166107937268`}, {
          17.01253910767611, 12.566475026985792`}, {16.34818429709308, 
          12.120792757139668`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.02848744550461, 13.100609397826865`}, {-1, -1}], LineBox[CompressedData["
1:eJwd0H0s1HEcB/BfpqFVdGSjh52nO86du1HKyvklpHWWrqEl7WqkPF5cuupc
V1vczmZkZGjmwkUrtJQ55JlWK5lCSh4P7dgxxRXp/euzfffd64/v5/P+fB0u
JgtjTAiC2IdD3YToEsqdJP4XkyQYW9OLGuFaT5qSAX+dPzttwyYJaYG+yxsu
tI6KT4ZX23rWPOBy5W5BJ5zT6RNpCXtsqO23c0hCqZKtDDNIYowTKwiGQze1
lN+HTSJ4BcmwdGOynA9/OZXKvQtfFm72nXLBO333r9vwrIC3XwWbB2VK4mDx
XHPIAbipUqINgFfFgfWLziQxWtLvZg27jnqrWuCJaUurEeRpOOqVUgavVe0K
UMOPH6WfLIHphG11Emwlu25fA6vrxrWB8MHZQxODsI2Zlx2bes8KWaZh3haN
hZsLLApluUbCMiZ9wJPyaGdwNayz+B4VRvUvjM8zwvqypfNZVD+2ub8/9psy
+Tn+Ee5jVuyUwwbdVKkzlZdw/6yBlxYfRMjhsZq9zVqY9Y7jMwxbZQ63v4Kf
DS2c43qQxFDztt5imLSzPXwLVlgGFV6Bs5O+OTTAoYmcLmdYl9SWNQGbP8mO
6keeDMPxq6uwUjpxTwrHSYsSjbA0v/b3Hrg74UaPDq7lvM94g/0VbHZEB+Wq
m9I7sKDiiCAPFqWl0E/AjrnLg5GweMQ/hgUvjK+3O8KGtKxUB5h4LlyZ51D/
ocrlwaaxfLMOuFejsAuH/8rk2ko4JzaqOBuWDFyrK4EN0WZOn2DGUsAODUw3
lD10Qr7QuKcXuuFWYVNYKizhHks3wgoLhawZDtfoqnwx38pv7sUfWGx7pi8b
HnP5ke+K/+BXbHyYofbppNmRMHu4bsafi/0HpvR+cN/baPciuLeeX86E1abC
9RlYtEZ7vYp+fqUJrdY85BM2Zr2E6SNGOQvuWzk9GeNC/gMYvjFU
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.237909540841969`, 9.564358586412357}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.166, 10.699}], 
          PointBox[{14.295, 13.488}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["22",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 13.499856222011616`}, {-1, 1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673122`, 6.7498562220116165`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.}, {13.69999999999527, 11.5}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          12.019762806798532`, 10.108951330844206`}, {11.631248461855627`, 
          10.808277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.202053456673124`, 11.500143777988384`}, {1, -1}], 
         LineBox[{{17.30000000000291, 13.500000000003638`}, {
          13.700000000004366`, 11.500000000003638`}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{15.647946543326878`, 11.749856222011616`}, {-1, 1}], LineBox[CompressedData["

1:eJwd0wtMU2cUB/ArQyGKrj4HtQuFGMXvu2AZRa5T8Dh8oAGpOLCCYmFOwE0s
qAgM67O1FXxSsAoqCEuYCmrEEo0o0S1WvGADPtrVzc4HVnQKilPXJnquN7m5
+eX77vlyzv/eoMy1ST96MQwTjbfwNP7B83x8ODCfLwoDFw9IjoehRYdmrbhB
gL8VG101TPDll5E1BEx3nOvkT2TA6HfIH+sI1Kf2esa3ol2VVx5oCPQVzyYB
FWjVw8CjegLqsuThr3PRnK9ndx3uX3Q+YlE8WrGwKLCTwMr+RvepMLR1R5pj
MIV6ieRl+FdC/fm6nfMoKEMT0s8OQRc+vZqyl4Le+99a4p4CjHOmT5Cdgk0z
+mLzOzSj8tRIWchu1jm/HIT74UHkuQwWPIxZm+6H7kvMLzex4Lv9VotRgm5I
SR1zlQWjceVCTia4VtvpYCHu+UBO9RzhfG7E7Mcs6HPmxfyfila66/77E/dv
+ZhWKvTTcNee28bCpeS7A3UatLoZ5JUsMGXyypsG4TzX7+Z0Frw3aF907UGH
2KvJ1yxIHyUlPClDWxL+Vt6hEO8s9K7Yjva/vzGrlIJ1nV/TP/nCPBNuVsVS
WFZcHMOlCfWrN4z3osBPvd/yc7TwfsyLpnYC++T0WaMY7TxSYsB89nW32566
cB5Wd28A5gP7V3fcNqGVpbk9JQRcu4oOnY5DS6dLC7QElEk+EUM8YdiPgcw/
hvn9FuP+xox2TQhzWgionDMDaDE6pGNN10cC2VGTzyYtQEu9ivqBgmvkjc1L
Q9Cq4DmFeszn+oWeoLHo96vc33VRsOTV2c+JhHqps/zELIi0UWtmiNGiTepT
aSwUfrF8uFiG1ivGLiln4X3evfqBRLQ/P6m7FdebYhXOArRlWesJGwumyF8P
62vRNf35iQ9x3lzvtQwena1ZH4751Oh47q83aFlmx3XMJ3tiifzoGOxX3aYz
HGTBsmLYtpMUvWXvh3sqzHuUe1PVVOF7emd0BLKg9u/hWyLQHBOcbKMgfWU0
jwtGi+xRq/ZQaEsub2wYJOyvGPf9XAoyzahvp3UL/dC8OG9cf14w7QcTWqno
HM0T4HYaAosWC/OrEL3FfDh1iCnTB62w5eTg/OM2D9bltYQC49u+1ecX/F8m
pThkWWhVSn/WNgJnQq0TZGI0ZKyVVOH6T7nWzm7su68y3HwN6xnU0UOxL8Zl
DXZ8IPAJxmF/uw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.788515046681937`, 15.139182060064893`}, {0, -1}], LineBox[CompressedData["

1:eJwVVHlUzQkUjjTSlDgMWgYtv+W931I5ljKlT4r2R1RiOqE0WabF1jIkSdJq
bHmVSlmqpzyyJLSd0JPSQvEqTMk7Q0Vkrcb8/HHPPfd859zl++69RhtCPDaO
VVNT8xXsh2/LD7XRSmfh19tk3PFVjDUt0srzy1h0erZWhvWL0eiYHEIOM5De
n1nncV8M91nnlEQZg9+i7GeV/S3GlNo6Sv0Ag/R7SQNnlomRYZ1UXRjIwMi4
49XO9yIohg2CCvwYzDgY5/9SKsIzmyJJUwiDYSrP+bqDCLbSNu3txxiYRBdv
ixqhcZm852OoYHCoJ3X14RoaynnXWvdqsthfZyVtlNLQLTk64O7G4sRbqfn5
fTTUbUOHa4+wGDw2qLgZTqN/X8fLrkcsGt5qGRZE0PhjYm2p2yQOsaVGZEsM
jdkta3X97DgE10wY0Uuh8fiioVNHEIciufKeeyaNbXG7bmrHcogaXH3Ep4jG
7eLm+LZkDgXzygL0y2kwtcnNDQkcSvX9TdTraSR1N6fa7uJgsVVel9glxHUp
1vpeHHYMp106NUijKk0n6o6Iw5N5KWfNfhKhauySgf/es5DXhNtp6YlQHOu6
7VEpix6XwEs9pAjm1UZ77wSzUNHekluMCGYBkRbfSRZvvGLaIkxEiPv+5urh
FwwyXkzTP68lgneC0aMtuQwcgpuXrXhOo2s0u01nKwMdGz2Pa6dpLKoZ56Jw
ZGDtIdW9sopGTO6B5c0LGTREpFWUjlIoid4/2cWeQVP68WL1MxS6bi0quhbA
YE+84asrEgqurdmWb04yCHqASGhSeLEkXUvewaAl132qpJVE5NzNFbVCfzoY
2TNYSkJiUHVOYwcL8XzV6+AiEp1DQz3eFSxMCdfKzCsCnhE5nhzLoTa+pfBg
IwnnVyk16204tId6/uz0gYT3yru6ZAgHP9W4XTd+pZCV3FnEpHM45BiUo+NE
YUWspqbDVQ4GuayPfRgF+nXhKy0FB7WIiOw/j1LYMlz15UEzB8uPxiH3iilc
Ts2SVTdyUO53jimqpHB96ReX3ysF/e4MONsrKPRPKzMpyOdg5Wu3iK6nMOnp
1rDk3Rz+neO/ufkOhRPh1I4+Fw4uOY/sLW9QGLBWF5+fymHg/j8zcwS+snYp
g2OVLGK7A5xl8RTITVbdmnnC/unNXNTgS6Gx6oGrZyiLnIic9pc0hb2JXkPl
ziyMnTKnj6hIKAh1xcL5LKxt/bf3ZpGQLW9KvivE5Ke7Ww66kOirW7Iz1F3Q
X0rZyb4SuJy31qd6N4uXz05sXlJCYEFqQXW4wO/ShhKRcyiBa4ljvDp1OIz2
jMbtcyCQ+dn3SYE/hw9x0ZohFgQC+ieEOJdxMOzT7u+YS0BVUppOafBYs3dE
8d6NwK2UVQnHHHicnZc4pyuCgO/rdRp24Ty+qnRuZssJVAxJJLMzeGwYqFH2
viVwXWtKwuOLPIIaC7bnWZBY9jBa41s5j+h368M8w0i05aUGulfwYNp9e8Uy
Erczcplpt3gc+mn000gHCfSl+F25yuOq7FPU6jGCXvKg+pMyHpvy339M1KPQ
vj6N/OuUUG+ylpPchMIB17FMUyKPhvw360pmUXg9YfGo13YearJytykTKSTU
MNZSbx6N06vilW9JaFuqjbO1EvqLV5ib3SXhYddiThrwyBx/vd/xiIBLfLqt
xvCweDa1N9KTxMbsITfdd8L9zklxD5tMYt3hSwpJn8Cfnf7MU3UE6MAmz+Gv
HOacS22PiSagnZSddVLI93zDlMKHCwiwlXMTwlfw+N7kWG78zRQDhRM/WEl5
yNVOP7WrN8XC3AtHH/cL8215XEPKTZGak68udjTDpMXFyj6ZKcqDgpbr5Zrh
e2v2YqLSFNKq4w+7+82wQ6XyvaAyhc7Jvhmlv5jDZpN0lakRgR9fO1psjvbG
zyt7NhL4H4dDj84=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.43251255449539, 11.086390602173136`}, {1, 1}], LineBox[CompressedData["
1:eJwVVHtUTXkUDiMJrXKTSh41uuece+4955dM3R74FtUkeiHUkB5zw0QXFVZS
hESmB3mUoodXpDQVqVzcynjcUlmTklJSGIq8382ZP/baa6/9rf34vr22eZBy
gWKolpbWcsH+9/fz1s3QPSxFQE+DRdtnCfwvuw/mXJPisxtvkTGTxSM6NdDy
hxQ1mk7jsZtZVG1TdfbPk0HeyfySns3i2dwdso48GejwCZmNJSwc5kZlTfsh
w13XtyWHi1gkzDKeWb+AQ0IJlbUljYWb5M2ZlAwO9t2xVUcVLNb4J6QpGzj4
eZfV9VMsvHcsbhv9hsMIPfvjM7okMPaNXFk+yMFlfeFS88MSGM6b1v3pK4fa
7Q60u4cEazLqPOr/5aDTXJLqpCuB+PQircdCvWMFBnsi6xisLRbLbv3FoVE2
6JyczmBA/PVdWBqH0Eyd191KBkmPeqyLN3OwG6nHnPNm8DZzqrY8kMMa/fZx
+mCQs/uw7p8eHJpzVTI9RwZ/y1UKJTjc1W/eZu3CQP79SONCGw5xOk1fFf4M
Kj3ih7/hORSdvxN7fQeDXJcjty/LOIxp/yP1QCkDK9MzrJ4Vh4rqkOGr+xlc
6xDlT3Pk0NAzeZgZJ0HFu+aup95Cv9rhmS1KCUwHa9kbERxmaSoqJxdJUPlb
05nwUxwOxMx06uqVINU+PL2zl0P7dE3BHQMWom9m6ivWPMxuVu55LGHRPnBz
6YoEHs68W5gnx0Ltb9nGdvBwaLg94GjKIlt21famEcGbXhP3iFcSBOf1hObZ
Evj8YzV6S7EELikD7Xc8CNxNc6aLQiQImc82ZfsRiFRDRtkaStDSZVDaK8Tf
IhTDWtUMovlVx055ERiFylaoNzHwMZRtaJ1BkOHAHG+1YcBc23h6cCoRdDsr
HxjCIMxO3KXWIbjYGCwVtdLYW/DziMD7PNye9y6ZXEmje/1g1YRsHn2G76zi
T9FYpV9xdYuSh+d2xei+dBqJTWFjAlx5DC6Olx9Ko3E9l+s5J+Wxq7+VjTpE
44bf3l0uE3n4k5gPomwaY/oVcgdTAa86OeJjMQ2vwWi37xY8bB29Wrw0NKiV
k5hqGx4TDTMnRvXTeLjMLf7hIh5FoQH3LxgxMD8UYP44mkerplpr+GwGsmHp
YacKeQTOCQ+NXCPs53meefycR7jsg8YmlYHHlyPiFoFfUXL0JYtzDMattHYe
4kigF9OV6FnGINbP/3JOEAEbofmkEfKapl9F0gSCEFFUkHIfA6NRTtFJ+QQN
T3d/37qEgbtRRPWZaoIqWUHKSH0GirLG5I57BKvNrG9VCHxFn/Bad7SFINTB
rlw3iEaWyCD8eKOgTyC/UncUjfr3c2TfrxJINxVN1VFRSL0Qx9WfJHjr/JFb
u43Cl4isK5W7CI7dK1423ofCSXVz7E5hPvflrtnOjhSmpg31zZlJ4J24nho6
nUKI9dpHJmbCvawKqn0gp1De2z175GceCQNe00tdKFz2Nbk0W80jIMahMcmX
Qt+U9Hd9B3nUlMtrp4RR6Nf2yb0ZyUPb1bjQOo6Cu2XGCvtgQb90xcLKNAqV
Dj5nrVbwSInRH3viJAVtxYXIr6t5qNTKJ7+XUciz030m3cEjaVh/obKGwuLx
d8zz83ncGlBFnWiioN8WtLWijUeZ1mmvlx0UakssxQdHCfzUXPEzekrB3qct
I0dK4DL5fsHQZxRKs+bmLPMkyK3hXMYL+Hrfcq5mPUHpi6T2V7UUgruKXkSm
ElxosJvEZVEwo+n9t88RjLOZ8JNTMIU6H9XrCSqCWfvE0hGmFFo27nrC3yCw
kDmJHtSKoWyusy1SE5js70t2XStGydiinXElQr+XjgkmxmKMU4AdSCf4Ybdz
iXa9JXy86jrXbSY4/jy/MPSgJVISb8jJAoJl/u8vzt9giUbha8dICFrqPy7s
VljiPwB5daY=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.02848744550461, 13.100609397826865`}, {-1, -1}], LineBox[CompressedData["
1:eJwd0H0s1HEcB/BfpqFVdGSjh52nO86du1HKyvklpHWWrqEl7WqkPF5cuupc
V1vczmZkZGjmwkUrtJQ55JlWK5lCSh4P7dgxxRXp/euzfffd64/v5/P+fB0u
JgtjTAiC2IdD3YToEsqdJP4XkyQYW9OLGuFaT5qSAX+dPzttwyYJaYG+yxsu
tI6KT4ZX23rWPOBy5W5BJ5zT6RNpCXtsqO23c0hCqZKtDDNIYowTKwiGQze1
lN+HTSJ4BcmwdGOynA9/OZXKvQtfFm72nXLBO333r9vwrIC3XwWbB2VK4mDx
XHPIAbipUqINgFfFgfWLziQxWtLvZg27jnqrWuCJaUurEeRpOOqVUgavVe0K
UMOPH6WfLIHphG11Emwlu25fA6vrxrWB8MHZQxODsI2Zlx2bes8KWaZh3haN
hZsLLApluUbCMiZ9wJPyaGdwNayz+B4VRvUvjM8zwvqypfNZVD+2ub8/9psy
+Tn+Ee5jVuyUwwbdVKkzlZdw/6yBlxYfRMjhsZq9zVqY9Y7jMwxbZQ63v4Kf
DS2c43qQxFDztt5imLSzPXwLVlgGFV6Bs5O+OTTAoYmcLmdYl9SWNQGbP8mO
6keeDMPxq6uwUjpxTwrHSYsSjbA0v/b3Hrg74UaPDq7lvM94g/0VbHZEB+Wq
m9I7sKDiiCAPFqWl0E/AjrnLg5GweMQ/hgUvjK+3O8KGtKxUB5h4LlyZ51D/
ocrlwaaxfLMOuFejsAuH/8rk2ko4JzaqOBuWDFyrK4EN0WZOn2DGUsAODUw3
lD10Qr7QuKcXuuFWYVNYKizhHks3wgoLhawZDtfoqnwx38pv7sUfWGx7pi8b
HnP5ke+K/+BXbHyYofbppNmRMHu4bsafi/0HpvR+cN/baPciuLeeX86E1abC
9RlYtEZ7vYp+fqUJrdY85BM2Zr2E6SNGOQvuWzk9GeNC/gMYvjFU
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.237909540841969`, 9.564358586412357}, {0, 1}], 
         {PointSize[0.04], PointBox[{16.166, 10.699}], 
          PointBox[{14.295, 13.488}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{13.700000000000001`, 11.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["23",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{5.500000000001819, 12.200000000002547`}, {
          3.3000000000029104`, 10.000000000003638`}}], 
         PolygonBox[{{4.824264068711929, 11.524264068711929`}, {
          3.692893218813453, 10.958578643762692`}, {4.258578643762691, 
          10.392893218813455`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.908784489160852, 11.591215510839149`}, {1, -1}], 
         LineBox[{{5.499999999996362, 12.200000000000728`}, {
          7.699999999995271, 10.000000000001819`}}], 
         PolygonBox[{{7.024264068711928, 10.675735931288072`}, {
          5.892893218813452, 11.241421356237312`}, {6.45857864376269, 
          11.80710678118655}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{6.108784489160851, 10.608784489160854`}, {1, 1}], LineBox[CompressedData["
1:eJxFlAlMFGcUx6e1XLFW2J2dEsKxMyzsFkRACimWNi+gyLVhoRxyCVvBFVBw
U6HlUEFAAtgtNRyLpqKAVo4uKPdhL1uiCLKlgAukdIEirVColAhFxH6T2o8v
mUx+ed838977/u9Pf5gcGPcqQRBh6GHf/y0KePIjaImAKDbKv5FAwVL94N1W
EeKYnaXl3RQoCgoetk0JgWjK3AxcpMAZOL32JYi1wQGvE2+CLf+Cv4cHYvVA
6d0lCoY/Lfbdt2wNhETsmPQ1Bf5W8cpzNYgNJ6w7kiiYlDut2kchJqrid+pQ
kOWiTNlugVg7FOeXz4Oy2s5fpAtWQJxwco7dIIEYfRA51otYW03qHiOhwjzP
5lAjYvhRsXuSC892+2TwriP+67i0LZwLZRvKlIg69nx39MzvHBi/dsCzvwsx
Qaa+8hkHMltlQ6EjLC9nGwRywP1vl6VDq+z+sPlwZw4YnDE7qzFj83X5rWgv
B2qe2mZk7kesFj8JlHIg6uR2n+UEtr7K9cIvOVBX6edGFCGOUdUGGHChUXeX
sLOKPW909P1sLoQuzY9ea2TrI1dU+iTkSeX/zLFc3O1vpSSBX9FTGsTuj8nR
3BTywG2Eqsg9x/ZHp+5wIw8i/McV45FsXIfsFFJQYL/qcMqazUc0b1BIwVy9
2OzALMrfwW5oVk3BwdDJ0oZLiPmtHtI1CspanL9R+iE2nDIWb1Jg+jzEe98z
Aaq3PS14moLTnsYDevUsD2TpXqfA8b312F3RiNXHJG1+FNi/0UHfNkH87exg
8AQP9gQUL0ZoLYG48lhlc5AHbz+P9DK9hRjGFKl9JDy62P4w7nPE2j43w70k
JG96KrSn2XhKw4qKC2JXt5jEdDYu1Yw4cOG7qCxT8xzEko/vfdXLgZD0814h
5YhPVPVHpHKgyFjP+0EL4uJt4oB9HFA1tChua9h49HygEwd8TFkdC3B8NNZX
7i4U4PPra2ttjr4C/H378o/aa48K8P9/aD4zM5UpwPm9o7lTWZInwPnzuJ1y
i2wBrq8y1Ut0MUmA65eQ1e0z4q3+HJ5LNqqx2Opf8DavSfdZS9zfoGGdSz9f
scT917s/t2ctyBLfTzxZ29b9miW+vxc5fSBrZvD92nXRCR/EMvj+L8fdmu4w
YbA+Nqb0B8+O0Vg/Nt9bya2qaawvb7c/jmSk0Vh/cZ94JNRH0VifRJNC11hC
Y/3aMeRb4Sy/1Pcjx2GNJ7v/pf7H5c13Xpyk8XyIvlhp3VFK4/mZMhnOzemi
8XxlXJg+JZymt+bP/J7hVX0Gz6dBpNLuhg2D57c+saPTaT+D53uRlBXWBDN4
/gunW3MfhzHYHxaW00euShjsH3nvOlZbuzLYX5ZknsxlLoP9R2dCVriipbE/
tff266uraOxfzgtlsrQIGvvbzZzklu4dNPa/8CfHm3p6+NgfJT/96apO5GP/
fJqfKvI252N/tb0fk1SisoD//fdXcELLAv4F/NIbGg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.4452, 10.}, {-1, 0}], 
         LineBox[{{5.5, 7.800000000001091}, {3.2999999999992724`, 
          10.000000000001819`}}], 
         PolygonBox[{{3.975735931288072, 9.324264068711928}, {
          5.107106781186548, 8.758578643762691}, {4.54142135623731, 
          8.192893218813454}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.891215510839149, 9.39121551083915}, \
{-1, -1}], 
         LineBox[{{5.500000000003638, 7.800000000001091}, {7.700000000002547, 
          10.}}], PolygonBox[{{6.175735931288071, 8.475735931288073}, {
          7.3071067811865476`, 9.04142135623731}, {6.741421356237309, 
          9.607106781186546}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.091215510839148, 8.408784489160851}, \
{-1, 1}], 
         {PointSize[0.04], PointBox[{5.5, 7.8}], 
          PointBox[{5.5, 12.200000000000001`}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["24",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 0}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{5.500000000001819, 12.200000000002547`}, {
          3.3000000000029104`, 10.000000000003638`}}], 
         PolygonBox[{{4.824264068711929, 11.524264068711929`}, {
          3.692893218813453, 10.958578643762692`}, {4.258578643762691, 
          10.392893218813455`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.908784489160852, 11.591215510839149`}, {1, -1}], 
         LineBox[{{5.499999999996362, 12.200000000000728`}, {
          7.699999999995271, 10.000000000001819`}}], 
         PolygonBox[{{7.024264068711928, 10.675735931288072`}, {
          5.892893218813452, 11.241421356237312`}, {6.45857864376269, 
          11.80710678118655}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.108784489160851, 10.608784489160854`}, {1, 1}], LineBox[CompressedData["
1:eJxFlAlMFGcUx6e1XLFW2J2dEsKxMyzsFkRACimWNi+gyLVhoRxyCVvBFVBw
U6HlUEFAAtgtNRyLpqKAVo4uKPdhL1uiCLKlgAukdIEirVColAhFxH6T2o8v
mUx+ed838977/u9Pf5gcGPcqQRBh6GHf/y0KePIjaImAKDbKv5FAwVL94N1W
EeKYnaXl3RQoCgoetk0JgWjK3AxcpMAZOL32JYi1wQGvE2+CLf+Cv4cHYvVA
6d0lCoY/Lfbdt2wNhETsmPQ1Bf5W8cpzNYgNJ6w7kiiYlDut2kchJqrid+pQ
kOWiTNlugVg7FOeXz4Oy2s5fpAtWQJxwco7dIIEYfRA51otYW03qHiOhwjzP
5lAjYvhRsXuSC892+2TwriP+67i0LZwLZRvKlIg69nx39MzvHBi/dsCzvwsx
Qaa+8hkHMltlQ6EjLC9nGwRywP1vl6VDq+z+sPlwZw4YnDE7qzFj83X5rWgv
B2qe2mZk7kesFj8JlHIg6uR2n+UEtr7K9cIvOVBX6edGFCGOUdUGGHChUXeX
sLOKPW909P1sLoQuzY9ea2TrI1dU+iTkSeX/zLFc3O1vpSSBX9FTGsTuj8nR
3BTywG2Eqsg9x/ZHp+5wIw8i/McV45FsXIfsFFJQYL/qcMqazUc0b1BIwVy9
2OzALMrfwW5oVk3BwdDJ0oZLiPmtHtI1CspanL9R+iE2nDIWb1Jg+jzEe98z
Aaq3PS14moLTnsYDevUsD2TpXqfA8b312F3RiNXHJG1+FNi/0UHfNkH87exg
8AQP9gQUL0ZoLYG48lhlc5AHbz+P9DK9hRjGFKl9JDy62P4w7nPE2j43w70k
JG96KrSn2XhKw4qKC2JXt5jEdDYu1Yw4cOG7qCxT8xzEko/vfdXLgZD0814h
5YhPVPVHpHKgyFjP+0EL4uJt4oB9HFA1tChua9h49HygEwd8TFkdC3B8NNZX
7i4U4PPra2ttjr4C/H378o/aa48K8P9/aD4zM5UpwPm9o7lTWZInwPnzuJ1y
i2wBrq8y1Ut0MUmA65eQ1e0z4q3+HJ5LNqqx2Opf8DavSfdZS9zfoGGdSz9f
scT917s/t2ctyBLfTzxZ29b9miW+vxc5fSBrZvD92nXRCR/EMvj+L8fdmu4w
YbA+Nqb0B8+O0Vg/Nt9bya2qaawvb7c/jmSk0Vh/cZ94JNRH0VifRJNC11hC
Y/3aMeRb4Sy/1Pcjx2GNJ7v/pf7H5c13Xpyk8XyIvlhp3VFK4/mZMhnOzemi
8XxlXJg+JZymt+bP/J7hVX0Gz6dBpNLuhg2D57c+saPTaT+D53uRlBXWBDN4
/gunW3MfhzHYHxaW00euShjsH3nvOlZbuzLYX5ZknsxlLoP9R2dCVriipbE/
tff266uraOxfzgtlsrQIGvvbzZzklu4dNPa/8CfHm3p6+NgfJT/96apO5GP/
fJqfKvI252N/tb0fk1SisoD//fdXcELLAv4F/NIbGg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.4452, 10.}, {-1, 0}], 
         LineBox[{{5.5, 7.800000000001091}, {3.2999999999992724`, 
          10.000000000001819`}}], 
         PolygonBox[{{3.975735931288072, 9.324264068711928}, {
          5.107106781186548, 8.758578643762691}, {4.54142135623731, 
          8.192893218813454}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.891215510839149, 9.39121551083915}, \
{-1, -1}], 
         LineBox[{{5.500000000003638, 7.800000000001091}, {7.700000000002547, 
          10.}}], PolygonBox[{{6.175735931288071, 8.475735931288073}, {
          7.3071067811865476`, 9.04142135623731}, {6.741421356237309, 
          9.607106781186546}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.091215510839148, 8.408784489160851}, \
{-1, 1}], 
         {PointSize[0.04], PointBox[{5.5, 7.8}], 
          PointBox[{5.5, 12.200000000000001`}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["25",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{5.500000000001819, 12.200000000002547`}, {
          3.3000000000029104`, 10.000000000003638`}}], 
         PolygonBox[{{4.824264068711929, 11.524264068711929`}, {
          3.692893218813453, 10.958578643762692`}, {4.258578643762691, 
          10.392893218813455`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.908784489160852, 11.591215510839149`}, {1, -1}], LineBox[CompressedData["
1:eJwd0Q0slHEcB/D/2nVR1HXR3dUpl6tuwsjLbo48CVHNzm3FIjVS3gpR1ClK
b16SIqnRUESUU+iuhWu19DbdnbymuqWXOzO7xGK9fp/+27Nnn+f5P/99f99H
EJUki5lFCPHBRd+J8hXWMoowaAgpkl1c6uEEqwv9LzFgytrySIkNRQwTpTlc
WByXlM6E+Rk7kyWw6EtcYh6fIm5dPJdEmPBOV6yEGS2Zb+tgFpdnObyUIiW+
FvpxmNtWu0EBT/6qcJSspIhS1+5wnXbaSHgeHNjoPrcF9ktoHBqA1a5Hj+nh
Fo3Uzm4VRc7qE28Lcb7Q5ZIxBiajuhQ5PF0VHXQNFnN0N97DhvwP48/hQIlb
6kbkTbOr9/gKl/myHt6By3ov+E3BsSPVqZaY15B4ZsEkrI7+y9gBszgLoz7B
+pPBty/Dojuhr7vg7ISGl0pY6B5SUUmfv2eLoB326vGuSYGT+Vl/qmHFLXuH
dXQesfn2eDg7vaLVnH7fWq+1gsumhnX9mM+s4F5xDfJUeh/+1gBTqT/WCuCi
yUdrcuG695F7z2GeiAftNgfo/TM1HSb0ITWfVsfCpO3Yehm8ZSE/ZD8scjVj
q5ZQxIo96JkDc5s/mjnAtlO/rW/C4kCvbc087BMua6D7FY3+XBMMl4elMBYj
H0t49yATruQxciPgjBXi6mEuRSzCO23r6f7ZpuNauOiXbOc03WeaXPIJ3p3U
dC9gNfIZc0LZ+F6hkl8poi2f+BgK2z4dfK6Dp0dVNU2wZuhN7FwReh/3TFiE
fA7lDEd3mNq6NSwT9iv0yQqBd9XKN43Azq61RZGwc3y153rMq2ntjYmgHdF0
7Tyc6XuzbzNMnFY8eQaPNWedcILVg/OXG2B1ilbFhFkNdrmjcJjALLIfeTQB
sn3dsMLmquo6rOB0viuG6+bFyZLggQ1dF71g1kGj0hsWR9n3vEQeaXLbNjYs
raoKCIDHrqzzMaEPaV9YK90v1bdk9hCsMD69bwWnGSNLtfCzGUn4IfTFD+8R
DNB9n+Kq33HwnwoKmOOwYe9MWwhcYPG4+//5ha5RA4vR33c3D3/Y1tShPwQ3
ttyX5MDKz45zXGjnMb1fwLE2+fbmsCZo81IO5jX9xbLG8z1YIuofelBfLQ==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.108784489160851, 10.608784489160854`}, {1, 1}], 
         LineBox[{{5.5, 7.8000000000031084`}, {5.5, 12.200000000004858`}}], 
         PolygonBox[{{5.5, 9.4}, {5.9, 10.6}, {5.1, 10.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.4452, 10.}, {-1, 0}], 
         LineBox[{{5.5, 7.800000000001091}, {3.2999999999992724`, 
          10.000000000001819`}}], 
         PolygonBox[{{3.975735931288072, 9.324264068711928}, {
          5.107106781186548, 8.758578643762691}, {4.54142135623731, 
          8.192893218813454}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.891215510839149, 9.39121551083915}, \
{-1, -1}], LineBox[CompressedData["
1:eJwd0w1MzGEcB/BHExHJS3XpUlcpldEWSl56JO+qu/KWl/XKTuKuOuR1TeRS
cmeWmrn9JZwr67au3TmRSWmW3FZ0OEblpdw4LRx5+z6e7dmzz57/8+z7++35
C9IlidsdCCHRmGwl5naMaZT8HwGUzKhbUtUMr+3t02n9KXF9r2q5AMcGcE8z
YLllvHU9HBYvesKH8xdXZg15U2Io7Xj8yo+StsxOt0Mw7+pWcS1sWKEb95lP
Se30QO4kTDoVORtg3VD2rRyY1+voZvCiZFyhffYuuOBbnocX3CXRqfLgitDQ
oqKplCjCzeZitp9V9YPAXWOEthrYLneMUnhSonYpTOxm9zWM/xwJL7ziYHVC
vrCigyl/eJRk6sOTFsEVgutLe+A+46XvUtg+ePtGH7zWeVcSB7epNE4jcV6+
xpPfBoul3rMWwfYXVPoWNqSHnC2BpQmSv1/hyEDRsX64tj06xM6+379fJkQ+
U9Az3kfY9FKZdhPWzVW1djC/mRTkjfrEsj0O1fBd0fN+GaxLycvIhl/3rtxr
hNcluYWHsPs3iZQf4F9Oq+Jeo75Uo2X1MDyn3GpXwvmacyE2WDu0yRgDm9VT
N7bCsfUX/b4JKOF6TNxROGxwrKoOzj+u2Mf6q7utj82FDV4DzRzyGcrO11FY
UXnfOBnmanK7+GzfPb7kIOozaRoGRsPqzkAXC/plqVJudoB9P+3eHg3bkjbq
nWFp8pPKGg9K+GVbWvxgLfWcHwTP7J5UsgymHv7DenecO7MvO4fdd645Nx0O
mBiRcJnl08Z1BMPC2Pst3fDdHQU1E2BbdvdiZ9QXliYYMQXWdgzMi4JnrDzq
Mxc2r/giToOddj56IIGrtZrhI7CwPzi5CZaNElw/zXyq0XUa8lBJ6KsyZmVF
8gnY1hhjKmDvreFn0Ce4YPoBeQZ7zz0/NotQX+mC3z0RrP/WFB813P5wVuFf
5LNXJMRbYZo10NgIi33P3+OhX+0jl8tYf8lPFy4YPuxy5Ks/64d7YDEfLi2P
eNDli/p09U02nJ9puXJADn+oXz77Glwd5743BubepafGwKbU+oeOsDpj2/pW
5JP7D07W+GDVNyVGwsQzLSoTFs6fV65i/RHe2bYQdmX/NRwZjuFD/wEPjWVZ

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.091215510839148, 8.408784489160851}, \
{-1, 1}], 
         {PointSize[0.04], PointBox[{5.5, 7.8}], 
          PointBox[{5.5, 12.200000000000001`}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["26",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{5.500000000001819, 12.200000000002547`}, {
          3.3000000000029104`, 10.000000000003638`}}], 
         PolygonBox[{{4.824264068711929, 11.524264068711929`}, {
          3.692893218813453, 10.958578643762692`}, {4.258578643762691, 
          10.392893218813455`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.908784489160852, 11.591215510839149`}, {1, -1}], LineBox[CompressedData["
1:eJwd0Q0slHEcB/D/2nVR1HXR3dUpl6tuwsjLbo48CVHNzm3FIjVS3gpR1ClK
b16SIqnRUESUU+iuhWu19DbdnbymuqWXOzO7xGK9fp/+27Nnn+f5P/99f99H
EJUki5lFCPHBRd+J8hXWMoowaAgpkl1c6uEEqwv9LzFgytrySIkNRQwTpTlc
WByXlM6E+Rk7kyWw6EtcYh6fIm5dPJdEmPBOV6yEGS2Zb+tgFpdnObyUIiW+
FvpxmNtWu0EBT/6qcJSspIhS1+5wnXbaSHgeHNjoPrcF9ktoHBqA1a5Hj+nh
Fo3Uzm4VRc7qE28Lcb7Q5ZIxBiajuhQ5PF0VHXQNFnN0N97DhvwP48/hQIlb
6kbkTbOr9/gKl/myHt6By3ov+E3BsSPVqZaY15B4ZsEkrI7+y9gBszgLoz7B
+pPBty/Dojuhr7vg7ISGl0pY6B5SUUmfv2eLoB326vGuSYGT+Vl/qmHFLXuH
dXQesfn2eDg7vaLVnH7fWq+1gsumhnX9mM+s4F5xDfJUeh/+1gBTqT/WCuCi
yUdrcuG695F7z2GeiAftNgfo/TM1HSb0ITWfVsfCpO3Yehm8ZSE/ZD8scjVj
q5ZQxIo96JkDc5s/mjnAtlO/rW/C4kCvbc087BMua6D7FY3+XBMMl4elMBYj
H0t49yATruQxciPgjBXi6mEuRSzCO23r6f7ZpuNauOiXbOc03WeaXPIJ3p3U
dC9gNfIZc0LZ+F6hkl8poi2f+BgK2z4dfK6Dp0dVNU2wZuhN7FwReh/3TFiE
fA7lDEd3mNq6NSwT9iv0yQqBd9XKN43Azq61RZGwc3y153rMq2ntjYmgHdF0
7Tyc6XuzbzNMnFY8eQaPNWedcILVg/OXG2B1ilbFhFkNdrmjcJjALLIfeTQB
sn3dsMLmquo6rOB0viuG6+bFyZLggQ1dF71g1kGj0hsWR9n3vEQeaXLbNjYs
raoKCIDHrqzzMaEPaV9YK90v1bdk9hCsMD69bwWnGSNLtfCzGUn4IfTFD+8R
DNB9n+Kq33HwnwoKmOOwYe9MWwhcYPG4+//5ha5RA4vR33c3D3/Y1tShPwQ3
ttyX5MDKz45zXGjnMb1fwLE2+fbmsCZo81IO5jX9xbLG8z1YIuofelBfLQ==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.108784489160851, 10.608784489160854`}, {1, 1}], 
         LineBox[{{5.5, 7.8000000000031084`}, {5.5, 12.200000000004858`}}], 
         PolygonBox[{{5.5, 9.4}, {5.9, 10.6}, {5.1, 10.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.4452, 10.}, {-1, 0}], 
         LineBox[{{5.5, 7.800000000001091}, {3.2999999999992724`, 
          10.000000000001819`}}], 
         PolygonBox[{{3.975735931288072, 9.324264068711928}, {
          5.107106781186548, 8.758578643762691}, {4.54142135623731, 
          8.192893218813454}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.891215510839149, 9.39121551083915}, \
{-1, -1}], LineBox[CompressedData["
1:eJwd0w1MzGEcB/BHExHJS3XpUlcpldEWSl56JO+qu/KWl/XKTuKuOuR1TeRS
cmeWmrn9JZwr67au3TmRSWmW3FZ0OEblpdw4LRx5+z6e7dmzz57/8+z7++35
C9IlidsdCCHRmGwl5naMaZT8HwGUzKhbUtUMr+3t02n9KXF9r2q5AMcGcE8z
YLllvHU9HBYvesKH8xdXZg15U2Io7Xj8yo+StsxOt0Mw7+pWcS1sWKEb95lP
Se30QO4kTDoVORtg3VD2rRyY1+voZvCiZFyhffYuuOBbnocX3CXRqfLgitDQ
oqKplCjCzeZitp9V9YPAXWOEthrYLneMUnhSonYpTOxm9zWM/xwJL7ziYHVC
vrCigyl/eJRk6sOTFsEVgutLe+A+46XvUtg+ePtGH7zWeVcSB7epNE4jcV6+
xpPfBoul3rMWwfYXVPoWNqSHnC2BpQmSv1/hyEDRsX64tj06xM6+379fJkQ+
U9Az3kfY9FKZdhPWzVW1djC/mRTkjfrEsj0O1fBd0fN+GaxLycvIhl/3rtxr
hNcluYWHsPs3iZQf4F9Oq+Jeo75Uo2X1MDyn3GpXwvmacyE2WDu0yRgDm9VT
N7bCsfUX/b4JKOF6TNxROGxwrKoOzj+u2Mf6q7utj82FDV4DzRzyGcrO11FY
UXnfOBnmanK7+GzfPb7kIOozaRoGRsPqzkAXC/plqVJudoB9P+3eHg3bkjbq
nWFp8pPKGg9K+GVbWvxgLfWcHwTP7J5UsgymHv7DenecO7MvO4fdd645Nx0O
mBiRcJnl08Z1BMPC2Pst3fDdHQU1E2BbdvdiZ9QXliYYMQXWdgzMi4JnrDzq
Mxc2r/giToOddj56IIGrtZrhI7CwPzi5CZaNElw/zXyq0XUa8lBJ6KsyZmVF
8gnY1hhjKmDvreFn0Ce4YPoBeQZ7zz0/NotQX+mC3z0RrP/WFB813P5wVuFf
5LNXJMRbYZo10NgIi33P3+OhX+0jl8tYf8lPFy4YPuxy5Ks/64d7YDEfLi2P
eNDli/p09U02nJ9puXJADn+oXz77Glwd5743BubepafGwKbU+oeOsDpj2/pW
5JP7D07W+GDVNyVGwsQzLSoTFs6fV65i/RHe2bYQdmX/NRwZjuFD/wEPjWVZ

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.091215510839148, 8.408784489160851}, \
{-1, 1}], 
         {PointSize[0.04], PointBox[{5.5, 7.8}], 
          PointBox[{5.5, 12.200000000000001`}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["27",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{3.300000000003109, 10.}, {7.700000000004858, 10.}}], 
         PolygonBox[{{6.1, 10.}, {4.9, 9.6}, {4.9, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{5.596999999999298, 9.743999999998778}, {
          3.3000000000006366`, 9.999999999996362}}], 
         PolygonBox[{{3.8521919537804865`, 9.938458363009227}, {
          5.089113621558999, 10.20308033447045}, {5.000502470880029, 
          9.408002939511096}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.620184104440504, 10.807662452733744`}, {0, -1}], 
         CircleBox[{5.045225, 9.511987500000002}, 0.598569503718031, {
          3.982260792372014, 6.681224561266225}], 
         PolygonBox[{{5.881252287050288, 9.372905416004308}, {
          5.136351323619248, 8.350593059295502}, {4.6719441022801735`, 
          9.001996108695888}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.687688991344457, 8.367306591786756}, \
{-1, 1}], LineBox[CompressedData["
1:eJwBIQLe/SFib1JlAgAAACEAAAACAAAAF9nO91NjFkCwcmiR7XwjQHlB6UsJ
QhZAm/kk+xrQI0D6DP3tmskVQJXVYRajGSRAOLxKWDYKFUBp/IPm1kQkQMB5
hswzLhRAuP5z7BhDJEBomGE1GWwTQBIqbaL3EyRAWD4lUq7wEkDaIcG03cUj
QAJmUr76zhJAjXLNn7NuI0AvA5M1Hf8SQIuDiEWjISNAoLY0YCNpE0DaEB12
QegiQFV3pAnn8RNAQ7NIKqrBIkDle5z4KoIUQAeGaLUIpiJAElkdJA4HFUAZ
idjTuYsiQN6/Nr36bxVAEHeAiQlsIkCbBeHa0a4VQOQAGGBkRiJAn3btYfa7
FUAk+Jzi9CAiQKWfUCdtmxVAb4vxXRIGIkAig4u1/F0VQDGRGx91/yFAe2On
2H0bFUBA1xbe3xEiQJgPCE7N6BRAbviU2Nk7IkCrc7g7GM8UQGR+rB3HdyJA
RKUkS63IFEBeKfkMK78iQIaWptB3wxRAgkxjLV4MI0BMWl83F6cUQC6f52GY
WSNAbGgMpildFEByflQja54jQPkTsBMm2xNAL4j4jEvOI0BaT8h+WysTQFjr
1kHI2iNApnm663NvEkAkON9S6LkjQHS0vLMP1xFAH8bh/4VtI0CMjaFJCowR
QOMMmsIYBiNAi0uyk1OfEUDIyodEi50iQK7jGYwXAxJA/cNmT6VMIkAv3SQG
gZUSQG8Sg8DKISJAl5/bYg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.5553110086555435`, 10.442693408213248`}, {1, -1}], 
         LineBox[{{4.646000000003369, 9.066000000002532}, {7.700000000004366, 
          10.}}], PolygonBox[{{5.599232828529638, 9.357525691501863}, {
          6.863750043802455, 9.32596286085123}, {6.629784299138269, 
          10.090985756145049`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.626308630286861, 8.655568140368231}, \
{0, 1}], 
         {PointSize[0.04], PointBox[{4.646, 9.066}], PointBox[{5.597, 9.744}],
           PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["28",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{3.300000000003109, 10.}, {7.700000000004858, 10.}}], 
         PolygonBox[{{4.9, 10.}, {6.1, 9.6}, {6.1, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{5.596999999999298, 9.743999999998778}, {
          3.3000000000006366`, 9.999999999996362}}], 
         PolygonBox[{{5.044808046219514, 9.805541636990773}, {
          3.8964975291199715`, 10.335997060488904`}, {3.8078863784410015`, 
          9.54091966552955}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.620184104440504, 10.807662452733744`}, {0, -1}], 
         CircleBox[{5.045225, 9.511987500000002}, 0.598569503718031, {
          3.982260792372014, 6.681224561266225}], 
         PolygonBox[{{4.904147712949711, 8.676294583995695}, {
          6.113455897719826, 9.047203891304115}, {5.649048676380751, 
          9.6986069407045}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.687688991344457, 8.367306591786756}, \
{-1, 1}], LineBox[CompressedData["
1:eJwBIQLe/SFib1JlAgAAACEAAAACAAAAF9nO91NjFkCwcmiR7XwjQHlB6UsJ
QhZAm/kk+xrQI0D6DP3tmskVQJXVYRajGSRAOLxKWDYKFUBp/IPm1kQkQMB5
hswzLhRAuP5z7BhDJEBomGE1GWwTQBIqbaL3EyRAWD4lUq7wEkDaIcG03cUj
QAJmUr76zhJAjXLNn7NuI0AvA5M1Hf8SQIuDiEWjISNAoLY0YCNpE0DaEB12
QegiQFV3pAnn8RNAQ7NIKqrBIkDle5z4KoIUQAeGaLUIpiJAElkdJA4HFUAZ
idjTuYsiQN6/Nr36bxVAEHeAiQlsIkCbBeHa0a4VQOQAGGBkRiJAn3btYfa7
FUAk+Jzi9CAiQKWfUCdtmxVAb4vxXRIGIkAig4u1/F0VQDGRGx91/yFAe2On
2H0bFUBA1xbe3xEiQJgPCE7N6BRAbviU2Nk7IkCrc7g7GM8UQGR+rB3HdyJA
RKUkS63IFEBeKfkMK78iQIaWptB3wxRAgkxjLV4MI0BMWl83F6cUQC6f52GY
WSNAbGgMpildFEByflQja54jQPkTsBMm2xNAL4j4jEvOI0BaT8h+WysTQFjr
1kHI2iNApnm663NvEkAkON9S6LkjQHS0vLMP1xFAH8bh/4VtI0CMjaFJCowR
QOMMmsIYBiNAi0uyk1OfEUDIyodEi50iQK7jGYwXAxJA/cNmT6VMIkAv3SQG
gZUSQG8Sg8DKISJAl5/bYg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.5553110086555435`, 10.442693408213248`}, {1, -1}], 
         LineBox[{{4.646000000003369, 9.066000000002532}, {7.700000000004366, 
          10.}}], PolygonBox[{{6.746767171470362, 9.70847430849814}, {
          5.716215700861731, 8.975014243854954}, {5.4822499561975455`, 
          9.740037139148772}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.626308630286861, 8.655568140368231}, \
{0, 1}], 
         {PointSize[0.04], PointBox[{4.646, 9.066}], PointBox[{5.597, 9.744}],
           PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["29",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{3.300000000003109, 10.}, {7.700000000004858, 10.}}], 
         PolygonBox[{{6.1, 10.}, {4.9, 9.6}, {4.9, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{5.596999999999298, 9.743999999998778}, {
          3.3000000000006366`, 9.999999999996362}}], 
         PolygonBox[{{3.8521919537804865`, 9.938458363009227}, {
          5.089113621558999, 10.20308033447045}, {5.000502470880029, 
          9.408002939511096}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.620184104440504, 10.807662452733744`}, {0, -1}], 
         CircleBox[{5.045225, 9.511987500000002}, 0.598569503718031, {
          3.982260792372014, 6.681224561266225}], 
         PolygonBox[{{5.881252287050288, 9.372905416004308}, {
          5.136351323619248, 8.350593059295502}, {4.6719441022801735`, 
          9.001996108695888}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.687688991344457, 8.367306591786756}, \
{-1, 1}], LineBox[CompressedData["
1:eJwBIQLe/SFib1JlAgAAACEAAAACAAAAF9nO91NjFkCwcmiR7XwjQHlB6UsJ
QhZAm/kk+xrQI0D6DP3tmskVQJXVYRajGSRAOLxKWDYKFUBp/IPm1kQkQMB5
hswzLhRAuP5z7BhDJEBomGE1GWwTQBIqbaL3EyRAWD4lUq7wEkDaIcG03cUj
QAJmUr76zhJAjXLNn7NuI0AvA5M1Hf8SQIuDiEWjISNAoLY0YCNpE0DaEB12
QegiQFV3pAnn8RNAQ7NIKqrBIkDle5z4KoIUQAeGaLUIpiJAElkdJA4HFUAZ
idjTuYsiQN6/Nr36bxVAEHeAiQlsIkCbBeHa0a4VQOQAGGBkRiJAn3btYfa7
FUAk+Jzi9CAiQKWfUCdtmxVAb4vxXRIGIkAig4u1/F0VQDGRGx91/yFAe2On
2H0bFUBA1xbe3xEiQJgPCE7N6BRAbviU2Nk7IkCrc7g7GM8UQGR+rB3HdyJA
RKUkS63IFEBeKfkMK78iQIaWptB3wxRAgkxjLV4MI0BMWl83F6cUQC6f52GY
WSNAbGgMpildFEByflQja54jQPkTsBMm2xNAL4j4jEvOI0BaT8h+WysTQFjr
1kHI2iNApnm663NvEkAkON9S6LkjQHS0vLMP1xFAH8bh/4VtI0CMjaFJCowR
QOMMmsIYBiNAi0uyk1OfEUDIyodEi50iQK7jGYwXAxJA/cNmT6VMIkAv3SQG
gZUSQG8Sg8DKISJAl5/bYg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.5553110086555435`, 10.442693408213248`}, {1, -1}], 
         LineBox[{{4.646000000003369, 9.066000000002532}, {7.700000000004366, 
          10.}}], PolygonBox[{{5.599232828529638, 9.357525691501863}, {
          6.863750043802455, 9.32596286085123}, {6.629784299138269, 
          10.090985756145049`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.626308630286861, 8.655568140368231}, \
{0, 1}], 
         {PointSize[0.04], PointBox[{4.646, 9.066}], PointBox[{5.597, 9.744}],
           PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["30",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHtIU1Ecxy8pPjGsNDbN3O45q2mFmkqFD34is6wJZZaSlGNlzke6cqiV
lpUPVCJYZWWiy5JePjIlG5hJ4iyTJJZpKjI1fDDJR5k2Z3Vuf5wOHC4fzrnn
/B7f7xHK0yLjVzEME0om943u6yZjIzD/BgaBzHi3NNgdGPGoasEHQ2VrmFDX
Rnj/ZuVCDAbT7NyN7M0CYBzD09l0DM4hKm+3M4T7fSwCL2IoOBlkWfmKsKDX
XJuNIWO7OemAtRCYyb1PbyVjaOf5GO9ICUPpni4phmu31XynEsKatOVWIQZ1
fMfXujbCBm133jSCIl3iWIeRsGPT6Yf1CCR+11P97VlglF3ujqkIzIdq1YqN
hL3HJ0Y8EfzoPbdVJiIs/hBxcIqF9pulg88F3P6ZuYIaFsIf1rzPX0N4j6dX
fCYLksKA2gu/yPm3z17OkrJwhb9cVD/AxbtO+WEbC0/s/iTnNBNuq5APb2Ah
+FOPwxU14VlDc6YLC7IGG5MylXB/4efmTSykWIdqpiMIx+zW2YSwUDz/bi7N
l8svP1asYCGxkb92i5CwYrhiRxkLNq9Hrk7xCGetbvr4iQXhilRmdiWc6yr2
cEYwIX+p99hCWCAGzxgEtqZ8iwdhhJfcqjVlCLCoJd2Ku1/Aqzz2BUGVXdbQ
EQ0XfyhzYg2GxVjrGTzErdeMrgIMvS1LeSp3kr/sewZzHIPBQXtkJYHwTpfI
Z1lkf6Cvr28jV69uo00uhn3yWvvE3//XDzeeGpkLQ/R/P6e4zJBiRM8viv9W
ZK9D9P73UU/m45YQjY/NMMjfuGMaf6hDnTowANP87NL5Vkl7Mc2fbcjrz47A
tD75OmlUgATT+t3LKVf93I5pfVsVSHeWh2n9u1qPpiwsItqf6bulMWN6RPsn
SUswTdYh2t9Hh5OHXpQg2v/xSU/tRAqi+hh9qy0zRCGqH8VAT3GjBFF9me4b
d1mFIKo/LxetfmU3ovocD9Avesciqt+qx+bLmnOI6vuO3b1LzdWI6r+8Ty+N
HETUH4po//PdfEz90xSUtOwUh6m/qm095J2PMPXfeq95//AfmPrT56Lq41Sw
iPp30LH9kEWBiPrbtU3Jc+0UUf/nWnLvwib6PvwFZfuZHg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 10.9452}, {0, -1}], 
         LineBox[{{3.300000000003109, 10.}, {7.700000000004858, 10.}}], 
         PolygonBox[{{4.9, 10.}, {6.1, 9.6}, {6.1, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{5.596999999999298, 9.743999999998778}, {
          3.3000000000006366`, 9.999999999996362}}], 
         PolygonBox[{{5.044808046219514, 9.805541636990773}, {
          3.8964975291199715`, 10.335997060488904`}, {3.8078863784410015`, 
          9.54091966552955}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.620184104440504, 10.807662452733744`}, {0, -1}], 
         CircleBox[{5.045225, 9.511987500000002}, 0.598569503718031, {
          3.982260792372014, 6.681224561266225}], 
         PolygonBox[{{4.904147712949711, 8.676294583995695}, {
          6.113455897719826, 9.047203891304115}, {5.649048676380751, 
          9.6986069407045}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], {5.687688991344457, 8.367306591786756}, \
{-1, 1}], LineBox[CompressedData["
1:eJwBIQLe/SFib1JlAgAAACEAAAACAAAAF9nO91NjFkCwcmiR7XwjQHlB6UsJ
QhZAm/kk+xrQI0D6DP3tmskVQJXVYRajGSRAOLxKWDYKFUBp/IPm1kQkQMB5
hswzLhRAuP5z7BhDJEBomGE1GWwTQBIqbaL3EyRAWD4lUq7wEkDaIcG03cUj
QAJmUr76zhJAjXLNn7NuI0AvA5M1Hf8SQIuDiEWjISNAoLY0YCNpE0DaEB12
QegiQFV3pAnn8RNAQ7NIKqrBIkDle5z4KoIUQAeGaLUIpiJAElkdJA4HFUAZ
idjTuYsiQN6/Nr36bxVAEHeAiQlsIkCbBeHa0a4VQOQAGGBkRiJAn3btYfa7
FUAk+Jzi9CAiQKWfUCdtmxVAb4vxXRIGIkAig4u1/F0VQDGRGx91/yFAe2On
2H0bFUBA1xbe3xEiQJgPCE7N6BRAbviU2Nk7IkCrc7g7GM8UQGR+rB3HdyJA
RKUkS63IFEBeKfkMK78iQIaWptB3wxRAgkxjLV4MI0BMWl83F6cUQC6f52GY
WSNAbGgMpildFEByflQja54jQPkTsBMm2xNAL4j4jEvOI0BaT8h+WysTQFjr
1kHI2iNApnm663NvEkAkON9S6LkjQHS0vLMP1xFAH8bh/4VtI0CMjaFJCowR
QOMMmsIYBiNAi0uyk1OfEUDIyodEi50iQK7jGYwXAxJA/cNmT6VMIkAv3SQG
gZUSQG8Sg8DKISJAl5/bYg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.5553110086555435`, 10.442693408213248`}, {1, -1}], 
         LineBox[{{4.646000000003369, 9.066000000002532}, {7.700000000004366, 
          10.}}], PolygonBox[{{6.746767171470362, 9.70847430849814}, {
          5.716215700861731, 8.975014243854954}, {5.4822499561975455`, 
          9.740037139148772}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.626308630286861, 8.655568140368231}, \
{0, 1}], 
         {PointSize[0.04], PointBox[{4.646, 9.066}], PointBox[{5.597, 9.744}],
           PointBox[{3.3000000000000003`, 10.}], PointBox[{11., 10.}], 
          PointBox[{7.7, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["31",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxFlHtsDFscx8/qPpRW17ba4qa3MkcpaukQdP84v6tLVCI2VTRRSVVNBbu2
qD9QSnQb3MZqV2u29JF61KuESj2iEqqKUdamHvHIiiq92avFqlR7d++p5IyT
TCafzJzf+T2+3zMua2PqmiEIoWT6DL7Xpgp0jQH0a2EISY8qGQiMBmSvNbyK
xxCumplaKlL+dLR27GIMW9Pmf57EU35ShurWY/C/9djvSNF0/7PpGTsw2M9v
Na4UKJtia1btwjBCofb9CEQBak1Wr96CofrU2o2lImXrkdbDGRh0F4o36XnK
mam95UkYWlOq6x9KkYCuqu1xOgxJ4gd3jkB59vvkho8chBTEVAQhyp5/pnhv
cYCKTI8qxVGAqheunFrFQUdDbMocnjJUno7Yx4FhxSyVW4oANNExN7SAg0J9
SdgGgXKsyRhZyEFmhnK/ElFe8OB2TTkH9d49x4+K4TT+POHEFQ7ipmnLZvCD
bPS4X3Hg6zYeliQdrTfoY58KQ32n2ZUtULb+mb9Fj0F98w/zf4GRNN6tPZeX
YOha3rKtTKScuYm/bsFwMq+lX89T9jgsB/Mx9JvvvLsvaQFNW249tBNDcRLW
ZQuUre67tbkYzib2tQ8EwgBFx1sS0zFox2Udms5T9ux8EjwTw5k13yMFYQSg
npzPmhAM+p7Q2U4xlNZ/I//COw7Gd0S9eSSFAHqxrqPuOgfm4x6NAlH+FBFG
KjhwGNpcPD+c/t98sNPGgckwqIdhNP820+tdtN8G/88jYrD8fbPF2/xQGirv
317S0O0PaOT4ubYmRyKvkc/v1T6/IghqOT/70nWGClEl5/892d37WFLK9dkU
e+uUSPm7fue84Dl8kNyftAOkfYMwRO5fWwt5XiUq5P46Bkq+uiQk9994uic+
CCF5PtaEQmOO4CdsfienpP91TxogbL62pGLNBL6fsPk3bS/KLRL7CNPHZL1p
RVfgB2H6SSu4sXeR0EuYvi6NjsxrkHyE6W+ua1l2DP+NMH3q/o1JOCB+IUy/
57Qqhy/QTZi+jX/rm9olL2H6961qrmwUuwjzR8Exb2O50EmYf7RPFaWb+feE
+Utp2RGWgjyE+U97d2l6gvMlYf6sNvt3N/Juwvx78V6cLct5nzB/I9fV8Dzn
NcL8/+taiC6/ye6H/wEYu5+a
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.3000000000000003`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.50205345667312, 15.000143777988384`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {14.599999999998545`, 8.}}], 
         PolygonBox[{{16.77550563432708, 6.791385758707179}, {
          18.018751538144375`, 6.558277151741437}, {17.63023719320147, 
          5.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.44794654332688, 7.2501437779883835`}, {-1, -1}], 
         LineBox[{{13.700000000008004`, 11.500000000003638`}, {
          12.800000000008367`, 11.000000000003638`}}], 
         PolygonBox[{{13.774494365672922`, 11.541385758707179`}, {
          12.531248461855625`, 11.308277151741436`}, {12.91976280679853, 
          10.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{13.102053456673119`, 12.000143777988384`}, {1, -1}], 
         CircleBox[{15.275, 12.905}, 2.1106041789023346`, {
          3.8700056003299204`, 6.568969369224131}], 
         PolygonBox[{{16.824494365672923`, 11.35138575870718}, {
          15.969762806798531`, 10.418951330844207`}, {15.581248461855626`, 
          11.118277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.44794654332688, 10.309856222011618`}, {-1, 1}], LineBox[CompressedData["
1:eJwtVQs0lVkbPhNK5BaViKlR45xv7+8731Eml6aeaIowyPVvzUjphv6i/lwO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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 14.690143777988382`}, {1, -1}], 
         LineBox[{{12.800000000010186`, 11.000000000003638`}, {
          6.600000000011278, 10.000000000003638`}}], 
         PolygonBox[{{10.292344672427589`, 10.595539463294772`}, {
          9.043962352042564, 10.799356984990286`}, {9.171348303102262, 
          10.00956408842017}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.453189719821838, 11.425423737104602`}, {0, -1}], LineBox[CompressedData["
1:eJwd1AtMU1cYB/CLIL04ijgY5SFdaR0Cm7PiA0TEK+2QRaObaASiPAVuEAMI
Sm1AxVfKtFpQl8IQLi6iBjY7iRUkOmQ8GrDIZNGKoJ2CNAqVACpBBf/Hk9zc
/HLP4zv/fLk+SZmbU2ZRFBWBh7yp4KUYixjq8xAxlKnOLnHkO4baUMpzEcBM
+tj3tbBu36mWlbCKd99uDzw5yduUButO/vWegUUpW3Mq4OAHUW2esKPx9vF+
eDR5YQUFb4++cEzsg3OsJ5ePf8tQuTI6YBesaOXpxuD8MOHs67C09io3DbvL
8ss+wirtGyc3rP9gZ5MfLmaoyIFjhUFwvZafUgir2o1OibBc8DqqHk5oTt+j
gc+ei+EGYe7wGd9/4PLmST4twfyNOw69Jd4b0SuEuSnOuBD3VykdW/2JY6U5
W2CuItCFWKFzmKOEh7saTniT9Tbhyl9hZ8Y+k+ynGFF8eRnOCuxlX+G8Q4p7
ZX/AZnlFgQFmi9RJ1bC80zBeBRtS+9eVwKxtYogSjr4a0ZANd6c1GqPgrDUm
3Q9w3zbOawmxOOjKXHjZmCvrCmtmHNX/kryd3tXNkLyyT1epYYVVVT0BG8qd
d0fAuSqLkpjruV5G8hyIDx8j82lecV89/N9U7HoB9nPmZRjz4GDPrjUhJE9H
bc9qWNMl5rEw1Uzb82G/L0riK2H3eIvrUABDXY71Ln8Cj6pqVhjhvivbbkhI
fmv3ezbB0cn89AwJqe+g+W84tNIo18Nar1/c78KTQ8s7p2GN20jbc1jKdp2W
LWCopq+LMmeT+kRhmw7D5pos38VwpLzaoQGuv5Z9MQ7OqqU9XsDd/SUpxaT+
i7dl9Dfox7DUxhZ4VLGkQwj7URMTb0g/xazK9Ie7m0byxKQ/PHqUfjCz1PRk
HekH/qB+PkzFDrjsJN8L8lN5cNPN9exemNZzQS9xHmcbcElJ+tF6RNoOO/vb
SHLg0d8eaDnYcuunW3GkP5+GmhWw4dyrodWwZdr2yGaYbv1QOQ8ejvrRuhiO
bBxPeoz6tld17HSB2Ywe7/Nwd4xm8HM+WysFMfDwY/VXE7Ch7+HDeXDomcBF
xAl1kjkdyO/u/zdfzsDu55MTjsIthb1FApLX/YO5MrjcS3gnhNSXtrbGAXa9
5NPKwqKCAx6P/JGzPu9nUr+Z35ZRB4ue9+ufwgn2xaZS2DIltC5AHobfO2k1
bHK79+duOFi60ZO4NveZ6gasLcWA7cj/xJf5BIVsjuk=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.975685965145555`, 9.307331579087336}, {1, 1}], 
         LineBox[{{6.600000000001273, 9.999999999996362}, {14.60000000000582, 
          7.999999999996362}}], 
         PolygonBox[{{10.0179144999128, 9.145521375021799}, {
          11.085071250072666`, 8.466421624920068}, {11.279099750101732`, 
          9.242535625036334}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.224069781193684`, 8.101079124774735}, {0, 1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 11.5}], 
          PointBox[{12.8, 11.}], PointBox[{6.6000000000000005`, 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{14.6, 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["32",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000010914`, 15.000000000007276`}, {
          14.600000000011278`, 12.000000000007276`}}], 
         PolygonBox[{{17.824494365672923`, 13.791385758707179`}, {
          16.581248461855626`, 13.558277151741436`}, {16.96976280679853, 
          12.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.152053456673123`, 14.250143777988384`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {14.599999999998545`, 8.}}], 
         PolygonBox[{{16.77550563432708, 6.791385758707179}, {
          18.018751538144375`, 6.558277151741437}, {17.63023719320147, 
          5.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.44794654332688, 7.2501437779883835`}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQLbP/DBDIOTCAgYpDfulF6QXJ8g4MHBHs6X4qDv+W
G59MElBwYBBIa5pdqeKQuTdQSrIMxBeT3Vat4mDhaLXj81UgfwWHfW6IisPb
CRkhB5QVHRhuZEnJc6o4vFvYEXokFMiX8L4nukTZwTBntqZgGpA/4dsueWNl
B/E9lmuDg4H8GXEfVh9VchAu/RXDpAjkb2B4tS5LyWHGhOsiT64Bzb8xS2+6
npKD7bf////XA/kM0dNnSSo5PIzWeC2tiZAPZuJ2+71XHq7fp6RnaUSkPNx8
cZGTa+4ay8Ptn7fiXUh5iDzcfRLb9cLrd8jD3e+4+qg7jxbCf1cs1j191onw
v73fGbO57xHh02ot/knfTxEefuunzQQCRXj4AgAUnYKF
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.25, 10.9452}, {0, -1}], 
         LineBox[{{8.799999999996544, 9.999999999996362}, {
          14.599999999993088`, 11.999999999996362`}}], 
         PolygonBox[{{12.267223788975762`, 11.195594409991642`}, {
          11.263172484351998`, 10.426256397357848`}, {11.002379937696475`, 
          11.182554782658867`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.205285559145075`, 10.139471878479277`}, {0, 1}], 
         LineBox[{{8.799999999996544, 10.}, {14.599999999993997`, 
          8.000000000003638}}], 
         PolygonBox[{{11.132776211024236`, 9.195594409991642}, {
          12.136827515648001`, 8.426256397357848}, {12.397620062303524`, 
          9.182554782658867}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.194714440854924`, 8.139471878479277}, {0, 1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         LineBox[CompressedData["
1:eJxF1A1MU1cUB/CnYoAGRAUUlk3E1/dayHuvfLoSKDl0bgsyQWBjG5qCXRop
okwKsjHGQGTWAobvkTrsB1YlkQ0EYY5JBw66iCKwDFvGujIVts5NYXHqhnO3
yXJp0jS/NO++e8/9nxMsz09TrCYIIhV9Xb9EWCT6sJDp7sJWIN5t5EU9YaD3
n4qymQRkYb9gw00Gjsg+ett0Enni/typfgak7/z1ufZnly3XFb0MWJx7ne/H
BAPRav3BNMLAmT5b9cRJZIftUpSTAaf+euYrduSwO0X6IBbijKRlgdwGxPoc
dYachW06SaNyL7L1k2lHJwu2pHXifVXIj2XCiKcsnE98Fn+oDVlo1C/v5EBT
LRtPPYP8YFzj1cDBkuH89JIWuS58v+EGB8n9v2mTK5HfWkwwP+Eg1dpxOiPb
tf70nQJfEfxtrEkYjEbuCn3DLVAEuzPG9g+4I0/8K/HjiYB/YSl3cMZ1npEU
4wIHdj8pSfcgd20XLvdwcL8z6fmHTch6+3xzAQdTWt2fNZXI4tb1NwUcsHWJ
xYUVyGry0jkrC/aiATlT53LDrVwNCwOWLJ1/l+v59Kl0KQulvz8cznW41v9x
JGotC5VetrhdQWg/4hof/i0GykOkaXeVyOUhydsHGAh4cYul4SvkrQvGxB4G
TG3fLIX6kkB8kdd318xAx9qYQFAiByyZIhwMJD8TSw9+ifz1Tyr3jSwQ+R4l
htV8VK/jbj5pLEi8NxYo4pHDYtlVn7LQ5BVt3XQA2Rrnds/JwqnxY3n+x5HV
N06oYzggY8On7PXIRC3RfJQDg60bNtcgl2c7uSEOkoa3HC0rRna0mOhFDjZc
mFyezEB+kO/7qo8Iuq8WXWtjkQOEZTWo/qM8V/5W/k+P10TLp0j8fImSW8N1
kHh9SJ/7sEJN4vfLhcnmX1Qk3l/OzDgbfIjE+1dGftDofI/E53vkMSb4tYnE
50/x+KNWO7RSnzWNj6pkT1fqJ9lkudLwMh/Xd8/y6Ly5hY/rP18c6/39PT6+
H5V9TnzwJQrf3zHrlc7nWih8v56egqsXHRS+f7UfT9IcRON8TO7SZztTaJyf
Tt2Otu/yaZyvrKo+dXUZjfN3uSTbS1VK43z6v957m3eAxvntJAZnba/RON85
Su/mBIrG+Y81R7XueEzh/qgvUA0mXKNw//SEZOzr0FG4v+JlforuUgr3n1b1
Zoi3gsL9uUpjCP9sD4X79/YYMVQvp3B/h/YtWk6XULj/c48M57W0U3g+fJs1
Kw6YpfD8MBZGmi4H03i+LMjoQtFhGs+fuM3t/h+P0ng+7bzouVvxggDPr9r2
8rOZhwV4vlkDyyqoQeT/51/EOlf+hPAfMSz6yA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.654799999999998`, 10.}, {1, 0}], 
         {PointSize[0.04], PointBox[{8.8, 10.}], PointBox[{7.7, 10.}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{14.6, 12.}], 
          PointBox[{14.6, 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["33",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000010914`, 15.000000000007276`}, {
          14.600000000011278`, 12.000000000007276`}}], 
         PolygonBox[{{17.824494365672923`, 13.791385758707179`}, {
          16.581248461855626`, 13.558277151741436`}, {16.96976280679853, 
          12.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.152053456673123`, 14.250143777988384`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {14.599999999998545`, 8.}}], 
         PolygonBox[{{16.77550563432708, 6.791385758707179}, {
          18.018751538144375`, 6.558277151741437}, {17.63023719320147, 
          5.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.44794654332688, 7.2501437779883835`}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQLbP/DBDIOTCAgYpDfulF6QXJ8g4MHBHs6X4qDv+W
G59MElBwYBBIa5pdqeKQuTdQSrIMxBeT3Vat4mDhaLXj81UgfwWHfW6IisPb
CRkhB5QVHRhuZEnJc6o4vFvYEXokFMiX8L4nukTZwTBntqZgGpA/4dsueWNl
B/E9lmuDg4H8GXEfVh9VchAu/RXDpAjkb2B4tS5LyWHGhOsiT64Bzb8xS2+6
npKD7bf////XA/kM0dNnSSo5PIzWeC2tiZAPZuJ2+71XHq7fp6RnaUSkPNx8
cZGTa+4ay8Ptn7fiXUh5iDzcfRLb9cLrd8jD3e+4+qg7jxbCf1cs1j191onw
v73fGbO57xHh02ot/knfTxEefuunzQQCRXj4AgAUnYKF
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.25, 10.9452}, {0, -1}], 
         LineBox[{{8.799999999996544, 9.999999999996362}, {
          14.599999999993088`, 11.999999999996362`}}], 
         PolygonBox[{{12.267223788975762`, 11.195594409991642`}, {
          11.263172484351998`, 10.426256397357848`}, {11.002379937696475`, 
          11.182554782658867`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.205285559145075`, 10.139471878479277`}, {0, 1}], 
         LineBox[{{8.799999999996544, 10.}, {14.599999999993997`, 
          8.000000000003638}}], 
         PolygonBox[{{11.132776211024236`, 9.195594409991642}, {
          12.136827515648001`, 8.426256397357848}, {12.397620062303524`, 
          9.182554782658867}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.194714440854924`, 8.139471878479277}, {0, 1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         LineBox[CompressedData["
1:eJxF1A1MU1cUB/CnYoAGRAUUlk3E1/dayHuvfLoSKDl0bgsyQWBjG5qCXRop
okwKsjHGQGTWAobvkTrsB1YlkQ0EYY5JBw66iCKwDFvGujIVts5NYXHqhnO3
yXJp0jS/NO++e8/9nxMsz09TrCYIIhV9Xb9EWCT6sJDp7sJWIN5t5EU9YaD3
n4qymQRkYb9gw00Gjsg+ett0Enni/typfgak7/z1ufZnly3XFb0MWJx7ne/H
BAPRav3BNMLAmT5b9cRJZIftUpSTAaf+euYrduSwO0X6IBbijKRlgdwGxPoc
dYachW06SaNyL7L1k2lHJwu2pHXifVXIj2XCiKcsnE98Fn+oDVlo1C/v5EBT
LRtPPYP8YFzj1cDBkuH89JIWuS58v+EGB8n9v2mTK5HfWkwwP+Eg1dpxOiPb
tf70nQJfEfxtrEkYjEbuCn3DLVAEuzPG9g+4I0/8K/HjiYB/YSl3cMZ1npEU
4wIHdj8pSfcgd20XLvdwcL8z6fmHTch6+3xzAQdTWt2fNZXI4tb1NwUcsHWJ
xYUVyGry0jkrC/aiATlT53LDrVwNCwOWLJ1/l+v59Kl0KQulvz8cznW41v9x
JGotC5VetrhdQWg/4hof/i0GykOkaXeVyOUhydsHGAh4cYul4SvkrQvGxB4G
TG3fLIX6kkB8kdd318xAx9qYQFAiByyZIhwMJD8TSw9+ifz1Tyr3jSwQ+R4l
htV8VK/jbj5pLEi8NxYo4pHDYtlVn7LQ5BVt3XQA2Rrnds/JwqnxY3n+x5HV
N06oYzggY8On7PXIRC3RfJQDg60bNtcgl2c7uSEOkoa3HC0rRna0mOhFDjZc
mFyezEB+kO/7qo8Iuq8WXWtjkQOEZTWo/qM8V/5W/k+P10TLp0j8fImSW8N1
kHh9SJ/7sEJN4vfLhcnmX1Qk3l/OzDgbfIjE+1dGftDofI/E53vkMSb4tYnE
50/x+KNWO7RSnzWNj6pkT1fqJ9lkudLwMh/Xd8/y6Ly5hY/rP18c6/39PT6+
H5V9TnzwJQrf3zHrlc7nWih8v56egqsXHRS+f7UfT9IcRON8TO7SZztTaJyf
Tt2Otu/yaZyvrKo+dXUZjfN3uSTbS1VK43z6v957m3eAxvntJAZnba/RON85
Su/mBIrG+Y81R7XueEzh/qgvUA0mXKNw//SEZOzr0FG4v+JlforuUgr3n1b1
Zoi3gsL9uUpjCP9sD4X79/YYMVQvp3B/h/YtWk6XULj/c48M57W0U3g+fJs1
Kw6YpfD8MBZGmi4H03i+LMjoQtFhGs+fuM3t/h+P0ng+7bzouVvxggDPr9r2
8rOZhwV4vlkDyyqoQeT/51/EOlf+hPAfMSz6yA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.654799999999998`, 10.}, {1, 0}], 
         {PointSize[0.04], PointBox[{8.8, 10.}], PointBox[{7.7, 10.}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{14.6, 12.}], 
          PointBox[{14.6, 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["34",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000010914`, 15.000000000007276`}, {
          14.600000000011278`, 12.000000000007276`}}], 
         PolygonBox[{{17.824494365672923`, 13.791385758707179`}, {
          16.581248461855626`, 13.558277151741436`}, {16.96976280679853, 
          12.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.152053456673123`, 14.250143777988384`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {14.599999999998545`, 8.}}], 
         PolygonBox[{{16.77550563432708, 6.791385758707179}, {
          18.018751538144375`, 6.558277151741437}, {17.63023719320147, 
          5.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.44794654332688, 7.2501437779883835`}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQLbP/DBDIOTCAgYpDfulF6QXJ8g4MHBHs6X4qDv+W
G59MElBwYBBIa5pdqeKQuTdQSrIMxBeT3Vat4mDhaLXj81UgfwWHfW6IisPb
CRkhB5QVHRhuZEnJc6o4vFvYEXokFMiX8L4nukTZwTBntqZgGpA/4dsueWNl
B/E9lmuDg4H8GXEfVh9VchAu/RXDpAjkb2B4tS5LyWHGhOsiT64Bzb8xS2+6
npKD7bf////XA/kM0dNnSSo5PIzWeC2tiZAPZuJ2+71XHq7fp6RnaUSkPNx8
cZGTa+4ay8Ptn7fiXUh5iDzcfRLb9cLrd8jD3e+4+qg7jxbCf1cs1j191onw
v73fGbO57xHh02ot/knfTxEefuunzQQCRXj4AgAUnYKF
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.25, 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwt1Qs0lGkYB/CPiCbZKWqEI8ZgSmrWCjuRb7eTSzdUVEpn0pZZ1eYUuSVT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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.205285559145075`, 10.139471878479277`}, {0, 1}], LineBox[CompressedData["
1:eJwd1As0VHkcB/CrozBhJTW0lZFnSjvJq9nU3ewSTZGkOmv3zIplqUwqpo20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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.194714440854924`, 8.139471878479277}, {0, 1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         LineBox[{{14.599999999998545`, 12.000000000006455`}, {
          14.599999999998545`, 8.000000000003343}}], 
         PolygonBox[{{14.6, 10.6}, {14.2, 9.4}, {15., 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.654799999999998`, 10.}, {1, 0}], 
         {PointSize[0.04], PointBox[{8.8, 10.}], PointBox[{7.7, 10.}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{14.6, 12.}], 
          PointBox[{14.6, 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["35",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000010914`, 15.000000000007276`}, {
          14.600000000011278`, 12.000000000007276`}}], 
         PolygonBox[{{17.824494365672923`, 13.791385758707179`}, {
          16.581248461855626`, 13.558277151741436`}, {16.96976280679853, 
          12.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.152053456673123`, 14.250143777988384`}, {1, -1}], 
         LineBox[{{19.99999999999818, 5.}, {14.599999999998545`, 8.}}], 
         PolygonBox[{{16.77550563432708, 6.791385758707179}, {
          18.018751538144375`, 6.558277151741437}, {17.63023719320147, 
          5.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.44794654332688, 7.2501437779883835`}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQLbP/DBDIOTCAgYpDfulF6QXJ8g4MHBHs6X4qDv+W
G59MElBwYBBIa5pdqeKQuTdQSrIMxBeT3Vat4mDhaLXj81UgfwWHfW6IisPb
CRkhB5QVHRhuZEnJc6o4vFvYEXokFMiX8L4nukTZwTBntqZgGpA/4dsueWNl
B/E9lmuDg4H8GXEfVh9VchAu/RXDpAjkb2B4tS5LyWHGhOsiT64Bzb8xS2+6
npKD7bf////XA/kM0dNnSSo5PIzWeC2tiZAPZuJ2+71XHq7fp6RnaUSkPNx8
cZGTa+4ay8Ptn7fiXUh5iDzcfRLb9cLrd8jD3e+4+qg7jxbCf1cs1j191onw
v73fGbO57xHh02ot/knfTxEefuunzQQCRXj4AgAUnYKF
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.25, 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwt1Qs0lGkYB/CPiCbZKWqEI8ZgSmrWCjuRb7eTSzdUVEpn0pZZ1eYUuSVT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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.205285559145075`, 10.139471878479277`}, {0, 1}], LineBox[CompressedData["
1:eJwd1As0VHkcB/CrozBhJTW0lZFnSjvJq9nU3ewSTZGkOmv3zIplqUwqpo20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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.194714440854924`, 8.139471878479277}, {0, 1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         LineBox[{{14.599999999998545`, 12.000000000006455`}, {
          14.599999999998545`, 8.000000000003343}}], 
         PolygonBox[{{14.6, 10.6}, {14.2, 9.4}, {15., 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.654799999999998`, 10.}, {1, 0}], 
         {PointSize[0.04], PointBox[{8.8, 10.}], PointBox[{7.7, 10.}], 
          PointBox[{3.3000000000000003`, 10.}], PointBox[{14.6, 12.}], 
          PointBox[{14.6, 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["36",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 0}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxFlHtsDFscx8/qPpRW17ba4qa3MkcpaukQdP84v6tLVCI2VTRRSVVNBbu2
qD9QSnQb3MZqV2u29JF61KuESj2iEqqKUdamHvHIiiq92avFqlR7d++p5IyT
TCafzJzf+T2+3zMua2PqmiEIoWT6DL7Xpgp0jQH0a2EISY8qGQiMBmSvNbyK
xxCumplaKlL+dLR27GIMW9Pmf57EU35ShurWY/C/9djvSNF0/7PpGTsw2M9v
Na4UKJtia1btwjBCofb9CEQBak1Wr96CofrU2o2lImXrkdbDGRh0F4o36XnK
mam95UkYWlOq6x9KkYCuqu1xOgxJ4gd3jkB59vvkho8chBTEVAQhyp5/pnhv
cYCKTI8qxVGAqheunFrFQUdDbMocnjJUno7Yx4FhxSyVW4oANNExN7SAg0J9
SdgGgXKsyRhZyEFmhnK/ElFe8OB2TTkH9d49x4+K4TT+POHEFQ7ipmnLZvCD
bPS4X3Hg6zYeliQdrTfoY58KQ32n2ZUtULb+mb9Fj0F98w/zf4GRNN6tPZeX
YOha3rKtTKScuYm/bsFwMq+lX89T9jgsB/Mx9JvvvLsvaQFNW249tBNDcRLW
ZQuUre67tbkYzib2tQ8EwgBFx1sS0zFox2Udms5T9ux8EjwTw5k13yMFYQSg
npzPmhAM+p7Q2U4xlNZ/I//COw7Gd0S9eSSFAHqxrqPuOgfm4x6NAlH+FBFG
KjhwGNpcPD+c/t98sNPGgckwqIdhNP820+tdtN8G/88jYrD8fbPF2/xQGirv
317S0O0PaOT4ubYmRyKvkc/v1T6/IghqOT/70nWGClEl5/892d37WFLK9dkU
e+uUSPm7fue84Dl8kNyftAOkfYMwRO5fWwt5XiUq5P46Bkq+uiQk9994uic+
CCF5PtaEQmOO4CdsfienpP91TxogbL62pGLNBL6fsPk3bS/KLRL7CNPHZL1p
RVfgB2H6SSu4sXeR0EuYvi6NjsxrkHyE6W+ua1l2DP+NMH3q/o1JOCB+IUy/
57Qqhy/QTZi+jX/rm9olL2H6961qrmwUuwjzR8Exb2O50EmYf7RPFaWb+feE
+Utp2RGWgjyE+U97d2l6gvMlYf6sNvt3N/Juwvx78V6cLct5nzB/I9fV8Dzn
NcL8/+taiC6/ye6H/wEYu5+a
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.3000000000000003`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000010914`, 15.000000000007276`}, {
          14.600000000011278`, 12.000000000007276`}}], 
         PolygonBox[{{17.824494365672923`, 13.791385758707179`}, {
          16.581248461855626`, 13.558277151741436`}, {16.96976280679853, 
          12.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.152053456673123`, 14.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{13.699999999998909`, 8.5}, {12.799999999999272`, 
          8.999999999996362}}], 
         PolygonBox[{{12.725505634327078`, 9.041385758707179}, {
          13.968751538144375`, 8.808277151741436}, {13.58023719320147, 
          8.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.397946543326878`, 9.500143777988384}, {-1, -1}], 
         CircleBox[{15.725, 7.904999999999999}, 2.1106041789023346`, {
          2.8558085915452467`, 5.554772360439458}], 
         PolygonBox[{{14.175505634327079`, 6.351385758707179}, {
          15.03023719320147, 5.418951330844206}, {15.418751538144376`, 
          6.118277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 5.309856222011617}, {1, 1}], LineBox[CompressedData["
1:eJwtVWdUVOcWnSDFAJGABHFwZu69xEK5M3fIkhLU2SE8ASVigRFQQCcPpCkC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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.44794654332688, 9.690143777988382}, \
{-1, -1}], 
         LineBox[{{12.800000000001546`, 9.000000000003638}, {
          6.6000000000026375`, 10.000000000003638`}}], 
         PolygonBox[{{9.107655327572413, 9.595539463294772}, {
          10.356037647957438`, 9.799356984990286}, {10.22865169689774, 
          9.00956408842017}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.946810280178164, 10.425423737104602`}, {0, -1}], LineBox[CompressedData["
1:eJwV1A1ME2cYB/DzA3oycCNlIqPUK6CWrwHVOb5mL5QQN52wloUCW+gYyCko
RtlsSDUwGVYJrNnCBhti/WrEEMMQSkXBDql0Ah2bBrF+tE5wzLnCtNHCpt3/
fZPL5Ze797nn+efNiYoq5CWLKYrKxEXuFLMOK46lbnix1rKUuvVAlxTWVP6u
7YVnxIuGiFtX/HJuJ8xdTXVL4DsLQlckPDfgqwqBgx7cNTrWsFRz1tbzT2NZ
qtu47wM9PF5cL7oCOw0iCwdTC1uUh+HyUOFPKbAufnv5JlhbYz8cDHMh7YOL
4Uv3+x55V7OUZ9bt7othKcE3kjg3zOyL69kLd95ueJ04WzV/Lh4+laebfgkr
A29t/yeapfyPXyvmo56ZNx57Eb6jeNcVD3s23J9qhK0vlxyUw9blmld3wVpH
mlMNd54cac+Dx3+YaCb9Mz+fWC0n+10HE4bhucCa4lyyv0f/9iNSf8+MiYMr
92u28Uh+Le6/a+HagGmjAHa+4sg5G03mul4lhnWjkzcn4MfK1Ioo2NOwLZAm
86gTrwlJnq7k0lSYOT0go2H6jYK1u+DsDjrkIb43yXxkOAqvzDAsu0D66ZB9
PAyfYTZmfQEnGB0+f8JLOdtXMtL/j+ndFPIc/eOtEZKPtuD9x/4wfWwswwiL
6xVdxEmfBuwoh80y0bde7O9oz+2NgHU7p3qnYS4/rPVeJPr0G7p6Gf7P7+vC
Y3CSef18I1xpC+dxsN6i8ubEkL5t0hTYZEmd48OmhfzNwTA9v0Jiw/xnCtOf
eCNwvuIvFdbBm7QzVW64uuCdGCmc5NIaiBPGLkv/jcK5Wf78PHmfPdsU0Q/T
T4K4INRjG9yhh+COXH1oIqwMq+MXwPrSi2MK2FkyW5sCi5d8klJFnsukO9YQ
j1ifHif7c14kroLNGW37rWQeh6IvEk6T2Lm/4ISRI/IN8JTUt4JGHkmCK4Ny
uNN2oT4Mpva05qnh7DYJP4qcz99ivKfg3VX+FmLxgYotE3D5Z5l2IUy/l2xY
hvlUg54AUk/lV5SfRs5PU55+Gt9jfLz9ZTAdPKswwdUzmZubSF7NASU1pL/1
N/uN8NTSRZp0WM2ExZI8J9M1378g+fAsAjucvVHo0wOb7km/u0XyldHRZbCZ
O/3mKCxQnvgyHNY+U/t2wXOjA3V3wzEPG/prPews/nBvGyy2LAznw90eXlYp
vLvIfpSBqz9vHEqGJ1dayxyY77UWHj8Y1rVgkbzI/0TE/g9NjpNY
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.424314034854444`, 10.307331579087336`}, {-1, 1}], 
         LineBox[{{6.600000000002183, 10.}, {14.60000000000673, 12.}}], 
         PolygonBox[{{11.182085500087199`, 11.145521375021799`}, {
          10.114928749927333`, 10.466421624920068`}, {9.920900249898267, 
          11.242535625036334`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{10.975930218806315`, 10.101079124774735`}, {0, 1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 8.5}], 
          PointBox[{12.8, 9.}], PointBox[{6.6000000000000005`, 10.}], 
          PointBox[{14.6, 12.}], PointBox[{17.3, 6.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["37",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000009095`, 15.000000000003638`}, {
          10.000000000009095`, 10.000000000003638`}}], 
         PolygonBox[{{15.536656314599949`, 12.768328157299974`}, {
          14.284458247200067`, 12.589442719099992`}, {14.642229123600035`, 
          11.87390096630006}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.00700331526404, 12.881839773783103`}, {-1, 1}], 
         LineBox[{{20., 4.999999999998181}, {14.499999999998181`, 
          9.999999999996362}}], 
         PolygonBox[{{16.806035955962432`, 7.903603676397788}, {
          17.963033161636094`, 7.39237235296059}, {17.42489492643904, 
          6.8004202942438345`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.687842830694283`, 8.042107113763713}, {-1, -1}], 
         CircleBox[{10.1125, 10.112499999999999`}, 4.613871747242222, {
          3.1659780627625116`, 4.688003571211971}], 
         PolygonBox[{{7.274264068711928, 6.425735931288071}, {
          6.142893218813452, 6.991421356237309}, {6.70857864376269, 
          7.5571067811865476`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.358784489160851, 6.358784489160851}, \
{1, 1}], LineBox[{{10., 5.499999999994685}, {10., 9.999999999995518}}], 
         PolygonBox[{{10., 8.35}, {10.4, 7.15}, {9.6, 7.15}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {10.9452, 7.75}, {-1, 0}], 
         LineBox[CompressedData["
1:eJw9VgkwlQ0X5iYRt1Duirtwt/e9hFJZqieFVsSHFhVKyRKFyhYlSxJCC7pU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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.64121551083915, 6.358784489160851}, \
{-1, 1}], 
         CircleBox[{10.1125, 9.887500000000001}, 4.613871747242222, {
          1.595181735967615, 3.1172072444170746`}], 
         PolygonBox[{{6.425735931288071, 12.725735931288073`}, {
          6.991421356237309, 13.857106781186546`}, {7.5571067811865476`, 
          13.29142135623731}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.358784489160851, 13.64121551083915}, \
{1, -1}], LineBox[CompressedData["
1:eJxFlAtMU2cUx68QsNDF8pSXKOzeC+1tsXMDeczKwSFigLGxaUHBlE0CruB4
zKLMIQ/BaObQFZA4BwrySAcLDESQyUOklMkAIdgiG1S2IgICyxjMRXFfo/l6
k5svv5x7v++c8/3P3/WTzyPijAiCOIhe/frqoWDc7NVKXD2sar1HQvrR5f+U
OxCXcBYbKkhI7Z9SZ+cgtmg9b3qOhObwKrMrA4g/CAmwziJBG3+untlMA2Ev
7TmUT0JeU+QWm6OIa2Qj8isk3CI4475tiCXfV+Z1kfD4o9bxciM3IDqntf5/
kwD7j9+vCNBz2gZXIQUHlm6FGMsQEx6itRQKhO0tGrtSxMHGT8JaKEjjiPIm
WhC3yKbEaxREmUu/ZikRs9Lr4oGGG5OdmS97ELsk1a1l0KD+5o+ayVbEKtt0
4x9o8JxRmh6pQJxV1Dv3Kw1PFy8VPD6NWKuamXhEQ41ro5NNLOKShIYBHQ1T
E71nF/3057UO5I3RoNo1ZnvZFnHyyWLFbRq+aLJh8ZZRfReW/JrlNJxp6Fk9
ib4jhpbV2hgaNprrHjiqEFvw4o230ODRHTBqdgfx8aptKQ8psAp/PnNTH5c4
9f0ip0DclJg8/Rvi+t4n/e9TEC3duly2hlgTOeHMoSBw981Cbw/9+YT1Jg0J
3mPrBBHx+nrfPVZfS8LHcnHNvwp9vgnm7gUkXG3i1vz0D2LJzhFONglzUb0L
1YHuQEDzNV4uCUfCO+eVhYgviInIIhJqQ7I/XJhC7NLYeL2ZBKvy8lQ7ARfp
Q9EVqiPh59yhvaeSELNmwzSbKfDKyP/93nXEEgFrhwTlG1ToIBhCrN0Zb1RF
gU7ESUhe0H/vmJMxT8GoVn3b8SXimc/emBXSkK+ZD4gz4iHOu9+RSMNKuXC7
/zMUd/E0ultGQ7+fKBR0+v0Set9T0iBbX2AlVyHWbOz3mqBhv0lSqboSsUox
uG+aBqU5knOWIe5lIY/lHjD8b7sh5Wmit2H/zod3Gjscufj84tShvk5TQ34m
Iuv2yy/ccf5CV+4uHWGor0q9dKnA0lD/W4PS6RGhoT/i70686XnQ0L/Bi8Lm
rG+5uL8ZaTHnZ0a4uP+TfZk/Wjrz8P2QSn9rXykP39+1Nh/LyDYevt95iVoc
Zs7g+/fJqRGv38dgffRz10bqShisn+ChYKlqmMH6EirusyXr+Fh/nrHyAgeS
j/V5QmlsF+zNx/o9FNHOrIj4WN8ythNPp4+/1r/JYT9Wrhsfz4d97NzqAzYf
z8/dYbfZuFkGz5diu4/MoofB8xetDf1zrpTB81lGJyTbf8ng+RV7+D4rjmbw
fN/I3yR2CWLw/Ku9/LMJPwb7wzvsorRVXwb7h4Npd4brHgb7S7R4j/tMLIP9
5216d9Pzswz2p9COM1uj2xnsX5VdQVTMCwb727DNX5QykI/971RYtXbuIh/7
Y23dsfDRR3zsn6M9USufbhNgf3Xc68zN/EqA/demmm3i2434tT/Xs/WrB/wP
ZtIz8w==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.0548, 12.25}, {1, 0}], 
         CircleBox[{9.8875, 9.887500000000001}, 4.613871747242222, {
          0.024385409172718475`, 1.546410917622178}], 
         PolygonBox[{{12.725735931288073`, 13.574264068711928`}, {
          13.857106781186546`, 13.008578643762691`}, {13.29142135623731, 
          12.442893218813454`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.071696286134426`, 14.07242891873301}, {-1, -1}], 
         {PointSize[0.04], PointBox[{10., 14.5}], PointBox[{10., 5.5}], 
          PointBox[{5.5, 10.}], PointBox[{10., 10.}], PointBox[{14.5, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["38",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/1d1e1f2d2e2f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/1d1e1f2d2e2f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{5.5, 10.}, {15., 15.}}], 
         PolygonBox[{{10.780950933429189`, 12.779447859699573`}, {
          9.90534763970386, 11.866584851347634`}, {9.532750493437762, 
          12.57451942925322}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.132893029109436`, 13.266823244692073`}, {1, -1}], 
         LineBox[{{8., 8.650000000001455}, {5.5, 10.}}], 
         PolygonBox[{{6.2220567614056845`, 9.61008934884093}, {
          7.468002804488269, 9.391872810221946}, {7.087883672700362, 
          8.687948492096192}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.6183191821609295`, 8.565946633631352}, {1, 1}], 
         CircleBox[{9.971875000000002, 7.024375}, 2.55557188438322, {
          6.036367865651755, 8.735331634545966}], 
         PolygonBox[{{10.589552334308433`, 9.57573196579911}, {
          11.840935642890976`, 9.391233144661939}, {11.479959688492166`, 
          8.677302923739848}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.219590911647696`, 10.083439803036555`}, {-1, -1}], LineBox[CompressedData["
1:eJwlVGs0lAsX1pxCdKFjLu+MmXnfmfGOSkm5dMLxJN+Q3KL7DYmkQqJcOpKI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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.230409088352307, 4.966560196963445}, \
{1, 1}], LineBox[{{12.44999999999527, 6.400000000005093}, {
          14.999999999996362`, 5.000000000007276}}], 
         PolygonBox[{{13.199052869532245`, 5.988755287315631}, {
          14.05844360559067, 5.0606132923725315`}, {14.443450655344845`, 
          5.761876132996207}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.583848841101286`, 4.946103246291629}, {1, 1}], LineBox[CompressedData["
1:eJxFlws0VNsfxycVQ5FG8sxrOObMzDlHj3+I+OneolJXbl091FXoKY8krzxK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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.6452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{12.450000000000001`, 6.4}], 
          PointBox[{8., 8.65}], PointBox[{5.5, 10.}], PointBox[{15., 15.}], 
          PointBox[{15., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["39",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/de1d21212fef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/de1d21212fef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000001286`, 15.}, {14.500000000003151`, 15.}}], 
         PolygonBox[{{17.85, 15.}, {16.65, 15.4}, {16.65, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 15.9452}, {0, -1}], 
         LineBox[{{20.000000000001286`, 5.}, {14.500000000003151`, 5.}}], 
         PolygonBox[{{16.65, 5.}, {17.85, 5.4}, {17.85, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999985448, 9.999999999992724}, {
          14.499999999989086`, 14.999999999992724`}}], 
         PolygonBox[{{10.524494365672922`, 12.791385758707179`}, {
          9.66976280679853, 11.858951330844206`}, {9.281248461855625, 
          12.558277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 13.250143777988384`}, {1, -1}], 
         LineBox[{{5.499999999987267, 10.000000000007276`}, {
          14.499999999990905`, 5.000000000003638}}], 
         PolygonBox[{{9.475505634327078, 7.791385758707179}, {
          10.33023719320147, 6.858951330844206}, {10.718751538144375`, 
          7.558277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 6.7498562220116165`}, {1, 1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6Dis4QbSeA8OF27d22es63Pu197zTbl0HBg6X
P27Zug5/c/qm5WUC+S/KVdjqdR0Onlps+JAfyF/xMjCkUtfhzt7qC3O26Dgw
MFStWBWl63BYc5HciVggPyEgYaqqrsP1Kzkp2oJA/oX5k2rv6TgcS37cuPWy
tgPDD9U0724dBz1T4/9CK4F8jqqMF0Y6Dqu2zHodPxnIz1BtW3NH28H2di37
70lA/oS9Nw/2aztsnJz+oXYZkP/A0qcnSNuB/cm5j7dOgdRvWKeope2gOFnh
79zfQH7FrGl3JbUdFj+ISvQ2Btr/Qvq7pLK2g5H2heQr2UC+wYYUIVdth09a
QTwS80H826lfarQdbnScLkk9DuRb6JVJH9N2qJLzPfPlPsj9IQeKpXQcdu4I
Y/J9CuRPuBHzu0jHYWvChhW9V4H8CIWo8BM6Dj86BZ5P2wDka5zVKRDWdTih
cVabpwIUPiadHH66Dm63V8QY6oP075BxKtZ1mDHjY8bSW0D3BuyQY23QdfCS
AUZHgzZc/m0F3/QoVW24fjuz8MXnLmjBzd/7wujK53YtuP297e8lMv214O67
sME/vEhLC+7+MxEr/26S0oL7j/vzJmNuBS24/9d0mHTPstCCh8/kZ2s/P0/Q
godf8DP9OdpTtODh+8hg6bwtZ7Tg4Z9Sd58piQURP7W6hnyqpoj4u1I7g78p
EhG/7mZNP1hyEfF/OyZVZEEOIn2E3znD8DZMG55+9q+VqfPS0Yanr7U2k2df
eacFT38Hn8c+NF+kBU+fK4Jb9JO9teDp94hbzDm5D5rw9N2xpDBz4jRNePpf
JuZksNcFyIfmjzvsIFrTAQBrWzbb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.4452, 13.75}, {-1, 0}], 
         CircleBox[{13.9375, 10.}, 2.5625, {4.93370342273248, 
          7.632667191626691}], 
         PolygonBox[{{16.5, 9.4}, {16.9, 10.6}, {16.1, 10.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.4452, 10.}, {-1, 0}], 
         CircleBox[{15.0625, 10.}, 2.5625, {1.7921107691426879`, 
          4.491074538036899}], 
         PolygonBox[{{12.5, 10.6}, {12.1, 9.4}, {12.9, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5548, 10.}, {1, 0}], 
         LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6DvuFQbSIA4OBiqrrAR2HEMOWhKpOIP/Hbuad
U3QcxHwfX1LSFXVgkOi4/bNBx6E+a8cc2wdAfoTYutVtOg4Zlzjij00Tc2Bg
aAsPXarjkNm6+1C+i7gDw4L187Nu6TgEhamZr7wL5Bsk5DYp6jqkClr+mRIh
4cDAcWvmlhJdh617ZjLeWgvk/2idLn5W12Hv3lMaDjeB/Bm3KoOV9Ry27mP5
yHkLyC9wUrMr0HNY999AnmUzkK9wbHPxWj0HhbVJMl5ZIPX+gfeu6jmcZ9jP
eoQDyO9IzVR6pudwOtXLNnMm0H6JJxzP7ug5rLkQbxGuAeRf8J/9dpeeg82P
6OfnDwLdf0FlFneznkO+TaaYaSGQf+Ji5xNLPQeew6npYg5AvsFq+56nug79
fUG/U02B/AL1xSy9ug7be+avexIC5K+4v3SFua4Dw9F7hpfmAfk3jC73v9Fx
aLyauuI5P9A+hjNl3zfqOJR4rnObuxDIL3B/vLdHx+HhVxm9rV5A925wf/ir
XsehRxQUH5JweTPTxxe9tkrC9S95GLgzKU4Kbv600FnhEh+k4Paz2yue002U
hrsvYV5CoPRqabj7O+asZZl1ShruP+6MWTNe7peG+3/Fu2kvT/VJw8PnttES
CQM7aXj4ZUctOL7iuhQ8fH2Obj00LU0KHv6WzI6/rL9LwuPne+3cx1unSsLj
L1P7/IPfPpLw+E3K+c33TVUSHv/RuabX96tJwtNHUl7aP9UgSXj62SR5nDl7
uSQ8fR29cvNEmJIUPP2tW+piO2G3FDx9ro/4vsAyUxqefmcccE8+JC8DT9+M
92R0716Sgaf/wz+r57eWyTrA8keEOIiWcwAAvAU6vA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.1452, 6.25}, {-1, 0}], 
         {PointSize[0.04], PointBox[{14.5, 7.5}], PointBox[{14.5, 12.5}], 
          PointBox[{5.5, 10.}], PointBox[{14.5, 15.}], PointBox[{14.5, 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["40",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/dedf1e21212f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/dedf1e21212f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000001286`, 15.}, {14.500000000003151`, 15.}}], 
         PolygonBox[{{17.85, 15.}, {16.65, 15.4}, {16.65, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 15.9452}, {0, -1}], 
         LineBox[{{20.000000000001286`, 5.}, {14.500000000003151`, 5.}}], 
         PolygonBox[{{16.65, 5.}, {17.85, 5.4}, {17.85, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999985448, 9.999999999992724}, {
          14.499999999989086`, 14.999999999992724`}}], 
         PolygonBox[{{10.524494365672922`, 12.791385758707179`}, {
          9.66976280679853, 11.858951330844206`}, {9.281248461855625, 
          12.558277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 13.250143777988384`}, {1, -1}], 
         LineBox[{{5.499999999987267, 10.000000000007276`}, {
          14.499999999990905`, 5.000000000003638}}], 
         PolygonBox[{{9.475505634327078, 7.791385758707179}, {
          10.33023719320147, 6.858951330844206}, {10.718751538144375`, 
          7.558277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 6.7498562220116165`}, {1, 1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6Dis4QbSeA8OF27d22es63Pu197zTbl0HBg6X
P27Zug5/c/qm5WUC+S/KVdjqdR0Onlps+JAfyF/xMjCkUtfhzt7qC3O26Dgw
MFStWBWl63BYc5HciVggPyEgYaqqrsP1Kzkp2oJA/oX5k2rv6TgcS37cuPWy
tgPDD9U0724dBz1T4/9CK4F8jqqMF0Y6Dqu2zHodPxnIz1BtW3NH28H2di37
70lA/oS9Nw/2aztsnJz+oXYZkP/A0qcnSNuB/cm5j7dOgdRvWKeope2gOFnh
79zfQH7FrGl3JbUdFj+ISvQ2Btr/Qvq7pLK2g5H2heQr2UC+wYYUIVdth09a
QTwS80H826lfarQdbnScLkk9DuRb6JVJH9N2qJLzPfPlPsj9IQeKpXQcdu4I
Y/J9CuRPuBHzu0jHYWvChhW9V4H8CIWo8BM6Dj86BZ5P2wDka5zVKRDWdTih
cVabpwIUPiadHH66Dm63V8QY6oP075BxKtZ1mDHjY8bSW0D3BuyQY23QdfCS
AUZHgzZc/m0F3/QoVW24fjuz8MXnLmjBzd/7wujK53YtuP297e8lMv214O67
sME/vEhLC+7+MxEr/26S0oL7j/vzJmNuBS24/9d0mHTPstCCh8/kZ2s/P0/Q
godf8DP9OdpTtODh+8hg6bwtZ7Tg4Z9Sd58piQURP7W6hnyqpoj4u1I7g78p
EhG/7mZNP1hyEfF/OyZVZEEOIn2E3znD8DZMG55+9q+VqfPS0Yanr7U2k2df
eacFT38Hn8c+NF+kBU+fK4Jb9JO9teDp94hbzDm5D5rw9N2xpDBz4jRNePpf
JuZksNcFyIfmjzvsIFrTAQBrWzbb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.4452, 13.75}, {-1, 0}], 
         CircleBox[{13.9375, 10.}, 2.5625, {4.93370342273248, 
          7.632667191626691}], 
         PolygonBox[{{16.5, 9.4}, {16.9, 10.6}, {16.1, 10.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.4452, 10.}, {-1, 0}], 
         CircleBox[{15.0625, 10.}, 2.5625, {1.7921107691426879`, 
          4.491074538036899}], 
         PolygonBox[{{12.5, 10.6}, {12.1, 9.4}, {12.9, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5548, 10.}, {1, 0}], 
         LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6DvuFQbSIA4OBiqrrAR2HEMOWhKpOIP/Hbuad
U3QcxHwfX1LSFXVgkOi4/bNBx6E+a8cc2wdAfoTYutVtOg4Zlzjij00Tc2Bg
aAsPXarjkNm6+1C+i7gDw4L187Nu6TgEhamZr7wL5Bsk5DYp6jqkClr+mRIh
4cDAcWvmlhJdh617ZjLeWgvk/2idLn5W12Hv3lMaDjeB/Bm3KoOV9Ry27mP5
yHkLyC9wUrMr0HNY999AnmUzkK9wbHPxWj0HhbVJMl5ZIPX+gfeu6jmcZ9jP
eoQDyO9IzVR6pudwOtXLNnMm0H6JJxzP7ug5rLkQbxGuAeRf8J/9dpeeg82P
6OfnDwLdf0FlFneznkO+TaaYaSGQf+Ji5xNLPQeew6npYg5AvsFq+56nug79
fUG/U02B/AL1xSy9ug7be+avexIC5K+4v3SFua4Dw9F7hpfmAfk3jC73v9Fx
aLyauuI5P9A+hjNl3zfqOJR4rnObuxDIL3B/vLdHx+HhVxm9rV5A925wf/ir
XsehRxQUH5JweTPTxxe9tkrC9S95GLgzKU4Kbv600FnhEh+k4Paz2yue002U
hrsvYV5CoPRqabj7O+asZZl1ShruP+6MWTNe7peG+3/Fu2kvT/VJw8PnttES
CQM7aXj4ZUctOL7iuhQ8fH2Obj00LU0KHv6WzI6/rL9LwuPne+3cx1unSsLj
L1P7/IPfPpLw+E3K+c33TVUSHv/RuabX96tJwtNHUl7aP9UgSXj62SR5nDl7
uSQ8fR29cvNEmJIUPP2tW+piO2G3FDx9ro/4vsAyUxqefmcccE8+JC8DT9+M
92R0716Sgaf/wz+r57eWyTrA8keEOIiWcwAAvAU6vA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.1452, 6.25}, {-1, 0}], 
         {PointSize[0.04], PointBox[{14.5, 7.5}], PointBox[{14.5, 12.5}], 
          PointBox[{5.5, 10.}], PointBox[{14.5, 15.}], PointBox[{14.5, 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["41",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/dedf1e21212f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/dedf1e21212f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000001286`, 15.}, {14.500000000003151`, 15.}}], 
         PolygonBox[{{17.85, 15.}, {16.65, 15.4}, {16.65, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 15.9452}, {0, -1}], 
         LineBox[{{20.000000000001286`, 5.}, {14.500000000003151`, 5.}}], 
         PolygonBox[{{16.65, 5.}, {17.85, 5.4}, {17.85, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999985448, 9.999999999992724}, {
          14.499999999989086`, 14.999999999992724`}}], 
         PolygonBox[{{10.524494365672922`, 12.791385758707179`}, {
          9.66976280679853, 11.858951330844206`}, {9.281248461855625, 
          12.558277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 13.250143777988384`}, {1, -1}], 
         LineBox[{{5.499999999987267, 10.000000000007276`}, {
          14.499999999990905`, 5.000000000003638}}], 
         PolygonBox[{{9.475505634327078, 7.791385758707179}, {
          10.33023719320147, 6.858951330844206}, {10.718751538144375`, 
          7.558277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 6.7498562220116165`}, {1, 1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6Dis4QbSeA8OF27d22es63Pu197zTbl0HBg6X
P27Zug5/c/qm5WUC+S/KVdjqdR0Onlps+JAfyF/xMjCkUtfhzt7qC3O26Dgw
MFStWBWl63BYc5HciVggPyEgYaqqrsP1Kzkp2oJA/oX5k2rv6TgcS37cuPWy
tgPDD9U0724dBz1T4/9CK4F8jqqMF0Y6Dqu2zHodPxnIz1BtW3NH28H2di37
70lA/oS9Nw/2aztsnJz+oXYZkP/A0qcnSNuB/cm5j7dOgdRvWKeope2gOFnh
79zfQH7FrGl3JbUdFj+ISvQ2Btr/Qvq7pLK2g5H2heQr2UC+wYYUIVdth09a
QTwS80H826lfarQdbnScLkk9DuRb6JVJH9N2qJLzPfPlPsj9IQeKpXQcdu4I
Y/J9CuRPuBHzu0jHYWvChhW9V4H8CIWo8BM6Dj86BZ5P2wDka5zVKRDWdTih
cVabpwIUPiadHH66Dm63V8QY6oP075BxKtZ1mDHjY8bSW0D3BuyQY23QdfCS
AUZHgzZc/m0F3/QoVW24fjuz8MXnLmjBzd/7wujK53YtuP297e8lMv214O67
sME/vEhLC+7+MxEr/26S0oL7j/vzJmNuBS24/9d0mHTPstCCh8/kZ2s/P0/Q
godf8DP9OdpTtODh+8hg6bwtZ7Tg4Z9Sd58piQURP7W6hnyqpoj4u1I7g78p
EhG/7mZNP1hyEfF/OyZVZEEOIn2E3znD8DZMG55+9q+VqfPS0Yanr7U2k2df
eacFT38Hn8c+NF+kBU+fK4Jb9JO9teDp94hbzDm5D5rw9N2xpDBz4jRNePpf
JuZksNcFyIfmjzvsIFrTAQBrWzbb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.4452, 13.75}, {-1, 0}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{13.9375, 10.}, 2.5625, {4.93370342273248, 
           7.632667191626691}]}, 
         PolygonBox[{{16.5, 9.4}, {16.9, 10.6}, {16.1, 10.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.4452, 10.}, {-1, 0}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{15.0625, 10.}, 2.5625, {1.7921107691426879`, 
           4.491074538036899}]}, 
         PolygonBox[{{12.5, 10.6}, {12.1, 9.4}, {12.9, 9.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5548, 10.}, {1, 0}], 
         LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6DvuFQbSIA4OBiqrrAR2HEMOWhKpOIP/Hbuad
U3QcxHwfX1LSFXVgkOi4/bNBx6E+a8cc2wdAfoTYutVtOg4Zlzjij00Tc2Bg
aAsPXarjkNm6+1C+i7gDw4L187Nu6TgEhamZr7wL5Bsk5DYp6jqkClr+mRIh
4cDAcWvmlhJdh617ZjLeWgvk/2idLn5W12Hv3lMaDjeB/Bm3KoOV9Ry27mP5
yHkLyC9wUrMr0HNY999AnmUzkK9wbHPxWj0HhbVJMl5ZIPX+gfeu6jmcZ9jP
eoQDyO9IzVR6pudwOtXLNnMm0H6JJxzP7ug5rLkQbxGuAeRf8J/9dpeeg82P
6OfnDwLdf0FlFneznkO+TaaYaSGQf+Ji5xNLPQeew6npYg5AvsFq+56nug79
fUG/U02B/AL1xSy9ug7be+avexIC5K+4v3SFua4Dw9F7hpfmAfk3jC73v9Fx
aLyauuI5P9A+hjNl3zfqOJR4rnObuxDIL3B/vLdHx+HhVxm9rV5A925wf/ir
XsehRxQUH5JweTPTxxe9tkrC9S95GLgzKU4Kbv600FnhEh+k4Paz2yue002U
hrsvYV5CoPRqabj7O+asZZl1ShruP+6MWTNe7peG+3/Fu2kvT/VJw8PnttES
CQM7aXj4ZUctOL7iuhQ8fH2Obj00LU0KHv6WzI6/rL9LwuPne+3cx1unSsLj
L1P7/IPfPpLw+E3K+c33TVUSHv/RuabX96tJwtNHUl7aP9UgSXj62SR5nDl7
uSQ8fR29cvNEmJIUPP2tW+piO2G3FDx9ro/4vsAyUxqefmcccE8+JC8DT9+M
92R0716Sgaf/wz+r57eWyTrA8keEOIiWcwAAvAU6vA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.1452, 6.25}, {-1, 0}], 
         {PointSize[0.04], PointBox[{14.5, 7.5}], PointBox[{14.5, 12.5}], 
          PointBox[{5.5, 10.}], PointBox[{14.5, 15.}], PointBox[{14.5, 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["42",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/dedf1e21212f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/dedf1e21212f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000001286`, 15.}, {14.500000000003151`, 15.}}], 
         PolygonBox[{{17.85, 15.}, {16.65, 15.4}, {16.65, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 15.9452}, {0, -1}], 
         LineBox[{{20.000000000001286`, 5.}, {14.500000000003151`, 5.}}], 
         PolygonBox[{{16.65, 5.}, {17.85, 5.4}, {17.85, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.25, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999985448, 9.999999999992724}, {
          14.499999999989086`, 14.999999999992724`}}], 
         PolygonBox[{{10.524494365672922`, 12.791385758707179`}, {
          9.66976280679853, 11.858951330844206`}, {9.281248461855625, 
          12.558277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{9.852053456673122, 13.250143777988384`}, {1, -1}], 
         LineBox[{{5.499999999987267, 10.000000000007276`}, {
          14.499999999990905`, 5.000000000003638}}], 
         PolygonBox[{{9.475505634327078, 7.791385758707179}, {
          10.33023719320147, 6.858951330844206}, {10.718751538144375`, 
          7.558277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.852053456673122, 6.7498562220116165`}, {1, 1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6Dis4QbSeA8OF27d22es63Pu197zTbl0HBg6X
P27Zug5/c/qm5WUC+S/KVdjqdR0Onlps+JAfyF/xMjCkUtfhzt7qC3O26Dgw
MFStWBWl63BYc5HciVggPyEgYaqqrsP1Kzkp2oJA/oX5k2rv6TgcS37cuPWy
tgPDD9U0724dBz1T4/9CK4F8jqqMF0Y6Dqu2zHodPxnIz1BtW3NH28H2di37
70lA/oS9Nw/2aztsnJz+oXYZkP/A0qcnSNuB/cm5j7dOgdRvWKeope2gOFnh
79zfQH7FrGl3JbUdFj+ISvQ2Btr/Qvq7pLK2g5H2heQr2UC+wYYUIVdth09a
QTwS80H826lfarQdbnScLkk9DuRb6JVJH9N2qJLzPfPlPsj9IQeKpXQcdu4I
Y/J9CuRPuBHzu0jHYWvChhW9V4H8CIWo8BM6Dj86BZ5P2wDka5zVKRDWdTih
cVabpwIUPiadHH66Dm63V8QY6oP075BxKtZ1mDHjY8bSW0D3BuyQY23QdfCS
AUZHgzZc/m0F3/QoVW24fjuz8MXnLmjBzd/7wujK53YtuP297e8lMv214O67
sME/vEhLC+7+MxEr/26S0oL7j/vzJmNuBS24/9d0mHTPstCCh8/kZ2s/P0/Q
godf8DP9OdpTtODh+8hg6bwtZ7Tg4Z9Sd58piQURP7W6hnyqpoj4u1I7g78p
EhG/7mZNP1hyEfF/OyZVZEEOIn2E3znD8DZMG55+9q+VqfPS0Yanr7U2k2df
eacFT38Hn8c+NF+kBU+fK4Jb9JO9teDp94hbzDm5D5rw9N2xpDBz4jRNePpf
JuZksNcFyIfmjzvsIFrTAQBrWzbb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.4452, 13.75}, {-1, 0}], 
         LineBox[CompressedData["
1:eJw9lgtUTukax8MISQmp1Ke993erSXtvRXL46l9mTnKZbqaJCemidDTdMROR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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.4452, 10.}, {-1, 0}], 
         LineBox[CompressedData["
1:eJxFlglUFFcWhhsTEQmCYEARGrp6ryqkURDjEfBvJmFzIUhCREWIIhADgy2b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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5548, 10.}, {1, 0}], 
         LineBox[CompressedData["
1:eJxTTMoPSmViYGAwBmIQDQG6DvuFQbSIA4OBiqrrAR2HEMOWhKpOIP/Hbuad
U3QcxHwfX1LSFXVgkOi4/bNBx6E+a8cc2wdAfoTYutVtOg4Zlzjij00Tc2Bg
aAsPXarjkNm6+1C+i7gDw4L187Nu6TgEhamZr7wL5Bsk5DYp6jqkClr+mRIh
4cDAcWvmlhJdh617ZjLeWgvk/2idLn5W12Hv3lMaDjeB/Bm3KoOV9Ry27mP5
yHkLyC9wUrMr0HNY999AnmUzkK9wbHPxWj0HhbVJMl5ZIPX+gfeu6jmcZ9jP
eoQDyO9IzVR6pudwOtXLNnMm0H6JJxzP7ug5rLkQbxGuAeRf8J/9dpeeg82P
6OfnDwLdf0FlFneznkO+TaaYaSGQf+Ji5xNLPQeew6npYg5AvsFq+56nug79
fUG/U02B/AL1xSy9ug7be+avexIC5K+4v3SFua4Dw9F7hpfmAfk3jC73v9Fx
aLyauuI5P9A+hjNl3zfqOJR4rnObuxDIL3B/vLdHx+HhVxm9rV5A925wf/ir
XsehRxQUH5JweTPTxxe9tkrC9S95GLgzKU4Kbv600FnhEh+k4Paz2yue002U
hrsvYV5CoPRqabj7O+asZZl1ShruP+6MWTNe7peG+3/Fu2kvT/VJw8PnttES
CQM7aXj4ZUctOL7iuhQ8fH2Obj00LU0KHv6WzI6/rL9LwuPne+3cx1unSsLj
L1P7/IPfPpLw+E3K+c33TVUSHv/RuabX96tJwtNHUl7aP9UgSXj62SR5nDl7
uSQ8fR29cvNEmJIUPP2tW+piO2G3FDx9ro/4vsAyUxqefmcccE8+JC8DT9+M
92R0716Sgaf/wz+r57eWyTrA8keEOIiWcwAAvAU6vA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.1452, 6.25}, {-1, 0}], 
         {PointSize[0.04], PointBox[{14.5, 7.5}], PointBox[{14.5, 12.5}], 
          PointBox[{5.5, 10.}], PointBox[{14.5, 15.}], PointBox[{14.5, 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["43",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/dedf1e21212f.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/dedf1e21212f.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9kwtMzVEcx0/3RglRXWaGxf8Ml0zdf+Vq6fxkrWuxJUuIPNI/V7MebELp
qnmOZrSrf00o1lWxSu8mYVb0Z0UaKvJqUiPyuOnq77Cd/287O/vsnJ3ze3y/
s7fHh8WoEEJ6uv7txyplGlMA/Q8M0/ev0ptFyqbQZ21eGNp7An0W8ZR3bjQ1
RWKAb3Vr7ksaylE+gfswrM5QV0QJlA31Q/1pGFQJhpBh2Q1QguNlYS8GTWJg
R5ZIefD9iox1GMrL0r568pSPO1msCzAMrI9e8EhyBfTRb/SKlYPsaRFuRoFy
Y5ApuYnyx2vh9oiyZelL74scVDnxDXmiC6Dmgn7+KAep/u8C9Tzl5x4hsSYO
rJkSapMmU76uqs3kYNbhA7xRoDx4oaD3BgfuqZfm2SHKjl1Huzo5MFQXRej4
STSfu0/SnTGYk0/YRwvOgKzlXzoDMKj9HhZniRNpfmV+wTEYTjn3GZukCfR8
rHPlQQzbpMdzRuTxgFodj+cewrAr//zpxTxl986kI7sx+MYHGHYIToD0g3dw
CIa1J5s/Z4vjAEFW3YcZGPQ/XL0eSY6AesYPuX3iwNWu+y1ClLdaKm7WcXA2
WFZ58w6Aatr9AswcaLT5NYIwlrL0Zk86B8sMm4JFcYzCjVrZ1iLZK/dJwuyZ
sqxW3pue0v9Yx6uV/0iGr0usoFLy2ZzXWJ0r2in5zo+LTWqVkFKPbrlntz1C
Sr0TRhd6JAqjhPXDvdBo6ZBshPXrt7b/51J+hLB+niuxhOaKw4T1u2BuyrYR
+Rdh8yieqtFFCj8Jm9eK1y22euk7YfP01CZ+n8UPETbvkEEvhyPiV8L0cNJU
FNYtfyFML+loeUy1NECYngb+vCoxi32E6a1miW1PstBLmB5L/XtXb+DfEaZX
fVDVVR/UQ5ieW8Prb2tyXhCm9zOGle+L+KeE+aG0IStiQ84DwvxyKe7eQGJO
LWF++ucy2FJ4i/ntL+8Xb1k=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.75, 9.0548}, {0, 1}], 
         LineBox[{{19.999999999998543`, 15.}, {14.999999999997012`, 15.}}], 
         PolygonBox[{{18.1, 15.}, {16.9, 15.4}, {16.9, 14.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 15.9452}, {0, -1}], 
         LineBox[{{19.999999999998543`, 5.}, {14.999999999997012`, 5.}}], 
         PolygonBox[{{16.9, 5.}, {18.1, 5.4}, {18.1, 4.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.5, 4.0548}, {0, 1}], 
         LineBox[{{5.499999999994543, 10.}, {14.999999999994543`, 5.}}], 
         PolygonBox[{{9.719049066570811, 7.779447859699573}, {
          10.59465236029614, 6.866584851347634}, {10.967249506562238`, 
          7.57451942925322}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.132893029109436`, 6.733176755307927}, {1, 1}], 
         LineBox[{{8.10000000000764, 11.35000000000582}, {5.500000000007276, 
          10.000000000007276`}}], 
         PolygonBox[{{7.332497646081044, 10.951489162388235`}, {
          6.083176245660133, 10.753509268332463`}, {6.451828462177779, 
          10.04351240689107}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.6905363304970615`, 11.445818919042697`}, {1, -1}], 
         CircleBox[{10.428125, 12.008125}, 2.419358289144045, {
          0.7181252438511612, 3.417089012745372}], 
         PolygonBox[{{9.802464247741284, 14.420974592148887`}, {
          8.556886024159462, 14.200668239678635`}, {8.938185480357978, 
          13.49738257602359}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.141032303615376, 14.892815973331647`}, {1, -1}], LineBox[CompressedData["
1:eJwlVgs0VlkbdokPlXtIruc7Z5/zfVRClKGeKDXllntTpNxDYqRUUk0XFUM3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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.20896769638463, 10.057184026668352`}, {-1, 1}], 
         LineBox[{{12.249999999998181`, 13.599999999998545`}, {
          14.999999999992724`, 14.999999999996362`}}], 
         PolygonBox[{{14.159697869360896`, 14.572209824401911`}, {
          13.271775346907045`, 13.671324929357493`}, {12.908828914371163`, 
          14.384255421838688`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.526591286961732`, 15.076502829182314`}, {1, -1}], LineBox[CompressedData["

1:eJxFlws0VNsfxycVQ5FG8sxrOObMzDlHj3+I+OneolJXbl091FXoKY8krzxK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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.6452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{12.25, 13.6}], PointBox[{8.1, 11.35}], 
          PointBox[{5.5, 10.}], PointBox[{15., 15.}], PointBox[{15., 5.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["44",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ad/becf/df1d21212eef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ad/becf/df1d21212eef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20.00000000000182, 5.000000000003638}, {
          18.200000000000728`, 6.000000000003638}}], 
         PolygonBox[{{18.57550563432708, 5.791385758707179}, {
          19.818751538144376`, 5.558277151741437}, {19.43023719320147, 
          4.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.24794654332688, 6.2501437779883835`}, {-1, -1}], 
         LineBox[{{11.000000000007276`, 9.999999999992724}, {
          12.799999999999272`, 8.999999999996362}}], 
         PolygonBox[{{11.375505634327078`, 9.791385758707179}, {
          12.23023719320147, 8.858951330844206}, {12.618751538144375`, 
          9.558277151741436}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.75205345667312, 8.749856222011616}, \
{1, 1}], LineBox[{{15.799999999997453`, 7.333000000002357}, {
          15.19999999999709, 7.667000000001281}}], 
         PolygonBox[{{14.975753109471105`, 7.791830769061085}, {
          16.21880073656962, 7.5576671579581784`}, {15.829693044488174`, 
          6.858671303919651}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.6490961534078, 8.249504467199646}, \
{-1, -1}], 
         CircleBox[{17.1499625, 6.9365}, 1.4069864965259078`, {
          2.8559147298784766`, 5.554878498772688}], 
         PolygonBox[{{15.942274710082684`, 5.997830088108243}, {
          16.797105231178485`, 5.065486385280214}, {17.18554534865614, 
          5.764853438503301}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.318997272387037`, 4.956276334380269}, {1, 1}], LineBox[CompressedData["
1:eJwdkg0w1HkcxjeXUndXXrP2/9/9092iUfn/VnuXIfO07ijV0aJJ5GVvVXTu
ToQwHemqo+ROXlKKjXNc8pbiutY6ZVW4SEUvaknoSpFWhe72ema+851n5jvP
d+Yzj5XsO2moHofDWaub/7fDe4nAeS8uIvqtM3Z5ijBPm5Gg9eZC/fB+3rUt
ImzMLdpuq+ai84P+17xYEaYvpclEPhYIaK+717xbhNTI+jy3MQtsPDtt9E+i
CGMhf+jzs3kosB9e/XOkCAv1q4fCHSjwU/1O1W4W4XqK95tsJYWihBIO7SrC
VNvep6uX0rguVXw741Ndvm+t6b5EGrK3tXFKPREema1sWVBCI4By+zFSQ1Aj
Vh/yq6BR7f7Mo6mR4PFvfo/TjtII77421VZEMBlwp+budt3907JOYRoBx0LR
ELWExoVhO/O30QRzYj66UjpAQVVjQiQygv3+veP+xyisD/r+xbCU4HeJojLe
h8L59ec6x9wIbBRzSs+ZUbj0PPxViguB5fLhE1/18WCYohQPOhJsuSe7zPmT
h2bL0eIYnU+IMzawPcVD6sGPDZJXECRprErn5vDgvP3t10t0ea9shMa+Oj4x
OyMK5d4ELXONm8pP8iBPftl6SU4QcKHS+YtqHpZ5xzXl7yIwztkT79jOg11s
vM14BsHVR9ukzeM8nLbZ+25fKYG5vyrH0JbCxKj1uls6Hovu54dUhFOoDD8S
s7ibwPGX8g0jOt7TylUrK54QnLEspEeENFSa+et2TxAkp/usaz1Bw0pctapv
iuAT7cKf9K35cCnXC059Q3A2c+CZay0fiYKgO21PCfS53ZZKVwHMy6PrNLcI
4g9MzV+iFiC3A3kv6ggmSsYPicUMVjg5uhlkEaTIOd2TexlYa5Jn3Y0gGJCJ
N+ypZGDhPzEzWMdjSl61YHctgzXPRhq2WhH0Th1duCOLgVZcpXCeQRDtNabl
+jKoslUMygZZWMjrJT6TAhgW511O6mKRkbBvfPywAB6Xh0e/aWHxYpt6cRlP
AKG6OM/vLxbD0qUdHgo+srwCe+c1skhStlYfWMxHh3RDg20Ti/oz4R6TF2k8
nFQd8VSzKJhRbBqwicYWx8IPQ1tZlJaFxV6dqetP5OrBwg4W24ZkNdoGCkMu
GiPFLRbLgo/tEqZTsIsunB14lwV7MTPXKJJCsGXvoPFDFlfqvU1VYRSkvo2h
/AEWFWt7e4LiKKz8bPPB9icshmYXFLgfp9AZW1QqGdX9t1/0g1kXBe51hYnk
NYvKQZMbfXwaW+fsFJJ/WeRUDCijYmk81uffNNIncNnvVbaql4bTwIPmbAMC
17DT+Xs8+QiNZte46XxQF//XfLXOF2b1lOkR3K4y7+BKBEjLcg48pGWR221o
41AvgLrobzj0s0i39Q56ImTQaOqema3jGdJulhmYxGD5Ml5bWgkLp8M9dsUq
Bn7uUuuWZBYjW+slow8YHJEvet3gxyK67mbYjR4G2acSczaLWDi6CPqdyxlU
crp3HJ/Hov1wl0obwODdPZNZTn32iGIeLAh6LsCmO8sleeft8bmXFT8qQoAc
pWfdppP2SL/NuIVo+Dj6XvZQfHnu5bQPH/8BOUwQsQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.68100272761296, 8.376723665619732}, \
{-1, -1}], 
         CircleBox[{13.850037500000001`, 8.063500000000001}, 
          1.4069864965259085`, {5.997507383468269, 8.69647115236248}], 
         PolygonBox[{{14.008674710082685`, 9.584830088108244}, {
          15.251945348656145`, 9.351853438503303}, {14.863505231178488`, 
          8.652486385280215}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.68100272761296, 10.043723665619734`}, {-1, -1}], LineBox[CompressedData["
1:eJwdkn9YzAccx5MTnot+PdHy/d73e1Ht+6NbP1bOkt45K+V0dp4zbiFRKzV1
HbWTiqInU41JWI6sp6So5MrE2SZmO8Y6STxFMlKxzLP8SGU3r+f5PJ/n/d/7
eT1vcUyyMtbayspKbrn//8H3MLB6jxvc5zTct13D4M322RcF7WKkzFtgivdl
0Hbay5R4SIyo13OmmhwYbCObrz9WiRG6tig01ZpBYFh5Q6eNGIr9gk7GhsHE
G26Z/xho/FzmldRFMEgerdwVEUfj8oD8bPECBi7zluUaaBpNk06NXdExEPhm
ODW1UlilPuc2eoaBdXXEgC6XQqCweGruMAOy/8LsvuUUFv6uHSKCWESapKmV
Syh8n+K+XreFhevnlR9sjKPAKuJt1jWxmFDuLp9VTCHMs1pt6GMxrnclCswU
8vmqVuE0DiOXdDO4KTSyDnToMwI5BJl3T1D60lgZQlG8mkPNvdpjF1U0dkft
VbumcngqyaiZtYlGGs06v87h8HztrvqOIhrdb3KKfQo4/LEnUv/uKA3tBdLg
WMihpeXl28RaGv2P/xoc3c4Bdsk1my0+8hihtELLwSbpWdotS3ZydjcGrOBw
aDhNxNTTkI66xKUFcFD+cOxBVyWNkhV2uku2HNINF+Y/PEDjuIPw1PR7LEoL
r7Yf2kGjsv6V9YsTLDaZKcOrRBr2REPsUouPRAcfRVUEjcPsbcdjChYPA463
5Vt8T8jZkHmUZaG8v+qhw10K2XO7Y8c7sMjW1DT4raFg3KffEGjDonNyV9rw
HRFyK4aDB2xZyJNcHc9HiDCxJWTROg8Wb3s974ecJ2GumPbtXCWLvPTxz81S
Etejsq6KilhMWi5MNZkI7NhZqB28xWJL6TtPQQaB00EjL8JncZihNx3PVBC4
Hu7wMtjiQ1Aa2bctgkCCu9tWbyMHt/FxXYXrCTgV65vzRjkU1E2Z3FtNQF97
zvsTHx69paHGGgGJ7kbTxrFlPDTbdUJhComkTTzhkczDfr/0puoRiSaNwSTP
4OFyslXSuFKE9lUyWU8mj60hfovTb4ose1u6rVbHI+yzfY2C+RQwKas9WsMj
ekw9UF5Bwa/Q+MT/Sx7LVSX5/m8pfEN85XM+ise1XZGVAotP75ygznNKHo8V
vrJnUhpV043PZeE84sVNfulhNOb9JPO+G8yDmNk+RC2iEaXK/PiyP4+qO/ZO
f8os+9v7aW8My8N2pO6jTh8aavszJzaTPLzjEzHkTOPG6uaODDseS1ye8mQP
Bd9nzo36cTwSlriGFORb+v17WdU5xOHi13WqM2IKO/2Rav6bg7FkaUxmvQgB
048kaAc5lHE9K2KkIsw2tsmrhzkwkU8kvUYS/W15skeOlr6egx3XQklomqKz
Ds7hob19osz5LoHsds0d8wYeZT79SUezCezZfzX2ST2PX1M82tTBBBZrvjid
PMLD8UCocSpNoCPgpN1WuReKwtgS3oNA2482rQtKvfDbLw+IFiWBI2MK1ys9
XvBL4YInlxMYN/Ppm0ukBEXddF+ZPYkK9yvG2IUSmOru9RR8RyIhpD6sIloC
7YfdZw+LRIh7jwSrm8NtrWtE+A8lIvCi
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.318997272387044`, 6.623276334380269}, {1, 1}], 
         {PointSize[0.04], PointBox[{15.8, 7.333}], 
          PointBox[{15.200000000000001`, 7.667}], PointBox[{11., 10.}], 
          PointBox[{18.2, 6.}], PointBox[{12.8, 9.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["45",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000007276`, 15.000000000007276`}, {
          18.200000000008004`, 14.000000000003638`}}], 
         PolygonBox[{{19.624494365672923`, 14.791385758707179`}, {
          18.381248461855627`, 14.558277151741436`}, {18.769762806798532`, 
          13.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.952053456673124`, 15.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], 
         LineBox[{{10.999999999990905`, 9.999999999992724}, {
          12.79999999998472, 10.999999999989086`}}], 
         PolygonBox[{{12.424494365672922`, 10.791385758707179`}, {
          11.569762806798531`, 9.858951330844206}, {11.181248461855626`, 
          10.558277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.04794654332688, 9.749856222011616}, \
{-1, 1}], 
         LineBox[{{15.799999999999272`, 12.666999999997643`}, {
          15.200000000002547`, 12.33299999999872}}], 
         PolygonBox[{{16.024246890528897`, 12.791830769061084`}, {
          14.781199263430382`, 12.55766715795818}, {15.170306955511828`, 
          11.858671303919651`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.3509038465922, 13.249504467199646`}, \
{1, -1}], 
         CircleBox[{16.8500375, 13.6035}, 1.4069864965259078`, {
          3.869899461996692, 6.568863230890903}], 
         PolygonBox[{{18.057725289917315`, 12.664830088108245`}, {
          17.202894768821515`, 11.732486385280215`}, {16.81445465134386, 
          12.431853438503303`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.68100272761296, 11.623276334380272`}, {-1, 1}], LineBox[CompressedData["
1:eJwdk3tUjWkUxg8pFU6FLpMmRTTnu/S97ynS0dQTchlNkQlTKjJRSRRFJHLP
oSbVMt2mi0TDqaTbmRCVmKgWmRRDNIxyzWXkVs1nnrX22utZa6/9x+/Z2zJg
nWfgUIlE4ibWl277v+QYIvkiDuUW7Uc9iRx9V+wK6WwOrYf/vie3kmP+nvIU
xUEOOwzexXRJ5XgSmpJEmjh4DFofWvmKwnhS4qrPGjxClu0KymuicPt8ZtN4
jkdoRttwFFAEq/WqQ2bwKKhSddjHUlgnxR2PduVx5eoj+calFE35BpmtU3g4
P60KDbanWF5f7FFiyMPBh9RcM6OICm+eEtXNITKkLjdChyJmnirNtJRDpl/I
ZK9+gnEVJWGHojhkKFvK1H0E/pdtOzwcOFwytXAIFv11g08/zOhnYfqxi3nx
kaDF4nufY/Us0t4duN0sofhUduSRIolFt+D77NxwiuHXi9ZF/MTCZr/P4nNS
Ckmk+pLSmUXNH3kbqCGF1YONbfkTWRT95pTeZUqxT2PsmxIDFjmdlibPx1OM
/d1df4Uui5tKv/e5VhRXqsNu6EhZmFw0f+Ero/ju0OvaZDMWrgkVJn42FCsX
tvnHTGGR3uD3Nt2WYndph8s/S1hYDHNeUzKNYtB3k37xThaCqhC3HCkO2nmf
jC1jsaY41/KVE8V6R6X/pacs7kzzDiz4liJf62xYwWQOFQojabLI80TSfrO2
AA693oGK7SxFn2nQjq1ZYn63fE2yTCgcE+Os2VYOuo3uvnUDBMGav8aNkvAo
G22AX+4RZN8JN1ZY8Hicr/0pW01QlLjNAITHbfPYD1aHCfISnGktz6OuvkHd
F0IQrjfp9YAxj9U2NYF6cwhYOr5pXC+HHk/lHHMZQajlxARjNYfdD3ImyccQ
yHt+Nlq+iUOtZ392pBaBznHj/gmEQ2eje+XXGgQuSysad3WzgHlvaZw2wXqr
Fu5xPgsNl8rzCwwJ9q6NHLs5iMWC1DXxrdYEbgVb6Xs7FjrSjlKZE0H91TcV
7WIei9q3DoleTFCtq+kZ38Pgz/172kesI0jx+FF42MLg6ql9jgF7Caq+WVWX
dZZBZuxI2+kZBHcV26yOFjE4cOCwX8YpguQu1eMlxxjUpWhm3hB5MOH2iSdz
GERteN0XW0vwsqncWyuXgaJT6h3fQLD2Kxtdw+MMemNcGk1EHzC/LSjuDIPp
+qm6MRcJEJ3mdaKBwbyiefwocd/MokLvZfcZFOck6hmpCPYMRPqdH2QAt2fN
57II3uimSATx/lyNbnZrKgnsTE4+8ZrLIn+MF397I0Fxz7+PFKEsUq3tNOQ+
BKoRo++HKVm4V40udxR5HD298MPsPHH+oXpLgjmB9mbtJl7FwnFEinTXRwGW
Qa4hIwtZ+HtXT+ivFMQ/2K5fn8yiUvv0wOVoAfys1Z+PifsbnrqpU2YJMM5u
3nlN5H0hhw+9aybgQue2LVNfMlhwZNhfMi0BMunQ+ESRz9yVCbMjhgp4mxq3
KN+dge97B2mjgQBVTYh9+qAMEU6zGoPkAnxkfdMN1TJYO5R2jlwhYOrS5zNj
tsuQnvZFAoS7VlpxXjL8B1cJ8VE=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.318997272387037`, 15.04372366561973}, {1, -1}], 
         CircleBox[{14.149962499999999`, 11.3965}, 1.4069864965259085`, {
          0.7283068084068978, 3.4272705773011087`}], 
         PolygonBox[{{13.991325289917317`, 12.917830088108243`}, {
          12.748054651343857`, 12.684853438503302`}, {13.136494768821514`, 
          11.985486385280213`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.318997272387044`, 13.376723665619732`}, {1, -1}], LineBox[CompressedData["

1:eJwdkglQUwcQhiEUr8pIrQ0eY73y3iO891JBaSngzC8Cyi3xiqFesaJjGAzI
iAdEiYiCEIkUOTxAMWipOlgolkNMkVAJGhQKymGhplTAEqBAaTxKn/1ndnZ2
ZmePb3eRbJ94F8/KyiqQs/c+N+e9hJj4XwSyq0N+lOYJMfWLho3GcgJh7W5e
uXeE+OGwv1y/n8BE58eTPF4Iwc+QRza6EJBZF/e4zXdCX7six+atAOxqi6jy
ayeU2gSe6m4WwLFePHS0xAnnfIe8M+8KUKq+GphuQ6MtUilJqhBAsWd5rkMI
jWVh5p2MQQCVUE2yGhr5HmuLQ/oFOOOXlxBeRyNZ7WxIdiDApw9VzR2gkVgm
Du8JImDwoks22DLo8QvRRCdzce/2c6n2DC67PSSL6wjEGu6lsR8xEJk/+O4u
j8SjD92Llkxj4Nz1/VWpBwnjzH5/2b80bAP2zNVHknDO96ESh2j8IwpNTztP
onBwzboTL2hYnfZ1j9GRUEVUzL7YQqPz09PCsE4S2iJVg7uBxsG3r6aHmEnU
oKqR0NH40+f24KUxEpJerxlVZTTcvDXs38MknsTXLPG+SaMs2m5FhIlErOms
cVUBjQu3hPnHDdx80V4OhmwacUFk9ZzrJHTjCcI8Ndf/yqnBtfEkCrQ3lKFJ
NGIatOsTAkjQl6uq5qtopN+PyXefRSJAHeBakkgjzLJKs7mNQGZKs6ktnYYk
MtYkvEhgkb/raetvaUzpyTqaKCOwM9S7UW+kUatQPvuFITAgl/kc4jE493qH
8q8JAbLbnx9t9GKwQ6Du1XcJsKzMPFWXykCsUJnvNQoQXGNjV9bBYGuUzrii
SYC4A9Gp9k4s1EPtdta9ArgUJCz/LYpFi7Hb5fcZBD6T9inu32Txza+dUb2+
BCye1aXn21gY+fv68k4RUEptdbNGWIxLennPGwl475Vpai0sqq/5ucY5cPzv
L8h9a2aR3rG0eXYYiTfxGtXWFhaHzqa8OplFQtbiOCApZlG6OdE4uYHEkCko
98ZxFp7Lp/tR3D2kIVPe2a1j4TP2+uYBPgWJeTS4dxELeXHh43lLKewtHEry
GGGw+mmOp2wVhXrey5zkegaL9aqOr0IpDPS3e9UXMoiYL559XUrhcOvoz1KO
R6v6gWL9Ngr8OfPseUcY2Lv4ay1bKaStfBEwEM0g448nY/0SCgsnxcuN+xlc
KhVmKQMppHS19E+OZ6BLCi/a9CUFRUGEtieNwd3QjdqCBRSuVnd3N2sZHLFO
ubzSioJqNyVLrGVgNyafeMz938HmTaLRPgYPX67/pOgOifJnG7QBfBah/Glz
9RkkWouzLOf9WKwbHVigiSERfqY9M4njsSX41qTyLVz+56nTTDUc75PjLk1i
EvpR+XCcrQhdWRWZIxtJXNF6P+etEaEp9tigRwQJ//Fga7ckLq7MtpFy9T1n
lfjKK0Xo2+4cv4zjvXJbU8ZikwhYqDpRb0/h9p43S1otIjwdfndAwfGpuBBV
nvZOhOl7lTfCbnP7Ta4bdjWLIEbmg11WjjjzU6Vr1iMRtK1Tavt9HdFRF7Rw
5JII9ddmVh0/5ojd4e8lwuPFHRZlkSP+AwpYHw4=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14.68100272761296, 9.956276334380266}, \
{-1, 1}], 
         {PointSize[0.04], PointBox[{15.8, 12.667}], 
          PointBox[{15.200000000000001`, 12.333}], PointBox[{11., 10.}], 
          PointBox[{18.2, 14.}], PointBox[{12.8, 11.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["46",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         CircleBox[{7.333500000000001, 10.329962499999999`}, 
          1.5031625000000002`, {3.3629070959375835`, 6.0618708648317945`}], 
         PolygonBox[{{6.733500000000001, 8.8268}, {7.9335, 8.4268}, {7.9335, 
          9.2268}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], CircleBox[{7.333500000000001, 9.670037500000001}, 
          1.5031625000000002`, {0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{7.9335, 11.1732}, {6.733500000000001, 11.5732}, {
          6.733500000000001, 10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["47",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         CircleBox[{7.333500000000001, 10.329962499999999`}, 
          1.5031625000000002`, {3.3629070959375835`, 6.0618708648317945`}], 
         PolygonBox[{{6.733500000000001, 8.8268}, {7.9335, 8.4268}, {7.9335, 
          9.2268}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], CircleBox[{7.333500000000001, 9.670037500000001}, 
          1.5031625000000002`, {0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{7.9335, 11.1732}, {6.733500000000001, 11.5732}, {
          6.733500000000001, 10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["48",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 0}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         CircleBox[{7.333500000000001, 10.329962499999999`}, 
          1.5031625000000002`, {3.3629070959375835`, 6.0618708648317945`}], 
         PolygonBox[{{6.733500000000001, 8.8268}, {7.9335, 8.4268}, {7.9335, 
          9.2268}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], CircleBox[{7.333500000000001, 9.670037500000001}, 
          1.5031625000000002`, {0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{7.9335, 11.1732}, {6.733500000000001, 11.5732}, {
          6.733500000000001, 10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["49",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         CircleBox[{7.333500000000001, 10.329962499999999`}, 
          1.5031625000000002`, {3.3629070959375835`, 6.0618708648317945`}], 
         PolygonBox[{{6.733500000000001, 8.8268}, {7.9335, 8.4268}, {7.9335, 
          9.2268}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], CircleBox[{7.333500000000001, 9.670037500000001}, 
          1.5031625000000002`, {0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{7.9335, 11.1732}, {6.733500000000001, 11.5732}, {
          6.733500000000001, 10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["50",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{7.333500000000001, 10.329962499999999`}, 
           1.5031625000000002`, {3.3629070959375835`, 6.0618708648317945`}]}, 
         PolygonBox[{{6.733500000000001, 8.8268}, {7.9335, 8.4268}, {7.9335, 
          9.2268}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{7.333500000000001, 9.670037500000001}, 
           1.5031625000000002`, {0.2213144423477913, 2.9202782112420023`}]}, 
         PolygonBox[{{7.9335, 11.1732}, {6.733500000000001, 11.5732}, {
          6.733500000000001, 10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["51",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{7.333500000000001, 10.329962499999999`}, 
           1.5031625000000002`, {3.3629070959375835`, 6.0618708648317945`}]}, 
         PolygonBox[{{6.733500000000001, 8.8268}, {7.9335, 8.4268}, {7.9335, 
          9.2268}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{7.333500000000001, 9.670037500000001}, 
           1.5031625000000002`, {0.2213144423477913, 2.9202782112420023`}]}, 
         PolygonBox[{{7.9335, 11.1732}, {6.733500000000001, 11.5732}, {
          6.733500000000001, 10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["52",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJxFVAs01Xccv5pHzNCU677+9//IpauYO4+tcD6adw9iYtnpgeg4EWJYR2FI
Mq5zo7GOIhutNoqyk0narBJqeWROlpB0WTlWbeXR1Pmd2/ec3/md3zm/8/t+
f58XE7YnYOciHo8XsLBe72Wlr4vBqzfFob2j8oR7C4Nup7DU6loOvHx57dbr
DPwrTHcrAzjscivakdnAwMMn26NzmkWQm4/VfAaDydnwhE3HWOS3vVqr/xGD
NQap8TZbWDw08IyNGaKR6tWh0rJncclA99+lOTTGJPGy9JUs6Lho13gFDWed
/YosNxa3L0+lbLkjxTGTFdW+MSy+YoXRm3dJUfe7BU9ew8LXvGZfvJkUqxxu
Wi5Ts/D2LxmYuUvBJ+5sxbfWHC57fFMtaKQQEDuikxLJQbxmpoJfSSH4wRWv
sWIOmXdM7/mVU1CEvlSMXuAwHOmcnPYThSGx/Nq6axwe9yWkjfdSsMuPyi1r
57AvNtTKfqHfcf7RmY0tHMr+e/+z8hgpcpovmi+q4XDT6fkDyz4pln8pl5dm
c/jCcHS+m6Uhv52693kwB213fSOvYBraQ7eMwykOXUEWcyXpNHpXXBVV9LP4
QzKj119CA96GF71zWazudG2IVNEItqq5O2jHQn1lpLAggYb/19eCdAcZ2Bbr
X9JzoHF52VVZlopB5ZSxqXOrFEmhdSqzbQzqVAXvTM5T2C17lPPMk4FJ9kn5
4BoK4Tb6seMbGHgnNDsWFkqwLa+07Eo8A+UYNca9K8H1UJvZnWcYtCdNxGRc
ECNSyyWo8RmDA37B2z2VYhhNtp4v9WCxvt2hea5QjLs31D84FLPwqf85I+mU
GLlP1IX+91iMJX5gYn1LjIaMCNkz6QIeiuGqZS/FsBmyGJkI5HB84sX054wE
Oo8Px8SncNjxsKqlwV2CJ75aYZ8e5jB/u2r9hsi355YuP8F3hyTQJffHD0ZF
TdZJYEveS/YQzPYNSHCe9DPTK8rL16I08yyZMvXTFlGaee9FT59uElIwJv85
4r61PP2pBFHkv53dctlfx9/iYe6d35sgkmA7wcvf4fv7y/eIEUHwrJU1LTY7
I0IMwbvp+scv2h4JNXzwlI7KaQjRSvjKs8+cHG0TwI/wqZZ2nLTPEiCE8P3L
/sDZTWkCuBE9/Gr0aknPKYFGL5sVTlN7ZwUaPSVUZBht2SmENdGbMjNQNtAj
hAXR42+tjvXhLiKNXofXPR7pPCLS6DlupK9A0CuCguh9f+B7xVmzItwnfhgc
dlh1zlCMD4lfPjmwMVq9WIwQ4qcTiWcjxv8RIZD4zSy63k63SwRf4sellRVK
5UI/G+LXwaO2lafXinCO+LngRYiquluIcuJ3eiCxUOophCvJg5EqF6q/XIBx
kheqqmZbS7W5Jk98bH1znqw21+RNvVNjk041H3+TPCpysvSZc+Nr8mpr5dMf
3c35mjzroLVvpKzko4fknfyg2Z9dyXzw3tRy/A/CkyFp
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], LineBox[CompressedData["
1:eJxFlAlMFFccxkFQhKqVcix7zc7izGJQgqXAekE/yGJB1KUSxdpmU0GQbEDk
KhQKCFWIle6CCFWKCLgWEC8UhUoRtJBUFCjHLkfUIlCRI4WiteVaCslkfMlk
8pLJ/N/7jp84MGJv8DIDA4O9i8/SO//80hLD0GBpUWh+UlIkqxdj428X0g/4
UTDMtL+heCTGDcV4TsQ1CqEe2YfSqsSovZtwz2k1jf0ePuv1qWJYGhV8fz2I
RmbTgqfpZjEa/0lXtV+mMWy242h4P4mMmo/C9I9p1Jut+NcynQR3QNWb0kmD
PKZ0j3Qi0TiT2pJ4n0ZHw2T8wW4RgiZ0B+6cofGtLU+5P1QEvy19C9oAGr42
ZYmR1iJ0Nm/qGbWi4eOX1zf7lEC1eo/icBeFBq9zpdxqAtezBTMZ5ykIts0W
c0oIlPPcarhKCmndFn/ICwm0apa38H0oDIRsj0u6ToAc1EqrpBQmdFFJr7QE
2qLPxQW7UPjm6OfrnRfnHXoValwJCvn/ffBZYbgICZ5ew/MBFNqkb/+004nw
9KRWG5JA4atVQ/pOWxI6h/RM0zIKxjLTNZ8EkJgTOU4WvKDQuo+ezztOYoNu
85DCjka7cNakJ49EffVrr+o4Gltb3KtCckiUdQess22lMfZwUK2KInEzSnpl
2lYCx1zT+yYuJD4e3dybGCbBpcn3LbY/EOG0Rh42UiRBZY7KaFxPIKfX+muz
nyUwP3nJ/tk2Ahfa32ZzbkngHVXnqlYLURwbEuymkiD7JfFy3XtCSDXtRvn+
EjyOHQtPvStAvv7hFW8zCVLkAV/uyBJgysJ9Z8g9GrubXerm1QJQzlb7mpWL
+t+uSY0tFyB+rdWxG+Si3zGb1m74XYBdyT/2mvUv6uE0oLGaEaBD1CewvErh
4tj01BdiIWbNY86oMigEDmvqq2RCmN/RX6iIoaDv0OzaHfJu39Aq514+JcQM
8/1IxpEj4zeFaGf+F+/FndP1CeHLzLM2yf4u05Bgz2MxaSE35hPsefuVUxW1
PAJ/M/fJlSkKj78R4jxz37ZOe8nzi+/0sPHO1EbxhShi9PJz+ekFFSFAAaPn
TUntSuurfFbvukdbpptGeKwfy7Jcs6bAAxi/TjunjQ81cVHJ+DkmenLJ+QQX
5Yzfdcn+c58mcdk8NK5ZMO8q57J5CXCSTkbPcdk8RRenrjkYzGPzlp3mL+nr
4rF5bHzgejvIjc/mddD3r8GWs3w2z5GDOhVXy2fznuy/OvfEHJ/tw7MBF4db
qwRoYfoiS9mjHF0pQBnTp6KYysOvXvPZvnGUtz9c0crHXaaP1iXFWVmL8zqY
vj7/wbGkwpMPOdNn9fSBnNJOHgKZvtv2xahFO3j4leHBkMaN6Cnksrw4q6lz
tBu1YXmy03Fn+sRWGzQxvKmSVtcuL+XAguHRGamdz7wHB78wvFKUvLkms+Gg
kuHZE9L4cfxGDhwY3tlnWPe2xnFYHv4PRdMzNA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["53",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGAQB2IQ7SM2EwgYHRjAQMWh9P1t4VNnGBwYPDoXapuqONjc
2S/6/f9/e4YX8eK7E1QckgWvhMWk/bNnYHA/c65KxeFM1Ofth8/8sWc4ICKz
oU7FQSVo2lQd49/2DAtEMldnqzjIWc/mmzbzpz2DwpRJBq4qDnvzOBz///9u
z2Ax7Z2xkIpDVA5nU3raN3uGH3Utb24rO5Q5erecO/PFnmFG1wbNNcoOFudz
U0yMP9szOCjZ93YrO7zhmgF070d7hopfHH/qlR0celvm/v3/Hi5/TPJM7Mkz
b+D6/x/2T5o+8yXcfJv9gU/S057B7a9U0/G0Mn4Md1/eD1NrXoYHcPfbnPv+
Zu/Mm3D/PfpcUhhtfBnuf4aJ3a8fzDwJDx+B79O2f5i50x4WfqBQPXB89V5Y
+AIAzAGaOw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.1, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQBWIQHTNpJhAoOjCAgYoDU+g9c+6TQH7GBhYObRWHTw0O
UXP+AfkVwnfT/VQc/Nrn1jJZKjkwXCirf56h4jB7pnvNiTIg38FMd26lisNt
ddOs6G1A/oOfDV/rVRy05mpWSHwD8hvUm3dUqzhcWWucHG2iDFS/djp3rorD
rdp/H3rygfwJsiYFISoOC38w/TRYBuQn1GbuN1Jx2HtoUd3na0A+g2amEqeK
g493XcYWoDsZJG67cdxUdqhgWJA7UQnIP2EfJLlE2eF+9YcHb62B/Au7yzVy
lB0uuLiKVfiA5Is1lY2VHa4/nmqoEwKS745j+K3kULJyh3tIMJCvsMhu/VEl
h/JfW863egH5HElCbFOVHGRYUqrnWAH5Dxbzb8pScvhk+WPJJhUgP+Bnv4aH
ksMmO+8p4lxA/gImk1l6Sg5HJp746PwG6F6Odykv5JQcuA6fAQIgX2P+hLmS
Sg7+Xy4sD12LkE9092Nt6leG638auvDLsxJluPm7u03l/sYpw+1PmPrkpLC/
Mtx9yc8eBLq6KcPdz7CGOSPXVRnuP2fHkGZFX2W4/024Vu5YHasMD5/ZrT+1
OcqV4eF33Era8OkMZXj4hh7+JrXsECL8zaZPW5DxGRE/fwVnr+DRUoHHH9Oj
VDO7VBV4/K75lK/YuFgFHv9/0lP07R6rwNNHxZtA2YtKqvD0c/Tfb+npCarw
9KXIGmn/d5YqPP0JGu5O7LugCk+f1X//A4EqPP0CANm3CIU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.9, 10.9452}, {0, -1}], 
         LineBox[{{5.1329999999994955`, 10.}, {5.356541552311122, 
          10.210745121607033`}, {5.63941016870233, 10.235478872007661`}, {
          5.885810126600888, 10.065996179877402`}, {6.012046481527392, 
          9.758520170838892}, {5.974292839226491, 9.415048727820249}, {
          5.783124135913633, 9.149516893248801}, {5.5, 9.050003367228783}, {
          5.216875864086367, 9.149516893248801}, {5.025707160773508, 
          9.415048727820249}, {4.987953518472608, 9.758520170838892}, {
          5.114189873394671, 10.065996179877402`}, {5.36058983129767, 
          10.235478872007661`}, {5.643458447688878, 10.210745121607033`}, {
          5.8670000000005045`, 10.}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJxFVAs01Xccv5pHzNCU677+9//IpauYO4+tcD6adw9iYtnpgeg4EWJYR2FI
Mq5zo7GOIhutNoqyk0narBJqeWROlpB0WTlWbeXR1Pmd2/ec3/md3zm/8/t+
f58XE7YnYOciHo8XsLBe72Wlr4vBqzfFob2j8oR7C4Nup7DU6loOvHx57dbr
DPwrTHcrAzjscivakdnAwMMn26NzmkWQm4/VfAaDydnwhE3HWOS3vVqr/xGD
NQap8TZbWDw08IyNGaKR6tWh0rJncclA99+lOTTGJPGy9JUs6Lho13gFDWed
/YosNxa3L0+lbLkjxTGTFdW+MSy+YoXRm3dJUfe7BU9ew8LXvGZfvJkUqxxu
Wi5Ts/D2LxmYuUvBJ+5sxbfWHC57fFMtaKQQEDuikxLJQbxmpoJfSSH4wRWv
sWIOmXdM7/mVU1CEvlSMXuAwHOmcnPYThSGx/Nq6axwe9yWkjfdSsMuPyi1r
57AvNtTKfqHfcf7RmY0tHMr+e/+z8hgpcpovmi+q4XDT6fkDyz4pln8pl5dm
c/jCcHS+m6Uhv52693kwB213fSOvYBraQ7eMwykOXUEWcyXpNHpXXBVV9LP4
QzKj119CA96GF71zWazudG2IVNEItqq5O2jHQn1lpLAggYb/19eCdAcZ2Bbr
X9JzoHF52VVZlopB5ZSxqXOrFEmhdSqzbQzqVAXvTM5T2C17lPPMk4FJ9kn5
4BoK4Tb6seMbGHgnNDsWFkqwLa+07Eo8A+UYNca9K8H1UJvZnWcYtCdNxGRc
ECNSyyWo8RmDA37B2z2VYhhNtp4v9WCxvt2hea5QjLs31D84FLPwqf85I+mU
GLlP1IX+91iMJX5gYn1LjIaMCNkz6QIeiuGqZS/FsBmyGJkI5HB84sX054wE
Oo8Px8SncNjxsKqlwV2CJ75aYZ8e5jB/u2r9hsi355YuP8F3hyTQJffHD0ZF
TdZJYEveS/YQzPYNSHCe9DPTK8rL16I08yyZMvXTFlGaee9FT59uElIwJv85
4r61PP2pBFHkv53dctlfx9/iYe6d35sgkmA7wcvf4fv7y/eIEUHwrJU1LTY7
I0IMwbvp+scv2h4JNXzwlI7KaQjRSvjKs8+cHG0TwI/wqZZ2nLTPEiCE8P3L
/sDZTWkCuBE9/Gr0aknPKYFGL5sVTlN7ZwUaPSVUZBht2SmENdGbMjNQNtAj
hAXR42+tjvXhLiKNXofXPR7pPCLS6DlupK9A0CuCguh9f+B7xVmzItwnfhgc
dlh1zlCMD4lfPjmwMVq9WIwQ4qcTiWcjxv8RIZD4zSy63k63SwRf4sellRVK
5UI/G+LXwaO2lafXinCO+LngRYiquluIcuJ3eiCxUOophCvJg5EqF6q/XIBx
kheqqmZbS7W5Jk98bH1znqw21+RNvVNjk041H3+TPCpysvSZc+Nr8mpr5dMf
3c35mjzroLVvpKzko4fknfyg2Z9dyXzw3tRy/A/CkyFp
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {7.333500000000001, 7.881600000000001}, \
{0, 1}], LineBox[CompressedData["
1:eJxFlAlMFFccxkFQhKqVcix7zc7izGJQgqXAekE/yGJB1KUSxdpmU0GQbEDk
KhQKCFWIle6CCFWKCLgWEC8UhUoRtJBUFCjHLkfUIlCRI4WiteVaCslkfMlk
8pLJ/N/7jp84MGJv8DIDA4O9i8/SO//80hLD0GBpUWh+UlIkqxdj428X0g/4
UTDMtL+heCTGDcV4TsQ1CqEe2YfSqsSovZtwz2k1jf0ePuv1qWJYGhV8fz2I
RmbTgqfpZjEa/0lXtV+mMWy242h4P4mMmo/C9I9p1Jut+NcynQR3QNWb0kmD
PKZ0j3Qi0TiT2pJ4n0ZHw2T8wW4RgiZ0B+6cofGtLU+5P1QEvy19C9oAGr42
ZYmR1iJ0Nm/qGbWi4eOX1zf7lEC1eo/icBeFBq9zpdxqAtezBTMZ5ykIts0W
c0oIlPPcarhKCmndFn/ICwm0apa38H0oDIRsj0u6ToAc1EqrpBQmdFFJr7QE
2qLPxQW7UPjm6OfrnRfnHXoValwJCvn/ffBZYbgICZ5ew/MBFNqkb/+004nw
9KRWG5JA4atVQ/pOWxI6h/RM0zIKxjLTNZ8EkJgTOU4WvKDQuo+ezztOYoNu
85DCjka7cNakJ49EffVrr+o4Gltb3KtCckiUdQess22lMfZwUK2KInEzSnpl
2lYCx1zT+yYuJD4e3dybGCbBpcn3LbY/EOG0Rh42UiRBZY7KaFxPIKfX+muz
nyUwP3nJ/tk2Ahfa32ZzbkngHVXnqlYLURwbEuymkiD7JfFy3XtCSDXtRvn+
EjyOHQtPvStAvv7hFW8zCVLkAV/uyBJgysJ9Z8g9GrubXerm1QJQzlb7mpWL
+t+uSY0tFyB+rdWxG+Si3zGb1m74XYBdyT/2mvUv6uE0oLGaEaBD1CewvErh
4tj01BdiIWbNY86oMigEDmvqq2RCmN/RX6iIoaDv0OzaHfJu39Aq514+JcQM
8/1IxpEj4zeFaGf+F+/FndP1CeHLzLM2yf4u05Bgz2MxaSE35hPsefuVUxW1
PAJ/M/fJlSkKj78R4jxz37ZOe8nzi+/0sPHO1EbxhShi9PJz+ekFFSFAAaPn
TUntSuurfFbvukdbpptGeKwfy7Jcs6bAAxi/TjunjQ81cVHJ+DkmenLJ+QQX
5Yzfdcn+c58mcdk8NK5ZMO8q57J5CXCSTkbPcdk8RRenrjkYzGPzlp3mL+nr
4rF5bHzgejvIjc/mddD3r8GWs3w2z5GDOhVXy2fznuy/OvfEHJ/tw7MBF4db
qwRoYfoiS9mjHF0pQBnTp6KYysOvXvPZvnGUtz9c0crHXaaP1iXFWVmL8zqY
vj7/wbGkwpMPOdNn9fSBnNJOHgKZvtv2xahFO3j4leHBkMaN6Cnksrw4q6lz
tBu1YXmy03Fn+sRWGzQxvKmSVtcuL+XAguHRGamdz7wHB78wvFKUvLkms+Gg
kuHZE9L4cfxGDhwY3tlnWPe2xnFYHv4PRdMzNA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.333500000000001, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 9.670037500000001}, 1.5031624999999997`, {
          0.2213144423477913, 2.9202782112420023`}], 
         PolygonBox[{{4.2665, 11.1732}, {3.0665, 11.5732}, {3.0665, 
          10.7732}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.6665, 12.118400000000001`}, {0, -1}], 
         CircleBox[{3.6665, 10.329962499999999`}, 1.5031624999999997`, {
          3.3629070959375844`, 6.061870864831795}], 
         PolygonBox[{{3.0665, 8.8268}, {4.2665, 8.4268}, {4.2665, 9.2268}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.6664999999999996`, 7.881600000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.867, 10.}], PointBox[{5.133, 10.}], 
          PointBox[{2.2, 10.}], PointBox[{11., 10.}], PointBox[{8.8, 10.}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["54",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQlQE1ccxh+HSOUKl1CsIu4Tj6ogi4Ii7r/WKl4dtA5ErRIRtoiKiBjx
LAgqMANFgZAFBaoOUkFoFUekIngCulAQj3iVCBa0VIGCRbnShzP7MpNkfrOT
7H+/933f3ylw+6pgfYTQdvIe/l5tRT6QM6BPLwyJKueMGnEioCjnN1tnYAjr
OlWWKhCWjbp/2A9DdGLA9Q084c7TYY0RGGIsvtJOZQkrlrTt+ZGw01T8QYcB
vZ6irNiLYZpByY47ImH7xkyfYAw9fuGJ6cIwHyvWW4AhYGn16GCecHzxu1Jb
DHfaOi+4s4Q1nrXKVgbm95sNGpC5UHV2bEEZA6FxMafuiwy5f0LEt2oG9C2y
QnMFwtH+OrM4BtytF3pv4wlrfb8pjGYg0qkVz2EJV5c8z05mYHxy1DJDRBit
fTBUwMDu1FXddeIEQOHvTe89YMDG5e1WQSDs69i+0wDDpjpnkyCe8K9X/U7O
JPO/KbNzYQlrZOoBOYZCJqGmT+cEyDXnERuJIWlzt3eVSFie45tE9Nj3Yqwy
TSCs0VTt34PhsXqBPJAn/KEh+mEgBj+bshJXlrA2K+r8fAx8ZMMKnW48mTe9
vMsSQ52/y8takbCiRFXWzEBZxfczTwiEXRe9Ul5mIGiMrXkoTxj9ctFExYBe
stdaD5bwa8sso1gGjNA1c0NEWOvXMkT0aMtfEpkrOAIqfYy9Uhg4f3F8nwdL
2PVq/tkiBqw684z/EMcBUjv9OU7DwFPlGgOeJ+xzwKDPCMPoo3E+g7qxgEJU
Kx65YyjPNMLpAuFS36Kd6zBo7166PZ0dvp7elLwLw8hz2fOqxC8A5df2q4ke
R5bE7d7IE5bZ5rBRGBRlnksHdGMAGdflKRQY1t7IU2YIhLXPzAK8MBxvfHvZ
jSVcH1g8x5zoYV5YWys6EH2iNtg1MZCrKTgcwhMO37IsqIQBmUfsDX1E+PXs
912pDNyULz54Qvic+CveZdohBgY3OtbMYod5YevBGOKDKq9bdaI9IJhxWu84
A7MUkPkDTzj/pjzyNwa2RKiUCBF2dWtOesZAeIfnAUGwI/7THj42CkN7Wugt
N5bw5AYHO08MGYr1K0VxNPF3duLf6zH8rPK153nCnq1zXykxpHkO54ywZpf2
BNEj/mlxmyDY0usRN04Vsawt/f245HSvWtGG/r/AVh3ieRt6/2tDmxN0Oms6
X4+sMlgtWNP5p5y75TWTtabPZ3pdO/uuaEWff9Hd0phNvBXVp699v/egzpLq
l6BevTdDsKT6rgtc7ebGWlL9M05a+dSKMno+xYo1CSG8jJ7fuZCqSwZIRs93
xMVemTdrQc9/YfLNIzt4c+qPY8vTI88IZtQ/jwu7Lz8UTam/Wj16w0YgU+q/
9fIVH1nWhPqzI/jCu438KOrfj3srTFKEz6i/j6Rss6sQjan/vZqNb3foRtJ8
WJhbZUxgR9L8xGoKqvx5I5qvyYuT5v4kjKD5c4j9el+1aEjzeUX2e7U+MqT5
7U+xn+7NGtB8X2p+n63k9Wn+J9bHTyoS9Gg/NOW4GbeIiPbHme+eP7FBiPaL
0a6pC47yQ5zUP/Er7+U1iQOc1E9x1hlqd7afk/orrHNLcLzwkZP6DY+Zd+aF
rpeT+s/wSr1mNv8fJ/Uj+vKlrUrs4aT+TFs+pWDArZuT+rU+pKc9VOjipP4N
CXJyqdR1cFI/a7P+rS8R/+Gk/pa3HEwtEt5wUr8jhyH5Wb6Vk/o//q/KV2q2
hZP2gycKKD+AtJy0PxSg7T2d+YST9kuKVvZgknsjJ+2f3JqV/pZZNZy0nyoN
G7a6Z13hpP31aa11rimX9tv/rFPClw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{18.50205345667312, 15.000143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], 
         LineBox[{{13.700000000004366`, 11.500000000003638`}, {
          11.000000000007276`, 10.000000000003638`}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          11.631248461855627`, 10.808277151741436`}, {12.019762806798532`, 
          10.108951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.20205345667312, 11.500143777988384`}, {1, -1}], 
         LineBox[{{13.69999999999709, 8.5}, {10.999999999996362`, 10.}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          13.068751538144376`, 9.308277151741436}, {12.68023719320147, 
          8.608951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.49794654332688, 10.000143777988384`}, {-1, -1}], 
         CircleBox[{15.725, 7.904999999999999}, 2.1106041789023346`, {
          2.8558085915452467`, 5.554772360439458}], 
         PolygonBox[{{14.175505634327079`, 6.351385758707179}, {
          15.03023719320147, 5.418951330844206}, {15.418751538144376`, 
          6.118277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 5.309856222011617}, {1, 1}], LineBox[CompressedData["
1:eJwtVWdUVOcWnSDFAJGABHFwZu69xEK5M3fIkhLU2SE8ASVigRFQQCcPpCkC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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.44794654332688, 9.690143777988382}, \
{-1, -1}], 
         CircleBox[{15.275, 12.905}, 2.1106041789023346`, {
          3.8700056003299204`, 6.568969369224131}], 
         PolygonBox[{{16.824494365672923`, 11.35138575870718}, {
          15.969762806798531`, 10.418951330844207`}, {15.581248461855626`, 
          11.118277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.44794654332688, 10.309856222011618`}, {-1, 1}], LineBox[CompressedData["
1:eJwtVQs0lVkbPhNK5BaViKlR45xv7+8731Eml6aeaIowyPVvzUjphv6i/lwO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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 14.690143777988382`}, {1, -1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 8.5}], 
          PointBox[{13.700000000000001`, 11.5}], PointBox[{11., 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{17.3, 6.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["55",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHkwlGEcx3dSEt0ZszuO3Wff9y1Hh+zsH4h+G6Ho0jbTaNhlxuQKlRkh
utw6pBJRqakcY3WYHKXs1KBikB0hm1khibCmWsdqe7Zpnn4z77zzmed939/1
/b4oKMo3eAGLxXLDl/7+bEaHgwHW36CBUx/NNm/C3G2ZZdxKga/21QGDVMyP
ui8ZF1NgnxTzfsQV82TV+U/nKSgUTQl3/KCBZd2qfX2aAjsLm6CmEswqWwPf
cxRMOITNmUows5+KQ65SYCsVJrRwMMtDq4WVFASoFD6+vRSwpNkLRX0U5C25
c6b6PmaehyBhDQ3KW9qzRfGYJ70vHtxLw6CTQmN/CHO2UDVxmYb+uRrFvCfm
dg7bqpMGWcqoo6EIc/fmJ+VmDPQ1tGVWbtc/v3J53H4GVE21N1RizDWd7TfS
Gch/Z1s7HIE5JDUxrZIB7rZwZVUWZvbYpc0dDDSfOTz7tQIzBN5EAwzcTsw9
OKDAvJLrlT7EgKjMP+KnBrP1CRuvHgZm/KnGODbur6TGUfOSAbGkMOaXA2Zp
/L2u6wzIn/uInbdjzkOBb4IZGKpN7j65C3O62nujHQMtCRzDsJ2YTw92ffhG
Q9KCipwtzvp5dtV3FtOQF6gKesXFbPTZTxJMw6KdZVOSaZxfNSy9s5YG3d7x
DJNGff11uoJxChzWSGJFmfrz/Zbz9RS03DrSr/bAXLQ0Q1dIgaOTzCT0Nx9Y
b4Zke9IoiLQWCASVmLMFpppTFNw1q/WbP/z/nNsznRzD5ZP3S10WT9BKRL6v
1MUq/YoQyf9lQ91xw0hE6nOfSDG454FI/QG7axQ2doj0x9X6SLXmiPSvye2/
MMJGZD4W1zir7RAi80tXv1VHCRCZ7xTHrWhsFyLzL3hgNButz/9vPy6KtmXn
chDZH+/X7/DEakT2+2L5XMbDj4js3y3VWZY0g4g+TB6VN6es4hP9zGfm9j7h
8Ym+NjXHr5cyfKK/WbEsJ8SKT/Rp5HIlUmjCJ/rVPQ4daBhFRN/ung2DFXJE
9H/heg7HNAsRfzyetB/N90HEP98ttWH7FiPiL/khl4W3X/CI/7auiLG3PMYj
/izvUF89uY5H/Gta6oEa5Vzi7/tHRwtyXbnE/+4VLTisyP/hD7GisvI=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 9.0548}, {0, 1}], 
         LineBox[{{13.69999999999709, 8.5}, {10.999999999996362`, 10.}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          13.068751538144376`, 9.308277151741436}, {12.68023719320147, 
          8.608951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.49794654332688, 10.000143777988384`}, {-1, -1}], 
         CircleBox[{15.725, 7.904999999999999}, 2.1106041789023346`, {
          2.8558085915452467`, 5.554772360439458}], 
         PolygonBox[{{14.175505634327079`, 6.351385758707179}, {
          15.03023719320147, 5.418951330844206}, {15.418751538144376`, 
          6.118277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 5.309856222011617}, {1, 1}], LineBox[CompressedData["
1:eJwtVWdUVOcWnSDFAJGABHFwZu69xEK5M3fIkhLU2SE8ASVigRFQQCcPpCkC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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.44794654332688, 9.690143777988382}, \
{-1, -1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 8.5}], 
          PointBox[{7.7, 10.}], PointBox[{3.3000000000000003`, 10.}], 
          PointBox[{11., 10.}], PointBox[{17.3, 6.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["56",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999818, 14.999999999996362`}, {
          10.999999999996362`, 9.999999999996362}}], 
         PolygonBox[{{16.024494365672922`, 12.791385758707179`}, {
          14.781248461855625`, 12.558277151741436`}, {15.16976280679853, 
          11.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 13.250143777988384`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 5.}, {17.299999999995634`, 6.5}}], 
         PolygonBox[{{18.125505634327077`, 6.041385758707179}, {
          19.368751538144373`, 5.808277151741437}, {18.980237193201468`, 
          5.108951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.797946543326876`, 6.5001437779883835`}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHkwlGEcx3dSEt0ZszuO3Wff9y1Hh+zsH4h+G6Ho0jbTaNhlxuQKlRkh
utw6pBJRqakcY3WYHKXs1KBikB0hm1khibCmWsdqe7Zpnn4z77zzmed939/1
/b4oKMo3eAGLxXLDl/7+bEaHgwHW36CBUx/NNm/C3G2ZZdxKga/21QGDVMyP
ui8ZF1NgnxTzfsQV82TV+U/nKSgUTQl3/KCBZd2qfX2aAjsLm6CmEswqWwPf
cxRMOITNmUows5+KQ65SYCsVJrRwMMtDq4WVFASoFD6+vRSwpNkLRX0U5C25
c6b6PmaehyBhDQ3KW9qzRfGYJ70vHtxLw6CTQmN/CHO2UDVxmYb+uRrFvCfm
dg7bqpMGWcqoo6EIc/fmJ+VmDPQ1tGVWbtc/v3J53H4GVE21N1RizDWd7TfS
Gch/Z1s7HIE5JDUxrZIB7rZwZVUWZvbYpc0dDDSfOTz7tQIzBN5EAwzcTsw9
OKDAvJLrlT7EgKjMP+KnBrP1CRuvHgZm/KnGODbur6TGUfOSAbGkMOaXA2Zp
/L2u6wzIn/uInbdjzkOBb4IZGKpN7j65C3O62nujHQMtCRzDsJ2YTw92ffhG
Q9KCipwtzvp5dtV3FtOQF6gKesXFbPTZTxJMw6KdZVOSaZxfNSy9s5YG3d7x
DJNGff11uoJxChzWSGJFmfrz/Zbz9RS03DrSr/bAXLQ0Q1dIgaOTzCT0Nx9Y
b4Zke9IoiLQWCASVmLMFpppTFNw1q/WbP/z/nNsznRzD5ZP3S10WT9BKRL6v
1MUq/YoQyf9lQ91xw0hE6nOfSDG454FI/QG7axQ2doj0x9X6SLXmiPSvye2/
MMJGZD4W1zir7RAi80tXv1VHCRCZ7xTHrWhsFyLzL3hgNButz/9vPy6KtmXn
chDZH+/X7/DEakT2+2L5XMbDj4js3y3VWZY0g4g+TB6VN6es4hP9zGfm9j7h
8Ym+NjXHr5cyfKK/WbEsJ8SKT/Rp5HIlUmjCJ/rVPQ4daBhFRN/ung2DFXJE
9H/heg7HNAsRfzyetB/N90HEP98ttWH7FiPiL/khl4W3X/CI/7auiLG3PMYj
/izvUF89uY5H/Gta6oEa5Vzi7/tHRwtyXbnE/+4VLTisyP/hD7GisvI=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 9.0548}, {0, 1}], 
         LineBox[{{13.69999999999709, 8.5}, {10.999999999996362`, 10.}}], 
         PolygonBox[{{11.82550563432708, 9.541385758707179}, {
          13.068751538144376`, 9.308277151741436}, {12.68023719320147, 
          8.608951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.49794654332688, 10.000143777988384`}, {-1, -1}], 
         CircleBox[{15.725, 7.904999999999999}, 2.1106041789023346`, {
          2.8558085915452467`, 5.554772360439458}], 
         PolygonBox[{{14.175505634327079`, 6.351385758707179}, {
          15.03023719320147, 5.418951330844206}, {15.418751538144376`, 
          6.118277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 5.309856222011617}, {1, 1}], LineBox[CompressedData["
1:eJwtVWdUVOcWnSDFAJGABHFwZu69xEK5M3fIkhLU2SE8ASVigRFQQCcPpCkC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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.44794654332688, 9.690143777988382}, \
{-1, -1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 8.5}], 
          PointBox[{7.7, 10.}], PointBox[{3.3000000000000003`, 10.}], 
          PointBox[{11., 10.}], PointBox[{17.3, 6.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["57",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.50205345667312, 15.000143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHkwlGEcx3dSEt0ZszuO3Wff9y1Hh+zsH4h+G6Ho0jbTaNhlxuQKlRkh
utw6pBJRqakcY3WYHKXs1KBikB0hm1khibCmWsdqe7Zpnn4z77zzmed939/1
/b4oKMo3eAGLxXLDl/7+bEaHgwHW36CBUx/NNm/C3G2ZZdxKga/21QGDVMyP
ui8ZF1NgnxTzfsQV82TV+U/nKSgUTQl3/KCBZd2qfX2aAjsLm6CmEswqWwPf
cxRMOITNmUows5+KQ65SYCsVJrRwMMtDq4WVFASoFD6+vRSwpNkLRX0U5C25
c6b6PmaehyBhDQ3KW9qzRfGYJ70vHtxLw6CTQmN/CHO2UDVxmYb+uRrFvCfm
dg7bqpMGWcqoo6EIc/fmJ+VmDPQ1tGVWbtc/v3J53H4GVE21N1RizDWd7TfS
Gch/Z1s7HIE5JDUxrZIB7rZwZVUWZvbYpc0dDDSfOTz7tQIzBN5EAwzcTsw9
OKDAvJLrlT7EgKjMP+KnBrP1CRuvHgZm/KnGODbur6TGUfOSAbGkMOaXA2Zp
/L2u6wzIn/uInbdjzkOBb4IZGKpN7j65C3O62nujHQMtCRzDsJ2YTw92ffhG
Q9KCipwtzvp5dtV3FtOQF6gKesXFbPTZTxJMw6KdZVOSaZxfNSy9s5YG3d7x
DJNGff11uoJxChzWSGJFmfrz/Zbz9RS03DrSr/bAXLQ0Q1dIgaOTzCT0Nx9Y
b4Zke9IoiLQWCASVmLMFpppTFNw1q/WbP/z/nNsznRzD5ZP3S10WT9BKRL6v
1MUq/YoQyf9lQ91xw0hE6nOfSDG454FI/QG7axQ2doj0x9X6SLXmiPSvye2/
MMJGZD4W1zir7RAi80tXv1VHCRCZ7xTHrWhsFyLzL3hgNButz/9vPy6KtmXn
chDZH+/X7/DEakT2+2L5XMbDj4js3y3VWZY0g4g+TB6VN6es4hP9zGfm9j7h
8Ym+NjXHr5cyfKK/WbEsJ8SKT/Rp5HIlUmjCJ/rVPQ4daBhFRN/ung2DFXJE
9H/heg7HNAsRfzyetB/N90HEP98ttWH7FiPiL/khl4W3X/CI/7auiLG3PMYj
/izvUF89uY5H/Gta6oEa5Vzi7/tHRwtyXbnE/+4VLTisyP/hD7GisvI=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 9.0548}, {0, 1}], 
         LineBox[{{13.700000000004366`, 11.500000000003638`}, {
          11.000000000007276`, 10.000000000003638`}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          11.631248461855627`, 10.808277151741436`}, {12.019762806798532`, 
          10.108951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.20205345667312, 11.500143777988384`}, {1, -1}], 
         CircleBox[{15.275, 12.905}, 2.1106041789023346`, {
          3.8700056003299204`, 6.568969369224131}], 
         PolygonBox[{{16.824494365672923`, 11.35138575870718}, {
          15.969762806798531`, 10.418951330844207`}, {15.581248461855626`, 
          11.118277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.44794654332688, 10.309856222011618`}, {-1, 1}], LineBox[CompressedData["
1:eJwtVQs0lVkbPhNK5BaViKlR45xv7+8731Eml6aeaIowyPVvzUjphv6i/lwO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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 14.690143777988382`}, {1, -1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 11.5}], 
          PointBox[{7.7, 10.}], PointBox[{3.3000000000000003`, 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{11., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["58",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAmIQ7eubBgRcDgxgoOJQdIX105SZnA4MAhZyy6xUHJpY
bPiPnuFwYGiw+eeYoeJgH2xk9/0/O5Av1Lq+TsVhx7KKUm1jIL8g9M/HChUH
wdUCxolpbA4MF5699o9QcWCzSVk5fSarA8MO1kYuFRWHCsuOX2fPsADljy47
8FzZoevkgkJmBiB/w905NzcoO+Q33V9tbszswGBwVbKrW9lhXeyR5VlpTA4M
FUJTZBuUHXbOPL1jzkxGBwaPEDW5icoOUXM4Us+dYQDK67Md2aXscOvvXXOg
RxwYFgivP/NZ2UH8wma91LR/9gwR7q879VUc5k0vCzh95o89UEXf5gQVB+l0
7QlGxr/tGQJmZETXqDhE7lxgM2vmTzj/WkdoDSPDD7j6Ox2S89LTvsHNO6gi
EHvmzBd7mH0Hnu9iMjD+bA9zD9Nbk+jJMz/aw9yre6rt17v/7+1h/tFZIfNq
15k39jD/HpASvdw986U9LDxYo8vXJ6Y9s4eF16UA9i9Wxo/tYeG5ZEroWRGG
B/aw8BbQK2i/PvOmPSw+7jyf61hpfNkeFl8NuxcUic46aQ+LT4FK9h69WTvt
YfENSgUNnfP2wtIDANJK1SM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.6500000000000001`, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 15.000000000003638`}, {
          17.30000000000655, 13.500000000003638`}}], 
         PolygonBox[{{19.17449436567292, 14.541385758707179`}, {
          17.931248461855624`, 14.308277151741436`}, {18.31976280679853, 
          13.608951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.50205345667312, 15.000143777988384`}, {1, -1}], 
         LineBox[{{20., 5.}, {11.000000000001819`, 10.}}], 
         PolygonBox[{{14.975505634327078`, 7.791385758707179}, {
          16.218751538144375`, 7.558277151741437}, {15.83023719320147, 
          6.858951330844206}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 8.250143777988384}, {-1, -1}], LineBox[CompressedData["
1:eJxFlHkwlGEcx3dSEt0ZszuO3Wff9y1Hh+zsH4h+G6Ho0jbTaNhlxuQKlRkh
utw6pBJRqakcY3WYHKXs1KBikB0hm1khibCmWsdqe7Zpnn4z77zzmed939/1
/b4oKMo3eAGLxXLDl/7+bEaHgwHW36CBUx/NNm/C3G2ZZdxKga/21QGDVMyP
ui8ZF1NgnxTzfsQV82TV+U/nKSgUTQl3/KCBZd2qfX2aAjsLm6CmEswqWwPf
cxRMOITNmUows5+KQ65SYCsVJrRwMMtDq4WVFASoFD6+vRSwpNkLRX0U5C25
c6b6PmaehyBhDQ3KW9qzRfGYJ70vHtxLw6CTQmN/CHO2UDVxmYb+uRrFvCfm
dg7bqpMGWcqoo6EIc/fmJ+VmDPQ1tGVWbtc/v3J53H4GVE21N1RizDWd7TfS
Gch/Z1s7HIE5JDUxrZIB7rZwZVUWZvbYpc0dDDSfOTz7tQIzBN5EAwzcTsw9
OKDAvJLrlT7EgKjMP+KnBrP1CRuvHgZm/KnGODbur6TGUfOSAbGkMOaXA2Zp
/L2u6wzIn/uInbdjzkOBb4IZGKpN7j65C3O62nujHQMtCRzDsJ2YTw92ffhG
Q9KCipwtzvp5dtV3FtOQF6gKesXFbPTZTxJMw6KdZVOSaZxfNSy9s5YG3d7x
DJNGff11uoJxChzWSGJFmfrz/Zbz9RS03DrSr/bAXLQ0Q1dIgaOTzCT0Nx9Y
b4Zke9IoiLQWCASVmLMFpppTFNw1q/WbP/z/nNsznRzD5ZP3S10WT9BKRL6v
1MUq/YoQyf9lQ91xw0hE6nOfSDG454FI/QG7axQ2doj0x9X6SLXmiPSvye2/
MMJGZD4W1zir7RAi80tXv1VHCRCZ7xTHrWhsFyLzL3hgNButz/9vPy6KtmXn
chDZH+/X7/DEakT2+2L5XMbDj4js3y3VWZY0g4g+TB6VN6es4hP9zGfm9j7h
8Ym+NjXHr5cyfKK/WbEsJ8SKT/Rp5HIlUmjCJ/rVPQ4daBhFRN/ung2DFXJE
9H/heg7HNAsRfzyetB/N90HEP98ttWH7FiPiL/khl4W3X/CI/7auiLG3PMYj
/izvUF89uY5H/Gta6oEa5Vzi7/tHRwtyXbnE/+4VLTisyP/hD7GisvI=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.35, 9.0548}, {0, 1}], 
         LineBox[{{13.700000000004366`, 11.500000000003638`}, {
          11.000000000007276`, 10.000000000003638`}}], 
         PolygonBox[{{12.874494365672923`, 11.041385758707179`}, {
          11.631248461855627`, 10.808277151741436`}, {12.019762806798532`, 
          10.108951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.20205345667312, 11.500143777988384`}, {1, -1}], 
         CircleBox[{15.275, 12.905}, 2.1106041789023346`, {
          3.8700056003299204`, 6.568969369224131}], 
         PolygonBox[{{16.824494365672923`, 11.35138575870718}, {
          15.969762806798531`, 10.418951330844207`}, {15.581248461855626`, 
          11.118277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.44794654332688, 10.309856222011618`}, {-1, 1}], LineBox[CompressedData["
1:eJwtVQs0lVkbPhNK5BaViKlR45xv7+8731Eml6aeaIowyPVvzUjphv6i/lwO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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.552053456673121`, 14.690143777988382`}, {1, -1}], 
         CircleBox[{5.5, 9.505}, 2.255, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{6.1, 11.76}, {4.9, 12.16}, {4.9, 11.36}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 12.705200000000001`}, {0, -1}], 
         CircleBox[{5.5, 10.495}, 2.255, {3.3629070959375844`, 
          6.061870864831795}], 
         PolygonBox[{{4.9, 8.24}, {6.1, 7.84}, {6.1, 8.64}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.5, 7.2948}, {0, 1}], 
         {PointSize[0.04], PointBox[{13.700000000000001`, 11.5}], 
          PointBox[{7.7, 10.}], PointBox[{3.3000000000000003`, 10.}], 
          PointBox[{17.3, 13.5}], PointBox[{11., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["59",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtU01UcB/ALgoJDGQmIQDpgsBmh05jOMv5/QYV8xFSqoWETc6CYYiiC
CE4Z8QhhvAshRZFmvuCYOBFQ8dEwXxFwwFcTxabzwfuh6Pre0/+cnf/57O7+
XveeuYVvXr7OnBAShg99k3smPG4sGaJvAUu+nWemvQRz5xur3IUsqdE730uD
ZZ5HTX7ULrsPS2Ftkef6YFhxqeelCxzd3uAohWPb5zgaeXB8QTgDN9lMelUH
qyNFEybD1aq4xjxYVDfAvEK+hV/0p0fBctXFjjNw7DHFXwHwT3o/ryQ443aL
jwv9vfRWehDM7ptaop/CkriwvGFXWu8e1RoVfGC/2RuTF0u2TpSpeTAxV3L6
4BBHnuvpySwxDOwOHIZ3Puq1CoBJkloxDvtHRjGB199Hf9WNNh/CtfdTNq2A
9ed3KILhzG0P5HddWSJUuA1Fw9y1Fo8VcNeS5pQMOHpCuN+QC/p7e727EK4c
vNuphrW1nN5c2OKTysUz4KDQ0oxE2CA/eb3NGfnzhQIZ3FxGdCmwpufGFi8a
75/57WI4zSzhzxeoV60c8XkyCXnPPs38nfYXH/gsD9aZ7Pi74ciOnqy5sLyh
fFooHOT58/IHTpjPo5wqljoqYk0crM/Ymz4HtvAufz0WVlYXCgNg/oL2iKKJ
LOGNv3xkJaxJvX9zCiyT5EsSYJsar84KR5Y4VYbW/QJvvOKQPR3WeJ6prqfx
+Gfjah3Qb32bWwscNyqqOhhOi7Jadp/Gby3a8cwe8wurym2Ghw6NNmbC8v1p
OTU0fmmRyyxY6/O9PIfW6+w7/fEExGmPF4V50Xv1VXE+zM2aec4DPtZrTJwP
q9nQWKMnzstGN677PeS79PHxGnhjSlZfEWyIsDhQCOfbDlbMhq1cQ3R74MtF
z4pv2eG+2HuHKel+zcWVcjt6rtmCLPh5/cAWIxfzOWY4roFV3HfiGDjS3aby
GmxgDZ19triHZUc7XsCVs5NtrWChneVtDurr0vqb+ONRz+rE+smwdvnhTcHj
WNJmbtrpDkuM88TpNohzsGTIHpZN5Yxp4uC+84p/GEA8UUFXkgCO9k1va6Tx
3ddsTRmLeTx2K8mDo1svZj+xxvwCmOZQmCviWy+AI1cvMveAldblsfusEP9r
06f9fOQTVOztHIPvW1+uaoHlMZa1PFi7znbeVdiw7e/kJaMR986GBB2sS36Y
ucES84kPnnYHDkorW59kgX6jdCGDsNX50ol7R2F/ttMyZ1rvQenSEnPU+VR8
ei6smSM+UWWGc34lFcro+otdZk0E+8WKzyJg3R/qVAJHxvhvVsCk/wYn8R1D
nBJmeofA+g9SP3J8yxDJNc+dM2h/3ftOXn3DkEqZ+1EzWD51f2TKa4a0qd61
NKIe3vnXiwKHGcLdek6ihkWKlU1kiCHqfwc/XwUrv7xW/usA1i1Vz0Vwl2SW
hW8/8gWt+s4Bvm1q6CjrZQjJGO3MgaU9CTHd3Yi/eFKrHawfEXgIuhgitRTy
BfCFKzl9Bc8ZYnD4LWkhfCC0ZOSmgSF6ZfjwepjdPhDD7WQIW3zD/keaf0WD
f/RD5BM7VRyi6z7J3uQeQ9IuBJ46QV365GZBC/IH6flHYDJs/XDpLeSv9vfN
pV6bXxV8Gfm3f5MaRe13xtB3iiHyQ/p+CTV9NHX//397sP8BoW0i+g==
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.163757790381083, 11.908289943285126`}, {0, -1}], 
         LineBox[{{20.000000000009095`, 17.000000000003638`}, {
          9.000000000009095, 12.000000000007276`}}], 
         PolygonBox[{{15.046219886477562`, 14.74828176658071}, {
          13.788258935801965`, 14.615864824404332`}, {14.11930129124291, 
          13.887571642434247`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.463405436333165`, 15.306268040067035`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 10.}, {13.499999999996362`, 
          6.499999999996362}}], 
         PolygonBox[{{16.221717340046695`, 7.96554010617899}, {
          17.0886427307393, 8.88664833378988}, {17.467922589167312`, 
          8.18227145385214}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.619945274295723`, 9.009930204879371}, {1, -1}], LineBox[CompressedData["
1:eJwt1gk0lXkfB/BHpSLbGFtFpMkoySWpLPWMNdrkvZYu6TLIa8rWMteWdvu4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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.880054725704277`, 5.509930204879372}, {-1, -1}], 
         LineBox[{{8.999999999998181, 12.000000000005457`}, {
          13.499999999996362`, 6.500000000005457}}], 
         PolygonBox[{{10.870057325845643`, 9.714374379521992}, {
          11.320359754473028`, 8.532330504375103}, {11.939525593835686`, 
          9.038920736580913}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.655166186234855`, 8.873281425101244}, {1, 1}], 
         {PointSize[0.04], PointBox[{9., 12.}], PointBox[{13.5, 6.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/becfdf/ef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/becfdf/ef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwd0wlQU0cYB/AFwQaDHDXGiBYTDonFI1pQUOE9FUtqi4BlbKIFM2gwSB1C
cVSM2OBFRLmRMhYLCrZBCxMoBRWh8ZhOhFLTKSgwtMUbihwiYlAk/a9vJvPN
7+3bb79vdyOKTdyktCWERONHI+m24hGxxIVCyBJp7mJ+B7y/rOdlABxQFXi9
Clbt25aeAPfc05gOw8N1a/3y4K5mXeTnsEDa41UBb3lSNjkPLjKG/lwFZ3T9
9aYXUeIv+KEM8cMio4W+V5xp5Z1EtFu53CUZ0aKNHY9HXPuwrnQ5Iod/8dDH
dJwsXzE2jyVqe7GXDzxqKyo7Avd+OvuuK1zAhKfPggUZU9248GdJI5M17lhP
unX3TLg3vpW/GRbaH30mgSeU8S3WD1iif/pq41bYGL5mURUs3tMQkPOuzo4L
MXDv0cn223Q/psyOcoKNMo9LNuhL6+eefHUuS2TN3geXwhNpjo0KWKJZ5hsF
y75bcc06B98lr02Mg8s5rLUQVvjHfbITNsQ+FvvALsMRYhmsrvqis9oN+fv8
f1kNc4RL9wTC5mxBpBu8Y9WSFuNslkQkmKJeoR59UPaVDbAhJXxxF+0/hN/U
KcB6Xbs0Jrhj/dtzibBZ6bzmN9p//8LA6XDO3cGt7bDpdWVIzSzs95fWoJew
39PdmmhYF7PB1hPr2fHmXXaEteuYNjn8aPed9Y18zH8kKs6HS2vT2a+o/U50
3Ia7dx2Y4MN64ZnjY/Azh23KhpnIf77YwvNgycLNlVIZPGxrPegB88pitAM8
3I+Y1CZ32LA92KKBWVd7M5eO/8vNtIdl5y49GEA+MmTTfGoG6vFwNDTDxaxD
Fg/O+am3Ug+bNgdfKHkf57c92yeL3tudM5US+DLPN1oL64c4cpMrxudGvbu3
5pQkrgrmqOxKC6lHVow4wUU3V1Zepf1weYcbXDCPle/tp/d6vl6ZAAuzljV4
or7Vj1vMc2BdZ4okGrbcMpvynZFn0deKXFhYP7LgiRPOtUSXexU2bTJP84UD
oqvz2ui4ZUG+Yjr2MYET+TfMqc90SnfE/KYOUTtclNx2v5SLed71dU2wIqPG
v2Ya7pdB3vg9LB4KiTU6IH9AQYAGluwJk/zJQX6nWxVbaL6K+mn/vYf9qSsU
r4N14qVBzrAlI/NEIKytbd2ybireP8ytZmF2pduD4/bId6N8sZzmc0zy7bDD
/9zq6pUGD7+dWv4RLL7Sd6qWjrN7wwqmoE4bTcsArGqYTBy1xX6PtCbN90S+
oHNxG2GXArGPDFbrSlRnbfCd/GxGKixwlvr9Q9DvNe6LPLjo3qojzrD+7e/P
C2GDejBxxyRDOGFtxzJgIj5+MG+CIWycyKKm+dN719e8ZsjwrwfiwmHJ4JSH
f1gYoqq77biQWi3pfjHGEJdDOXHTYXP5N9m+LxlCXqWFjqPeiNKSptQXDBHY
armjcM7oQMrgc4YYSmzeWOn55MqzU4cZoovOH59L13t+3iNskCHSiDsnpHDp
stN9N/oZcrkneP4hWHvhoiK0jyH7j15/UE9dtU918wlDIhrHmCFYuCTRSfSI
IXqVZIa7F/pRxhZvvM8QbcrpWAY2nvxRdLgb63Xe4EfAbFp3eH87Q8zfeseH
0+9vho6bWhmi9r5kDaaWbLN6NjFEGNJf7UFNH30j4dDoxf4PufsRZQ==
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.163757790381084, 8.091710056714874}, \
{0, 1}], LineBox[{{20.000000000009095`, 17.000000000003638`}, {
          15.000000000009095`, 14.500000000003638`}}], 
         PolygonBox[{{18.03665631459995, 16.018328157299976`}, {
          16.784458247200067`, 15.839442719099992`}, {17.142229123600035`, 
          15.12390096630006}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.411618926975066`, 16.53156214604987}, {1, -1}], 
         LineBox[{{20.00000000000955, 10.}, {9.000000000009095, 8.}}], 
         PolygonBox[{{13.909678053940056`, 8.89266873708001}, {
          15.01876777077995, 9.500879226959952}, {15.161876121339938`, 
          8.713783298880028}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.544992164264901`, 7.858382411310172}, {0, 1}], LineBox[CompressedData["
1:eJwl1wk4lNsfB/BBZS1LRLLvSjXJWrqmqKa0SIWEcOmO7EqWqMlSFBoqFEVS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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.23329719907116, 5.22601020508129}, \
{1, 1}], LineBox[{{8.999999999996362, 7.999999999998181}, {15., 
          14.500000000003638`}}], 
         PolygonBox[{{12.406968061639839`, 11.690882066776492`}, {
          11.886953316211157`, 10.537805892130281`}, {11.299110560509165`, 
          11.080429974316734`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.465854660827393`, 11.696534159236254`}, {1, -1}], 
         {PointSize[0.04], PointBox[{9., 8.}], PointBox[{15., 14.5}]}, 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcedf/ef.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcedf/ef.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {57., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {61.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {66., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {70.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["g",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {75., 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1AswXFcYB/APayUeQULbqU6yVTQRbTeUoUl6T5MQGtKtImRCxEiimg6p
lXp3E6FUp6U0WrNqdehEEGTGIxJmJYpqPDpjVISyiWejgshabGz/t9Mzc+fM
b77z/845987cVyNi/E/rE5E3Hn6mBR2GNaP/xhuMZJulfefhkTa7ZUP46hmH
C/9YMTpnmTTJmznPsmT48h1V25oTI41xY7kI1ox5uKhg5jIXNbaNUddeY/sm
WKZLDaqHRwK7n8hgxfODRXK4vWrVloNn3rlY8TUcqU5Ur+xmtHBT9GMyXOaW
llgHKw9ZV56ErVo7Ss7DTQ8rB8RwSsjOj/bz1jTT31sZzW2qu/4iLNl14EE2
3B7ocUmP7+dZLNwGD/3+YM+qIyP3kyt1OZaMslyj1Wuwwie9VGvBKFbSLeXX
ixze64uB3VXZBUI+v57XMGrOyHSftVQAxx770MoA9tUeC1AjrwwLswnZgvNm
TgfdhyV/3chTmeH+nssLVbDY96fgK/CmYIFeHByepT0VDTtNWVx+Cx7f433r
CDz0duuz6V3ov7vyhAM8MFwj/RlW9KQfmTbFdzA93hMJh5t3fPcFfG97TZ4L
72vuRVMmjCbcDCMsYXFeR5oNHLBvJH1jJ/pX6bvZGzO6/ZJ2UAOzJx98brkZ
+c9KHLVweLT/lSUj/jxBk/rIk66ybETIKCEjbFIIK5/O+z00RG7U5wJfV+bL
H1vCmlv5rovIL6xyXnEC+FPfO/2wJLA4VAgLGs++UgqLM9qi7hrgPgbKskjY
Qj5vJocntve27IBrvfNLU+H+6g2bodfRv0qyNQCe8X9NlQ/XrmtNzWBBfPsv
wXB4Ye9wkT7ef3dHsSMsHv4tU62HXMjFKRM4VhG/LILPpefs1TpgPx+P27bE
SPrtD1oNrExYLLfe4MhiubrwOTyedO300DpHSo/9TIC8qLDp1/pVjsqUiX8K
YcVhz4o/VjjyNvlGpcN6mbDV0UvNUUKE5NRjmD0yzdzxjCNfP6uETr5fXOqb
x5c5arr6fW0Bv5+YQh895ShKfnA2kK+nZZIcFsxIVOYwdTYUhfJmmUld9nCz
oxfBh/SLejJgxeD95fAlju41z7v6weyrrLkzixx1XVpKsrXnv29sudECRwXe
De5GfN3ZdTx3jqPaozEla3bo59HJkmc4khis56/CshsWrVsmOGLVR510fF1e
c9l5jKOskhCpIfKyCpN3BwZxP132CSF/HrP3c17o42hCT1Gl5dd/El328V3c
d6p3YIL3fGB8fz1HotnYlBbeuYcNzpZzpKjWNH7JO1bxckouznvz+qiX3f//
mdwWEvCzHfsXy6mzBg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{2.3732814251012426`, 12.844833813765145`}, {1, -1}], 
         LineBox[{{20.000000000002274`, 17.}, {5.5000000000027285`, 14.5}}], 
         PolygonBox[{{13.341276069446609`, 15.851944149904588`}, {
          12.090761163950333`, 16.042239896393152`}, {12.226686697156449`, 
          15.253871803797672`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.48664427941315, 16.672663179403738`}, {0, -1}], 
         LineBox[{{19.99999999999909, 10.000000000003638`}, {
          15.50000000000091, 10.5}}], 
         PolygonBox[{{17.15366975919583, 10.316258915644909`}, {
          18.39050285123411, 10.581294578224538`}, {18.302157630374232`, 
          9.786187590485644}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.92116886541601, 11.185719788744109`}, {0, -1}], LineBox[CompressedData["
1:eJwtlgs0lekaxz+Sy2aakFNKQpNpEKI0KL1R6eJeNpqh3SBEiaY1FTN9SEI4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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.617875450609192`, 1.6086826603562592`}, {0, 1}], 
         LineBox[{{5.5, 14.500000000009095`}, {7.5, 10.500000000010914`}}], 
         PolygonBox[{{6.231671842700025, 13.036656314599949`}, {
          6.410557280900009, 11.784458247200067`}, {7.1260990336999415`, 
          12.142229123600035`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.71843785395013, 12.411618926975063`}, \
{1, 1}], LineBox[{{15.500000000003638`, 10.500000000009095`}, {
          11.500000000003638`, 3.500000000009095}}], 
         PolygonBox[{{13.2023166369859, 6.479054114725325}, {
          13.450386106164316`, 7.719401460617409}, {14.144980619863883`, 
          7.322490309931942}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.240976870292911`, 6.835784645546908}, {-1, 1}], LineBox[CompressedData["
1:eJxFlgs0lVkbx08kVpNLxYRB8r7vcd7jdZlGQuIp15BLKJfJNYNyxmXGJcQx
CRmDoTCVW6mE5JII+RTClDqU0DBdjMs3vkqMqMjsM2vW/s5aZ531W/vsc553
7+f5//9bfEP2+YuwWKwS9BZ+rvm0gl7KwPrnRUFi3mjlZxEqwMofDG/QpuCG
m6Oi3HvE1VEu510oYPkFlh5K3QysIR2ecigFljkPP/hKqwIr68lfovEUpCXE
mv7mgTi0aHcL4pHDY202lxF7nzliHE7BaEgiL/QN4mqTs7JuFCxlSd+kt24B
1nOXL1x0KbB2c1x9ioc40MAuV4wCaZs9J5SKEU+dUh19QMK2VY2V2d2I23qv
J2eToBadbbttAvGMnomWEwnJpXqvmxYRF+vpZm0kgW2esOTMUgMWX1U8+AkB
/GDFoqvLaN3Kpdu/kIBLK9u9nr9F7Jo1O8UjYHLZcJL3O+JFolzSkgCHS4r+
Gp2IQx+eTtQgwPakoaVWGWKdNY4RSgR0fiz/qv6kcP2iWaYiAUMWpvrSwcL6
AqL+oAhgJPPMf3AUcuTFhZ0EKH6dIbXVSFiv2pVWHwJ0Ujb5SmghTl31/mom
Ad885awu5iDmxLT0dBIwGyt4s1UTsapMU/MqEp5W8u+mC/d7y2q+AhKqnraL
iBwQns9O/wt8EvZcnrcPjhOe59055VYSrIb7tZlK4XmEyde/I2H8xLoR7XHh
fjGVBXUKDvtull+h0PnopPi026P765wPvReMWPWrYZ9gCnQS/qOX1IBYMPV5
bhwFkdslnF+JEmh/7VxlArrPaHOxRnvEoSp7kyIoqBvV9zqXjzi1Iz7dkwIu
J/Vk+AjitpKU9h0UOMdsz7VWIBGvWV8uRUF6hdLrxL2IZ5IFW4dJUGDt9h+I
Qix/f6mygIT9AjvtkDzEqVfbyz1J6LR8pDBTgbh7w8gfKiQU2TZc+q4eseqm
b5VeEHBF/8FwxnXhfsPCpcsE2F/T7NpTLvy+67rlCAI4S89o2Xzhem66xV4C
emj3xPoExKxKy7PaBGg/DvM38EMsEWltqkKA/5tdfw6ZIYY4j08KBKwodUm+
ZCPmfKr9qEbAHfcM48LPEC8uzAzqo36IaJU1mEfPyyoLKHEnQDPoF+eyccT5
mj++TCZg3plnmvs74sbFDbXNBCSWHzS2R3WzOI8Vu98RcDDtlGfda8SLsu55
20hIK23vYcSF9Vi6B0WQ0OvgrFjBRSyTEBFXR4KXLG9DqSvioVsOaq9IMPk8
IlshC7HAL3F0MwV9Kgk/lz4U1j8gd2wPBa0BmYVJcmiOZVoFuoEUvDc+ZFLn
hTj/2dqaGDSPy8n7P1Qgngr0ovgUtDS3569bFHI03yWaAvKhmWjRLjaqV6Y/
zw/pxc0A/R+SEXsXVPXtpuAXidjypg7EU+HxvE0UaKU8S839gNihV9pmjISm
mv/FtbHVUb9VjpSXkVAq4RNgZ4VYZoCQO0xCe21S7D1PxN4JUzvRc87JnrDa
eRhx9SGpyWkCYt/VHSWOCPd3jNyuIcCzQ+AQ7otY/3sDx3gCnie3b0p1RNwt
P97rRIBxWOcnZUPEgZ7TkroEFCj5SW1TEa6X7g9VJf7VPcR8L/EH6H6Lhs+9
/fUFG6+3ZxiuetnJxvu/DjGaiqpi49+/sEP72k8FbPz/d1ZskkdOsXF9ktMx
DXdOs3H9MaaxSbolbPx8BxcK82sa2Pj5Le02HjcZZuPzeRStGhUkpo7PL3Ah
ycV8hzo+X75Fftexo+r4/Ed1mpe8W9Xx/ZiZGQXMiXPw/YmJHL0i5szB9ytx
+2ybbTEH379pfqhP7J8c3B86L/jfeujQuH/u/DeNNxFC4/46P539yuMyjftP
91rFuPMAjfszysNI79Eijfu3NOJIie16Lu5vvsv529UqXNz/6yqyW4K2cPF8
PM79puKAIhfPj1Harum7a7l4vowvht32+YvG8ycypj+rPEzj+Rxa0yqn3kzj
+e3XyDhz9xyN5/v7+xfjrfk0nv8hUdUn0wE01ofXFv36N1xorB+uXV+yw2xo
rC+FV6zc0q1prD9GjcyEtDON9WmCqhFkBdJYv6ZWbp3hptJY32qsx9bH19JY
/ypPdAfVT9BYHzc+iIr8ieBi/bRT6yr2C+BifSVTmj5YVnOx/v7YvVCxYYmL
9VnAh55YK43/67fh2kzdHA2s77kFzydZTzWw/ivFzBECRQb7QyNbU/DIicH+
0ei6tjrnOIP9ZR9ltLxcxmD/UXdKMPiuncH+JJ2kLdPbz2D/OksNfvQaZLC/
9UVPFSz0Mdj/Ji+Eve0V7v/XHwdmA5uOVTHYPzk3W7pmchjsr0arXe5bRzLY
f+sjOxv7DjDYn6vdxs88NmSwfyuKdMpnb2Gwv3uwuos3SzHY/xNzxr33iTA4
H/zWxBetWNHA+aHBsWe9uTiD88WzFF7drwoMzh8S85ZOvO0Mziez12/EjHkx
OL/Mnqv6eCSbwflGLGTG4kkvg/OPd3J8dqaMJs5H1Gklk+uumjg/1UUmtZCl
mjhfNQXriBXMaOL8davj3tXeHVo4n7GdBsMjjmvh/HaM+zMd06mF892nf/RP
G+e/vwEv4OIs
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5, 11.4452}, {0, -1}], 
         LineBox[{{11.499999999996362`, 3.500000000001819}, {7.5, 
          10.500000000001819`}}], 
         PolygonBox[{{9.2023166369859, 7.520945885274675}, {
          10.144980619863883`, 6.677509690068058}, {9.450386106164316, 
          6.280598539382591}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.759023129707089, 6.835784645546908}, \
{1, 1}], 
         {PointSize[0.04], PointBox[{5.5, 14.5}], PointBox[{15.5, 10.5}], 
          PointBox[{11.5, 3.5}], PointBox[{7.5, 10.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/becfdg/ehfgfhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/becfdg/ehfgfhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1AswXFcYB/APayUeQULbqU6yVTQRbTeUoUl6T5MQGtKtImRCxEiimg6p
lXp3E6FUp6U0WrNqdehEEGTGIxJmJYpqPDpjVISyiWejgshabGz/t9Mzc+fM
b77z/845987cVyNi/E/rE5E3Hn6mBR2GNaP/xhuMZJulfefhkTa7ZUP46hmH
C/9YMTpnmTTJmznPsmT48h1V25oTI41xY7kI1ox5uKhg5jIXNbaNUddeY/sm
WKZLDaqHRwK7n8hgxfODRXK4vWrVloNn3rlY8TUcqU5Ur+xmtHBT9GMyXOaW
llgHKw9ZV56ErVo7Ss7DTQ8rB8RwSsjOj/bz1jTT31sZzW2qu/4iLNl14EE2
3B7ocUmP7+dZLNwGD/3+YM+qIyP3kyt1OZaMslyj1Wuwwie9VGvBKFbSLeXX
ixze64uB3VXZBUI+v57XMGrOyHSftVQAxx770MoA9tUeC1AjrwwLswnZgvNm
TgfdhyV/3chTmeH+nssLVbDY96fgK/CmYIFeHByepT0VDTtNWVx+Cx7f433r
CDz0duuz6V3ov7vyhAM8MFwj/RlW9KQfmTbFdzA93hMJh5t3fPcFfG97TZ4L
72vuRVMmjCbcDCMsYXFeR5oNHLBvJH1jJ/pX6bvZGzO6/ZJ2UAOzJx98brkZ
+c9KHLVweLT/lSUj/jxBk/rIk66ybETIKCEjbFIIK5/O+z00RG7U5wJfV+bL
H1vCmlv5rovIL6xyXnEC+FPfO/2wJLA4VAgLGs++UgqLM9qi7hrgPgbKskjY
Qj5vJocntve27IBrvfNLU+H+6g2bodfRv0qyNQCe8X9NlQ/XrmtNzWBBfPsv
wXB4Ye9wkT7ef3dHsSMsHv4tU62HXMjFKRM4VhG/LILPpefs1TpgPx+P27bE
SPrtD1oNrExYLLfe4MhiubrwOTyedO300DpHSo/9TIC8qLDp1/pVjsqUiX8K
YcVhz4o/VjjyNvlGpcN6mbDV0UvNUUKE5NRjmD0yzdzxjCNfP6uETr5fXOqb
x5c5arr6fW0Bv5+YQh895ShKfnA2kK+nZZIcFsxIVOYwdTYUhfJmmUld9nCz
oxfBh/SLejJgxeD95fAlju41z7v6weyrrLkzixx1XVpKsrXnv29sudECRwXe
De5GfN3ZdTx3jqPaozEla3bo59HJkmc4khis56/CshsWrVsmOGLVR510fF1e
c9l5jKOskhCpIfKyCpN3BwZxP132CSF/HrP3c17o42hCT1Gl5dd/El328V3c
d6p3YIL3fGB8fz1HotnYlBbeuYcNzpZzpKjWNH7JO1bxckouznvz+qiX3f//
mdwWEvCzHfsXy6mzBg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{2.3732814251012426`, 12.844833813765145`}, {1, -1}], 
         LineBox[{{20.000000000002274`, 17.}, {5.5000000000027285`, 14.5}}], 
         PolygonBox[{{13.341276069446609`, 15.851944149904588`}, {
          12.090761163950333`, 16.042239896393152`}, {12.226686697156449`, 
          15.253871803797672`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.48664427941315, 16.672663179403738`}, {0, -1}], 
         LineBox[{{19.99999999999909, 10.000000000003638`}, {
          15.50000000000091, 10.5}}], 
         PolygonBox[{{17.15366975919583, 10.316258915644909`}, {
          18.39050285123411, 10.581294578224538`}, {18.302157630374232`, 
          9.786187590485644}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.92116886541601, 11.185719788744109`}, {0, -1}], LineBox[CompressedData["
1:eJwtlgs0lekaxz+Sy2aakFNKQpNpEKI0KL1R6eJeNpqh3SBEiaY1FTN9SEI4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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.617875450609192`, 1.6086826603562592`}, {0, 1}], 
         LineBox[{{5.5, 14.500000000009095`}, {7.5, 10.500000000010914`}}], 
         PolygonBox[{{6.231671842700025, 13.036656314599949`}, {
          6.410557280900009, 11.784458247200067`}, {7.1260990336999415`, 
          12.142229123600035`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.71843785395013, 12.411618926975063`}, \
{1, 1}], LineBox[CompressedData["
1:eJwd1ns41PkeB/BJlMqp2dxGSa4Z9xF1KJffpiSPyxCnIctULmOXTFtpijSk
behCoqSth0RD0riG1I5UhEqJpjaHom1IGVqXThfn/d0/Zn7Pa763z+c932cw
2B7nH6FEo9Eu4UWeNFXyZkrR2OSpRtEkZntuf1pB0Vi31kuZ/8LnNiYZxArD
KjpnIT7veaeYgTM3df2cuoiiqab7vWBgvWrdWnt/Op4u5jXOMK/ELuc53NpR
vTMWZiYXhfj8QNE8lqftL4YlzRzrKpiVlDTwFxnvcf519mKK1q/zIdqSSdEE
Snf5zjAzLjRrDywLzIvgwvK6i/caYf7uhaIYmFc9N0bJjKLlG5Wr7oClNx7P
c4PpwxuU3eF8JWZwItzv3h7DgGXuSrwyMj/MZqgX58sYcYVP4GndA9fy4Drj
svFhMr/ntddmmKPPvD0Fyy26j6rBmbGfJ4nzTzMY7eiPNSrUHoGZM1YHT8P6
y1fqdcOKfY7t0bDMsz6lGmZvP5DkB7M3TZoch1unXAK94XzeAedQWF+hsA2D
eV7Ra61gD5WUOSIyPtVc9QX9CuWMthaSb+qPVAfMSo/bvwz18EfkOwrI+OI+
g6MkT+1LcQdheWLB5Cz0m/t8iLed5BW6LS6d5MG59cmPrK9v26mrTtHEXuWU
N0xzEMRVwezOgZkAmJHgX+Opge9B90xQNMz0dU/uhVUdt4ykkfky3Q1Rmtj3
17VPasj5c1y7B2CWZnXRezLfc/0ZXy3kVyrJM0M/fN/gy1dghts2USzJ92XP
+BAsfdqkUgNzs/O2a2pj/boPKd/gXAb7kRlMO8+IWG+O+1I6S40JdxYrGx2F
eaf+HvgBZkZd5zYTJ1t/I/tlamvPmzIn+R+QVsGirIlwfQuK5mBqbrcbVqgn
WrnA/Y4W7ZbE39fl+cKZIWyXt6hfUtQdHQAL9d74FMDivsRBL1jftORsBCz6
YvbCERZJxUGrYUZxYbYuWc8o99aGHWYFdk3i/Pz68oQFsOB6UHkb7OFfoaMJ
07JzUvNgaUXJ7lWw1OFpPw/mRPj2xpA8Q7/+5ACLLqWVN5Dz7knSF8D0gbOe
y1Cv/I3Pw0HkIz1WYptFfOrOpvvk/i430dVCHtMso7YKWNa+V68IZpebXhXD
jB5Z1xoG9l+14eFVsv7DlpNP4VybpcdukvuXPJgWpYNcAnJqnsOdf0boT8B8
9Y85M+T7YSZ/jV+CevwTJmxRT91qhccQLHNVmMfCuWHbXnosRR+sw5bXYf45
Lf/TsMjctniCzDd78O8WmH/LQp3kzxuunTcIs923vjhiQc4zWzcMTy97ENEG
c145+/8J00JvesyzRF3aeln1MBXiy6bg6Y+RpiJYmsTPj7Uk+95940nG35u0
ZMCc4xu15sCS7KCOIrJ+pV3kbdTLPxHjdB1W5fqeEcCdLiZWZWT99/5AR5h1
u8npAkw1pqqpwJk/ORw7DIvqGKV9yIMzX3GNC4sfNSV0wLzI+IurYfba13Mf
wTIrxeW5sMelw+NvYeEKG9ce0t/K+Pca2I+ZeFCliNyv5HvFwTD9yMaT8TBr
w3RfDSwNfuHnDXONBqKMSV6fwjMsYL7pIadCWLZ/n706GXf9o8RaF/dLZ7hF
hYw/+L7rDszZax+kDMuXsL8GLcN5hs/3LCJ5n6j4TQFTD0P0TGFJpXJwsh7G
C+w6PUl9f2V8m7scuaZxk/fDAufnUUJYPPNQUkHud0Wt+Ts491rD3jH4bkls
h64+5ntmZ5P++V3CsTWwuDT4bBLJ4+arHG8yflAov0/yzPmlNQCmGxqPqllh
/y5NTT+40yHGwgcOyWhNcoNFYU/9RLBoMmqPFUwtWRleD8uaVzTR4Uxheksf
rG/9auy/qMfD50jg/2Buy5bmszBvhXK0qjXqfDFTzSb9GFu8nUf8pOzwQlg4
fn71DOYzhjRvdKN/4UiXzxBMaYXwS4kH+s49gIWJE06ZJJ+G7o8FsGRmcDQN
znS8lbAHVowOppyDxe7N19zgTt1umhRWtbyptZjUU7nlwDdYZrbI7w36V2zV
UfEjeZb6D9TCwtcd9+pg+rky4yxyXwdXK+mgP1W7jUbxcHXl5OUQmFn96kQ4
uU+7Z7sVwQ5O/ykIhZXfPbqggGkDG5ZGwF7erzlOBjjn2WNbAVxmU6j+G0yN
Bp86C8uHAw63w5Klhtw7sNMKxvH5hjg//NDGKdhY+HnLj3DdkmMp9qg/O/X7
zzthhV3c2D445veQqkxD8rtznvUHyafUNOcKLNzXWzkX+apdmVZU/jPu3OsD
a+S7mVTB09a7jbJgJ4nP/BJY7nHD6DEsXDcpzyH7+96XzrahaB2thTEJsKh3
wsEazk+ln9oKy4a2OXrDrc/OH3cg9R3kjoTBlm0GdRrkvDUXlSJhr5GgvePo
jx8srCXj9l+6NZ7BNNcrD8h6jqCx8SYsbihLtIX/LvT3u0ry0BAkq8HhjwO5
l+FceaO8n/SzRoUuhuU3di6ogMv25iy+QcZXLd91CBY7nbn4BGarhlWTfr82
XYiZIs7wiTSA5fTNDaaoj8nKGf+MvPjptf7bYCpyLPUlHLDV5/wlWBCfMnYX
5uy4UjRE+ht9lN8AN1b8st7eCPMDtwsbyXwh00gIi5il020w00Da1wazzZLl
b8n+il3Gi42xPvVcsRrOr3u6tWkznCmoVXWBU3eG3j0Be+yI5AtgyUDoydvG
5O9h5at62D7t2+Ag/M//g8hDgzxNqP8DT9CdfA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.848747069780032`, 6.4884873886971235`}, {-1, 1}], 
         LineBox[{{15.500000000000451`, 10.5}, {7.499999999999551, 10.5}}], 
         PolygonBox[{{10.9, 10.5}, {12.1, 10.9}, {12.1, 10.1}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5, 11.4452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwl1gs8lOkeB/C3ktQRk+uoUdON6UJitaNGTYW1bIyRT0MOw6Ys0hTJocu4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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.151252930219968, 6.4884873886971235`}, {1, 1}], 
         {PointSize[0.04], PointBox[{5.5, 14.5}], PointBox[{15.5, 10.5}], 
          PointBox[{11.5, 3.5}], PointBox[{7.5, 10.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/becfdg/ehfgfhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/becfdg/ehfgfhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwV0gtQlFUUB/ALiDxdCFEWWWh5yCOdyQCVKe37iodICLjRQBm0NBEMiCy6
8hpGdgqVVkgUUiLUjUlYimxzImBVWlSISGh5DGK4I88VcITlFQu4bv/7zeyc
+X3nnnvOvfu5f5opSDYlhEThRyMZMuLZwhI+hQNLwlt2967AqZI6vhROzXyp
18K5l5WXbDazZCg0crkdLvMseVIAywZX0s9TV1fJhuDic5mjLKzmhSfyHFlC
fsreNEwj57I6HDH1vbq7SYjxRde6aLTXrKUPo063x6w5A5Y7tb4RA+e6draJ
YcU5r/FOzFGm7G0qhIW9VX8dgnWtLhOX4BEyWzr5Cksq//1nTgXrzr+MrYCL
jbNGE8xB5iP9j8C6RI3PEdjSZ59ZEJw63RDSBXeW65q203qDdVjEVpbEODkS
V5jf5zTcD4c/yw7ZAktMeUcTnVii9w59YA4Hbe1lRmHROy1arT3W/+pxPZaL
vleq5m/Com8aV2/C8upi5xSY5QWmrcKK344pzag1twW+zpg/UlrLtYOTK/2D
YZGPxJXDQZ87jwJj4F2HfhxdsWWJSjSQEAeXuUdLn9rg/kLy85Jg2VTd5HNr
zJnGm8uGFVGht7hws2llzVVYbVAui6xY4vvHDisNrOo4cHzVEvV2hy8EbEPf
Cs61G3BQz2iWDFZYzyydgX0rNj11c0G/gCv7xLC8xzBQR526UZwMq/3ylP48
rPf8weEgdc7O/t9h/n1zoR1c5jL57k5XvM9MLGmwQF/1Q/dSWCX7JXgbHB70
etggrNjTdTp+I+59zFRh4YbzzsXXZpizhOsg+MoLHhEsiJM3oH+gmdLfjX5n
6/XRZpivKsrtTVjl49Kz3xTvzU/mB8O67+6Nsyb4fyPORsfBrFuNRwrBPY17
thTA/PUbXmIjQ9hCTX4z3d9KKM1/yZDKupoE21eRl2fdHjYwpGypfyEbFtpP
T56GLS/4tC/CqrbDOQlwbnmafj8f9f993s/Ceivh2S9hUpHUZQvHl9f4dcB8
8WLE9y+w/qO9GhN3fEeZavnkOkOKou6XBMCq2EHt32sMmYjqzv2Q5u9JKwSr
DBmSfXziBDwSF2IsWmGIrNxD+wXNW4oUecsMeXDyrRop9YFPjkYvMqTTxlry
NSxc3sH1m2fIhgj98EWYPZiyXTDLkOrNV02qab+lmbcfzTBEkev97c80v5Zj
tJtiCDe88LUuul/fqVMWWpyfrHQsUC9r+E0TDFGndk94e+B8faWthnHsH3Yx
6zNYNc1yamGdzROuHJY0Oqsy4IZjhsbnNM9/EcuBB66/P7rLE/UZzUVhY+j/
wa0zGbBknF0YG2HIUnuTSEbdXm8l1TBE9Gy98U+6fvddffdDhth61q6NUR+v
T29RM+Tx3sKCRWoTnWNIG/Z/XDW1Tk0f2R1iS6MX+z+s7NlG
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{2.2587844891608513`, 6.758784489160852}, {1, 1}], 
         LineBox[{{20.000000000000455`, 17.000000000003638`}, {
          10.999999999999545`, 18.000000000003638`}}], 
         PolygonBox[{{16.09633024080417, 17.43374108435509}, {
          14.947842369625768`, 17.963812409514357`}, {14.85949714876589, 
          17.16870542177546}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.671168865416012`, 18.43571978874411}, {0, -1}], 
         LineBox[{{20.000000000009095`, 10.000000000003638`}, {
          5.500000000009095, 4.500000000003638}}], 
         PolygonBox[{{12.18900148420933, 7.037207459527677}, {
          13.169136822142455`, 7.836791550999437}, {13.452860209438885`, 
          7.088793529945209}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.588731357443775`, 5.370907180173249}, {0, 1}], LineBox[CompressedData["
1:eJwt1gtUjOkfB/A31ZoQ05g0kprpYqebZpNE0UuhRRqZTGi7kFZWacmaUhq6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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.79984196030395, 5.1006801186071264`}, {1, 1}], 
         LineBox[{{5.5, 4.5000000000027285`}, {7., 11.000000000005457`}}], 
         PolygonBox[{{6.384915704019255, 8.334634717416773}, {
          6.504840774258593, 7.075421479903725}, {5.7253278177028974`, 
          7.2553090852627315`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.361029083676273, 9.056371337494769}, \
{1, 0}], LineBox[{{10.999999999992724`, 18.000000000007276`}, {
          14.999999999992724`, 11.000000000007276`}}], 
         PolygonBox[{{12.7023166369859, 15.020945885274676`}, {
          12.950386106164316`, 13.780598539382591`}, {13.644980619863883`, 
          14.177509690068057`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740976870292911`, 14.664215354453091`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1ns8VHkfB/DTpE2Zp6bY0m1yNxiMlKaiTigSorSLJsmiUampVSlCSU3Y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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.259023129707089, 14.664215354453091`}, {1, -1}], 
         LineBox[{{15.000000000000451`, 11.}, {6.999999999999551, 11.}}], 
         PolygonBox[{{11.6, 11.}, {10.4, 11.4}, {10.4, 10.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11., 10.0548}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.5, 4.5}], PointBox[{11., 18.}], 
          PointBox[{15., 11.}], PointBox[{7., 11.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["3",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcedg/ehfgfhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcedg/ehfgfhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwV0gtQlFUUB/ALiDxdCFEWWWh5yCOdyQCVKe37iodICLjRQBm0NBEMiCy6
8hpGdgqVVkgUUiLUjUlYimxzImBVWlSISGh5DGK4I88VcITlFQu4bv/7zeyc
+X3nnnvOvfu5f5opSDYlhEThRyMZMuLZwhI+hQNLwlt2967AqZI6vhROzXyp
18K5l5WXbDazZCg0crkdLvMseVIAywZX0s9TV1fJhuDic5mjLKzmhSfyHFlC
fsreNEwj57I6HDH1vbq7SYjxRde6aLTXrKUPo063x6w5A5Y7tb4RA+e6draJ
YcU5r/FOzFGm7G0qhIW9VX8dgnWtLhOX4BEyWzr5Cksq//1nTgXrzr+MrYCL
jbNGE8xB5iP9j8C6RI3PEdjSZ59ZEJw63RDSBXeW65q203qDdVjEVpbEODkS
V5jf5zTcD4c/yw7ZAktMeUcTnVii9w59YA4Hbe1lRmHROy1arT3W/+pxPZaL
vleq5m/Com8aV2/C8upi5xSY5QWmrcKK344pzag1twW+zpg/UlrLtYOTK/2D
YZGPxJXDQZ87jwJj4F2HfhxdsWWJSjSQEAeXuUdLn9rg/kLy85Jg2VTd5HNr
zJnGm8uGFVGht7hws2llzVVYbVAui6xY4vvHDisNrOo4cHzVEvV2hy8EbEPf
Cs61G3BQz2iWDFZYzyydgX0rNj11c0G/gCv7xLC8xzBQR526UZwMq/3ylP48
rPf8weEgdc7O/t9h/n1zoR1c5jL57k5XvM9MLGmwQF/1Q/dSWCX7JXgbHB70
etggrNjTdTp+I+59zFRh4YbzzsXXZpizhOsg+MoLHhEsiJM3oH+gmdLfjX5n
6/XRZpivKsrtTVjl49Kz3xTvzU/mB8O67+6Nsyb4fyPORsfBrFuNRwrBPY17
thTA/PUbXmIjQ9hCTX4z3d9KKM1/yZDKupoE21eRl2fdHjYwpGypfyEbFtpP
T56GLS/4tC/CqrbDOQlwbnmafj8f9f993s/Ceivh2S9hUpHUZQvHl9f4dcB8
8WLE9y+w/qO9GhN3fEeZavnkOkOKou6XBMCq2EHt32sMmYjqzv2Q5u9JKwSr
DBmSfXziBDwSF2IsWmGIrNxD+wXNW4oUecsMeXDyrRop9YFPjkYvMqTTxlry
NSxc3sH1m2fIhgj98EWYPZiyXTDLkOrNV02qab+lmbcfzTBEkev97c80v5Zj
tJtiCDe88LUuul/fqVMWWpyfrHQsUC9r+E0TDFGndk94e+B8faWthnHsH3Yx
6zNYNc1yamGdzROuHJY0Oqsy4IZjhsbnNM9/EcuBB66/P7rLE/UZzUVhY+j/
wa0zGbBknF0YG2HIUnuTSEbdXm8l1TBE9Gy98U+6fvddffdDhth61q6NUR+v
T29RM+Tx3sKCRWoTnWNIG/Z/XDW1Tk0f2R1iS6MX+z+s7NlG
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{2.2587844891608513`, 6.758784489160852}, {1, 1}], 
         LineBox[{{20.000000000000455`, 17.000000000003638`}, {
          10.999999999999545`, 18.000000000003638`}}], 
         PolygonBox[{{16.09633024080417, 17.43374108435509}, {
          14.947842369625768`, 17.963812409514357`}, {14.85949714876589, 
          17.16870542177546}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.671168865416012`, 18.43571978874411}, {0, -1}], 
         LineBox[{{20.000000000009095`, 10.000000000003638`}, {
          5.500000000009095, 4.500000000003638}}], 
         PolygonBox[{{12.18900148420933, 7.037207459527677}, {
          13.169136822142455`, 7.836791550999437}, {13.452860209438885`, 
          7.088793529945209}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.588731357443775`, 5.370907180173249}, {0, 1}], LineBox[CompressedData["
1:eJwt1gtUjOkfB/A31ZoQ05g0kprpYqebZpNE0UuhRRqZTGi7kFZWacmaUhq6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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.79984196030395, 5.1006801186071264`}, {1, 1}], 
         LineBox[{{5.5, 4.5000000000027285`}, {7., 11.000000000005457`}}], 
         PolygonBox[{{6.384915704019255, 8.334634717416773}, {
          6.504840774258593, 7.075421479903725}, {5.7253278177028974`, 
          7.2553090852627315`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.361029083676273, 9.056371337494769}, \
{1, 0}], LineBox[CompressedData["
1:eJwd1gtUzOkbB/DRShNSSppkqS10mTIlTLoYROlUqg3pwnTdKaWRaFojE1kp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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.348747069780032`, 15.011512611302877`}, {-1, -1}], 
         LineBox[{{11.000000000003638`, 18.000000000009095`}, {
          7.000000000003638, 11.000000000009095`}}], 
         PolygonBox[{{9.2976833630141, 15.020945885274676`}, {
          8.355019380136117, 14.177509690068057`}, {9.049613893835684, 
          13.780598539382591`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{8.259023129707089, 14.664215354453091`}, {1, -1}], LineBox[CompressedData["
1:eJxFlgs0lWkXx08kVpNLxYRB4n1f7/G6nGkkJHa5JrnEKZfJJRkU4zLjEuKY
hIzBqDCVW6mE5JIImYrjMqUOJTRENS7f+GrEiMplnjPr6/nOWmed9VvP2efs
dz97//9708Hgvb4iLBarCL2Fn6uWltHrC2D9+6IgIWeo/LNwJWDl9oXV6VJw
09VRUe494spI7gUuBSwf/+JDKcrA6ucEKYdQkE4VzPJVVICV+fRv0TgKZJff
Jd2sQxxSsLMJcX1vW0KO3UZgeZ09YhJGQcOwyO8jzxFXmp6TdaVgIVP6Fnuz
KrBGuF9w9SiwcXVceToIsb+hXbYYBdK7d51QKkQ8cVp16CEJW1bUl2d1IL7T
dSMpiwS1qCzbLWOIp/RNdZxISCrWf9Mwj7hQXy9zPQmURfyCM2sTsHiq4oFP
CeAFKhZcW0Tn1twO33wCLi9v9Rx5i9glc3oiiIDxRaPxoOeI59VLJa0IcLis
6MvwEYc8OpPAEGB70shKpwQxZ5VjuBIB/I+lX9WeFJ5fMs9QJKDf0sxAOlCY
n1/kHyQBWpI5Fj84Cjni0tx2AhS/TpfabCzMV+1qszcBnOQNByV0EKeseH8t
g4BvntErC2nEdHRTJ5+A6RjBX5u1EavKNDSuIOFZOa8tTRjvJav9GkioeNYi
IrJfWJ/tvhd5JOy6MmsfGCusZ9uMcjMJ1gM9ulrlwnqEyte+I2H0xJpB3VFh
vJjKnAYFhw9ulF8mUX04yd4t9hSk8mdD7gciVv1qwDuQAk78r/qJdYgFE59n
x1IQsVXC+bWoGoqvnimPp2AoykKs3h5xiMqexHAKaoYMPM/nIk5pjUvzoECT
TjkZNoj4TlFyyzYKnKO3ZtsoqCNetbZUioK0MqU3CXsQTyUJNg+QoMDa6dsb
iVj+wUJ5Hgn7BHa6wTmIU661lHqQwLd6rDBVhrhj3eAfKiQU2NZd/q4WseqG
b5VeEHDV4OFA+g1hvFH+whUC7K9rt+8qFX7fZc1iOAH0wjBbNld4np1muYeA
TrZbQm08Yla51TldAnSfhPoa+iCWiLAxUyHA968df/abI4ZY9yUFApaV2iVf
UojppeqPagTcc0s3yf8M8fzcVJ8B6ofwZlnDWfS8rBK/IjcCtAN+cS4ZRZyr
/ePLJAJmnYPMsp8jrp9fV91IQELpARP7F4jpJ4od7wg4kHrao+YN4nlZt5wt
JKQWt3RqiQvzsXILCCehy8FZsUwTsUx8eGwNCZ6yQeuKXRD333ZQe02C6efh
WQqZiAU+CUMbKehWif+5+JEw/165Y7soaPbLyE+UI1B8s0DPn4L3JodMazwR
5w6vropG87iYtO9DGeIJf0+SR0FTY0vumnkhR/G4URQQj8xFC3aQKF+Znhwf
pBe3/Ax+SELslVfRvZOCXyRiShtaEU+ExQVtoEAneTgl+wNihy7p3a9IaKj6
b+wdCukIp3ywtISEYglvPztrxDK96nKHSWipToy5j/qG5RU/sV2ThBnZE9bb
DyOuPCQ1PklAzLuao+pHhPGtg3erCPBoFTiEHURs8L2hYxwBI0ktG1IcEXfI
j3Y5EWASyl9SNkLs7zEpqUdAnpKP1BYV4XnxvhBVAusei+cp/hDdb8HA+be/
vSDxeUu60YqXfBLHfx1sPBFZQeLfv7hN9/pPqE8//f+95d1Jg6dJnJ/kZHTd
vTMkzj/aLCZRr4jEz3dgLj+3qo7Ez29lt/646QCJ6/M4SjUyAOngp/r5zyVy
LdAcfaovzzK3/dhRCtd/iNO44NVM4fsxNzf2mxHXwPcnJnL0qpizBr5fibvn
7tgWauD7N8sN8Y75UwP3B+cF71t3Do37595/UoPGgmncXxcms167X6Fx/+ld
Lxt17qVxf0a6G+s/nqdx/xaHHymyXcvG/c3jXrhbqcLG/b+mLKspYBMbz8eT
7G/K9iuy8fwYp+6YbFvNxvNlcin0rvffNJ4/kVcG08oDNJ7P/lXNchqNNJ7f
Hib9bNt5Gs/39w8uxdnwaDz//aKqTyf9aKwPbyx7DG5yaawfLu1fUqG7aawv
+VetXdNsaKw/xvVaY9LONNanMbJKkOlPY/2aWL59VjOFxvpWZfNqbVw1jfWv
/ERHQO0YjfVx/cPIiJ/U2Vg/7dTaC3382FhfieSGD1aVbKy/P3bMla1bYGN9
FvCgM8Za8//6bbQ6Q++UJtb37LyRcdYzTaz/StEz6gJFBvtDPaUteOzEYP+o
d1ldeeo4g/1lL2m8uFjCYP/RcIo3/K6Fwf4knagr09XDYP86R/Z99OxjsL91
R03kzXUz2P/GL4a+7RLG/88fe6f9G45VMNg/6VtN7VOnGOyvxiu5D2wiGOy/
tRH8+u79DPbnStfRs0+MGOzfiiJ8+axNDPZ3d1ZH4UYpBvt/wqlRr70iDN4P
fm/giZYta+L9oc6xc62FOIP3i+HkoJrfFBi8f0jMWjkFbWXwfjJ942b0K08G
7y/T5ys+Hsli8H4jFjxl+bSLwfuPV1JcVoaMFt6PyDNKpjdctPD+VBOR2EQU
a+H9qiGQI5Y3pYX3r9ut9691bdPG+xnl1BcWflwb72/HNH9mR/O18X639K/+
6WAd/AcNXeO1
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11., 9.354800000000001}, {0, 1}], 
         {PointSize[0.04], PointBox[{5.5, 4.5}], PointBox[{11., 18.}], 
          PointBox[{15., 11.}], PointBox[{7., 11.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["4",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcedg/ehfgfhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcedg/ehfgfhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt0wss1VEcB/Af6V6Pcr3arPW4uHm0yJU8cu2cUV5rY4bWYzczrx6bx50w
JSSPycZQirge5TF5N1qq22SubGIYSR5x1/1TiDuJRr9/dbb/fvvsnPP7/s/5
728SEuUfpg4ADviwFeZ3cBhS+Dv4FGj62/YedGTOtLM7Wq5vw6Sj87Sc/JPR
95Z5jk5op7Z3WfXod+V2rpMGFBJchqu60fPLLnbRaNuR0Ro5+pJokrOuTyH6
qKF6F1o0lqkfi5YunTxchVY5c/wX9CjwGentu+jSgKHrYeiZgoyYa+jaxB2j
YR6FlWTD2kvs+0yEF6/rUtA7pPwmRsuUixqn0Hm3agRxaD+JT37pXgqD/R5D
ZWilNeeYMVp6ItZiHB0t5dyp3ENhIyKsbb8J7stRddugjVddY8ToQV+3viYd
PH+8vVUJWml3LnA/WiqsHulDFxcvbYZrU/AK8P6gQMOE/HKBFuYCI/3OGpjK
Sk3MmbJynUM3a2iMlnIpWLb0NvajvQZMHuVyMMc0Q9TI9n8xfzxtN4XODp51
AZsv8rHJ1KCQFRRxOY1dfyE99OkurKk/VKlsflmvjUId73vzoGcR218RkuuO
ptPzgV1svnadoEsN+36R966hbYdbjpxB+5mKk06a4rmuSDaGAO93rTI/Ea3p
phAHojV1fnZ3oG3HqiMzdwgE69T2K9GwFTVgtU3A9r1uGdcMv0NKfVPObwLN
9x94GqDzmE9vWjYJdLZlU22038Xwj/UbBJSzR61VuD9Y6J2Qu05gpWSf0Rha
Wu3PSFQEZDe3jr1E89vTh+NX0fYVSbVo2VTR52cr2K+vXvKEdUno+dNLBOic
78xz1sLsuoZFAnqPs8kE28+hySKNISC90drAw3xwlMjLvxLgzziJ/dAz215u
IwoCKR68ikJ23vzV1YU5ApZN1hYjaL6dh6h1FvN4wdFcAYUUh7O/9KdwXs1c
zwoNPTnur8cIjJsFPXRAy3QWmbhBAnImkytk13fJuVMyAgnC4MIDgv//UfGr
f1VA/wB3f0Mb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.075655548348861, 9.834338870872152}, \
{0, -1}], 
         LineBox[{{19.999999999996362`, 17.}, {14.999999999992724`, 
          13.499999999998181`}}], 
         PolygonBox[{{17.991539152311425`, 15.594077406617997`}, {
          16.779075909943245`, 15.233615361589619`}, {17.237845785433905`, 
          14.578229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.21593336623684, 15.91500947680451}, \
{1, -1}], LineBox[{{20., 10.}, {15., 13.5}}], 
         PolygonBox[{{17.008460847688575`, 12.094077406617997`}, {
          18.220924090056755`, 11.733615361589619`}, {17.762154214566095`, 
          11.078229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.784066633763157`, 12.415009476804514`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1gs01fkWB/C/Z4qDo6TyOjKTR0RDI1T+ck6r3KxUborUKZOkIuU5MfOv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          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.72092184717428, 1.8721747773942476`}, {0, 1}], LineBox[CompressedData["
1:eJw91AtIU1EcBvDrjLY0dVo6e2imlctHM5KQXHbViTNKnVhKz1laI3oYWmpJ
XXM+yoQR1pYtHSXZw2yIOSmVykfLoCfS1HQrKzeqsWjUrKy+U9FgfPx2Lv97
9nHPnb9tX1oOi6IoEb4k/3749P/k1l/MmYss7HdKTkBGdo9eFyEzXcNa5Ejp
zwB9AVmPSLlwG6ldFh/UimR+XXWMkzlZwXe+Ik1x+W6cxTRV3J2XRCMND0r6
ZyNFP5MFVUh90/L2eUht/PPeAaSNz3XmIR1xj7L9QmgqaqKmyAlmYtMrtsLm
tq5jJsy9zLphOQtHpKW4tcHpA6KGfph+bSgk+9OkRMXZYK7z8aJkuGZV67Np
oZgXwnb4wPYuH7sP3D4xZjUFY87iAgMPVu2MTLwBM0pJvBvMXzt5Uw7vfpLH
ODCPCnJRb4f1Vcr0YVg62GRZC8uGpWYd/GRpGzsBrmTdalbAss+6viRYO9GR
KINzGzLWbYJ/9JdmxcK+ygUxxWTdLFH5wzaby57LcEOZIoNFrv8VKDTCb9Lc
O63oIfOpqMIf+xdmGo6+Jf11FvRuhw2cF8PjcKXFI7gJ1i8LLP0CUytn6uyw
ovbWaU/Maw+WHlmB39dYXTVRsGbIeOkQ6ZtO0uwi/SWKzrfAmm3ZXpfgO+I6
gYms2zxZFtK/x4FGJ3Ldc+fApehHqpg/6g0H3EsYPwwreopXzyXrqQcb+0JJ
DzFssi4ey293D8O8ms6nFFw86KqQwAZJiNMrzM/f4f/6JGzqqT2ogwOqP9/u
ggv36srLYbXPhnNvYfN99mgq/PLdKYYKpylObmjWLFhBh39zh02SIpcx/N8I
RhziCWt3nu/TkuczerKHDVOKQUkZvMBKceyYx48u5/3pTzl+7QUsM6Z7r4E/
7Gh53ApH7b3SHUtM1RmryX5LgqaQ8yDM2liYA2tKF/HWk+exlyOOgbmTL1X5
5Ly832CdTcwMldXD8trckUn0wfAXJg7AnIf3hBbYd4TXO4M8/x0nHo/CEYKV
TCYsL3pkNMK2Z9+bNXA2103wETZvbd1igaf7xX6cSu7vff+MAH0KRxrawmBa
/epwLuy4mabaDPvGf9p/BZ6+uWqOivRHL4o2kHPTPCdjCNYLlV7fYTU9Y2og
+lFJxHe5uE9HSeeF3bBm17kmX9ierZbrYFn4kREveAp5Tyz5974IpX8DNP98
/A==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.411618926975063`, 13.531562146049868`}, {1, -1}], 
         LineBox[{{4.999999999992724, 8.000000000003638}, {
          11.999999999998181`, 4.000000000003638}}], 
         PolygonBox[{{7.979054114725325, 6.2976833630141}, {8.822490309931942,
           5.355019380136117}, {9.21940146061741, 6.049613893835684}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.335784645546907, 5.259023129707088}, \
{1, 1}], LineBox[{{4.999999999992724, 7.999999999992724}, {12., 
          11.999999999996362`}}], 
         PolygonBox[{{9.020945885274676, 10.2976833630141}, {
          8.177509690068058, 9.355019380136117}, {7.780598539382591, 
          10.049613893835684`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.335784645546907, 10.740976870292911`}, {1, -1}], 
         LineBox[{{12., 3.9999999999902114`}, {12., 11.999999999982228`}}], 
         PolygonBox[{{12., 7.4}, {12.4, 8.6}, {11.6, 8.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.9452, 8.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{5., 8.}], PointBox[{15., 13.5}], 
          PointBox[{12., 4.}], PointBox[{12., 12.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["5",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcfdg/fhegehgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcfdg/fhegehgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt0wss1VEcB/Af6V6Pcr3arPW4uHm0yJU8cu2cUV5rY4bWYzczrx6bx50w
JSSPycZQirge5TF5N1qq22SubGIYSR5x1/1TiDuJRr9/dbb/fvvsnPP7/s/5
728SEuUfpg4ADviwFeZ3cBhS+Dv4FGj62/YedGTOtLM7Wq5vw6Sj87Sc/JPR
95Z5jk5op7Z3WfXod+V2rpMGFBJchqu60fPLLnbRaNuR0Ro5+pJokrOuTyH6
qKF6F1o0lqkfi5YunTxchVY5c/wX9CjwGentu+jSgKHrYeiZgoyYa+jaxB2j
YR6FlWTD2kvs+0yEF6/rUtA7pPwmRsuUixqn0Hm3agRxaD+JT37pXgqD/R5D
ZWilNeeYMVp6ItZiHB0t5dyp3ENhIyKsbb8J7stRddugjVddY8ToQV+3viYd
PH+8vVUJWml3LnA/WiqsHulDFxcvbYZrU/AK8P6gQMOE/HKBFuYCI/3OGpjK
Sk3MmbJynUM3a2iMlnIpWLb0NvajvQZMHuVyMMc0Q9TI9n8xfzxtN4XODp51
AZsv8rHJ1KCQFRRxOY1dfyE99OkurKk/VKlsflmvjUId73vzoGcR218RkuuO
ptPzgV1svnadoEsN+36R966hbYdbjpxB+5mKk06a4rmuSDaGAO93rTI/Ea3p
phAHojV1fnZ3oG3HqiMzdwgE69T2K9GwFTVgtU3A9r1uGdcMv0NKfVPObwLN
9x94GqDzmE9vWjYJdLZlU22038Xwj/UbBJSzR61VuD9Y6J2Qu05gpWSf0Rha
Wu3PSFQEZDe3jr1E89vTh+NX0fYVSbVo2VTR52cr2K+vXvKEdUno+dNLBOic
78xz1sLsuoZFAnqPs8kE28+hySKNISC90drAw3xwlMjLvxLgzziJ/dAz215u
IwoCKR68ikJ23vzV1YU5ApZN1hYjaL6dh6h1FvN4wdFcAYUUh7O/9KdwXs1c
zwoNPTnur8cIjJsFPXRAy3QWmbhBAnImkytk13fJuVMyAgnC4MIDgv//UfGr
f1VA/wB3f0Mb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.075655548348861, 9.834338870872152}, \
{0, -1}], 
         LineBox[{{19.999999999996362`, 17.}, {14.999999999992724`, 
          13.499999999998181`}}], 
         PolygonBox[{{17.991539152311425`, 15.594077406617997`}, {
          16.779075909943245`, 15.233615361589619`}, {17.237845785433905`, 
          14.578229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.21593336623684, 15.91500947680451}, \
{1, -1}], LineBox[{{20., 10.}, {15., 13.5}}], 
         PolygonBox[{{17.008460847688575`, 12.094077406617997`}, {
          18.220924090056755`, 11.733615361589619`}, {17.762154214566095`, 
          11.078229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.784066633763157`, 12.415009476804514`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1gs01fkWB/C/Z4qDo6TyOjKTR0RDI1T+ck6r3KxUborUKZOkIuU5MfOv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          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.72092184717428, 1.8721747773942476`}, {0, 1}], LineBox[CompressedData["
1:eJw91AtIU1EcBvDrjLY0dVo6e2imlctHM5KQXHbViTNKnVhKz1laI3oYWmpJ
XXM+yoQR1pYtHSXZw2yIOSmVykfLoCfS1HQrKzeqsWjUrKy+U9FgfPx2Lv97
9nHPnb9tX1oOi6IoEb4k/3749P/k1l/MmYss7HdKTkBGdo9eFyEzXcNa5Ejp
zwB9AVmPSLlwG6ldFh/UimR+XXWMkzlZwXe+Ik1x+W6cxTRV3J2XRCMND0r6
ZyNFP5MFVUh90/L2eUht/PPeAaSNz3XmIR1xj7L9QmgqaqKmyAlmYtMrtsLm
tq5jJsy9zLphOQtHpKW4tcHpA6KGfph+bSgk+9OkRMXZYK7z8aJkuGZV67Np
oZgXwnb4wPYuH7sP3D4xZjUFY87iAgMPVu2MTLwBM0pJvBvMXzt5Uw7vfpLH
ODCPCnJRb4f1Vcr0YVg62GRZC8uGpWYd/GRpGzsBrmTdalbAss+6viRYO9GR
KINzGzLWbYJ/9JdmxcK+ygUxxWTdLFH5wzaby57LcEOZIoNFrv8VKDTCb9Lc
O63oIfOpqMIf+xdmGo6+Jf11FvRuhw2cF8PjcKXFI7gJ1i8LLP0CUytn6uyw
ovbWaU/Maw+WHlmB39dYXTVRsGbIeOkQ6ZtO0uwi/SWKzrfAmm3ZXpfgO+I6
gYms2zxZFtK/x4FGJ3Ldc+fApehHqpg/6g0H3EsYPwwreopXzyXrqQcb+0JJ
DzFssi4ey293D8O8ms6nFFw86KqQwAZJiNMrzM/f4f/6JGzqqT2ogwOqP9/u
ggv36srLYbXPhnNvYfN99mgq/PLdKYYKpylObmjWLFhBh39zh02SIpcx/N8I
RhziCWt3nu/TkuczerKHDVOKQUkZvMBKceyYx48u5/3pTzl+7QUsM6Z7r4E/
7Gh53ApH7b3SHUtM1RmryX5LgqaQ8yDM2liYA2tKF/HWk+exlyOOgbmTL1X5
5Ly832CdTcwMldXD8trckUn0wfAXJg7AnIf3hBbYd4TXO4M8/x0nHo/CEYKV
TCYsL3pkNMK2Z9+bNXA2103wETZvbd1igaf7xX6cSu7vff+MAH0KRxrawmBa
/epwLuy4mabaDPvGf9p/BZ6+uWqOivRHL4o2kHPTPCdjCNYLlV7fYTU9Y2og
+lFJxHe5uE9HSeeF3bBm17kmX9ierZbrYFn4kREveAp5Tyz5974IpX8DNP98
/A==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.411618926975063`, 13.531562146049868`}, {1, -1}], 
         LineBox[{{4.999999999992724, 8.000000000003638}, {
          11.999999999998181`, 4.000000000003638}}], 
         PolygonBox[{{9.020945885274676, 5.7023166369859}, {7.780598539382591,
           5.950386106164316}, {8.177509690068058, 6.644980619863883}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.335784645546907, 5.259023129707088}, \
{1, 1}], LineBox[{{4.999999999992724, 7.999999999992724}, {12., 
          11.999999999996362`}}], 
         PolygonBox[{{7.979054114725325, 9.7023166369859}, {9.21940146061741, 
          9.950386106164316}, {8.822490309931942, 10.644980619863883`}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.335784645546907, 10.740976870292911`}, {1, -1}], 
         LineBox[{{12., 3.9999999999902114`}, {12., 11.999999999982228`}}], 
         PolygonBox[{{12., 8.6}, {12.4, 7.4}, {11.6, 7.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.9452, 8.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{5., 8.}], PointBox[{15., 13.5}], 
          PointBox[{12., 4.}], PointBox[{12., 12.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["6",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcfdg/fhegehgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcfdg/fhegehgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt0wss1VEcB/Af6V6Pcr3arPW4uHm0yJU8cu2cUV5rY4bWYzczrx6bx50w
JSSPycZQirge5TF5N1qq22SubGIYSR5x1/1TiDuJRr9/dbb/fvvsnPP7/s/5
728SEuUfpg4ADviwFeZ3cBhS+Dv4FGj62/YedGTOtLM7Wq5vw6Sj87Sc/JPR
95Z5jk5op7Z3WfXod+V2rpMGFBJchqu60fPLLnbRaNuR0Ro5+pJokrOuTyH6
qKF6F1o0lqkfi5YunTxchVY5c/wX9CjwGentu+jSgKHrYeiZgoyYa+jaxB2j
YR6FlWTD2kvs+0yEF6/rUtA7pPwmRsuUixqn0Hm3agRxaD+JT37pXgqD/R5D
ZWilNeeYMVp6ItZiHB0t5dyp3ENhIyKsbb8J7stRddugjVddY8ToQV+3viYd
PH+8vVUJWml3LnA/WiqsHulDFxcvbYZrU/AK8P6gQMOE/HKBFuYCI/3OGpjK
Sk3MmbJynUM3a2iMlnIpWLb0NvajvQZMHuVyMMc0Q9TI9n8xfzxtN4XODp51
AZsv8rHJ1KCQFRRxOY1dfyE99OkurKk/VKlsflmvjUId73vzoGcR218RkuuO
ptPzgV1svnadoEsN+36R966hbYdbjpxB+5mKk06a4rmuSDaGAO93rTI/Ea3p
phAHojV1fnZ3oG3HqiMzdwgE69T2K9GwFTVgtU3A9r1uGdcMv0NKfVPObwLN
9x94GqDzmE9vWjYJdLZlU22038Xwj/UbBJSzR61VuD9Y6J2Qu05gpWSf0Rha
Wu3PSFQEZDe3jr1E89vTh+NX0fYVSbVo2VTR52cr2K+vXvKEdUno+dNLBOic
78xz1sLsuoZFAnqPs8kE28+hySKNISC90drAw3xwlMjLvxLgzziJ/dAz215u
IwoCKR68ikJ23vzV1YU5ApZN1hYjaL6dh6h1FvN4wdFcAYUUh7O/9KdwXs1c
zwoNPTnur8cIjJsFPXRAy3QWmbhBAnImkytk13fJuVMyAgnC4MIDgv//UfGr
f1VA/wB3f0Mb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.075655548348861, 9.834338870872152}, \
{0, -1}], 
         LineBox[{{19.999999999996362`, 17.}, {14.999999999992724`, 
          13.499999999998181`}}], 
         PolygonBox[{{17.991539152311425`, 15.594077406617997`}, {
          16.779075909943245`, 15.233615361589619`}, {17.237845785433905`, 
          14.578229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.21593336623684, 15.91500947680451}, \
{1, -1}], LineBox[{{20., 10.}, {15., 13.5}}], 
         PolygonBox[{{17.008460847688575`, 12.094077406617997`}, {
          18.220924090056755`, 11.733615361589619`}, {17.762154214566095`, 
          11.078229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.784066633763157`, 12.415009476804514`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1gs01fkWB/C/Z4qDo6TyOjKTR0RDI1T+ck6r3KxUborUKZOkIuU5MfOv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          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.72092184717428, 1.8721747773942476`}, {0, 1}], LineBox[CompressedData["
1:eJw91AtIU1EcBvDrjLY0dVo6e2imlctHM5KQXHbViTNKnVhKz1laI3oYWmpJ
XXM+yoQR1pYtHSXZw2yIOSmVykfLoCfS1HQrKzeqsWjUrKy+U9FgfPx2Lv97
9nHPnb9tX1oOi6IoEb4k/3749P/k1l/MmYss7HdKTkBGdo9eFyEzXcNa5Ejp
zwB9AVmPSLlwG6ldFh/UimR+XXWMkzlZwXe+Ik1x+W6cxTRV3J2XRCMND0r6
ZyNFP5MFVUh90/L2eUht/PPeAaSNz3XmIR1xj7L9QmgqaqKmyAlmYtMrtsLm
tq5jJsy9zLphOQtHpKW4tcHpA6KGfph+bSgk+9OkRMXZYK7z8aJkuGZV67Np
oZgXwnb4wPYuH7sP3D4xZjUFY87iAgMPVu2MTLwBM0pJvBvMXzt5Uw7vfpLH
ODCPCnJRb4f1Vcr0YVg62GRZC8uGpWYd/GRpGzsBrmTdalbAss+6viRYO9GR
KINzGzLWbYJ/9JdmxcK+ygUxxWTdLFH5wzaby57LcEOZIoNFrv8VKDTCb9Lc
O63oIfOpqMIf+xdmGo6+Jf11FvRuhw2cF8PjcKXFI7gJ1i8LLP0CUytn6uyw
ovbWaU/Maw+WHlmB39dYXTVRsGbIeOkQ6ZtO0uwi/SWKzrfAmm3ZXpfgO+I6
gYms2zxZFtK/x4FGJ3Ldc+fApehHqpg/6g0H3EsYPwwreopXzyXrqQcb+0JJ
DzFssi4ey293D8O8ms6nFFw86KqQwAZJiNMrzM/f4f/6JGzqqT2ogwOqP9/u
ggv36srLYbXPhnNvYfN99mgq/PLdKYYKpylObmjWLFhBh39zh02SIpcx/N8I
RhziCWt3nu/TkuczerKHDVOKQUkZvMBKceyYx48u5/3pTzl+7QUsM6Z7r4E/
7Gh53ApH7b3SHUtM1RmryX5LgqaQ8yDM2liYA2tKF/HWk+exlyOOgbmTL1X5
5Ly832CdTcwMldXD8trckUn0wfAXJg7AnIf3hBbYd4TXO4M8/x0nHo/CEYKV
TCYsL3pkNMK2Z9+bNXA2103wETZvbd1igaf7xX6cSu7vff+MAH0KRxrawmBa
/epwLuy4mabaDPvGf9p/BZ6+uWqOivRHL4o2kHPTPCdjCNYLlV7fYTU9Y2og
+lFJxHe5uE9HSeeF3bBm17kmX9ierZbrYFn4kREveAp5Tyz5974IpX8DNP98
/A==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.411618926975063`, 13.531562146049868`}, {1, -1}], 
         LineBox[{{4.999999999992724, 8.000000000003638}, {
          11.999999999998181`, 4.000000000003638}}], 
         PolygonBox[{{7.979054114725325, 6.2976833630141}, {8.822490309931942,
           5.355019380136117}, {9.21940146061741, 6.049613893835684}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.335784645546907, 5.259023129707088}, \
{1, 1}], LineBox[{{4.999999999992724, 7.999999999992724}, {12., 
          11.999999999996362`}}], 
         PolygonBox[{{9.020945885274676, 10.2976833630141}, {
          8.177509690068058, 9.355019380136117}, {7.780598539382591, 
          10.049613893835684`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.335784645546907, 10.740976870292911`}, {1, -1}], 
         LineBox[{{12., 3.9999999999902114`}, {12., 11.999999999982228`}}], 
         PolygonBox[{{12., 7.4}, {12.4, 8.6}, {11.6, 8.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.9452, 8.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{5., 8.}], PointBox[{15., 13.5}], 
          PointBox[{12., 4.}], PointBox[{12., 12.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["7",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcfdg/fhegehgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcfdg/fhegehgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt0wss1VEcB/Af6V6Pcr3arPW4uHm0yJU8cu2cUV5rY4bWYzczrx6bx50w
JSSPycZQirge5TF5N1qq22SubGIYSR5x1/1TiDuJRr9/dbb/fvvsnPP7/s/5
728SEuUfpg4ADviwFeZ3cBhS+Dv4FGj62/YedGTOtLM7Wq5vw6Sj87Sc/JPR
95Z5jk5op7Z3WfXod+V2rpMGFBJchqu60fPLLnbRaNuR0Ro5+pJokrOuTyH6
qKF6F1o0lqkfi5YunTxchVY5c/wX9CjwGentu+jSgKHrYeiZgoyYa+jaxB2j
YR6FlWTD2kvs+0yEF6/rUtA7pPwmRsuUixqn0Hm3agRxaD+JT37pXgqD/R5D
ZWilNeeYMVp6ItZiHB0t5dyp3ENhIyKsbb8J7stRddugjVddY8ToQV+3viYd
PH+8vVUJWml3LnA/WiqsHulDFxcvbYZrU/AK8P6gQMOE/HKBFuYCI/3OGpjK
Sk3MmbJynUM3a2iMlnIpWLb0NvajvQZMHuVyMMc0Q9TI9n8xfzxtN4XODp51
AZsv8rHJ1KCQFRRxOY1dfyE99OkurKk/VKlsflmvjUId73vzoGcR218RkuuO
ptPzgV1svnadoEsN+36R966hbYdbjpxB+5mKk06a4rmuSDaGAO93rTI/Ea3p
phAHojV1fnZ3oG3HqiMzdwgE69T2K9GwFTVgtU3A9r1uGdcMv0NKfVPObwLN
9x94GqDzmE9vWjYJdLZlU22038Xwj/UbBJSzR61VuD9Y6J2Qu05gpWSf0Rha
Wu3PSFQEZDe3jr1E89vTh+NX0fYVSbVo2VTR52cr2K+vXvKEdUno+dNLBOic
78xz1sLsuoZFAnqPs8kE28+hySKNISC90drAw3xwlMjLvxLgzziJ/dAz215u
IwoCKR68ikJ23vzV1YU5ApZN1hYjaL6dh6h1FvN4wdFcAYUUh7O/9KdwXs1c
zwoNPTnur8cIjJsFPXRAy3QWmbhBAnImkytk13fJuVMyAgnC4MIDgv//UfGr
f1VA/wB3f0Mb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {3.075655548348861, 9.834338870872152}, \
{0, -1}], 
         LineBox[{{19.999999999996362`, 17.}, {14.999999999992724`, 
          13.499999999998181`}}], 
         PolygonBox[{{17.991539152311425`, 15.594077406617997`}, {
          16.779075909943245`, 15.233615361589619`}, {17.237845785433905`, 
          14.578229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.21593336623684, 15.91500947680451}, \
{1, -1}], LineBox[{{20., 10.}, {15., 13.5}}], 
         PolygonBox[{{17.008460847688575`, 12.094077406617997`}, {
          18.220924090056755`, 11.733615361589619`}, {17.762154214566095`, 
          11.078229825174388`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.784066633763157`, 12.415009476804514`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1gs01fkWB/C/Z4qDo6TyOjKTR0RDI1T+ck6r3KxUborUKZOkIuU5MfOv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          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.72092184717428, 1.8721747773942476`}, {0, 1}], LineBox[CompressedData["
1:eJw91AtIU1EcBvDrjLY0dVo6e2imlctHM5KQXHbViTNKnVhKz1laI3oYWmpJ
XXM+yoQR1pYtHSXZw2yIOSmVykfLoCfS1HQrKzeqsWjUrKy+U9FgfPx2Lv97
9nHPnb9tX1oOi6IoEb4k/3749P/k1l/MmYss7HdKTkBGdo9eFyEzXcNa5Ejp
zwB9AVmPSLlwG6ldFh/UimR+XXWMkzlZwXe+Ik1x+W6cxTRV3J2XRCMND0r6
ZyNFP5MFVUh90/L2eUht/PPeAaSNz3XmIR1xj7L9QmgqaqKmyAlmYtMrtsLm
tq5jJsy9zLphOQtHpKW4tcHpA6KGfph+bSgk+9OkRMXZYK7z8aJkuGZV67Np
oZgXwnb4wPYuH7sP3D4xZjUFY87iAgMPVu2MTLwBM0pJvBvMXzt5Uw7vfpLH
ODCPCnJRb4f1Vcr0YVg62GRZC8uGpWYd/GRpGzsBrmTdalbAss+6viRYO9GR
KINzGzLWbYJ/9JdmxcK+ygUxxWTdLFH5wzaby57LcEOZIoNFrv8VKDTCb9Lc
O63oIfOpqMIf+xdmGo6+Jf11FvRuhw2cF8PjcKXFI7gJ1i8LLP0CUytn6uyw
ovbWaU/Maw+WHlmB39dYXTVRsGbIeOkQ6ZtO0uwi/SWKzrfAmm3ZXpfgO+I6
gYms2zxZFtK/x4FGJ3Ldc+fApehHqpg/6g0H3EsYPwwreopXzyXrqQcb+0JJ
DzFssi4ey293D8O8ms6nFFw86KqQwAZJiNMrzM/f4f/6JGzqqT2ogwOqP9/u
ggv36srLYbXPhnNvYfN99mgq/PLdKYYKpylObmjWLFhBh39zh02SIpcx/N8I
RhziCWt3nu/TkuczerKHDVOKQUkZvMBKceyYx48u5/3pTzl+7QUsM6Z7r4E/
7Gh53ApH7b3SHUtM1RmryX5LgqaQ8yDM2liYA2tKF/HWk+exlyOOgbmTL1X5
5Ly832CdTcwMldXD8trckUn0wfAXJg7AnIf3hBbYd4TXO4M8/x0nHo/CEYKV
TCYsL3pkNMK2Z9+bNXA2103wETZvbd1igaf7xX6cSu7vff+MAH0KRxrawmBa
/epwLuy4mabaDPvGf9p/BZ6+uWqOivRHL4o2kHPTPCdjCNYLlV7fYTU9Y2og
+lFJxHe5uE9HSeeF3bBm17kmX9ierZbrYFn4kREveAp5Tyz5974IpX8DNP98
/A==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.411618926975063`, 13.531562146049868`}, {1, -1}], 
         LineBox[{{4.999999999992724, 8.000000000003638}, {
          11.999999999998181`, 4.000000000003638}}], 
         PolygonBox[{{9.020945885274676, 5.7023166369859}, {7.780598539382591,
           5.950386106164316}, {8.177509690068058, 6.644980619863883}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.335784645546907, 5.259023129707088}, \
{1, 1}], LineBox[{{4.999999999992724, 7.999999999992724}, {12., 
          11.999999999996362`}}], 
         PolygonBox[{{7.979054114725325, 9.7023166369859}, {9.21940146061741, 
          9.950386106164316}, {8.822490309931942, 10.644980619863883`}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.335784645546907, 10.740976870292911`}, {1, -1}], 
         LineBox[{{12., 3.9999999999902114`}, {12., 11.999999999982228`}}], 
         PolygonBox[{{12., 8.6}, {12.4, 7.4}, {11.6, 7.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.9452, 8.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{5., 8.}], PointBox[{15., 13.5}], 
          PointBox[{12., 4.}], PointBox[{12., 12.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["8",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcfdg/fhegehgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcfdg/fhegehgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGBQAWIQHaEDJBh4HBjAQMXhRa5yk54xtwODQ2Jyh7mKw3mv
kNK4NC4HBgXNJwIpKg5ntv4MnTCT04Gh4KDX8RoVB4asgzpHznA4MCx4EH8M
yL+cUzPj5392oP42ZT6g+g239d/pGwP5Gr/PNAPNy3nWnJaWxubAYHF18s//
yg76QY6+c2ayOjDsKM3OP6HsYKRyS/3iGRagfjv3p7OVHWycbp5iYQDyDY7G
6TYrOzybVhllbszswMAR8lIHyGd7+so8M40JyH/97DFQ/ccOu4ezZjIC+Y4K
uUDz2vJunzlzhsGB4cT+2G9A+4rfy0759/+/PcMG1jX1QPfEM366k5L2z56B
YaM4J9C9ixMcNM+c+WPPEKGdeQDonybtA9eMjX/D+ce6+JRnz/wJV795zqNs
RoYfcPOk1nsC/ffNHmafS1Wz7akzX+xh7pE+pDVB1/izPcy90XG83hNnfrSH
+SeOf86VN//f28P829I+I2H3mTf2sPDoP//7VN/Ml/aw8JpT9zE1Le2ZPSw8
e1Y6ujsYP7aHhTdDKv88SYYH9rD4mPHg28x7M2/aw+LLpMkkoNb4sj0sPi22
7F8lMeukPSy+E3rX7daftdMelh5AqaRhV6ctLL0AAP+Y434=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.75, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000009095`, 17.000000000007276`}, {
          15.500000000009095`, 14.500000000007276`}}], 
         PolygonBox[{{18.274494365672922`, 16.04138575870718}, {
          17.031248461855625`, 15.808277151741436`}, {17.41976280679853, 
          15.108951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.602053456673122`, 16.50014377798838}, {1, -1}], 
         LineBox[{{20.000000000009095`, 10.000000000003638`}, {
          10.500000000009095`, 6.000000000003638}}], 
         PolygonBox[{{14.697018774917161`, 7.76716579996512}, {
          15.647758425059585`, 8.601488350090106}, {15.958204025106092`, 
          7.8641800499796535`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.580390455651742`, 5.944902635795507}, {-1, 1}], LineBox[CompressedData["
1:eJwl1wk01GsfB/DJErJmqZFtkmRpxjaVJI2deKMiLYosJamsJSTJFirKelMh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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.628160615326795`, 12.09148827813548}, {-1, 0}], 
         LineBox[{{15.500000000004547`, 14.5}, {10.500000000005002`, 14.}}], 
         PolygonBox[{{13.597022314125994`, 14.309702231412599`}, {
          12.363176198265608`, 14.588312644671397`}, {12.442779173482405`, 
          13.792282892503405`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.845769235517452`, 15.187507644825484`}, {0, -1}], 
         LineBox[{{3.499999999992724, 10.000000000003638`}, {
          10.499999999998181`, 6.000000000003638}}], 
         PolygonBox[{{6.479054114725325, 8.2976833630141}, {7.322490309931942,
           7.355019380136117}, {7.719401460617409, 8.049613893835684}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.835784645546908, 7.259023129707088}, \
{1, 1}], LineBox[{{3.499999999992724, 9.999999999992724}, {10.5, 
          13.999999999996362`}}], 
         PolygonBox[{{7.520945885274675, 12.2976833630141}, {
          6.677509690068058, 11.355019380136117`}, {6.280598539382591, 
          12.049613893835684`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.835784645546908, 12.740976870292911`}, {1, -1}], LineBox[CompressedData["
1:eJxFlg001Fkfx6esOFqlsJhmxvz9///Si9hn1aLi11aIiFDIegtDjEEGIS/l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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.4452, 10.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{3.5, 10.}], PointBox[{15.5, 14.5}], 
          PointBox[{10.5, 6.}], PointBox[{10.5, 14.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["9",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdf/fhegehgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdf/fhegehgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl03sslWEcB/AfXQ6HnBOyEyVJbnOKSp0We59NlNWiFmM1jpqxVKQL7aRk
XXT5Q3NSpnQkc7KUSDgu3TlSoVSicrptx7pM7mLp+6xne/fss9/7fH/P793e
+dsSN8WaEtFyPHwn4ySWDaNRvrsx6qt5IDkIp/oo1XPcGdmGxBSMWjO633tS
5ANnBjQ5HIPbFoTnLodd/BIMjrDUxFXqAWetTKlunsmoTKJMEsGyLdryDNjp
ltytA/mjPXdTAuGkIEmBmvvi7+mWvG5QRIXCU2eLa1uljDQHym9I4EvvXuw4
DUsvnxK6XBkpd77atwJWfAkxVMKhUt+Z1yXo11yytwjWFPnbPbZiVL2sQKWF
I0rDLRpn4P6Hxr1qYYrrDqixZGTMm2X7lvvl+uwSC0adnzzlg7BRbO1ULEZe
xSkmRn/FPU93nTkjbc6FNdawxkQv+26GuZ5a5Ytgmbp0nR8cYaFtMeK8tt6+
oETEyMxi5FEVzHLOihbBhoGr5w7C6V4bunTTGa11jlL5wmVD4d0BsPGzvmlg
Ib7H7MvJddPQt+drWB0si1G028NJf+auOQ/HF6nyIqfi/NHfg0dhw8PK9tQp
yA2Pi86A7wc6tKWZYv4qiTybv+/zMTbZhJHe+YTfTZ7/Z9xzO+F732662QIr
J+U63V+B4j96+H/hXrebvZsQKJR6NT9heu4sXjIukPfm4NZvcFbuxrCWMYE0
Ptc6muGyaNUuzahAypRlHnmwt9Eu9eqIQNJ+/z1RcHpD/5huWKC+uNgKe+7C
geLWIYGylia7dbrgvrKED42DArW1BLXnwwa5+/CZAYEy04pd9vN68JJW636B
zByNP6JgqaXJm8g+5B220W7l9cj9uR0/BOrMPrEnAXZa/CKx1CiQrFdz5DjP
W/1gb+JX9PvlO6+Q28GxOcSAeT1tTOtgNsneT3QK5N7xulgPp0+89vJ7ifqq
V4WPYDI3dRzTo17xOLOE+5jqzpMG5JkrNh3mfuZ/ZWEp5j/TszKAmy91/f//
aQH7B6xhRyU=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.075655548348861, 10.165661129127848`}, {0, 1}], 
         LineBox[{{19.999999999999545`, 16.999999999996362`}, {
          12.000000000001364`, 15.999999999996362`}}], 
         PolygonBox[{{16.5953667260282, 16.574420840753525`}, {
          15.355019380136117`, 16.822490309931943`}, {15.454247167807484`, 
          16.028668008561006`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.807746161386726`, 17.433230708906184`}, {0, -1}], 
         LineBox[{{19.999999999996362`, 10.}, {14.999999999992724`, 
          6.499999999998181}}], 
         PolygonBox[{{17.008460847688575`, 7.905922593382003}, {
          17.762154214566095`, 8.921770174825612}, {18.220924090056755`, 
          8.266384638410381}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.78406663376316, 7.584990523195488}, \
{-1, 1}], LineBox[CompressedData["
1:eJwt1Qs0lGkYB/DPZUvjstO4JClsk4owzYzqVPKlVlLNmSlsxRZmMiJF26Ko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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.81450972518251, 3.511528178832159}, \
{1, 1}], LineBox[{{15.000000000001819`, 6.5}, {12.000000000001819`, 
          7.999999999996362}}], 
         PolygonBox[{{12.963343685400051`, 7.518328157299975}, {
          14.215541752799933`, 7.339442719099991}, {13.857770876399965`, 
          6.6239009663000585`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.411618926975066`, 6.46843785395013}, \
{1, 1}], LineBox[{{4.999999999992724, 11.999999999992724`}, {12., 
          15.999999999996362`}}], 
         PolygonBox[{{9.020945885274676, 14.2976833630141}, {
          8.177509690068058, 13.355019380136117`}, {7.780598539382591, 
          14.049613893835684`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.335784645546907, 14.740976870292911`}, {1, -1}], 
         LineBox[{{4.999999999992724, 12.000000000003638`}, {
          11.999999999998181`, 8.000000000003638}}], 
         PolygonBox[{{7.979054114725325, 10.2976833630141}, {
          8.822490309931942, 9.355019380136117}, {9.21940146061741, 
          10.049613893835684`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.335784645546907, 9.259023129707089}, \
{1, 1}], LineBox[CompressedData["
1:eJxFlgs41OkexydySErXJ6RSNv/L+BtsFzTVTzgJ5Z42G7mFmNxOua47acql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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.645199999999999`, 12.}, {-1, 0}], 
         {PointSize[0.04], PointBox[{5., 12.}], PointBox[{12., 16.}], 
          PointBox[{15., 6.5}], PointBox[{12., 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["10",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdg/ghefehfh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdg/ghefehfh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQA2IQ7fXmPxAIOjCAgYpDTs8xg+kzgfwJp+8f1VZx2LD3
3kM9YyBf4hq3S5CKQ52BncvxMwIODDMWnrqap+JgFKidl5AG5K/IPcdRp+Lw
bdfJLV/+8zswWBQuYAfyzdR5rmkaA/kPVjtcAqrf0yxyPz6Nz4FB4ZmqHdC8
b3vORU6byQu0/8HcvUD7upS5z545w+PAECCe9PW/ssMRoeazTAxAvoCoue4V
ZYfrvPO5LYy5HRgqkh6uWavs8GexVUxOGpcDw4ZaKcmJyg4lNoW758/kdGDQ
UDd40qjssLU33fPSGQ4Hhh/qMx8D+d9WPXNkYQDyPRpCxYHq/cLCP5gaszsw
JKS/Wwk0j0NedVF6GhtQvYyYJtA+/YYdM2fOZHVg+KAk/AHoHuVMfpYzZ1iA
7vmgsB3o3qlnyj79+88MdP/vKjOgf+yafSKNjIH8CUfTTwL9e8H2rHBaGhPQ
vv6Zf2tVHJhyjN7MnMkI509V27zs7BkGuHrbux28jAwMcPPiGh98T0/7Zw+z
zy2478yZM3/sYe4R9du228D4tz3MvRzLA+SnzPxpD/PPjr+GNd//f7eH+dfS
wiA3Ou2bPSw8dt+S6Txw5os9LLzWcne5qxt/toeF51fzgy/6Z360h4W3Rcyn
k1/+v7eHxYeC5KXTZ868sYfFV9HmwzOWzXxpD4vPmBlrmhvSntnD4juP69zy
COPH9rD0ULT7uo4uwwN7WHopsPaQ/Tnzpj0sPW1xk9rTbXzZHpbeFJ4rHVGe
ddIelh4T5IPOGc/aaQ9Lr+BknDh9Lyw9AwDd+iN6
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.25, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000012733`, 17.}, {10.000000000014552`, 15.5}}], 
         PolygonBox[{{15.593361811720978`, 16.339004271758146`}, {
          14.347302007106924`, 16.556570269389173`}, {14.46597436945112, 
          15.765421187094535`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.770072297958121`, 17.17805134694586}, {0, -1}], 
         LineBox[{{19.99999999999818, 10.}, {9.999999999996362, 4.5}}], 
         PolygonBox[{{14.474270454794011`, 6.960848750136706}, {
          15.332962045297126`, 7.889637613333953}, {15.718497045114852`, 
          7.188664886392635}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.706414924119443`, 6.025685606932476}, {-1, 1}], LineBox[CompressedData["
1:eJwt1gs0lOsaB/CRSVPuuWQzbkUcZCasSKVpQje1pWErl0O2HVKNojOJs0cq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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.160447344128674`, 5.134637948213067}, {-1, -1}], 
         LineBox[{{4.499999999994543, 9.999999999992724}, {9.999999999998181, 
          15.499999999996362`}}], 
         PolygonBox[{{7.674264068711929, 13.174264068711928`}, {
          7.108578643762691, 12.042893218813454`}, {6.542893218813452, 
          12.60857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.758784489160852, 13.241215510839147`}, {1, -1}], 
         LineBox[{{4.499999999992724, 10.000000000005457`}, {
          9.999999999990905, 4.500000000007276}}], 
         PolygonBox[{{6.825735931288071, 7.674264068711929}, {
          7.391421356237309, 6.542893218813452}, {7.957106781186548, 
          7.108578643762691}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], {6.758784489160852, 6.758784489160852}, \
{1, 1}], LineBox[CompressedData["
1:eJw1lg1QzHkcxrduZcuimXKTae+E1GIj4W6d7tqj06plmpMTtwhxdeXsEdZh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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.567205028802515`, 13.066824174393599`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1gs0lHkfB/BBtZTL5DoihlwmEW+kaakm5ZLQJOXSbDu5RbGN2+5sIatJ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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.763809742330267`, 8.23619025766973}, \
{1, -1}], 
         {PointSize[0.04], PointBox[{4.5, 10.}], PointBox[{10., 15.5}], 
          PointBox[{10., 4.5}], PointBox[{15.5, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["11",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdh/efegfhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdh/efegfhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9k21IU1Ecxo9v+ZLpysg0leIe0gLLvCEa5vl/iBI1EF+QKajJvLpKViQV
ldqLWZmWErruUnCmETosR5JiSh+SFO5KzQr7kCPFaS60cjlLt47C2YHL5cfh
nvv8n+c5u/JUKfnOCCGePmtv9ZKdLj9A6wtDifU01ItrfC2tNArDTq8ninCe
suygR2MeBkNMzsiAtAXQxYZZy2UMDfcCW3IEylbuflcJhkPcqP6vfTMgo6W0
ogDDUplJrBcphwUXaw9jaCqtHIvkKQ9vfWp3xzCv/zH+TpIB8piI/zTMgTHJ
Gn1SoBxfI2tt5iC9bkLniihHH9NFVXCQMTN6PJb3BdS9krp4lYPcgriHRYIP
5ZGQrw840CifDTeKm6je3Ucyezkov6R8YZC86flva6bmOeCzx5dX7RsBNRW2
XQnFoK/8vGc/T/n5nLI6A0NOT0/FCcGL6jt3e+E81dv8yq1O9KT7oRd0ZRhS
ao2ZQ5IHoOT0cv8zGJbjAvxsdndAYPoWlojBIq/viuQp5/aGzvhjKMzWTQrC
BupnS/82IwfT7dUqjehG/emsTOvgILz9scIguQIa9NkXWMXB7NhaHi50npCb
c2UctNqyHkXyLo598+rZmnzB2fF9myk1SBSdHOcHm5UJBgk5/h/gbjI7IeTQ
53wgLqFIsBGmX+FTsfhRWiFsvmnj3V+x/D/C5oe4ZJ1WXCbMH6JO8nZFVsL8
C1KHqxXCH8L8XejWZr+RFgnzX9F34z3mfxOWj+W6XHFL/ElYfg0rTuIX+zxh
+cZDbHSHZCYs/1Hf2og74ixh/UjQt2UphWnC+lO1fSAikZ8krF/FNpVyLzIS
1r/Bo/J+N804Yf183en5vZn/QFh/ZfKYU6maIcL6jXa8nFJpegjr//q1yND2
sfvxH189TgY=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.5, 9.0548}, {0, 1}], 
         LineBox[{{19.99999999999727, 16.999999999996362`}, {
          13.999999999995453`, 14.999999999996362`}}], 
         PolygonBox[{{17.569209978830308`, 16.1897366596101}, {
          16.304298914762956`, 16.1897366596101}, {16.557281127576427`, 
          15.430790021169692`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.5098469626739, 16.865659111978296`}, \
{0, -1}], LineBox[{{20.000000000000668`, 10.}, {13.999999999999329`, 10.}}], 
         PolygonBox[{{16.4, 10.}, {17.6, 10.4}, {17.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {18.0000434305083, 9.355145491133467}, \
{0, 1}], LineBox[CompressedData["
1:eJwt1QtUjGkYB/BXmDQxhVxOF9vuuiWFwtAmX9NUWnLNZTIyExLRhTYpmko3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          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.288487526462113`, 2.470262579386344}, {0, 1}], 
         LineBox[{{4.999999999985448, 9.999999999992724}, {
          13.999999999989086`, 14.999999999992724`}}], 
         PolygonBox[{{10.024494365672922`, 12.791385758707179`}, {
          9.16976280679853, 11.858951330844206`}, {8.781248461855625, 
          12.558277151741436`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.352053456673122, 13.250143777988384`}, {1, -1}], 
         LineBox[{{4.999999999987267, 10.000000000007276`}, {
          13.999999999990905`, 5.000000000003638}}], 
         PolygonBox[{{8.975505634327078, 7.791385758707179}, {
          9.83023719320147, 6.858951330844206}, {10.218751538144375`, 
          7.558277151741437}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.352053456673122, 6.7498562220116165`}, {1, 1}], LineBox[CompressedData["
1:eJxFlQtQVFUYx288BScegsMKzPIIYXfPvSuBCsJqHzUjlCssIKOEQCHIMoAC
hmALgstDStcHEqzBDq4WGqCSqEAprfEMMCMkMAQZWwV5lBFri2h07tQc78yd
M7+5891zvsf/f1zi9oYlGFAUlYBfdv3voeGGHbu6AdWt8Er/CcH0bOTafl/M
8OhEZS2CqwcnqZUHMHN27eeeQNDhFU3pWjGnFVxylCPwEu17M3rZaqAutCyu
OYTgmvR594pwzEqDvDgFguJlgZf6qjDra4OENQh8LR6UiX/DPGxpV9CDwFaV
UZ3l5g7UuI/Rz3oEdTpxbmoMy3Nq7hoatrSvN1Idx0wVdNsm0zD+uSx+8Drm
kk0811oagiNFfuEDLIcWlT6hoSqqfaZbi9nzZOJFdwYm1G9TcdOYs1P0AzEM
tK6/u+7WBMuOXarjDIRY/HJs4Vc2Psrnr+sM6A4V5ss7MU9+lu02wIBypuae
Rz1m3yqzFw8ZmAl0Xv5MgVly82XvIwYcnNrq3tjjzuafnjnCwKlhU+FCMPu/
zrGkdga2frLQ3+SFedh+Y42aAc+L5o/zHTDrP0QXPmKgy9R196blmDnmpdFv
MVD3VcCCwhCzcirT0ogBmWbz2svG7H7lD4M6aNhlL9dmrsCcplz6poiG2Po7
ehCw+S7KnYJo2MqB1Gdi9jw5HfaWNMw1Ze02+hgzPNVwRxEoe5bSEhvY+DAf
yVUEzyI2n7n2OxufzN9bjiDAYJ+N0NsD11tR3VaIIHaqY12fDLP0aIVlPoIF
gd1EdCdm5yNXiksQzGa+UFZa8YDaYfJPrwpBU0HJ9oAdmE8cfKprRWAdllzr
UYm5YTJVOYXgnXCdKGoI8/Bgwl0uDVnRz4O8lvOBsmqOTYmkQZIfeC9mHWbP
YpvRChoaPNY4vNyGWVJKHRii4ZSFo6sqCXND6UDRSgbs/DdeCdyHeZll5Psh
DAxdNB/LzWDj9zTK5Azc6xPJvKWYNcZlzfUMbLiSVN8egXkyRRzQx0CYNH7u
sQjzmdcavn3AgNuf0qRyJ8ycounQxwwEc1l9vPqetaWiPGmcR+JvaSpfX9XG
I/8f2V8e7l/LI/v77Wwr7KvgkfOJZDoXuYJHzv9y9vutzUd5JD8zoXY0rpxH
8p+27dca1PFIfcQ3mt3VPa/qV9Vr3HR7/lV9+wfN277g8Un9DYfKIobj+aQ/
O7/MutB6nk/659FomBj/B5/095z1E87ZDQLSf1GYX493oYDMR5SfYqa/T0Dm
hzvQEvueNSLzlRNDHVeGIDJ/ruklpxeLEJnPLitNqVkjIvN7LcfT9LtBRObb
1/Wgw6eziMy/c05Lr+Q5IvqQ5rWaZLxARD9il7HTt+YR0ZdWpluVp0VEf53V
JsKg24joM1+i7zf5GhH96s02brt/EhF9X55/MG+WgYj+DSWR7sPhiPiD03lZ
Xu4GRPyj3V6QbrIaEX+JmEM9mziI+A9z+OaSyhYRf8q+7W/j44CIf3Xeaeuo
pxHxtxvqwkrBu4j4X2j2D4txaYj4o9S7UV6qRsQ/l44oPigZQcRf79u7JDc6
0sR/jzm3hHrvook/x2qMfWbraeLfD2eO/nj2b5r4+5DZuSXzAIb4/8RNu6Wp
YobcDzz/ml5hF+b/74/t1uwqhH8Bc7d2WA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.645199999999999`, 12.5}, {-1, 0}], 
         LineBox[{{14., 10.000000000007885`}, {14., 5.000000000010794}}], 
         PolygonBox[{{14., 6.9}, {13.6, 8.1}, {14.4, 8.1}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14.9452, 7.5}, {-1, 0}], 
         {PointSize[0.04], PointBox[{5., 10.}], PointBox[{14., 15.}], 
          PointBox[{14., 10.}], PointBox[{14., 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["12",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdh/efehfggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdh/efehfggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 0}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {57., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {61.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {66., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {70.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["g",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {75., 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGDQA2IQ7fXmPxAIOjCAgYpDTs8xg+kzgfwJp+8f1VZx2LD3
3kM9YyBf4hq3S5CKQ52BncvxMwIODDMWnrqap+JgFKidl5AG5K/IPcdRp+Lw
bdfJLV/+8zswWBQuYAfyzdR5rmkaA/kPVjtcAqrf0yxyPz6Nz4FB4ZmqHdC8
b3vORU6byQu0/8HcvUD7upS5z545w+PAECCe9PW/ssMRoeazTAxAvoCoue4V
ZYfrvPO5LYy5HRgqkh6uWavs8GexVUxOGpcDw4ZaKcmJyg4lNoW758/kdGDQ
UDd40qjssLU33fPSGQ4Hhh/qMx8D+d9WPXNkYQDyPRpCxYHq/cLCP5gaszsw
JKS/Wwk0j0NedVF6GhtQvYyYJtA+/YYdM2fOZHVg+KAk/AHoHuVMfpYzZ1iA
7vmgsB3o3qlnyj79+88MdP/vKjOgf+yafSKNjIH8CUfTTwL9e8H2rHBaGhPQ
vv6Zf2tVHJhyjN7MnMkI509V27zs7BkGuHrbux28jAwMcPPiGh98T0/7Zw+z
zy2478yZM3/sYe4R9du228D4tz3MvRzLA+SnzPxpD/PPjr+GNd//f7eH+dfS
wiA3Ou2bPSw8dt+S6Txw5os9LLzWcne5qxt/toeF51fzgy/6Z360h4W3Rcyn
k1/+v7eHxYeC5KXTZ868sYfFV9HmwzOWzXxpD4vPmBlrmhvSntnD4juP69zy
COPH9rD0ULT7uo4uwwN7WHopsPaQ/Tnzpj0sPW1xk9rTbXzZHpbeFJ4rHVGe
ddIelh4T5IPOGc/aaQ9Lr+BknDh9Lyw9AwDd+iN6
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {2.25, 10.9452}, {0, -1}], 
         LineBox[{{20.000000000007276`, 17.000000000003638`}, {
          15.500000000007276`, 10.000000000001819`}}], 
         PolygonBox[{{18.074454554788097`, 14.004707085225933`}, {
          17.089074055061282`, 13.211595951299467`}, {17.762016835362523`, 
          12.778989878248668`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.160082313252435`, 14.866097218463327`}, {-1, 1}], 
         LineBox[{{20., 10.000000000003638`}, {10., 15.500000000003638`}}], 
         PolygonBox[{{14.474270454794011`, 13.039151249863293`}, {
          15.718497045114852`, 12.811335113607365`}, {15.332962045297126`, 
          12.110362386666049`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.705759138948043`, 13.973984608400851`}, {-1, -1}], LineBox[CompressedData["
1:eJwt1ws0VfkXB/DjlWceYTVzUdeQiPKoqUR1VCqVEnrII3MN1WRQQ4lJx2Mi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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.666233980906949`, 2.1296932060463307`}, {0, 1}], 
         LineBox[{{4.499999999994543, 9.999999999992724}, {9.999999999998181, 
          15.499999999996362`}}], 
         PolygonBox[{{6.825735931288071, 12.325735931288072`}, {
          7.957106781186548, 12.89142135623731}, {7.391421356237309, 
          13.457106781186546`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{6.758784489160852, 13.241215510839147`}, {1, -1}], 
         LineBox[{{4.499999999992724, 10.000000000005457`}, {
          9.999999999990905, 4.500000000007276}}], 
         PolygonBox[{{7.674264068711929, 6.825735931288071}, {
          6.542893218813452, 7.391421356237309}, {7.108578643762691, 
          7.957106781186548}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {6.758784489160852, 6.758784489160852}, \
{1, 1}], LineBox[CompressedData["
1:eJxNlg1Q03UYx//IspmoE8XIV6SBC8dL5yQKtD+Ll4GoNDHAW94EMiw4Uae3
ahDnLUOFJFqIRjrzpaXcKQo6a+bKpeSBEWGgoVHb2dLdOWXiMKS+/+P53cUd
99yHsd/L83ue7/eZW7Be/cYYjuMa8SvE0R8pPxpjeU4c0xi4FOxpyyz1x/Bc
keOyogFckr67ywlOulpw4yY462Fo4ffg9sf7DLMjeK7WPv9vE1j7cK8rByw5
EHhOAz4xPSCnEtz+U3fLDHDLv9Yd+8HaxNhJ3dE8V5V6sLcZbD5WnVcFtllb
Is6AbYUZSxPBOcmTB46DS5avO3tXznNBB59Z+TlYV3841wLOWnbylrD+CXta
85vghrL80xpwWUAaFwfuK7kTvQCsTw8JCQR3frBgcKxw3jpl8p/zsa53d891
3MfccS/9CtijV+9pBrc1O3MugV2dZ/prwIfmbHrUDja0TwzeCA6tMMffBHdn
rlVowYqv71YMgd0xrdfywfZlNV0zsZ8o9fUg4XP+luvnVLCx5rZyE9g/3XR9
A9gs2b/qE7CsvCxyH5gvHb/nvLCfbYHnEtix+UDLoPA+6UsrPWB7WJ8zHuef
2v6abhzy07k/8345OCjg7KmZYOOsD1f9KOTbH9MoBfd98VfhtEjkcWjs5DDw
8KGdygKw+8iiqonghgtZCUfB0q5eiRfruw5PbL0D9vXGq38Q9g+/p5PO4znV
0VXGXWBLC/eZGmzkd63JBhe/E3VJBw4aGh/8FLi7oyxyB7j4zvbg80L+NAcM
deC8gMFpG8FFM6bP+wjsXrnndgQ4qd4QWgHujI5dfiMK71a4taMQ3JZf0rcX
3GTakpgMVmzc8ocWXDsYfjZUWK/oZMbzYLky2enBeR2vRm2bAO6XtZbbwQn7
TIbB55An55C6Huz95oLBA3Y8UE7ZAK699ijpLljH/6JYIXy/Un3jMdg7MvLr
YnCK/In707Ce3RX91UKw5liUKQFsWRnYkwC2bJ2lF85jDa9/pAIP92wVV4Pz
bueHrAVb+08brWDTtgm1u8BNF1+Y6wRnlx5PdoBlqz+WjcP9D2VHlo8R7uNL
nSEDh254LFWB7YuTOhaBue1ZDiF/tm2576vAviVHjv0u5JsP8qWBLQNWiVyG
+/Aj5QnCemXy2M3guCnDT84CGyclrrMK/OxQyAPsb1tvXDsAFkv/+fY7sO6t
47FS3Fu/RGk2gosnfZqaATZee+Xyy2DPZufJNWB/Vq7lAWJnfvHTpYh5EQtd
FiHmVjS9LeRt+e40DaKnsVi5GtE+pbQ+GFG7whafhmj1SfVXhHMMqIoihM9/
q1lWBzaWLpowgjjVc262FlE0uNPehdjQmqF/EdEbdzHxCGJWrbNgDqIs4Grm
e4j+lxLVwYi96eIvcxBVKU1hAtvfHdOjQNSdGpLM/l9kf0+h/+uj77F12Loq
2sdH+5roHGI6FztnFZ2b3UNB97LRPYvo3jbKg5fywvKkobyxPLK8DlOe3ZT3
SnqHEnoX9k56ejcRvaOd3lVO71xF7y6nOmB1UU110kR1I6Y68lNdsToTUd2x
OpxJdcnq1Ex1K6U6zqG6tlCdN1Ddt1AfaKgv/NQnrG9YH7G+Yn3moL5jfeij
vrRTn+qpb03Ux6yvWZ+zvmc60Ee6wHQijnSD6YiJdCWOdOYE6Y6DdEhHusR0
iifdcpGOMV1jOucm3WM62Eu6KCGdZLpZSTrKdDWFdLaNdNdLOuwmXQ4jnTaR
brtIx7nDo7qeTTrPdL+NfKCSfEFMPtFNvsF8xEa+ImI+Q77DfIj5EvOpRvIt
H/lYFflaEvkcR75nIx/0kC8yn2S+OZV8VE++aiafdZPvOsiH3eTLjeTTzLeZ
j7eTrxvI5/vJ90toDmBzAZsTsmluKKM5QkJzBZsz8mjuYHOIneYSNqdU09yi
oTmGzTVtbM6huYfNQUU0F7E5ic1N2TRH8TRXsTlLS3OXm+awIJrLRDSniUbn
tv8AHLt2wg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.763809742330267`, 11.763809742330267`}, {1, 1}], 
         LineBox[{{15.500000000005457`, 10.000000000007276`}, {
          10.000000000007276`, 4.500000000009095}}], 
         PolygonBox[{{13.174264068711928`, 7.674264068711929}, {
          12.042893218813454`, 7.108578643762691}, {12.60857864376269, 
          6.542893218813452}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.258784489160853`, 7.741215510839148}, {1, -1}], 
         {PointSize[0.04], PointBox[{4.5, 10.}], PointBox[{15.5, 10.}], 
          PointBox[{10., 15.5}], PointBox[{10., 4.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["13",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdh/egehfgfh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdh/egehfgfh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1As0VHkcB/D/pBiFPDONx8p7VTNTuyG13ZsklbBy2mqtHFSTPMuraI1S
KO0SrdmlmhCyHK/a9MAQsi0WW7YHRZHHrlJ6KLbZ73/P3nPmzvnc3+//u/f/
+//vXeAX6rlzBiEkCD/6T7j0tID9DxIRSwyN7ApqTVjy91qXDOrXh1/dj4eV
bTJj42GXdt3BtfA+sWJ5LM1/7pWuDa+MDsjbB4cdvPOy7ROWxGVtVATAvKc3
uo7BzfYf+Z6w/TGdK26wdmWTmwM83sBXMoNHw25HGMPVqy51KYxZ0hI5+4JC
iPud8fPtg6dXCJ/XwbwG6ZE6OO2hmnkCPP7SebkUTjy/7KQzbK2XZL0L7pin
wdGh+QblepZwcsLpnGEB6nuO2nUbsYS7L+J4C0yC9oqj4D43/kAlzB2KtuTC
TnEGZQUwqyYP/c6QJeKHeu/Ow/c6l4SrwxKnEE4uHd8VyjtuwJIb/idP58Fh
oV+nq8Ir6808qdlLwpo0PvrZk6d8Fra+5lVhBpc3Vqikw5qcdRub5rNE+qE1
8iDcoS8fiIK5c5+FbIMneZNCezjZvc5CABepVzIq1BdKe94vRn2y6lofD9ej
nRpr4ZZvbJbKYY8mn/p42L4sJlsGD/x5xISFZR9sO+PgDnHGKiXYV1hq5AG3
jMqFrYvQ13bVK/NgUZB2yBnYN/WiWbs+S1LzO7Ji4Hvh2bGRsLi2zMkHlnpm
S9RhzSnjt5vg6g8qx6XzMN8TPubOcHLJkDMf9nDl7HaCY+74e2fpscTEtPjS
Wpg90DmqC5c0BlfRfE3pP7OkuizRNUjRovmkvElgASfeZ3odYNHmJw/rdHAf
t7Yua1rPy844EO6x3CpVo/VSzAPNYXOtX/oHFyKe+e2JMW2sZ7xrzGVYNh3c
1ACnPvvC8hA8WT3Ml8GLSmcnMXDYqXXVEph75USDwgb7x0Vt/w44R7ZneT21
Rq6nHcwbq1qYBN/b9lpnJiye6PbzgkUX2Z8atFgybNk9ZAPb34tZHQ675u5t
mwNPbqpQ1YRlS8/pvPsUdQ4YZck0WRKhaCz8C24xyFa2hHPIgGQIHv/ttO36
uehXYuZdao/i6GxHDfS1vXHnCBwWVZTmqI51MLzZMUzjVVY1nmroX0Th8yd0
fP4FfuwcXN/+yOwubHJgl2n9bIwfFxXUwGnvfVZawa1zrEqyYc3po/Elqtgf
L3oSQmGR4sWvrnCcnkWCAyxJtBVwYJFd89hHa+yP2w+K5VyWNJonX62DRaJS
XgLMdXW0lcDypafs7GH5sqmaNbDM4UFNjwr6sES4XZ3m+xi57Ya9ewP3P7bC
fJSEDn8oYz61KenXYUk/URjD7FXRZ7nUgoIg11lYf43L8lOwiY5XlO9MfCcq
b2efgD1WbynzU8J7Nrh9w3EaH0o87D2DJTOt9appvK920n4rhyVbh14lpcId
hY0jO/AWVfPZ4BRYnl/dO/mRIb5TE4fj6fjx1DfifxgykPOlNAhmqy62608z
hPfiaaA7rRcQfE55iiHlgu5bVjAxfUxEHxjSMsM2/a0l+nPo8qHk9wwJUM4s
qIElnT/XKyYZ4r0nuCoelt+9tf4wrLsmfpihXtkjnnrHkMy9noUc2MTrceBX
cEDkdesWC8Rvxp09+pYhatPfW/1A3T2hcfANQ+Tz/aXBsG9dulzwmiEu0/WR
7jQu1XDPf8WQnoENBSvgvuLg3v5xhogTXRUimGTwKofHGJLjVeS/mOaPdTk/
GGHI65DmKWpJnk6qfBDP09wpEcLsiL68sZ8hHp9HZgjo+Dvztfg9mC+zQM+a
WpDECb/LkI795w7Opw42UA37nSFcrz1zOdTe/ROKJoYYbm41fWQO801qp28w
pKhNyK+gfjszeX0F6jn+uCuWOmeLjpKMIVuZRB+G+v9jWoHDjP0XKXxuLA==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.9132523332652642`, 13.876454641114655`}, {1, -1}], 
         LineBox[{{20.000000000005798`, 16.999999999996362`}, {
          8.500000000004206, 16.499999999996362`}}], 
         PolygonBox[{{14.849433695013035`, 16.776062334565783`}, {
          13.633191415276443`, 17.12356012877624}, {13.667941194697487`, 
          16.324315202092194`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.182672302371722`, 17.69373704545034}, {0, -1}], 
         LineBox[{{20.000000000007276`, 10.000000000003638`}, {
          14.000000000003638`, 6.5}}], 
         PolygonBox[{{16.48173265946094, 7.947677384685548}, {
          17.316718930329426`, 8.897834175673825}, {17.719815750748698`, 
          8.206811054955079}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.176200089562332`, 7.515942703607428}, {-1, 1}], LineBox[CompressedData["
1:eJwd1gs01GkfB/D/WGQHhSh6baiNWpYJp8u+NP/Z7Fsum2vbtso7yKW0EhYr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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.471090192570806`, 3.411297472978525}, {1, 1}], 
         LineBox[{{8.499999999996362, 16.500000000007276`}, {
          9.999999999996362, 14.000000000009095`}}], 
         PolygonBox[{{8.941302546743485, 15.764495755427527`}, {
          9.215700282971497, 14.529705942401462`}, {9.901694623541534, 
          14.941302546743485`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.525685965145556, 15.057331579087336`}, {1, 1}], 
         LineBox[{{14., 6.500000000001819}, {12.5, 9.000000000001819}}], 
         PolygonBox[{{12.941302546743485`, 8.264495755427527}, {
          13.901694623541534`, 7.441302546743484}, {13.215700282971497`, 
          7.0297059424014625`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.525685965145556`, 7.557331579087334}, {1, 1}], 
         CircleBox[{11.8125, 11.78125}, 2.8649620961716056`, {
          2.255758378143494, 4.954722147037705}], 
         PolygonBox[{{8.981671842700026, 11.036656314599949`}, {
          9.160557280900008, 9.784458247200067}, {9.87609903369994, 
          10.142229123600035`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.468437853950132, 10.411618926975063`}, {1, 1}], LineBox[CompressedData["
1:eJwtlQs0ltkax0WTu1xChFze7/3eyxeRQUh/GjOVZhpCoRSmRIjQ0ZfK/RKG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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.031562146049868`, 12.588381073024934`}, {-1, -1}], 
         {PointSize[0.04], PointBox[{8.5, 16.5}], PointBox[{14., 6.5}], 
          PointBox[{10., 14.}], PointBox[{12.5, 9.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["14",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/becfdf/egfhghgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/becfdf/egfhghgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl0gs0lGkYB/A3FRIWiSGXaaPcKjoq9+9zbc5WdtCirZh2FckRbYXWaoRw
Ukm6qEwTVjqk6YZEpoylk0KULrKW7Eg5mZJrjv2/Z+ecmff8zvu8z/d8/3cW
/7LHf4cSIWQvvnQl6vSHy5I5dF3Fklb5jkNxpiw50B7TPW3HElXZPa8xE5ZI
82SlE3B27IrUI7CPm/UFBRwjUsRawX8p14a+o/XTJmf6jVkSXZMU1AZLBJPx
lfBtU764iu57Rd28CO910tK9AAcfuPnzCdjDOM7mILVN5ZN0ONGA9ywI7uls
vfIHbFhnZbIGrvLxeBMBN2UELNCn88gvVXvA+7oUW7/ZsiQ+qfy5MszZMa0n
h8lRu6xbRiwpKki92gmXHL/h+QNcE+d07wmsWrHOsWUR3i9nxqUJDk4I/8qD
I80Gsxrg4c2y3+oM6bzeZfWwLYdd5AznXD2WLIUdNpKSegPk0NJvXA0PuOVa
h8LDncrT5fT8G0cvbdgiKmqtCG6KtvvzNQf5rmHE6XQ+/zMzVbCeiPt4Nxzh
U+17BbYa0Z9YD1fFmieJ4IFB/U+WsHSwMeAMfJ4ff08FztA2FB6BJTJ539OV
6O+rdCMcjj8qf3IarnKfSrGHebk6/+6EeVW2DoP6yGGyJcsTHlDLnpcJS2+5
F1nDUrmKpS4cn26aawwLbcb9T+qhv2tckh6tf71fWxUedksW6cD8APlD4ULc
7xdegwY87rgqbTbMOarxbjbdL2Msz+qyJOxjQo5iBUsEJ9L9WPjckq2HOmBh
uOwlgZser68thcc9RLmdC1iipQj6Jw6OX+OYKIVVkxo+OcHSx6XL78CsyhaD
8eXof8TkwTV43Pul3y24tW3D08swd9L5SQxc5Zt/OBPu2VSzzx6WVvYWhsDq
lRE7ZsHCA4EhxjBHxSbzlQ3WupX1Mh3ktDspqBZu0kuUB8K5GSMNpbBwreul
l9rI4fvSyAI4XuPWWz4cLzFqzofJ0L6WRi2WeLlqdeRR12s2e8PvFI2BOfT8
arPUou+Q18OxiymwbVTq4X5N/H/vSrojYb6ha9ommPSHXPKBBV+5WQoNlpR9
Hu3Wg8UOPUUVcJZo3OCtNfrNr7t+Di7J9OZchMVux1oy4KiEpSeDYLI/4KcD
cETYdskCuPXt8gE+rR9xz+qwwn6AcoEaLORGiPLh1r4f3+Sp471lHfxYeHjJ
tsjJ+SyxKdIs58PiMs3lK+CS4BtuzrBEX+uxixryPvuKs5J6QFS1ah761saY
WsLZfafOm6oiT0OrXeawUHWQ1VVhycewvgemtL+z+CBXGXXdktW6sFZbq5rf
XNQ355+ascR9xZ5QKp/Dkq0OLp1/w2JfToUj/M5SsPs23W9/tGd4Ns717XRN
ggV2XYNNcFhgloiBe+ZMpd2Bb+t4j01ZwPOWKBfDA+9TAithSW/y9UyYp574
YT/MtZuI9oNzJ9f+7gxnL5Q2K5RQXyjvU4PJ9TvhYbBNqNL9/mW4r2fbm0pm
IY+XScua4ZghQYyMoN/Y57oaWDp6kJc6w5AyNaMXt2Gxf7Lm9mmG8NYNSSXU
94XR7BRDRroKgstovenoRs8Jhlh02gsLqfPUv6WNMUTGC7+WQ+3qmGk0yhBd
nb5tcbAgOWel1leGnOtZHcmnnle4OGSEIVGvZqxNYG64uXzsC/p/Cd3buxTe
XObcAGu1G/uIYfHdbXZXYIHnhrQtMPtc2TYRNjPf+isHFqQFbzCDOxyWdb82
R34v7oSnfWZIiZPcsRiWlltmH1cwpFVqcToRJo69tk7DDGHVJQUhsFB+/LTx
EEMkXSE566mNS5xc3jPEKFffxh1m275VzO1nSNcuMwMXuh9sFFvdw5AeWbG/
A3W4YsrpFebpflRsR60//qGzjSHNLYc9zejz2i+zmxoZIk4ettegZic8xPcw
r61K46AZLEy5+qiMIV6NA7U11Jy8IX4eQ1RNCralUseEJagno7/6RtaLmn6E
rv+vZux/4Q+Z0Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.491215510839148, 13.508784489160853`}, {-1, 1}], 
         LineBox[{{20.000000000002046`, 17.}, {8.000000000002501, 18.}}], 
         PolygonBox[{{14.597927454946928`, 17.450172712087756`}, {
          13.435290736994567`, 17.948445591210195`}, {13.368854353111576`, 
          17.151208984614293`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14.128720493773297`, 18.439845925279563`},
           {0, -1}], 
         LineBox[{{20.000000000008185`, 10.000000000003638`}, {
          9.000000000008185, 7.000000000003638}}], 
         PolygonBox[{{13.92114170725736, 8.342129556524736}, {
          14.973611330425795`, 9.043775971970357}, {15.184105255059483`, 
          8.271964914980172}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.092168021022232`, 9.390583922918484}, {0, -1}], LineBox[CompressedData["
1:eJwd1H1Q03UcB/DveFh7QEARIRpPd1HJRbihNuKGP/RKcOBsncUxCYRAQATD
QW1wHNeDkRxKFDhzTeQUFGVaUG48xEw4QUX9w0UhcuMqggOmiAEGHr0//e52
v3v9Pt/f5+l3t9CMQnWWC2NsO350Z1G42jZw7O4tXN4cu1DcE2iHmWhYyVZz
TNweOrIIp++Nb5DA9qvLSYFSnB+ccgmBecEqZRzMSWK6hLDCKfo+E95lCXp1
GPnKN3184FM4xHq/rxbWZw1XnaK4dTvbDBvWD8degW2lfxQf9eLY2vm4yQG4
onJAVueJ537mF4fI9lcufrmKYy25f/eNUr6e7sASD47N7LSOOmAW1hKQJeaY
dlVk9wO4YUUVsFfEMbcFbsJO55c0fK2QY8KffQNuUD3Tb7ebBBx74dI6eSed
NzkEPHi3LCKhheJBPsbK5zg2PvN1noE8r/wkDparf7x2mPqJ1hgjYFH9SV0J
nG6v/S4RVtzWN+ZQvT6BlxFeODsS/T7tpzgvzw/5q5KdvsmU7yeF0gL3vpE1
+i655qP2EvTnMbbHXQM75g/tS0D/R8vLfvmA6oUb41/CfLGpBm8t9ZujHXTB
/HNvT7ZW0vwTGYX34JUnQ0GN/8d3Bn2FfdWmKdp6KO660RyJfYr3OappX1y+
KOYcvPXhg1ZXGaz+ofwJvHap7unLcMU2pa83voezsmlHIpye6da6iPjDzg/1
hXCILS3YDC/Ymwpq4IaFe0UyOHTPm2Fm2FFwtZe+Fy//eM91sqVe0oH+TNUX
qkYo/2Bk+E3Ms9Q2Lp6m+Kj/4z7Mu2Fas/wP5Q9L2d+OfeQfu7T7Xzp/fm7N
aeyr6NvspKewzeecrh77dZs2vTNH70sf3zXxOfYoN8Fzks6bhWXX3LFvaYD6
PvU/az7vAUscpc/fpPMp65Yq3DjGN0zorFRvTVdMGKy6ITE0U/4j0ohlV45Z
FGdO1lO+GkkuH3F5f3zjF7SvLeaUrfD6A1NtpWTv31UXYZt74pVDVG/8yOlY
1DNoo3MOwqxp0+wsfMe/TlVEFgt13ei3P6OiW0f1H2n3H8c8XjNzfYepv44/
U4sxL3/HX1tOUP6o8I5E7OPOjNv1yxRf7JX6YF8Hs3m2W1Rvc/BrPbDX2dA0
2qdtZVEej/3uMjove0bh+1hrkpvh6mfVZ2SwY5tRMwS3Z6b6vQfbpn5dbYeb
P1eN6eEQS2fBKTjK+tmYEeaOzX7zOlyprxN1wax/eqAB9crnBQVD9L5/ktqB
/qJP6N9yUv7sMvkz9M+jP5mNyEd3AfcfyiqUyA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.879701747784395`, 4.451879806407078}, {0, -1}], 
         LineBox[{{8.000000000003638, 18.00000000000705}, {9.000000000003638, 
          7.000000000011141}}], 
         PolygonBox[{{8.445678552374488, 13.097535923880622`}, {
          8.15596416503843, 11.86624977770237}, {8.952678730212593, 
          11.938678374536385`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.561165529975058, 12.359669593634095`}, {1, 0}], LineBox[CompressedData["
1:eJwV1A9Qk3UYB/A3jhNFVJKJoGAb2xqGKFugO0h9Bbbhn2zg6jLucvwHgZxI
SaI2+SPmuXNeLkSEgKmHwcULiXTAXZOAFhM2J9LCDheQLQpvGr5Roev7+929
t/vc8+x5n+f53b289EMpWV4URSnwkF8qtwYnnKaUr+Ospqkyzqn+t2EuK7zp
CaQpprpXsRw21XcdmIKjji/KHBAhX8J5/hDuvp2WdwJ2T/UnsHAuf6ZOCpsU
v9wTol7HxRXCf15FPYUgthC2Re92mWCn/BvPD3B96+6yC7AyfvmANIimLgYc
jyyA3XG12Z1wus14KxlWG/5YHxdMUyc3iIvjYT2X96QfZtN08dtgZvbLojfX
0NSd6cbKRJJPOYqs8PjLJzepSL6XfUG2lqZay0ct+bCN+exSK6xvtJ//lPTD
nC/1wNLaa4UtpJ6GbY8Loal5vuM1K2zyHt92AC5KNZQ8JfUz7OI8mLln9gvA
vBrD5vb9sCrDt2cjrK8rnRfDk1/LzyUQDwztmkP9kb380bdgRpA2boQzz8Yk
pJB9bem/lgh/tVX+bCext3zPA/Tf6+VTuhl2zrQ/KIBVMuvEahK3yGpeYP6F
1Mirj9GP/7m9ZgPcukrZ10P2Z3FotsCX6coaLen/xa/hLuyzW7x9muyLGjo2
2QKreIlX54Xwj9dNlSS+ZNSPgdXvv/fsKDw223EwB9aPnLhRDv9bFbKUB/tz
le5m+M99HWEPBbi/CsPgI7hC8WGnEWaSvggl9zUquqvVEBtvPmmCr083u+Qk
P9h1cA3mYaslLhGJ6/9aVAeXiJJ8OLB+NmVIiH2wzgbVUlgdcfqjNtgxJMla
Rv7PRnwXjf2aRX3vBsNRv52uZuCfIwqnImF3+H871oWi7hVL6k5YWxXe8jH8
+fbFw3nkffbism/hZutGPx3M/ZuXPgtL8wetDGyKzfn9pXU09U5Fl9kOO7sO
WxYQVx/6oMdN6jmb7RNwgXD9pcVkHyUBjhuwci7nQhDZpy62TQ33Buo2vULi
kVk+PrBk5L56LdxwrDShEf1Gx3oq/GDuGdkuMk+3MKj4Keo3NC1LHsC8Y/kx
2mHyvkczTfvhQAF7uB62lYfmsNiXS/L8k2yYynLnNsBjnb07yD652vKpVJj+
Pqx+ko++jF37NsBm7+lbl+GoCXEKB9ZIZXeTYfos5+gqWDun3uoL+wuakqLg
Ah37xmAY7kNaFZMJm3zvz1TB/leW1LbBfSsf85Nh25Gfxlagv+xh3yN8mLlz
O+MU7PTg8JBPPjYh9P+oBpxX
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.120298252215608`, 5.548120193592922}, {0, 1}], 
         CircleBox[{14.32, 4.505000000000003}, 2.399463481697522, {
          6.15572874594347, 8.854692514837682}], 
         PolygonBox[{{15.703876054093064`, 6.554954162147977}, {
          14.71282117114162, 7.340963207247397}, {14.439426720672255`, 
          6.589128468456647}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.669701747784393`, 7.611879806407077}, {0, -1}], 
         CircleBox[{14.68, 5.494999999999997}, 2.399463481697522, {
          3.014136092353677, 5.713099861247888}], 
         PolygonBox[{{13.296123945906936`, 3.4450458378520237`}, {
          14.28717882885838, 2.659036792752604}, {14.560573279327745`, 
          3.410871531543353}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{13.330298252215607`, 2.3881201935929237`}, {0, 1}], 
         {PointSize[0.04], PointBox[{8., 18.}], PointBox[{9., 7.}], 
          PointBox[{16.7, 4.2}], PointBox[{12.3, 5.8}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["15",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl0gs0lGkYB/A3FRIWiSGXaaPcKjoq9+9zbc5WdtCirZh2FckRbYXWaoRw
Ukm6qEwTVjqk6YZEpoylk0KULrKW7Eg5mZJrjv2/Z+ecmff8zvu8z/d8/3cW
/7LHf4cSIWQvvnQl6vSHy5I5dF3Fklb5jkNxpiw50B7TPW3HElXZPa8xE5ZI
82SlE3B27IrUI7CPm/UFBRwjUsRawX8p14a+o/XTJmf6jVkSXZMU1AZLBJPx
lfBtU764iu57Rd28CO910tK9AAcfuPnzCdjDOM7mILVN5ZN0ONGA9ywI7uls
vfIHbFhnZbIGrvLxeBMBN2UELNCn88gvVXvA+7oUW7/ZsiQ+qfy5MszZMa0n
h8lRu6xbRiwpKki92gmXHL/h+QNcE+d07wmsWrHOsWUR3i9nxqUJDk4I/8qD
I80Gsxrg4c2y3+oM6bzeZfWwLYdd5AznXD2WLIUdNpKSegPk0NJvXA0PuOVa
h8LDncrT5fT8G0cvbdgiKmqtCG6KtvvzNQf5rmHE6XQ+/zMzVbCeiPt4Nxzh
U+17BbYa0Z9YD1fFmieJ4IFB/U+WsHSwMeAMfJ4ff08FztA2FB6BJTJ539OV
6O+rdCMcjj8qf3IarnKfSrGHebk6/+6EeVW2DoP6yGGyJcsTHlDLnpcJS2+5
F1nDUrmKpS4cn26aawwLbcb9T+qhv2tckh6tf71fWxUedksW6cD8APlD4ULc
7xdegwY87rgqbTbMOarxbjbdL2Msz+qyJOxjQo5iBUsEJ9L9WPjckq2HOmBh
uOwlgZser68thcc9RLmdC1iipQj6Jw6OX+OYKIVVkxo+OcHSx6XL78CsyhaD
8eXof8TkwTV43Pul3y24tW3D08swd9L5SQxc5Zt/OBPu2VSzzx6WVvYWhsDq
lRE7ZsHCA4EhxjBHxSbzlQ3WupX1Mh3ktDspqBZu0kuUB8K5GSMNpbBwreul
l9rI4fvSyAI4XuPWWz4cLzFqzofJ0L6WRi2WeLlqdeRR12s2e8PvFI2BOfT8
arPUou+Q18OxiymwbVTq4X5N/H/vSrojYb6ha9ommPSHXPKBBV+5WQoNlpR9
Hu3Wg8UOPUUVcJZo3OCtNfrNr7t+Di7J9OZchMVux1oy4KiEpSeDYLI/4KcD
cETYdskCuPXt8gE+rR9xz+qwwn6AcoEaLORGiPLh1r4f3+Sp471lHfxYeHjJ
tsjJ+SyxKdIs58PiMs3lK+CS4BtuzrBEX+uxixryPvuKs5J6QFS1ah761saY
WsLZfafOm6oiT0OrXeawUHWQ1VVhycewvgemtL+z+CBXGXXdktW6sFZbq5rf
XNQ355+ascR9xZ5QKp/Dkq0OLp1/w2JfToUj/M5SsPs23W9/tGd4Ns717XRN
ggV2XYNNcFhgloiBe+ZMpd2Bb+t4j01ZwPOWKBfDA+9TAithSW/y9UyYp574
YT/MtZuI9oNzJ9f+7gxnL5Q2K5RQXyjvU4PJ9TvhYbBNqNL9/mW4r2fbm0pm
IY+XScua4ZghQYyMoN/Y57oaWDp6kJc6w5AyNaMXt2Gxf7Lm9mmG8NYNSSXU
94XR7BRDRroKgstovenoRs8Jhlh02gsLqfPUv6WNMUTGC7+WQ+3qmGk0yhBd
nb5tcbAgOWel1leGnOtZHcmnnle4OGSEIVGvZqxNYG64uXzsC/p/Cd3buxTe
XObcAGu1G/uIYfHdbXZXYIHnhrQtMPtc2TYRNjPf+isHFqQFbzCDOxyWdb82
R34v7oSnfWZIiZPcsRiWlltmH1cwpFVqcToRJo69tk7DDGHVJQUhsFB+/LTx
EEMkXSE566mNS5xc3jPEKFffxh1m275VzO1nSNcuMwMXuh9sFFvdw5AeWbG/
A3W4YsrpFebpflRsR60//qGzjSHNLYc9zejz2i+zmxoZIk4ettegZic8xPcw
r61K46AZLEy5+qiMIV6NA7U11Jy8IX4eQ1RNCralUseEJagno7/6RtaLmn6E
rv+vZux/4Q+Z0Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.491215510839148, 13.508784489160853`}, {-1, 1}], 
         LineBox[{{20.000000000002046`, 17.}, {8.000000000002501, 18.}}], 
         PolygonBox[{{14.597927454946928`, 17.450172712087756`}, {
          13.435290736994567`, 17.948445591210195`}, {13.368854353111576`, 
          17.151208984614293`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.128720493773297`, 18.439845925279563`}, {0, -1}], 
         LineBox[{{20.000000000008185`, 10.000000000003638`}, {
          9.000000000008185, 7.000000000003638}}], 
         PolygonBox[{{13.92114170725736, 8.342129556524736}, {
          14.973611330425795`, 9.043775971970357}, {15.184105255059483`, 
          8.271964914980172}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.092168021022232`, 9.390583922918484}, {0, -1}], LineBox[CompressedData["
1:eJwd1H1Q03UcB/DveFh7QEARIRpPd1HJRbihNuKGP/RKcOBsncUxCYRAQATD
QW1wHNeDkRxKFDhzTeQUFGVaUG48xEw4QUX9w0UhcuMqggOmiAEGHr0//e52
v3v9Pt/f5+l3t9CMQnWWC2NsO350Z1G42jZw7O4tXN4cu1DcE2iHmWhYyVZz
TNweOrIIp++Nb5DA9qvLSYFSnB+ccgmBecEqZRzMSWK6hLDCKfo+E95lCXp1
GPnKN3184FM4xHq/rxbWZw1XnaK4dTvbDBvWD8degW2lfxQf9eLY2vm4yQG4
onJAVueJ537mF4fI9lcufrmKYy25f/eNUr6e7sASD47N7LSOOmAW1hKQJeaY
dlVk9wO4YUUVsFfEMbcFbsJO55c0fK2QY8KffQNuUD3Tb7ebBBx74dI6eSed
NzkEPHi3LCKhheJBPsbK5zg2PvN1noE8r/wkDparf7x2mPqJ1hgjYFH9SV0J
nG6v/S4RVtzWN+ZQvT6BlxFeODsS/T7tpzgvzw/5q5KdvsmU7yeF0gL3vpE1
+i655qP2EvTnMbbHXQM75g/tS0D/R8vLfvmA6oUb41/CfLGpBm8t9ZujHXTB
/HNvT7ZW0vwTGYX34JUnQ0GN/8d3Bn2FfdWmKdp6KO660RyJfYr3OappX1y+
KOYcvPXhg1ZXGaz+ofwJvHap7unLcMU2pa83voezsmlHIpye6da6iPjDzg/1
hXCILS3YDC/Ymwpq4IaFe0UyOHTPm2Fm2FFwtZe+Fy//eM91sqVe0oH+TNUX
qkYo/2Bk+E3Ms9Q2Lp6m+Kj/4z7Mu2Fas/wP5Q9L2d+OfeQfu7T7Xzp/fm7N
aeyr6NvspKewzeecrh77dZs2vTNH70sf3zXxOfYoN8Fzks6bhWXX3LFvaYD6
PvU/az7vAUscpc/fpPMp65Yq3DjGN0zorFRvTVdMGKy6ITE0U/4j0ohlV45Z
FGdO1lO+GkkuH3F5f3zjF7SvLeaUrfD6A1NtpWTv31UXYZt74pVDVG/8yOlY
1DNoo3MOwqxp0+wsfMe/TlVEFgt13ei3P6OiW0f1H2n3H8c8XjNzfYepv44/
U4sxL3/HX1tOUP6o8I5E7OPOjNv1yxRf7JX6YF8Hs3m2W1Rvc/BrPbDX2dA0
2qdtZVEej/3uMjove0bh+1hrkpvh6mfVZ2SwY5tRMwS3Z6b6vQfbpn5dbYeb
P1eN6eEQS2fBKTjK+tmYEeaOzX7zOlyprxN1wax/eqAB9crnBQVD9L5/ktqB
/qJP6N9yUv7sMvkz9M+jP5mNyEd3AfcfyiqUyA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.879701747784395`, 4.451879806407078}, {0, -1}], 
         LineBox[{{8.000000000003638, 18.00000000000705}, {9.000000000003638, 
          7.000000000011141}}], 
         PolygonBox[{{8.445678552374488, 13.097535923880622`}, {
          8.15596416503843, 11.86624977770237}, {8.952678730212593, 
          11.938678374536385`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.561165529975058, 12.359669593634095`}, {1, 0}], LineBox[CompressedData["
1:eJwV1A9Qk3UYB/A3jhNFVJKJoGAb2xqGKFugO0h9Bbbhn2zg6jLucvwHgZxI
SaI2+SPmuXNeLkSEgKmHwcULiXTAXZOAFhM2J9LCDheQLQpvGr5Roev7+929
t/vc8+x5n+f53b289EMpWV4URSnwkF8qtwYnnKaUr+Ospqkyzqn+t2EuK7zp
CaQpprpXsRw21XcdmIKjji/KHBAhX8J5/hDuvp2WdwJ2T/UnsHAuf6ZOCpsU
v9wTol7HxRXCf15FPYUgthC2Re92mWCn/BvPD3B96+6yC7AyfvmANIimLgYc
jyyA3XG12Z1wus14KxlWG/5YHxdMUyc3iIvjYT2X96QfZtN08dtgZvbLojfX
0NSd6cbKRJJPOYqs8PjLJzepSL6XfUG2lqZay0ct+bCN+exSK6xvtJ//lPTD
nC/1wNLaa4UtpJ6GbY8Loal5vuM1K2zyHt92AC5KNZQ8JfUz7OI8mLln9gvA
vBrD5vb9sCrDt2cjrK8rnRfDk1/LzyUQDwztmkP9kb380bdgRpA2boQzz8Yk
pJB9bem/lgh/tVX+bCext3zPA/Tf6+VTuhl2zrQ/KIBVMuvEahK3yGpeYP6F
1Mirj9GP/7m9ZgPcukrZ10P2Z3FotsCX6coaLen/xa/hLuyzW7x9muyLGjo2
2QKreIlX54Xwj9dNlSS+ZNSPgdXvv/fsKDw223EwB9aPnLhRDv9bFbKUB/tz
le5m+M99HWEPBbi/CsPgI7hC8WGnEWaSvggl9zUquqvVEBtvPmmCr083u+Qk
P9h1cA3mYaslLhGJ6/9aVAeXiJJ8OLB+NmVIiH2wzgbVUlgdcfqjNtgxJMla
Rv7PRnwXjf2aRX3vBsNRv52uZuCfIwqnImF3+H871oWi7hVL6k5YWxXe8jH8
+fbFw3nkffbism/hZutGPx3M/ZuXPgtL8wetDGyKzfn9pXU09U5Fl9kOO7sO
WxYQVx/6oMdN6jmb7RNwgXD9pcVkHyUBjhuwci7nQhDZpy62TQ33Buo2vULi
kVk+PrBk5L56LdxwrDShEf1Gx3oq/GDuGdkuMk+3MKj4Keo3NC1LHsC8Y/kx
2mHyvkczTfvhQAF7uB62lYfmsNiXS/L8k2yYynLnNsBjnb07yD652vKpVJj+
Pqx+ko++jF37NsBm7+lbl+GoCXEKB9ZIZXeTYfos5+gqWDun3uoL+wuakqLg
Ah37xmAY7kNaFZMJm3zvz1TB/leW1LbBfSsf85Nh25Gfxlagv+xh3yN8mLlz
O+MU7PTg8JBPPjYh9P+oBpxX
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.120298252215608`, 5.548120193592922}, {0, 1}], 
         CircleBox[{14.32, 4.505000000000003}, 2.399463481697522, {
          6.15572874594347, 8.854692514837682}], 
         PolygonBox[{{15.703876054093064`, 6.554954162147977}, {
          14.71282117114162, 7.340963207247397}, {14.439426720672255`, 
          6.589128468456647}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.669701747784393`, 7.611879806407077}, {0, -1}], 
         CircleBox[{14.68, 5.494999999999997}, 2.399463481697522, {
          3.014136092353677, 5.713099861247888}], 
         PolygonBox[{{13.296123945906936`, 3.4450458378520237`}, {
          14.28717882885838, 2.659036792752604}, {14.560573279327745`, 
          3.410871531543353}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.330298252215607`, 2.3881201935929237`}, {0, 1}], 
         {PointSize[0.04], PointBox[{8., 18.}], PointBox[{9., 7.}], 
          PointBox[{16.7, 4.2}], PointBox[{12.3, 5.8}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["16",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl0gs0lGkYB/A3FRIWiSGXaaPcKjoq9+9zbc5WdtCirZh2FckRbYXWaoRw
Ukm6qEwTVjqk6YZEpoylk0KULrKW7Eg5mZJrjv2/Z+ecmff8zvu8z/d8/3cW
/7LHf4cSIWQvvnQl6vSHy5I5dF3Fklb5jkNxpiw50B7TPW3HElXZPa8xE5ZI
82SlE3B27IrUI7CPm/UFBRwjUsRawX8p14a+o/XTJmf6jVkSXZMU1AZLBJPx
lfBtU764iu57Rd28CO910tK9AAcfuPnzCdjDOM7mILVN5ZN0ONGA9ywI7uls
vfIHbFhnZbIGrvLxeBMBN2UELNCn88gvVXvA+7oUW7/ZsiQ+qfy5MszZMa0n
h8lRu6xbRiwpKki92gmXHL/h+QNcE+d07wmsWrHOsWUR3i9nxqUJDk4I/8qD
I80Gsxrg4c2y3+oM6bzeZfWwLYdd5AznXD2WLIUdNpKSegPk0NJvXA0PuOVa
h8LDncrT5fT8G0cvbdgiKmqtCG6KtvvzNQf5rmHE6XQ+/zMzVbCeiPt4Nxzh
U+17BbYa0Z9YD1fFmieJ4IFB/U+WsHSwMeAMfJ4ff08FztA2FB6BJTJ539OV
6O+rdCMcjj8qf3IarnKfSrGHebk6/+6EeVW2DoP6yGGyJcsTHlDLnpcJS2+5
F1nDUrmKpS4cn26aawwLbcb9T+qhv2tckh6tf71fWxUedksW6cD8APlD4ULc
7xdegwY87rgqbTbMOarxbjbdL2Msz+qyJOxjQo5iBUsEJ9L9WPjckq2HOmBh
uOwlgZser68thcc9RLmdC1iipQj6Jw6OX+OYKIVVkxo+OcHSx6XL78CsyhaD
8eXof8TkwTV43Pul3y24tW3D08swd9L5SQxc5Zt/OBPu2VSzzx6WVvYWhsDq
lRE7ZsHCA4EhxjBHxSbzlQ3WupX1Mh3ktDspqBZu0kuUB8K5GSMNpbBwreul
l9rI4fvSyAI4XuPWWz4cLzFqzofJ0L6WRi2WeLlqdeRR12s2e8PvFI2BOfT8
arPUou+Q18OxiymwbVTq4X5N/H/vSrojYb6ha9ommPSHXPKBBV+5WQoNlpR9
Hu3Wg8UOPUUVcJZo3OCtNfrNr7t+Di7J9OZchMVux1oy4KiEpSeDYLI/4KcD
cETYdskCuPXt8gE+rR9xz+qwwn6AcoEaLORGiPLh1r4f3+Sp471lHfxYeHjJ
tsjJ+SyxKdIs58PiMs3lK+CS4BtuzrBEX+uxixryPvuKs5J6QFS1ah761saY
WsLZfafOm6oiT0OrXeawUHWQ1VVhycewvgemtL+z+CBXGXXdktW6sFZbq5rf
XNQ355+ascR9xZ5QKp/Dkq0OLp1/w2JfToUj/M5SsPs23W9/tGd4Ns717XRN
ggV2XYNNcFhgloiBe+ZMpd2Bb+t4j01ZwPOWKBfDA+9TAithSW/y9UyYp574
YT/MtZuI9oNzJ9f+7gxnL5Q2K5RQXyjvU4PJ9TvhYbBNqNL9/mW4r2fbm0pm
IY+XScua4ZghQYyMoN/Y57oaWDp6kJc6w5AyNaMXt2Gxf7Lm9mmG8NYNSSXU
94XR7BRDRroKgstovenoRs8Jhlh02gsLqfPUv6WNMUTGC7+WQ+3qmGk0yhBd
nb5tcbAgOWel1leGnOtZHcmnnle4OGSEIVGvZqxNYG64uXzsC/p/Cd3buxTe
XObcAGu1G/uIYfHdbXZXYIHnhrQtMPtc2TYRNjPf+isHFqQFbzCDOxyWdb82
R34v7oSnfWZIiZPcsRiWlltmH1cwpFVqcToRJo69tk7DDGHVJQUhsFB+/LTx
EEMkXSE566mNS5xc3jPEKFffxh1m275VzO1nSNcuMwMXuh9sFFvdw5AeWbG/
A3W4YsrpFebpflRsR60//qGzjSHNLYc9zejz2i+zmxoZIk4ettegZic8xPcw
r61K46AZLEy5+qiMIV6NA7U11Jy8IX4eQ1RNCralUseEJagno7/6RtaLmn6E
rv+vZux/4Q+Z0Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.491215510839148, 13.508784489160853`}, {-1, 1}], 
         LineBox[{{20.000000000002046`, 17.}, {8.000000000002501, 18.}}], 
         PolygonBox[{{14.597927454946928`, 17.450172712087756`}, {
          13.435290736994567`, 17.948445591210195`}, {13.368854353111576`, 
          17.151208984614293`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.128720493773297`, 18.439845925279563`}, {0, -1}], 
         LineBox[{{20.000000000008185`, 10.000000000003638`}, {
          9.000000000008185, 7.000000000003638}}], 
         PolygonBox[{{13.92114170725736, 8.342129556524736}, {
          14.973611330425795`, 9.043775971970357}, {15.184105255059483`, 
          8.271964914980172}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.092168021022232`, 9.390583922918484}, {0, -1}], LineBox[CompressedData["
1:eJwd1H1Q03UcB/DveFh7QEARIRpPd1HJRbihNuKGP/RKcOBsncUxCYRAQATD
QW1wHNeDkRxKFDhzTeQUFGVaUG48xEw4QUX9w0UhcuMqggOmiAEGHr0//e52
v3v9Pt/f5+l3t9CMQnWWC2NsO350Z1G42jZw7O4tXN4cu1DcE2iHmWhYyVZz
TNweOrIIp++Nb5DA9qvLSYFSnB+ccgmBecEqZRzMSWK6hLDCKfo+E95lCXp1
GPnKN3184FM4xHq/rxbWZw1XnaK4dTvbDBvWD8degW2lfxQf9eLY2vm4yQG4
onJAVueJ537mF4fI9lcufrmKYy25f/eNUr6e7sASD47N7LSOOmAW1hKQJeaY
dlVk9wO4YUUVsFfEMbcFbsJO55c0fK2QY8KffQNuUD3Tb7ebBBx74dI6eSed
NzkEPHi3LCKhheJBPsbK5zg2PvN1noE8r/wkDparf7x2mPqJ1hgjYFH9SV0J
nG6v/S4RVtzWN+ZQvT6BlxFeODsS/T7tpzgvzw/5q5KdvsmU7yeF0gL3vpE1
+i655qP2EvTnMbbHXQM75g/tS0D/R8vLfvmA6oUb41/CfLGpBm8t9ZujHXTB
/HNvT7ZW0vwTGYX34JUnQ0GN/8d3Bn2FfdWmKdp6KO660RyJfYr3OappX1y+
KOYcvPXhg1ZXGaz+ofwJvHap7unLcMU2pa83voezsmlHIpye6da6iPjDzg/1
hXCILS3YDC/Ymwpq4IaFe0UyOHTPm2Fm2FFwtZe+Fy//eM91sqVe0oH+TNUX
qkYo/2Bk+E3Ms9Q2Lp6m+Kj/4z7Mu2Fas/wP5Q9L2d+OfeQfu7T7Xzp/fm7N
aeyr6NvspKewzeecrh77dZs2vTNH70sf3zXxOfYoN8Fzks6bhWXX3LFvaYD6
PvU/az7vAUscpc/fpPMp65Yq3DjGN0zorFRvTVdMGKy6ITE0U/4j0ohlV45Z
FGdO1lO+GkkuH3F5f3zjF7SvLeaUrfD6A1NtpWTv31UXYZt74pVDVG/8yOlY
1DNoo3MOwqxp0+wsfMe/TlVEFgt13ei3P6OiW0f1H2n3H8c8XjNzfYepv44/
U4sxL3/HX1tOUP6o8I5E7OPOjNv1yxRf7JX6YF8Hs3m2W1Rvc/BrPbDX2dA0
2qdtZVEej/3uMjove0bh+1hrkpvh6mfVZ2SwY5tRMwS3Z6b6vQfbpn5dbYeb
P1eN6eEQS2fBKTjK+tmYEeaOzX7zOlyprxN1wax/eqAB9crnBQVD9L5/ktqB
/qJP6N9yUv7sMvkz9M+jP5mNyEd3AfcfyiqUyA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.879701747784395`, 4.451879806407078}, {0, -1}], 
         LineBox[{{8.000000000003638, 18.00000000000705}, {9.000000000003638, 
          7.000000000011141}}], 
         PolygonBox[{{8.445678552374488, 13.097535923880622`}, {
          8.15596416503843, 11.86624977770237}, {8.952678730212593, 
          11.938678374536385`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.561165529975058, 12.359669593634095`}, {1, 0}], LineBox[CompressedData["
1:eJwV1A9Qk3UYB/A3jhNFVJKJoGAb2xqGKFugO0h9Bbbhn2zg6jLucvwHgZxI
SaI2+SPmuXNeLkSEgKmHwcULiXTAXZOAFhM2J9LCDheQLQpvGr5Roev7+929
t/vc8+x5n+f53b289EMpWV4URSnwkF8qtwYnnKaUr+Ospqkyzqn+t2EuK7zp
CaQpprpXsRw21XcdmIKjji/KHBAhX8J5/hDuvp2WdwJ2T/UnsHAuf6ZOCpsU
v9wTol7HxRXCf15FPYUgthC2Re92mWCn/BvPD3B96+6yC7AyfvmANIimLgYc
jyyA3XG12Z1wus14KxlWG/5YHxdMUyc3iIvjYT2X96QfZtN08dtgZvbLojfX
0NSd6cbKRJJPOYqs8PjLJzepSL6XfUG2lqZay0ct+bCN+exSK6xvtJ//lPTD
nC/1wNLaa4UtpJ6GbY8Loal5vuM1K2zyHt92AC5KNZQ8JfUz7OI8mLln9gvA
vBrD5vb9sCrDt2cjrK8rnRfDk1/LzyUQDwztmkP9kb380bdgRpA2boQzz8Yk
pJB9bem/lgh/tVX+bCext3zPA/Tf6+VTuhl2zrQ/KIBVMuvEahK3yGpeYP6F
1Mirj9GP/7m9ZgPcukrZ10P2Z3FotsCX6coaLen/xa/hLuyzW7x9muyLGjo2
2QKreIlX54Xwj9dNlSS+ZNSPgdXvv/fsKDw223EwB9aPnLhRDv9bFbKUB/tz
le5m+M99HWEPBbi/CsPgI7hC8WGnEWaSvggl9zUquqvVEBtvPmmCr083u+Qk
P9h1cA3mYaslLhGJ6/9aVAeXiJJ8OLB+NmVIiH2wzgbVUlgdcfqjNtgxJMla
Rv7PRnwXjf2aRX3vBsNRv52uZuCfIwqnImF3+H871oWi7hVL6k5YWxXe8jH8
+fbFw3nkffbism/hZutGPx3M/ZuXPgtL8wetDGyKzfn9pXU09U5Fl9kOO7sO
WxYQVx/6oMdN6jmb7RNwgXD9pcVkHyUBjhuwci7nQhDZpy62TQ33Buo2vULi
kVk+PrBk5L56LdxwrDShEf1Gx3oq/GDuGdkuMk+3MKj4Keo3NC1LHsC8Y/kx
2mHyvkczTfvhQAF7uB62lYfmsNiXS/L8k2yYynLnNsBjnb07yD652vKpVJj+
Pqx+ko++jF37NsBm7+lbl+GoCXEKB9ZIZXeTYfos5+gqWDun3uoL+wuakqLg
Ah37xmAY7kNaFZMJm3zvz1TB/leW1LbBfSsf85Nh25Gfxlagv+xh3yN8mLlz
O+MU7PTg8JBPPjYh9P+oBpxX
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.120298252215608`, 5.548120193592922}, {0, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{14.32, 4.505000000000003}, 2.399463481697522, {
           6.15572874594347, 8.854692514837682}]}, 
         PolygonBox[{{15.703876054093064`, 6.554954162147977}, {
          14.71282117114162, 7.340963207247397}, {14.439426720672255`, 
          6.589128468456647}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.669701747784393`, 7.611879806407077}, {0, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{14.68, 5.494999999999997}, 2.399463481697522, {
           3.014136092353677, 5.713099861247888}]}, 
         PolygonBox[{{13.296123945906936`, 3.4450458378520237`}, {
          14.28717882885838, 2.659036792752604}, {14.560573279327745`, 
          3.410871531543353}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.330298252215607`, 2.3881201935929237`}, {0, 1}], 
         {PointSize[0.04], PointBox[{8., 18.}], PointBox[{9., 7.}], 
          PointBox[{16.7, 4.2}], PointBox[{12.3, 5.8}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["17",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl0gs0lGkYB/A3FRIWiSGXaaPcKjoq9+9zbc5WdtCirZh2FckRbYXWaoRw
Ukm6qEwTVjqk6YZEpoylk0KULrKW7Eg5mZJrjv2/Z+ecmff8zvu8z/d8/3cW
/7LHf4cSIWQvvnQl6vSHy5I5dF3Fklb5jkNxpiw50B7TPW3HElXZPa8xE5ZI
82SlE3B27IrUI7CPm/UFBRwjUsRawX8p14a+o/XTJmf6jVkSXZMU1AZLBJPx
lfBtU764iu57Rd28CO910tK9AAcfuPnzCdjDOM7mILVN5ZN0ONGA9ywI7uls
vfIHbFhnZbIGrvLxeBMBN2UELNCn88gvVXvA+7oUW7/ZsiQ+qfy5MszZMa0n
h8lRu6xbRiwpKki92gmXHL/h+QNcE+d07wmsWrHOsWUR3i9nxqUJDk4I/8qD
I80Gsxrg4c2y3+oM6bzeZfWwLYdd5AznXD2WLIUdNpKSegPk0NJvXA0PuOVa
h8LDncrT5fT8G0cvbdgiKmqtCG6KtvvzNQf5rmHE6XQ+/zMzVbCeiPt4Nxzh
U+17BbYa0Z9YD1fFmieJ4IFB/U+WsHSwMeAMfJ4ff08FztA2FB6BJTJ539OV
6O+rdCMcjj8qf3IarnKfSrGHebk6/+6EeVW2DoP6yGGyJcsTHlDLnpcJS2+5
F1nDUrmKpS4cn26aawwLbcb9T+qhv2tckh6tf71fWxUedksW6cD8APlD4ULc
7xdegwY87rgqbTbMOarxbjbdL2Msz+qyJOxjQo5iBUsEJ9L9WPjckq2HOmBh
uOwlgZser68thcc9RLmdC1iipQj6Jw6OX+OYKIVVkxo+OcHSx6XL78CsyhaD
8eXof8TkwTV43Pul3y24tW3D08swd9L5SQxc5Zt/OBPu2VSzzx6WVvYWhsDq
lRE7ZsHCA4EhxjBHxSbzlQ3WupX1Mh3ktDspqBZu0kuUB8K5GSMNpbBwreul
l9rI4fvSyAI4XuPWWz4cLzFqzofJ0L6WRi2WeLlqdeRR12s2e8PvFI2BOfT8
arPUou+Q18OxiymwbVTq4X5N/H/vSrojYb6ha9ommPSHXPKBBV+5WQoNlpR9
Hu3Wg8UOPUUVcJZo3OCtNfrNr7t+Di7J9OZchMVux1oy4KiEpSeDYLI/4KcD
cETYdskCuPXt8gE+rR9xz+qwwn6AcoEaLORGiPLh1r4f3+Sp471lHfxYeHjJ
tsjJ+SyxKdIs58PiMs3lK+CS4BtuzrBEX+uxixryPvuKs5J6QFS1ah761saY
WsLZfafOm6oiT0OrXeawUHWQ1VVhycewvgemtL+z+CBXGXXdktW6sFZbq5rf
XNQ355+ascR9xZ5QKp/Dkq0OLp1/w2JfToUj/M5SsPs23W9/tGd4Ns717XRN
ggV2XYNNcFhgloiBe+ZMpd2Bb+t4j01ZwPOWKBfDA+9TAithSW/y9UyYp574
YT/MtZuI9oNzJ9f+7gxnL5Q2K5RQXyjvU4PJ9TvhYbBNqNL9/mW4r2fbm0pm
IY+XScua4ZghQYyMoN/Y57oaWDp6kJc6w5AyNaMXt2Gxf7Lm9mmG8NYNSSXU
94XR7BRDRroKgstovenoRs8Jhlh02gsLqfPUv6WNMUTGC7+WQ+3qmGk0yhBd
nb5tcbAgOWel1leGnOtZHcmnnle4OGSEIVGvZqxNYG64uXzsC/p/Cd3buxTe
XObcAGu1G/uIYfHdbXZXYIHnhrQtMPtc2TYRNjPf+isHFqQFbzCDOxyWdb82
R34v7oSnfWZIiZPcsRiWlltmH1cwpFVqcToRJo69tk7DDGHVJQUhsFB+/LTx
EEMkXSE566mNS5xc3jPEKFffxh1m275VzO1nSNcuMwMXuh9sFFvdw5AeWbG/
A3W4YsrpFebpflRsR60//qGzjSHNLYc9zejz2i+zmxoZIk4ettegZic8xPcw
r61K46AZLEy5+qiMIV6NA7U11Jy8IX4eQ1RNCralUseEJagno7/6RtaLmn6E
rv+vZux/4Q+Z0Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.491215510839148, 13.508784489160853`}, {-1, 1}], 
         LineBox[{{20.000000000002046`, 17.}, {8.000000000002501, 18.}}], 
         PolygonBox[{{14.597927454946928`, 17.450172712087756`}, {
          13.435290736994567`, 17.948445591210195`}, {13.368854353111576`, 
          17.151208984614293`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.128720493773297`, 18.439845925279563`}, {0, -1}], 
         LineBox[{{20.000000000008185`, 10.000000000003638`}, {
          9.000000000008185, 7.000000000003638}}], 
         PolygonBox[{{13.92114170725736, 8.342129556524736}, {
          14.973611330425795`, 9.043775971970357}, {15.184105255059483`, 
          8.271964914980172}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.092168021022232`, 9.390583922918484}, {0, -1}], LineBox[CompressedData["
1:eJwd1H1Q03UcB/DveFh7QEARIRpPd1HJRbihNuKGP/RKcOBsncUxCYRAQATD
QW1wHNeDkRxKFDhzTeQUFGVaUG48xEw4QUX9w0UhcuMqggOmiAEGHr0//e52
v3v9Pt/f5+l3t9CMQnWWC2NsO350Z1G42jZw7O4tXN4cu1DcE2iHmWhYyVZz
TNweOrIIp++Nb5DA9qvLSYFSnB+ccgmBecEqZRzMSWK6hLDCKfo+E95lCXp1
GPnKN3184FM4xHq/rxbWZw1XnaK4dTvbDBvWD8degW2lfxQf9eLY2vm4yQG4
onJAVueJ537mF4fI9lcufrmKYy25f/eNUr6e7sASD47N7LSOOmAW1hKQJeaY
dlVk9wO4YUUVsFfEMbcFbsJO55c0fK2QY8KffQNuUD3Tb7ebBBx74dI6eSed
NzkEPHi3LCKhheJBPsbK5zg2PvN1noE8r/wkDparf7x2mPqJ1hgjYFH9SV0J
nG6v/S4RVtzWN+ZQvT6BlxFeODsS/T7tpzgvzw/5q5KdvsmU7yeF0gL3vpE1
+i655qP2EvTnMbbHXQM75g/tS0D/R8vLfvmA6oUb41/CfLGpBm8t9ZujHXTB
/HNvT7ZW0vwTGYX34JUnQ0GN/8d3Bn2FfdWmKdp6KO660RyJfYr3OappX1y+
KOYcvPXhg1ZXGaz+ofwJvHap7unLcMU2pa83voezsmlHIpye6da6iPjDzg/1
hXCILS3YDC/Ymwpq4IaFe0UyOHTPm2Fm2FFwtZe+Fy//eM91sqVe0oH+TNUX
qkYo/2Bk+E3Ms9Q2Lp6m+Kj/4z7Mu2Fas/wP5Q9L2d+OfeQfu7T7Xzp/fm7N
aeyr6NvspKewzeecrh77dZs2vTNH70sf3zXxOfYoN8Fzks6bhWXX3LFvaYD6
PvU/az7vAUscpc/fpPMp65Yq3DjGN0zorFRvTVdMGKy6ITE0U/4j0ohlV45Z
FGdO1lO+GkkuH3F5f3zjF7SvLeaUrfD6A1NtpWTv31UXYZt74pVDVG/8yOlY
1DNoo3MOwqxp0+wsfMe/TlVEFgt13ei3P6OiW0f1H2n3H8c8XjNzfYepv44/
U4sxL3/HX1tOUP6o8I5E7OPOjNv1yxRf7JX6YF8Hs3m2W1Rvc/BrPbDX2dA0
2qdtZVEej/3uMjove0bh+1hrkpvh6mfVZ2SwY5tRMwS3Z6b6vQfbpn5dbYeb
P1eN6eEQS2fBKTjK+tmYEeaOzX7zOlyprxN1wax/eqAB9crnBQVD9L5/ktqB
/qJP6N9yUv7sMvkz9M+jP5mNyEd3AfcfyiqUyA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.879701747784395`, 4.451879806407078}, {0, -1}], 
         LineBox[{{8.000000000003638, 18.00000000000705}, {9.000000000003638, 
          7.000000000011141}}], 
         PolygonBox[{{8.445678552374488, 13.097535923880622`}, {
          8.15596416503843, 11.86624977770237}, {8.952678730212593, 
          11.938678374536385`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.561165529975058, 12.359669593634095`}, {1, 0}], LineBox[CompressedData["
1:eJwV1A9Qk3UYB/A3jhNFVJKJoGAb2xqGKFugO0h9Bbbhn2zg6jLucvwHgZxI
SaI2+SPmuXNeLkSEgKmHwcULiXTAXZOAFhM2J9LCDheQLQpvGr5Roev7+929
t/vc8+x5n+f53b289EMpWV4URSnwkF8qtwYnnKaUr+Ospqkyzqn+t2EuK7zp
CaQpprpXsRw21XcdmIKjji/KHBAhX8J5/hDuvp2WdwJ2T/UnsHAuf6ZOCpsU
v9wTol7HxRXCf15FPYUgthC2Re92mWCn/BvPD3B96+6yC7AyfvmANIimLgYc
jyyA3XG12Z1wus14KxlWG/5YHxdMUyc3iIvjYT2X96QfZtN08dtgZvbLojfX
0NSd6cbKRJJPOYqs8PjLJzepSL6XfUG2lqZay0ct+bCN+exSK6xvtJ//lPTD
nC/1wNLaa4UtpJ6GbY8Loal5vuM1K2zyHt92AC5KNZQ8JfUz7OI8mLln9gvA
vBrD5vb9sCrDt2cjrK8rnRfDk1/LzyUQDwztmkP9kb380bdgRpA2boQzz8Yk
pJB9bem/lgh/tVX+bCext3zPA/Tf6+VTuhl2zrQ/KIBVMuvEahK3yGpeYP6F
1Mirj9GP/7m9ZgPcukrZ10P2Z3FotsCX6coaLen/xa/hLuyzW7x9muyLGjo2
2QKreIlX54Xwj9dNlSS+ZNSPgdXvv/fsKDw223EwB9aPnLhRDv9bFbKUB/tz
le5m+M99HWEPBbi/CsPgI7hC8WGnEWaSvggl9zUquqvVEBtvPmmCr083u+Qk
P9h1cA3mYaslLhGJ6/9aVAeXiJJ8OLB+NmVIiH2wzgbVUlgdcfqjNtgxJMla
Rv7PRnwXjf2aRX3vBsNRv52uZuCfIwqnImF3+H871oWi7hVL6k5YWxXe8jH8
+fbFw3nkffbism/hZutGPx3M/ZuXPgtL8wetDGyKzfn9pXU09U5Fl9kOO7sO
WxYQVx/6oMdN6jmb7RNwgXD9pcVkHyUBjhuwci7nQhDZpy62TQ33Buo2vULi
kVk+PrBk5L56LdxwrDShEf1Gx3oq/GDuGdkuMk+3MKj4Keo3NC1LHsC8Y/kx
2mHyvkczTfvhQAF7uB62lYfmsNiXS/L8k2yYynLnNsBjnb07yD652vKpVJj+
Pqx+ko++jF37NsBm7+lbl+GoCXEKB9ZIZXeTYfos5+gqWDun3uoL+wuakqLg
Ah37xmAY7kNaFZMJm3zvz1TB/leW1LbBfSsf85Nh25Gfxlagv+xh3yN8mLlz
O+MU7PTg8JBPPjYh9P+oBpxX
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.120298252215608`, 5.548120193592922}, {0, 1}], LineBox[CompressedData["
1:eJwdVgs0lekalhqdXMO+/XvvCSPa9u3/SyqVPCSXbiT3EJtUioxECYmikpB0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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.669701747784393`, 7.611879806407077}, {0, -1}], LineBox[CompressedData["
1:eJwdVnk01AsbVh+GMWN+M1OGNlIy+0wSQnn4RJQWiny2i7QqGhS3jaRra6Ed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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.330298252215607`, 2.3881201935929237`}, {0, 1}], 
         {PointSize[0.04], PointBox[{8., 18.}], PointBox[{9., 7.}], 
          PointBox[{16.7, 4.2}], PointBox[{12.3, 5.8}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["18",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl0gs0lGkYB/A3FRIWiSGXaaPcKjoq9+9zbc5WdtCirZh2FckRbYXWaoRw
Ukm6qEwTVjqk6YZEpoylk0KULrKW7Eg5mZJrjv2/Z+ecmff8zvu8z/d8/3cW
/7LHf4cSIWQvvnQl6vSHy5I5dF3Fklb5jkNxpiw50B7TPW3HElXZPa8xE5ZI
82SlE3B27IrUI7CPm/UFBRwjUsRawX8p14a+o/XTJmf6jVkSXZMU1AZLBJPx
lfBtU764iu57Rd28CO910tK9AAcfuPnzCdjDOM7mILVN5ZN0ONGA9ywI7uls
vfIHbFhnZbIGrvLxeBMBN2UELNCn88gvVXvA+7oUW7/ZsiQ+qfy5MszZMa0n
h8lRu6xbRiwpKki92gmXHL/h+QNcE+d07wmsWrHOsWUR3i9nxqUJDk4I/8qD
I80Gsxrg4c2y3+oM6bzeZfWwLYdd5AznXD2WLIUdNpKSegPk0NJvXA0PuOVa
h8LDncrT5fT8G0cvbdgiKmqtCG6KtvvzNQf5rmHE6XQ+/zMzVbCeiPt4Nxzh
U+17BbYa0Z9YD1fFmieJ4IFB/U+WsHSwMeAMfJ4ff08FztA2FB6BJTJ539OV
6O+rdCMcjj8qf3IarnKfSrGHebk6/+6EeVW2DoP6yGGyJcsTHlDLnpcJS2+5
F1nDUrmKpS4cn26aawwLbcb9T+qhv2tckh6tf71fWxUedksW6cD8APlD4ULc
7xdegwY87rgqbTbMOarxbjbdL2Msz+qyJOxjQo5iBUsEJ9L9WPjckq2HOmBh
uOwlgZser68thcc9RLmdC1iipQj6Jw6OX+OYKIVVkxo+OcHSx6XL78CsyhaD
8eXof8TkwTV43Pul3y24tW3D08swd9L5SQxc5Zt/OBPu2VSzzx6WVvYWhsDq
lRE7ZsHCA4EhxjBHxSbzlQ3WupX1Mh3ktDspqBZu0kuUB8K5GSMNpbBwreul
l9rI4fvSyAI4XuPWWz4cLzFqzofJ0L6WRi2WeLlqdeRR12s2e8PvFI2BOfT8
arPUou+Q18OxiymwbVTq4X5N/H/vSrojYb6ha9ommPSHXPKBBV+5WQoNlpR9
Hu3Wg8UOPUUVcJZo3OCtNfrNr7t+Di7J9OZchMVux1oy4KiEpSeDYLI/4KcD
cETYdskCuPXt8gE+rR9xz+qwwn6AcoEaLORGiPLh1r4f3+Sp471lHfxYeHjJ
tsjJ+SyxKdIs58PiMs3lK+CS4BtuzrBEX+uxixryPvuKs5J6QFS1ah761saY
WsLZfafOm6oiT0OrXeawUHWQ1VVhycewvgemtL+z+CBXGXXdktW6sFZbq5rf
XNQ355+ascR9xZ5QKp/Dkq0OLp1/w2JfToUj/M5SsPs23W9/tGd4Ns717XRN
ggV2XYNNcFhgloiBe+ZMpd2Bb+t4j01ZwPOWKBfDA+9TAithSW/y9UyYp574
YT/MtZuI9oNzJ9f+7gxnL5Q2K5RQXyjvU4PJ9TvhYbBNqNL9/mW4r2fbm0pm
IY+XScua4ZghQYyMoN/Y57oaWDp6kJc6w5AyNaMXt2Gxf7Lm9mmG8NYNSSXU
94XR7BRDRroKgstovenoRs8Jhlh02gsLqfPUv6WNMUTGC7+WQ+3qmGk0yhBd
nb5tcbAgOWel1leGnOtZHcmnnle4OGSEIVGvZqxNYG64uXzsC/p/Cd3buxTe
XObcAGu1G/uIYfHdbXZXYIHnhrQtMPtc2TYRNjPf+isHFqQFbzCDOxyWdb82
R34v7oSnfWZIiZPcsRiWlltmH1cwpFVqcToRJo69tk7DDGHVJQUhsFB+/LTx
EEMkXSE566mNS5xc3jPEKFffxh1m275VzO1nSNcuMwMXuh9sFFvdw5AeWbG/
A3W4YsrpFebpflRsR60//qGzjSHNLYc9zejz2i+zmxoZIk4ettegZic8xPcw
r61K46AZLEy5+qiMIV6NA7U11Jy8IX4eQ1RNCralUseEJagno7/6RtaLmn6E
rv+vZux/4Q+Z0Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.491215510839148, 13.508784489160853`}, {-1, 1}], 
         LineBox[{{20.000000000002046`, 17.}, {8.000000000002501, 18.}}], 
         PolygonBox[{{14.597927454946928`, 17.450172712087756`}, {
          13.435290736994567`, 17.948445591210195`}, {13.368854353111576`, 
          17.151208984614293`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.128720493773297`, 18.439845925279563`}, {0, -1}], 
         LineBox[{{20.000000000003638`, 10.}, {16.700000000001637`, 
          9.100000000002183}}], 
         PolygonBox[{{17.771141707257364`, 9.392129556524736}, {
          18.823611330425795`, 10.093775971970357`}, {19.034105255059483`, 
          9.321964914980173}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.942168021022233`, 10.440583922918485`}, {0, -1}], LineBox[CompressedData["

1:eJwd13k8lGsbB/AnW0JjSQgxZc061mRriDpRSKUzddJky5btaLNkouN1skRR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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.029701747784392`, 5.851879806407078}, {0, -1}], 
         LineBox[{{8.000000000003638, 18.00000000000705}, {9.000000000003638, 
          7.000000000011141}}], 
         PolygonBox[{{8.445678552374488, 13.097535923880622`}, {
          8.15596416503843, 11.86624977770237}, {8.952678730212593, 
          11.938678374536385`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{7.561165529975058, 12.359669593634095`}, {1, 0}], 
         LineBox[{{9.00000000000091, 7.}, {12.299999999997453`, 
          7.899999999997817}}], 
         PolygonBox[{{10.071141707257361`, 7.292129556524735}, {
          11.334105255059484`, 7.221964914980172}, {11.123611330425796`, 
          7.993775971970359}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.057831978977768`, 6.559416077081516}, {0, 1}], 
         CircleBox[{14.635, 8.005}, 2.3373596214532317`, {
          0.48756649149871567`, 3.1865302603929266`}], 
         PolygonBox[{{13.441141707257362`, 10.102129556524735`}, {
          14.493611330425797`, 10.803775971970357`}, {14.704105255059485`, 
          10.031964914980172`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.612168021022233`, 11.150583922918486`},
           {0, -1}], LineBox[CompressedData["
1:eJwlUwk0lAsbjqFFuijEzDffN4SYGd9MIRXpsRMqy6CucMmeLF2hkRRpnxAt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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.387831978977768`, 5.849416077081515}, {0, 1}], 
         {PointSize[0.04], PointBox[{8., 18.}], PointBox[{16.7, 9.1}], 
          PointBox[{9., 7.}], PointBox[{12.3, 7.9}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["19",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwt1glUU1caB/CA1AZkU3YMlk1UNqmgUAUaaQxYEaKURQ2FkcUQ1rIoaBQQ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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.885562118843275`, 6.925817748240936}, {-1, 0}], 
         LineBox[{{9.999999999996698, 10.}, {12.399999999993506`, 10.}}], 
         PolygonBox[{{11.799999999999999`, 10.}, {10.6, 9.6}, {10.6, 10.4}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.2, 9.0548}, {0, 1}], 
         LineBox[{{17.999999999998405`, 10.}, {15.599999999997154`, 10.}}], 
         PolygonBox[{{17.400000000000002`, 10.}, {16.2, 10.4}, {16.2, 9.6}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.8, 10.9452}, {0, -1}], 
         CircleBox[{14., 10.36}, 1.6399999999999995`, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{14.6, 8.72}, {13.4, 8.32}, {13.4, 9.120000000000001}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 7.774800000000001}, {0, 1}], 
         LineBox[CompressedData["
1:eJw9lA1UjXccx2+pjg0phVReyvM83fu83FtMc1B9s9QMCautNxUnpm7slmrb
XVkOZ0rKcdtZHEJL5mVSYVSSG02GiinphV6Xbu9LO/S2y27/3zn/8z/Pec7/
7fv7fj82W3ZtDNUXiUS+2vFuXvy+ZBC9LwouczNtd/nIsKNs5/7VayiUyj/Z
/GaLDEc0ccYXTlMYmhll2uItg1+jmV7VMIWRM3nDG2UyyL851/jCi4ZKmVfw
ZECK4tUnHsxKp3HW4A+z8mwplPVnD8fdp2EhBN8Rb5KiNrrplXUrjcHObVMb
xwWM7nuavVP7LbL0EvnnC5BUZAfpP6Lh/axl8KZCgFdTwtDRMzTWYFFC7goB
Jg/TLOYqaFCFByu7zASoDdKUlx1pbFxbX2cyxON6eVip6QiFDQaHolVNPL6Q
Fxo7lFGwWlQ7FvWUB1t0MLYvncKB28YLUqp47FFveP0mhkJxRI0qvppHvtHm
kpwQCs/FubXSNh6Kv3y8PvWlEJy2LnrJGI/fy119bAIpKGbzT1xsBLTGJa1w
llPIc6y2v7RWQFrzD3sf7KdwIuuXlYo4AZG7veyV5yi06bcYUjkCIsI9/JU1
FOSa1HHvegFhRd3+ViY0xrJCsvW1ekx+jOMzNtEw9/x3mZOpFEYFTaFXM2g4
t8+5bm4sRX2u43hALw3rkI5lxwa0629nfrfehcGqEdXWR8UChqJHkj//kUFv
tP29+TECJu1xs9LcYmCf+MpcNl9AcLOso6KZQX9wSMwdNY9Q8yuVX3Yx6L7R
72gfxqPM0M5CWs/A6KQm6tBcHgWlRy3FVxlQCZXV25s5WDYcurjzewbXg650
WxRy6OlIaV++lMEBz4jqzmwOBbYxi6/10PhMpXpelcnBVuE+vFjbv4E0dXrU
eQ5pGUvsxoNoHFzg9pXjDQ4Os+r6Y21pzHNt0Gf+5LCjb1OOXy+F/TbuU5Ib
OARe/vZS3z0Kx9qsx+x6OZSlTE40yaV0vuUhCUhVv/6VwjbLFr1nJjz5r2y8
MRhly5P1zcW2o7Uf8WT/6BeFP5e787DXnW8c7rQ92Zsn97sml9PSQJ7cPzb3
iGSmL0/e53FqzYy/3Xjy/rp4mSRdq9eEPm+DzZMjtXpN6Gfmblxe8RNH9J1+
MiNxuRNH9K+MmaWpfMmS/pgmVXz8MpEl/SuYJD/u4MqS/rZmzL4bOo0l/VfP
affQ9EuIPzS5k++qeyQw1PlH3z/Dd1TEEn+pzL8WnV/IEv9ZjVq/rfNiiT+3
+gfcTNnLEv96lksqNuSzxN+pHU/4kkaW+P9UeF7SfT2O5IOp8WSqLDmSn5VH
xPIECYc8Xb5inftfKKQcyV/bQ0YvmeVIPnOGI5Pk8zj46PJ7KX8h1fvBu/3/
z3eD226ZWReLEl3+r956GO9ZymK6jg8X7KynrExhsV7HD732nPGstSzEOr5M
nXn0HxutHhP86Xh8cnnNRQnhk9/w0i5HXwnhV8S+dav4aRLCNwO/dQa/VYoR
oePfh53JPdbnxYSPTqqi00YnxFDp+FnSHNhw+JyY8PXB+xJDT8ff/wBtQWLi

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 12.225200000000001`}, {0, -1}], 
         {PointSize[0.04], PointBox[{10., 10.}], PointBox[{18., 10.}], 
          PointBox[{12.4, 10.}], PointBox[{15.6, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["20",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwd0WtIk1EYB/CTgs05ddmW81LN2ypZMtTSpPDF6EMasdwgs6g3MTfqg2SF
pQsxoSZGzNQw29DVCMVKwUojlUWho9nWRaeTrJUT5l0m2aR96H88cHj4cS7P
eZ4TV1SafyGAECLFpJEUl2CkMURNo4AhmZuyU27BPKlaFQfrOh1/W2BlzrvX
1q0MYbUL3i6YHTbkqmCxoaTMDLu/HTSuRTDk8wud1wqbU2Y7NLDF9n37143z
Ux3+LQyRd6zbqf3/eDsqYBE3PtIGj/5ZaPHxEcsfpn2An7nYzkq4+Elh0yvq
uYDuIHhEFDrxGCaJM1UHwuFa1cBdmC95IKoOQx0tqZKrsHix/p4nFOtDUm4h
vCLjiK/AffnRpdmwb493KB5mxxTnkuh9o+MSH48hWn2yiwd7xMu3vfBoW6Vn
LRX3GT9OhGN/osBy0g3LFKesx2DB26SlMZj5dGfWBKvLTxwZgV3jqvcReI9l
OcE6DDc/LxLrYFPFowELfN35NHMb3n/G0brbDrNC86oBfpnXODMJy30hQj7q
7U7p18/DyvvVq7mwdrrYQmi/E24qNbBpf9+bKFqPwXHcCEvruK37YNnp6Uu9
cGzPxToFrNNnDPbDlva9VWWw/Py11S5YnLfrcj2s/1LD6mh//VkF9H9XggY3
n4U1jvl0+p8FPT+bo2j+mMZoN10PZMe0eK/ftpSxTvM5j06ZUZ9fa/oRnI48
kt+HXOgHZ7ipVgjLeg/bF9BPV0P7rxiYdOZELoZg/85FcyysXk8O9XBxLsuf
JoILJkuc7mDsC6y5EQa3S9iGOQ5DNgbMoZHD/Afu8QPs
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{20.585562118843278`, 4.475817748240935}, {-1, 0}], 
         LineBox[{{9.999999999990212, 10.}, {17.99999999998223, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 9.6}, {13.4, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 9.0548}, {0, 1}], 
         LineBox[CompressedData["
1:eJwV0n1I00EYB/BfbYy0LDXBUMuZpmws3Npoy/7YL52mIs3ENJyp+UZoaC9r
RhlEbr4Umn+UTQtb5RttpoUlSYgrB3NtlallY4XUzKj+MNemydS+d3AcH+6O
57nnuYjCyoyS9RRF8TDJSmlbMfbQVDNZw2lqdHVk/BL8d+KW9gt8KpOzoQS2
5N2NCmHTlK5qq186LAt7nJMGOwpbG6VwWMvxPjUsOlzpJ4LbT2zUG2GJIPLr
bvhNXoB+XQRN9Wex+MTMK//m42Gn+/MPAeyraXhSA/PqS6374a7rcu5L+E5u
Z1kq3HN1tdsL8+cWYo7B3mcsnWAn4qm08WfhMyr783xYW2L7dA0eUCscGpi/
b8LdCZtdK+IHcHLvbPMIXH0u4dAgrOMadHZ4wTfVY4SVRVPhLnhSLMt7BfPy
1cE+Qpoayrb5vICnf4l7QuE4PcdmgH/L7AFc+OGmdlMLiV//aEgIL/lEyy/C
En+HSAJrM+cWFXCupjheDLelfDgdB3dwCj7yyf3lLTXbSH4D5SG7YAfjgnsR
700fTygKgi17pbV2WBl2W7yG/PwrjFNGOHfsYNJ3eLn/xtE+eFLpLrDAWRmb
UzvgmbSbSgM8LHDq7sGi7PeXG+FieSKrh+wPTDdVwHHBw28HSX0Zpm7S3yYv
2/sOVnOCXpN+JkU3C13weWf5bCip/2q3ZTvyZQoCzCxY19BrlsNUlSLCI6Cp
HFXdfB28JClT/YQlgbHWUZh9MnuHk/hI7BAzEudnAqXf4PsHnnoTYZGJYZ2D
S1O6xmph5kp17R842ROjMMHeNQyYRT5vFP0fl0kBaQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.414437881156726`, 8.524182251759061}, {1, 0}], 
         CircleBox[{18.685000000000002`, 6.41}, 1.492422527302506, {
          5.212003081737592, 7.910966850631803}], 
         PolygonBox[{{20.28483267673843, 6.243085631415506}, {
          20.3397769023179, 7.506802819743446}, {19.570557744205246`, 
          7.2870259174255425`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{21.005562118843276`, 7.245817748240937}, {-1, 0}], 
         CircleBox[{19.314999999999998`, 6.59}, 1.492422527302506, {
          2.0704104281477993`, 4.76937419704201}], 
         PolygonBox[{{17.71516732326157, 6.756914368584494}, {
          17.6602230976821, 5.493197180256555}, {18.429442255794754`, 
          5.712974082574457}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.994437881156724`, 5.7541822517590635`}, {1, 0}], 
         {PointSize[0.04], PointBox[{10., 10.}], PointBox[{18., 10.}], 
          PointBox[{19.400000000000002`, 5.1000000000000005`}], 
          PointBox[{18.6, 7.9}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["21",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwd0WtIk1EYB/CTgs05ddmW81LN2ypZMtTSpPDF6EMasdwgs6g3MTfqg2SF
pQsxoSZGzNQw29DVCMVKwUojlUWho9nWRaeTrJUT5l0m2aR96H88cHj4cS7P
eZ4TV1SafyGAECLFpJEUl2CkMURNo4AhmZuyU27BPKlaFQfrOh1/W2BlzrvX
1q0MYbUL3i6YHTbkqmCxoaTMDLu/HTSuRTDk8wud1wqbU2Y7NLDF9n37143z
Ux3+LQyRd6zbqf3/eDsqYBE3PtIGj/5ZaPHxEcsfpn2An7nYzkq4+Elh0yvq
uYDuIHhEFDrxGCaJM1UHwuFa1cBdmC95IKoOQx0tqZKrsHix/p4nFOtDUm4h
vCLjiK/AffnRpdmwb493KB5mxxTnkuh9o+MSH48hWn2yiwd7xMu3vfBoW6Vn
LRX3GT9OhGN/osBy0g3LFKesx2DB26SlMZj5dGfWBKvLTxwZgV3jqvcReI9l
OcE6DDc/LxLrYFPFowELfN35NHMb3n/G0brbDrNC86oBfpnXODMJy30hQj7q
7U7p18/DyvvVq7mwdrrYQmi/E24qNbBpf9+bKFqPwXHcCEvruK37YNnp6Uu9
cGzPxToFrNNnDPbDlva9VWWw/Py11S5YnLfrcj2s/1LD6mh//VkF9H9XggY3
n4U1jvl0+p8FPT+bo2j+mMZoN10PZMe0eK/ftpSxTvM5j06ZUZ9fa/oRnI48
kt+HXOgHZ7ipVgjLeg/bF9BPV0P7rxiYdOZELoZg/85FcyysXk8O9XBxLsuf
JoILJkuc7mDsC6y5EQa3S9iGOQ5DNgbMoZHD/Afu8QPs
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{20.585562118843278`, 4.475817748240935}, {-1, 0}], 
         LineBox[{{9.999999999990212, 10.}, {17.99999999998223, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 9.6}, {13.4, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 9.0548}, {0, 1}], 
         LineBox[CompressedData["
1:eJwV0n1I00EYB/BfbYy0LDXBUMuZpmws3Npoy/7YL52mIs3ENJyp+UZoaC9r
RhlEbr4Umn+UTQtb5RttpoUlSYgrB3NtlallY4XUzKj+MNemydS+d3AcH+6O
57nnuYjCyoyS9RRF8TDJSmlbMfbQVDNZw2lqdHVk/BL8d+KW9gt8KpOzoQS2
5N2NCmHTlK5qq186LAt7nJMGOwpbG6VwWMvxPjUsOlzpJ4LbT2zUG2GJIPLr
bvhNXoB+XQRN9Wex+MTMK//m42Gn+/MPAeyraXhSA/PqS6374a7rcu5L+E5u
Z1kq3HN1tdsL8+cWYo7B3mcsnWAn4qm08WfhMyr783xYW2L7dA0eUCscGpi/
b8LdCZtdK+IHcHLvbPMIXH0u4dAgrOMadHZ4wTfVY4SVRVPhLnhSLMt7BfPy
1cE+Qpoayrb5vICnf4l7QuE4PcdmgH/L7AFc+OGmdlMLiV//aEgIL/lEyy/C
En+HSAJrM+cWFXCupjheDLelfDgdB3dwCj7yyf3lLTXbSH4D5SG7YAfjgnsR
700fTygKgi17pbV2WBl2W7yG/PwrjFNGOHfsYNJ3eLn/xtE+eFLpLrDAWRmb
UzvgmbSbSgM8LHDq7sGi7PeXG+FieSKrh+wPTDdVwHHBw28HSX0Zpm7S3yYv
2/sOVnOCXpN+JkU3C13weWf5bCip/2q3ZTvyZQoCzCxY19BrlsNUlSLCI6Cp
HFXdfB28JClT/YQlgbHWUZh9MnuHk/hI7BAzEudnAqXf4PsHnnoTYZGJYZ2D
S1O6xmph5kp17R842ROjMMHeNQyYRT5vFP0fl0kBaQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.414437881156726`, 8.524182251759061}, {1, 0}], 
         CircleBox[{18.685000000000002`, 6.41}, 1.492422527302506, {
          5.212003081737592, 7.910966850631803}], 
         PolygonBox[{{20.28483267673843, 6.243085631415506}, {
          20.3397769023179, 7.506802819743446}, {19.570557744205246`, 
          7.2870259174255425`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{21.005562118843276`, 7.245817748240937}, {-1, 0}], 
         CircleBox[{19.314999999999998`, 6.59}, 1.492422527302506, {
          2.0704104281477993`, 4.76937419704201}], 
         PolygonBox[{{17.71516732326157, 6.756914368584494}, {
          17.6602230976821, 5.493197180256555}, {18.429442255794754`, 
          5.712974082574457}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.994437881156724`, 5.7541822517590635`}, {1, 0}], 
         {PointSize[0.04], PointBox[{10., 10.}], PointBox[{18., 10.}], 
          PointBox[{19.400000000000002`, 5.1000000000000005`}], 
          PointBox[{18.6, 7.9}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["22",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwd0WtIk1EYB/CTgs05ddmW81LN2ypZMtTSpPDF6EMasdwgs6g3MTfqg2SF
pQsxoSZGzNQw29DVCMVKwUojlUWho9nWRaeTrJUT5l0m2aR96H88cHj4cS7P
eZ4TV1SafyGAECLFpJEUl2CkMURNo4AhmZuyU27BPKlaFQfrOh1/W2BlzrvX
1q0MYbUL3i6YHTbkqmCxoaTMDLu/HTSuRTDk8wud1wqbU2Y7NLDF9n37143z
Ux3+LQyRd6zbqf3/eDsqYBE3PtIGj/5ZaPHxEcsfpn2An7nYzkq4+Elh0yvq
uYDuIHhEFDrxGCaJM1UHwuFa1cBdmC95IKoOQx0tqZKrsHix/p4nFOtDUm4h
vCLjiK/AffnRpdmwb493KB5mxxTnkuh9o+MSH48hWn2yiwd7xMu3vfBoW6Vn
LRX3GT9OhGN/osBy0g3LFKesx2DB26SlMZj5dGfWBKvLTxwZgV3jqvcReI9l
OcE6DDc/LxLrYFPFowELfN35NHMb3n/G0brbDrNC86oBfpnXODMJy30hQj7q
7U7p18/DyvvVq7mwdrrYQmi/E24qNbBpf9+bKFqPwXHcCEvruK37YNnp6Uu9
cGzPxToFrNNnDPbDlva9VWWw/Py11S5YnLfrcj2s/1LD6mh//VkF9H9XggY3
n4U1jvl0+p8FPT+bo2j+mMZoN10PZMe0eK/ftpSxTvM5j06ZUZ9fa/oRnI48
kt+HXOgHZ7ipVgjLeg/bF9BPV0P7rxiYdOZELoZg/85FcyysXk8O9XBxLsuf
JoILJkuc7mDsC6y5EQa3S9iGOQ5DNgbMoZHD/Afu8QPs
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{20.585562118843278`, 4.475817748240935}, {-1, 0}], 
         LineBox[{{9.999999999990212, 10.}, {17.99999999998223, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 9.6}, {13.4, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 9.0548}, {0, 1}], 
         LineBox[CompressedData["
1:eJwV0n1I00EYB/BfbYy0LDXBUMuZpmws3Npoy/7YL52mIs3ENJyp+UZoaC9r
RhlEbr4Umn+UTQtb5RttpoUlSYgrB3NtlallY4XUzKj+MNemydS+d3AcH+6O
57nnuYjCyoyS9RRF8TDJSmlbMfbQVDNZw2lqdHVk/BL8d+KW9gt8KpOzoQS2
5N2NCmHTlK5qq186LAt7nJMGOwpbG6VwWMvxPjUsOlzpJ4LbT2zUG2GJIPLr
bvhNXoB+XQRN9Wex+MTMK//m42Gn+/MPAeyraXhSA/PqS6374a7rcu5L+E5u
Z1kq3HN1tdsL8+cWYo7B3mcsnWAn4qm08WfhMyr783xYW2L7dA0eUCscGpi/
b8LdCZtdK+IHcHLvbPMIXH0u4dAgrOMadHZ4wTfVY4SVRVPhLnhSLMt7BfPy
1cE+Qpoayrb5vICnf4l7QuE4PcdmgH/L7AFc+OGmdlMLiV//aEgIL/lEyy/C
En+HSAJrM+cWFXCupjheDLelfDgdB3dwCj7yyf3lLTXbSH4D5SG7YAfjgnsR
700fTygKgi17pbV2WBl2W7yG/PwrjFNGOHfsYNJ3eLn/xtE+eFLpLrDAWRmb
UzvgmbSbSgM8LHDq7sGi7PeXG+FieSKrh+wPTDdVwHHBw28HSX0Zpm7S3yYv
2/sOVnOCXpN+JkU3C13weWf5bCip/2q3ZTvyZQoCzCxY19BrlsNUlSLCI6Cp
HFXdfB28JClT/YQlgbHWUZh9MnuHk/hI7BAzEudnAqXf4PsHnnoTYZGJYZ2D
S1O6xmph5kp17R842ROjMMHeNQyYRT5vFP0fl0kBaQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.414437881156726`, 8.524182251759061}, {1, 0}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{18.685000000000002`, 6.41}, 1.492422527302506, {
           5.212003081737592, 7.910966850631803}]}, 
         PolygonBox[{{20.28483267673843, 6.243085631415506}, {
          20.3397769023179, 7.506802819743446}, {19.570557744205246`, 
          7.2870259174255425`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{21.005562118843276`, 7.245817748240937}, {-1, 0}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          CircleBox[{19.314999999999998`, 6.59}, 1.492422527302506, {
           2.0704104281477993`, 4.76937419704201}]}, 
         PolygonBox[{{17.71516732326157, 6.756914368584494}, {
          17.6602230976821, 5.493197180256555}, {18.429442255794754`, 
          5.712974082574457}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.994437881156724`, 5.7541822517590635`}, {1, 0}], 
         {PointSize[0.04], PointBox[{10., 10.}], PointBox[{18., 10.}], 
          PointBox[{19.400000000000002`, 5.1000000000000005`}], 
          PointBox[{18.6, 7.9}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["23",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwd0WtIk1EYB/CTgs05ddmW81LN2ypZMtTSpPDF6EMasdwgs6g3MTfqg2SF
pQsxoSZGzNQw29DVCMVKwUojlUWho9nWRaeTrJUT5l0m2aR96H88cHj4cS7P
eZ4TV1SafyGAECLFpJEUl2CkMURNo4AhmZuyU27BPKlaFQfrOh1/W2BlzrvX
1q0MYbUL3i6YHTbkqmCxoaTMDLu/HTSuRTDk8wud1wqbU2Y7NLDF9n37143z
Ux3+LQyRd6zbqf3/eDsqYBE3PtIGj/5ZaPHxEcsfpn2An7nYzkq4+Elh0yvq
uYDuIHhEFDrxGCaJM1UHwuFa1cBdmC95IKoOQx0tqZKrsHix/p4nFOtDUm4h
vCLjiK/AffnRpdmwb493KB5mxxTnkuh9o+MSH48hWn2yiwd7xMu3vfBoW6Vn
LRX3GT9OhGN/osBy0g3LFKesx2DB26SlMZj5dGfWBKvLTxwZgV3jqvcReI9l
OcE6DDc/LxLrYFPFowELfN35NHMb3n/G0brbDrNC86oBfpnXODMJy30hQj7q
7U7p18/DyvvVq7mwdrrYQmi/E24qNbBpf9+bKFqPwXHcCEvruK37YNnp6Uu9
cGzPxToFrNNnDPbDlva9VWWw/Py11S5YnLfrcj2s/1LD6mh//VkF9H9XggY3
n4U1jvl0+p8FPT+bo2j+mMZoN10PZMe0eK/ftpSxTvM5j06ZUZ9fa/oRnI48
kt+HXOgHZ7ipVgjLeg/bF9BPV0P7rxiYdOZELoZg/85FcyysXk8O9XBxLsuf
JoILJkuc7mDsC6y5EQa3S9iGOQ5DNgbMoZHD/Afu8QPs
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{20.585562118843278`, 4.475817748240935}, {-1, 0}], 
         LineBox[{{9.999999999990212, 10.}, {17.99999999998223, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 9.6}, {13.4, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 9.0548}, {0, 1}], 
         LineBox[CompressedData["
1:eJwV0n1I00EYB/BfbYy0LDXBUMuZpmws3Npoy/7YL52mIs3ENJyp+UZoaC9r
RhlEbr4Umn+UTQtb5RttpoUlSYgrB3NtlallY4XUzKj+MNemydS+d3AcH+6O
57nnuYjCyoyS9RRF8TDJSmlbMfbQVDNZw2lqdHVk/BL8d+KW9gt8KpOzoQS2
5N2NCmHTlK5qq186LAt7nJMGOwpbG6VwWMvxPjUsOlzpJ4LbT2zUG2GJIPLr
bvhNXoB+XQRN9Wex+MTMK//m42Gn+/MPAeyraXhSA/PqS6374a7rcu5L+E5u
Z1kq3HN1tdsL8+cWYo7B3mcsnWAn4qm08WfhMyr783xYW2L7dA0eUCscGpi/
b8LdCZtdK+IHcHLvbPMIXH0u4dAgrOMadHZ4wTfVY4SVRVPhLnhSLMt7BfPy
1cE+Qpoayrb5vICnf4l7QuE4PcdmgH/L7AFc+OGmdlMLiV//aEgIL/lEyy/C
En+HSAJrM+cWFXCupjheDLelfDgdB3dwCj7yyf3lLTXbSH4D5SG7YAfjgnsR
700fTygKgi17pbV2WBl2W7yG/PwrjFNGOHfsYNJ3eLn/xtE+eFLpLrDAWRmb
UzvgmbSbSgM8LHDq7sGi7PeXG+FieSKrh+wPTDdVwHHBw28HSX0Zpm7S3yYv
2/sOVnOCXpN+JkU3C13weWf5bCip/2q3ZTvyZQoCzCxY19BrlsNUlSLCI6Cp
HFXdfB28JClT/YQlgbHWUZh9MnuHk/hI7BAzEudnAqXf4PsHnnoTYZGJYZ2D
S1O6xmph5kp17R842ROjMMHeNQyYRT5vFP0fl0kBaQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.414437881156726`, 8.524182251759061}, {1, 0}], LineBox[CompressedData["
1:eJwlU3k01XkUf6kY6Vif937v/bzFVCdRWb4/lSE+lijRosmSOpVsJYTRYjmv
RaKkHvPeDNma8qLQSCJaSKZCw4wxSZYWUqiZFp1RmPk1fc655/5x77n3fD73
c40DIr2C1Dgczlo2PufsrM+wwpcsQdrmnOTYWiuMJWqdL9GQokKxtj1m0Aqz
c3/QeWgqxR3ngdxxisBylpuCtpCipm36y+4VBM+iZUUHDaWQL43LL40hkN3O
Say6JsHU9F07veUE1X2qSc4cCT52yUfEBQQBS5z6N+0VY17RtLyP+QRjDi/P
/9wkQl+NPJrOJBgq/UA+mYrQYrwyRpRIYKaZma88bYSD6T7qJgEE5/W2eKpb
GkGpVV5m60pwYcmM2PhOGv7KPfmV8wmOfMqyqVDSqPVyU07wCSr008hkGA39
ZmbPBk22v/zjjdsbaPgow7mfOATxASLxsWAa0n/6ahwnrTBDW3GIl0FjVcNy
A3e2bn1ilyy7h0ZD46T2fnWCYN5Gr4JlRojaEucyU4+g86ZSp+2WEUSLZnU8
kxLcMU3J8HAXYTDQ5MX2xQSZkQ+Slv0qgpNBUmbVeoICgzf6N+zFcFNlbqyP
J+DlPnV/d1SMre0Jm4rPEVTqarg/PCNGsXneYnkHgetN+cNsmRjBA2VVUg6D
XreXK8QiMa7saDYvN2aw7vVW31eHRUjt84uptWYw6ZOQJrlvhCz73b5jNgxC
dOpks7WMEO/c1nJvIYPL9VP2dfjQGLrT3t/AY+BrILo1o1EI1QHdwV9GCfit
RSdP+AvRfUpx+t9WAtXO5+k8Wogiv+Q6+2KC8eyzq7LGBRh4WyupTiKwob6+
+npUgCrzH8M1gggO33UVB04IUN19ztfRg+C+T8v9iJlC3Myh3g/bEjxO00l4
IhXi9vHKU48ZgqjE1tAIOyGidbfbLmH10rYpVlQGCHH2lZlgmxNB/45MV6VS
iJM9NZ6RrH4q163dlQ+EGOl76ugfSdBlPOWRmjENifXcXocTBHuzOy+mhtLA
gjWzFl4msNh3afexAhrLo/z2N3YTOFy+sPDMVRrmG47YXVFnkGew8sO9Qhrv
e3/nylh9DBx7Uob9Wb8MXNIuXc3g7vuIlDWPhFj+vUtbeRCD/bGqPaoFQrTM
Tw/khTMQra2NDQwWYG5cWV96IINv5/wloPIpBK40/zvck4Ge7U/1KW/4mPvk
bMhSMwbPV/9ZWxfFh7+GpUmJGoMhTzcTI1M+5N7bU1d3EqwJCimIEPJxoL1J
q+Yi6/9to/JnNnzwnEZM5h8jMD2UOuETx0dnmevd9jCW3+4dXc+b+dBJPR7h
7cXyCz1ODUooTDiUeqxzICi//sZfEEZBnu2+uICweprVVQ6rKJR7cdQsLAjS
DflS1W8U7HO/uX7dmsCRFA95v6LQsyXgTJQLQYeHXkLhBIUHXcqZpf7sPLt+
R7PpAnCPBkzTjGPv67m+EVMFcK90FuzKJSh5G+ti+I7Cdx3KGssG9r+tI6/G
tVBwGiv0m/qCIMnl0UmPZArjOuuEml8xSMjI7x0QUNCXnzalJAz4zsGya8l8
RBeLOb7zGBRtNh4e+oOH6o1uA1azGYTqBpWEG/KQo3A+5KzDYK9x3pPQzYYo
N02JaRohUBQWxo82ceHUGq/tW8/u4z5+uiiMixa9t0xFBkG0XUjz2AouQoI/
g/3n/8HFfzovGzM=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{21.005562118843276`, 7.245817748240937}, {-1, 0}], LineBox[CompressedData["
1:eJwlk3s0lHkcxpG0qeyWubwz77ilUrLL+xu6SDwVraYoKiEbKXVmlnFJJDYO
posjt5Tc0im5bUrEuiVFSLZUNLvVulVEGmxl083Ots853/P88fz1fJ7zNfD2
d/ZRU1FRcVLef57+VQT/ux7MFbPubPAjON5w2lyjQg/XzWLH8q0JIl7p6dxO
1YNjZ2i8hTaBxN09VtNCD7sltRG1owziunbqpp3VxbrzlfzfnzDIGsnYnDKk
gwSX3sn8Rwxcz3ax2It1oJDu6BzoYfCyNj1qTZgA6qMXX4z/w0Dm52XA76dR
p+Ge94lL8EvNPP+yMBrxxmK7u1YE01qT1sqENMpFtf2J3gRbZsXVDM2kMbw/
++O4jEBYWmR5/RMfi+aL3zjnEmxsoOs6PvBhmJPtkVBDsLhMEeOuRuOm963V
q1sJrre5sAvn0HAsnvx87x5B0NDLbrcfaPj7lLd4thEIOmxELDca6vXOZRfr
CRxXa+X2p9IYCwlKLiwmqE8LGqB6aCx0qlpqkkpw79jJ+TxzAYauacsfBhPI
I0LUCmQCRJpYvLXbRNCZpSkVNAngX3gqpMSIoCTjtGfyCwFy39+Vb/7MgF1a
KX/WKcCfuttP3+xgwC9ZZGuTIEDW1fZvU0oZjDT+ZB2qpcxlTvqBmQyKHn/X
P+BFw/xhYOZgCoO+S7Zxmel8VP5cY+qYweBQnNtRtwc8KPgOY84lDJbNSA69
PI+HmQbfv4q6z2C4bm4oK5fCgwsHGtZNMPAevjp9iTuFqoS8SEtDZb/1W+57
rKVQ/+Dy0/vrlf0OOxwL9qTQ2/qHQX0AwREfo00haRRY+jp1uUkERl1f5n3u
ohAw1640qYjAveqMoWghDx5zOvqzrhHQJ32rxGIeEkZONTa3EPzd5CIRZfNg
FSTarXeHIDHaLC25gYdVuwZmsBoJ9PeNxuh282Bv6JpfV6bct0BIkt/wsK59
b8r7DILwpVW9mWp80LNse34MJ2ge+GvtbE0+XPPCbG64EKj7bL+yZ4py/92S
HSpmBG4neIGDz3lwPzRCLdckGDTb5h+QywNnmQltNcggb7hv5TtLHmS1psIl
Sj4brL8cNi6gsPfGsXyNJgbpozZRT95ycUA/L962lYFKQcy+HsKF6KR5u2MX
g+LZZ/LnHOQgg+9k3z3JwKCSU9Hbx8a29p0tUmOCakgX+ESxkWMSc87FlUA0
8r58gQcb1W6+561jCYa8dcpn7mejbjjc117Jz2tY6/WqG2zwMuXcU0pefkkL
w6sZDp7dfmos6CZY3hnqV1bOQWKOrOaNQskf4lHTlVystFQdi5wgyBg8X+xZ
zcUVq9+GyycJWgMSo3zmUuiZpm9XpyrE+JgkYtCPQqE45ZW6Mpdl9/UGp1NQ
mN85emGcYKvj1FUrLlCwK64Uq70kcHi9r001lUKe+KiX6CHBQfU9TRvEFCRv
HZ0aq5T/2PwuLtaIQmRgwYQ8m8BFkVM6cZcLi66r8X3RBI/Gfb3EO7gwsJMe
f+5LwHae3qbeyYF828edU70IvqnnbN1owYHGfKl2rydBdvM1+twRNr4EJp4o
lBLY67upTj5mwf+mzMs5niB6xa5o+zUs2Jl+4BhWEFjyVS9/6NDG8lq+95QB
gjNVgZ0tOdpwkFwyiaeEqLjV3uNQpI3gs7vC4myF0FIf/DVvVBt7vkoIn6/O
wr+QFjMG
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.994437881156724`, 5.7541822517590635`}, {1, 0}], 
         {PointSize[0.04], PointBox[{10., 10.}], PointBox[{18., 10.}], 
          PointBox[{19.400000000000002`, 5.1000000000000005`}], 
          PointBox[{18.6, 7.9}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["24",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 0}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {57., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {61.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {66., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {70.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["g",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {75., 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 10.9452}, {0, -1}], 
         LineBox[{{20., 17.00000000000182}, {19.399999999997817`, 
          14.899999999997817`}}], 
         PolygonBox[{{19.86483267673843, 16.526914368584492`}, {
          19.150557744205244`, 15.482974082574458`}, {19.919776902317906`, 
          15.263197180256554`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{20.585562118843278`, 15.524182251759065`}, {-1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 9.0548}, {0, 1}], 
         LineBox[CompressedData["
1:eJwt1glUU1caB/CA1AZkU3YMlk1UNqmgUAUaaQxYEaKURQ2FkcUQ1rIoaBQQ
ZJBNlmADMhoVFNoURJagIATFpRUGBIZFAYOCWIuVzYE6qfR/Zybn5LzzO+/l
vu+7/3vvidHh6P0h8hQKpRBfcqXIlvH5nE7578eUTunRu1xutIVOydhpsKPX
hE5hlCb+4Az3pzrlC+ClUNcGb9hGb+n5N7CI6XIrBJasuV2gCLO/WK8YBXeM
RsTdMqZTpHKeAcRufreKjsBCe+54KCzKfamrA7/OjV3nB+8/FKb92IhO4c/b
fPoVvNR2MSgD7rxVNL4RPq6UkLAHDt6Ya78KVraJdqLBPKOAu7+h/glBstoH
QzpFIZPr1QlfibDOn4TLUm83V8NbOhd/HIM5l/l1fJh50FswAffc9LmXDGun
BI+9hxechq2/gw8+LolSx/gR78InwuABr8SWz+EyPiuGCx8/PyT1IfX8xn4f
C7/ZPifhwW4uZhZpMHc/zUEI2yR1/60UpppquLTCEkdmthjuiMk26YcNPeN2
D8Nlotm85zDL/BLlI9zaH3pvDJ65o/HWBP3m1B4o74H744a274Y/Jpx6VA+X
s3avi4Crqw4mZ8OZkc7TWfAIVb/Lm7yv55RNORw4/pKtBT/KZLU3wqYpHG4n
+qXXretuJ/k19V89CQ+1HH7XAfuYSj6xhC1/dTKWwLzxxZCGz1Bf3bHeOpjq
18bXg8u0x7uFsINUkJS1DvbgMDLh/HP2e1RgimFMeTgcwTfwvGaAek85XXAn
v1czF/rBVMW0BUuYlvRUwQym8R3SVcn90ehBVZg/vEJ1HvNRuctebg08rX/S
agRuSbFVtII7JvPoj+FyucNb2bCDRMO7Ha6nPGsphdPHKvXb4CWpOfUlzA6L
H+6AAzeky6xRX/zXh6J64Zyewb2JMCd02H0K5hme6WuCZfFBGfKox09m/ewN
nOL8bsAYphzwT1BE/7zuPxqYsFRvQqwOc7rMwkke0sqCuT/xPKsmaH0hbJgT
0NcN27X4apD5n2l2TMyA042HTgyQfN6mRJrAhpt1/jULD3FpJVWod+G6z/BK
W8xnwwUNA1i9caVAE37tumr2NA374Dn3tD5sE3JZJl2L9fxQP04XpgY84DrB
jNo/alRhy9LiMxf1cbXMevUnqU/O0U8Jpid0G0/B5d9f98/QQ30M+/bHMOP6
P7fpwPXi3eyfyH7zj4qQ6GKdTGmJc8n6KjjzZTqsefZkBdnfneOK2UHkvvDi
/D5ivcgmf/iGxrW9X8Cai9kPY+AYBmuTGck3uDSsBJ7YxbypB8sSrlj1wcr/
CLitCS/Uftmii/f7mZdV68DxT+I+C4b72dwCMv+Oi91rauDA0owPdmQ+Y+/4
zsPs3ASaB8yvSFK0QH8OjZ7dJA+GK3uaBYvM9wrPwcEazvWBcGcefbKe9Ffk
7uwLa4b9p/cpPLGT3rwFZkmiipfJ+TfV/HAO46uf/YpjhPlU78r+TgCzqMJq
Z3hpkLfJDJbte0r1gRVoZ69cQj/UsfnQUNhBV0OoCC8dntSOhnknuhrCdJBj
/Icu4o6Satl9beQibRVxYMf0FxJTOOKpm9NB+JGmilWOFp3i/vHf4btg9+37
FuXgG+MrTljAMznHNbM18Vz8+VkVWKDmlWABy3bY/v13st95FU8mNdC3/lXl
bvjGTF1aM5yTneFVS/LXVTkqglNmRrW+J3k03zQVw4Gre8ZSYbvFlNoheHrD
1PtYcp6P3lVSwfid/Gd7yf7mBb418oQt8/qOEVM7hZtLYKrJvQtxJC+1ALlJ
OOcTr6NpJN/RF+6WqL/fjcsuhV9/ECuEwdOpFgwxyW/kalUxTOFN9w6RemwF
/J/g9I6IaLJ+WcXuW4nznXx8TdAv+5SdSyE8celVoxtM+3bP4CFYoSWtPZzk
NaZ9WgmeGXv3cxY8HeO4XYh6BHVDfuWw6S/KroawULmALibj5XE0zqHfFtVp
0V14QuQ5N7sGV+dg6wfwSMuOQQ+Y7bLU1k72X5TSStFqjHMg6mEDzBmpblWD
dZc/qF0l+Xvo/ZqsjnqaCtvI+6mN7fflYGn/Udco2O1HBQ9bNezfbV0sT5h+
aDayWBXzv8/t2maSf+p8iRW8sEGZthpOia2qkqlgfio0j70n58XJbwspuD/i
yPEeI/nK55y2g6UyH90u4rlq4wK48sybJ3fJ81LqkjJ5X2LKNnK+Z96hTZTB
5QOsovtwpdJaf33Up14W8EsvOb+X82eDYAlN/RtyXpTRdq6ogA2nGEx5sn4r
wocGYbcYUaAxnPOi4+giXOlQbMCEM3mRr5ZhxrYjD0ge+cwDR36HN+6qySog
+Sh5n2+HX0syBGT+BNPi5CTYRllt6wBZ7yfuXdSBe6rEfbOwRD6uyAn15jdp
WXxqh3MiN0THDP0Jv7ZjasH5I9Ejc8rIKzm+YC0888Mst2YV9rkc64UeHPNz
6HqOEuryLU5ShzcyWy9vUqT/7/+Q7f//D1HpfwGyzEhM
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.44137111780815, 5.881877462230899}, \
{1, 0}], LineBox[{{9.999999999990212, 10.}, {17.99999999998223, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 9.6}, {13.4, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 10.9452}, {0, -1}], 
         LineBox[{{18., 9.999999999997272}, {18.599999999998545`, 
          12.099999999992178`}}], 
         PolygonBox[{{18.46483267673843, 11.626914368584494`}, {
          18.5197769023179, 10.363197180256556`}, {17.750557744205246`, 
          10.58297408257446}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.414437881156726`, 11.475817748240937`}, {1, 0}], 
         CircleBox[{18.685000000000002`, 13.59}, 
          1.4924225273025065`, {-1.6277815434522172`, 1.0711822254419938`}], 
         PolygonBox[{{20.28483267673843, 13.756914368584493`}, {
          20.3397769023179, 12.493197180256555`}, {19.570557744205246`, 
          12.712974082574458`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{21.005562118843276`, 12.754182251759064`}, {-1, 0}], LineBox[CompressedData["

1:eJwdk3s0lWkYxXWjIlpTSlLRSTnf9x35XqREduWMS1EoSSacRO6XUqkJXazQ
KTLp4lImOSSmcimFM8jIpdxjMkaawlRGoVySzFd7rXc9f+w/3vXs57c1RP62
eyfLyMjYcO/bvPJdBLrfpQW9vlk1m30J1lzN7fqyTAu/65zqT19HsGFmQGCi
vhasnx0S688hqG5sLD6yVwtuXkU/F31g8XB2XUDebS2YpxaoPv2LBYkkLfdm
8nHO/uVEeguL5tuxwcFefPT57X7W08lCrJFrsvsRH1M/ZHUNDbMwrtryfkCB
glTWUfJlPsH6677myzdSEFOewlojgi8J952M3SnkWxZ1x4gI7EcNr0eHUOgN
Th4biiDolmSKUo5T0NL0HLS9QeCweCAl/gQF3rVkp3OFBN6S6dLGUAploj82
bKgm+PXKtuBbBylYZ0+M19URPI6Wn920j4L/3vxK5ycEA8qdbcF2FKaU2OZm
lRA87xjJFutT6D8YdP5mNoF63o/iQSXuP5sHBswFApW1+r6DXXy8LZ7T2nSA
4EJp+Ve6mI8wRv+jcCtB/N3Plf1JfPjfvHjwzgqCjBzdrQZn+bgxUttqN87i
k4Q3VzGej+eLd10ua2YxuOLdT50FfCTl1SvF5bCIXNTgUj7C+RE26oGJLDLj
hhXEWyjoNQUmvoljEZPxtyHJpVDgXbjSOoFFqWHT2qZ5NPpUrfpt77BwO2Jg
muFLQ0FD8C68gUV0il/5p2wajWmHH1mMsqiL0hx73Uij4JwkzJDH7de2/VJD
O42SxtvtDZsILhvtLBTW03hZ/adGSQCBkVVvq2kejbnqi6Q3Ygl6Y8N9j8fS
CFgqzInNJLBL1elr86bh9ENzd1IxQVC+BW/TJhpn318sf1xJcDqpudqRpWEU
ZOm2pIbggdRDz2AxjfV7euTnlhMEVMhJA2bTMOc5pEtzufvS3e4GsjQs6j3i
RhI4Ht14SjlfKSycZdppdpRATeG80/BnCg6SEJNSe4JQK9v0zZzPc/PaLaND
sG3VfrMzs2g4hr5XWTOTQOyxrfKxgIbyamah0RsW2VGKrhEiGhFFK3VXcfm0
dYYXvb1Fw6M0Kl22gkWeXKdJ3DQGh9QlYtNqFjLMuooYdwaW8Xr11h0s+MM6
WUNlDBJUbcxfTLD4kPpxWYyiADvqXSv9KILUF5U+T00EuMqcvG7vQLArckpa
r40AD3f6pK47RXDGOjAtRCiAtPeojzmXX33U4RnOSwRYkNg6/yKXV/V977Kl
PQxeVbVTai8I3JRNLZanMYi5FlE42Efwn3ON3D5XBsaGk/rDRrk+y2vz65cy
uGt0vzd/giDRrMDkaR+NTjl1oXSSLqadrl4vX0UjwzPu3VTON76rGCPM5XjR
q4lMGyJocTjmLOL4EGYXeE7+l8BxRsWCiXwaEs9IF8smgkfjT5J21tDw+mht
U/6A66NaxEPjXhqhgRmjrckEz6bfWX5ShYF+R574nxMc/6eqttRsZaAh9Dv7
2odgyH7syrF4Bi07xlynuXD9vnZgu0MXA1lNvzkvnQmImsaAtpEA44Exv9z0
I2j3twsLjxPAvyzCxVZMYCKqle7qEEC48vM83j2CV0FhGSeUtbG6SFU0pYeA
5ZnFL9TRhpXXb4xYRReayZeUnlPa2J+yJyTaVBejx+Qzs+S04f5duqh98k0C
/A93gDxA
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.994437881156724`, 14.245817748240936`}, {1, 0}], 
         {PointSize[0.04], PointBox[{10., 10.}], 
          PointBox[{19.400000000000002`, 14.9}], PointBox[{18., 10.}], 
          PointBox[{18.6, 12.1}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["25",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "Automatic", ")"}]],
       Annotation[#, 
        FeynArts`VFlip[Automatic], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGCQB2IQvcICSDBwODCAgYrDjEsX5+gZszswROzg+W+v4iCT
N5s1Po3NgSHhluiNPBUHic/JeybOZHVg+NEY8bdexcHn1Ly+Y2dYHBg8TJqv
AeVL2M12/vnP7MBgkf38N1C/1EdXY2NjID+Cy+P3f2WHcywTWDPTmID2Fez8
tV/ZweTH7klzZzI6MKz45a8wUdmhpqd+94UzDA4ME/rO/axXdkhl7bgGdKcD
Q8OyNnmgfNTXqUD9/+wZNE6o/QTqz1p5/8yZM3/sGTr43/wAmj87a+tjQ+Pf
9gwCcod+AO0v8uAsnzHzpz0Dx9Qpl4DuC3oE8ucPewaDCQt+AN2/2fl7R3ra
N7g8V5SPzLkzX+D69zxSALr/M9x83/PZv6bN/Ai3fzfTw6Vf/r+3h7kvIZxV
aO+ZN/Yw99eZC27tnPnSHuY/sbLDL6PTntnD/P88+KWRsfFje1j4VGRrS/My
PLCHhZ/NPe7u0zNv2sPC90zlc5ki48v2sPAvqPddxj/rpD0sfgq+P0zVmbXT
HhZ/4Gi9VmULi18Awg/Ksg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.5, 10.9452}, {0, -1}], 
         LineBox[{{20.00000000000182, 17.000000000003638`}, {
          14.500000000001819`, 13.500000000003638`}}], 
         PolygonBox[{{17.756196892639263`, 15.572125295315896`}, {
          16.529052910483472`, 15.265339299776947`}, {16.958553304238002`, 
          14.59041010959126}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.022642987100603`, 15.952875305984765`}, {1, -1}], 
         LineBox[{{20.00000000000182, 10.000000000003638`}, {
          14.500000000001819`, 13.500000000003638`}}], 
         PolygonBox[{{16.743803107360737`, 12.072125295315896`}, {
          17.970947089516528`, 11.765339299776947`}, {17.541446695761998`, 
          11.09041010959126}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.477357012899397`, 12.452875305984765`}, {-1, -1}], LineBox[CompressedData["
1:eJwl1ws41OkeB/B/KENEcu3mmh2xGpKI8t9KJEmKRqlFSJKoIZXNJawKTUIS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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.723433619591429`, 5.328757510903266}, {1, 1}], LineBox[CompressedData["
1:eJwt1XtM01cUB/AfCOIUw8SWl0ULVigvrSloMeA6qEmDMB2Kz84hL6tUQEVW
BBSw46GoIFUeguBARSyb4xESKROZimKjlSmUCVJBWSngGsWlAsK+d9kf7c0n
53fvuffc88vPKSI+NNqUoqhI/MhIzSN/LnzKkowsPpW6mFExF87v7HHcCEv/
aR20h8f2rMuQw8KateW+sGBGW9ALJ2TvtY6G+d6jOtsVfIqZZEJdghVrtkpC
YIbFKkUvWW/OZIwUVn62+rTUlU9VBy0vK4LF/gvSxLCxrjqsFjau/+5yA2x3
dI7kFpmf1XXrM6wp4QlJnHLYShOwkY9ODRST+D1dqAzO4eqvpsG81c/9lHAf
S5G5i6zXet5EDxu745pWk+dr5AJLNz4l2UK7Yw6PzVd5OsP8qRIBOY8nRxTn
AWu7Xxl+hiceTIWwYXmS67ZcWJw8VG4Pq35tUR+AVR9iGylYMTAeG0bi0gei
AeTzO36nMwguizje3gw/HPaiBxOfrNuRBzNOTOt3wg/FM63fw3lehaJEkn+l
8rI3fC81QF8KW3ZvuLgQ1g18zFXB6q8CFo2jHjkrFzdYYP/aPSOrX8CG0dn4
IOLrFc0dMOuwLPE8rH40qb8PNx5ayOuHpWLaOTUcZaiwZuNeRCNdmr9gqanP
7gRY7Z4nWYB8mq+dsxrgthDHiHXkPDu5AX/Db9qFm46wSV1cC50wT2Yc2t4I
R7VH64Tw5pQlgVMwK+3Vyihitz/lG1Af2uS2/sMw7RGnNh++p82ZTIT7HJJ7
ekn9uC8YsXDb3ZdZS935VHF8XuQ2OH9ZX5gIlljbH+DButxiTgHMl1X4LCYu
zd7fQp5vees4gv0xS6pUvTDt7al6JRzOdasbhauj7pbnw7yO4HEDzHjT++M+
0tebz86OwHYZU+8C4byJstc9cPqCyKWkPmYnzBYpYc4vhygbeGKJy7FiWMNr
sLaCxZfWR8TDClV3KA02OhwNCYQNMSKuC5wocBi0gT2jpRcFsColvHMM52Un
7XsaB1eOD1/ogFO1y2urYFn+kbIbsGg9K/k1WV/lPiSHOW+z2C44745lP5if
huetfRoWD0+b33A+A7d1r4i9TeqR6fTsEqyUvVGa4R4U/qW2zXCOBaNmI2zn
q5G8Ivn9m2+RPtRcU7V+if0F1w+W34e9f+tPDob9XhwpnICrP3dWn4PbbLxc
7TCvxip8nNRHXjSasAoWn33f4eyB+WvM6etg3ZmU0wdhzo5wIQ/2pFH5TbDf
FZsn7rDs2Viwkbit9Z01zGLOT+J64r12v73/PfIJT7bSo2Hts8C9j+H0LknV
aVhYZBJQCVMZc+jX4PTU0OukD4X2XfoGOLE8w18IazP4Pk1wY7ZvFZP0JdNa
XkvmW8o3zpJ+/Eaw6gKJZzEihknf52puJMEMX7P6HtjbzcnqW5L/aVXBc1hl
FtPiAjM/cv/ohysVo/xJ7N+wPbblA3nP3AqCnsCsgIMzdOTjvNySdhXWPPH4
PRCWq+i7MmHKz7YuBVbpZ5n7YMvhm6MtsCCmozmMxI95bTJFPYynpgs3wbS7
P+0NJu9N6O5aEq9un8gsghOa+iLF8GZTw9zXcPq0eUk2PGFUb3HBvYgf3Sys
h8dyZ2ajYPVjmq0OrpReHi8lfWuyLI6c5/n2hxEPYOaV+8djYfkXAY7DcJ4P
e7ARrhwbCv8E//d98Pp/9OD/C+a3Nhg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.736190257669733`, 10.763809742330267`}, {-1, 1}], LineBox[CompressedData["

1:eJxFlA1MzGEcx/9FXGfI61Sul//zhDq9CR1Gv1qpEy25xHkZlyJ1vaiGzZ27
NJpsnUqGXhAWt+KYJOuFSeelF7pOrHFUtkutbJ1eZHmO9Xi2/5599uz/PL+X
7/fnLEmKiLFkGCaYfOZd/GuCLB4wfxcGo9eawCdSB2DCJXP2e2Lonv90yK+X
sKKY5xiJIT9lbWF/iiMwnDq3kiQMr4d220cyTsDsqLtdL8NQ5WiTNF1IWBVW
Y5RjyPPd++m1irChsGFmKoZN/lVFh9sJD340losxFHVluzbPdwZmmepS22oM
LQ6eo7ZbCNer/a2sCfsUPK89QVgRfW9VGwL+MtPJ7OuEW790VlxE4MmNFlXX
EN43NjqyA8HlynjOaBPhQ+ldtbYItK5CWUIb4ayQuZpOFjZrnOzvNhNWzRId
usFCkahl/at6wiMXMoWpLDx29+h1U5vPt8niQ1nIiRzbmJtDmInbrvBgYcgi
K8MymXBVw+AzHgspuzxNZWGEDRtyjbYs6OQRiS/cCYeEujggFmKk1eUFswkP
PkweF7BwkDM+T/CT5M9RiwvELBinBuTmdxEeecqRn2bhml2UbvMHwsc+JbZX
s6CvuHXmVgfhkNJSiYmF2X18hfwrYRvglfkgcBJsjfAbJrxI1zxxBEFmnljb
udAcX8DEo3sI8jOWxujAnL+yxKsXwZ0wu4EVaYQFfYbhxRj0PY6NxXfN9Q7U
dW/E0BsovxL7g3CyZMAjFsOM4KP6PF+W5BdyXHkc/9OJwsyW1lwFBlNVlmuC
9v/5No3ywKVZiP4fa2MQ7A9H9P4oTY9NUzai70tlHZf5dYjGJ0u46YuMiMbP
q60or+dgmp8HV3dSycM0f/7Wncd2LsW0Pvt6yvgPlmBaP9GPGmk8eWeyvvr+
KX2NRFeT9e/ICoqIG0S0P+7K8vNr3iHaP9WNVm6UBtH+CoeOyn7nINr/iqtW
j6XJiOojU/t2QbsIUf0cUHs9W+WHqL4ehn3uDvdGVH/pbyoz8pYjqs9Sv/zh
aSsQ1e87U6E6KABRfevfcxVpexDVv/ebuHX+pxD1R+L71mnf7iPqn6tz0/Te
3xH1V2VtsOSCG6b+k2qvOa8kPp70p9AwoBqpxNS/RyxKu+9MYOrv1Jqv3tbB
LtT/OeklVoazLnQ+nDvYtYd96ULnh8g8ZJgldL78AU4rvNg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.25, 8.354800000000001}, {0, 1}], 
         CircleBox[{5.25, 9.49375}, 2.30625, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{4.65, 11.8}, {5.85, 12.200000000000001`}, {5.85, 
          11.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 12.7452}, {0, -1}], 
         CircleBox[{5.25, 10.50625}, 2.30625, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{5.85, 8.2}, {4.65, 7.799999999999999}, {4.65, 8.6}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 7.2547999999999995`}, {0, 1}], 
         {PointSize[0.04], PointBox[{3., 10.}], PointBox[{14.5, 13.5}], 
          PointBox[{11., 10.}], PointBox[{7.5, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["26",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"ae/bfcgdg/fgfheheh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["ae/bfcgdg/fgfheheh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGCQB2IQvcICSDBwODCAgYrDjEsX5+gZszswROzg+W+v4iCT
N5s1Po3NgSHhluiNPBUHic/JeybOZHVg+NEY8bdexcHn1Ly+Y2dYHBg8TJqv
AeVL2M12/vnP7MBgkf38N1C/1EdXY2NjID+Cy+P3f2WHcywTWDPTmID2Fez8
tV/ZweTH7klzZzI6MKz45a8wUdmhpqd+94UzDA4ME/rO/axXdkhl7bgGdKcD
Q8OyNnmgfNTXqUD9/+wZNE6o/QTqz1p5/8yZM3/sGTr43/wAmj87a+tjQ+Pf
9gwCcod+AO0v8uAsnzHzpz0Dx9Qpl4DuC3oE8ucPewaDCQt+AN2/2fl7R3ra
N7g8V5SPzLkzX+D69zxSALr/M9x83/PZv6bN/Ai3fzfTw6Vf/r+3h7kvIZxV
aO+ZN/Yw99eZC27tnPnSHuY/sbLDL6PTntnD/P88+KWRsfFje1j4VGRrS/My
PLCHhZ/NPe7u0zNv2sPC90zlc5ki48v2sPAvqPddxj/rpD0sfgq+P0zVmbXT
HhZ/4Gi9VmULi18Awg/Ksg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.5, 10.9452}, {0, -1}], 
         LineBox[{{20.00000000000182, 17.000000000003638`}, {
          14.500000000001819`, 13.500000000003638`}}], 
         PolygonBox[{{17.756196892639263`, 15.572125295315896`}, {
          16.529052910483472`, 15.265339299776947`}, {16.958553304238002`, 
          14.59041010959126}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.022642987100603`, 15.952875305984765`}, {1, -1}], 
         LineBox[{{20.00000000000182, 10.000000000003638`}, {
          14.500000000001819`, 13.500000000003638`}}], 
         PolygonBox[{{16.743803107360737`, 12.072125295315896`}, {
          17.970947089516528`, 11.765339299776947`}, {17.541446695761998`, 
          11.09041010959126}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.477357012899397`, 12.452875305984765`}, {-1, -1}], LineBox[CompressedData["
1:eJwl1ws41OkeB/B/KENEcu3mmh2xGpKI8t9KJEmKRqlFSJKoIZXNJawKTUIS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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.723433619591429`, 5.328757510903266}, {1, 1}], LineBox[CompressedData["
1:eJwt1XtM01cUB/AfCOIUw8SWl0ULVigvrSloMeA6qEmDMB2Kz84hL6tUQEVW
BBSw46GoIFUeguBARSyb4xESKROZimKjlSmUCVJBWSngGsWlAsK+d9kf7c0n
53fvuffc88vPKSI+NNqUoqhI/MhIzSN/LnzKkowsPpW6mFExF87v7HHcCEv/
aR20h8f2rMuQw8KateW+sGBGW9ALJ2TvtY6G+d6jOtsVfIqZZEJdghVrtkpC
YIbFKkUvWW/OZIwUVn62+rTUlU9VBy0vK4LF/gvSxLCxrjqsFjau/+5yA2x3
dI7kFpmf1XXrM6wp4QlJnHLYShOwkY9ODRST+D1dqAzO4eqvpsG81c/9lHAf
S5G5i6zXet5EDxu745pWk+dr5AJLNz4l2UK7Yw6PzVd5OsP8qRIBOY8nRxTn
AWu7Xxl+hiceTIWwYXmS67ZcWJw8VG4Pq35tUR+AVR9iGylYMTAeG0bi0gei
AeTzO36nMwguizje3gw/HPaiBxOfrNuRBzNOTOt3wg/FM63fw3lehaJEkn+l
8rI3fC81QF8KW3ZvuLgQ1g18zFXB6q8CFo2jHjkrFzdYYP/aPSOrX8CG0dn4
IOLrFc0dMOuwLPE8rH40qb8PNx5ayOuHpWLaOTUcZaiwZuNeRCNdmr9gqanP
7gRY7Z4nWYB8mq+dsxrgthDHiHXkPDu5AX/Db9qFm46wSV1cC50wT2Yc2t4I
R7VH64Tw5pQlgVMwK+3Vyihitz/lG1Af2uS2/sMw7RGnNh++p82ZTIT7HJJ7
ekn9uC8YsXDb3ZdZS935VHF8XuQ2OH9ZX5gIlljbH+DButxiTgHMl1X4LCYu
zd7fQp5vees4gv0xS6pUvTDt7al6JRzOdasbhauj7pbnw7yO4HEDzHjT++M+
0tebz86OwHYZU+8C4byJstc9cPqCyKWkPmYnzBYpYc4vhygbeGKJy7FiWMNr
sLaCxZfWR8TDClV3KA02OhwNCYQNMSKuC5wocBi0gT2jpRcFsColvHMM52Un
7XsaB1eOD1/ogFO1y2urYFn+kbIbsGg9K/k1WV/lPiSHOW+z2C44745lP5if
huetfRoWD0+b33A+A7d1r4i9TeqR6fTsEqyUvVGa4R4U/qW2zXCOBaNmI2zn
q5G8Ivn9m2+RPtRcU7V+if0F1w+W34e9f+tPDob9XhwpnICrP3dWn4PbbLxc
7TCvxip8nNRHXjSasAoWn33f4eyB+WvM6etg3ZmU0wdhzo5wIQ/2pFH5TbDf
FZsn7rDs2Viwkbit9Z01zGLOT+J64r12v73/PfIJT7bSo2Hts8C9j+H0LknV
aVhYZBJQCVMZc+jX4PTU0OukD4X2XfoGOLE8w18IazP4Pk1wY7ZvFZP0JdNa
XkvmW8o3zpJ+/Eaw6gKJZzEihknf52puJMEMX7P6HtjbzcnqW5L/aVXBc1hl
FtPiAjM/cv/ohysVo/xJ7N+wPbblA3nP3AqCnsCsgIMzdOTjvNySdhXWPPH4
PRCWq+i7MmHKz7YuBVbpZ5n7YMvhm6MtsCCmozmMxI95bTJFPYynpgs3wbS7
P+0NJu9N6O5aEq9un8gsghOa+iLF8GZTw9zXcPq0eUk2PGFUb3HBvYgf3Sys
h8dyZ2ajYPVjmq0OrpReHi8lfWuyLI6c5/n2hxEPYOaV+8djYfkXAY7DcJ4P
e7ARrhwbCv8E//d98Pp/9OD/C+a3Nhg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.736190257669733`, 10.763809742330267`}, {-1, 1}], LineBox[CompressedData["

1:eJxFlA1MzGEcx/9FXGfI61Sul//zhDq9CR1Gv1qpEy25xHkZlyJ1vaiGzZ27
NJpsnUqGXhAWt+KYJOuFSeelF7pOrHFUtkutbJ1eZHmO9Xi2/5599uz/PL+X
7/fnLEmKiLFkGCaYfOZd/GuCLB4wfxcGo9eawCdSB2DCJXP2e2Lonv90yK+X
sKKY5xiJIT9lbWF/iiMwnDq3kiQMr4d220cyTsDsqLtdL8NQ5WiTNF1IWBVW
Y5RjyPPd++m1irChsGFmKoZN/lVFh9sJD340losxFHVluzbPdwZmmepS22oM
LQ6eo7ZbCNer/a2sCfsUPK89QVgRfW9VGwL+MtPJ7OuEW790VlxE4MmNFlXX
EN43NjqyA8HlynjOaBPhQ+ldtbYItK5CWUIb4ayQuZpOFjZrnOzvNhNWzRId
usFCkahl/at6wiMXMoWpLDx29+h1U5vPt8niQ1nIiRzbmJtDmInbrvBgYcgi
K8MymXBVw+AzHgspuzxNZWGEDRtyjbYs6OQRiS/cCYeEujggFmKk1eUFswkP
PkweF7BwkDM+T/CT5M9RiwvELBinBuTmdxEeecqRn2bhml2UbvMHwsc+JbZX
s6CvuHXmVgfhkNJSiYmF2X18hfwrYRvglfkgcBJsjfAbJrxI1zxxBEFmnljb
udAcX8DEo3sI8jOWxujAnL+yxKsXwZ0wu4EVaYQFfYbhxRj0PY6NxXfN9Q7U
dW/E0BsovxL7g3CyZMAjFsOM4KP6PF+W5BdyXHkc/9OJwsyW1lwFBlNVlmuC
9v/5No3ywKVZiP4fa2MQ7A9H9P4oTY9NUzai70tlHZf5dYjGJ0u46YuMiMbP
q60or+dgmp8HV3dSycM0f/7Wncd2LsW0Pvt6yvgPlmBaP9GPGmk8eWeyvvr+
KX2NRFeT9e/ICoqIG0S0P+7K8vNr3iHaP9WNVm6UBtH+CoeOyn7nINr/iqtW
j6XJiOojU/t2QbsIUf0cUHs9W+WHqL4ehn3uDvdGVH/pbyoz8pYjqs9Sv/zh
aSsQ1e87U6E6KABRfevfcxVpexDVv/ebuHX+pxD1R+L71mnf7iPqn6tz0/Te
3xH1V2VtsOSCG6b+k2qvOa8kPp70p9AwoBqpxNS/RyxKu+9MYOrv1Jqv3tbB
LtT/OeklVoazLnQ+nDvYtYd96ULnh8g8ZJgldL78AU4rvNg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.25, 8.354800000000001}, {0, 1}], 
         CircleBox[{5.25, 9.49375}, 2.30625, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{4.65, 11.8}, {5.85, 12.200000000000001`}, {5.85, 
          11.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 12.7452}, {0, -1}], 
         CircleBox[{5.25, 10.50625}, 2.30625, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{5.85, 8.2}, {4.65, 7.799999999999999}, {4.65, 8.6}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 7.2547999999999995`}, {0, 1}], 
         {PointSize[0.04], PointBox[{3., 10.}], PointBox[{14.5, 13.5}], 
          PointBox[{11., 10.}], PointBox[{7.5, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["27",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"ae/bfcgdg/fgfheheh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["ae/bfcgdg/fgfheheh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lQtUFFUcxu8iCIoPhDVA0FzvqCSJyqy6eJT5g0JIaWJIpLzicUNDo0BF
UCREMF9nEVyZRWLNB3RSXqZgFnIgUNcRMPKIlmKZuSqiYCYgsl1OZ+6cs2fO
78zMztzv/33fVUV/tjLOAiEUR39D59hbZnpgQL3/n8sMFUqjNMTjdoa6czDS
d70uX6RsKC3XBnHQrDE2RhLKK9x98QYO1FcXSe48ZTtBUb6Ng8EI55pe8xR6
PUsRns5B2L4naU0S5YqOVU6JHGRZBHTni5QD/tqlDeagwu3DOzGE8t003fBZ
HOicS1578pSfLdnwaBgHxz/tGW6BKGs7jga3Yyh48LC6VVIBsjnqkVCJoSv2
j9++FilH/bzRmIehozgjKIFQrtvs/CwTgzN/c7QXT9mtq+tWBoYV0QZXK0TZ
4FFzeR8GsvR4Uas0GZBpU426FMOZ1c8vFIqU3S5YDVAdNlqGnCeE8t3wiU0v
MRzucL08h6esfbJ3B+agf79qQb/5Tfq+RZNs3uUgfZupNlcc4p61JWs5sHPL
DprBU27fuaw+jYPdaxabG6RJgFLqd6u3cxAZs7gqglC+pNUfTKJ6BpJ5/eaJ
gDSPbPxWc9D5pMxHJ1J2un0P5nGgWXU/3ZOnvEKafsWWg0v6wYJmyRUQ0t8x
dWDwfPui3zoyxA0exmoMQbGpy60QZW120MwCDErN0oPFogug+ExFSxaGlacW
O3rxlA0Ja/q2Y/g1xPDomjSBrt8q13s3BrjR7LWOUA7tLP7kGwxjs9ULLRDl
Vp2j2Iihza9ypl50ps8HHPz2KYZdVm7+njzl0MAfi1048P9ybpVRcqJ6RDQG
LOHg7AHL/TGEsmZa+/VYDrZX3/5zwOwIqPRI3votHDi91pfoRMpwSmug+nj3
X2yYzVM2tTie3sRBzbEW5RXpDUA1Cz84HUn1fz8kOY5Q1swJ/2gRB5Zx52vN
5vGAEsXu7x04eOeLunZRpFywLL3qAYYTHjtO8zzlS+NdsmvpvLPnLr8qKal/
0mzGFGGIV8edIYQydAYLORiCDw/lwoHen6oso/r4WOoDRdGBXVciY6Un78Ce
x6Fb35Mke/b/mUphOiH27P1NqiU+CNmz7ys3le7Ri+PY93fnzXqs5sex9SV4
z1/YItmx9Vd9FRu6ltgxfQJbVBOHITum3zTTi3safizT18Y2v3AdGcP09yVl
SYXiaDafk/cLj12RRrH5JSt0Oa/Mtmy+9pEf73Xnbdn8TZG1M8LISOaPrqrN
1/aJI5h/4osSrtdJNsxf5aUjYl6YrZn/upwSk915a+ZPRbuiOZoMZ/5Vxagt
CkUr5u9zc45sbZMsmf/f2uKZZ4ssWT6u+6cU+vLDWH68f9isSCEWLF9qVU/Y
d6KC5e+QvUfm7xJi+Sx6en6PLUIsv/6THKs/J4OCnO/Ul8umGqUBQc7/4Kho
NIV/Jcj9kHLoF/s0sU+Q+8O1tW/CDfNLQe6XsEx17nzyryD3zwzXvqQi6R9B
7qezz3smj+CfC3J/xd8pg1SxW5D7rXdkm7bD/FSQ+6/XfdfUVqlTkPsRFuRM
qBcfCnJ/Oo3Z+riC/C3I/ZoRhX0K+HuC3L9R/iQ9Gd0V5H6e3eSdkq2/Kcj9
/SzxpL+1uk2Q+73ugC72lv6yIPd/Xdu2khf6c4K8P9AtBGUMnvhJ3j/+A78J
hDo=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{19.999999999998018`, 10.}, {16.99999999999754, 10.}}], 
         PolygonBox[{{17.9, 10.}, {19.1, 10.4}, {19.1, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {18.5, 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwt1glUU1caB/CA1AZkU3YMlk1UNqmgUAUaaQxYEaKURQ2FkcUQ1rIoaBQQ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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.885562118843275`, 6.925817748240936}, {-1, 0}], 
         LineBox[{{9.999999999990212, 10.}, {17.99999999998223, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 9.6}, {13.4, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 9.0548}, {0, 1}], 
         LineBox[{{10.000000000001526`, 10.}, {13.000000000006441`, 10.}}], 
         PolygonBox[{{10.9, 10.}, {12.1, 9.6}, {12.1, 10.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.5, 9.0548}, {0, 1}], 
         CircleBox[{15., 9.55}, 2.05, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{14.4, 11.6}, {15.6, 12.}, {15.6, 11.2}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 12.545200000000001`}, {0, -1}], 
         LineBox[CompressedData["
1:eJw9lXk0lWkcx+9FKsaUe2nRdu97Kdx3uVosk+pbHSrcSphImVKElK3TQiRL
tJMwMdqE0jK5lvYk0yaV0OlKplQzZ0ZXVDoaWeZ2es77nvOe54/nvO9v+34/
P6l/2JIAHYFAEKZ9v53fH2t8P80RE11jbtBshdMPleoXJTKMye+ar/vQCnFK
33Ut3jJcLfQMuqi2QufMMEbZS6GtX7yuu88KiKCnN5+k8Oqz3wd/B2skzTXu
+NGXQmrI7cZPSdZY5n0sosKCQqsZDOgma9Qc0n9QNYTC7r70zBRWjl0+k4Lq
9Sh491UaVsbJEai+cSxhJIWdmt8vJ9yWwzJgpIh1pLB02XTrNAGNLTWtDppQ
CpoWZeBMjkaOy9V52/Ip9Fz4Qa/anUbB8tGDhM8pmO7ZnhcTROPR8ZPvdxrK
8P78x4AtUTSOPB4V4z1VBrHoSW9AJI22PYUdE5bIkK8JvlO3hobtQtvhdatl
mGd659cjLjQezHlQ6h4sQ0duYd5+c228wn1G+tr7o94P1mZ1yWHcEmEw112G
FcMkJwbfkMOpc2dizxQZHC8IymKT5UgMSphRbiDDzQwU5CyW46mlfV1nI4UU
286NmRba+sXG8bUHKViNWRR6yEAOk9gWkzonCs628ab1enKEXNWbnPhBimbv
3MATI+VwzHG/vjdXikbbvuUH5sqRmV2fwrpK8d/p+57N2njGCV80E4dIITAL
yDmnlsN2RLjXrCYJvDS7qnun0LCf78x035Jgjolfp94BGqqtqQcr7kgw8HaE
qLWVxq4mRN18LcHWD9V59RMZvBvr++qJSIp9hz5u/3sZA+Hk+xek7lLIAp8O
fhPLYGVa2mP/bCkUpccLT6QzCO822pH8UoozBjKnCYcZRD+yNZeZU/hkp0zf
nM0gKNPq3amVFEImuZ6RH2Dw2M6iODCdwjUvD9XlHQzGjlron1dKYb2dsPlU
GIOGXp+z6bcpZLe5OH/21uajLl5z8A6Fo7/ZTCufwaDmzCDf1goKOsKIN1Xj
Gdi6Vt2MyaZgkn6hW9GvrW/o69hnwRQWPBMnb3tJo7Hheu8ljsK2V4Yhyrs0
5BnF+rPeSVGk/GtK+xUaDWqTFZZHpfhz+bKusEoaczJ1zKYukmI0hAdK1DTY
NkmHv0CKuxUZpUuGMEjbOO7D80sSpDYry7rcGHha3PRbsF2C9NS+mJ7jDJZK
LQbqfCTEZyxqvz1uEoRt1kka5MfiZ3K/xVzlaljK8t/bLQh1O9fN8v//p7I/
fOEkjo8vzS8wTAOH2SQ/H5VZrciJQz3JP2lCV6bKnoM1qe+ipSY6UcLx9bdv
WDTksZBDCelPf//+sTNbWUwj/Vu1+lGNyy2W72+IafllgyKW7/+he/3mS9NY
fj7OxWeVznEsxpP5ZanLPBvDWX6+RvdUG1KDWH7+np9fXHkVyPL6qHPz81mx
gUUk0U/LYesvYxNYrCL6ilxnFD+6gIUO0V/yj47+TBMLDdHngNm/7ePHc9hN
9Fv57viwr+Ecr+9i8SZH3VoODkT/gqiHk0+LFbw/dOW5XqIZCgwn/slcs8Pd
3k+Bg8RfGR6qWnajAtOJ/1rezvqcFatAMPFnaKT6fnm0AiLi3+tvU5ymrVcg
lfhbUvHLOdVSBRqI/98UDVjM1sZLIHxY8qWD7ZUoeH6UeJjVlg3S5kP4Ilrj
4x/RzPH8udLmsHZVOcfzyU+c5cDlcjy/2ttZ3Zx9HM83kbtRTspejuefZmO8
f1I2hzrCR/1Fn746nudQRPgpplSHq+o45BK+nir442nYAIdowt9rQy3zZ45T
wIrw2UMoiT4GBdYSfv8ktJt2LUiB3YTv+z++7AnOUOA+4f9VlzzTndcV/H4Y
lrw4ZeEbbT/I/tgj2OSiJ7Th98uJs09dT4pt0EH2T9T5ZN+40Tb8fqq+ldMe
LbJBEdlf3/1mw++3/wEyGNnI
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 7.4548000000000005`}, {0, 1}], 
         {PointSize[0.04], PointBox[{10., 10.}], PointBox[{18., 10.}], 
          PointBox[{17., 10.}], PointBox[{13., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["28",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGCQB2IQvcICSDBwODCAgYrDjEsX5+gZszswROzg+W+v4iCT
N5s1Po3NgSHhluiNPBUHic/JeybOZHVg+NEY8bdexcHn1Ly+Y2dYHBg8TJqv
AeVL2M12/vnP7MBgkf38N1C/1EdXY2NjID+Cy+P3f2WHcywTWDPTmID2Fez8
tV/ZweTH7klzZzI6MKz45a8wUdmhpqd+94UzDA4ME/rO/axXdkhl7bgGdKcD
Q8OyNnmgfNTXqUD9/+wZNE6o/QTqz1p5/8yZM3/sGTr43/wAmj87a+tjQ+Pf
9gwCcod+AO0v8uAsnzHzpz0Dx9Qpl4DuC3oE8ucPewaDCQt+AN2/2fl7R3ra
N7g8V5SPzLkzX+D69zxSALr/M9x83/PZv6bN/Ai3fzfTw6Vf/r+3h7kvIZxV
aO+ZN/Yw99eZC27tnPnSHuY/sbLDL6PTntnD/P88+KWRsfFje1j4VGRrS/My
PLCHhZ/NPe7u0zNv2sPC90zlc5ki48v2sPAvqPddxj/rpD0sfgq+P0zVmbXT
HhZ/4Gi9VmULi18Awg/Ksg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwt1glUU1caB/CA1AZkU3YMlk1UNqmgUAUaaQxYEaKURQ2FkcUQ1rIoaBQQ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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.885562118843275`, 6.925817748240936}, {-1, 0}], 
         LineBox[{{18.00000000000045, 10.}, {9.99999999999955, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 10.4}, {13.4, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJw9k3lIVFEUxp9iOiWKRppSLve+5zJKOhpaOqNzQjNTwtFKQjAnl3LLMRtT
cV8oUwKr0WiRSZPcN6IwtdJUxMrBFhDatCnEVFxTwammO/1xL1weP959951z
vu9DsbKIBH2GYfzI1j0917Vk7QHm/+Lgy2osGOTsBSa9Yi3Wm4Oy2/yW6G02
wDQ9uhEYw8EB+YrEUknYmd93JJuDqQxhSIbYFpjlbsFWIQeMxeSZ41OEexSG
c5kczJnvLqzPsgOmK8S7LYqD8JxIyWfGHhjptN4Hdw5KHRTt00LC2a11nzZZ
8EuPuFmeTJg3WaN9zMKLLwP+XpWEiyxfmqax4CptnbGsJWxma+DJsbAVMtN7
8D7h2aaWjs8Y1vq2+xZVEy5v8mm8iyHqScKssoCw5OjBr/EYSgKH9/2NJjwg
qywQYrh0tUZ13Zswo1EsIAwCvzLTJ8aEnX3H660xNOwo3bE6SOoX8NV7WAzn
Mi14qouEB7TK134Yfjv2u7TtJyxZNnZKxFAbaOafZERYKvbpU2LYq/jB9K/p
5sG7Ev4Ng5d4KXpJQ9isy82Oz8LzQ3XN3YicTy+O+5PBwqUO+x5NPGFnh5GH
vSykRZn5mw8TPqW4taFhIbTqlGjcRlf/GyvH/RzwRP1vk+N0/XskGkRzkCqz
6VrsJBw8OOSQxYFoQ6cv4QlZjFERByb56jBzCaLvL3QFuFXfQ/T71YTGIU81
ovcbjm1KN20w/b++coktOYZpfadPurs6yTCt/0EkL5xXgml/7m3XtOgypv0v
vt+UhuZhOp8X6ro5/llM51chXNdYBWE63/n051uGdpjO3+Rp7srRX4jqM5qa
5JE2hqh+vd3Cd3r1iOr7XXxnt7QYUf1DfDdGCpMR9UdAq/Xh0RhE/bOz5oT1
z1hE/VVgJR58JUfUf8zl9tSsakT92TuREmE3jKh/F/KaO1ktov4OPi8fcAvE
1P/PUnaFRVZhmg+VSP0xaBrT/CDRlUm5gKX5yuXPu4TnszR/qjzVVP4QS/PZ
oNHpy9L8/gNTRnP8
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.5, 10.9452}, {0, -1}], 
         CircleBox[{5., 10.45}, 2.05, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{4.4, 8.4}, {5.6, 8.}, {5.6, 8.8}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 7.4548000000000005`}, {0, 1}], 
         CircleBox[{5., 9.55}, 2.05, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{5.6, 11.6}, {4.4, 12.}, {4.4, 11.2}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 12.545200000000001`}, {0, -1}], 
         {PointSize[0.04], PointBox[{3., 10.}], PointBox[{18., 10.}], 
          PointBox[{10., 10.}], PointBox[{7., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["29",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGCQB2IQvcICSDBwODCAgYrDjEsX5+gZszswROzg+W+v4iCT
N5s1Po3NgSHhluiNPBUHic/JeybOZHVg+NEY8bdexcHn1Ly+Y2dYHBg8TJqv
AeVL2M12/vnP7MBgkf38N1C/1EdXY2NjID+Cy+P3f2WHcywTWDPTmID2Fez8
tV/ZweTH7klzZzI6MKz45a8wUdmhpqd+94UzDA4ME/rO/axXdkhl7bgGdKcD
Q8OyNnmgfNTXqUD9/+wZNE6o/QTqz1p5/8yZM3/sGTr43/wAmj87a+tjQ+Pf
9gwCcod+AO0v8uAsnzHzpz0Dx9Qpl4DuC3oE8ucPewaDCQt+AN2/2fl7R3ra
N7g8V5SPzLkzX+D69zxSALr/M9x83/PZv6bN/Ai3fzfTw6Vf/r+3h7kvIZxV
aO+ZN/Yw99eZC27tnPnSHuY/sbLDL6PTntnD/P88+KWRsfFje1j4VGRrS/My
PLCHhZ/NPe7u0zNv2sPC90zlc5ki48v2sPAvqPddxj/rpD0sfgq+P0zVmbXT
HhZ/4Gi9VmULi18Awg/Ksg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000003638`, 17.000000000006366`}, {
          18.000000000003638`, 10.000000000003638`}}], 
         PolygonBox[{{19.164832676738428`, 14.076914368584493`}, {
          18.45055774420524, 13.032974082574459`}, {19.219776902317903`, 
          12.813197180256555`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.114437881156725`, 13.925817748240936`}, {1, 0}], 
         LineBox[{{20.00000000000112, 10.000000000003638`}, {
          10.000000000003325`, 10.000000000003638`}}], 
         PolygonBox[{{14.4, 10.}, {15.6, 10.4}, {15.6, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwt1glUU1caB/CA1AZkU3YMlk1UNqmgUAUaaQxYEaKURQ2FkcUQ1rIoaBQQ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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{19.885562118843275`, 6.925817748240936}, {-1, 0}], 
         LineBox[{{18.00000000000045, 10.}, {9.99999999999955, 10.}}], 
         PolygonBox[{{14.6, 10.}, {13.4, 10.4}, {13.4, 9.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {14., 10.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJw9k3lIVFEUxp9iOiWKRppSLve+5zJKOhpaOqNzQjNTwtFKQjAnl3LLMRtT
cV8oUwKr0WiRSZPcN6IwtdJUxMrBFhDatCnEVFxTwammO/1xL1weP959951z
vu9DsbKIBH2GYfzI1j0917Vk7QHm/+Lgy2osGOTsBSa9Yi3Wm4Oy2/yW6G02
wDQ9uhEYw8EB+YrEUknYmd93JJuDqQxhSIbYFpjlbsFWIQeMxeSZ41OEexSG
c5kczJnvLqzPsgOmK8S7LYqD8JxIyWfGHhjptN4Hdw5KHRTt00LC2a11nzZZ
8EuPuFmeTJg3WaN9zMKLLwP+XpWEiyxfmqax4CptnbGsJWxma+DJsbAVMtN7
8D7h2aaWjs8Y1vq2+xZVEy5v8mm8iyHqScKssoCw5OjBr/EYSgKH9/2NJjwg
qywQYrh0tUZ13Zswo1EsIAwCvzLTJ8aEnX3H660xNOwo3bE6SOoX8NV7WAzn
Mi14qouEB7TK134Yfjv2u7TtJyxZNnZKxFAbaOafZERYKvbpU2LYq/jB9K/p
5sG7Ev4Ng5d4KXpJQ9isy82Oz8LzQ3XN3YicTy+O+5PBwqUO+x5NPGFnh5GH
vSykRZn5mw8TPqW4taFhIbTqlGjcRlf/GyvH/RzwRP1vk+N0/XskGkRzkCqz
6VrsJBw8OOSQxYFoQ6cv4QlZjFERByb56jBzCaLvL3QFuFXfQ/T71YTGIU81
ovcbjm1KN20w/b++coktOYZpfadPurs6yTCt/0EkL5xXgml/7m3XtOgypv0v
vt+UhuZhOp8X6ro5/llM51chXNdYBWE63/n051uGdpjO3+Rp7srRX4jqM5qa
5JE2hqh+vd3Cd3r1iOr7XXxnt7QYUf1DfDdGCpMR9UdAq/Xh0RhE/bOz5oT1
z1hE/VVgJR58JUfUf8zl9tSsakT92TuREmE3jKh/F/KaO1ktov4OPi8fcAvE
1P/PUnaFRVZhmg+VSP0xaBrT/CDRlUm5gKX5yuXPu4TnszR/qjzVVP4QS/PZ
oNHpy9L8/gNTRnP8
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.5, 10.9452}, {0, -1}], 
         CircleBox[{5., 10.45}, 2.05, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{4.4, 8.4}, {5.6, 8.}, {5.6, 8.8}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 7.4548000000000005`}, {0, 1}], 
         CircleBox[{5., 9.55}, 2.05, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{5.6, 11.6}, {4.4, 12.}, {4.4, 11.2}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5., 12.545200000000001`}, {0, -1}], 
         {PointSize[0.04], PointBox[{3., 10.}], PointBox[{18., 10.}], 
          PointBox[{10., 10.}], PointBox[{7., 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["30",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl0g0w1GkcB/BHS7TIcqzFdlznrc5ce5eTbO7/RzWuTuP1rJSRiWGutGoy
TkjczZTzXopK9pquVpjbs5bVm1U4d8QSZ+eY0FjteTkqnLdy33/3zOw889nn
N7+X//N8FHMiOHYdIeQAfsxOdm/H+pAmpWVYLjTxP8lKadxEE0nbL0H34Oos
sxJfOLDV17IflnVTej18muiSNjWMwUOmdEYcLJI3/qiFORl5QyxYUDsxxMRP
ldQMS+1oMlLnvKUeduD77A2FBbZV5hfgsYbhORbsOtBIQuDSbv0vFbY0KbTS
FdvAudnJ3rGwQ3DJwAtn5J9c8DKH26Mss+rg6yxrwwYbmqgsbF0uwqvCjj1h
sOBqjOQc7LlPfX2SR5MUs7iqTJh2/LUpFc7cZaXOh1V2feH6cHvParEU5o2r
OnKsaaL8036qHZaZCqY5MEcue/03rHugLi3j0kR8+8ltA6bf2LJQF1h9jPuI
y8wbUbl03wrzFlWImP7HLnicFsG88t7vjZn4vhnpqiXyXFw4N4180TXsE3dg
cV59QzN8Kyyp+RuY525yNBe+xNa8M4Rl/afZIcx8r7N2Kz/AfEOtDXawSU6L
+igs097cPuVEE/5DiyPGsKqSFv4Of33X5exdC5pI+RXDClhar631gTMfj/Dl
cNp3ouIec9TLr1Y2wbtuyGoiYXEr95gG5h0Yj3jBwfzZqSGLzLnO/kg8XEof
LLFBfSP3ub2jZujnluFWD/j8s7aElY3Iv3aq0h9efObD/hiW1CpZAXDpH8eF
B01xH6yHxj6w63r2W4kJTTTc9GZHWDBzVThvjP872cHLqFfYtd86BBYH+T9t
hWfntsUr2HCc37VcmFymE7mM42uehML+zpIS8QbUz2A7ODox72bRvckI93vf
6NKaI+bYWeS2Zoi+58rjJuCRsKnAbbAywK1XCxvxa4oD1+M9n7ny1ywsHU9/
HmuAe4ocTTRCvpHWnY2n9PGOunt7XGBNctXPZ1not+KB8CvYc/pkfME67D80
VcQx33veyU+qBw8EpKYy/aVN0l0E9x8TlZfJzLc1rZoFK6c9k5JhzqPB7KJ3
FHG1zB+PhAVexwN2vKUIna17/jmsvjkd9GqFIrqIcK/38xi4n1EsU8Tz8Ra/
Nji6eVGUskQRqTBqKQ8WR/je+XSRIpICzZ5DsMOgtqtrgSIp3tUCD5hOH3AL
nqeI/8boL+wZF0jnFG/gNE4nj/k+h/UKV15RRGWyb8dmOPPp4UTXWeTz3V/u
Cav+Xb6cM0URjX1gQjhMqnPZ93ToN/6flylMvjR50MIYRcR9lRuuMPHqn9Sh
oxQRTXR9Vs3Ev2mp7xxE/k/IgpyJ9zWbSeinyOy1yboq5vylisfupsj5Cs5s
MVNfoTXb3IL5zG8c+pY5L7IT/ibH/B2ulUw/71eh9/+7I/0fdJbJ3g==
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.4155382286327365`, 11.802942429901577`}, {0, -1}], 
         LineBox[{{20.00000000000182, 17.}, {16.50000000000091, 15.5}}], 
         PolygonBox[{{18.801487018010835`, 16.486351579147502`}, {
          17.540945262557496`, 16.38130643285972}, {17.856080701420833`, 
          15.645990408845275`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.24422508720229, 17.069874796527987`}, {1, -1}], 
         LineBox[{{20.000000000007276`, 10.}, {12.500000000007276`, 7.5}}], 
         PolygonBox[{{15.680790021169692`, 8.560263340389897}, {
          16.692718872423573`, 9.319209978830308}, {16.945701085237044`, 
          8.560263340389898}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.7401530373261, 7.884340888021704}, \
{0, 1}], LineBox[CompressedData["
1:eJwt1gs01HkbB/B/KeyQpkFpE0OFymVc2lTSX+mijAiVJQ3lVikKiXfLvd2x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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.697184550288068`, 3.925440917146776}, {1, 1}], 
         LineBox[{{7.549999999999272, 11.800000000001091`}, {
          12.500000000001819`, 7.5}}], 
         PolygonBox[{{9.57203974204145, 10.043480628125609`}, {
          10.215639839208144`, 8.954545866568692}, {10.740280676708956`, 
          9.558492877180091}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], {9.613308377342179, 9.084652666940414}, \
{1, 1}], LineBox[{{7.549999999995634, 11.799999999995634`}, {
          11.649999999995998`, 13.549999999995634`}}], 
         PolygonBox[{{10.151834507290827`, 12.910539118965598`}, {
          9.20519157201957, 12.071571209507185`}, {8.891139413398774, 
          12.807350552561621`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.596323942672207, 13.495772477167971`}, {1, -1}], 
         CircleBox[{14.294374999999999`, 13.979375000000001`}, 
          2.6790076579304505`, {0.6035970165265399, 3.302560785420751}], 
         PolygonBox[{{13.851689338152955`, 16.688823548329538`}, {
          12.589094962960687`, 16.612302677105767`}, {12.887526360733403`, 
          15.87005022623516}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {12.716789166815358`, 17.2983141235618}, \
{0, -1}], LineBox[CompressedData["
1:eJwdVA04V+cbJompIRWR5qOZ877nnN85aSlCdx/+pEgSRlitWiGVIqkJIVol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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.43321083318464, 11.751685876438199`}, {0, 1}], 
         {PointSize[0.04], PointBox[{7.55, 11.8}], PointBox[{16.5, 15.5}], 
          PointBox[{12.5, 7.5}], PointBox[{11.65, 13.55}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["31",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdg/egehfhfh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdg/egehfhfh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGCQB2IQvcICSDBwODCAgYrDjEsX5+gZszswROzg+W+v4iCT
N5s1Po3NgSHhluiNPBUHic/JeybOZHVg+NEY8bdexcHn1Ly+Y2dYHBg8TJqv
AeVL2M12/vnP7MBgkf38N1C/1EdXY2NjID+Cy+P3f2WHcywTWDPTmID2Fez8
tV/ZweTH7klzZzI6MKz45a8wUdmhpqd+94UzDA4ME/rO/axXdkhl7bgGdKcD
Q8OyNnmgfNTXqUD9/+wZNE6o/QTqz1p5/8yZM3/sGTr43/wAmj87a+tjQ+Pf
9gwCcod+AO0v8uAsnzHzpz0Dx9Qpl4DuC3oE8ucPewaDCQt+AN2/2fl7R3ra
N7g8V5SPzLkzX+D69zxSALr/M9x83/PZv6bN/Ai3fzfTw6Vf/r+3h7kvIZxV
aO+ZN/Yw99eZC27tnPnSHuY/sbLDL6PTntnD/P88+KWRsfFje1j4VGRrS/My
PLCHhZ/NPe7u0zNv2sPC90zlc5ki48v2sPAvqPddxj/rpD0sfgq+P0zVmbXT
HhZ/4Gi9VmULi18Awg/Ksg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000005457`, 17.000000000003638`}, {
          11.000000000007276`, 10.000000000005457`}}], 
         PolygonBox[{{15.973611330425795`, 13.868364368108953`}, {
          14.780812424168237`, 13.447376518841578`}, {15.271964914980172`, 
          12.815894744940517`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.153192049051874`, 14.118695936933307`}, {1, -1}], 
         LineBox[{{20.00000000000182, 10.000000000003638`}, {
          14.500000000001819`, 6.500000000003638}}], 
         PolygonBox[{{16.743803107360737`, 7.927874704684104}, {
          17.541446695761998`, 8.90958989040874}, {17.970947089516528`, 
          8.234660700223053}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.477357012899397`, 7.547124694015234}, {-1, 1}], LineBox[CompressedData["
1:eJwd1g001PkaB/B/k5CXdory0rCSVWQH6RYRfxHu9EbZ05I0KR1FoXVGadJo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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.646830142565392`, 3.4565616526027583`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.000000000001819`}, {
          14.499999999994543`, 6.500000000001819}}], 
         PolygonBox[{{12.325735931288072`, 8.674264068711928}, {
          12.89142135623731, 7.542893218813452}, {13.457106781186546`, 
          8.10857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.241215510839147`, 8.741215510839147}, {-1, -1}], LineBox[CompressedData["
1:eJxFlA1MzGEcx/9FXGfI61Sul//zhDq9CR1Gv1qpEy25xHkZlyJ1vaiGzZ27
NJpsnUqGXhAWt+KYJOuFSeelF7pOrHFUtkutbJ1eZHmO9Xi2/5599uz/PL+X
7/fnLEmKiLFkGCaYfOZd/GuCLB4wfxcGo9eawCdSB2DCJXP2e2Lonv90yK+X
sKKY5xiJIT9lbWF/iiMwnDq3kiQMr4d220cyTsDsqLtdL8NQ5WiTNF1IWBVW
Y5RjyPPd++m1irChsGFmKoZN/lVFh9sJD340losxFHVluzbPdwZmmepS22oM
LQ6eo7ZbCNer/a2sCfsUPK89QVgRfW9VGwL+MtPJ7OuEW790VlxE4MmNFlXX
EN43NjqyA8HlynjOaBPhQ+ldtbYItK5CWUIb4ayQuZpOFjZrnOzvNhNWzRId
usFCkahl/at6wiMXMoWpLDx29+h1U5vPt8niQ1nIiRzbmJtDmInbrvBgYcgi
K8MymXBVw+AzHgspuzxNZWGEDRtyjbYs6OQRiS/cCYeEujggFmKk1eUFswkP
PkweF7BwkDM+T/CT5M9RiwvELBinBuTmdxEeecqRn2bhml2UbvMHwsc+JbZX
s6CvuHXmVgfhkNJSiYmF2X18hfwrYRvglfkgcBJsjfAbJrxI1zxxBEFmnljb
udAcX8DEo3sI8jOWxujAnL+yxKsXwZ0wu4EVaYQFfYbhxRj0PY6NxXfN9Q7U
dW/E0BsovxL7g3CyZMAjFsOM4KP6PF+W5BdyXHkc/9OJwsyW1lwFBlNVlmuC
9v/5No3ywKVZiP4fa2MQ7A9H9P4oTY9NUzai70tlHZf5dYjGJ0u46YuMiMbP
q60or+dgmp8HV3dSycM0f/7Wncd2LsW0Pvt6yvgPlmBaP9GPGmk8eWeyvvr+
KX2NRFeT9e/ICoqIG0S0P+7K8vNr3iHaP9WNVm6UBtH+CoeOyn7nINr/iqtW
j6XJiOojU/t2QbsIUf0cUHs9W+WHqL4ehn3uDvdGVH/pbyoz8pYjqs9Sv/zh
aSsQ1e87U6E6KABRfevfcxVpexDVv/ebuHX+pxD1R+L71mnf7iPqn6tz0/Te
3xH1V2VtsOSCG6b+k2qvOa8kPp70p9AwoBqpxNS/RyxKu+9MYOrv1Jqv3tbB
LtT/OeklVoazLnQ+nDvYtYd96ULnh8g8ZJgldL78AU4rvNg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.25, 10.9452}, {0, -1}], 
         CircleBox[{5.25, 10.50625}, 2.30625, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{4.65, 8.2}, {5.85, 7.799999999999999}, {5.85, 8.6}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 7.2547999999999995`}, {0, 1}], 
         CircleBox[{5.25, 9.49375}, 2.30625, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{5.85, 11.8}, {4.65, 12.200000000000001`}, {4.65, 
          11.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 12.7452}, {0, -1}], 
         {PointSize[0.04], PointBox[{3., 10.}], PointBox[{11., 10.}], 
          PointBox[{14.5, 6.5}], PointBox[{7.5, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["32",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdg/fgfheheh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdg/fgfheheh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJxTTMoPSmViYGCQB2IQvcICSDBwODCAgYrDjEsX5+gZszswROzg+W+v4iCT
N5s1Po3NgSHhluiNPBUHic/JeybOZHVg+NEY8bdexcHn1Ly+Y2dYHBg8TJqv
AeVL2M12/vnP7MBgkf38N1C/1EdXY2NjID+Cy+P3f2WHcywTWDPTmID2Fez8
tV/ZweTH7klzZzI6MKz45a8wUdmhpqd+94UzDA4ME/rO/axXdkhl7bgGdKcD
Q8OyNnmgfNTXqUD9/+wZNE6o/QTqz1p5/8yZM3/sGTr43/wAmj87a+tjQ+Pf
9gwCcod+AO0v8uAsnzHzpz0Dx9Qpl4DuC3oE8ucPewaDCQt+AN2/2fl7R3ra
N7g8V5SPzLkzX+D69zxSALr/M9x83/PZv6bN/Ai3fzfTw6Vf/r+3h7kvIZxV
aO+ZN/Yw99eZC27tnPnSHuY/sbLDL6PTntnD/P88+KWRsfFje1j4VGRrS/My
PLCHhZ/NPe7u0zNv2sPC90zlc5ki48v2sPAvqPddxj/rpD0sfgq+P0zVmbXT
HhZ/4Gi9VmULi18Awg/Ksg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {1.5, 9.0548}, {0, 1}], 
         LineBox[{{20.000000000005457`, 17.000000000003638`}, {
          11.000000000007276`, 10.000000000005457`}}], 
         PolygonBox[{{15.973611330425795`, 13.868364368108953`}, {
          14.780812424168237`, 13.447376518841578`}, {15.271964914980172`, 
          12.815894744940517`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.153192049051874`, 14.118695936933307`}, {1, -1}], 
         LineBox[{{20.00000000000182, 10.000000000003638`}, {
          14.500000000001819`, 6.500000000003638}}], 
         PolygonBox[{{16.743803107360737`, 7.927874704684104}, {
          17.541446695761998`, 8.90958989040874}, {17.970947089516528`, 
          8.234660700223053}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.477357012899397`, 7.547124694015234}, {-1, 1}], LineBox[CompressedData["
1:eJwd1g001PkaB/B/k5CXdory0rCSVWQH6RYRfxHu9EbZ05I0KR1FoXVGadJo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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.646830142565392`, 3.4565616526027583`}, {1, 1}], 
         LineBox[{{10.999999999996362`, 10.000000000001819`}, {
          14.499999999994543`, 6.500000000001819}}], 
         PolygonBox[{{12.325735931288072`, 8.674264068711928}, {
          12.89142135623731, 7.542893218813452}, {13.457106781186546`, 
          8.10857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.241215510839147`, 8.741215510839147}, {-1, -1}], LineBox[CompressedData["
1:eJxFlA1MzGEcx/9FXGfI61Sul//zhDq9CR1Gv1qpEy25xHkZlyJ1vaiGzZ27
NJpsnUqGXhAWt+KYJOuFSeelF7pOrHFUtkutbJ1eZHmO9Xi2/5599uz/PL+X
7/fnLEmKiLFkGCaYfOZd/GuCLB4wfxcGo9eawCdSB2DCJXP2e2Lonv90yK+X
sKKY5xiJIT9lbWF/iiMwnDq3kiQMr4d220cyTsDsqLtdL8NQ5WiTNF1IWBVW
Y5RjyPPd++m1irChsGFmKoZN/lVFh9sJD340losxFHVluzbPdwZmmepS22oM
LQ6eo7ZbCNer/a2sCfsUPK89QVgRfW9VGwL+MtPJ7OuEW790VlxE4MmNFlXX
EN43NjqyA8HlynjOaBPhQ+ldtbYItK5CWUIb4ayQuZpOFjZrnOzvNhNWzRId
usFCkahl/at6wiMXMoWpLDx29+h1U5vPt8niQ1nIiRzbmJtDmInbrvBgYcgi
K8MymXBVw+AzHgspuzxNZWGEDRtyjbYs6OQRiS/cCYeEujggFmKk1eUFswkP
PkweF7BwkDM+T/CT5M9RiwvELBinBuTmdxEeecqRn2bhml2UbvMHwsc+JbZX
s6CvuHXmVgfhkNJSiYmF2X18hfwrYRvglfkgcBJsjfAbJrxI1zxxBEFmnljb
udAcX8DEo3sI8jOWxujAnL+yxKsXwZ0wu4EVaYQFfYbhxRj0PY6NxXfN9Q7U
dW/E0BsovxL7g3CyZMAjFsOM4KP6PF+W5BdyXHkc/9OJwsyW1lwFBlNVlmuC
9v/5No3ywKVZiP4fa2MQ7A9H9P4oTY9NUzai70tlHZf5dYjGJ0u46YuMiMbP
q60or+dgmp8HV3dSycM0f/7Wncd2LsW0Pvt6yvgPlmBaP9GPGmk8eWeyvvr+
KX2NRFeT9e/ICoqIG0S0P+7K8vNr3iHaP9WNVm6UBtH+CoeOyn7nINr/iqtW
j6XJiOojU/t2QbsIUf0cUHs9W+WHqL4ehn3uDvdGVH/pbyoz8pYjqs9Sv/zh
aSsQ1e87U6E6KABRfevfcxVpexDVv/ebuHX+pxD1R+L71mnf7iPqn6tz0/Te
3xH1V2VtsOSCG6b+k2qvOa8kPp70p9AwoBqpxNS/RyxKu+9MYOrv1Jqv3tbB
LtT/OeklVoazLnQ+nDvYtYd96ULnh8g8ZJgldL78AU4rvNg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.25, 10.9452}, {0, -1}], 
         CircleBox[{5.25, 10.50625}, 2.30625, {3.3629070959375835`, 
          6.0618708648317945`}], 
         PolygonBox[{{4.65, 8.2}, {5.85, 7.799999999999999}, {5.85, 8.6}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 7.2547999999999995`}, {0, 1}], 
         CircleBox[{5.25, 9.49375}, 2.30625, {0.2213144423477913, 
          2.9202782112420023`}], 
         PolygonBox[{{5.85, 11.8}, {4.65, 12.200000000000001`}, {4.65, 
          11.4}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {5.25, 12.7452}, {0, -1}], 
         {PointSize[0.04], PointBox[{3., 10.}], PointBox[{11., 10.}], 
          PointBox[{14.5, 6.5}], PointBox[{7.5, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["33",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"ae/bfcgdg/fgfheheh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "ae/bfcgdg/fgfheheh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 0}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {57., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {61.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {66., 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {70.5, 47.959999999999994`}, {0, 0}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["g",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {75., 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1A1UjFkYB/A7fTBRtpWWaqtpVFI+Bqtd1sx7RYQi0TQlmhTJKoWlVltT
26Z1iiiUsoaWzaZSURzUYCUtq0XpUExHHyNpR/QhY9v/PbznvOee33nufZ7n
3vue1279Vp8NeoSQQLxs/PgI6MdxKiXKHP9tu2wp0ZT7GJrCItsDKX02lNCj
Oxwt4W+mbepOYC7T+VrAUQskBy3hjL5y81HwxcYi81vWlMhsSH/XFEoES5Wp
ybAywr3uCnOKOHMF3JRduzcJlunUr4SwSLZGQ2G1Zbnd4JeUqJ66nfjgQolH
wvc7amHNqkslVbDid01LOou/fb71F/hcy4PnC+DaOSuag2FVkq9YY4V69Xd0
S2BNu6vjbvhifHDefLgpKVQ6bElJalu99yI4e971X2NhD3L+/irmzT/O0VpQ
4hScnhMKF7i2d4TD2c3HTm1j+dZNGuiegD6KKgNi4HpXfmg8TPpbPaPhCaUb
rOxgD/5vrQGw0+LgrIbxlMS064q/Yv2HhFflwKQ6fi9h9UrSPMNhkVfYouvO
lPC/XuMjYXGp2DEBLsgZshoLq/Pu2YjhGEm+8OkX2O8zrY4HCxadG8qDVYYd
B+5PxpgrtVgGm2oK9ctg/vpHvp3mmH/DJuAE3BR002w7LFPK/Y/Dyjd+pr3j
4FWPiwvg1EiDx+Gw6mTE5otwzNm7hq1muN/bmWW1sDzXhVsHk4VVOQ/gwWLV
6s6x2Me+g/3M2ef3hCpg/uoZghpYKy1wd4FVMy5sZ/mjKv3cOz+nxFvxfP9u
mNj70lJYVpYucYNFt+TmybBgfvMYHqyoDJT7w2p9XslVJ5zHnMgQZzjVTN86
Hq6nnTqtKb7Hlc53FsNy656UAripcPpTGzhjqWHicjhVlJJmCGtjjzTe/Az3
kuX66P0k9DHuZVr9GFjYkjwMqy/oompMsD62brsx5gtGzqgpN8a99lSKrGHV
HgP+6dFYF3L4qgvsbUL/OzMK+26cuGUmLIqo67pthD69hUVT2PzbP5kZw8rS
MxXj2fzDsS3RfOxrZ+uLXtTTLjM+OjAS53A5/0U1THsXmhyCBfn31iWx+LHZ
RyWw+tVf2zgWDznS1TEC96mWmH1wRP6pSxoUcH3YPYtrcFTiWgkP3iSt3pkB
K/Z4n5Ub4nyf1cRFwtT8D1meAeqtbQ4LZPGMWYcq9HFOrtNE/nBGqXRXhR76
9blUEsziehcaTvNQX2hnHg2rC+uuZBHc+7dbMhNhQUK7g9UwRwbFubPTYSWP
C5r7gSMZA28q97H+dj2SZr/niCZrlEcSix8SeImHODLhtYvpRtaP4GHopHfw
z8t+mMvy2fE6lw5ifexLgR6bv2RjT+YAR2SeDlHXHND3/lniF/0c8RgOMIiH
VadqG6bD8stpLWJY/UT8YHkfRxRhcckjYFoX2TfzLfLv+5s22yN/n33QP70c
KZjXHXeNuUYqnPwa9Y63RVTAqpUmf6b0YD/VzaMvwYoqtVFoF0fUG+wCb8J0
i9FKyw6O8Dd66TWyeHXi+metHGl63KDfbs/+A22Nd5+gXpDtyW5mzxha8ZAj
3n5GUzTMDsK2gLsciao2O/OQ2fXftOQbHIlJuOJeyrz6XdZ3lRzRBudHJzA7
paXfOIX+bCcXutl/+k8rr5LBYTwT6f/y6hvZ
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.8448950785695333`, 12.996142082543052`}, {1, -1}], 
         LineBox[{{20.000000000004547`, 18.000000000003638`}, {
          8.000000000002728, 14.500000000003638`}}], 
         PolygonBox[{{14.576, 16.418}, {13.312, 16.466}, {13.536, 15.698}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.565999999999999`, 17.1332}, {0, -1}],
          LineBox[{{19.999999999996362`, 11.999999999996362`}, {
          15.499999999995453`, 9.999999999996362}}], 
         PolygonBox[{{17.201713070827847`, 10.756316920367931`}, {
          18.135831542750775`, 11.609207699080171`}, {18.46074231559353, 
          10.878158460183966`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.725285377617155`, 11.811607900361398`}, {1, -1}], LineBox[CompressedData["

1:eJwt0w8wlHkYB/A3dXIl1ZTa3LUsdnKuzbLO6iq9K0J/haRu1Y4/JXuupSsb
OkuxSnVr1Fi5ao3UalUqbYs9QyqLEyXsYtjmuqlQTFyn6Y/7/q57Z3be+czv
eZ73eZ7fLCtqX0isBUVR0fiR9+fHjf78dqapZ72XIwNhSVtV13z41KCbOh92
sOwQLIZrxR7RA8Qe0sMMWNl/IXs+l6biHE4ILWH2LQGTD+vSNeueO9GU2MyJ
DIfbo3YG6OE9PmGbEmHjlWs92fBpTjdHDo++LZ0aBJtamxMLYK4klfUFXJAa
2V4Mq44eNP7uSFMTrk5rLsF15fqxA3D2jJ3TS2GDLNqfCzdsERSdJ/kLPbNH
WDR1smqjLh9WzFV4aWFbf6v1WaS/yt6MbDjCJz7uZzj4xC12FLxKcEQQBStv
WO1cBxv1PbbBpN9Wy+7V8EjuUvZqkt8VU+oLZ55yKubCc3yl/iHwpOftC2wy
r79tcQIs1FbZ2MPmhe++yocDczaVLSbnUu6lOrjNL22EBeckqzvG4LAf0h2W
kviwvn++wTzW5nH+SnhiV4yfCD7TNfP1Fpi+YvUyH84b1FX/CKutXvLuwrOX
GXpzSXzTuooXcOMteeQ1mHqp8SX7FLHV7A6YUdxDL4DrHbsnJ0h+y8ejduQ+
4o9omO40JfOvTJkL8+M5g74wxb/Nfo96r1JUJ2NgkaSUZYRn2CRlZcJWTU5F
ZfD1LUuHimCDrdA3EQ61KbSugCOE1Xx3eMW4dn8tzGAUjQ1h3pTNn4YewHSZ
jetF2O6V7mgTqd/zRiOE49JDFOTccK9hLQOOlJWpSH7ESP5FkwNNyQeOy27C
ZnUhpwT+3rXEtQRuT7w6fBC2f7qhNg9Wnh1LDocV0/VJ6WS+0zX9vvBM5aJj
CWS+4ehzK+EYJccnEq7jtBasgdOzVug3w6OMIrtt8PKMm7PWknk7vvvzAHyV
FzRXQPIb3JadJ/GMWRJiSbW4pRV+WNF1JgBWiRbzLdB/rUFaHgqrt2drveHc
/qeFsfDE0yXJP8FqJ/8ph0m9xGnXVfCUyVUBhfCLOzL9H7Dkba2yilie7DcK
68bSxvvI/OUuG7/Efuv3iL0sPDB/+cOwBXAm64GzC2w1yxRFHN4975eN5HyK
5TwSL3fPmJ4IG7P6Cl+j3iBz9Gwe7F0aercR1ijsPK/But76Z2fgGwEeH+7D
FUFp3Ttg7zJpqBGWVnWayP1MPhEZn8GGhoSUx5i/LG5QPASbX/196Di8nTeV
N0ziLWo0ZL9PvjWffg7Ltu7opGBbncirn8RPc2Ees6ep3dwbb9pJv26BGxxh
0SfZozo4uHlw2iMmTbnsWvme9GfOq2QpYUlSu+Y3cr4tJ/YQXGlUOOeSfr/e
E3cATv0wcC4VppmXFblM8r9etXwfrPBrW1QNdzKE47thSaPB5RP8fKTlXDSs
ksuGQ/H9wt33W2JJvQtHmrRwSX9SZwIcWBTSOAf9l2Zat5H6KmfXiAD4r+TR
nF/JPps/rkmDw+i7eWqYq13fXA47eh9a8QB2uNgm7IKbCpxLyD5cTFX6t/C7
vdY1M3m4P96O+dbYL91VE+4OM24usraFeXyx9zZYKn9smgNnXebap8GBjO6g
SeTfqwvrPA9XlNP1ZlhcsMSnFlbvFS/Xwvuv3IkyEQd2eGbAWwWmtBE4OOvR
VXI//z2e/78d6H8BUeA4Uw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.440988451379745`, 7.548724015604422}, {1, 1}], LineBox[CompressedData["
1:eJwt1ws41OkXB/CfUq4xVC6lmhW5MyS1XcxvpdJGRQlJhqI7KqFiGlEu0QpJ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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{14.681479476010207`, 2.8029766133560106`}, {1, 1}], 
         LineBox[{{7.999999999996362, 14.500000000006366`}, {10., 
          6.500000000001819}}], 
         PolygonBox[{{8.854478624978201, 11.0820855000872}, {
          8.757464374963666, 9.820900249898267}, {9.533578375079932, 
          10.014928749927334`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.101079124774735, 10.124069781193684`}, {1, 0}], 
         LineBox[{{15.500000000001819`, 10.000000000003638`}, {
          10.000000000001819`, 6.500000000003638}}], 
         PolygonBox[{{12.243803107360735`, 7.927874704684104}, {
          13.041446695762001`, 8.90958989040874}, {13.470947089516528`, 
          8.234660700223053}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.522642987100603`, 8.952875305984765}, {1, -1}], 
         {PointSize[0.04], PointBox[{8., 14.5}], PointBox[{15.5, 10.}], 
          PointBox[{10., 6.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bfcgdgeh/fhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bfcgdgeh/fhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwV1Qs0lGkYB/DXdENyb1yrSXIJRTUlWd9HLBbtUAnVaWpxRqXcwuRgumzZ
LqLSPbIkc0haXaYWkcolpJRxtonZbK0tiUSE7P/9zpnznd/3PM/7PO/7fefM
3K27AsI4hJBI/OiddE/gmsuSPnpfxJKNAbcnKmHbQ4oDHHuWGM40cJfAovfH
AyfBGk0NbR7UR7V6CTzsn+GvA0sshPZfUC/qz7/ZxWOJLLrQWgG3jFb7VsAC
t1dr78I3T68pugTXlUjXH4UnG5fd2A8LZ1d3BsJpRq4lkbR+vV2CCRxXsy0t
GB4+tcHt9UL02W/21R0ujIsuyYGFq93L7Wi819UwHL45q6NDBxYFvo9bAss+
d5V3zmFJ+5ahyepwmvpo2TFYtaTd6IMdS84W5frZwY4NhQNtcKH5Y+n92Szh
zcu91wjbK0SsJ9zybUpWAyxpjMmtncWSRE9neRNcNcvM0AMm9Qu9W+G+RzEr
H5rief+gWA5brZLHe8OXs92NqJVLNp1pN0H+tV/9nsIt0neiOFiW0DNeQftz
UyNnw907B1R/hx2X2N5uM0Z9RtWWJLhbg//3BTiqI1bHB854Md1lO9zS/IOp
Hs1X4eu5wmn6I75ttjj3qXoThrDhznM9Z2BWvSai1wj91zTt2ADLDiaNlcMS
TrKmORxVtlp3L40POR/7YoN1Qju4jnCfjnd7M3w2OLReaYh+Ti7KMtiR67op
CdbuKk7Jg4VeNXFqsChyQnwRTsy6PnTMAHMa/HyF2v5S+wMNuGplj2EuzR/Y
6XeYi/UbFckFMGl0tZkOX7a+GiqFS7eUnj85E+f257lsGmd5W2dawsMrUjMu
0Hjs18Fafbxffkj8QdiK2+GUACvveWaG0/kjUnYshb2SlE7OcNVzoXRcD/vs
5ClUYZ6nlvwpXOf8sK55Ab6L7PLPBdTRTziZsNdve8UpcEt8QE8AnME//9of
FvDzPxnAgvKKMVMaTxdbdFlj7pEjRxS6qOeGq92B+1QWbE+HvVIuL8qCE3dv
VDrAwqyi+SmwqvxQQa0OzstNd3O0Nd2XVOEP1y2TukfCwk/85lZtnFtv16Zd
cGHfnkv+cOEb9cw4WLs6pDRHC+/BozlWDPNirTgFmni+flr+Hlj57SK/cgbO
TX5lzm5Yku2Y2quBfSQqToTT+oiUXCc48cnFU34wcXQ9kjcdcZnLE2s4SlfE
WMKygfODo1aod8lRrVbHXI+K5I/gqPjbAxFwxvEijcOw/VvmgBHsGBRc4wNr
txplPlbDd5u2arsmLDydulgEl65K2fDSEh6Klw6pYv/XXyblwYImE+sdsETf
d2QPjT8PiXw0DdZLn7EJ5pU0a3Fg5V7jTm+YtTU7YT4V534oOJmFqyb+M108
BfF81REXWp9wcsRhMvpPSE67wS0Jk1/aTkK/5Ap9LzhKaNFhw2FJ0Mf3Kr6w
RJORL1dB/o0efR9arx2wOIhgrn3399F65d6fcswnGCLx1PK0p/NcM/rn2ThD
HGW5QXow+SWk984YQ7zEqx5+sMA8CwZvtYwyJHEoqege9bLlGjowybG23AeT
KTGVEd+wHsPR8YCVN7/w6kcYolqRencqzFv641+mcB2/MrNhPuIvVarWDDNE
1q32/CQsvP/x1OavDBkWOliEweQCJ2nZEPKPGotYWJL3YqD1C+arL9hlAbPj
getWDGB9u7oxLlx1eM7VyH6G9JWqVGrT+s5Q+UgvLJDG6ND63pqWj+8Zwv4b
PZ/ms2Fvua/ewRIfsQmtb3NZ+u4N+q9rJLNpfb/N+i0dDAnqdDtjTOudfM2K
5NivQ/gfmjS+7XVW0zPkPwjOHzaH14Z9L65niPZ3Qa2c2uPW0cRqnEdsaHox
dbFRdtIthghcusSJ1PnvzgbnM0RpIuj7gZpekgoyTP+/5rH/A9AjhWk=
          "]], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.7549853041491055`, 12.315122043776341`}, {-1, 1}], 
         LineBox[{{20.00000000000955, 18.}, {9.000000000009095, 16.}}], 
         PolygonBox[{{15.090321946059944`, 17.10733126291999}, {
          13.838123878660062`, 17.286216701119976`}, {13.98123222922005, 
          16.499120773040048`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.222727570790026`, 17.920198360654858`}, {0, -1}], 
         LineBox[{{20., 12.}, {15., 4.}}], 
         PolygonBox[{{17.182000635998094`, 7.491201017596947}, {
          17.47880004239987, 8.720798558404326}, {18.15719868560394, 
          8.296799406401782}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {16.790402628792116`, 8.21669835700493}, \
{1, -1}], LineBox[CompressedData["
1:eJw9l3tcTGkYx0epptANKzozc+ii29Q5J0Q3D+1WhIZCqVxDF11MkUrpokyl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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15., 8.9452}, {0, -1}], 
         LineBox[CompressedData["
1:eJwllXs01HkYxr+mwboOuVdqYtF2GdeiVPOzywklpQm5jktSOiu62LZ2G7da
WpvS2raUSTdCUYkilKQ2yXHpIjROF84ibUvFFvu8u3/M/M7nvO/3+7zP885l
Znisz3oeY2wzXvT8/03EMR166nBs/FG+rw+4NsJvkRP401KjR/nEvhd7VoCr
Z/zqOQFedSPvhDM4KtswcbU1x1jNQCIf/F531/5jYE67REUi4NhbUSzrAMt8
v74cps0xtw+xJso24JHSQLEWx5IzqkQGYKmFZ8hnDY61lJc9nwYWrv1x6gV1
jpXtXCP5j6cuvx+pxrFLN8IW6YNrRUOuDl9w7ELO9ut0X+2WpsZ5qhwbyLD+
4yX0FIH9/4SqcGxHbLFdBVjoc6ulXZlj+UeqGtPA0q5MlWzwqifKbcHEgpDD
eWBLcU2QE/U3FMk/gs3iN+maEie8PJOJ+5IC3D9oU3/2njch0ON9575Eh/x/
4NZIMM9cy0xjM/L7g9Q2FPM6PX9a4kp10eS0zfAjutdvmgCW9y602wi/ZtXe
O8qJ46/ofaPJsZ7d1glK8MONxHv8CTaa3c6bQ/4KKy19kBczHwr1pfrBWfx4
8JTIuvhEqtd7NbqBJeLlX54nnmS8tAnnVTbNSWkCy+8UKnTAwwOOfQOUZ2t6
gRH0z1onq/JtMa/V1YC3mDejRyzQAyviDE4Vwo+lmVmJCdW9/QVB8BuUFmVu
DOYc0zr1kYekap6VgOrm22v6+ByL9Yt+N0H3j1Zf6ZvEsUGh8UzSk7n2DpiB
O0asjrbTvs/ee3WSB/3W5C3V1O8kqo0B+zltuJlPddmSzkTwKzXlQ9nkp9Wy
TgEOl129kEb1hZGhMtzXfKI+I5l4koZMAv0Mu2LHVDATPDrhhf2Zt3ovOEB5
9Qd0rsO8C2QLfj5JejWbzKLgxzVo/ZVKuv+AXUAo/HrcuXe+E6yw3Zxnjzx+
TL32iUf+st7EtYJvnE5QzAULC746bId9qkpcfPyJw1KG3MF7x9TnJ1N+4Tw1
PfDbwKfbCsHSkLbFh3DefelPsx5Sv5ZzfiP0vIyelgxQXdXiTiPm4b5tbuHZ
QS/hyW055v1bqey+Dpht3ZW7Dn6Gk2KCDcG1iunq6vAbaRnVokd1txaXeuTj
mNL3UQ2s6NVoO67EsTjvQYMx3M9MFsedZhzTlCU3vwLX3vetfDAhZjPa3hnS
PNzNIvPucTELiQiaKKf+mH6bVPC4WvuZPOoPGoq+DbacJlQ/SKziUpCF8/7n
FgylgOUnj7Zvwf1uNuer91Beest+8YS+MXc5Lom4K9p0GeZrtZQmZlD/MXmM
BPtLffGyW055eQSuDoWfF9K04UrKZ1CrSQK/17daXewiPcfPi82Rh4ZW7gG+
HX3/3m+tBDsXW9SJwHKn+mu6yO9D+msuAMwFV6ZMB2faT+5PAQvfO4x0o79K
7FpaRPmoLOevBYeXCS49pPt8ayP2QY/vlblugDjPVjcR80wXC3Yp2UP/+tZT
qzBvUPSUSG2wfObOYFX4CfZ/3DuZ+FmTQQX8T43uCBNQv7+S2gDyWimUlPCJ
Nc8ZnPskZgk2wxbDtD++t+6tMTHbNrph23Pi7ten/UbFbKywJfMu7f9ZfOn3
H8Us5/h4Vin5UZKazge/yVlxJIf6V5Ta7wNrbUz/bT/15yzK3YDzRXc7PGX0
eYhx5N3G/TK1Cu3d1N9xUpgE/eJpjxv3EOtL0xMxn4eLzyCdV1QJ21di/mfO
mqvkpKdvremP/SV1lK+sojxMdFM94Hc0JKKhm/rXOHTqIo9rs7pEKvAnS5d4
5oEfRDYU2IBZaMHhIdpvRd3sIPIv2Fj3DjynoTFrL1j6l2pWAf0exOU+Kabz
53i/G4IvzeILm6n+vtPEBXoBdR09/VTvYVkOmEeJ/qccsE96Mu5fS+RWFg==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.075655548348863`, 3.834338870872152}, {0, -1}], 
         LineBox[{{9., 16.000000000008185`}, {10., 8.000000000010004}}], 
         PolygonBox[{{9.425579159246475, 12.5953667260282}, {
          9.177509690068058, 11.355019380136117`}, {9.971331991438992, 
          11.454247167807484`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.566769291093816, 11.807746161386726`}, {1, 0}], 
         LineBox[{{15., 4.}, {10., 7.999999999998181}}], 
         PolygonBox[{{12.031478714334181`, 6.374817028532655}, {
          13.218399304687589`, 5.937530495244557}, {12.71864326664405, 
          5.312835447690134}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.863477323709358`, 6.605546654636698}, {-1, -1}], 
         {PointSize[0.04], PointBox[{9., 16.}], PointBox[{15., 4.}], 
          PointBox[{10., 8.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuCF1BjAx0VQIFYJATVAwQiKqLT3llgFbacUp4YWBMQXoiBF
xBilBkERRocoLYKKjfUVeZmqKKg4EUVRBGMpFm3QDJUYW5AAClRb7b/jncns
fHP2nN09e3M9EjZErrIlhMTiR0cy/A6PiHs/+nIkJJpZWwcn9g3wnD/iyB4i
2JUJC1z0i8TwqOnzK0Jh/WXtkymw1uSZNQZu99LKJsIRPvZZ190xv6jkGYHD
/GQLFHCqW3XC76jPNPxoCIZV3t3LjsKWxuU8G9h18XDQChr/ZuOFpsmIS5vW
iGHBwtlJhXCELPDWUx+O6JR/jFwC68e3XjkDm3o9CvmwKOiznFw4tVFTXT4J
888HlqTA2p5kJhQmFwvjV8GJbUx404cc4Vw7hEk0P2Pw3iK48ZK7xxY6vyw0
sWEiRzQyzcq9sPXK6iwOjvLf536U1g+dueWakCO7XW8fqYIVd6NMn8OiFz2C
X2B17aNHXW4ckaR1jTwBa2pjzu6F29clfZwHF08Xn5ZRVxnTYmFR55DTCFga
2JA5ja4Xvvlisyv6sbPFaPHG+cOeRxfDgtZqoRZWZRdnx8GKouOj1sGmTJ3P
ZBpPX86fDXNmRYbBBeOwt4iBrSG2IemwtbH01LAE+57eudgOVtnI7llh6eHT
in0TUM9Nc2oQHn4YIGdg1TZjAh/5NYuV1/LHoy9tFQ+FsP7SBD8+rEtZzfrD
lu83dec6Y73oE91zYfXSp+nOcLzc6BVC9/frnfoqJ9zHmweX6Xxr9NI+Odxu
LspxhKNetTmNhVU9a5VdWL/muMPS++Owjqr0SCVsNdjFHYTD6jMLUuDE8knP
V8LMgldmP7jRyV7pC+vifsvom4a+yZPFzx0x/puReRUm5hk+B2FVdf6LYlh9
XLU/BGYODFl2wJJkRbSBQR/+vLF+G80fPS1HTp3sEp0Npwrz88rHIq8tXF8I
My/T3TRjsP9Km1gtbKr8uuLAaNzzocSEy7C1wav58Cj0c2Zu4C2Yy5b41Tgg
//67AGp15Lz0XnuMJC2uBjZckG6dD0f5bq8rgaU2XzlU29H/iVC5nu7/u1rf
ebDWnHEtkMa/YPqbBTivWhLxcir69NPYSwmwwf9E8zlYZx9f8mIkxjIdXwGr
LTYda2BOe068gDp8VswtPt6X3GO9HrDKfCTNDlZIHJeNpnHjzu3+I3A+S3aw
gLrlbvvcD+CCmJkMbIpJCQzi4d6rvgym+dZ+3yIfW/RtxVbrLJjxDngjtkHe
TXVuGI2n7jR4E7wPQm3aEjh+v5dLwVuWtMt7AiJhTvoJj/cfS7THqvs/pftJ
7fn27BuW7D65x0Lrm0j60IHXLFGoz6cNTEH86Z3wU/+whNyQP6uFuZgsx5Zh
lkTk97srYVH9rkeDQyyRHHp8bw6syTL85QDH/+xR99YLcUHenNevWCKacGaw
GdaPkwjOv8R69096lsOks6I+YIAl+hYdr5jOd97QtbkP9YM6Z/xA4/si97v1
skRnlx1QSvM9Kmts/2aJNLT1dhXMLRzobjWzpHH7UMF1Gk8qFtd2ol6GsbKN
5ss2el7sYInhcdjpJ9RVFX41D1gSPEXm1QGrHs9OLjUg/qxsUxONT+3Ne3IT
6/fO4JdR59h2x15hicXIW6WkNl5NHFEOX1jkyFHTR133/nvuyf0PQQYoFA==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.194814593688687, 12.79766898841154}, \
{1, -1}], LineBox[{{20.000000000008185`, 18.000000000007276`}, {
          8.500000000009095, 14.000000000007276`}}], 
         PolygonBox[{{14.816698078023695`, 16.197112374964764`}, {
          13.551893671999796`, 16.1806863437177}, {13.814710171952814`, 
          15.425088906352773`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740793031341028`, 16.859170034894547`}, {0, -1}], 
         LineBox[{{20.000000000006366`, 12.000000000003638`}, {
          11.500000000005457`, 8.500000000003638}}], 
         PolygonBox[{{15.19519314091517, 10.021550116847424`}, {
          16.152506936983112`, 10.848321122542462`}, {16.45710678118655, 
          10.10857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.159837801855844`, 11.07845105263581}, {0, -1}], LineBox[CompressedData["
1:eJwt1Qs01HkUB/C/iLI5pLwKeeWZx7CLyPRv2tqSmAqRiqZIoRfb8aht9GDX
EdNTSVEzSRSzW6atvEJbbZKIJMkWRYUphU3Y7+3snGOcz/n/f/d+7zX/YSzY
ujx0HMMwUfih34wSvdmzzLfXVJaJnOC5UAfWaFo3KxdO7ebx5sJ1h9SD52qx
TM1ZJmMnnJN19GAVHBlx++MN2GGT9KuNNsu0FjGZSg4sEzK+tCAaTuVKHtjB
UtdESQY8KSldcRXczo+2Jg8Uyafth4UFcX/R/V6rGpUKYA02yo4D61aNddTQ
+bDOiY3opzR6yL7LgfKMZoTBTfNkt4Zhtq3EvRd5RYk3+5U5yLOaHYiBa5WS
xBNh+S6TJ4o0n2KhsxIs5MkjxFNY5seH1kcHKa9apUsQrJ/Vp9UJiy5sdefA
5rbB0bVUv7jxmBWsmddzvpj6XzHmLYRLEiX3smgebfehJDjA8pfhA5Q3l+V0
w1/O/O4SDfPF9T7h6D+0R8UtDM5h8rpHYEkBRxACG4lbbM9hnnDHF6cFlEea
us4L8xfK3gZG0X4CzSL64UUeogQhefaO9Wk6LPPYXkdwiuqZXzPV0mUZ3rwm
eQnlW1F0RwiHhH7IeQXLxydPr4EX7pedVMP87XbPe/vhj3en1LrCRpa5Pw3C
Xcwyg/W0H9mt7U/ggHSFn1PoetCStxmwhDP54yVYtLnPggsv39A6epf2bVE2
/xHyqJ6NiXxO5/V7kgPhyp7dB7rhuobj714gf8zmOL0emF92bV8ULPG5t6eL
6mkv/6gKO0u4h1vJ6uWt5dhH/Pnk7L+p/5qK7YdhS6OV/ldhxq9bORlmakX5
p8hb9jZkw9xJ5rK9dH7E6UErPKp4JCMC1hjMz3NH/dWrNfkBcMVW29FrsKrg
nNiT6vs1py9BXudnT7/Oh9kEi8u9sMTF238B2fX45RO0L+tP1j5kB1k8q8cy
jnExW9bB0nIT4+cwe7vvYALVt1ocHDaNZX7dOjMuiz5/ghlWjbDRvogrlZTX
7Iiu2XSWafkzPOI95ZtaWOUDZx3TOqHjiHpe6v3+8M6NP+jxYCM132I3WH6O
fboJbk+f7T2Kekl8cW8qXJFwYNwFuN3V48ZFun/y5yp3eG2Kh0I51ftje2U1
8r0cLmVqYKY+utsTzl3UGVcHb1O9voz+3mEqdcV0XVpqI90BD3UqJN6i62uV
75vBlTPyDKV0PrB5pB/7yeHGfp9JeY6YPe6A0+ojpImwvKGr4l84LHn4Yjgs
jBINOuK83cMqdhnswKsIT4OlJbnxHnCdpv0rFeTZop5sag/z1ff4n4SlHtlS
C5qvqc+J5nn/1cPKEhb5d7zugA93BGlzYI07N9ccpP1kasTTvuRasZk2+nj+
l257GPTtuh3vGuz1pkFzF/W/vsLEygD1O1Tqz8Ls2jRxAjwuNHj+feo31GdS
ADu2uwkGqJ/BFxsZ7PnOTd3YCfs43ZR+Gp72Rt9xMRzyfmnABnibwcxnkTDr
45WhDpe5KQhSYCnvq1Yu+vvKnIU5MN94b5c9LFeMDS2k83f0eoqR/0tdVvUV
Or8zmzsPfj22pVkKi+b4iVswb5sbE5ILt3t/YFJoHxWVh4/BFYcutfnBkdam
akJYmG85woMNHZplG+G6BXmqvvCc84kSb+p3kx/7G3w1O37Ameo3G0T/A68S
9clMYA3dISYA/avHfK9PoXoTDNk3cOvnBbu/g+VtvdXJyP9a9ZQ2OaR/rgYH
8x6f4KpP9+esfKHdAnPHic+YknkNj+IM8bwJh83n0HzPAmtVZrDMrDRBWRDl
Hyu6ugtedWzNCOVvH3EprYeHvLkm+eR3AcqKRvh+/GRl2wgvsnkZqwqPjuHl
9P//LyP2P1TTb9U=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 7.184877956223658}, {1, -1}], LineBox[CompressedData["
1:eJwt1Qs01FkcB/A7TKs0MYkeCKHSQx69Ntn8V1LTEdNziml2stk6pfT0iGWS
2NSm1YNVU0oPoVB6UAmh15okdJqKWc+yKntsmzLMfn97cs6Y85l7//d+f797
z8yYwJAlQXqMsU140Tvj0z8njv3/Z8qxb0fknFsOK1aU/20KCwP33MqG857n
C76BF9R5DeiHNX0z79QM41jtVuX39s4cK9HlFkTB2hj+i7mw88tVKgO4vE+6
Rgof8nz2NMKEY5s/3W8OJksah6uHcsz849q3O2Bxt4fQBbYxLIoldzW8L08Q
csyLZXvT/BKhu0JlzLHMEKFoNcycu3h5Rhzzle8dJIJt8me3lQ3hWKV/dIAL
nB56ussEbnoyXjSSvF9slingmEN2YauO8i9b/GUvPC16l7qW6rVbdScb9knU
eVyFS5RNPFM8zzFbv5PUn4gs72uwpawyNBXmOuf8HIv9bXqMp2fQ/INT1RuR
b+XhZF0xLB4T+8YJ+Q8k9s3+C0731Jqdh5Ob1PbDqD9+PinGqPfNyOaqGbC8
c9HKtfAV04ouGaxQ1icp4WrrCuU+8ugvk67DesfTbxdQP3o+2GbDYosexwaY
kx3MiIRvjpz8Tt8FuQqX2k6Ap11JcraHhfeneN7A/qZBH/Z+B3eZhadNgkMt
zr71hdnBsFvmyG+uX1YvIV96UPUQ9fqWyf1WwvJLjetC0Z8jXeGvxLDCqj/H
dTDHlsxdtdeTPFZWaWzIsRpn97tTaH6c4qnlII6lhvNXjKD1Xtp8CBqI/RyP
+PVT/gtPd3cbYF5pwuNW8nhR5kPYWyP58wmsyc/rbYcNzfuji2n8tuCYCM+n
9zgtvEr1RkrkrXDgDfP4XDpf020Zl7HfOL6njsZLMtprjiLPiZ1l/SXUv6Dm
uH3Ia3+/bF493RfbjoQw1KPIlQ7tpvFPhgN8Ua/R6rvjhiOvxnr7CS2cmPqo
jfrFmWdu2Inz3rChY9168u9WbZdgh/kpRw5Tvx2kCadgTdQRnztU74JxLgvg
qEPq+Ga4pGLOkiysV+wm3MZ3Rd1Lk4v/wP6q+P2/WZFj5nkWIZ+wbnCdM6yQ
B39UIL/sQf5FN/KWyQEuqE9lP07fnfyvr/9r9EerXC2dDttk5VmnfMOxrIx3
ueNpvUZ/bcgAjvG1wetNYfZ++eIwPsfKul+I+ih/77q0Qn187hqyvIXyvuJf
8IBrGzMvPqZ6smqGGMGJ/uHya+QiRbM9rAoo3ZRB862jC2Ngad4MRQqNO55t
McP6yUfj25LpPkSFGLyEvU1idx2j+3DB8fld5Pki23/4DD0/S218FXlzjtvx
btC4b1TCGboPqvdDntF4XYNfFM43a0t/4kfKe/1YrRPqN9FdbrKkejftaabz
FszsdJlPFvskvYeLIkU/bYNLVhe9bofPpBn9cALWDDaamQqXFU8wvkf9uyBL
4sGTq98oW2Bu0eb5E7HflbKQOB2tt72l0hJ5VojMzptMxfNXFtS3If/JKpvA
0XC6fNS9VNTnv0yltIbl3tcCvNAPgaKnYhTM8cTPenm4v2t8Hwhghfe5U9X4
Nsl6vpv/hfKl6VWc1Xkwn+iAGc10PgGdGot+D9YlMfF/RHmexA8c1efB6mvS
D+TTeN0KyXmtB+v4UdiRRj5Z/K4cvu1ct/EXmj99YGMw5idrFz+MpHHD0txf
sZ5bb4Q0lPb77HXZAPuHR3sJImj+TUXnW9ihWDQnjvpxOyLsFfJqeZurU+B0
uz6nKj18P7kzi6s0v+rRsCzUV3BuYm8t9UehlgaifoHYuLWX9pN0H2iHeyK2
aMdSf063LZuCfu0oDXBdTPVv1RM5wVZiXVAk9dNwl4TmL1mojD9D/aqRf5bA
lrkxBeXkWRMsErCf2/4GmQYusbMujkYeVa9a+g/Mcu6t9OF9/d2a9vWdcf8B
K5Fq1Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.106724558493237`, 2.232686837739852}, {1, 1}], 
         LineBox[{{8.499999999996362, 14.000000000001819`}, {
          11.499999999996362`, 8.500000000005457}}], 
         PolygonBox[{{9.712688721591656, 11.77673734374863}, {
          9.936153049242591, 10.53172180397914}, {10.638469507574097`, 
          10.9148035085236}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.244061861982704, 11.112579197445111`}, {1, 1}], LineBox[CompressedData["
1:eJwd1QlQU1cUBuCHsqmIQVEChEVFCAhuLAYbypNlpDWsA5KxVFYFIWrKIlgV
scaKCmNErAGRRUHj1JSAqDBSQIq1FSooaQ0iCKhMqKGgRokI0v+amTeZb87N
vec/72Xe0rg9YdtnURSVgot8U+Mz+KyiKSaBKU0JUpw+1MP8NrcUXXjqGjM3
FZY6RhdYw4b3D/26Cs7Zn5cfBBeKV4vHXGiKSk4sOwPLr/b/eROWctO4ariF
V/3uKPF7e4sti2kqvMosdBuccyrc/wEsqbzTTMMB7ynv0CU0ZTSubHUh+50q
sO6F2wxY6Svg3OsG9YlmNFU+PCBygMVxzW1v4K6rSza6wmw///lZTJrSNtv/
tAmWJ+QWvIVN3VOt4uGQZyN1seY0ZRej6ySCtaqqyGaY26tRSuEugdfMHAuc
s0+5vBO2Db414g0n9OZpNDCzqsMmBmZYXvBkIj8Vws9OhkX+WZMeZB6TslvR
sMKxLjkY5qS28zbCnMTyj9GwpPtytgmcvjw7JZHUJ9mcbpzP77v3MAFm2//8
+jhMHTKRRpL97YxPecIlc5w6vcn8jW96DiOPtIkntiL1t33rC2Fb0y3Fr0me
6q7WTXB9pdnHRjjrm7I+XdiPbbT4MJnvXyqXTsxLMdx52huWDK5VXoVVNYOh
k85Y77sy9hzsrLd563WYkWFytghu2T/UIoBtHZh7a2DN0ZFzbDhAL3/dM1hQ
2dGjWgmfX7h8Gc6Lcu+IqIaVtboPyf3QNdxcehBmVtzTHYQbn+75EA5zhHt7
I5Gv7kt5uAcslFk+74GVV5q+WAYnXYrhxpP517l1M2F53M6WMZgqcnlsDmvd
Uhz3W6Ke2X/BDhbnC7k6LJo6YLHImgPHWNe+OQgPiII/hMEhH8fLR2E3c9fj
afBARKhloBVNsQKLx86R/q5EKorgSqNBdhM5r6nn3UN4jccZmxckT8B4qgbO
k91vMUD+nAbPhTrWuB/aNIYDLD0xV6hFXb7jOpeG+QcztL1weoSsOpTMi72j
RwY3Tsy22UrWZ0aMCuGA7WkSPix8n1jGhp/2MU7yYAlD/74S/cqnH3W4k/uR
6nb6KMx44lxiCsf4sTa4wvzAaxdG0F9uSZDOMOZxrbT27xswY2Lr6wpY0iAq
/Z7kKQq8nQx3PV903hOm1kkKfGHRq5nLGifkbL07fzWpj+1uk8HCFc6fbWcr
MEyEVUMFUj9Y65P/4zI4YPoHOyEcNVPFHXKkKVooHZPDA7uDZ67AHONoqR76
U383KykDrk9zcxIQj6Td3QzH/PPk4iC8Rt1u7Ex8sVkWh/wChvuTJcSD6zNe
wXwf20gj8vvfpx9nYt7hazN4xJwy8enZNujT6I4FWa/1UsefgEvMg1rZcLpB
jq2uLZ7bdJXaB57yZTdw4Q7f8ck44hdl23bBA+mei47BB44H3imE1cVW7jK4
cB9LXQOrvBbwumGJ9ePANjjBecx3AvY7W9PUDvuZf2tlhnlMBT44/AecZOSl
sxZWvJH53Cb1mdowX7i++JJxFdxl8rKJBzPX5H99DC45/O9Xnx3rVhUPK+x3
8jfCmgV9+qRfRdGRGyvhys6K7IVwnnl/4Tw454U06hHyFgoviwfJPDKDUkQw
a7rm0i+wvEGxwRu283hlkw7nBoUe04eN5ggqXOGQ4faCPjLP0dZP/7HxP6+I
17sHUz5THlJY3Bze3QaXJ9lsi4XpZ08dlbDQvz+XBYdM1MfrYD9No1d9rwOe
h1qZHm1D+vt0txwe53B/K4AN9Y+E7YKzoiJOTsAMR3WdD1ySvOmBO/Kw502O
LoWFcXO7MmCW2Lt0HswccnjZAH9+P9ljH/J+Wkr/D3mMX8U=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.084117857803797`, 6.188306123204339}, {1, 1}], 
         {PointSize[0.04], PointBox[{8.5, 14.}], PointBox[{11.5, 8.5}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["3",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bfcgdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bfcgdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1Q0w1GkcB/DHe16KRLTkLcU1LCc126r+/zo6GsTmYhuycRsp7F11bcc6
yktTlLJJm9emZHrnCBdm43R7l7qtEatXq00qV97p1X3/3f1ndp75zD6/t+f/
7KxjTDJPqE0I2YYPs5L303gW0OTLs5gmnoeVba2wtRG9dB7slKajyoar2CVG
bPhNrmV8EKxKqZBx4SSnIb4VE38k5YUPrE/avJ870URx/ibbE9bo+5jXwyGD
SpYVrJhbeyoXnir8cGv4K5p0SWrnCGGRit3YClsEVyVQcG9OZ0IeXKg+Ecli
vn/Sb7URrgw/KR1zpAmH3e/hCE9E2t7rgB3EZU9GXGnSX/fB+DScPrnpxB3Y
+Ye0vJ2wKoRbew1em25rtBq2bo7wvgjzFj8+awQfsO8ruwy7mQuHlQ40Edez
LZvggTinB1KYE7FLche2qODlh8OqnQHPXsFbIxqFLMas/Wod9BPpzg9ttqeJ
/2ignTXs36M9vQ4ub9JpXwA7J9XkP7XDOV+WDDMOzDivlQpHmOpqWTHzR9WI
HGCVh5P8I/KLXzbadsxHf3+uFXUx/ex/v1cCO1jIzlTCu7p8o7xgAU/9TgSb
NLVwBmxxznEbqpfDov7tscVweU/ycj3Y/9al2EBY/LO6utuFJoNCqfeEDU3M
2om4Fk4tTEgrggVxRYoS2Dc7uMkLrrL3yZDCA5sa17WzkG9Sb0aRC5MnOioU
dvh7d2QlrLHJeqyah/twU7y+CdYN/cl7M2wdcPLRfThC88lAY02TeEmG7A2s
PLxqaAfcO5yzTBv91QYYe3ywwvnE3/jHBB6SVIvyYcFkxIQxTFbZl7jD1l9X
3f6E+LCsrJZ7c2lCz8+M6YMDa1eQX2AVz87qOpxvYXTTEzbL9JMdhHP3cHue
W+L+Va4pDmP6zztdeQoWaR2Kmw9/jG3KXg+n+0njXy6iiZz32WDaAg4dqm6C
Q/gBS6rgA/xVB2VwmGSkNQAe6NlulgWbDG1Z+2wO6nYFEwnsmxnD3g2nj1iX
ZjDxaoWcwAPDfX/lww17OsZyzNHPutn0OdhbM+UzExYnmL+Ww03dzSsLZmPf
t4nRPfAKHivPFg6ZPMsbhHtHMwUXzHCOiQkWE/CZmLENq2FBm3p8FFaGlO5p
McUcJUqpBva/v8XGCA6hN2krmPrX7NpiZ2H/Y4cbZbAgaOVcxUyaFMneyZJg
B93Io16wsquTx4FFOfW6pSaYr13oo8XkL01cqAeL60xbbi/E++Vdc4w1ZuYP
f1oBl7s9cqk1wv15GH0nHc7/w+PuuCHWjoZDO2Dx8W+ELrCndH21EJ7ys3wT
MAPv+c7xLduZ+KNDLIEBvK8lOgUWLPEZT9SnyQzD+OBjcNGx75PFelgXqGZe
hBVj4Q8ydXE/5qV4/A6L5CdDinQQ56PzYxecXsh3qddGvTwp5ymssrfdrdFC
359H3j6EXTPXXHeEXTMqNt6CGwY115MI9lvMKr8Ey904WgXTFH6/w0mZsJlq
winqM0VUhzPO8Zj+rJIven2iSHznLjtb2NPtqrvhR4o0cH9zf+6M/Xlym673
FBlQyrg1cHqVevLIO4rIT5dyc2Alay/fY4oiHMcLvnEwSTzQfGWCIuTXVG4Y
nJ+mpWc2TpGrdSMFwbDn5kVS3ihFFNPDod/BgkvZppJhioi4QaytjFm9KR5v
KVKVcXavhKnH5q7mv6ZIetgs3xNweep0Xe4L9B/QZ3iFsX7p7CfPsD+bSOVM
fRO13+ZeivgnP+QrGLd2+3s9oEiRZ5KmDabNDrq4dlJEWew358s8xsd7l96G
X81aUcDYIGjfaCvyp3EctzHxxt2yshqK0O4X5Muc///fKW7+b3Wm/wVrzFvh

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{5.445389222627836, 10.620169257907214`}, {0, 1}], 
         LineBox[{{19.99999999999818, 17.999999999996362`}, {
          10.999999999996362`, 12.999999999996362`}}], 
         PolygonBox[{{16.024494365672922`, 15.791385758707179`}, {
          14.781248461855625`, 15.558277151741436`}, {15.16976280679853, 
          14.858951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.352053456673119`, 16.25014377798838}, {1, -1}], 
         LineBox[{{19.999999999997726`, 11.999999999996362`}, {
          9.999999999997726, 12.999999999996362`}}], 
         PolygonBox[{{14.402977685874006`, 12.559702231412599`}, {
          15.636823801734392`, 12.838312644671397`}, {15.557220826517595`, 
          12.042282892503405`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.154230764482548`, 13.437507644825484`}, {0, -1}], LineBox[CompressedData["

1:eJwt1ws01OkbB/Apw9rCzoo1ihqXNLlkGrSSMi0xQo2olUuptk3briS2yWWM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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.647946543326878`, 11.250143777988384`}, {-1, -1}], LineBox[CompressedData["
1:eJwdlwk4lfkXx19LlMZSlizJ1USULQmlvL/SIllCRf+auhQpokiWpJcRIppB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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.154230764482548`, 3.437507644825483}, {0, -1}], 
         LineBox[{{9.999999999996362, 13.000000000001576`}, {
          9.999999999996362, 2.99999999999934}}], 
         PolygonBox[{{10., 7.4}, {9.6, 8.6}, {10.4, 8.6}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.0548, 8.}, {1, 0}], 
         LineBox[{{11.000000000003638`, 13.000000000002728`}, {
          10.000000000003638`, 3.000000000006821}}], 
         PolygonBox[{{10.559702231412599`, 8.597022314125994}, {
          10.042282892503405`, 7.442779173482406}, {10.838312644671397`, 
          7.363176198265607}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.562492355174516, 8.154230764482548}, \
{1, 0}], 
         {PointSize[0.04], PointBox[{10., 13.}], PointBox[{11., 13.}], 
          PointBox[{10., 3.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["4",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1QlUjWkYB/BX0o6bRm41rdosM2ih7fR9aREn0zJyouO4JSNLizbXUCpm
isKVe1RSrjalJBVFUoY2UQ3itnHLVhL3pEiL5v+e+c655zu/732e93ne5Zyr
7x/itVOGELIHP/omH2fwGLBknL6Xs6RCxulrLSz5+iho/i8scfJMehoH85v8
xJrUm6I3rIO5nqVXdOCmF61XVOFSPk+PC/N8rGv69Fki5nSdmQPbtX2+Wglz
a2ouvcP8gV8k8mfhlc5DLnfgUp3198JgXtOsskRY6G3j403zp1epusPH3JWm
rGCOhdhKDWaHzWu1YH49Y9O5DPP4ej6a1GOJj7T7Sy4slDuS9gJWeDCWfxAm
j+/OLYEF/VZZW2C3zNaCGFhqd+uTM1xcYJ+wHg48aLHVHm7SdgqYD8eG13k7
wHXZjGKOLvp0WGHhBsfGx5sshkPfZIT4wiKtpqFsHfTDt1HaS/P/XWJqDA9E
y09F0fjuAPdSbZZYyQkuHaL9ymuoO8EFkZ906XiVf6jp658xjx5bHkDrFS6d
Ogmn5Zm8coE5lpP3HeGCj8aHtWGfbcK/ZGC+2E/0YSlL2nkyHo1aqMO6/nEd
Dhw4knwKdm38LTUSlnSo+22CA3/0OtrCodIozYVwU3rO73Kw+FPemceaLPFo
WKvZtQROzIo/BLMXBiar4Com5aE2LB0VLsuFuRa6CRUaqHfG6EQGTI7qqTnA
Kx2vFmbCouHQvnouvm+uHr0Mt3tetHaBrfp8x27CCrzW3vpF6KdvVUUDbOq/
L8cVFpRHGj2l9d07FZ+oI++uYZoYbnLb0BUAS3vDtJ7RfkS23TLUbelt9+n4
k4KLxQvR/7ctGrRemvW6A34wVzmvLIbaPiJOF04sLjjnBnsMtkpe/4R9dN/W
ogYnJo8oFsGJMfvniU3R//hLw/2wleBwcCbMeqZHmcHilrjuXbDV85LGITXs
f031bRuYcz7tznk4kNl4nEvj3Y6q28Npu3tSZWCxYkJHxwL0MxL8btwE9ZVz
rHbAHJPdhhMw2xOdMaCKfSi69mw24gdCUi13w1ULlFMWwAUa/fcGOLh/nKh9
BrCgneHupTb28l5B65U1J7yfj+9eb09ZwqXOHx2V4bQBUzUz2l9ZiZHzPKyL
J3lH89vtVp88Nxf37K6Khzwcu8Nu0w8V5K15M/gK/cSqFw7zYb7nEmEJ3C67
vJHAsU5FTBR17tb+U8pYb0NUpg1MJvqFWnDd6x23ZoxxLrarfC8qYR2X35xo
hiWVQ/lqcKKZbGIGrHdd82y4ItZXut6FT8fv5e25o4C8WTfM/WCOQeXIsDz6
SYgI2Qzz5ijuV4K5L5MaqOusQ4PU5XAOpxufbYdJrY8ldw7ugUOESzDc7pvy
6yJZ3LPkFvMYWDAuq6I1G/EKLslJcOhJ3YmlMqj358a/U+h4dGWn2yzch3Oe
8gI4Njgu6xjBeZiofY+j9QNGH1yeYQhbMx2/C2ZV2sLDfzCk/f4ufQca/9gr
fec0QwpUbdI4sCh1XUD4FENEl4L+ERuh3/B65dOTDOFHCR0zYbJWfkPhBEPq
JJ2L/eFYkTYp/45x49V+y2C9bREdWeMMKfW4HTdliPo9p/P9vzGE13HbTAzr
teRmjI8xJNTUfqYWllwzaN4+inqDQqMKuC7yq+uFEYboqRQKymFRQFvzTSlD
uJLqsWqYhMg4HBhmCOeGn9pDGr917YewQYYMeI0rdcFsUAzvyFuGeDyffvie
Wkmaf6sP66/sjflM8+Pdi8J7GLIyx3x82JDu14ic/nOGmCatkeujznabOt6O
eL70eiONn+yW82pEvy8qfbKpdcwMhTUMEWRYZIdRWwimRcWIj+7zsqWmj6jm
//+jxex/Zul1/w==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.691266165654481, 13.045993650284984`}, {1, -1}], 
         LineBox[{{20., 18.000000000006366`}, {15.500000000003638`, 
          4.500000000010004}}], 
         PolygonBox[{{17.9397366596101, 11.819209978830308`}, {
          17.180790021169692`, 10.807281127576427`}, {17.939736659610105`, 
          10.554298914762956`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.36530281910303, 16.579218250098275`}, {1, 0}], 
         LineBox[{{19.99999999999818, 12.000000000003638`}, {
          9.499999999997272, 14.500000000003638`}}], 
         PolygonBox[{{14.16631627118878, 13.388972316383624`}, {
          15.426331939733636`, 13.500150169490523`}, {15.241035517888804`, 
          12.72190519774223}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.557648941662487`, 14.638733081970063`}, {0, -1}], LineBox[CompressedData["

1:eJwt1gk01OsbB/Bff7JlTZZCjShEcq0t0i9L9ij/alCZ3IpuymSppDRUdqVE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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {18.83483736847501, 6.669566028483164}, \
{0, 1}], LineBox[CompressedData["
1:eJwt1Qk41HkYB/CfYxu60KJQGsXudjyKQbuk+XcbO1nRIxtZR+tIkSWVY51p
Siph04WNso51TK6KGtG9SztJsqnpWCWtHMuWdez37Wmex8zzeX7v+/7e92X+
jLyDnb5XZoz54Yc+mSq9LeLYh9c0jtU63x1cC8v0d3Xrw2+UF5WmwmXXho8a
wq4GWa/byTEmferwo4zOSs3FHDuS2ck6tDjW4nJ8vxkcG1Ndlwk38WyGHWCF
dKmeEA5q8LD0p/jBK51yTY6lJ6tmRMFlRe51G+FOaci5ZNhTV5CXq8Exu7Zu
0wzKHxv/KXkq7os1Nj1OTp5sFDKFY4ZNNo+PwfxPz87zmMyxGaJN5amUb9Vv
5z6JY8pW/4n3UbzfRe+wiZhLpJewB85JS7crUefYmp69ogDKNwvw04QvZ1wv
3QjLUvMTstU4Nt/k+b3VdD5tOMkDdjh5stWC6jkneznAFi5xzSZkLz4LguXh
IfV6lK+7Y+NFuOph12EtOhdziaaov0y7YHAK1ds2NNYAq60aidKA2aNIxQ/o
T9I0KUKXrJdjaY7+nRb48OZSPcH+ze/gINc2VUvab75D8CXMq33i8QwxzZt+
6AztI6nIqsKP6hf8WqWFfY1l3p9O88feitmQBsfPbnQvoPpLnXzfwl9m50b/
Tv3tqvTmY9/ay8O8+qleYm/SHNj87PyXumbIF1eMDSD+7h8n+r6Cc/wC/02H
t0isQzfBnnabsnhwyY1le3bBrDBX3RH9cLxsyyMwf7HxlDD0e8AtnZdH+SZ/
j0RgnjfV1vFSWOb76rvtmD9xxdSQSxRf2pqxAfvJdNJpvwwrOjzCrLHPXnHO
aC3Ft6x48gUP8+qfca6i+6KUPOdN4FieR11SEcwtEdSv+gT7vmO9/TT153+9
PF6VY+9GzVkKzePYE9qlwrHXK0XrI6memsHJaFhW7RMTQPFF0j+XwytuPSh0
pX7jJT1W8I6efV0iuk9FuN8NVm4OtllG+e0vAothee70WCuqX2KbYIz7wt5v
3i740M86wzp4R/e7fZaUv7XXIgj9hbeNzrKleN9TtaboX2JbsNKenHqNNwi/
qJDI3WkfGWxhNeb9ZVQjMIz6a8jfGoh93A6X7Kb9KnZWHFPDvrIKzZaUUb/5
hyKSYL6vU62c7otuMX4M/+j21/gQ+bO9NhOx7/niyJcG5uhP/8AiHn0/Op+s
52CZq8jnHuJrgibM9YFZ1nmbYPimUPZNAuzprT+zA/efC65/lU35xe/FRnAL
P8a4xpx+37Km5eg3zybG7w6d54Xqr8E8xjX19g+pnrRJboX5q2rHs59Svft6
AbOwn8batNAXFN8Ud0EF+7RKGJr0jPrp4OUOKOH5kN/a2Q4r5rlkDzF8X1d2
nW8mhzR46MAT6gvX1VN8iuhhyZiQ1fq6l5XBsfb+V/aOCpm4+fTbLPK35wQ1
I0K2ZabmghTqR5Ql+BqeI4+Ij6Jza4OrQviN/1v1IOrHJejRUbgx3/+yN9V/
vviELeodqdwc6Ubn4cbhlrjP9GpVnyvV+9ysbtu4kLVlxufQOV959+EA9Hfz
nwvNtE9Zb+fCBZinRr42JYTySx9E98GxOuHFiXCOQdoxqTLtrfHGKerHzev6
FuxDrXL66guUr9M7MAYzey132qfCyJa3B/tTqDgbjtD9dzStf4OHV62P4wvg
27GeA3DURMeA1bBnuGR2NzzzoE5lAJzjtFNTCvebPBAdpPPs8mY7uPj4M8Ni
Af09Ow6V4z7pTondTQE9L38e60N/p+xalBSUH/hSWwNO6iqq6Cc/NW/VVPr4
f8Xi4yfj/gcg30hb
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.26210340484808, 1.8879461287265402`}, {1, 1}], 
         LineBox[{{9.499999999996362, 14.500000000004547`}, {
          11.499999999996362`, 9.500000000004547}}], 
         PolygonBox[{{10.722834405812462`, 11.442913985468845`}, {
          9.905774917833435, 12.408529743989513`}, {10.648556270541642`, 
          12.705642285072797`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.665661129127848, 11.42434445165114}, \
{1, 0}], LineBox[{{15.5, 4.500000000001819}, {11.5, 9.500000000001819}}], 
         PolygonBox[{{13.874817028532654`, 6.531478714334182}, {
          13.437530495244559`, 7.718399304687588}, {12.812835447690134`, 
          7.218643266644048}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.894453345363303`, 6.6365226762906415`}, {1, 1}], 
         {PointSize[0.04], PointBox[{9.5, 14.5}], PointBox[{15.5, 4.5}], 
          PointBox[{11.5, 9.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["5",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bgcfdheg/fhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bgcfdheg/fhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuAFRQlECIhCKUoAxVRoCwYRRHsvIhicqiFqBRwfZUCDIGYK
WipCqYw0UIsUrKLyUuShWAEDAuIDrE9GLVBrmY6SoBFSChoaIxgJ6b/TO5PZ
+XI2e/7dTeIWs0cSZ04I2YoXHckbEx43lozTUcCSK4IEs5twZJ2CuH7EEkn/
66ZDMHurV70EPvrU8+fPYUvPTXGhsLvYWO0Ik27e/RB4+8Dh2EE+SwIqDTt8
YVY2WdIG8/V8lQ28IeB8biEs33+Bo0Q/+6Yy5W5YpBjMqYKHZD1GESz1vbs4
AW78umOhB6xVuYUKYZ532wm9K0u6VPHfTIUf3FaGlcBFwrwNLxag3nhdHQS3
D6em9cCx61qWdM9liU/xuShqcYdStBWWlsjt++HejvvBL+fg/TkpNw3w01oD
Px7WPKmc5oz1l/2xMmLEhSWputjWAPhqxrZFe2HpjMf3xDTfiHnFFFh1gGO9
BY40lhef/JAl+e0Z/0bBsrH4V0thgaO/VRg8fH26m9oZ8/3PCtxhy7qh7KNw
jXWCZBT9a4Je7QuHUwPV4haat3P9OdMHyBOpskqnefNjzivgdq89l0NpXu70
9TFwZsETkyPMFb7358BdtpEh454YNyW9q3Jiyb2Dv7/RwKLJ0gNB8LjnzNVD
cGbJ4Mu7jiwpP5EsNMAuoQ1718JFeZoKB6zHcm4F9MzGfq7nzPWDvYmi5At4
XLtxnwT2E2gn+mbh/tNNlfGwtnrAIwn2qV1zJgW+asbTWcDyS5utZDTvd1+6
VDtgn3eqSqLhxt+mWkfAMvNHOxfD8ypqx8xg0Q7r4il0/ZFmz+aZLHEKjJnf
iXx8r6urdlObbhvyYPWx1X/zYX73xbWRsGXcytFH9ji3Jz9e9oIfz1/O2Qun
5tbvnAHLXRZo7WDpcb+zpvnoo3evqrTD+WTPTTRHvX5DaqIv3NIamT4TJpda
TjTzsK/ek2c+oesXckRBcP6FlLC1sCworftbW/TbtMs/ERa42xX9ZIN86zLT
DsLt4cnRDTMwRtUsPAyrAgqrnnNxrseKedn0PjqzKz1hgUzhtofeh62ZLs0a
56+Ulorgms8MD/uskHdZ+eAsuPeX/lPhMHE4bvkU+TN7ZReaOMg9WTntNOy0
LdbDBSb6ZzYJcPlzoWafJT53JLCUoZby0jqmI6/znx3u8HZ3x4J309AnT5oz
G/ZhV8a7wiIX7xvOcNHosiuLLdBfsOK1N61H87YET8XvLGv7NREsiNCWiqbg
XEJ90qVwTZExTmKOPqLc8UO0X4Q/G2eG7290lugU3B6WlJFNcB//VFdUwES+
QP+ViSFiFWtRDLN3/EInjQwZr39wLAvuyuLW104whA1c9XAzrNJxIpLfM0Rl
F+wvgHk3FHXhBjik9f3IPMwfKBC7vWOIdFftpwrYp0d8eHiMIfIU3Vg63G7x
KLHsLUNSW/Y3r4cz9fbrfPUM6VISSQCdP2BrV6pjSH2y79mPad3GeHpgFOtv
LFwhhOt/re7naRlSdE0oD4PL68Kr1gwzxGk5VxdD1+/rS9+lYUiv3/cdB2H2
QJKmUI38tWfaymCV8uKDFyqGaMXLuY20HrLm9LO/GCIoU3PbaL9JGS/rMUN8
fpBoGmAiVb/d/xD5BxKNx2m903HFxE3044fok2i9fFHzkUvoN8TJWUpNH/m1
///PPdj/ALrMLZg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.914847988868447, 10.361347240073592`}, {0, 1}], 
         LineBox[{{20.00000000000182, 18.000000000007276`}, {
          12.500000000001819`, 8.000000000003638}}], 
         PolygonBox[{{16.61, 13.48}, {15.57, 12.76}, {16.21, 12.28}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.403022683139532`, 16.54583280199964}, {1, -1}], 
         LineBox[{{19.999999999999545`, 12.}, {9.000000000000341, 12.5}}], 
         PolygonBox[{{13.900618875873821`, 12.27724459655119}, {
          15.117544188493639`, 12.62234281953293}, {15.081218059758719`, 
          11.82316798736469}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.488212026308705`, 10.854216844881332`}, {0, 1}], LineBox[CompressedData["
1:eJwt1Qs01HkUB/C/iLI5pLwKeeWZx7CLyPRv2tqSmAqRiqZIoRfb8aht9GDX
EdNTSVEzSRSzW6atvEJbbZKIJMkWRYUphU3Y7+3snGOcz/n/f/d+7zX/YSzY
ujx0HMMwUfih34wSvdmzzLfXVJaJnOC5UAfWaFo3KxdO7ebx5sJ1h9SD52qx
TM1ZJmMnnJN19GAVHBlx++MN2GGT9KuNNsu0FjGZSg4sEzK+tCAaTuVKHtjB
UtdESQY8KSldcRXczo+2Jg8Uyafth4UFcX/R/V6rGpUKYA02yo4D61aNddTQ
+bDOiY3opzR6yL7LgfKMZoTBTfNkt4Zhtq3EvRd5RYk3+5U5yLOaHYiBa5WS
xBNh+S6TJ4o0n2KhsxIs5MkjxFNY5seH1kcHKa9apUsQrJ/Vp9UJiy5sdefA
5rbB0bVUv7jxmBWsmddzvpj6XzHmLYRLEiX3smgebfehJDjA8pfhA5Q3l+V0
w1/O/O4SDfPF9T7h6D+0R8UtDM5h8rpHYEkBRxACG4lbbM9hnnDHF6cFlEea
us4L8xfK3gZG0X4CzSL64UUeogQhefaO9Wk6LPPYXkdwiuqZXzPV0mUZ3rwm
eQnlW1F0RwiHhH7IeQXLxydPr4EX7pedVMP87XbPe/vhj3en1LrCRpa5Pw3C
Xcwyg/W0H9mt7U/ggHSFn1PoetCStxmwhDP54yVYtLnPggsv39A6epf2bVE2
/xHyqJ6NiXxO5/V7kgPhyp7dB7rhuobj714gf8zmOL0emF92bV8ULPG5t6eL
6mkv/6gKO0u4h1vJ6uWt5dhH/Pnk7L+p/5qK7YdhS6OV/ldhxq9bORlmakX5
p8hb9jZkw9xJ5rK9dH7E6UErPKp4JCMC1hjMz3NH/dWrNfkBcMVW29FrsKrg
nNiT6vs1py9BXudnT7/Oh9kEi8u9sMTF238B2fX45RO0L+tP1j5kB1k8q8cy
jnExW9bB0nIT4+cwe7vvYALVt1ocHDaNZX7dOjMuiz5/ghlWjbDRvogrlZTX
7Iiu2XSWafkzPOI95ZtaWOUDZx3TOqHjiHpe6v3+8M6NP+jxYCM132I3WH6O
fboJbk+f7T2Kekl8cW8qXJFwYNwFuN3V48ZFun/y5yp3eG2Kh0I51ftje2U1
8r0cLmVqYKY+utsTzl3UGVcHb1O9voz+3mEqdcV0XVpqI90BD3UqJN6i62uV
75vBlTPyDKV0PrB5pB/7yeHGfp9JeY6YPe6A0+ojpImwvKGr4l84LHn4Yjgs
jBINOuK83cMqdhnswKsIT4OlJbnxHnCdpv0rFeTZop5sag/z1ff4n4SlHtlS
C5qvqc+J5nn/1cPKEhb5d7zugA93BGlzYI07N9ccpP1kasTTvuRasZk2+nj+
l257GPTtuh3vGuz1pkFzF/W/vsLEygD1O1Tqz8Ls2jRxAjwuNHj+feo31GdS
ADu2uwkGqJ/BFxsZ7PnOTd3YCfs43ZR+Gp72Rt9xMRzyfmnABnibwcxnkTDr
45WhDpe5KQhSYCnvq1Yu+vvKnIU5MN94b5c9LFeMDS2k83f0eoqR/0tdVvUV
Or8zmzsPfj22pVkKi+b4iVswb5sbE5ILt3t/YFJoHxWVh4/BFYcutfnBkdam
akJYmG85woMNHZplG+G6BXmqvvCc84kSb+p3kx/7G3w1O37Ameo3G0T/A68S
9clMYA3dISYA/avHfK9PoXoTDNk3cOvnBbu/g+VtvdXJyP9a9ZQ2OaR/rgYH
8x6f4KpP9+esfKHdAnPHic+YknkNj+IM8bwJh83n0HzPAmtVZrDMrDRBWRDl
Hyu6ugtedWzNCOVvH3EprYeHvLkm+eR3AcqKRvh+/GRl2wgvsnkZqwqPjuHl
9P//LyP2P1TTb9U=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.393275441506763`, 5.232686837739852}, {-1, 1}], LineBox[CompressedData["
1:eJwt1Qs01FkcB/A7TKs0MYkeCKHSQx69Ntn8V1LTEdNziml2stk6pfT0iGWS
2NSm1YNVU0oPoVB6UAmh15okdJqKWc+yKntsmzLMfn97cs6Y85l7//d+f797
z8yYwJAlQXqMsU140Tvj0z8njv3/Z8qxb0fknFsOK1aU/20KCwP33MqG857n
C76BF9R5DeiHNX0z79QM41jtVuX39s4cK9HlFkTB2hj+i7mw88tVKgO4vE+6
Rgof8nz2NMKEY5s/3W8OJksah6uHcsz849q3O2Bxt4fQBbYxLIoldzW8L08Q
csyLZXvT/BKhu0JlzLHMEKFoNcycu3h5Rhzzle8dJIJt8me3lQ3hWKV/dIAL
nB56ussEbnoyXjSSvF9slingmEN2YauO8i9b/GUvPC16l7qW6rVbdScb9knU
eVyFS5RNPFM8zzFbv5PUn4gs72uwpawyNBXmOuf8HIv9bXqMp2fQ/INT1RuR
b+XhZF0xLB4T+8YJ+Q8k9s3+C0731Jqdh5Ob1PbDqD9+PinGqPfNyOaqGbC8
c9HKtfAV04ouGaxQ1icp4WrrCuU+8ugvk67DesfTbxdQP3o+2GbDYosexwaY
kx3MiIRvjpz8Tt8FuQqX2k6Ap11JcraHhfeneN7A/qZBH/Z+B3eZhadNgkMt
zr71hdnBsFvmyG+uX1YvIV96UPUQ9fqWyf1WwvJLjetC0Z8jXeGvxLDCqj/H
dTDHlsxdtdeTPFZWaWzIsRpn97tTaH6c4qnlII6lhvNXjKD1Xtp8CBqI/RyP
+PVT/gtPd3cbYF5pwuNW8nhR5kPYWyP58wmsyc/rbYcNzfuji2n8tuCYCM+n
9zgtvEr1RkrkrXDgDfP4XDpf020Zl7HfOL6njsZLMtprjiLPiZ1l/SXUv6Dm
uH3Ia3+/bF493RfbjoQw1KPIlQ7tpvFPhgN8Ua/R6rvjhiOvxnr7CS2cmPqo
jfrFmWdu2Inz3rChY9168u9WbZdgh/kpRw5Tvx2kCadgTdQRnztU74JxLgvg
qEPq+Ga4pGLOkiysV+wm3MZ3Rd1Lk4v/wP6q+P2/WZFj5nkWIZ+wbnCdM6yQ
B39UIL/sQf5FN/KWyQEuqE9lP07fnfyvr/9r9EerXC2dDttk5VmnfMOxrIx3
ueNpvUZ/bcgAjvG1wetNYfZ++eIwPsfKul+I+ih/77q0Qn187hqyvIXyvuJf
8IBrGzMvPqZ6smqGGMGJ/uHya+QiRbM9rAoo3ZRB862jC2Ngad4MRQqNO55t
McP6yUfj25LpPkSFGLyEvU1idx2j+3DB8fld5Pki23/4DD0/S218FXlzjtvx
btC4b1TCGboPqvdDntF4XYNfFM43a0t/4kfKe/1YrRPqN9FdbrKkejftaabz
FszsdJlPFvskvYeLIkU/bYNLVhe9bofPpBn9cALWDDaamQqXFU8wvkf9uyBL
4sGTq98oW2Bu0eb5E7HflbKQOB2tt72l0hJ5VojMzptMxfNXFtS3If/JKpvA
0XC6fNS9VNTnv0yltIbl3tcCvNAPgaKnYhTM8cTPenm4v2t8Hwhghfe5U9X4
Nsl6vpv/hfKl6VWc1Xkwn+iAGc10PgGdGot+D9YlMfF/RHmexA8c1efB6mvS
D+TTeN0KyXmtB+v4UdiRRj5Z/K4cvu1ct/EXmj99YGMw5idrFz+MpHHD0txf
sZ5bb4Q0lPb77HXZAPuHR3sJImj+TUXnW9ihWDQnjvpxOyLsFfJqeZurU+B0
uz6nKj18P7kzi6s0v+rRsCzUV3BuYm8t9UehlgaifoHYuLWX9pN0H2iHeyK2
aMdSf063LZuCfu0oDXBdTPVv1RM5wVZiXVAk9dNwl4TmL1mojD9D/aqRf5bA
lrkxBeXkWRMsErCf2/4GmQYusbMujkYeVa9a+g/Mcu6t9OF9/d2a9vWdcf8B
K5Fq1Q==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.106724558493237`, 2.232686837739852}, {1, 1}], 
         LineBox[{{9.000000000001819, 12.500000000003638`}, {
          12.500000000001819`, 8.}}], 
         PolygonBox[{{11.118364368108953`, 9.776388669574205}, {
          10.065894744940517`, 10.478035085019828`}, {10.697376518841578`, 
          10.969187575831763`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.131304063066693`, 9.903192049051874}, {1, 1}], LineBox[CompressedData["
1:eJwV1AlMU3ccB/Ang8omR40cBQvtFoxACxYEhgTic2GIkriC3HSjdNUVBGGI
rqBREBI7hXFsMCEM0OpWVu57rIxCgHUExjGFogHZBooHhxxCIs59/02a5pP/
ey/f37e/vPclySGnDCiKCsOX/FIrb/FxpSkWgQVNia87X1XD8hDD0V3wAjNo
M4qYGWDkA9N/jyWawzf3q8q/hil1UvaIC01xZ0S5m3DkmtrrB7ghx27tsiVN
qR5V1GXAgWU7pDZWNCU4FWUfD6cooo53w6H+QY0yuKDzYqzMmqbYXRUKOWzs
IeKZsGjqzu9aXQm81cxR3IZF7xxp08KUQR3XwYamOp4cnV2FI2uL/7kO9xWb
hTshn04YJRqH2SOsuVhY4JlssAXnSNPDiuCVgP7w13A5m9eqgRvmpCvT8Oix
azcewtwrWokSFr7mWLwg/cicGcHwNrtTRixIHb+zSPJphyrI9SyvkN1ZcO7y
r/d/gTNP8mv3wNz9qx4KOLK6/HEt5rvXszEVCHfYVrZGwEOv8tv/Q35dIYfL
IudjBrJamL47ObSOfuS5FhuRsP5ZhvolbDERWmpE+gxVtDNxPe2ZcLmdjz5n
5FPBMPMj6aVU4u94C01wlRU1cwhucE0fPIA8rOErLkx4xXa+rRfW6QSmmzya
8mbm5Ukxn+9Thv0yzOXMBpjaor+GyQ1yXhXyeUoLLHIb3kfu9zZbqQ/ei/58
aoSesCwz7eNpWPhv0SdSeOFWeLaQTVN8zZqkFKaL+F4/wsLf/kocJedtmRsP
YHFmj4DMc3NIE7kIt7Qqiw+S+eU7TWdgrV48H0H6aewZrIcNNdmMZLjqzFHe
F7Dg9Gc1F8h+UNlx78GiGG1GEjx76aepKuQxPma19yTZL/mTfndYL1QfcSR9
Duyr1GEeFt15e4n0dT7yVgJM7dyaVJF+JGVBXNij2s07mswnrn++jH7SzFOU
DNLnuKnvDLxybSCgCf3oC+Nsl2BRXmpMHCwrextN7vf93nfeGpa7UJpzpM9B
y7QpZ+xDjm5rDuYma16o4I6Ee9tnSV7nvhgFLK6pdt+FeULPd92Xw442Ts8a
4ZzxM6czYL3SaDTajvx/5sxceHSRk/wG9s8beaCGZ7OfNxfaYx9DT6xNwPIT
c/FMDvbVqWTiXeQJfPSV+hyH7EG1gib5skr92uFt7/jxC7DAJ21MD68ftMxW
kfleHXYklh8WN44Rf2tmTa6nU923luCUQ+EJ6bBQKVG/gYV/rP7sAOeUfbhj
m5xXuOm6kUeo4rs+hbU8u5oQuGDi+Jf9sGokXzuP/HNUBLMInmVcTcyCZ6fM
/YNJvvysQhd4wabOkkGel+G3+yX68JjuOtBC5lWUSMZgdh//bixcMNjkOw4b
D1NNJrCi2+/TDbKPvd+kdzthXx5HJPngeetnXQcuwoLwh4xKuGFP6J/+cEev
csoe+bjGSgM7mPqgZKkZdnAyWDAi91uEGUaQ/sj705GmTMj7k0v/D9JU6NU=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.508784489160853`, 6.008784489160852}, {1, 1}], 
         {PointSize[0.04], PointBox[{9., 12.5}], PointBox[{12.5, 8.}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["6",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bgcfdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bgcfdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJw9lAtQVFUYxw+wsNAKKogECenc04C5hHFjJEnPh48ApxIphw1GHkt7KUNe
Qw9sYHlo4OgKEcLeZWFwWgqofAw45RjG+ABX7yJJMRjGENkilsOijLaC3A7O
3Htm7tz5zT33fud+3///X6PNS9K5IoR20WvxPnxapOt5QE8Xhg1Xk8OP8ZQ9
a5w71Bgu5TLB61jKGelo1dsYKivE6AtCCCBNQObXeXS/dv7nFG6RVWZtKYbm
vEu9s2IwIND5l5ZgcCgigmp5yu9/dHHFPgyuprFyNUv504Gz4TsxtB211ViF
VYDiaxb2h2JwDqtsHEe5dr/x9mMGXvSzXHRDlBPP7Lg3wEDiUJV3K/8cIEdP
hKKDgYgvLBkxLOXVlRdGDAy0bDvR8JsQBKjM2NhXxoDaNnIyl6M8bg/6sJKB
q77coAeiHHawqPAYAxW6b15t4QMB/ae0H+9mwH0XT6JYypoETcgfDLxlOfWa
IDxL69/JsHthGPyh9mAWR/mOS2d+FIY9XHLanBhA+7c5NDUFQ27W0P0vecpX
6vL8izC4PzicFM5S7r2hLtBjCMuy5/cLKwG1s30en2FQ/R66VstR7g35NTMb
Q3F5+fYnoj+tZxibfh3D47gnVUaecrxXdVcwBtj9dxfLUtb0Wi47GDjyKKFh
QFhB64/e67rMwIbcwiUfcJTb62N/OU771Z+jdEOU199aazjEQPGp9CNm3g9Q
tUfJpJ4BTx/r6SiWcrutcPnnDDRPbzINCL60vnYqromBu32Wfdkc5ej6A//8
RJ+fUyUhRNnR/e2QnYGvZuN1PL+czvMmc8OPnk+Z9mMkS7m23yV5M4amjYlv
CsIyul9zcnUmBoPyfADHUV7/MHbiEwyd0Yv6W0rnt7HuGu1P8ULFIMsulZ/v
Mabpsjkf+X1fcUuNifeWv+9dnRNlE5bI9TUNk6miqJLP16M/g19mVfL5V9ZV
XNNyz8j/F7s7x1zPe8n/3/odmewTPOX+3C99YcYpKuX+TdQHToazSrm/MRV7
h7Sch9x/v6xXjEbeXZ5PS4nCMSAo5Pll5s4VuSOFPN9NPSe+j2Hd5PlviYsM
KeBcZX1Uh06mtPEusn4+7kwOGhGQrC+TZedWL4Rk/dljzk0VcAtE0ufdQzHe
14V5Iun3aEujyzp2jkj6Nr+RHljFO4mk/4XruvO3xUdE8keR3QTbuYdE8o96
cGtbhzBLJH/Zm/eW+LIPiOS/cduYVc/PEMmf8+bRiT/FaSL5t8y7qdEq/Esk
f+NEQ3kXP0Uk/0dzypkmzk6kfLjic+AlPfsXkfLjvfhh57tonEj5EpYQid4x
3SRS/mRsQ/m32CEi5VMrqkw1m6xEyq8y82hHt+kskfLtaexVtfVI+fc/8NwC
Cg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4., 9.0548}, {0, 1}], 
         LineBox[{{20.000000000007276`, 18.000000000003638`}, {
          14.000000000005457`, 13.000000000003638`}}], 
         PolygonBox[{{17.460932767758425`, 15.884110639798688`}, {
          16.282993472375782`, 15.423177872040261`}, {16.795140992107367`, 
          14.808600848362362`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.61251418052006, 16.085942983375933`}, {1, -1}], 
         LineBox[{{20.000000000007276`, 12.000000000003638`}, {
          14.000000000005457`, 7.000000000003638}}], 
         PolygonBox[{{16.539067232241575`, 9.115889360201312}, {
          17.204859007892633`, 10.191399151637638`}, {17.717006527624218`, 
          9.576822127959739}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{16.61251418052006, 10.085942983375933`}, {1, -1}], LineBox[CompressedData["
1:eJwl1gs0lOkfB/C3TE4pS7IntUMjyaVh0STbtHppstqkaZLQJKxqRKi1mkRZ
tyRJals2tWq1ycotu5FpkEmkP8UUuYXGZYV0UUMj/+9zds6Z857P+c3zu81z
5oyhX4hgz0yKoi7hTZ7/vdj0f09zmlL6XfvoAvuIk0TTZjQVN+m+LQPWNfzY
8gYOPW0m6oWtVq9waIGzZRuLjC1oSmGXYFkI8x86j/jAlZz08ihYftDTLg0W
frJPdICH/V2WlsLK9ZXXpkxh/qaKJzD9THKsBE7/NrW4C46T6JaLYLfNmTYv
4LCbaob6sNWwRF0Oz9bvDJOboN4hVZQULuG98DgLR/cO7M6CMzecs3SHtb9p
bzkKn/cy+LQMTtx3f+c2eLDi/pbp5TTFqzh5xozE9f5Z3Q8r6jctnwGrentW
tMGcWxrVHZiX0xCr2Q77aE48l8LR2XEmg7Be2Pi5G7A4ocSWQn4hJ+Tvy7Db
+GGxIexf1nbyd7h2nX3FRmLnbHke7JHXfCoc1uvYUlsDS+wDzK/BVg4WjBHi
OqqpiZhbnm6AftIXPJap4Md9CTpeMCtiaSgL+5htO+fUJbJP2imJC4vapdJ+
ONnVzMIFznt9xGSlJU05p4UM8+Gg3PJF0cQvZ/Q7w6FDlS8fwokRbo9WwvTx
1mHtrzEff+Tf+XBWxoXmLXAeb4aBAvXDeuNPxsOp4i1Nf5F9P1kQWAiPqQqH
9pN+C68IGuFHjswCI7jEvuNaD/m8gFnUin3Js6TCAZidP3YzGWZ8Hp1B4uIz
RcGO8DxpaB45H2f86rrKmKbs3oy/LYEVp0zb7sKq0h+ep8FrLTs4ifAjyrcs
kOTL+xi5C+ZEUiaOMM/PS2kPZ8+VNemReX6ctdsCNvV03/8W87K/qmoyhfn8
vUeewKbc851WJL//Fy6lcObgksM8mBm/RJ5D4uVZ2n6wYig+LRtO/0nj+gm4
1b0hPg8WSgPbiuHhggsGVTBnbdKcHnhsa+iaXthqW3ioFubLSQox1kI/2aa3
nLlw5Eb3q07wsF7Uv75wq+CTMoHMx7vdHw2znjlpNMDJPnoWF2AbdnDJPCv0
2ee6OAvuLruxZiXMjldLvQTPvr3T3hMOje2gUmDvuPHAKJjlPxR+kLjYIf4y
8bTNOWdY56fnIeVw5mafBF34Xu2EQzNsZzg4swX9d7Uv3dwH0zNdzdJgnznq
y9/AvK/2hDjD2pp7b3wg8fXFsaplyFdYODVO8q3lWhbBXTvWnX0N1+paWgXA
3QrBlwo4uqfVyATmec8Kl8M5Lnu9R4xoKn/jmcYqWDdfKr8LJ17d5XsTTn1U
y/gNVlanHE+HJWIdRQwcpH88No6c98qVi2GN3Yskh+DKnXlZR+GZh7O8/eFW
fklkEmzJyhZ7kX2FLJ6fDTuaVL/cTub5PuVkDSw5W8fdQfo13Jn7Gu4WXWrx
hhkx3kNM9Dvo5nI/CPa5GJu6CU7o31cXDXdLg5rFcIdGqe9vZL6PbRNX4PzA
ZfWlpJ93XhEymGGaqtFO8j2s0euCDXQclNOwfM0GagQWqLluXm6N7/Hxr62v
4ZpV4mIXWFJV8bQPdp1V3XgQVu5dX9VAzt/VFZ6Hu81zVLmwY6Y5+xaJGzGi
jsL+DBv6f7Ceoq3CATaNFjX2wDTnQdk05kty4G4dg7X/KK+4DQe3HY1Vwuer
ewODYGrU98AknNXudZxF9t1ju/uDNbnfFveeLaUp6aGC0WGYNzzeexa++lB/
rAuOfHhrhxus41EfROoLC1Y4sGC/9qELZXDr3BWCj4b4PZnIL8+GVXvryttg
CdeEkwqHUilL6uHe5HPbo+CxgMWJtXDEAYXvATKfVYGoCe6gmEd8Yb5J6h8D
cGXDN8leZB/zPl9QR72apRsFnuT8iKWVBTzMmk97w2sHRWqepP+FD2IC4HQv
E9EJuOkv6nME6fe4WvrfsLpsyiWN5LObaOiGv9jKfZAPiwdSA9WxD9fBRYmN
pP4qseYy+I5o3ZF3pL8246LVcKUBd/ciG9ybHKcH62BZ8vOBdXDhly+duPB+
wc9le2zIfQkoM4PzTvfHnITDDHTpuXCplrdGLiyf9k0j9flzDdbWwNSUbDwX
1ttaY9sJB/k+tQyC36+4vH0ULhVUBS2Hl2Vo203ALCvF1x3YD79Yq2YKTrTO
Z6fCHrZsaxUsG7Rkb4C7Tri9fU/ieYWvplg0FfPjdtkgrFvQqZLAb+XCi62w
qX7P9Tj4H0ZPxn2YH/zisTv8bEOwWhH83mCyhwMXN76zyIRrzbvuLIFnhnMP
JcKZ1pl/LiTx0zTzMCwxpj2Y8OyFTGEAzJvDH2XDV3etsvOBHzPnTH0HJ4p3
/CyEPTIUlwPhUk3bOm9YzzA76Rc4Lqxx5z5Y2dxnUQ3ztSe44SQfb279G2I1
e6NTMLtpYYYB5h2rnx9/jdSfXKBwgu2k1r/ISHwyki2CEwSv+AMwQ2sbLxo+
drk9ad5K7P/bMNFpWBr/u9IGDt3D3J4Cq6c9dfCEGZMX3WLIfZ64p3UMzlFk
fBdA3NnHzIIfBUb/uR4OTalYL4WFIYxIHfhDgXlyCywuVvdsQb8R41dsX8Gl
HkLH83C60XuZEubcEWRsIvsk/9s4qEeeLPr/maqS4g==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.38748581947994, 11.085942983375933`}, {-1, -1}], LineBox[CompressedData["
1:eJwl1gtUjPkbB/A30k1pUruKtLl1sWFKOjH+vMV/iWiaraSmMbXFdNHF2hTS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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.38748581947994, 5.085942983375933}, \
{-1, -1}], 
         LineBox[{{7.999999999996362, 9.999999999996362}, {
          13.999999999992724`, 12.999999999996362`}}], 
         PolygonBox[{{11.536656314599949`, 11.768328157299974`}, {
          10.642229123600035`, 10.87390096630006}, {10.284458247200067`, 
          11.589442719099992`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{11.088381073024934`, 10.718437853950132`}, {-1, 1}], 
         LineBox[{{7.999999999998181, 9.999999999996362}, {
          13.999999999992724`, 7.}}], 
         PolygonBox[{{10.463343685400051`, 8.768328157299974}, {
          11.357770876399965`, 7.873900966300059}, {11.715541752799933`, 
          8.589442719099992}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.911618926975066`, 7.718437853950131}, {1, 1}], 
         {PointSize[0.04], PointBox[{8., 10.}], PointBox[{14., 13.}], 
          PointBox[{14., 7.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["7",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQE1cUBuALDRgjKo+AQQEJaBuhlug4FMG6K2IRlIqAAtKWFF8YqEQG
MBakGaA26lh1wIrKywc2rbQNJKOoiKEBQYgSYaJhqBoUFZEClVQMYuy/485k
7ny595x79tzd5adkxGy1JYQk48eM5PU7XHz6/biAJroAdboG9l7VFDHHjyYJ
eQc6CmBlQ3mjEK6LG+9YDQufuq8PgZe/FJ52YeLPu6d+CpeG+/Q89qYJWWZJ
9IVjfLIjLsOynIxqJv8jS4hrMUx7VZpuw8O5e9oz4GjlSdExWJlWXBwFrz67
sDoBtlq4OX6woOVymRf8neSrn1nMvHNUz5CAJtKY5GH1XJo4cmZ0tcJZN8MT
N8JBWTmRdbBQsml8yAvxvq1jNTCpf63dwzglfsNFuCF/C+edJ+ovGbS2M/Ed
obNk8GpujPUJHPfb5VYCR788I7fB/s3hlN0PHljflKKZBVcZmilHmL0vIN8X
Hmie+/rcHJq02Tnb8GEyM1UcBktWDe6cARfda9g8PBvrC3QuTP3y+y4Rp+G2
vrvxDbAo2e+LTfDood70QliymzK4wgrLw8bP4bUq3WCnO+73ekWwA6ycqtYV
wansuwU9HyHPni7xInhgQhqggpe0tnV289CHb+TCMri/3rI5DZa7Fu84BnuE
tdtYZtFEk7G2uBx2EGYY8+EjSyO218IsSQCfwHITx1EHqwt4tTI39E+bfPsZ
nC55lM2CRfet16wwe1Q+8pMrTYzFvMppTH3aw/U+sPz60H4ObE56+XETlyYm
+21ZFqw351WHiGHvkoe7jXCeX1+qB2zk/nXrV8aCfmJwQbw+rXAX/Dc7NesI
LHU4PLIENhWIuyJhXu6hN5YPEZdmt47AQQuXJt6Aq7jRLb87ox/+tpoKeODq
sM96mDw9sWYfLP1x9I/nTqjzeZlTPqzsKfSRwkGD0skiJj5wyRGrI/q7s2j0
OMzytPGUwbLg2MA6eAuvrYTAxltOHrdh3miLfspMjJfi65/Apm8vRM2ZgXMp
364eg4NeeU9ZMR39vmiwNcOjE1xzrgPyPfAX98NtHp6Buml4Pz5h3WmGHbus
3ABY8OXOFcz+pZrrSZUc/K/wP5sMCxff4rjBbGduoC+sEKSFH5yKuiu/f/Bs
Pt7f2LTUcTae3wuqJBUsPBc6GQvXJ4YVHmDmd+2uLpuCeh4oT0lgEjH2Z6c9
zvNe3NA2WK8IWzNshzix4mg6LIlgp0yykH/yl1P5sGbdMxcCO8bHZpbC7BqX
fvIB+si/m6lm5p+E1tvbYp/5iUodM39G1s6zwTloLjX2wsbp5tplBP09YZt+
HzZ1rwx/YaWI92NWdCeznz5zqvwtRXj+FV61jFcKrkROUkQfNFJbxHhfevyC
NxSR975tj4KrauznuU1QeH7Djzox+1cc3GtjQb714keGeeirgh7rHaeILDJ3
fxWsHNmwsfQVRSQTemkOLFH35S38jyKC4JveSbBelJB7fIwiSgN/Yh0s2lXh
Z/wX+y2qjI1jrDp2558RigwEX92xlbGbV/C7FxQRZi9uyYdlN4do2wGKmLJ1
3SdhctRQ6NZPEcWA+byKmc/QxiSaKCLaW9OoZerj9+3X9uD+7tS13GDWL3+z
Krcb+cza+VcYR/d81tiBel64i8oZd3JUV5ooEr32a69MxrMTtiyvpUjp7DZL
CGPmKrn2/vvuS/8PVyA4dA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.990153037326099, 10.634340888021704`}, {0, 1}], 
         LineBox[{{20.00000000000182, 18.}, {11.000000000001819`, 9.}}], 
         PolygonBox[{{15.924264068711928`, 13.924264068711928`}, {
          14.792893218813454`, 13.35857864376269}, {15.35857864376269, 
          12.792893218813454`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.008784489160853`, 13.991215510839147`}, {1, -1}], 
         LineBox[{{20.00000000001046, 12.000000000003638`}, {
          2.0000000000118234`, 9.000000000003638}}], 
         PolygonBox[{{10.408163645700714`, 10.401360607616786`}, {
          11.526076759377142`, 10.993196961916071`}, {11.65759594922143, 
          10.204081822850357`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.745181569676697`, 11.424110581939825`}, {0, -1}], LineBox[CompressedData["

1:eJwl2Hk8VN//B/BLiCJClHUkW7KrpORGQiWjUKQSWZMmpYQYilYlWizZSyoh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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {11.085978530414911`, 9.44281354746839}, \
{0, -1}], LineBox[CompressedData["
1:eJwl1ws4lGkbB/DXoUyWHHLu4Jh1+jQkUcpbSpJERVOpdWxVki1JJaewSpiE
JDSrMEVIo8bpSyXpZKdSKCaHpckhttBUlu//7DfX5Zrrdz3Pcz/3/Tczhq7v
wc0BkhRF9eGHPP//YUhTk9N46NJU7cJzZs4wO+bIMKVHU235Fz3SYTmHvG0T
WN8qfrmyAx7bvEC1C6Z1mA+0f6Ypu/tLj/wXHpqYOegFS2dcsEuDWZIBVBo8
tDgg1wt2ca3j1MG82H4TbVjjuM1TITwWZvP4nQ5NtfQrD47BBmkNF9Nhs8e5
ndMwvWVF3wbYYHbBl39tM0NaGr56y/TTOOwdIEqK1Kap6COciR7ixkHe9AKa
ykp+l9YEi05/OpwKG7T6ZHBhp2ua+hZwuBFbNh4OjHT0+zCfptJzftP3Jut7
31I8eOjMXG072OaOVfcF2ED12OK5cL178Lc0mDV8JGwaeXAznrhehc/OCm0a
gMtrW6gncFvKQL0QHo2+5yWB+9r6xnNJfnzZjd8cYfGZ2KV9cHTxeHsmrDiV
OSGGNbp/zhqCrfhqL9RJf7cq2A6Yz+VvY+dVsI5h2ZwLsIqe49FQOFpmKL0H
Ljl0ZVY5PDkSOqxI8smt2/kZFtTQGUZwHMMj1tYIv99W/m1ruCmf9j4Fm03p
GS+DVcYdap/DcY8WWlrCoe9lWGrGNMXJc01eADPKnWR2wqMnr3dO4b7yH+sb
L8IMW0qqGGYrcZsFMCUVdZf021VVdm0KdtqfrPgM87CiI+31TGBHy9x1cMPs
TT3LYHaKJ7saeclVRXg4wQKLygPGcNft8T+diRnWvVnzsF8g+X01nKWemSUP
158/0W8Bsy65TSfNRZ5hN2Q04cBLVQYasFtroM530o9IP7FSC7mHN1u0wjqx
9yf2wKF3L+6pIPMsrA42h11u100mw25BybLyMJNpZx4EM+vujEvCbksMtTaR
edlD+xTgEpNbpTYwd4Z8LRNueOGWYQJH18T4+cNDb47pGsI2nWsXFMFxP6Qc
zGBWU6b8ONmvdzhiBUyXDQnXo19pzcrO7eT+O5fFHPhqeIRUDBxSHnRtHKaD
Vs0l/SZWfGI5YH6Gzt3BQdJfj8adeJgXmNlthvkTGx6K+LB4+3ntw3DbqdSy
dthgdPhcLSzS9Ujqg9vU70rJmGL+bqWWt7B07LMlbrBG8o8n5LxKw93xDFgc
ZJ0fRdZft1m9gbl2fx5nwsyVFvMVzPA++q6w/Tn6G9XkGdjDtDtXYTss4s6c
HwCLbGP0WjFv+kCtfTRslLs3fwMcNLHTmg3bXJsjVaOJ979D6/wMUs9OddgY
Lm/3j0+FOe8b5XM1kEMeL5Ccp4o+92vA4dWNC/fCTSqOnD/U8f5b/yR4Azlf
+2vHMtgqPdPGGM7i7j//QQ3PmqGF0jDzq/HYNZjjP3PsPeZJpHgbT8GBomW7
62DmaYFEKJyubyp3GXbaIRtzApbmd4kT4PruY25ZsNvfXcIwmKE7W7MJ3vpk
s2EI8co7d2Vwvzjn6JvDJL/AY/FbYKddckmxcJbJfPMimPOFCs4h+bcNm07B
/MzGJffI/oL7zVswn1Wlo9EobJQdFHoFpib89IzQP8s+gy+C5aT5V34l82/X
ztdGXkwtVWEJmbfQ6tkaOGK3adM4yU+fddoDpjjnLOn/4PdhoDy9CZ5czxKe
hhM/qdsuhnleK9IFMDfXaPE06itePbdc2Rzr62zbebDd8Z4RF9go+tM/HrC/
QqJKFFwveLiuD/1TIxsDimC+VIXqHjjnwGmph3BIzlqvDuTTdkktqpWsP3wV
4ULyfyx8K4QDWeYaNarIr73q7DvilhahGcyoKg5vhrt+i/UuVMH7wavsAx8u
f6DQaQqzzwcfz4Ft4lI07s2hKS+LEI0TsNt1f6u9cEjCvg5PWMNcnGIIR+fG
T1nADIZt8Xdl3D90oG42zK2eodsPr9lksG4E87M9f7oggkMrj+e/hvlzvRsp
nDf6uDD7Acnn1K3fzYgD7K5Xw23Xkj4GwpRh92gNHFJw3bUCPpvP12uC6eW+
56XRb6jU5VPvyX7+VpudcFfLQK8E7heUfGTfhBmnWfnmsPhV5hdJzO+vmVTh
D3MSZAec4clmYVE+Waccf4mHs/IXNPTDoplPH5eS/GLjZi1ahHWBkVsDfLXW
edMxWLQiuP8BbGP7fdsDuL7Eza8EplPWhc5gYv9ls7Mn4R7W/RFDuGvUT90G
3r3owMK1cDlDoqAT/Xl9VVv5C1zvNHgjBO6/Ip8cCvP9duwbw7zhH9QuxMEs
Mbs8CI7I26ueCnPY1s87kWd1VKhkOqw4a9DTBVbx8Ug4D7ddEU7VKtFUqfMk
IwVmVrG8reDsEnkDUk/0SvfFbUV4R2leGBxeVmvlBPecOhawB45uCue2KtDU
mZdforfCdFpJuBMceFgou5qsr5rr2D8bn7+72vSYsJunUkkFXOv6oEeb7M85
438Dtvb036JM5i2dsaEFjtxc6cMg+49ePaSPenSex2tJ0r9vpWsG7JUc4UQs
TpUIU0I/Ic/rdWXgpraN83zhrLaxVCWS16wMUz4cZOGQqEvy0Xf4Zzbmmz2a
22gNsxf19fjAbPOTAjeYYZzgWQJHqgq+BcMa2fuMB+DJ7OH9bJg7ua5MHflt
9l1hzoN15tj+ZQG/9LVntsNG1f1J1vDnZ2e2TpH53+c7GcAueZsj9CywPznh
2nfU29orTltjQb7HUIuqYRcJ24t+MK3MXesPT2QcKYuE63Vilv5A/4o77i7I
gPl6LJVYWK5gqXch3BVhY/ANeRSaxepVwJwlCs87kd/xMvmNfFJPvHJpjTz+
Hl5QkqiCBcU597Ll8PfI1zqLR+prfePF/4TP28oJTjHpb47NX3GyNBVbKWzL
g5kNpu6cWTSVr9ZnkAqHBHa3Cxk0ZWn4n8STsGIn7++1cKLnOc395P6JmXvb
ZfD6CPIVs0h/9u/WXoLz0v+knch55y+lSXCD8kFqGelnauUAFxaJd5kuIvU+
3wwegW1Uf00zIv08fMjYgfqfazLfGpJ+RpmKPTDTNS3MhNR7XVQShf6szcV/
LIbZN3sMTdB//buvu1fDo5ZXV7fBWsrLwzzJvFu186Ix7+ounWMHyX2evcpa
yIPTuLjxLByd6q+SAz/pLkm9AXvbuzOmYevJ+l4BOW+mm2GHPCt6fW6Pw273
b/VugqfcFYa0LDHPXx4vlsiT7xHrAuxhzjdjxWGcr+bpyfrC5TX5VmGw+fbc
1zEw0zO+7BX64bvrd+XC3jfLebPgZlVRYCU82lOqsAD9G3iVdT4i+1VeflDH
vKFbeF0tpP4up7Bp5CVIOeTSAbPn6d/snEm+5xsvEcKK2+o/1cygqe9fZEVv
yX7rLfFXpJGrcJ7oJUwHhuzLlUJ/2iOiRji6NPbNTUn0z9v56A65f1XvyAcJ
vO4jc6ILYZ2DkjFr4bA2ZaN0Uv9gdm0zhc+PFJMOMk95oIdyAuyr87Q6hJyv
KVANghvyml/7kPNjr45GwXJO4z97kv35Uap18EDvozOucNczGXUd1C9d/eLG
BlI/56xePly757QCWacPJ+QuR3+he5aLPcj+oo0/emGNYq2jfuS+3wJV0zBP
enGBTxjsFqdwxg7zKsZ7eaSQ+V0057+FrV5V1Vwn9Wt81X2RT0THGsETmLr5
jvMMFp2I5A6Tfp3TleYgT/HMbSuUF6PeYJw1E1ZLfnpkKSwoX/+7Dmzeamfp
BZe7Fzj34Ly11E+sKDi60J17BN7z8J+OyzB7f0BRB+5X85Z6WQsz7/+iORdm
WLQGvYbrl1ccskb/hZSZ4kdy/oZP2hLMx82esfkrTH+WPkHy+fdhhfnJ/4/T
9v8D+6QibA==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.846807950948126`, 6.118695936933306}, {-1, -1}], 
         LineBox[{{9., 13.000000000009095`}, {11., 9.000000000010914}}], 
         PolygonBox[{{10.268328157299974`, 10.463343685400051`}, {
          9.37390096630006, 11.357770876399965`}, {10.089442719099992`, 
          11.715541752799933`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.218437853950132, 10.911618926975063`}, {1, 1}], 
         LineBox[{{9., 13.000000000003638`}, {2., 9.000000000003638}}], 
         PolygonBox[{{6.020945885274675, 11.2976833630141}, {
          4.780598539382591, 11.049613893835684`}, {5.177509690068058, 
          10.355019380136117`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{5.335784645546908, 11.740976870292911`}, {1, -1}], 
         {PointSize[0.04], PointBox[{9., 13.}], PointBox[{11., 9.}], 
          PointBox[{2., 9.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["8",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "Automatic"],
       Annotation[#, Automatic, "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {54.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["g",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["g",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {77.25, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuCF1BjAx0VQIFYJATVAwQiKqLT3llgFbacUp4YWBMQXoiBF
xBilBkERRocoLYKKjfUVeZmqKKg4EUVRBGMpFm3QDJUYW5AAClRb7b/jncns
fHP2nN09e3M9EjZErrIlhMTiR0cy/A6PiHs/+nIkJJpZWwcn9g3wnD/iyB4i
2JUJC1z0i8TwqOnzK0Jh/WXtkymw1uSZNQZu99LKJsIRPvZZ190xv6jkGYHD
/GQLFHCqW3XC76jPNPxoCIZV3t3LjsKWxuU8G9h18XDQChr/ZuOFpsmIS5vW
iGHBwtlJhXCELPDWUx+O6JR/jFwC68e3XjkDm3o9CvmwKOiznFw4tVFTXT4J
888HlqTA2p5kJhQmFwvjV8GJbUx404cc4Vw7hEk0P2Pw3iK48ZK7xxY6vyw0
sWEiRzQyzcq9sPXK6iwOjvLf536U1g+dueWakCO7XW8fqYIVd6NMn8OiFz2C
X2B17aNHXW4ckaR1jTwBa2pjzu6F29clfZwHF08Xn5ZRVxnTYmFR55DTCFga
2JA5ja4Xvvlisyv6sbPFaPHG+cOeRxfDgtZqoRZWZRdnx8GKouOj1sGmTJ3P
ZBpPX86fDXNmRYbBBeOwt4iBrSG2IemwtbH01LAE+57eudgOVtnI7llh6eHT
in0TUM9Nc2oQHn4YIGdg1TZjAh/5NYuV1/LHoy9tFQ+FsP7SBD8+rEtZzfrD
lu83dec6Y73oE91zYfXSp+nOcLzc6BVC9/frnfoqJ9zHmweX6Xxr9NI+Odxu
LspxhKNetTmNhVU9a5VdWL/muMPS++Owjqr0SCVsNdjFHYTD6jMLUuDE8knP
V8LMgldmP7jRyV7pC+vifsvom4a+yZPFzx0x/puReRUm5hk+B2FVdf6LYlh9
XLU/BGYODFl2wJJkRbSBQR/+vLF+G80fPS1HTp3sEp0Npwrz88rHIq8tXF8I
My/T3TRjsP9Km1gtbKr8uuLAaNzzocSEy7C1wav58Cj0c2Zu4C2Yy5b41Tgg
//67AGp15Lz0XnuMJC2uBjZckG6dD0f5bq8rgaU2XzlU29H/iVC5nu7/u1rf
ebDWnHEtkMa/YPqbBTivWhLxcir69NPYSwmwwf9E8zlYZx9f8mIkxjIdXwGr
LTYda2BOe068gDp8VswtPt6X3GO9HrDKfCTNDlZIHJeNpnHjzu3+I3A+S3aw
gLrlbvvcD+CCmJkMbIpJCQzi4d6rvgym+dZ+3yIfW/RtxVbrLJjxDngjtkHe
TXVuGI2n7jR4E7wPQm3aEjh+v5dLwVuWtMt7AiJhTvoJj/cfS7THqvs/pftJ
7fn27BuW7D65x0Lrm0j60IHXLFGoz6cNTEH86Z3wU/+whNyQP6uFuZgsx5Zh
lkTk97srYVH9rkeDQyyRHHp8bw6syTL85QDH/+xR99YLcUHenNevWCKacGaw
GdaPkwjOv8R69096lsOks6I+YIAl+hYdr5jOd97QtbkP9YM6Z/xA4/si97v1
skRnlx1QSvM9Kmts/2aJNLT1dhXMLRzobjWzpHH7UMF1Gk8qFtd2ol6GsbKN
5ss2el7sYInhcdjpJ9RVFX41D1gSPEXm1QGrHs9OLjUg/qxsUxONT+3Ne3IT
6/fO4JdR59h2x15hicXIW6WkNl5NHFEOX1jkyFHTR133/nvuyf0PQQYoFA==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.194814593688687, 12.79766898841154}, \
{1, -1}], 
         LineBox[{{20.000000000008185`, 18.000000000007276`}, {
          8.500000000009095, 14.000000000007276`}}], 
         PolygonBox[{{14.816698078023695`, 16.197112374964764`}, {
          13.551893671999796`, 16.1806863437177}, {13.814710171952814`, 
          15.425088906352773`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740793031341028`, 16.859170034894547`}, {0, -1}], 
         LineBox[{{20.000000000006366`, 12.000000000003638`}, {
          11.500000000005457`, 8.500000000003638}}], 
         PolygonBox[{{15.19519314091517, 10.021550116847424`}, {
          16.152506936983112`, 10.848321122542462`}, {16.45710678118655, 
          10.10857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.159837801855844`, 11.07845105263581}, {0, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          LineBox[{{19.99999999999818, 8.}, {15.500000000001819`, 5.}}]}, 
         PolygonBox[{{18.249230176602705`, 6.832820117735137}, {
          17.028889744907204`, 6.500000000000001}, {17.472649901887387`, 
          5.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 7.184877956223658}, {1, -1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          LineBox[{{20., 2.}, {15.500000000001819`, 5.}}]}, 
         PolygonBox[{{17.250769823397295`, 3.8328201177351375`}, {
          18.471110255092796`, 3.5}, {18.027350098112613`, 
          2.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 2.8151220437763422`}, {1, 1}], 
         LineBox[{{8.499999999996362, 14.000000000001819`}, {
          11.499999999996362`, 8.500000000005457}}], 
         PolygonBox[{{9.712688721591656, 11.77673734374863}, {
          9.936153049242591, 10.53172180397914}, {10.638469507574097`, 
          10.9148035085236}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.244061861982704, 11.112579197445111`}, {1, 1}], LineBox[CompressedData["
1:eJwd1QlQU1cUBuCHsqmIQVEChEVFCAhuLAYbypNlpDWsA5KxVFYFIWrKIlgV
scaKCmNErAGRRUHj1JSAqDBSQIq1FSooaQ0iCKhMqKGgRokI0v+amTeZb87N
vec/72Xe0rg9YdtnURSVgot8U+Mz+KyiKSaBKU0JUpw+1MP8NrcUXXjqGjM3
FZY6RhdYw4b3D/26Cs7Zn5cfBBeKV4vHXGiKSk4sOwPLr/b/eROWctO4ariF
V/3uKPF7e4sti2kqvMosdBuccyrc/wEsqbzTTMMB7ynv0CU0ZTSubHUh+50q
sO6F2wxY6Svg3OsG9YlmNFU+PCBygMVxzW1v4K6rSza6wmw///lZTJrSNtv/
tAmWJ+QWvIVN3VOt4uGQZyN1seY0ZRej6ySCtaqqyGaY26tRSuEugdfMHAuc
s0+5vBO2Db414g0n9OZpNDCzqsMmBmZYXvBkIj8Vws9OhkX+WZMeZB6TslvR
sMKxLjkY5qS28zbCnMTyj9GwpPtytgmcvjw7JZHUJ9mcbpzP77v3MAFm2//8
+jhMHTKRRpL97YxPecIlc5w6vcn8jW96DiOPtIkntiL1t33rC2Fb0y3Fr0me
6q7WTXB9pdnHRjjrm7I+XdiPbbT4MJnvXyqXTsxLMdx52huWDK5VXoVVNYOh
k85Y77sy9hzsrLd563WYkWFytghu2T/UIoBtHZh7a2DN0ZFzbDhAL3/dM1hQ
2dGjWgmfX7h8Gc6Lcu+IqIaVtboPyf3QNdxcehBmVtzTHYQbn+75EA5zhHt7
I5Gv7kt5uAcslFk+74GVV5q+WAYnXYrhxpP517l1M2F53M6WMZgqcnlsDmvd
Uhz3W6Ke2X/BDhbnC7k6LJo6YLHImgPHWNe+OQgPiII/hMEhH8fLR2E3c9fj
afBARKhloBVNsQKLx86R/q5EKorgSqNBdhM5r6nn3UN4jccZmxckT8B4qgbO
k91vMUD+nAbPhTrWuB/aNIYDLD0xV6hFXb7jOpeG+QcztL1weoSsOpTMi72j
RwY3Tsy22UrWZ0aMCuGA7WkSPix8n1jGhp/2MU7yYAlD/74S/cqnH3W4k/uR
6nb6KMx44lxiCsf4sTa4wvzAaxdG0F9uSZDOMOZxrbT27xswY2Lr6wpY0iAq
/Z7kKQq8nQx3PV903hOm1kkKfGHRq5nLGifkbL07fzWpj+1uk8HCFc6fbWcr
MEyEVUMFUj9Y65P/4zI4YPoHOyEcNVPFHXKkKVooHZPDA7uDZ67AHONoqR76
U383KykDrk9zcxIQj6Td3QzH/PPk4iC8Rt1u7Ex8sVkWh/wChvuTJcSD6zNe
wXwf20gj8vvfpx9nYt7hazN4xJwy8enZNujT6I4FWa/1UsefgEvMg1rZcLpB
jq2uLZ7bdJXaB57yZTdw4Q7f8ck44hdl23bBA+mei47BB44H3imE1cVW7jK4
cB9LXQOrvBbwumGJ9ePANjjBecx3AvY7W9PUDvuZf2tlhnlMBT44/AecZOSl
sxZWvJH53Cb1mdowX7i++JJxFdxl8rKJBzPX5H99DC45/O9Xnx3rVhUPK+x3
8jfCmgV9+qRfRdGRGyvhys6K7IVwnnl/4Tw454U06hHyFgoviwfJPDKDUkQw
a7rm0i+wvEGxwRu283hlkw7nBoUe04eN5ggqXOGQ4faCPjLP0dZP/7HxP6+I
17sHUz5THlJY3Bze3QaXJ9lsi4XpZ08dlbDQvz+XBYdM1MfrYD9No1d9rwOe
h1qZHm1D+vt0txwe53B/K4AN9Y+E7YKzoiJOTsAMR3WdD1ySvOmBO/Kw502O
LoWFcXO7MmCW2Lt0HswccnjZAH9+P9ljH/J+Wkr/D3mMX8U=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.084117857803797`, 6.188306123204339}, {1, 1}], 
         {PointSize[0.04], PointBox[{8.5, 14.}], PointBox[{11.5, 8.5}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bfcgdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bfcgdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuAFRQlECIhCKUoAxVRoCwYRRHsvIhicqiFqBRwfZUCDIGYK
WipCqYw0UIsUrKLyUuShWAEDAuIDrE9GLVBrmY6SoBFSChoaIxgJ6b/TO5PZ
+XI2e/7dTeIWs0cSZ04I2YoXHckbEx43lozTUcCSK4IEs5twZJ2CuH7EEkn/
66ZDMHurV70EPvrU8+fPYUvPTXGhsLvYWO0Ik27e/RB4+8Dh2EE+SwIqDTt8
YVY2WdIG8/V8lQ28IeB8biEs33+Bo0Q/+6Yy5W5YpBjMqYKHZD1GESz1vbs4
AW78umOhB6xVuYUKYZ532wm9K0u6VPHfTIUf3FaGlcBFwrwNLxag3nhdHQS3
D6em9cCx61qWdM9liU/xuShqcYdStBWWlsjt++HejvvBL+fg/TkpNw3w01oD
Px7WPKmc5oz1l/2xMmLEhSWputjWAPhqxrZFe2HpjMf3xDTfiHnFFFh1gGO9
BY40lhef/JAl+e0Z/0bBsrH4V0thgaO/VRg8fH26m9oZ8/3PCtxhy7qh7KNw
jXWCZBT9a4Je7QuHUwPV4haat3P9OdMHyBOpskqnefNjzivgdq89l0NpXu70
9TFwZsETkyPMFb7358BdtpEh454YNyW9q3Jiyb2Dv7/RwKLJ0gNB8LjnzNVD
cGbJ4Mu7jiwpP5EsNMAuoQ1718JFeZoKB6zHcm4F9MzGfq7nzPWDvYmi5At4
XLtxnwT2E2gn+mbh/tNNlfGwtnrAIwn2qV1zJgW+asbTWcDyS5utZDTvd1+6
VDtgn3eqSqLhxt+mWkfAMvNHOxfD8ypqx8xg0Q7r4il0/ZFmz+aZLHEKjJnf
iXx8r6urdlObbhvyYPWx1X/zYX73xbWRsGXcytFH9ji3Jz9e9oIfz1/O2Qun
5tbvnAHLXRZo7WDpcb+zpvnoo3evqrTD+WTPTTRHvX5DaqIv3NIamT4TJpda
TjTzsK/ek2c+oesXckRBcP6FlLC1sCworftbW/TbtMs/ERa42xX9ZIN86zLT
DsLt4cnRDTMwRtUsPAyrAgqrnnNxrseKedn0PjqzKz1hgUzhtofeh62ZLs0a
56+Ulorgms8MD/uskHdZ+eAsuPeX/lPhMHE4bvkU+TN7ZReaOMg9WTntNOy0
LdbDBSb6ZzYJcPlzoWafJT53JLCUoZby0jqmI6/znx3u8HZ3x4J309AnT5oz
G/ZhV8a7wiIX7xvOcNHosiuLLdBfsOK1N61H87YET8XvLGv7NREsiNCWiqbg
XEJ90qVwTZExTmKOPqLc8UO0X4Q/G2eG7290lugU3B6WlJFNcB//VFdUwES+
QP+ViSFiFWtRDLN3/EInjQwZr39wLAvuyuLW104whA1c9XAzrNJxIpLfM0Rl
F+wvgHk3FHXhBjik9f3IPMwfKBC7vWOIdFftpwrYp0d8eHiMIfIU3Vg63G7x
KLHsLUNSW/Y3r4cz9fbrfPUM6VISSQCdP2BrV6pjSH2y79mPad3GeHpgFOtv
LFwhhOt/re7naRlSdE0oD4PL68Kr1gwzxGk5VxdD1+/rS9+lYUiv3/cdB2H2
QJKmUI38tWfaymCV8uKDFyqGaMXLuY20HrLm9LO/GCIoU3PbaL9JGS/rMUN8
fpBoGmAiVb/d/xD5BxKNx2m903HFxE3044fok2i9fFHzkUvoN8TJWUpNH/m1
///PPdj/ALrMLZg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.914847988868447, 10.361347240073592`}, {0, 1}], 
         LineBox[{{20.00000000000182, 18.000000000007276`}, {
          12.500000000001819`, 8.000000000003638}}], 
         PolygonBox[{{16.61, 13.48}, {15.57, 12.76}, {16.21, 12.28}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.403022683139532`, 16.54583280199964}, {1, -1}], 
         LineBox[{{19.999999999999545`, 12.}, {9.000000000000341, 12.5}}], 
         PolygonBox[{{13.900618875873821`, 12.27724459655119}, {
          15.117544188493639`, 12.62234281953293}, {15.081218059758719`, 
          11.82316798736469}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.488212026308705`, 10.854216844881332`}, {0, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          LineBox[{{19.99999999999818, 8.}, {15.500000000001819`, 5.}}]}, 
         PolygonBox[{{18.249230176602705`, 6.832820117735137}, {
          17.028889744907204`, 6.500000000000001}, {17.472649901887387`, 
          5.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.004985304149102`, 5.815122043776342}, {-1, 1}], 
         {Dashing[{0.005, 0.030000000000000002`}], 
          LineBox[{{20., 2.}, {15.500000000001819`, 5.}}]}, 
         PolygonBox[{{17.250769823397295`, 3.8328201177351375`}, {
          18.471110255092796`, 3.5}, {18.027350098112613`, 
          2.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "g"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 2.8151220437763422`}, {1, 1}], 
         LineBox[{{9.000000000001819, 12.500000000003638`}, {
          12.500000000001819`, 8.}}], 
         PolygonBox[{{11.118364368108953`, 9.776388669574205}, {
          10.065894744940517`, 10.478035085019828`}, {10.697376518841578`, 
          10.969187575831763`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.131304063066693`, 9.903192049051874}, {1, 1}], LineBox[CompressedData["
1:eJwV1AlMU3ccB/Ang8omR40cBQvtFoxACxYEhgTic2GIkriC3HSjdNUVBGGI
rqBREBI7hXFsMCEM0OpWVu57rIxCgHUExjGFogHZBooHhxxCIs59/02a5pP/
ey/f37e/vPclySGnDCiKCsOX/FIrb/FxpSkWgQVNia87X1XD8hDD0V3wAjNo
M4qYGWDkA9N/jyWawzf3q8q/hil1UvaIC01xZ0S5m3DkmtrrB7ghx27tsiVN
qR5V1GXAgWU7pDZWNCU4FWUfD6cooo53w6H+QY0yuKDzYqzMmqbYXRUKOWzs
IeKZsGjqzu9aXQm81cxR3IZF7xxp08KUQR3XwYamOp4cnV2FI2uL/7kO9xWb
hTshn04YJRqH2SOsuVhY4JlssAXnSNPDiuCVgP7w13A5m9eqgRvmpCvT8Oix
azcewtwrWokSFr7mWLwg/cicGcHwNrtTRixIHb+zSPJphyrI9SyvkN1ZcO7y
r/d/gTNP8mv3wNz9qx4KOLK6/HEt5rvXszEVCHfYVrZGwEOv8tv/Q35dIYfL
IudjBrJamL47ObSOfuS5FhuRsP5ZhvolbDERWmpE+gxVtDNxPe2ZcLmdjz5n
5FPBMPMj6aVU4u94C01wlRU1cwhucE0fPIA8rOErLkx4xXa+rRfW6QSmmzya
8mbm5Ukxn+9Thv0yzOXMBpjaor+GyQ1yXhXyeUoLLHIb3kfu9zZbqQ/ei/58
aoSesCwz7eNpWPhv0SdSeOFWeLaQTVN8zZqkFKaL+F4/wsLf/kocJedtmRsP
YHFmj4DMc3NIE7kIt7Qqiw+S+eU7TWdgrV48H0H6aewZrIcNNdmMZLjqzFHe
F7Dg9Gc1F8h+UNlx78GiGG1GEjx76aepKuQxPma19yTZL/mTfndYL1QfcSR9
Duyr1GEeFt15e4n0dT7yVgJM7dyaVJF+JGVBXNij2s07mswnrn++jH7SzFOU
DNLnuKnvDLxybSCgCf3oC+Nsl2BRXmpMHCwrextN7vf93nfeGpa7UJpzpM9B
y7QpZ+xDjm5rDuYma16o4I6Ee9tnSV7nvhgFLK6pdt+FeULPd92Xw442Ts8a
4ZzxM6czYL3SaDTajvx/5sxceHSRk/wG9s8beaCGZ7OfNxfaYx9DT6xNwPIT
c/FMDvbVqWTiXeQJfPSV+hyH7EG1gib5skr92uFt7/jxC7DAJ21MD68ftMxW
kfleHXYklh8WN44Rf2tmTa6nU923luCUQ+EJ6bBQKVG/gYV/rP7sAOeUfbhj
m5xXuOm6kUeo4rs+hbU8u5oQuGDi+Jf9sGokXzuP/HNUBLMInmVcTcyCZ6fM
/YNJvvysQhd4wabOkkGel+G3+yX68JjuOtBC5lWUSMZgdh//bixcMNjkOw4b
D1NNJrCi2+/TDbKPvd+kdzthXx5HJPngeetnXQcuwoLwh4xKuGFP6J/+cEev
csoe+bjGSgM7mPqgZKkZdnAyWDAi91uEGUaQ/sj705GmTMj7k0v/D9JU6NU=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.508784489160853`, 6.008784489160852}, {1, 1}], 
         {PointSize[0.04], PointBox[{9., 12.5}], PointBox[{12.5, 8.}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bgcfdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bgcfdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {54.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "g"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "g"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {77.25, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1A1UjFkYB/A7fTBRtpWWaqtpVFI+Bqtd1sx7RYQi0TQlmhTJKoWlVltT
26Z1iiiUsoaWzaZSURzUYCUtq0XpUExHHyNpR/QhY9v/PbznvOee33nufZ7n
3vue1279Vp8NeoSQQLxs/PgI6MdxKiXKHP9tu2wp0ZT7GJrCItsDKX02lNCj
Oxwt4W+mbepOYC7T+VrAUQskBy3hjL5y81HwxcYi81vWlMhsSH/XFEoES5Wp
ybAywr3uCnOKOHMF3JRduzcJlunUr4SwSLZGQ2G1Zbnd4JeUqJ66nfjgQolH
wvc7amHNqkslVbDid01LOou/fb71F/hcy4PnC+DaOSuag2FVkq9YY4V69Xd0
S2BNu6vjbvhifHDefLgpKVQ6bElJalu99yI4e971X2NhD3L+/irmzT/O0VpQ
4hScnhMKF7i2d4TD2c3HTm1j+dZNGuiegD6KKgNi4HpXfmg8TPpbPaPhCaUb
rOxgD/5vrQGw0+LgrIbxlMS064q/Yv2HhFflwKQ6fi9h9UrSPMNhkVfYouvO
lPC/XuMjYXGp2DEBLsgZshoLq/Pu2YjhGEm+8OkX2O8zrY4HCxadG8qDVYYd
B+5PxpgrtVgGm2oK9ctg/vpHvp3mmH/DJuAE3BR002w7LFPK/Y/Dyjd+pr3j
4FWPiwvg1EiDx+Gw6mTE5otwzNm7hq1muN/bmWW1sDzXhVsHk4VVOQ/gwWLV
6s6x2Me+g/3M2ef3hCpg/uoZghpYKy1wd4FVMy5sZ/mjKv3cOz+nxFvxfP9u
mNj70lJYVpYucYNFt+TmybBgfvMYHqyoDJT7w2p9XslVJ5zHnMgQZzjVTN86
Hq6nnTqtKb7Hlc53FsNy656UAripcPpTGzhjqWHicjhVlJJmCGtjjzTe/Az3
kuX66P0k9DHuZVr9GFjYkjwMqy/oompMsD62brsx5gtGzqgpN8a99lSKrGHV
HgP+6dFYF3L4qgvsbUL/OzMK+26cuGUmLIqo67pthD69hUVT2PzbP5kZw8rS
MxXj2fzDsS3RfOxrZ+uLXtTTLjM+OjAS53A5/0U1THsXmhyCBfn31iWx+LHZ
RyWw+tVf2zgWDznS1TEC96mWmH1wRP6pSxoUcH3YPYtrcFTiWgkP3iSt3pkB
K/Z4n5Ub4nyf1cRFwtT8D1meAeqtbQ4LZPGMWYcq9HFOrtNE/nBGqXRXhR76
9blUEsziehcaTvNQX2hnHg2rC+uuZBHc+7dbMhNhQUK7g9UwRwbFubPTYSWP
C5r7gSMZA28q97H+dj2SZr/niCZrlEcSix8SeImHODLhtYvpRtaP4GHopHfw
z8t+mMvy2fE6lw5ifexLgR6bv2RjT+YAR2SeDlHXHND3/lniF/0c8RgOMIiH
VadqG6bD8stpLWJY/UT8YHkfRxRhcckjYFoX2TfzLfLv+5s22yN/n33QP70c
KZjXHXeNuUYqnPwa9Y63RVTAqpUmf6b0YD/VzaMvwYoqtVFoF0fUG+wCb8J0
i9FKyw6O8Dd66TWyeHXi+metHGl63KDfbs/+A22Nd5+gXpDtyW5mzxha8ZAj
3n5GUzTMDsK2gLsciao2O/OQ2fXftOQbHIlJuOJeyrz6XdZ3lRzRBudHJzA7
paXfOIX+bCcXutl/+k8rr5LBYTwT6f/y6hvZ
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.8448950785695333`, 12.996142082543052`}, {1, -1}], 
         LineBox[{{20.000000000004547`, 18.000000000003638`}, {
          8.000000000002728, 14.500000000003638`}}], 
         PolygonBox[{{14.576, 16.418}, {13.312, 16.466}, {13.536, 15.698}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.565999999999999`, 17.1332}, {0, -1}],
          LineBox[{{19.999999999996362`, 11.999999999996362`}, {
          15.499999999995453`, 9.999999999996362}}], 
         PolygonBox[{{17.201713070827847`, 10.756316920367931`}, {
          18.135831542750775`, 11.609207699080171`}, {18.46074231559353, 
          10.878158460183966`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.725285377617155`, 11.811607900361398`}, {1, -1}], 
         LineBox[{{20., 8.}, {15.5, 10.}}], 
         PolygonBox[{{18.298286929172153`, 8.756316920367931}, {
          17.364168457249225`, 9.609207699080171}, {17.03925768440647, 
          8.878158460183966}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.72528537761716, 8.188392099638602}, \
{1, 1}], LineBox[{{20., 1.999999999996362}, {10., 6.5}}], 
         PolygonBox[{{14.452847096894937`, 4.4962188063972786`}, {
          15.711298774036582`, 4.368549795672764}, {15.383007032173545`, 
          3.639012591532679}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.968734750140364`, 3.4413216669785855`}, {1, 1}], 
         LineBox[{{7.999999999996362, 14.500000000006366`}, {10., 
          6.500000000001819}}], 
         PolygonBox[{{8.854478624978201, 11.0820855000872}, {
          8.757464374963666, 9.820900249898267}, {9.533578375079932, 
          10.014928749927334`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{8.101079124774735, 10.124069781193684`}, {1, 0}], LineBox[CompressedData["
1:eJwl1ntUjGkcB/C3tsKaGES5bcmoSbdx2wbFS2VnMzGlpZhsqE1JKpMmdVK6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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.522642987100603`, 8.952875305984765}, {1, -1}], 
         {PointSize[0.04], PointBox[{8., 14.5}], PointBox[{15.5, 10.}], 
          PointBox[{10., 6.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bfcgdgeh/fhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bfcgdgeh/fhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuCF1BjAx0VQIFYJATVAwQiKqLT3llgFbacUp4YWBMQXoiBF
xBilBkERRocoLYKKjfUVeZmqKKg4EUVRBGMpFm3QDJUYW5AAClRb7b/jncns
fHP2nN09e3M9EjZErrIlhMTiR0cy/A6PiHs/+nIkJJpZWwcn9g3wnD/iyB4i
2JUJC1z0i8TwqOnzK0Jh/WXtkymw1uSZNQZu99LKJsIRPvZZ190xv6jkGYHD
/GQLFHCqW3XC76jPNPxoCIZV3t3LjsKWxuU8G9h18XDQChr/ZuOFpsmIS5vW
iGHBwtlJhXCELPDWUx+O6JR/jFwC68e3XjkDm3o9CvmwKOiznFw4tVFTXT4J
888HlqTA2p5kJhQmFwvjV8GJbUx404cc4Vw7hEk0P2Pw3iK48ZK7xxY6vyw0
sWEiRzQyzcq9sPXK6iwOjvLf536U1g+dueWakCO7XW8fqYIVd6NMn8OiFz2C
X2B17aNHXW4ckaR1jTwBa2pjzu6F29clfZwHF08Xn5ZRVxnTYmFR55DTCFga
2JA5ja4Xvvlisyv6sbPFaPHG+cOeRxfDgtZqoRZWZRdnx8GKouOj1sGmTJ3P
ZBpPX86fDXNmRYbBBeOwt4iBrSG2IemwtbH01LAE+57eudgOVtnI7llh6eHT
in0TUM9Nc2oQHn4YIGdg1TZjAh/5NYuV1/LHoy9tFQ+FsP7SBD8+rEtZzfrD
lu83dec6Y73oE91zYfXSp+nOcLzc6BVC9/frnfoqJ9zHmweX6Xxr9NI+Odxu
LspxhKNetTmNhVU9a5VdWL/muMPS++Owjqr0SCVsNdjFHYTD6jMLUuDE8knP
V8LMgldmP7jRyV7pC+vifsvom4a+yZPFzx0x/puReRUm5hk+B2FVdf6LYlh9
XLU/BGYODFl2wJJkRbSBQR/+vLF+G80fPS1HTp3sEp0Npwrz88rHIq8tXF8I
My/T3TRjsP9Km1gtbKr8uuLAaNzzocSEy7C1wav58Cj0c2Zu4C2Yy5b41Tgg
//67AGp15Lz0XnuMJC2uBjZckG6dD0f5bq8rgaU2XzlU29H/iVC5nu7/u1rf
ebDWnHEtkMa/YPqbBTivWhLxcir69NPYSwmwwf9E8zlYZx9f8mIkxjIdXwGr
LTYda2BOe068gDp8VswtPt6X3GO9HrDKfCTNDlZIHJeNpnHjzu3+I3A+S3aw
gLrlbvvcD+CCmJkMbIpJCQzi4d6rvgym+dZ+3yIfW/RtxVbrLJjxDngjtkHe
TXVuGI2n7jR4E7wPQm3aEjh+v5dLwVuWtMt7AiJhTvoJj/cfS7THqvs/pftJ
7fn27BuW7D65x0Lrm0j60IHXLFGoz6cNTEH86Z3wU/+whNyQP6uFuZgsx5Zh
lkTk97srYVH9rkeDQyyRHHp8bw6syTL85QDH/+xR99YLcUHenNevWCKacGaw
GdaPkwjOv8R69096lsOks6I+YIAl+hYdr5jOd97QtbkP9YM6Z/xA4/si97v1
skRnlx1QSvM9Kmts/2aJNLT1dhXMLRzobjWzpHH7UMF1Gk8qFtd2ol6GsbKN
5ss2el7sYInhcdjpJ9RVFX41D1gSPEXm1QGrHs9OLjUg/qxsUxONT+3Ne3IT
6/fO4JdR59h2x15hicXIW6WkNl5NHFEOX1jkyFHTR133/nvuyf0PQQYoFA==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.194814593688687, 12.79766898841154}, \
{1, -1}], 
         LineBox[{{20.000000000008185`, 18.000000000007276`}, {
          8.500000000009095, 14.000000000007276`}}], 
         PolygonBox[{{14.816698078023695`, 16.197112374964764`}, {
          13.551893671999796`, 16.1806863437177}, {13.814710171952814`, 
          15.425088906352773`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740793031341028`, 16.859170034894547`}, {0, -1}], 
         LineBox[{{20.000000000006366`, 12.000000000003638`}, {
          11.500000000005457`, 8.500000000003638}}], 
         PolygonBox[{{15.19519314091517, 10.021550116847424`}, {
          16.152506936983112`, 10.848321122542462`}, {16.45710678118655, 
          10.10857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.159837801855844`, 11.07845105263581}, {0, -1}], 
         LineBox[{{19.99999999999818, 8.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{18.249230176602705`, 6.832820117735137}, {
          17.028889744907204`, 6.500000000000001}, {17.472649901887387`, 
          5.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 7.184877956223658}, {1, -1}], 
         LineBox[{{20., 2.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{17.250769823397295`, 3.8328201177351375`}, {
          18.471110255092796`, 3.5}, {18.027350098112613`, 
          2.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 2.8151220437763422`}, {1, 1}], 
         LineBox[{{8.499999999996362, 14.000000000001819`}, {
          11.499999999996362`, 8.500000000005457}}], 
         PolygonBox[{{9.712688721591656, 11.77673734374863}, {
          9.936153049242591, 10.53172180397914}, {10.638469507574097`, 
          10.9148035085236}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.244061861982704, 11.112579197445111`}, {1, 1}], LineBox[CompressedData["
1:eJwd1QlQU1cUBuCHsqmIQVEChEVFCAhuLAYbypNlpDWsA5KxVFYFIWrKIlgV
scaKCmNErAGRRUHj1JSAqDBSQIq1FSooaQ0iCKhMqKGgRokI0v+amTeZb87N
vec/72Xe0rg9YdtnURSVgot8U+Mz+KyiKSaBKU0JUpw+1MP8NrcUXXjqGjM3
FZY6RhdYw4b3D/26Cs7Zn5cfBBeKV4vHXGiKSk4sOwPLr/b/eROWctO4ariF
V/3uKPF7e4sti2kqvMosdBuccyrc/wEsqbzTTMMB7ynv0CU0ZTSubHUh+50q
sO6F2wxY6Svg3OsG9YlmNFU+PCBygMVxzW1v4K6rSza6wmw///lZTJrSNtv/
tAmWJ+QWvIVN3VOt4uGQZyN1seY0ZRej6ySCtaqqyGaY26tRSuEugdfMHAuc
s0+5vBO2Db414g0n9OZpNDCzqsMmBmZYXvBkIj8Vws9OhkX+WZMeZB6TslvR
sMKxLjkY5qS28zbCnMTyj9GwpPtytgmcvjw7JZHUJ9mcbpzP77v3MAFm2//8
+jhMHTKRRpL97YxPecIlc5w6vcn8jW96DiOPtIkntiL1t33rC2Fb0y3Fr0me
6q7WTXB9pdnHRjjrm7I+XdiPbbT4MJnvXyqXTsxLMdx52huWDK5VXoVVNYOh
k85Y77sy9hzsrLd563WYkWFytghu2T/UIoBtHZh7a2DN0ZFzbDhAL3/dM1hQ
2dGjWgmfX7h8Gc6Lcu+IqIaVtboPyf3QNdxcehBmVtzTHYQbn+75EA5zhHt7
I5Gv7kt5uAcslFk+74GVV5q+WAYnXYrhxpP517l1M2F53M6WMZgqcnlsDmvd
Uhz3W6Ke2X/BDhbnC7k6LJo6YLHImgPHWNe+OQgPiII/hMEhH8fLR2E3c9fj
afBARKhloBVNsQKLx86R/q5EKorgSqNBdhM5r6nn3UN4jccZmxckT8B4qgbO
k91vMUD+nAbPhTrWuB/aNIYDLD0xV6hFXb7jOpeG+QcztL1weoSsOpTMi72j
RwY3Tsy22UrWZ0aMCuGA7WkSPix8n1jGhp/2MU7yYAlD/74S/cqnH3W4k/uR
6nb6KMx44lxiCsf4sTa4wvzAaxdG0F9uSZDOMOZxrbT27xswY2Lr6wpY0iAq
/Z7kKQq8nQx3PV903hOm1kkKfGHRq5nLGifkbL07fzWpj+1uk8HCFc6fbWcr
MEyEVUMFUj9Y65P/4zI4YPoHOyEcNVPFHXKkKVooHZPDA7uDZ67AHONoqR76
U383KykDrk9zcxIQj6Td3QzH/PPk4iC8Rt1u7Ex8sVkWh/wChvuTJcSD6zNe
wXwf20gj8vvfpx9nYt7hazN4xJwy8enZNujT6I4FWa/1UsefgEvMg1rZcLpB
jq2uLZ7bdJXaB57yZTdw4Q7f8ck44hdl23bBA+mei47BB44H3imE1cVW7jK4
cB9LXQOrvBbwumGJ9ePANjjBecx3AvY7W9PUDvuZf2tlhnlMBT44/AecZOSl
sxZWvJH53Cb1mdowX7i++JJxFdxl8rKJBzPX5H99DC45/O9Xnx3rVhUPK+x3
8jfCmgV9+qRfRdGRGyvhys6K7IVwnnl/4Tw454U06hHyFgoviwfJPDKDUkQw
a7rm0i+wvEGxwRu283hlkw7nBoUe04eN5ggqXOGQ4faCPjLP0dZP/7HxP6+I
17sHUz5THlJY3Bze3QaXJ9lsi4XpZ08dlbDQvz+XBYdM1MfrYD9No1d9rwOe
h1qZHm1D+vt0txwe53B/K4AN9Y+E7YKzoiJOTsAMR3WdD1ySvOmBO/Kw502O
LoWFcXO7MmCW2Lt0HswccnjZAH9+P9ljH/J+Wkr/D3mMX8U=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.084117857803797`, 6.188306123204339}, {1, 1}], 
         {PointSize[0.04], PointBox[{8.5, 14.}], PointBox[{11.5, 8.5}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bfcgdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bfcgdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1QlUjWkYB/BX0o6bRm41rdosM2ih7fR9aREn0zJyouO4JSNLizbXUCpm
isKVe1RSrjalJBVFUoY2UQ3itnHLVhL3pEiL5v+e+c655zu/732e93ne5Zyr
7x/itVOGELIHP/omH2fwGLBknL6Xs6RCxulrLSz5+iho/i8scfJMehoH85v8
xJrUm6I3rIO5nqVXdOCmF61XVOFSPk+PC/N8rGv69Fki5nSdmQPbtX2+Wglz
a2ouvcP8gV8k8mfhlc5DLnfgUp3198JgXtOsskRY6G3j403zp1epusPH3JWm
rGCOhdhKDWaHzWu1YH49Y9O5DPP4ej6a1GOJj7T7Sy4slDuS9gJWeDCWfxAm
j+/OLYEF/VZZW2C3zNaCGFhqd+uTM1xcYJ+wHg48aLHVHm7SdgqYD8eG13k7
wHXZjGKOLvp0WGHhBsfGx5sshkPfZIT4wiKtpqFsHfTDt1HaS/P/XWJqDA9E
y09F0fjuAPdSbZZYyQkuHaL9ymuoO8EFkZ906XiVf6jp658xjx5bHkDrFS6d
Ogmn5Zm8coE5lpP3HeGCj8aHtWGfbcK/ZGC+2E/0YSlL2nkyHo1aqMO6/nEd
Dhw4knwKdm38LTUSlnSo+22CA3/0OtrCodIozYVwU3rO73Kw+FPemceaLPFo
WKvZtQROzIo/BLMXBiar4Com5aE2LB0VLsuFuRa6CRUaqHfG6EQGTI7qqTnA
Kx2vFmbCouHQvnouvm+uHr0Mt3tetHaBrfp8x27CCrzW3vpF6KdvVUUDbOq/
L8cVFpRHGj2l9d07FZ+oI++uYZoYbnLb0BUAS3vDtJ7RfkS23TLUbelt9+n4
k4KLxQvR/7ctGrRemvW6A34wVzmvLIbaPiJOF04sLjjnBnsMtkpe/4R9dN/W
ogYnJo8oFsGJMfvniU3R//hLw/2wleBwcCbMeqZHmcHilrjuXbDV85LGITXs
f031bRuYcz7tznk4kNl4nEvj3Y6q28Npu3tSZWCxYkJHxwL0MxL8btwE9ZVz
rHbAHJPdhhMw2xOdMaCKfSi69mw24gdCUi13w1ULlFMWwAUa/fcGOLh/nKh9
BrCgneHupTb28l5B65U1J7yfj+9eb09ZwqXOHx2V4bQBUzUz2l9ZiZHzPKyL
J3lH89vtVp88Nxf37K6Khzwcu8Nu0w8V5K15M/gK/cSqFw7zYb7nEmEJ3C67
vJHAsU5FTBR17tb+U8pYb0NUpg1MJvqFWnDd6x23ZoxxLrarfC8qYR2X35xo
hiWVQ/lqcKKZbGIGrHdd82y4ItZXut6FT8fv5e25o4C8WTfM/WCOQeXIsDz6
SYgI2Qzz5ijuV4K5L5MaqOusQ4PU5XAOpxufbYdJrY8ldw7ugUOESzDc7pvy
6yJZ3LPkFvMYWDAuq6I1G/EKLslJcOhJ3YmlMqj358a/U+h4dGWn2yzch3Oe
8gI4Njgu6xjBeZiofY+j9QNGH1yeYQhbMx2/C2ZV2sLDfzCk/f4ufQca/9gr
fec0QwpUbdI4sCh1XUD4FENEl4L+ERuh3/B65dOTDOFHCR0zYbJWfkPhBEPq
JJ2L/eFYkTYp/45x49V+y2C9bREdWeMMKfW4HTdliPo9p/P9vzGE13HbTAzr
teRmjI8xJNTUfqYWllwzaN4+inqDQqMKuC7yq+uFEYboqRQKymFRQFvzTSlD
uJLqsWqYhMg4HBhmCOeGn9pDGr917YewQYYMeI0rdcFsUAzvyFuGeDyffvie
Wkmaf6sP66/sjflM8+Pdi8J7GLIyx3x82JDu14ic/nOGmCatkeujznabOt6O
eL70eiONn+yW82pEvy8qfbKpdcwMhTUMEWRYZIdRWwimRcWIj+7zsqWmj6jm
//+jxex/Zul1/w==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.691266165654481, 13.045993650284984`}, {1, -1}], 
         LineBox[{{20., 18.000000000006366`}, {15.500000000003638`, 
          4.500000000010004}}], 
         PolygonBox[{{17.9397366596101, 11.819209978830308`}, {
          17.180790021169692`, 10.807281127576427`}, {17.939736659610105`, 
          10.554298914762956`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            
            14.940711462450594`]& ], \
{18.36530281910303, 16.579218250098275`}, {1, 0}], 
         LineBox[{{19.99999999999818, 12.000000000003638`}, {
          9.499999999997272, 14.500000000003638`}}], 
         PolygonBox[{{14.16631627118878, 13.388972316383624`}, {
          15.426331939733636`, 13.500150169490523`}, {15.241035517888804`, 
          12.72190519774223}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.557648941662487`, 14.638733081970063`}, {0, -1}], 
         LineBox[{{20.00000000000091, 8.}, {11.500000000003183`, 9.5}}], 
         PolygonBox[{{16.34087013529076, 8.645728799654572}, {
          15.228643998272856`, 9.248184623872604}, {15.089615731145619`, 
          8.460357776818253}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {18.956487102211344`, 7.35891451965572}, \
{0, 1}], LineBox[{{20., 2.}, {15.500000000001819`, 4.5}}], 
         PolygonBox[{{17.225505634327078`, 3.5413857587071793`}, {
          18.468751538144375`, 3.3082771517414358`}, {18.08023719320147, 
          2.6089513308442056`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.602053456673122`, 2.4998562220116165`}, {1, 1}], 
         LineBox[{{9.499999999996362, 14.500000000004547`}, {
          11.499999999996362`, 9.500000000004547}}], 
         PolygonBox[{{10.722834405812462`, 11.442913985468845`}, {
          9.905774917833435, 12.408529743989513`}, {10.648556270541642`, 
          12.705642285072797`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.665661129127848, 11.42434445165114}, \
{1, 0}], LineBox[CompressedData["
1:eJwl1QtUk2UYB/BPoaAUBAEdblwECiSYaIqaHPkIoYWYQCRrZ9Aml42Mi4KK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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.347845178753182`, 6.199236143002545}, {1, 1}], 
         {PointSize[0.04], PointBox[{9.5, 14.5}], PointBox[{15.5, 4.5}], 
          PointBox[{11.5, 9.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["3",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bgcfdheg/fhgh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bgcfdheg/fhgh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuAFRQlECIhCKUoAxVRoCwYRRHsvIhicqiFqBRwfZUCDIGYK
WipCqYw0UIsUrKLyUuShWAEDAuIDrE9GLVBrmY6SoBFSChoaIxgJ6b/TO5PZ
+XI2e/7dTeIWs0cSZ04I2YoXHckbEx43lozTUcCSK4IEs5twZJ2CuH7EEkn/
66ZDMHurV70EPvrU8+fPYUvPTXGhsLvYWO0Ik27e/RB4+8Dh2EE+SwIqDTt8
YVY2WdIG8/V8lQ28IeB8biEs33+Bo0Q/+6Yy5W5YpBjMqYKHZD1GESz1vbs4
AW78umOhB6xVuYUKYZ532wm9K0u6VPHfTIUf3FaGlcBFwrwNLxag3nhdHQS3
D6em9cCx61qWdM9liU/xuShqcYdStBWWlsjt++HejvvBL+fg/TkpNw3w01oD
Px7WPKmc5oz1l/2xMmLEhSWputjWAPhqxrZFe2HpjMf3xDTfiHnFFFh1gGO9
BY40lhef/JAl+e0Z/0bBsrH4V0thgaO/VRg8fH26m9oZ8/3PCtxhy7qh7KNw
jXWCZBT9a4Je7QuHUwPV4haat3P9OdMHyBOpskqnefNjzivgdq89l0NpXu70
9TFwZsETkyPMFb7358BdtpEh454YNyW9q3Jiyb2Dv7/RwKLJ0gNB8LjnzNVD
cGbJ4Mu7jiwpP5EsNMAuoQ1718JFeZoKB6zHcm4F9MzGfq7nzPWDvYmi5At4
XLtxnwT2E2gn+mbh/tNNlfGwtnrAIwn2qV1zJgW+asbTWcDyS5utZDTvd1+6
VDtgn3eqSqLhxt+mWkfAMvNHOxfD8ypqx8xg0Q7r4il0/ZFmz+aZLHEKjJnf
iXx8r6urdlObbhvyYPWx1X/zYX73xbWRsGXcytFH9ji3Jz9e9oIfz1/O2Qun
5tbvnAHLXRZo7WDpcb+zpvnoo3evqrTD+WTPTTRHvX5DaqIv3NIamT4TJpda
TjTzsK/ek2c+oesXckRBcP6FlLC1sCworftbW/TbtMs/ERa42xX9ZIN86zLT
DsLt4cnRDTMwRtUsPAyrAgqrnnNxrseKedn0PjqzKz1hgUzhtofeh62ZLs0a
56+Ulorgms8MD/uskHdZ+eAsuPeX/lPhMHE4bvkU+TN7ZReaOMg9WTntNOy0
LdbDBSb6ZzYJcPlzoWafJT53JLCUoZby0jqmI6/znx3u8HZ3x4J309AnT5oz
G/ZhV8a7wiIX7xvOcNHosiuLLdBfsOK1N61H87YET8XvLGv7NREsiNCWiqbg
XEJ90qVwTZExTmKOPqLc8UO0X4Q/G2eG7290lugU3B6WlJFNcB//VFdUwES+
QP+ViSFiFWtRDLN3/EInjQwZr39wLAvuyuLW104whA1c9XAzrNJxIpLfM0Rl
F+wvgHk3FHXhBjik9f3IPMwfKBC7vWOIdFftpwrYp0d8eHiMIfIU3Vg63G7x
KLHsLUNSW/Y3r4cz9fbrfPUM6VISSQCdP2BrV6pjSH2y79mPad3GeHpgFOtv
LFwhhOt/re7naRlSdE0oD4PL68Kr1gwzxGk5VxdD1+/rS9+lYUiv3/cdB2H2
QJKmUI38tWfaymCV8uKDFyqGaMXLuY20HrLm9LO/GCIoU3PbaL9JGS/rMUN8
fpBoGmAiVb/d/xD5BxKNx2m903HFxE3044fok2i9fFHzkUvoN8TJWUpNH/m1
///PPdj/ALrMLZg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.914847988868447, 10.361347240073592`}, {0, 1}], 
         LineBox[{{20.00000000000182, 18.000000000007276`}, {
          12.500000000001819`, 8.000000000003638}}], 
         PolygonBox[{{16.61, 13.48}, {15.57, 12.76}, {16.21, 12.28}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.403022683139532`, 16.54583280199964}, {1, -1}], 
         LineBox[{{19.999999999999545`, 12.}, {9.000000000000341, 12.5}}], 
         PolygonBox[{{13.900618875873821`, 12.27724459655119}, {
          15.117544188493639`, 12.62234281953293}, {15.081218059758719`, 
          11.82316798736469}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.488212026308705`, 10.854216844881332`}, {0, 1}], 
         LineBox[{{19.99999999999818, 8.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{18.249230176602705`, 6.832820117735137}, {
          17.028889744907204`, 6.500000000000001}, {17.472649901887387`, 
          5.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.004985304149102`, 5.815122043776342}, {-1, 1}], 
         LineBox[{{20., 2.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{17.250769823397295`, 3.8328201177351375`}, {
          18.471110255092796`, 3.5}, {18.027350098112613`, 
          2.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 2.8151220437763422`}, {1, 1}], 
         LineBox[{{9.000000000001819, 12.500000000003638`}, {
          12.500000000001819`, 8.}}], 
         PolygonBox[{{11.118364368108953`, 9.776388669574205}, {
          10.065894744940517`, 10.478035085019828`}, {10.697376518841578`, 
          10.969187575831763`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.131304063066693`, 9.903192049051874}, {1, 1}], LineBox[CompressedData["
1:eJwV1AlMU3ccB/Ang8omR40cBQvtFoxACxYEhgTic2GIkriC3HSjdNUVBGGI
rqBREBI7hXFsMCEM0OpWVu57rIxCgHUExjGFogHZBooHhxxCIs59/02a5pP/
ey/f37e/vPclySGnDCiKCsOX/FIrb/FxpSkWgQVNia87X1XD8hDD0V3wAjNo
M4qYGWDkA9N/jyWawzf3q8q/hil1UvaIC01xZ0S5m3DkmtrrB7ghx27tsiVN
qR5V1GXAgWU7pDZWNCU4FWUfD6cooo53w6H+QY0yuKDzYqzMmqbYXRUKOWzs
IeKZsGjqzu9aXQm81cxR3IZF7xxp08KUQR3XwYamOp4cnV2FI2uL/7kO9xWb
hTshn04YJRqH2SOsuVhY4JlssAXnSNPDiuCVgP7w13A5m9eqgRvmpCvT8Oix
azcewtwrWokSFr7mWLwg/cicGcHwNrtTRixIHb+zSPJphyrI9SyvkN1ZcO7y
r/d/gTNP8mv3wNz9qx4KOLK6/HEt5rvXszEVCHfYVrZGwEOv8tv/Q35dIYfL
IudjBrJamL47ObSOfuS5FhuRsP5ZhvolbDERWmpE+gxVtDNxPe2ZcLmdjz5n
5FPBMPMj6aVU4u94C01wlRU1cwhucE0fPIA8rOErLkx4xXa+rRfW6QSmmzya
8mbm5Ukxn+9Thv0yzOXMBpjaor+GyQ1yXhXyeUoLLHIb3kfu9zZbqQ/ei/58
aoSesCwz7eNpWPhv0SdSeOFWeLaQTVN8zZqkFKaL+F4/wsLf/kocJedtmRsP
YHFmj4DMc3NIE7kIt7Qqiw+S+eU7TWdgrV48H0H6aewZrIcNNdmMZLjqzFHe
F7Dg9Gc1F8h+UNlx78GiGG1GEjx76aepKuQxPma19yTZL/mTfndYL1QfcSR9
Duyr1GEeFt15e4n0dT7yVgJM7dyaVJF+JGVBXNij2s07mswnrn++jH7SzFOU
DNLnuKnvDLxybSCgCf3oC+Nsl2BRXmpMHCwrextN7vf93nfeGpa7UJpzpM9B
y7QpZ+xDjm5rDuYma16o4I6Ee9tnSV7nvhgFLK6pdt+FeULPd92Xw442Ts8a
4ZzxM6czYL3SaDTajvx/5sxceHSRk/wG9s8beaCGZ7OfNxfaYx9DT6xNwPIT
c/FMDvbVqWTiXeQJfPSV+hyH7EG1gib5skr92uFt7/jxC7DAJ21MD68ftMxW
kfleHXYklh8WN44Rf2tmTa6nU923luCUQ+EJ6bBQKVG/gYV/rP7sAOeUfbhj
m5xXuOm6kUeo4rs+hbU8u5oQuGDi+Jf9sGokXzuP/HNUBLMInmVcTcyCZ6fM
/YNJvvysQhd4wabOkkGel+G3+yX68JjuOtBC5lWUSMZgdh//bixcMNjkOw4b
D1NNJrCi2+/TDbKPvd+kdzthXx5HJPngeetnXQcuwoLwh4xKuGFP6J/+cEev
csoe+bjGSgM7mPqgZKkZdnAyWDAi91uEGUaQ/sj705GmTMj7k0v/D9JU6NU=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.508784489160853`, 6.008784489160852}, {1, 1}], 
         {PointSize[0.04], PointBox[{9., 12.5}], PointBox[{12.5, 8.}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["4",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bgcfdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bgcfdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1AtUk2UYB/BXhDkgxtgQJimXbeoUKxBFSU7f1wzcoZBFXgbLJC84RXDl
BRCwCYlASoalgOCZctEClcQAU3QSCl7goCGiQnwcCVBRBwKuwXH9v/I7Z+c5
vz3nfd7neb+L15ot4eutCCFf4MdGMmTB5UWT/y53mpzO/4F5ABvuCgXb4Tr9
K945mKbTn3TCS+OkzlmwZ9N983seNBmNT+5Tw1qXkpa18J2fRgSzYOUH16fs
gI1O8/8Y9EQ8tlOkgWeIEq9UwboVQfEUTARiVSJs2MhLs4GH1h2Zvgimd4p6
r2I/Pe9Qwxj+VyZt4O+HAy/qVqbDhrIZe6LhTzgXk0Swtm+NQg37yBeVVyIa
Lk8o1iJqdt2SqhCN2pG4E4iKm087rBE9H/e+GEOUBj8sq55GE6Z1xepo1OHz
hT/Gwoa9X796CDOptd4SNu/0ob8MfTF/jQvvTkWdgCVNavij5rNxOthUcetQ
GtwxrUougVUMbVMAJ3c0Gi6+jXmfnigqgq3v9awOhY2RaYoj8EK/CU1tbjSp
UGSZ9rD17u2tUMP8IJ8UDXsud+I/7ZpCE1FaVK2CdXq3eD2cIJO/mAPXd636
2ChCP1PnXHaDi+OG/dPgimxNpguccTNe7gEr3e5dEbPm9X5X74o5vg84SsGK
ltDnX8HckU5eDJx7sGHWdLjidQnnGDwgc5z70AX3yfkwt4Ndz1jlHYC1gfq+
ybjPUxO+4chhfZfmqAI+lx94amAyTWTaSq8tbH5tVXk2rDxUwE+HxyOvhc+E
DREnZ++DG+Ne+VU7Y74wXVIq3Pql3JWGo1Zu8t8Mt2wera4Xoq/ybcFL2Xzo
kmgFbGrPP/4uLCvbe7BZQJOa86oUIbzPrjMiEm5Jd99sBS971O743AnPweF5
xRb0z31QN5oF67IqNjiw9fMv8HxhVdv+Km94WPObSxcfz8ft00tV7Dyti2ty
WFuumrPhYoHivoJ1wJrpN+D2mpP8cUesj7YvmCjGeUaNPl4Fa62aN8yHTWOB
Db/ycJ7XSgsjYf1Y15RhB8xzVm2nhRvPzK2RwAvLQo9vY9ef8eDI30KfKZaS
jbDSklcfbo86xuU7wmHZs/hstR1NDlzKdJ8HJ2wvdF5ri3rZ/UXOcE29zzwt
F/vkbfUzs+/pxDpO5iTcv5nCkCewSClxOMvB3Kl/DvfDuXm2Oc9sYEfVYhNs
/HswYhGsy2mzuKKepvP9slxr5L23VAXB3BjVKQL3qxi7FDafIRHETkQfAT3K
Gph/bPZQmxX6sI8JH4SZ3cbaBTDjXywTS1CHP0eSNQFzuPrbBcNcm0vXbhD0
Ydg1FAHrFcNbBy0UYZJt7VfBOutUf+/XFNE4tDYqWf8y8HvIOEV0L9edXwgb
wyYnq80U6W8r4bjBeDBMm0wUUU3bVmfG/j5mXUf6KEW0hRmCbrafK+5eF4Yp
4lPwc8Qd2PiOTSz/JUUaBxKSWNP9up6sQYro/bKXPWLXhwnzZhkp0s5sTLRG
fXp3/2dNzyhCur2C/GBPY7tY+xT9+TbMj2H7P2XNET+miLKyL7OU7Tfx8+UN
vRTJ2Flu2wUzZpFncA9FZCOeDE+K9yDXtzqzmyJ0iTnaF9b9s5zYd1KEe5t/
fTFMYqTq2DaK8GesXB/E5ksnfZvTTBFFfXvPAjafFxNiuUSRqDOVxEP65jt+
oPb/KKX/BRvvHSQ=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.914847988868447, 9.638652759926408}, \
{0, -1}], LineBox[{{19.99999999999818, 18.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{18.249230176602705`, 16.83282011773514}, {
          17.028889744907204`, 16.5}, {17.472649901887387`, 
          15.834359764529724`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 17.18487795622366}, {1, -1}], 
         LineBox[{{20., 12.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{17.250769823397295`, 13.832820117735137`}, {
          18.471110255092796`, 13.5}, {18.027350098112613`, 
          12.834359764529726`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.004985304149102`, 14.184877956223659`}, {-1, -1}], 
         LineBox[{{20.00000000000466, 8.}, {9.000000000005457, 7.5}}], 
         PolygonBox[{{15.099381124126179`, 7.77724459655119}, {
          13.882455811506361`, 8.12234281953293}, {13.918781940241281`, 
          7.32316798736469}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {18.48749849569805, 9.147412217828109}, \
{0, -1}], 
         LineBox[{{20.00000000000182, 2.000000000001819}, {
          12.500000000001819`, 12.}}], 
         PolygonBox[{{15.89, 7.48}, {16.93, 6.76}, {16.29, 
          6.279999999999999}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.40465226442821, 3.4544723157306363`}, {1, 1}], 
         LineBox[{{9.000000000001819, 7.5}, {12.499999999998181`, 
          11.999999999996362`}}], 
         PolygonBox[{{10.381635631891047`, 9.276388669574205}, {
          11.434105255059483`, 9.978035085019828}, {10.802623481158422`, 
          10.469187575831763`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.131304063066693`, 10.096807950948126`}, {1, -1}], LineBox[CompressedData["

1:eJwt1Hs0m2ccB/CXKurExHGpg6ruqElEXMrQWsWlnVuVCtrKVhwcjENVSy1b
barSlta0qBq1SbtYTKPHpa0g3RQz90tUZxvDijqVcRQdte+zsz+S7/nked/3
/J7v++bdE5l0LFqZoqhgfEhSdPLF4FA0khYcisNQnt0O3zo78hkXdlGvVN4F
GyzWqd+Du3W4vZ6wzIR9QwHTktjtn8JUcAHfBsn98dr5VqSIxv05EmnaQz2g
MfG7aURgNuzSxrp7Cu6/fW+yCPb7rdm6DhY11fkXw+qGNXqqlhxqwj0iLQde
G9PyDIR5Tq6OcbCKUii9ABZutHi7whK5z+mnsJfBw92acKyIxZ6DneL/rpJj
vq9vOyptwOM2L4wqYH5lpHgT5pde+j4R5qXYrb2Co6aGfTzgm67htQMwqyE9
1Qym2+tliWCJysRLHdjAOW7wHKwydaiYmGZUJeCQ+Vz3RZDjpYwi+Q44le7J
PAxLuP7Lcuxv/JnmL2lweF7fczHMOqhh3ghPd2rRc2H7HcleSphfOPhJTAZZ
Z28WH4OTzzaPpMMXz7zU+A42Y7aFCeAFhjjmDekzSboggqVdhV94IZ269FLH
kIrm0qw80nuxy8xOMt/c1p52mJ+XEhYBT9tlDi7CxoL2ww9IX459S+rIfK3k
SlUW7q/++iQd5lbZOp2Ep49aaJB1Y7W1UREs7LV7Ts635zz+9hUc/jTofhe8
FiB2s7DiUNXGiZJS2Mxb3ygI7vaglcXA6WGyjERY4sa6YQ0HxHxcfQ6uuHCw
j+zHWJqlmgT7+fDiu+HZZ+EzoTDnvTuGIri/vrJwH7zhwV/Ph/mZT2y2k+uX
fBN0GU5oGUroJ/MJXbMLyPMlXY4sgdca5qarSZ9TBUejYNP73JZRWLq/OsAB
nm099JE2mc+2Zy8dptbVLEJJn5LRlVXsuzOkt+Muk/SoarII5yqu89ZJz502
mmRdPd350RHSY88FK3J+RUqesNyS9Ej97gy3xf5kOEueK/cofgrp02H+qjnp
zSH+80bkRL6HTiiSJeeWkP2sjRwQpsG3cnVXjpM+r7gzc2DFgHLsD3C61c2S
bDL36+uNb63I/VMsniHrT1bGvNjYV4rNXi6L9LZkL4ATxLU0BixabBp5BPPk
cYzXZD5Zd/mvMH0mJK0VNtj1OHYBVukosroMJ/yzW5eY1j3PD4UV3jlXyfEC
2R1/NvzQ2LJeCsvyetS04eqo88EFsIRXxlWGXfoMpk/B/OjOt9vIc1f4TgiD
rIcODejDwy8+nFjC/AmM+mv7yf9mm1uDDNbdEi0mkvk6ZJ1FsKm5oqmGXK+5
XJBO1v8KnF8nx3v61saS+/+ue78v9ucXfcIkDp44MHSyAu78o7wmg/Rpt3Vk
Gc5cFr9fBg9/cPoSB2ksTszqQ9K+GnC+iDw+OZCghfkCFsS2TbCKtu/qCXhi
lbf5J5x5hfmlGB6Xb+58A//3/rT+//3J5vwLaMzlPQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.508784489160853`, 13.991215510839147`}, {1, -1}], 
         {PointSize[0.04], PointBox[{9., 7.5}], PointBox[{15.5, 15.}], 
          PointBox[{12.5, 12.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["5",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"af/bgcfdheh/fggh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["af/bgcfdheh/fggh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {88, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1AlQU0cYB/AFaqBFMQjIpSUcFugUTAeLUyy+pxFIUYegqLQUCkgMkIpR
O63lKHgVdLAFaVMRkIigURFw6o1oqlZoFQTPKCBoE4xAlJs3QaX/bftmMju/
+Xa/79vd9+KeuGGF1JwQEocfHQk3iUfAkn8fJ5bwF/1q0sHd63/xyIENJu/F
jdTnLf8agaNF2q2VcN3jtVOjnFnyilXaZ1DfWtWnhDP3710YDueE/yC+AN+V
6kJn0HhD4YrLcN5QdbvaDXHjuZjjMKdM2DYHVk2+cMuF1TG96vJ3WaLYcbgo
Fj534nTzLDjHjicNgluDKnkVs5Hvq2trfeDoxqiIALigLMFSCAsnRl/fnoXx
0PKw1XR+Q9nGrbDkgzhjCazoWjolGNYsEaW/hiURcZ+aw/GBDeWbXVjSpC7P
uOmK/PZ6ZgQ+F7rHTwnz82pbNrvSPj3MP8dI3FP0PRg1h5ak2GEULHduEyFP
02Yhcxnr4sPk+u0w2x4Y9AVcV/tIWknrhDU49qGuZp5EfxiO7lu3LY32vcd9
eiGN83pP9eKchdfHuuVwcpu4JwlujS1XLIYL9hUXPXXEerfzyQJq22BNMnWb
5UY+rMqvC52Yif4iPS64wE4x+shSmG8T1kX722edbbsMLpDIQvLhPFPEb1aw
YlXCtAGYm8tVtDggX2x4iQznyj4c5xXDCvdIlQGODvNMTYQlzcZtMtwLm70m
fQ4s2P3RiJbeW6XsSqc91q9+kx6Ae923+mlIHsya6S4pYPaS/LmnPb2fSemP
cHfp+uKTdug3xUZGbfC55zsPlsxvu/YN3LozZLh2BvLFWFyU0PdkKGjB+7Dw
eFGON6wZ0maobVH3kf8ZHly3N7dyLhx/fddDDv2IY7LXX+Vjvv/LtimIa51b
7klhgdyh80O6Xr23cybMckZTJizQ9lkfnY58l9XGp7Q/9TW/QNgpdLTHT0Dv
oX/RDRv0dyXrpzSYPWNYGQs35ZQdOEG/I4co7ZNpOFdjSrqezg9bOVsCa3uU
O2zdWTLwWWp6xVSc48T9en+4YMX8mQ+s0e+aDq9P4G4zB++Bd2jfVf1BsHD3
7arht/Eepq1j6Hw22Tqr1wr7vFf90AVuXbo/sssS9TI7EnmwpLHappMHO6uO
jKG+SrUg2TgFNhPdGqD97JzIdYS5prIjHO0/wMbuy7dQj/MV8Gl/FuXHrlrg
ffk6gTcfruNaOBbOU1ZOlcOqDX7iW+aI3zntoqb15baFSbBgZ0uHAVYs4GtH
zdB3wB9Z3h74blX1J7bAhpqOTXGwIj+0REewT3lq8C64ID88fB7ss0lvWQWr
mku/v/mGIT4vjFYnYc2ov6vpFUPynP48UEPj1lLTaxNDov0L3Q7SeNRoXC/H
ELVIlbQH7vZ8T3N3jCFNF9zcv6M+erH99ghDxE6dLql0vlh8aXyIIVZni+Kl
1J2ZvpGDyHcqoDiNOntg3Y2XqB8YsiMXZn29fpcZGdLqcOdiLd3P9ZCapD6G
xIsCGnU0HihdVvEc/QlvyDw8ET87qBt/xhDBUi5wLczej8z4uAdx3/7Yg7Dm
557mYB1DtjifTnxA49U1438/YQhfWfyMeGG9YXv73McM0XqpRa4wmZX+7bEH
2E+9umsOdf/kwohWzHfUhHtQ+z2vr9Ig3+Cwhb3X///jBQ3/jV7sP0RmLSA=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.194814593688687, 7.202331011588459}, \
{1, 1}], LineBox[{{19.99999999999818, 18.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{18.249230176602705`, 16.83282011773514}, {
          17.028889744907204`, 16.5}, {17.472649901887387`, 
          15.834359764529724`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 17.18487795622366}, {1, -1}], 
         LineBox[{{20., 12.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{17.250769823397295`, 13.832820117735137`}, {
          18.471110255092796`, 13.5}, {18.027350098112613`, 
          12.834359764529726`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 12.815122043776341`}, {1, 1}], 
         LineBox[{{19.99999999999909, 8.000000000003638}, {
          11.499999999997272`, 11.500000000003638`}}], 
         PolygonBox[{{16.30480685908483, 9.521550116847424}, {
          15.34749306301689, 10.348321122542462`}, {15.042893218813452`, 
          9.60857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.159837801855844`, 8.92154894736419}, \
{0, 1}], LineBox[{{20., 2.}, {8.50000000000091, 6.}}], 
         PolygonBox[{{13.683301921976305`, 4.197112374964764}, {
          14.948106328000204`, 4.1806863437177}, {14.685289828047186`, 
          3.425088906352773}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740793031341028`, 3.1408299651054534`}, {0, 1}], 
         LineBox[{{8.499999999992724, 5.999999999989086}, {
          11.499999999992724`, 11.499999999985448`}}], 
         PolygonBox[{{10.287311278408344`, 9.27673734374863}, {
          10.063846950757409`, 8.03172180397914}, {9.361530492425903, 
          8.4148035085236}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.244061861982704, 8.887420802554889}, \
{1, -1}], LineBox[CompressedData["
1:eJwt1X8003scx/FVQtduTRNWRFnZ/HYwlGqY3LKGHbuWIyFKTTeHVbo3SS1x
0239dJO5uknrRGmVM/fiSkrJjW7d025+RFTKsO5dx+/c18fpj3mex77f8/1+
vu/v58yS+F3CxJkUCkWCDymFSv7Ycym0L/WUrykyQSmLNQvXot+YjA0sQZWK
T64/om5s7e0AVHZW7d6EsgwPC3ei7eGiFQYsLqWM8fl+MVzcE1fEgc1qVqnb
YP1y9uRmOC+oM8cKTZHZCTJRdZJNTByaLjftP426BUxwlGh0seicApUO1H8e
RH1cBRmFqNOc5IsebC5FRy3gnIR5nH4bKfxcHFm9H/abvHnpOsy9dDY+Fq5I
KZV3w/sbxE2rYCWj84yJA5eiOX7tqTmsP3/4jiMsdYvJGphe53axP8zULR29
B1dnfVjHhym/PI4qhHWmb2qn3Zp6Jx2OnVIJA+HYxOiPUXAOddUMd5j3tbCG
B0sZm6wtYJbK/YA3Oe40FDhM1rtD4O8JG6+ex38GH+xmqVYSX7D2KifO7qEL
yHvY7Us/CpfsUa2WwNpA3Vg8bLUo9uIpmCvt0wYQV79YS9arZgjqWeT8kcxX
47CcenehOTnePefJCjJP9918KplHhbf3AeIA20jiaqsZZg1wDs3rJTk/aV5C
0FdomPbxXCe018X2qgA1/pB2PwTtO3PK5QRKTZ2ZtBv1e/DmwyOUsit06grp
vlbHSTR96rLfa3Sk0THHHnOwDDyevZTMO6t2IBgNs0023oZW/FR5bROqG/qu
oZzMc2D7viS0pH+qVE/mbWM9bV0P/YKvI67z7hYnBlZXyYq+h7keG39dDzcX
qf6+A8tlQiNXcp8NZYr38POb4ikq7PnSqdDMCXPs+Bj/Futqnvyr2QumNSZ6
15B1dtCO8eHm4YmC02QOiWk/i+EuA36FBG4fnmUTBfN4otxgWC4qvxEO6/r5
dDY5vvWWHxdO978/az78cCSNZg83/DbWOAsuK2+qM4LFdj17pjDvMM5pm16s
T9l29MFsHG+gdrNq4Tpb+mwLMtcNBUP5cJ7T8wVk34cxPHLTYJYoMnUjmfOR
0FEhnHLoqGsuXLiQvtgH7tOVGt6FZXs7FUw4rKoxbgq2Pe/8ggEXJ66XBGAe
stuezyzJ/QSPbXLgkSu1K5fCSddb4p7AXasrIjjk+uzUuXS0vX3XaAQ60cib
T8o0DinKIOtzLcw4ge73ahbdIM9TmpF/j3xf0vxPHypV7RkcRM2y3+ez8PzS
jKqtNLT3h9d1yWie9j8emRN/dkjULVQetNfLAzVWdYePodExO0I9Uf3blpNr
nHHevHd0RziBRV2QBaecN9luCWtKLMar4Wih28lJsi/Mvi34CFsp8xe3ww9r
+XJrF8zJ0LSX7JPqOQ4ta1zIe2L25MG2mabKSNhyMN1nC5ze0fg0ATbY6WC0
Ej5jd0CyDZ441L/FHE7YVjy+GY4+V8YdJvuTfXtHKDk/uGDjK7g4KH2MA2tm
spmtsNsiha8lbKV5VtcEH2zL0+uxvrxyhq4FVu7T2LXAMjHTpAuObtNrlHDF
ujjKOPldiTVwkMGFJs4Rtri/p1eq9RY4mXZsWSiZb93yc8Fw3Z+/c7Jh7VVz
fw/YIGbTKHkfdW+7ZPZwq6BYYUj2tZGVdBkcsXzCjux7H52m3hlWqitTTsHq
krt/cGG9LDP2BXlvly3CY8h81aIOS1TDv/HpCNog0/4bQa57tfNRJXHcg+5c
sm65q3wINQvhu5HvmU2ZNS54/oh63yoNeS/lljmpME/SWqkjc5A4jKphrSKx
aRKe/v/k+qUu3P8Bx5tWhg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.084117857803797`, 13.811693876795658`}, {1, -1}], 
         {PointSize[0.04], PointBox[{8.5, 6.}], PointBox[{15.5, 15.}], 
          PointBox[{11.5, 11.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["6",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"af/bfcgdheh/fggh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["af/bfcgdheh/fggh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {110, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1QtUzFkcB/Cr9JQM0YMeM9tDHqvpnJak/P+UtCtJkWhpopRQszWjpJiN
PYZajQ1570SlUnbIoQ4xpa1Ep2gdQ8WUJKSd7EQPbfu9uzvnzPmdz/zuvb97
7//+7/A2xwdF6RBCYvGlkWjG8PmKJf9+zFlSEq8T/gqWD1fn74Lzztx8WgPz
F2zzeQ1rs7puXoA1V+Kv+liw5OFpG8UeWKZRb8mCd+yo/xwAB5rHm1TDJfvz
N1vTvG+0phVmx2RRXTw47tboC5hjNWhUCCtKRo43wW7Pwx7FwJp8zbTrsFb1
Z4YzzI+953QM9jerdXrLhZVJ75NgZ9eOXSVwoGUQLxquKWrVF8HCllkrYuDU
rRm9LKwJTP17H7xm/LzLZrCyWBylgCU9F75T2SHminSHYHnOIJMNk5U7qkMs
MV+XvaVrYVlV2fIaWHXB1Msa5nxwPuhhxZKTVn1Or2xRz7v1wFVY4H/5VAEc
aC/i86azpNzTxH4LrDjYtDQdLuyuNjWHZSkjek2w1O+uotIG/QKcPurOYMkg
65C9Hm6ee9GUCxsu1KS/s8b8B06bOML8/aO7E2A5J9PSAlY9uhGjReRY5kR/
wXisPPh1Iizf35PZBqs7l9/XIkpss4ruIlqm7alLoe1uW/xciqiMy9M3QhRM
evuiBLHZtrQhF+vQ6MZGV8M9BQnzfWF5RHamhubtD8/9hH3gRhlyFqBOc0R0
kAJuXnV/WjasTD/yTQIsq6tLGINDSx8fcYfVmR88xJh36Oi2s2PYZ+6hkH29
sEQ059A9eg4uBjlvxLqVjR1PJLAsvr7lLt2XjsgKN+o0K7kp9i25wHlnO86h
sErZuAzuaTPalAKzj1Y7RcKFTjM8TWB+foPNdpjvonxzbBrGWZlXvAnmdtss
toCFvF6VD3wyK6bhxFSsz9HcwRGW2KdHWsCCkPAiA9hZ+33dWTM8z4DHflob
+hySZ8+G2a72h/2wtCz1bdUUuKXcRY8+56ho7yhYvff0FFdYOmu1+1RYk5Ik
FsPsla2HGicjn6N7s4nOx4DXegQWcofaPHCuBhMmfloDB452WitgobmrnTUs
WLXxAT23/KqMvWoOxi0pPOEPO9eWS8/D3An51/bC5Ot1gmA49PN6q0uwpFL6
7fAknKPmU0336HsQtX3FfJi8SJjxB8yGr+wNM8V87jicVNHxF82siJuI/S4T
O7bQ/swWH5EJ9r3D9Xot7R9WlZk4gSXuHWEDN+h752DrJTLG+CG3tLSebOjO
szQj5L1Li87RfE5V/wlD/H7U8fAZmt/XGltjgPYDx+bkwc0fqocNYb+zPSPl
sJp3vS1SH/OvXTr9OSy3nfLgiR7anboYrI97oDn4sdc6WPh3u/ciem/MI286
x6N/XUCOGFZ7jYT/AEtZv61X4UCfp9phXVgVIX8HC44bJSfDil6nVBt675UU
G3TpoE7+zJe+MMfD7reFcCGXLYuk915/Y0vSOFjcZ7eL3mN+8yN+JainL8ul
9x4pDw2PHmNIT5rBF5oXZL0MFI0yRJbsYbyd5gcGIhJHGJLcdSY+DOaKxrUm
DjGkfImLmz+s5LxsP/CZIdJE5dol9N6tMK4tG2CIcLfbhsU0PzWjwljLkEHP
ir5ltH+O0S9ZHxki73Q/v57Wu2/g6d3PEMuagYLdsMRRfI6nYYhA0/pXHq0/
u8rEpw/1f2c8ntHx9MIYYS/au6nczezhvA2pxe8Yohp1nbwKZs3i3rf0oH/9
uGtSWFKn+bGpmyHuy9773oaJoS7vpy6G8Csrc7tp++TYuV0dDFFznh/Vc0Be
17GgoR3rSxZwLWFJ4lqF+VOGcFcrim1p/tIcm4wmhkjqI1TTqVeMxekosX+f
Hu6c5PD//5Cs8r/owP4DEF1nnw==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.691266165654482, 6.954006349715015}, \
{1, 1}], LineBox[{{20.000000000009095`, 18.000000000007276`}, {
          15.500000000009095`, 15.500000000007276`}}], 
         PolygonBox[{{18.274494365672922`, 17.04138575870718}, {
          17.031248461855625`, 16.808277151741436`}, {17.41976280679853, 
          16.108951330844206`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.602053456673122`, 17.50014377798838}, {1, -1}], 
         LineBox[{{19.999999999997726`, 12.}, {11.49999999999909, 
          10.500000000003638`}}], 
         PolygonBox[{{15.159129864709238`, 11.145728799654572`}, {
          16.271356001727142`, 11.748184623872604`}, {16.41038426885438, 
          10.960357776818253`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.955774102824826`, 12.642409183862153`}, {0, -1}], 
         LineBox[{{20.000000000006366`, 8.000000000003638}, {
          9.500000000005457, 5.500000000003638}}], 
         PolygonBox[{{15.33368372881122, 6.888972316383624}, {
          14.073668060266364`, 7.000150169490523}, {14.258964482111196`, 
          6.221905197742229}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.558362036635682`, 5.360757775890333}, {0, 1}], 
         LineBox[{{20., 1.9999999999990905`}, {15.500000000003638`, 
          15.499999999998181`}}], 
         PolygonBox[{{17.5602633403899, 9.319209978830308}, {
          18.319209978830308`, 8.307281127576427}, {17.560263340389895`, 
          8.054298914762956}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.367441432857568`, 3.4208830744510736`}, {1, 0}], 
         LineBox[{{9.5, 5.499999999994543}, {11.5, 10.499999999993634`}}], 
         PolygonBox[{{10.277165594187538`, 7.442913985468844}, {
          11.094225082166565`, 8.408529743989513}, {10.351443729458358`, 
          8.705642285072797}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.665661129127848, 8.57565554834886}, \
{1, 0}], LineBox[CompressedData["
1:eJwt1X9cy3kcB/AvOhVhKUp1hJRSiVZCmLM0tVOXZDL6oaj8qBQNqSikG8m5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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.894453345363303`, 13.363477323709358`}, {1, -1}], 
         {PointSize[0.04], PointBox[{9.5, 5.5}], PointBox[{15.5, 15.5}], 
          PointBox[{11.5, 10.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["7",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"af/bgcfdheg/fhgh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["af/bgcfdheg/fhgh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 0}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwd1A1Q02UcB/AHnIDzbYokG0hDEKnDcysjMOH/FwTzfY4XES0HEqC2wzc6
vOWcUB6R5KBUJM3hIUK8OIHFIMQpJkPpWFK6hIMhIssO2ykvw4b0ffrf7Z77
/H//5/f7PS833+QM6SfOhJCd+NGR2KfwvMkSHsUbLMkdepL1OyxipPFfwiZB
1dZLsKk1euA1LDpR1L0bJlNZXycvZIn9eW+gB2xQLpVWwJ4K+/f1PizR1sxP
MMEqbcHiSFjSEr7mT1hkvOp+exFL1LpK17twwv1e3SpY2FHMqYJ51lZlnTfq
3YvQ58GaOOXS5TBRNS48CEs2723QedH+2rs/hc23gss3UB8xT2TT+KbSiJcC
2PvCcT1szRVtqIJ5nXa32Z7o60E+9zCs/Yz/swL29K7TRsGa1zo3B6yXp3K8
qIvWW7L5LClqYKKHMYpePljNwXvT+N2An6i5Z0f+z/OooWo/LHy8TdgB23zX
P+PDlrc81Bz0pWdE6deQl/fPUMYimKc5ERwK2xTB+wSwraK4RIc+SZY5YBps
US45HUTXHRXx9mPa95XY26XYd8nlg7taaV/vp+r86TnId8pq4ZDRH7uueaB+
k75TD3s2P6zYAmtb8id64HTmN9nUApxPlXJYiPxkz47TN2Bt5w/vZsPs5vq8
k7Dm5tGAV3CgeLBWRt19dlCJfTec+OjIe7Ak+uJ1Ds5J9GjiHQ5sy1zbdQzW
roic1+aOPodvdQ3CdkH42BFYsjtGvhznnttezPWC2ZI1idvhkIk+Ujkfrl5r
ToHZ5sl/g2D1nSDFNnpPPh/yLpuH9z0ZSSthW+nxNm/Y8PK7IU/YnpnXeJaH
+JqmGVOoZ4moX7AIlgXKa8bpvdo7Opk8F30mzeLPxPfkF78XNXMwf9uu5jCY
pxS7+MCqS7FFp2j+k36hlbMxBrisdMAGdTSzCTa5On+swL2WHAj/Y3wW9nd1
zxwnIerq445dgCW1+eERsEUtyRTDCaqBUwpYJr16uGIm9kWsO1QOq87Ygqe4
OJdYsfAObChRxovhom8Kx7pgdXv58cgZWFdkyzlqU1nKSJgb4u3f1hphkd/4
hiBX5Lsk36uH93tK+rxcUC/mUQ3Nz2MzzfzpqKuRbb9Iv1+1NVbMQZ1Wn8QS
aqP01z3TUN9aOa2Wzj90kXvHmSXG6U8L7tN+shubImFzn83h5Is8HVxjtxPW
db7TJwxmC9ftyIFtcWEBx2CV5mnqMjhkU1r0LVgbsiyng7DE7YYyjyzGdwfC
HsbC6YOOmhUw+WBPoWKKIRK30v5EmF3Sl+HymiH20f6NB2BLTrLfRgdDcqsb
ErNgg/++lIRXDNH0XLh8EBbGcwWMnSHCpPxzKdSzx7/gjyEfqb8fQz0qL3Me
YYjqiUkSTeen+ae5v2Dwf5R0fjUsG6ng7rIxpNzU4VgHq9yT+i3DDNGHbulJ
ov21NvEK/maIcWQg4ysa58TubP6LIdaYxqstNB6Y9iLcinhvRMkk7d/brrY+
ZUhIwg4r64e4yT01ZxDxImNeNmyonmzqHmCIVh51t5nG1xVI2/oZku5Dxp7B
qntePfG9WO9onQfXH3HB0VAnM0M+rHXE8annPj9zxcQQUfHhAAF1W0f9vJvY
rzrpdB41fTTXiRsd/dn/ACt9Fmg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{3.8448950785695333`, 7.003857917456949}, {1, 1}], 
         LineBox[{{20.000000000009095`, 18.000000000003638`}, {
          10.000000000007276`, 13.5}}], 
         PolygonBox[{{15.547152903105063`, 15.996218806397279`}, {
          14.288701225963418`, 15.868549795672765`}, {14.616992967826455`, 
          15.139012591532678`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{14.968734750140364`, 16.558678333021415`}, {1, -1}], 
         LineBox[{{19.999999999996362`, 11.999999999996362`}, {
          15.499999999995453`, 9.999999999996362}}], 
         PolygonBox[{{17.201713070827847`, 10.756316920367931`}, {
          18.135831542750775`, 11.609207699080171`}, {18.46074231559353, 
          10.878158460183966`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.725285377617155`, 11.811607900361398`}, {1, -1}], 
         LineBox[{{20., 8.}, {15.5, 10.}}], 
         PolygonBox[{{18.298286929172153`, 8.756316920367931}, {
          17.364168457249225`, 9.609207699080171}, {17.03925768440647, 
          8.878158460183966}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {17.72528537761716, 8.188392099638602}, \
{1, 1}], LineBox[{{20.00000000000182, 2.}, {8., 5.5}}], 
         PolygonBox[{{13.424, 3.9179999999999997`}, {14.688, 3.966}, {14.464, 
          3.1980000000000004`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {13.566, 2.8667999999999996`}, {0, 1}], 
         LineBox[{{8.000000000003638, 5.4999999999972715`}, {
          10.000000000003638`, 13.50000000000091}}], 
         PolygonBox[{{9.145521375021799, 10.0820855000872}, {
          9.242535625036334, 8.820900249898267}, {8.466421624920068, 
          9.014928749927334}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {8.101079124774735, 9.875930218806316}, \
{1, 0}], LineBox[CompressedData["
1:eJwd1g841HccB/AfQyg9Kss1bawQusLYkqVuOXW45E9brDydYRjj0ilWa9f8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          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{12.522642987100603`, 11.047124694015235`}, {1, 1}], 
         {PointSize[0.04], PointBox[{8., 5.5}], PointBox[{10., 13.5}], 
          PointBox[{15.5, 10.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["8",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"af/bfcgdgeh/fhgh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["af/bfcgdgeh/fhgh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 0}, {0, 0}, {22, 22}], InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {54.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "l"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "m"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {77.25, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  GraphicsBox[{InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuCF1BjAx0VQIFYJATVAwQiKqLT3llgFbacUp4YWBMQXoiBF
xBilBkERRocoLYKKjfUVeZmqKKg4EUVRBGMpFm3QDJUYW5AAClRb7b/jncns
fHP2nN09e3M9EjZErrIlhMTiR0cy/A6PiHs/+nIkJJpZWwcn9g3wnD/iyB4i
2JUJC1z0i8TwqOnzK0Jh/WXtkymw1uSZNQZu99LKJsIRPvZZ190xv6jkGYHD
/GQLFHCqW3XC76jPNPxoCIZV3t3LjsKWxuU8G9h18XDQChr/ZuOFpsmIS5vW
iGHBwtlJhXCELPDWUx+O6JR/jFwC68e3XjkDm3o9CvmwKOiznFw4tVFTXT4J
888HlqTA2p5kJhQmFwvjV8GJbUx404cc4Vw7hEk0P2Pw3iK48ZK7xxY6vyw0
sWEiRzQyzcq9sPXK6iwOjvLf536U1g+dueWakCO7XW8fqYIVd6NMn8OiFz2C
X2B17aNHXW4ckaR1jTwBa2pjzu6F29clfZwHF08Xn5ZRVxnTYmFR55DTCFga
2JA5ja4Xvvlisyv6sbPFaPHG+cOeRxfDgtZqoRZWZRdnx8GKouOj1sGmTJ3P
ZBpPX86fDXNmRYbBBeOwt4iBrSG2IemwtbH01LAE+57eudgOVtnI7llh6eHT
in0TUM9Nc2oQHn4YIGdg1TZjAh/5NYuV1/LHoy9tFQ+FsP7SBD8+rEtZzfrD
lu83dec6Y73oE91zYfXSp+nOcLzc6BVC9/frnfoqJ9zHmweX6Xxr9NI+Odxu
LspxhKNetTmNhVU9a5VdWL/muMPS++Owjqr0SCVsNdjFHYTD6jMLUuDE8knP
V8LMgldmP7jRyV7pC+vifsvom4a+yZPFzx0x/puReRUm5hk+B2FVdf6LYlh9
XLU/BGYODFl2wJJkRbSBQR/+vLF+G80fPS1HTp3sEp0Npwrz88rHIq8tXF8I
My/T3TRjsP9Km1gtbKr8uuLAaNzzocSEy7C1wav58Cj0c2Zu4C2Yy5b41Tgg
//67AGp15Lz0XnuMJC2uBjZckG6dD0f5bq8rgaU2XzlU29H/iVC5nu7/u1rf
ebDWnHEtkMa/YPqbBTivWhLxcir69NPYSwmwwf9E8zlYZx9f8mIkxjIdXwGr
LTYda2BOe068gDp8VswtPt6X3GO9HrDKfCTNDlZIHJeNpnHjzu3+I3A+S3aw
gLrlbvvcD+CCmJkMbIpJCQzi4d6rvgym+dZ+3yIfW/RtxVbrLJjxDngjtkHe
TXVuGI2n7jR4E7wPQm3aEjh+v5dLwVuWtMt7AiJhTvoJj/cfS7THqvs/pftJ
7fn27BuW7D65x0Lrm0j60IHXLFGoz6cNTEH86Z3wU/+whNyQP6uFuZgsx5Zh
lkTk97srYVH9rkeDQyyRHHp8bw6syTL85QDH/+xR99YLcUHenNevWCKacGaw
GdaPkwjOv8R69096lsOks6I+YIAl+hYdr5jOd97QtbkP9YM6Z/xA4/si97v1
skRnlx1QSvM9Kmts/2aJNLT1dhXMLRzobjWzpHH7UMF1Gk8qFtd2ol6GsbKN
5ss2el7sYInhcdjpJ9RVFX41D1gSPEXm1QGrHs9OLjUg/qxsUxONT+3Ne3IT
6/fO4JdR59h2x15hicXIW6WkNl5NHFEOX1jkyFHTR133/nvuyf0PQQYoFA==

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.194814593688687, 12.79766898841154}, \
{1, -1}], 
         LineBox[{{20.000000000008185`, 18.000000000007276`}, {
          8.500000000009095, 14.000000000007276`}}], 
         PolygonBox[{{14.816698078023695`, 16.197112374964764`}, {
          13.551893671999796`, 16.1806863437177}, {13.814710171952814`, 
          15.425088906352773`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740793031341028`, 16.859170034894547`}, {0, -1}], 
         LineBox[{{20.000000000006366`, 12.000000000003638`}, {
          11.500000000005457`, 8.500000000003638}}], 
         PolygonBox[{{15.19519314091517, 10.021550116847424`}, {
          16.152506936983112`, 10.848321122542462`}, {16.45710678118655, 
          10.10857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{15.159837801855844`, 11.07845105263581}, {0, -1}], 
         LineBox[{{19.99999999999818, 8.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{18.249230176602705`, 6.832820117735137}, {
          17.028889744907204`, 6.500000000000001}, {17.472649901887387`, 
          5.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 7.184877956223658}, {1, -1}], 
         LineBox[{{20., 2.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{17.250769823397295`, 3.8328201177351375`}, {
          18.471110255092796`, 3.5}, {18.027350098112613`, 
          2.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 2.8151220437763422`}, {1, 1}], 
         LineBox[{{8.499999999996362, 14.000000000001819`}, {
          11.499999999996362`, 8.500000000005457}}], 
         PolygonBox[{{9.712688721591656, 11.77673734374863}, {
          9.936153049242591, 10.53172180397914}, {10.638469507574097`, 
          10.9148035085236}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{9.244061861982704, 11.112579197445111`}, {1, 1}], LineBox[CompressedData["
1:eJwd1QlQU1cUBuCHsqmIQVEChEVFCAhuLAYbypNlpDWsA5KxVFYFIWrKIlgV
scaKCmNErAGRRUHj1JSAqDBSQIq1FSooaQ0iCKhMqKGgRokI0v+amTeZb87N
vec/72Xe0rg9YdtnURSVgot8U+Mz+KyiKSaBKU0JUpw+1MP8NrcUXXjqGjM3
FZY6RhdYw4b3D/26Cs7Zn5cfBBeKV4vHXGiKSk4sOwPLr/b/eROWctO4ariF
V/3uKPF7e4sti2kqvMosdBuccyrc/wEsqbzTTMMB7ynv0CU0ZTSubHUh+50q
sO6F2wxY6Svg3OsG9YlmNFU+PCBygMVxzW1v4K6rSza6wmw///lZTJrSNtv/
tAmWJ+QWvIVN3VOt4uGQZyN1seY0ZRej6ySCtaqqyGaY26tRSuEugdfMHAuc
s0+5vBO2Db414g0n9OZpNDCzqsMmBmZYXvBkIj8Vws9OhkX+WZMeZB6TslvR
sMKxLjkY5qS28zbCnMTyj9GwpPtytgmcvjw7JZHUJ9mcbpzP77v3MAFm2//8
+jhMHTKRRpL97YxPecIlc5w6vcn8jW96DiOPtIkntiL1t33rC2Fb0y3Fr0me
6q7WTXB9pdnHRjjrm7I+XdiPbbT4MJnvXyqXTsxLMdx52huWDK5VXoVVNYOh
k85Y77sy9hzsrLd563WYkWFytghu2T/UIoBtHZh7a2DN0ZFzbDhAL3/dM1hQ
2dGjWgmfX7h8Gc6Lcu+IqIaVtboPyf3QNdxcehBmVtzTHYQbn+75EA5zhHt7
I5Gv7kt5uAcslFk+74GVV5q+WAYnXYrhxpP517l1M2F53M6WMZgqcnlsDmvd
Uhz3W6Ke2X/BDhbnC7k6LJo6YLHImgPHWNe+OQgPiII/hMEhH8fLR2E3c9fj
afBARKhloBVNsQKLx86R/q5EKorgSqNBdhM5r6nn3UN4jccZmxckT8B4qgbO
k91vMUD+nAbPhTrWuB/aNIYDLD0xV6hFXb7jOpeG+QcztL1weoSsOpTMi72j
RwY3Tsy22UrWZ0aMCuGA7WkSPix8n1jGhp/2MU7yYAlD/74S/cqnH3W4k/uR
6nb6KMx44lxiCsf4sTa4wvzAaxdG0F9uSZDOMOZxrbT27xswY2Lr6wpY0iAq
/Z7kKQq8nQx3PV903hOm1kkKfGHRq5nLGifkbL07fzWpj+1uk8HCFc6fbWcr
MEyEVUMFUj9Y65P/4zI4YPoHOyEcNVPFHXKkKVooHZPDA7uDZ67AHONoqR76
U383KykDrk9zcxIQj6Td3QzH/PPk4iC8Rt1u7Ex8sVkWh/wChvuTJcSD6zNe
wXwf20gj8vvfpx9nYt7hazN4xJwy8enZNujT6I4FWa/1UsefgEvMg1rZcLpB
jq2uLZ7bdJXaB57yZTdw4Q7f8ck44hdl23bBA+mei47BB44H3imE1cVW7jK4
cB9LXQOrvBbwumGJ9ePANjjBecx3AvY7W9PUDvuZf2tlhnlMBT44/AecZOSl
sxZWvJH53Cb1mdowX7i++JJxFdxl8rKJBzPX5H99DC45/O9Xnx3rVhUPK+x3
8jfCmgV9+qRfRdGRGyvhys6K7IVwnnl/4Tw454U06hHyFgoviwfJPDKDUkQw
a7rm0i+wvEGxwRu283hlkw7nBoUe04eN5ggqXOGQ4faCPjLP0dZP/7HxP6+I
17sHUz5THlJY3Bze3QaXJ9lsi4XpZ08dlbDQvz+XBYdM1MfrYD9No1d9rwOe
h1qZHm1D+vt0txwe53B/K4AN9Y+E7YKzoiJOTsAMR3WdD1ySvOmBO/Kw502O
LoWFcXO7MmCW2Lt0HswccnjZAH9+P9ljH/J+Wkr/D3mMX8U=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.084117857803797`, 6.188306123204339}, {1, 1}], 
         {PointSize[0.04], PointBox[{8.5, 14.}], PointBox[{11.5, 8.5}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["1",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bfcgdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bfcgdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {0, 22}, {0, 0}, {22, 22}], InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwl1AtQU0cUBuAFRQlECIhCKUoAxVRoCwYRRHsvIhicqiFqBRwfZUCDIGYK
WipCqYw0UIsUrKLyUuShWAEDAuIDrE9GLVBrmY6SoBFSChoaIxgJ6b/TO5PZ
+XI2e/7dTeIWs0cSZ04I2YoXHckbEx43lozTUcCSK4IEs5twZJ2CuH7EEkn/
66ZDMHurV70EPvrU8+fPYUvPTXGhsLvYWO0Ik27e/RB4+8Dh2EE+SwIqDTt8
YVY2WdIG8/V8lQ28IeB8biEs33+Bo0Q/+6Yy5W5YpBjMqYKHZD1GESz1vbs4
AW78umOhB6xVuYUKYZ532wm9K0u6VPHfTIUf3FaGlcBFwrwNLxag3nhdHQS3
D6em9cCx61qWdM9liU/xuShqcYdStBWWlsjt++HejvvBL+fg/TkpNw3w01oD
Px7WPKmc5oz1l/2xMmLEhSWputjWAPhqxrZFe2HpjMf3xDTfiHnFFFh1gGO9
BY40lhef/JAl+e0Z/0bBsrH4V0thgaO/VRg8fH26m9oZ8/3PCtxhy7qh7KNw
jXWCZBT9a4Je7QuHUwPV4haat3P9OdMHyBOpskqnefNjzivgdq89l0NpXu70
9TFwZsETkyPMFb7358BdtpEh454YNyW9q3Jiyb2Dv7/RwKLJ0gNB8LjnzNVD
cGbJ4Mu7jiwpP5EsNMAuoQ1718JFeZoKB6zHcm4F9MzGfq7nzPWDvYmi5At4
XLtxnwT2E2gn+mbh/tNNlfGwtnrAIwn2qV1zJgW+asbTWcDyS5utZDTvd1+6
VDtgn3eqSqLhxt+mWkfAMvNHOxfD8ypqx8xg0Q7r4il0/ZFmz+aZLHEKjJnf
iXx8r6urdlObbhvyYPWx1X/zYX73xbWRsGXcytFH9ji3Jz9e9oIfz1/O2Qun
5tbvnAHLXRZo7WDpcb+zpvnoo3evqrTD+WTPTTRHvX5DaqIv3NIamT4TJpda
TjTzsK/ek2c+oesXckRBcP6FlLC1sCworftbW/TbtMs/ERa42xX9ZIN86zLT
DsLt4cnRDTMwRtUsPAyrAgqrnnNxrseKedn0PjqzKz1hgUzhtofeh62ZLs0a
56+Ulorgms8MD/uskHdZ+eAsuPeX/lPhMHE4bvkU+TN7ZReaOMg9WTntNOy0
LdbDBSb6ZzYJcPlzoWafJT53JLCUoZby0jqmI6/znx3u8HZ3x4J309AnT5oz
G/ZhV8a7wiIX7xvOcNHosiuLLdBfsOK1N61H87YET8XvLGv7NREsiNCWiqbg
XEJ90qVwTZExTmKOPqLc8UO0X4Q/G2eG7290lugU3B6WlJFNcB//VFdUwES+
QP+ViSFiFWtRDLN3/EInjQwZr39wLAvuyuLW104whA1c9XAzrNJxIpLfM0Rl
F+wvgHk3FHXhBjik9f3IPMwfKBC7vWOIdFftpwrYp0d8eHiMIfIU3Vg63G7x
KLHsLUNSW/Y3r4cz9fbrfPUM6VISSQCdP2BrV6pjSH2y79mPad3GeHpgFOtv
LFwhhOt/re7naRlSdE0oD4PL68Kr1gwzxGk5VxdD1+/rS9+lYUiv3/cdB2H2
QJKmUI38tWfaymCV8uKDFyqGaMXLuY20HrLm9LO/GCIoU3PbaL9JGS/rMUN8
fpBoGmAiVb/d/xD5BxKNx2m903HFxE3044fok2i9fFHzkUvoN8TJWUpNH/m1
///PPdj/ALrMLZg=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{4.914847988868447, 10.361347240073592`}, {0, 1}], 
         LineBox[{{20.00000000000182, 18.000000000007276`}, {
          12.500000000001819`, 8.000000000003638}}], 
         PolygonBox[{{16.61, 13.48}, {15.57, 12.76}, {16.21, 12.28}}], 
         InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.403022683139532`, 16.54583280199964}, {1, -1}], 
         LineBox[{{19.999999999999545`, 12.}, {9.000000000000341, 12.5}}], 
         PolygonBox[{{13.900618875873821`, 12.27724459655119}, {
          15.117544188493639`, 12.62234281953293}, {15.081218059758719`, 
          11.82316798736469}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.488212026308705`, 10.854216844881332`}, {0, 1}], 
         LineBox[{{19.99999999999818, 8.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{18.249230176602705`, 6.832820117735137}, {
          17.028889744907204`, 6.500000000000001}, {17.472649901887387`, 
          5.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.004985304149102`, 5.815122043776342}, {-1, 1}], 
         LineBox[{{20., 2.}, {15.500000000001819`, 5.}}], 
         PolygonBox[{{17.250769823397295`, 3.8328201177351375`}, {
          18.471110255092796`, 3.5}, {18.027350098112613`, 
          2.834359764529725}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 2.8151220437763422`}, {1, 1}], 
         LineBox[{{9.000000000001819, 12.500000000003638`}, {
          12.500000000001819`, 8.}}], 
         PolygonBox[{{11.118364368108953`, 9.776388669574205}, {
          10.065894744940517`, 10.478035085019828`}, {10.697376518841578`, 
          10.969187575831763`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.131304063066693`, 9.903192049051874}, {1, 1}], LineBox[CompressedData["
1:eJwV1AlMU3ccB/Ang8omR40cBQvtFoxACxYEhgTic2GIkriC3HSjdNUVBGGI
rqBREBI7hXFsMCEM0OpWVu57rIxCgHUExjGFogHZBooHhxxCIs59/02a5pP/
ey/f37e/vPclySGnDCiKCsOX/FIrb/FxpSkWgQVNia87X1XD8hDD0V3wAjNo
M4qYGWDkA9N/jyWawzf3q8q/hil1UvaIC01xZ0S5m3DkmtrrB7ghx27tsiVN
qR5V1GXAgWU7pDZWNCU4FWUfD6cooo53w6H+QY0yuKDzYqzMmqbYXRUKOWzs
IeKZsGjqzu9aXQm81cxR3IZF7xxp08KUQR3XwYamOp4cnV2FI2uL/7kO9xWb
hTshn04YJRqH2SOsuVhY4JlssAXnSNPDiuCVgP7w13A5m9eqgRvmpCvT8Oix
azcewtwrWokSFr7mWLwg/cicGcHwNrtTRixIHb+zSPJphyrI9SyvkN1ZcO7y
r/d/gTNP8mv3wNz9qx4KOLK6/HEt5rvXszEVCHfYVrZGwEOv8tv/Q35dIYfL
IudjBrJamL47ObSOfuS5FhuRsP5ZhvolbDERWmpE+gxVtDNxPe2ZcLmdjz5n
5FPBMPMj6aVU4u94C01wlRU1cwhucE0fPIA8rOErLkx4xXa+rRfW6QSmmzya
8mbm5Ukxn+9Thv0yzOXMBpjaor+GyQ1yXhXyeUoLLHIb3kfu9zZbqQ/ei/58
aoSesCwz7eNpWPhv0SdSeOFWeLaQTVN8zZqkFKaL+F4/wsLf/kocJedtmRsP
YHFmj4DMc3NIE7kIt7Qqiw+S+eU7TWdgrV48H0H6aewZrIcNNdmMZLjqzFHe
F7Dg9Gc1F8h+UNlx78GiGG1GEjx76aepKuQxPma19yTZL/mTfndYL1QfcSR9
Duyr1GEeFt15e4n0dT7yVgJM7dyaVJF+JGVBXNij2s07mswnrn++jH7SzFOU
DNLnuKnvDLxybSCgCf3oC+Nsl2BRXmpMHCwrextN7vf93nfeGpa7UJpzpM9B
y7QpZ+xDjm5rDuYma16o4I6Ee9tnSV7nvhgFLK6pdt+FeULPd92Xw442Ts8a
4ZzxM6czYL3SaDTajvx/5sxceHSRk/wG9s8beaCGZ7OfNxfaYx9DT6xNwPIT
c/FMDvbVqWTiXeQJfPSV+hyH7EG1gib5skr92uFt7/jxC7DAJ21MD68ftMxW
kfleHXYklh8WN44Rf2tmTa6nU923luCUQ+EJ6bBQKVG/gYV/rP7sAOeUfbhj
m5xXuOm6kUeo4rs+hbU8u5oQuGDi+Jf9sGokXzuP/HNUBLMInmVcTcyCZ6fM
/YNJvvysQhd4wabOkkGel+G3+yX68JjuOtBC5lWUSMZgdh//bixcMNjkOw4b
D1NNJrCi2+/TDbKPvd+kdzthXx5HJPngeetnXQcuwoLwh4xKuGFP6J/+cEev
csoe+bjGSgM7mPqgZKkZdnAyWDAi91uEGUaQ/sj705GmTMj7k0v/D9JU6NU=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.508784489160853`, 6.008784489160852}, {1, 1}], 
         {PointSize[0.04], PointBox[{9., 12.5}], PointBox[{12.5, 8.}], 
          PointBox[{15.5, 5.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["2",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        "\"af/bgcfdheh/fggh.m\"",
        TooltipStyle->"TextStyling"],
       Annotation[#, "af/bgcfdheh/fggh.m", "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {22, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1AtUk2UYB/BXhDkgxtgQJimXbeoUKxBFSU7f1wzcoZBFXgbLJC84RXDl
BRCwCYlASoalgOCZctEClcQAU3QSCl7goCGiQnwcCVBRBwKuwXH9v/I7Z+c5
vz3nfd7neb+L15ot4eutCCFf4MdGMmTB5UWT/y53mpzO/4F5ABvuCgXb4Tr9
K945mKbTn3TCS+OkzlmwZ9N983seNBmNT+5Tw1qXkpa18J2fRgSzYOUH16fs
gI1O8/8Y9EQ8tlOkgWeIEq9UwboVQfEUTARiVSJs2MhLs4GH1h2Zvgimd4p6
r2I/Pe9Qwxj+VyZt4O+HAy/qVqbDhrIZe6LhTzgXk0Swtm+NQg37yBeVVyIa
Lk8o1iJqdt2SqhCN2pG4E4iKm087rBE9H/e+GEOUBj8sq55GE6Z1xepo1OHz
hT/Gwoa9X796CDOptd4SNu/0ob8MfTF/jQvvTkWdgCVNavij5rNxOthUcetQ
GtwxrUougVUMbVMAJ3c0Gi6+jXmfnigqgq3v9awOhY2RaYoj8EK/CU1tbjSp
UGSZ9rD17u2tUMP8IJ8UDXsud+I/7ZpCE1FaVK2CdXq3eD2cIJO/mAPXd636
2ChCP1PnXHaDi+OG/dPgimxNpguccTNe7gEr3e5dEbPm9X5X74o5vg84SsGK
ltDnX8HckU5eDJx7sGHWdLjidQnnGDwgc5z70AX3yfkwt4Ndz1jlHYC1gfq+
ybjPUxO+4chhfZfmqAI+lx94amAyTWTaSq8tbH5tVXk2rDxUwE+HxyOvhc+E
DREnZ++DG+Ne+VU7Y74wXVIq3Pql3JWGo1Zu8t8Mt2wera4Xoq/ybcFL2Xzo
kmgFbGrPP/4uLCvbe7BZQJOa86oUIbzPrjMiEm5Jd99sBS971O743AnPweF5
xRb0z31QN5oF67IqNjiw9fMv8HxhVdv+Km94WPObSxcfz8ft00tV7Dyti2ty
WFuumrPhYoHivoJ1wJrpN+D2mpP8cUesj7YvmCjGeUaNPl4Fa62aN8yHTWOB
Db/ycJ7XSgsjYf1Y15RhB8xzVm2nhRvPzK2RwAvLQo9vY9ef8eDI30KfKZaS
jbDSklcfbo86xuU7wmHZs/hstR1NDlzKdJ8HJ2wvdF5ri3rZ/UXOcE29zzwt
F/vkbfUzs+/pxDpO5iTcv5nCkCewSClxOMvB3Kl/DvfDuXm2Oc9sYEfVYhNs
/HswYhGsy2mzuKKepvP9slxr5L23VAXB3BjVKQL3qxi7FDafIRHETkQfAT3K
Gph/bPZQmxX6sI8JH4SZ3cbaBTDjXywTS1CHP0eSNQFzuPrbBcNcm0vXbhD0
Ydg1FAHrFcNbBy0UYZJt7VfBOutUf+/XFNE4tDYqWf8y8HvIOEV0L9edXwgb
wyYnq80U6W8r4bjBeDBMm0wUUU3bVmfG/j5mXUf6KEW0hRmCbrafK+5eF4Yp
4lPwc8Qd2PiOTSz/JUUaBxKSWNP9up6sQYro/bKXPWLXhwnzZhkp0s5sTLRG
fXp3/2dNzyhCur2C/GBPY7tY+xT9+TbMj2H7P2XNET+miLKyL7OU7Tfx8+UN
vRTJ2Flu2wUzZpFncA9FZCOeDE+K9yDXtzqzmyJ0iTnaF9b9s5zYd1KEe5t/
fTFMYqTq2DaK8GesXB/E5ksnfZvTTBFFfXvPAjafFxNiuUSRqDOVxEP65jt+
oPb/KKX/BRvvHSQ=
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.914847988868447, 9.638652759926408}, \
{0, -1}], LineBox[{{19.99999999999818, 18.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{18.249230176602705`, 16.83282011773514}, {
          17.028889744907204`, 16.5}, {17.472649901887387`, 
          15.834359764529724`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 17.18487795622366}, {1, -1}], 
         LineBox[{{20., 12.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{17.250769823397295`, 13.832820117735137`}, {
          18.471110255092796`, 13.5}, {18.027350098112613`, 
          12.834359764529726`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{18.004985304149102`, 14.184877956223659`}, {-1, -1}], 
         LineBox[{{20.00000000000466, 8.}, {9.000000000005457, 7.5}}], 
         PolygonBox[{{15.099381124126179`, 7.77724459655119}, {
          13.882455811506361`, 8.12234281953293}, {13.918781940241281`, 
          7.32316798736469}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {18.48749849569805, 9.147412217828109}, \
{0, -1}], 
         LineBox[{{20.00000000000182, 2.000000000001819}, {
          12.500000000001819`, 12.}}], 
         PolygonBox[{{15.89, 7.48}, {16.93, 6.76}, {16.29, 
          6.279999999999999}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.40465226442821, 3.4544723157306363`}, {1, 1}], 
         LineBox[{{9.000000000001819, 7.5}, {12.499999999998181`, 
          11.999999999996362`}}], 
         PolygonBox[{{10.381635631891047`, 9.276388669574205}, {
          11.434105255059483`, 9.978035085019828}, {10.802623481158422`, 
          10.469187575831763`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{10.131304063066693`, 10.096807950948126`}, {1, -1}], LineBox[CompressedData["

1:eJwt1Hs0m2ccB/CXKurExHGpg6ruqElEXMrQWsWlnVuVCtrKVhwcjENVSy1b
barSlta0qBq1SbtYTKPHpa0g3RQz90tUZxvDijqVcRQdte+zsz+S7/nked/3
/J7v++bdE5l0LFqZoqhgfEhSdPLF4FA0khYcisNQnt0O3zo78hkXdlGvVN4F
GyzWqd+Du3W4vZ6wzIR9QwHTktjtn8JUcAHfBsn98dr5VqSIxv05EmnaQz2g
MfG7aURgNuzSxrp7Cu6/fW+yCPb7rdm6DhY11fkXw+qGNXqqlhxqwj0iLQde
G9PyDIR5Tq6OcbCKUii9ABZutHi7whK5z+mnsJfBw92acKyIxZ6DneL/rpJj
vq9vOyptwOM2L4wqYH5lpHgT5pde+j4R5qXYrb2Co6aGfTzgm67htQMwqyE9
1Qym2+tliWCJysRLHdjAOW7wHKwydaiYmGZUJeCQ+Vz3RZDjpYwi+Q44le7J
PAxLuP7Lcuxv/JnmL2lweF7fczHMOqhh3ghPd2rRc2H7HcleSphfOPhJTAZZ
Z28WH4OTzzaPpMMXz7zU+A42Y7aFCeAFhjjmDekzSboggqVdhV94IZ269FLH
kIrm0qw80nuxy8xOMt/c1p52mJ+XEhYBT9tlDi7CxoL2ww9IX459S+rIfK3k
SlUW7q/++iQd5lbZOp2Ep49aaJB1Y7W1UREs7LV7Ts635zz+9hUc/jTofhe8
FiB2s7DiUNXGiZJS2Mxb3ygI7vaglcXA6WGyjERY4sa6YQ0HxHxcfQ6uuHCw
j+zHWJqlmgT7+fDiu+HZZ+EzoTDnvTuGIri/vrJwH7zhwV/Ph/mZT2y2k+uX
fBN0GU5oGUroJ/MJXbMLyPMlXY4sgdca5qarSZ9TBUejYNP73JZRWLq/OsAB
nm099JE2mc+2Zy8dptbVLEJJn5LRlVXsuzOkt+Muk/SoarII5yqu89ZJz502
mmRdPd350RHSY88FK3J+RUqesNyS9Ej97gy3xf5kOEueK/cofgrp02H+qjnp
zSH+80bkRL6HTiiSJeeWkP2sjRwQpsG3cnVXjpM+r7gzc2DFgHLsD3C61c2S
bDL36+uNb63I/VMsniHrT1bGvNjYV4rNXi6L9LZkL4ATxLU0BixabBp5BPPk
cYzXZD5Zd/mvMH0mJK0VNtj1OHYBVukosroMJ/yzW5eY1j3PD4UV3jlXyfEC
2R1/NvzQ2LJeCsvyetS04eqo88EFsIRXxlWGXfoMpk/B/OjOt9vIc1f4TgiD
rIcODejDwy8+nFjC/AmM+mv7yf9mm1uDDNbdEi0mkvk6ZJ1FsKm5oqmGXK+5
XJBO1v8KnF8nx3v61saS+/+ue78v9ucXfcIkDp44MHSyAu78o7wmg/Rpt3Vk
Gc5cFr9fBg9/cPoSB2ksTszqQ9K+GnC+iDw+OZCghfkCFsS2TbCKtu/qCXhi
lbf5J5x5hfmlGB6Xb+58A//3/rT+//3J5vwLaMzlPQ==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.508784489160853`, 13.991215510839147`}, {1, -1}], 
         {PointSize[0.04], PointBox[{9., 7.5}], PointBox[{15.5, 15.}], 
          PointBox[{12.5, 12.}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["3",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"af/bgcfdheh/fggh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["af/bgcfdheh/fggh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {44, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     GraphicsBox[
      TagBox[
       TooltipBox[
        {Thickness[0.005], LineBox[CompressedData["
1:eJwt1AlQU0cYB/AFaqBFMQjIpSUcFugUTAeLUyy+pxFIUYegqLQUCkgMkIpR
O63lKHgVdLAFaVMRkIigURFw6o1oqlZoFQTPKCBoE4xAlJs3QaX/bftmMju/
+Xa/79vd9+KeuGGF1JwQEocfHQk3iUfAkn8fJ5bwF/1q0sHd63/xyIENJu/F
jdTnLf8agaNF2q2VcN3jtVOjnFnyilXaZ1DfWtWnhDP3710YDueE/yC+AN+V
6kJn0HhD4YrLcN5QdbvaDXHjuZjjMKdM2DYHVk2+cMuF1TG96vJ3WaLYcbgo
Fj534nTzLDjHjicNgluDKnkVs5Hvq2trfeDoxqiIALigLMFSCAsnRl/fnoXx
0PKw1XR+Q9nGrbDkgzhjCazoWjolGNYsEaW/hiURcZ+aw/GBDeWbXVjSpC7P
uOmK/PZ6ZgQ+F7rHTwnz82pbNrvSPj3MP8dI3FP0PRg1h5ak2GEULHduEyFP
02Yhcxnr4sPk+u0w2x4Y9AVcV/tIWknrhDU49qGuZp5EfxiO7lu3LY32vcd9
eiGN83pP9eKchdfHuuVwcpu4JwlujS1XLIYL9hUXPXXEerfzyQJq22BNMnWb
5UY+rMqvC52Yif4iPS64wE4x+shSmG8T1kX722edbbsMLpDIQvLhPFPEb1aw
YlXCtAGYm8tVtDggX2x4iQznyj4c5xXDCvdIlQGODvNMTYQlzcZtMtwLm70m
fQ4s2P3RiJbeW6XsSqc91q9+kx6Ae923+mlIHsya6S4pYPaS/LmnPb2fSemP
cHfp+uKTdug3xUZGbfC55zsPlsxvu/YN3LozZLh2BvLFWFyU0PdkKGjB+7Dw
eFGON6wZ0maobVH3kf8ZHly3N7dyLhx/fddDDv2IY7LXX+Vjvv/LtimIa51b
7klhgdyh80O6Xr23cybMckZTJizQ9lkfnY58l9XGp7Q/9TW/QNgpdLTHT0Dv
oX/RDRv0dyXrpzSYPWNYGQs35ZQdOEG/I4co7ZNpOFdjSrqezg9bOVsCa3uU
O2zdWTLwWWp6xVSc48T9en+4YMX8mQ+s0e+aDq9P4G4zB++Bd2jfVf1BsHD3
7arht/Eepq1j6Hw22Tqr1wr7vFf90AVuXbo/sssS9TI7EnmwpLHappMHO6uO
jKG+SrUg2TgFNhPdGqD97JzIdYS5prIjHO0/wMbuy7dQj/MV8Gl/FuXHrlrg
ffk6gTcfruNaOBbOU1ZOlcOqDX7iW+aI3zntoqb15baFSbBgZ0uHAVYs4GtH
zdB3wB9Z3h74blX1J7bAhpqOTXGwIj+0REewT3lq8C64ID88fB7ss0lvWQWr
mku/v/mGIT4vjFYnYc2ov6vpFUPynP48UEPj1lLTaxNDov0L3Q7SeNRoXC/H
ELVIlbQH7vZ8T3N3jCFNF9zcv6M+erH99ghDxE6dLql0vlh8aXyIIVZni+Kl
1J2ZvpGDyHcqoDiNOntg3Y2XqB8YsiMXZn29fpcZGdLqcOdiLd3P9ZCapD6G
xIsCGnU0HihdVvEc/QlvyDw8ET87qBt/xhDBUi5wLczej8z4uAdx3/7Yg7Dm
557mYB1DtjifTnxA49U1438/YQhfWfyMeGG9YXv73McM0XqpRa4wmZX+7bEH
2E+9umsOdf/kwohWzHfUhHtQ+z2vr9Ig3+Cwhb3X///jBQ3/jV7sP0RmLSA=

          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["\[Gamma]",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {4.194814593688687, 7.202331011588459}, \
{1, 1}], LineBox[{{19.99999999999818, 18.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{18.249230176602705`, 16.83282011773514}, {
          17.028889744907204`, 16.5}, {17.472649901887387`, 
          15.834359764529724`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox["u",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 17.18487795622366}, {1, -1}], 
         LineBox[{{20., 12.}, {15.500000000001819`, 15.}}], 
         PolygonBox[{{17.250769823397295`, 13.832820117735137`}, {
          18.471110255092796`, 13.5}, {18.027350098112613`, 
          12.834359764529726`}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["u", "k"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{17.495014695850898`, 12.815122043776341`}, {1, 1}], 
         LineBox[{{19.99999999999909, 8.000000000003638}, {
          11.499999999997272`, 11.500000000003638`}}], 
         PolygonBox[{{16.30480685908483, 9.521550116847424}, {
          15.34749306301689, 10.348321122542462`}, {15.042893218813452`, 
          9.60857864376269}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {15.159837801855844`, 8.92154894736419}, \
{0, 1}], LineBox[{{20., 2.}, {8.50000000000091, 6.}}], 
         PolygonBox[{{13.683301921976305`, 4.197112374964764}, {
          14.948106328000204`, 4.1806863437177}, {14.685289828047186`, 
          3.425088906352773}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "m"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.740793031341028`, 3.1408299651054534`}, {0, 1}], 
         LineBox[{{8.499999999992724, 5.999999999989086}, {
          11.499999999992724`, 11.499999999985448`}}], 
         PolygonBox[{{10.287311278408344`, 9.27673734374863}, {
          10.063846950757409`, 8.03172180397914}, {9.361530492425903, 
          8.4148035085236}}], InsetBox[
          TagBox[
           StyleBox[
            TagBox[
             SubscriptBox["d", "l"],
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], {9.244061861982704, 8.887420802554889}, \
{1, -1}], LineBox[CompressedData["
1:eJwt1X8003scx/FVQtduTRNWRFnZ/HYwlGqY3LKGHbuWIyFKTTeHVbo3SS1x
0239dJO5uknrRGmVM/fiSkrJjW7d025+RFTKsO5dx+/c18fpj3mex77f8/1+
vu/v58yS+F3CxJkUCkWCDymFSv7Ycym0L/WUrykyQSmLNQvXot+YjA0sQZWK
T64/om5s7e0AVHZW7d6EsgwPC3ei7eGiFQYsLqWM8fl+MVzcE1fEgc1qVqnb
YP1y9uRmOC+oM8cKTZHZCTJRdZJNTByaLjftP426BUxwlGh0seicApUO1H8e
RH1cBRmFqNOc5IsebC5FRy3gnIR5nH4bKfxcHFm9H/abvHnpOsy9dDY+Fq5I
KZV3w/sbxE2rYCWj84yJA5eiOX7tqTmsP3/4jiMsdYvJGphe53axP8zULR29
B1dnfVjHhym/PI4qhHWmb2qn3Zp6Jx2OnVIJA+HYxOiPUXAOddUMd5j3tbCG
B0sZm6wtYJbK/YA3Oe40FDhM1rtD4O8JG6+ex38GH+xmqVYSX7D2KifO7qEL
yHvY7Us/CpfsUa2WwNpA3Vg8bLUo9uIpmCvt0wYQV79YS9arZgjqWeT8kcxX
47CcenehOTnePefJCjJP9918KplHhbf3AeIA20jiaqsZZg1wDs3rJTk/aV5C
0FdomPbxXCe018X2qgA1/pB2PwTtO3PK5QRKTZ2ZtBv1e/DmwyOUsit06grp
vlbHSTR96rLfa3Sk0THHHnOwDDyevZTMO6t2IBgNs0023oZW/FR5bROqG/qu
oZzMc2D7viS0pH+qVE/mbWM9bV0P/YKvI67z7hYnBlZXyYq+h7keG39dDzcX
qf6+A8tlQiNXcp8NZYr38POb4ikq7PnSqdDMCXPs+Bj/Futqnvyr2QumNSZ6
15B1dtCO8eHm4YmC02QOiWk/i+EuA36FBG4fnmUTBfN4otxgWC4qvxEO6/r5
dDY5vvWWHxdO978/az78cCSNZg83/DbWOAsuK2+qM4LFdj17pjDvMM5pm16s
T9l29MFsHG+gdrNq4Tpb+mwLMtcNBUP5cJ7T8wVk34cxPHLTYJYoMnUjmfOR
0FEhnHLoqGsuXLiQvtgH7tOVGt6FZXs7FUw4rKoxbgq2Pe/8ggEXJ66XBGAe
stuezyzJ/QSPbXLgkSu1K5fCSddb4p7AXasrIjjk+uzUuXS0vX3XaAQ60cib
T8o0DinKIOtzLcw4ge73ahbdIM9TmpF/j3xf0vxPHypV7RkcRM2y3+ez8PzS
jKqtNLT3h9d1yWie9j8emRN/dkjULVQetNfLAzVWdYePodExO0I9Uf3blpNr
nHHevHd0RziBRV2QBaecN9luCWtKLMar4Wih28lJsi/Mvi34CFsp8xe3ww9r
+XJrF8zJ0LSX7JPqOQ4ta1zIe2L25MG2mabKSNhyMN1nC5ze0fg0ATbY6WC0
Ej5jd0CyDZ441L/FHE7YVjy+GY4+V8YdJvuTfXtHKDk/uGDjK7g4KH2MA2tm
spmtsNsiha8lbKV5VtcEH2zL0+uxvrxyhq4FVu7T2LXAMjHTpAuObtNrlHDF
ujjKOPldiTVwkMGFJs4Rtri/p1eq9RY4mXZsWSiZb93yc8Fw3Z+/c7Jh7VVz
fw/YIGbTKHkfdW+7ZPZwq6BYYUj2tZGVdBkcsXzCjux7H52m3hlWqitTTsHq
krt/cGG9LDP2BXlvly3CY8h81aIOS1TDv/HpCNog0/4bQa57tfNRJXHcg+5c
sm65q3wINQvhu5HvmU2ZNS54/oh63yoNeS/lljmpME/SWqkjc5A4jKphrSKx
aRKe/v/k+qUu3P8Bx5tWhg==
          "]], InsetBox[
          TagBox[
           StyleBox[
            TagBox["g",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->14.940711462450594`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            14.940711462450594`]& ], \
{13.084117857803797`, 13.811693876795658`}, {1, -1}], 
         {PointSize[0.04], PointBox[{8.5, 6.}], PointBox[{15.5, 15.}], 
          PointBox[{11.5, 11.5}]}, InsetBox[
          TagBox[
           StyleBox[
            TagBox["4",
             DisplayForm],
            FontFamily->"Helvetica",
            FontSize->11.952569169960476`],
           
           StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
            11.952569169960476`]& ], {10., -0.5}, {0, -1}]},
        RowBox[{"VFlip", "(", "\"af/bfcgdheh/fggh.m\"", ")"}]],
       Annotation[#, 
        FeynArts`VFlip["af/bfcgdheh/fggh.m"], "Tooltip"]& ],
      AspectRatio->1,
      PlotRange->{{-1, 21}, {-1, 21}}], {66, 22}, {0, 0}, {22, 22}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[Gamma]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {54.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["\[RightArrow]",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {59.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox["u",
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {63.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["u", "k"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {68.25, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["d", "l"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {72.75, 47.959999999999994`}, {0, 0}], 
    InsetBox[
     TagBox[
      StyleBox[
       TagBox[
        SubscriptBox["d", "m"],
        DisplayForm],
       FontFamily->"Helvetica",
       FontSize->17.92885375494071],
      StyleForm[#, FontFamily -> "Helvetica", FontSize -> 
       17.92885375494071]& ], {77.25, 47.959999999999994`}, {0, 0}]},
   AspectRatio->0.3833333333333333,
   ImageSize->{800, 300},
   PlotRange->{{0, 132}, {0, 50.599999999999994`}}], 
  TraditionalForm]], "Print"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"RemoveMasslessOnShellPropagators", "[", 
   RowBox[{"ex_", ",", "momenta_List"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"ex", "//", "FeynAmpDenominatorSplit"}], "//", 
    RowBox[{"ReplaceAll", "[", 
     RowBox[{
      RowBox[{"FeynAmpDenominator", "[", 
       RowBox[{"PropagatorDenominator", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"Momentum", "[", 
           RowBox[{
            RowBox[{"Alternatives", "@@", "momenta"}], ",", "_"}], "]"}], "|", 
          RowBox[{"-", 
           RowBox[{"Momentum", "[", 
            RowBox[{
             RowBox[{"Alternatives", "@@", "momenta"}], ",", "_"}], "]"}]}]}],
          ",", "0"}], "]"}], "]"}], "\[Rule]", "0"}], "]"}]}], "//", 
   "FeynAmpDenominatorCombine"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"W", "=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
     "\"\<uU\>\"", "+", "\"\<uUg\>\"", "+", "\"\<uUgg\>\"", "-", 
      "\"\<uUhH\>\"", "+", "\"\<uUuU\>\"", "+", "\"\<uUdD\>\""}], ")"}], "//", 
    RowBox[{"ReplaceAll", "[", "\[IndentingNewLine]", 
     RowBox[{"particles_String", "\[RuleDelayed]", 
      RowBox[{"Sum", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"amp", "[", 
          RowBox[{"particles", ",", "i"}], "]"}], "*", 
         RowBox[{"camp", "[", 
          RowBox[{"particles", ",", "j"}], "]"}], "*", 
         RowBox[{"P", "[", "particles", "]"}]}], ",", 
        RowBox[{"{", 
         RowBox[{"i", ",", "0", ",", "4"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"j", ",", "0", ",", 
          RowBox[{"4", "-", "i"}]}], "}"}]}], "]"}]}], "\[IndentingNewLine]", 
     "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"W", "=", 
   RowBox[{
    RowBox[{"W", "//", "Expand"}], "//", 
    RowBox[{
     RowBox[{"RemoveMasslessOnShellPropagators", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4"}], "}"}]}], "]"}], 
     "&"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"W", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["502", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["\<\"e\"\>", "2"], " ", 
     SubsuperscriptBox["g", "s", "4"], " ", 
     RowBox[{"P", "(", "\<\"uUdD\"\>", ")"}], " ", 
     SubsuperscriptBox["Q", "u", "2"], " ", 
     RowBox[{"A", "(", 
      RowBox[{"\<\"uUdD\"\>", ",", "2", ",", "2"}], ")"}], " ", 
     RowBox[{"B", "(", 
      RowBox[{"\<\"uUdD\"\>", ",", "2", ",", "2"}], ")"}], " ", 
     FormBox[
      SuperscriptBox["g", 
       RowBox[{
        FormBox[
         FormBox[
          FormBox["Lor2",
           TraditionalForm],
          TraditionalForm],
         TraditionalForm], 
        FormBox[
         FormBox[
          FormBox["Lor3",
           TraditionalForm],
          TraditionalForm],
         TraditionalForm]}]],
      TraditionalForm], " ", 
     SubsuperscriptBox["T", 
      RowBox[{
       FormBox[
        FormBox["Col2",
         TraditionalForm],
        TraditionalForm], 
       FormBox[
        FormBox["Col3",
         TraditionalForm],
        TraditionalForm]}], 
      FormBox[
       FormBox["Glu6",
        TraditionalForm],
       TraditionalForm]], " ", 
     SubsuperscriptBox["T", 
      RowBox[{
       FormBox[
        FormBox["Col4",
         TraditionalForm],
        TraditionalForm], 
       FormBox[
        FormBox["Col5",
         TraditionalForm],
        TraditionalForm]}], 
      FormBox[
       FormBox["Glu6",
        TraditionalForm],
       TraditionalForm]], " ", 
     RowBox[{"IndexDelta", "(", 
      RowBox[{"1", ",", 
       FormBox[
        InterpretationBox[
         RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
         SequenceForm["Gen", 3],
         Editable->False],
        TraditionalForm]}], ")"}], " ", 
     RowBox[{"IndexDelta", "(", 
      RowBox[{
       FormBox[
        InterpretationBox[
         RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
         SequenceForm["Gen", 4],
         Editable->False],
        TraditionalForm], ",", 
       FormBox[
        InterpretationBox[
         RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
         SequenceForm["Gen", 5],
         Editable->False],
        TraditionalForm]}], ")"}], " ", 
     FormBox[
      SuperscriptBox["g", 
       RowBox[{
        FormBox[
         FormBox[
          FormBox[
           RowBox[{"$AL", "\[InvisibleApplication]", 
            RowBox[{"(", "$836", ")"}]}],
           TraditionalForm],
          TraditionalForm],
         TraditionalForm], 
        FormBox[
         FormBox[
          FormBox[
           RowBox[{"$AL", "\[InvisibleApplication]", 
            RowBox[{"(", "$837", ")"}]}],
           TraditionalForm],
          TraditionalForm],
         TraditionalForm]}]],
      TraditionalForm], " ", 
     SuperscriptBox[
      RowBox[{"SumOver", "(", 
       RowBox[{
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
          SequenceForm["Gen", 3],
          Editable->False],
         TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"SumOver", "(", 
       RowBox[{
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
          SequenceForm["Gen", 4],
          Editable->False],
         TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"SumOver", "(", 
       RowBox[{
        FormBox[
         InterpretationBox[
          RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
          SequenceForm["Gen", 5],
          Editable->False],
         TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
     SubsuperscriptBox["T", 
      RowBox[{
       FormBox[
        FormBox["Col3",
         TraditionalForm],
        TraditionalForm], 
       FormBox[
        FormBox["Col2",
         TraditionalForm],
        TraditionalForm]}], 
      FormBox[
       FormBox[
        RowBox[{"$AL", "\[InvisibleApplication]", 
         RowBox[{"(", "$835", ")"}]}],
        TraditionalForm],
       TraditionalForm]], " ", 
     SubsuperscriptBox["T", 
      RowBox[{
       FormBox[
        FormBox["Col5",
         TraditionalForm],
        TraditionalForm], 
       FormBox[
        FormBox["Col4",
         TraditionalForm],
        TraditionalForm]}], 
      FormBox[
       FormBox[
        RowBox[{"$AL", "\[InvisibleApplication]", 
         RowBox[{"(", "$835", ")"}]}],
        TraditionalForm],
       TraditionalForm]], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         FormBox["k3",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}], ".", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox["Lor3",
         TraditionalForm],
        TraditionalForm]], ".", 
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{"-", 
          FormBox["k4",
           TraditionalForm]}],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}]}], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{"-", 
          FormBox["k4",
           TraditionalForm]}],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}], ".", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox[
         RowBox[{"$AL", "\[InvisibleApplication]", 
          RowBox[{"(", "$837", ")"}]}],
         TraditionalForm],
        TraditionalForm]], ".", 
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         FormBox["k3",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}]}], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         FormBox["k1",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}], ".", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox["Lor2",
         TraditionalForm],
        TraditionalForm]], ".", 
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
        FormBox[
         FormBox[
          RowBox[{
           FormBox["k1",
            TraditionalForm], "+", 
           FormBox["k3",
            TraditionalForm], "+", 
           FormBox["k4",
            TraditionalForm]}],
          TraditionalForm],
         TraditionalForm], ")"}], ")"}], ".", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox["mu",
         TraditionalForm],
        TraditionalForm]], ".", 
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{"-", 
          FormBox["k2",
           TraditionalForm]}],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}]}], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         RowBox[{"-", 
          FormBox["k2",
           TraditionalForm]}],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}], ".", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox["nu",
         TraditionalForm],
        TraditionalForm]], ".", 
      RowBox[{"(", 
       RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
        FormBox[
         FormBox[
          RowBox[{
           FormBox["k1",
            TraditionalForm], "+", 
           FormBox["k3",
            TraditionalForm], "+", 
           FormBox["k4",
            TraditionalForm]}],
          TraditionalForm],
         TraditionalForm], ")"}], ")"}], ".", 
      SuperscriptBox["\[Gamma]", 
       FormBox[
        FormBox[
         RowBox[{"$AL", "\[InvisibleApplication]", 
          RowBox[{"(", "$836", ")"}]}],
         TraditionalForm],
        TraditionalForm]], ".", 
      RowBox[{"(", 
       RowBox[{
        FormBox["\<\"\[CurlyPhi]\"\>",
         TraditionalForm], 
        FormBox["\<\"(\"\>",
         TraditionalForm], 
        FormBox[
         FormBox["k1",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\")\"\>",
         TraditionalForm]}], ")"}]}]}], ")"}], "/", 
   RowBox[{
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       FormBox["k3",
        TraditionalForm], "+", 
       FormBox["k4",
        TraditionalForm]}], ")"}], "2"], ".", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       FormBox["k3",
        TraditionalForm], "+", 
       FormBox["k4",
        TraditionalForm]}], ")"}], "2"], ".", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FormBox["k1",
         TraditionalForm]}], "-", 
       FormBox["k3",
        TraditionalForm], "-", 
       FormBox["k4",
        TraditionalForm]}], ")"}], "2"], ".", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        FormBox["k1",
         TraditionalForm]}], "-", 
       FormBox["k3",
        TraditionalForm], "-", 
       FormBox["k4",
        TraditionalForm]}], ")"}], "2"]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "To", " ", "change", " ", "propagators", " ", "of", " ", "the", " ", 
    "form", " ", 
    RowBox[{"1", "/", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "k1"}], "-", "k2"}], ")"}], "^", " ", "into"}]}], " ", 
    RowBox[{"1", "/", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"k1", "+", "k2"}], ")"}], "^", "2"}]}]}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"W", "=", 
    RowBox[{"W", "//", "FeynAmpDenominatorSimplify"}]}], ";"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"W", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"2", " ", 
       SuperscriptBox["\<\"e\"\>", "2"], " ", 
       SubsuperscriptBox["g", "s", "4"], " ", 
       RowBox[{"P", "(", "\<\"uUgg\"\>", ")"}], " ", 
       SubsuperscriptBox["Q", "u", "2"], " ", 
       RowBox[{"A", "(", 
        RowBox[{"\<\"uUgg\"\>", ",", "2", ",", "4"}], ")"}], " ", 
       RowBox[{"B", "(", 
        RowBox[{"\<\"uUgg\"\>", ",", "2", ",", "3"}], ")"}], " ", 
       FormBox[
        RowBox[{
         SuperscriptBox[
          SuperscriptBox[
           FormBox["\<\"\[CurlyEpsilon]\"\>",
            TraditionalForm], "*"], 
          FormBox[
           FormBox[
            FormBox["Lor2",
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]], "(", 
         FormBox["k3",
          TraditionalForm], ")"}],
        TraditionalForm], " ", 
       FormBox[
        RowBox[{
         SuperscriptBox[
          SuperscriptBox[
           FormBox["\<\"\[CurlyEpsilon]\"\>",
            TraditionalForm], "*"], 
          FormBox[
           FormBox[
            FormBox["Lor3",
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]], "(", 
         FormBox["k4",
          TraditionalForm], ")"}],
        TraditionalForm], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       FormBox[
        SuperscriptBox["g", 
         RowBox[{
          FormBox[
           FormBox[
            FormBox[
             RowBox[{"$AL", "\[InvisibleApplication]", 
              RowBox[{"(", "$871", ")"}]}],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm], 
          FormBox[
           FormBox[
            FormBox[
             RowBox[{"$AL", "\[InvisibleApplication]", 
              RowBox[{"(", "$872", ")"}]}],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]}]],
        TraditionalForm], " ", 
       FormBox[
        SuperscriptBox["g", 
         RowBox[{
          FormBox[
           FormBox[
            FormBox[
             RowBox[{"$AL", "\[InvisibleApplication]", 
              RowBox[{"(", "$873", ")"}]}],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm], 
          FormBox[
           FormBox[
            FormBox[
             RowBox[{"$AL", "\[InvisibleApplication]", 
              RowBox[{"(", "$874", ")"}]}],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]}]],
        TraditionalForm], " ", 
       FormBox[
        RowBox[{
         SuperscriptBox[
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], 
          FormBox[
           FormBox[
            FormBox[
             RowBox[{"$AL", "\[InvisibleApplication]", 
              RowBox[{"(", "$871", ")"}]}],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]], "(", 
         FormBox["k3",
          TraditionalForm], ")"}],
        TraditionalForm], " ", 
       FormBox[
        SuperscriptBox[
         RowBox[{"(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k3",
              TraditionalForm], "-", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], 
         FormBox[
          FormBox[
           FormBox[
            RowBox[{"$AL", "\[InvisibleApplication]", 
             RowBox[{"(", "$874", ")"}]}],
            TraditionalForm],
           TraditionalForm],
          TraditionalForm]],
        TraditionalForm], " ", 
       FormBox[
        RowBox[{
         SuperscriptBox[
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], 
          FormBox[
           FormBox[
            FormBox[
             RowBox[{"$AL", "\[InvisibleApplication]", 
              RowBox[{"(", "$872", ")"}]}],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]], "(", 
         FormBox["k4",
          TraditionalForm], ")"}],
        TraditionalForm], " ", 
       SubscriptBox[
        RowBox[{"(", 
         SuperscriptBox["T", 
          FormBox[
           FormBox["Glu4",
            TraditionalForm],
           TraditionalForm]], 
         SuperscriptBox["T", 
          FormBox[
           FormBox["Glu5",
            TraditionalForm],
           TraditionalForm]], ")"}], 
        RowBox[{
         FormBox[
          FormBox["Col2",
           TraditionalForm],
          TraditionalForm], 
         FormBox[
          FormBox["Col3",
           TraditionalForm],
          TraditionalForm]}]], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SubsuperscriptBox["T", 
        RowBox[{
         FormBox[
          FormBox["Col3",
           TraditionalForm],
          TraditionalForm], 
         FormBox[
          FormBox["Col2",
           TraditionalForm],
          TraditionalForm]}], 
        FormBox[
         FormBox[
          RowBox[{"$AL", "\[InvisibleApplication]", 
           RowBox[{"(", "$870", ")"}]}],
          TraditionalForm],
         TraditionalForm]], " ", 
       RowBox[{"(", 
        RowBox[{
         FormBox["\<\"tr\"\>",
          TraditionalForm], 
         FormBox["\<\"(\"\>",
          TraditionalForm], 
         FormBox[
          RowBox[{
           SuperscriptBox["T", 
            FormBox[
             FormBox[
              RowBox[{"$AL", "\[InvisibleApplication]", 
               RowBox[{"(", "$870", ")"}]}],
              TraditionalForm],
             TraditionalForm]], ".", 
           SuperscriptBox["T", 
            FormBox[
             FormBox["Glu5",
              TraditionalForm],
             TraditionalForm]], ".", 
           SuperscriptBox["T", 
            FormBox[
             FormBox["Glu4",
              TraditionalForm],
             TraditionalForm]]}],
          TraditionalForm], 
         FormBox["\<\")\"\>",
          TraditionalForm]}], ")"}], " ", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox["Lor2",
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k1",
              TraditionalForm], "+", 
             FormBox["k3",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox["Lor3",
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k1",
              TraditionalForm], "+", 
             FormBox["k3",
              TraditionalForm], "+", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox["mu",
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           RowBox[{"-", 
            FormBox["k2",
             TraditionalForm]}],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}]}], " ", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           RowBox[{"-", 
            FormBox["k2",
             TraditionalForm]}],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox[
           RowBox[{"$AL", "\[InvisibleApplication]", 
            RowBox[{"(", "$873", ")"}]}],
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             RowBox[{"-", 
              FormBox["k2",
               TraditionalForm]}], "-", 
             FormBox["k3",
              TraditionalForm], "-", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox["nu",
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}]}]}], ")"}], "/", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k1",
          TraditionalForm], "+", 
         FormBox["k3",
          TraditionalForm]}], ")"}], "2"], ".", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"], ".", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k1",
          TraditionalForm], "+", 
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"], ".", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k2",
          TraditionalForm], "+", 
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"]}]}], ")"}]}], 
  TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Tensor reduction", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"TransverseProjector", "[", 
   RowBox[{
   "mu_Symbol", ",", "nu_Symbol", ",", "q_", ",", "k_", ",", "d_Symbol"}], 
   "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"kq", ",", "q2", ",", "kmu", ",", "knu"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"kq", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"Momentum", "[", 
         RowBox[{"k", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"q", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"q2", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"Momentum", "[", 
         RowBox[{"q", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"q", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"kmu", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"LorentzIndex", "[", 
         RowBox[{"mu", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"k", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"knu", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"LorentzIndex", "[", 
         RowBox[{"nu", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"k", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], 
       RowBox[{
        RowBox[{"kq", "/", "q2"}], "/", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}]}], " ", 
       RowBox[{"MetricTensor", "[", 
        RowBox[{"mu", ",", "nu", ",", 
         RowBox[{"Dimension", "\[Rule]", "d"}]}], "]"}]}], "-", 
      RowBox[{
       RowBox[{
        RowBox[{"2", "/", "kq"}], " ", "/", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}]}], "kmu", " ", "knu"}]}]}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"LongitudinalProjector", "[", 
   RowBox[{
   "mu_Symbol", ",", "nu_Symbol", ",", "q_", ",", "k_", ",", "d_Symbol"}], 
   "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"kq", ",", "kmu", ",", "knu"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"kq", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"Momentum", "[", 
         RowBox[{"k", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"q", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"kmu", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"LorentzIndex", "[", 
         RowBox[{"mu", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"k", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"knu", "=", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"LorentzIndex", "[", 
         RowBox[{"nu", ",", "d"}], "]"}], ",", 
        RowBox[{"Momentum", "[", 
         RowBox[{"k", ",", "d"}], "]"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"1", "/", "kq"}], " ", "kmu", " ", "knu"}]}]}], 
   "\[IndentingNewLine]", "]"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{"Project", " ", "from", " ", 
    RowBox[{"W", "^", 
     RowBox[{"{", 
      RowBox[{"mu", ",", "nu"}], "}"}]}], " ", "into", " ", "the", " ", 
    "scalar", " ", "F_T"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"FF", "=", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"W", "*", 
           RowBox[{"TransverseProjector", "[", 
            RowBox[{"mu", ",", "nu", ",", "q", ",", "k1", ",", "d"}], "]"}]}],
           "//", 
          RowBox[{"ReplaceAll", "[", 
           RowBox[{
            RowBox[{"_A", "|", "_B"}], "\[Rule]", "1"}], "]"}]}], "//", 
         "Expand"}], "//", "TM"}], "//", 
       RowBox[{"MapWithProgress", "[", "Contract", "]"}]}], "//", "TM"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{"FF", "//", "Length"}]}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"Expand", "[", 
      RowBox[{
       RowBox[{"ReplaceAll", "[", 
        RowBox[{
         RowBox[{"_A", "|", "_B"}], "\[Rule]", "1"}], "]"}], "[", 
       RowBox[{"W", " ", 
        RowBox[{"TransverseProjector", "(", 
         RowBox[{"mu", ",", "nu", ",", "q", ",", "k1", ",", "d"}], ")"}]}], 
       "]"}], "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.122046`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     Expand[
      ReplaceAll[Alternatives[
         Blank[$CellContext`A], 
         Blank[$CellContext`B]] -> 
       1][$CellContext`W $CellContext`TransverseProjector[$CellContext`mu, \
$CellContext`nu, $CellContext`q, $CellContext`k1, $CellContext`d]]]], ": ", 
    0.122046, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"MapWithProgress", "(", "Contract", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", 
       RowBox[{"TM", "(", 
        RowBox[{"Expand", "[", 
         RowBox[{
          RowBox[{"ReplaceAll", "[", 
           RowBox[{
            RowBox[{"_A", "|", "_B"}], "\[Rule]", "1"}], "]"}], "[", 
          RowBox[{"W", " ", 
           RowBox[{"TransverseProjector", "(", 
            RowBox[{"mu", ",", "nu", ",", "q", ",", "k1", ",", "d"}], ")"}]}],
           "]"}], "]"}], ")"}], ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "15.690373`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`MapWithProgress[FeynCalc`Contract][
      $CellContext`TM[
       Expand[
        ReplaceAll[Alternatives[
           Blank[$CellContext`A], 
           Blank[$CellContext`B]] -> 
         1][$CellContext`W $CellContext`TransverseProjector[$CellContext`mu, \
$CellContext`nu, $CellContext`q, $CellContext`k1, $CellContext`d]]]]]], ": ", 
    15.690373, " sec"],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["1188", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FF", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"2", " ", 
       SuperscriptBox["\<\"e\"\>", "2"], " ", "N", " ", 
       SubsuperscriptBox["g", "s", "4"], " ", 
       RowBox[{"P", "(", "\<\"uUgg\"\>", ")"}], " ", 
       SubsuperscriptBox["Q", "u", "2"], " ", 
       SubsuperscriptBox["T", 
        RowBox[{
         FormBox[
          FormBox["Col2",
           TraditionalForm],
          TraditionalForm], 
         FormBox[
          FormBox["Col3",
           TraditionalForm],
          TraditionalForm]}], 
        FormBox[
         FormBox["Glu6",
          TraditionalForm],
         TraditionalForm]], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FormBox[
           FormBox["k3",
            TraditionalForm],
           TraditionalForm], 
          FormBox["\<\"\[CenterDot]\"\>",
           TraditionalForm], 
          FormBox[
           RowBox[{
            SuperscriptBox[
             FormBox["\<\"\[CurlyEpsilon]\"\>",
              TraditionalForm], "*"], "(", 
            FormBox["k3",
             TraditionalForm], ")"}],
           TraditionalForm]}], "+", 
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            FormBox[
             FormBox["k4",
              TraditionalForm],
             TraditionalForm], 
            FormBox["\<\"\[CenterDot]\"\>",
             TraditionalForm], 
            FormBox[
             RowBox[{
              SuperscriptBox[
               FormBox["\<\"\[CurlyEpsilon]\"\>",
                TraditionalForm], "*"], "(", 
              FormBox["k3",
               TraditionalForm], ")"}],
             TraditionalForm]}], ")"}]}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          FormBox[
           FormBox["k3",
            TraditionalForm],
           TraditionalForm], 
          FormBox["\<\"\[CenterDot]\"\>",
           TraditionalForm], 
          FormBox[
           RowBox[{
            FormBox["\<\"\[CurlyEpsilon]\"\>",
             TraditionalForm], "(", 
            FormBox["k3",
             TraditionalForm], ")"}],
           TraditionalForm]}], "+", 
         RowBox[{"2", " ", 
          RowBox[{"(", 
           RowBox[{
            FormBox[
             FormBox["k4",
              TraditionalForm],
             TraditionalForm], 
            FormBox["\<\"\[CenterDot]\"\>",
             TraditionalForm], 
            FormBox[
             RowBox[{
              FormBox["\<\"\[CurlyEpsilon]\"\>",
               TraditionalForm], "(", 
              FormBox["k3",
               TraditionalForm], ")"}],
             TraditionalForm]}], ")"}]}]}], ")"}], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       SuperscriptBox["\[Delta]", 
        RowBox[{
         FormBox[
          FormBox["Glu6",
           TraditionalForm],
          TraditionalForm], 
         FormBox[
          FormBox[
           RowBox[{"$AL", "\[InvisibleApplication]", 
            RowBox[{"(", "$870", ")"}]}],
           TraditionalForm],
          TraditionalForm]}]], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SubsuperscriptBox["T", 
        RowBox[{
         FormBox[
          FormBox["Col3",
           TraditionalForm],
          TraditionalForm], 
         FormBox[
          FormBox["Col2",
           TraditionalForm],
          TraditionalForm]}], 
        FormBox[
         FormBox[
          RowBox[{"$AL", "\[InvisibleApplication]", 
           RowBox[{"(", "$870", ")"}]}],
          TraditionalForm],
         TraditionalForm]], " ", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           RowBox[{
            SuperscriptBox[
             FormBox["\<\"\[CurlyEpsilon]\"\>",
              TraditionalForm], "*"], "(", 
            FormBox["k4",
             TraditionalForm], ")"}],
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k1",
              TraditionalForm], "+", 
             FormBox["k3",
              TraditionalForm], "+", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           RowBox[{"-", 
            FormBox["k2",
             TraditionalForm]}],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}]}], " ", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           RowBox[{"-", 
            FormBox["k2",
             TraditionalForm]}],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k1",
              TraditionalForm], "+", 
             FormBox["k3",
              TraditionalForm], "+", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           RowBox[{
            FormBox["\<\"\[CurlyEpsilon]\"\>",
             TraditionalForm], "(", 
            FormBox["k4",
             TraditionalForm], ")"}],
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{
          FormBox["\<\"\[CurlyPhi]\"\>",
           TraditionalForm], 
          FormBox["\<\"(\"\>",
           TraditionalForm], 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm], 
          FormBox["\<\")\"\>",
           TraditionalForm]}], ")"}]}]}], ")"}], "/", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"d", "-", "2"}], ")"}], " ", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          FormBox[
           RowBox[{
            FormBox["k3",
             TraditionalForm], "+", 
            FormBox["k4",
             TraditionalForm]}],
           TraditionalForm], ")"}], "2"], ".", 
        SuperscriptBox[
         RowBox[{"(", 
          FormBox[
           RowBox[{
            FormBox["k3",
             TraditionalForm], "+", 
            FormBox["k4",
             TraditionalForm]}],
           TraditionalForm], ")"}], "2"], ".", 
        SuperscriptBox[
         RowBox[{"(", 
          FormBox[
           RowBox[{
            FormBox["k1",
             TraditionalForm], "+", 
            FormBox["k3",
             TraditionalForm], "+", 
            FormBox["k4",
             TraditionalForm]}],
           TraditionalForm], ")"}], "2"], ".", 
        SuperscriptBox[
         RowBox[{"(", 
          FormBox[
           RowBox[{
            FormBox["k1",
             TraditionalForm], "+", 
            FormBox["k3",
             TraditionalForm], "+", 
            FormBox["k4",
             TraditionalForm]}],
           TraditionalForm], ")"}], "2"]}], " ", 
       RowBox[{"(", 
        RowBox[{
         FormBox[
          FormBox["k1",
           TraditionalForm],
          TraditionalForm], 
         FormBox["\<\"\[CenterDot]\"\>",
          TraditionalForm], 
         FormBox[
          FormBox["q",
           TraditionalForm],
          TraditionalForm]}], ")"}]}], ")"}]}], ")"}]}], 
  TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Quark generation factors", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SumOverGenerationIndices", "[", 
   RowBox[{"ex_", ",", 
    RowBox[{"Ngenerations_Symbol:", "Ngen"}]}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"e", ",", "sumoverindices"}], "}"}], ",", "\[IndentingNewLine]", 
    
    RowBox[{
     RowBox[{"sumoverindices", "=", 
      RowBox[{
       RowBox[{"{", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
          RowBox[{"Cases", "[", 
           RowBox[{"ex", ",", 
            RowBox[{
             RowBox[{"SumOver", "[", 
              RowBox[{
               RowBox[{"idx", ":", 
                RowBox[{"Index", "[", 
                 RowBox[{"Generation", ",", "_"}], "]"}]}], ",", "_", ",", 
               "External"}], "]"}], "\[RuleDelayed]", "idx"}], ",", 
            RowBox[{"{", 
             RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "//", "Union"}], 
         ",", "\[IndentingNewLine]", 
         RowBox[{"Cases", "[", 
          RowBox[{"ex", ",", 
           RowBox[{
            RowBox[{"SumOver", "[", 
             RowBox[{
              RowBox[{"idx", ":", 
               RowBox[{"Index", "[", 
                RowBox[{"Generation", ",", "_"}], "]"}]}], ",", "_"}], "]"}], 
            "\[RuleDelayed]", "idx"}], ",", 
           RowBox[{"{", 
            RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], ",", 
         "\[IndentingNewLine]", 
         RowBox[{"Cases", "[", 
          RowBox[{"ex", ",", 
           RowBox[{
            RowBox[{
             RowBox[{"SumOver", "[", 
              RowBox[{
               RowBox[{"idx", ":", 
                RowBox[{"Index", "[", 
                 RowBox[{"Generation", ",", "_"}], "]"}]}], ",", "_"}], "]"}],
              "^", "n_"}], "\[RuleDelayed]", 
            RowBox[{"Table", "[", 
             RowBox[{"idx", ",", 
              RowBox[{"n", "-", "1"}]}], "]"}]}], ",", 
           RowBox[{"{", 
            RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}]}], 
        "\[IndentingNewLine]", "}"}], "//", "Flatten"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Length", "[", "sumoverindices", "]"}], "===", "0"}], ",", 
       RowBox[{"Return", "[", "ex", "]"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"e", "=", 
      RowBox[{"ex", "/.", 
       RowBox[{
        RowBox[{"SumOver", "[", 
         RowBox[{
          RowBox[{"idx", ":", 
           RowBox[{"Index", "[", 
            RowBox[{"Generation", ",", "_"}], "]"}]}], ",", "__"}], "]"}], 
        "\[Rule]", "1"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"e", "=", 
      RowBox[{"e", "/.", 
       RowBox[{
        RowBox[{"IndexDelta", "[", 
         RowBox[{
          RowBox[{"i1", ":", 
           RowBox[{"Index", "[", 
            RowBox[{"Generation", ",", "_"}], "]"}]}], ",", 
          RowBox[{"i2", ":", 
           RowBox[{"Index", "[", 
            RowBox[{"Generation", ",", "_"}], "]"}]}]}], "]"}], 
        "\[RuleDelayed]", 
        RowBox[{"DiscreteDelta", "[", 
         RowBox[{"i1", "-", "i2"}], "]"}]}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"e", "=", 
      RowBox[{"e", "/.", 
       RowBox[{
        RowBox[{"IndexDelta", "[", 
         RowBox[{"i1_Integer", ",", 
          RowBox[{"i2", ":", 
           RowBox[{"Index", "[", 
            RowBox[{"Generation", ",", "_"}], "]"}]}]}], "]"}], 
        "\[RuleDelayed]", 
        RowBox[{"DiscreteDelta", "[", 
         RowBox[{"i1", "-", "i2"}], "]"}]}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Assuming", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"Ngenerations", ">", "999"}], "}"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"e", "//", 
         RowBox[{
          RowBox[{"Join", "[", 
           RowBox[{
            RowBox[{"{", "#", "}"}], ",", 
            RowBox[{"Map", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"{", 
                RowBox[{"#", ",", "1", ",", "Ngenerations"}], "}"}], "&"}], 
              ",", "sumoverindices"}], "]"}]}], "]"}], "&"}]}], "//", 
        RowBox[{"Apply", "[", "Sum", "]"}]}]}], "\[IndentingNewLine]", 
      "]"}]}]}], "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CalcGenerationFactor", "[", "ex_", "]"}], ":=", 
  RowBox[{"Bracket", "[", 
   RowBox[{"ex", ",", 
    RowBox[{"x_", "/;", 
     RowBox[{"Not", "[", 
      RowBox[{"FreeQ", "[", 
       RowBox[{"x", ",", "Generation"}], "]"}], "]"}]}], ",", 
    RowBox[{"#", "&"}], ",", "SumOverGenerationIndices"}], "]"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"FF", "//", 
     RowBox[{
      RowBox[{"Bracket", "[", 
       RowBox[{"#", ",", 
        RowBox[{"x_", "/;", 
         RowBox[{"Not", "[", 
          RowBox[{"FreeQ", "[", 
           RowBox[{"x", ",", "Generation"}], "]"}], "]"}]}], ",", 
        RowBox[{"1", "&"}]}], "]"}], "&"}]}], "//", 
    RowBox[{"Apply", "[", "List", "]"}]}], "//", 
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"#", ",", 
       RowBox[{"CalcGenerationFactor", "[", "#", "]"}]}], "}"}], "&"}], 
    "]"}]}], "//", "TableForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"]}], 
      "1"},
     {
      RowBox[{
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       RowBox[{"SumOver", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
           SequenceForm["Gen", 4],
           Editable->False],
          TraditionalForm], ",", "3"}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"]}], 
      "Ngen"},
     {
      RowBox[{
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
           SequenceForm["Gen", 5],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm], ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
           SequenceForm["Gen", 4],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
            SequenceForm["Gen", 4],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
            SequenceForm["Gen", 5],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"]}], 
      "Ngen"},
     {
      RowBox[{
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
           SequenceForm["Gen", 4],
           Editable->False],
          TraditionalForm], ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
           SequenceForm["Gen", 5],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
            SequenceForm["Gen", 4],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
            SequenceForm["Gen", 5],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"]}], 
      "Ngen"},
     {
      RowBox[{
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{"1", ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
           SequenceForm["Gen", 5],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
           SequenceForm["Gen", 3],
           Editable->False],
          TraditionalForm], ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
           SequenceForm["Gen", 4],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       RowBox[{"IndexDelta", "(", 
        RowBox[{
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
           SequenceForm["Gen", 4],
           Editable->False],
          TraditionalForm], ",", 
         FormBox[
          InterpretationBox[
           RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
           SequenceForm["Gen", 5],
           Editable->False],
          TraditionalForm]}], ")"}], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "3"}],
            SequenceForm["Gen", 3],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "4"}],
            SequenceForm["Gen", 4],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"], " ", 
       SuperscriptBox[
        RowBox[{"SumOver", "(", 
         RowBox[{
          FormBox[
           InterpretationBox[
            RowBox[{"\<\"Gen\"\>", "\[InvisibleSpace]", "5"}],
            SequenceForm["Gen", 5],
            Editable->False],
           TraditionalForm], ",", "3", ",", "External"}], ")"}], "2"]}], "1"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"FFg", "=", 
   RowBox[{
    RowBox[{"FF", "//", "CalcGenerationFactor"}], "//", "Expand"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFg", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["1348", TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Color factors", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"FactorRelatedToColorQ", "[", "ex_", "]"}], ":=", 
  RowBox[{"Not", "[", 
   RowBox[{"FreeQ", "[", 
    RowBox[{"ex", ",", 
     RowBox[{
     "ExplicitSUNFIndex", "|", "ExplicitSUNIndex", "|", "SUND", "|", 
      "SUNDelta", "|", "SUNDeltaContract", "|", "SUNF", "|", "SUNFDelta", "|",
       "SUNFDeltaContract", "|", "SUNFIndex", "|", "SUNFIndexNames", "|", 
      "SUNFJacobi", "|", "SUNFSimplify", "|", "SUNIndex", "|", 
      "SUNIndexNames", "|", "SUNIndexRename", "|", "SUNN", "|", "SUNNToCACF", 
      "|", "SUNSimplify", "|", "SUNT", "|", "SUNTF", "|", "SUNTrace"}]}], 
    "]"}], "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"FFgc", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"FFg", "//", 
       RowBox[{
        RowBox[{"Bracket", "[", 
         RowBox[{"#", ",", 
          RowBox[{"_", "?", "FactorRelatedToColorQ"}], ",", 
          RowBox[{"#", "&"}], ",", 
          RowBox[{"CalcColorFactor", "/*", "SUNSimplify"}]}], "]"}], "&"}]}], 
      "//", "TM"}], "//", "Expand"}], "//", "TM"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFgc", "//", "Length"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Bracket", "(", 
         RowBox[{"#1", ",", 
          RowBox[{"_", "?", "FactorRelatedToColorQ"}], ",", 
          RowBox[{"#1", "&"}], ",", 
          RowBox[{"CalcColorFactor", "/*", "SUNSimplify"}]}], ")"}], "&"}], 
       ")"}], "\[InvisibleApplication]", 
      RowBox[{"(", "FFg", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.967545`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     (FeynCalc`Bracket[#, 
      PatternTest[
       Blank[], $CellContext`FactorRelatedToColorQ], #& , 
      RightComposition[
      FeynCalc`CalcColorFactor, FeynCalc`SUNSimplify]]& )[$CellContext`FFg]], 
    ": ", 0.967545, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"Expand", "[", 
      RowBox[{"TM", "(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"Bracket", "(", 
           RowBox[{"#1", ",", 
            RowBox[{"_", "?", "FactorRelatedToColorQ"}], ",", 
            RowBox[{"#1", "&"}], ",", 
            RowBox[{"CalcColorFactor", "/*", "SUNSimplify"}]}], ")"}], "&"}], 
         ")"}], "\[InvisibleApplication]", 
        RowBox[{"(", "FFg", ")"}]}], ")"}], "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "1.038521`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     Expand[
      $CellContext`TM[
       (FeynCalc`Bracket[#, 
        PatternTest[
         Blank[], $CellContext`FactorRelatedToColorQ], #& , 
        RightComposition[
        FeynCalc`CalcColorFactor, 
         FeynCalc`SUNSimplify]]& )[$CellContext`FFg]]]], ": ", 1.038521, 
    " sec"],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["980", TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Quark spin summation", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"FFgcs", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"FFgc", "//", 
       RowBox[{"MapWithProgress", "[", "FermionSpinSum", "]"}]}], "//", 
      "TM"}], "//", "Expand"}], "//", "TM"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFgcs", "//", "Length"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"MapWithProgress", "(", "FermionSpinSum", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "FFgc", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "13.565704`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`MapWithProgress[
     FeynCalc`FermionSpinSum][$CellContext`FFgc]], ": ", 13.565704, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"Expand", "[", 
      RowBox[{"TM", "(", 
       RowBox[{
        RowBox[{"MapWithProgress", "(", "FermionSpinSum", ")"}], 
        "\[InvisibleApplication]", 
        RowBox[{"(", "FFgc", ")"}]}], ")"}], "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "13.596886`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     Expand[
      $CellContext`TM[
       $CellContext`MapWithProgress[
       FeynCalc`FermionSpinSum][$CellContext`FFgc]]]], ": ", 13.596886, 
    " sec"],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["980", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FFgcs", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox[
    RowBox[{"4", " ", 
     SuperscriptBox["\<\"e\"\>", "2"], " ", 
     SubsuperscriptBox["C", "A", "2"], " ", 
     SubscriptBox["C", "F"], " ", 
     SubsuperscriptBox["g", "s", "4"], " ", 
     RowBox[{"P", "(", "\<\"uUgg\"\>", ")"}], " ", 
     SubsuperscriptBox["Q", "u", "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       FormBox[
        FormBox["k1",
         TraditionalForm],
        TraditionalForm], 
       FormBox["\<\"\[CenterDot]\"\>",
        TraditionalForm], 
       FormBox[
        FormBox["q",
         TraditionalForm],
        TraditionalForm]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       FormBox[
        FormBox["k3",
         TraditionalForm],
        TraditionalForm], 
       FormBox["\<\"\[CenterDot]\"\>",
        TraditionalForm], 
       FormBox[
        RowBox[{
         SuperscriptBox[
          FormBox["\<\"\[CurlyEpsilon]\"\>",
           TraditionalForm], "*"], "(", 
         FormBox["k3",
          TraditionalForm], ")"}],
        TraditionalForm]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       FormBox[
        FormBox["k4",
         TraditionalForm],
        TraditionalForm], 
       FormBox["\<\"\[CenterDot]\"\>",
        TraditionalForm], 
       FormBox[
        RowBox[{
         FormBox["\<\"\[CurlyEpsilon]\"\>",
          TraditionalForm], "(", 
         FormBox["k3",
          TraditionalForm], ")"}],
        TraditionalForm]}], ")"}], " ", 
     RowBox[{"tr", "(", 
      FormBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           FormBox["k1",
            TraditionalForm],
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           RowBox[{
            SuperscriptBox[
             FormBox["\<\"\[CurlyEpsilon]\"\>",
              TraditionalForm], "*"], "(", 
            FormBox["k4",
             TraditionalForm], ")"}],
           TraditionalForm]}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k1",
              TraditionalForm], "+", 
             FormBox["k3",
              TraditionalForm], "+", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox["nu",
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           FormBox["k2",
            TraditionalForm],
           TraditionalForm]}], ")"}], ".", 
        SuperscriptBox["\[Gamma]", 
         FormBox[
          FormBox["nu",
           TraditionalForm],
          TraditionalForm]], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", "(", 
          FormBox[
           FormBox[
            RowBox[{
             FormBox["k1",
              TraditionalForm], "+", 
             FormBox["k3",
              TraditionalForm], "+", 
             FormBox["k4",
              TraditionalForm]}],
            TraditionalForm],
           TraditionalForm], ")"}], ")"}], ".", 
        RowBox[{"(", 
         RowBox[{"\[Gamma]", "\[CenterDot]", 
          FormBox[
           RowBox[{
            FormBox["\<\"\[CurlyEpsilon]\"\>",
             TraditionalForm], "(", 
            FormBox["k4",
             TraditionalForm], ")"}],
           TraditionalForm]}], ")"}]}],
       TraditionalForm], ")"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"d", "-", "2"}], ")"}], " ", 
     SuperscriptBox[
      FormBox[
       FormBox["q",
        TraditionalForm],
       TraditionalForm], "2"], " ", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"], ".", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"], ".", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k1",
          TraditionalForm], "+", 
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"], ".", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         FormBox["k1",
          TraditionalForm], "+", 
         FormBox["k3",
          TraditionalForm], "+", 
         FormBox["k4",
          TraditionalForm]}], ")"}], "2"]}]}]]}], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Traces over gamma matrices", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"MemoizedTR", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{"MemoizedTR", "[", "t", "]"}], "=", 
   RowBox[{
    RowBox[{"t", "//", "TR"}], "//", 
    "Contract"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"lorentzindices", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"L", "[", "i", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "100"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FFgcst", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"FFgcs", "//", "\[IndentingNewLine]", 
         RowBox[{"MapWithProgress", "[", 
          RowBox[{
           RowBox[{"FCCanonicalizeDummyIndices", "[", 
            RowBox[{"#", ",", 
             RowBox[{"DotSimplify", "\[Rule]", "False"}], ",", 
             RowBox[{"FCTraceExpand", "\[Rule]", "False"}], ",", 
             RowBox[{"LorentzIndexNames", "\[Rule]", "lorentzindices"}]}], 
            "]"}], "&"}], "]"}]}], "//", "TM"}], "//", 
       RowBox[{
        RowBox[{"MapWithProgress", "[", 
         RowBox[{
          RowBox[{"ReplaceAll", "[", 
           RowBox[{"DiracTrace", "\[Rule]", "MemoizedTR"}], "]"}], ",", "#"}],
          "]"}], "&"}]}], "//", "TM"}], "//", "Expand"}], "//", "TM"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFgcst", "//", "Length"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"MapWithProgress", "(", 
       RowBox[{
        RowBox[{"FCCanonicalizeDummyIndices", "(", 
         RowBox[{"#1", ",", 
          RowBox[{"DotSimplify", "\[Rule]", "False"}], ",", 
          RowBox[{"FCTraceExpand", "\[Rule]", "False"}], ",", 
          RowBox[{"LorentzIndexNames", "\[Rule]", "lorentzindices"}]}], ")"}],
         "&"}], ")"}], "\[InvisibleApplication]", 
      RowBox[{"(", "FFgcs", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "7.120652`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`MapWithProgress[
     FeynCalc`FCCanonicalizeDummyIndices[#, FeynCalc`DotSimplify -> False, 
       FeynCalc`FCTraceExpand -> False, 
       FeynCalc`LorentzIndexNames -> $CellContext`lorentzindices]& \
][$CellContext`FFgcs]], ": ", 7.120652, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"MapWithProgress", "(", 
         RowBox[{
          RowBox[{"ReplaceAll", "[", 
           RowBox[{"DiracTrace", "\[Rule]", "MemoizedTR"}], "]"}], ",", 
          "#1"}], ")"}], "&"}], ")"}], "\[InvisibleApplication]", 
      RowBox[{"(", 
       RowBox[{"TM", "(", 
        RowBox[{
         RowBox[{"MapWithProgress", "(", 
          RowBox[{
           RowBox[{"FCCanonicalizeDummyIndices", "(", 
            RowBox[{"#1", ",", 
             RowBox[{"DotSimplify", "\[Rule]", "False"}], ",", 
             RowBox[{"FCTraceExpand", "\[Rule]", "False"}], ",", 
             RowBox[{"LorentzIndexNames", "\[Rule]", "lorentzindices"}]}], 
            ")"}], "&"}], ")"}], "\[InvisibleApplication]", 
         RowBox[{"(", "FFgcs", ")"}]}], ")"}], ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "7.440176`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ($CellContext`MapWithProgress[
      ReplaceAll[FeynCalc`DiracTrace -> $CellContext`MemoizedTR], #]& )[
      $CellContext`TM[
       $CellContext`MapWithProgress[
       FeynCalc`FCCanonicalizeDummyIndices[#, FeynCalc`DotSimplify -> False, 
         FeynCalc`FCTraceExpand -> False, 
         FeynCalc`LorentzIndexNames -> $CellContext`lorentzindices]& \
][$CellContext`FFgcs]]]], ": ", 7.440176, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"Expand", "[", 
      RowBox[{"TM", "(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"MapWithProgress", "(", 
           RowBox[{
            RowBox[{"ReplaceAll", "[", 
             RowBox[{"DiracTrace", "\[Rule]", "MemoizedTR"}], "]"}], ",", 
            "#1"}], ")"}], "&"}], ")"}], "\[InvisibleApplication]", 
        RowBox[{"(", 
         RowBox[{"TM", "(", 
          RowBox[{
           RowBox[{"MapWithProgress", "(", 
            RowBox[{
             RowBox[{"FCCanonicalizeDummyIndices", "(", 
              RowBox[{"#1", ",", 
               RowBox[{"DotSimplify", "\[Rule]", "False"}], ",", 
               RowBox[{"FCTraceExpand", "\[Rule]", "False"}], ",", 
               RowBox[{"LorentzIndexNames", "\[Rule]", "lorentzindices"}]}], 
              ")"}], "&"}], ")"}], "\[InvisibleApplication]", 
           RowBox[{"(", "FFgcs", ")"}]}], ")"}], ")"}]}], ")"}], "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "16.750289`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     Expand[
      $CellContext`TM[
       ($CellContext`MapWithProgress[
        ReplaceAll[FeynCalc`DiracTrace -> $CellContext`MemoizedTR], #]& )[
        $CellContext`TM[
         $CellContext`MapWithProgress[
         FeynCalc`FCCanonicalizeDummyIndices[#, FeynCalc`DotSimplify -> False,
            FeynCalc`FCTraceExpand -> False, 
           FeynCalc`LorentzIndexNames -> $CellContext`lorentzindices]& \
][$CellContext`FFgcs]]]]]], ": ", 16.750289, " sec"],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["75992", TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"DumpSave", "[", 
   RowBox[{"\"\<MemoizedTR.mx\>\"", ",", "MemoizedTR"}], "]"}], 
  ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Get", "[", "\"\<MemoizedTR.mx\>\"", "]"}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Gluon polarization summation", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"DoAllPolarizationSumsInATerm", "[", "term_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "momenta", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"momenta", "=", 
      RowBox[{
       RowBox[{"Cases", "[", 
        RowBox[{"term", ",", 
         RowBox[{
          RowBox[{"Momentum", "[", 
           RowBox[{
            RowBox[{"Polarization", "[", 
             RowBox[{"m_", ",", "_"}], "]"}], ",", "_"}], "]"}], 
          "\[RuleDelayed]", "m"}], ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "//", "Union"}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"Fold", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"DoPolarizationSums", "[", 
         RowBox[{"#1", ",", "#2", ",", "0"}], "]"}], "&"}], ",", "term", ",", 
       "momenta"}], "]"}]}]}], "\[IndentingNewLine]", 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CalcPolarizationSums", "[", "ex_", "]"}], ":=", 
  RowBox[{"ex", "//", 
   RowBox[{
    RowBox[{"Bracket", "[", 
     RowBox[{"#", ",", 
      RowBox[{"x_", "/;", 
       RowBox[{"Not", "[", 
        RowBox[{"FreeQ", "[", 
         RowBox[{"x", ",", "Polarization"}], "]"}], "]"}]}], ",", 
      RowBox[{"#", "&"}], ",", "DoAllPolarizationSumsInATerm"}], "]"}], 
    "&"}]}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"FFgcstp", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"FFgcst", "//", "CalcPolarizationSums"}], "//", "TM"}], "//", 
     "Expand"}], "//", "TM"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFgcstp", "//", "Length"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"CalcPolarizationSums", "(", "FFgcst", ")"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "10.244158`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`CalcPolarizationSums[$CellContext`FFgcst]], ": ", 10.244158,
     " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"Expand", "[", 
      RowBox[{"TM", "(", 
       RowBox[{"CalcPolarizationSums", "(", "FFgcst", ")"}], ")"}], "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "11.873678`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     Expand[
      $CellContext`TM[
       $CellContext`CalcPolarizationSums[$CellContext`FFgcst]]]], ": ", 
    11.873678, " sec"],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["22010", TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Final cleanup and export", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"FFgcstpc", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"FFgcstp", "//", 
       RowBox[{
        RowBox[{"Bracket", "[", 
         RowBox[{"#", ",", 
          RowBox[{"x_", "/;", 
           RowBox[{"Not", "[", 
            RowBox[{"FreeQ", "[", 
             RowBox[{"x", ",", "LorentzIndex"}], "]"}], "]"}]}], ",", 
          RowBox[{"#", "&"}], ",", "Contract"}], "]"}], "&"}]}], "//", "TM"}],
      "//", "Expand"}], "//", "TM"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFgcstpc", "//", "Length"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Bracket", "(", 
         RowBox[{"#1", ",", 
          RowBox[{"x_", "/;", 
           RowBox[{"\[Not]", 
            RowBox[{"FreeQ", "[", 
             RowBox[{"x", ",", "LorentzIndex"}], "]"}]}]}], ",", 
          RowBox[{"#1", "&"}], ",", "Contract"}], ")"}], "&"}], ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "FFgcstp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "4.285861`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     (FeynCalc`Bracket[#, 
      Condition[
       Pattern[$CellContext`x, 
        Blank[]], 
       Not[
        FreeQ[$CellContext`x, FeynCalc`LorentzIndex]]], #& , 
      FeynCalc`Contract]& )[$CellContext`FFgcstp]], ": ", 4.285861, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{"Expand", "[", 
      RowBox[{"TM", "(", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"Bracket", "(", 
           RowBox[{"#1", ",", 
            RowBox[{"x_", "/;", 
             RowBox[{"\[Not]", 
              RowBox[{"FreeQ", "[", 
               RowBox[{"x", ",", "LorentzIndex"}], "]"}]}]}], ",", 
            RowBox[{"#1", "&"}], ",", "Contract"}], ")"}], "&"}], ")"}], 
        "\[InvisibleApplication]", 
        RowBox[{"(", "FFgcstp", ")"}]}], ")"}], "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "4.784128`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     Expand[
      $CellContext`TM[
       (FeynCalc`Bracket[#, 
        Condition[
         Pattern[$CellContext`x, 
          Blank[]], 
         Not[
          FreeQ[$CellContext`x, FeynCalc`LorentzIndex]]], #& , 
        FeynCalc`Contract]& )[$CellContext`FFgcstp]]]], ": ", 4.784128, 
    " sec"],
   Editable->False], TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["9943", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FFgcstpc", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{"240", " ", "\[ImaginaryI]", " ", "d", " ", 
    SuperscriptBox["\<\"e\"\>", "2"], " ", 
    SubsuperscriptBox["C", "A", "2"], " ", 
    SubscriptBox["C", "F"], " ", 
    SubsuperscriptBox["g", "s", "4"], " ", 
    RowBox[{"P", "(", "\<\"uUg\"\>", ")"}], " ", 
    SubsuperscriptBox["Q", "u", "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\"\[CenterDot]\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["l1",
        TraditionalForm],
       TraditionalForm]}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      FormBox[
       FormBox["k1",
        TraditionalForm],
       TraditionalForm], 
      FormBox["\<\"\[CenterDot]\"\>",
       TraditionalForm], 
      FormBox[
       FormBox["q",
        TraditionalForm],
       TraditionalForm]}], ")"}], " ", 
    SuperscriptBox[
     RowBox[{
      FormBox["\<\"(\"\>",
       TraditionalForm], 
      FormBox[
       RowBox[{
        FormBox[
         FormBox["k2",
          TraditionalForm],
         TraditionalForm], 
        FormBox["\<\"\[CenterDot]\"\>",
         TraditionalForm], 
        FormBox[
         FormBox["k3",
          TraditionalForm],
         TraditionalForm]}],
       TraditionalForm], 
      FormBox["\<\")\"\>",
       TraditionalForm]}], "2"]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"d", "-", "2"}], ")"}], " ", 
    SuperscriptBox[
     FormBox[
      FormBox["q",
       TraditionalForm],
      TraditionalForm], "2"], " ", 
    RowBox[{
     SuperscriptBox[
      FormBox["l1",
       TraditionalForm], "2"], ".", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        FormBox["k1",
         TraditionalForm], "+", 
        FormBox["k3",
         TraditionalForm]}], ")"}], "2"], ".", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        FormBox["k2",
         TraditionalForm], "+", 
        FormBox["k3",
         TraditionalForm]}], ")"}], "2"], ".", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        FormBox["k3",
         TraditionalForm], "-", 
        FormBox["l1",
         TraditionalForm]}], ")"}], "2"], ".", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        FormBox["k2",
         TraditionalForm], "+", 
        FormBox["l1",
         TraditionalForm]}], ")"}], "2"]}]}]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"FFgcstpc2", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"FFgcstpc", "//", 
       RowBox[{"ReplaceAll", "[", 
        RowBox[{"d", "\[Rule]", 
         RowBox[{"2", "+", "dd"}]}], "]"}]}], "//", "Expand"}], "//", 
     RowBox[{"ReplaceAll", "[", 
      RowBox[{"dd", "\[Rule]", 
       RowBox[{"d", "-", "2"}]}], "]"}]}], "//", "TM"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"FFgcstpc2", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ReplaceAll", "[", 
       RowBox[{"dd", "\[Rule]", 
        RowBox[{"d", "-", "2"}]}], "]"}], "[", 
      RowBox[{"Expand", "[", 
       RowBox[{
        RowBox[{"ReplaceAll", "[", 
         RowBox[{"d", "\[Rule]", 
          RowBox[{"2", "+", "dd"}]}], "]"}], "[", "FFgcstpc", "]"}], "]"}], 
      "]"}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "1.54395`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[$CellContext`dd -> $CellContext`d - 2][
      Expand[
       ReplaceAll[$CellContext`d -> 
        2 + $CellContext`dd][$CellContext`FFgcstpc]]]], ": ", 1.54395, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["6701", TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"FFgcstpc2", ",", "\"\<FT.m\>\""}], "]"}], ";"}]], "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3. IBP machinery for the hadronic tensor", "Section"],

Cell[CellGroupData[{

Cell["Import from FeynCalc notation into den[p,m] and sp[p,q]", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{"Setup", " ", 
    RowBox[{"sp", "[", 
     RowBox[{"a", ",", "b"}], "]"}], " ", "for", " ", "future", " ", "use"}], 
   " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"SetAttributes", "[", 
     RowBox[{"sp", ",", "Orderless"}], "]"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"MakeBoxes", "[", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"a_", ",", "b_"}], "]"}], ",", "StandardForm"}], "]"}], ":=", 
    RowBox[{"RowBox", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Parenthesize", "[", 
        RowBox[{"a", ",", "StandardForm", ",", "Times", ",", "Left"}], "]"}], 
       ",", 
       RowBox[{"AdjustmentBox", "[", 
        RowBox[{"\"\<\[CenterDot]\>\"", ",", 
         RowBox[{"BoxMargins", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{"-", "0.3"}], ",", 
              RowBox[{"-", "0.3"}]}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"0", ",", "0"}], "}"}]}], "}"}]}]}], "]"}], ",", 
       RowBox[{"Parenthesize", "[", 
        RowBox[{"b", ",", "StandardForm", ",", "Times", ",", "Right"}], 
        "]"}]}], "}"}], "]"}]}], "\n", 
   RowBox[{
    RowBox[{"MakeBoxes", "[", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"a_", ",", "b_"}], "]"}], ",", "TraditionalForm"}], "]"}], ":=", 
    RowBox[{"RowBox", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Parenthesize", "[", 
        RowBox[{"a", ",", "StandardForm", ",", "Times", ",", "Left"}], "]"}], 
       ",", "\"\<\[CenterDot]\>\"", ",", 
       RowBox[{"Parenthesize", "[", 
        RowBox[{"b", ",", "StandardForm", ",", "Times", ",", "Right"}], 
        "]"}]}], "}"}], "]"}]}], "\n", 
   RowBox[{
    RowBox[{"MakeExpression", "[", 
     RowBox[{
      RowBox[{"RowBox", "[", 
       RowBox[{"{", 
        RowBox[{"a_", ",", "\"\<\[CenterDot]\>\"", ",", "b_"}], "}"}], "]"}], 
      ",", 
      RowBox[{"StandardForm", "|", "TraditionalForm"}]}], "]"}], ":=", 
    RowBox[{"MakeExpression", "[", 
     RowBox[{
      RowBox[{"RowBox", "[", 
       RowBox[{"{", 
        RowBox[{"\"\<sp\>\"", ",", "\"\<[\>\"", ",", 
         RowBox[{"RowBox", "[", 
          RowBox[{"{", 
           RowBox[{"a", ",", "\"\<,\>\"", ",", "b"}], "}"}], "]"}], ",", 
         "\"\<]\>\""}], "}"}], "]"}], ",", "StandardForm"}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"Also", " ", "setup", " ", 
     RowBox[{"B", "[", 
      RowBox[{"id", ",", 
       RowBox[{"{", 
        RowBox[{"idx", ",", "..."}], "}"}]}], "]"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"MakeBoxes", "[", 
     RowBox[{
      RowBox[{"B", "[", 
       RowBox[{"id_Integer", ",", "idx_List"}], "]"}], ",", 
      "TraditionalForm"}], "]"}], ":=", 
    RowBox[{"SubsuperscriptBox", "[", 
     RowBox[{"\"\<B\>\"", ",", 
      RowBox[{"RowBox", "[", 
       RowBox[{
        RowBox[{"idx", "//", 
         RowBox[{"Map", "[", 
          RowBox[{
           RowBox[{"MakeBoxes", "[", 
            RowBox[{"#", ",", "TraditionalForm"}], "]"}], "&"}], "]"}]}], "//", 
        RowBox[{
         RowBox[{"Riffle", "[", 
          RowBox[{"#", ",", "\"\<,\>\""}], "]"}], "&"}]}], "]"}], ",", 
      RowBox[{"MakeBoxes", "[", 
       RowBox[{"id", ",", "TraditionalForm"}], "]"}]}], "]"}]}]}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"B", "[", 
  RowBox[{"1", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 FormBox[
  SubsuperscriptBox["B", 
   RowBox[{"1", ",", "2", ",", "3"}], "1"], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Mfull", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Get", "[", "\"\<FT.m\>\"", "]"}], "//.", 
     RowBox[{"{", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"FeynAmpDenominator", "\[Rule]", "Times"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Pair", "[", 
         RowBox[{
          RowBox[{"Momentum", "[", 
           RowBox[{"p1_", ",", "d"}], "]"}], ",", 
          RowBox[{"Momentum", "[", 
           RowBox[{"p2_", ",", "d"}], "]"}]}], "]"}], "\[RuleDelayed]", 
        RowBox[{"sp", "[", 
         RowBox[{"p1", ",", "p2"}], "]"}]}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"PropagatorDenominator", "[", 
         RowBox[{"mom_", ",", "0"}], "]"}], "\[RuleDelayed]", 
        RowBox[{"den", "[", 
         RowBox[{
          RowBox[{"mom", "/.", 
           RowBox[{
            RowBox[{"Momentum", "[", 
             RowBox[{"p_", ",", "d"}], "]"}], ":>", "p"}]}], ",", "0"}], 
         "]"}]}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"SMP", "[", "\"\<Q_u\>\"", "]"}], "\[RuleDelayed]", "qu"}], 
       ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"SMP", "[", "\"\<Q_d\>\"", "]"}], "\[RuleDelayed]", "qd"}], 
       ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"SMP", "[", "\"\<g_s\>\"", "]"}], "\[RuleDelayed]", "gs"}], 
       ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"SMP", "[", "\"\<e\>\"", "]"}], "\[RuleDelayed]", "1"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"sp", "[", 
         RowBox[{"q", ",", "k1"}], "]"}], "\[Rule]", 
        RowBox[{"x", " ", 
         RowBox[{"Q2", "/", "2"}]}]}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"sp", "[", 
         RowBox[{"q", ",", "q"}], "]"}], "\[Rule]", "Q2"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"P", "[", "s_String", "]"}], "\[RuleDelayed]", 
        RowBox[{"PS", "[", 
         RowBox[{"StringLength", "[", "s", "]"}], "]"}]}]}], 
      "\[IndentingNewLine]", "}"}]}], "//", 
    RowBox[{
     RowBox[{"Bracket", "[", 
      RowBox[{"#", ",", 
       RowBox[{"_den", "|", "_sp"}], ",", "Factor"}], "]"}], "&"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Mfull", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["2800", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Mfull", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{"128", " ", "\[ImaginaryI]", " ", "CF", " ", 
    SuperscriptBox["gs", "4"], " ", 
    RowBox[{"PS", "(", "3", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", 
    RowBox[{"den", "(", 
     RowBox[{"r1", ",", "0"}], ")"}], " ", 
    RowBox[{"k1", "\[CenterDot]", "k2"}], " ", 
    RowBox[{"k1", "\[CenterDot]", "k3"}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"k1", "\[CenterDot]", "r1"}], ")"}], "2"], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
   RowBox[{"Q2", " ", "x"}]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Cases", "[", 
   RowBox[{"Mfull", ",", "_Symbol", ",", 
    RowBox[{"-", "1"}]}], "]"}], "//", "Union"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Cases", "[", 
   RowBox[{"Mfull", ",", 
    RowBox[{"x_", "\[RuleDelayed]", 
     RowBox[{"Head", "[", "x", "]"}]}], ",", 
    RowBox[{"-", "1"}]}], "]"}], "//", "Union"}]}], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
   "CA", ",", "CF", ",", "d", ",", "gs", ",", "k1", ",", "k2", ",", "k3", ",",
     "k4", ",", "l1", ",", "l2", ",", "Ngen", ",", "Q2", ",", "qd", ",", "qu",
     ",", "r1", ",", "r2", ",", "x"}], "}"}], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
   "Complex", ",", "den", ",", "Integer", ",", "Plus", ",", "Power", ",", 
    "PS", ",", "sp", ",", "Symbol", ",", "Times"}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Integrand symmetries", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"AllPermutationRules", "[", "symbols_List", "]"}], ":=", 
  RowBox[{
   RowBox[{"symbols", "//", "Permutations"}], "//", 
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{"MapThread", "[", 
      RowBox[{"Rule", ",", 
       RowBox[{"{", 
        RowBox[{"symbols", ",", "#"}], "}"}]}], "]"}], "&"}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CanonicalPermutationRules", "[", 
   RowBox[{"term_", ",", "symmetricsymbols_List"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"symmetricsymbols", "//", "AllPermutationRules"}], "//", 
   RowBox[{"MinimalBy", "[", 
    RowBox[{
     RowBox[{"ReplaceAll", "[", 
      RowBox[{"term", ",", "#"}], "]"}], "&"}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Denominators", "[", "term_", "]"}], ":=", 
  RowBox[{
   RowBox[{"term", "//", 
    RowBox[{
     RowBox[{"Cases", "[", 
      RowBox[{"#", ",", 
       RowBox[{
        RowBox[{"(", "_den", ")"}], "^", "n_."}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "&"}]}], "//", 
   RowBox[{"Apply", "[", "Times", "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CanonicalTerm", "[", 
   RowBox[{"term_", ",", "symmetricmomenta_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "rules", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"rules", "=", 
      RowBox[{"CanonicalPermutationRules", "[", 
       RowBox[{
        RowBox[{"Denominators", "[", "term", "]"}], ",", "symmetricmomenta"}],
        "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"term", "/.", "rules"}], "//", "Sort"}], "//", "First"}]}]}], 
   "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CanonicalTerm", "[", "symmetricmomenta_List", "]"}], ":=", 
  RowBox[{
   RowBox[{"CanonicalTerm", "[", 
    RowBox[{"#", ",", "symmetricmomenta"}], "]"}], "&"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Mfull", "=", 
   RowBox[{
    RowBox[{"Mfull", "//", 
     RowBox[{"MapTerms", "[", 
      RowBox[{
       RowBox[{"CanonicalTerm", "[", 
        RowBox[{"{", 
         RowBox[{"k2", ",", "k3", ",", "k4"}], "}"}], "]"}], "/*", 
       RowBox[{"CanonicalTerm", "[", 
        RowBox[{"{", 
         RowBox[{"l1", ",", "l2"}], "}"}], "]"}], "/*", 
       RowBox[{"CanonicalTerm", "[", 
        RowBox[{"{", 
         RowBox[{"r1", ",", "r2"}], "}"}], "]"}]}], "]"}]}], "//", "TM"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Mfull", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"MapTerms", "(", 
       RowBox[{
        RowBox[{"CanonicalTerm", "(", 
         RowBox[{"{", 
          RowBox[{"k2", ",", "k3", ",", "k4"}], "}"}], ")"}], "/*", 
        RowBox[{"CanonicalTerm", "(", 
         RowBox[{"{", 
          RowBox[{"l1", ",", "l2"}], "}"}], ")"}], "/*", 
        RowBox[{"CanonicalTerm", "(", 
         RowBox[{"{", 
          RowBox[{"r1", ",", "r2"}], "}"}], ")"}]}], ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "Mfull", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "1.440462`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`MapTerms[
      RightComposition[
       $CellContext`CanonicalTerm[{$CellContext`k2, $CellContext`k3, \
$CellContext`k4}], 
       $CellContext`CanonicalTerm[{$CellContext`l1, $CellContext`l2}], 
       $CellContext`CanonicalTerm[{$CellContext`r1, \
$CellContext`r2}]]][$CellContext`Mfull]], ": ", 1.440462, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["2153", TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Groups by loop number and phase space size", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Mfull", "//", "Terms"}], "//", 
    RowBox[{"Map", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_PS", ",", 
           RowBox[{"-", "1"}]}], "]"}], "//", 
         RowBox[{"Apply", "[", "Times", "]"}]}], ")"}], "*", 
       "\[IndentingNewLine]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", 
           RowBox[{"gs", "^", "n_"}], ",", 
           RowBox[{"-", "1"}]}], "]"}], "//", 
         RowBox[{"Apply", "[", "Times", "]"}]}], ")"}], "*", 
       "\[IndentingNewLine]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"Cases", "[", 
           RowBox[{"#", ",", 
            RowBox[{"l1", "|", "l2", "|", "r1", "|", "r2"}], ",", 
            RowBox[{"-", "1"}]}], "]"}], "//", "Union"}], "//", 
         RowBox[{"Apply", "[", "L", "]"}]}], ")"}], "*", 
       "\[IndentingNewLine]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"Cases", "[", 
           RowBox[{"#", ",", 
            RowBox[{"k1", "|", "k2", "|", "k3", "|", "k4"}], ",", 
            RowBox[{"-", "1"}]}], "]"}], "//", "Union"}], "//", 
         RowBox[{"Apply", "[", "PS", "]"}]}], ")"}]}], "\[IndentingNewLine]", 
      "&"}], "]"}]}], "//", "Union"}], "//", "TableForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       RowBox[{
        RowBox[{"L", "(", ")"}], " ", 
        RowBox[{"PS", "(", "2", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "2", ")"}], " ", 
        RowBox[{"L", "(", "l1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "2", ")"}], " ", 
        RowBox[{"L", "(", "r1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "4"], " ", 
        RowBox[{"PS", "(", "2", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}], " ", 
        RowBox[{"L", "(", 
         RowBox[{"l1", ",", "l2"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "4"], " ", 
        RowBox[{"PS", "(", "2", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}], " ", 
        RowBox[{"L", "(", 
         RowBox[{"l1", ",", "r1"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "4"], " ", 
        RowBox[{"PS", "(", "2", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}], " ", 
        RowBox[{"L", "(", 
         RowBox[{"r1", ",", "r2"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"L", "(", ")"}], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "4"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        RowBox[{"L", "(", "l1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "4"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        RowBox[{"L", "(", "r1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}]},
      {
       RowBox[{
        SuperscriptBox["gs", "4"], " ", 
        RowBox[{"L", "(", ")"}], " ", 
        RowBox[{"PS", "(", "4", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4"}], ")"}]}]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"M", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Mfull", "//", "Terms"}], "//", 
     RowBox[{
      RowBox[{"GroupBy", "[", 
       RowBox[{"#", ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"Cases", "[", 
              RowBox[{"#", ",", 
               RowBox[{"l1", "|", "l2", "|", "r1", "|", "r2"}], ",", 
               RowBox[{"-", "1"}]}], "]"}], "//", "Union"}], "//", 
            RowBox[{"Apply", "[", "L", "]"}]}], ")"}], "\[IndentingNewLine]", 
          
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"Cases", "[", 
              RowBox[{"#", ",", 
               RowBox[{"k1", "|", "k2", "|", "k3", "|", "k4"}], ",", 
               RowBox[{"-", "1"}]}], "]"}], "//", "Union"}], "//", 
            RowBox[{"Apply", "[", "PS", "]"}]}], ")"}]}], 
         "\[IndentingNewLine]", "&"}]}], "]"}], "&"}]}], "//", 
    RowBox[{"Map", "[", 
     RowBox[{"Apply", "[", "Plus", "]"}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Keys", "[", "M", "]"}], "//", "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       RowBox[{
        RowBox[{"L", "(", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"L", "(", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"L", "(", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"L", "(", "l1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"L", "(", "r1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"L", "(", "l1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}], " ", 
        RowBox[{"L", "(", 
         RowBox[{"l1", ",", "l2"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"L", "(", "r1", ")"}], " ", 
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}], " ", 
        RowBox[{"L", "(", 
         RowBox[{"r1", ",", "r2"}], ")"}]}]},
      {
       RowBox[{
        RowBox[{"PS", "(", 
         RowBox[{"k1", ",", "k2"}], ")"}], " ", 
        RowBox[{"L", "(", 
         RowBox[{"l1", ",", "r1"}], ")"}]}]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"M", ",", "\"\<M.m\>\""}], "]"}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Partial fraction decomposition", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ExpandSP", "[", "ex_", "]"}], ":=", 
  RowBox[{
   RowBox[{"ex", "/.", 
    RowBox[{
     RowBox[{"sp", "[", 
      RowBox[{"k1_", ",", "k2_"}], "]"}], "\[RuleDelayed]", 
     RowBox[{"Distribute", "[", 
      RowBox[{"sp", "[", 
       RowBox[{
        RowBox[{"k1", "//", "Expand"}], ",", 
        RowBox[{"k2", "//", "Expand"}]}], "]"}], "]"}]}]}], "//.", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"c_", "?", "NumberQ"}], "*", "k1_"}], ",", "k2_"}], "]"}], 
      "\[RuleDelayed]", 
      RowBox[{"c", " ", 
       RowBox[{"sp", "[", 
        RowBox[{"k1", ",", "k2"}], "]"}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1_", ",", 
        RowBox[{
         RowBox[{"c_", "?", "NumberQ"}], "*", "k2_"}]}], "]"}], 
      "\[RuleDelayed]", 
      RowBox[{"c", " ", 
       RowBox[{"sp", "[", 
        RowBox[{"k1", ",", "k2"}], "]"}]}]}]}], "\[IndentingNewLine]", 
    "}"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"PartialFractionTermOnce", "[", 
   RowBox[{"term_", ",", "externalsprules_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "dens", ",", "nums", ",", "allsps", ",", "numspcoeffs", ",", "constants",
       ",", "nsrows", ",", "nsc", ",", "row", ",", "c", ",", "n", ",", "s", 
      ",", "zero", ",", "i"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"dens", "=", 
      RowBox[{
       RowBox[{"Cases", "[", 
        RowBox[{"term", ",", "_den", ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "//", "Union"}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"dens", "===", 
        RowBox[{"{", "}"}]}], ",", 
       RowBox[{"Return", "[", "term", "]"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"nums", "=", 
      RowBox[{
       RowBox[{
        RowBox[{"dens", "/.", 
         RowBox[{
          RowBox[{"den", "[", 
           RowBox[{"k_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"k", ",", "k"}], "]"}], "-", "m2"}]}]}], "//", 
        "ExpandSP"}], "//", 
       RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"allsps", "=", 
      RowBox[{
       RowBox[{"nums", "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_sp", ",", 
           RowBox[{"{", 
            RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "&"}]}], "//", 
       "Union"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"numspcoeffs", "=", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"Coefficient", "[", 
         RowBox[{"n", ",", "s"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"n", ",", "nums"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"s", ",", "allsps"}], "}"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"constants", "=", 
      RowBox[{"nums", "/.", 
       RowBox[{"_sp", "\[Rule]", "0"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"AnyTrue", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"numspcoeffs", ".", "allsps"}], "+", "constants", "-", 
          "nums"}], ",", 
         RowBox[{"#", "=!=", "0"}]}], "]"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Error", "[", 
         RowBox[{
         "\"\<Failed to linearly decompose \>\"", ",", "dens", ",", 
          "\"\< into \>\"", ",", "allsps"}], "]"}], ";"}]}], 
      "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"nsrows", "=", 
      RowBox[{
       RowBox[{"numspcoeffs", "//", "Transpose"}], "//", "NullSpace"}]}], ";",
      "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"nsrows", "===", 
        RowBox[{"{", "}"}]}], ",", 
       RowBox[{"Return", "[", "term", "]"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"nsrows", "=", 
      RowBox[{
       RowBox[{"nsrows", "//", "RowReduce"}], "//", "Expand"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"nsc", "=", 
      RowBox[{"nsrows", ".", "constants"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"AnyTrue", "[", 
        RowBox[{"nsc", ",", 
         RowBox[{
          RowBox[{"#", "=!=", "0"}], "&"}]}], "]"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{
         RowBox[{"{", 
          RowBox[{"row", ",", "c"}], "}"}], "=", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{"{", 
               RowBox[{"nsrows", ",", "nsc"}], "}"}], "//", "Transpose"}], "//", 
             RowBox[{
              RowBox[{"Cases", "[", 
               RowBox[{"#", ",", 
                RowBox[{"{", 
                 RowBox[{"_", ",", 
                  RowBox[{"Except", "[", "0", "]"}]}], "}"}]}], "]"}], 
              "&"}]}], "//", 
            RowBox[{
             RowBox[{"MinimalBy", "[", 
              RowBox[{"#", ",", 
               RowBox[{
                RowBox[{"Count", "[", 
                 RowBox[{
                  RowBox[{"#", "[", 
                   RowBox[{"[", "1", "]"}], "]"}], ",", 
                  RowBox[{"Except", "[", "0", "]"}]}], "]"}], "&"}]}], "]"}], 
             "&"}]}], "//", "Sort"}], "//", "First"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"row", ".", 
            RowBox[{"(", 
             RowBox[{"1", "/", "dens"}], ")"}]}], "/", "c"}], "*", "term"}], "//",
          "Expand"}]}], "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"row", "=", 
         RowBox[{
          RowBox[{
           RowBox[{"nsrows", "//", 
            RowBox[{
             RowBox[{"MinimalBy", "[", 
              RowBox[{"#", ",", 
               RowBox[{"Count", "[", 
                RowBox[{"Except", "[", "0", "]"}], "]"}]}], "]"}], "&"}]}], "//",
            "Sort"}], "//", "First"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"zero", "=", 
         RowBox[{
          RowBox[{
           RowBox[{"{", 
            RowBox[{"row", ",", "dens"}], "}"}], "//", "Transpose"}], "//", 
          RowBox[{"Cases", "[", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"Except", "[", "0", "]"}], ",", "_"}], "}"}], "]"}]}]}], 
        ";", "\[IndentingNewLine]", 
        RowBox[{"i", "=", 
         RowBox[{
          RowBox[{"FirstPosition", "[", 
           RowBox[{"zero", ",", 
            RowBox[{"{", 
             RowBox[{"_", ",", 
              RowBox[{"den", "[", 
               RowBox[{"_", ",", "_"}], "]"}]}], "}"}], ",", "$Failed", ",", 
            RowBox[{"{", "1", "}"}]}], "]"}], "//", "First"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"zero", "=", 
         RowBox[{"zero", "//", 
          RowBox[{"Map", "[", 
           RowBox[{"Apply", "[", "Divide", "]"}], "]"}]}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
          RowBox[{"1", "/", 
           RowBox[{"zero", "[", 
            RowBox[{"[", "i", "]"}], "]"}]}], "*", 
          RowBox[{"-", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"zero", "//", 
              RowBox[{
               RowBox[{"Delete", "[", 
                RowBox[{"#", ",", "i"}], "]"}], "&"}]}], "//", 
             RowBox[{"Apply", "[", "Plus", "]"}]}], ")"}]}], "*", "term"}], "//",
          "Expand"}]}]}], "\[IndentingNewLine]", "]"}]}]}], 
   "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"PartialFraction", "[", 
   RowBox[{"ex_", ",", "externalsprules_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"oldex", ",", "newex"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"oldex", "=", 
      RowBox[{"ex", "//", "Expand"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"While", "[", 
      RowBox[{"True", ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"newex", "=", 
         RowBox[{"oldex", "//", 
          RowBox[{"MapTerms", "[", 
           RowBox[{
            RowBox[{"PartialFractionTermOnce", "[", 
             RowBox[{"#", ",", "externalsprules"}], "]"}], "&"}], "]"}]}]}], 
        ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"newex", "===", "oldex"}], ",", 
          RowBox[{"Break", "[", "]"}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"oldex", "=", "newex"}], ";"}]}], "\[IndentingNewLine]", 
      "]"}], ";", "\[IndentingNewLine]", "newex"}]}], "\[IndentingNewLine]", 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"PartialFraction", "[", "externalsprules_List", "]"}], ":=", 
  RowBox[{
   RowBox[{"PartialFraction", "[", 
    RowBox[{"#", ",", "externalsprules"}], "]"}], "&"}]}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["IBP topologies", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Topologies", "[", "ex_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"densets", ",", "denset", ",", "topologies"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"densets", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"ex", "//", "Terms"}], "//", 
          RowBox[{"Map", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"Cases", "[", 
              RowBox[{"#", ",", "_den", ",", 
               RowBox[{"{", 
                RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "&"}], "/*", 
            "Union"}], "]"}]}], "//", "Union"}], "//", 
        RowBox[{"SortBy", "[", "Length", "]"}]}], "//", "Reverse"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"topologies", "=", 
      RowBox[{"{", "}"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Do", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"AnyTrue", "[", 
           RowBox[{"topologies", ",", 
            RowBox[{
             RowBox[{"SubsetQ", "[", 
              RowBox[{"#", ",", "denset"}], "]"}], "&"}]}], "]"}], ",", 
          RowBox[{"Continue", "[", "]"}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"AppendTo", "[", 
         RowBox[{"topologies", ",", "denset"}], "]"}], ";"}], 
       "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"denset", ",", "densets"}], "}"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"topologies", "//", "Sort"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"MakePrettyTopologyGraph", "[", 
   RowBox[{"ex_", ",", "momenta_List"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"ex", "//", 
     RowBox[{
      RowBox[{"Cases", "[", 
       RowBox[{"#", ",", "_den", ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "&"}]}], "//", 
    "Union"}], "//", 
   RowBox[{
    RowBox[{"FindGraph", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"Plus", "@@", "momenta"}], "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"Plus", "@@", "momenta"}], "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"Join", "[", 
       RowBox[{"momenta", ",", 
        RowBox[{"#", "/.", 
         RowBox[{
          RowBox[{"den", "[", 
           RowBox[{"p_", ",", "0"}], "]"}], "\[RuleDelayed]", "p"}]}]}], 
       "]"}], ",", "\[IndentingNewLine]", 
      RowBox[{"EdgeFn", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"#1", ",", 
           RowBox[{"#2", "/.", 
            RowBox[{"{", "\[IndentingNewLine]", 
             RowBox[{
              RowBox[{
               RowBox[{"Internal", "[", 
                RowBox[{"momenta", "[", 
                 RowBox[{"[", "1", "]"}], "]"}], "]"}], "\[Rule]", 
               RowBox[{"{", 
                RowBox[{"Thick", ",", "Red", ",", "Dashed"}], "}"}]}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{
               RowBox[{"Internal", "[", 
                RowBox[{"Alternatives", "@@", "momenta"}], "]"}], "\[Rule]", 
               "Dashed"}], ",", "\[IndentingNewLine]", 
              RowBox[{"_Incoming", "\[Rule]", "Gray"}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{"_Outgoing", "\[Rule]", "Gray"}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{"_", "\[Rule]", "Thick"}]}], "\[IndentingNewLine]", 
             "}"}]}]}], "]"}], "&"}], ")"}]}]}], "\[IndentingNewLine]", "]"}],
     "&"}]}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Do", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Print", "[", 
      RowBox[{"Style", "[", 
       RowBox[{"key", ",", "Bold", ",", "Larger"}], "]"}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"M", "[", "key", "]"}], "//", "Topologies"}], "//", 
        RowBox[{"Map", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"Times", "@@", "#"}], ",", 
            RowBox[{"MakePrettyTopologyGraph", "[", 
             RowBox[{"#", ",", 
              RowBox[{"key", "/.", 
               RowBox[{
                RowBox[{"l_L", "*", "ps_PS"}], "\[RuleDelayed]", 
                RowBox[{"List", "@@", "ps"}]}]}]}], "]"}]}], "}"}], "&"}], 
         "]"}]}], "//", "TableForm"}], "//", "Print"}], ";"}], 
    "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
    RowBox[{"{", 
     RowBox[{"key", ",", 
      RowBox[{
       RowBox[{"Keys", "[", "M", "]"}], "//", "Sort"}]}], "}"}]}], "]"}], 
  ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {"1", 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1]}, {Null, {{1, 2}, {2, 3}, {2, 3}, {3, 4}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 2] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBUQKu/SFib1JlAgAAACQAAAACAAAAS5rUjjXGB0D4lKVRQzPDP0ELKF6Q
t/8/AAAAAAAAAAAAjJu0A9PvPzNhaQeQfsU/AAAAAAAAAACAJtdiPBi4P9gf
tFrR8v0/4MLlt8w6k78hthp24n/9P4A0wP4yZ5e/Zue4IEgM/T/ApoU8tzCa
v44D0MxGmPw/AMJ5+7KVm78bfJwpIyT8PyDtn95SlZu/CC+r+iGw+z/AO+Ae
ly+av5+qIu+HPPs/YLvlilNll79PhRh5mcn6P4A4tgkvOJO/A9BaRC139D8Q
TJogAsioPzUoTXAuBfQ/oIxOVciOqz+zSti8YpTzP8B1DKgdBK8/elrA8gwl
8z942iv0+pKxP0SjZv1ut/I/+K3TYO/4sz8sAMLDyUvyPxhEHpeAsrY/R3Lx
AV3i8T9In8uSEb65P8WnfyNne/E/wE/FwtQZvT98KfYUqyX+P3DlGpiWxqs/
+16ENrW+/T8sIwd8Dj+xPxbRs3RIVf0/XH60d59KtD/+LQ87o+n8P4AU/60w
BLc/x3a1RQV8/D/456YaJWq5P46GnXuvDPw/jIfMOhF7uz8NqSjI45v7Pxx8
K+S7Nb0/PwEb9OQp+z9knIX+Hpm+P/NLXb941/Q/RCig6JXlxz+jJlNJimT0
P6AYxng6a8g/OaLKPfDw8z+waEXrgsTIPyZV2Q7vfPM/2l49YzrxyD+0zaVr
ywjzP3KZ2GZG8cg/3Om8F8qU8j8MFvrupsTIPyEbW8IvIfI/xGdBZ3ZryD9p
scHdQK7xP44ZZp7p5cc/QUYc1A==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03003681795644597]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.460798315079948, -0.07847324791946825}, 3}], 
                 0.03003681795644597]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.5160257020134742`, 0.2463981323522549}, 3}], 
                 0.03003681795644597]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03003681795644597]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06007363591289194, 0.06007363591289194}], 
              {GrayLevel[0], DiskBox[2, 0.049862173792727754`]}, 
              {GrayLevel[0], DiskBox[3, 0.049862173792727754`]}, InsetBox[
               GraphicsBox[{}], 4, 
               Automatic, {0.06007363591289194, 0.06007363591289194}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", "l1", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 4}, {4, 5}, {3, 5}, {5, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.814781421808071, 0.4877084622328541}, {
            2.825162389780683, 0.453296561078493}, {1.9077379198057707`, 
            0.913800986380224}, {1.9070977558442062`, 0.}, {
            0.9896915982171941, 0.4559929121833412}, {0., 
            0.48656376120221734`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.035810735981566255`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.035810735981566255`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.035810735981566255`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.035810735981566255`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07162147196313251, 0.07162147196313251}], 
              {GrayLevel[0], DiskBox[2, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[3, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[4, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[5, 0.04569006053069333]}, InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.07162147196313251, 0.07162147196313251}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", "r1", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 4}, {4, 5}, {3, 5}, {5, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.814781421808071, 0.4877084622328541}, {
            2.825162389780683, 0.453296561078493}, {1.9077379198057707`, 
            0.913800986380224}, {1.9070977558442062`, 0.}, {
            0.9896915982171941, 0.4559929121833412}, {0., 
            0.48656376120221734`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.035810735981566255`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.035810735981566255`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.035810735981566255`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.035810735981566255`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07162147196313251, 0.07162147196313251}], 
              {GrayLevel[0], DiskBox[2, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[3, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[4, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[5, 0.04569006053069333]}, InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.07162147196313251, 0.07162147196313251}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2"}], ")"}], " ", 
    RowBox[{"L", "(", 
     RowBox[{"l1", ",", "l2"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 5}, {4, 6}, {3, 6}, {5, 4}, {5, 
            4}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAKMVgjURFC0Ak5XRK27D7P6iYea/Y
ZgRAEDCkz5759T/W+c9O3jj9P5QH2Az6s+U/QvkUKE/j+T+pqYOXAl/1P+RM
cS0aQPk/AAAAAAAAAAB8sdmVtmjqP0b6rGVUX/U/AAAAAAAAAAAEt1Wc4KL8
P9+NUNJOnPo/K67dOCkX8z/0+SCYh8f6PxIf8K1sgvI/eq/M7ozr+j8Fg70y
yOvxP6Aqi4JJCPs/7feV+JRT8T9hGatPrB37P4+p4BwtuvA/QUilrKgr+z94
2sBz6x/wP+zhm1E2Mvs/wXadpFYK7z9ukUFdUTH7P0PQB7GQ1O0/SBkA+Qjw
+j8Xkdaig3fZP6on5PKx5/o/GSy1LNsM1z/GOCLL7df6P/VwYHOEpNQ/1dZ1
18XA+j9LphBj7D/SP+OHs81Hovo/C27GYPvAzz/Ysqq7hXz6P9bGNANAD8s/
lCJ1/JVP+j8uVFgzbm3GP0J7OiuTG/o/LPo8C0TewT/myksq1gf5P2MKHBY6
I/M/kiMRWdPT+D8inxjRVJHyP06T25njpvg/zRAdlxr98T9EvtKHIYH4P+fb
aivjZvE/UW8QfqNi+D8VgL9+B8/wP2ANZIp7S/g/ao2reuE18D99HqJitzv4
P0O9rJiXN+8/3yyGXGAz+D/DCpxdQwLuP7u0RPgX8vc/GQb/++jS2T89ZOoD
M/H3Px250xRdZ9c/6P3gqMD39z+9PAuPXPzUP8ks2wW9Bfg/ZQCM6lWT0j+I
G/vSHxv4P+rGtnu2LdA/r5a5Ztw3+D8aNTEm05nLPzZMZb3hW/g/qlScTK/k
xj9KuDWDGof4P+TbL/XKPsI/M8499A==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03310124558093995]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03310124558093995]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.817715508625982, 0.6612957005613861}, 5}], 
                 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.37842986765586, 0.6744001305971647}, 5}], 
                 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03310124558093995]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03310124558093995]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0662024911618799, 0.0662024911618799}], 
              {GrayLevel[0], DiskBox[2, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[3, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[4, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[5, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[6, 0.03448552094633272]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.0662024911618799, 0.0662024911618799}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 4}, {3, 4}, {5, 6}, {4, 5}, {3, 
            6}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAA/E05uD4zDUB2tkrqZ1T0PwXH685H
5QVAixDEe8vb7z+a/Nlh/7/9P1r9omt6MvQ/s2cDn4OaAUDgdpS4p2jOPw5+
YsIIyfI/AAAAAAAAAADJyEoVCKPrP6n+tjdxFvA/AAAAAAAAAAATRd9v/T33
P+pCa5p/zgBAVDqwF0hXxj/RfXYcIpoAQKynG1v0YMQ/+n5ZtVFkAEA0K19y
fZPCP+7U1EEuLQBA0Pn2lvTvwD9eMM/OsOn/P9D04F2k7r4/eTP2AeN2/z+Q
ka9z6lS8P+5PRhY3Av8/wHdemUUUuj/BNv0f8ov+P8g2fjoLLrg/n3y7IVn6
9z8AWEf8EtdCv2BwRifBgvc/gMQYZ+IFe7+qXcz3HAr3P8BU1sTQ9Ia/JxtB
AbSQ9j8A4LJPeH2Nv7ajGibOFvY/ALCRvYSMkL8YVcGSs5z1P2Dju7mw4pC/
Q2bUkqwi9T+AuBIuGoKPv3ziW2YBqfQ/wETUpYpQir98tQZNh6oAQJbdeDHY
BtA/mHPKtrFtAEA10bqtZDDQPy380zauMABApvnl535C0D+8qU7aQef/P3BW
IygcPdA/TDIo/1tt/z9v0seePyDQP8nvnAjz8/4/KtzhxPTXzz8S3SLZTnv+
PwA9zcvWQM8/09Ct3rYD/j82vpDLfnvOP7MWbOAdcvc/eFtVG6JRwj+G/SLq
2Pv2P/w65euEXsE/+hlz/iyH9j8Urrx+Mj7APxUdmjFfFPY/7PhHE6vivT+Y
o798s6P1PxD6OkNm8bo/fk+2lWw19T9Ml2qMVKq3P9NRfMfLyfQ/YJ7xumYP
tD+gx5LLEGH0Pwh5yEG/IrA/vb47Nw==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0347263707657203]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0347263707657203]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.7263294841374197`, -0.04999193827856571}, 5}], 
                 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.6481974949339981`, 0.2875607603066587}, 5}], 
                 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0347263707657203]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0347263707657203]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0694527415314406, 0.0694527415314406}], 
              {GrayLevel[0], DiskBox[2, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[3, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[4, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[5, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[6, 0.04586111767247925]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.0694527415314406, 0.0694527415314406}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 3}, {4, 3}, {5, 6}, {3, 5}, {4, 
            6}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAwKGOZlUzDUD4HFcBpFjHP9WdSE8r
5QVAPSFs0qFC3T8p4Xv8dpoBQLL6QvsIcvM/RtLd+Wq//T/mVvX7ynHIPyiW
wMvTyPI/9ATbc4RA9z+gGh+qH6HrP0srvLIRqdw/AAAAAAAAAAAAAAAAAAAA
AClN15NyqgBAAhPovIw98z9XpcQOm20AQLKpZgczM/M/EZ522pUwAECyeTYP
ti7zPywt5y8O5/8/uhWgfBgw8z9jkdTbJW3/P1zsyX1ZN/M/YoSU5rrz/j+S
hzTHdETzPzliyzMVe/4/yIpFlmJX8z8SU1IyfAP+P+Lt37UXcPM/aEE8INhx
9z+Wi+my4vX0P+jr1Z+S+/Y/rFoQglAU9T/6WEGn5ob2P3wMkPxkOPU/6tvD
ShkU9j/uTqnFCmL1PyOLQoNuo/U/o/i2NCmR9T8WgQMGKTX1P4+ix2OkxfU/
NrkuHYrJ9D+URCJAXf/1P9PrJYHRYPQ/dg2snDE+9j9UNskheM4AQDLycdJb
dPQ/os/E0xuaAEAWu/suMLP0P7NrWl9MZABAGl1WC+ns9D+s5rqgKS0AQAYH
ZzpkIfU/kXz0eajp/z+6sHSpglD1P4D/dh3bdv8/LPONcih69T+SbOIkLwL/
P/ykDe08nvU/Ehd8pOmL/j8SdDS8qrz1P2gFZpJF+vc/xhE+uXVC9z9A9uyQ
rIL3P+F02NgqW/c/GNQj3gYK9z8WeOmnGG73PxbH4+ibkPY/TBNU8TN79z9N
K9GUsxb2P+7pffJ0gvc/VhzLD5ac9T/1hedf14P3P8wNL6eLIvU/9lW3Z1p/
9z8nvgmd3Kj0P6bsNbIAdfc/zCE0hA==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03472665818833397]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03472665818833397]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.6481039685042613`, 1.165515264345253}, 5}], 
                 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.7263484016032933`, 1.5030766489276046`}, 5}], 
                 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03472665818833397]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03472665818833397]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06945331637666793, 0.06945331637666793}], 
              {GrayLevel[0], DiskBox[2, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[3, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[4, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[5, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[6, 0.045857674204805216`]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06945331637666793, 0.06945331637666793}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 5}, {6, 7}, {3, 7}, {4, 5}, {5, 
            6}, {4, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.82775950748957, 0.34102566225772063`}, {
            2.857279458924138, 0.3642453300352074}, {1.9008765210392695`, 
            0.}, {2.4882573319258685`, 1.450220578471928}, {
            1.9176099356458205`, 1.0859896894096428`}, {1.3410753682205372`, 
            1.4692240547736586`}, {0.9581671951634747, 0.4009557497035342}, {
            0., 0.32507801395474695`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358952851820334]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358952851820334]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0358952851820334]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358952851820334]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0717905703640668, 0.0717905703640668}], 
              {GrayLevel[0], DiskBox[2, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[3, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[4, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[5, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[6, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[7, 0.033849024780202346`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0717905703640668, 0.0717905703640668}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {4, 5}, {6, 7}, {4, 7}, {3, 5}, {5, 
            6}, {3, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
                Thickness[Large]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.374022807750645, 1.877601037910519}, {
            2.6200116887705454`, 1.2861863965154487`}, {2.317945018414745, 
            0.31556980190522654`}, {1.6954182340426815`, 
            1.3509692689724844`}, {1.6888887135352744`, 0.}, {
            1.0670885557293825`, 0.32525130093028776`}, {0.7660058071489272, 
            1.2967119800808657`}, {0., 1.8657165026996045`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03286298885040069]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03286298885040069]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03286298885040069]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03286298885040069]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06572597770080138, 0.06572597770080138}], 
              {GrayLevel[0], DiskBox[2, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[3, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[4, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[5, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[6, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[7, 0.035086459104769915`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06572597770080138, 0.06572597770080138}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "l1", "+", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 5, 6, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 6}, {7, 5}, {3, 7}, {3, 6}, {6, 
            4}, {4, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 5] -> {
                Thickness[Large]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274196709203682, 0.5136246102169043}, {
            3.3670820276555293`, 0.5053864523625655}, {2.530169512788247, 
            0.}, {2.5304984937516943`, 1.0131836739104354`}, {
            0.9071271193738002, 0.5077856682990879}, {1.743514219500912, 
            0.0006579432279606445}, {1.7442270282497168`, 
            1.0137999850147033`}, {0., 0.5122562334524448}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.038729380658446846`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.038729380658446846`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.038729380658446846`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.038729380658446846`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07745876131689369, 0.07745876131689369}], 
              {GrayLevel[0], DiskBox[2, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[3, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[4, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[5, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[6, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[7, 0.03931358535233097]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07745876131689369, 0.07745876131689369}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 5, 6, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 6}, {7, 5}, {4, 3}, {3, 7}, {4, 
            6}, {6, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598366806140282, 0.5457743356918155}, {
            3.633548263833968, 0.5250650316016542}, {2.8217992036599027`, 
            0.}, {2.821231543079123, 1.0517221476263738`}, {
            0.9648178152185232, 0.5256291579082695}, {1.7773076412298578`, 
            1.0519135692253971`}, {1.7764749771256565`, 
            0.0005493950476227472}, {0., 0.5453390276145662}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070062802933476]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070062802933476]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070062802933476]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070062802933476]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140125605866952, 0.08140125605866952}], 
              {GrayLevel[0], DiskBox[2, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[3, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[4, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[5, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[6, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[7, 0.04825095583319784]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140125605866952, 0.08140125605866952}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2"}], ")"}], " ", 
    RowBox[{"L", "(", 
     RowBox[{"l1", ",", "r1"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 6, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {5, 6}, {5, 3}, {2, 5}, {2, 3}, {4, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598716580880777, 0.5111844757682964}, {
            3.633781226626953, 0.5248792070768915}, {2.820926459245353, 0.}, {
            1.7768649313204117`, 0.0017248097633325576`}, {
            2.8224137730810934`, 1.0514880756581495`}, {1.777143959522975, 
            1.05308695777259}, {0.9648670034137008, 0.5287805535604434}, {0., 
            0.5085688457319367}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.04070271725560909]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.04070271725560909]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070271725560909]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070271725560909]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140543451121818, 0.08140543451121818}], 
              {GrayLevel[0], DiskBox[2, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[3, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[4, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[5, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[6, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[7, 0.048251626489543516`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140543451121818, 0.08140543451121818}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2"}], ")"}], " ", 
    RowBox[{"L", "(", 
     RowBox[{"r1", ",", "r2"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1", "-", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 5}, {4, 6}, {3, 6}, {5, 4}, {5, 
            4}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAKMVgjURFC0Ak5XRK27D7P6iYea/Y
ZgRAEDCkz5759T/W+c9O3jj9P5QH2Az6s+U/QvkUKE/j+T+pqYOXAl/1P+RM
cS0aQPk/AAAAAAAAAAB8sdmVtmjqP0b6rGVUX/U/AAAAAAAAAAAEt1Wc4KL8
P9+NUNJOnPo/K67dOCkX8z/0+SCYh8f6PxIf8K1sgvI/eq/M7ozr+j8Fg70y
yOvxP6Aqi4JJCPs/7feV+JRT8T9hGatPrB37P4+p4BwtuvA/QUilrKgr+z94
2sBz6x/wP+zhm1E2Mvs/wXadpFYK7z9ukUFdUTH7P0PQB7GQ1O0/SBkA+Qjw
+j8Xkdaig3fZP6on5PKx5/o/GSy1LNsM1z/GOCLL7df6P/VwYHOEpNQ/1dZ1
18XA+j9LphBj7D/SP+OHs81Hovo/C27GYPvAzz/Ysqq7hXz6P9bGNANAD8s/
lCJ1/JVP+j8uVFgzbm3GP0J7OiuTG/o/LPo8C0TewT/myksq1gf5P2MKHBY6
I/M/kiMRWdPT+D8inxjRVJHyP06T25njpvg/zRAdlxr98T9EvtKHIYH4P+fb
aivjZvE/UW8QfqNi+D8VgL9+B8/wP2ANZIp7S/g/ao2reuE18D99HqJitzv4
P0O9rJiXN+8/3yyGXGAz+D/DCpxdQwLuP7u0RPgX8vc/GQb/++jS2T89ZOoD
M/H3Px250xRdZ9c/6P3gqMD39z+9PAuPXPzUP8ks2wW9Bfg/ZQCM6lWT0j+I
G/vSHxv4P+rGtnu2LdA/r5a5Ztw3+D8aNTEm05nLPzZMZb3hW/g/qlScTK/k
xj9KuDWDGof4P+TbL/XKPsI/M8499A==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03310124558093995]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03310124558093995]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.817715508625982, 0.6612957005613861}, 5}], 
                 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.37842986765586, 0.6744001305971647}, 5}], 
                 0.03310124558093995]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03310124558093995]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03310124558093995]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0662024911618799, 0.0662024911618799}], 
              {GrayLevel[0], DiskBox[2, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[3, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[4, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[5, 0.03448552094633272]}, 
              {GrayLevel[0], DiskBox[6, 0.03448552094633272]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.0662024911618799, 0.0662024911618799}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None,
       ImageCache->GraphicsData["CompressedBitmap", "\<\
eJyFWgl4VdW1PiWECB2sFKvVVob2VR+ttbWtRYtlqMHXKlQftjgiAREQGQS0
ilJEKoggM8isQMgAhJBAwpSEJAwJhCEzyBSGBIUkkBAgCSSsrn/vtc495xrp
/b6be8/ee83zvukZMWbwwGERY4YMiGj3+KiIkYOHDBjdrseIUbwU8i3HcU7x
e1w7B9+Jv8qfU8Qv3mzqY5D5CKHr169TYWEhPp1/2q1XzUcopaSkUHh4OHXo
0IEef/xxysjIcN6yRwYKdH19PXXr1s2Qa2b+hlCPHj0MVpwY4z9+8OBBateu
ne84sIOBJo4De9euXX3HH3ssPBi78rt0RSx1uPdX9J1b29AP7v459Rk8Qfm1
R1rQxowCGjBuBXUMH0EP9ZlMPQbOodr6a2/6T81auZ1efT+Knh6xiDo9P5Ue
fmEapWYfftPLXRg13rhBIyavMSc/XraNHu073ZwePXWd4zt6C5356qI5hveO
/ceo98jF5uhfBs+jyqorY/ynt+894p7+quKSQY7TnV/6hJIyC4NOL167y5wc
/XEcVmh37gnqFjHLQIz8aI3vdEt6e8Z6c/rTmExz+tLlWpebXq8voJKySme0
BXnFfDSnC9VXXHa27j7UTJYnzE82iukaMZOik/cFAe0tOOkCHT9T3kxMBP1D
5yA37MNYamhsdEb5IaOScgzU0H/H0vWGRoU8dfYC/WP0UgMJhnMPl/ohQ2ji
gk0GcupnKfxkwdhGMAj98cVPKJwJL1y9U8EGCMHaums0eEK0AY1PzWsukCsS
99Bfh8ynR1hGbF+prXfe8EMWHfvSFbLgSJlC5h8poxfe+syw+uyYZZS572gw
ZOL2fAM1yCJWyGvXG4zcf3p5hqE9Zek2P2QIzYvONJDj5mxkT7Vg05enGuN1
7z+LMUZR+YUaZ6QfbNKizQZsxoo0F4ytRi+PXWF0g70N6QXBYG99Em+2lsVn
CVgLSs+x7glqEe+tpHdnJULNIyxkf5HwBi8N/sBqdV1KrgJ/UXLOLMGEz735
mfl+7HR5EHB1Ta2r2NTsLxS4+rJd7j8ukv7+xhLzPXLD3iDgk2crXeD9xafD
BPiNKXHuMt4jP1oL/xruB877otTrt2GiB4SgCTKO7Wb4xmuMm+TlvO7HwhZ3
sVRcvKwsrGSXwtIIJv3GlLXm+7zojCAW1DfwZgbDJN1syzrsLi+RmB/yQQzV
XKlzhnkxhBjXlZTACcJiLTx21gVPyTqsQqTt+cIVYqgfy5xV6eb0B58mC5Yw
JCsXC9LUe7M3mO/sXsFM/HuhjUZOpgJuo3HYpNVmHXEOj1NOviyvdjl5zY9q
DAcwQJYnZHMKs5x8KA4N32d/N34vEazWiBDROcmYQMP2+rQ8i+EWWrZut1mC
j+OlNkaKZO8d6kfizYIcAS1Fok07itxlDmQT6yoRfF0lGuLHVlJa4YIdOHRG
Wco9fMZd5qgwGJEP8Pz5+uwglhjQPc152yJpaUqGLu86eNzQ57Rnnodzvaq/
1qAKVkx78gPZuvxiTUvxV2RGXYageHGAqHzI8Srf4G9GWHruYissW7AWlJxp
ir0xy2ix7KzI7UEsIYkqBi6frcQV1MycUhytRMhjWEO64cwRpOoCD6Kjp84r
Iq3ccdsOuohOlgUyB1e6IEQM7O4eOXlOESGPYGnVxhypOiE0fu5Gs8YVKdjy
x04HsLCMikXrMqczwdKckj2+xSoIUjGr1d1loyiid2Yk2BQRt8stgvBdBMDA
8StpWXSS6fUG+bGVnatysbHxFBuKPJYWrdnpaJGavTCSunbr7msLX/Vj86aJ
7LwSxcaZxCxxZnFV9d86yFduGoaKWWOX/UisGYLESJPnrKJ72rblp683m0GY
r3qcnVNbKwlJbibcDg+vfUWnjNeii7zvlw8qYvTJ38Ay95feKGoly570Sp/G
7ggx0RFKr78zjbp06UJt27Y1euEmXLXbPxBFIWgE3eC7RSSeG5VBfx+1BI2O
o9E2ittCDVTo+sCBA3Tt2jVVPxUG0ER4yJhQc6u3m3UvXrpqaj5am679Zqpv
oI10wkQw2riR6MkniRYtIk6kQVg17tZuPehoaUNZQFuMPbTEwP7nAbOR4ES2
FmCe6JVXLGa8584F9n5+7N5gbCHYOaGh7XFt8MRr8w0V1ozTSrFXVBANGhTA
Pn06UWPjy37sGqQrOUibyxr6HFRCLmui8+b0z+nr2W+sR7JM35bAoQsXiIYO
tQRmzQL7ThAFLZozV6ZJhLSic5WXXC+0PWOU6STgsNA91rgH+45atJrr6IoV
YN/xaSeMlsbZigf+8MIEw3kOJUZ4D6OxMxPp0Imv1CloqkQW+7/ja7RCTSJ3
6w8XtMmLt7i1D86DvFxXf90NdW0WkExaK7e5uURXrpin573obXkbOD6Knhq+
kPqNXaGYx83ZQEc4FYuzGSxaI2M373du0529e4meeorbfx5Namp8yMNM86ZT
EkIGHf7arQfQgcNSBgOqr8q3Zdeh29SMkyYF/GT4cCjcec5i7yfYuW2jhLR8
xhpjZiRQ6fvOcjpRWuGGSdn5QMbNzi+5TT0RSWTKlACF17gwVlU9awm8LASg
7vfnJ7kIuvEQ1mvYApodme52fMXHA9MJW7S1ss9l13gfkL/7LgLLEewviTox
hnB8oqC6Po3hasrSreptSgQNhhL5qqL6B2pWhPR69rK6OtXNi+bj26YAwoYK
9drEGErKKDTtAHRmKjlbA5ZQXXm7LE6n4ofPC7G6a9cpZtM++JywG0qTFm8x
GtYXp0DXNzGNy0tsHeL2EAjju1SI4mKio0fNU29L8jnRPyIEkaJMwQHZRRBR
bvn2Dv1MXkeJBbE7zBLrwPmR+uqRI0R9+th3cbEQ6yMQaMO4L0DguzKgHKHR
C0RBc2M2TdqsSSX40RJrtU8+TxWCLPSmTUQ9e1oneOYZooIC5/8t1X+IiAVH
z5pEoSIM+3C1mRY44TmKGqMf9jBV65q3AblLyW3dStSrlyXXu7ch97Ql94yA
Idg+i88y4qmQprfn7E1BNQ7o35+X5JKEC2FtzZYDzt2qkowMor/9zSbcqiqh
1luEO8i5Zcy0da5w6IszuQKx2zpaoLkzNHsYshobb2hWuMRthYLxTPZjdZZ9
+2wV4Sch9rRsAUJuaVzRZkduN72MjYmvxxJPQ0rQ60c5hadkQn9K8gVqO65+
9ATmyqzcE98T4BP+QUNx5h4OjLnc17ZXIc6eJe44zNOTlk4v2aqquUrzYzJ9
QuAZozpOKEEEo9qDg1IJxqfkupHN/txOrVReTtSvn/XFxESh2VNkw1WWVnS8
0XexBrR2QIeSjtw41kKJzru9eh9i+NlnAyl19WrnCUvpCQGDKQSdKx1MBtPh
RGvhV+8+8D57vkpDHfleaxDPtR2U8HGevl54IUA4Ls75qyX8FyGMPJmx7ygc
0CX85rR4MwniRBuh4O2oORsq4eNnyr055qeq1tOnifr2JXqRs25pqdD8P6GJ
kEJoeYXl6dKEIE78UM5p84/SrwSR5JQgK+1n6jilpUQnTZepEvZwJbyBOxzk
D5caGqlCzi84cadgXufxkMtX67WLF58DiF7chUsQc73BDZmvMUKeZP+SzOMf
2niuUawaMWxjp6PKcOmSbZwaGoSE82fZ4vBnyYs9jVIIjeXcd7jENEqSdFqY
hkovdjj/YRVcKgj6RHk596qPXL5sWxT4x8SJqP/OY5Z6d5ETsxnqWUDOaIrd
tB+h5twjaEBNwx9c7DxwvMl6d69wyh0R0ahRAdf8179QrK3ATleRGzUZV7Yc
0E6g/9po7iZwoq35e4tv3IfGlydka/MLbjqqrFevEr39doDohAmiYqeLyHr6
ywvuPe6rMtMjEaMbaC9n2DvcC1t4DTjCd9w0/VIjoI4Dd9w4rTZC5VGRC/G6
PjXP0+BYby8pqzQnfirndKzkztU9B32oNn+hroOm7dgx8ySkOguziNDx8wKd
GqoJ2hxmwfkfMYa2N7gsxnQlHZtcqXcS7ekvNbV19ah0UIHLFOYGMbPzc8Gq
FwRPDplL4S9NoEHvr8Ldz4NentEHnjvn/MlS+oP5aOn74afjAw9Rz4hxbquG
pp/Dwflf4WrQW1Ppvvt/R7ffcTfd/2AncxvwG7V3Q4P1adi6f3+UFzGB85Bw
0dT4f/s9Hem5MUtp2ucpHJzoLTgZr9lx02uCX4u6admygHtxBrxeUuL+vPWI
n3RTP0Z9r/Ud9MzQj41WO4punx82NehcKN3W5k7qO2qmSQKwLvJIc9E7KmLc
tlxcbtLmnUUm6vcXnfbe8SNzVVRdNhkXjvBj5T4y0nCe0qkThd91F3Vo3969
YnnYcv974f4CD5ObN2+mzp07u7cbsMNEngo+XLiZucffEJqyeFPQb2ih9Mij
XWhP3nEkCcrIOYrxw+UeTSV+HMLEjM4RwcgDm05anoLjOh9yk0ZEM0HDxdHU
mNmr0s3l4/KEPagdXh3hJgXX7ZhTcGd//kKNNnn10dHUrU2bb7S2dVTnt+Kv
V3iA3LVrF0UzWGZmJiUkJBiZvXco94v6t2zZ8rW91sIV0jw4w30yEhGEPVlW
qUyjScDlPIy7kuvMMm5XOZRVNbg9RIfiLeV6a+RVDZInuho01xiJX39vHt3e
5o6mrqhETo0nvb+JiYnhjnorN0/l7m2X527HeaCJNU1V6M9wIyVJwHCEm7ID
PBCzH5ifLjam59Nvf/+wctS9e3favXs3XE45ekA4auQCcejQIVq7di1t2LCB
zpw5I7cErb6m6fU8D6amplJUVJThvoRDk+F/ZzH+SjCe5C4iMTGRW6U4noqO
4IT8BOaXR7VZVVXFrfc+WrNmjYHJzc3lknrZZjkpB6Gc484ZfmJjYykvLw+Y
nLESdHCqo9wiJicnGw/iYONEddYMOlZORwsYiMHBcCw7OxuO54wXRs6fP2+c
EMaBKqAWJmP9zrlPMFzlEpiTk2MwbN++3eh0omCora01loeiwCgoVFZWWstJ
WrdloKCggJvY1TzFbQKrzmQRBb0Wm4DS0tKMmiHyiRMnOA83CAfOz4QYjh7j
khUfH28IHudeldemyvYlbob2799vDAu1IlnW1NRYJqQ+hlJZWRklJSUZ5YNz
0JnhUSsIgEdIm56ebs5DrRbebdCRy1BzoDoop66uzpkrjLCPG+fDHvyimIdy
3reg0oCEuhkAQsNAzL6zQDDwaSoqKjJZAViysrJ4UKuwoM5PbhJbSzxqLeUO
FwaDKMi7EA3i2tQtHWAgIKAzDYjlHp2CBHSKN77zmm1V3Vm5qQCIFEaYoDEU
GICoYAj2hk7vsGjuvInHxwojLL3RAvagFdNV1NbebjH88CbOHufRKSwBRqEx
6J4DwF7/yNDStLMneHQKd4BbYB9ugihEof6+RaM3Vk15fLKmLu5j4ZvQFrwQ
qaC6utpOo4LoG519i/ACoyFMoC+oFVnqNI9QUKsdzp1bhRckhmCPTxMqHKu0
Z88eo1ZQAc+sVnt37nz3Js6eKRjghPAeeA72d+7cCUPaew5H75GbcvYss9fc
8KdG43V7B+7or5Xq5OBQnTxHkiuyj7xkmpb7laadusge0f8lgEsFe22x94gl
kJ+fbzzxsN262T8oyYfzrf8Ag2NWVw==\
\>"]]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "r1", "-", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 4}, {3, 4}, {5, 6}, {4, 5}, {3, 
            6}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAA/E05uD4zDUB2tkrqZ1T0PwXH685H
5QVAixDEe8vb7z+a/Nlh/7/9P1r9omt6MvQ/s2cDn4OaAUDgdpS4p2jOPw5+
YsIIyfI/AAAAAAAAAADJyEoVCKPrP6n+tjdxFvA/AAAAAAAAAAATRd9v/T33
P+pCa5p/zgBAVDqwF0hXxj/RfXYcIpoAQKynG1v0YMQ/+n5ZtVFkAEA0K19y
fZPCP+7U1EEuLQBA0Pn2lvTvwD9eMM/OsOn/P9D04F2k7r4/eTP2AeN2/z+Q
ka9z6lS8P+5PRhY3Av8/wHdemUUUuj/BNv0f8ov+P8g2fjoLLrg/n3y7IVn6
9z8AWEf8EtdCv2BwRifBgvc/gMQYZ+IFe7+qXcz3HAr3P8BU1sTQ9Ia/JxtB
AbSQ9j8A4LJPeH2Nv7ajGibOFvY/ALCRvYSMkL8YVcGSs5z1P2Dju7mw4pC/
Q2bUkqwi9T+AuBIuGoKPv3ziW2YBqfQ/wETUpYpQir98tQZNh6oAQJbdeDHY
BtA/mHPKtrFtAEA10bqtZDDQPy380zauMABApvnl535C0D+8qU7aQef/P3BW
IygcPdA/TDIo/1tt/z9v0seePyDQP8nvnAjz8/4/KtzhxPTXzz8S3SLZTnv+
PwA9zcvWQM8/09Ct3rYD/j82vpDLfnvOP7MWbOAdcvc/eFtVG6JRwj+G/SLq
2Pv2P/w65euEXsE/+hlz/iyH9j8Urrx+Mj7APxUdmjFfFPY/7PhHE6vivT+Y
o798s6P1PxD6OkNm8bo/fk+2lWw19T9Ml2qMVKq3P9NRfMfLyfQ/YJ7xumYP
tD+gx5LLEGH0Pwh5yEG/IrA/vb47Nw==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0347263707657203]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0347263707657203]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.7263294841374197`, -0.04999193827856571}, 5}], 
                 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.6481974949339981`, 0.2875607603066587}, 5}], 
                 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0347263707657203]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0347263707657203]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0694527415314406, 0.0694527415314406}], 
              {GrayLevel[0], DiskBox[2, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[3, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[4, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[5, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[6, 0.04586111767247925]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.0694527415314406, 0.0694527415314406}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "r1", "-", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 3}, {4, 3}, {5, 6}, {3, 5}, {4, 
            6}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAwKGOZlUzDUD4HFcBpFjHP9WdSE8r
5QVAPSFs0qFC3T8p4Xv8dpoBQLL6QvsIcvM/RtLd+Wq//T/mVvX7ynHIPyiW
wMvTyPI/9ATbc4RA9z+gGh+qH6HrP0srvLIRqdw/AAAAAAAAAAAAAAAAAAAA
AClN15NyqgBAAhPovIw98z9XpcQOm20AQLKpZgczM/M/EZ522pUwAECyeTYP
ti7zPywt5y8O5/8/uhWgfBgw8z9jkdTbJW3/P1zsyX1ZN/M/YoSU5rrz/j+S
hzTHdETzPzliyzMVe/4/yIpFlmJX8z8SU1IyfAP+P+Lt37UXcPM/aEE8INhx
9z+Wi+my4vX0P+jr1Z+S+/Y/rFoQglAU9T/6WEGn5ob2P3wMkPxkOPU/6tvD
ShkU9j/uTqnFCmL1PyOLQoNuo/U/o/i2NCmR9T8WgQMGKTX1P4+ix2OkxfU/
NrkuHYrJ9D+URCJAXf/1P9PrJYHRYPQ/dg2snDE+9j9UNskheM4AQDLycdJb
dPQ/os/E0xuaAEAWu/suMLP0P7NrWl9MZABAGl1WC+ns9D+s5rqgKS0AQAYH
ZzpkIfU/kXz0eajp/z+6sHSpglD1P4D/dh3bdv8/LPONcih69T+SbOIkLwL/
P/ykDe08nvU/Ehd8pOmL/j8SdDS8qrz1P2gFZpJF+vc/xhE+uXVC9z9A9uyQ
rIL3P+F02NgqW/c/GNQj3gYK9z8WeOmnGG73PxbH4+ibkPY/TBNU8TN79z9N
K9GUsxb2P+7pffJ0gvc/VhzLD5ac9T/1hedf14P3P8wNL6eLIvU/9lW3Z1p/
9z8nvgmd3Kj0P6bsNbIAdfc/zCE0hA==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03472665818833397]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03472665818833397]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.6481039685042613`, 1.165515264345253}, 5}], 
                 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.7263484016032933`, 1.5030766489276046`}, 5}], 
                 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03472665818833397]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03472665818833397]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03472665818833397]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06945331637666793, 0.06945331637666793}], 
              {GrayLevel[0], DiskBox[2, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[3, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[4, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[5, 0.045857674204805216`]}, 
              {GrayLevel[0], DiskBox[6, 0.045857674204805216`]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06945331637666793, 0.06945331637666793}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1", "-", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 5}, {6, 7}, {3, 7}, {4, 5}, {5, 
            6}, {4, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.82775950748957, 0.34102566225772063`}, {
            2.857279458924138, 0.3642453300352074}, {1.9008765210392695`, 
            0.}, {2.4882573319258685`, 1.450220578471928}, {
            1.9176099356458205`, 1.0859896894096428`}, {1.3410753682205372`, 
            1.4692240547736586`}, {0.9581671951634747, 0.4009557497035342}, {
            0., 0.32507801395474695`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358952851820334]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358952851820334]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0358952851820334]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358952851820334]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0717905703640668, 0.0717905703640668}], 
              {GrayLevel[0], DiskBox[2, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[3, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[4, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[5, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[6, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[7, 0.033849024780202346`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0717905703640668, 0.0717905703640668}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1", "-", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 5, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {4, 5}, {6, 7}, {4, 7}, {3, 5}, {5, 
            6}, {3, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
                Thickness[Large]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.374022807750645, 1.877601037910519}, {
            2.6200116887705454`, 1.2861863965154487`}, {2.317945018414745, 
            0.31556980190522654`}, {1.6954182340426815`, 
            1.3509692689724844`}, {1.6888887135352744`, 0.}, {
            1.0670885557293825`, 0.32525130093028776`}, {0.7660058071489272, 
            1.2967119800808657`}, {0., 1.8657165026996045`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03286298885040069]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03286298885040069]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03286298885040069]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03286298885040069]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03286298885040069]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06572597770080138, 0.06572597770080138}], 
              {GrayLevel[0], DiskBox[2, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[3, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[4, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[5, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[6, 0.035086459104769915`]}, 
              {GrayLevel[0], DiskBox[7, 0.035086459104769915`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06572597770080138, 0.06572597770080138}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"r1", "+", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "r1", "+", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 5, 6, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 6}, {7, 5}, {3, 7}, {3, 6}, {6, 
            4}, {4, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 5] -> {
                Thickness[Large]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274196709203682, 0.5136246102169043}, {
            3.3670820276555293`, 0.5053864523625655}, {2.530169512788247, 
            0.}, {2.5304984937516943`, 1.0131836739104354`}, {
            0.9071271193738002, 0.5077856682990879}, {1.743514219500912, 
            0.0006579432279606445}, {1.7442270282497168`, 
            1.0137999850147033`}, {0., 0.5122562334524448}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.038729380658446846`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.038729380658446846`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.038729380658446846`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.038729380658446846`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.038729380658446846`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07745876131689369, 0.07745876131689369}], 
              {GrayLevel[0], DiskBox[2, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[3, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[4, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[5, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[6, 0.03931358535233097]}, 
              {GrayLevel[0], DiskBox[7, 0.03931358535233097]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07745876131689369, 0.07745876131689369}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"r2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"r1", "+", "r2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 5, 6, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {5, 6}, {7, 5}, {4, 3}, {3, 7}, {4, 
            6}, {6, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598366806140282, 0.5457743356918155}, {
            3.633548263833968, 0.5250650316016542}, {2.8217992036599027`, 
            0.}, {2.821231543079123, 1.0517221476263738`}, {
            0.9648178152185232, 0.5256291579082695}, {1.7773076412298578`, 
            1.0519135692253971`}, {1.7764749771256565`, 
            0.0005493950476227472}, {0., 0.5453390276145662}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070062802933476]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070062802933476]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070062802933476]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.04070062802933476]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070062802933476]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140125605866952, 0.08140125605866952}], 
              {GrayLevel[0], DiskBox[2, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[3, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[4, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[5, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[6, 0.04825095583319784]}, 
              {GrayLevel[0], DiskBox[7, 0.04825095583319784]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140125605866952, 0.08140125605866952}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {3, 5}, {2, 4}, {5, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.814781421808071, 0.4877084622328541}, {
            2.825162389780683, 0.453296561078493}, {1.9077379198057707`, 
            0.913800986380224}, {1.9070977558442062`, 0.}, {
            0.9896915982171941, 0.4559929121833412}, {0., 
            0.48656376120221734`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.035810735981566255`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.035810735981566255`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.035810735981566255`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.035810735981566255`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.035810735981566255`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.035810735981566255`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07162147196313251, 0.07162147196313251}], 
              {GrayLevel[0], DiskBox[2, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[3, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[4, 0.04569006053069333]}, 
              {GrayLevel[0], DiskBox[5, 0.04569006053069333]}, InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.07162147196313251, 0.07162147196313251}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None,
       ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", "l1", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 5, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {6, 5}, {2, 4}, {3, 6}, {3, 
            6}, {5, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdi3tIUwEche/svTArn8s/UoeVyy1Dhpnl72YyFRJmRTNWJrKBUzbdnIWa
ZlaOfCydFEOm+YKJD9CtB0UrCZ1rpk1yFrQRJptECeJ9mY/d1r8dOBz4Pk50
QclFaQCCIEf9/beS+5IGQWIgygxuNM2LJkHwpFPA9uxAXfYXus40N+xzLKjl
7QxUn7HK8CkxGBbPZUvebYFJ07WrWWmDzRmTld/9C5haVw1tdAHVEU9+s61C
sJqh6UEpQP5LDBb0WjaAoIe9+yM1ZzHIjoiQ6uQIGlPPtfASMRCfYcmaExBU
Xfc43MzDIP7TyvSWl4YQ3DYbloDBhuw2M0xOgxfiWFeTMNCN5z0fxHzgtZd3
l6f7fZF7dGeVD7RFIqNEhEGXRaZQbPPBUNAf/FApBvlXuKhTSYFbI71OEjh8
r2A1tWyQcKY4PEcaQkBZ9aDcU0/C5vSap/IUAYMNxrWIEBL64nvfRhcQMDwg
iy3p9fM2x8lzrQQk5Br7OXwCsvRErmWCAJ5pZarpAw7pQmvULZoA/ejuDGc+
Dl9rDeILqSSkWss+ls4g6JfTIu5lNQ7YBL0ntB1Bi4q5hXFmHFZVfZECCYJK
2L4Ti79xMGedd7O5CLr9GFt0I5qAQnPmUsdPGpiPhPYaIQHdY+mq2Zs0jCyv
98RWEsAfn790JIAGldCVlNJJwLOl908XGn3gSNQeH3hDgKq2Xpm5TgFjaL3t
QRoF4Z8P8kbqKIjKMSykiCm4W9HK2cukwLZSvcRR+P1cmbOrhQRblb0ko4oC
j4I+8CqMBP0k5+VD/+/ahCXQYCAgOTlPuXyPgsrAsVBdDAFNfOuU8g4FDpGj
eLEfB63zh5xVTsFfp783Jg==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03472646987638081]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03472646987638081]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.6480923366178761`, 1.1655286418924933`}, 6}], 
                 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.7263496444684259`, 1.5030814874346081`}, 6}], 
                 0.03472646987638081]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03472646987638081]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03472646987638081]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06945293975276162, 0.06945293975276162}], 
              {GrayLevel[0], DiskBox[2, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[3, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[4, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[5, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[6, 0.04585734142274525]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06945293975276162, 0.06945293975276162}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 2, 4, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {3, 5}, {2, 6}, {6, 4}, {6, 
            4}, {5, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAAAAAAAAAAACwV1o0nlP3P7IYRDLD
j+s/Gyo9c/EW8D+hWiEb6Lj9PylI2In/J/Q/IcCR5yWVAUDMcntc2R3OP0zs
itgE4gVARHwpXJC97z8wjewtU7ryPwAAAAAAAAAAgedR2SgqDUCi4TJB6m70
P51fxOC/yABA+DgtMjMTxj+8s+FCSpQAQFymUjmcHsQ/Gz8dGWNeAED4ynMf
9FLCP85qsU0qJwBA+LgHDEuxwD9RxYAlgd3/PxgVFJIwdb4/qJDwnY5q/z/I
KyFhc9+7P78AmRPB9f4/sGjhUuaiuT/InZ+uXX/+P6DDg2bcwLc/jzwjbi3s
9z8AQFkRLMFVv+8B+kZ6dPc/AFJGLOHWfb+my948vvv2PwDzsxx+Ooi/jg/l
y0CC9j8AycQfy5+Ov8m6ouJJCPY/4NV+GscLkb+jQpm3IY71P8Ay4N3cT5G/
SiB0nhAU9T/gsS+K/huQvxoINt1emvQ/QBTaz8Thir+tAu0PIKUAQBAUean1
y88/lPZNL0doAECFtOCWrBDQP2hlu6I+KwBAlLwbfOoj0D+rUm0aVdz/P8Sm
5R+pH9A/5v0qMV5i/z+t3zsH6wPQP8xBMcDg6P4/+rFGPoGhzz+ECxa2JHD+
P1ql3WWQDM8/5NDsjnH4/T9KJZ60W0nOP6tvcE5BZfc/+pA5KWs9wj+zDHfp
3e72P3K+CjNmTME/ynwfXxB69j/o3OqrHy7APyJIj9cdB/Y/gNDiJoLGvT/X
N61hSpb1P5xz56Ac2bo/PY/Vytgn9T+gTw96ypW3P/ylTHcKvPQ/2JhRRnr+
sz86Toc7H1P0P6BznFRMFbA/Pw4tvA==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03469688878550463]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03469688878550463]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03469688878550463]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.7228484355785176`, -0.051293057415546905`}, 6}], 
                 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.6454672501955165`, 0.28657897866236537`}, 6}], 
                 0.03469688878550463]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03469688878550463]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06939377757100926, 0.06939377757100926}], 
              {GrayLevel[0], DiskBox[2, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[3, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[4, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[5, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[6, 0.04588423454946072]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06939377757100926, 0.06939377757100926}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {2, 5}, {2, 3}, {5, 6}, {4, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598716580880777, 0.5111844757682964}, {
            3.633781226626953, 0.5248792070768915}, {2.820926459245353, 0.}, {
            1.7768649313204117`, 0.0017248097633325576`}, {
            2.8224137730810934`, 1.0514880756581495`}, {1.777143959522975, 
            1.05308695777259}, {0.9648670034137008, 0.5287805535604434}, {0., 
            0.5085688457319367}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070271725560909]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.04070271725560909]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070271725560909]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070271725560909]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070271725560909]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140543451121818, 0.08140543451121818}], 
              {GrayLevel[0], DiskBox[2, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[3, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[4, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[5, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[6, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[7, 0.048251626489543516`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140543451121818, 0.08140543451121818}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 2, 6, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 6}, {7, 6}, {2, 4}, {5, 7}, {5, 
            3}, {3, 7}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.419833699470418, 1.8208204045965677`}, {
            2.636945124344794, 1.2764098748254205`}, {2.334130367630822, 
            0.3079455830245873}, {1.7076096131859468`, 1.3637847989951255`}, {
            1.7119561078649905`, 0.}, {0.7777370382325235, 
            1.269212434284803}, {1.0875891149617871`, 0.3045481913998621}, {
            0., 1.822743944098868}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03317643529084821]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.03317643529084821]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03317643529084821]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03317643529084821]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 8}, 0.03317643529084821]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06635287058169642, 0.06635287058169642}], 
              {GrayLevel[0], DiskBox[2, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[3, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[4, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[5, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[6, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[7, 0.0347106356762417]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06635287058169642, 0.06635287058169642}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {2, 6}, {2, 3}, {4, 5}, {5, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.893775166347438, 1.058545006665348}, {
            2.915694528247547, 1.0756291081466522`}, {2.5068717703198495`, 
            0.007184978087345639}, {1.3899492395455988`, 0.}, {
            1.9430078188089612`, 0.44168242913944494`}, {1.9379763628633724`, 
            1.4588714772995928`}, {0.9748993901709971, 1.0664818021657314`}, {
            0., 1.0609833747799051`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03632343081060238]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03632343081060238]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03632343081060238]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07264686162120476, 0.07264686162120476}], 
              {GrayLevel[0], DiskBox[2, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[3, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[4, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[5, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[6, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[7, 0.03538916360650989]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07264686162120476, 0.07264686162120476}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 3}, {3, 6}, {2, 4}, {5, 6}, {6, 
            7}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598696981543726, 0.521344888717441}, {
            3.633713364196972, 0.5235056604646946}, {2.822648355334171, 
            1.0516008511869326`}, {2.820674110126016, 0.}, {1.77636106386888, 
            0.0029744306467108794`}, {1.7776543428734137`, 
            1.0544039818428559`}, {0.964921633754265, 0.5308134530541693}, {
            0., 0.5165385052401519}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070260018966251]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.04070260018966251]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.04070260018966251]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070260018966251]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070260018966251]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140520037932503, 0.08140520037932503}], 
              {GrayLevel[0], DiskBox[2, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[3, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[4, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[5, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[6, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[7, 0.04824930182609304]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140520037932503, 0.08140520037932503}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 2, 5, 4, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {6, 3}, {3, 5}, {2, 7}, {7, 6}, {4, 
            6}, {7, 4}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.3599246399282907`, 1.881029516389429}, {
            2.626345691904735, 1.2620069147053299`}, {1.6919525138296148`, 
            1.354835865971224}, {1.0589154244234025`, 0.31939702102541234`}, {
            0.7604519880339898, 1.287040548571826}, {1.679325626871987, 0.}, {
            2.3078762131568675`, 0.30680165584071795`}, {0., 
            1.8660106823856832`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03276618788109682]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 7}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03276618788109682]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03276618788109682]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03276618788109682]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03276618788109682]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06553237576219365, 0.06553237576219365}], 
              {GrayLevel[0], DiskBox[2, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[3, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[4, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[5, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[6, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[7, 0.034889943978967684`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06553237576219365, 0.06553237576219365}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {2, 5}, {3, 6}, {2, 3}, {4, 5}, {5, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274355433643473, 0.49794596333510094`}, {
            3.367218488780101, 0.5062556253957285}, {2.530062777037424, 0.}, {
            1.7439065021935352`, 0.00047664907637468}, {2.5307680447593563`, 
            1.0130617528447177`}, {1.7439509204073738`, 
            1.0135216012509871`}, {0.9071226324241428, 0.5077858595970015}, {
            0., 0.49707056543331124`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036324537664]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036324537664]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036324537664]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07746072649075328, 0.07746072649075328}], 
              {GrayLevel[0], DiskBox[2, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[3, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[4, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[5, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[6, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[7, 0.03930782096704047]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07746072649075328, 0.07746072649075328}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {2, 5}, {3, 6}, {2, 3}, {5, 4}, {4, 6}, {6, 
            7}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.333473537124406, 1.9341571361054695`}, {
            2.6014431696053224`, 1.3201914931841159`}, {2.2941142165255872`, 
            0.34844678885844815`}, {1.6617920929561945`, 0.}, {
            1.697397851815909, 1.3601012411757878`}, {1.0553594643335766`, 
            0.3733174362865064}, {0.7615739842896245, 1.3499586448882352`}, {
            0., 1.9146693310666216`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03258413672108591]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03258413672108591]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03258413672108591]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03258413672108591]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03258413672108591]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06516827344217183, 0.06516827344217183}], 
              {GrayLevel[0], DiskBox[2, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[3, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[4, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[5, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[6, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[7, 0.03560640536805378]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06516827344217183, 0.06516827344217183}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 6}, {2, 4}, {3, 5}, {5, 6}, {6, 
            7}, {3, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.82775950748957, 0.34102566225772063`}, {
            2.857279458924138, 0.3642453300352074}, {1.9008765210392695`, 
            0.}, {2.4882573319258685`, 1.450220578471928}, {
            1.9176099356458205`, 1.0859896894096428`}, {1.3410753682205372`, 
            1.4692240547736586`}, {0.9581671951634747, 0.4009557497035342}, {
            0., 0.32507801395474695`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358952851820334]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358952851820334]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358952851820334]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0358952851820334]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358952851820334]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0717905703640668, 0.0717905703640668}], 
              {GrayLevel[0], DiskBox[2, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[3, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[4, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[5, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[6, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[7, 0.033849024780202346`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0717905703640668, 0.0717905703640668}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 6}, {2, 4}, {5, 3}, {3, 6}, {6, 
            7}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.2742494026600495`, 0.5081113472152592}, {
            3.3671062154547293`, 0.5037619324437881}, {2.5295135910223037`, 
            0.}, {2.5311983977440184`, 1.01321578245788}, {
            1.7431352717894992`, 0.001849348364290515}, {1.7446354037863188`, 
            1.0149037074850855`}, {0.9071260127032055, 0.5106663320150644}, {
            0., 0.5042416907204542}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0387297068603034]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.0387297068603034]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0387297068603034]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0387297068603034]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0387297068603034]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0774594137206068, 0.0774594137206068}], 
              {GrayLevel[0], DiskBox[2, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[3, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[4, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[5, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[6, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[7, 0.03931902469062473]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0774594137206068, 0.0774594137206068}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", "r1", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2", ",", "k3"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 4, 2, 5, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {6, 5}, {2, 4}, {3, 6}, {3, 
            6}, {5, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdi3tIUwEche/svTArn8s/UoeVyy1Dhpnl72YyFRJmRTNWJrKBUzbdnIWa
ZlaOfCydFEOm+YKJD9CtB0UrCZ1rpk1yFrQRJptECeJ9mY/d1r8dOBz4Pk50
QclFaQCCIEf9/beS+5IGQWIgygxuNM2LJkHwpFPA9uxAXfYXus40N+xzLKjl
7QxUn7HK8CkxGBbPZUvebYFJ07WrWWmDzRmTld/9C5haVw1tdAHVEU9+s61C
sJqh6UEpQP5LDBb0WjaAoIe9+yM1ZzHIjoiQ6uQIGlPPtfASMRCfYcmaExBU
Xfc43MzDIP7TyvSWl4YQ3DYbloDBhuw2M0xOgxfiWFeTMNCN5z0fxHzgtZd3
l6f7fZF7dGeVD7RFIqNEhEGXRaZQbPPBUNAf/FApBvlXuKhTSYFbI71OEjh8
r2A1tWyQcKY4PEcaQkBZ9aDcU0/C5vSap/IUAYMNxrWIEBL64nvfRhcQMDwg
iy3p9fM2x8lzrQQk5Br7OXwCsvRErmWCAJ5pZarpAw7pQmvULZoA/ejuDGc+
Dl9rDeILqSSkWss+ls4g6JfTIu5lNQ7YBL0ntB1Bi4q5hXFmHFZVfZECCYJK
2L4Ti79xMGedd7O5CLr9GFt0I5qAQnPmUsdPGpiPhPYaIQHdY+mq2Zs0jCyv
98RWEsAfn790JIAGldCVlNJJwLOl908XGn3gSNQeH3hDgKq2Xpm5TgFjaL3t
QRoF4Z8P8kbqKIjKMSykiCm4W9HK2cukwLZSvcRR+P1cmbOrhQRblb0ko4oC
j4I+8CqMBP0k5+VD/+/ahCXQYCAgOTlPuXyPgsrAsVBdDAFNfOuU8g4FDpGj
eLEfB63zh5xVTsFfp783Jg==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03472646987638081]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03472646987638081]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.6480923366178761`, 1.1655286418924933`}, 6}], 
                 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.7263496444684259`, 1.5030814874346081`}, 6}], 
                 0.03472646987638081]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03472646987638081]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03472646987638081]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03472646987638081]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06945293975276162, 0.06945293975276162}], 
              {GrayLevel[0], DiskBox[2, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[3, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[4, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[5, 0.04585734142274525]}, 
              {GrayLevel[0], DiskBox[6, 0.04585734142274525]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06945293975276162, 0.06945293975276162}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 2, 4, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {3, 5}, {2, 6}, {6, 4}, {6, 
            4}, {5, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAAAAAAAAAAACwV1o0nlP3P7IYRDLD
j+s/Gyo9c/EW8D+hWiEb6Lj9PylI2In/J/Q/IcCR5yWVAUDMcntc2R3OP0zs
itgE4gVARHwpXJC97z8wjewtU7ryPwAAAAAAAAAAgedR2SgqDUCi4TJB6m70
P51fxOC/yABA+DgtMjMTxj+8s+FCSpQAQFymUjmcHsQ/Gz8dGWNeAED4ynMf
9FLCP85qsU0qJwBA+LgHDEuxwD9RxYAlgd3/PxgVFJIwdb4/qJDwnY5q/z/I
KyFhc9+7P78AmRPB9f4/sGjhUuaiuT/InZ+uXX/+P6DDg2bcwLc/jzwjbi3s
9z8AQFkRLMFVv+8B+kZ6dPc/AFJGLOHWfb+my948vvv2PwDzsxx+Ooi/jg/l
y0CC9j8AycQfy5+Ov8m6ouJJCPY/4NV+GscLkb+jQpm3IY71P8Ay4N3cT5G/
SiB0nhAU9T/gsS+K/huQvxoINt1emvQ/QBTaz8Thir+tAu0PIKUAQBAUean1
y88/lPZNL0doAECFtOCWrBDQP2hlu6I+KwBAlLwbfOoj0D+rUm0aVdz/P8Sm
5R+pH9A/5v0qMV5i/z+t3zsH6wPQP8xBMcDg6P4/+rFGPoGhzz+ECxa2JHD+
P1ql3WWQDM8/5NDsjnH4/T9KJZ60W0nOP6tvcE5BZfc/+pA5KWs9wj+zDHfp
3e72P3K+CjNmTME/ynwfXxB69j/o3OqrHy7APyJIj9cdB/Y/gNDiJoLGvT/X
N61hSpb1P5xz56Ac2bo/PY/Vytgn9T+gTw96ypW3P/ylTHcKvPQ/2JhRRnr+
sz86Toc7H1P0P6BznFRMFbA/Pw4tvA==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03469688878550463]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03469688878550463]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03469688878550463]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.7228484355785176`, -0.051293057415546905`}, 6}], 
                 0.03469688878550463]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.6454672501955165`, 0.28657897866236537`}, 6}], 
                 0.03469688878550463]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03469688878550463]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06939377757100926, 0.06939377757100926}], 
              {GrayLevel[0], DiskBox[2, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[3, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[4, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[5, 0.04588423454946072]}, 
              {GrayLevel[0], DiskBox[6, 0.04588423454946072]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06939377757100926, 0.06939377757100926}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {2, 5}, {2, 3}, {5, 6}, {4, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598716580880777, 0.5111844757682964}, {
            3.633781226626953, 0.5248792070768915}, {2.820926459245353, 0.}, {
            1.7768649313204117`, 0.0017248097633325576`}, {
            2.8224137730810934`, 1.0514880756581495`}, {1.777143959522975, 
            1.05308695777259}, {0.9648670034137008, 0.5287805535604434}, {0., 
            0.5085688457319367}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070271725560909]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.04070271725560909]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070271725560909]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070271725560909]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070271725560909]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070271725560909]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140543451121818, 0.08140543451121818}], 
              {GrayLevel[0], DiskBox[2, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[3, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[4, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[5, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[6, 0.048251626489543516`]}, 
              {GrayLevel[0], DiskBox[7, 0.048251626489543516`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140543451121818, 0.08140543451121818}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 2, 6, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 6}, {7, 6}, {2, 4}, {5, 7}, {5, 
            3}, {3, 7}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.419833699470418, 1.8208204045965677`}, {
            2.636945124344794, 1.2764098748254205`}, {2.334130367630822, 
            0.3079455830245873}, {1.7076096131859468`, 1.3637847989951255`}, {
            1.7119561078649905`, 0.}, {0.7777370382325235, 
            1.269212434284803}, {1.0875891149617871`, 0.3045481913998621}, {
            0., 1.822743944098868}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03317643529084821]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.03317643529084821]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03317643529084821]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03317643529084821]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03317643529084821]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 8}, 0.03317643529084821]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06635287058169642, 0.06635287058169642}], 
              {GrayLevel[0], DiskBox[2, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[3, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[4, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[5, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[6, 0.0347106356762417]}, 
              {GrayLevel[0], DiskBox[7, 0.0347106356762417]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06635287058169642, 0.06635287058169642}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {2, 6}, {2, 3}, {4, 5}, {5, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.893775166347438, 1.058545006665348}, {
            2.915694528247547, 1.0756291081466522`}, {2.5068717703198495`, 
            0.007184978087345639}, {1.3899492395455988`, 0.}, {
            1.9430078188089612`, 0.44168242913944494`}, {1.9379763628633724`, 
            1.4588714772995928`}, {0.9748993901709971, 1.0664818021657314`}, {
            0., 1.0609833747799051`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03632343081060238]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03632343081060238]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03632343081060238]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07264686162120476, 0.07264686162120476}], 
              {GrayLevel[0], DiskBox[2, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[3, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[4, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[5, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[6, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[7, 0.03538916360650989]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07264686162120476, 0.07264686162120476}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 3}, {3, 6}, {2, 4}, {5, 6}, {6, 
            7}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598696981543726, 0.521344888717441}, {
            3.633713364196972, 0.5235056604646946}, {2.822648355334171, 
            1.0516008511869326`}, {2.820674110126016, 0.}, {1.77636106386888, 
            0.0029744306467108794`}, {1.7776543428734137`, 
            1.0544039818428559`}, {0.964921633754265, 0.5308134530541693}, {
            0., 0.5165385052401519}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070260018966251]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.04070260018966251]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.04070260018966251]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070260018966251]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070260018966251]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070260018966251]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140520037932503, 0.08140520037932503}], 
              {GrayLevel[0], DiskBox[2, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[3, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[4, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[5, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[6, 0.04824930182609304]}, 
              {GrayLevel[0], DiskBox[7, 0.04824930182609304]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140520037932503, 0.08140520037932503}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 2, 5, 4, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {6, 3}, {3, 5}, {2, 7}, {7, 6}, {4, 
            6}, {7, 4}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.3599246399282907`, 1.881029516389429}, {
            2.626345691904735, 1.2620069147053299`}, {1.6919525138296148`, 
            1.354835865971224}, {1.0589154244234025`, 0.31939702102541234`}, {
            0.7604519880339898, 1.287040548571826}, {1.679325626871987, 0.}, {
            2.3078762131568675`, 0.30680165584071795`}, {0., 
            1.8660106823856832`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03276618788109682]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 7}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03276618788109682]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03276618788109682]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03276618788109682]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.03276618788109682]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03276618788109682]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06553237576219365, 0.06553237576219365}], 
              {GrayLevel[0], DiskBox[2, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[3, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[4, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[5, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[6, 0.034889943978967684`]}, 
              {GrayLevel[0], DiskBox[7, 0.034889943978967684`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06553237576219365, 0.06553237576219365}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {2, 5}, {3, 6}, {2, 3}, {4, 5}, {5, 6}, {6, 
            7}, {4, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274355433643473, 0.49794596333510094`}, {
            3.367218488780101, 0.5062556253957285}, {2.530062777037424, 0.}, {
            1.7439065021935352`, 0.00047664907637468}, {2.5307680447593563`, 
            1.0130617528447177`}, {1.7439509204073738`, 
            1.0135216012509871`}, {0.9071226324241428, 0.5077858595970015}, {
            0., 0.49707056543331124`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036324537664]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036324537664]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036324537664]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07746072649075328, 0.07746072649075328}], 
              {GrayLevel[0], DiskBox[2, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[3, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[4, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[5, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[6, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[7, 0.03930782096704047]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07746072649075328, 0.07746072649075328}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {2, 5}, {3, 6}, {2, 3}, {5, 4}, {4, 6}, {6, 
            7}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.333473537124406, 1.9341571361054695`}, {
            2.6014431696053224`, 1.3201914931841159`}, {2.2941142165255872`, 
            0.34844678885844815`}, {1.6617920929561945`, 0.}, {
            1.697397851815909, 1.3601012411757878`}, {1.0553594643335766`, 
            0.3733174362865064}, {0.7615739842896245, 1.3499586448882352`}, {
            0., 1.9146693310666216`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03258413672108591]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03258413672108591]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03258413672108591]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03258413672108591]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03258413672108591]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03258413672108591]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06516827344217183, 0.06516827344217183}], 
              {GrayLevel[0], DiskBox[2, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[3, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[4, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[5, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[6, 0.03560640536805378]}, 
              {GrayLevel[0], DiskBox[7, 0.03560640536805378]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06516827344217183, 0.06516827344217183}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 6}, {2, 4}, {3, 5}, {5, 6}, {6, 
            7}, {3, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.82775950748957, 0.34102566225772063`}, {
            2.857279458924138, 0.3642453300352074}, {1.9008765210392695`, 
            0.}, {2.4882573319258685`, 1.450220578471928}, {
            1.9176099356458205`, 1.0859896894096428`}, {1.3410753682205372`, 
            1.4692240547736586`}, {0.9581671951634747, 0.4009557497035342}, {
            0., 0.32507801395474695`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358952851820334]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358952851820334]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358952851820334]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358952851820334]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0358952851820334]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358952851820334]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0717905703640668, 0.0717905703640668}], 
              {GrayLevel[0], DiskBox[2, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[3, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[4, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[5, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[6, 0.033849024780202346`]}, 
              {GrayLevel[0], DiskBox[7, 0.033849024780202346`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0717905703640668, 0.0717905703640668}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"r1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "r1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "r1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 6, 4, 5, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 6}, {2, 4}, {5, 3}, {3, 6}, {6, 
            7}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.2742494026600495`, 0.5081113472152592}, {
            3.3671062154547293`, 0.5037619324437881}, {2.5295135910223037`, 
            0.}, {2.5311983977440184`, 1.01321578245788}, {
            1.7431352717894992`, 0.001849348364290515}, {1.7446354037863188`, 
            1.0149037074850855`}, {0.9071260127032055, 0.5106663320150644}, {
            0., 0.5042416907204542}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0387297068603034]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.0387297068603034]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0387297068603034]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.0387297068603034]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0387297068603034]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0387297068603034]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0774594137206068, 0.0774594137206068}], 
              {GrayLevel[0], DiskBox[2, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[3, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[4, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[5, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[6, 0.03931902469062473]}, 
              {GrayLevel[0], DiskBox[7, 0.03931902469062473]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0774594137206068, 0.0774594137206068}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  StyleBox[
   RowBox[{
    RowBox[{"L", "(", ")"}], " ", 
    RowBox[{"PS", "(", 
     RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4"}], ")"}]}],
   StripOnInput->False,
   FontSize->Larger,
   FontWeight->Bold], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 5}, {4, 3}, {5, 6}, {3, 6}, {2, 
            4}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAA/E05uD4zDUB2tkrqZ1T0PwXH685H
5QVAixDEe8vb7z+a/Nlh/7/9P1r9omt6MvQ/s2cDn4OaAUDgdpS4p2jOPw5+
YsIIyfI/AAAAAAAAAADJyEoVCKPrP6n+tjdxFvA/AAAAAAAAAAATRd9v/T33
P+pCa5p/zgBAVDqwF0hXxj/RfXYcIpoAQKynG1v0YMQ/+n5ZtVFkAEA0K19y
fZPCP+7U1EEuLQBA0Pn2lvTvwD9eMM/OsOn/P9D04F2k7r4/eTP2AeN2/z+Q
ka9z6lS8P+5PRhY3Av8/wHdemUUUuj/BNv0f8ov+P8g2fjoLLrg/n3y7IVn6
9z8AWEf8EtdCv2BwRifBgvc/gMQYZ+IFe7+qXcz3HAr3P8BU1sTQ9Ia/JxtB
AbSQ9j8A4LJPeH2Nv7ajGibOFvY/ALCRvYSMkL8YVcGSs5z1P2Dju7mw4pC/
Q2bUkqwi9T+AuBIuGoKPv3ziW2YBqfQ/wETUpYpQir98tQZNh6oAQJbdeDHY
BtA/mHPKtrFtAEA10bqtZDDQPy380zauMABApvnl535C0D+8qU7aQef/P3BW
IygcPdA/TDIo/1tt/z9v0seePyDQP8nvnAjz8/4/KtzhxPTXzz8S3SLZTnv+
PwA9zcvWQM8/09Ct3rYD/j82vpDLfnvOP7MWbOAdcvc/eFtVG6JRwj+G/SLq
2Pv2P/w65euEXsE/+hlz/iyH9j8Urrx+Mj7APxUdmjFfFPY/7PhHE6vivT+Y
o798s6P1PxD6OkNm8bo/fk+2lWw19T9Ml2qMVKq3P9NRfMfLyfQ/YJ7xumYP
tD+gx5LLEGH0Pwh5yEG/IrA/vb47Nw==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0347263707657203]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.0347263707657203]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.0347263707657203]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.7263294841374197`, -0.04999193827856571}, 5}], 
                 0.0347263707657203]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 4, {1.6481974949339981`, 0.2875607603066587}, 5}], 
                 0.0347263707657203]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0347263707657203]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.0347263707657203]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0694527415314406, 0.0694527415314406}], 
              {GrayLevel[0], DiskBox[2, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[3, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[4, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[5, 0.04586111767247925]}, 
              {GrayLevel[0], DiskBox[6, 0.04586111767247925]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.0694527415314406, 0.0694527415314406}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {6, 4}, {2, 5}, {6, 3}, {4, 7}, {2, 
            6}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.391204699197312, 1.8438947129965138`}, {
            2.6217725564957175`, 1.2811053633583065`}, {1.7012645945519833`, 
            0.}, {1.074202049265616, 0.30585220640588695`}, {
            1.6932997325655055`, 1.3639617706279485`}, {2.3202638767233874`, 
            0.3159755466814528}, {0.7622649356690391, 1.270322347714335}, {0.,
             1.847211895852448}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.032980747396119475`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.032980747396119475`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.032980747396119475`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.032980747396119475`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.032980747396119475`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.032980747396119475`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.032980747396119475`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.032980747396119475`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.032980747396119475`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.032980747396119475`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06596149479223895, 0.06596149479223895}], 
              {GrayLevel[0], DiskBox[2, 0.03474912435693008]}, 
              {GrayLevel[0], DiskBox[3, 0.03474912435693008]}, 
              {GrayLevel[0], DiskBox[4, 0.03474912435693008]}, 
              {GrayLevel[0], DiskBox[5, 0.03474912435693008]}, 
              {GrayLevel[0], DiskBox[6, 0.03474912435693008]}, 
              {GrayLevel[0], DiskBox[7, 0.03474912435693008]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06596149479223895, 0.06596149479223895}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {2, 4}, {6, 7}, {6, 3}, {4, 5}, {5, 
            7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[1, 3] -> {
                Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.5032872287971495`, 0.0003346141677250669}, {
            2.6769991669233297`, 0.47072605311572774`}, {1.7515332825353969`, 
            1.7414595812870004`}, {2.3724813237658715`, 1.44443327890605}, {
            1.1305529226961912`, 1.4445737463534254`}, {1.7514371336353263`, 
            0.3680517900471344}, {0.8258396839749501, 0.47116929705690014`}, {
            0., 0.}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03374314900398022]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03374314900398022]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03374314900398022]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03374314900398022]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03374314900398022]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03374314900398022]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03374314900398022]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03374314900398022]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03374314900398022]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03374314900398022]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06748629800796044, 0.06748629800796044}], 
              {GrayLevel[0], DiskBox[2, 0.034415033280901945`]}, 
              {GrayLevel[0], DiskBox[3, 0.034415033280901945`]}, 
              {GrayLevel[0], DiskBox[4, 0.034415033280901945`]}, 
              {GrayLevel[0], DiskBox[5, 0.034415033280901945`]}, 
              {GrayLevel[0], DiskBox[6, 0.034415033280901945`]}, 
              {GrayLevel[0], DiskBox[7, 0.034415033280901945`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06748629800796044, 0.06748629800796044}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {3, 5}, {2, 5}, {6, 7}, {6, 3}, {5, 4}, {4, 
            7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{0., 0.041783056965116305`}, {
            0.8101025613207566, 0.5795952167671551}, {1.6815595845361242`, 
            1.8915139217192074`}, {2.3182283471807077`, 
            1.5995850699464622`}, {1.0734050654168052`, 
            1.5561298946839506`}, {1.7231623419423252`, 0.53388210916044}, {
            2.6542623631603717`, 0.6493455329211022}, {3.3674951147066268`, 
            0.}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03281818822513381]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03281818822513381]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03281818822513381]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03281818822513381]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03281818822513381]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03281818822513381]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03281818822513381]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03281818822513381]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03281818822513381]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03281818822513381]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06563637645026763, 0.06563637645026763}], 
              {GrayLevel[0], DiskBox[2, 0.03472514811759757]}, 
              {GrayLevel[0], DiskBox[3, 0.03472514811759757]}, 
              {GrayLevel[0], DiskBox[4, 0.03472514811759757]}, 
              {GrayLevel[0], DiskBox[5, 0.03472514811759757]}, 
              {GrayLevel[0], DiskBox[6, 0.03472514811759757]}, 
              {GrayLevel[0], DiskBox[7, 0.03472514811759757]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06563637645026763, 0.06563637645026763}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {6, 7}, {3, 5}, {6, 4}, {5, 
            7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.510061966473745, 0.000250684455576744}, {
            2.6801048896397273`, 0.4637499996366458}, {2.3757597393804795`, 
            1.438054776414318}, {1.7549325209020046`, 1.734250717792647}, {
            1.1341829702727326`, 1.4381476807487799`}, {1.7548449873859506`, 
            0.3602285070392836}, {0.8295773500174981, 0.46415477242340875`}, {
            0., 0.}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03378891349164677]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03378891349164677]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03378891349164677]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03378891349164677]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03378891349164677]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03378891349164677]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03378891349164677]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03378891349164677]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03378891349164677]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03378891349164677]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06757782698329354, 0.06757782698329354}], 
              {GrayLevel[0], DiskBox[2, 0.03438775265821557]}, 
              {GrayLevel[0], DiskBox[3, 0.03438775265821557]}, 
              {GrayLevel[0], DiskBox[4, 0.03438775265821557]}, 
              {GrayLevel[0], DiskBox[5, 0.03438775265821557]}, 
              {GrayLevel[0], DiskBox[6, 0.03438775265821557]}, 
              {GrayLevel[0], DiskBox[7, 0.03438775265821557]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06757782698329354, 0.06757782698329354}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 6, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 3}, {5, 6}, {4, 7}, {3, 6}, {6, 7}, {5, 
            4}, {2, 5}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274355433643473, 0.49794596333510094`}, {
            3.367218488780101, 0.5062556253957285}, {2.530062777037424, 0.}, {
            1.7439065021935352`, 0.00047664907637468}, {2.5307680447593563`, 
            1.0130617528447177`}, {1.7439509204073738`, 
            1.0135216012509871`}, {0.9071226324241428, 0.5077858595970015}, {
            0., 0.49707056543331124`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036324537664]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036324537664]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036324537664]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07746072649075328, 0.07746072649075328}], 
              {GrayLevel[0], DiskBox[2, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[3, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[4, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[5, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[6, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[7, 0.03930782096704047]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07746072649075328, 0.07746072649075328}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 5, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {5, 6}, {5, 4}, {2, 6}, {4, 7}, {3, 5}, {2, 
            3}, {6, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.893775166347438, 1.058545006665348}, {
            2.915694528247547, 1.0756291081466522`}, {2.5068717703198495`, 
            0.007184978087345639}, {1.3899492395455988`, 0.}, {
            1.9430078188089612`, 0.44168242913944494`}, {1.9379763628633724`, 
            1.4588714772995928`}, {0.9748993901709971, 1.0664818021657314`}, {
            0., 1.0609833747799051`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03632343081060238]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03632343081060238]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03632343081060238]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03632343081060238]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03632343081060238]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07264686162120476, 0.07264686162120476}], 
              {GrayLevel[0], DiskBox[2, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[3, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[4, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[5, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[6, 0.03538916360650989]}, 
              {GrayLevel[0], DiskBox[7, 0.03538916360650989]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07264686162120476, 0.07264686162120476}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 6, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {5, 6}, {3, 6}, {5, 4}, {2, 3}, {6, 7}, {4, 
            7}, {2, 5}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 5] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274355433643473, 0.49794596333510094`}, {
            3.367218488780101, 0.5062556253957285}, {2.530062777037424, 0.}, {
            1.7439065021935352`, 0.00047664907637468}, {2.5307680447593563`, 
            1.0130617528447177`}, {1.7439509204073738`, 
            1.0135216012509871`}, {0.9071226324241428, 0.5077858595970015}, {
            0., 0.49707056543331124`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036324537664]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.03873036324537664]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036324537664]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036324537664]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036324537664]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07746072649075328, 0.07746072649075328}], 
              {GrayLevel[0], DiskBox[2, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[3, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[4, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[5, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[6, 0.03930782096704047]}, 
              {GrayLevel[0], DiskBox[7, 0.03930782096704047]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07746072649075328, 0.07746072649075328}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {4, 3}, {6, 5}, {3, 7}, {6, 4}, {5, 
            7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.8146970080157248`, 0.270094964547705}, {
            2.8546429121838464`, 0.36272128098849865`}, {1.9019585089099569`, 
            0.}, {1.9076115834066782`, 1.1013769563057294`}, {
            1.3260564267646497`, 1.4479493200813462`}, {2.4907264540435166`, 
            1.4419671898819246`}, {0.9548673470545616, 0.3758312776987519}, {
            0., 0.26530661354951324`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03581018563525899]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03581018563525899]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.03581018563525899]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03581018563525899]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.03581018563525899]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03581018563525899]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03581018563525899]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03581018563525899]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03581018563525899]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03581018563525899]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07162037127051799, 0.07162037127051799}], 
              {GrayLevel[0], DiskBox[2, 0.0337647989615407]}, 
              {GrayLevel[0], DiskBox[3, 0.0337647989615407]}, 
              {GrayLevel[0], DiskBox[4, 0.0337647989615407]}, 
              {GrayLevel[0], DiskBox[5, 0.0337647989615407]}, 
              {GrayLevel[0], DiskBox[6, 0.0337647989615407]}, 
              {GrayLevel[0], DiskBox[7, 0.0337647989615407]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07162037127051799, 0.07162037127051799}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 3, 6, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {5, 6}, {2, 4}, {3, 7}, {4, 6}, {6, 7}, {5, 
            3}, {2, 5}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[5, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[1, 3] -> {
                Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.2743558218519055`, 0.4982697553927159}, {
            3.367216636566389, 0.5060420742650629}, {1.7438360815419336`, 
            0.0005509861483867784}, {2.5300152499060187`, 0.}, {
            2.5308131464576706`, 1.0130325055265477`}, {1.7440033387581189`, 
            1.013618635197326}, {0.9071208517169125, 0.5080258956640598}, {0.,
             0.49714741741886176`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036564856769]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03873036564856769]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036564856769]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036564856769]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03873036564856769]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.03873036564856769]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03873036564856769]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036564856769]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036564856769]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036564856769]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07746073129713538, 0.07746073129713538}], 
              {GrayLevel[0], DiskBox[2, 0.039308968072037334`]}, 
              {GrayLevel[0], DiskBox[3, 0.039308968072037334`]}, 
              {GrayLevel[0], DiskBox[4, 0.039308968072037334`]}, 
              {GrayLevel[0], DiskBox[5, 0.039308968072037334`]}, 
              {GrayLevel[0], DiskBox[6, 0.039308968072037334`]}, 
              {GrayLevel[0], DiskBox[7, 0.039308968072037334`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07746073129713538, 0.07746073129713538}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {6, 3}, {4, 7}, {3, 5}, {5, 7}, {6, 
            4}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 3] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.598522327325652, 0.5247790208320846}, {
            3.6335632241293476`, 0.5235496942757668}, {2.820645753909042, 
            0.}, {1.7776569213671092`, 1.0544959210609217`}, {
            1.776245356626343, 0.0028303389894658304`}, {2.8224211992699, 
            1.0515878913232055`}, {0.9649264735376399, 0.5306936622436798}, {
            0., 0.5200584649020372}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070155697690686]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070155697690686]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.04070155697690686]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.04070155697690686]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.04070155697690686]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.04070155697690686]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070155697690686]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.04070155697690686]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070155697690686]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070155697690686]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.08140311395381372, 0.08140311395381372}], 
              {GrayLevel[0], DiskBox[2, 0.048247994312851955`]}, 
              {GrayLevel[0], DiskBox[3, 0.048247994312851955`]}, 
              {GrayLevel[0], DiskBox[4, 0.048247994312851955`]}, 
              {GrayLevel[0], DiskBox[5, 0.048247994312851955`]}, 
              {GrayLevel[0], DiskBox[6, 0.048247994312851955`]}, 
              {GrayLevel[0], DiskBox[7, 0.048247994312851955`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.08140311395381372, 0.08140311395381372}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 6, 3, 5, 4, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {3, 4}, {5, 6}, {2, 6}, {5, 7}, {6, 4}, {4, 7}, {3, 
            5}, {2, 3}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[6, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{4.274283190916821, 0.49739354893378535`}, {
            3.367157851379004, 0.5058124693940612}, {2.5298516488140805`, 
            0.}, {1.743742504212162, 0.0008015201330868749}, {
            1.7440445027709608`, 1.0138428854165964`}, {2.5308946448538796`, 
            1.0131150143193053`}, {0.9070967461466173, 0.5085668274137312}, {
            0., 0.49584558482859165`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.038729916027433794`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.038729916027433794`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.038729916027433794`]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.038729916027433794`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.038729916027433794`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.038729916027433794`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.038729916027433794`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.038729916027433794`]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.038729916027433794`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.038729916027433794`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.07745983205486759, 0.07745983205486759}], 
              {GrayLevel[0], DiskBox[2, 0.039305477660921635`]}, 
              {GrayLevel[0], DiskBox[3, 0.039305477660921635`]}, 
              {GrayLevel[0], DiskBox[4, 0.039305477660921635`]}, 
              {GrayLevel[0], DiskBox[5, 0.039305477660921635`]}, 
              {GrayLevel[0], DiskBox[6, 0.039305477660921635`]}, 
              {GrayLevel[0], DiskBox[7, 0.039305477660921635`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.07745983205486759, 0.07745983205486759}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
            $CellContext`OO[1]}, {
           Null, {{1, 2}, {2, 3}, {4, 5}, {6, 5}, {4, 7}, {5, 3}, {3, 7}, {6, 
            4}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[6, 5] -> {
                Thickness[Large]}, UndirectedEdge[5, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
                Dashing[{Small, Small}]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {{
                 Thickness[Large], 
                 RGBColor[1, 0, 0], 
                 Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding",
             VertexShape -> {$CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {
             6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {6 -> {
                GrayLevel[0]}, 1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{3.823380060986201, 0.3202142547795689}, {
            2.8581060538573357`, 0.3716978129524196}, {1.902524331809892, 
            0.}, {1.342897087240333, 1.4612105976674619`}, {
            1.9083078849663895`, 1.0909562586778705`}, {2.4956078100275993`, 
            1.453584263786467}, {0.9597268034309243, 0.3911244194666886}, {0.,
             0.3178130764376545}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358667681098368]}}, 
              {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358667681098368]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.0358667681098368]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358667681098368]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358667681098368]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358667681098368]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358667681098368]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.0358667681098368]}}, 
              {Dashing[{Small, Small}], 
               {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358667681098368]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358667681098368]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0717335362196736, 0.0717335362196736}], 
              {GrayLevel[0], DiskBox[2, 0.033792663616746564`]}, 
              {GrayLevel[0], DiskBox[3, 0.033792663616746564`]}, 
              {GrayLevel[0], DiskBox[4, 0.033792663616746564`]}, 
              {GrayLevel[0], DiskBox[5, 0.033792663616746564`]}, 
              {GrayLevel[0], DiskBox[6, 0.033792663616746564`]}, 
              {GrayLevel[0], DiskBox[7, 0.033792663616746564`]}, InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0717335362196736, 0.0717335362196736}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None,
       ImageCache->GraphicsData["CompressedBitmap", "\<\
eJyNWQlYlVUa/vOqYI3zNNoy1TzltFhTatk085ipM9VY1uRkUz3VzOSoZaFm
NuWYuRBJapZlZZpm2pgFKiIqq4Agiws7CCIIyCIg+44sAt987/m/83Pv5YLy
PFz4z3/O++3L+e702Ys95i6cvXjea7NHPfnu7Hc85r323qgnFr3LS7arDMMo
5t9Zowz8T/yvfBQT//BLV38WqT826u7uJvmRNWOh+jOEysrKyNvbmyZPnkxr
166lCxcuOG5xow+WLSd3d3eQo2HDhqltVFpK1NMje4z56s9g6rzURSfSz9Hq
LSF0022jbYrFITT5kUeo+/nniTw8iPz8iBoajAXmSQ9hsbqumfaFp9G7n/ob
g+gNLx8+Nn7KDBoyZIgiDAa8Z83infTMM/j09bVEmmdCzRUmenp6KCO3lL7+
KZre/MjXptCupsXr99P3vodpjsdCumXU3TR52suUEJtA5O9vMjZ9OlFV1WBh
iFpbLQKvmwRek1dNre0UFn+aPvjyIJMDujt99kMkJWYW0aUuU9WnzpbRXM+f
8JbyiqsGadSiIgt1kJiAurqIZs4kWrGC6OhRos5OIWXMlmMF56tp+/7jNN97
t2HKY6O31/jRz0FJVFpZbwwW0UEVFIWqm1oeSnTyJNTW+/vKK0Tbt88xicwU
Iu2dlyguJZ+8t4SKWDby2hRM0Ylnqa29EziKSG5RZV8iLFp5OdHWrQCHaKaV
1q2zxJ1lUvuXbK+oaaTdocm0aK2fJdK8VbvpO794UDCGCTWmYE9NL7OSTGVB
aWlpltWys4lOn1ZP/zQJviInurt7KDW7hL7YeYTdwkfTfP+LAxQUk0mNzW3G
Na5p/kobaulSU33wlv37lRub6MZLQr+h+SIFHs2kJZ8HiFSDQYw2/BiliHf3
9Ax3TeRaLUFdHdGMGUqBlhrxjIjjHSYl40XZnlNYQVv3xkN1lhrf+WQf7QlL
YRU3GddeRqT6ejMiIRJEW7BAi/R3ocCmp6iEXPpwU7DlFuwiFJ+aTx2dXcYI
FxTOnKsYqeXhaKTMTKKUFMtNVRQsW0YUHU3U0fGcSXCGnDhfUU8/BSXSwjV7
LZHY8WkHB8C50hrjehcE2fFv1s6+bZujo2/ZQlRS8qxJZLoQQZwiXj/dEWFJ
tezLQ3T4WDY1t7YbNw5ERNIPwO2tFBxsObtJxvirbK9paKGAIxm0+LP9dm7h
Sxt/PkqZHLOcsm5yTfAWbSj4e2ys6e8vvEDU0mJYTm8mY/U0zSQ8TdCQC7Py
y2mTbyx56FxoQ5Yl/4h0qq5vNn7jmvCtGrypqW/GamszmbDL50+ahJ+QYy0X
Oyj8+Bla8XWgpd9Pvg+nExmF0L1xq2uqd2gjhoc7ZixEwIYNJrrxFyFSWFZL
Ow+epAUf77E8ZeHqvfTjoQQquVBn/FaIcIrsS8QGtnX27zXi5s2WuCYZQ2fR
3NxcKiws5DxXTrW1tWyBFrp06dJDgobadyytgNZ8F2ZJvHJjEEWezKHWtg7j
DtfMjNbMZGVBQqKCgl7DHjtGFBOjKsJ4teamanl7ezs1NzdzqqijyspKVcSL
i4uFUytRY2tISAgFBATQ7t27beTj42MMx79qCa8iIyMpLi6OkpKSKCMjg3Jy
cpSMQKypqVFUOpn6XcJRZW0Tl+lU+s+6fZYje6zypW/3xFF2wQV44V2u5Rwr
/NP8+b2hiXLB1PDDzBpmdI1UXtvR0aE0XM8Zqrq6Go0JZKT8/HzFZSbnlNTU
VEpMTKT4+HiuBUcpIiKCQkND6dChQ7Rv3z5Iyt4KqW20d+9eztn7tX9BT1er
P8MhoCIFo8K4UABInDp1SikG8EeOHLF0uWcP3K1XmYAN5uDXygRLUOaZM2fo
3LlzSpkQoYlDicUyRgoTui3rloblq112DYuN/rs+gA5GnaK6xlbjHtdKHaMd
parKKR1xZzd1KnlzxbLv7EzBRyiJL168SI2NjUpqKLekpERxCzc/zSU0PT1d
SXL8+HFOPLFKA2FhYRQYGKgkZoWauh2hlODPgQS9Q/9QBOxxjP0XEGlpacpe
UCrsBzuCJJTSwEEI5bNiTHtI6h2O2OImrNUySxFXDTAHsyQnJyv0qKgoRfHA
gQMOZvHl/KzNAq+AAAkJCUoomKWgoICzZqniAErgiJKo69sS7/SPsgJ6/f8i
KSmrmLrYW+8dyNF/6dihcG/pNWmSQxvt6emp2D98+DAFBQUpEfz8/EydjlQS
wIcPHjyopAjnjIjtcEZIkpKSojSRzR1PXl6e0ik4Rz5AXkDkIk8grswUb9wp
iu3ihhOKxTaYQSsWJtKKjebSDGuDK7a0g2JhanAMxc6ZM8dBquWeXqqyvb3W
j3yCk6i8qsEY60JRKadLHtCl+Px5mjB+vLorDFKf19DEiRNVdMMRz549q0Ly
PO+rqKhQ/oBQauMqxD5i3G9KZyZIdyUdfBuJA9uhGEDA9OPGjTPvJDZl3dTs
Ymrv0AiODCacKhovDrHk/aV8Dxnaew/x9jYeNolOkC0r2MhDhzpuGT8wqs+B
aHpwynN0822jacnSZTCeMypCEsgTJkxQ7oItA6KaNRZLH2w4YDxiok3UeuZ4
0qWtH5QHZJlTjmpbsRwQmS5Ab8vb8upGV4ciTuRYy8XldU6Hahta+x7q5Rd9
wiTzxCR5xUWTYpLyUA0sCzmh2C/b8esExK29KlK4VlwGKIJbFr3MHYQATRag
r/hGuSswUT2NGwjIRmu3HbbkmuyIghaI66dqg/pB0cs19S2WXAeOZDgBseuq
dic9p9QlUExynl4Ot5OL+2sBmiJAHIzK5Hga6wLoaNLZcbKVuxu1tHJjoBx3
aWE+MdaVDFEZV3YIbbgdv1dyyJ63IONP5ok/yyt0FHI1N+4TlMaWNnuUMbLM
PbG1zOXXCQgFAMzZA13kG5ILIIwI9DLHmRMQbom4Rl0WyEarRS7Pb4LkuIWC
sQmUiyFAPyh6uaquV65D0acE6FG7OPSPSFNP9w4MZC9XWWWDE1Bscp4aS/QH
FJ0oQDb6eGuoWsLN8lFHC9ufCI07fa9rGa7sUEhcL79cjq7kkHnDxZIX8/bY
gCd+J8tojfUyG1cOPS54fHFVjODpnoGBQmKzevmtbnQCQpeN2zBfsCwg7kb6
AtnJsDlYjlsoaDmXfB4Aaxlq+yBpcEJNHtW20QJfUeMgmUDpS5jf4VRa/0Nk
r2QDQQXbyXahutEJijOu8ma+Ll8Oykarvg1RMB9tDtE3tanyCuOyHrkQ390f
jr6nYBClWQqKyRSoRfbHgmOy+hzDzEgfYwhXx/ZHpjsds4FbdYS5F3b7P3On
kLpgV2+ZFTn2pCDibbcIe9floJB1erluEih9aceoB77FPmaquX8oG7zKkuQJ
R5Y2+cbQDwEnHATH1rkf/gwoS4f2jQR7hhMK+mwMOTFkMomavfmsNxbRmPt/
T+8tWa4uNa5k42B0QmMUhcaoFpqXl5fzZNsSj5OSQsI0VGYb0+QVAhSTVjzd
LmtortCm9TaxZospdzxrFFFW1WAxyflAkN9Sf9yV06KNeunVNx3ayaeen42Z
2u0CwqnPXlIBMSfq1yiHXrc9XEnrFxRrdZDLlq+gjTuDMQdSs7SmljZjlJb2
myCFxilZj5D0TB88wdYYYXItN2dAbvbdpIo3hCMGj1n55dZwh6uDfV6aZs+n
jeqbLsrUSt/IbWpICwVwhbdQ7KXlAiAoT8sJvI3kE3jSBvIUaVZvDTOectyO
BISpDJ5+LWtFZbVKYZiGjZA1pNm0M+fVk54FYjCpGeEi6ISsS3FhaY1xgzgY
ZkIYl6D7/N7/GCftdmteilYT69/4xCDh3SDLuPFrIgwpRPQgEWz6hiSrp+vk
BBIAt2Mqn2fkll6vQ4WV/+mOCNUhhsVnqzb6OoHBdqWg78KEf4sCF3o0rXYU
OC90dNJeTvMYY90oa7AcRsDb9h2zOHFUULYTMiIQCsHTVDnBTW2vtPXNcuJV
OYECiW8b8KSHdnjWXjNc1sAvEDB6M6nJ1wyD1XQVilnD/Tgf0xGEbvet1XvU
1zYckr+QZRQgzQ53eYL1D9EoLiQ7OK3BanzLgDBWKkHXCEsifWKy2t55SeZL
g9VeIKrSVt8iqHoSjW+TtKddLWtQNgKKr1FS5G2UynvwZQUr3dpnL/gzjrCc
LhSj3OVa2/ENT5ckJN0k4BKha6W7MMxkLTXwzcEJGbvBSOTJHCtuxSLq6UFZ
g60Rz+aTS1UI8t9kD74AhKF6T9iUpfBNXmxKvvGQ3T6oJ7+k2tq37MuDCpkv
Xnr6bsIOs9yyvKpecQhm0aZiRoG+lx1AoIchKZo25iS561AC5RUU0UpPL7rt
zvvoj4+/THsC4xzh3VVN+jfXpIcfnkirV6+hj772U0GHrMS60spHjYYHIYjc
1KcN3zSovMG/vIZPc8RaVF6LX17Dp9k9IiOhMLupTxtl5ZVT8uli/HLFwacb
BRxJVx449dmZVmFzdx9GXqu8Hbke2qf2Pf3i6zTPe/cgJHFGX8S6ERcw/iBd
wOsrd9GYceP7rXFOju2qJKJuowsA4ET1OYTemL+oz+zEgdm+45V33lvKiuc4
Y0i4FK4k/MsawKdN3dBgYbS3burTppKV1peb+rSpr+Sc9ZpbWOmkfzflQbBT
fkFhn4GME6uuZjZT7DSio0SOzXDx6gvzlcuv9I2r/g+0wUj0\
\>"]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Irreducible numerators", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"CompleteDenominatorSet", "[", 
   RowBox[{
   "dens_List", ",", "externalmomenta_List", ",", "loopmomenta_List"}], "]"}],
   ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "externalsps", ",", "internalsps", ",", "nums", ",", "candidates", ",", 
      "m", ",", "r", ",", "denset", ",", "newnum", ",", "newden", ",", "newr",
       ",", "newm", ",", "EXT"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"externalsps", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"a", ",", "b"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"a", ",", "externalmomenta"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"b", ",", "externalmomenta"}], "}"}]}], "]"}], "//", 
         "ExpandSP"}], "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_sp", ",", 
           RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"internalsps", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"a", ",", "b"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"a", ",", "loopmomenta"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"b", ",", 
             RowBox[{"Join", "[", 
              RowBox[{"externalmomenta", ",", "loopmomenta"}], "]"}]}], 
            "}"}]}], "]"}], "//", "ExpandSP"}], "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_sp", ",", 
           RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"nums", "=", 
      RowBox[{
       RowBox[{"dens", "/.", 
        RowBox[{
         RowBox[{"den", "[", 
          RowBox[{"k_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
         RowBox[{
          RowBox[{"sp", "[", 
           RowBox[{"k", ",", "k"}], "]"}], "-", "m2"}]}]}], "//", 
       "ExpandSP"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"candidates", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Join", "[", 
          RowBox[{"loopmomenta", ",", "externalmomenta"}], "]"}], "//", 
         "Subsets"}], "//", 
        RowBox[{
         RowBox[{"#", "[", 
          RowBox[{"[", 
           RowBox[{"2", ";;"}], "]"}], "]"}], "&"}]}], "//", 
       RowBox[{"Map", "[", 
        RowBox[{
         RowBox[{"den", "[", 
          RowBox[{
           RowBox[{"Plus", "@@", "#"}], ",", "0", ",", "irred"}], "]"}], 
         "&"}], "]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"m", "=", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"Coefficient", "[", 
         RowBox[{"n", ",", "s"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"n", ",", "nums"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"s", ",", "internalsps"}], "}"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"r", "=", 
      RowBox[{"MatrixRank", "[", "m", "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"denset", "=", "dens"}], ";", "\[IndentingNewLine]", 
     RowBox[{"Do", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"r", ">=", 
           RowBox[{"Length", "[", "internalsps", "]"}]}], ",", 
          RowBox[{"Break", "[", "]"}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"newnum", "=", 
         RowBox[{
          RowBox[{"newden", "/.", 
           RowBox[{
            RowBox[{"den", "[", 
             RowBox[{"k_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
            
            RowBox[{
             RowBox[{"sp", "[", 
              RowBox[{"k", ",", "k"}], "]"}], "-", "m2"}]}]}], "//", 
          "ExpandSP"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"newm", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"m", ",", 
           RowBox[{"Table", "[", 
            RowBox[{
             RowBox[{"Coefficient", "[", 
              RowBox[{"newnum", ",", "s"}], "]"}], ",", 
             RowBox[{"{", 
              RowBox[{"s", ",", "internalsps"}], "}"}]}], "]"}]}], "]"}]}], 
        ";", "\[IndentingNewLine]", 
        RowBox[{"newr", "=", 
         RowBox[{"MatrixRank", "[", "newm", "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"newr", ">", "r"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"m", "=", "newm"}], ";", "\[IndentingNewLine]", 
           RowBox[{"r", "=", "newr"}], ";", "\[IndentingNewLine]", 
           RowBox[{"denset", "=", 
            RowBox[{"Append", "[", 
             RowBox[{"denset", ",", "newden"}], "]"}]}], ";"}]}], 
         "\[IndentingNewLine]", "]"}], ";"}], "\[IndentingNewLine]", ",", 
       "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"newden", ",", "candidates"}], "}"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"r", "=!=", 
        RowBox[{"Length", "[", "internalsps", "]"}]}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Error", "[", 
         RowBox[{
         "\"\<Failed to complete denominator set \>\"", ",", "dens", ",", 
          "\"\< to \>\"", ",", 
          RowBox[{"Length", "[", "internalsps", "]"}]}], "]"}], ";"}]}], 
      "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", "denset"}]}], 
   "\[IndentingNewLine]", "]"}]}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Conversion into IBP-friendly notation", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ToTopologies", "[", 
   RowBox[{
   "ex_", ",", "topologies_List", ",", "externalmomenta_List", ",", 
    "loopmomenta_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "topologieswithids", ",", "externalsps", ",", "internalsps", ",", "nums",
       ",", "m", ",", "c", ",", "den2DEN", ",", "sp2DEN", ",", "denset", ",", 
      "topid", ",", "indices"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"topologieswithids", "=", 
      RowBox[{"topologies", "//", 
       RowBox[{"MapIndexed", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"#2", "[", 
            RowBox[{"[", "1", "]"}], "]"}], ",", "#1"}], "}"}], "&"}], 
        "]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"den2DEN", "=", 
      RowBox[{"topologies", "//", 
       RowBox[{"Map", "[", 
        RowBox[{"MapIndexed", "[", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"#1", "\[Rule]", 
            RowBox[{"DEN", "@@", "#2"}]}], ")"}], "&"}], "]"}], "]"}]}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"internalsps", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"a", ",", "b"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"a", ",", "loopmomenta"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"b", ",", 
             RowBox[{"Join", "[", 
              RowBox[{"externalmomenta", ",", "loopmomenta"}], "]"}]}], 
            "}"}]}], "]"}], "//", "ExpandSP"}], "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_sp", ",", 
           RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"sp2DEN", "=", 
      RowBox[{"topologies", "//", 
       RowBox[{"Map", "[", 
        RowBox[{
         RowBox[{"(", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"nums", "=", 
            RowBox[{
             RowBox[{"#", "/.", 
              RowBox[{
               RowBox[{"den", "[", 
                RowBox[{"k_", ",", "m2_", ",", "___"}], "]"}], 
               "\[RuleDelayed]", 
               RowBox[{
                RowBox[{"sp", "[", 
                 RowBox[{"k", ",", "k"}], "]"}], "-", "m2"}]}]}], "//", 
             "ExpandSP"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"m", "=", 
            RowBox[{"Table", "[", 
             RowBox[{
              RowBox[{"Coefficient", "[", 
               RowBox[{"n", ",", "s"}], "]"}], ",", 
              RowBox[{"{", 
               RowBox[{"n", ",", "nums"}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{"s", ",", "internalsps"}], "}"}]}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"c", "=", 
            RowBox[{"nums", "/.", 
             RowBox[{
              RowBox[{"Alternatives", "@@", "internalsps"}], "\[Rule]", 
              "0"}]}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"(*", " ", 
            RowBox[{"nums", " ", "\[Equal]", " ", 
             RowBox[{
              RowBox[{"m", ".", "internalsps"}], " ", "+", " ", "c"}]}], " ", 
            "*)"}], "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{
               RowBox[{"Range", "[", 
                RowBox[{"Length", "[", "internalsps", "]"}], "]"}], "//", 
               RowBox[{"Map", "[", 
                RowBox[{
                 RowBox[{"1", "/", 
                  RowBox[{"DEN", "[", "#", "]"}]}], "&"}], "]"}]}], "//", 
              RowBox[{
               RowBox[{
                RowBox[{"Inverse", "[", "m", "]"}], ".", 
                RowBox[{"(", 
                 RowBox[{"#", "-", "c"}], ")"}]}], "&"}]}], "//", "Expand"}], 
            "//", 
            RowBox[{
             RowBox[{"MapThread", "[", 
              RowBox[{"Rule", ",", 
               RowBox[{"{", 
                RowBox[{"internalsps", ",", "#"}], "}"}]}], "]"}], "&"}]}]}], 
          "\[IndentingNewLine]", ")"}], "&"}], "]"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"ex", "//", "ExpandSP"}], "//", "Expand"}], "//", "Terms"}], 
       "//", 
       RowBox[{"MapWithProgress", "[", 
        RowBox[{
         RowBox[{"(", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"denset", "=", 
            RowBox[{
             RowBox[{"Cases", "[", 
              RowBox[{"#", ",", "_den", ",", 
               RowBox[{"{", 
                RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "//", 
             "Union"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"topid", "=", 
            RowBox[{"FirstCase", "[", 
             RowBox[{"topologieswithids", ",", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{"id_", ",", 
                 RowBox[{"dens_", "/;", 
                  RowBox[{"SubsetQ", "[", 
                   RowBox[{"dens", ",", "denset"}], "]"}]}]}], "}"}], 
               "\[RuleDelayed]", "id"}], ",", "$Failed"}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"topid", "===", "$Failed"}], ",", "\[IndentingNewLine]", 
             
             RowBox[{
              RowBox[{"Error", "[", 
               RowBox[{
               "\"\<Failed to map \>\"", ",", "denset", ",", "\"\< to \>\"", 
                ",", "topologieswithids"}], "]"}], ";"}]}], 
            "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"#", "/.", 
                RowBox[{"den2DEN", "[", 
                 RowBox[{"[", "topid", "]"}], "]"}]}], "/.", 
               RowBox[{"sp2DEN", "[", 
                RowBox[{"[", "topid", "]"}], "]"}]}], ")"}], "//", "Expand"}],
             "//", 
            RowBox[{"MapTerms", "[", 
             RowBox[{
              RowBox[{"(", "\[IndentingNewLine]", 
               RowBox[{
                RowBox[{"indices", "=", 
                 RowBox[{"Table", "[", 
                  RowBox[{"0", ",", 
                   RowBox[{"Length", "[", "internalsps", "]"}]}], "]"}]}], 
                ";", "\[IndentingNewLine]", 
                RowBox[{
                 RowBox[{"#", "/.", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"DEN", "[", "i_", "]"}], "^", "n_."}], 
                   "\[RuleDelayed]", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"indices", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "+=", "n"}], ";", "1"}], 
                    ")"}]}]}], "//", 
                 RowBox[{
                  RowBox[{"#", "*", 
                   RowBox[{"B", "[", 
                    RowBox[{"topid", ",", "indices"}], "]"}]}], "&"}]}]}], 
               "\[IndentingNewLine]", ")"}], "&"}], "]"}]}]}], 
          "\[IndentingNewLine]", ")"}], "&"}], "]"}]}], "//", 
      RowBox[{"Apply", "[", "Plus", "]"}]}]}]}], "\[IndentingNewLine]", 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FromTopologies", "[", 
   RowBox[{"ex_", ",", "topologies_List"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"ex", "/.", 
     RowBox[{
      RowBox[{"B", "[", 
       RowBox[{"topid_", ",", "indices_List"}], "]"}], "\[RuleDelayed]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"MapThread", "[", 
          RowBox[{"Power", ",", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"topologies", "[", 
              RowBox[{"[", "topid", "]"}], "]"}], ",", "indices"}], "}"}]}], 
          "]"}], "//", 
         RowBox[{"Apply", "[", "Times", "]"}]}], "//", "Expand"}], ")"}]}]}], 
    "//", 
    RowBox[{"ReplaceAll", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"den", "[", 
        RowBox[{"k_", ",", "m_"}], "]"}], "^", 
       RowBox[{"n_", "?", "Negative"}]}], "\[RuleDelayed]", 
      RowBox[{"ExpandSP", "[", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"sp", "[", 
           RowBox[{"k", ",", "k"}], "]"}], "-", "m"}], ")"}], "^", 
        RowBox[{"(", 
         RowBox[{"-", "n"}], ")"}]}], "]"}]}], "]"}]}], "//", 
   "Expand"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FromTopologies", "[", "topologies_List", "]"}], ":=", 
  RowBox[{
   RowBox[{"FromTopologies", "[", 
    RowBox[{"#", ",", "topologies"}], "]"}], "&"}]}]}], "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4. Leading order: PS[k1,k2] L[]", "Section"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"externalsprules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k1"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k2"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k2"}], "]"}], "\[Rule]", 
      RowBox[{"Q2", "/", "2"}]}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", "\"\<M.m\>\"", "]"}], "[", 
      RowBox[{
       RowBox[{"PS", "[", 
        RowBox[{"k1", ",", "k2"}], "]"}], 
       RowBox[{"L", "[", "]"}]}], "]"}], "/.", "externalsprules"}], "/.", 
    RowBox[{"x", "\[Rule]", "1"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"expr", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "=", 
  RowBox[{"expr", "//", "Topologies"}]}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"MakePrettyTopologyGraph", "[", 
     RowBox[{"#", ",", 
      RowBox[{"{", 
       RowBox[{"k1", ",", "k2"}], "}"}]}], "]"}], "&"}], "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox["5", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{"{", "}"}], "}"}], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   GraphicsBox[
    NamespaceBox["NetworkGraphics",
     DynamicModuleBox[{Typeset`graph = HoldComplete[
       Graph[{
         $CellContext`II[1], 1, 2, 
         $CellContext`OO[1]}, {Null, {{1, 2}, {2, 3}, {2, 3}, {3, 4}}}, {
        EdgeStyle -> {UndirectedEdge[2, 
             $CellContext`OO[1]] -> {
             GrayLevel[0.5]}, UndirectedEdge[1, 2] -> {{
              Thickness[Large], 
              RGBColor[1, 0, 0], 
              Dashing[{Small, Small}]}}, UndirectedEdge[
             $CellContext`II[1], 1] -> {
             GrayLevel[0.5]}}, GraphLayout -> "SpringEmbedding", 
         VertexShape -> {$CellContext`II[1] -> 
           Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
         VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
             GrayLevel[0]}, 2 -> {
             GrayLevel[0]}}}]]}, 
      TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBUQKu/SFib1JlAgAAACQAAAACAAAAS5rUjjXGB0D4lKVRQzPDP0ELKF6Q
t/8/AAAAAAAAAAAAjJu0A9PvPzNhaQeQfsU/AAAAAAAAAACAJtdiPBi4P9gf
tFrR8v0/4MLlt8w6k78hthp24n/9P4A0wP4yZ5e/Zue4IEgM/T/ApoU8tzCa
v44D0MxGmPw/AMJ5+7KVm78bfJwpIyT8PyDtn95SlZu/CC+r+iGw+z/AO+Ae
ly+av5+qIu+HPPs/YLvlilNll79PhRh5mcn6P4A4tgkvOJO/A9BaRC139D8Q
TJogAsioPzUoTXAuBfQ/oIxOVciOqz+zSti8YpTzP8B1DKgdBK8/elrA8gwl
8z942iv0+pKxP0SjZv1ut/I/+K3TYO/4sz8sAMLDyUvyPxhEHpeAsrY/R3Lx
AV3i8T9In8uSEb65P8WnfyNne/E/wE/FwtQZvT98KfYUqyX+P3DlGpiWxqs/
+16ENrW+/T8sIwd8Dj+xPxbRs3RIVf0/XH60d59KtD/+LQ87o+n8P4AU/60w
BLc/x3a1RQV8/D/456YaJWq5P46GnXuvDPw/jIfMOhF7uz8NqSjI45v7Pxx8
K+S7Nb0/PwEb9OQp+z9knIX+Hpm+P/NLXb941/Q/RCig6JXlxz+jJlNJimT0
P6AYxng6a8g/OaLKPfDw8z+waEXrgsTIPyZV2Q7vfPM/2l49YzrxyD+0zaVr
ywjzP3KZ2GZG8cg/3Om8F8qU8j8MFvrupsTIPyEbW8IvIfI/xGdBZ3ZryD9p
scHdQK7xP44ZZp7p5cc/QUYc1A==
         "], {
          {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
           {GrayLevel[0.5], 
            {Arrowheads[0.], ArrowBox[{1, 2}, 0.03003681795644597]}}, 
           {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
            {Arrowheads[0.], 
             ArrowBox[BezierCurveBox[{
              2, {1.460798315079948, -0.07847324791946825}, 3}], 
              0.03003681795644597]}}, 
           {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
            {Arrowheads[0.], 
             ArrowBox[BezierCurveBox[{
              2, {1.5160257020134742`, 0.2463981323522549}, 3}], 
              0.03003681795644597]}}, 
           {GrayLevel[0.5], 
            {Arrowheads[0.], ArrowBox[{3, 4}, 0.03003681795644597]}}}, 
          {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
           InsetBox[
            GraphicsBox[{}], 1, 
            Automatic, {0.06007363591289194, 0.06007363591289194}], 
           {GrayLevel[0], DiskBox[2, 0.049862173792727754`]}, 
           {GrayLevel[0], DiskBox[3, 0.049862173792727754`]}, InsetBox[
            GraphicsBox[{}], 4, 
            Automatic, {0.06007363591289194, 0.06007363591289194}]}}]],
       MouseAppearanceTag["NetworkGraphics"]],
      AllowKernelInitialization->False]],
    DefaultBaseStyle->{
     "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
    
    FormatType->TraditionalForm,
    FrameTicks->None], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["expr"], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"2", " ", 
   RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
   SuperscriptBox["qu", "2"], " ", 
   SubscriptBox["C", "A"]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Omega", "[", "k_", "]"}], ":=", 
  RowBox[{"2", 
   RowBox[{
    RowBox[{"Pi", "^", 
     RowBox[{"(", 
      RowBox[{"k", "/", "2"}], ")"}]}], "/", 
    RowBox[{"Gamma", "[", 
     RowBox[{"k", "/", "2"}], "]"}]}]}]}], "\[IndentingNewLine]", 
 RowBox[{"ps2factor", "=", 
  RowBox[{
   RowBox[{"Q2", "^", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"d", "/", "2"}], "-", "2"}], ")"}]}], 
   RowBox[{"2", "^", 
    RowBox[{"(", 
     RowBox[{"1", "-", "d"}], ")"}]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"2", "Pi"}], ")"}], "^", 
    RowBox[{"(", 
     RowBox[{"2", "-", "d"}], ")"}]}], 
   RowBox[{"Omega", "[", 
    RowBox[{"d", "-", "1"}], "]"}], 
   RowBox[{"(*", " ", 
    RowBox[{"delta", "[", 
     RowBox[{"x", "-", "1"}], "]"}], " ", "*)"}]}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{
    SuperscriptBox["4", 
     RowBox[{"2", "-", "d"}]], " ", 
    SuperscriptBox["\[Pi]", 
     RowBox[{
      FractionBox[
       RowBox[{"d", "-", "1"}], "2"], "-", "d", "+", "2"}]], " ", 
    SuperscriptBox["Q2", 
     RowBox[{
      FractionBox["d", "2"], "-", "2"}]]}], 
   TemplateBox[{FractionBox[
      RowBox[{"d", "-", "1"}], "2"]},
    "Gamma"]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"result", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"expr", "*", 
      RowBox[{"ps2factor", "/", 
       RowBox[{"PS", "[", "2", "]"}]}]}], "/.", 
     RowBox[{"d", "\[Rule]", 
      RowBox[{"4", "-", 
       RowBox[{"2", "ep"}]}]}]}], "//", 
    RowBox[{
     RowBox[{"#", "+", 
      RowBox[{
       RowBox[{"O", "[", "ep", "]"}], "^", "2"}]}], "&"}]}], "//", 
   RowBox[{"MapSeries", "[", 
    RowBox[{"FunctionExpand", "/*", "Factor", "/*", "PowerExpand"}], 
    "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    FractionBox[
     RowBox[{"Q2", " ", 
      SuperscriptBox["qu", "2"], " ", 
      SubscriptBox["C", "A"]}], 
     RowBox[{"4", " ", "\[Pi]"}]], "+", 
    FractionBox[
     RowBox[{"ep", " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", 
      SubscriptBox["C", "A"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"log", "(", "Q2", ")"}]}], "-", 
        TagBox["\[DoubledGamma]",
         Function[{}, EulerGamma]], "+", "2", "+", 
        RowBox[{"log", "(", "\[Pi]", ")"}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", "2", ")"}]}]}], ")"}]}], 
     RowBox[{"4", " ", "\[Pi]"}]], "+", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "2"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 0, 2, 1],
     Editable->False]}],
   SeriesData[$CellContext`ep, 0, {
    Rational[1, 4] FeynCalc`CA Pi^(-1) $CellContext`Q2 $CellContext`qu^2, 
     Rational[1, 4] FeynCalc`CA 
     Pi^(-1) $CellContext`Q2 $CellContext`qu^2 (2 - EulerGamma + 2 Log[2] + 
      Log[Pi] - Log[$CellContext`Q2])}, 0, 2, 1],
   Editable->False], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"result", ",", "\"\<result.ps2.l0.m\>\""}], "]"}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["5. PS[k1,k2] L[l1]: IBP reduction, hand-rolled", "Section"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"externalsprules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k1"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k2"}], "]"}], "\[Rule]", 
      RowBox[{"Q2", "/", "2"}]}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k2"}], "]"}], "\[Rule]", "0"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", "\"\<M.m\>\"", "]"}], "[", 
      RowBox[{
       RowBox[{"PS", "[", 
        RowBox[{"k1", ",", "k2"}], "]"}], 
       RowBox[{"L", "[", "l1", "]"}]}], "]"}], "/.", "externalsprules"}], "/.", 
    RowBox[{"x", "\[Rule]", "1"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expr", "//", 
   RowBox[{
    RowBox[{"Bracket", "[", 
     RowBox[{"#", ",", 
      RowBox[{"_sp", "|", "_den"}]}], "]"}], "&"}]}], "//", 
  "Length"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "=", 
  RowBox[{"expr", "//", "Topologies"}]}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"MakePrettyTopologyGraph", "[", 
     RowBox[{"#", ",", 
      RowBox[{"{", 
       RowBox[{"k1", ",", "k2"}], "}"}]}], "]"}], "&"}], "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox["4", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   GraphicsBox[
    NamespaceBox["NetworkGraphics",
     DynamicModuleBox[{Typeset`graph = HoldComplete[
       Graph[{
         $CellContext`II[1], 1, 3, 4, 2, 
         $CellContext`OO[1]}, {
        Null, {{1, 2}, {2, 3}, {2, 4}, {3, 4}, {4, 5}, {3, 5}, {5, 6}}}, {
        EdgeStyle -> {UndirectedEdge[2, 
             $CellContext`OO[1]] -> {
             GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
             Thickness[Large]}, UndirectedEdge[3, 4] -> {
             Thickness[Large]}, UndirectedEdge[1, 4] -> {
             Dashing[{Small, Small}]}, UndirectedEdge[
             $CellContext`II[1], 1] -> {
             GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
             Thickness[Large]}, UndirectedEdge[1, 3] -> {{
              Thickness[Large], 
              RGBColor[1, 0, 0], 
              Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
         VertexShape -> {$CellContext`II[1] -> 
           Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
         VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
         VertexStyle -> {1 -> {
             GrayLevel[0]}, 2 -> {
             GrayLevel[0]}, 3 -> {
             GrayLevel[0]}, 4 -> {
             GrayLevel[0]}}}]]}, 
      TagBox[
       GraphicsGroupBox[
        GraphicsComplexBox[{{3.814781421808071, 0.4877084622328541}, {
         2.825162389780683, 0.453296561078493}, {1.9077379198057707`, 
         0.913800986380224}, {1.9070977558442062`, 0.}, {0.9896915982171941, 
         0.4559929121833412}, {0., 0.48656376120221734`}}, {
          {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
           {GrayLevel[0.5], 
            {Arrowheads[0.], ArrowBox[{1, 2}, 0.035810735981566255`]}}, 
           {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
            {Arrowheads[0.], ArrowBox[{2, 3}, 0.035810735981566255`]}}, 
           {Dashing[{Small, Small}], 
            {Arrowheads[0.], ArrowBox[{2, 4}, 0.035810735981566255`]}}, 
           {Thickness[Large], 
            {Arrowheads[0.], ArrowBox[{3, 4}, 0.035810735981566255`]}}, 
           {Thickness[Large], 
            {Arrowheads[0.], ArrowBox[{3, 5}, 0.035810735981566255`]}}, 
           {Thickness[Large], 
            {Arrowheads[0.], ArrowBox[{4, 5}, 0.035810735981566255`]}}, 
           {GrayLevel[0.5], 
            {Arrowheads[0.], ArrowBox[{5, 6}, 0.035810735981566255`]}}}, 
          {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
           InsetBox[
            GraphicsBox[{}], 1, 
            Automatic, {0.07162147196313251, 0.07162147196313251}], 
           {GrayLevel[0], DiskBox[2, 0.04569006053069333]}, 
           {GrayLevel[0], DiskBox[3, 0.04569006053069333]}, 
           {GrayLevel[0], DiskBox[4, 0.04569006053069333]}, 
           {GrayLevel[0], DiskBox[5, 0.04569006053069333]}, InsetBox[
            GraphicsBox[{}], 6, 
            Automatic, {0.07162147196313251, 0.07162147196313251}]}}]],
       MouseAppearanceTag["NetworkGraphics"]],
      AllowKernelInitialization->False]],
    DefaultBaseStyle->{
     "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
    
    FormatType->TraditionalForm,
    FrameTicks->None], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "//", 
  RowBox[{
   RowBox[{"Bracket", "[", 
    RowBox[{"#", ",", 
     RowBox[{"_sp", "|", "_den"}]}], "]"}], "&"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{
    RowBox[{"den", "(", 
     RowBox[{"l1", ",", "0"}], ")"}], " ", 
    RowBox[{"k1", "\[CenterDot]", "l1"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"4", " ", "\[ImaginaryI]", " ", "d", " ", 
       SuperscriptBox["gs", "2"], " ", 
       RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
       SuperscriptBox["qu", "2"], " ", 
       SubscriptBox["C", "A"], " ", 
       SubscriptBox["C", "F"]}], "-", 
      RowBox[{"16", " ", "\[ImaginaryI]", " ", 
       SuperscriptBox["gs", "2"], " ", 
       RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
       SuperscriptBox["qu", "2"], " ", 
       SubscriptBox["C", "A"], " ", 
       SubscriptBox["C", "F"]}]}], ")"}]}], "+", 
   RowBox[{
    RowBox[{"l1", "\[CenterDot]", "l1"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{"l1", ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"8", " ", "\[ImaginaryI]", " ", 
       SuperscriptBox["gs", "2"], " ", 
       RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
       SuperscriptBox["qu", "2"], " ", 
       SubscriptBox["C", "A"], " ", 
       SubscriptBox["C", "F"]}], "-", 
      RowBox[{"2", " ", "\[ImaginaryI]", " ", "d", " ", 
       SuperscriptBox["gs", "2"], " ", 
       RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
       SuperscriptBox["qu", "2"], " ", 
       SubscriptBox["C", "A"], " ", 
       SubscriptBox["C", "F"]}]}], ")"}]}], "+", 
   RowBox[{"8", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"], " ", 
    RowBox[{"den", "(", 
     RowBox[{"l1", ",", "0"}], ")"}], " ", 
    RowBox[{"k2", "\[CenterDot]", "l1"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}]}], "-", 
   RowBox[{"16", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"], " ", 
    RowBox[{"den", "(", 
     RowBox[{"l1", ",", "0"}], ")"}], " ", 
    RowBox[{"k1", "\[CenterDot]", "l1"}], " ", 
    RowBox[{"k2", "\[CenterDot]", "l1"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}]}]}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"topologies", "=", 
  RowBox[{"topologies", "//", 
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{"CompleteDenominatorSet", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"k1", ",", "k2"}], "}"}], ",", 
       RowBox[{"{", "l1", "}"}]}], "]"}], "&"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"exprb", "=", 
  RowBox[{
   RowBox[{"expr", "//", 
    RowBox[{
     RowBox[{"ToTopologies", "[", 
      RowBox[{"#", ",", "topologies", ",", 
       RowBox[{"{", 
        RowBox[{"k1", ",", "k2"}], "}"}], ",", 
       RowBox[{"{", "l1", "}"}]}], "]"}], "&"}]}], "//", 
   RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"-", 
    RowBox[{"2", " ", "\[ImaginaryI]", " ", "d", " ", 
     SuperscriptBox["gs", "2"], " ", 
     RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
     SuperscriptBox["qu", "2"], " ", 
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "1", ",", "1"}], "1"], " ", 
     SubscriptBox["C", "A"], " ", 
     SubscriptBox["C", "F"]}]}], "+", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", 
    SuperscriptBox["Q2", "2"], " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "1"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "+", 
   RowBox[{"16", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "1"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "-", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "0", ",", "1"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "-", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "0"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "+", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "1", ",", "1"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "-", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "0", ",", "1"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "-", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "0"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}], "+", 
   RowBox[{"4", " ", "\[ImaginaryI]", " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "0", ",", "0"}], "1"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"]}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprb", "//", 
   RowBox[{
    RowBox[{"Cases", "[", 
     RowBox[{"#", ",", "_B", ",", 
      RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "1", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "0", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "0"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "0", ",", "0"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "0", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "0"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "1"}], "1"]}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["IBP Relation Construction", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ScalarProductsToDenominators", "[", 
   RowBox[{
   "dens_List", ",", "externalmomenta_List", ",", "loopmomenta_List"}], "]"}],
   ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"nums", ",", "internalsps", ",", "m", ",", "c"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"nums", "=", 
      RowBox[{
       RowBox[{"dens", "/.", 
        RowBox[{
         RowBox[{"den", "[", 
          RowBox[{"k_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
         RowBox[{
          RowBox[{"sp", "[", 
           RowBox[{"k", ",", "k"}], "]"}], "-", "m2"}]}]}], "//", 
       "ExpandSP"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"internalsps", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"a", ",", "b"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"a", ",", "loopmomenta"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"b", ",", 
             RowBox[{"Join", "[", 
              RowBox[{"externalmomenta", ",", "loopmomenta"}], "]"}]}], 
            "}"}]}], "]"}], "//", "ExpandSP"}], "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_sp", ",", 
           RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"m", "=", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"Coefficient", "[", 
         RowBox[{"n", ",", "s"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"n", ",", "nums"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"s", ",", "internalsps"}], "}"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"c", "=", 
      RowBox[{"nums", "/.", 
       RowBox[{
        RowBox[{"Alternatives", "@@", "internalsps"}], "\[Rule]", "0"}]}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"nums", " ", "\[Equal]", " ", 
       RowBox[{
        RowBox[{"m", ".", "internalsps"}], " ", "+", " ", "c"}]}], " ", 
      "*)"}], "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Range", "[", 
          RowBox[{"Length", "[", "internalsps", "]"}], "]"}], "//", 
         RowBox[{"Map", "[", 
          RowBox[{
           RowBox[{"DEN", "[", "#", "]"}], "&"}], "]"}]}], "//", 
        RowBox[{
         RowBox[{
          RowBox[{"Inverse", "[", "m", "]"}], ".", 
          RowBox[{"(", 
           RowBox[{"#", "-", "c"}], ")"}]}], "&"}]}], "//", "Expand"}], "//", 
      
      RowBox[{
       RowBox[{"MapThread", "[", 
        RowBox[{"Rule", ",", 
         RowBox[{"{", 
          RowBox[{"internalsps", ",", "#"}], "}"}]}], "]"}], "&"}]}]}]}], 
   "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DENsToB", "[", 
   RowBox[{"ex_", ",", "topid_Integer", ",", "dencount_Integer"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "indices", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"ex", "//", "Expand"}], "//", 
     RowBox[{"MapTerms", "[", 
      RowBox[{
       RowBox[{"(", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"indices", "=", 
          RowBox[{"Table", "[", 
           RowBox[{"0", ",", "dencount"}], "]"}]}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"#", "/.", 
           RowBox[{
            RowBox[{
             RowBox[{"DEN", "[", "i_", "]"}], "^", "n_."}], "\[RuleDelayed]", 
            
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"indices", "[", 
                RowBox[{"[", "i", "]"}], "]"}], "-=", "n"}], ";", "1"}], 
             ")"}]}]}], "//", 
          RowBox[{
           RowBox[{"#", "*", 
            RowBox[{"B", "[", 
             RowBox[{"topid", ",", "indices"}], "]"}]}], "&"}]}]}], 
        "\[IndentingNewLine]", ")"}], "&"}], "]"}]}]}], "\[IndentingNewLine]",
    "]"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ScalarProductsToDenominators", "[", 
  RowBox[{
   RowBox[{"topologies", "[", 
    RowBox[{"[", "1", "]"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"k1", ",", "k2"}], "}"}], ",", 
   RowBox[{"{", "l1", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"k1", "\[CenterDot]", "l1"}], "\[Rule]", 
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        RowBox[{"DEN", "(", "1", ")"}], "2"]}], "+", 
      FractionBox[
       RowBox[{"DEN", "(", "3", ")"}], "2"], "+", 
      FractionBox[
       RowBox[{"k1", "\[CenterDot]", "k1"}], "2"]}]}], ",", 
    RowBox[{
     RowBox[{"k2", "\[CenterDot]", "l1"}], "\[Rule]", 
     RowBox[{
      FractionBox[
       RowBox[{"DEN", "(", "1", ")"}], "2"], "-", 
      FractionBox[
       RowBox[{"DEN", "(", "2", ")"}], "2"], "+", 
      RowBox[{"k1", "\[CenterDot]", "k2"}], "+", 
      FractionBox[
       RowBox[{"k2", "\[CenterDot]", "k2"}], "2"]}]}], ",", 
    RowBox[{
     RowBox[{"l1", "\[CenterDot]", "l1"}], "\[Rule]", 
     RowBox[{"DEN", "(", "3", ")"}]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"IBPRelations", "[", 
   RowBox[{
   "topid_Integer", ",", " ", "dens_List", ",", " ", "externalmomenta_List", 
    ",", "loopmomenta_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "den2DEN", ",", "internalsps", ",", "nums", ",", "indices", ",", 
      "sp2DEN", ",", "rels"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"den2DEN", "=", 
      RowBox[{"dens", "//", 
       RowBox[{"MapIndexed", "[", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"#1", "\[Rule]", 
           RowBox[{"DEN", "@@", "#2"}]}], ")"}], "&"}], "]"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"internalsps", "=", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"a", ",", "b"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"a", ",", "loopmomenta"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"b", ",", 
             RowBox[{"Join", "[", 
              RowBox[{"externalmomenta", ",", "loopmomenta"}], "]"}]}], 
            "}"}]}], "]"}], "//", "ExpandSP"}], "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_sp", ",", 
           RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"nums", "=", 
      RowBox[{
       RowBox[{"dens", "/.", 
        RowBox[{
         RowBox[{"den", "[", 
          RowBox[{"k_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
         RowBox[{
          RowBox[{"sp", "[", 
           RowBox[{"k", ",", "k"}], "]"}], "-", "m2"}]}]}], "//", 
       "ExpandSP"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"sp2DEN", "=", 
      RowBox[{"ScalarProductsToDenominators", "[", 
       RowBox[{"dens", ",", "externalmomenta", ",", "loopmomenta"}], "]"}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"indices", "=", 
      RowBox[{
       RowBox[{
        RowBox[{"dens", "//", "Length"}], "//", "Range"}], "//", 
       RowBox[{"Map", "[", "n", "]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"Table", "[", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Product", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"DEN", "[", "i", "]"}], "^", 
             RowBox[{"(", 
              RowBox[{"-", 
               RowBox[{"n", "[", "i", "]"}]}], ")"}]}], ",", 
            RowBox[{"{", 
             RowBox[{"i", ",", 
              RowBox[{"Length", "[", "dens", "]"}]}], "}"}]}], "]"}], "*", 
          RowBox[{"(", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"Sum", "[", "\[IndentingNewLine]", 
             RowBox[{
              RowBox[{
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{"-", "2"}], 
                  RowBox[{"n", "[", "i", "]"}], 
                  RowBox[{"D", "[", 
                   RowBox[{
                    RowBox[{"dens", "[", 
                    RowBox[{"[", 
                    RowBox[{"i", ",", "1"}], "]"}], "]"}], ",", "l"}], "]"}], 
                  
                  RowBox[{
                   RowBox[{"sp", "[", 
                    RowBox[{"v", ",", 
                    RowBox[{"dens", "[", 
                    RowBox[{"[", 
                    RowBox[{"i", ",", "1"}], "]"}], "]"}]}], "]"}], "/", 
                   RowBox[{"DEN", "[", "i", "]"}]}]}], "//", "ExpandSP"}], "//", 
                RowBox[{"ReplaceAll", "[", 
                 RowBox[{"sp", "\[Rule]", "sp"}], "]"}]}], "//", 
               RowBox[{"ReplaceAll", "[", "sp2DEN", "]"}]}], 
              "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
              RowBox[{"{", 
               RowBox[{"i", ",", 
                RowBox[{"Length", "[", "dens", "]"}]}], "}"}]}], "]"}], "+", 
            "\[IndentingNewLine]", 
            RowBox[{"If", "[", 
             RowBox[{
              RowBox[{"v", "===", "l"}], ",", "d", ",", "0"}], "]"}]}], 
           "\[IndentingNewLine]", ")"}]}], "//", 
         RowBox[{
          RowBox[{"DENsToB", "[", 
           RowBox[{"#", ",", "topid", ",", 
            RowBox[{"Length", "[", "dens", "]"}]}], "]"}], "&"}]}], 
        "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"l", ",", "loopmomenta"}], "}"}], ",", "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"v", ",", 
          RowBox[{"Join", "[", 
           RowBox[{"externalmomenta", ",", "loopmomenta"}], "]"}]}], "}"}]}], 
       "]"}], "//", "Flatten"}]}]}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"relations", "=", 
   RowBox[{
    RowBox[{"IBPRelations", "[", 
     RowBox[{"1", ",", 
      RowBox[{"topologies", "[", 
       RowBox[{"[", "1", "]"}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"k1", ",", "k2"}], "}"}], ",", 
      RowBox[{"{", "l1", "}"}]}], "]"}], "//", 
    RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"relations", "//", 
   RowBox[{"ReplaceAll", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "[", "1", "]"}], "\[Rule]", "a"}], ",", 
      RowBox[{
       RowBox[{"n", "[", "2", "]"}], "\[Rule]", "b"}], ",", 
      RowBox[{
       RowBox[{"n", "[", "3", "]"}], "\[Rule]", "c"}]}], "}"}], "]"}]}], "//",
   "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       RowBox[{
        RowBox[{"b", " ", "Q2", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", 
           RowBox[{"b", "+", "1"}], ",", "c"}], "1"]}], "+", 
        RowBox[{"c", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "-", "1"}], ",", "b", ",", 
           RowBox[{"c", "+", "1"}]}], "1"]}], "+", 
        RowBox[{"b", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "-", "1"}], ",", 
           RowBox[{"b", "+", "1"}], ",", "c"}], "1"]}], "+", 
        RowBox[{"a", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "-", 
        RowBox[{"c", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "-", 
        RowBox[{"b", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", 
           RowBox[{"b", "+", "1"}], ",", 
           RowBox[{"c", "-", "1"}]}], "1"]}], "-", 
        RowBox[{"a", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "+", "1"}], ",", "b", ",", 
           RowBox[{"c", "-", "1"}]}], "1"]}]}]},
      {
       RowBox[{
        RowBox[{"-", 
         RowBox[{"c", " ", "Q2", " ", 
          SubsuperscriptBox["B", 
           RowBox[{"a", ",", "b", ",", 
            RowBox[{"c", "+", "1"}]}], "1"]}]}], "-", 
        RowBox[{"c", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "-", "1"}], ",", "b", ",", 
           RowBox[{"c", "+", "1"}]}], "1"]}], "-", 
        RowBox[{"b", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "-", "1"}], ",", 
           RowBox[{"b", "+", "1"}], ",", "c"}], "1"]}], "+", 
        RowBox[{"c", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", 
           RowBox[{"b", "-", "1"}], ",", 
           RowBox[{"c", "+", "1"}]}], "1"]}], "-", 
        RowBox[{"a", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "+", 
        RowBox[{"b", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "+", 
        RowBox[{"a", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "+", "1"}], ",", 
           RowBox[{"b", "-", "1"}], ",", "c"}], "1"]}]}]},
      {
       RowBox[{
        RowBox[{"d", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "+", 
        RowBox[{"b", " ", "Q2", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", 
           RowBox[{"b", "+", "1"}], ",", "c"}], "1"]}], "-", 
        RowBox[{"a", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "-", 
        RowBox[{"b", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "-", 
        RowBox[{"2", " ", "c", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", "b", ",", "c"}], "1"]}], "-", 
        RowBox[{"b", " ", 
         SubsuperscriptBox["B", 
          RowBox[{"a", ",", 
           RowBox[{"b", "+", "1"}], ",", 
           RowBox[{"c", "-", "1"}]}], "1"]}], "-", 
        RowBox[{"a", " ", 
         SubsuperscriptBox["B", 
          RowBox[{
           RowBox[{"a", "+", "1"}], ",", "b", ",", 
           RowBox[{"c", "-", "1"}]}], "1"]}]}]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["IBP Solution", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"MakeIndexSubstitutionRules", "[", 
   RowBox[{"nindices_Integer", ",", "maxr_Integer", ",", "maxs_Integer"}], 
   "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"r", ",", "s"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"Tuples", "[", 
       RowBox[{
        RowBox[{"Range", "[", 
         RowBox[{
          RowBox[{"-", "maxs"}], ",", "maxr"}], "]"}], ",", "nindices"}], 
       "]"}], "//", 
      RowBox[{"Select", "[", 
       RowBox[{
        RowBox[{"(", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"r", "=", 
           RowBox[{
            RowBox[{"#", "*", 
             RowBox[{"HeavisideTheta", "[", 
              RowBox[{"#", "-", 
               RowBox[{"1", "/", "2"}]}], "]"}]}], "//", 
            RowBox[{"Apply", "[", "Plus", "]"}]}]}], ";", 
          "\[IndentingNewLine]", 
          RowBox[{"s", "=", 
           RowBox[{
            RowBox[{
             RowBox[{"-", "#"}], "*", 
             RowBox[{"HeavisideTheta", "[", 
              RowBox[{
               RowBox[{"1", "/", "2"}], "-", "#"}], "]"}]}], "//", 
            RowBox[{"Apply", "[", "Plus", "]"}]}]}], ";", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"r", "<", "maxr"}], "&&", 
           RowBox[{"s", "<", "maxs"}]}]}], "\[IndentingNewLine]", ")"}], 
        "&"}], "]"}]}], "//", 
     RowBox[{"Map", "[", 
      RowBox[{"MapIndexed", "[", 
       RowBox[{
        RowBox[{"Rule", "[", 
         RowBox[{
          RowBox[{"n", "@@", "#2"}], ",", "#1"}], "]"}], "&"}], "]"}], 
      "]"}]}]}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MakeIndexSubstitutionRules", "[", 
  RowBox[{"3", ",", 
   RowBox[{"3", "+", "1"}], ",", "1"}], "]"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "2"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "3"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "2"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "2"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "2"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "3"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "2"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "2"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "2"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "2"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "1"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "2"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "1"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "(", "1", ")"}], "\[Rule]", "3"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "2", ")"}], "\[Rule]", "0"}], ",", 
      RowBox[{
       RowBox[{"n", "(", "3", ")"}], "\[Rule]", "0"}]}], "}"}]}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"BPositiveIndexSum", "[", 
   RowBox[{"B", "[", 
    RowBox[{"_", ",", "idx_List"}], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"idx", "//", 
    RowBox[{"Select", "[", "Positive", "]"}]}], "//", 
   RowBox[{"Apply", "[", "Plus", "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BNegativeIndexSum", "[", 
   RowBox[{"B", "[", 
    RowBox[{"_", ",", "idx_List"}], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "idx"}], "//", 
    RowBox[{"Select", "[", "Positive", "]"}]}], "//", 
   RowBox[{"Apply", "[", "Plus", "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BPositiveIndexCount", "[", 
   RowBox[{"B", "[", 
    RowBox[{"_", ",", "idx_List"}], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"idx", "//", 
    RowBox[{"Select", "[", "Positive", "]"}]}], "//", 
   "Length"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BNegativeIndexCount", "[", 
   RowBox[{"B", "[", 
    RowBox[{"_", ",", "idx_List"}], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "idx"}], "//", 
    RowBox[{"Select", "[", "Positive", "]"}]}], "//", 
   "Length"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BPositiveDotSum", "[", 
   RowBox[{"B", "[", 
    RowBox[{"_", ",", "idx_List"}], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"idx", "//", 
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{"#", "-", "1"}], "&"}], "]"}]}], "//", 
    RowBox[{"Select", "[", "Positive", "]"}]}], "//", 
   RowBox[{"Apply", "[", "Plus", "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BNegativeDotSum", "[", 
   RowBox[{"B", "[", 
    RowBox[{"_", ",", "idx_List"}], "]"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"idx", "//", 
     RowBox[{"Map", "[", 
      RowBox[{
       RowBox[{"#", "-", "1"}], "&"}], "]"}]}], "//", 
    RowBox[{"Select", "[", "Positive", "]"}]}], "//", 
   RowBox[{"Apply", "[", "Plus", "]"}]}]}]}], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ibpzeromap", "=", 
   RowBox[{"{", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"B", "[", 
       RowBox[{"_", ",", 
        RowBox[{"{", 
         RowBox[{"_", ",", "_", ",", 
          RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "\[Rule]", 
      "0"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"B", "[", 
       RowBox[{"_", ",", 
        RowBox[{"{", 
         RowBox[{"_", ",", 
          RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], 
      "\[Rule]", "0"}]}], "\[IndentingNewLine]", "}"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"irels", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"relations", "/.", 
      RowBox[{"MakeIndexSubstitutionRules", "[", 
       RowBox[{"3", ",", 
        RowBox[{"3", "+", "1", "+", "1"}], ",", 
        RowBox[{"1", "+", "1"}]}], "]"}]}], "/.", "ibpzeromap"}], "//", 
    "Flatten"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"irels", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{"bs", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"irels", "//", 
      RowBox[{
       RowBox[{"Cases", "[", 
        RowBox[{"#", ",", "_B", ",", 
         RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}], "//", 
    RowBox[{"SortBy", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"BPositiveIndexSum", "[", "#", "]"}], "+", 
         RowBox[{"BNegativeIndexSum", "[", "#", "]"}]}], ",", "#"}], "}"}], 
      "&"}], "]"}]}], "//", "Reverse"}]}], "\[IndentingNewLine]", 
 RowBox[{"bs", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["240", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "4", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "3", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "2", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "1", ",", "4"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "4", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "3", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "2", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "1", ",", "4"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "3", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "2", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "1", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"2", ",", "2", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"2", ",", "1", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "3", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "2", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "4", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "3", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "2", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "4"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "3", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "2", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "1", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "2", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "1", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"2", ",", "1", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "2", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "3", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "2", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "3"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "2", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "1", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "2", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "2"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{
      RowBox[{"-", "1"}], ",", "1", ",", "1"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "1", ",", "1"}], "1"]}], "}"}], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox["38", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"sys", "=", 
  RowBox[{
   RowBox[{"irels", "//", 
    RowBox[{
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"Coefficient", "[", 
        RowBox[{"r", ",", "b"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"r", ",", "#"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"b", ",", "bs"}], "}"}]}], "]"}], "&"}]}], "//", 
   "RowReduce"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"1", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["d", "2"], "-", 
          RowBox[{"2", " ", "d"}]}], ")"}]}], 
       RowBox[{"12", " ", "Q2"}]]},
     {"0", "1", "0", 
      RowBox[{"-", "3"}], "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "2"}], ")"}]}], 
        RowBox[{"2", " ", "Q2"}]]}]},
     {"0", "0", "1", "3", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["d", "2"], "-", 
          RowBox[{"4", " ", "d"}], "+", "4"}], ")"}]}], 
       RowBox[{"4", " ", "Q2"}]]},
     {"0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "2"}], ")"}]}], 
        RowBox[{"12", " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "6"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{"4", " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "6"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{"4", " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "2"}], ")"}]}], 
        RowBox[{"12", " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "1", "0", 
      RowBox[{"-", "1"}], "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "1", "2", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         FractionBox["1", "2"]}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", "d"}], "-", "1"}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{"4", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "6"}], ")"}], " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{"4", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "6"}], ")"}], " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       SuperscriptBox["Q2", "3"]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["d", "2"], "-", 
          RowBox[{"13", " ", "d"}], "+", "40"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "6"}], ")"}], " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       SuperscriptBox["Q2", "3"]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{"6", " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["d", "2"], "-", 
          RowBox[{"13", " ", "d"}], "+", "40"}], ")"}]}], 
       RowBox[{"2", " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["d", "2"], "-", 
          RowBox[{"13", " ", "d"}], "+", "40"}], ")"}]}], 
       RowBox[{"2", " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{"6", " ", 
        SuperscriptBox["Q2", "3"]}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}]}], 
       RowBox[{"4", " ", "Q2"}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{"2", " ", "Q2"}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "1", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}]}], 
       RowBox[{"4", " ", "Q2"}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", "1", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{"d", " ", "Q2"}], "2"]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"4", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "6"}], ")"}], " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        SuperscriptBox["Q2", "2"]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "1", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        SuperscriptBox["Q2", "2"]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "1", "0", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{"2", " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "1", "0", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "6"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        SuperscriptBox["Q2", "2"]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "1", "0", "0", "0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{"2", " ", 
         SuperscriptBox["Q2", "2"]}]]}]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "1", "0", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{"2", "-", "d"}], "2"]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "1", "0", "0", "0", "0", 
      FractionBox[
       RowBox[{"2", "-", "d"}], "2"]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "1", "0", "0", "0", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", "Q2"}]]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "1", "0", "0", 
      FractionBox[
       RowBox[{"d", "-", "3"}], "Q2"]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "1", "0", 
      FractionBox[
       RowBox[{"d", "-", "3"}], "Q2"]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "1", 
      FractionBox["Q2", "2"]},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"},
     {"0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", "0", 
      "0", "0", "0", "0", "0", "0", "0", "0", "0", "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Module", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"i", "=", "1"}], "}"}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"masters", "=", 
     RowBox[{"{", "}"}]}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Do", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"While", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"i", "<=", 
            RowBox[{"Length", "[", "row", "]"}]}], "&&", 
           RowBox[{
            RowBox[{"row", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "===", "0"}]}], ",", 
          RowBox[{
           RowBox[{"masters", "=", 
            RowBox[{"{", 
             RowBox[{"masters", ",", "i"}], "}"}]}], ";", 
           RowBox[{"i", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"i", "++"}], ";"}], "\[IndentingNewLine]", ",", 
       "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"row", ",", "sys"}], "}"}]}], "]"}], "\[IndentingNewLine]", 
     RowBox[{"bselection", "=", 
      RowBox[{"Select", "[", 
       RowBox[{"FreeQ", "[", 
        RowBox[{"b_B", "/;", 
         RowBox[{
          RowBox[{
           RowBox[{"BPositiveIndexSum", "[", "b", "]"}], ">", 
           RowBox[{"3", "+", "1"}]}], "||", 
          RowBox[{
           RowBox[{"BNegativeIndexSum", "[", "b", "]"}], ">", "1"}]}]}], 
        "]"}], "]"}]}]}], ";"}]}], "\[IndentingNewLine]", 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"masters", "=", 
   RowBox[{
    RowBox[{"bs", "[", 
     RowBox[{"[", 
      RowBox[{
       RowBox[{"masters", "//", "Flatten"}], "//", "Reverse"}], "]"}], "]"}], 
    "//", "bselection"}]}], 
  ";"}], "\[IndentingNewLine]", "masters", "\[IndentingNewLine]", 
 RowBox[{"ibpmap", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"sys", ".", "bs"}], "//", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{"#", "\[Equal]", "0"}], "&"}], "]"}]}], "//", "bselection"}], 
    "//", 
    RowBox[{
     RowBox[{"Solve", "[", 
      RowBox[{"#", ",", 
       RowBox[{"Complement", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"Cases", "[", 
           RowBox[{"#", ",", "_B", ",", 
            RowBox[{"-", "1"}]}], "]"}], "//", "Union"}], ",", "masters"}], 
        "]"}]}], "]"}], "&"}]}], "//", "First"}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   SubsuperscriptBox["B", 
    RowBox[{"0", ",", "1", ",", "1"}], "1"], "}"}], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "1", ",", "1"}], "1"], "\[Rule]", 
     RowBox[{
      RowBox[{"-", 
       FractionBox["1", "2"]}], " ", "Q2", " ", 
      SubsuperscriptBox["B", 
       RowBox[{"0", ",", "1", ",", "1"}], "1"]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "1", ",", "2"}], "1"], "\[Rule]", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{"d", "-", "2"}], ")"}], " ", 
      SubsuperscriptBox["B", 
       RowBox[{"0", ",", "1", ",", "1"}], "1"]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "1", ",", "3"}], "1"], "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}], " ", 
        SubsuperscriptBox["B", 
         RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
       RowBox[{"4", " ", "Q2"}]]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "2", ",", "1"}], "1"], "\[Rule]", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{"d", "-", "2"}], ")"}], " ", 
      SubsuperscriptBox["B", 
       RowBox[{"0", ",", "1", ",", "1"}], "1"]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "2", ",", "2"}], "1"], "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        SubsuperscriptBox["B", 
         RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
       RowBox[{"2", " ", "Q2"}]]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "3", ",", "1"}], "1"], "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}], " ", 
        SubsuperscriptBox["B", 
         RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
       RowBox[{"4", " ", "Q2"}]]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "1", ",", "2"}], "1"], "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        SubsuperscriptBox["B", 
         RowBox[{"0", ",", "1", ",", "1"}], "1"]}], "Q2"]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "1", ",", "3"}], "1"], "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"d", "-", "4"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"d", "-", "3"}], ")"}], " ", 
       SubsuperscriptBox["B", 
        RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
      RowBox[{"2", " ", 
       SuperscriptBox["Q2", "2"]}]]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "2", ",", "1"}], "1"], "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        SubsuperscriptBox["B", 
         RowBox[{"0", ",", "1", ",", "1"}], "1"]}], "Q2"]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "2", ",", "2"}], "1"], "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"d", "-", "6"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"d", "-", "3"}], ")"}], " ", 
       SubsuperscriptBox["B", 
        RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
      SuperscriptBox["Q2", "2"]]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "3", ",", "1"}], "1"], "\[Rule]", 
     FractionBox[
      RowBox[{
       RowBox[{"(", 
        RowBox[{"d", "-", "4"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"d", "-", "3"}], ")"}], " ", 
       SubsuperscriptBox["B", 
        RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
      RowBox[{"2", " ", 
       SuperscriptBox["Q2", "2"]}]]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"1", ",", "1", ",", "1"}], "1"], "\[Rule]", 
     RowBox[{"-", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        SubsuperscriptBox["B", 
         RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", "Q2"}]]}]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"1", ",", "1", ",", "2"}], "1"], "\[Rule]", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"d", "-", "3"}], ")"}], " ", 
       SubsuperscriptBox["B", 
        RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
      SuperscriptBox["Q2", "2"]]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"1", ",", "2", ",", "1"}], "1"], "\[Rule]", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{"d", "-", "3"}], ")"}], " ", 
       SubsuperscriptBox["B", 
        RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
      SuperscriptBox["Q2", "2"]]}], ",", 
    RowBox[{
     SubsuperscriptBox["B", 
      RowBox[{"2", ",", "1", ",", "1"}], "1"], "\[Rule]", 
     FractionBox[
      RowBox[{"4", " ", 
       RowBox[{"(", 
        RowBox[{"d", "-", "3"}], ")"}], " ", 
       SubsuperscriptBox["B", 
        RowBox[{"0", ",", "1", ",", "1"}], "1"]}], 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"d", "-", "6"}], ")"}], " ", 
       SuperscriptBox["Q2", "2"]}]]}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"exprb", "/.", "ibpzeromap"}], "/.", "ibpmap"}], "//", 
  "Factor"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox[
    RowBox[{"2", " ", "\[ImaginaryI]", " ", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["d", "2"], "-", 
       RowBox[{"7", " ", "d"}], "+", "16"}], ")"}], " ", 
     SuperscriptBox["gs", "2"], " ", 
     RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
     SuperscriptBox["qu", "2"], " ", 
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "1", ",", "1"}], "1"], " ", 
     SubscriptBox["C", "A"], " ", 
     SubscriptBox["C", "F"]}], 
    RowBox[{"d", "-", "4"}]]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"masters", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{"#", ",", "\[IndentingNewLine]", 
      RowBox[{"#", "//", 
       RowBox[{"FromTopologies", "[", "topologies", "]"}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"#", "//", 
          RowBox[{"FromTopologies", "[", "topologies", "]"}]}], "//", 
         "Factors"}], "//", 
        RowBox[{"ReplaceAll", "[", 
         RowBox[{
          RowBox[{"den", "[", 
           RowBox[{"p_", ",", "__"}], "]"}], "\[RuleDelayed]", "p"}], "]"}]}],
        "//", 
       RowBox[{
        RowBox[{"FindGraph", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"k1", "+", "k2"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"k1", ",", "k2"}], "}"}], ",", "#"}], "]"}], "&"}]}]}], 
     "\[IndentingNewLine]", "}"}], "&"}], "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      SubsuperscriptBox["B", 
       RowBox[{"0", ",", "1", ",", "1"}], "1"], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2]}, {
           Null, {{1, 2}, {2, 3}, {2, 3}, {3, 4}, {3, 5}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[2]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdin0slAEAh9+wrrj0j7JxviIOdbMdLUv9pGR1heiDolvJZGeH7nxNk1CT
uzbSUe1WKN86My2V0xFl1FqZFP7QfN57974J99HlbtK/Pduz55/H41J6TLIV
QRA71/3XIlGXtMfZJvREd/jdg5e1aBFNrjq6mrDsUJ0bb6/FTyIq9UfaNMAh
awZPakH8z7upvXnMQigPNKeaUlbxdsOp1hGhEVz2xT1iNYmE+M0vZgkjphU6
n+QnJCao6ihhhQH+gvk75iskti62Z3e4GxCjumBJCCbx0PGY9pdCD/Zwf1Od
A4lcvx2MjBA9VDcZUTN/1EjPGpNdHdLheccgh0OpUTnV3xZyRodKdoRFolYj
sGog1ntFC97HwIkSaxL5Wc1rX5haMOXcN0ZbEnH8iNsuPhqcXWyKzHch4Wes
XyoPI1FYypfZ7CMxvvaorJOvRlKYaFtREgmiXNVgd30BclXr18kqElJh71HX
2nnwvRXClVESaYqBceH7OfTesLCeumigkdf0R2caIe4rMl8LozChrLMzWRsx
xhpp2S2h4L7LM9FKZsD94tFD4UMUuGKuQOFlwGdii81jMwXp7/YudOrR+dq9
hLWThvjwlMAUqsdk7Lfh+jAaScGKALdPOrAbj5/3PU2D0VyzUX9OB3U2D/IE
Gk6+BYxqsxa1TsEsWc76b4kL8t+uRWHFS82yiEZkZV9MQIAG2T25w/uTaTBf
ZZZxeCSYecqMAh4NT9sG0bMUNaaLPVgD3jSk9mN93FsL2PT9gcTLSKE0aEV3
pHEe0fckS91KCsrZRFvBhzkQzm4zbTkU/gJkWihb
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.026646348233343067`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.0714652955152804`, 0.6861538745929865}, 3}], 
                 0.026646348233343067`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.0741723974113162`, 1.0130806585327994`}, 3}], 
                 0.026646348233343067`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.026646348233343067`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.026646348233343067`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.053292696466686135`, 0.053292696466686135`}], 
              {GrayLevel[0], DiskBox[2, 0.04718781515691621]}, 
              {GrayLevel[0], DiskBox[3, 0.04718781515691621]}, InsetBox[
               GraphicsBox[{}], 4, 
               Automatic, {0.053292696466686135`, 0.053292696466686135`}], 
              InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.053292696466686135`, 0.053292696466686135`}]}}]],
          
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Master evaluation", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"Taken", " ", "from", " ", 
     RowBox[{"e", ".", "g", ".", " ", 
      RowBox[{"Smirnov", "'"}]}], "s", " ", "book"}], ",", " ", "10.7"}], " ",
    "*)"}], "\[IndentingNewLine]", 
  RowBox[{"result", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"exprb", "/.", "ibpzeromap"}], "/.", "ibpmap"}], "//", 
     "Factor"}], "//", 
    RowBox[{"ReplaceAll", "[", 
     RowBox[{"{", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"B", "[", 
        RowBox[{"1", ",", 
         RowBox[{"{", 
          RowBox[{"0", ",", "1", ",", "1"}], "}"}]}], "]"}], "\[Rule]", 
       RowBox[{"I", " ", 
        RowBox[{
         RowBox[{"Pi", "^", 
          RowBox[{"(", 
           RowBox[{"d", "/", "2"}], ")"}]}], "/", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"2", "Pi"}], ")"}], "^", "d"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"-", "Q2"}], ")"}], "^", 
         RowBox[{"(", 
          RowBox[{"2", "-", 
           RowBox[{"d", "/", "2"}]}], ")"}]}], 
        RowBox[{
         RowBox[{"Gamma", "[", 
          RowBox[{
           RowBox[{"d", "/", "2"}], "-", "1"}], "]"}], "^", "2"}], 
        RowBox[{
         RowBox[{"Gamma", "[", 
          RowBox[{"2", "-", 
           RowBox[{"d", "/", "2"}]}], "]"}], "/", 
         RowBox[{"Gamma", "[", 
          RowBox[{"d", "-", "2"}], "]"}]}]}]}], "\[IndentingNewLine]", "}"}], 
     "]"}]}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{
    SuperscriptBox["2", 
     RowBox[{"1", "-", "d"}]], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["d", "2"], "-", 
      RowBox[{"7", " ", "d"}], "+", "16"}], ")"}], " ", 
    SuperscriptBox["\[Pi]", 
     RowBox[{
      RowBox[{"-", "d"}], "/", "2"}]], " ", 
    SuperscriptBox["gs", "2"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
    SuperscriptBox["qu", "2"], " ", 
    SubscriptBox["C", "A"], " ", 
    SubscriptBox["C", "F"], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"-", "Q2"}], ")"}], 
     RowBox[{"2", "-", 
      FractionBox["d", "2"]}]], " ", 
    TemplateBox[{RowBox[{"2", "-", 
        FractionBox["d", "2"]}]},
     "Gamma"], " ", 
    SuperscriptBox[
     TemplateBox[{RowBox[{
         FractionBox["d", "2"], "-", "1"}]},
      "Gamma"], "2"]}], 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"d", "-", "4"}], ")"}], " ", 
    TemplateBox[{RowBox[{"d", "-", "2"}]},
     "Gamma"]}]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Omega", "[", "k_", "]"}], ":=", 
  RowBox[{"2", 
   RowBox[{
    RowBox[{"Pi", "^", 
     RowBox[{"(", 
      RowBox[{"k", "/", "2"}], ")"}]}], "/", 
    RowBox[{"Gamma", "[", 
     RowBox[{"k", "/", "2"}], "]"}]}]}]}], "\[IndentingNewLine]", 
 RowBox[{"ps2factor", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"Q2", "^", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"d", "/", "2"}], "-", "2"}], ")"}]}], 
    RowBox[{"2", "^", 
     RowBox[{"(", 
      RowBox[{"1", "-", "d"}], ")"}]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"2", "Pi"}], ")"}], "^", 
     RowBox[{"(", 
      RowBox[{"2", "-", "d"}], ")"}]}], 
    RowBox[{"Omega", "[", 
     RowBox[{"d", "-", "1"}], "]"}]}], "//", "ExpandAll"}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{
    SuperscriptBox["4", 
     RowBox[{"2", "-", "d"}]], " ", 
    SuperscriptBox["\[Pi]", 
     RowBox[{
      FractionBox["3", "2"], "-", 
      FractionBox["d", "2"]}]], " ", 
    SuperscriptBox["Q2", 
     RowBox[{
      FractionBox["d", "2"], "-", "2"}]]}], 
   TemplateBox[{RowBox[{
       FractionBox["d", "2"], "-", 
       FractionBox["1", "2"]}]},
    "Gamma"]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"result", "*", 
     RowBox[{"ps2factor", "/", 
      RowBox[{"PS", "[", "2", "]"}]}]}], "/.", 
    RowBox[{"d", "\[Rule]", 
     RowBox[{"4", "-", 
      RowBox[{"2", "ep"}]}]}]}], "//", 
   RowBox[{
    RowBox[{"#", "+", 
     RowBox[{
      RowBox[{"O", "[", "ep", "]"}], "^", "1"}]}], "&"}]}], "//", 
  RowBox[{"MapSeries", "[", 
   RowBox[{"FunctionExpand", "/*", "FullSimplify"}], "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      RowBox[{
       SuperscriptBox["gs", "2"], " ", "Q2", " ", 
       SuperscriptBox["qu", "2"], " ", 
       SubscriptBox["C", "A"], " ", 
       SubscriptBox["C", "F"]}], 
      RowBox[{"32", " ", 
       SuperscriptBox["\[Pi]", "3"], " ", 
       SuperscriptBox["ep", "2"]}]]}], "+", 
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "2"], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"2", " ", 
          RowBox[{"log", "(", 
           RowBox[{"-", "Q2"}], ")"}]}]}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", "Q2", ")"}]}], "+", 
        RowBox[{"4", " ", 
         TagBox["\[DoubledGamma]",
          Function[{}, EulerGamma]]}], "-", "7", "-", 
        RowBox[{"4", " ", 
         RowBox[{"log", "(", "\[Pi]", ")"}]}], "-", 
        RowBox[{"8", " ", 
         RowBox[{"log", "(", "2", ")"}]}]}], ")"}]}], 
     RowBox[{"64", " ", 
      SuperscriptBox["\[Pi]", "3"], " ", "ep"}]], "-", 
    RowBox[{
     FractionBox["1", 
      RowBox[{"192", " ", 
       SuperscriptBox["\[Pi]", "3"]}]], 
     RowBox[{
      SuperscriptBox["gs", "2"], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         RowBox[{"log", "(", "Q2", ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"log", "(", "Q2", ")"}], "-", 
           RowBox[{"2", " ", 
            RowBox[{"log", "(", 
             RowBox[{
              RowBox[{"-", "16"}], " ", 
              SuperscriptBox["\[Pi]", "2"], " ", "Q2"}], ")"}]}], "+", 
           RowBox[{"4", " ", 
            TagBox["\[DoubledGamma]",
             Function[{}, EulerGamma]]}], "-", "7"}], ")"}]}], "+", 
        RowBox[{"3", " ", 
         RowBox[{"log", "(", 
          RowBox[{"-", "Q2"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"log", "(", 
            RowBox[{
             RowBox[{"-", "256"}], " ", 
             SuperscriptBox["\[Pi]", "4"], " ", "Q2"}], ")"}], "-", 
           RowBox[{"4", " ", 
            TagBox["\[DoubledGamma]",
             Function[{}, EulerGamma]]}], "+", "7"}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"33", "+", 
           RowBox[{"6", " ", 
            SuperscriptBox[
             TagBox["\[DoubledGamma]",
              Function[{}, EulerGamma]], "2"]}], "-", 
           SuperscriptBox["\[Pi]", "2"], "+", 
           RowBox[{
            RowBox[{"log", "(", "16", ")"}], " ", 
            RowBox[{"log", "(", "64", ")"}]}], "-", 
           RowBox[{"3", " ", 
            TagBox["\[DoubledGamma]",
             Function[{}, EulerGamma]], " ", 
            RowBox[{"(", 
             RowBox[{"7", "+", 
              RowBox[{"log", "(", "256", ")"}], "+", 
              RowBox[{"4", " ", 
               RowBox[{"log", "(", "\[Pi]", ")"}]}]}], ")"}]}], "+", 
           RowBox[{"21", " ", 
            RowBox[{"log", "(", 
             RowBox[{"4", " ", "\[Pi]"}], ")"}]}], "+", 
           RowBox[{"3", " ", 
            RowBox[{"log", "(", "\[Pi]", ")"}], " ", 
            RowBox[{"log", "(", 
             RowBox[{"256", " ", 
              SuperscriptBox["\[Pi]", "2"]}], ")"}]}]}], ")"}]}]}], ")"}]}]}],
     "+", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "1"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, -2, 1, 1],
     Editable->False]}],
   SeriesData[$CellContext`ep, 0, {
    Rational[-1, 32] FeynCalc`CA FeynCalc`CF $CellContext`gs^2 
     Pi^(-3) $CellContext`Q2 $CellContext`qu^2, Rational[1, 64] FeynCalc`CA 
     FeynCalc`CF $CellContext`gs^2 
     Pi^(-3) $CellContext`Q2 $CellContext`qu^2 (-7 + 4 EulerGamma - 8 Log[2] - 
      4 Log[Pi] - 2 Log[-$CellContext`Q2] + 2 Log[$CellContext`Q2]), 
     Rational[-1, 192] FeynCalc`CA FeynCalc`CF $CellContext`gs^2 
     Pi^(-3) $CellContext`Q2 $CellContext`qu^2 (
      2 (33 + 6 EulerGamma^2 - Pi^2 + Log[16] Log[64] - 3 
        EulerGamma (7 + Log[256] + 4 Log[Pi]) + 21 Log[4 Pi] + 
        3 Log[Pi] Log[256 Pi^2]) + 
      3 Log[$CellContext`Q2] (-7 + 4 EulerGamma + Log[$CellContext`Q2] - 2 
        Log[(-16) Pi^2 $CellContext`Q2]) + 
      3 Log[-$CellContext`Q2] (7 - 4 EulerGamma + 
        Log[(-256) Pi^4 $CellContext`Q2]))}, -2, 1, 1],
   Editable->False], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Put", "[", 
  RowBox[{"result", ",", "\"\<result.ps2.l1.m\>\""}], "]"}]], "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6. PS[k1,k2,k3] L[]: no reduction needed", "Section"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"externalsprules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k1"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k2"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k3", ",", "k3"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k2"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", "-", "s13"}], ")"}], " ", 
       RowBox[{"Q2", "/", "2"}]}]}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k3"}], "]"}], "\[Rule]", 
      RowBox[{"s13", " ", 
       RowBox[{"Q2", "/", "2"}]}]}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k3"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "x"}], ")"}], " ", 
       RowBox[{"Q2", "/", "2"}]}]}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{"Get", "[", "\"\<M.m\>\"", "]"}], "[", 
    RowBox[{
     RowBox[{"PS", "[", 
      RowBox[{"k1", ",", "k2", ",", "k3"}], "]"}], 
     RowBox[{"L", "[", "]"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"expr", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"expr", "//", "Topologies"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"MakePrettyTopologyGraph", "[", 
     RowBox[{"#", ",", 
      RowBox[{"{", 
       RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}]}], "]"}], "&"}], 
   "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox["7", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   GraphicsBox[
    NamespaceBox["NetworkGraphics",
     DynamicModuleBox[{Typeset`graph = HoldComplete[
       Graph[{
         $CellContext`II[1], 1, 3, 4, 2, 
         $CellContext`OO[1]}, {
        Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {3, 5}, {2, 4}, {5, 6}}}, {
        EdgeStyle -> {UndirectedEdge[2, 
             $CellContext`OO[1]] -> {
             GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
             Thickness[Large]}, UndirectedEdge[4, 3] -> {
             Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
             Thickness[Large]}, UndirectedEdge[
             $CellContext`II[1], 1] -> {
             GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
             Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
              Thickness[Large], 
              RGBColor[1, 0, 0], 
              Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
         VertexShape -> {$CellContext`II[1] -> 
           Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
         VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
         VertexStyle -> {1 -> {
             GrayLevel[0]}, 2 -> {
             GrayLevel[0]}, 3 -> {
             GrayLevel[0]}, 4 -> {
             GrayLevel[0]}}}]]}, 
      TagBox[
       GraphicsGroupBox[
        GraphicsComplexBox[{{3.814781421808071, 0.4877084622328541}, {
         2.825162389780683, 0.453296561078493}, {1.9077379198057707`, 
         0.913800986380224}, {1.9070977558442062`, 0.}, {0.9896915982171941, 
         0.4559929121833412}, {0., 0.48656376120221734`}}, {
          {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
           {GrayLevel[0.5], 
            {Arrowheads[0.], ArrowBox[{1, 2}, 0.035810735981566255`]}}, 
           {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
            {Arrowheads[0.], ArrowBox[{2, 3}, 0.035810735981566255`]}}, 
           {Thickness[Large], 
            {Arrowheads[0.], ArrowBox[{2, 4}, 0.035810735981566255`]}}, 
           {Dashing[{Small, Small}], 
            {Arrowheads[0.], ArrowBox[{3, 4}, 0.035810735981566255`]}}, 
           {Thickness[Large], 
            {Arrowheads[0.], ArrowBox[{3, 5}, 0.035810735981566255`]}}, 
           {Dashing[{Small, Small}], 
            {Arrowheads[0.], ArrowBox[{4, 5}, 0.035810735981566255`]}}, 
           {GrayLevel[0.5], 
            {Arrowheads[0.], ArrowBox[{5, 6}, 0.035810735981566255`]}}}, 
          {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
           InsetBox[
            GraphicsBox[{}], 1, 
            Automatic, {0.07162147196313251, 0.07162147196313251}], 
           {GrayLevel[0], DiskBox[2, 0.04569006053069333]}, 
           {GrayLevel[0], DiskBox[3, 0.04569006053069333]}, 
           {GrayLevel[0], DiskBox[4, 0.04569006053069333]}, 
           {GrayLevel[0], DiskBox[5, 0.04569006053069333]}, InsetBox[
            GraphicsBox[{}], 6, 
            Automatic, {0.07162147196313251, 0.07162147196313251}]}}]],
       MouseAppearanceTag["NetworkGraphics"]],
      AllowKernelInitialization->False]],
    DefaultBaseStyle->{
     "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
    
    FormatType->TraditionalForm,
    FrameTicks->None], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"expr", "//", 
    RowBox[{
     RowBox[{"Bracket", "[", 
      RowBox[{"#", ",", 
       RowBox[{"_sp", "|", "_den"}], ",", "Factor"}], "]"}], "&"}]}], "//", 
   "Terms"}], "//", "TableForm"}]], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       RowBox[{"32", " ", 
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        SuperscriptBox["qu", "2"], " ", "x", " ", 
        SubscriptBox["C", "A"], " ", 
        SubscriptBox["C", "F"], " ", 
        RowBox[{"k1", "\[CenterDot]", "k2"}], " ", 
        RowBox[{"k1", "\[CenterDot]", "k3"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}]}]},
      {
       RowBox[{"-", 
        FractionBox[
         RowBox[{"64", " ", 
          SuperscriptBox["gs", "2"], " ", 
          RowBox[{"PS", "(", "3", ")"}], " ", 
          SuperscriptBox["qu", "2"], " ", 
          SubscriptBox["C", "A"], " ", 
          SubscriptBox["C", "F"], " ", 
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{"k1", "\[CenterDot]", "k2"}], ")"}], "2"], " ", 
          RowBox[{"k1", "\[CenterDot]", "k3"}], " ", 
          SuperscriptBox[
           RowBox[{"den", "(", 
            RowBox[{
             RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], "2"]}], 
         RowBox[{"Q2", " ", "x"}]]}]},
      {
       RowBox[{"32", " ", 
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        SuperscriptBox["qu", "2"], " ", "x", " ", 
        SubscriptBox["C", "A"], " ", 
        SubscriptBox["C", "F"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"k1", "\[CenterDot]", "k3"}], ")"}], "2"], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}]}]},
      {
       RowBox[{"8", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}], " ", 
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        SuperscriptBox["qu", "2"], " ", "x", " ", 
        SubscriptBox["C", "A"], " ", 
        SubscriptBox["C", "F"], " ", 
        RowBox[{"k1", "\[CenterDot]", "k2"}], " ", 
        RowBox[{"k2", "\[CenterDot]", "k3"}], " ", 
        SuperscriptBox[
         RowBox[{"den", "(", 
          RowBox[{
           RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], "2"]}]},
      {
       RowBox[{"16", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        SuperscriptBox["qu", "2"], " ", "x", " ", 
        SubscriptBox["C", "A"], " ", 
        SubscriptBox["C", "F"], " ", 
        RowBox[{"k1", "\[CenterDot]", "k2"}], " ", 
        RowBox[{"k2", "\[CenterDot]", "k3"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}]}]},
      {
       RowBox[{"8", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "2"}], ")"}], " ", 
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        SuperscriptBox["qu", "2"], " ", "x", " ", 
        SubscriptBox["C", "A"], " ", 
        SubscriptBox["C", "F"], " ", 
        RowBox[{"k1", "\[CenterDot]", "k2"}], " ", 
        RowBox[{"k2", "\[CenterDot]", "k3"}], " ", 
        SuperscriptBox[
         RowBox[{"den", "(", 
          RowBox[{
           RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], "2"]}]},
      {
       RowBox[{"32", " ", 
        SuperscriptBox["gs", "2"], " ", 
        RowBox[{"PS", "(", "3", ")"}], " ", 
        SuperscriptBox["qu", "2"], " ", "x", " ", 
        SubscriptBox["C", "A"], " ", 
        SubscriptBox["C", "F"], " ", 
        RowBox[{"k1", "\[CenterDot]", "k3"}], " ", 
        RowBox[{"k2", "\[CenterDot]", "k3"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
        RowBox[{"den", "(", 
         RowBox[{
          RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}]}]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ps3factor", "=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"OM", ",", 
      RowBox[{"d", "=", 
       RowBox[{"4", "-", 
        RowBox[{"2", "ep"}]}]}]}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"OM", "[", "k_", "]"}], ":=", 
      RowBox[{"2", 
       RowBox[{
        RowBox[{"Pi", "^", 
         RowBox[{"(", 
          RowBox[{"k", "/", "2"}], ")"}]}], "/", 
        RowBox[{"Gamma", "[", 
         RowBox[{"k", "/", "2"}], "]"}]}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Q2", "^", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"d", "/", "2"}]}], ")"}]}], 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"2", "Pi"}], ")"}], "^", 
         RowBox[{"(", 
          RowBox[{"3", "-", 
           RowBox[{"3", "d"}]}], ")"}]}], "/", 
        RowBox[{"2", "^", 
         RowBox[{"(", 
          RowBox[{"1", "+", "d"}], ")"}]}]}], 
       RowBox[{"OM", "[", 
        RowBox[{"d", "-", "1"}], "]"}], 
       RowBox[{"OM", "[", 
        RowBox[{"d", "-", "2"}], "]"}], 
       RowBox[{"Q2", "^", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"3", 
           RowBox[{"d", "/", "2"}]}], "-", "6"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"x", "-", "s13"}], ")"}], " ", "s13", " ", 
          RowBox[{"(", 
           RowBox[{"1", "-", "x"}], ")"}]}], ")"}], "^", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"d", "/", "2"}], "-", "2"}], ")"}]}], 
       RowBox[{"Q2", "^", "2"}]}], "//", "FunctionExpand"}]}]}], 
   "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{
    SuperscriptBox["2", 
     RowBox[{
      RowBox[{"6", " ", "ep"}], "-", "11"}]], " ", 
    SuperscriptBox["\[Pi]", 
     RowBox[{
      RowBox[{"4", " ", "ep"}], "-", "7"}]], " ", 
    SuperscriptBox["Q2", 
     RowBox[{"1", "-", 
      RowBox[{"2", " ", "ep"}]}]], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{"s13", " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "x"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{"x", "-", "s13"}], ")"}]}], ")"}], 
     RowBox[{"-", "ep"}]]}], 
   TemplateBox[{RowBox[{"2", "-", 
       RowBox[{"2", " ", "ep"}]}]},
    "Gamma"]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"exprint", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"expr", "/.", 
          RowBox[{
           RowBox[{"den", "[", 
            RowBox[{"p_", ",", "m_"}], "]"}], "\[RuleDelayed]", 
           RowBox[{"1", "/", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"sp", "[", 
               RowBox[{"p", ",", "p"}], "]"}], "-", "m"}], ")"}]}]}]}], "//", 
         "ExpandSP"}], "//", 
        RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}], "//", 
       RowBox[{"ReplaceAll", "[", 
        RowBox[{"d", "\[Rule]", 
         RowBox[{"4", "-", 
          RowBox[{"2", "ep"}]}]}], "]"}]}], "//", 
      RowBox[{
       RowBox[{"#", "*", 
        RowBox[{"ps3factor", "/", 
         RowBox[{"PS", "[", "3", "]"}]}]}], "&"}]}], "//", "Factor"}], "//", 
    "PowerExpand"}], "//", 
   RowBox[{
    RowBox[{"Integrate", "[", 
     RowBox[{"#", ",", "s13"}], "]"}], "&"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"ep", "-", "2"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"ep", "-", "1"}], ")"}], " ", "x", " ", 
      TemplateBox[{RowBox[{"2", "-", 
          RowBox[{"2", " ", "ep"}]}]},
       "Gamma"]}]]}], 
   SuperscriptBox["8", 
    RowBox[{
     RowBox[{"2", " ", "ep"}], "-", "3"}]], " ", 
   SuperscriptBox["\[Pi]", 
    RowBox[{
     RowBox[{"4", " ", "ep"}], "-", "7"}]], " ", 
   SuperscriptBox["gs", "2"], " ", 
   SuperscriptBox["qu", "2"], " ", 
   SubscriptBox["C", "A"], " ", 
   SubscriptBox["C", "F"], " ", 
   SuperscriptBox["Q2", 
    RowBox[{"1", "-", 
     RowBox[{"2", " ", "ep"}]}]], " ", 
   SuperscriptBox["s13", 
    RowBox[{"1", "-", "ep"}]], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", "x"}], ")"}], 
    RowBox[{
     RowBox[{"-", "ep"}], "-", "1"}]], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", 
      FractionBox["s13", "x"]}], ")"}], "ep"], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"x", "-", "s13"}], ")"}], 
    RowBox[{"-", "ep"}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["ep", "2"], "-", "ep", "-", "2"}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{"x", "-", "2"}], ")"}], " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{
       RowBox[{"1", "-", "ep"}],"ep",RowBox[{"2", "-", "ep"}],FractionBox[
        "s13", "x"]},
       "Hypergeometric2F1"]}], "-", 
     RowBox[{"x", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox["ep", "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"x", "-", "1"}], ")"}], "2"]}], "+", 
        RowBox[{"ep", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"3", " ", 
             SuperscriptBox["x", "2"]}]}], "+", 
           RowBox[{"4", " ", "x"}], "-", "3"}], ")"}]}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           SuperscriptBox["x", "2"], "+", "1"}], ")"}]}]}], ")"}], " ", 
      TemplateBox[{
       RowBox[{"1", "-", "ep"}],RowBox[{"ep", "+", "1"}],
        RowBox[{"2", "-", "ep"}],FractionBox["s13", "x"]},
       "Hypergeometric2F1"]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"ep", "-", "1"}], ")"}], " ", "s13", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"ep", "-", "1"}], ")"}], " ", 
         SuperscriptBox["x", "2"]}], "+", 
        RowBox[{"2", " ", "x"}], "-", "2"}], ")"}], " ", 
      TemplateBox[{
       RowBox[{"2", "-", "ep"}],"ep",RowBox[{"3", "-", "ep"}],FractionBox[
        "s13", "x"]},
       "Hypergeometric2F1"]}]}], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Assuming", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], "<", "ep", "<", "1"}], "}"}], ",", 
   RowBox[{"Limit", "[", 
    RowBox[{"exprint", ",", 
     RowBox[{"s13", "\[Rule]", "0"}]}], "]"}]}], "]"}]], "Input"],

Cell[BoxData[
 FormBox["0", TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"result", "=", 
   RowBox[{"Assuming", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], "<", "ep", "<", "1"}], "}"}], ",", 
     RowBox[{"Limit", "[", 
      RowBox[{"exprint", ",", 
       RowBox[{"s13", "\[Rule]", "x"}]}], "]"}]}], "]"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"result", "//", "Factor"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   FractionBox["1", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"ep", "-", "1"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"x", "-", "1"}], ")"}], " ", 
     TemplateBox[{RowBox[{"2", "-", 
         RowBox[{"2", " ", "ep"}]}]},
      "Gamma"], " ", 
     TemplateBox[{RowBox[{
         FractionBox["1", "2"], "-", "ep"}]},
      "Gamma"], " ", 
     TemplateBox[{RowBox[{
         FractionBox["3", "2"], "-", "ep"}]},
      "Gamma"]}]], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"-", "1"}], ")"}], 
    RowBox[{"-", "ep"}]], " ", 
   SuperscriptBox["2", 
    RowBox[{
     RowBox[{"8", " ", "ep"}], "-", "11"}]], " ", 
   SuperscriptBox["\[Pi]", 
    RowBox[{
     RowBox[{"4", " ", "ep"}], "-", 
     FractionBox["13", "2"]}]], " ", 
   SuperscriptBox["gs", "2"], " ", 
   SuperscriptBox["qu", "2"], " ", 
   SubscriptBox["C", "A"], " ", 
   SubscriptBox["C", "F"], " ", 
   SuperscriptBox["Q2", 
    RowBox[{"1", "-", 
     RowBox[{"2", " ", "ep"}]}]], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"1", "-", "x"}], ")"}], 
    RowBox[{"-", "ep"}]], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"-", 
      FractionBox["1", "x"]}], ")"}], "ep"], " ", 
   SuperscriptBox["x", 
    RowBox[{"1", "-", "ep"}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      SuperscriptBox["ep", "3"], " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"2", " ", 
      SuperscriptBox["ep", "2"], " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"4", " ", 
      SuperscriptBox["ep", "2"], " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["ep", "2"], " ", "x", " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"8", " ", 
      SuperscriptBox["ep", "2"], " ", "x", " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"2", " ", 
      SuperscriptBox["ep", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"4", " ", 
      SuperscriptBox["ep", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"2", " ", "ep", " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"2", "-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"ep", " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"8", " ", "ep", " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"2", "-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"4", " ", 
      SuperscriptBox["x", "2"], " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"4", " ", "ep", " ", "x", " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"2", "-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"2", " ", "ep", " ", "x", " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"8", " ", "ep", " ", "x", " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"4", " ", "x", " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"2", "-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"2", " ", "ep", " ", 
      TemplateBox[{RowBox[{
          FractionBox["1", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "-", 
     RowBox[{"8", " ", "ep", " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}], "+", 
     RowBox[{"4", " ", 
      TemplateBox[{RowBox[{
          FractionBox["3", "2"], "-", "ep"}]},
       "Gamma"], " ", 
      TemplateBox[{RowBox[{"-", "ep"}]},
       "Gamma"]}]}], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"result", "+", 
   RowBox[{"O", "[", "ep", "]"}]}], "//", 
  RowBox[{"MapSeries", "[", 
   RowBox[{"FunctionExpand", "/*", "PowerExpand", "/*", "Factor"}], 
   "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "2"], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["x", "2"], "+", "1"}], ")"}], " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"]}], 
     RowBox[{"512", " ", 
      SuperscriptBox["\[Pi]", "7"], " ", "ep", " ", 
      RowBox[{"(", 
       RowBox[{"x", "-", "1"}], ")"}]}]], "-", 
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "2"], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         SuperscriptBox["x", "2"], " ", 
         RowBox[{"log", "(", "Q2", ")"}]}], "+", 
        RowBox[{"4", " ", 
         RowBox[{"log", "(", "Q2", ")"}]}], "+", 
        RowBox[{"4", " ", 
         TagBox["\[DoubledGamma]",
          Function[{}, EulerGamma]], " ", 
         SuperscriptBox["x", "2"]}], "-", 
        SuperscriptBox["x", "2"], "+", 
        RowBox[{"2", " ", 
         SuperscriptBox["x", "2"], " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", "x"}], ")"}]}], "+", 
        RowBox[{"4", " ", 
         SuperscriptBox["x", "2"], " ", 
         RowBox[{"log", "(", "x", ")"}]}], "-", 
        RowBox[{"8", " ", 
         SuperscriptBox["x", "2"], " ", 
         RowBox[{"log", "(", "\[Pi]", ")"}]}], "-", 
        RowBox[{"12", " ", 
         SuperscriptBox["x", "2"], " ", 
         RowBox[{"log", "(", "2", ")"}]}], "-", 
        RowBox[{"6", " ", "x"}], "+", 
        RowBox[{"2", " ", 
         RowBox[{"log", "(", 
          RowBox[{"1", "-", "x"}], ")"}]}], "+", 
        RowBox[{"4", " ", 
         RowBox[{"log", "(", "x", ")"}]}], "+", 
        RowBox[{"4", " ", 
         TagBox["\[DoubledGamma]",
          Function[{}, EulerGamma]]}], "-", "4", "-", 
        RowBox[{"8", " ", 
         RowBox[{"log", "(", "\[Pi]", ")"}]}], "-", 
        RowBox[{"12", " ", 
         RowBox[{"log", "(", "2", ")"}]}]}], ")"}]}], 
     RowBox[{"1024", " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Pi]", "7"], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}]}], ")"}]}]], "+", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "1"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, -1, 1, 1],
     Editable->False]}],
   SeriesData[$CellContext`ep, 0, {
    Rational[1, 512] FeynCalc`CA FeynCalc`CF $CellContext`gs^2 
     Pi^(-7) $CellContext`Q2 $CellContext`qu^2 (-1 + $CellContext`x)^(-1) \
$CellContext`x (1 + $CellContext`x^2), Rational[-1, 1024] FeynCalc`CA 
     FeynCalc`CF $CellContext`gs^2 
     Pi^(-7) $CellContext`Q2 $CellContext`qu^2 (-1 + $CellContext`x)^(-1) \
$CellContext`x (-4 + 4 EulerGamma - 6 $CellContext`x - $CellContext`x^2 + 
      4 EulerGamma $CellContext`x^2 - 12 Log[2] - 12 $CellContext`x^2 Log[2] - 
      8 Log[Pi] - 8 $CellContext`x^2 Log[Pi] + 4 Log[$CellContext`Q2] + 
      4 $CellContext`x^2 Log[$CellContext`Q2] + 2 Log[1 - $CellContext`x] + 
      2 $CellContext`x^2 Log[1 - $CellContext`x] + 4 Log[$CellContext`x] + 
      4 $CellContext`x^2 Log[$CellContext`x])}, -1, 1, 1],
   Editable->False], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"result", "+", 
      RowBox[{"O", "[", "ep", "]"}]}], "//", 
     RowBox[{"MapSeries", "[", 
      RowBox[{"FunctionExpand", "/*", "PowerExpand", "/*", "Factor"}], 
      "]"}]}], ",", "\"\<result.ps3.l0.m\>\""}], "]"}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["7. IBP reduction for PS[k1, k2] L[l1, l2], LiteRed", "Section"],

Cell[CellGroupData[{

Cell["Topologies and the B notation", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"externalsprules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k1"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k2"}], "]"}], "\[Rule]", 
      RowBox[{"Q2", "/", "2"}]}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k2"}], "]"}], "\[Rule]", "0"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Get", "[", "\"\<M.m\>\"", "]"}], "[", 
     RowBox[{
      RowBox[{"PS", "[", 
       RowBox[{"k1", ",", "k2"}], "]"}], 
      RowBox[{"L", "[", 
       RowBox[{"l1", ",", "l2"}], "]"}]}], "]"}], "//", 
    RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"expr", "//", "Topologies"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"expr", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", "TableForm"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"MakePrettyTopologyGraph", "[", 
     RowBox[{"#", ",", 
      RowBox[{"{", 
       RowBox[{"k1", ",", "k2"}], "}"}]}], "]"}], "&"}], "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox["201", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  InterpretationBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "l1", "+", "l2"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   TableForm[{{
      $CellContext`den[$CellContext`k1 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 - $CellContext`l1 - $CellContext`l2, 
       0], 
      $CellContext`den[$CellContext`l2, 0]}, {
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 - $CellContext`l2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l2, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`l1 - \
$CellContext`l2, 0], 
      $CellContext`den[$CellContext`l2, 0]}, {
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k2 - $CellContext`l2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l2, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`l1 - \
$CellContext`l2, 0], 
      $CellContext`den[$CellContext`l2, 0]}, {
      $CellContext`den[$CellContext`k1 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k2 - $CellContext`l2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1 - \
$CellContext`l2, 0], 
      $CellContext`den[$CellContext`l2, 0]}, {
      $CellContext`den[$CellContext`k2 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 - $CellContext`l2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1 - \
$CellContext`l2, 0], 
      $CellContext`den[$CellContext`l2, 0]}, {
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`k1 - $CellContext`l2, 0], 
      $CellContext`den[$CellContext`l2, 0], 
      $CellContext`den[$CellContext`l1 + $CellContext`l2, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`l1 + $CellContext`l2, 
       0]}, {
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`k2 - $CellContext`l2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l2, 
       0], 
      $CellContext`den[$CellContext`l2, 0], 
      $CellContext`den[$CellContext`l1 + $CellContext`l2, 0]}}]], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 5, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {3, 5}, {4, 6}, {3, 6}, {5, 4}, {5, 
          4}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[3, 5] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAKMVgjURFC0Ak5XRK27D7P6iYea/Y
ZgRAEDCkz5759T/W+c9O3jj9P5QH2Az6s+U/QvkUKE/j+T+pqYOXAl/1P+RM
cS0aQPk/AAAAAAAAAAB8sdmVtmjqP0b6rGVUX/U/AAAAAAAAAAAEt1Wc4KL8
P9+NUNJOnPo/K67dOCkX8z/0+SCYh8f6PxIf8K1sgvI/eq/M7ozr+j8Fg70y
yOvxP6Aqi4JJCPs/7feV+JRT8T9hGatPrB37P4+p4BwtuvA/QUilrKgr+z94
2sBz6x/wP+zhm1E2Mvs/wXadpFYK7z9ukUFdUTH7P0PQB7GQ1O0/SBkA+Qjw
+j8Xkdaig3fZP6on5PKx5/o/GSy1LNsM1z/GOCLL7df6P/VwYHOEpNQ/1dZ1
18XA+j9LphBj7D/SP+OHs81Hovo/C27GYPvAzz/Ysqq7hXz6P9bGNANAD8s/
lCJ1/JVP+j8uVFgzbm3GP0J7OiuTG/o/LPo8C0TewT/myksq1gf5P2MKHBY6
I/M/kiMRWdPT+D8inxjRVJHyP06T25njpvg/zRAdlxr98T9EvtKHIYH4P+fb
aivjZvE/UW8QfqNi+D8VgL9+B8/wP2ANZIp7S/g/ao2reuE18D99HqJitzv4
P0O9rJiXN+8/3yyGXGAz+D/DCpxdQwLuP7u0RPgX8vc/GQb/++jS2T89ZOoD
M/H3Px250xRdZ9c/6P3gqMD39z+9PAuPXPzUP8ks2wW9Bfg/ZQCM6lWT0j+I
G/vSHxv4P+rGtnu2LdA/r5a5Ztw3+D8aNTEm05nLPzZMZb3hW/g/qlScTK/k
xj9KuDWDGof4P+TbL/XKPsI/M8499A==
          "], {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03310124558093995]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03310124558093995]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.03310124558093995]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03310124558093995]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03310124558093995]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               4, {1.817715508625982, 0.6612957005613861}, 5}], 
               0.03310124558093995]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               4, {1.37842986765586, 0.6744001305971647}, 5}], 
               0.03310124558093995]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03310124558093995]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03310124558093995]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0662024911618799, 0.0662024911618799}], 
            {GrayLevel[0], DiskBox[2, 0.03448552094633272]}, 
            {GrayLevel[0], DiskBox[3, 0.03448552094633272]}, 
            {GrayLevel[0], DiskBox[4, 0.03448552094633272]}, 
            {GrayLevel[0], DiskBox[5, 0.03448552094633272]}, 
            {GrayLevel[0], DiskBox[6, 0.03448552094633272]}, InsetBox[
             GraphicsBox[{}], 7, 
             Automatic, {0.0662024911618799, 0.0662024911618799}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 5, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {5, 4}, {3, 4}, {5, 6}, {4, 5}, {3, 
          6}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Thickness[Large]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[3, 4] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[4, 5] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAA/E05uD4zDUB2tkrqZ1T0PwXH685H
5QVAixDEe8vb7z+a/Nlh/7/9P1r9omt6MvQ/s2cDn4OaAUDgdpS4p2jOPw5+
YsIIyfI/AAAAAAAAAADJyEoVCKPrP6n+tjdxFvA/AAAAAAAAAAATRd9v/T33
P+pCa5p/zgBAVDqwF0hXxj/RfXYcIpoAQKynG1v0YMQ/+n5ZtVFkAEA0K19y
fZPCP+7U1EEuLQBA0Pn2lvTvwD9eMM/OsOn/P9D04F2k7r4/eTP2AeN2/z+Q
ka9z6lS8P+5PRhY3Av8/wHdemUUUuj/BNv0f8ov+P8g2fjoLLrg/n3y7IVn6
9z8AWEf8EtdCv2BwRifBgvc/gMQYZ+IFe7+qXcz3HAr3P8BU1sTQ9Ia/JxtB
AbSQ9j8A4LJPeH2Nv7ajGibOFvY/ALCRvYSMkL8YVcGSs5z1P2Dju7mw4pC/
Q2bUkqwi9T+AuBIuGoKPv3ziW2YBqfQ/wETUpYpQir98tQZNh6oAQJbdeDHY
BtA/mHPKtrFtAEA10bqtZDDQPy380zauMABApvnl535C0D+8qU7aQef/P3BW
IygcPdA/TDIo/1tt/z9v0seePyDQP8nvnAjz8/4/KtzhxPTXzz8S3SLZTnv+
PwA9zcvWQM8/09Ct3rYD/j82vpDLfnvOP7MWbOAdcvc/eFtVG6JRwj+G/SLq
2Pv2P/w65euEXsE/+hlz/iyH9j8Urrx+Mj7APxUdmjFfFPY/7PhHE6vivT+Y
o798s6P1PxD6OkNm8bo/fk+2lWw19T9Ml2qMVKq3P9NRfMfLyfQ/YJ7xumYP
tD+gx5LLEGH0Pwh5yEG/IrA/vb47Nw==
          "], {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.0347263707657203]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.0347263707657203]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               4, {1.7263294841374197`, -0.04999193827856571}, 5}], 
               0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               4, {1.6481974949339981`, 0.2875607603066587}, 5}], 
               0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.0347263707657203]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.0347263707657203]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0694527415314406, 0.0694527415314406}], 
            {GrayLevel[0], DiskBox[2, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[3, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[4, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[5, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[6, 0.04586111767247925]}, InsetBox[
             GraphicsBox[{}], 7, 
             Automatic, {0.0694527415314406, 0.0694527415314406}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 5, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {5, 3}, {4, 3}, {5, 6}, {3, 5}, {4, 
          6}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAAwKGOZlUzDUD4HFcBpFjHP9WdSE8r
5QVAPSFs0qFC3T8p4Xv8dpoBQLL6QvsIcvM/RtLd+Wq//T/mVvX7ynHIPyiW
wMvTyPI/9ATbc4RA9z+gGh+qH6HrP0srvLIRqdw/AAAAAAAAAAAAAAAAAAAA
AClN15NyqgBAAhPovIw98z9XpcQOm20AQLKpZgczM/M/EZ522pUwAECyeTYP
ti7zPywt5y8O5/8/uhWgfBgw8z9jkdTbJW3/P1zsyX1ZN/M/YoSU5rrz/j+S
hzTHdETzPzliyzMVe/4/yIpFlmJX8z8SU1IyfAP+P+Lt37UXcPM/aEE8INhx
9z+Wi+my4vX0P+jr1Z+S+/Y/rFoQglAU9T/6WEGn5ob2P3wMkPxkOPU/6tvD
ShkU9j/uTqnFCmL1PyOLQoNuo/U/o/i2NCmR9T8WgQMGKTX1P4+ix2OkxfU/
NrkuHYrJ9D+URCJAXf/1P9PrJYHRYPQ/dg2snDE+9j9UNskheM4AQDLycdJb
dPQ/os/E0xuaAEAWu/suMLP0P7NrWl9MZABAGl1WC+ns9D+s5rqgKS0AQAYH
ZzpkIfU/kXz0eajp/z+6sHSpglD1P4D/dh3bdv8/LPONcih69T+SbOIkLwL/
P/ykDe08nvU/Ehd8pOmL/j8SdDS8qrz1P2gFZpJF+vc/xhE+uXVC9z9A9uyQ
rIL3P+F02NgqW/c/GNQj3gYK9z8WeOmnGG73PxbH4+ibkPY/TBNU8TN79z9N
K9GUsxb2P+7pffJ0gvc/VhzLD5ac9T/1hedf14P3P8wNL6eLIvU/9lW3Z1p/
9z8nvgmd3Kj0P6bsNbIAdfc/zCE0hA==
          "], {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03472665818833397]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03472665818833397]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.03472665818833397]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03472665818833397]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               3, {1.6481039685042613`, 1.165515264345253}, 5}], 
               0.03472665818833397]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               3, {1.7263484016032933`, 1.5030766489276046`}, 5}], 
               0.03472665818833397]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03472665818833397]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03472665818833397]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03472665818833397]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06945331637666793, 0.06945331637666793}], 
            {GrayLevel[0], DiskBox[2, 0.045857674204805216`]}, 
            {GrayLevel[0], DiskBox[3, 0.045857674204805216`]}, 
            {GrayLevel[0], DiskBox[4, 0.045857674204805216`]}, 
            {GrayLevel[0], DiskBox[5, 0.045857674204805216`]}, 
            {GrayLevel[0], DiskBox[6, 0.045857674204805216`]}, InsetBox[
             GraphicsBox[{}], 7, 
             Automatic, {0.06945331637666793, 0.06945331637666793}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 5, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {3, 5}, {6, 7}, {3, 7}, {4, 5}, {5, 
          6}, {4, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Thickness[Large]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
              Thickness[Large]}, UndirectedEdge[4, 5] -> {
              Thickness[Large]}, UndirectedEdge[3, 5] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.82775950748957, 0.34102566225772063`}, {
          2.857279458924138, 0.3642453300352074}, {1.9008765210392695`, 0.}, {
          2.4882573319258685`, 1.450220578471928}, {1.9176099356458205`, 
          1.0859896894096428`}, {1.3410753682205372`, 1.4692240547736586`}, {
          0.9581671951634747, 0.4009557497035342}, {0., 
          0.32507801395474695`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358952851820334]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358952851820334]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.0358952851820334]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358952851820334]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358952851820334]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358952851820334]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358952851820334]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358952851820334]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.0358952851820334]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358952851820334]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0717905703640668, 0.0717905703640668}], 
            {GrayLevel[0], DiskBox[2, 0.033849024780202346`]}, 
            {GrayLevel[0], DiskBox[3, 0.033849024780202346`]}, 
            {GrayLevel[0], DiskBox[4, 0.033849024780202346`]}, 
            {GrayLevel[0], DiskBox[5, 0.033849024780202346`]}, 
            {GrayLevel[0], DiskBox[6, 0.033849024780202346`]}, 
            {GrayLevel[0], DiskBox[7, 0.033849024780202346`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.0717905703640668, 0.0717905703640668}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 5, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {4, 5}, {6, 7}, {4, 7}, {3, 5}, {5, 
          6}, {3, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
              Thickness[Large]}, UndirectedEdge[3, 6] -> {
              Thickness[Large]}, UndirectedEdge[4, 5] -> {
              Thickness[Large]}, UndirectedEdge[3, 5] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.374022807750645, 1.877601037910519}, {
          2.6200116887705454`, 1.2861863965154487`}, {2.317945018414745, 
          0.31556980190522654`}, {1.6954182340426815`, 1.3509692689724844`}, {
          1.6888887135352744`, 0.}, {1.0670885557293825`, 
          0.32525130093028776`}, {0.7660058071489272, 1.2967119800808657`}, {
          0., 1.8657165026996045`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03286298885040069]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03286298885040069]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.03286298885040069]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03286298885040069]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03286298885040069]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03286298885040069]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03286298885040069]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03286298885040069]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03286298885040069]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03286298885040069]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06572597770080138, 0.06572597770080138}], 
            {GrayLevel[0], DiskBox[2, 0.035086459104769915`]}, 
            {GrayLevel[0], DiskBox[3, 0.035086459104769915`]}, 
            {GrayLevel[0], DiskBox[4, 0.035086459104769915`]}, 
            {GrayLevel[0], DiskBox[5, 0.035086459104769915`]}, 
            {GrayLevel[0], DiskBox[6, 0.035086459104769915`]}, 
            {GrayLevel[0], DiskBox[7, 0.035086459104769915`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.06572597770080138, 0.06572597770080138}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 2, 5, 6, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {5, 6}, {7, 5}, {3, 7}, {3, 6}, {6, 
          4}, {4, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[3, 6] -> {
              Thickness[Large]}, UndirectedEdge[3, 5] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[2, 5] -> {
              Thickness[Large]}, UndirectedEdge[4, 6] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.274196709203682, 0.5136246102169043}, {
          3.3670820276555293`, 0.5053864523625655}, {2.530169512788247, 0.}, {
          2.5304984937516943`, 1.0131836739104354`}, {0.9071271193738002, 
          0.5077856682990879}, {1.743514219500912, 0.0006579432279606445}, {
          1.7442270282497168`, 1.0137999850147033`}, {0., 
          0.5122562334524448}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.038729380658446846`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.038729380658446846`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.038729380658446846`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.038729380658446846`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.038729380658446846`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.038729380658446846`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.038729380658446846`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.038729380658446846`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.038729380658446846`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{5, 8}, 0.038729380658446846`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07745876131689369, 0.07745876131689369}], 
            {GrayLevel[0], DiskBox[2, 0.03931358535233097]}, 
            {GrayLevel[0], DiskBox[3, 0.03931358535233097]}, 
            {GrayLevel[0], DiskBox[4, 0.03931358535233097]}, 
            {GrayLevel[0], DiskBox[5, 0.03931358535233097]}, 
            {GrayLevel[0], DiskBox[6, 0.03931358535233097]}, 
            {GrayLevel[0], DiskBox[7, 0.03931358535233097]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07745876131689369, 0.07745876131689369}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 2, 5, 6, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {5, 6}, {7, 5}, {4, 3}, {3, 7}, {4, 
          6}, {6, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 6] -> {
              Thickness[Large]}, UndirectedEdge[4, 5] -> {
              Thickness[Large]}, UndirectedEdge[3, 6] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[2, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.598366806140282, 0.5457743356918155}, {
          3.633548263833968, 0.5250650316016542}, {2.8217992036599027`, 0.}, {
          2.821231543079123, 1.0517221476263738`}, {0.9648178152185232, 
          0.5256291579082695}, {1.7773076412298578`, 1.0519135692253971`}, {
          1.7764749771256565`, 0.0005493950476227472}, {0., 
          0.5453390276145662}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070062802933476]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070062802933476]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.04070062802933476]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.04070062802933476]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.04070062802933476]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070062802933476]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.04070062802933476]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070062802933476]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{5, 8}, 0.04070062802933476]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.04070062802933476]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.08140125605866952, 0.08140125605866952}], 
            {GrayLevel[0], DiskBox[2, 0.04825095583319784]}, 
            {GrayLevel[0], DiskBox[3, 0.04825095583319784]}, 
            {GrayLevel[0], DiskBox[4, 0.04825095583319784]}, 
            {GrayLevel[0], DiskBox[5, 0.04825095583319784]}, 
            {GrayLevel[0], DiskBox[6, 0.04825095583319784]}, 
            {GrayLevel[0], DiskBox[7, 0.04825095583319784]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.08140125605866952, 0.08140125605866952}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"topologies", "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"CompleteDenominatorSet", "[", 
       RowBox[{"#", ",", 
        RowBox[{"{", 
         RowBox[{"k1", ",", "k2"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"l1", ",", "l2"}], "}"}]}], "]"}], "&"}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0", ",", "irred"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "l2"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0", ",", "irred"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0", ",", "irred"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "l1", "+", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l2", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"exprb", "=", 
   RowBox[{
    RowBox[{"expr", "//", 
     RowBox[{
      RowBox[{"ToTopologies", "[", 
       RowBox[{"#", ",", "topologies", ",", 
        RowBox[{"{", 
         RowBox[{"k1", ",", "k2"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"l1", ",", "l2"}], "}"}]}], "]"}], "&"}]}], "//", 
    RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"exprb", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"exprb", "//", 
    RowBox[{
     RowBox[{"Cases", "[", 
      RowBox[{"#", ",", "_B", ",", 
       RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}], "//", 
  "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["3036", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox["469", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprb", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"8", " ", 
   SuperscriptBox["gs", "4"], " ", 
   RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
   SuperscriptBox["qu", "2"], " ", "x", " ", 
   SubsuperscriptBox["B", 
    RowBox[{"1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", ",", 
     RowBox[{"-", "1"}]}], "4"], " ", 
   SubscriptBox["C", "A"], " ", 
   SubsuperscriptBox["C", "F", "2"]}], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["IBP reduction via LiteRed", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SetAttributes", "[", 
   RowBox[{"RunInATemporaryKernel", ",", 
    RowBox[{"{", "HoldFirst", "}"}]}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RunInATemporaryKernel", "[", "script_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"kern", ",", "result"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"{", "kern", "}"}], "=", 
      RowBox[{"LaunchKernels", "[", "1", "]"}]}], ";", "\[IndentingNewLine]", 
     
     RowBox[{"result", "=", 
      RowBox[{
       RowBox[{"ParallelEvaluate", "[", 
        RowBox[{"script", ",", 
         RowBox[{"{", "kern", "}"}]}], "]"}], "//", "First"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"CloseKernels", "[", "kern", "]"}], ";", "\[IndentingNewLine]", 
     "result"}]}], "\[IndentingNewLine]", "]"}]}]}], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"topologynames", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"topologies", "//", "Length"}], "//", "Range"}], "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"ToExpression", "[", 
       RowBox[{"\"\<ps2l2t\>\"", "<>", 
        RowBox[{"ToString", "[", "#", "]"}]}], "]"}], "&"}], "]"}]}]}], ";", 
  RowBox[{"exprj", "=", 
   RowBox[{"exprb", "/.", 
    RowBox[{
     RowBox[{"B", "[", 
      RowBox[{"topid_", ",", "indices_List"}], "]"}], "\[RuleDelayed]", 
     RowBox[{"j", "[", 
      RowBox[{
       RowBox[{"topologynames", "[", 
        RowBox[{"[", "topid", "]"}], "]"}], ",", 
       RowBox[{"Sequence", "@@", "indices"}]}], "]"}]}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprj", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  FractionBox[
   RowBox[{"56", " ", 
    SuperscriptBox["gs", "4"], " ", 
    RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
    SuperscriptBox["qu", "2"], " ", "x", " ", 
    SubsuperscriptBox["C", "A", "2"], " ", 
    SubscriptBox["C", "F"], " ", 
    RowBox[{"j", "(", 
     RowBox[{
     "ps2l2t6", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", 
      ",", "0"}], ")"}]}], 
   RowBox[{"d", "-", "2"}]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
    "We", " ", "could", " ", "have", " ", "called", " ", "LiteRed", " ", 
     "inline"}], ",", " ", 
    RowBox[{
    "but", " ", "then", " ", "it", " ", "would", " ", "pollute", " ", "the", 
     " ", "global", " ", "name", " ", "space", " ", "with", " ", "its", " ", 
     RowBox[{"definitions", "."}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"Put", "[", 
    RowBox[{
     RowBox[{"<|", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"\"\<topology-denominators\>\"", "\[Rule]", "topologies"}], "/.", 
        RowBox[{
         RowBox[{"den", "[", 
          RowBox[{"p_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
         RowBox[{
          RowBox[{"sp", "[", "p", "]"}], "-", "m2"}]}]}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"\"\<topology-names\>\"", "\[Rule]", "topologynames"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"\"\<topology-loop-momenta\>\"", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"l1", ",", "l2"}], "}"}]}], ",", "\[IndentingNewLine]", 
       RowBox[{"\"\<expression\>\"", "\[Rule]", "exprj"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"\"\<external-sp-rules\>\"", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"k1", ",", "k1"}], "]"}], "\[Rule]", "0"}], ",", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"k2", ",", "k2"}], "]"}], "\[Rule]", "0"}], ",", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"k1", ",", "k2"}], "]"}], "\[Rule]", 
           RowBox[{"Q2", "/", "2"}]}]}], "}"}]}], ",", "\[IndentingNewLine]", 
       
       RowBox[{
       "\"\<output-file\>\"", "\[Rule]", 
        "\"\<ps2.l2.litered.result.m\>\""}]}], "\[IndentingNewLine]", "|>"}], 
     ",", "\"\<ps2.l2.litered.job.m\>\""}], "]"}], ";"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"RunInATemporaryKernel", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"JOBFILE", "=", "\"\<ps2.l2.litered.job.m\>\""}], ";", 
    RowBox[{"Get", "[", "\"\<litered-ibpreduce.m\>\"", "]"}]}], 
   "\[IndentingNewLine]", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expribp", "=", 
   RowBox[{
    RowBox[{"Get", "[", "\"\<ps2.l2.litered.result.m\>\"", "]"}], "/.", 
    RowBox[{
     RowBox[{"j", "[", 
      RowBox[{"topo_", ",", "indices__"}], "]"}], "\[RuleDelayed]", 
     RowBox[{"B", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Position", "[", 
         RowBox[{"topologynames", ",", "topo"}], "]"}], "[", 
        RowBox[{"[", 
         RowBox[{"1", ",", "1"}], "]"}], "]"}], ",", 
       RowBox[{"{", "indices", "}"}]}], "]"}]}]}]}], ";"}]}], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"**************** \"\>", "\[InvisibleSpace]", 
    StyleBox["\<\"LiteRed v1.82\"\>",
     StripOnInput->False,
     LineColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     FrontFaceColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     BackFaceColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     GraphicsColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     FontWeight->Bold,
     FontColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]]], 
    "\[InvisibleSpace]", "\<\" ********************\\nAuthor: Roman N. Lee, \
Budker Institute of Nuclear Physics, Novosibirsk.\\nRelease Date: \
01.06.2015\\nLiteRed stands for Loop InTEgrals REDuction.\\nThe package is \
designed for the search and application of the Integration-By-Part reduction \
rules. It also contains some other useful tools.\\nSee ?LiteRed`* for a list \
of functions. \"\>"}],
   SequenceForm["**************** ", 
    Style["LiteRed v1.82", {Bold, 
      RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]]}], 
    " ********************\nAuthor: Roman N. Lee, Budker Institute of Nuclear \
Physics, Novosibirsk.\nRelease Date: 01.06.2015\nLiteRed stands for Loop \
InTEgrals REDuction.\nThe package is designed for the search and application \
of the Integration-By-Part reduction rules. It also contains some other \
useful tools.\nSee ?LiteRed`* for a list of functions. "],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"SetSharedFunction", "::", "subnopar"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Parallel programming is not available in subkernels of \
another parallel computation.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"SetSharedFunction", "::", "subnopar"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Parallel programming is not available in subkernels of \
another parallel computation.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t1", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t1, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t1] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t1::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t2", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t2, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t2] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t2::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t3", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t3, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t3] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t3::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t4", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t4, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t4] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t4::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t5", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t5, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t5] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t5::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t6", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t6, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t6] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t6::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the definitions for \"\>", "\[InvisibleSpace]", 
    "ps2l2t7", "\[InvisibleSpace]", "\<\" from disk\"\>"}],
   SequenceForm[
   "* Loading the definitions for ", $CellContext`ps2l2t7, " from disk"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"2-loop basis with 2 external momenta.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Use Information[ps2l2t7] to get basis summary.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"This string can be redefined by setting ps2l2t7::usage=\\\"\
\[Ellipsis]\\\"\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t1", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t1, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{
     "ps2l2t1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", 
      ",", "0"}], ")"}], "\[InvisibleSpace]", "\<\" = \"\>", 
    "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox["l1", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "-", "l1", "-", "l2"}], ")"}], "2"], " ", 
      SuperscriptBox["l2", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps2l2t1, 0, 1, 1, 1, 1, 0, 0], " = ", 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, \
$CellContext`k1 + $CellContext`k2 - $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`l1, $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 - $CellContext`l1 - $CellContext`l2, \
$CellContext`k1 - $CellContext`l1 - $CellContext`l2]^(-1)/
    Vectors`sp[$CellContext`l2, $CellContext`l2]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t2", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t2, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{
     "ps2l2t2", ",", "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "0", 
      ",", "0"}], ")"}], "\[InvisibleSpace]", "\<\" = \"\>", 
    "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox["l1", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ")"}], "2"], " ", 
      SuperscriptBox["l2", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps2l2t2, 1, 0, 0, 1, 1, 0, 0], " = ", 
    Vectors`sp[$CellContext`l1, $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`l1 - \
$CellContext`l2, $CellContext`k1 + $CellContext`k2 + $CellContext`l1 - \
$CellContext`l2]^(-1)/Vectors`sp[$CellContext`l2, $CellContext`l2]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t3", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t3, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t4", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t4, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t5", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t5, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{
     "ps2l2t5", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", 
      ",", "0"}], ")"}], "\[InvisibleSpace]", "\<\" = \"\>", 
    "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox["l1", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "-", "l2"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ")"}], "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps2l2t5, 0, 1, 1, 1, 1, 0, 0], " = ", 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, \
$CellContext`k1 + $CellContext`k2 - $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`l1, $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 - $CellContext`l2, $CellContext`k1 - \
$CellContext`l2]^(-1)/
    Vectors`sp[$CellContext`k1 + $CellContext`k2 - $CellContext`l1 - \
$CellContext`l2, $CellContext`k1 + $CellContext`k2 - $CellContext`l1 - \
$CellContext`l2]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t6", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t6, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{
     "ps2l2t6", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", 
      ",", "0"}], ")"}], "\[InvisibleSpace]", "\<\" = \"\>", 
    "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox["l1", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "-", "l2"}], ")"}], "2"], " ", 
      SuperscriptBox["l2", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"l1", "+", "l2"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k2", "+", "l1", "+", "l2"}], ")"}], "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps2l2t6, 1, 1, 1, 1, 1, 1, 0], " = ", 
    Vectors`sp[$CellContext`l1, $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`l1, \
$CellContext`k1 + $CellContext`k2 + $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 - $CellContext`l2, $CellContext`k1 - \
$CellContext`l2]^(-1) Vectors`sp[$CellContext`l2, $CellContext`l2]^(-1) 
    Vectors`sp[$CellContext`l1 + $CellContext`l2, $CellContext`l1 + \
$CellContext`l2]^(-1)/
    Vectors`sp[$CellContext`k2 + $CellContext`l1 + $CellContext`l2, \
$CellContext`k2 + $CellContext`l1 + $CellContext`l2]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps2l2t7", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps2l2t7, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{
     "ps2l2t7", ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "0", 
      ",", "0"}], ")"}], "\[InvisibleSpace]", "\<\" = \"\>", 
    "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox["l1", "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "-", "l2"}], ")"}], "2"], " ", 
      SuperscriptBox["l2", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps2l2t7, 1, 1, 0, 1, 1, 0, 0], " = ", 
    Vectors`sp[$CellContext`l1, $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`l1, \
$CellContext`k1 + $CellContext`k2 + $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 - $CellContext`l2, \
$CellContext`k1 + $CellContext`k2 - $CellContext`l2]^(-1)/
    Vectors`sp[$CellContext`l2, $CellContext`l2]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"* IBPReduce\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"FrontEndObject", "::", "notavail"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"A front end is not available; certain operations require \
a front end.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"MIs: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"j", "(", 
       RowBox[{
       "ps2l2t6", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", 
        ",", "0"}], ")"}], ",", 
      RowBox[{"j", "(", 
       RowBox[{
       "ps2l2t7", ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "0", 
        ",", "0"}], ")"}], ",", 
      RowBox[{"j", "(", 
       RowBox[{
       "ps2l2t5", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", 
        ",", "0"}], ")"}], ",", 
      RowBox[{"j", "(", 
       RowBox[{
       "ps2l2t1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", 
        ",", "0"}], ")"}], ",", 
      RowBox[{"j", "(", 
       RowBox[{
       "ps2l2t2", ",", "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "0", 
        ",", "0"}], ")"}]}], "}"}]}],
   SequenceForm["MIs: ", {
     LiteRed`j[$CellContext`ps2l2t6, 1, 1, 1, 1, 1, 1, 0], 
     LiteRed`j[$CellContext`ps2l2t7, 1, 1, 0, 1, 1, 0, 0], 
     LiteRed`j[$CellContext`ps2l2t5, 0, 1, 1, 1, 1, 0, 0], 
     LiteRed`j[$CellContext`ps2l2t1, 0, 1, 1, 1, 1, 0, 0], 
     LiteRed`j[$CellContext`ps2l2t2, 1, 0, 0, 1, 1, 0, 0]}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"FrontEndObject", "::", "notavail"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"A front end is not available; certain operations require \
a front end.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Save to \"\>", 
    "\[InvisibleSpace]", "\<\"ps2.l2.litered.result.m\"\>"}],
   SequenceForm["* Save to ", "ps2.l2.litered.result.m"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"* Done\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"masters", "=", 
  RowBox[{
   RowBox[{"expribp", "//", 
    RowBox[{
     RowBox[{"Cases", "[", 
      RowBox[{"#", ",", "_B", ",", 
       RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    SubsuperscriptBox["B", 
     RowBox[{
     "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", ",", "0"}], "1"], 
    ",", 
    SubsuperscriptBox["B", 
     RowBox[{
     "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "0", ",", "0"}], "2"], 
    ",", 
    SubsuperscriptBox["B", 
     RowBox[{
     "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", ",", "0"}], "5"], 
    ",", 
    SubsuperscriptBox["B", 
     RowBox[{
     "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0"}], "6"], 
    ",", 
    SubsuperscriptBox["B", 
     RowBox[{
     "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "0", ",", "0"}], 
     "7"]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"masters", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"#", "//", "InputForm"}], ",", "\[IndentingNewLine]", 
      RowBox[{"#", "//", 
       RowBox[{"FromTopologies", "[", "topologies", "]"}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"#", "//", 
        RowBox[{"FromTopologies", "[", "topologies", "]"}]}], "//", 
       RowBox[{
        RowBox[{"FindGraph", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"k1", "+", "k2"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"k1", ",", "k2"}], "}"}], ",", 
          RowBox[{"Cases", "[", 
           RowBox[{"#", ",", 
            RowBox[{
             RowBox[{"den", "[", 
              RowBox[{"p_", ",", "__"}], "]"}], "\[RuleDelayed]", "p"}], ",", 
            
            RowBox[{"{", 
             RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}]}], "]"}], 
        "&"}]}]}], "\[IndentingNewLine]", "}"}], "&"}], "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"1", ",", " ", 
         RowBox[{"{", 
          RowBox[{
          "0", ",", " ", "1", ",", " ", "1", ",", " ", "1", ",", " ", "1", 
           ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {3, 4}, {3, 4}, {4, 5}, {3, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 
                $CellContext`OO[2]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 3] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdxW0s1AEAx/Fbt5KLrciKWXGjoiVJNpx+VxpiGjqiLkm4ZOdphZ0lK90L
eSjuaNnJnKe8sOtmFzsPbbRWTlbSMC7hPHUt9/S/O/c/V73tu3329UzPS8jc
QaFQjvz17/+Xykn8KndTQxlUOyiNX4B9VqiL3bgVuWzbNCVlAfX5h2oC1g2I
Fh8tPzVuQAz3/EjOmgF0yTtW0S4qs5d7TjqzuZNJ0UWJowrzcCdbtThURKLP
yqxoaFvEbugl254kBtannlwKW8Jph+Zs6mcLyGKqwDa/BNfX923kIws6ZXEl
0fxl5LNLHV1DLeD4+wjGzqggcVDIHxJb4O6fs7qpVZg4nNGRLttCsoA+quha
gdpJParkbaGWZu9XkLMKrwO8Y5sJRvTG/mgO8NEgQjr55ri7ESnxHP6Fag1e
/mZ4R24QSOOJFmZ1GnSb3C03Bgk8Z+3d7GNrwWjkTzwVErjsLRW1vNei9R4v
UFNIYMalWZQQqEP3A+0rYSIBeXCdoKZVB3Pn0LPaMAIFTNpZZyc92vya9nWX
WOFFD7keNrEEXRLHLE6zou6krySiZxmWtO+5UeFWsIqdmocqVUjD5GSXhxU3
kzdE9IwVaOcZXTMmElPmE2ZN8CqsK+5l2jESkT3OB3v2rCHJi//T8QUJFqVc
MTu9hpLh77yLt0iI7KqatlvWIRnhKq+GmsFQtCaaw7X4JDOQ32wmXLFrqIxp
1+LXh+pV2bAJsZH93DaqDr48pdCj1IS4/jq6IlWHCu85MSPABH+iTFIl0+H2
R3lwkMqIx8J2upSmR31jpjZEYEQ7e6CDdk2PAO7dtzwY0fmF353Vqccfo+07
Jw==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.02836475564592497]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.02836475564592497]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.02836475564592497]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.4759349373872694`, 0.8556408993098347}, 4}], 
                 0.02836475564592497]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.7560536243284057`, 1.017819798273857}, 4}], 
                 0.02836475564592497]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.02836475564592497]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.02836475564592497]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.05672951129184994, 0.05672951129184994}], 
              {GrayLevel[0], DiskBox[2, 0.049198926075657456`]}, 
              {GrayLevel[0], DiskBox[3, 0.049198926075657456`]}, 
              {GrayLevel[0], DiskBox[4, 0.049198926075657456`]}, InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.05672951129184994, 0.05672951129184994}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.05672951129184994, 0.05672951129184994}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"2", ",", " ", 
         RowBox[{"{", 
          RowBox[{
          "1", ",", " ", "0", ",", " ", "0", ",", " ", "1", ",", " ", "1", 
           ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1", "-", "l2"}], ",", "0"}], ")"}], 
       " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2]}, {
           Null, {{1, 2}, {3, 2}, {2, 3}, {2, 3}, {3, 4}, {3, 5}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[2, 1] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[2]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdin0slAEAh9+wrrj0j7JxviIOdbMdLUv9pGR1heiDolvJZGeH7nxNk1CT
uzbSUe1WKN86My2V0xFl1FqZFP7QfN57974J99HlbtK/Pduz55/H41J6TLIV
QRA71/3XIlGXtMfZJvREd/jdg5e1aBFNrjq6mrDsUJ0bb6/FTyIq9UfaNMAh
awZPakH8z7upvXnMQigPNKeaUlbxdsOp1hGhEVz2xT1iNYmE+M0vZgkjphU6
n+QnJCao6ihhhQH+gvk75iskti62Z3e4GxCjumBJCCbx0PGY9pdCD/Zwf1Od
A4lcvx2MjBA9VDcZUTN/1EjPGpNdHdLheccgh0OpUTnV3xZyRodKdoRFolYj
sGog1ntFC97HwIkSaxL5Wc1rX5haMOXcN0ZbEnH8iNsuPhqcXWyKzHch4Wes
XyoPI1FYypfZ7CMxvvaorJOvRlKYaFtREgmiXNVgd30BclXr18kqElJh71HX
2nnwvRXClVESaYqBceH7OfTesLCeumigkdf0R2caIe4rMl8LozChrLMzWRsx
xhpp2S2h4L7LM9FKZsD94tFD4UMUuGKuQOFlwGdii81jMwXp7/YudOrR+dq9
hLWThvjwlMAUqsdk7Lfh+jAaScGKALdPOrAbj5/3PU2D0VyzUX9OB3U2D/IE
Gk6+BYxqsxa1TsEsWc76b4kL8t+uRWHFS82yiEZkZV9MQIAG2T25w/uTaTBf
ZZZxeCSYecqMAh4NT9sG0bMUNaaLPVgD3jSk9mN93FsL2PT9gcTLSKE0aEV3
pHEe0fckS91KCsrZRFvBhzkQzm4zbTkU/gJkWihb
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.026646348233343067`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.0714652955152804`, 0.6861538745929865}, 3}], 
                 0.026646348233343067`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.026646348233343067`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.0741723974113162`, 1.0130806585327994`}, 3}], 
                 0.026646348233343067`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.026646348233343067`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.026646348233343067`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.053292696466686135`, 0.053292696466686135`}], 
              {GrayLevel[0], DiskBox[2, 0.04718781515691621]}, 
              {GrayLevel[0], DiskBox[3, 0.04718781515691621]}, InsetBox[
               GraphicsBox[{}], 4, 
               Automatic, {0.053292696466686135`, 0.053292696466686135`}], 
              InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.053292696466686135`, 0.053292696466686135`}]}}]],
          
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"5", ",", " ", 
         RowBox[{"{", 
          RowBox[{
          "0", ",", " ", "1", ",", " ", "1", ",", " ", "1", ",", " ", "1", 
           ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1", "-", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 3, 
            $CellContext`OO[1], 
            $CellContext`OO[2]}, {
           Null, {{1, 2}, {2, 3}, {2, 4}, {4, 3}, {3, 4}, {3, 5}, {4, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 
                $CellContext`OO[2]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJw9jnlIkwEAxedFacss6dKZmeZFiCiirvItj/5QWGowS0beNWhYaVaoIaQu
NTUxo8ga6kpFMzTNIyeYiVtl4hVoTUXSLY/020p0+461IHrwePx4PHguyVdi
08xZLJa7yX/zcMeMIujkBlj/9Ekd6FZ6YwNi+661ihAC4vTguSjpClrKNC2v
SRq6S3SDZRWNiImUAy1y+v9uJFefUhphwTsB0dvmOSteo9OD4bBpc55Dj5lX
4+Ya/LJPvQ8rYZDo1uTgMb4OH+nSYHszg27vOu5IHIHP29pDpxUMtrfNlA4o
CNguJlbKZhkYj3/gvgrSImRXRYAVwcCuuk3Eb9IiOrZj/hfJoKTHcjSBo0Nn
pK9EZG7EC6W7RlquQ2RIQXSChRHFoQ6pF2a2IOm/M++1zKCsKlly94we4q37
qFtksDPXkxP3To+p8PU56TfTn1vKgWI/AxpvxrceHWbQYyy+rqozQDDk9RVd
DHg8fit7Dwmxh0auf8ogM42omcoj0TBvUCffZpBkxQs3rpA4LMtil59jUOnf
US9YW8Nk35eesQQaVzPP/pQNrYOfFz1HcWksz0ZpnGIIFAWnWl90pJHhbOfK
7ycgfeKy6mxGI7N+bLLGVwv+ErGXu0rhkOxazkStFhyOUN2voiAtCdwts9eh
SNQr6RunUKvK8x8u1CG9RrkQMEKhUJDZem9yC01HFFmsN6Y+7aHi42k9ePsc
eZsmdiz44aPq1uO57aOFy3IKx76PCnO9DeiM6a2IV1JwTZfbiKoN0OQs+8un
KWT83hQKd5DwPP84X0JQINkRNvuzSSgGksxesmk8Cz2YL9CQGFRbW3B8aPwB
l6M8tg==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.028302669793239324`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.028302669793239324`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.028302669793239324`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.358385386192534, 2.0283234811938304`}, 4}], 
                 0.028302669793239324`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {1.3533251547653695`, 1.7041897315724497`}, 4}], 
                 0.028302669793239324`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.028302669793239324`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.028302669793239324`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.05660533958647865, 0.05660533958647865}], 
              {GrayLevel[0], DiskBox[2, 0.049146698858177566`]}, 
              {GrayLevel[0], DiskBox[3, 0.049146698858177566`]}, 
              {GrayLevel[0], DiskBox[4, 0.049146698858177566`]}, InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.05660533958647865, 0.05660533958647865}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.05660533958647865, 0.05660533958647865}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"6", ",", " ", 
         RowBox[{"{", 
          RowBox[{
          "1", ",", " ", "1", ",", " ", "1", ",", " ", "1", ",", " ", "1", 
           ",", " ", "1", ",", " ", "0"}], "}"}]}], "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"l1", "+", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "l1", "+", "l2"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 5, 4, 2, 3, 
            $CellContext`OO[1], 
            $CellContext`OO[2]}, {
           Null, {{1, 2}, {3, 2}, {2, 4}, {4, 5}, {3, 5}, {6, 3}, {4, 6}, {5, 
            7}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[5, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 
                $CellContext`OO[2]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[3, 5] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[5, 1] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}, 5 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{1.6649597986768259`, 2.769192289750192}, {
            1.7630627110852506`, 1.795856642030814}, {1.661281867926384, 
            1.1373810513289846`}, {1.7748088592199869`, 0.5396003636332672}, {
            0.8392881066780625, 0.3676263734944357}, {2.6188394988279633`, 
            0.42189776170155924`}, {0., 0.}, {3.4620853418687254`, 
            0.04498353017925516}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03346404529752639]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 3}, 0.03346404529752639]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.03346404529752639]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.03346404529752639]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.03346404529752639]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03346404529752639]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03346404529752639]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{5, 7}, 0.03346404529752639]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{6, 8}, 0.03346404529752639]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06692809059505278, 0.06692809059505278}], 
              {GrayLevel[0], DiskBox[2, 0.030423269397541027`]}, 
              {GrayLevel[0], DiskBox[3, 0.030423269397541027`]}, 
              {GrayLevel[0], DiskBox[4, 0.030423269397541027`]}, 
              {GrayLevel[0], DiskBox[5, 0.030423269397541027`]}, 
              {GrayLevel[0], DiskBox[6, 0.030423269397541027`]}, InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.06692809059505278, 0.06692809059505278}], 
              InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.06692809059505278, 0.06692809059505278}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"7", ",", " ", 
         RowBox[{"{", 
          RowBox[{
          "1", ",", " ", "1", ",", " ", "0", ",", " ", "1", ",", " ", "1", 
           ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l2"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l2", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2]}, {
           Null, {{1, 2}, {3, 2}, {2, 3}, {3, 4}, {3, 4}, {4, 5}, {4, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 1] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[2]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 3] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJw1kn9UjXccx5+n3dsvzUQqKarFlNq62rSb9P7WbX6kKGEUI65VpjKp02jq
MGzSj42SuMau8mtF5YRpNuQuwpVQmuqms+p5bs+j52bKvZe7uz98znmfz3md
9zmffz4vj7Vpi9dbUBQ1x5z/99uJcej0721lMfqlC1QiHrovLiVuOzGApLs0
81r6Bq0FKU1qvRZXq7w+meVkRSa25znH7dXC5fK1cyvG2RGX2BJFxQoDqn3r
0vOFUeTtXfndCtmLY4NYJlikFtVwKMxZXVRiK2C2o0/gPYHD4VWdqqMZAnbo
NXTzVB4RAYY9d7sFLHT2+uqbGB7SRd6qZwt1KKvwk1R9zWPiDLlWUa/DhafJ
YSv38GiyTTH97jOE+8HXGjN/4mETH3aPPjSEiJkzJMNmTni26RLOvcLonkqd
8yIeyl3MyHo3PbZGSoQ34eY+ZGW6X54e6xL7cqP9eTzvDZ62dESPabH5mgn2
PMTJpZlH5AY0WM3OS+njoK5ZHnZDbUDI+9l+8loOP9BBLnulRuwQHStxzeAw
zTp+Vc1xIxqfFFZc8eUQQT4+W1U2CAuHkn2Fm7Vg5izlH1sImMfsOvKhjxbK
4pkqxUYB+26fFvdpWaR2vfyxoVWAVW//yY6LLHYmLNaMCtdBWVa6WFLAwt42
oDigWoek3SE1+lQW6uvl7xkmDSHm+IvQyDhz/+WkaI/8IVy1zSIzolkkc5FB
q5WvkHtxiF5ix+LkbE8bpb0ehvPyDQ/N/LnjdLcbOXo42tcP8k4spK/Wh23n
9JDMHaEVfiy8xx/I3R9nwLeP6MbmKBat7PnQ6yoDFgyFS7Zmsehp/CjrtsSI
wlaDIbvS7M+2/oaNh43IMHre/G2AhacgUm3WmlB/0N2915VDVZ/T0X99KVJT
8LpneI2ZWY2gk1PEt9LRb1mZ2YeYzsAiBUVkGuGdrkYOd9I62LOPKbLlU/+/
Mwc4ZGjuB3mNoYnj+M6frcU8pigeubfOp8luOXdz11ge5XnL03/ZSZPrxQ+i
ex14hGZQcXsuiMiZ3LNprPl/0pwRzTxeRNrtnId9uziclnpv6PxATER31CeX
NXGQu63tmJIgJqdKSkn6rxzyY6uiXx8SkwrRnFHFOzikeya7yh6ISUbiP80P
ozhMdpr+/JGNJTEmjbkfM5pD4uTvug8QS1JnSj0maRiAy0BYUUSfCcHJe2Uj
f7JYsn1rkNyHIi9N/ePaMllQ6cHvDidQJJt2GLs0kEVf6HBbcBlFIuW1bQcs
WIzJkkU6tFBkwYM1yro2Bk3x2X/E29Fk8Ih3cMtlBrMcK+cJn9FkwsruAWM5
gwD3cnVNDk3m90dtgYKB2iaq7uB5EZG0TN0v28Rg0Yjaeh0rIikiu0knNjMw
XXjIj/cSk6exE8M9chmcmtIYUrBKTHL8b1nXljJ4IvtLOF0iJtXfW2ki6hnc
PNMQOFctJreuPKtrYc1cmZ8XbmVJZO0Gk8TL7IOrUpUbYkmkzW03ZEks/gPG
USL0
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.03378477028029864]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.417026337472927, 1.0411180601685286`}, 3}], 
                 0.03378477028029864]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.410214638323126, 0.7283269775766625}, 3}], 
                 0.03378477028029864]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {2.3900082230283752`, 1.0314805483863017`}, 4}], 
                 0.03378477028029864]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 3, {2.385493338485248, 0.7035247781332522}, 4}], 
                 0.03378477028029864]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{4, 5}, 0.03378477028029864]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.03378477028029864]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.06756954056059727, 0.06756954056059727}], 
              {GrayLevel[0], DiskBox[2, 0.04698830824134602]}, 
              {GrayLevel[0], DiskBox[3, 0.04698830824134602]}, 
              {GrayLevel[0], DiskBox[4, 0.04698830824134602]}, InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.06756954056059727, 0.06756954056059727}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.06756954056059727, 0.06756954056059727}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{
    RowBox[{"<|", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"\"\<topologies\>\"", "\[Rule]", "topologies"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"\"\<masters\>\"", "\[Rule]", "masters"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"\"\<expression\>\"", "\[Rule]", "expribp"}]}], 
     "\[IndentingNewLine]", "|>"}], ",", "\"\<reduced.ps2.l2.m\>\""}], "]"}], 
  ";"}]], "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8. IBP reduction for PS[k1,k2,k3] L[l1], LiteRed", "Section"],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{
   RowBox[{
    RowBox[{"We", "'"}], "ll", " ", "IBP", " ", "reduce", " ", "the", " ", 
    "loop", " ", "integration"}], ",", " ", 
   RowBox[{
   "but", " ", "leave", " ", "the", " ", "phase", " ", "space", " ", 
    "integration", " ", 
    RowBox[{"intact", ".", " ", "It"}], " ", "can", " ", "be", " ", 
    "resolved", " ", "later"}], ",", " ", 
   RowBox[{
    RowBox[{"because", " ", "3"}], "-", 
    RowBox[{"particle", " ", "phase", " ", "space", " ", "is", " ", 
     RowBox[{"simple", "."}]}]}]}], " ", "*)"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"externalsprules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k1"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k2"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k3", ",", "k3"}], "]"}], "\[Rule]", "0"}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k2"}], "]"}], "\[Rule]", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"1", "-", "s13", "-", "s23"}], ")"}], " ", 
       RowBox[{"Q2", "/", "2"}]}]}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k1", ",", "k3"}], "]"}], "\[Rule]", 
      RowBox[{"s13", " ", 
       RowBox[{"Q2", "/", "2"}]}]}], ",", 
     RowBox[{
      RowBox[{"sp", "[", 
       RowBox[{"k2", ",", "k3"}], "]"}], "\[Rule]", 
      RowBox[{"s23", " ", 
       RowBox[{"Q2", "/", "2"}]}]}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", "\"\<M.m\>\"", "]"}], "[", 
      RowBox[{
       RowBox[{"PS", "[", 
        RowBox[{"k1", ",", "k2", ",", "k3"}], "]"}], 
       RowBox[{"L", "[", "l1", "]"}]}], "]"}], "//", 
     RowBox[{"ReplaceAll", "[", 
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"den", "[", 
         RowBox[{
          RowBox[{"p_", "/;", 
           RowBox[{"FreeQ", "[", 
            RowBox[{"p", ",", "l1"}], "]"}]}], ",", "m2_", ",", "___"}], 
         "]"}], "\[RuleDelayed]", 
        RowBox[{"1", "/", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"p", ",", "p"}], "]"}], "-", "m2"}], ")"}]}]}], 
       "\[IndentingNewLine]", "}"}], "]"}]}], "//", 
    RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"expr", "//", "Topologies"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"expr", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", "TableForm"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"MakePrettyTopologyGraph", "[", 
     RowBox[{"#", ",", 
      RowBox[{"{", 
       RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}]}], "]"}], "&"}], 
   "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox["463", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  InterpretationBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], "\<\"\"\>", "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0"}], ")"}], "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "l1"}], ",", "0"}], ")"}], "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   TableForm[{{
      $CellContext`den[$CellContext`k2 + $CellContext`k3 - $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`l1, 0]}, {
      $CellContext`den[$CellContext`k2 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`l1, 0]}, {
      $CellContext`den[$CellContext`k3 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`l1, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`l1, 0]}, {
      $CellContext`den[$CellContext`k1 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1, 0], 
      $CellContext`den[$CellContext`l1, 0]}, {
      $CellContext`den[$CellContext`k2 - $CellContext`l1, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1, 0], 
      $CellContext`den[$CellContext`l1, 0]}}]], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 2, 3, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {2, 4}, {4, 3}, {4, 3}, {3, 5}}}, {
         EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 2] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 3] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdkH80FAYAx22VhzXcrsyPq6HpF24esR/U10RFr5lrM3liyo+tEibmmaTL
zo9yqj06ZMjWpiQO582JO3TOryn9WJxuOvdTOw7n7nB3rf2773uf9/nr88/X
6VgyJf5NExOToNf8Z+ZQfmeqwNQ/NSkvoeoNNSyKil5QBo0gcNu/DMl/CY5h
LSFRoMSRNI/zDvSXuE+ZGiBPaGDyvx2sO/vU43XvvDuOv5huRG5iM7vHTI5z
f7W3So8aYRaQf6tmmwy9XneDv/c3gvx8ZzAhWIp44t9q1iYjEhobjjKSJNC2
1AZ3LBmQ9cR2or1UjMsOcjWDb4CzjFYyx51GHc37blKZAa9iTok6F0SwDBxx
C48xoJcenVuxXQR6v7Mqy3MFfdUdwu6yYTyeSh29plmGhZWJb3jeACi6rh/X
s5Zh2lNiLyzk4c6wa8irlGXQSt+as6/vBWd0wyvGtmWwCNXznGkOMnwm7dZN
6OBxPflTeUYnup+SC1MLdHi/pIVOP96OLquxcksvHSL4QZsyLjNBHGEQ0lIN
WGtmddXXWwGCqZ+zNckAt9AXF55VyBF1k34le1CPopHCQ0k6Gd55O7xBnq3H
DX2oXeRnMrhYI5bqpUejQF2+pkqKXPFtl/TZVdAfEEfHxBIYVnUCxZ1VHLqn
5e3aKkHgxgNs0+9WsdnOPLoiVoweRlWOa5AO5XHMWzH3xvBBuLeeaKmDU0+Z
1H7HQwRwqSy5QIuOxF7bkxtHIdg+ENvfqEV44w+zwbwhnBq4yBqkaXEjxG2q
P4EPy3p2rEWcFrklXp5iXR8Or2vRVOzTwl7PbbiZw0VTtel0MVmLzGHPB35q
Nkod9v3++T4VIi5F7LTyVUDE0y/VVqpQoK8bPVUjB3sbM85ZqcIuHxqhzyDD
eqGaWfvJPBzN2vakH5bh65kDxvkL85Bs2eMuqZVCWkwRyvjzOGH7yJitkMCP
VZgTab6AwY+KRPU7JYi+l8DaErSAyyEfP56MF+OP2ws+BSlLyBg7eCaFM4Zn
DrSavN+WMMhZzhlyfYhWm9N3fhpfQibR6ZKN/Sh+1pI6rplq0PStdVvI4BDy
ePHvZZA1yN76a/KOE3zY2QYGvBumAZ9CjWau9oG05mltwmkNmo+QapW5XCSL
z01TaBqctirEEy0byRP3n2++MItvwjDuaCUHscouVsucxUyUwnbKXQZS2m7H
MuUsfhGWKMihUgTarMQ/85xD+479LlfSJCC2XvQopc5hJuKfRwWVYkjTFcFV
k3PIJG8XlvZP48+2yeajjirMX23heutEmHVJcYj8SgVSkcX+aXcRQjboK7uo
i9hcN9myUjOMEw7jj/K7FuExw8jiFQ+Au9f6zIpxEa0n082zrvJwQGb3hWqv
GvzdA6HjTb24vphU7V+sxmSD0rx8hgNlfZOye0INR0XYWTa1EzyLhCh/t9f/
sY5fSk9qB5FFunXl/BL0H3azjjGY+Beg+j2N
          "], {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.02920626822107371]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.02920626822107371]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               2, {1.8177939716803648`, 0.3485487818549059}, 4}], 
               0.02920626822107371]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               2, {1.5348687533861491`, 0.5117161559442401}, 4}], 
               0.02920626822107371]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               3, {1.322249361298571, 0.5119853193234354}, 4}], 
               0.02920626822107371]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               3, {1.0390504603837425`, 0.3491117802885821}, 4}], 
               0.02920626822107371]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.02920626822107371]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.05841253644214742, 0.05841253644214742}], 
            {GrayLevel[0], DiskBox[2, 0.04956816263236978]}, 
            {GrayLevel[0], DiskBox[3, 0.04956816263236978]}, 
            {GrayLevel[0], DiskBox[4, 0.04956816263236978]}, InsetBox[
             GraphicsBox[{}], 5, 
             Automatic, {0.05841253644214742, 0.05841253644214742}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {2, 5}, {4, 5}, {4, 3}, {3, 5}, {5, 
          6}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{2.787211297380792, 1.4121275865846938`}, {
          1.8810348058035467`, 0.8602220240217593}, {1.7298213341240785`, 
          0.}, {1.0515015239254766`, 0.011683175959522396`}, {
          0.9208154431153655, 0.8777390321240717}, {0., 
          1.4075710614726358`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.028693857686064472`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.028693857686064472`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.028693857686064472`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.028693857686064472`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.028693857686064472`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.028693857686064472`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.028693857686064472`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.028693857686064472`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.057387715372128945`, 0.057387715372128945`}], 
            {GrayLevel[0], DiskBox[2, 0.03392102082442274]}, 
            {GrayLevel[0], DiskBox[3, 0.03392102082442274]}, 
            {GrayLevel[0], DiskBox[4, 0.03392102082442274]}, 
            {GrayLevel[0], DiskBox[5, 0.03392102082442274]}, InsetBox[
             GraphicsBox[{}], 6, 
             Automatic, {0.057387715372128945`, 0.057387715372128945`}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 2, 3, 4, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {2, 4}, {2, 5}, {5, 3}, {5, 4}, {4, 3}, {3, 
          6}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 2] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{2.822819072046311, 1.3716639222126714`}, {
          1.8892004662918367`, 0.8673845744815409}, {0.9336128192702793, 
          0.8674145120768175}, {1.0736812633121395`, 
          0.00001134504651911783}, {1.7490938277877257`, 0.}, {0., 
          1.3716738339968648`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.028955280495466626`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.028955280495466626`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.028955280495466626`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.028955280495466626`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.028955280495466626`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.028955280495466626`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.028955280495466626`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.028955280495466626`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.05791056099093325, 0.05791056099093325}], 
            {GrayLevel[0], DiskBox[2, 0.03377062822854344]}, 
            {GrayLevel[0], DiskBox[3, 0.03377062822854344]}, 
            {GrayLevel[0], DiskBox[4, 0.03377062822854344]}, 
            {GrayLevel[0], DiskBox[5, 0.03377062822854344]}, InsetBox[
             GraphicsBox[{}], 6, 
             Automatic, {0.05791056099093325, 0.05791056099093325}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 2, 4, 5, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 4}, {6, 4}, {5, 3}, {6, 5}, {2, 6}, {2, 
          3}, {4, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {
              Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.586217995715037, 0.912528793619743}, {
          2.656479969295118, 0.6881629973857788}, {1.719388911187862, 
          1.1262808267537263`}, {0.9226725624057439, 0.7249287345585744}, {
          1.135476513566023, 0.0014242319546774462`}, {1.8760274209539214`, 
          0.}, {0., 1.0819939273958494`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03430089216185016]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03430089216185016]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03430089216185016]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03430089216185016]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03430089216185016]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03430089216185016]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03430089216185016]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03430089216185016]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03430089216185016]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06860178432370032, 0.06860178432370032}], 
            {GrayLevel[0], DiskBox[2, 0.03702761384661792]}, 
            {GrayLevel[0], DiskBox[3, 0.03702761384661792]}, 
            {GrayLevel[0], DiskBox[4, 0.03702761384661792]}, 
            {GrayLevel[0], DiskBox[5, 0.03702761384661792]}, 
            {GrayLevel[0], DiskBox[6, 0.03702761384661792]}, InsetBox[
             GraphicsBox[{}], 7, 
             Automatic, {0.06860178432370032, 0.06860178432370032}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 4, 2, 3, 5, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 4}, {6, 4}, {3, 5}, {6, 3}, {2, 6}, {2, 
          5}, {4, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[1, 3] -> {
              Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.614586969208987, 0.2801901952532879}, {
          2.6708146755016107`, 0.42431688999994477`}, {1.1633673360135584`, 
          1.1341432243800158`}, {0.9411673256710975, 0.4139830562124596}, {
          1.8999179334021181`, 1.1331643305395331`}, {1.7323097985831406`, 
          0.}, {0., 0.11241670898640505`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03449045861985295]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.03449045861985295]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03449045861985295]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03449045861985295]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03449045861985295]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03449045861985295]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03449045861985295]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03449045861985295]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03449045861985295]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0689809172397059, 0.0689809172397059}], 
            {GrayLevel[0], DiskBox[2, 0.036827562393762746`]}, 
            {GrayLevel[0], DiskBox[3, 0.036827562393762746`]}, 
            {GrayLevel[0], DiskBox[4, 0.036827562393762746`]}, 
            {GrayLevel[0], DiskBox[5, 0.036827562393762746`]}, 
            {GrayLevel[0], DiskBox[6, 0.036827562393762746`]}, InsetBox[
             GraphicsBox[{}], 7, 
             Automatic, {0.0689809172397059, 0.0689809172397059}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"topologies", "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"CompleteDenominatorSet", "[", 
       RowBox[{"#", ",", 
        RowBox[{"{", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}], ",", 
        RowBox[{"{", "l1", "}"}]}], "]"}], "&"}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0", ",", "irred"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k3", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "l1"}], ",", "0", ",", "irred"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"l1", ",", "0"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"exprb", "=", 
   RowBox[{
    RowBox[{"expr", "//", 
     RowBox[{
      RowBox[{"ToTopologies", "[", 
       RowBox[{"#", ",", "topologies", ",", 
        RowBox[{"{", 
         RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}], ",", 
        RowBox[{"{", "l1", "}"}]}], "]"}], "&"}]}], "//", 
    RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"exprb", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"exprb", "//", 
    RowBox[{
     RowBox[{"Cases", "[", 
      RowBox[{"#", ",", "_B", ",", 
       RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}], "//", 
  "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["4393", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox["99", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprb", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox[
    RowBox[{"32", " ", "\[ImaginaryI]", " ", 
     SuperscriptBox["gs", "4"], " ", 
     RowBox[{"PS", "(", "3", ")"}], " ", 
     SuperscriptBox["qu", "2"], " ", 
     SuperscriptBox["s23", "2"], " ", 
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "0", ",", "1", ",", "1"}], "4"], " ", 
     SubscriptBox["C", "A"], " ", 
     SubsuperscriptBox["C", "F", "2"]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"d", "-", "2"}], ")"}], " ", "x", " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "s13"}], "-", "s23", "+", "1"}], ")"}]}]]}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"topologynames", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"topologies", "//", "Length"}], "//", "Range"}], "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"ToExpression", "[", 
       RowBox[{"\"\<ps3l1t\>\"", "<>", 
        RowBox[{"ToString", "[", "#", "]"}]}], "]"}], "&"}], "]"}]}]}], ";", 
  RowBox[{"exprj", "=", 
   RowBox[{"exprb", "/.", 
    RowBox[{
     RowBox[{"B", "[", 
      RowBox[{"topid_", ",", "indices_List"}], "]"}], "\[RuleDelayed]", 
     RowBox[{"j", "[", 
      RowBox[{
       RowBox[{"topologynames", "[", 
        RowBox[{"[", "topid", "]"}], "]"}], ",", 
       RowBox[{"Sequence", "@@", "indices"}]}], "]"}]}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprj", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox[
    RowBox[{"12", " ", "\[ImaginaryI]", " ", 
     SuperscriptBox["gs", "4"], " ", 
     RowBox[{"PS", "(", "3", ")"}], " ", "Q2", " ", 
     SuperscriptBox["qu", "2"], " ", "s23", " ", "x", " ", 
     SubsuperscriptBox["C", "A", "2"], " ", 
     SubscriptBox["C", "F"], " ", 
     RowBox[{"j", "(", 
      RowBox[{"ps3l1t4", ",", "1", ",", "1", ",", "1", ",", "0"}], ")"}]}], 
    RowBox[{
     RowBox[{"-", "s13"}], "-", "s23", "+", "1"}]]}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{
    RowBox[{"<|", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"\"\<topology-denominators\>\"", "\[Rule]", "topologies"}], "/.", 
       RowBox[{
        RowBox[{"den", "[", 
         RowBox[{"p_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
        RowBox[{
         RowBox[{"sp", "[", "p", "]"}], "-", "m2"}]}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"\"\<topology-names\>\"", "\[Rule]", "topologynames"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"\"\<topology-loop-momenta\>\"", "\[Rule]", 
       RowBox[{"{", "l1", "}"}]}], ",", "\[IndentingNewLine]", 
      RowBox[{"\"\<expression\>\"", "\[Rule]", "exprj"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"\"\<external-sp-rules\>\"", "\[Rule]", "externalsprules"}], 
      ",", "\[IndentingNewLine]", 
      RowBox[{
      "\"\<output-file\>\"", "\[Rule]", "\"\<ps3.l1.litered.result.m\>\""}]}],
      "\[IndentingNewLine]", "|>"}], ",", "\"\<ps3.l1.litered.job.m\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RunInATemporaryKernel", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"JOBFILE", "=", "\"\<ps3.l1.litered.job.m\>\""}], ";", 
    RowBox[{"Get", "[", "\"\<litered-ibpreduce.m\>\"", "]"}]}], 
   "\[IndentingNewLine]", "]"}], ";"}]}], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"**************** \"\>", "\[InvisibleSpace]", 
    StyleBox["\<\"LiteRed v1.82\"\>",
     StripOnInput->False,
     LineColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     FrontFaceColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     BackFaceColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     GraphicsColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]],
     FontWeight->Bold,
     FontColor->RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]]], 
    "\[InvisibleSpace]", "\<\" ********************\\nAuthor: Roman N. Lee, \
Budker Institute of Nuclear Physics, Novosibirsk.\\nRelease Date: \
01.06.2015\\nLiteRed stands for Loop InTEgrals REDuction.\\nThe package is \
designed for the search and application of the Integration-By-Part reduction \
rules. It also contains some other useful tools.\\nSee ?LiteRed`* for a list \
of functions. \"\>"}],
   SequenceForm["**************** ", 
    Style["LiteRed v1.82", {Bold, 
      RGBColor[1, 
       Rational[1, 3], 
       Rational[1, 3]]}], 
    " ********************\nAuthor: Roman N. Lee, Budker Institute of Nuclear \
Physics, Novosibirsk.\nRelease Date: 01.06.2015\nLiteRed stands for Loop \
InTEgrals REDuction.\nThe package is designed for the search and application \
of the Integration-By-Part reduction rules. It also contains some other \
useful tools.\nSee ?LiteRed`* for a list of functions. "],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"SetSharedFunction", "::", "subnopar"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Parallel programming is not available in subkernels of \
another parallel computation.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"SetSharedFunction", "::", "subnopar"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Parallel programming is not available in subkernels of \
another parallel computation.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* NewBasis[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* NewBasis[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Valid basis.\\n    Ds[ps3l1t1] \[LongDash] denominators,\\n    \
SPs[ps3l1t1] \[LongDash] scalar products involving loop momenta,\\n    \
LMs[ps3l1t1] \[LongDash] loop momenta,\\n    EMs[ps3l1t1] \[LongDash] \
external momenta,\\n    Toj[ps3l1t1] \[LongDash] rules to transform scalar \
products to denominators.\\nThe definitions of the basis will be saved in \
litered.ps3l1t1.dir\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"DiskSave", "::", "dir"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"The directory \
\\!\\(\\*FormBox[\\\"\\\\\\\"litered.ps3l1t1.dir\\\\\\\"\\\", \
TraditionalForm]\\) has been created.\"\>"}], TraditionalForm]], "Message", \
"MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* GenerateIBP[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* GenerateIBP[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Integration-By-Part&Lorentz-Invariance identities are \
generated.\\n    IBP[ps3l1t1] \[LongDash] integration-by-part identities,\\n  \
  LI[ps3l1t1] \[LongDash] Lorentz invariance identities.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* AnalyzeSectors[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* AnalyzeSectors[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Analyze pattern: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "0", ",", "0"}], "}"}]}],
   SequenceForm["Analyze pattern: ", {
     Blank[], 
     Blank[], 0, 0}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"FrontEndObject", "::", "notavail"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"A front end is not available; certain operations require \
a front end.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"FrontEndObject", "::", "notavail"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"A front end is not available; certain operations require \
a front end.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox["\<\"Found 3 zero sectors out of 4.\\n    ZeroSectors[ps3l1t1] \
\[LongDash] zero sectors,\\n    NonZeroSectors[ps3l1t1] \[LongDash] nonzero \
sectors,\\n    SimpleSectors[ps3l1t1] \[LongDash] simple sectors (no nonzero \
subsectors),\\n    BasisSectors[ps3l1t1] \[LongDash] basis sectors (at least \
one immediate subsector is zero),\\n    ZerojRule[ps3l1t1] \[LongDash] a rule \
to nullify all zero j[ps3l1t1\[Ellipsis]],\\n    CutDs[ps3l1t1] \[LongDash] a \
flag list of cut denominatorsj (1=cut).\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindExtSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindExtSymmetries[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{"{", "ps3l1t1", "}"}], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindSymmetries[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing Feynman parametrization...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Using Feynman parametrization to determine equivalent simple \
sectors...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"FrontEndObject", "::", "notavail"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"A front end is not available; certain operations require \
a front end.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"General", "::", "stop"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Further output of \
\\!\\(\\*FormBox[StyleBox[RowBox[{\\\"FrontEndObject\\\", \\\"::\\\", \
\\\"notavail\\\"}], \\\"MessageName\\\"], TraditionalForm]\\) will be \
suppressed during this calculation.\"\>"}], TraditionalForm]], "Message", \
"MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox["\<\"Constructing shifts for simple sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 1 unique simple sectors with degeneracies 1 \
respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for all nonzero sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 1 unique sectors with degeneracies 1 \
respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming rules...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors and 1 unique sectors.\\n    \
UniqueSectors[ps3l1t1] \[LongDash] unique sectors.\\n    \
MappedSectors[ps3l1t1] \[LongDash] mapped sectors.\\n    SR[ps3l1t1][\
\[Ellipsis]] \[LongDash] symmetry relations for j[ps3l1t1,\[Ellipsis]] from \
UniqueSectors[ps3l1t1].\\n    jSymmetries[ps3l1t1,\[Ellipsis]] \[LongDash] \
symmetry rules for the sector js[ps3l1t1,\[Ellipsis]] in \
UniqueSectors[ps3l1t1].\\n    jRules[ps3l1t1,\[Ellipsis]] \[LongDash] \
reduction rules for j[ps3l1t1,\[Ellipsis]] from MappedSectors[ps3l1t1].\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* SolvejSector[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* SolvejSector[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t1", ",", "1", ",", "1", ",", "0", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t1, 1, 1, 0, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found master integral {1, 1, 0, 0}\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    1 master integrals found:\\nj[ps3l1t1, 1, 1, 0, 0].\\n    \
jRules[ps3l1t1, 1, 1, 0, 0] \[LongDash] reduction rules for the sector.\\n    \
MIs[ps3l1t1] \[LongDash] updated list of the masters.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* DiskSave[\"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* DiskSave[", $CellContext`ps3l1t1, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"DiskSave", "::", "overwrite"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"The file \
\\!\\(\\*FormBox[\\\"\\\\\\\"litered.ps3l1t1.dir/ps3l1t1\\\\\\\"\\\", \
TraditionalForm]\\) has been overwritten.\"\>"}], TraditionalForm]], "Message",\
 "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* NewBasis[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* NewBasis[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Valid basis.\\n    Ds[ps3l1t2] \[LongDash] denominators,\\n    \
SPs[ps3l1t2] \[LongDash] scalar products involving loop momenta,\\n    \
LMs[ps3l1t2] \[LongDash] loop momenta,\\n    EMs[ps3l1t2] \[LongDash] \
external momenta,\\n    Toj[ps3l1t2] \[LongDash] rules to transform scalar \
products to denominators.\\nThe definitions of the basis will be saved in \
litered.ps3l1t2.dir\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"DiskSave", "::", "dir"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"The directory \
\\!\\(\\*FormBox[\\\"\\\\\\\"litered.ps3l1t2.dir\\\\\\\"\\\", \
TraditionalForm]\\) has been created.\"\>"}], TraditionalForm]], "Message", \
"MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* GenerateIBP[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* GenerateIBP[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Integration-By-Part&Lorentz-Invariance identities are \
generated.\\n    IBP[ps3l1t2] \[LongDash] integration-by-part identities,\\n  \
  LI[ps3l1t2] \[LongDash] Lorentz invariance identities.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* AnalyzeSectors[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* AnalyzeSectors[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Analyze pattern: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "0"}], "}"}]}],
   SequenceForm["Analyze pattern: ", {
     Blank[], 
     Blank[], 
     Blank[], 0}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 6 zero sectors out of 8.\\n    ZeroSectors[ps3l1t2] \
\[LongDash] zero sectors,\\n    NonZeroSectors[ps3l1t2] \[LongDash] nonzero \
sectors,\\n    SimpleSectors[ps3l1t2] \[LongDash] simple sectors (no nonzero \
subsectors),\\n    BasisSectors[ps3l1t2] \[LongDash] basis sectors (at least \
one immediate subsector is zero),\\n    ZerojRule[ps3l1t2] \[LongDash] a rule \
to nullify all zero j[ps3l1t2\[Ellipsis]],\\n    CutDs[ps3l1t2] \[LongDash] a \
flag list of cut denominatorsj (1=cut).\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindExtSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindExtSymmetries[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{"ps3l1t2", ",", "ps3l1t1"}], "}"}], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors.\\n    ExtUniqueSectors[ps3l1t2] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t2] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t2,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t2,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t2].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindSymmetries[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing Feynman parametrization...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Using Feynman parametrization to determine equivalent simple \
sectors...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for simple sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 1 unique simple sectors with degeneracies 1 \
respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for all nonzero sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 2 unique sectors with degeneracies 1, 1 respectively.\
\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming rules...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors and 2 unique sectors.\\n    \
UniqueSectors[ps3l1t2] \[LongDash] unique sectors.\\n    \
MappedSectors[ps3l1t2] \[LongDash] mapped sectors.\\n    SR[ps3l1t2][\
\[Ellipsis]] \[LongDash] symmetry relations for j[ps3l1t2,\[Ellipsis]] from \
UniqueSectors[ps3l1t2].\\n    jSymmetries[ps3l1t2,\[Ellipsis]] \[LongDash] \
symmetry rules for the sector js[ps3l1t2,\[Ellipsis]] in \
UniqueSectors[ps3l1t2].\\n    jRules[ps3l1t2,\[Ellipsis]] \[LongDash] \
reduction rules for j[ps3l1t2,\[Ellipsis]] from MappedSectors[ps3l1t2].\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* SolvejSector[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* SolvejSector[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t2", ",", "1", ",", "0", ",", "1", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t2, 1, 0, 1, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found master integral {1, 0, 1, 0}\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    1 master integrals found:\\nj[ps3l1t2, 1, 0, 1, 0].\\n    \
jRules[ps3l1t2, 1, 0, 1, 0] \[LongDash] reduction rules for the sector.\\n    \
MIs[ps3l1t2] \[LongDash] updated list of the masters.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t2", ",", "1", ",", "1", ",", "1", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t2, 1, 1, 1, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t2, 1, 1, 1, 0] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t2] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* DiskSave[\"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* DiskSave[", $CellContext`ps3l1t2, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"DiskSave", "::", "overwrite"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"The file \
\\!\\(\\*FormBox[\\\"\\\\\\\"litered.ps3l1t2.dir/ps3l1t2\\\\\\\"\\\", \
TraditionalForm]\\) has been overwritten.\"\>"}], TraditionalForm]], "Message",\
 "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* NewBasis[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* NewBasis[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Valid basis.\\n    Ds[ps3l1t3] \[LongDash] denominators,\\n    \
SPs[ps3l1t3] \[LongDash] scalar products involving loop momenta,\\n    \
LMs[ps3l1t3] \[LongDash] loop momenta,\\n    EMs[ps3l1t3] \[LongDash] \
external momenta,\\n    Toj[ps3l1t3] \[LongDash] rules to transform scalar \
products to denominators.\\nThe definitions of the basis will be saved in \
litered.ps3l1t3.dir\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"DiskSave", "::", "dir"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"The directory \
\\!\\(\\*FormBox[\\\"\\\\\\\"litered.ps3l1t3.dir\\\\\\\"\\\", \
TraditionalForm]\\) has been created.\"\>"}], TraditionalForm]], "Message", \
"MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"General", "::", "stop"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Further output of \
\\!\\(\\*FormBox[StyleBox[RowBox[{\\\"DiskSave\\\", \\\"::\\\", \
\\\"dir\\\"}], \\\"MessageName\\\"], TraditionalForm]\\) will be suppressed \
during this calculation.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* GenerateIBP[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* GenerateIBP[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Integration-By-Part&Lorentz-Invariance identities are \
generated.\\n    IBP[ps3l1t3] \[LongDash] integration-by-part identities,\\n  \
  LI[ps3l1t3] \[LongDash] Lorentz invariance identities.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* AnalyzeSectors[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* AnalyzeSectors[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Analyze pattern: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "0"}], "}"}]}],
   SequenceForm["Analyze pattern: ", {
     Blank[], 
     Blank[], 
     Blank[], 0}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 6 zero sectors out of 8.\\n    ZeroSectors[ps3l1t3] \
\[LongDash] zero sectors,\\n    NonZeroSectors[ps3l1t3] \[LongDash] nonzero \
sectors,\\n    SimpleSectors[ps3l1t3] \[LongDash] simple sectors (no nonzero \
subsectors),\\n    BasisSectors[ps3l1t3] \[LongDash] basis sectors (at least \
one immediate subsector is zero),\\n    ZerojRule[ps3l1t3] \[LongDash] a rule \
to nullify all zero j[ps3l1t3\[Ellipsis]],\\n    CutDs[ps3l1t3] \[LongDash] a \
flag list of cut denominatorsj (1=cut).\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindExtSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindExtSymmetries[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{"ps3l1t3", ",", "ps3l1t2", ",", "ps3l1t1"}], "}"}], 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 1 mapped sectors.\\n    ExtUniqueSectors[ps3l1t3] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t3] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t3,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t3,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t3].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors.\\n    ExtUniqueSectors[ps3l1t3] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t3] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t3,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t3,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t3].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindSymmetries[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing Feynman parametrization...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Using Feynman parametrization to determine equivalent simple \
sectors...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for simple sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 1 unique simple sectors with degeneracies 1 \
respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for all nonzero sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 2 unique sectors with degeneracies 1, 1 respectively.\
\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming rules...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors and 1 unique sectors.\\n    \
UniqueSectors[ps3l1t3] \[LongDash] unique sectors.\\n    \
MappedSectors[ps3l1t3] \[LongDash] mapped sectors.\\n    SR[ps3l1t3][\
\[Ellipsis]] \[LongDash] symmetry relations for j[ps3l1t3,\[Ellipsis]] from \
UniqueSectors[ps3l1t3].\\n    jSymmetries[ps3l1t3,\[Ellipsis]] \[LongDash] \
symmetry rules for the sector js[ps3l1t3,\[Ellipsis]] in \
UniqueSectors[ps3l1t3].\\n    jRules[ps3l1t3,\[Ellipsis]] \[LongDash] \
reduction rules for j[ps3l1t3,\[Ellipsis]] from MappedSectors[ps3l1t3].\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* SolvejSector[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* SolvejSector[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t3", ",", "1", ",", "1", ",", "1", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t3, 1, 1, 1, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t3, 1, 1, 1, 0] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t3] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* DiskSave[\"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* DiskSave[", $CellContext`ps3l1t3, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"DiskSave", "::", "overwrite"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"The file \
\\!\\(\\*FormBox[\\\"\\\\\\\"litered.ps3l1t3.dir/ps3l1t3\\\\\\\"\\\", \
TraditionalForm]\\) has been overwritten.\"\>"}], TraditionalForm]], "Message",\
 "MSG",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{
   StyleBox[
    RowBox[{"General", "::", "stop"}], "MessageName"], 
   RowBox[{
   ":", " "}], "\<\"Further output of \
\\!\\(\\*FormBox[StyleBox[RowBox[{\\\"DiskSave\\\", \\\"::\\\", \\\"overwrite\
\\\"}], \\\"MessageName\\\"], TraditionalForm]\\) will be suppressed during \
this calculation.\"\>"}], TraditionalForm]], "Message", "MSG",
 ShowCellLabel->True],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* NewBasis[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* NewBasis[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Valid basis.\\n    Ds[ps3l1t4] \[LongDash] denominators,\\n    \
SPs[ps3l1t4] \[LongDash] scalar products involving loop momenta,\\n    \
LMs[ps3l1t4] \[LongDash] loop momenta,\\n    EMs[ps3l1t4] \[LongDash] \
external momenta,\\n    Toj[ps3l1t4] \[LongDash] rules to transform scalar \
products to denominators.\\nThe definitions of the basis will be saved in \
litered.ps3l1t4.dir\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* GenerateIBP[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* GenerateIBP[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Integration-By-Part&Lorentz-Invariance identities are \
generated.\\n    IBP[ps3l1t4] \[LongDash] integration-by-part identities,\\n  \
  LI[ps3l1t4] \[LongDash] Lorentz invariance identities.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* AnalyzeSectors[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* AnalyzeSectors[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Analyze pattern: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_"}], "}"}]}],
   SequenceForm["Analyze pattern: ", {
     Blank[], 
     Blank[], 
     Blank[], 
     Blank[]}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 8 zero sectors out of 16.\\n    ZeroSectors[ps3l1t4] \
\[LongDash] zero sectors,\\n    NonZeroSectors[ps3l1t4] \[LongDash] nonzero \
sectors,\\n    SimpleSectors[ps3l1t4] \[LongDash] simple sectors (no nonzero \
subsectors),\\n    BasisSectors[ps3l1t4] \[LongDash] basis sectors (at least \
one immediate subsector is zero),\\n    ZerojRule[ps3l1t4] \[LongDash] a rule \
to nullify all zero j[ps3l1t4\[Ellipsis]],\\n    CutDs[ps3l1t4] \[LongDash] a \
flag list of cut denominatorsj (1=cut).\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindExtSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindExtSymmetries[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{"ps3l1t4", ",", "ps3l1t3", ",", "ps3l1t2", ",", "ps3l1t1"}], "}"}],
   TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 1 mapped sectors.\\n    ExtUniqueSectors[ps3l1t4] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t4] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t4,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t4,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t4].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 2 mapped sectors.\\n    ExtUniqueSectors[ps3l1t4] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t4] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t4,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t4,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t4].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors.\\n    ExtUniqueSectors[ps3l1t4] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t4] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t4,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t4,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t4].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindSymmetries[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing Feynman parametrization...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Using Feynman parametrization to determine equivalent simple \
sectors...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for simple sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 3 unique simple sectors with degeneracies 1, 1, 1 \
respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for all nonzero sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 7 unique sectors with degeneracies 1, 1, 1, 1, 1, 1, \
1 respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming rules...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors and 5 unique sectors.\\n    \
UniqueSectors[ps3l1t4] \[LongDash] unique sectors.\\n    \
MappedSectors[ps3l1t4] \[LongDash] mapped sectors.\\n    SR[ps3l1t4][\
\[Ellipsis]] \[LongDash] symmetry relations for j[ps3l1t4,\[Ellipsis]] from \
UniqueSectors[ps3l1t4].\\n    jSymmetries[ps3l1t4,\[Ellipsis]] \[LongDash] \
symmetry rules for the sector js[ps3l1t4,\[Ellipsis]] in \
UniqueSectors[ps3l1t4].\\n    jRules[ps3l1t4,\[Ellipsis]] \[LongDash] \
reduction rules for j[ps3l1t4,\[Ellipsis]] from MappedSectors[ps3l1t4].\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* SolvejSector[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* SolvejSector[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t4", ",", "0", ",", "0", ",", "1", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t4, 0, 0, 1, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found master integral {0, 0, 1, 1}\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    1 master integrals found:\\nj[ps3l1t4, 0, 0, 1, 1].\\n    \
jRules[ps3l1t4, 0, 0, 1, 1] \[LongDash] reduction rules for the sector.\\n    \
MIs[ps3l1t4] \[LongDash] updated list of the masters.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t4", ",", "0", ",", "1", ",", "1", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t4, 0, 1, 1, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t4, 0, 1, 1, 1] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t4] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t4", ",", "1", ",", "0", ",", "1", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t4, 1, 0, 1, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t4, 1, 0, 1, 1] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t4] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t4", ",", "1", ",", "1", ",", "1", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t4, 1, 1, 1, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t4, 1, 1, 1, 0] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t4] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t4", ",", "1", ",", "1", ",", "1", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t4, 1, 1, 1, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found master integral {1, 1, 1, 1}\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    1 master integrals found:\\nj[ps3l1t4, 1, 1, 1, 1].\\n    \
jRules[ps3l1t4, 1, 1, 1, 1] \[LongDash] reduction rules for the sector.\\n    \
MIs[ps3l1t4] \[LongDash] updated list of the masters.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* DiskSave[\"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* DiskSave[", $CellContext`ps3l1t4, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* NewBasis[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* NewBasis[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Valid basis.\\n    Ds[ps3l1t5] \[LongDash] denominators,\\n    \
SPs[ps3l1t5] \[LongDash] scalar products involving loop momenta,\\n    \
LMs[ps3l1t5] \[LongDash] loop momenta,\\n    EMs[ps3l1t5] \[LongDash] \
external momenta,\\n    Toj[ps3l1t5] \[LongDash] rules to transform scalar \
products to denominators.\\nThe definitions of the basis will be saved in \
litered.ps3l1t5.dir\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* GenerateIBP[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* GenerateIBP[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Integration-By-Part&Lorentz-Invariance identities are \
generated.\\n    IBP[ps3l1t5] \[LongDash] integration-by-part identities,\\n  \
  LI[ps3l1t5] \[LongDash] Lorentz invariance identities.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* AnalyzeSectors[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* AnalyzeSectors[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Analyze pattern: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_"}], "}"}]}],
   SequenceForm["Analyze pattern: ", {
     Blank[], 
     Blank[], 
     Blank[], 
     Blank[]}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 8 zero sectors out of 16.\\n    ZeroSectors[ps3l1t5] \
\[LongDash] zero sectors,\\n    NonZeroSectors[ps3l1t5] \[LongDash] nonzero \
sectors,\\n    SimpleSectors[ps3l1t5] \[LongDash] simple sectors (no nonzero \
subsectors),\\n    BasisSectors[ps3l1t5] \[LongDash] basis sectors (at least \
one immediate subsector is zero),\\n    ZerojRule[ps3l1t5] \[LongDash] a rule \
to nullify all zero j[ps3l1t5\[Ellipsis]],\\n    CutDs[ps3l1t5] \[LongDash] a \
flag list of cut denominatorsj (1=cut).\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindExtSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindExtSymmetries[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
   "ps3l1t5", ",", "ps3l1t4", ",", "ps3l1t3", ",", "ps3l1t2", ",", 
    "ps3l1t1"}], "}"}], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors.\\n    ExtUniqueSectors[ps3l1t5] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t5] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t5,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t5,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t5].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 1 mapped sectors.\\n    ExtUniqueSectors[ps3l1t5] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t5] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t5,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t5,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t5].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors.\\n    ExtUniqueSectors[ps3l1t5] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t5] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t5,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t5,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t5].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming signatures for unique sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Searching for mappings...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 2 mapped sectors.\\n    ExtUniqueSectors[ps3l1t5] \
\[LongDash] new sectors of the basis.\\n    ExtMappedSectors[ps3l1t5] \
\[LongDash] a list of mapped sectors,\\n    jExtRules[ps3l1t5,\[Ellipsis]] \
\[LongDash] reduction rules for j[ps3l1t5,\[Ellipsis]] from \
ExtMappedSectors[ps3l1t5].\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* FindSymmetries[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* FindSymmetries[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing Feynman parametrization...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Using Feynman parametrization to determine equivalent simple \
sectors...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for simple sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 3 unique simple sectors with degeneracies 1, 1, 1 \
respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Constructing shifts for all nonzero sectors...\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    Found 7 unique sectors with degeneracies 1, 1, 1, 1, 1, 1, \
1 respectively.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Forming rules...\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found 0 mapped sectors and 5 unique sectors.\\n    \
UniqueSectors[ps3l1t5] \[LongDash] unique sectors.\\n    \
MappedSectors[ps3l1t5] \[LongDash] mapped sectors.\\n    SR[ps3l1t5][\
\[Ellipsis]] \[LongDash] symmetry relations for j[ps3l1t5,\[Ellipsis]] from \
UniqueSectors[ps3l1t5].\\n    jSymmetries[ps3l1t5,\[Ellipsis]] \[LongDash] \
symmetry rules for the sector js[ps3l1t5,\[Ellipsis]] in \
UniqueSectors[ps3l1t5].\\n    jRules[ps3l1t5,\[Ellipsis]] \[LongDash] \
reduction rules for j[ps3l1t5,\[Ellipsis]] from MappedSectors[ps3l1t5].\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* SolvejSector[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* SolvejSector[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "0", ",", "1", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t5, 1, 0, 1, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found master integral {1, 0, 1, 0}\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    1 master integrals found:\\nj[ps3l1t5, 1, 0, 1, 0].\\n    \
jRules[ps3l1t5, 1, 0, 1, 0] \[LongDash] reduction rules for the sector.\\n    \
MIs[ps3l1t5] \[LongDash] updated list of the masters.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "0", ",", "1", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t5, 1, 0, 1, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t5, 1, 0, 1, 1] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t5] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "1", ",", "0", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t5, 1, 1, 0, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t5, 1, 1, 0, 1] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t5] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "1", ",", "1", ",", "0"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t5, 1, 1, 1, 0]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    0 master integrals found.\\n    jRules[ps3l1t5, 1, 1, 1, 0] \
\[LongDash] reduction rules for the sector.\\n    MIs[ps3l1t5] \[LongDash] \
updated list of the masters.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Sector \"\>", "\[InvisibleSpace]", 
    RowBox[{"js", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "1", ",", "1", ",", "1"}], ")"}]}],
   SequenceForm["Sector ", 
    LiteRed`js[$CellContext`ps3l1t5, 1, 1, 1, 1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Parameters {d, Q2, s13, s23} are assumed to be \
independent.\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"Found master integral {1, 1, 1, 1}\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"    1 master integrals found:\\nj[ps3l1t5, 1, 1, 1, 1].\\n    \
jRules[ps3l1t5, 1, 1, 1, 1] \[LongDash] reduction rules for the sector.\\n    \
MIs[ps3l1t5] \[LongDash] updated list of the masters.\"\>", 
  TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* DiskSave[\"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\"]\"\>"}],
   SequenceForm["* DiskSave[", $CellContext`ps3l1t5, "]"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps3l1t1", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps3l1t1, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{"ps3l1t1", ",", "1", ",", "1", ",", "0", ",", "0"}], ")"}], 
    "\[InvisibleSpace]", "\<\" = \"\>", "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k2", "+", "k3", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox["l1", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps3l1t1, 1, 1, 0, 0], " = ", 
    Vectors`sp[$CellContext`k2 + $CellContext`k3 - $CellContext`l1, \
$CellContext`k2 + $CellContext`k3 - $CellContext`l1]^(-1)/
    Vectors`sp[$CellContext`l1, $CellContext`l1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps3l1t2", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps3l1t2, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{"ps3l1t2", ",", "1", ",", "0", ",", "1", ",", "0"}], ")"}], 
    "\[InvisibleSpace]", "\<\" = \"\>", "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "l1"}], ")"}], "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps3l1t2, 1, 0, 1, 0], " = ", 
    Vectors`sp[$CellContext`k2 - $CellContext`l1, $CellContext`k2 - \
$CellContext`l1]^(-1)/
    Vectors`sp[$CellContext`k1 + $CellContext`l1, $CellContext`k1 + \
$CellContext`l1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps3l1t3", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps3l1t3, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps3l1t4", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps3l1t4, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{"ps3l1t4", ",", "0", ",", "0", ",", "1", ",", "1"}], ")"}], 
    "\[InvisibleSpace]", "\<\" = \"\>", "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox["l1", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps3l1t4, 0, 0, 1, 1], " = ", 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1, $CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1]^(-1)/Vectors`sp[$CellContext`l1, $CellContext`l1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{"ps3l1t4", ",", "1", ",", "1", ",", "1", ",", "1"}], ")"}], 
    "\[InvisibleSpace]", "\<\" = \"\>", "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox["l1", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps3l1t4, 1, 1, 1, 1], " = ", 
    Vectors`sp[$CellContext`k1 - $CellContext`l1, $CellContext`k1 - \
$CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, \
$CellContext`k1 + $CellContext`k2 - $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1, $CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1]^(-1)/Vectors`sp[$CellContext`l1, $CellContext`l1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters for \"\>", "\[InvisibleSpace]", "ps3l1t5", 
    "\[InvisibleSpace]", "\<\":\"\>"}],
   SequenceForm["Masters for ", $CellContext`ps3l1t5, ":"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "0", ",", "1", ",", "0"}], ")"}], 
    "\[InvisibleSpace]", "\<\" = \"\>", "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ")"}], "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps3l1t5, 1, 0, 1, 0], " = ", 
    Vectors`sp[$CellContext`k2 - $CellContext`l1, $CellContext`k2 - \
$CellContext`l1]^(-1)/
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1, $CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"- \"\>", "\[InvisibleSpace]", 
    RowBox[{"j", "(", 
     RowBox[{"ps3l1t5", ",", "1", ",", "1", ",", "1", ",", "1"}], ")"}], 
    "\[InvisibleSpace]", "\<\" = \"\>", "\[InvisibleSpace]", 
    FractionBox["1", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ")"}], "2"], " ", 
      SuperscriptBox["l1", "2"]}]]}],
   SequenceForm["- ", 
    LiteRed`j[$CellContext`ps3l1t5, 1, 1, 1, 1], " = ", 
    Vectors`sp[$CellContext`k2 - $CellContext`l1, $CellContext`k2 - \
$CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 - $CellContext`l1, \
$CellContext`k1 + $CellContext`k2 - $CellContext`l1]^(-1) 
    Vectors`sp[$CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1, $CellContext`k1 + $CellContext`k2 + $CellContext`k3 - \
$CellContext`l1]^(-1)/Vectors`sp[$CellContext`l1, $CellContext`l1]],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"* IBPReduce\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"MIs: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"j", "(", 
       RowBox[{"ps3l1t5", ",", "1", ",", "1", ",", "1", ",", "1"}], ")"}], 
      ",", 
      RowBox[{"j", "(", 
       RowBox[{"ps3l1t4", ",", "1", ",", "1", ",", "1", ",", "1"}], ")"}], 
      ",", 
      RowBox[{"j", "(", 
       RowBox[{"ps3l1t5", ",", "1", ",", "0", ",", "1", ",", "0"}], ")"}], 
      ",", 
      RowBox[{"j", "(", 
       RowBox[{"ps3l1t4", ",", "0", ",", "0", ",", "1", ",", "1"}], ")"}], 
      ",", 
      RowBox[{"j", "(", 
       RowBox[{"ps3l1t2", ",", "1", ",", "0", ",", "1", ",", "0"}], ")"}], 
      ",", 
      RowBox[{"j", "(", 
       RowBox[{"ps3l1t1", ",", "1", ",", "1", ",", "0", ",", "0"}], ")"}]}], 
     "}"}]}],
   SequenceForm["MIs: ", {
     LiteRed`j[$CellContext`ps3l1t5, 1, 1, 1, 1], 
     LiteRed`j[$CellContext`ps3l1t4, 1, 1, 1, 1], 
     LiteRed`j[$CellContext`ps3l1t5, 1, 0, 1, 0], 
     LiteRed`j[$CellContext`ps3l1t4, 0, 0, 1, 1], 
     LiteRed`j[$CellContext`ps3l1t2, 1, 0, 1, 0], 
     LiteRed`j[$CellContext`ps3l1t1, 1, 1, 0, 0]}],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Save to \"\>", 
    "\[InvisibleSpace]", "\<\"ps3.l1.litered.result.m\"\>"}],
   SequenceForm["* Save to ", "ps3.l1.litered.result.m"],
   Editable->False], TraditionalForm]], "Print",
 ShowCellLabel->True],

Cell[BoxData[
 FormBox["\<\"* Done\"\>", TraditionalForm]], "Print",
 ShowCellLabel->True]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"expribp", "=", 
   RowBox[{
    RowBox[{"Get", "[", "\"\<ps3.l1.litered.result.m\>\"", "]"}], "/.", 
    RowBox[{
     RowBox[{"j", "[", 
      RowBox[{"topo_", ",", "indices__"}], "]"}], "\[RuleDelayed]", 
     RowBox[{"B", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Position", "[", 
         RowBox[{"topologynames", ",", "topo"}], "]"}], "[", 
        RowBox[{"[", 
         RowBox[{"1", ",", "1"}], "]"}], "]"}], ",", 
       RowBox[{"{", "indices", "}"}]}], "]"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expribp", "=", 
   RowBox[{"expribp", "//", "Expand"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"expribp", "//", "Length"}], "\[IndentingNewLine]", 
 RowBox[{"masters", "=", 
  RowBox[{
   RowBox[{"expribp", "//", 
    RowBox[{
     RowBox[{"Cases", "[", 
      RowBox[{"#", ",", "_B", ",", 
       RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}]}], "Input"],

Cell[BoxData[
 FormBox["1878", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "0", ",", "0"}], "1"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "0", ",", "1", ",", "0"}], "2"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"0", ",", "0", ",", "1", ",", "1"}], "4"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "4"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "0", ",", "1", ",", "0"}], "5"], ",", 
    SubsuperscriptBox["B", 
     RowBox[{"1", ",", "1", ",", "1", ",", "1"}], "5"]}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"expribp", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox[
    RowBox[{"2960", " ", "\[ImaginaryI]", " ", "d", " ", 
     SuperscriptBox["gs", "4"], " ", 
     RowBox[{"PS", "(", "3", ")"}], " ", 
     SuperscriptBox["qu", "2"], " ", 
     SuperscriptBox["s13", "2"], " ", 
     SuperscriptBox["s23", "4"], " ", 
     SubsuperscriptBox["B", 
      RowBox[{"0", ",", "0", ",", "1", ",", "1"}], "4"], " ", 
     SubsuperscriptBox["C", "A", "2"], " ", 
     SubscriptBox["C", "F"]}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"d", "-", "4"}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{"d", "-", "2"}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"s13", "-", "1"}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"s23", "-", "1"}], ")"}], "2"], " ", "x", " ", 
     RowBox[{"(", 
      RowBox[{"s13", "+", "s23", "-", "1"}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"s13", "+", "s23"}], ")"}], "2"]}]]}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"masters", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"#", "//", "InputForm"}], ",", "\[IndentingNewLine]", 
      RowBox[{"#", "//", 
       RowBox[{"FromTopologies", "[", "topologies", "]"}]}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"#", "//", 
        RowBox[{"FromTopologies", "[", "topologies", "]"}]}], "//", 
       RowBox[{
        RowBox[{"FindGraph", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"k1", "+", "k2", "+", "k3"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}], ",", 
          RowBox[{"Cases", "[", 
           RowBox[{"#", ",", 
            RowBox[{
             RowBox[{"den", "[", 
              RowBox[{"p_", ",", "__"}], "]"}], "\[RuleDelayed]", "p"}], ",", 
            
            RowBox[{"{", 
             RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], ",", 
          RowBox[{"EdgeFn", "\[Rule]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"Style", "[", 
              RowBox[{"#1", ",", 
               RowBox[{"#2", "/.", 
                RowBox[{"{", "\[IndentingNewLine]", 
                 RowBox[{
                  RowBox[{"_Incoming", "\[Rule]", 
                   RowBox[{"{", 
                    RowBox[{"Thick", ",", "Black"}], "}"}]}], ",", 
                  "\[IndentingNewLine]", 
                  RowBox[{
                   RowBox[{"Outgoing", "[", "k1", "]"}], "\[Rule]", "Red"}], 
                  ",", "\[IndentingNewLine]", 
                  RowBox[{"_Outgoing", "\[Rule]", "Gray"}], ",", 
                  "\[IndentingNewLine]", 
                  RowBox[{"_Internal", "\[Rule]", "Thick"}]}], 
                 "\[IndentingNewLine]", "}"}]}]}], "]"}], "&"}], ")"}]}]}], 
         "]"}], "&"}]}]}], "\[IndentingNewLine]", "}"}], "&"}], 
   "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"1", ",", " ", 
         RowBox[{"{", 
          RowBox[{"1", ",", " ", "1", ",", " ", "0", ",", " ", "0"}], "}"}]}],
         "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2], 
            $CellContext`OO[3]}, {
           Null, {{1, 2}, {2, 3}, {2, 3}, {2, 4}, {3, 5}, {3, 6}}}, {
           EdgeStyle -> {UndirectedEdge[1, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[3]] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 
                $CellContext`OO[2]] -> {
                GrayLevel[0.5]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {{
                 Thickness[Large], 
                 GrayLevel[0]}}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[3] -> 
              Graphics[{}], $CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJw1z3lIk3EYB/CXVUYyj9QZgXiVkpbaMjAb+Z0dK/MKMzVa2jWscI4kk1Zp
qB2UFCpoHl22QiU1HEl5oJFTKUozUmvp5tG77W0umuzoTatff/jAwwMPz/Pw
fPyOyZIkHIqiAkn+r72Tgw7qfRzh+nVu1YJCO9Yu6ytQiFjMe/CVu6v1cB32
kLe1zCDwisCYU6OHf7SXXCniCKnFCKVOfIvKAsX9ndKasAeL7etdlyc8i+zY
zKnTjVbaMVixyqLO0qGWG1bLCbajr6I9fSZAhx5au7qq3YZkb2SXGmiEDqlS
x2NsiJf6qB1e0DBL+Ty3MStmZOfZo6U0DMqIVrnEiqLOZ00vc2nkPxhYfshs
QcvVqWRvCY2sU8Ww5VuQUBKuqsqgIe7M8dRFmVDmcy935REa3s1BJp+0WWQm
nJx7QuYNs28/TJ4xYro8duP+czSkWq3785LvUMz7x3PLaNTXeEj8GhiY2K+a
V+SfBqu2qKnfgJD08AUxQ6P7r8o4odcjdes2UU+gDsd50TcjnfQobhv+1S/V
QT4dZpsi/rrpqsdmsheXLRW5Ev9QRthB1pnBiFnD1hM/K4isOxzLIM1X5PiD
+Lfcv91tv8jAuqDZFED8hXdjkpofMYjIYz7eIv4dDn6yzG4GXpUdKhnxd8Uu
feo+xOCPSyWfW2BBufx9UMMnBoUr7rCuMGHUKfKL7wgDcY3L6Rjiz3O/sbeW
zF87y+c55hjxcLzjJ9XDIJeKSlQTvzB4TpGmIPcEvDc7iX9sQ3yv8hKDz86a
Xe+IY/BCCD8ojsGa19ujWeJPX9IqHHFhMDcqbkwk/pQlBxonBgz4B9/FNH8=

            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0], Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.024917760460904242`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.1443245568905105`, 0.6051546881450159}, 3}], 
                 0.024917760460904242`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.1443714818853157`, 0.9326540270800798}, 3}], 
                 0.024917760460904242`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.024917760460904242`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.024917760460904242`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.024917760460904242`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.049835520921808485`, 0.049835520921808485`}], 
              {GrayLevel[0], DiskBox[2, 0.04978987080543341]}, 
              {GrayLevel[0], DiskBox[3, 0.04978987080543341]}, InsetBox[
               GraphicsBox[{}], 4, 
               Automatic, {0.049835520921808485`, 0.049835520921808485`}], 
              InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.049835520921808485`, 0.049835520921808485`}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.049835520921808485`, 0.049835520921808485`}]}}]],
          
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None,
       ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"2", ",", " ", 
         RowBox[{"{", 
          RowBox[{"1", ",", " ", "0", ",", " ", "1", ",", " ", "0"}], "}"}]}],
         "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2], 
            $CellContext`OO[3]}, {
           Null, {{1, 2}, {2, 3}, {2, 3}, {3, 4}, {3, 5}, {2, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[2]] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 
                $CellContext`OO[3]] -> {
                GrayLevel[0.5]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {{
                 Thickness[Large], 
                 GrayLevel[0]}}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[3] -> 
              Graphics[{}], $CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJw1z30s1HEcB/BzmbtItZlbNY1KXFeu+qOla7yPC4myXA/maTHkqZ2HNJIi
D7OcjmIsmadpHOWP7lo0hgv5QxNZhXnY6Xd330pdOBfa+vaHz/be55/33ttr
X7QsOJbNYrFcaP7/zRtzG7zFum+GZUav1amRBRgyZdsDVXrsEVhWcS+vocV3
5oFZrcdhu44TUQFsMWta7dnZfQmlstAFUShb3Ji37qkuMIPl9T6g0KIQm7t+
Ux8bmgR6PJniv/07b8CIRjnjbWVAvWOJlHOIYC3hXYLNVwPCYlI1ybEEd81n
wrIGCa603rhoXUVQc8RZxFF+w2fBwdr+bgIZPyJAqPiOfBt1590vBMeX6ubj
Mn6AfczntJAQuCQrn/PDF6GS9ImHfhI4DPMG0kpXMBLd1Zz2i8DxQO/KOMuE
YknbjJH25R0NTsZ0E+Yi3EpDJwnytJV7ixgTno6yC171EDw6v8s+KGQV9dn9
TrxqguHylBbu8CqMSa7eZXEEvLnK3SnuZsijikTeAgK+38s0yTMzUtJTe4K0
BhhF0UEa6t8fOD25XKSD3dZai0Tq50QORbLFOrxo1mndqb9J6nJVtUUHB2G1
dTv1/5aUZArHGHiwa/94UP/q4jn7h20MXBMLl65TvyrKaRtRMLDIP+vfSP22
dxq7gu8x4J6sUcZQ/1D446UPWQzkQoWmjPpns+UhObcZcJIqWo3UX67gRvJy
GcSPqpdtb5qwU1U9O1DGQHohOKeB+t9U8Dfy2hlc0/clx1M/hfqJxhk0Hc2N
daR+97DpiU+WOrxm72jLp/7lun4fDy8dNkxZ/lLqX9c6T/gW6/APgwArcg==

            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0], Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.024953558064746473`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.1478722775104355`, 0.9358461378891384}, 3}], 
                 0.024953558064746473`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.1479394739452318`, 0.6083760257134779}, 3}], 
                 0.024953558064746473`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{2, 6}, 0.024953558064746473`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.024953558064746473`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.024953558064746473`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.049907116129492946`, 0.049907116129492946`}], 
              {GrayLevel[0], DiskBox[2, 0.049785427985258444`]}, 
              {GrayLevel[0], DiskBox[3, 0.049785427985258444`]}, InsetBox[
               GraphicsBox[{}], 4, 
               Automatic, {0.049907116129492946`, 0.049907116129492946`}], 
              InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.049907116129492946`, 0.049907116129492946`}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.049907116129492946`, 0.049907116129492946`}]}}]],
          
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"4", ",", " ", 
         RowBox[{"{", 
          RowBox[{"0", ",", " ", "0", ",", " ", "1", ",", " ", "1"}], "}"}]}],
         "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
       " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2], 
            $CellContext`OO[3]}, {
           Null, {{1, 2}, {2, 3}, {2, 3}, {3, 4}, {3, 5}, {3, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[3]] -> {
                GrayLevel[0.5]}, UndirectedEdge[2, 
                $CellContext`OO[2]] -> {
                GrayLevel[0.5]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {{
                 Thickness[Large], 
                 GrayLevel[0]}}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[3] -> 
              Graphics[{}], $CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJw1jW0s1AEAxu/k1qaOc95mSVFMckXmvMw8N4yacmltQqapFnOx6TLZbmnL
SDGhKS+rQ0dKnLRbOkkZYxySlBKnk8bxfzm1cy9JHzzbs9+eL7/HLTXr1EUL
BoPhudn/3IqngvgefZKEF2u4N7VhBaP1gbsz91NI7eRI2Lf+wsZVmSkoInGj
61Ky7ZHtAmurpxUf5ggUng16J7RjCsSJdMbjTIZgytc/b17KEGx5X/qoBs/n
k5g+FNrU3EwhJlzT7rBK4lxxSVruAIWm4Ttn3JMpzNA1Wv0chdHOF4yjIxQ+
t5udd/2mMPFIEx4ZRmNVzGRNbKOxmFET/LGNRrlf9x8nNo3qL7NNnW462Jv5
2RoOjUqOg1t9hQ6fVhoD+TY0OuITfAzf1lFUcOV29xiFxbsHQuWxBuhryh7c
G6QQnJJTutpjQCzJ4umVFLiTAU5ZvkbkS351a1so3OR4hJAPjShgnTBKKig4
FQfRh61NSGxd63uVQ4Gv7tBr80zItnHcJz9NocwwtPxzwQSVNDo0l0chXXmQ
nhKT4O55JuRfJ/F+MmJhfJaESD6TZo4nofZ5s1x6fPPHbqN/LYDEVdl4g6qL
QsJrWZmvM4lF9dt+oTeNlIgdql4miZ74fOFYNQ136ddu0SwBrseIlL1TB9Wx
fo1fP4GolDp5n0SH1hwGc15B4IJMrXQZWIedrK6ZDCGQMuPItwo24H7hNf/K
QAK2MUmNSy2b2/WyNw8EPB25NXkuRmiqNGxFHAFvi4iZnBIjbPdqf3hmEYhb
Si+oNRthWdsWllFF4LmYy3siMmFDNBwlGiIQWexlnzRtAqtcoWFZkvgHmgI1
rQ==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0], Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.029514059841334164`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.4235973845618026`, 1.2138538095678728`}, 3}], 
                 0.029514059841334164`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.4110572966876305`, 0.8908048386270714}, 3}], 
                 0.029514059841334164`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.029514059841334164`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.029514059841334164`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.029514059841334164`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.05902811968266833, 0.05902811968266833}], 
              {GrayLevel[0], DiskBox[2, 0.04638208156008704]}, 
              {GrayLevel[0], DiskBox[3, 0.04638208156008704]}, InsetBox[
               GraphicsBox[{}], 4, 
               Automatic, {0.05902811968266833, 0.05902811968266833}], 
              InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.05902811968266833, 0.05902811968266833}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.05902811968266833, 0.05902811968266833}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"4", ",", " ", 
         RowBox[{"{", 
          RowBox[{"1", ",", " ", "1", ",", " ", "1", ",", " ", "1"}], "}"}]}],
         "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
       " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 3, 2, 4, 
            $CellContext`OO[1], 
            $CellContext`OO[2], 
            $CellContext`OO[3]}, {
           Null, {{1, 2}, {3, 4}, {5, 3}, {2, 5}, {2, 4}, {4, 6}, {3, 7}, {5, 
            8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[3, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[3, 
                $CellContext`OO[2]] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 
                $CellContext`OO[3]] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 3] -> {
                Thickness[Large]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {{
                 Thickness[Large], 
                 GrayLevel[0]}}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[3] -> 
              Graphics[{}], $CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{0., 0.0012564034136794966`}, {
            0.7333358249056314, 0.6805492113840973}, {1.831806252896569, 
            0.6805032234930137}, {1.8315078504121374`, 1.3112329068658821`}, {
            0.7334185495101785, 1.3109998533975507`}, {2.5626843854583186`, 
            1.992870368389799}, {2.5637387120234836`, 0.}, {
            0.0008567719563303822, 1.9914859037620511`}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0], Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.027027045538954894`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.027027045538954894`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.027027045538954894`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.027027045538954894`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.027027045538954894`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 7}, 0.027027045538954894`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{4, 6}, 0.027027045538954894`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.027027045538954894`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.05405409107790979, 0.05405409107790979}], 
              {GrayLevel[0], DiskBox[2, 0.03152253237204047]}, 
              {GrayLevel[0], DiskBox[3, 0.03152253237204047]}, 
              {GrayLevel[0], DiskBox[4, 0.03152253237204047]}, 
              {GrayLevel[0], DiskBox[5, 0.03152253237204047]}, InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.05405409107790979, 0.05405409107790979}], 
              InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.05405409107790979, 0.05405409107790979}], 
              InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.05405409107790979, 0.05405409107790979}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"5", ",", " ", 
         RowBox[{"{", 
          RowBox[{"1", ",", " ", "0", ",", " ", "1", ",", " ", "0"}], "}"}]}],
         "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 
            $CellContext`OO[1], 
            $CellContext`OO[2], 
            $CellContext`OO[3]}, {
           Null, {{1, 2}, {3, 2}, {2, 3}, {3, 4}, {2, 5}, {3, 6}}}, {
           EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[1, 2] -> {
                Thickness[Large]}, UndirectedEdge[2, 1] -> {
                Thickness[Large]}, UndirectedEdge[2, 
                $CellContext`OO[3]] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 
                $CellContext`OO[2]] -> {
                GrayLevel[0.5]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {{
                 Thickness[Large], 
                 GrayLevel[0]}}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[3] -> 
              Graphics[{}], $CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {2 -> 0.1, 1 -> 0.1}, VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJxdz2tIU3EYBvDjrL7kKsN0+CGtZKQzxJIYXXzsgileIqdODFbOXIbmrZs4
nTLNS1qUl0mmdINkZqww8YaGWzZxS9OUjKVIydnZmQU1L6dl2b+vvfDy8nx5
Xn475NlxaTyKooRk/13OkqDPOsULM0xWewWoOVRbtow3hDshCXwj0DUz6NaJ
4ht181hU6MC0MKBCpMbbPZmgVieDqhTRqLmX+GFDOC+M+m/kqsuFrmUc7C9s
s8ONHB55peuPZ1txZNZc5PDnsCnRszJmtxU1He5MVe8KunwfLDntNPTfxz1G
I1eQ7xahz+mlceFchoE3vQz7ScdRcx2N0SGnJCNtGWOfTUl++TQeX7l2P+rH
ElzrXqZWnaeR9LBizKpaQq5Ho3i9nMYtt7liY+g3FA0lK9UkiwTeMVTSVygL
EmSCdBrDv3xE+twFfFGYqSHSdyOrMLOhxo6Nl2SlqnoaTx2WQRcti5mpfmVI
H411bKftptEGLij9z9gCjTbX62dfMQwMGnH7Pn8rNssH3L35DIrLTYaDOVak
qE/kTRK/ya2t/vSIDXXB8yMuARwqVw+1p7qzUAxElTcRvySVn2yMYVHbI26a
I/6L/V2f5MUsdhUyrz2JP9TXa0rwhEVZpyathPgVgVrpxCALSevd+BTiHxX4
OlUTLHY+29/3m/jNb71nPKZZbPcLlTqI/6Nq1qeW5AqlrEtE/BHU4CL/PYu9
FmnsAvEXPNeu5elZrLZcze4nftkx9TYT+RdZqjULiX+8fI9GWMKie81fqCP+
O3lT8uZYFtFtHU4L8R84fOZd5FYWP3kF4mDib1XyNXEmG/4Cbjs33w==
            "], {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0], Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.024919964705658965`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.1443090372615017`, 0.6054513133632251}, 3}], 
                 0.024919964705658965`]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], 
                ArrowBox[BezierCurveBox[{
                 2, {1.1443581466710708`, 0.9329509319407084}, 3}], 
                 0.024919964705658965`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.024919964705658965`]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.024919964705658965`]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.024919964705658965`]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.04983992941131793, 0.04983992941131793}], 
              {GrayLevel[0], DiskBox[2, 0.04978991336856662]}, 
              {GrayLevel[0], DiskBox[3, 0.04978991336856662]}, InsetBox[
               GraphicsBox[{}], 4, Automatic, {0.04983992941131793, 0.04983992941131793}],\
 InsetBox[
               GraphicsBox[{}], 5, 
               Automatic, {0.04983992941131793, 0.04983992941131793}], 
              InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.04983992941131793, 0.04983992941131793}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]},
     {
      StyleBox[
       RowBox[{"B", "[", 
        RowBox[{"5", ",", " ", 
         RowBox[{"{", 
          RowBox[{"1", ",", " ", "1", ",", " ", "1", ",", " ", "1"}], "}"}]}],
         "]"}],
       ShowStringCharacters->True,
       NumberMarks->True], 
      RowBox[{
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "-", "l1"}], ",", "0"}], ")"}], " ", 
       RowBox[{"den", "(", 
        RowBox[{
         RowBox[{"k1", "+", "k2", "+", "k3", "-", "l1"}], ",", "0"}], ")"}], 
       " ", 
       RowBox[{"den", "(", 
        RowBox[{"l1", ",", "0"}], ")"}]}], 
      GraphicsBox[
       NamespaceBox["NetworkGraphics",
        DynamicModuleBox[{Typeset`graph = HoldComplete[
          Graph[{
            $CellContext`II[1], 1, 2, 3, 4, 
            $CellContext`OO[1], 
            $CellContext`OO[2], 
            $CellContext`OO[3]}, {
           Null, {{1, 2}, {3, 4}, {5, 3}, {2, 5}, {2, 4}, {3, 6}, {4, 7}, {5, 
            8}}}, {EdgeStyle -> {UndirectedEdge[2, 
                $CellContext`OO[1]] -> {
                RGBColor[1, 0, 0]}, UndirectedEdge[3, 
                $CellContext`OO[2]] -> {
                GrayLevel[0.5]}, UndirectedEdge[4, 
                $CellContext`OO[3]] -> {
                GrayLevel[0.5]}, UndirectedEdge[1, 4] -> {
                Thickness[Large]}, UndirectedEdge[
                $CellContext`II[1], 1] -> {{
                 Thickness[Large], 
                 GrayLevel[0]}}, UndirectedEdge[4, 2] -> {
                Thickness[Large]}, UndirectedEdge[1, 3] -> {
                Thickness[Large]}, UndirectedEdge[2, 3] -> {
                Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
            VertexShape -> {$CellContext`OO[3] -> 
              Graphics[{}], $CellContext`OO[2] -> 
              Graphics[{}], $CellContext`II[1] -> 
              Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
            VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1}, 
            VertexStyle -> {1 -> {
                GrayLevel[0]}, 2 -> {
                GrayLevel[0]}, 3 -> {
                GrayLevel[0]}, 4 -> {
                GrayLevel[0]}}}]]}, 
         TagBox[
          GraphicsGroupBox[
           GraphicsComplexBox[{{2.3942448104587584`, 2.4598004051739846`}, {
            1.7356041367058292`, 1.7640771799830035`}, {0.7241969615921047, 
            0.6957238693592844}, {0.6957239829692918, 1.7356041368634492`}, {
            1.7640772003883838`, 0.7241968143163054}, {0.06555699797190084, 
            0.}, {0., 2.3942439748381172`}, {2.4598015437844465`, 
            0.06555726598161593}}, {
             {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
              {GrayLevel[0], Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{1, 2}, 0.02623610385430375]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 4}, 0.02623610385430375]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{2, 5}, 0.02623610385430375]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 4}, 0.02623610385430375]}}, 
              {Thickness[Large], 
               {Arrowheads[0.], ArrowBox[{3, 5}, 0.02623610385430375]}}, 
              {RGBColor[1, 0, 0], 
               {Arrowheads[0.], ArrowBox[{3, 6}, 0.02623610385430375]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{4, 7}, 0.02623610385430375]}}, 
              {GrayLevel[0.5], 
               {Arrowheads[0.], ArrowBox[{5, 8}, 0.02623610385430375]}}}, 
             {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
              InsetBox[
               GraphicsBox[{}], 1, 
               Automatic, {0.0524722077086075, 0.0524722077086075}], 
              {GrayLevel[0], DiskBox[2, 0.04790193880851673]}, 
              {GrayLevel[0], DiskBox[3, 0.04790193880851673]}, 
              {GrayLevel[0], DiskBox[4, 0.04790193880851673]}, 
              {GrayLevel[0], DiskBox[5, 0.04790193880851673]}, InsetBox[
               GraphicsBox[{}], 6, 
               Automatic, {0.0524722077086075, 0.0524722077086075}], 
              InsetBox[
               GraphicsBox[{}], 7, 
               Automatic, {0.0524722077086075, 0.0524722077086075}], 
              InsetBox[
               GraphicsBox[{}], 8, 
               Automatic, {0.0524722077086075, 0.0524722077086075}]}}]],
          MouseAppearanceTag["NetworkGraphics"]],
         AllowKernelInitialization->False]],
       DefaultBaseStyle->{
        "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> 
         Hue[0.8, 1., 0.6]},
       FormatType->TraditionalForm,
       FrameTicks->None]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Put", "[", 
  RowBox[{
   RowBox[{"<|", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"\"\<topologies\>\"", "\[Rule]", "topologies"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"\"\<masters\>\"", "\[Rule]", "masters"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"\"\<expression\>\"", "\[Rule]", "expribp"}]}], 
    "\[IndentingNewLine]", "|>"}], ",", "\"\<reduced.ps3.l1.m\>\""}], 
  "]"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
9. IBP reduction & the differential equations for PS[k1,k2,k3,k4] L[], Kira\
\>", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"We", "'"}], "ll", " ", "reduce", " ", "the", " ", "whole", " ", 
    "matrix", " ", "element", " ", "with", " ", "the", " ", "phase", " ", 
    "space", " ", "integration", " ", "translated", " ", "into", " ", "cut", 
    " ", 
    RowBox[{"propagators", "."}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"externalsprules", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"sp", "[", 
        RowBox[{"q", ",", "q"}], "]"}], "\[Rule]", "Q2"}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"expr", "=", 
     RowBox[{
      RowBox[{
       RowBox[{"Get", "[", "\"\<M.m\>\"", "]"}], "[", 
       RowBox[{
        RowBox[{"PS", "[", 
         RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4"}], "]"}], 
        RowBox[{"L", "[", "]"}]}], "]"}], "//", 
      RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"topologies", "=", 
     RowBox[{"expr", "//", "Topologies"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"expr", "//", "Length"}], "\[IndentingNewLine]", 
   RowBox[{"topologies", "//", "TableForm"}], "\[IndentingNewLine]", 
   RowBox[{"topologies", "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"MakePrettyTopologyGraph", "[", 
       RowBox[{"#", ",", 
        RowBox[{"{", 
         RowBox[{"k1", ",", "k2", ",", "k3", ",", "k4"}], "}"}]}], "]"}], 
      "&"}], "]"}]}]}]}]], "Input"],

Cell[BoxData[
 FormBox["770", TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  InterpretationBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}], "\<\"\"\>"},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k4"}], ",", "0"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k4"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   TableForm[{{
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k2, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k2, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k2, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k1 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k4, 
       0]}, {
      $CellContext`den[$CellContext`k2 + $CellContext`k3, 0], 
      $CellContext`den[$CellContext`k1 + $CellContext`k2 + $CellContext`k3, 
       0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k4, 0], 
      $CellContext`den[$CellContext`k2 + $CellContext`k3 + $CellContext`k4, 
       0]}}]], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 5, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 5}, {4, 5}, {4, 3}, {5, 6}, {3, 6}, {2, 
          4}, {6, 7}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[GraphicsGroupBox[GraphicsComplexBox[CompressedData["
1:eJwBgQJ+/SFib1JlAgAAACcAAAACAAAA/E05uD4zDUB2tkrqZ1T0PwXH685H
5QVAixDEe8vb7z+a/Nlh/7/9P1r9omt6MvQ/s2cDn4OaAUDgdpS4p2jOPw5+
YsIIyfI/AAAAAAAAAADJyEoVCKPrP6n+tjdxFvA/AAAAAAAAAAATRd9v/T33
P+pCa5p/zgBAVDqwF0hXxj/RfXYcIpoAQKynG1v0YMQ/+n5ZtVFkAEA0K19y
fZPCP+7U1EEuLQBA0Pn2lvTvwD9eMM/OsOn/P9D04F2k7r4/eTP2AeN2/z+Q
ka9z6lS8P+5PRhY3Av8/wHdemUUUuj/BNv0f8ov+P8g2fjoLLrg/n3y7IVn6
9z8AWEf8EtdCv2BwRifBgvc/gMQYZ+IFe7+qXcz3HAr3P8BU1sTQ9Ia/JxtB
AbSQ9j8A4LJPeH2Nv7ajGibOFvY/ALCRvYSMkL8YVcGSs5z1P2Dju7mw4pC/
Q2bUkqwi9T+AuBIuGoKPv3ziW2YBqfQ/wETUpYpQir98tQZNh6oAQJbdeDHY
BtA/mHPKtrFtAEA10bqtZDDQPy380zauMABApvnl535C0D+8qU7aQef/P3BW
IygcPdA/TDIo/1tt/z9v0seePyDQP8nvnAjz8/4/KtzhxPTXzz8S3SLZTnv+
PwA9zcvWQM8/09Ct3rYD/j82vpDLfnvOP7MWbOAdcvc/eFtVG6JRwj+G/SLq
2Pv2P/w65euEXsE/+hlz/iyH9j8Urrx+Mj7APxUdmjFfFPY/7PhHE6vivT+Y
o798s6P1PxD6OkNm8bo/fk+2lWw19T9Ml2qMVKq3P9NRfMfLyfQ/YJ7xumYP
tD+gx5LLEGH0Pwh5yEG/IrA/vb47Nw==
          "], {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.0347263707657203]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.0347263707657203]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.0347263707657203]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               4, {1.7263294841374197`, -0.04999193827856571}, 5}], 
               0.0347263707657203]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], 
              ArrowBox[BezierCurveBox[{
               4, {1.6481974949339981`, 0.2875607603066587}, 5}], 
               0.0347263707657203]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.0347263707657203]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.0347263707657203]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0694527415314406, 0.0694527415314406}], 
            {GrayLevel[0], DiskBox[2, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[3, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[4, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[5, 0.04586111767247925]}, 
            {GrayLevel[0], DiskBox[6, 0.04586111767247925]}, InsetBox[
             GraphicsBox[{}], 7, 
             Automatic, {0.0694527415314406, 0.0694527415314406}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {3, 5}, {6, 4}, {2, 5}, {6, 3}, {4, 7}, {2, 
          6}, {5, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.391204699197312, 1.8438947129965138`}, {
          2.6217725564957175`, 1.2811053633583065`}, {1.7012645945519833`, 
          0.}, {1.074202049265616, 0.30585220640588695`}, {
          1.6932997325655055`, 1.3639617706279485`}, {2.3202638767233874`, 
          0.3159755466814528}, {0.7622649356690391, 1.270322347714335}, {0., 
          1.847211895852448}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.032980747396119475`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.032980747396119475`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.032980747396119475`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.032980747396119475`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.032980747396119475`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.032980747396119475`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.032980747396119475`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.032980747396119475`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.032980747396119475`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.032980747396119475`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06596149479223895, 0.06596149479223895}], 
            {GrayLevel[0], DiskBox[2, 0.03474912435693008]}, 
            {GrayLevel[0], DiskBox[3, 0.03474912435693008]}, 
            {GrayLevel[0], DiskBox[4, 0.03474912435693008]}, 
            {GrayLevel[0], DiskBox[5, 0.03474912435693008]}, 
            {GrayLevel[0], DiskBox[6, 0.03474912435693008]}, 
            {GrayLevel[0], DiskBox[7, 0.03474912435693008]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.06596149479223895, 0.06596149479223895}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {3, 5}, {2, 4}, {6, 7}, {6, 3}, {4, 5}, {5, 
          7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[1, 3] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.5032872287971495`, 0.0003346141677250669}, {
          2.6769991669233297`, 0.47072605311572774`}, {1.7515332825353969`, 
          1.7414595812870004`}, {2.3724813237658715`, 1.44443327890605}, {
          1.1305529226961912`, 1.4445737463534254`}, {1.7514371336353263`, 
          0.3680517900471344}, {0.8258396839749501, 0.47116929705690014`}, {
          0., 0.}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03374314900398022]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.03374314900398022]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03374314900398022]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03374314900398022]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03374314900398022]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03374314900398022]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03374314900398022]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.03374314900398022]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03374314900398022]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03374314900398022]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06748629800796044, 0.06748629800796044}], 
            {GrayLevel[0], DiskBox[2, 0.034415033280901945`]}, 
            {GrayLevel[0], DiskBox[3, 0.034415033280901945`]}, 
            {GrayLevel[0], DiskBox[4, 0.034415033280901945`]}, 
            {GrayLevel[0], DiskBox[5, 0.034415033280901945`]}, 
            {GrayLevel[0], DiskBox[6, 0.034415033280901945`]}, 
            {GrayLevel[0], DiskBox[7, 0.034415033280901945`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.06748629800796044, 0.06748629800796044}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 5, 3, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {3, 5}, {2, 5}, {6, 7}, {6, 3}, {5, 4}, {4, 
          7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{0., 0.041783056965116305`}, {0.8101025613207566,
           0.5795952167671551}, {1.6815595845361242`, 1.8915139217192074`}, {
          2.3182283471807077`, 1.5995850699464622`}, {1.0734050654168052`, 
          1.5561298946839506`}, {1.7231623419423252`, 0.53388210916044}, {
          2.6542623631603717`, 0.6493455329211022}, {3.3674951147066268`, 
          0.}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03281818822513381]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.03281818822513381]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03281818822513381]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03281818822513381]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03281818822513381]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03281818822513381]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03281818822513381]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03281818822513381]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03281818822513381]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03281818822513381]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06563637645026763, 0.06563637645026763}], 
            {GrayLevel[0], DiskBox[2, 0.03472514811759757]}, 
            {GrayLevel[0], DiskBox[3, 0.03472514811759757]}, 
            {GrayLevel[0], DiskBox[4, 0.03472514811759757]}, 
            {GrayLevel[0], DiskBox[5, 0.03472514811759757]}, 
            {GrayLevel[0], DiskBox[6, 0.03472514811759757]}, 
            {GrayLevel[0], DiskBox[7, 0.03472514811759757]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.06563637645026763, 0.06563637645026763}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 3}, {4, 5}, {6, 7}, {3, 5}, {6, 4}, {5, 
          7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[6, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.510061966473745, 0.000250684455576744}, {
          2.6801048896397273`, 0.4637499996366458}, {2.3757597393804795`, 
          1.438054776414318}, {1.7549325209020046`, 1.734250717792647}, {
          1.1341829702727326`, 1.4381476807487799`}, {1.7548449873859506`, 
          0.3602285070392836}, {0.8295773500174981, 0.46415477242340875`}, {
          0., 0.}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03378891349164677]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03378891349164677]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03378891349164677]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03378891349164677]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03378891349164677]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03378891349164677]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03378891349164677]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.03378891349164677]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03378891349164677]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03378891349164677]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06757782698329354, 0.06757782698329354}], 
            {GrayLevel[0], DiskBox[2, 0.03438775265821557]}, 
            {GrayLevel[0], DiskBox[3, 0.03438775265821557]}, 
            {GrayLevel[0], DiskBox[4, 0.03438775265821557]}, 
            {GrayLevel[0], DiskBox[5, 0.03438775265821557]}, 
            {GrayLevel[0], DiskBox[6, 0.03438775265821557]}, 
            {GrayLevel[0], DiskBox[7, 0.03438775265821557]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.06757782698329354, 0.06757782698329354}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 5, 6, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 3}, {5, 6}, {4, 7}, {3, 6}, {6, 7}, {5, 
          4}, {2, 5}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[6, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.274355433643473, 0.49794596333510094`}, {
          3.367218488780101, 0.5062556253957285}, {2.530062777037424, 0.}, {
          1.7439065021935352`, 0.00047664907637468}, {2.5307680447593563`, 
          1.0130617528447177`}, {1.7439509204073738`, 1.0135216012509871`}, {
          0.9071226324241428, 0.5077858595970015}, {0., 
          0.49707056543331124`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036324537664]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036324537664]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03873036324537664]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03873036324537664]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036324537664]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036324537664]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07746072649075328, 0.07746072649075328}], 
            {GrayLevel[0], DiskBox[2, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[3, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[4, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[5, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[6, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[7, 0.03930782096704047]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07746072649075328, 0.07746072649075328}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 6, 3, 5, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 6}, {5, 4}, {2, 6}, {4, 7}, {3, 5}, {2, 
          3}, {6, 7}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.893775166347438, 1.058545006665348}, {
          2.915694528247547, 1.0756291081466522`}, {2.5068717703198495`, 
          0.007184978087345639}, {1.3899492395455988`, 0.}, {
          1.9430078188089612`, 0.44168242913944494`}, {1.9379763628633724`, 
          1.4588714772995928`}, {0.9748993901709971, 1.0664818021657314`}, {
          0., 1.0609833747799051`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03632343081060238]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03632343081060238]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03632343081060238]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03632343081060238]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03632343081060238]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03632343081060238]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03632343081060238]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03632343081060238]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03632343081060238]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03632343081060238]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07264686162120476, 0.07264686162120476}], 
            {GrayLevel[0], DiskBox[2, 0.03538916360650989]}, 
            {GrayLevel[0], DiskBox[3, 0.03538916360650989]}, 
            {GrayLevel[0], DiskBox[4, 0.03538916360650989]}, 
            {GrayLevel[0], DiskBox[5, 0.03538916360650989]}, 
            {GrayLevel[0], DiskBox[6, 0.03538916360650989]}, 
            {GrayLevel[0], DiskBox[7, 0.03538916360650989]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07264686162120476, 0.07264686162120476}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 5, 3, 6, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 6}, {3, 6}, {5, 4}, {2, 3}, {6, 7}, {4, 
          7}, {2, 5}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.274355433643473, 0.49794596333510094`}, {
          3.367218488780101, 0.5062556253957285}, {2.530062777037424, 0.}, {
          1.7439065021935352`, 0.00047664907637468}, {2.5307680447593563`, 
          1.0130617528447177`}, {1.7439509204073738`, 1.0135216012509871`}, {
          0.9071226324241428, 0.5077858595970015}, {0., 
          0.49707056543331124`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036324537664]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036324537664]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.03873036324537664]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03873036324537664]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036324537664]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036324537664]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036324537664]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07746072649075328, 0.07746072649075328}], 
            {GrayLevel[0], DiskBox[2, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[3, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[4, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[5, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[6, 0.03930782096704047]}, 
            {GrayLevel[0], DiskBox[7, 0.03930782096704047]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07746072649075328, 0.07746072649075328}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 5}, {4, 3}, {6, 5}, {3, 7}, {6, 4}, {5, 
          7}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.8146970080157248`, 0.270094964547705}, {
          2.8546429121838464`, 0.36272128098849865`}, {1.9019585089099569`, 
          0.}, {1.9076115834066782`, 1.1013769563057294`}, {
          1.3260564267646497`, 1.4479493200813462`}, {2.4907264540435166`, 
          1.4419671898819246`}, {0.9548673470545616, 0.3758312776987519}, {0.,
           0.26530661354951324`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03581018563525899]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03581018563525899]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.03581018563525899]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03581018563525899]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.03581018563525899]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03581018563525899]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03581018563525899]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03581018563525899]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.03581018563525899]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03581018563525899]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07162037127051799, 0.07162037127051799}], 
            {GrayLevel[0], DiskBox[2, 0.0337647989615407]}, 
            {GrayLevel[0], DiskBox[3, 0.0337647989615407]}, 
            {GrayLevel[0], DiskBox[4, 0.0337647989615407]}, 
            {GrayLevel[0], DiskBox[5, 0.0337647989615407]}, 
            {GrayLevel[0], DiskBox[6, 0.0337647989615407]}, 
            {GrayLevel[0], DiskBox[7, 0.0337647989615407]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07162037127051799, 0.07162037127051799}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 5, 3, 6, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 6}, {2, 4}, {3, 7}, {4, 6}, {6, 7}, {5, 
          3}, {2, 5}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[6, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[5, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[1, 3] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.2743558218519055`, 0.4982697553927159}, {
          3.367216636566389, 0.5060420742650629}, {1.7438360815419336`, 
          0.0005509861483867784}, {2.5300152499060187`, 0.}, {
          2.5308131464576706`, 1.0130325055265477`}, {1.7440033387581189`, 
          1.013618635197326}, {0.9071208517169125, 0.5080258956640598}, {0., 
          0.49714741741886176`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03873036564856769]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.03873036564856769]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.03873036564856769]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03873036564856769]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03873036564856769]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.03873036564856769]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03873036564856769]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03873036564856769]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03873036564856769]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.03873036564856769]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07746073129713538, 0.07746073129713538}], 
            {GrayLevel[0], DiskBox[2, 0.039308968072037334`]}, 
            {GrayLevel[0], DiskBox[3, 0.039308968072037334`]}, 
            {GrayLevel[0], DiskBox[4, 0.039308968072037334`]}, 
            {GrayLevel[0], DiskBox[5, 0.039308968072037334`]}, 
            {GrayLevel[0], DiskBox[6, 0.039308968072037334`]}, 
            {GrayLevel[0], DiskBox[7, 0.039308968072037334`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07746073129713538, 0.07746073129713538}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 5}, {6, 3}, {4, 7}, {3, 5}, {5, 7}, {6, 
          4}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[6, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 4] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.598522327325652, 0.5247790208320846}, {
          3.6335632241293476`, 0.5235496942757668}, {2.820645753909042, 0.}, {
          1.7776569213671092`, 1.0544959210609217`}, {1.776245356626343, 
          0.0028303389894658304`}, {2.8224211992699, 1.0515878913232055`}, {
          0.9649264735376399, 0.5306936622436798}, {0., 0.5200584649020372}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.04070155697690686]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.04070155697690686]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.04070155697690686]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.04070155697690686]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.04070155697690686]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.04070155697690686]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.04070155697690686]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.04070155697690686]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.04070155697690686]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.04070155697690686]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.08140311395381372, 0.08140311395381372}], 
            {GrayLevel[0], DiskBox[2, 0.048247994312851955`]}, 
            {GrayLevel[0], DiskBox[3, 0.048247994312851955`]}, 
            {GrayLevel[0], DiskBox[4, 0.048247994312851955`]}, 
            {GrayLevel[0], DiskBox[5, 0.048247994312851955`]}, 
            {GrayLevel[0], DiskBox[6, 0.048247994312851955`]}, 
            {GrayLevel[0], DiskBox[7, 0.048247994312851955`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.08140311395381372, 0.08140311395381372}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 6, 3, 5, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 6}, {2, 6}, {5, 7}, {6, 4}, {4, 7}, {3, 
          5}, {2, 3}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.274283190916821, 0.49739354893378535`}, {
          3.367157851379004, 0.5058124693940612}, {2.5298516488140805`, 0.}, {
          1.743742504212162, 0.0008015201330868749}, {1.7440445027709608`, 
          1.0138428854165964`}, {2.5308946448538796`, 1.0131150143193053`}, {
          0.9070967461466173, 0.5085668274137312}, {0., 
          0.49584558482859165`}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.038729916027433794`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.038729916027433794`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.038729916027433794`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.038729916027433794`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.038729916027433794`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.038729916027433794`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.038729916027433794`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.038729916027433794`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.038729916027433794`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.038729916027433794`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07745983205486759, 0.07745983205486759}], 
            {GrayLevel[0], DiskBox[2, 0.039305477660921635`]}, 
            {GrayLevel[0], DiskBox[3, 0.039305477660921635`]}, 
            {GrayLevel[0], DiskBox[4, 0.039305477660921635`]}, 
            {GrayLevel[0], DiskBox[5, 0.039305477660921635`]}, 
            {GrayLevel[0], DiskBox[6, 0.039305477660921635`]}, 
            {GrayLevel[0], DiskBox[7, 0.039305477660921635`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07745983205486759, 0.07745983205486759}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 5, 4, 6, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 5}, {6, 5}, {4, 7}, {5, 3}, {3, 7}, {6, 
          4}, {2, 6}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[6, 4] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {{
               Thickness[Large], 
               RGBColor[1, 0, 0], 
               Dashing[{Small, Small}]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.823380060986201, 0.3202142547795689}, {
          2.8581060538573357`, 0.3716978129524196}, {1.902524331809892, 0.}, {
          1.342897087240333, 1.4612105976674619`}, {1.9083078849663895`, 
          1.0909562586778705`}, {2.4956078100275993`, 1.453584263786467}, {
          0.9597268034309243, 0.3911244194666886}, {0., 0.3178130764376545}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.0358667681098368]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.0358667681098368]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.0358667681098368]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.0358667681098368]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.0358667681098368]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.0358667681098368]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.0358667681098368]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.0358667681098368]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.0358667681098368]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.0358667681098368]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0717335362196736, 0.0717335362196736}], 
            {GrayLevel[0], DiskBox[2, 0.033792663616746564`]}, 
            {GrayLevel[0], DiskBox[3, 0.033792663616746564`]}, 
            {GrayLevel[0], DiskBox[4, 0.033792663616746564`]}, 
            {GrayLevel[0], DiskBox[5, 0.033792663616746564`]}, 
            {GrayLevel[0], DiskBox[6, 0.033792663616746564`]}, 
            {GrayLevel[0], DiskBox[7, 0.033792663616746564`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.0717335362196736, 0.0717335362196736}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"exprpf", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"expr", " ", 
      RowBox[{"den", "[", 
       RowBox[{"k1", ",", "0", ",", "cut"}], "]"}], 
      RowBox[{"den", "[", 
       RowBox[{"k2", ",", "0", ",", "cut"}], "]"}], 
      RowBox[{"den", "[", 
       RowBox[{"k3", ",", "0", ",", "cut"}], "]"}], 
      RowBox[{"den", "[", 
       RowBox[{"k4", ",", "0", ",", "cut"}], "]"}], 
      RowBox[{"den", "[", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], "]"}]}], "/.", 
     RowBox[{"k4", "\[Rule]", 
      RowBox[{"q", "-", "k1", "-", "k2", "-", "k3"}]}]}], "//", 
    RowBox[{"PartialFraction", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"exprpf", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["1879", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprpf", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   FractionBox["1", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"d", "-", "2"}], ")"}], " ", "Q2", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", "x"}], ")"}]}]], 
   RowBox[{"16", " ", "CF", " ", 
    SuperscriptBox["d", "3"], " ", 
    SuperscriptBox["gs", "4"], " ", 
    RowBox[{"PS", "(", "4", ")"}], " ", 
    SuperscriptBox["qu", "2"], " ", "x", " ", 
    RowBox[{"k1", "\[CenterDot]", "k3"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], " ", 
    RowBox[{"k1", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ")"}]}], " ", 
    RowBox[{"k2", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ")"}]}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",", 
      "cut"}], ")"}], " ", 
    RowBox[{"den", "(", 
     RowBox[{
      RowBox[{"q", "-", "k1"}], ",", 
      RowBox[{"Q2", " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}]}]}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"exprpf", "=", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"exprpf", "//", "Expand"}], "//", "Terms"}], "//", 
         RowBox[{"Select", "[", 
          RowBox[{"NotFreeQ", "[", 
           RowBox[{"den", "[", 
            RowBox[{"k1", ",", "0", ",", "cut"}], "]"}], "]"}], "]"}]}], "//", 
        RowBox[{"Select", "[", 
         RowBox[{"NotFreeQ", "[", 
          RowBox[{"den", "[", 
           RowBox[{"k2", ",", "0", ",", "cut"}], "]"}], "]"}], "]"}]}], "//", 
       
       RowBox[{"Select", "[", 
        RowBox[{"NotFreeQ", "[", 
         RowBox[{"den", "[", 
          RowBox[{"k3", ",", "0", ",", "cut"}], "]"}], "]"}], "]"}]}], "//", 
      RowBox[{"Select", "[", 
       RowBox[{"NotFreeQ", "[", 
        RowBox[{"den", "[", 
         RowBox[{
          RowBox[{"q", "-", "k1", "-", "k2", "-", "k3"}], ",", "0", ",", 
          "cut"}], "]"}], "]"}], "]"}]}], "//", 
     RowBox[{"Select", "[", 
      RowBox[{"NotFreeQ", "[", 
       RowBox[{"den", "[", 
        RowBox[{
         RowBox[{"q", "-", "k1"}], ",", 
         RowBox[{"Q2", 
          RowBox[{"(", 
           RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], "]"}], "]"}], 
      "]"}]}], "//", 
    RowBox[{"Apply", "[", "Plus", "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"exprpf", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["1264", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprpf", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   RowBox[{
    FractionBox["1", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"d", "-", "2"}], ")"}], " ", "Q2", " ", "x"}]], 
    RowBox[{"512", " ", "CA", " ", 
     SuperscriptBox["CF", "2"], " ", 
     SuperscriptBox["gs", "4"], " ", 
     RowBox[{"PS", "(", "4", ")"}], " ", 
     SuperscriptBox["qu", "2"], " ", 
     RowBox[{"k1", "\[CenterDot]", "k2"}], " ", 
     RowBox[{"k1", "\[CenterDot]", "k3"}], " ", 
     RowBox[{"k2", "\[CenterDot]", "k3"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], " ", 
     SuperscriptBox[
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], "2"], " ", 
     RowBox[{"k1", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ")"}]}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",", 
       "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{"q", "-", "k1"}], ",", 
       RowBox[{"Q2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}]}]}]}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"exprpf", "//", "Topologies"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", "TableForm"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{"FindGraph", "[", 
     RowBox[{
      RowBox[{"{", "q", "}"}], ",", 
      RowBox[{"{", "q", "}"}], ",", 
      RowBox[{
       RowBox[{"#", "//", 
        RowBox[{"DeleteCases", "[", 
         RowBox[{"den", "[", 
          RowBox[{
           RowBox[{"q", "-", "k1"}], ",", "__"}], "]"}], "]"}]}], "//", 
       RowBox[{"ReplaceAll", "[", 
        RowBox[{
         RowBox[{"den", "[", 
          RowBox[{"p_", ",", "__"}], "]"}], "\[RuleDelayed]", "p"}], "]"}]}], 
      ",", 
      RowBox[{"EdgeFn", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"#1", ",", 
           RowBox[{"#2", "/.", 
            RowBox[{"{", "\[IndentingNewLine]", 
             RowBox[{
              RowBox[{"_Incoming", "\[Rule]", "Gray"}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{"_Outgoing", "\[Rule]", "Gray"}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{
               RowBox[{"Internal", "[", "k1", "]"}], "\[Rule]", 
               RowBox[{"{", 
                RowBox[{"Thick", ",", "Dashed", ",", "Red"}], "}"}]}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{
               RowBox[{"Internal", "[", 
                RowBox[{"k2", "|", "k3", "|", 
                 RowBox[{"q", "-", "k1", "-", "k2", "-", "k3"}]}], "]"}], 
               "\[Rule]", "Dashed"}], ",", "\[IndentingNewLine]", 
              RowBox[{"_Internal", "\[Rule]", "Thick"}]}], 
             "\[IndentingNewLine]", "}"}]}]}], "]"}], "&"}], ")"}]}]}], "]"}],
     "&"}], "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k2"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k2"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 6, 3, 5, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 6}, {2, 5}, {4, 7}, {6, 3}, {5, 3}, {6, 
          7}, {2, 4}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[5, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[4, 6] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.810891326176811, 0.21288146965883242`}, {
          2.8563428287313455`, 0.34451597763968467`}, {1.9055666902986421`, 
          1.0973876253626975`}, {1.8989693623132038`, 0.}, {
          2.4957558301097023`, 1.4249052775343602`}, {1.3229619512482746`, 
          1.43545069072827}, {0.948033828305193, 0.3631179111642913}, {0., 
          0.2067379649175119}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.035785368503188336`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.035785368503188336`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.035785368503188336`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.035785368503188336`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.035785368503188336`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.035785368503188336`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07157073700637667, 0.07157073700637667}], 
            {GrayLevel[0], DiskBox[2, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[3, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[4, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[5, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[6, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[7, 0.03367918192772481]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07157073700637667, 0.07157073700637667}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 6, 3, 4, 2, 5, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {2, 5}, {4, 6}, {2, 7}, {5, 3}, {7, 3}, {5, 
          4}, {7, 6}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {
              Thickness[Large]}, UndirectedEdge[4, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 4] -> {
              Thickness[Large]}, UndirectedEdge[5, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{0., 0.}, {0.7382446548809886, 
          0.5784974052200702}, {1.6629358999316872`, 1.875679959796779}, {
          2.287372646435766, 1.5579567693172267`}, {1.038820631453675, 
          1.5499687047380497`}, {2.60207252679494, 0.5903693499925401}, {
          1.6700266986001722`, 0.5146738548880586}, {3.3556708784027465`, 
          0.008544624635647868}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03273694923765445]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.03273694923765445]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 7}, 0.03273694923765445]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03273694923765445]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03273694923765445]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.03273694923765445]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03273694923765445]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.03273694923765445]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03273694923765445]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 8}, 0.03273694923765445]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.0654738984753089, 0.0654738984753089}], 
            {GrayLevel[0], DiskBox[2, 0.035031031820127186`]}, 
            {GrayLevel[0], DiskBox[3, 0.035031031820127186`]}, 
            {GrayLevel[0], DiskBox[4, 0.035031031820127186`]}, 
            {GrayLevel[0], DiskBox[5, 0.035031031820127186`]}, 
            {GrayLevel[0], DiskBox[6, 0.035031031820127186`]}, 
            {GrayLevel[0], DiskBox[7, 0.035031031820127186`]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.0654738984753089, 0.0654738984753089}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 6, 3, 5, 4, 2, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {5, 6}, {2, 5}, {4, 7}, {5, 3}, {6, 3}, {6, 
          7}, {2, 4}, {7, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[3, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[6, 3] -> {
              Thickness[Large]}, UndirectedEdge[5, 4] -> {
              Thickness[Large]}, UndirectedEdge[5, 6] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[4, 6] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.810891326176811, 0.21288146965883242`}, {
          2.8563428287313455`, 0.34451597763968467`}, {1.9055666902986421`, 
          1.0973876253626975`}, {1.8989693623132038`, 0.}, {
          2.4957558301097023`, 1.4249052775343602`}, {1.3229619512482746`, 
          1.43545069072827}, {0.948033828305193, 0.3631179111642913}, {0., 
          0.2067379649175119}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.035785368503188336`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 4}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 5}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.035785368503188336`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.035785368503188336`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 6}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.035785368503188336`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.035785368503188336`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.035785368503188336`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{7, 8}, 0.035785368503188336`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07157073700637667, 0.07157073700637667}], 
            {GrayLevel[0], DiskBox[2, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[3, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[4, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[5, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[6, 0.03367918192772481]}, 
            {GrayLevel[0], DiskBox[7, 0.03367918192772481]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07157073700637667, 0.07157073700637667}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 6, 4, 2, 5, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 5}, {5, 6}, {2, 7}, {3, 5}, {7, 4}, {3, 
          4}, {7, 6}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 5] -> {
              Thickness[Large]}, UndirectedEdge[3, 4] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}, UndirectedEdge[6, 4] -> {
              Thickness[Large]}, UndirectedEdge[5, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 3] -> {
              Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{3.503475862146792, 0.0002776932267058996}, {
          2.6771075294591515`, 0.47052418469640717`}, {2.372636774743663, 
          1.4442709762467936`}, {1.7516089055095183`, 1.7411872850428}, {
          1.1306712447094447`, 1.4443571673849505`}, {0.8259705445550408, 
          0.47092413791231874`}, {1.751535428665432, 0.36780995684569506`}, {
          0., 0.}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.03374442374167848]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.03374442374167848]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 7}, 0.03374442374167848]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.03374442374167848]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.03374442374167848]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.03374442374167848]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.03374442374167848]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.03374442374167848]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.03374442374167848]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 8}, 0.03374442374167848]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.06748884748335696, 0.06748884748335696}], 
            {GrayLevel[0], DiskBox[2, 0.03441190554694117]}, 
            {GrayLevel[0], DiskBox[3, 0.03441190554694117]}, 
            {GrayLevel[0], DiskBox[4, 0.03441190554694117]}, 
            {GrayLevel[0], DiskBox[5, 0.03441190554694117]}, 
            {GrayLevel[0], DiskBox[6, 0.03441190554694117]}, 
            {GrayLevel[0], DiskBox[7, 0.03441190554694117]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.06748884748335696, 0.06748884748335696}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 3, 6, 2, 4, 5, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {2, 3}, {4, 5}, {2, 6}, {7, 5}, {3, 7}, {6, 4}, {3, 
          4}, {6, 7}, {5, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Thickness[Large]}, UndirectedEdge[6, 2] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Thickness[Large]}, UndirectedEdge[4, 5] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 5] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[4, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[1, 3] -> {
              Thickness[Large]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.264430820254448, 0.5960692886569563}, {
          3.3554596140144017`, 0.45415229183597833`}, {2.5427183402114975`, 
          1.041451105136273}, {1.753186013537033, 1.0382297698644312`}, {
          0.9022616220512496, 0.5599894078040809}, {2.5177957777238724`, 
          0.}, {1.7333500568722469`, 0.05688177970724889}, {0., 
          0.5360114633065131}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.038668891321340965`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.038668891321340965`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 6}, 0.038668891321340965`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.038668891321340965`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 7}, 0.038668891321340965`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{4, 5}, 0.038668891321340965`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.038668891321340965`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.038668891321340965`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{5, 8}, 0.038668891321340965`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{6, 7}, 0.038668891321340965`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07733778264268193, 0.07733778264268193}], 
            {GrayLevel[0], DiskBox[2, 0.03932526623638709]}, 
            {GrayLevel[0], DiskBox[3, 0.03932526623638709]}, 
            {GrayLevel[0], DiskBox[4, 0.03932526623638709]}, 
            {GrayLevel[0], DiskBox[5, 0.03932526623638709]}, 
            {GrayLevel[0], DiskBox[6, 0.03932526623638709]}, 
            {GrayLevel[0], DiskBox[7, 0.03932526623638709]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07733778264268193, 0.07733778264268193}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 4, 3, 5, 2, 6, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {2, 3}, {5, 6}, {7, 5}, {4, 7}, {3, 5}, {4, 
          6}, {2, 7}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Thickness[Large]}, UndirectedEdge[3, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Thickness[Large]}, UndirectedEdge[4, 5] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[3, 6] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.274383468606779, 0.6079698427074953}, {
          3.3594381930130948`, 0.46400877526934037`}, {2.5274866522307633`, 
          1.0404137095355401`}, {1.7545879004029217`, 1.0276001728483508`}, {
          1.7397898185093301`, 0.044434553297110935`}, {0.9012210391017212, 
          0.5351020633384811}, {2.5252102302361283`, 0.}, {0., 
          0.5655151777563268}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.038730536794593984`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 7}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.038730536794593984`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.038730536794593984`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.038730536794593984`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.038730536794593984`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 8}, 0.038730536794593984`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07746107358918797, 0.07746107358918797}], 
            {GrayLevel[0], DiskBox[2, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[3, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[4, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[5, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[6, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[7, 0.03865024797137366]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07746107358918797, 0.07746107358918797}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None], ",", 
    GraphicsBox[
     NamespaceBox["NetworkGraphics",
      DynamicModuleBox[{Typeset`graph = HoldComplete[
        Graph[{
          $CellContext`II[1], 1, 4, 3, 5, 2, 6, 
          $CellContext`OO[1]}, {
         Null, {{1, 2}, {3, 4}, {2, 3}, {5, 6}, {7, 5}, {3, 5}, {4, 7}, {4, 
          6}, {2, 7}, {6, 8}}}, {EdgeStyle -> {UndirectedEdge[2, 
              $CellContext`OO[1]] -> {
              GrayLevel[0.5]}, UndirectedEdge[5, 2] -> {
              Thickness[Large]}, UndirectedEdge[3, 2] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[4, 3] -> {
              Thickness[Large]}, UndirectedEdge[6, 5] -> {
              Thickness[Large]}, UndirectedEdge[1, 4] -> {
              Thickness[Large]}, UndirectedEdge[3, 6] -> {
              Dashing[{Small, Small}]}, UndirectedEdge[4, 5] -> {{
               Thickness[Large], 
               Dashing[{Small, Small}], 
               RGBColor[1, 0, 0]}}, UndirectedEdge[
              $CellContext`II[1], 1] -> {
              GrayLevel[0.5]}, UndirectedEdge[1, 6] -> {
              Dashing[{Small, Small}]}}, GraphLayout -> "SpringEmbedding", 
          VertexShape -> {$CellContext`II[1] -> 
            Graphics[{}], $CellContext`OO[1] -> Graphics[{}]}, 
          VertexSize -> {
           6 -> 0.1, 4 -> 0.1, 3 -> 0.1, 2 -> 0.1, 1 -> 0.1, 5 -> 0.1}, 
          VertexStyle -> {6 -> {
              GrayLevel[0]}, 1 -> {
              GrayLevel[0]}, 2 -> {
              GrayLevel[0]}, 3 -> {
              GrayLevel[0]}, 4 -> {
              GrayLevel[0]}, 5 -> {
              GrayLevel[0]}}}]]}, 
       TagBox[
        GraphicsGroupBox[
         GraphicsComplexBox[{{4.274383468606779, 0.6079698427074953}, {
          3.3594381930130948`, 0.46400877526934037`}, {2.5274866522307633`, 
          1.0404137095355401`}, {1.7545879004029217`, 1.0276001728483508`}, {
          1.7397898185093301`, 0.044434553297110935`}, {0.9012210391017212, 
          0.5351020633384811}, {2.5252102302361283`, 0.}, {0., 
          0.5655151777563268}}, {
           {Hue[0.6, 0.7, 0.5], Opacity[0.7], 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{1, 2}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{2, 3}, 0.038730536794593984`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{2, 7}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{3, 4}, 0.038730536794593984`]}}, 
            {RGBColor[1, 0, 0], Thickness[Large], Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{3, 5}, 0.038730536794593984`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 6}, 0.038730536794593984`]}}, 
            {Dashing[{Small, Small}], 
             {Arrowheads[0.], ArrowBox[{4, 7}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 6}, 0.038730536794593984`]}}, 
            {Thickness[Large], 
             {Arrowheads[0.], ArrowBox[{5, 7}, 0.038730536794593984`]}}, 
            {GrayLevel[0.5], 
             {Arrowheads[0.], ArrowBox[{6, 8}, 0.038730536794593984`]}}}, 
           {Hue[0.6, 0.2, 0.8], EdgeForm[{GrayLevel[0], Opacity[0.7]}], 
            InsetBox[
             GraphicsBox[{}], 1, 
             Automatic, {0.07746107358918797, 0.07746107358918797}], 
            {GrayLevel[0], DiskBox[2, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[3, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[4, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[5, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[6, 0.03865024797137366]}, 
            {GrayLevel[0], DiskBox[7, 0.03865024797137366]}, InsetBox[
             GraphicsBox[{}], 8, 
             Automatic, {0.07746107358918797, 0.07746107358918797}]}}]],
        MouseAppearanceTag["NetworkGraphics"]],
       AllowKernelInitialization->False]],
     DefaultBaseStyle->{
      "NetworkGraphics", FrontEnd`GraphicsHighlightColor -> Hue[0.8, 1., 0.6]},
     FormatType->TraditionalForm,
     FrameTicks->None]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "Why", " ", "do", " ", "we", " ", "get", " ", "less", " ", "topologies", 
    " ", "than", " ", 
    RowBox[{"before", "?"}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"den", "[", 
      RowBox[{
       RowBox[{"k2", "+", "k3"}], ",", "0"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{
       RowBox[{"k1", "+", "k4"}], ",", "0"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{
       RowBox[{"k2", "+", "k3", "+", "k4"}], ",", "0"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{"k1", ",", "0", ",", "cut"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{"k2", ",", "0", ",", "cut"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{"k3", ",", "0", ",", "cut"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{"k4", ",", "0", ",", "cut"}], "]"}], 
     RowBox[{"den", "[", 
      RowBox[{
       RowBox[{"q", "-", "k1"}], ",", 
       RowBox[{"Q2", 
        RowBox[{"(", 
         RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], "]"}]}], "/.", 
    RowBox[{"k4", "\[Rule]", 
     RowBox[{"q", "-", "k1", "-", "k2", "-", "k3"}]}]}], "//", 
   RowBox[{"PartialFraction", "[", "externalsprules", "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   FractionBox[
    RowBox[{
     RowBox[{"den", "(", 
      RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "k2"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",", 
       "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{"q", "-", "k1"}], ",", 
       RowBox[{"Q2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}]}], 
    RowBox[{"Q2", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", "x"}], ")"}]}]], "-", 
   FractionBox[
    RowBox[{
     RowBox[{"den", "(", 
      RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{"q", "-", "k1"}], ",", "0"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "k2"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], " ", 
     RowBox[{"den", "(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",", 
       "cut"}], ")"}]}], 
    RowBox[{"Q2", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", "x"}], ")"}]}]]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"topologies", "=", 
   RowBox[{"topologies", "//", 
    RowBox[{"Map", "[", 
     RowBox[{
      RowBox[{"CompleteDenominatorSet", "[", 
       RowBox[{"#", ",", 
        RowBox[{"{", "q", "}"}], ",", 
        RowBox[{"-", 
         RowBox[{"{", 
          RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}]}]}], "]"}], "&"}], 
     "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"topologies", "//", "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[GridBox[{
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k2"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k2"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]},
     {
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"k1", "+", "k2", "+", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k3"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k3", "+", "q"}], ",", "0"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k1", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k2", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{"k3", ",", "0", ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{"q", "-", "k1"}], ",", 
        RowBox[{"Q2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", "x"}], ")"}]}], ",", "cut"}], ")"}], 
      RowBox[{"den", "(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "k1"}], "-", "k2", "-", "k3", "+", "q"}], ",", "0", ",",
         "cut"}], ")"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"exprb", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"exprpf", "//", 
      RowBox[{
       RowBox[{"ToTopologies", "[", 
        RowBox[{"#", ",", "topologies", ",", 
         RowBox[{"{", "q", "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}]}], "]"}], "&"}]}], "//",
      "TM"}], "//", 
    RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"exprb", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"ToTopologies", "(", 
         RowBox[{"#1", ",", "topologies", ",", 
          RowBox[{"{", "q", "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"k1", ",", "k2", ",", "k3"}], "}"}]}], ")"}], "&"}], ")"}],
       "\[InvisibleApplication]", 
      RowBox[{"(", "exprpf", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "12.297571`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ($CellContext`ToTopologies[#, $CellContext`topologies, {$CellContext`q}, \
{$CellContext`k1, $CellContext`k2, $CellContext`k3}]& )[$CellContext`exprpf]],
     ": ", 12.297571, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["30602", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprb", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"4", " ", "CA", " ", 
   SuperscriptBox["CF", "2"], " ", 
   SuperscriptBox["d", "2"], " ", 
   SuperscriptBox["gs", "4"], " ", 
   RowBox[{"PS", "(", "4", ")"}], " ", 
   SuperscriptBox["qu", "2"], " ", "x", " ", 
   SubsuperscriptBox["B", 
    RowBox[{
    "0", ",", "1", ",", "0", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1",
      ",", 
     RowBox[{"-", "1"}]}], "5"]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"CutPropagatorZeroPattern", "[", "topologies_List", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"i", ",", "j"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"Table", "[", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"MatchQ", "[", 
           RowBox[{
            RowBox[{"topologies", "[", 
             RowBox[{"[", 
              RowBox[{"i", ",", "j"}], "]"}], "]"}], ",", 
            RowBox[{"den", "[", 
             RowBox[{"_", ",", "_", ",", "cut"}], "]"}]}], "]"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"B", "[", 
           RowBox[{"i", ",", 
            RowBox[{
             RowBox[{"Range", "[", 
              RowBox[{"Length", "[", 
               RowBox[{"topologies", "[", 
                RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "/.", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"j", "\[Rule]", 
                RowBox[{"_", "?", "NonPositive"}]}], ",", 
               RowBox[{"_Integer", "\[Rule]", "_"}]}], "}"}]}]}], "]"}], ",", 
          "\[IndentingNewLine]", "Nothing"}], "]"}], "\[IndentingNewLine]", 
        ",", "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"i", ",", 
          RowBox[{"Length", "[", "topologies", "]"}]}], "}"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"{", 
         RowBox[{"j", ",", 
          RowBox[{"Length", "[", 
           RowBox[{"topologies", "[", 
            RowBox[{"[", "i", "]"}], "]"}], "]"}]}], "}"}]}], 
       "\[IndentingNewLine]", "]"}], "//", "Flatten"}], "//", 
     RowBox[{"Apply", "[", "Alternatives", "]"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CutPropagatorZeroPattern", "[", "topologies", "]"}], "//", 
  "StandardForm"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"B", "[", 
   RowBox[{"1", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"1", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"1", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"1", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"1", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"2", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"2", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"2", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"2", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"2", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"3", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"3", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"3", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"3", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"3", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"4", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"4", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"4", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"4", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"4", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"5", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"5", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"5", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"5", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"5", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"6", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"6", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"6", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"6", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"6", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"7", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_", ",", 
      "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"7", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_", ",", "_"}], 
     "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"7", ",", 
    RowBox[{"{", 
     RowBox[{"_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_", ",", "_"}], "}"}]}], "]"}],
   "|", 
  RowBox[{"B", "[", 
   RowBox[{"7", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      RowBox[{"_", "?", "NonPositive"}], ",", "_"}], "}"}]}], "]"}], "|", 
  RowBox[{"B", "[", 
   RowBox[{"7", ",", 
    RowBox[{"{", 
     RowBox[{
     "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", "_", ",", 
      "_", ",", 
      RowBox[{"_", "?", "NonPositive"}]}], "}"}]}], "]"}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"exprb", "=", 
   RowBox[{"exprb", "/.", 
    RowBox[{
     RowBox[{"CutPropagatorZeroPattern", "[", "topologies", "]"}], "\[Rule]", 
     "0"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"exprb", "//", "Length"}]}], "Input"],

Cell[BoxData[
 FormBox["4107", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Kira config files", "Subsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"IndexStatistics", "[", "ex_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "minidx", ",", "maxidx", ",", "maxs", ",", "maxr", ",", "maxdots", ",", 
      "topid", ",", "idx", ",", "r", ",", "s", ",", "dots"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"minidx", "=", 
      RowBox[{"maxidx", "=", 
       RowBox[{"maxs", "=", 
        RowBox[{"maxr", "=", 
         RowBox[{"maxdots", "=", 
          RowBox[{"<|", "|>"}]}]}]}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"ex", "//", 
        RowBox[{
         RowBox[{"Cases", "[", 
          RowBox[{"#", ",", "_B", ",", 
           RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}], "//", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{"(", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"{", 
            RowBox[{"topid", ",", "idx"}], "}"}], "=", 
           RowBox[{"List", "@@", "#"}]}], ";", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"minidx", "[", "topid", "]"}], "=", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"MatchQ", "[", 
              RowBox[{
               RowBox[{"minidx", "[", "topid", "]"}], ",", "_Missing"}], 
              "]"}], ",", "idx", ",", 
             RowBox[{"MapThread", "[", 
              RowBox[{"Min", ",", 
               RowBox[{"{", 
                RowBox[{
                 RowBox[{"minidx", "[", "topid", "]"}], ",", "idx"}], "}"}]}],
               "]"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"maxidx", "[", "topid", "]"}], "=", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"MatchQ", "[", 
              RowBox[{
               RowBox[{"maxidx", "[", "topid", "]"}], ",", "_Missing"}], 
              "]"}], ",", "idx", ",", 
             RowBox[{"MapThread", "[", 
              RowBox[{"Max", ",", 
               RowBox[{"{", 
                RowBox[{
                 RowBox[{"maxidx", "[", "topid", "]"}], ",", "idx"}], "}"}]}],
               "]"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
          RowBox[{"r", "=", 
           RowBox[{"Sum", "[", 
            RowBox[{
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{"i", ">", "0"}], ",", "i", ",", "0"}], "]"}], ",", 
             RowBox[{"{", 
              RowBox[{"i", ",", "idx"}], "}"}]}], "]"}]}], ";", 
          "\[IndentingNewLine]", 
          RowBox[{"s", "=", 
           RowBox[{"Sum", "[", 
            RowBox[{
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "i"}], ">", "0"}], ",", 
               RowBox[{"-", "i"}], ",", "0"}], "]"}], ",", 
             RowBox[{"{", 
              RowBox[{"i", ",", "idx"}], "}"}]}], "]"}]}], ";", 
          "\[IndentingNewLine]", 
          RowBox[{"dots", "=", 
           RowBox[{"Sum", "[", 
            RowBox[{
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{"i", ">", "1"}], ",", 
               RowBox[{"i", "-", "1"}], ",", "0"}], "]"}], ",", 
             RowBox[{"{", 
              RowBox[{"i", ",", "idx"}], "}"}]}], "]"}]}], ";", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"maxr", "[", "topid", "]"}], "=", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"MatchQ", "[", 
              RowBox[{
               RowBox[{"maxr", "[", "topid", "]"}], ",", "_Missing"}], "]"}], 
             ",", "r", ",", 
             RowBox[{"Max", "[", 
              RowBox[{
               RowBox[{"maxr", "[", "topid", "]"}], ",", "r"}], "]"}]}], 
            "]"}]}], ";", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"maxs", "[", "topid", "]"}], "=", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"MatchQ", "[", 
              RowBox[{
               RowBox[{"maxs", "[", "topid", "]"}], ",", "_Missing"}], "]"}], 
             ",", "s", ",", 
             RowBox[{"Max", "[", 
              RowBox[{
               RowBox[{"maxs", "[", "topid", "]"}], ",", "s"}], "]"}]}], 
            "]"}]}], ";", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"maxdots", "[", "topid", "]"}], "=", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"MatchQ", "[", 
              RowBox[{
               RowBox[{"maxdots", "[", "topid", "]"}], ",", "_Missing"}], 
              "]"}], ",", "dots", ",", 
             RowBox[{"Max", "[", 
              RowBox[{
               RowBox[{"maxdots", "[", "topid", "]"}], ",", "dots"}], "]"}]}],
             "]"}]}], ";"}], "\[IndentingNewLine]", ")"}], "&"}], "]"}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"Table", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"<|", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"\"\<minidx\>\"", "\[Rule]", 
          RowBox[{"minidx", "[", "i", "]"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"\"\<maxidx\>\"", "\[Rule]", 
          RowBox[{"maxidx", "[", "i", "]"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"\"\<maxs\>\"", "\[Rule]", 
          RowBox[{"maxs", "[", "i", "]"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"\"\<maxr\>\"", "\[Rule]", 
          RowBox[{"maxr", "[", "i", "]"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"\"\<maxdots\>\"", "\[Rule]", 
          RowBox[{"maxdots", "[", "i", "]"}]}]}], "\[IndentingNewLine]", 
        "|>"}], ",", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"i", ",", 
         RowBox[{
          RowBox[{"maxr", "//", "Keys"}], "//", 
          RowBox[{"Apply", "[", "Max", "]"}]}]}], "}"}]}], "]"}]}]}], 
   "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"exprb", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"16", " ", "CF", " ", 
   SuperscriptBox["gs", "4"], " ", 
   RowBox[{"PS", "(", "4", ")"}], " ", 
   SuperscriptBox["Q2", "2"], " ", 
   SuperscriptBox["qu", "2"], " ", 
   SuperscriptBox["x", "2"], " ", 
   SubsuperscriptBox["B", 
    RowBox[{
    "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1",
      ",", "1"}], "1"]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "To", " ", "know", " ", "how", " ", "big", " ", "the", " ", "IBP", " ", 
    "table", " ", "should", " ", "we", " ", "request"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{"stats", "=", 
   RowBox[{"IndexStatistics", "[", "exprb", "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "4"}], ",", 
         RowBox[{"-", "1"}], ",", 
         RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "2", ",", "1", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "4"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "9"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "2"}]}], "\[RightAssociation]"}], 
    ",", 
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "1"}], ",", 
         RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "3"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "8"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "0"}]}], "\[RightAssociation]"}], 
    ",", 
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "2"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "2", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "3"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "9"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "2"}]}], "\[RightAssociation]"}], 
    ",", 
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2"}], ",", 
         RowBox[{"-", "2"}], ",", 
         RowBox[{"-", "2"}], ",", 
         RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "2"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "8"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "0"}]}], "\[RightAssociation]"}], 
    ",", 
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2"}], ",", 
         RowBox[{"-", "1"}], ",", 
         RowBox[{"-", "1"}], ",", 
         RowBox[{"-", "2"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "2"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "9"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "0"}]}], "\[RightAssociation]"}], 
    ",", 
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "1"}], ",", 
         RowBox[{"-", "1"}], ",", 
         RowBox[{"-", "3"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "3"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "9"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "0"}]}], "\[RightAssociation]"}], 
    ",", 
    RowBox[{"\[LeftAssociation]", 
     RowBox[{
      RowBox[{"\<\"minidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3"}], ",", 
         RowBox[{"-", "2"}], ",", 
         RowBox[{"-", "2"}], ",", 
         RowBox[{"-", "3"}], ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxidx\"\>", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "}"}]}], ",", 
      RowBox[{"\<\"maxs\"\>", "\[Rule]", "3"}], ",", 
      RowBox[{"\<\"maxr\"\>", "\[Rule]", "9"}], ",", 
      RowBox[{"\<\"maxdots\"\>", "\[Rule]", "0"}]}], 
     "\[RightAssociation]"}]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Do", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"toponame", "=", 
      RowBox[{"\"\<ps4l0t\>\"", "<>", 
       RowBox[{"ToString", "[", "i", "]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"dirname", "=", 
      RowBox[{"\"\<kira.\>\"", "<>", "toponame", "<>", "\"\<.dir\>\""}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"dens", "=", 
      RowBox[{"topologies", "[", 
       RowBox[{"[", "i", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Quiet", "[", 
      RowBox[{
       RowBox[{"CreateDirectory", "[", 
        RowBox[{"dirname", "<>", "\"\</config\>\""}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"CreateDirectory", "::", "filex"}], "}"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"topsector", "=", 
      RowBox[{"Sum", "[", 
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"MatchQ", "[", 
           RowBox[{
            RowBox[{"dens", "[", 
             RowBox[{"[", "i", "]"}], "]"}], ",", 
            RowBox[{"den", "[", 
             RowBox[{"_", ",", "_", ",", "irred"}], "]"}]}], "]"}], ",", "0", 
          ",", 
          RowBox[{"2", "^", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}]}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"i", ",", 
          RowBox[{"Length", "[", "dens", "]"}]}], "}"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"config", "/", 
       RowBox[{"kinematics", ".", "yaml"}]}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"Note", ":", " ", 
       RowBox[{
        RowBox[{"can", "'"}], "t", " ", "use", " ", "uppercase", " ", 
        "letters", " ", "in", " ", "variables", " ", "names", " ", "with", 
        " ", 
        RowBox[{"Kira", ".", " ", "Will"}], " ", "rename", " ", "Q2", " ", 
        "into", " ", 
        RowBox[{"q2", "."}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
     RowBox[{"Put", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"MkString", "[", "\[IndentingNewLine]", 
         RowBox[{
         "\"\<kinematics:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\< incoming_momenta: [q]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\< kinematic_invariants:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<  - [x, 0]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<  - [q2, 2]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\< scalarproduct_rules:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<  - [[q,q], q2]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\< symbol_to_replace_by_one: q2\\n\>\""}], "\[IndentingNewLine]",
          "]"}], "//", "OutputForm"}], ",", 
       RowBox[{"dirname", "<>", "\"\</config/kinematics.yaml\>\""}]}], "]"}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"config", "/", 
       RowBox[{"integralfamilies", ".", "yaml"}]}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"Put", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"MkString", "[", "\[IndentingNewLine]", 
         RowBox[{
         "\"\<integralfamilies:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\< - name: \\\"\>\"", ",", "toponame", ",", "\"\<\\\"\\n\>\"", 
          ",", "\[IndentingNewLine]", 
          "\"\<   loop_momenta: [k1, k2, k3]\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<   top_level_sectors: [\>\"", ",", 
          "topsector", ",", "\"\<]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<   magic_relations: true\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<   propagators:\\n\>\"", ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"topologies", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "/.", 
            RowBox[{"Q2", "\[Rule]", "q2"}]}], "/.", 
           RowBox[{
            RowBox[{"den", "[", 
             RowBox[{"p_", ",", "m_", ",", "___"}], "]"}], "\[RuleDelayed]", 
            RowBox[{"{", "\[IndentingNewLine]", 
             RowBox[{"\"\<    - [\\\"\>\"", ",", 
              RowBox[{"CForm", "[", "p", "]"}], ",", "\"\<\\\", \\\"\>\"", 
              ",", 
              RowBox[{"CForm", "[", "m", "]"}], ",", "\"\<\\\"]\\n\>\""}], 
             "\[IndentingNewLine]", "}"}]}]}], ",", "\[IndentingNewLine]", 
          "\"\<   cut_propagators: [\>\"", ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"topologies", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "//", 
            RowBox[{
             RowBox[{"MapIndexed", "[", 
              RowBox[{
               RowBox[{
                RowBox[{"If", "[", 
                 RowBox[{
                  RowBox[{"MatchQ", "[", 
                   RowBox[{"#1", ",", 
                    RowBox[{"den", "[", 
                    RowBox[{"_", ",", "_", ",", "cut"}], "]"}]}], "]"}], ",", 
                  
                  RowBox[{"#2", "[", 
                   RowBox[{"[", "1", "]"}], "]"}], ",", "Nothing"}], "]"}], 
                "&"}], ",", "#"}], "]"}], "&"}]}], "//", 
           RowBox[{
            RowBox[{"Riffle", "[", 
             RowBox[{"#", ",", "\"\<, \>\""}], "]"}], "&"}]}], ",", 
          "\[IndentingNewLine]", "\"\<]\\n\>\""}], "\[IndentingNewLine]", 
         "]"}], "//", "OutputForm"}], ",", 
       RowBox[{"dirname", "<>", "\"\</config/integralfamilies.yaml\>\""}]}], 
      "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"jobs", ".", "yaml"}], " ", "*)"}], "\[IndentingNewLine]", 
     RowBox[{"topr", "=", 
      RowBox[{"Max", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Length", "[", 
          RowBox[{"dens", "//", 
           RowBox[{"DeleteCases", "[", 
            RowBox[{"den", "[", 
             RowBox[{"_", ",", "_", ",", "irred"}], "]"}], "]"}]}], "]"}], 
         "+", "1"}], ",", 
        RowBox[{
         RowBox[{"stats", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "[", "\"\<maxr\>\"", "]"}]}], 
       "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"tops", "=", 
      RowBox[{"Max", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"stats", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "[", "\"\<maxs\>\"", "]"}], ",", 
        "1"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"topd", "=", 
      RowBox[{"Max", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"stats", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "[", "\"\<maxdots\>\"", "]"}], ",", 
        "1"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Put", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"MkString", "[", 
         RowBox[{
         "\"\<jobs:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\< - reduce_sectors:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<    reduce:\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<     - {topologies: [\>\"", ",", "toponame", ",", 
          "\"\<], sectors: [\>\"", ",", "topsector", ",", "\"\<], r: \>\"", 
          ",", "topr", ",", "\"\<, s: \>\"", ",", "tops", ",", "\"\<}\\n\>\"",
           ",", "\[IndentingNewLine]", "\"\<    select_integrals:\\n\>\"", 
          ",", "\[IndentingNewLine]", 
          "\"\<     select_mandatory_list:\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<      - [\>\"", ",", "toponame", ",", 
          "\"\<, \\\"\>\"", ",", "toponame", ",", 
          "\"\<.integrals\\\"]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<     select_mandatory_recursively:\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<      - {topologies: [\>\"", ",", 
          "toponame", ",", "\"\<], sectors: [\>\"", ",", "topsector", ",", 
          "\"\<], r: \>\"", ",", "topr", ",", "\"\<, s: \>\"", ",", "tops", 
          ",", "\"\<, d: \>\"", ",", "topd", ",", "\"\<}\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<    integral_ordering: 1\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\< - kira2math:\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<    target:\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<     - [\>\"", ",", "toponame", ",", 
          "\"\<, \\\"\>\"", ",", "toponame", ",", 
          "\"\<.integrals\\\"]\\n\>\"", ",", "\[IndentingNewLine]", 
          "\"\<     - {topologies: [\>\"", ",", "toponame", ",", 
          "\"\<], sectors: [\>\"", ",", "topsector", ",", "\"\<], r: \>\"", 
          ",", "topr", ",", "\"\<, s: \>\"", ",", "tops", ",", 
          "\"\<, d: \>\"", ",", "topd", ",", "\"\<}\\n\>\"", ",", 
          "\[IndentingNewLine]", "\"\<    reconstruct_mass: false\\n\>\""}], 
         " ", 
         RowBox[{"(*", " ", 
          RowBox[{
           RowBox[{"Way", " ", "too", " ", "slow"}], ";", " ", 
           RowBox[{
           "will", " ", "need", " ", "to", " ", "restore", " ", "Q2", " ", 
            "factors", " ", "manually"}]}], " ", "*)"}], 
         "\[IndentingNewLine]", "]"}], "//", "OutputForm"}], ",", 
       RowBox[{"dirname", "<>", "\"\</jobs.yaml\>\""}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"The", " ", "integral", " ", "list"}], " ", "*)"}], 
     "\[IndentingNewLine]", 
     RowBox[{"Put", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"MkString", "[", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"exprb", "//", 
            RowBox[{
             RowBox[{"Cases", "[", 
              RowBox[{"#", ",", 
               RowBox[{
                RowBox[{"B", "[", 
                 RowBox[{"i", ",", "idx_List"}], "]"}], "\[RuleDelayed]", 
                "idx"}], ",", 
               RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}], "//", 
          RowBox[{"Map", "[", 
           RowBox[{
            RowBox[{"{", "\[IndentingNewLine]", 
             RowBox[{"\"\<- [\>\"", ",", 
              RowBox[{"Riffle", "[", 
               RowBox[{"#", ",", "\"\<,\>\""}], "]"}], ",", "\"\<]\\n\>\""}], 
             "\[IndentingNewLine]", "}"}], "&"}], "]"}]}], "]"}], "//", 
        "OutputForm"}], ",", 
       RowBox[{
       "dirname", "<>", "\"\</\>\"", "<>", "toponame", "<>", 
        "\"\<.integrals\>\""}]}], "]"}], ";"}], "\[IndentingNewLine]", ",", 
    "\[IndentingNewLine]", 
    RowBox[{"{", 
     RowBox[{"i", ",", 
      RowBox[{"Length", "[", "topologies", "]"}]}], "}"}]}], 
   "\[IndentingNewLine]", "]"}], ";"}]], "Input"]
}, Open  ]],

Cell["\<\
IBP reduction: run runkira.sh to execute Kira on each topology\
\>", "Subsection"],

Cell[CellGroupData[{

Cell["Extraction of the IBP tables from Kira", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ApplyBMapFaster", "[", 
   RowBox[{"ex_", ",", "bmap_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "BB", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"bmap", "//", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"#", "/.", 
           RowBox[{"B", "\[Rule]", "BB"}]}], "//", 
          RowBox[{"Apply", "[", "Set", "]"}]}], ")"}], "&"}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"ex", "/.", 
       RowBox[{"B", "\[Rule]", "BB"}]}], "/.", 
      RowBox[{"BB", "\[Rule]", "B"}]}]}]}], "\[IndentingNewLine]", 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ApplyBMapFaster", "[", "bmap_List", "]"}], ":=", 
  RowBox[{
   RowBox[{"ApplyBMapFaster", "[", 
    RowBox[{"#", ",", "bmap"}], "]"}], "&"}]}]}], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"IntegralDiffByMass", "[", 
   RowBox[{
    RowBox[{"B", "[", 
     RowBox[{"topid_", ",", "indices_List"}], "]"}], ",", "x_", ",", 
    "topologies_List"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"dens", ",", "nums", ",", "i", ",", "k"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"dens", "=", 
      RowBox[{"topologies", "[", 
       RowBox[{"[", "topid", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"nums", "=", 
      RowBox[{
       RowBox[{
        RowBox[{"dens", "/.", 
         RowBox[{
          RowBox[{"den", "[", 
           RowBox[{"p_", ",", "m2_", ",", "___"}], "]"}], "\[RuleDelayed]", 
          RowBox[{
           RowBox[{"sp", "[", 
            RowBox[{"p", ",", "p"}], "]"}], "-", "m2"}]}]}], "//", 
        "ExpandSP"}], "//", 
       RowBox[{"ReplaceAll", "[", "externalsprules", "]"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"Sum", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         RowBox[{"indices", "[", 
          RowBox[{"[", "i", "]"}], "]"}]}], 
        RowBox[{"D", "[", 
         RowBox[{
          RowBox[{"nums", "[", 
           RowBox[{"[", "i", "]"}], "]"}], ",", "x"}], "]"}], 
        RowBox[{"B", "[", 
         RowBox[{"topid", ",", 
          RowBox[{"indices", "+", 
           RowBox[{"Table", "[", 
            RowBox[{
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{"k", "===", "i"}], ",", "1", ",", "0"}], "]"}], ",", 
             RowBox[{"{", 
              RowBox[{"k", ",", 
               RowBox[{"Length", "[", "dens", "]"}]}], "}"}]}], "]"}]}]}], 
         "]"}]}], "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"i", ",", 
         RowBox[{"Length", "[", "dens", "]"}]}], "}"}]}], "]"}]}]}], 
   "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "Make", " ", "all", " ", "B", " ", "to", " ", "temporarily", " ", "have", 
    " ", "the", " ", "same", " ", "dimensionality"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"expribp", "=", 
     RowBox[{"exprb", "/.", 
      RowBox[{
       RowBox[{"B", "[", 
        RowBox[{"topid_", ",", "idx_List"}], "]"}], "\[RuleDelayed]", 
       RowBox[{
        RowBox[{"B", "[", 
         RowBox[{"topid", ",", "idx"}], "]"}], "*", 
        RowBox[{"Q2", "^", 
         RowBox[{"(", 
          RowBox[{"-", 
           RowBox[{"Plus", "@@", "idx"}]}], ")"}]}]}]}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"topologymasters", "=", 
     RowBox[{"<|", "|>"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"diffequations", "=", 
     RowBox[{"<|", "|>"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Do", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"toponame", "=", 
        RowBox[{"\"\<ps4l0t\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], ";", "\[IndentingNewLine]", 
       
       RowBox[{"Print", "[", 
        RowBox[{"\"\<* Loading the IBP tables for \>\"", ",", "toponame"}], 
        "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"dirname", "=", 
        RowBox[{"\"\<kira.\>\"", "<>", "toponame", "<>", "\"\<.dir\>\""}]}], 
       ";", "\[IndentingNewLine]", 
       RowBox[{"topo", "=", 
        RowBox[{"toponame", "//", "ToExpression"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{"ibpmap", "=", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Get", "[", 
            RowBox[{
            "dirname", "<>", "\"\</results/\>\"", "<>", "toponame", "<>", 
             "\"\</kira_\>\"", "<>", "toponame", "<>", "\"\<.m\>\""}], "]"}], 
           "/.", 
           RowBox[{
            RowBox[{"topo", "[", "idx__", "]"}], "\[RuleDelayed]", 
            RowBox[{"B", "[", 
             RowBox[{"i", ",", 
              RowBox[{"{", "idx", "}"}]}], "]"}]}]}], "//", "TM"}], "//", 
         "Union"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{"Print", "[", 
        RowBox[{"\"\<Table size: \>\"", ",", 
         RowBox[{"ibpmap", "//", "Length"}]}], "]"}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{"masters", "=", 
        RowBox[{
         RowBox[{
          RowBox[{"ibpmap", "[", 
           RowBox[{"[", 
            RowBox[{";;", ",", "2"}], "]"}], "]"}], "//", 
          RowBox[{
           RowBox[{"Cases", "[", 
            RowBox[{"#", ",", "_B", ",", 
             RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{"Print", "[", 
        RowBox[{"\"\<Masters: \>\"", ",", "masters"}], "]"}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"topologymasters", "[", "toponame", "]"}], "=", "masters"}], 
       ";", "\[IndentingNewLine]", 
       RowBox[{"expribp", "=", 
        RowBox[{
         RowBox[{"expribp", "//", 
          RowBox[{"ApplyBMapFaster", "[", "ibpmap", "]"}]}], "//", "TM"}]}], 
       ";", "\[IndentingNewLine]", 
       RowBox[{"(*", " ", 
        RowBox[{
        "Differential", " ", "equation", " ", "in", " ", "x", " ", "for", " ",
          "the", " ", "masters"}], " ", "*)"}], "\[IndentingNewLine]", 
       RowBox[{"dmasters", "=", 
        RowBox[{
         RowBox[{"masters", "//", 
          RowBox[{"Map", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"IntegralDiffByMass", "[", 
              RowBox[{"#", ",", "x", ",", "topologies"}], "]"}], "/", "Q2"}], 
            "&"}], "]"}]}], "//", 
         RowBox[{"Map", "[", "Expand", "]"}]}]}], ";", "\[IndentingNewLine]", 
       
       RowBox[{"dmastersibp", "=", 
        RowBox[{"dmasters", "//", 
         RowBox[{"ApplyBMapFaster", "[", "ibpmap", "]"}]}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{"mx", "=", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Coefficient", "[", 
            RowBox[{"dm", ",", "m"}], "]"}], "//", "Factor"}], ",", 
          RowBox[{"{", 
           RowBox[{"dm", ",", "dmastersibp"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"m", ",", "masters"}], "}"}]}], "]"}]}], ";", 
       "\[IndentingNewLine]", 
       RowBox[{"Print", "[", "mx", "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{"dmastersibp", "-", 
               RowBox[{"mx", ".", "masters"}]}], "//", "Expand"}], "//", 
             "Together"}], "//", "Flatten"}], "//", "Union"}], "//", 
          RowBox[{
           RowBox[{"#", "=!=", 
            RowBox[{"{", "0", "}"}]}], "&"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
          "Error", "[", 
           "\"\<Failed to decompose d/dx*masters into the masters\>\"", "]"}],
           ";"}]}], "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
       
       RowBox[{
        RowBox[{"diffequations", "[", "toponame", "]"}], "=", 
        RowBox[{"mx", "/.", 
         RowBox[{"d", "\[Rule]", 
          RowBox[{"4", "-", 
           RowBox[{"2", "ep"}]}]}]}]}], ";", "\[IndentingNewLine]", 
       RowBox[{"Put", "[", 
        RowBox[{
         RowBox[{"diffequations", "[", "toponame", "]"}], ",", 
         RowBox[{"toponame", "<>", "\"\<.diffeq.m\>\""}]}], "]"}], ";"}], 
      "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"i", ",", 
        RowBox[{"Length", "[", "topologies", "]"}]}], "}"}]}], 
     "\[IndentingNewLine]", "]"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"Back", " ", "to", " ", "B", " ", "being", " ", "dimensionful"}], 
    " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"expribp", "=", 
     RowBox[{
      RowBox[{"expribp", "/.", 
       RowBox[{
        RowBox[{"B", "[", 
         RowBox[{"topid_", ",", "idx_List"}], "]"}], "\[RuleDelayed]", 
        RowBox[{
         RowBox[{"B", "[", 
          RowBox[{"topid", ",", "idx"}], "]"}], "*", 
         RowBox[{"Q2", "^", 
          RowBox[{"(", 
           RowBox[{"Plus", "@@", "idx"}], ")"}]}]}]}]}], "//", "Expand"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{"expribp", "//", "Length"}]}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t1\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t1"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "6.127387`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 6.127387, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "788196"}],
   SequenceForm["Table size: ", 788196],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", 
        ",", "1", ",", "1", ",", "1"}], "1"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "1"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "1"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "1"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[1, {-1, 0, 1, 1, 1, 1, 1, 1, 1}], 
     $CellContext`B[1, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[1, {0, 0, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[1, {0, 0, 1, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "6.431145`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 6.431145, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", "x"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{"x", "-", "1"}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", "x"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]]},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "10"}], 
       RowBox[{"2", " ", "x"}]], "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{"3", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"d", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"4", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"2", " ", "d"}], "+", "7"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t2\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t2"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.913659`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 0.913659, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "120864"}],
   SequenceForm["Table size: ", 120864],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "2"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "2"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "2"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "2"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[2, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[2, {0, 0, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[2, {0, 0, 1, 1, 1, 1, 1, 1, 1}], 
     $CellContext`B[2, {1, 0, 1, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "1.00039`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 1.00039, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0"},
     {
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "10"}], 
       RowBox[{"2", " ", "x"}]], "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "10"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}], 
      FractionBox[
       RowBox[{"d", "-", "4"}], 
       RowBox[{"x", "-", "1"}]], "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "10"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "8"}], ")"}]}], 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"d", "-", "4"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"x", "-", "1"}], ")"}], "2"], " ", 
         SuperscriptBox["x", "2"]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"3", " ", "d"}], "-", "10"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        SuperscriptBox["x", "2"]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"d", "-", "4"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"2", " ", "d"}], "-", "9"}], "x"]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t3\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t3"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "3.150466`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 3.150466, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "361009"}],
   SequenceForm["Table size: ", 361009],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", 
        ",", "1", ",", "1", ",", "1"}], "3"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "3"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "3"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "3"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "3"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "3"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[3, {-1, 0, 1, 1, 1, 1, 1, 1, 1}], 
     $CellContext`B[3, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[3, {0, 0, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[3, {0, 0, 1, 1, 1, 1, 1, 1, 1}], 
     $CellContext`B[3, {0, 1, 0, 1, 1, 1, 1, 1, 1}], 
     $CellContext`B[3, {1, 1, 0, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "2.974712`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 2.974712, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", "x"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{"x", "-", "1"}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", "x"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "10"}], 
       RowBox[{"2", " ", "x"}]], "0", "0", "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{"3", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"d", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"4", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"2", " ", "d"}], "+", "7"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], "0", "0"},
     {"0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "10"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}], "0", 
      FractionBox[
       RowBox[{"d", "-", "4"}], 
       RowBox[{"x", "-", "1"}]], "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "10"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "8"}], ")"}]}], 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"d", "-", "4"}], ")"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"x", "-", "1"}], ")"}], "2"], " ", 
         SuperscriptBox["x", "2"]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"3", " ", "d"}], "-", "10"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        SuperscriptBox["x", "2"]}]], "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"d", "-", "4"}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"2", " ", "d"}], "-", "9"}], "x"]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t4\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t4"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.37759`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 0.37759, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "46534"}],
   SequenceForm["Table size: ", 46534],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "4"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "4"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "4"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[4, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[4, {0, 0, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[4, {0, 0, 1, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.397016`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 0.397016, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0"},
     {
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "10"}], 
       RowBox[{"2", " ", "x"}]], "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "10"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}], 
      FractionBox[
       RowBox[{"d", "-", "4"}], 
       RowBox[{"x", "-", "1"}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t5\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t5"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.35085`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 0.35085, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "46535"}],
   SequenceForm["Table size: ", 46535],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "5"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "5"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[5, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[5, {1, 1, 1, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.426971`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 0.426971, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t6\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t6"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.939089`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 0.939089, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "120863"}],
   SequenceForm["Table size: ", 120863],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", 
        ",", "1", ",", "1", ",", "1"}], "6"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "6"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "6"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "6"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "6"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[6, {-1, 1, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[6, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[6, {0, 1, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[6, {0, 1, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[6, {1, 1, 1, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.925905`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 0.925905, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{"x", "-", "1"}]], "0", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{"x", "-", "1"}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}]}], 
        RowBox[{"x", "-", "1"}]]}], "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "10"}], 
       RowBox[{"2", " ", "x"}]], "0", "0"},
     {
      FractionBox[
       RowBox[{"6", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"d", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"4", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"3", " ", "d", " ", "x"}], "+", 
        RowBox[{"9", " ", "x"}], "-", 
        RowBox[{"2", " ", "d"}], "+", "7"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], "0"},
     {
      FractionBox[
       RowBox[{"12", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["x", "2"], "+", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        SuperscriptBox["x", "3"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}], "3"]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"3", " ", "d"}], "-", "8"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"7", " ", 
           SuperscriptBox["x", "2"], " ", 
           SuperscriptBox["d", "2"]}], "+", 
          RowBox[{"12", " ", "x", " ", 
           SuperscriptBox["d", "2"]}], "+", 
          RowBox[{"7", " ", 
           SuperscriptBox["d", "2"]}], "-", 
          RowBox[{"46", " ", 
           SuperscriptBox["x", "2"], " ", "d"}], "-", 
          RowBox[{"76", " ", "x", " ", "d"}], "-", 
          RowBox[{"46", " ", "d"}], "+", 
          RowBox[{"76", " ", 
           SuperscriptBox["x", "2"]}], "+", 
          RowBox[{"120", " ", "x"}], "+", "76"}], ")"}]}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", 
        SuperscriptBox["x", "3"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}], "3"]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"d", " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"2", " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"6", " ", "d", " ", "x"}], "+", 
           RowBox[{"20", " ", "x"}], "+", "d", "-", "2"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", 
         SuperscriptBox["x", "3"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"x", "+", "1"}], ")"}], "2"]}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", "d", " ", 
            SuperscriptBox["x", "4"]}], "-", 
           RowBox[{"7", " ", 
            SuperscriptBox["x", "4"]}], "+", 
           RowBox[{"3", " ", "d", " ", 
            SuperscriptBox["x", "3"]}], "-", 
           RowBox[{"9", " ", 
            SuperscriptBox["x", "3"]}], "-", 
           RowBox[{"2", " ", "d", " ", 
            SuperscriptBox["x", "2"]}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox["x", "2"]}], "+", 
           RowBox[{"3", " ", "d", " ", "x"}], "-", 
           RowBox[{"9", " ", "x"}], "+", 
           RowBox[{"2", " ", "d"}], "-", "7"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", 
         SuperscriptBox["x", "3"], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"x", "+", "1"}], ")"}], "3"]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "+", "1"}], ")"}]}], 
       RowBox[{"x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"* Loading the IBP tables for \"\>", 
    "\[InvisibleSpace]", "\<\"ps4l0t7\"\>"}],
   SequenceForm["* Loading the IBP tables for ", "ps4l0t7"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{
       "dirname", "<>", "\<\"/results/\"\>", "<>", "toponame", 
        "<>", "\<\"/kira_\"\>", "<>", "toponame", "<>", "\<\".m\"\>"}], "]"}],
       "/.", "\[VeryThinSpace]", 
      RowBox[{
       RowBox[{"topo", "(", "idx__", ")"}], "\[RuleDelayed]", 
       RowBox[{"B", "(", 
        RowBox[{"i", ",", 
         RowBox[{"{", "idx", "}"}]}], ")"}]}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.952692`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     ReplaceAll[
      Get[
       StringJoin[$CellContext`dirname, "/results/", $CellContext`toponame, 
        "/kira_", $CellContext`toponame, ".m"]], $CellContext`topo[
        Pattern[$CellContext`idx, 
         BlankSequence[]]] :> $CellContext`B[$CellContext`i, \
{$CellContext`idx}]]], ": ", 0.952692, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Table size: \"\>", "\[InvisibleSpace]", "120862"}],
   SequenceForm["Table size: ", 120862],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{"\<\"Masters: \"\>", "\[InvisibleSpace]", 
    RowBox[{"{", 
     RowBox[{
      SubsuperscriptBox["B", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", 
        ",", "1", ",", "1", ",", "1"}], "7"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "7"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "7"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "0", ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "7"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "1", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "7"], ",", 
      SubsuperscriptBox["B", 
       RowBox[{
       "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
        "1", ",", "1"}], "7"]}], "}"}]}],
   SequenceForm["Masters: ", {
     $CellContext`B[7, {-1, 1, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[7, {0, 0, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[7, {0, 1, 0, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[7, {0, 1, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[7, {1, 0, 1, 0, 1, 1, 1, 1, 1}], 
     $CellContext`B[7, {1, 1, 1, 1, 1, 1, 1, 1, 1}]}],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    TagBox[
     RowBox[{
      RowBox[{"ApplyBMapFaster", "(", "ibpmap", ")"}], 
      "\[InvisibleApplication]", 
      RowBox[{"(", "expribp", ")"}]}],
     HoldForm], "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
    "0.954849`", "\[InvisibleSpace]", "\<\" sec\"\>"}],
   SequenceForm[
    HoldForm[
     $CellContext`ApplyBMapFaster[$CellContext`ibpmap][$CellContext`expribp]],
     ": ", 0.954849, " sec"],
   Editable->False], TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      FractionBox[
       RowBox[{"3", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{"x", "+", "1"}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"3", " ", "d"}], "-", "8"}], 
        RowBox[{"2", " ", "x", " ", 
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}]}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"3", " ", "x", " ", "d"}], "-", 
         RowBox[{"3", " ", "d"}], "-", 
         RowBox[{"8", " ", "x"}], "+", "10"}], 
        RowBox[{"2", " ", "x", " ", 
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}]}], 
       RowBox[{"x", "+", "1"}]], "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "10"}], 
       RowBox[{"2", " ", "x"}]], "0", "0", "0"},
     {
      FractionBox[
       RowBox[{"6", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{
        RowBox[{"3", " ", "d"}], "-", "8"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "2"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{
        RowBox[{"d", " ", 
         SuperscriptBox["x", "2"]}], "-", 
        RowBox[{"4", " ", 
         SuperscriptBox["x", "2"]}], "+", 
        RowBox[{"3", " ", "d", " ", "x"}], "-", 
        RowBox[{"9", " ", "x"}], "-", 
        RowBox[{"2", " ", "d"}], "+", "7"}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], "0", "0"},
     {"0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "3"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"3", " ", "d"}], "-", "10"}], ")"}]}], 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}], "0", 
      FractionBox[
       RowBox[{"d", "-", "4"}], 
       RowBox[{"x", "-", "1"}]], "0"},
     {
      FractionBox[
       RowBox[{"12", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          SuperscriptBox["x", "2"], "+", "1"}], ")"}]}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
        SuperscriptBox["x", "3"], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"3", " ", "d"}], "-", "8"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"4", " ", 
           SuperscriptBox["x", "2"], " ", 
           SuperscriptBox["d", "2"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["d", "2"]}], "-", 
          RowBox[{"27", " ", 
           SuperscriptBox["x", "2"], " ", "d"}], "+", 
          RowBox[{"11", " ", "d"}], "+", 
          RowBox[{"46", " ", 
           SuperscriptBox["x", "2"]}], "-", "14"}], ")"}]}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"d", "-", "4"}], ")"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
        SuperscriptBox["x", "3"], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "3"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"x", "-", "2"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"5", " ", "d", " ", 
           SuperscriptBox["x", "2"]}], "-", 
          RowBox[{"18", " ", 
           SuperscriptBox["x", "2"]}], "-", "d", "+", "2"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
        SuperscriptBox["x", "3"], " ", 
        RowBox[{"(", 
         RowBox[{"x", "+", "1"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", "d", " ", 
            SuperscriptBox["x", "4"]}], "-", 
           RowBox[{"7", " ", 
            SuperscriptBox["x", "4"]}], "-", 
           RowBox[{"3", " ", "d", " ", 
            SuperscriptBox["x", "3"]}], "+", 
           RowBox[{"9", " ", 
            SuperscriptBox["x", "3"]}], "-", 
           RowBox[{"2", " ", "d", " ", 
            SuperscriptBox["x", "2"]}], "+", 
           RowBox[{"8", " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"3", " ", "d", " ", "x"}], "+", 
           RowBox[{"9", " ", "x"}], "+", 
           RowBox[{"2", " ", "d"}], "-", "7"}], ")"}]}], 
        RowBox[{
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
         SuperscriptBox["x", "3"], " ", 
         RowBox[{"(", 
          RowBox[{"x", "+", "1"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"d", "-", "4"}], ")"}]}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "-", "1"}], ")"}], "2"], " ", "x"}]], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{"d", "-", "5"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Print"]
}, Open  ]],

Cell[BoxData[
 FormBox["8919", TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"expribp", "//", "Terms"}], "//", "RandomChoice"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox[
    RowBox[{"48", " ", "CF", " ", 
     SuperscriptBox["gs", "4"], " ", 
     RowBox[{"PS", "(", "4", ")"}], " ", "Q2", " ", 
     SuperscriptBox["qu", "2"], " ", "x", " ", 
     SubsuperscriptBox["B", 
      RowBox[{
      "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
       "1", ",", "1"}], "3"]}], 
    RowBox[{"d", "-", "3"}]]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["topologymasters"], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAssociation]", 
   RowBox[{
    RowBox[{"\<\"ps4l0t1\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
         RowBox[{"-", "1"}], ",", "0", ",", "1", ",", "1", ",", "1", ",", "1",
          ",", "1", ",", "1", ",", "1"}], "1"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "1"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "1"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "1"]}], "}"}]}], ",", 
    RowBox[{"\<\"ps4l0t2\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "2"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "2"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "2"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "2"]}], "}"}]}], ",", 
    RowBox[{"\<\"ps4l0t3\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
         RowBox[{"-", "1"}], ",", "0", ",", "1", ",", "1", ",", "1", ",", "1",
          ",", "1", ",", "1", ",", "1"}], "3"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "3"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "3"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "3"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "3"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "3"]}], "}"}]}], ",", 
    RowBox[{"\<\"ps4l0t4\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "4"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "4"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "4"]}], "}"}]}], ",", 
    RowBox[{"\<\"ps4l0t5\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "5"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "5"]}], "}"}]}], ",", 
    RowBox[{"\<\"ps4l0t6\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
         RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "0", ",", "1", ",", "1",
          ",", "1", ",", "1", ",", "1"}], "6"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "6"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "6"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "6"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "6"]}], "}"}]}], ",", 
    RowBox[{"\<\"ps4l0t7\"\>", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       SubsuperscriptBox["B", 
        RowBox[{
         RowBox[{"-", "1"}], ",", "1", ",", "1", ",", "0", ",", "1", ",", "1",
          ",", "1", ",", "1", ",", "1"}], "7"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "7"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "7"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "0", ",", "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "7"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "1", ",", "0", ",", "1", ",", "0", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "7"], ",", 
       SubsuperscriptBox["B", 
        RowBox[{
        "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
         "1", ",", "1"}], "7"]}], "}"}]}]}], "\[RightAssociation]"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["diffequations"], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAssociation]", 
   RowBox[{
    RowBox[{"\<\"ps4l0t1\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{"3", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{"x", "-", "1"}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]]},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], 
          RowBox[{"2", " ", "x"}]], "0"},
        {
         RowBox[{"-", 
          FractionBox[
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", "7"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}], ",", 
    RowBox[{"\<\"ps4l0t2\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0"},
        {
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], 
          RowBox[{"2", " ", "x"}]], "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
           
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"2", " ", "ep", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", "ep"}], 
           RowBox[{"x", "-", "1"}]]}], "0"},
        {
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], ")"}], " ", 
            
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{"2", " ", 
            SuperscriptBox["ep", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"x", "-", "1"}], ")"}], "2"], " ", 
            SuperscriptBox["x", "2"]}]]}], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{"2", " ", "ep", " ", 
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], " ", 
            SuperscriptBox["x", "2"]}]]}], 
         FractionBox[
          RowBox[{"2", " ", "ep"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "9"}], "x"]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}], ",", 
    RowBox[{"\<\"ps4l0t3\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{"3", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{"x", "-", "1"}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], 
          RowBox[{"2", " ", "x"}]], "0", "0", "0"},
        {
         RowBox[{"-", 
          FractionBox[
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", "7"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], "0", "0"},
        {"0", "0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
           
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"2", " ", "ep", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", "ep"}], 
           RowBox[{"x", "-", "1"}]]}], "0"},
        {"0", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], ")"}], " ", 
            
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{"2", " ", 
            SuperscriptBox["ep", "2"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"x", "-", "1"}], ")"}], "2"], " ", 
            SuperscriptBox["x", "2"]}]]}], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], 
           RowBox[{"2", " ", "ep", " ", 
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], " ", 
            SuperscriptBox["x", "2"]}]]}], "0", 
         FractionBox[
          RowBox[{"2", " ", "ep"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "9"}], "x"]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}], ",", 
    RowBox[{"\<\"ps4l0t4\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0"},
        {
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], 
          RowBox[{"2", " ", "x"}]], "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
           
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"2", " ", "ep", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", "ep"}], 
           RowBox[{"x", "-", "1"}]]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}], ",", 
    RowBox[{"\<\"ps4l0t5\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", "ep"}], "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}], ",", 
    RowBox[{"\<\"ps4l0t6\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{"3", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"x", "-", "1"}]], "0", 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"x", "-", "1"}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}], 
           RowBox[{"x", "-", "1"}]]}], "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], 
          RowBox[{"2", " ", "x"}]], "0", "0"},
        {
         FractionBox[
          RowBox[{"6", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], "+", 
           RowBox[{"9", " ", "x"}], "-", 
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", "7"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], "0"},
        {
         FractionBox[
          RowBox[{"12", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox["x", "2"], "+", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           SuperscriptBox["x", "3"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}], "3"]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"7", " ", 
              SuperscriptBox["x", "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], "2"]}], "+", 
             RowBox[{"12", " ", "x", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], "2"]}], "+", 
             RowBox[{"7", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], "2"]}], "-", 
             RowBox[{"46", " ", 
              SuperscriptBox["x", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", 
             RowBox[{"76", " ", "x", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", 
             RowBox[{"46", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", 
             RowBox[{"76", " ", 
              SuperscriptBox["x", "2"]}], "+", 
             RowBox[{"120", " ", "x"}], "+", "76"}], ")"}]}], 
          RowBox[{"2", " ", 
           SuperscriptBox["ep", "2"], " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           SuperscriptBox["x", "3"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}], "3"]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
              SuperscriptBox["x", "2"]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox["x", "2"]}], "-", 
             RowBox[{"6", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], "+", 
             RowBox[{"20", " ", "x"}], "-", 
             RowBox[{"2", " ", "ep"}], "+", "2"}], ")"}]}], 
          RowBox[{"ep", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", 
           SuperscriptBox["x", "3"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}], "2"]}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "4"]}], "-", 
              RowBox[{"7", " ", 
               SuperscriptBox["x", "4"]}], "+", 
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "3"]}], "-", 
              RowBox[{"9", " ", 
               SuperscriptBox["x", "3"]}], "-", 
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "2"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["x", "2"]}], "+", 
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], "-", 
              RowBox[{"9", " ", "x"}], "+", 
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "7"}], ")"}]}], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], " ", 
            SuperscriptBox["x", "3"], " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"x", "+", "1"}], ")"}], "3"]}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", "ep"}], "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "+", "1"}], ")"}]}], 
          RowBox[{"x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}], ",", 
    RowBox[{"\<\"ps4l0t7\"\>", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         FractionBox[
          RowBox[{"3", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"x", "+", "1"}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"3", " ", 
             RowBox[{"(", 
              RowBox[{"4", "-", 
               RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
           RowBox[{"2", " ", "x", " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}]]}], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"3", " ", "x", " ", 
             RowBox[{"(", 
              RowBox[{"4", "-", 
               RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", 
            RowBox[{"3", " ", 
             RowBox[{"(", 
              RowBox[{"4", "-", 
               RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", 
            RowBox[{"8", " ", "x"}], "+", "10"}], 
           RowBox[{"2", " ", "x", " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}]}], 
          RowBox[{"x", "+", "1"}]], "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", "0", "0", "0"},
        {"0", 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], 
          RowBox[{"2", " ", "x"}]], "0", "0", "0"},
        {
         FractionBox[
          RowBox[{"6", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{"2", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "2"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
            SuperscriptBox["x", "2"]}], "-", 
           RowBox[{"4", " ", 
            SuperscriptBox["x", "2"]}], "+", 
           RowBox[{"3", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], "-", 
           RowBox[{"9", " ", "x"}], "-", 
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{"4", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", "7"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x", " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], "0", "0"},
        {"0", "0", 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "10"}], ")"}], " ", 
           
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}]}], 
          RowBox[{"2", " ", "ep", " ", 
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]], "0", 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", "ep"}], 
           RowBox[{"x", "-", "1"}]]}], "0"},
        {
         FractionBox[
          RowBox[{"12", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox["x", "2"], "+", "1"}], ")"}]}], 
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
           SuperscriptBox["x", "3"], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "8"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", " ", 
              SuperscriptBox["x", "2"], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], "2"]}], "-", 
             RowBox[{"2", " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], "2"]}], "-", 
             RowBox[{"27", " ", 
              SuperscriptBox["x", "2"], " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", 
             RowBox[{"11", " ", 
              RowBox[{"(", 
               RowBox[{"4", "-", 
                RowBox[{"2", " ", "ep"}]}], ")"}]}], "+", 
             RowBox[{"46", " ", 
              SuperscriptBox["x", "2"]}], "-", "14"}], ")"}]}], 
          RowBox[{"2", " ", 
           SuperscriptBox["ep", "2"], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
           SuperscriptBox["x", "3"], " ", 
           RowBox[{"(", 
            RowBox[{"x", "+", "1"}], ")"}]}]], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{"x", "-", "2"}], ")"}], " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"5", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "2"]}], "-", 
              RowBox[{"18", " ", 
               SuperscriptBox["x", "2"]}], "+", 
              RowBox[{"2", " ", "ep"}], "-", "2"}], ")"}]}], 
           RowBox[{"ep", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
            SuperscriptBox["x", "3"], " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}]]}], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"2", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "4"]}], "-", 
              RowBox[{"7", " ", 
               SuperscriptBox["x", "4"]}], "-", 
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "3"]}], "+", 
              RowBox[{"9", " ", 
               SuperscriptBox["x", "3"]}], "-", 
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
               SuperscriptBox["x", "2"]}], "+", 
              RowBox[{"8", " ", 
               SuperscriptBox["x", "2"]}], "-", 
              RowBox[{"3", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}], " ", "x"}], "+", 
              RowBox[{"9", " ", "x"}], "+", 
              RowBox[{"2", " ", 
               RowBox[{"(", 
                RowBox[{"4", "-", 
                 RowBox[{"2", " ", "ep"}]}], ")"}]}], "-", "7"}], ")"}]}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"x", "-", "1"}], ")"}], "3"], " ", 
            SuperscriptBox["x", "3"], " ", 
            RowBox[{"(", 
             RowBox[{"x", "+", "1"}], ")"}]}]]}], 
         RowBox[{"-", 
          FractionBox[
           RowBox[{"4", " ", "ep"}], 
           RowBox[{
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"x", "-", "1"}], ")"}], "2"], " ", "x"}]]}], 
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", "ep"}], "-", "1"}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", "x"}], "-", "1"}], ")"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "-", "1"}], ")"}], " ", "x"}]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.7]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}]}], 
   "\[RightAssociation]"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{
    RowBox[{"<|", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"\"\<topologies\>\"", "\[Rule]", "topologies"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"\"\<expression\>\"", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{"expribp", "//", 
         RowBox[{
          RowBox[{"Bracket", "[", 
           RowBox[{"#", ",", "_B", ",", "Factor"}], "]"}], "&"}]}], ")"}]}]}],
      "\[IndentingNewLine]", "|>"}], ",", "\"\<reduced.ps4.l0.m\>\""}], "]"}],
   ";"}]], "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
10. Master evaluation for PS[k1, k2] L[l1, l2], featuring HypExp\
\>", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"From", " ", 
     RowBox[{"https", ":"}]}], "//", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"arxiv", ".", "org"}], "/", "abs"}], "/", "hep"}], "-", 
     RowBox[{"ph", "/", "0507061"}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"mastervalues", "=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"(*", 
         RowBox[{
          RowBox[{"ep", "=", 
           RowBox[{"2", "-", 
            RowBox[{"d", "/", "2"}]}]}], ","}], "*)"}], 
        RowBox[{"Sgamma", ",", "A2", ",", "A3", ",", "A4", ",", "A6"}], "}"}],
        ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Sgamma", "=", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"4", "Pi"}], ")"}], "^", "ep"}], "/", 
          RowBox[{"(", 
           RowBox[{"16", 
            RowBox[{"Pi", "^", "2"}], 
            RowBox[{"Gamma", "[", 
             RowBox[{"1", "-", "ep"}], "]"}]}], ")"}]}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"A2", "=", 
         RowBox[{"Sgamma", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"-", "Q2"}], ")"}], "^", 
           RowBox[{"(", 
            RowBox[{"-", "ep"}], ")"}]}], " ", 
          RowBox[{"Gamma", "[", 
           RowBox[{"1", "+", "ep"}], "]"}], 
          RowBox[{
           RowBox[{
            RowBox[{"Gamma", "[", 
             RowBox[{"1", "-", "ep"}], "]"}], "^", "3"}], "/", 
           RowBox[{"Gamma", "[", 
            RowBox[{"2", "-", 
             RowBox[{"2", "ep"}]}], "]"}]}], 
          RowBox[{"I", "/", "ep"}]}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"A3", "=", 
         RowBox[{
          RowBox[{"Sgamma", "^", "2"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"-", "Q2"}], ")"}], "^", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"2", "ep"}]}], ")"}]}], 
          RowBox[{"Gamma", "[", 
           RowBox[{"1", "+", 
            RowBox[{"2", "ep"}]}], "]"}], 
          RowBox[{
           RowBox[{
            RowBox[{"Gamma", "[", 
             RowBox[{"1", "-", "ep"}], "]"}], "^", "5"}], "/", 
           RowBox[{"Gamma", "[", 
            RowBox[{"3", "-", 
             RowBox[{"3", "ep"}]}], "]"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"-", "1"}], ")"}], "/", 
           RowBox[{"(", 
            RowBox[{"2", 
             RowBox[{"(", 
              RowBox[{"1", "-", 
               RowBox[{"2", "ep"}]}], ")"}], "ep"}], ")"}]}]}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"A4", "=", 
         RowBox[{
          RowBox[{"Sgamma", "^", "2"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"-", "Q2"}], ")"}], "^", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "ep"}], ")"}]}], 
          RowBox[{"Gamma", "[", 
           RowBox[{"1", "-", 
            RowBox[{"2", "ep"}]}], "]"}], 
          RowBox[{"Gamma", "[", 
           RowBox[{"1", "+", "ep"}], "]"}], 
          RowBox[{
           RowBox[{"Gamma", "[", 
            RowBox[{"1", "-", "ep"}], "]"}], "^", "4"}], 
          RowBox[{
           RowBox[{"Gamma", "[", 
            RowBox[{"1", "+", 
             RowBox[{"2", "ep"}]}], "]"}], "/", 
           RowBox[{"Gamma", "[", 
            RowBox[{"2", "-", 
             RowBox[{"3", "ep"}]}], "]"}]}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"-", "1"}], ")"}], "/", 
           RowBox[{"(", 
            RowBox[{"2", 
             RowBox[{"(", 
              RowBox[{"1", "-", 
               RowBox[{"2", "ep"}]}], ")"}], 
             RowBox[{"ep", "^", "2"}]}], ")"}]}]}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"A6", "=", 
         RowBox[{
          RowBox[{"Sgamma", "^", "2"}], 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"-", "Q2"}], ")"}], "^", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "2"}], "-", 
             RowBox[{"2", "ep"}]}], ")"}]}], 
          RowBox[{"(", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
             RowBox[{"-", 
              RowBox[{
               RowBox[{"Gamma", "[", 
                RowBox[{"1", "-", "ep"}], "]"}], "^", "3"}]}], 
             RowBox[{"Gamma", "[", 
              RowBox[{"1", "+", "ep"}], "]"}], 
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "-", 
                RowBox[{"2", "ep"}]}], "]"}], "^", "4"}], 
             RowBox[{
              RowBox[{
               RowBox[{"Gamma", "[", 
                RowBox[{"1", "+", 
                 RowBox[{"2", "ep"}]}], "]"}], "^", "3"}], "/", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"ep", "^", "4"}], 
                RowBox[{
                 RowBox[{"Gamma", "[", 
                  RowBox[{"1", "-", 
                   RowBox[{"4", "ep"}]}], "]"}], "^", "2"}], 
                RowBox[{"Gamma", "[", 
                 RowBox[{"1", "+", 
                  RowBox[{"4", "ep"}]}], "]"}]}], ")"}]}]}], 
            "\[IndentingNewLine]", "+", 
            RowBox[{
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "-", "ep"}], "]"}], "^", "4"}], 
             RowBox[{"Gamma", "[", 
              RowBox[{"1", "+", "ep"}], "]"}], 
             RowBox[{"Gamma", "[", 
              RowBox[{"1", "-", 
               RowBox[{"2", "ep"}]}], "]"}], 
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "+", 
                RowBox[{"2", "ep"}]}], "]"}], "/", 
              RowBox[{"(", 
               RowBox[{"2", 
                RowBox[{"ep", "^", "4"}], 
                RowBox[{"Gamma", "[", 
                 RowBox[{"1", "-", 
                  RowBox[{"3", "ep"}]}], "]"}]}], ")"}]}], 
             RowBox[{"HypergeometricPFQ", "[", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{"1", ",", 
                 RowBox[{
                  RowBox[{"-", "4"}], "ep"}], ",", 
                 RowBox[{
                  RowBox[{"-", "2"}], "ep"}]}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{
                 RowBox[{"1", "-", 
                  RowBox[{"3", "ep"}]}], ",", 
                 RowBox[{"1", "-", 
                  RowBox[{"2", "ep"}]}]}], "}"}], ",", "1"}], "]"}]}], 
            "\[IndentingNewLine]", "-", 
            RowBox[{"4", 
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "-", "ep"}], "]"}], "^", "4"}], 
             RowBox[{"Gamma", "[", 
              RowBox[{"1", "-", 
               RowBox[{"2", "ep"}]}], "]"}], 
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "+", 
                RowBox[{"2", "ep"}]}], "]"}], "/", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"ep", "^", "2"}], 
                RowBox[{"(", 
                 RowBox[{"1", "+", "ep"}], ")"}], 
                RowBox[{"(", 
                 RowBox[{"1", "+", 
                  RowBox[{"2", "ep"}]}], ")"}], 
                RowBox[{"Gamma", "[", 
                 RowBox[{"1", "-", 
                  RowBox[{"4", "ep"}]}], "]"}]}], ")"}]}], 
             RowBox[{"HypergeometricPFQ", "[", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{"1", ",", "1", ",", 
                 RowBox[{"1", "+", 
                  RowBox[{"2", "ep"}]}]}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{
                 RowBox[{"2", "+", "ep"}], ",", 
                 RowBox[{"2", "+", 
                  RowBox[{"2", "ep"}]}]}], "}"}], ",", "1"}], "]"}]}], 
            "\[IndentingNewLine]", "-", 
            RowBox[{
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "-", "ep"}], "]"}], "^", "5"}], 
             RowBox[{
              RowBox[{"Gamma", "[", 
               RowBox[{"1", "+", 
                RowBox[{"2", "ep"}]}], "]"}], "/", 
              RowBox[{"(", 
               RowBox[{"2", 
                RowBox[{"ep", "^", "4"}], 
                RowBox[{"Gamma", "[", 
                 RowBox[{"1", "-", 
                  RowBox[{"3", "ep"}]}], "]"}]}], ")"}]}], 
             RowBox[{"HypergeometricPFQ", "[", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{"1", ",", 
                 RowBox[{"1", "-", "ep"}], ",", 
                 RowBox[{
                  RowBox[{"-", "4"}], "ep"}], ",", 
                 RowBox[{
                  RowBox[{"-", "2"}], "ep"}]}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{
                 RowBox[{"1", "-", 
                  RowBox[{"3", "ep"}]}], ",", 
                 RowBox[{"1", "-", 
                  RowBox[{"2", "ep"}]}], ",", 
                 RowBox[{"1", "-", 
                  RowBox[{"2", "ep"}]}]}], "}"}], ",", "1"}], "]"}]}]}], 
           "\[IndentingNewLine]", ")"}]}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"{", "\[IndentingNewLine]", 
         RowBox[{
          StyleBox[
           RowBox[{
            RowBox[{"B", "[", 
             RowBox[{"1", ",", " ", 
              RowBox[{"{", 
               RowBox[{
               "0", ",", " ", "1", ",", " ", "1", ",", " ", "1", ",", " ", 
                "1", ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}], 
            "\[Rule]", "A4"}],
           ShowStringCharacters->True,
           NumberMarks->True], 
          StyleBox[",",
           ShowStringCharacters->True,
           NumberMarks->True], 
          StyleBox["\[IndentingNewLine]",
           ShowStringCharacters->True,
           NumberMarks->True], 
          RowBox[{
           StyleBox[
            RowBox[{"B", "[", 
             RowBox[{"2", ",", " ", 
              RowBox[{"{", 
               RowBox[{
               "1", ",", " ", "0", ",", " ", "0", ",", " ", "1", ",", " ", 
                "1", ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
            ShowStringCharacters->True,
            NumberMarks->True], 
           StyleBox["\[Rule]",
            ShowStringCharacters->True,
            NumberMarks->True], "A3"}], 
          StyleBox[",",
           ShowStringCharacters->True,
           NumberMarks->True], 
          StyleBox["\[IndentingNewLine]",
           ShowStringCharacters->True,
           NumberMarks->True], 
          RowBox[{
           StyleBox[
            RowBox[{"B", "[", 
             RowBox[{"5", ",", " ", 
              RowBox[{"{", 
               RowBox[{
               "0", ",", " ", "1", ",", " ", "1", ",", " ", "1", ",", " ", 
                "1", ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
            ShowStringCharacters->True,
            NumberMarks->True], 
           StyleBox["\[Rule]",
            ShowStringCharacters->True,
            NumberMarks->True], "A4"}], 
          StyleBox[",",
           ShowStringCharacters->True,
           NumberMarks->True], 
          StyleBox["\[IndentingNewLine]",
           ShowStringCharacters->True,
           NumberMarks->True], 
          RowBox[{
           StyleBox[
            RowBox[{"B", "[", 
             RowBox[{"6", ",", " ", 
              RowBox[{"{", 
               RowBox[{
               "1", ",", " ", "1", ",", " ", "1", ",", " ", "1", ",", " ", 
                "1", ",", " ", "1", ",", " ", "0"}], "}"}]}], "]"}],
            ShowStringCharacters->True,
            NumberMarks->True], 
           StyleBox["\[Rule]",
            ShowStringCharacters->True,
            NumberMarks->True], "A6"}], 
          StyleBox[",",
           ShowStringCharacters->True,
           NumberMarks->True], 
          StyleBox["\[IndentingNewLine]",
           ShowStringCharacters->True,
           NumberMarks->True], 
          RowBox[{
           StyleBox[
            RowBox[{"B", "[", 
             RowBox[{"7", ",", " ", 
              RowBox[{"{", 
               RowBox[{
               "1", ",", " ", "1", ",", " ", "0", ",", " ", "1", ",", " ", 
                "1", ",", " ", "0", ",", " ", "0"}], "}"}]}], "]"}],
            ShowStringCharacters->True,
            NumberMarks->True], 
           StyleBox["\[Rule]",
            ShowStringCharacters->True,
            NumberMarks->True], 
           RowBox[{"A2", "^", "2"}]}]}], "\[IndentingNewLine]", "}"}]}]}], 
      "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"mastervalues", "[", 
    RowBox[{"[", 
     RowBox[{"4", ",", "2"}], "]"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{
   FractionBox["1", 
    SuperscriptBox[
     TemplateBox[{RowBox[{"1", "-", "ep"}]},
      "Gamma"], "2"]], 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"4", " ", "\[Pi]"}], ")"}], 
    RowBox[{
     RowBox[{"2", " ", "ep"}], "-", "4"}]], " ", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{"-", "Q2"}], ")"}], 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"2", " ", "ep"}]}], "-", "2"}]], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{
        TemplateBox[{RowBox[{
            RowBox[{"2", " ", "ep"}], "+", "1"}]},
         "Gamma"], " ", 
        SuperscriptBox[
         TemplateBox[{RowBox[{"1", "-", "ep"}]},
          "Gamma"], "5"], " ", 
        TagBox[
         TagBox[
          RowBox[{
           RowBox[{
            SubscriptBox["\[InvisiblePrefixScriptBase]", "4"], 
            SubscriptBox["F", "3"]}], "\[InvisibleApplication]", 
           RowBox[{"(", 
            RowBox[{
             TagBox[
              TagBox[
               RowBox[{
                TagBox["1",
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"-", "4"}], " ", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"-", "2"}], " ", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox[
              TagBox[
               RowBox[{
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"3", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"2", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"2", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox["1",
              HypergeometricPFQ,
              Editable->True,
              Selectable->True]}], ")"}]}],
          InterpretTemplate[HypergeometricPFQ[#, #2, #3]& ],
          Editable->False,
          Selectable->False],
         HypergeometricPFQ]}], 
       RowBox[{"2", " ", 
        SuperscriptBox["ep", "4"], " ", 
        TemplateBox[{RowBox[{"1", "-", 
            RowBox[{"3", " ", "ep"}]}]},
         "Gamma"]}]]}], "+", 
     FractionBox[
      RowBox[{
       TemplateBox[{RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}]},
        "Gamma"], " ", 
       TemplateBox[{RowBox[{"ep", "+", "1"}]},
        "Gamma"], " ", 
       TemplateBox[{RowBox[{
           RowBox[{"2", " ", "ep"}], "+", "1"}]},
        "Gamma"], " ", 
       SuperscriptBox[
        TemplateBox[{RowBox[{"1", "-", "ep"}]},
         "Gamma"], "4"], " ", 
       TagBox[
        TagBox[
         RowBox[{
          RowBox[{
           SubscriptBox["\[InvisiblePrefixScriptBase]", "3"], 
           SubscriptBox["F", "2"]}], "\[InvisibleApplication]", 
          RowBox[{"(", 
           RowBox[{
            TagBox[
             TagBox[
              RowBox[{
               TagBox["1",
                HypergeometricPFQ,
                Editable->True,
                Selectable->True], ",", 
               TagBox[
                RowBox[{
                 RowBox[{"-", "4"}], " ", "ep"}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True], ",", 
               TagBox[
                RowBox[{
                 RowBox[{"-", "2"}], " ", "ep"}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True]}],
              InterpretTemplate[{
                SlotSequence[1]}& ]],
             HypergeometricPFQ,
             Editable->False,
             Selectable->False], ";", 
            TagBox[
             TagBox[
              RowBox[{
               TagBox[
                RowBox[{"1", "-", 
                 RowBox[{"3", " ", "ep"}]}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True], ",", 
               TagBox[
                RowBox[{"1", "-", 
                 RowBox[{"2", " ", "ep"}]}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True]}],
              InterpretTemplate[{
                SlotSequence[1]}& ]],
             HypergeometricPFQ,
             Editable->False,
             Selectable->False], ";", 
            TagBox["1",
             HypergeometricPFQ,
             Editable->True,
             Selectable->True]}], ")"}]}],
         InterpretTemplate[HypergeometricPFQ[#, #2, #3]& ],
         Editable->False,
         Selectable->False],
        HypergeometricPFQ]}], 
      RowBox[{"2", " ", 
       SuperscriptBox["ep", "4"], " ", 
       TemplateBox[{RowBox[{"1", "-", 
           RowBox[{"3", " ", "ep"}]}]},
        "Gamma"]}]], "-", 
     FractionBox[
      RowBox[{"4", " ", 
       TemplateBox[{RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}]},
        "Gamma"], " ", 
       TemplateBox[{RowBox[{
           RowBox[{"2", " ", "ep"}], "+", "1"}]},
        "Gamma"], " ", 
       SuperscriptBox[
        TemplateBox[{RowBox[{"1", "-", "ep"}]},
         "Gamma"], "4"], " ", 
       TagBox[
        TagBox[
         RowBox[{
          RowBox[{
           SubscriptBox["\[InvisiblePrefixScriptBase]", "3"], 
           SubscriptBox["F", "2"]}], "\[InvisibleApplication]", 
          RowBox[{"(", 
           RowBox[{
            TagBox[
             TagBox[
              RowBox[{
               TagBox["1",
                HypergeometricPFQ,
                Editable->True,
                Selectable->True], ",", 
               TagBox["1",
                HypergeometricPFQ,
                Editable->True,
                Selectable->True], ",", 
               TagBox[
                RowBox[{
                 RowBox[{"2", " ", "ep"}], "+", "1"}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True]}],
              InterpretTemplate[{
                SlotSequence[1]}& ]],
             HypergeometricPFQ,
             Editable->False,
             Selectable->False], ";", 
            TagBox[
             TagBox[
              RowBox[{
               TagBox[
                RowBox[{"ep", "+", "2"}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True], ",", 
               TagBox[
                RowBox[{
                 RowBox[{"2", " ", "ep"}], "+", "2"}],
                HypergeometricPFQ,
                Editable->True,
                Selectable->True]}],
              InterpretTemplate[{
                SlotSequence[1]}& ]],
             HypergeometricPFQ,
             Editable->False,
             Selectable->False], ";", 
            TagBox["1",
             HypergeometricPFQ,
             Editable->True,
             Selectable->True]}], ")"}]}],
         InterpretTemplate[HypergeometricPFQ[#, #2, #3]& ],
         Editable->False,
         Selectable->False],
        HypergeometricPFQ]}], 
      RowBox[{
       SuperscriptBox["ep", "2"], " ", 
       RowBox[{"(", 
        RowBox[{"ep", "+", "1"}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "ep"}], "+", "1"}], ")"}], " ", 
       TemplateBox[{RowBox[{"1", "-", 
           RowBox[{"4", " ", "ep"}]}]},
        "Gamma"]}]], "-", 
     FractionBox[
      RowBox[{
       SuperscriptBox[
        TemplateBox[{RowBox[{"1", "-", 
            RowBox[{"2", " ", "ep"}]}]},
         "Gamma"], "4"], " ", 
       TemplateBox[{RowBox[{"ep", "+", "1"}]},
        "Gamma"], " ", 
       SuperscriptBox[
        TemplateBox[{RowBox[{
            RowBox[{"2", " ", "ep"}], "+", "1"}]},
         "Gamma"], "3"], " ", 
       SuperscriptBox[
        TemplateBox[{RowBox[{"1", "-", "ep"}]},
         "Gamma"], "3"]}], 
      RowBox[{
       SuperscriptBox["ep", "4"], " ", 
       SuperscriptBox[
        TemplateBox[{RowBox[{"1", "-", 
            RowBox[{"4", " ", "ep"}]}]},
         "Gamma"], "2"], " ", 
       TemplateBox[{RowBox[{
           RowBox[{"4", " ", "ep"}], "+", "1"}]},
        "Gamma"]}]]}], ")"}]}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["HypergeometricPFQ expansion via HypExp+HPL", "Subsection"],

Cell["\<\
See HypExp home page for the installation and documentation:
https://www.physik.uzh.ch/data/HypExp/

Also see HPL home page for the same:
https://www.physik.uzh.ch/data/HPL/\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"<<", "HypExp`"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$HPLAutoConvertToKnownFunctions", "=", "False"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"$HPLAutoReduceToMinimalSet", "=", "False"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"$HypExpPolyLogRules", "=", "False"}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 FormBox["\<\"*-*-*-*-*-* HPL 2.0 *-*-*-*-*-*\\n\\nAuthor: Daniel Maitre, \
University of Zurich\"\>", TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"Rules for minimal set loaded for weights: 2, 3, 4, 5, 6, \
7.\"\>", TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"Rules for minimal set for + - weights loaded for weights: 2, 3, \
4, 5, 6, 7.\"\>", TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"Table of MZVs loaded up to weight 7\"\>", 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"Table of values at I loaded up to weight 6\"\>", 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"$HPLFunctions gives a list of the functions of the \
package.\\n$HPLOptions gives a list of the options of the package.\"\>", 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"More info in hep-ph/0507152, hep-ph/0703052 and at \\n \
http://krone.physik.unizh.ch/~maitreda/HPL/\"\>", TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"***********************************\\n***********  HypExp 2.0  \
************\\n***********************************\\nAuthors:\\n Tobias \
Huber:  RWTH Aachen,\\n\\nDaniel Maitre: SLAC, University of Zurich.\\n\"\>", 
  TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"HypExp loaded! It allows the expansion of hypergeometric \
functions around their parameters. \\n The new provided commands are:\\n - \
HypExp\\n - HypExpInt\\n - HypExpU\\n - HypExpAddToLib\\n - \
HypExpIsKnownToOrder\"\>", TraditionalForm]], "Print"],

Cell[BoxData[
 FormBox["\<\"More info in hep-ph/0507094 and at \\n \
http://krone.physik.unizh.ch/~maitreda/HypExp/\"\>", 
  TraditionalForm]], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"hpfqmap", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"mastervalues", "//", 
      RowBox[{
       RowBox[{"Cases", "[", 
        RowBox[{"#", ",", "_HypergeometricPFQ", ",", 
         RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}], "//", 
    RowBox[{"MapWithProgress", "[", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"#", "\[Rule]", 
        RowBox[{
         RowBox[{"HypExp", "[", 
          RowBox[{"#", ",", "ep", ",", "7"}], "]"}], "+", 
         RowBox[{
          RowBox[{"O", "[", "ep", "]"}], "^", "8"}]}]}], ")"}], "&"}], 
     "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"hpfqmap", "//", "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       RowBox[{
        TagBox[
         TagBox[
          RowBox[{
           RowBox[{
            SubscriptBox["\[InvisiblePrefixScriptBase]", "3"], 
            SubscriptBox["F", "2"]}], "\[InvisibleApplication]", 
           RowBox[{"(", 
            RowBox[{
             TagBox[
              TagBox[
               RowBox[{
                TagBox["1",
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox["1",
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"2", " ", "ep"}], "+", "1"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox[
              TagBox[
               RowBox[{
                TagBox[
                 RowBox[{"ep", "+", "2"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"2", " ", "ep"}], "+", "2"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox["1",
              HypergeometricPFQ,
              Editable->True,
              Selectable->True]}], ")"}]}],
          InterpretTemplate[HypergeometricPFQ[#, #2, #3]& ],
          Editable->False,
          Selectable->False],
         HypergeometricPFQ], "\[Rule]", 
        InterpretationBox[
         RowBox[{
          FractionBox[
           SuperscriptBox["\[Pi]", "2"], "6"], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              SuperscriptBox["\[Pi]", "2"], "2"], "-", 
             RowBox[{"4", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", "ep"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              SuperscriptBox["\[Pi]", "2"], "3"], "+", 
             FractionBox[
              RowBox[{"19", " ", 
               SuperscriptBox["\[Pi]", "4"]}], "180"], "-", 
             RowBox[{"12", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              RowBox[{"19", " ", 
               SuperscriptBox["\[Pi]", "4"]}], "60"], "-", 
             RowBox[{"8", " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "+", 
             FractionBox[
              RowBox[{
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "3"], "-", 
             RowBox[{"23", " ", 
              TemplateBox[{"5"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"342", " ", 
               SuperscriptBox["\[Pi]", "4"]}], "+", 
              RowBox[{"73", " ", 
               SuperscriptBox["\[Pi]", "6"]}], "+", 
              RowBox[{"1620", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "-", 
              RowBox[{"6480", " ", 
               SuperscriptBox[
                TemplateBox[{"3"},
                 "Zeta"], "2"]}], "-", 
              RowBox[{"111780", " ", 
               TemplateBox[{"5"},
                "Zeta"]}]}], ")"}], " ", 
            SuperscriptBox["ep", "4"]}], "1620"], "+", 
          RowBox[{
           FractionBox["1", "540"], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"73", " ", 
              SuperscriptBox["\[Pi]", "6"]}], "+", 
             RowBox[{"360", " ", 
              SuperscriptBox["\[Pi]", "2"], " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "+", 
             RowBox[{"114", " ", 
              SuperscriptBox["\[Pi]", "4"], " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "-", 
             RowBox[{"6480", " ", 
              SuperscriptBox[
               TemplateBox[{"3"},
                "Zeta"], "2"]}], "-", 
             RowBox[{"24840", " ", 
              TemplateBox[{"5"},
               "Zeta"]}], "+", 
             RowBox[{"540", " ", 
              SuperscriptBox["\[Pi]", "2"], " ", 
              TemplateBox[{"5"},
               "Zeta"]}], "-", 
             RowBox[{"56835", " ", 
              TemplateBox[{"7"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "5"]}], "+", 
          FractionBox[
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "6"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "5"}]]}], "+", 
              RowBox[{"3402000", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"5", ",", "3"}]]}], "+", 
              RowBox[{"7030800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"6", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "5"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "2", ",", "4"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "3", ",", "3"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "4", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "5", ",", "1"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "1", ",", "4"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "2", ",", "3"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "3", ",", "2"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "4", ",", "1"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "1", ",", "3"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "2", ",", "2"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "3", ",", "1"}]]}], "+", 
              RowBox[{"3402000", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"5", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"3402000", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"5", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"7030800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"6", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "1", ",", "4"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "2", ",", "3"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "3", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "4", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "2", ",", "1", ",", "3"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "2", ",", "3", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "3", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "3", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "4", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "1", ",", "1", ",", "3"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "1", ",", "2", ",", "2"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "2", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "2", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "3", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "1", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "1", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "2", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"3402000", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"5", ",", "1", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "1", ",", "1", ",", "3"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "1", ",", "2", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "1", ",", "3", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "2", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "2", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "1", ",", "3", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "2", ",", "1", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "2", ",", "1", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "2", ",", "2", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"2", ",", "3", ",", "1", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "1", ",", "1", ",", "1", ",", "2"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "1", ",", "1", ",", "2", ",", "1"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "1", ",", "2", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"680400", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"3", ",", "2", ",", "1", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"1587600", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{"4", ",", "1", ",", "1", ",", "1", ",", "1"}]]}], "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{
                "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", "2"}]]}], 
              "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{
                "2", ",", "1", ",", "1", ",", "1", ",", "2", ",", "1"}]]}], 
              "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{
                "2", ",", "1", ",", "1", ",", "2", ",", "1", ",", "1"}]]}], 
              "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{
                "2", ",", "1", ",", "2", ",", "1", ",", "1", ",", "1"}]]}], 
              "+", 
              RowBox[{"226800", " ", 
               SubscriptBox["\[Zeta]", 
                RowBox[{
                "2", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1"}]]}], 
              "-", 
              RowBox[{"71612100", " ", 
               TemplateBox[{"7"},
                "Zeta"]}], "-", 
              RowBox[{"14968800", " ", 
               TemplateBox[{"3"},
                "Zeta"], " ", 
               TemplateBox[{"5"},
                "Zeta"]}], "+", 
              RowBox[{"680400", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"5"},
                "Zeta"]}], "-", 
              RowBox[{"1814400", " ", 
               SuperscriptBox[
                TemplateBox[{"3"},
                 "Zeta"], "2"]}], "+", 
              RowBox[{"143640", " ", 
               SuperscriptBox["\[Pi]", "4"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "+", 
              RowBox[{"5067", " ", 
               SuperscriptBox["\[Pi]", "8"]}], "+", 
              RowBox[{"20440", " ", 
               SuperscriptBox["\[Pi]", "6"]}]}], ")"}], " ", 
            SuperscriptBox["ep", "6"]}], "226800"], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"2", ",", "7"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "6"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "5"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "4"}]]}], "-", 
             RowBox[{"31", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"6", ",", "3"}]]}], "-", 
             RowBox[{"63", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"7", ",", "2"}]]}], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "6"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "5"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "4"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "4", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "5", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "6", ",", "1"}]], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "5"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "4"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "3", ",", "3"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "4", ",", "2"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "5", ",", "1"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "4"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "2", ",", "3"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "3", ",", "2"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "4", ",", "1"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "1", ",", "3"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "2", ",", "2"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "3", ",", "1"}]]}], "-", 
             RowBox[{"31", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"6", ",", "1", ",", "2"}]]}], "-", 
             RowBox[{"31", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"6", ",", "2", ",", "1"}]]}], "-", 
             RowBox[{"63", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"7", ",", "1", ",", "1"}]]}], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "1", ",", "5"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "2", ",", "4"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "3", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "4", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "5", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "1", ",", "4"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "2", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "3", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "4", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "1", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "2", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "3", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "4", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "4", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "5", ",", "1", ",", "1"}]], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "1", ",", "4"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "2", ",", "3"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "3", ",", "2"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "4", ",", "1"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "1", ",", "3"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "2", ",", "2"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "3", ",", "1"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "3", ",", "1", ",", "2"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "3", ",", "2", ",", "1"}]]}], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "4", ",", "1", ",", "1"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "1", ",", "3"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "2", ",", "2"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "3", ",", "1"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "2", ",", "1", ",", "2"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "2", ",", "2", ",", "1"}]]}], "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "3", ",", "1", ",", "1"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "1", ",", "1", ",", "2"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "1", ",", "2", ",", "1"}]]}], "-", 
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "2", ",", "1", ",", "1"}]]}], "-", 
             RowBox[{"31", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"6", ",", "1", ",", "1", ",", "1"}]]}], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "1", ",", "1", ",", "4"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "1", ",", "2", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "1", ",", "3", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "1", ",", "4", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "2", ",", "1", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "2", ",", "2", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "2", ",", "3", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "3", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "3", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "1", ",", "4", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "1", ",", "1", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "1", ",", "2", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "1", ",", "3", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "2", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "2", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "2", ",", "3", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "1", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "1", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "3", ",", "2", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{"2", ",", "4", ",", "1", ",", "1", ",", "1"}]], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "1", ",", "1", ",", "3"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "1", ",", "2", ",", "2"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "1", ",", "3", ",", "1"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "2", ",", "1", ",", "2"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "2", ",", "2", ",", "1"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "1", ",", "3", ",", "1", ",", "1"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "1", ",", "1", ",", "2"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "1", ",", "2", ",", "1"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "2", ",", "2", ",", "1", ",", "1"}]]}], "-", 
             
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"3", ",", "3", ",", "1", ",", "1", ",", "1"}]]}], "-", 
             
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "1", ",", "1", ",", "2"}]]}], "-", 
             
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "1", ",", "2", ",", "1"}]]}], "-", 
             
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "1", ",", "2", ",", "1", ",", "1"}]]}], "-", 
             
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"4", ",", "2", ",", "1", ",", "1", ",", "1"}]]}], "-", 
             
             RowBox[{"15", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{"5", ",", "1", ",", "1", ",", "1", ",", "1"}]]}], "-", 
             
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", "3"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "1", ",", "2", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "1", ",", "3", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "2", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "2", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "3", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "2", ",", "1", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "2", ",", "1", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "2", ",", "2", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "3", ",", "1", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "2", ",", "1", ",", "1", ",", "1", ",", "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "2", ",", "1", ",", "1", ",", "2", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "2", ",", "1", ",", "2", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "2", ",", "2", ",", "1", ",", "1", ",", "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "3", ",", "1", ",", "1", ",", "1", ",", "1"}]], "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{
               "3", ",", "1", ",", "1", ",", "1", ",", "1", ",", "2"}]]}], 
             "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{
               "3", ",", "1", ",", "1", ",", "1", ",", "2", ",", "1"}]]}], 
             "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{
               "3", ",", "1", ",", "1", ",", "2", ",", "1", ",", "1"}]]}], 
             "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{
               "3", ",", "1", ",", "2", ",", "1", ",", "1", ",", "1"}]]}], 
             "-", 
             RowBox[{"3", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{
               "3", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1"}]]}], 
             "-", 
             RowBox[{"7", " ", 
              SubscriptBox["\[Zeta]", 
               RowBox[{
               "4", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1"}]]}], 
             "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
               "2"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", "2", ",", 
               "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "1", ",", "2", ",", "1", ",", 
               "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "1", ",", "2", ",", "1", ",", "1", ",", 
               "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "1", ",", "2", ",", "1", ",", "1", ",", "1", ",", 
               "1"}]], "-", 
             SubscriptBox["\[Zeta]", 
              RowBox[{
              "2", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", 
               "1"}]], "+", 
             RowBox[{"3", " ", 
              RowBox[{"(", 
               RowBox[{
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "6"}]], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "5"}]]}], "+", 
                RowBox[{"15", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"5", ",", "3"}]]}], "+", 
                RowBox[{"31", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"6", ",", "2"}]]}], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "5"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "2", ",", "4"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "3", ",", "3"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "4", ",", "2"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "5", ",", "1"}]], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "1", ",", "4"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "2", ",", "3"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "3", ",", "2"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "4", ",", "1"}]]}], "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "1", ",", "3"}]]}], "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "2", ",", "2"}]]}], "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "3", ",", "1"}]]}], "+", 
                RowBox[{"15", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"5", ",", "1", ",", "2"}]]}], "+", 
                RowBox[{"15", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"5", ",", "2", ",", "1"}]]}], "+", 
                RowBox[{"31", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"6", ",", "1", ",", "1"}]]}], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "1", ",", "4"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "2", ",", "3"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "3", ",", "2"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "4", ",", "1"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "2", ",", "1", ",", "3"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "2", ",", "3", ",", "1"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "3", ",", "1", ",", "2"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "3", ",", "2", ",", "1"}]], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "4", ",", "1", ",", "1"}]], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "1", ",", "1", ",", "3"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "1", ",", "2", ",", "2"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "2", ",", "1", ",", "2"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "2", ",", "2", ",", "1"}]]}], "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "3", ",", "1", ",", "1"}]]}], "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "1", ",", "1", ",", "2"}]]}], "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "1", ",", "2", ",", "1"}]]}], "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "2", ",", "1", ",", "1"}]]}], "+", 
                RowBox[{"15", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"5", ",", "1", ",", "1", ",", "1"}]]}], "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "1", ",", "1", ",", "3"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "1", ",", "2", ",", "2"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "1", ",", "3", ",", "1"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "2", ",", "1", ",", "2"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "2", ",", "2", ",", "1"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "1", ",", "3", ",", "1", ",", "1"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "2", ",", "1", ",", "1", ",", "2"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "2", ",", "1", ",", "2", ",", "1"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "2", ",", "2", ",", "1", ",", "1"}]], "+", 
                
                SubscriptBox["\[Zeta]", 
                 RowBox[{"2", ",", "3", ",", "1", ",", "1", ",", "1"}]], "+", 
                
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "1", ",", "1", ",", "1", ",", "2"}]]}], 
                "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "1", ",", "1", ",", "2", ",", "1"}]]}], 
                "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "1", ",", "2", ",", "1", ",", "1"}]]}], 
                "+", 
                RowBox[{"3", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"3", ",", "2", ",", "1", ",", "1", ",", "1"}]]}], 
                "+", 
                RowBox[{"7", " ", 
                 SubscriptBox["\[Zeta]", 
                  RowBox[{"4", ",", "1", ",", "1", ",", "1", ",", "1"}]]}], 
                "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{
                 "2", ",", "1", ",", "1", ",", "1", ",", "1", ",", "2"}]], 
                "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{
                 "2", ",", "1", ",", "1", ",", "1", ",", "2", ",", "1"}]], 
                "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{
                 "2", ",", "1", ",", "1", ",", "2", ",", "1", ",", "1"}]], 
                "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{
                 "2", ",", "1", ",", "2", ",", "1", ",", "1", ",", "1"}]], 
                "+", 
                SubscriptBox["\[Zeta]", 
                 RowBox[{
                 "2", ",", "2", ",", "1", ",", "1", ",", "1", ",", "1"}]], 
                "+", 
                RowBox[{"63", " ", 
                 TemplateBox[{"7"},
                  "Zeta"]}], "+", 
                RowBox[{"63", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   TemplateBox[{"7"},
                    "Zeta"]}], ")"}]}], "+", 
                RowBox[{"66", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    SuperscriptBox["\[Pi]", "8"], "2700"], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox["\[Pi]", "8"], "2100"]}], "-", 
                    RowBox[{"2", " ", 
                    TemplateBox[{"3"},
                    "Zeta"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"]}], "+", 
                   RowBox[{"3", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{"16", " ", 
                 TemplateBox[{"5"},
                  "Zeta"]}], "+", 
                RowBox[{"16", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   TemplateBox[{"5"},
                    "Zeta"]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                   RowBox[{"3", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "36"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"23", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "540"], "-", 
                   RowBox[{"48", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "-", 
                   RowBox[{"24", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                    FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "+", 
                   RowBox[{"48", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{
                 FractionBox["3", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    SuperscriptBox["\[Pi]", "6"], "90"], "-", 
                   RowBox[{"12", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "-", 
                   RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                    FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "+", 
                   RowBox[{"12", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2160"], "-", 
                   RowBox[{"4", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "+", 
                   RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2160"]}], "+", 
                   RowBox[{"4", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "-", 
                   RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "4"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox["\[Pi]", "6"], "90"]}], "+", 
                   RowBox[{"12", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "+", 
                   RowBox[{"8", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                    FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{"12", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "36"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"23", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "540"]}], "+", 
                   RowBox[{"48", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "+", 
                   RowBox[{"24", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                    FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{"48", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    SuperscriptBox["\[Pi]", "6"], "180"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "+", 
                   RowBox[{"3", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2268"], "-", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    SuperscriptBox["\[Pi]", "6"], "180"], "-", 
                   RowBox[{"2", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "-", 
                   RowBox[{"3", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2268"], "-", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{"14", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"4", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2835"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2268"], "-", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}], "-", 
                RowBox[{"5", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "-", 
                FractionBox[
                 RowBox[{"19", " ", 
                  SuperscriptBox[
                   TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"], "+", 
                RowBox[{"4", " ", 
                 TemplateBox[{"3"},
                  "Zeta"]}], "+", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   TemplateBox[{"3"},
                    "Zeta"]}], ")"}]}], "-", 
                RowBox[{"127", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "8"], "9450"]}], ")"}]}], "+", 
                RowBox[{"32", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "6"], "945"]}], ")"}]}], "+", 
                RowBox[{"8", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "4"], "90"]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "2"], "6"]}], ")"}]}], "+", 
                FractionBox[
                 RowBox[{"13", " ", 
                  SuperscriptBox["\[Pi]", "8"]}], "75600"], "+", 
                FractionBox[
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Pi]", "6"]}], "105"], "+", 
                FractionBox[
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Pi]", "4"]}], "45"], "+", 
                FractionBox[
                 SuperscriptBox["\[Pi]", "2"], "3"], "-", "125"}], ")"}]}], 
             "-", 
             RowBox[{"130", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "6"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "945"]}], "-", 
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "45"], "-", 
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "3"]}], ")"}]}], "+", 
                RowBox[{"4", " ", 
                 TemplateBox[{"9"},
                  "Zeta"]}]}], ")"}]}], "-", 
             TemplateBox[{"9"},
              "Zeta"], "+", 
             RowBox[{"255", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                TemplateBox[{"9"},
                 "Zeta"]}], ")"}]}], "+", 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{"61", "-", 
                FractionBox[
                 SuperscriptBox["\[Pi]", "2"], "3"], "-", 
                FractionBox[
                 RowBox[{"4", " ", 
                  SuperscriptBox["\[Pi]", "4"]}], "45"], "-", 
                FractionBox[
                 RowBox[{"31", " ", 
                  SuperscriptBox["\[Pi]", "6"]}], "945"], "-", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "2"], "6"]}], ")"}]}], "-", 
                RowBox[{"8", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "4"], "90"]}], ")"}]}], "-", 
                RowBox[{"31", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   FractionBox[
                    SuperscriptBox["\[Pi]", "6"], "945"]}], ")"}]}], "-", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   TemplateBox[{"3"},
                    "Zeta"]}], ")"}]}], "-", 
                RowBox[{"4", " ", 
                 TemplateBox[{"3"},
                  "Zeta"]}], "+", 
                FractionBox[
                 RowBox[{"17", " ", 
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "120"], "-", 
                RowBox[{
                 FractionBox["1", "3"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"16", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   TemplateBox[{"5"},
                    "Zeta"]}], ")"}]}], "-", 
                RowBox[{"16", " ", 
                 TemplateBox[{"5"},
                  "Zeta"]}], "-", 
                FractionBox[
                 RowBox[{"79", " ", 
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "12"], "-", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                   RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "-", 
                RowBox[{"15", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   RowBox[{"11", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "-", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                RowBox[{"63", " ", 
                 RowBox[{"(", 
                  RowBox[{"1", "-", 
                   TemplateBox[{"7"},
                    "Zeta"]}], ")"}]}], "-", 
                RowBox[{"7", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "24"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "+", 
                FractionBox[
                 RowBox[{"107", " ", 
                  TemplateBox[{"7"},
                   "Zeta"]}], "2"], "-", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "-", 
                RowBox[{
                 FractionBox["11", "36"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"151", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "180"], "-", 
                   RowBox[{
                    FractionBox["3", "2"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{
                    FractionBox["17", "6"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                    FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{"54", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                    FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                   RowBox[{"18", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                    FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                   RowBox[{"72", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "+", 
                   RowBox[{"24", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "24"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "-", 
                   RowBox[{"102", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"37", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "180"], "-", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{
                    FractionBox["5", "6"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                    FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{"12", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                    FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                    FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                   RowBox[{"12", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "+", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "24"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "-", 
                   RowBox[{"24", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"17", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   RowBox[{
                    FractionBox["1", "6"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                    FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                   RowBox[{"3", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                    FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                   RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "-", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["3", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "180"]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                    FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                    FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                    FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                   RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "+", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{
                 FractionBox["7", "6"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"37", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "180"]}], "+", 
                   RowBox[{
                    FractionBox["1", "2"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                   RowBox[{
                    FractionBox["5", "6"], " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                    FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                   RowBox[{"12", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                    FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                    FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                   RowBox[{"12", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "-", 
                   RowBox[{"6", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "24"]}], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "+", 
                   RowBox[{"24", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                    FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                    FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "+", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "-", 
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                   RowBox[{"3", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                    RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "+", 
                   RowBox[{"2", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"]}], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                    RowBox[{"10", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}], "-", 
                RowBox[{
                 FractionBox["17", "10"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "30"], "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                   RowBox[{"5", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                    RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "-", 
                   RowBox[{"10", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"], "+", 
                    FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                    RowBox[{"11", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "+", 
                   RowBox[{"4", " ", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"]}], "-", 
                    FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                    RowBox[{"10", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "5"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "30"], "-", 
                FractionBox[
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "3"], "+", 
                RowBox[{"5", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                   RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "-", 
                RowBox[{"10", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   RowBox[{"11", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "+", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"]}], "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                   RowBox[{"10", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "90"], "-", 
                FractionBox[
                 RowBox[{
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "3"], "+", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                   RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"]}], "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                   RowBox[{"10", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "90"]}], "+", 
                FractionBox[
                 RowBox[{
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "3"], "-", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                   RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"]}], "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                   RowBox[{"10", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "5"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "30"]}], "+", 
                FractionBox[
                 RowBox[{"2", " ", 
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "3"], "-", 
                RowBox[{"5", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "90"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "-", 
                   RowBox[{"18", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "+", 
                RowBox[{"10", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   RowBox[{"11", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}], "-", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "45"]}], "-", 
                   FractionBox[
                    RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "3"], "+", 
                   RowBox[{"10", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "9"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"151", " ", 
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "180"]}], "+", 
                RowBox[{
                 FractionBox["3", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["17", "6"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{"54", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                   FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                RowBox[{"18", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                RowBox[{"72", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "-", 
                RowBox[{"24", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "24"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "+", 
                RowBox[{"102", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{"5", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"17", " ", 
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "360"]}], "+", 
                RowBox[{
                 FractionBox["1", "6"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                   FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "+", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{
              FractionBox["3", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"13", " ", 
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "180"]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "+", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                   FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "+", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"13", " ", 
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "180"], "-", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{
                 FractionBox["1", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"2", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                   FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "+", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "-", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{"6", " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"17", " ", 
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "360"], "-", 
                RowBox[{
                 FractionBox["1", "6"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                   FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "-", 
                RowBox[{"6", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "9"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"151", " ", 
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "180"], "-", 
                RowBox[{
                 FractionBox["3", "2"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "2"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{
                 FractionBox["17", "6"], " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                   FractionBox[
                    RowBox[{"9", " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"54", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                   RowBox[{"2", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "-", 
                   FractionBox[
                    RowBox[{"291", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                RowBox[{"18", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "6"], "-", 
                   FractionBox[
                    RowBox[{"109", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "16"]}], ")"}]}], "-", 
                RowBox[{"72", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    TemplateBox[{"7"},
                    "Zeta"], "4"]}], ")"}]}], "+", 
                RowBox[{"24", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "24"]}], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"5", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}], "-", 
                RowBox[{"102", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"7", " ", 
                    SuperscriptBox["\[Pi]", "4"], " ", 
                    TemplateBox[{"3"},
                    "Zeta"]}], "360"], "-", 
                   FractionBox[
                    RowBox[{"11", " ", 
                    SuperscriptBox["\[Pi]", "2"], " ", 
                    TemplateBox[{"5"},
                    "Zeta"]}], "12"], "+", 
                   FractionBox[
                    RowBox[{"61", " ", 
                    TemplateBox[{"7"},
                    "Zeta"]}], "8"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{"7", " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "360"], "-", 
                FractionBox[
                 RowBox[{"11", " ", 
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "12"], "+", 
                FractionBox[
                 RowBox[{"61", " ", 
                  TemplateBox[{"7"},
                   "Zeta"]}], "8"]}], ")"}]}], "-", 
             FractionBox[
              RowBox[{"851", " ", 
               TemplateBox[{"7"},
                "Zeta"]}], "16"], "-", 
             RowBox[{"11", " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{
                  SuperscriptBox["\[Pi]", "4"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "360"], "-", 
                FractionBox[
                 TemplateBox[{"7"},
                  "Zeta"], "4"]}], ")"}]}], "-", 
             RowBox[{"64", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                TemplateBox[{"7"},
                 "Zeta"]}], ")"}]}], "+", 
             RowBox[{"12", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "72"]}], "+", 
                FractionBox[
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "6"], "-", 
                FractionBox[
                 RowBox[{"109", " ", 
                  TemplateBox[{"7"},
                   "Zeta"]}], "16"]}], ")"}]}], "+", 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   SuperscriptBox["\[Pi]", "4"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "60"]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox["\[Pi]", "2"], " ", 
                 TemplateBox[{"5"},
                  "Zeta"]}], "-", 
                FractionBox[
                 RowBox[{"291", " ", 
                  TemplateBox[{"7"},
                   "Zeta"]}], "16"]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["5", "6"], " ", 
              SuperscriptBox["\[Pi]", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   SuperscriptBox["\[Pi]", "2"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                FractionBox[
                 RowBox[{"9", " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "2"]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{
                   SuperscriptBox["\[Pi]", "2"], " ", 
                   TemplateBox[{"3"},
                    "Zeta"]}], "3"]}], "+", 
                FractionBox[
                 RowBox[{"9", " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "2"]}], ")"}]}], "-", 
             FractionBox[
              RowBox[{"13", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"5"},
                "Zeta"]}], "12"], "-", 
             RowBox[{"7", " ", 
              TemplateBox[{"5"},
               "Zeta"]}], "-", 
             RowBox[{"16", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                TemplateBox[{"5"},
                 "Zeta"]}], ")"}]}], "+", 
             RowBox[{
              SuperscriptBox["\[Pi]", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{
                  SuperscriptBox["\[Pi]", "2"], " ", 
                  TemplateBox[{"3"},
                   "Zeta"]}], "2"], "-", 
                FractionBox[
                 RowBox[{"11", " ", 
                  TemplateBox[{"5"},
                   "Zeta"]}], "2"]}], ")"}]}], "-", 
             RowBox[{
              FractionBox["35", "36"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"23", " ", 
                  SuperscriptBox["\[Pi]", "6"]}], "540"], "-", 
                RowBox[{"48", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "-", 
                RowBox[{"24", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "+", 
                RowBox[{"48", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 SuperscriptBox["\[Pi]", "6"], "90"], "-", 
                RowBox[{"12", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "-", 
                RowBox[{"8", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "+", 
                RowBox[{"12", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{"10", " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"7", " ", 
                  SuperscriptBox["\[Pi]", "6"]}], "2160"], "-", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "+", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{"10", " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"7", " ", 
                   SuperscriptBox["\[Pi]", "6"]}], "2160"]}], "+", 
                RowBox[{"4", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "-", 
                RowBox[{"4", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{
              FractionBox["1", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  SuperscriptBox["\[Pi]", "6"], "90"]}], "+", 
                RowBox[{"12", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "+", 
                RowBox[{"8", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"12", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{
              FractionBox["35", "36"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  RowBox[{"23", " ", 
                   SuperscriptBox["\[Pi]", "6"]}], "540"]}], "+", 
                RowBox[{"48", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "+", 
                RowBox[{"24", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"53", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "22680"], "-", 
                   FractionBox[
                    RowBox[{"3", " ", 
                    SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], "2"]}], ")"}]}], "-", 
                RowBox[{"48", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{"-", 
                    FractionBox[
                    RowBox[{"13", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "15120"]}], "+", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{
              FractionBox["27", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", 
                 FractionBox[
                  SuperscriptBox["\[Pi]", "6"], "180"]}], "+", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "+", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2268"], "-", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "-", 
             RowBox[{
              FractionBox["25", "2"], " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 SuperscriptBox["\[Pi]", "6"], "180"], "-", 
                RowBox[{"2", " ", 
                 SuperscriptBox[
                  TemplateBox[{"3"},
                   "Zeta"], "2"]}], "-", 
                RowBox[{"3", " ", 
                 RowBox[{"(", 
                  RowBox[{
                   FractionBox[
                    RowBox[{"5", " ", 
                    SuperscriptBox["\[Pi]", "6"]}], "2268"], "-", 
                   SuperscriptBox[
                    TemplateBox[{"3"},
                    "Zeta"], "2"]}], ")"}]}]}], ")"}]}], "+", 
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{
                FractionBox[
                 RowBox[{"5", " ", 
                  SuperscriptBox["\[Pi]", "6"]}], "2268"], "-", 
                SuperscriptBox[
                 TemplateBox[{"3"},
                  "Zeta"], "2"]}], ")"}]}], "+", 
             RowBox[{"3", " ", 
              SuperscriptBox[
               TemplateBox[{"3"},
                "Zeta"], "2"]}], "-", 
             FractionBox[
              RowBox[{"37", " ", 
               SuperscriptBox["\[Pi]", "4"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "360"], "-", 
             FractionBox[
              RowBox[{"2", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "3"], "-", 
             RowBox[{"4", " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "-", 
             RowBox[{"4", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                TemplateBox[{"3"},
                 "Zeta"]}], ")"}]}], "-", 
             RowBox[{"127", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                FractionBox[
                 SuperscriptBox["\[Pi]", "8"], "9450"]}], ")"}]}], "-", 
             RowBox[{"32", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                FractionBox[
                 SuperscriptBox["\[Pi]", "6"], "945"]}], ")"}]}], "-", 
             RowBox[{"8", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                FractionBox[
                 SuperscriptBox["\[Pi]", "4"], "90"]}], ")"}]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{"(", 
               RowBox[{"1", "-", 
                FractionBox[
                 SuperscriptBox["\[Pi]", "2"], "6"]}], ")"}]}], "-", 
             FractionBox[
              RowBox[{"127", " ", 
               SuperscriptBox["\[Pi]", "8"]}], "9450"], "-", 
             FractionBox[
              RowBox[{"118", " ", 
               SuperscriptBox["\[Pi]", "6"]}], "2835"], "-", 
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox["\[Pi]", "4"]}], "45"], "-", 
             FractionBox[
              SuperscriptBox["\[Pi]", "2"], "3"], "+", "253"}], ")"}], " ", 
           SuperscriptBox["ep", "7"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "8"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, 0, 8, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 0, {
          Rational[1, 6] Pi^2, Rational[1, 2] Pi^2 - 4 Zeta[3], 
           Rational[1, 3] Pi^2 + Rational[19, 180] Pi^4 - 12 Zeta[3], 
           Rational[19, 60] Pi^4 - 8 Zeta[3] + Rational[1, 3] Pi^2 Zeta[3] - 
           23 Zeta[5], 
           Rational[1, 1620] (342 Pi^4 + 73 Pi^6 + 1620 Pi^2 Zeta[3] - 6480 
            Zeta[3]^2 - 111780 Zeta[5]), 
           Rational[1, 540] (73 Pi^6 + 360 Pi^2 Zeta[3] + 114 Pi^4 Zeta[3] - 
            6480 Zeta[3]^2 - 24840 Zeta[5] + 540 Pi^2 Zeta[5] - 56835 
            Zeta[7]), 
           Rational[1, 226800] (20440 Pi^6 + 5067 Pi^8 + 
            226800 $CellContext`MZV[{2, 6}] + 680400 $CellContext`MZV[{3, 5}] + 
            3402000 $CellContext`MZV[{5, 3}] + 
            7030800 $CellContext`MZV[{6, 2}] + 
            226800 $CellContext`MZV[{2, 1, 5}] + 
            226800 $CellContext`MZV[{2, 2, 4}] + 
            226800 $CellContext`MZV[{2, 3, 3}] + 
            226800 $CellContext`MZV[{2, 4, 2}] + 
            226800 $CellContext`MZV[{2, 5, 1}] + 
            680400 $CellContext`MZV[{3, 1, 4}] + 
            680400 $CellContext`MZV[{3, 2, 3}] + 
            680400 $CellContext`MZV[{3, 3, 2}] + 
            680400 $CellContext`MZV[{3, 4, 1}] + 
            1587600 $CellContext`MZV[{4, 1, 3}] + 
            1587600 $CellContext`MZV[{4, 2, 2}] + 
            1587600 $CellContext`MZV[{4, 3, 1}] + 
            3402000 $CellContext`MZV[{5, 1, 2}] + 
            3402000 $CellContext`MZV[{5, 2, 1}] + 
            7030800 $CellContext`MZV[{6, 1, 1}] + 
            226800 $CellContext`MZV[{2, 1, 1, 4}] + 
            226800 $CellContext`MZV[{2, 1, 2, 3}] + 
            226800 $CellContext`MZV[{2, 1, 3, 2}] + 
            226800 $CellContext`MZV[{2, 1, 4, 1}] + 
            226800 $CellContext`MZV[{2, 2, 1, 3}] + 
            226800 $CellContext`MZV[{2, 2, 3, 1}] + 
            226800 $CellContext`MZV[{2, 3, 1, 2}] + 
            226800 $CellContext`MZV[{2, 3, 2, 1}] + 
            226800 $CellContext`MZV[{2, 4, 1, 1}] + 
            680400 $CellContext`MZV[{3, 1, 1, 3}] + 
            680400 $CellContext`MZV[{3, 1, 2, 2}] + 
            680400 $CellContext`MZV[{3, 2, 1, 2}] + 
            680400 $CellContext`MZV[{3, 2, 2, 1}] + 
            680400 $CellContext`MZV[{3, 3, 1, 1}] + 
            1587600 $CellContext`MZV[{4, 1, 1, 2}] + 
            1587600 $CellContext`MZV[{4, 1, 2, 1}] + 
            1587600 $CellContext`MZV[{4, 2, 1, 1}] + 
            3402000 $CellContext`MZV[{5, 1, 1, 1}] + 
            226800 $CellContext`MZV[{2, 1, 1, 1, 3}] + 
            226800 $CellContext`MZV[{2, 1, 1, 2, 2}] + 
            226800 $CellContext`MZV[{2, 1, 1, 3, 1}] + 
            226800 $CellContext`MZV[{2, 1, 2, 1, 2}] + 
            226800 $CellContext`MZV[{2, 1, 2, 2, 1}] + 
            226800 $CellContext`MZV[{2, 1, 3, 1, 1}] + 
            226800 $CellContext`MZV[{2, 2, 1, 1, 2}] + 
            226800 $CellContext`MZV[{2, 2, 1, 2, 1}] + 
            226800 $CellContext`MZV[{2, 2, 2, 1, 1}] + 
            226800 $CellContext`MZV[{2, 3, 1, 1, 1}] + 
            680400 $CellContext`MZV[{3, 1, 1, 1, 2}] + 
            680400 $CellContext`MZV[{3, 1, 1, 2, 1}] + 
            680400 $CellContext`MZV[{3, 1, 2, 1, 1}] + 
            680400 $CellContext`MZV[{3, 2, 1, 1, 1}] + 
            1587600 $CellContext`MZV[{4, 1, 1, 1, 1}] + 
            226800 $CellContext`MZV[{2, 1, 1, 1, 1, 2}] + 
            226800 $CellContext`MZV[{2, 1, 1, 1, 2, 1}] + 
            226800 $CellContext`MZV[{2, 1, 1, 2, 1, 1}] + 
            226800 $CellContext`MZV[{2, 1, 2, 1, 1, 1}] + 
            226800 $CellContext`MZV[{2, 2, 1, 1, 1, 1}] + 143640 Pi^4 Zeta[3] - 
            1814400 Zeta[3]^2 + 680400 Pi^2 Zeta[5] - 14968800 Zeta[3] 
            Zeta[5] - 71612100 Zeta[7]), 253 + Rational[-1, 3] Pi^2 + 
           Rational[-4, 45] Pi^4 + Rational[-118, 2835] Pi^6 + 
           Rational[-127, 9450] Pi^8 - 2 (1 + Rational[-1, 6] Pi^2) - 
           8 (1 + Rational[-1, 90] Pi^4) - 32 (1 + Rational[-1, 945] Pi^6) - 
           127 (1 + Rational[-1, 9450] Pi^8) - $CellContext`MZV[{2, 7}] - 
           3 $CellContext`MZV[{3, 6}] - 7 $CellContext`MZV[{4, 5}] - 
           15 $CellContext`MZV[{5, 4}] - 31 $CellContext`MZV[{6, 3}] - 
           63 $CellContext`MZV[{7, 2}] - $CellContext`MZV[{2, 1, 
            6}] - $CellContext`MZV[{2, 2, 5}] - $CellContext`MZV[{2, 3, 
            4}] - $CellContext`MZV[{2, 4, 3}] - $CellContext`MZV[{2, 5, 
            2}] - $CellContext`MZV[{2, 6, 1}] - 3 $CellContext`MZV[{3, 1, 5}] - 
           3 $CellContext`MZV[{3, 2, 4}] - 3 $CellContext`MZV[{3, 3, 3}] - 
           3 $CellContext`MZV[{3, 4, 2}] - 3 $CellContext`MZV[{3, 5, 1}] - 
           7 $CellContext`MZV[{4, 1, 4}] - 7 $CellContext`MZV[{4, 2, 3}] - 
           7 $CellContext`MZV[{4, 3, 2}] - 7 $CellContext`MZV[{4, 4, 1}] - 
           15 $CellContext`MZV[{5, 1, 3}] - 15 $CellContext`MZV[{5, 2, 2}] - 
           15 $CellContext`MZV[{5, 3, 1}] - 31 $CellContext`MZV[{6, 1, 2}] - 
           31 $CellContext`MZV[{6, 2, 1}] - 
           63 $CellContext`MZV[{7, 1, 1}] - $CellContext`MZV[{2, 1, 1, 
            5}] - $CellContext`MZV[{2, 1, 2, 4}] - $CellContext`MZV[{2, 1, 3, 
            3}] - $CellContext`MZV[{2, 1, 4, 2}] - $CellContext`MZV[{2, 1, 5, 
            1}] - $CellContext`MZV[{2, 2, 1, 4}] - $CellContext`MZV[{2, 2, 2, 
            3}] - $CellContext`MZV[{2, 2, 3, 2}] - $CellContext`MZV[{2, 2, 4, 
            1}] - $CellContext`MZV[{2, 3, 1, 3}] - $CellContext`MZV[{2, 3, 2, 
            2}] - $CellContext`MZV[{2, 3, 3, 1}] - $CellContext`MZV[{2, 4, 1, 
            2}] - $CellContext`MZV[{2, 4, 2, 1}] - $CellContext`MZV[{2, 5, 1, 
            1}] - 3 $CellContext`MZV[{3, 1, 1, 4}] - 
           3 $CellContext`MZV[{3, 1, 2, 3}] - 
           3 $CellContext`MZV[{3, 1, 3, 2}] - 
           3 $CellContext`MZV[{3, 1, 4, 1}] - 
           3 $CellContext`MZV[{3, 2, 1, 3}] - 
           3 $CellContext`MZV[{3, 2, 2, 2}] - 
           3 $CellContext`MZV[{3, 2, 3, 1}] - 
           3 $CellContext`MZV[{3, 3, 1, 2}] - 
           3 $CellContext`MZV[{3, 3, 2, 1}] - 
           3 $CellContext`MZV[{3, 4, 1, 1}] - 
           7 $CellContext`MZV[{4, 1, 1, 3}] - 
           7 $CellContext`MZV[{4, 1, 2, 2}] - 
           7 $CellContext`MZV[{4, 1, 3, 1}] - 
           7 $CellContext`MZV[{4, 2, 1, 2}] - 
           7 $CellContext`MZV[{4, 2, 2, 1}] - 
           7 $CellContext`MZV[{4, 3, 1, 1}] - 
           15 $CellContext`MZV[{5, 1, 1, 2}] - 
           15 $CellContext`MZV[{5, 1, 2, 1}] - 
           15 $CellContext`MZV[{5, 2, 1, 1}] - 
           31 $CellContext`MZV[{6, 1, 1, 1}] - $CellContext`MZV[{2, 1, 1, 1, 
            4}] - $CellContext`MZV[{2, 1, 1, 2, 3}] - $CellContext`MZV[{2, 1, 
            1, 3, 2}] - $CellContext`MZV[{2, 1, 1, 4, 1}] - $CellContext`MZV[{
            2, 1, 2, 1, 3}] - $CellContext`MZV[{2, 1, 2, 2, 
            2}] - $CellContext`MZV[{2, 1, 2, 3, 1}] - $CellContext`MZV[{2, 1, 
            3, 1, 2}] - $CellContext`MZV[{2, 1, 3, 2, 1}] - $CellContext`MZV[{
            2, 1, 4, 1, 1}] - $CellContext`MZV[{2, 2, 1, 1, 
            3}] - $CellContext`MZV[{2, 2, 1, 2, 2}] - $CellContext`MZV[{2, 2, 
            1, 3, 1}] - $CellContext`MZV[{2, 2, 2, 1, 2}] - $CellContext`MZV[{
            2, 2, 2, 2, 1}] - $CellContext`MZV[{2, 2, 3, 1, 
            1}] - $CellContext`MZV[{2, 3, 1, 1, 2}] - $CellContext`MZV[{2, 3, 
            1, 2, 1}] - $CellContext`MZV[{2, 3, 2, 1, 1}] - $CellContext`MZV[{
            2, 4, 1, 1, 1}] - 3 $CellContext`MZV[{3, 1, 1, 1, 3}] - 
           3 $CellContext`MZV[{3, 1, 1, 2, 2}] - 
           3 $CellContext`MZV[{3, 1, 1, 3, 1}] - 
           3 $CellContext`MZV[{3, 1, 2, 1, 2}] - 
           3 $CellContext`MZV[{3, 1, 2, 2, 1}] - 
           3 $CellContext`MZV[{3, 1, 3, 1, 1}] - 
           3 $CellContext`MZV[{3, 2, 1, 1, 2}] - 
           3 $CellContext`MZV[{3, 2, 1, 2, 1}] - 
           3 $CellContext`MZV[{3, 2, 2, 1, 1}] - 
           3 $CellContext`MZV[{3, 3, 1, 1, 1}] - 
           7 $CellContext`MZV[{4, 1, 1, 1, 2}] - 
           7 $CellContext`MZV[{4, 1, 1, 2, 1}] - 
           7 $CellContext`MZV[{4, 1, 2, 1, 1}] - 
           7 $CellContext`MZV[{4, 2, 1, 1, 1}] - 
           15 $CellContext`MZV[{5, 1, 1, 1, 1}] - $CellContext`MZV[{2, 1, 1, 
            1, 1, 3}] - $CellContext`MZV[{2, 1, 1, 1, 2, 
            2}] - $CellContext`MZV[{2, 1, 1, 1, 3, 1}] - $CellContext`MZV[{2, 
            1, 1, 2, 1, 2}] - $CellContext`MZV[{2, 1, 1, 2, 2, 
            1}] - $CellContext`MZV[{2, 1, 1, 3, 1, 1}] - $CellContext`MZV[{2, 
            1, 2, 1, 1, 2}] - $CellContext`MZV[{2, 1, 2, 1, 2, 
            1}] - $CellContext`MZV[{2, 1, 2, 2, 1, 1}] - $CellContext`MZV[{2, 
            1, 3, 1, 1, 1}] - $CellContext`MZV[{2, 2, 1, 1, 1, 
            2}] - $CellContext`MZV[{2, 2, 1, 1, 2, 1}] - $CellContext`MZV[{2, 
            2, 1, 2, 1, 1}] - $CellContext`MZV[{2, 2, 2, 1, 1, 
            1}] - $CellContext`MZV[{2, 3, 1, 1, 1, 1}] - 
           3 $CellContext`MZV[{3, 1, 1, 1, 1, 2}] - 
           3 $CellContext`MZV[{3, 1, 1, 1, 2, 1}] - 
           3 $CellContext`MZV[{3, 1, 1, 2, 1, 1}] - 
           3 $CellContext`MZV[{3, 1, 2, 1, 1, 1}] - 
           3 $CellContext`MZV[{3, 2, 1, 1, 1, 1}] - 
           7 $CellContext`MZV[{4, 1, 1, 1, 1, 1}] - $CellContext`MZV[{2, 1, 1,
             1, 1, 1, 2}] - $CellContext`MZV[{2, 1, 1, 1, 1, 2, 
            1}] - $CellContext`MZV[{2, 1, 1, 1, 2, 1, 1}] - $CellContext`MZV[{
            2, 1, 1, 2, 1, 1, 1}] - $CellContext`MZV[{2, 1, 2, 1, 1, 1, 
            1}] - $CellContext`MZV[{2, 2, 1, 1, 1, 1, 1}] - 4 (1 - Zeta[3]) - 
           4 Zeta[3] + Rational[-2, 3] Pi^2 Zeta[3] + 
           Rational[-37, 360] Pi^4 Zeta[3] + 3 Zeta[3]^2 + 
           4 (Rational[5, 2268] Pi^6 - Zeta[3]^2) + 
           Rational[-25, 2] (Rational[1, 180] Pi^6 - 2 Zeta[3]^2 - 
             3 (Rational[5, 2268] Pi^6 - Zeta[3]^2)) + 
           Rational[-27, 2] (Rational[-1, 180] Pi^6 + 2 Zeta[3]^2 + 
             3 (Rational[5, 2268] Pi^6 - Zeta[3]^2)) + 
           Rational[-35, 36] (Rational[-23, 540] Pi^6 + 48 Zeta[3]^2 + 
             24 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) - 
             48 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
           Rational[1, 2] (Rational[-1, 90] Pi^6 + 12 Zeta[3]^2 + 
             8 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) - 
             12 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) - 
           10 (Rational[-7, 2160] Pi^6 + 4 Zeta[3]^2 - 
            4 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) - 
           10 (Rational[7, 2160] Pi^6 - 4 Zeta[3]^2 + 
            4 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
           Rational[1, 2] (Rational[1, 90] Pi^6 - 12 Zeta[3]^2 - 
             8 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) + 
             12 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
           Rational[-35, 36] (Rational[23, 540] Pi^6 - 48 Zeta[3]^2 - 
             24 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) + 
             48 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
           Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) - 
           16 (1 - Zeta[5]) - 7 Zeta[5] + Rational[-13, 12] Pi^2 Zeta[5] - 
           2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
           Rational[5, 6] 
            Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
           2 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
             Rational[-291, 16] Zeta[7]) + 
           12 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
             Rational[-109, 16] Zeta[7]) - 64 (1 - Zeta[7]) - 
           11 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
           Rational[-851, 16] Zeta[7] + 
           7 (Rational[7, 360] Pi^4 Zeta[3] + Rational[-11, 12] Pi^2 Zeta[5] + 
             Rational[61, 8] Zeta[7]) + 
           3 (-125 + Rational[1, 3] Pi^2 + Rational[4, 45] Pi^4 + 
             Rational[4, 105] Pi^6 + Rational[13, 75600] Pi^8 + 
             2 (1 + Rational[-1, 6] Pi^2) + 8 (1 + Rational[-1, 90] Pi^4) + 
             32 (1 + Rational[-1, 945] Pi^6) - 
             127 (1 + Rational[-1, 9450] Pi^8) + $CellContext`MZV[{2, 6}] + 
             3 $CellContext`MZV[{3, 5}] + 15 $CellContext`MZV[{5, 3}] + 
             31 $CellContext`MZV[{6, 2}] + $CellContext`MZV[{2, 1, 
               5}] + $CellContext`MZV[{2, 2, 4}] + $CellContext`MZV[{2, 3, 
               3}] + $CellContext`MZV[{2, 4, 2}] + $CellContext`MZV[{2, 5, 
               1}] + 3 $CellContext`MZV[{3, 1, 4}] + 
             3 $CellContext`MZV[{3, 2, 3}] + 3 $CellContext`MZV[{3, 3, 2}] + 
             3 $CellContext`MZV[{3, 4, 1}] + 7 $CellContext`MZV[{4, 1, 3}] + 
             7 $CellContext`MZV[{4, 2, 2}] + 7 $CellContext`MZV[{4, 3, 1}] + 
             15 $CellContext`MZV[{5, 1, 2}] + 15 $CellContext`MZV[{5, 2, 1}] + 
             31 $CellContext`MZV[{6, 1, 1}] + $CellContext`MZV[{2, 1, 1, 
               4}] + $CellContext`MZV[{2, 1, 2, 3}] + $CellContext`MZV[{2, 1, 
               3, 2}] + $CellContext`MZV[{2, 1, 4, 1}] + $CellContext`MZV[{2, 
               2, 1, 3}] + $CellContext`MZV[{2, 2, 3, 1}] + $CellContext`MZV[{
               2, 3, 1, 2}] + $CellContext`MZV[{2, 3, 2, 
               1}] + $CellContext`MZV[{2, 4, 1, 1}] + 
             3 $CellContext`MZV[{3, 1, 1, 3}] + 
             3 $CellContext`MZV[{3, 1, 2, 2}] + 
             3 $CellContext`MZV[{3, 2, 1, 2}] + 
             3 $CellContext`MZV[{3, 2, 2, 1}] + 
             3 $CellContext`MZV[{3, 3, 1, 1}] + 
             7 $CellContext`MZV[{4, 1, 1, 2}] + 
             7 $CellContext`MZV[{4, 1, 2, 1}] + 
             7 $CellContext`MZV[{4, 2, 1, 1}] + 
             15 $CellContext`MZV[{5, 1, 1, 1}] + $CellContext`MZV[{2, 1, 1, 1,
                3}] + $CellContext`MZV[{2, 1, 1, 2, 2}] + $CellContext`MZV[{2,
                1, 1, 3, 1}] + $CellContext`MZV[{2, 1, 2, 1, 
               2}] + $CellContext`MZV[{2, 1, 2, 2, 1}] + $CellContext`MZV[{2, 
               1, 3, 1, 1}] + $CellContext`MZV[{2, 2, 1, 1, 
               2}] + $CellContext`MZV[{2, 2, 1, 2, 1}] + $CellContext`MZV[{2, 
               2, 2, 1, 1}] + $CellContext`MZV[{2, 3, 1, 1, 1}] + 
             3 $CellContext`MZV[{3, 1, 1, 1, 2}] + 
             3 $CellContext`MZV[{3, 1, 1, 2, 1}] + 
             3 $CellContext`MZV[{3, 1, 2, 1, 1}] + 
             3 $CellContext`MZV[{3, 2, 1, 1, 1}] + 
             7 $CellContext`MZV[{4, 1, 1, 1, 1}] + $CellContext`MZV[{2, 1, 1, 
               1, 1, 2}] + $CellContext`MZV[{2, 1, 1, 1, 2, 
               1}] + $CellContext`MZV[{2, 1, 1, 2, 1, 1}] + $CellContext`MZV[{
               2, 1, 2, 1, 1, 1}] + $CellContext`MZV[{2, 2, 1, 1, 1, 1}] + 
             4 (1 - Zeta[3]) + 4 Zeta[3] + Rational[-19, 2] Zeta[3]^2 - 
             5 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) + 
             2 (Rational[5, 2268] Pi^6 - Zeta[3]^2) + 
             2 (Rational[-4, 2835] Pi^6 + Zeta[3]^2) + 
             14 (Rational[-13, 15120] Pi^6 + Zeta[3]^2) + 
             Rational[1, 2] (Rational[1, 180] Pi^6 - 2 Zeta[3]^2 - 
               3 (Rational[5, 2268] Pi^6 - Zeta[3]^2)) + 
             Rational[1, 2] (Rational[-1, 180] Pi^6 + 2 Zeta[3]^2 + 
               3 (Rational[5, 2268] Pi^6 - Zeta[3]^2)) + 
             Rational[1, 36] (Rational[-23, 540] Pi^6 + 48 Zeta[3]^2 + 
               24 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) - 
               48 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
             Rational[1, 4] (Rational[-1, 90] Pi^6 + 12 Zeta[3]^2 + 
               8 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) - 
               12 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
             2 (Rational[-7, 2160] Pi^6 + 4 Zeta[3]^2 - 
               4 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
             2 (Rational[7, 2160] Pi^6 - 4 Zeta[3]^2 + 
               4 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
             Rational[-3, 4] (Rational[1, 90] Pi^6 - 12 Zeta[3]^2 - 
               8 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) + 
               12 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
             Rational[1, 36] (Rational[23, 540] Pi^6 - 48 Zeta[3]^2 - 
               24 (Rational[53, 22680] Pi^6 + Rational[-3, 2] Zeta[3]^2) + 
               48 (Rational[-13, 15120] Pi^6 + Zeta[3]^2)) + 
             Rational[1, 2] (Rational[1, 2] Pi^2 Zeta[3] - 
               3 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
               Rational[-9, 2] Zeta[5]) + 16 (1 - Zeta[5]) + 16 Zeta[5] + 
             Rational[1, 2] (Rational[-1, 2] Pi^2 Zeta[3] + 
               3 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
               Rational[9, 2] Zeta[5]) + 
             66 (Rational[1, 2700] Pi^8 + 
               Rational[1, 2] (Rational[-1, 2100] Pi^8 - 2 Zeta[3] Zeta[5])) + 
             63 (1 - Zeta[7]) + 63 Zeta[7]) + 
           Rational[1, 9] (Rational[151, 180] Pi^4 Zeta[3] + 
             Rational[-3, 2] 
              Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
             Rational[-17, 6] 
              Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
             54 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
              Rational[-291, 16] Zeta[7]) - 
             18 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
              Rational[-109, 16] Zeta[7]) - 
             72 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
             24 (Rational[-1, 24] Pi^4 Zeta[3] + Rational[5, 12] Pi^2 Zeta[5] + 
               Rational[5, 8] Zeta[7]) - 
             102 (Rational[7, 360] Pi^4 Zeta[3] + 
              Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) - 
           6 (Rational[17, 360] Pi^4 Zeta[3] + 
            Rational[-1, 6] 
             Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
            3 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
             Rational[-291, 16] Zeta[7]) - 
            4 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) - 
            6 (Rational[7, 360] Pi^4 Zeta[3] + Rational[-11, 12] Pi^2 Zeta[5] + 
             Rational[61, 8] Zeta[7])) + 
           Rational[1, 2] (Rational[13, 180] Pi^4 Zeta[3] + 
             Rational[-1, 2] 
              Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
             Rational[-1, 2] 
              Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
             2 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
              Rational[-291, 16] Zeta[7]) - 
             6 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
              Rational[-109, 16] Zeta[7]) + 
             4 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) - 
             6 (Rational[7, 360] Pi^4 Zeta[3] + 
              Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) + 
           Rational[-3, 2] (Rational[-13, 180] Pi^4 Zeta[3] + 
             Rational[1, 2] 
              Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
             Rational[1, 2] 
              Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
             2 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
               Rational[-291, 16] Zeta[7]) + 
             6 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
               Rational[-109, 16] Zeta[7]) - 
             4 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
             6 (Rational[7, 360] Pi^4 Zeta[3] + 
               Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) - 
           5 (Rational[-17, 360] Pi^4 Zeta[3] + 
            Rational[1, 6] 
             Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
            3 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
              Rational[-291, 16] Zeta[7]) + 
            4 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
            6 (Rational[7, 360] Pi^4 Zeta[3] + Rational[-11, 12] Pi^2 Zeta[5] + 
              Rational[61, 8] Zeta[7])) + 
           Rational[1, 9] (Rational[-151, 180] Pi^4 Zeta[3] + 
             Rational[3, 2] 
              Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
             Rational[17, 6] 
              Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
             54 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
               Rational[-291, 16] Zeta[7]) + 
             18 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
               Rational[-109, 16] Zeta[7]) + 
             72 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) - 
             24 (Rational[-1, 24] Pi^4 Zeta[3] + Rational[5, 12] Pi^2 Zeta[5] + 
              Rational[5, 8] Zeta[7]) + 
             102 (Rational[7, 360] Pi^4 Zeta[3] + 
               Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) + 
           Rational[1, 5] (Rational[-1, 30] Pi^4 Zeta[3] + 
             Rational[2, 3] Pi^2 Zeta[5] - 
             5 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
              18 Zeta[7]) + 
             10 (Rational[1, 45] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] - 
               11 Zeta[7]) - 
             4 (Rational[-1, 45] Pi^4 Zeta[3] + Rational[-2, 3] Pi^2 Zeta[5] + 
              10 Zeta[7])) + 
           Rational[1, 2] (Rational[-1, 90] Pi^4 Zeta[3] + 
             Rational[1, 3] Pi^2 Zeta[5] - 
             3 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
              18 Zeta[7]) - 
             2 (Rational[-1, 45] Pi^4 Zeta[3] + Rational[-2, 3] Pi^2 Zeta[5] + 
              10 Zeta[7])) + 
           Rational[1, 2] (Rational[1, 90] Pi^4 Zeta[3] + 
             Rational[-1, 3] Pi^2 Zeta[5] + 
             3 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
               18 Zeta[7]) + 
             2 (Rational[-1, 45] Pi^4 Zeta[3] + Rational[-2, 3] Pi^2 Zeta[5] + 
               10 Zeta[7])) + 
           Rational[1, 5] (Rational[1, 30] Pi^4 Zeta[3] + 
             Rational[-2, 3] Pi^2 Zeta[5] + 
             5 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
               18 Zeta[7]) - 
             10 (Rational[1, 45] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] - 
              11 Zeta[7]) + 
             4 (Rational[-1, 45] Pi^4 Zeta[3] + Rational[-2, 3] Pi^2 Zeta[5] + 
               10 Zeta[7])) + 
           2 (61 + Rational[-1, 3] Pi^2 + Rational[-4, 45] Pi^4 + 
             Rational[-31, 945] Pi^6 - 2 (1 + Rational[-1, 6] Pi^2) - 
             8 (1 + Rational[-1, 90] Pi^4) - 31 (1 + Rational[-1, 945] Pi^6) - 
             4 (1 - Zeta[3]) - 4 Zeta[3] + Rational[17, 120] Pi^4 Zeta[3] + 
             Rational[-1, 3] 
              Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) - 
             16 (1 - Zeta[5]) - 16 Zeta[5] + Rational[-79, 12] Pi^2 Zeta[5] + 
             Rational[-1, 2] 
              Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
             3 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
              18 Zeta[7]) - 
             15 (Rational[1, 45] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] - 
              11 Zeta[7]) - 
             3 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
              Rational[-109, 16] Zeta[7]) + 63 (1 - Zeta[7]) - 
             7 (Rational[-1, 24] Pi^4 Zeta[3] + Rational[5, 12] Pi^2 Zeta[5] + 
              Rational[5, 8] Zeta[7]) + Rational[107, 2] Zeta[7] - 
             6 (Rational[7, 360] Pi^4 Zeta[3] + 
              Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7]) + 
             Rational[-11, 36] (Rational[151, 180] Pi^4 Zeta[3] + 
               Rational[-3, 2] 
                Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
               Rational[-17, 6] 
                Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
               54 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
                Rational[-291, 16] Zeta[7]) - 
               18 (Rational[-1, 72] Pi^4 Zeta[3] + 
                Rational[5, 6] Pi^2 Zeta[5] + Rational[-109, 16] Zeta[7]) - 
               72 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
               24 (Rational[-1, 24] Pi^4 Zeta[3] + 
                 Rational[5, 12] Pi^2 Zeta[5] + Rational[5, 8] Zeta[7]) - 
               102 (Rational[7, 360] Pi^4 Zeta[3] + 
                Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) + 
             Rational[1, 2] (Rational[37, 180] Pi^4 Zeta[3] + 
               Rational[-1, 2] 
                Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
               Rational[-5, 6] 
                Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
               12 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
                Rational[-291, 16] Zeta[7]) - 
               6 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
                Rational[-109, 16] Zeta[7]) - 
               12 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
               6 (Rational[-1, 24] Pi^4 Zeta[3] + 
                 Rational[5, 12] Pi^2 Zeta[5] + Rational[5, 8] Zeta[7]) - 
               24 (Rational[7, 360] Pi^4 Zeta[3] + 
                Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) - 
             3 (Rational[17, 360] Pi^4 Zeta[3] + 
              Rational[-1, 6] 
               Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) - 
              3 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
               Rational[-291, 16] Zeta[7]) - 
              4 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) - 
              6 (Rational[7, 360] Pi^4 Zeta[3] + 
               Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) + 
             Rational[3, 2] (Rational[-13, 180] Pi^4 Zeta[3] + 
               Rational[1, 2] 
                Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
               Rational[1, 2] 
                Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
               2 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
                 Rational[-291, 16] Zeta[7]) + 
               6 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
                 Rational[-109, 16] Zeta[7]) - 
               4 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) + 
               6 (Rational[7, 360] Pi^4 Zeta[3] + 
                 Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) + 
             Rational[-7, 6] (Rational[-37, 180] Pi^4 Zeta[3] + 
               Rational[1, 2] 
                Pi^2 (Rational[1, 2] Pi^2 Zeta[3] + Rational[-11, 2] Zeta[5]) + 
               Rational[5, 6] 
                Pi^2 (Rational[-1, 3] Pi^2 Zeta[3] + Rational[9, 2] Zeta[5]) + 
               12 (Rational[-1, 60] Pi^4 Zeta[3] + 2 Pi^2 Zeta[5] + 
                 Rational[-291, 16] Zeta[7]) + 
               6 (Rational[-1, 72] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] + 
                 Rational[-109, 16] Zeta[7]) + 
               12 (Rational[1, 360] Pi^4 Zeta[3] + Rational[-1, 4] Zeta[7]) - 
               6 (Rational[-1, 24] Pi^4 Zeta[3] + 
                Rational[5, 12] Pi^2 Zeta[5] + Rational[5, 8] Zeta[7]) + 
               24 (Rational[7, 360] Pi^4 Zeta[3] + 
                 Rational[-11, 12] Pi^2 Zeta[5] + Rational[61, 8] Zeta[7])) + 
             4 (Rational[1, 90] Pi^4 Zeta[3] + Rational[-1, 3] Pi^2 Zeta[5] + 
               3 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
                 18 Zeta[7]) + 
               2 (Rational[-1, 45] Pi^4 Zeta[3] + 
                 Rational[-2, 3] Pi^2 Zeta[5] + 10 Zeta[7])) + 
             Rational[-17, 10] (Rational[1, 30] Pi^4 Zeta[3] + 
               Rational[-2, 3] Pi^2 Zeta[5] + 
               5 (Rational[1, 90] Pi^4 Zeta[3] + Rational[5, 3] Pi^2 Zeta[5] - 
                 18 Zeta[7]) - 
               10 (Rational[1, 45] Pi^4 Zeta[3] + Rational[5, 6] Pi^2 Zeta[5] - 
                11 Zeta[7]) + 
               4 (Rational[-1, 45] Pi^4 Zeta[3] + 
                 Rational[-2, 3] Pi^2 Zeta[5] + 10 Zeta[7]))) + 
           255 (1 - Zeta[9]) - Zeta[9] - 
           130 (Rational[1, 2] (Rational[-2, 945] Pi^6 Zeta[3] + 
              Rational[-1, 45] Pi^4 Zeta[5] + Rational[-1, 3] Pi^2 Zeta[7]) + 
            4 Zeta[9])}, 0, 8, 1],
         Editable->False]}]},
      {
       RowBox[{
        TagBox[
         TagBox[
          RowBox[{
           RowBox[{
            SubscriptBox["\[InvisiblePrefixScriptBase]", "3"], 
            SubscriptBox["F", "2"]}], "\[InvisibleApplication]", 
           RowBox[{"(", 
            RowBox[{
             TagBox[
              TagBox[
               RowBox[{
                TagBox["1",
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"-", "4"}], " ", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"-", "2"}], " ", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox[
              TagBox[
               RowBox[{
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"3", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"2", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox["1",
              HypergeometricPFQ,
              Editable->True,
              Selectable->True]}], ")"}]}],
          InterpretTemplate[HypergeometricPFQ[#, #2, #3]& ],
          Editable->False,
          Selectable->False],
         HypergeometricPFQ], "\[Rule]", 
        InterpretationBox[
         RowBox[{"1", "+", 
          FractionBox[
           RowBox[{"4", " ", 
            SuperscriptBox["\[Pi]", "2"], " ", 
            SuperscriptBox["ep", "2"]}], "3"], "+", 
          RowBox[{"32", " ", 
           TemplateBox[{"3"},
            "Zeta"], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          FractionBox[
           RowBox[{"22", " ", 
            SuperscriptBox["\[Pi]", "4"], " ", 
            SuperscriptBox["ep", "4"]}], "15"], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"40", " ", 
                SuperscriptBox["\[Pi]", "2"], " ", 
                TemplateBox[{"3"},
                 "Zeta"]}], "3"]}], "+", 
             RowBox[{"600", " ", 
              TemplateBox[{"5"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "5"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              RowBox[{"566", " ", 
               SuperscriptBox["\[Pi]", "6"]}], "315"], "-", 
             RowBox[{"160", " ", 
              SuperscriptBox[
               TemplateBox[{"3"},
                "Zeta"], "2"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "6"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"44", " ", 
                SuperscriptBox["\[Pi]", "4"], " ", 
                TemplateBox[{"3"},
                 "Zeta"]}], "3"]}], "-", 
             RowBox[{"200", " ", 
              SuperscriptBox["\[Pi]", "2"], " ", 
              TemplateBox[{"5"},
               "Zeta"]}], "+", 
             RowBox[{"8382", " ", 
              TemplateBox[{"7"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "7"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "8"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, 0, 8, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 0, {
          1, 0, Rational[4, 3] Pi^2, 32 Zeta[3], Rational[22, 15] Pi^4, 
           Rational[-40, 3] Pi^2 Zeta[3] + 600 Zeta[5], 
           Rational[566, 315] Pi^6 - 160 Zeta[3]^2, 
           Rational[-44, 3] Pi^4 Zeta[3] - 200 Pi^2 Zeta[5] + 8382 Zeta[7]}, 
          0, 8, 1],
         Editable->False]}]},
      {
       RowBox[{
        TagBox[
         TagBox[
          RowBox[{
           RowBox[{
            SubscriptBox["\[InvisiblePrefixScriptBase]", "4"], 
            SubscriptBox["F", "3"]}], "\[InvisibleApplication]", 
           RowBox[{"(", 
            RowBox[{
             TagBox[
              TagBox[
               RowBox[{
                TagBox["1",
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"-", "4"}], " ", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{
                  RowBox[{"-", "2"}], " ", "ep"}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox[
              TagBox[
               RowBox[{
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"3", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"2", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True], ",", 
                TagBox[
                 RowBox[{"1", "-", 
                  RowBox[{"2", " ", "ep"}]}],
                 HypergeometricPFQ,
                 Editable->True,
                 Selectable->True]}],
               InterpretTemplate[{
                 SlotSequence[1]}& ]],
              HypergeometricPFQ,
              Editable->False,
              Selectable->False], ";", 
             TagBox["1",
              HypergeometricPFQ,
              Editable->True,
              Selectable->True]}], ")"}]}],
          InterpretTemplate[HypergeometricPFQ[#, #2, #3]& ],
          Editable->False,
          Selectable->False],
         HypergeometricPFQ], "\[Rule]", 
        InterpretationBox[
         RowBox[{"1", "+", 
          FractionBox[
           RowBox[{"4", " ", 
            SuperscriptBox["\[Pi]", "2"], " ", 
            SuperscriptBox["ep", "2"]}], "3"], "+", 
          RowBox[{"48", " ", 
           TemplateBox[{"3"},
            "Zeta"], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          FractionBox[
           RowBox[{"98", " ", 
            SuperscriptBox["\[Pi]", "4"], " ", 
            SuperscriptBox["ep", "4"]}], "45"], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"64", " ", 
                SuperscriptBox["\[Pi]", "2"], " ", 
                TemplateBox[{"3"},
                 "Zeta"]}], "3"]}], "+", 
             RowBox[{"952", " ", 
              TemplateBox[{"5"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "5"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              RowBox[{"1742", " ", 
               SuperscriptBox["\[Pi]", "6"]}], "567"], "-", 
             RowBox[{"352", " ", 
              SuperscriptBox[
               TemplateBox[{"3"},
                "Zeta"], "2"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "6"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"448", " ", 
                SuperscriptBox["\[Pi]", "4"], " ", 
                TemplateBox[{"3"},
                 "Zeta"]}], "15"]}], "-", 
             FractionBox[
              RowBox[{"760", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"5"},
                "Zeta"]}], "3"], "+", 
             RowBox[{"13898", " ", 
              TemplateBox[{"7"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "7"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "8"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, 0, 8, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 0, {
          1, 0, Rational[4, 3] Pi^2, 48 Zeta[3], Rational[98, 45] Pi^4, 
           Rational[-64, 3] Pi^2 Zeta[3] + 952 Zeta[5], 
           Rational[1742, 567] Pi^6 - 352 Zeta[3]^2, 
           Rational[-448, 15] Pi^4 Zeta[3] + Rational[-760, 3] Pi^2 Zeta[5] + 
           13898 Zeta[7]}, 0, 8, 1],
         Editable->False]}]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"commonfactor", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"4", "Pi"}], ")"}], "^", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", "ep"}], "-", "4"}], ")"}]}], 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"-", "Q2"}], ")"}], "^", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", "2"}], "ep"}], ")"}]}], "/", 
     RowBox[{
      RowBox[{"Gamma", "[", 
       RowBox[{"1", "-", "ep"}], "]"}], "^", "2"}]}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"masterseries", "=", 
   RowBox[{"mastervalues", "/.", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"m_", "\[Rule]", "v_"}], ")"}], ":>", 
     RowBox[{"(", 
      RowBox[{"m", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Series", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{"v", "/", "commonfactor"}], ")"}], "/.", "hpfqmap"}], 
           ",", 
           RowBox[{"{", 
            RowBox[{"ep", ",", "0", ",", "3"}], "}"}]}], "]"}], "//", 
         RowBox[{"MapSeries", "[", 
          RowBox[{"FunctionExpand", "/*", "Expand"}], "]"}]}], ")"}]}], 
      ")"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"masterseries", "//", "TableForm"}]}], "Input"],

Cell[BoxData[
 FormBox[
  TagBox[
   TagBox[GridBox[{
      {
       RowBox[{
        SubsuperscriptBox["B", 
         RowBox[{
         "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", ",", "0"}], 
         "1"], "\[Rule]", 
        InterpretationBox[
         RowBox[{
          RowBox[{"-", 
           FractionBox["1", 
            RowBox[{"2", " ", 
             SuperscriptBox["ep", "2"]}]]}], "-", 
          FractionBox["5", 
           RowBox[{"2", " ", "ep"}]], "+", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", 
             FractionBox["19", "2"]}], "-", 
            FractionBox[
             SuperscriptBox["\[Pi]", "2"], "6"]}], ")"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["65", "2"]}], "-", 
             FractionBox[
              RowBox[{"5", " ", 
               SuperscriptBox["\[Pi]", "2"]}], "6"], "+", 
             RowBox[{"4", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", "ep"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["211", "2"]}], "-", 
             FractionBox[
              RowBox[{"19", " ", 
               SuperscriptBox["\[Pi]", "2"]}], "6"], "+", 
             FractionBox[
              SuperscriptBox["\[Pi]", "4"], "30"], "+", 
             RowBox[{"20", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["665", "2"]}], "-", 
             FractionBox[
              RowBox[{"65", " ", 
               SuperscriptBox["\[Pi]", "2"]}], "6"], "+", 
             FractionBox[
              SuperscriptBox["\[Pi]", "4"], "6"], "+", 
             RowBox[{"76", " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "+", 
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "3"], "+", 
             RowBox[{"24", " ", 
              TemplateBox[{"5"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "4"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, -2, 4, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 0, {
           Rational[-1, 2], 
           Rational[-5, 2], Rational[-19, 2] + Rational[-1, 6] Pi^2, 
           Rational[-65, 2] + Rational[-5, 6] Pi^2 + 4 Zeta[3], 
           Rational[-211, 2] + Rational[-19, 6] Pi^2 + Rational[1, 30] Pi^4 + 
           20 Zeta[3], Rational[-665, 2] + Rational[-65, 6] Pi^2 + 
           Rational[1, 6] Pi^4 + 76 Zeta[3] + Rational[4, 3] Pi^2 Zeta[3] + 
           24 Zeta[5]}, -2, 4, 1],
         Editable->False]}]},
      {
       RowBox[{
        SubsuperscriptBox["B", 
         RowBox[{
         "1", ",", "0", ",", "0", ",", "1", ",", "1", ",", "0", ",", "0"}], 
         "2"], "\[Rule]", 
        InterpretationBox[
         RowBox[{
          FractionBox["Q2", 
           RowBox[{"4", " ", "ep"}]], "+", 
          FractionBox[
           RowBox[{"13", " ", "Q2"}], "8"], "+", 
          FractionBox[
           RowBox[{"115", " ", "Q2", " ", "ep"}], "16"], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              RowBox[{"865", " ", "Q2"}], "32"], "-", 
             FractionBox[
              RowBox[{"5", " ", "Q2", " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "2"]}], ")"}], " ", 
           SuperscriptBox["ep", "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"65", " ", 
                TemplateBox[{"3"},
                 "Zeta"], " ", "Q2"}], "4"]}], "-", 
             FractionBox[
              RowBox[{
               SuperscriptBox["\[Pi]", "4"], " ", "Q2"}], "24"], "+", 
             FractionBox[
              RowBox[{"5971", " ", "Q2"}], "64"]}], ")"}], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "4"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, -1, 4, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 0, {
          Rational[1, 4] $CellContext`Q2, Rational[13, 8] $CellContext`Q2, 
           Rational[115, 16] $CellContext`Q2, 
           Rational[865, 32] $CellContext`Q2 + 
           Rational[-5, 2] $CellContext`Q2 Zeta[3], 
           Rational[5971, 64] $CellContext`Q2 + 
           Rational[-1, 24] Pi^4 $CellContext`Q2 + 
           Rational[-65, 4] $CellContext`Q2 Zeta[3]}, -1, 4, 1],
         Editable->False]}]},
      {
       RowBox[{
        SubsuperscriptBox["B", 
         RowBox[{
         "0", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0", ",", "0"}], 
         "5"], "\[Rule]", 
        InterpretationBox[
         RowBox[{
          RowBox[{"-", 
           FractionBox["1", 
            RowBox[{"2", " ", 
             SuperscriptBox["ep", "2"]}]]}], "-", 
          FractionBox["5", 
           RowBox[{"2", " ", "ep"}]], "+", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", 
             FractionBox["19", "2"]}], "-", 
            FractionBox[
             SuperscriptBox["\[Pi]", "2"], "6"]}], ")"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["65", "2"]}], "-", 
             FractionBox[
              RowBox[{"5", " ", 
               SuperscriptBox["\[Pi]", "2"]}], "6"], "+", 
             RowBox[{"4", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", "ep"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["211", "2"]}], "-", 
             FractionBox[
              RowBox[{"19", " ", 
               SuperscriptBox["\[Pi]", "2"]}], "6"], "+", 
             FractionBox[
              SuperscriptBox["\[Pi]", "4"], "30"], "+", 
             RowBox[{"20", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["665", "2"]}], "-", 
             FractionBox[
              RowBox[{"65", " ", 
               SuperscriptBox["\[Pi]", "2"]}], "6"], "+", 
             FractionBox[
              SuperscriptBox["\[Pi]", "4"], "6"], "+", 
             RowBox[{"76", " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "+", 
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], "3"], "+", 
             RowBox[{"24", " ", 
              TemplateBox[{"5"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "4"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, -2, 4, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 0, {
           Rational[-1, 2], 
           Rational[-5, 2], Rational[-19, 2] + Rational[-1, 6] Pi^2, 
           Rational[-65, 2] + Rational[-5, 6] Pi^2 + 4 Zeta[3], 
           Rational[-211, 2] + Rational[-19, 6] Pi^2 + Rational[1, 30] Pi^4 + 
           20 Zeta[3], Rational[-665, 2] + Rational[-65, 6] Pi^2 + 
           Rational[1, 6] Pi^4 + 76 Zeta[3] + Rational[4, 3] Pi^2 Zeta[3] + 
           24 Zeta[5]}, -2, 4, 1],
         Editable->False]}]},
      {
       RowBox[{
        SubsuperscriptBox["B", 
         RowBox[{
         "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "1", ",", "0"}], 
         "6"], "\[Rule]", 
        InterpretationBox[
         RowBox[{
          RowBox[{"-", 
           FractionBox["1", 
            RowBox[{
             SuperscriptBox["Q2", "2"], " ", 
             SuperscriptBox["ep", "4"]}]]}], "+", 
          FractionBox[
           RowBox[{"5", " ", 
            SuperscriptBox["\[Pi]", "2"]}], 
           RowBox[{"6", " ", 
            SuperscriptBox["Q2", "2"], " ", 
            SuperscriptBox["ep", "2"]}]], "+", 
          FractionBox[
           RowBox[{"27", " ", 
            TemplateBox[{"3"},
             "Zeta"]}], 
           RowBox[{
            SuperscriptBox["Q2", "2"], " ", "ep"}]], "+", 
          FractionBox[
           RowBox[{"23", " ", 
            SuperscriptBox["\[Pi]", "4"]}], 
           RowBox[{"36", " ", 
            SuperscriptBox["Q2", "2"]}]], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              RowBox[{"117", " ", 
               TemplateBox[{"5"},
                "Zeta"]}], 
              SuperscriptBox["Q2", "2"]], "-", 
             FractionBox[
              RowBox[{"8", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], 
              SuperscriptBox["Q2", "2"]]}], ")"}], " ", "ep"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             FractionBox[
              RowBox[{"19", " ", 
               SuperscriptBox["\[Pi]", "6"]}], 
              RowBox[{"315", " ", 
               SuperscriptBox["Q2", "2"]}]], "-", 
             FractionBox[
              RowBox[{"267", " ", 
               SuperscriptBox[
                TemplateBox[{"3"},
                 "Zeta"], "2"]}], 
              SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
           SuperscriptBox["ep", "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox[
               RowBox[{"6", " ", 
                TemplateBox[{"7"},
                 "Zeta"]}], 
               SuperscriptBox["Q2", "2"]]}], "-", 
             FractionBox[
              RowBox[{"40", " ", 
               SuperscriptBox["\[Pi]", "2"], " ", 
               TemplateBox[{"5"},
                "Zeta"]}], 
              SuperscriptBox["Q2", "2"]], "-", 
             FractionBox[
              RowBox[{"109", " ", 
               SuperscriptBox["\[Pi]", "4"], " ", 
               TemplateBox[{"3"},
                "Zeta"]}], 
              RowBox[{"10", " ", 
               SuperscriptBox["Q2", "2"]}]]}], ")"}], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "4"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, -4, 4, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 
          0, {(-1) $CellContext`Q2^(-2), 0, Rational[5, 6] 
           Pi^2 $CellContext`Q2^(-2), 27 $CellContext`Q2^(-2) Zeta[3], 
           Rational[23, 36] 
           Pi^4 $CellContext`Q2^(-2), (-8) Pi^2 $CellContext`Q2^(-2) Zeta[3] + 
           117 $CellContext`Q2^(-2) Zeta[5], 
           Rational[19, 315] Pi^6 $CellContext`Q2^(-2) - 
           267 $CellContext`Q2^(-2) Zeta[3]^2, 
           Rational[-109, 10] Pi^4 $CellContext`Q2^(-2) Zeta[3] - 40 
           Pi^2 $CellContext`Q2^(-2) Zeta[5] - 6 $CellContext`Q2^(-2) 
           Zeta[7]}, -4, 4, 1],
         Editable->False]}]},
      {
       RowBox[{
        SubsuperscriptBox["B", 
         RowBox[{
         "1", ",", "1", ",", "0", ",", "1", ",", "1", ",", "0", ",", "0"}], 
         "7"], "\[Rule]", 
        InterpretationBox[
         RowBox[{
          RowBox[{"-", 
           FractionBox["1", 
            SuperscriptBox["ep", "2"]]}], "-", 
          FractionBox["4", "ep"], "-", "12", "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "32"}], "+", 
             RowBox[{"4", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", "ep"}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "80"}], "+", 
             FractionBox[
              SuperscriptBox["\[Pi]", "4"], "15"], "+", 
             RowBox[{"16", " ", 
              TemplateBox[{"3"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "2"]}], "+", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", "192"}], "+", 
             FractionBox[
              RowBox[{"4", " ", 
               SuperscriptBox["\[Pi]", "4"]}], "15"], "+", 
             RowBox[{"48", " ", 
              TemplateBox[{"3"},
               "Zeta"]}], "+", 
             RowBox[{"12", " ", 
              TemplateBox[{"5"},
               "Zeta"]}]}], ")"}], " ", 
           SuperscriptBox["ep", "3"]}], "+", 
          InterpretationBox[
           RowBox[{"O", "(", 
            SuperscriptBox["ep", "4"], ")"}],
           SeriesData[$CellContext`ep, 0, {}, -2, 4, 1],
           Editable->False]}],
         SeriesData[$CellContext`ep, 
          0, {-1, -4, -12, -32 + 4 Zeta[3], -80 + Rational[1, 15] Pi^4 + 
           16 Zeta[3], -192 + Rational[4, 15] Pi^4 + 48 Zeta[3] + 
           12 Zeta[5]}, -2, 4, 1],
         Editable->False]}]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Column],
   Function[BoxForm`e$, 
    TableForm[BoxForm`e$]]], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Final result", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"exprseries", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", "\"\<reduced.ps2.l2.m\>\"", "]"}], "[", 
      "\"\<expression\>\"", "]"}], "/.", 
     RowBox[{"d", "\[Rule]", 
      RowBox[{"4", "-", 
       RowBox[{"2", "ep"}]}]}]}], "/.", "masterseries"}], "//", 
   RowBox[{"MapSeries", "[", "Factor", "]"}]}]}]], "Input"],

Cell[BoxData[
 FormBox[
  InterpretationBox[
   RowBox[{
    FractionBox[
     RowBox[{"4", " ", 
      SuperscriptBox["gs", "4"], " ", 
      RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      SubscriptBox["C", "A"], " ", 
      SubsuperscriptBox["C", "F", "2"]}], 
     SuperscriptBox["ep", "4"]], "-", 
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "4"], " ", 
      RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"11", " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"36", " ", 
         SubscriptBox["C", "F"]}], "-", 
        RowBox[{"4", " ", "Ngen"}]}], ")"}]}], 
     RowBox[{"3", " ", 
      SuperscriptBox["ep", "3"]}]], "+", 
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "4"], " ", 
      RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"166", " ", 
         SubscriptBox["C", "A"]}], "+", 
        RowBox[{"369", " ", 
         SubscriptBox["C", "F"]}], "+", 
        RowBox[{"56", " ", "Ngen"}]}], ")"}]}], 
     RowBox[{"9", " ", 
      SuperscriptBox["ep", "2"]}]], "+", 
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "4"], " ", 
      RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"1404", " ", 
         TemplateBox[{"3"},
          "Zeta"], " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"66", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"4129", " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"2160", " ", 
         TemplateBox[{"3"},
          "Zeta"], " ", 
         SubscriptBox["C", "F"]}], "+", 
        RowBox[{"5967", " ", 
         SubscriptBox["C", "F"]}], "+", 
        RowBox[{"108", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", 
         SubscriptBox["C", "F"]}], "+", 
        RowBox[{"1412", " ", "Ngen"}], "+", 
        RowBox[{"24", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", "Ngen"}]}], ")"}]}], 
     RowBox[{"54", " ", "ep"}]], "+", 
    FractionBox[
     RowBox[{
      SuperscriptBox["gs", "4"], " ", 
      RowBox[{"PS", "(", "2", ")"}], " ", "Q2", " ", 
      SuperscriptBox["qu", "2"], " ", "x", " ", 
      SubscriptBox["C", "A"], " ", 
      SubscriptBox["C", "F"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"164160", " ", 
         TemplateBox[{"3"},
          "Zeta"], " ", 
         SubscriptBox["C", "A"]}], "+", 
        RowBox[{"882", " ", 
         SuperscriptBox["\[Pi]", "4"], " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"12120", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"445865", " ", 
         SubscriptBox["C", "A"]}], "-", 
        RowBox[{"174960", " ", 
         TemplateBox[{"3"},
          "Zeta"], " ", 
         SubscriptBox["C", "F"]}], "+", 
        RowBox[{"466155", " ", 
         SubscriptBox["C", "F"]}], "-", 
        RowBox[{"1224", " ", 
         SuperscriptBox["\[Pi]", "4"], " ", 
         SubscriptBox["C", "F"]}], "+", 
        RowBox[{"15660", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", 
         SubscriptBox["C", "F"]}], "-", 
        RowBox[{"17280", " ", "Ngen", " ", 
         TemplateBox[{"3"},
          "Zeta"]}], "+", 
        RowBox[{"150820", " ", "Ngen"}], "+", 
        RowBox[{"3360", " ", 
         SuperscriptBox["\[Pi]", "2"], " ", "Ngen"}]}], ")"}]}], "1620"], "+", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "1"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, -4, 1, 1],
     Editable->False]}],
   SeriesData[$CellContext`ep, 0, {
    4 FeynCalc`CA 
     FeynCalc`CF^2 $CellContext`gs^4 $CellContext`Q2 $CellContext`qu^2 \
$CellContext`x $CellContext`PS[2], Rational[-1, 3] FeynCalc`CA 
     FeynCalc`CF $CellContext`gs^4 (11 FeynCalc`CA - 36 FeynCalc`CF - 
      4 $CellContext`Ngen) $CellContext`Q2 $CellContext`qu^2 $CellContext`x \
$CellContext`PS[2], Rational[1, 9] FeynCalc`CA 
     FeynCalc`CF $CellContext`gs^4 ((-166) FeynCalc`CA + 369 FeynCalc`CF + 
      56 $CellContext`Ngen + 
      3 FeynCalc`CA 
       Pi^2) $CellContext`Q2 $CellContext`qu^2 $CellContext`x $CellContext`PS[
      2], Rational[1, 54] FeynCalc`CA 
     FeynCalc`CF $CellContext`gs^4 $CellContext`Q2 $CellContext`qu^2 \
$CellContext`x $CellContext`PS[2] ((-4129) FeynCalc`CA + 5967 FeynCalc`CF + 
      1412 $CellContext`Ngen - 66 FeynCalc`CA Pi^2 + 108 FeynCalc`CF Pi^2 + 
      24 $CellContext`Ngen Pi^2 + 1404 FeynCalc`CA Zeta[3] - 2160 FeynCalc`CF 
      Zeta[3]), Rational[1, 1620] FeynCalc`CA 
     FeynCalc`CF $CellContext`gs^4 $CellContext`Q2 $CellContext`qu^2 \
$CellContext`x $CellContext`PS[2] ((-445865) FeynCalc`CA + 466155 FeynCalc`CF + 
      150820 $CellContext`Ngen - 12120 FeynCalc`CA Pi^2 + 
      15660 FeynCalc`CF Pi^2 + 3360 $CellContext`Ngen Pi^2 + 
      882 FeynCalc`CA Pi^4 - 1224 FeynCalc`CF Pi^4 + 
      164160 FeynCalc`CA Zeta[3] - 174960 FeynCalc`CF Zeta[3] - 
      17280 $CellContext`Ngen Zeta[3])}, -4, 1, 1],
   Editable->False], TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"exprseries", ",", "\"\<result.ps2.l2.m\>\""}], "]"}], 
  ";"}]], "Input"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
11. Differential equations for PS[k1,k2,k3] L[l1], from LiteRed\
\>", "Section"],

Cell["\<\
Manual step: open and run the ps3l1-differential-equations.nb notebook.\
\>", "Subsection"],

Cell[CellGroupData[{

Cell["Load the resulting differential equation", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"masters", "=", 
   RowBox[{"Get", "[", "\"\<ps3l1.masters\>\"", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"My", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Get", "[", "\"\<ps3l1.ms13\>\"", "]"}], "/.", 
     RowBox[{"s13", "\[Rule]", "y"}]}], "/.", 
    RowBox[{"s23", "\[Rule]", "z"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Mz", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Get", "[", "\"\<ps3l1.ms23\>\"", "]"}], "/.", 
     RowBox[{"s13", "\[Rule]", "y"}]}], "/.", 
    RowBox[{"s23", "\[Rule]", "z"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"T", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Get", "[", "\"\<ps3l1.ep.t\>\"", "]"}], "/.", 
     RowBox[{"s13", "\[Rule]", "y"}]}], "/.", 
    RowBox[{"s23", "\[Rule]", "z"}]}]}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Sy", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Inverse", "[", "T", "]"}], ".", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"My", ".", "T"}], "-", 
       RowBox[{"D", "[", 
        RowBox[{"T", ",", "y"}], "]"}]}], ")"}]}], "/", "ep"}], "//", 
   "Factor"}]}], "\[IndentingNewLine]", 
 RowBox[{"Sz", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Inverse", "[", "T", "]"}], ".", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"Mz", ".", "T"}], "-", 
       RowBox[{"D", "[", 
        RowBox[{"T", ",", "z"}], "]"}]}], ")"}]}], "/", "ep"}], "//", 
   "Factor"}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{"y", "+", "z", "-", "1"}]]}], "0", "0", "0", "0"},
     {"0", 
      FractionBox["1", 
       RowBox[{"y", "+", "z", "-", "1"}]], "0", "0", "0", "0"},
     {
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "y"}]]}], "0", 
      FractionBox["z", 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", "y", " ", 
        RowBox[{"(", 
         RowBox[{"y", "+", "z"}], ")"}]}]], 
      FractionBox[
       RowBox[{"z", "-", "1"}], 
       RowBox[{"y", " ", 
        RowBox[{"(", 
         RowBox[{"y", "+", "z", "-", "1"}], ")"}]}]], "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["1", "y"]}], 
      RowBox[{"-", 
       FractionBox["1", "y"]}], "0", 
      RowBox[{"-", 
       FractionBox["1", "y"]}], "0"},
     {"0", "0", 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{"y", "+", "z"}], ")"}]}]]}], "0", 
      FractionBox["1", 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"y", "-", "1"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"2", " ", "y"}], "+", "z", "-", "1"}], 
        RowBox[{"y", " ", 
         RowBox[{"(", 
          RowBox[{"y", "+", "z", "-", "1"}], ")"}]}]]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", 
       FractionBox["1", "z"]}], "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{"y", "+", "z", "-", "1"}]]}], "0", "0", "0", "0"},
     {"0", 
      FractionBox["1", 
       RowBox[{"y", "+", "z", "-", "1"}]], "0", "0", "0", "0"},
     {
      FractionBox["1", 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", 
        RowBox[{"(", 
         RowBox[{"z", "-", "1"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{"z", "-", "1"}], ")"}]}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"y", "+", 
         RowBox[{"2", " ", "z"}], "-", "1"}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{"z", "-", "1"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"y", "+", "z"}], ")"}]}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"y", "+", 
         RowBox[{"2", " ", "z"}], "-", "1"}], 
        RowBox[{"z", " ", 
         RowBox[{"(", 
          RowBox[{"y", "+", "z", "-", "1"}], ")"}]}]]}], "0", "0"},
     {"0", "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "z"}]]}], 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{"y", "+", "z"}], ")"}]}]]}], "0", 
      RowBox[{"-", 
       FractionBox["1", 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "z"}]]}], 
      FractionBox[
       RowBox[{"y", "-", "1"}], 
       RowBox[{"z", " ", 
        RowBox[{"(", 
         RowBox[{"y", "+", "z", "-", "1"}], ")"}]}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Epsilon Form solution construction", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
   RowBox[{"S_", ",", "x_", ",", "0"}], "]"}], ":=", 
  RowBox[{"IdentityMatrix", "[", 
   RowBox[{"Length", "[", "S", "]"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
   RowBox[{"S_", ",", "x_", ",", "order_Integer"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "Sa", ",", "Sn", ",", "Sym", ",", "Sym0", ",", "G", ",", "G0", ",", "a", 
      ",", "b"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Sa", "=", 
      RowBox[{
       RowBox[{
        RowBox[{"Apart", "[", 
         RowBox[{"S", ",", "x"}], "]"}], "/.", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"num_.", "/", "den_"}], ")"}], "/;", 
           RowBox[{"Not", "[", 
            RowBox[{"FreeQ", "[", 
             RowBox[{"den", ",", "x"}], "]"}], "]"}]}], ")"}], 
         "\[RuleDelayed]", 
         RowBox[{"(", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"{", 
             RowBox[{"a", ",", "b"}], "}"}], "=", 
            RowBox[{"CoefficientList", "[", 
             RowBox[{"den", ",", "x"}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"num", "/", "b"}], " ", 
            RowBox[{"Sym", "[", 
             RowBox[{
              RowBox[{"-", "a"}], "/", "b"}], "]"}]}]}], 
          "\[IndentingNewLine]", ")"}]}]}], "//", 
       RowBox[{
        RowBox[{"Collect", "[", 
         RowBox[{"#", ",", "_Sym", ",", "Factor"}], "]"}], "&"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"Not", "[", 
        RowBox[{"FreeQ", "[", 
         RowBox[{"Sa", ",", "x"}], "]"}], "]"}], ",", 
       RowBox[{"Throw", "[", "$FailedToMapSymbols", "]"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"G", "/:", 
      RowBox[{
       RowBox[{"Sym", "[", "a_", "]"}], "*", 
       RowBox[{"G", "[", "b___", "]"}]}], ":=", 
      RowBox[{"G", "[", 
       RowBox[{"a", ",", "b"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Sn", "=", 
      RowBox[{"Sa", "/.", 
       RowBox[{"Sym", "\[Rule]", "G"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Do", "[", 
      RowBox[{
       RowBox[{"Sn", "=", 
        RowBox[{
         RowBox[{
          RowBox[{"Sa", ".", "Sn"}], "//", "Expand"}], "//", 
         RowBox[{"ReplaceAll", "[", 
          RowBox[{"Sym", "\[Rule]", "G"}], "]"}]}]}], ",", 
       RowBox[{"order", "-", "1"}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"Sn", "/.", 
      RowBox[{
       RowBox[{"G", "[", "a___", "]"}], "\[RuleDelayed]", 
       RowBox[{"Hlog", "[", 
        RowBox[{"x", ",", 
         RowBox[{"{", "a", "}"}]}], "]"}]}]}]}]}], "\[IndentingNewLine]", 
   "]"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
  RowBox[{"Sy", ",", "y", ",", "0"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
  RowBox[{"Sy", ",", "y", ",", "1"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
  RowBox[{"Sy", ",", "y", ",", "2"}], "]"}]}], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"1", "0", "0", "0", "0", "0"},
     {"0", "1", "0", "0", "0", "0"},
     {"0", "0", "1", "0", "0", "0"},
     {"0", "0", "0", "1", "0", "0"},
     {"0", "0", "0", "0", "1", "0"},
     {"0", "0", "0", "0", "0", "1"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       RowBox[{"G", "(", 
        RowBox[{"1", "-", "z"}], ";", "y", ")"}]}], "0", "0", "0", "0"},
     {"0", 
      RowBox[{"G", "(", 
       RowBox[{"1", "-", "z"}], ";", "y", ")"}], "0", "0", "0", "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{"G", "(", "0", ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], "0", 
      RowBox[{
       FractionBox[
        RowBox[{"G", "(", "0", ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"-", "z"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], 
      RowBox[{
       RowBox[{"G", "(", "0", ";", "y", ")"}], "-", 
       RowBox[{"G", "(", 
        RowBox[{"1", "-", "z"}], ";", "y", ")"}]}], "0", "0"},
     {"0", 
      RowBox[{"-", 
       RowBox[{"G", "(", "0", ";", "y", ")"}]}], 
      RowBox[{"-", 
       RowBox[{"G", "(", "0", ";", "y", ")"}]}], "0", 
      RowBox[{"-", 
       RowBox[{"G", "(", "0", ";", "y", ")"}]}], "0"},
     {"0", "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"-", "z"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], "0", 
      FractionBox[
       RowBox[{"G", "(", "1", ";", "y", ")"}], 
       SuperscriptBox["Q2", "2"]], 
      RowBox[{
       RowBox[{"-", 
        RowBox[{"G", "(", "0", ";", "y", ")"}]}], "-", 
       RowBox[{"G", "(", 
        RowBox[{"1", "-", "z"}], ";", "y", ")"}]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"G", "(", 
       RowBox[{
        RowBox[{"1", "-", "z"}], ",", 
        RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       RowBox[{"G", "(", 
        RowBox[{
         RowBox[{"1", "-", "z"}], ",", 
         RowBox[{"1", "-", "z"}]}], ";", "y", ")"}]}], "0", "0", "0", "0"},
     {
      RowBox[{
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"0", ",", "0"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], 
      RowBox[{
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"0", ",", 
          RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"-", "z"}], ",", 
          RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], 
      RowBox[{
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"0", ",", "0"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"0", ",", 
          RowBox[{"-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "+", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"1", "-", "z"}], ",", 
          RowBox[{"-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], 
      RowBox[{
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0"}], ";", "y", ")"}], "-", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", 
         RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], "-", 
       RowBox[{"G", "(", 
        RowBox[{
         RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], "+", 
       RowBox[{"G", "(", 
        RowBox[{
         RowBox[{"1", "-", "z"}], ",", 
         RowBox[{"1", "-", "z"}]}], ";", "y", ")"}]}], "0", "0"},
     {"0", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0"}], ";", "y", ")"}], 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0"}], ";", "y", ")"}], "0", 
      RowBox[{"G", "(", 
       RowBox[{"0", ",", "0"}], ";", "y", ")"}], "0"},
     {"0", 
      RowBox[{
       RowBox[{"-", 
        FractionBox[
         RowBox[{"G", "(", 
          RowBox[{"1", ",", "0"}], ";", "y", ")"}], 
         SuperscriptBox["Q2", "2"]]}], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"-", "z"}], ",", 
          RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], 
      RowBox[{
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"0", ",", 
          RowBox[{"-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"1", ",", "0"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "+", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"1", "-", "z"}], ",", 
          RowBox[{"-", "z"}]}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], "0", 
      RowBox[{
       RowBox[{"-", 
        FractionBox[
         RowBox[{"G", "(", 
          RowBox[{"0", ",", "1"}], ";", "y", ")"}], 
         SuperscriptBox["Q2", "2"]]}], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{"1", ",", "0"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]], "-", 
       FractionBox[
        RowBox[{"G", "(", 
         RowBox[{
          RowBox[{"1", "-", "z"}], ",", "1"}], ";", "y", ")"}], 
        SuperscriptBox["Q2", "2"]]}], 
      RowBox[{
       RowBox[{"G", "(", 
        RowBox[{"0", ",", "0"}], ";", "y", ")"}], "+", 
       RowBox[{"G", "(", 
        RowBox[{"0", ",", 
         RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], "+", 
       RowBox[{"G", "(", 
        RowBox[{
         RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], "+", 
       RowBox[{"G", "(", 
        RowBox[{
         RowBox[{"1", "-", "z"}], ",", 
         RowBox[{"1", "-", "z"}]}], ";", "y", ")"}]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
   RowBox[{"S_", ",", "x_", ",", "x0_", ",", "0"}], "]"}], ":=", 
  RowBox[{"IdentityMatrix", "[", 
   RowBox[{"Length", "[", "S", "]"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
   RowBox[{"S_", ",", "x_", ",", "x0_", ",", "order_Integer"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "Sa", ",", "Sn", ",", "Sym", ",", "Sym0", ",", "G", ",", "a", ",", "b"}],
      "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Sa", "=", 
      RowBox[{
       RowBox[{
        RowBox[{"Apart", "[", 
         RowBox[{"S", ",", "x"}], "]"}], "/.", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{"num_.", "/", "den_"}], ")"}], "/;", 
           RowBox[{"Not", "[", 
            RowBox[{"FreeQ", "[", 
             RowBox[{"den", ",", "x"}], "]"}], "]"}]}], ")"}], 
         "\[RuleDelayed]", 
         RowBox[{"(", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"{", 
             RowBox[{"a", ",", "b"}], "}"}], "=", 
            RowBox[{"CoefficientList", "[", 
             RowBox[{"den", ",", "x"}], "]"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
             RowBox[{"num", "/", "b"}], " ", 
             RowBox[{"Sym", "[", 
              RowBox[{
               RowBox[{"-", "a"}], "/", "b"}], "]"}]}], "-", 
            RowBox[{
             RowBox[{"num", "/", "b"}], " ", 
             RowBox[{"Sym0", "[", 
              RowBox[{
               RowBox[{"-", "a"}], "/", "b"}], "]"}]}]}]}], 
          "\[IndentingNewLine]", ")"}]}]}], "//", 
       RowBox[{
        RowBox[{"Collect", "[", 
         RowBox[{"#", ",", "_Sym", ",", "Factor"}], "]"}], "&"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"Not", "[", 
        RowBox[{"FreeQ", "[", 
         RowBox[{"Sa", ",", "x"}], "]"}], "]"}], ",", 
       RowBox[{"Throw", "[", "$FailedToMapSymbols", "]"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"G", "/:", 
      RowBox[{
       RowBox[{"Sym", "[", "a_", "]"}], "*", 
       RowBox[{"G", "[", "b___", "]"}]}], ":=", 
      RowBox[{"G", "[", 
       RowBox[{"a", ",", "b"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"G", "/:", 
      RowBox[{
       RowBox[{"Sym0", "[", "a_", "]"}], "*", 
       RowBox[{"G", "[", "b___", "]"}]}], ":=", 
      RowBox[{"Hlog", "[", 
       RowBox[{"x0", ",", 
        RowBox[{"{", 
         RowBox[{"a", ",", "b"}], "}"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"Sn", "=", 
      RowBox[{"Sa", "//", 
       RowBox[{"ReplaceAll", "[", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"Sym", "\[Rule]", "G"}], ",", 
          RowBox[{
           RowBox[{"Sym0", "[", "a_", "]"}], "\[RuleDelayed]", 
           RowBox[{"Hlog", "[", 
            RowBox[{"x0", ",", 
             RowBox[{"{", "a", "}"}]}], "]"}]}]}], "}"}], "]"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"Do", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"Sn", "=", 
         RowBox[{
          RowBox[{
           RowBox[{"Sa", ".", "Sn"}], "//", "Expand"}], "//", 
          RowBox[{"ReplaceAll", "[", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"Sym", "\[Rule]", "G"}], ",", 
             RowBox[{
              RowBox[{"Sym0", "[", "a_", "]"}], "\[RuleDelayed]", 
              RowBox[{"Hlog", "[", 
               RowBox[{"x0", ",", 
                RowBox[{"{", "a", "}"}]}], "]"}]}]}], "}"}], "]"}]}]}], ";"}],
        "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"i", ",", 
         RowBox[{"order", "-", "1"}]}], "}"}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"Sn", "/.", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"G", "[", "a___", "]"}], "\[RuleDelayed]", 
        RowBox[{"Hlog", "[", 
         RowBox[{"x", ",", 
          RowBox[{"{", "a", "}"}]}], "]"}]}], "}"}]}]}]}], 
   "\[IndentingNewLine]", "]"}]}]}], "Input"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"SeriesOfMatrices", "[", 
   RowBox[{"mxlist_List", ",", "var_"}], "]"}], ":=", 
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{"SeriesData", "[", 
     RowBox[{"var", ",", "0", ",", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"mxlist", "[", 
         RowBox[{"[", 
          RowBox[{"k", ",", "i", ",", "j"}], "]"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"k", ",", 
          RowBox[{"Length", "[", "mxlist", "]"}]}], "}"}]}], "]"}], ",", "0", 
      ",", 
      RowBox[{"Length", "[", "mxlist", "]"}], ",", "1"}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", 
      RowBox[{"Length", "[", 
       RowBox[{"mxlist", "[", 
        RowBox[{"[", "1", "]"}], "]"}], "]"}]}], "}"}], ",", 
    RowBox[{"{", 
     RowBox[{"j", ",", 
      RowBox[{"Length", "[", 
       RowBox[{"mxlist", "[", 
        RowBox[{"[", 
         RowBox[{"1", ",", "1"}], "]"}], "]"}], "]"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"EpsilonFormFundamentalSolution", "[", 
    RowBox[{"S_", ",", "x_", ",", "maxord_Integer"}], "]"}], ":=", 
   RowBox[{"SeriesOfMatrices", "[", 
    RowBox[{
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
        RowBox[{"S", ",", "x", ",", "n"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"n", ",", "0", ",", "maxord"}], "}"}]}], "]"}], ",", "ep"}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"EpsilonFormFundamentalSolution", "[", 
   RowBox[{"S_", ",", "x_", ",", "x0_", ",", "maxord_Integer"}], "]"}], ":=", 
  
  RowBox[{"SeriesOfMatrices", "[", 
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"EpsilonFormFundamentalSolutionOrd", "[", 
       RowBox[{"S", ",", "x", ",", "x0", ",", "n"}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"n", ",", "0", ",", "maxord"}], "}"}]}], "]"}], ",", "ep"}], 
   "]"}]}], "\[IndentingNewLine]"}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"EpsilonFormFundamentalSolution", "[", 
  RowBox[{"Sy", ",", "y", ",", "2"}], "]"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      InterpretationBox[
       RowBox[{"1", "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 0, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {1}, 0, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False]},
     {
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"1", "-", 
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{"1", "-", "z"}], ";", "y", ")"}], " ", "ep"}], "+", 
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{
           RowBox[{"1", "-", "z"}], ",", 
           RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 0, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {
        1, -$CellContext`Hlog[$CellContext`y, {1 - $CellContext`z}], 
         $CellContext`Hlog[$CellContext`y, {
          1 - $CellContext`z, 1 - $CellContext`z}]}, 0, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False]},
     {
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{"1", "-", "z"}], ";", "y", ")"}], " ", "ep"}], "-", 
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{
           RowBox[{"1", "-", "z"}], ",", 
           RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {
         $CellContext`Hlog[$CellContext`y, {
          1 - $CellContext`z}], -$CellContext`Hlog[$CellContext`y, {
           1 - $CellContext`z, 1 - $CellContext`z}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"1", "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 0, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {1}, 0, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False]},
     {
      InterpretationBox[
       RowBox[{
        RowBox[{"-", 
         FractionBox[
          RowBox[{
           RowBox[{"G", "(", "0", ";", "y", ")"}], " ", "ep"}], 
          SuperscriptBox["Q2", "2"]]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"0", ",", "0"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {-$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
           0}], -$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {0, 
            0}] + $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
            1 - $CellContext`z, 0}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"0", ",", 
              RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"-", "z"}], ",", 
              RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 2, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
            0, 1 - $CellContext`z}] - $CellContext`Q2^(-2) \
$CellContext`Hlog[$CellContext`y, {-$CellContext`z, 1 - $CellContext`z}]}, 2, 
        3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           FractionBox[
            RowBox[{"G", "(", "0", ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"-", "z"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", "ep"}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"0", ",", "0"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"0", ",", 
              RowBox[{"-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "+", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"1", "-", "z"}], ",", 
              RowBox[{"-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
            0}] - $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, \
{-$CellContext`z}], $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {0,
             0}] - $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
           0, -$CellContext`z}] - $CellContext`Q2^(-2) \
$CellContext`Hlog[$CellContext`y, {
           1 - $CellContext`z, 
            0}] + $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
            1 - $CellContext`z, -$CellContext`z}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"1", "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"G", "(", "0", ";", "y", ")"}], "-", 
           RowBox[{"G", "(", 
            RowBox[{"1", "-", "z"}], ";", "y", ")"}]}], ")"}], " ", "ep"}], 
        "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"G", "(", 
            RowBox[{"0", ",", "0"}], ";", "y", ")"}], "-", 
           RowBox[{"G", "(", 
            RowBox[{"0", ",", 
             RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], "-", 
           RowBox[{"G", "(", 
            RowBox[{
             RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], "+", 
           RowBox[{"G", "(", 
            RowBox[{
             RowBox[{"1", "-", "z"}], ",", 
             RowBox[{"1", "-", "z"}]}], ";", "y", ")"}]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 0, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {
        1, $CellContext`Hlog[$CellContext`y, {
           0}] - $CellContext`Hlog[$CellContext`y, {
          1 - $CellContext`z}], $CellContext`Hlog[$CellContext`y, {0, 
           0}] - $CellContext`Hlog[$CellContext`y, {
          0, 1 - $CellContext`z}] - $CellContext`Hlog[$CellContext`y, {
          1 - $CellContext`z, 0}] + $CellContext`Hlog[$CellContext`y, {
           1 - $CellContext`z, 1 - $CellContext`z}]}, 0, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False]},
     {
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          RowBox[{"G", "(", "0", ";", "y", ")"}], " ", "ep"}]}], "+", 
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{"0", ",", "0"}], ";", "y", ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {-$CellContext`Hlog[$CellContext`y, {0}], 
         $CellContext`Hlog[$CellContext`y, {0, 0}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          RowBox[{"G", "(", "0", ";", "y", ")"}], " ", "ep"}]}], "+", 
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{"0", ",", "0"}], ";", "y", ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {-$CellContext`Hlog[$CellContext`y, {0}], 
         $CellContext`Hlog[$CellContext`y, {0, 0}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"1", "-", 
        RowBox[{
         RowBox[{"G", "(", "0", ";", "y", ")"}], " ", "ep"}], "+", 
        RowBox[{
         RowBox[{"G", "(", 
          RowBox[{"0", ",", "0"}], ";", "y", ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 0, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {
        1, -$CellContext`Hlog[$CellContext`y, {0}], 
         $CellContext`Hlog[$CellContext`y, {0, 0}]}, 0, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False]},
     {
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{"G", "(", 
              RowBox[{"1", ",", "0"}], ";", "y", ")"}], 
             SuperscriptBox["Q2", "2"]]}], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"-", "z"}], ",", 
              RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 2, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {-$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {1, 
            0}] - $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, \
{-$CellContext`z, 1 - $CellContext`z}]}, 2, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        RowBox[{"-", 
         FractionBox[
          RowBox[{
           RowBox[{"G", "(", 
            RowBox[{"-", "z"}], ";", "y", ")"}], " ", "ep"}], 
          SuperscriptBox["Q2", "2"]]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"0", ",", 
              RowBox[{"-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"1", ",", "0"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "+", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"1", "-", "z"}], ",", 
              RowBox[{"-", "z"}]}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {-$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, \
{-$CellContext`z}], $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
            0, -$CellContext`z}] - $CellContext`Q2^(-2) \
$CellContext`Hlog[$CellContext`y, {1, 
           0}] + $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
            1 - $CellContext`z, -$CellContext`z}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"O", "(", 
        SuperscriptBox["ep", "3"], ")"}],
       SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{
        FractionBox[
         RowBox[{
          RowBox[{"G", "(", "1", ";", "y", ")"}], " ", "ep"}], 
         SuperscriptBox["Q2", "2"]], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            FractionBox[
             RowBox[{"G", "(", 
              RowBox[{"0", ",", "1"}], ";", "y", ")"}], 
             SuperscriptBox["Q2", "2"]]}], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{"1", ",", "0"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]], "-", 
           FractionBox[
            RowBox[{"G", "(", 
             RowBox[{
              RowBox[{"1", "-", "z"}], ",", "1"}], ";", "y", ")"}], 
            SuperscriptBox["Q2", "2"]]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 1, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 
        0, {$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
           1}], -$CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {0, 
            1}] - $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {1, 
           0}] - $CellContext`Q2^(-2) $CellContext`Hlog[$CellContext`y, {
           1 - $CellContext`z, 1}]}, 1, 3, 1],
       Editable->False], 
      InterpretationBox[
       RowBox[{"1", "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"G", "(", "0", ";", "y", ")"}]}], "-", 
           RowBox[{"G", "(", 
            RowBox[{"1", "-", "z"}], ";", "y", ")"}]}], ")"}], " ", "ep"}], 
        "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"G", "(", 
            RowBox[{"0", ",", "0"}], ";", "y", ")"}], "+", 
           RowBox[{"G", "(", 
            RowBox[{"0", ",", 
             RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], "+", 
           RowBox[{"G", "(", 
            RowBox[{
             RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], "+", 
           RowBox[{"G", "(", 
            RowBox[{
             RowBox[{"1", "-", "z"}], ",", 
             RowBox[{"1", "-", "z"}]}], ";", "y", ")"}]}], ")"}], " ", 
         SuperscriptBox["ep", "2"]}], "+", 
        InterpretationBox[
         RowBox[{"O", "(", 
          SuperscriptBox["ep", "3"], ")"}],
         SeriesData[$CellContext`ep, 0, {}, 0, 3, 1],
         Editable->False]}],
       SeriesData[$CellContext`ep, 0, {
        1, -$CellContext`Hlog[$CellContext`y, {
            0}] - $CellContext`Hlog[$CellContext`y, {
          1 - $CellContext`z}], $CellContext`Hlog[$CellContext`y, {0, 
           0}] + $CellContext`Hlog[$CellContext`y, {
           0, 1 - $CellContext`z}] + $CellContext`Hlog[$CellContext`y, {
           1 - $CellContext`z, 0}] + $CellContext`Hlog[$CellContext`y, {
           1 - $CellContext`z, 1 - $CellContext`z}]}, 0, 3, 1],
       Editable->False]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"EpsilonFormSolution", "[", 
   RowBox[{"S_", ",", "x_", ",", "x0_", ",", "maxord_Integer"}], "]"}], ":=", 
  
  RowBox[{
   RowBox[{"EpsilonFormFundamentalSolution", "[", 
    RowBox[{"S", ",", "x", ",", "x0", ",", "maxord"}], "]"}], ".", 
   RowBox[{"ConstantSeriesVector", "[", 
    RowBox[{"JC", ",", 
     RowBox[{"Length", "[", "S", "]"}], ",", "maxord"}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"EpsilonFormSolution", "[", 
   RowBox[{"S_", ",", "x_", ",", "maxord_Integer"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"EpsilonFormFundamentalSolution", "[", 
    RowBox[{"S", ",", "x", ",", "maxord"}], "]"}], ".", 
   RowBox[{"ConstantSeriesVector", "[", 
    RowBox[{"JC", ",", 
     RowBox[{"Length", "[", "S", "]"}], ",", "maxord"}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ConstantSeriesVector", "[", 
   RowBox[{"C_", ",", "n_", ",", "maxord_"}], "]"}], ":=", 
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{"SeriesData", "[", 
     RowBox[{"ep", ",", "0", ",", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"C", "[", 
         RowBox[{"i", ",", "o"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"o", ",", "0", ",", "maxord"}], "}"}]}], "]"}], ",", "0", 
      ",", 
      RowBox[{"maxord", "+", "1"}], ",", "1"}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "n"}], "}"}]}], "]"}]}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Solutions", "Subsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"solutionyz", "[", "nords_", "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"EpsilonFormFundamentalSolution", "[", 
      RowBox[{"Sy", ",", "y", ",", "y0", ",", 
       RowBox[{"nords", "-", "1"}]}], "]"}], ".", 
     RowBox[{"EpsilonFormFundamentalSolution", "[", 
      RowBox[{
       RowBox[{"Sz", "/.", 
        RowBox[{"y", "\[Rule]", "y0"}]}], ",", "z", ",", "z0", ",", 
       RowBox[{"nords", "-", "1"}]}], "]"}], ".", 
     RowBox[{"ConstantSeriesVector", "[", 
      RowBox[{"C", ",", 
       RowBox[{"Length", "[", "Sy", "]"}], ",", 
       RowBox[{"nords", "-", "1"}]}], "]"}]}]}], "\[IndentingNewLine]", 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"solutionzy", "[", "nords_", "]"}], ":=", 
  RowBox[{"With", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"EpsilonFormFundamentalSolution", "[", 
      RowBox[{"Sz", ",", "z", ",", "z0", ",", 
       RowBox[{"nords", "-", "1"}]}], "]"}], ".", 
     RowBox[{"EpsilonFormFundamentalSolution", "[", 
      RowBox[{
       RowBox[{"Sy", "/.", 
        RowBox[{"z", "\[Rule]", "z0"}]}], ",", "y", ",", "y0", ",", 
       RowBox[{"nords", "-", "1"}]}], "]"}], ".", 
     RowBox[{"ConstantSeriesVector", "[", 
      RowBox[{"C", ",", 
       RowBox[{"Length", "[", "Sy", "]"}], ",", 
       RowBox[{"nords", "-", "1"}]}], "]"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"solutionyz", "[", "3", "]"}], "//", 
   RowBox[{
    RowBox[{"Cases", "[", 
     RowBox[{"#", ",", "_Hlog", ",", 
      RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{"G", "(", "0", ";", "y", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "z"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "z"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "z"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", "1"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", 
      RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", 
      RowBox[{"-", "z"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "z"}], ",", 
      RowBox[{"1", "-", "z"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", "0", ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "z"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "z"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "z"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "z"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", "0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", "1"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", 
      RowBox[{"1", "-", "z"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z"}], ",", 
      RowBox[{"-", "z"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "z"}], ",", 
      RowBox[{"1", "-", "z"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", "0", ";", "z", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "y0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "y0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "y0"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "y0"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", "0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", "1"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", 
      RowBox[{"1", "-", "y0"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", 
      RowBox[{"-", "y0"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "y0"}], ",", 
      RowBox[{"1", "-", "y0"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", "0", ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "y0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "y0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "y0"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "y0"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", "0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", "1"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", 
      RowBox[{"1", "-", "y0"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y0"}], ",", 
      RowBox[{"-", "y0"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "y0"}], ",", 
      RowBox[{"1", "-", "y0"}]}], ";", "z0", ")"}]}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"solutionzy", "[", "3", "]"}], "//", 
   RowBox[{
    RowBox[{"Cases", "[", 
     RowBox[{"#", ",", "_Hlog", ",", 
      RowBox[{"-", "1"}]}], "]"}], "&"}]}], "//", "Union"}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    RowBox[{"G", "(", "0", ";", "y", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "z0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "z0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "z0"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "z0"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", "0"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", "1"}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", 
      RowBox[{"1", "-", "z0"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", 
      RowBox[{"-", "z0"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "z0"}], ",", 
      RowBox[{"1", "-", "z0"}]}], ";", "y", ")"}], ",", 
    RowBox[{"G", "(", "0", ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "z0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "z0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "z0"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "z0"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", "0"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", "1"}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", 
      RowBox[{"1", "-", "z0"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "z0"}], ",", 
      RowBox[{"-", "z0"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "z0"}], ",", 
      RowBox[{"1", "-", "z0"}]}], ";", "y0", ")"}], ",", 
    RowBox[{"G", "(", "0", ";", "z", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "y"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "y"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "y"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "y"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", "0"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", "1"}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", 
      RowBox[{"1", "-", "y"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", 
      RowBox[{"-", "y"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "y"}], ",", 
      RowBox[{"1", "-", "y"}]}], ";", "z", ")"}], ",", 
    RowBox[{"G", "(", "0", ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", "1", ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", "-", "y"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"-", "y"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", "1"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"1", "-", "y"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"0", ",", 
      RowBox[{"-", "y"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{"1", ",", "0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", "0"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", "1"}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", 
      RowBox[{"1", "-", "y"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"1", "-", "y"}], ",", 
      RowBox[{"-", "y"}]}], ";", "z0", ")"}], ",", 
    RowBox[{"G", "(", 
     RowBox[{
      RowBox[{"-", "y"}], ",", 
      RowBox[{"1", "-", "y"}]}], ";", "z0", ")"}]}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Solution Checks", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"solutionyz", "[", "3", "]"}], "-", 
    RowBox[{"solutionzy", "[", "3", "]"}]}], "//", 
   RowBox[{"MapSeries", "[", "Expand", "]"}]}], "//", 
  RowBox[{"MapSeries", "[", 
   RowBox[{"HyperIntFibrationBasis", "[", 
    RowBox[{"{", 
     RowBox[{"y", ",", "z", ",", "y0", ",", "z0"}], "}"}], "]"}], 
   "]"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "3"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "3"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "3"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "3"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "3"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "3"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 3, 3, 1],
     Editable->False]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"solutionyz", "[", "2", "]"}], "-", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"solutionzy", "[", "2", "]"}], "/.", 
            RowBox[{"C", "\[Rule]", "C2"}]}], ")"}]}], ")"}], "[", 
        RowBox[{"[", 
         RowBox[{";;", ",", "3"}], "]"}], "]"}], "//", "Flatten"}], "//", 
      "Union"}], "//", "ExpandAll"}], "//", 
    RowBox[{"Map", "[", 
     RowBox[{"HyperIntFibrationBasis", "[", 
      RowBox[{"{", 
       RowBox[{"y", ",", "z", ",", "y0", ",", "z0"}], "}"}], "]"}], "]"}]}], "//", 
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{"#", "\[Equal]", "0"}], "&"}], "]"}]}], "//", 
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{"#", ",", 
     RowBox[{
      RowBox[{"Cases", "[", 
       RowBox[{"#", ",", "_C2", ",", 
        RowBox[{"-", "1"}]}], "]"}], "//", "Union"}]}], "]"}], 
   "&"}]}]], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"1", ",", "0"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"1", ",", "1"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"1", ",", "1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"2", ",", "0"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"2", ",", "1"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"2", ",", "1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"3", ",", "0"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"3", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"3", ",", "1"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"3", ",", "1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"4", ",", "0"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"4", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"4", ",", "1"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"4", ",", "1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"5", ",", "0"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"5", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"5", ",", "1"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"5", ",", "1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"6", ",", "0"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"6", ",", "0"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"C2", "(", 
       RowBox[{"6", ",", "1"}], ")"}], "\[Rule]", 
      RowBox[{"C", "[", 
       RowBox[{"6", ",", "1"}], "]"}]}]}], "}"}], "}"}], 
  TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"HlogD", "[", 
   RowBox[{
    RowBox[{"solutionzy", "[", "4", "]"}], ",", "y"}], "]"}], "-", 
  RowBox[{"ep", "*", 
   RowBox[{"Sy", ".", 
    RowBox[{"solutionzy", "[", "4", "]"}]}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"HlogD", "[", 
   RowBox[{
    RowBox[{"solutionyz", "[", "4", "]"}], ",", "y"}], "]"}], "-", 
  RowBox[{"ep", "*", 
   RowBox[{"Sy", ".", 
    RowBox[{"solutionyz", "[", "4", "]"}]}]}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False]}], "}"}], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"HlogD", "[", 
   RowBox[{
    RowBox[{"solutionzy", "[", "4", "]"}], ",", "z"}], "]"}], "-", 
  RowBox[{"ep", "*", 
   RowBox[{"Sz", ".", 
    RowBox[{"solutionzy", "[", "4", "]"}]}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"HlogD", "[", 
    RowBox[{
     RowBox[{"solutionyz", "[", "4", "]"}], ",", "z"}], "]"}], "-", 
   RowBox[{"ep", "*", 
    RowBox[{"Sz", ".", 
     RowBox[{"solutionyz", "[", "4", "]"}]}]}]}], "//", 
  RowBox[{"MapSeries", "[", "Together", "]"}]}]}], "Input"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False]}], "}"}], TraditionalForm]], "Output"],

Cell[BoxData[
 FormBox[
  RowBox[{"{", 
   RowBox[{
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False], ",", 
    InterpretationBox[
     RowBox[{"O", "(", 
      SuperscriptBox["ep", "4"], ")"}],
     SeriesData[$CellContext`ep, 0, {}, 4, 4, 1],
     Editable->False]}], "}"}], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]]
}, Closed]]
},
WindowSize->{1920, 1046},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Magnification:>1.25 Inherited,
FrontEndVersion->"10.4 for Linux x86 (64-bit) (April 11, 2016)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 115, 3, 82, "Section"],
Cell[698, 27, 213, 4, 97, "Text"],
Cell[914, 33, 124, 3, 41, "Text"],
Cell[1041, 38, 236, 6, 89, "Code",
 Evaluatable->False,
 InitializationCell->False],
Cell[1280, 46, 155, 3, 69, "Text"],
Cell[CellGroupData[{
Cell[1460, 53, 139, 4, 67, "Input"],
Cell[CellGroupData[{
Cell[1624, 61, 2282, 63, 36, "Print"],
Cell[3909, 126, 1067, 29, 36, "Print"],
Cell[4979, 157, 223, 4, 34, "Print"],
Cell[5205, 163, 208, 4, 34, "Print"],
Cell[5416, 169, 1868, 53, 36, "Print"]
}, Open  ]]
}, Open  ]],
Cell[7311, 226, 76, 2, 39, "Input"],
Cell[7390, 230, 217, 6, 67, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[7644, 241, 247, 9, 66, "Section"],
Cell[CellGroupData[{
Cell[7916, 254, 51, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[7992, 258, 478, 15, 128, "Input"],
Cell[8473, 275, 1896, 52, 721, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[10406, 332, 3408, 86, 1078, "Input"],
Cell[13817, 420, 5509, 130, 721, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[19375, 556, 60, 0, 110, "Subsection"],
Cell[19438, 558, 438, 13, 128, "Input"],
Cell[CellGroupData[{
Cell[19901, 575, 263, 8, 75, "Input"],
Cell[20167, 585, 13472, 363, 660, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[33676, 953, 2241, 50, 603, "Input"],
Cell[35920, 1005, 3466, 121, 147, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[39423, 1131, 129, 3, 75, "Input"],
Cell[39555, 1136, 3082, 108, 128, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[42674, 1249, 1350, 36, 395, "Input"],
Cell[44027, 1287, 2178, 78, 77, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[46254, 1371, 60, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[46339, 1375, 188, 5, 75, "Input"],
Cell[46530, 1382, 3228, 117, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[49795, 1504, 60, 1, 75, "Input"],
Cell[49858, 1507, 483, 9, 243, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[50378, 1521, 848, 23, 288, "Input"],
Cell[51229, 1546, 2236, 85, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[53502, 1636, 87, 2, 75, "Input"],
Cell[53592, 1640, 2037, 76, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[55666, 1721, 90, 2, 75, "Input"],
Cell[55759, 1725, 1215, 40, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[57011, 1770, 156, 4, 75, "Input"],
Cell[57170, 1776, 2886, 103, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[60093, 1884, 274, 8, 75, "Input"],
Cell[60370, 1894, 534, 19, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[60941, 1918, 353, 11, 75, "Input"],
Cell[61297, 1931, 295, 8, 87, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[61629, 1944, 383, 13, 88, "Subsubsection"],
Cell[62015, 1959, 1687, 60, 202, "Text"]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[63763, 2026, 297, 11, 69, "Section"],
Cell[CellGroupData[{
Cell[64085, 2041, 42, 0, 110, "Subsection"],
Cell[64130, 2043, 10353, 255, 2453, "Input"],
Cell[74486, 2300, 2001, 56, 446, "Input"],
Cell[CellGroupData[{
Cell[76512, 2360, 13434, 363, 1976, "Input"],
Cell[CellGroupData[{
Cell[89971, 2727, 5502, 132, 751, "Print"],
Cell[95476, 2861, 37226, 813, 751, "Print"],
Cell[132705, 3676, 124425, 2650, 751, "Print"],
Cell[257133, 6328, 129889, 2775, 751, "Print"],
Cell[387025, 9105, 109149, 2417, 751, "Print"],
Cell[496177, 11524, 121287, 2600, 751, "Print"],
Cell[617467, 14126, 101958, 2232, 751, "Print"],
Cell[719428, 16360, 18630, 392, 751, "Print"],
Cell[738061, 16754, 142772, 2914, 751, "Print"],
Cell[880836, 19670, 130262, 2722, 751, "Print"],
Cell[1011101, 22394, 95430, 1997, 751, "Print"],
Cell[1106534, 24393, 99881, 2013, 751, "Print"],
Cell[1206418, 26408, 18268, 416, 751, "Print"],
Cell[1224689, 26826, 68954, 1499, 751, "Print"],
Cell[1293646, 28327, 34062, 760, 751, "Print"]
}, Open  ]]
}, Open  ]],
Cell[1327735, 29091, 795, 21, 237, "Input"],
Cell[CellGroupData[{
Cell[1328555, 29116, 1258, 34, 288, "Input"],
Cell[1329816, 29152, 58, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1329911, 29158, 89, 2, 75, "Input"],
Cell[1330003, 29162, 9913, 356, 245, "Output"]
}, Open  ]],
Cell[1339931, 29521, 561, 18, 128, "Input"],
Cell[CellGroupData[{
Cell[1340517, 29543, 89, 2, 75, "Input"],
Cell[1340609, 29547, 10943, 370, 253, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1351601, 29923, 38, 0, 110, "Subsection"],
Cell[1351642, 29925, 3220, 97, 657, "Input"],
Cell[CellGroupData[{
Cell[1354887, 30026, 901, 26, 183, "Input"],
Cell[CellGroupData[{
Cell[1355813, 30056, 899, 25, 62, "Print"],
Cell[1356715, 30083, 1186, 32, 62, "Print"]
}, Open  ]],
Cell[1357916, 30118, 59, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1358012, 30124, 90, 2, 75, "Input"],
Cell[1358105, 30128, 9054, 302, 176, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1367208, 30436, 46, 0, 110, "Subsection"],
Cell[1367257, 30438, 4603, 124, 815, "Input"],
Cell[CellGroupData[{
Cell[1371885, 30566, 602, 19, 75, "Input"],
Cell[1372490, 30587, 8163, 240, 345, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1380690, 30832, 208, 6, 128, "Input"],
Cell[1380901, 30840, 59, 1, 72, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1381009, 30847, 35, 0, 110, "Subsection"],
Cell[1381047, 30849, 643, 13, 187, "Input"],
Cell[CellGroupData[{
Cell[1381715, 30866, 502, 15, 128, "Input"],
Cell[CellGroupData[{
Cell[1382242, 30885, 878, 25, 56, "Print"],
Cell[1383123, 30912, 1025, 30, 56, "Print"]
}, Open  ]],
Cell[1384163, 30945, 58, 1, 72, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1384270, 30952, 42, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[1384337, 30956, 319, 10, 128, "Input"],
Cell[CellGroupData[{
Cell[1384681, 30970, 538, 15, 56, "Print"],
Cell[1385222, 30987, 662, 20, 56, "Print"]
}, Open  ]],
Cell[1385899, 31010, 58, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1385994, 31016, 93, 2, 75, "Input"],
Cell[1386090, 31020, 4771, 165, 171, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1390910, 31191, 48, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[1390983, 31195, 1374, 39, 340, "Input"],
Cell[CellGroupData[{
Cell[1392382, 31238, 988, 24, 56, "Print"],
Cell[1393373, 31264, 1511, 37, 103, "Print"],
Cell[1394887, 31303, 1674, 41, 103, "Print"]
}, Open  ]],
Cell[1396576, 31347, 60, 1, 72, "Output"]
}, Open  ]],
Cell[1396651, 31351, 135, 4, 75, "Input"],
Cell[1396789, 31357, 94, 2, 75, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1396920, 31364, 50, 0, 110, "Subsection"],
Cell[1396973, 31366, 1354, 38, 288, "Input"],
Cell[CellGroupData[{
Cell[1398352, 31408, 280, 8, 128, "Input"],
Cell[CellGroupData[{
Cell[1398657, 31420, 430, 12, 56, "Print"],
Cell[1399090, 31434, 541, 16, 56, "Print"]
}, Open  ]],
Cell[1399646, 31453, 60, 1, 72, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1399755, 31460, 46, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[1399826, 31464, 569, 16, 128, "Input"],
Cell[CellGroupData[{
Cell[1400420, 31484, 942, 28, 56, "Print"],
Cell[1401365, 31514, 1092, 33, 56, "Print"]
}, Open  ]],
Cell[1402472, 31550, 59, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1402568, 31556, 96, 2, 75, "Input"],
Cell[1402667, 31560, 2427, 92, 171, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1405131, 31657, 473, 14, 128, "Input"],
Cell[1405607, 31673, 798, 23, 56, "Print"],
Cell[1406408, 31698, 59, 1, 72, "Output"]
}, Open  ]],
Cell[1406482, 31702, 117, 3, 75, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1406648, 31711, 59, 0, 64, "Section"],
Cell[CellGroupData[{
Cell[1406732, 31715, 77, 0, 57, "Subsection"],
Cell[1406812, 31717, 3417, 94, 257, "Input"],
Cell[CellGroupData[{
Cell[1410254, 31815, 136, 4, 39, "Input"],
Cell[1410393, 31821, 122, 3, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1410552, 31829, 2312, 60, 365, "Input"],
Cell[1412867, 31891, 59, 1, 38, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1412963, 31897, 93, 2, 39, "Input"],
Cell[1413059, 31901, 954, 26, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1414050, 31932, 370, 10, 67, "Input"],
Cell[1414423, 31944, 284, 6, 38, "Output"],
Cell[1414710, 31952, 217, 6, 38, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1414976, 31964, 42, 0, 57, "Subsection"],
Cell[1415021, 31966, 1945, 54, 229, "Input"],
Cell[CellGroupData[{
Cell[1416991, 32024, 598, 17, 67, "Input"],
Cell[1417592, 32043, 1120, 30, 29, "Print"],
Cell[1418715, 32075, 59, 1, 38, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1418823, 32082, 64, 0, 57, "Subsection"],
Cell[CellGroupData[{
Cell[1418912, 32086, 1442, 40, 177, "Input"],
Cell[1420357, 32128, 2957, 89, 318, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1423351, 32222, 1191, 33, 150, "Input"],
Cell[1424545, 32257, 2138, 70, 282, "Output"]
}, Open  ]],
Cell[1426698, 32330, 108, 3, 39, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1426843, 32338, 52, 0, 57, "Subsection"],
Cell[1426898, 32340, 9089, 248, 1066, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1436024, 32593, 36, 0, 57, "Subsection"],
Cell[1436063, 32595, 1676, 45, 283, "Input"],
Cell[1437742, 32642, 2032, 52, 337, "Input"],
Cell[CellGroupData[{
Cell[1439799, 32698, 1082, 31, 148, "Input"],
Cell[CellGroupData[{
Cell[1440906, 32733, 242, 9, 32, "Print"],
Cell[1441151, 32744, 3894, 81, 45, "Print"],
Cell[1445048, 32827, 248, 9, 32, "Print"],
Cell[1445299, 32838, 4242, 91, 78, "Print"],
Cell[1449544, 32931, 248, 9, 32, "Print"],
Cell[1449795, 32942, 4242, 91, 78, "Print"],
Cell[1454040, 33035, 275, 10, 32, "Print"],
Cell[1454318, 33047, 35583, 725, 725, "Print"],
Cell[1489904, 33774, 275, 10, 32, "Print"],
Cell[1490182, 33786, 5412, 114, 75, "Print"],
Cell[1495597, 33902, 275, 10, 32, "Print"],
Cell[1495875, 33914, 41353, 821, 725, "Print"],
Cell[1537231, 34737, 253, 9, 32, "Print"],
Cell[1537487, 34748, 7797, 150, 78, "Print"],
Cell[1545287, 34900, 259, 9, 32, "Print"],
Cell[1545549, 34911, 53881, 1092, 1110, "Print"],
Cell[1599433, 36005, 259, 9, 32, "Print"],
Cell[1599695, 36016, 53881, 1092, 1110, "Print"],
Cell[1653579, 37110, 264, 9, 32, "Print"],
Cell[1653846, 37121, 67274, 1336, 1289, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1721181, 38464, 44, 0, 57, "Subsection"],
Cell[1721228, 38466, 5628, 150, 715, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1726893, 38621, 59, 0, 57, "Subsection"],
Cell[1726955, 38623, 8664, 229, 769, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1735668, 38858, 50, 0, 64, "Section"],
Cell[CellGroupData[{
Cell[1735743, 38862, 1210, 35, 288, "Input"],
Cell[1736956, 38899, 56, 1, 72, "Output"],
Cell[1737015, 38902, 98, 3, 72, "Output"],
Cell[1737116, 38907, 3258, 67, 102, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1740411, 38979, 30, 0, 75, "Input"],
Cell[1740444, 38981, 189, 5, 81, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1740670, 38991, 811, 28, 128, "Input"],
Cell[1741484, 39021, 437, 15, 227, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1741958, 39041, 524, 17, 75, "Input"],
Cell[1742485, 39060, 1189, 33, 157, "Output"]
}, Open  ]],
Cell[1743689, 39096, 125, 3, 75, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[1743851, 39104, 65, 0, 64, "Section"],
Cell[CellGroupData[{
Cell[1743941, 39108, 1359, 42, 290, "Input"],
Cell[1745303, 39152, 56, 1, 72, "Output"],
Cell[1745362, 39155, 802, 23, 77, "Output"],
Cell[1746167, 39180, 3246, 67, 166, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1749450, 39252, 162, 5, 78, "Input"],
Cell[1749615, 39259, 3031, 82, 206, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1752683, 39346, 319, 9, 75, "Input"],
Cell[1753005, 39357, 802, 23, 77, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1753844, 39385, 363, 10, 75, "Input"],
Cell[1754210, 39397, 2981, 77, 157, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1757228, 39479, 194, 6, 75, "Input"],
Cell[1757425, 39487, 748, 21, 90, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1758210, 39513, 47, 0, 110, "Subsection"],
Cell[1758260, 39515, 4043, 116, 762, "Input"],
Cell[CellGroupData[{
Cell[1762328, 39635, 253, 7, 75, "Input"],
Cell[1762584, 39644, 874, 26, 150, "Output"]
}, Open  ]],
Cell[1763473, 39673, 4710, 121, 972, "Input"],
Cell[CellGroupData[{
Cell[1768208, 39798, 790, 24, 128, "Input"],
Cell[1769001, 39824, 3947, 110, 260, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1772997, 39940, 34, 0, 110, "Subsection"],
Cell[1773034, 39942, 1683, 48, 393, "Input"],
Cell[CellGroupData[{
Cell[1774742, 39994, 136, 3, 75, "Input"],
Cell[1774881, 39999, 5466, 164, 221, "Output"]
}, Open  ]],
Cell[1780362, 40166, 1915, 58, 340, "Input"],
Cell[1782280, 40226, 612, 18, 235, "Input"],
Cell[CellGroupData[{
Cell[1782917, 40248, 999, 29, 235, "Input"],
Cell[1783919, 40279, 58, 1, 72, "Output"],
Cell[1783980, 40282, 3544, 99, 157, "Output"],
Cell[1787527, 40383, 57, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1787621, 40389, 395, 13, 75, "Input"],
Cell[1788019, 40404, 58496, 984, 13316, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1846552, 41393, 2338, 69, 657, "Input"],
Cell[1848893, 41464, 150, 5, 90, "Output"],
Cell[1849046, 41471, 5765, 178, 507, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1854848, 41654, 125, 4, 75, "Input"],
Cell[1854976, 41660, 570, 16, 160, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1855583, 41681, 947, 27, 288, "Input"],
Cell[1856533, 41710, 4467, 97, 343, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1861049, 41813, 39, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[1861113, 41817, 1479, 45, 235, "Input"],
Cell[1862595, 41864, 1028, 34, 210, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1863660, 41903, 757, 26, 128, "Input"],
Cell[1864420, 41931, 451, 16, 227, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1864908, 41952, 464, 15, 75, "Input"],
Cell[1865375, 41969, 4609, 121, 378, "Output"]
}, Open  ]],
Cell[1869999, 42093, 106, 2, 75, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1870154, 42101, 59, 0, 64, "Section"],
Cell[CellGroupData[{
Cell[1870238, 42105, 1723, 53, 288, "Input"],
Cell[1871964, 42160, 56, 1, 72, "Output"],
Cell[1872023, 42163, 3260, 67, 166, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1875320, 42235, 250, 8, 78, "Input"],
Cell[1875573, 42245, 4695, 128, 508, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1880305, 42378, 1730, 55, 235, "Input"],
Cell[1882038, 42435, 677, 23, 165, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1882752, 42463, 1008, 31, 128, "Input"],
Cell[1883763, 42496, 2845, 90, 288, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1886645, 42591, 263, 8, 75, "Input"],
Cell[1886911, 42601, 56, 1, 72, "Output"]
}, Open  ]],
Cell[1886982, 42605, 324, 10, 75, "Input"],
Cell[CellGroupData[{
Cell[1887331, 42619, 60, 1, 75, "Input"],
Cell[1887394, 42622, 5648, 176, 596, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1893079, 42803, 212, 6, 75, "Input"],
Cell[1893294, 42811, 3324, 84, 337, "Output"]
}, Open  ]],
Cell[1896633, 42898, 323, 9, 75, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[1896993, 42912, 69, 0, 64, "Section"],
Cell[CellGroupData[{
Cell[1897087, 42916, 51, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[1897163, 42920, 1314, 39, 340, "Input"],
Cell[1898480, 42961, 58, 1, 72, "Output"],
Cell[1898541, 42964, 6837, 178, 402, "Output"],
Cell[1905381, 43144, 29975, 589, 296, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1935393, 43738, 449, 13, 128, "Input"],
Cell[1935845, 43753, 5313, 155, 402, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1941195, 43913, 718, 22, 182, "Input"],
Cell[1941916, 43937, 59, 1, 72, "Output"],
Cell[1941978, 43940, 58, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1942073, 43946, 93, 2, 75, "Input"],
Cell[1942169, 43950, 417, 10, 90, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1942635, 43966, 47, 0, 110, "Subsection"],
Cell[1942685, 43968, 890, 23, 393, "Input"],
Cell[1943578, 43993, 710, 20, 128, "Input"],
Cell[CellGroupData[{
Cell[1944313, 44017, 93, 2, 75, "Input"],
Cell[1944409, 44021, 475, 13, 157, "Output"]
}, Open  ]],
Cell[1944899, 44037, 1956, 48, 500, "Input"],
Cell[CellGroupData[{
Cell[1946880, 44089, 823, 21, 235, "Input"],
Cell[1947706, 44112, 1578, 39, 291, "Print"],
Cell[1949287, 44153, 303, 8, 51, "Message"],
Cell[1949593, 44163, 303, 8, 51, "Message"],
Cell[CellGroupData[{
Cell[1949921, 44175, 341, 8, 56, "Print"],
Cell[1950265, 44185, 124, 3, 56, "Print"],
Cell[1950392, 44190, 133, 3, 56, "Print"],
Cell[1950528, 44195, 160, 3, 56, "Print"],
Cell[1950691, 44200, 341, 8, 56, "Print"],
Cell[1951035, 44210, 124, 3, 56, "Print"],
Cell[1951162, 44215, 133, 3, 56, "Print"],
Cell[1951298, 44220, 160, 3, 56, "Print"],
Cell[1951461, 44225, 341, 8, 56, "Print"],
Cell[1951805, 44235, 124, 3, 56, "Print"],
Cell[1951932, 44240, 133, 3, 56, "Print"],
Cell[1952068, 44245, 160, 3, 56, "Print"],
Cell[1952231, 44250, 341, 8, 56, "Print"],
Cell[1952575, 44260, 124, 3, 56, "Print"],
Cell[1952702, 44265, 133, 3, 56, "Print"],
Cell[1952838, 44270, 160, 3, 56, "Print"],
Cell[1953001, 44275, 341, 8, 56, "Print"],
Cell[1953345, 44285, 124, 3, 56, "Print"],
Cell[1953472, 44290, 133, 3, 56, "Print"],
Cell[1953608, 44295, 160, 3, 56, "Print"],
Cell[1953771, 44300, 341, 8, 56, "Print"],
Cell[1954115, 44310, 124, 3, 56, "Print"],
Cell[1954242, 44315, 133, 3, 56, "Print"],
Cell[1954378, 44320, 160, 3, 56, "Print"],
Cell[1954541, 44325, 341, 8, 56, "Print"],
Cell[1954885, 44335, 124, 3, 56, "Print"],
Cell[1955012, 44340, 133, 3, 56, "Print"],
Cell[1955148, 44345, 160, 3, 56, "Print"],
Cell[1955311, 44350, 283, 7, 56, "Print"],
Cell[1955597, 44359, 1146, 28, 143, "Print"],
Cell[1956746, 44389, 283, 7, 56, "Print"],
Cell[1957032, 44398, 953, 23, 143, "Print"],
Cell[1957988, 44423, 283, 7, 56, "Print"],
Cell[1958274, 44432, 283, 7, 56, "Print"],
Cell[1958560, 44441, 283, 7, 56, "Print"],
Cell[1958846, 44450, 1291, 32, 143, "Print"],
Cell[1960140, 44484, 283, 7, 56, "Print"],
Cell[1960426, 44493, 1521, 37, 143, "Print"],
Cell[1961950, 44532, 283, 7, 56, "Print"],
Cell[1962236, 44541, 1146, 28, 143, "Print"],
Cell[1963385, 44571, 95, 2, 56, "Print"]
}, Open  ]],
Cell[1963495, 44576, 285, 8, 51, "Message"],
Cell[1963783, 44586, 1295, 33, 56, "Print"],
Cell[1965081, 44621, 285, 8, 51, "Message"],
Cell[CellGroupData[{
Cell[1965391, 44633, 269, 7, 56, "Print"],
Cell[1965663, 44642, 90, 2, 56, "Print"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1965802, 44650, 230, 7, 58, "Input"],
Cell[1966035, 44659, 739, 23, 90, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1966811, 44687, 1034, 28, 288, "Input"],
Cell[1967848, 44717, 25497, 544, 1746, "Output"]
}, Open  ]],
Cell[1993360, 45264, 467, 12, 288, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1993876, 45282, 67, 0, 64, "Section"],
Cell[1993946, 45284, 578, 14, 128, "Input"],
Cell[CellGroupData[{
Cell[1994549, 45302, 2395, 71, 446, "Input"],
Cell[1996947, 45375, 58, 1, 72, "Output"],
Cell[1997008, 45378, 3178, 83, 306, "Output"],
Cell[2000189, 45463, 19161, 378, 275, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2019387, 45846, 429, 12, 128, "Input"],
Cell[2019819, 45860, 2522, 75, 306, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2022378, 45940, 698, 21, 182, "Input"],
Cell[2023079, 45963, 59, 1, 72, "Output"],
Cell[2023141, 45966, 57, 1, 72, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2023235, 45972, 93, 2, 75, "Input"],
Cell[2023331, 45976, 644, 19, 168, "Output"]
}, Open  ]],
Cell[2023990, 45998, 710, 20, 128, "Input"],
Cell[CellGroupData[{
Cell[2024725, 46022, 93, 2, 75, "Input"],
Cell[2024821, 46026, 532, 14, 157, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2025390, 46045, 1358, 31, 603, "Input"],
Cell[2026751, 46078, 1578, 39, 291, "Print"],
Cell[2028332, 46119, 303, 8, 51, "Message"],
Cell[2028638, 46129, 303, 8, 51, "Message"],
Cell[CellGroupData[{
Cell[2028966, 46141, 281, 7, 56, "Print"],
Cell[2029250, 46150, 466, 7, 338, "Print"]
}, Open  ]],
Cell[2029731, 46160, 322, 10, 51, "Message"],
Cell[CellGroupData[{
Cell[2030078, 46174, 287, 7, 56, "Print"],
Cell[2030368, 46183, 279, 5, 151, "Print"],
Cell[2030650, 46190, 293, 7, 56, "Print"],
Cell[2030946, 46199, 330, 10, 56, "Print"]
}, Open  ]],
Cell[2031291, 46212, 285, 8, 51, "Message"],
Cell[2031579, 46222, 285, 8, 51, "Message"],
Cell[CellGroupData[{
Cell[2031889, 46234, 579, 8, 338, "Print"],
Cell[2032471, 46244, 299, 7, 56, "Print"],
Cell[2032773, 46253, 106, 3, 56, "Print"],
Cell[2032882, 46258, 293, 7, 56, "Print"],
Cell[2033178, 46267, 126, 3, 56, "Print"],
Cell[2033307, 46272, 157, 3, 56, "Print"]
}, Open  ]],
Cell[2033479, 46278, 285, 8, 51, "Message"],
Cell[2033767, 46288, 404, 11, 51, "Message"],
Cell[CellGroupData[{
Cell[2034196, 46303, 128, 3, 56, "Print"],
Cell[2034327, 46308, 153, 3, 56, "Print"],
Cell[2034483, 46313, 133, 3, 56, "Print"],
Cell[2034619, 46318, 146, 3, 56, "Print"],
Cell[2034768, 46323, 100, 2, 56, "Print"],
Cell[2034871, 46327, 627, 10, 291, "Print"],
Cell[2035501, 46339, 289, 7, 56, "Print"],
Cell[2035793, 46348, 345, 9, 56, "Print"],
Cell[2036141, 46359, 145, 3, 56, "Print"],
Cell[2036289, 46364, 121, 3, 56, "Print"],
Cell[2036413, 46369, 284, 5, 197, "Print"],
Cell[2036700, 46376, 281, 7, 56, "Print"]
}, Open  ]],
Cell[2036996, 46386, 335, 10, 51, "Message"],
Cell[CellGroupData[{
Cell[2037356, 46400, 281, 7, 56, "Print"],
Cell[2037640, 46409, 466, 7, 338, "Print"]
}, Open  ]],
Cell[2038121, 46419, 322, 10, 51, "Message"],
Cell[CellGroupData[{
Cell[2038468, 46433, 287, 7, 56, "Print"],
Cell[2038758, 46442, 279, 5, 151, "Print"],
Cell[2039040, 46449, 293, 7, 56, "Print"],
Cell[2039336, 46458, 342, 11, 56, "Print"],
Cell[2039681, 46471, 579, 8, 338, "Print"],
Cell[2040263, 46481, 299, 7, 56, "Print"],
Cell[2040565, 46490, 136, 4, 56, "Print"],
Cell[2040704, 46496, 127, 3, 56, "Print"],
Cell[2040834, 46501, 109, 2, 56, "Print"],
Cell[2040946, 46505, 379, 6, 197, "Print"],
Cell[2041328, 46513, 293, 7, 56, "Print"],
Cell[2041624, 46522, 126, 3, 56, "Print"],
Cell[2041753, 46527, 157, 3, 56, "Print"],
Cell[2041913, 46532, 128, 3, 56, "Print"],
Cell[2042044, 46537, 153, 3, 56, "Print"],
Cell[2042200, 46542, 133, 3, 56, "Print"],
Cell[2042336, 46547, 149, 3, 56, "Print"],
Cell[2042488, 46552, 100, 2, 56, "Print"],
Cell[2042591, 46556, 627, 10, 291, "Print"],
Cell[2043221, 46568, 289, 7, 56, "Print"],
Cell[2043513, 46577, 345, 9, 56, "Print"],
Cell[2043861, 46588, 145, 3, 56, "Print"],
Cell[2044009, 46593, 121, 3, 56, "Print"],
Cell[2044133, 46598, 284, 5, 197, "Print"],
Cell[2044420, 46605, 345, 9, 56, "Print"],
Cell[2044768, 46616, 145, 3, 56, "Print"],
Cell[2044916, 46621, 255, 4, 151, "Print"],
Cell[2045174, 46627, 281, 7, 56, "Print"]
}, Open  ]],
Cell[2045470, 46637, 335, 10, 51, "Message"],
Cell[CellGroupData[{
Cell[2045830, 46651, 281, 7, 56, "Print"],
Cell[2046114, 46660, 466, 7, 338, "Print"]
}, Open  ]],
Cell[2046595, 46670, 322, 10, 51, "Message"],
Cell[2046920, 46682, 391, 10, 51, "Message"],
Cell[CellGroupData[{
Cell[2047336, 46696, 287, 7, 56, "Print"],
Cell[2047626, 46705, 279, 5, 151, "Print"],
Cell[2047908, 46712, 293, 7, 56, "Print"],
Cell[2048204, 46721, 342, 11, 56, "Print"],
Cell[2048549, 46734, 579, 8, 338, "Print"],
Cell[2049131, 46744, 299, 7, 56, "Print"],
Cell[2049433, 46753, 155, 5, 56, "Print"],
Cell[2049591, 46760, 127, 3, 56, "Print"],
Cell[2049721, 46765, 109, 2, 56, "Print"],
Cell[2049833, 46769, 379, 6, 197, "Print"],
Cell[2050215, 46777, 127, 3, 56, "Print"],
Cell[2050345, 46782, 109, 2, 56, "Print"],
Cell[2050457, 46786, 379, 6, 197, "Print"],
Cell[2050839, 46794, 293, 7, 56, "Print"],
Cell[2051135, 46803, 126, 3, 56, "Print"],
Cell[2051264, 46808, 157, 3, 56, "Print"],
Cell[2051424, 46813, 128, 3, 56, "Print"],
Cell[2051555, 46818, 153, 3, 56, "Print"],
Cell[2051711, 46823, 133, 3, 56, "Print"],
Cell[2051847, 46828, 149, 3, 56, "Print"],
Cell[2051999, 46833, 100, 2, 56, "Print"],
Cell[2052102, 46837, 627, 10, 291, "Print"],
Cell[2052732, 46849, 289, 7, 56, "Print"],
Cell[2053024, 46858, 345, 9, 56, "Print"],
Cell[2053372, 46869, 145, 3, 56, "Print"],
Cell[2053520, 46874, 255, 4, 151, "Print"],
Cell[2053778, 46880, 281, 7, 56, "Print"]
}, Open  ]],
Cell[2054074, 46890, 335, 10, 51, "Message"],
Cell[2054412, 46902, 397, 10, 51, "Message"],
Cell[CellGroupData[{
Cell[2054834, 46916, 281, 7, 56, "Print"],
Cell[2055118, 46925, 466, 7, 338, "Print"],
Cell[2055587, 46934, 287, 7, 56, "Print"],
Cell[2055877, 46943, 279, 5, 151, "Print"],
Cell[2056159, 46950, 293, 7, 56, "Print"],
Cell[2056455, 46959, 354, 12, 56, "Print"],
Cell[2056812, 46973, 580, 8, 338, "Print"],
Cell[2057395, 46983, 299, 7, 56, "Print"],
Cell[2057697, 46992, 171, 5, 56, "Print"],
Cell[2057871, 46999, 127, 3, 56, "Print"],
Cell[2058001, 47004, 109, 2, 56, "Print"],
Cell[2058113, 47008, 379, 6, 197, "Print"],
Cell[2058495, 47016, 127, 3, 56, "Print"],
Cell[2058625, 47021, 109, 2, 56, "Print"],
Cell[2058737, 47025, 379, 6, 197, "Print"],
Cell[2059119, 47033, 127, 3, 56, "Print"],
Cell[2059249, 47038, 109, 2, 56, "Print"],
Cell[2059361, 47042, 379, 6, 197, "Print"],
Cell[2059743, 47050, 293, 7, 56, "Print"],
Cell[2060039, 47059, 126, 3, 56, "Print"],
Cell[2060168, 47064, 157, 3, 56, "Print"],
Cell[2060328, 47069, 128, 3, 56, "Print"],
Cell[2060459, 47074, 159, 3, 56, "Print"],
Cell[2060621, 47079, 133, 3, 56, "Print"],
Cell[2060757, 47084, 164, 3, 56, "Print"],
Cell[2060924, 47089, 100, 2, 56, "Print"],
Cell[2061027, 47093, 627, 10, 291, "Print"],
Cell[2061657, 47105, 289, 7, 56, "Print"],
Cell[2061949, 47114, 345, 9, 56, "Print"],
Cell[2062297, 47125, 145, 3, 56, "Print"],
Cell[2062445, 47130, 121, 3, 56, "Print"],
Cell[2062569, 47135, 284, 5, 197, "Print"],
Cell[2062856, 47142, 345, 9, 56, "Print"],
Cell[2063204, 47153, 145, 3, 56, "Print"],
Cell[2063352, 47158, 255, 4, 151, "Print"],
Cell[2063610, 47164, 345, 9, 56, "Print"],
Cell[2063958, 47175, 145, 3, 56, "Print"],
Cell[2064106, 47180, 255, 4, 151, "Print"],
Cell[2064364, 47186, 345, 9, 56, "Print"],
Cell[2064712, 47197, 145, 3, 56, "Print"],
Cell[2064860, 47202, 255, 4, 151, "Print"],
Cell[2065118, 47208, 345, 9, 56, "Print"],
Cell[2065466, 47219, 145, 3, 56, "Print"],
Cell[2065614, 47224, 121, 3, 56, "Print"],
Cell[2065738, 47229, 284, 5, 197, "Print"],
Cell[2066025, 47236, 281, 7, 56, "Print"],
Cell[2066309, 47245, 281, 7, 56, "Print"],
Cell[2066593, 47254, 466, 7, 338, "Print"],
Cell[2067062, 47263, 287, 7, 56, "Print"],
Cell[2067352, 47272, 279, 5, 151, "Print"],
Cell[2067634, 47279, 293, 7, 56, "Print"],
Cell[2067930, 47288, 354, 12, 56, "Print"],
Cell[2068287, 47302, 580, 8, 338, "Print"],
Cell[2068870, 47312, 299, 7, 56, "Print"],
Cell[2069172, 47321, 193, 6, 56, "Print"],
Cell[2069368, 47329, 127, 3, 56, "Print"],
Cell[2069498, 47334, 109, 2, 56, "Print"],
Cell[2069610, 47338, 379, 6, 197, "Print"],
Cell[2069992, 47346, 127, 3, 56, "Print"],
Cell[2070122, 47351, 109, 2, 56, "Print"],
Cell[2070234, 47355, 379, 6, 197, "Print"],
Cell[2070616, 47363, 127, 3, 56, "Print"],
Cell[2070746, 47368, 109, 2, 56, "Print"],
Cell[2070858, 47372, 379, 6, 197, "Print"],
Cell[2071240, 47380, 127, 3, 56, "Print"],
Cell[2071370, 47385, 109, 2, 56, "Print"],
Cell[2071482, 47389, 379, 6, 197, "Print"],
Cell[2071864, 47397, 293, 7, 56, "Print"],
Cell[2072160, 47406, 126, 3, 56, "Print"],
Cell[2072289, 47411, 157, 3, 56, "Print"],
Cell[2072449, 47416, 128, 3, 56, "Print"],
Cell[2072580, 47421, 159, 3, 56, "Print"],
Cell[2072742, 47426, 133, 3, 56, "Print"],
Cell[2072878, 47431, 164, 3, 56, "Print"],
Cell[2073045, 47436, 100, 2, 56, "Print"],
Cell[2073148, 47440, 627, 10, 291, "Print"],
Cell[2073778, 47452, 289, 7, 56, "Print"],
Cell[2074070, 47461, 345, 9, 56, "Print"],
Cell[2074418, 47472, 145, 3, 56, "Print"],
Cell[2074566, 47477, 121, 3, 56, "Print"],
Cell[2074690, 47482, 284, 5, 197, "Print"],
Cell[2074977, 47489, 345, 9, 56, "Print"],
Cell[2075325, 47500, 145, 3, 56, "Print"],
Cell[2075473, 47505, 255, 4, 151, "Print"],
Cell[2075731, 47511, 345, 9, 56, "Print"],
Cell[2076079, 47522, 145, 3, 56, "Print"],
Cell[2076227, 47527, 255, 4, 151, "Print"],
Cell[2076485, 47533, 345, 9, 56, "Print"],
Cell[2076833, 47544, 145, 3, 56, "Print"],
Cell[2076981, 47549, 255, 4, 151, "Print"],
Cell[2077239, 47555, 345, 9, 56, "Print"],
Cell[2077587, 47566, 145, 3, 56, "Print"],
Cell[2077735, 47571, 121, 3, 56, "Print"],
Cell[2077859, 47576, 284, 5, 197, "Print"],
Cell[2078146, 47583, 281, 7, 56, "Print"],
Cell[2078430, 47592, 283, 7, 56, "Print"],
Cell[2078716, 47601, 763, 19, 143, "Print"],
Cell[2079482, 47622, 283, 7, 56, "Print"],
Cell[2079768, 47631, 812, 22, 143, "Print"],
Cell[2080583, 47655, 283, 7, 56, "Print"],
Cell[2080869, 47664, 283, 7, 56, "Print"],
Cell[2081155, 47673, 807, 19, 143, "Print"],
Cell[2081965, 47694, 1234, 29, 143, "Print"],
Cell[2083202, 47725, 283, 7, 56, "Print"],
Cell[2083488, 47734, 908, 23, 143, "Print"],
Cell[2084399, 47759, 1234, 29, 143, "Print"],
Cell[2085636, 47790, 95, 2, 56, "Print"],
Cell[2085734, 47794, 1213, 32, 56, "Print"],
Cell[2086950, 47828, 269, 7, 56, "Print"],
Cell[2087222, 47837, 90, 2, 56, "Print"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2087361, 47845, 946, 27, 220, "Input"],
Cell[2088310, 47874, 59, 1, 72, "Output"],
Cell[2088372, 47877, 626, 16, 90, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2089035, 47898, 95, 2, 75, "Input"],
Cell[2089133, 47902, 1019, 30, 172, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2090189, 47937, 1930, 48, 551, "Input"],
Cell[2092122, 47987, 34311, 715, 2103, "Output"]
}, Open  ]],
Cell[2126448, 48705, 440, 11, 288, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[2126925, 48721, 102, 2, 64, "Section"],
Cell[CellGroupData[{
Cell[2127052, 48727, 1502, 40, 202, "Input"],
Cell[2128557, 48769, 58, 1, 38, "Output"],
Cell[2128618, 48772, 9598, 251, 357, "Output"],
Cell[2138219, 49025, 53965, 1049, 278, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2192221, 50079, 869, 24, 67, "Input"],
Cell[2193093, 50105, 59, 1, 38, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2193189, 50111, 94, 2, 39, "Input"],
Cell[2193286, 50115, 1699, 50, 111, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2195022, 50170, 1412, 40, 121, "Input"],
Cell[2196437, 50212, 59, 1, 38, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2196533, 50218, 94, 2, 39, "Input"],
Cell[2196630, 50222, 1576, 45, 115, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2198243, 50272, 1851, 47, 257, "Input"],
Cell[2200097, 50321, 7844, 231, 214, "Output"],
Cell[2207944, 50554, 28970, 563, 152, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2236951, 51122, 1235, 35, 94, "Input"],
Cell[2238189, 51159, 1885, 57, 155, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2240111, 51221, 457, 13, 67, "Input"],
Cell[2240571, 51236, 7844, 231, 214, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2248452, 51472, 507, 15, 67, "Input"],
Cell[2248962, 51489, 821, 22, 31, "Print"],
Cell[2249786, 51513, 60, 1, 38, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2249883, 51519, 93, 2, 39, "Input"],
Cell[2249979, 51523, 443, 12, 47, "Output"]
}, Open  ]],
Cell[2250437, 51538, 1789, 46, 283, "Input"],
Cell[CellGroupData[{
Cell[2252251, 51588, 124, 3, 39, "Input"],
Cell[2252378, 51593, 7975, 218, 368, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2260390, 51816, 262, 7, 67, "Input"],
Cell[2260655, 51825, 59, 1, 38, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2260751, 51831, 39, 0, 57, "Subsection"],
Cell[2260793, 51833, 5827, 147, 634, "Input"],
Cell[CellGroupData[{
Cell[2266645, 51984, 93, 2, 39, "Input"],
Cell[2266741, 51988, 413, 11, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2267191, 52004, 320, 8, 67, "Input"],
Cell[2267514, 52014, 5271, 136, 202, "Output"]
}, Open  ]],
Cell[2272800, 52153, 10359, 232, 1660, "Input"]
}, Open  ]],
Cell[2283174, 52388, 92, 2, 57, "Subsection"],
Cell[CellGroupData[{
Cell[2283291, 52394, 60, 0, 46, "Subsection"],
Cell[2283354, 52396, 890, 26, 148, "Input"],
Cell[2284247, 52424, 1939, 54, 229, "Input"],
Cell[CellGroupData[{
Cell[2286211, 52482, 6430, 168, 850, "Input"],
Cell[CellGroupData[{
Cell[2292666, 52654, 253, 6, 29, "Print"],
Cell[2292922, 52662, 1010, 27, 29, "Print"],
Cell[2293935, 52691, 202, 5, 29, "Print"],
Cell[2294140, 52698, 1093, 27, 39, "Print"],
Cell[2295236, 52727, 529, 15, 29, "Print"],
Cell[2295768, 52744, 3241, 107, 197, "Print"],
Cell[2299012, 52853, 253, 6, 29, "Print"],
Cell[2299268, 52861, 1010, 27, 29, "Print"],
Cell[2300281, 52890, 202, 5, 29, "Print"],
Cell[2300486, 52897, 1076, 27, 39, "Print"],
Cell[2301565, 52926, 527, 15, 29, "Print"],
Cell[2302095, 52943, 2921, 95, 194, "Print"],
Cell[2305019, 53040, 253, 6, 29, "Print"],
Cell[2305275, 53048, 1010, 27, 29, "Print"],
Cell[2306288, 53077, 202, 5, 29, "Print"],
Cell[2306493, 53084, 1523, 37, 39, "Print"],
Cell[2308019, 53123, 529, 15, 29, "Print"],
Cell[2308551, 53140, 5070, 166, 292, "Print"],
Cell[2313624, 53308, 253, 6, 29, "Print"],
Cell[2313880, 53316, 1008, 27, 29, "Print"],
Cell[2314891, 53345, 200, 5, 29, "Print"],
Cell[2315094, 53352, 861, 22, 39, "Print"],
Cell[2315958, 53376, 529, 15, 29, "Print"],
Cell[2316490, 53393, 1638, 53, 146, "Print"],
Cell[2318131, 53448, 253, 6, 29, "Print"],
Cell[2318387, 53456, 1008, 27, 29, "Print"],
Cell[2319398, 53485, 200, 5, 29, "Print"],
Cell[2319601, 53492, 646, 17, 39, "Print"],
Cell[2320250, 53511, 529, 15, 29, "Print"],
Cell[2320782, 53528, 1131, 36, 100, "Print"],
Cell[2321916, 53566, 253, 6, 29, "Print"],
Cell[2322172, 53574, 1010, 27, 29, "Print"],
Cell[2323185, 53603, 202, 5, 29, "Print"],
Cell[2323390, 53610, 1308, 32, 39, "Print"],
Cell[2324701, 53644, 529, 15, 29, "Print"],
Cell[2325233, 53661, 6480, 199, 252, "Print"],
Cell[2331716, 53862, 253, 6, 29, "Print"],
Cell[2331972, 53870, 1010, 27, 29, "Print"],
Cell[2332985, 53899, 202, 5, 29, "Print"],
Cell[2333190, 53906, 1523, 37, 39, "Print"],
Cell[2334716, 53945, 529, 15, 29, "Print"],
Cell[2335248, 53962, 7485, 233, 302, "Print"]
}, Open  ]],
Cell[2342748, 54198, 59, 1, 38, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2342844, 54204, 95, 2, 39, "Input"],
Cell[2342942, 54208, 444, 12, 83, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2343423, 54225, 41, 0, 39, "Input"],
Cell[2343467, 54227, 5751, 145, 115, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2349255, 54377, 39, 0, 39, "Input"],
Cell[2349297, 54379, 38868, 1117, 863, "Output"]
}, Open  ]],
Cell[2388180, 55499, 547, 15, 121, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2388776, 55520, 91, 2, 64, "Section"],
Cell[CellGroupData[{
Cell[2388892, 55526, 12617, 347, 1132, "Input"],
Cell[2401512, 55875, 9177, 284, 331, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2410726, 56164, 64, 0, 110, "Subsection"],
Cell[2410793, 56166, 197, 6, 296, "Text"],
Cell[CellGroupData[{
Cell[2411015, 56176, 308, 7, 235, "Input"],
Cell[CellGroupData[{
Cell[2411348, 56187, 144, 2, 151, "Print"],
Cell[2411495, 56191, 123, 2, 56, "Print"],
Cell[2411621, 56195, 139, 2, 56, "Print"],
Cell[2411763, 56199, 100, 2, 56, "Print"],
Cell[2411866, 56203, 107, 2, 56, "Print"],
Cell[2411976, 56207, 184, 3, 103, "Print"],
Cell[2412163, 56212, 162, 2, 103, "Print"],
Cell[2412328, 56216, 275, 4, 385, "Print"],
Cell[2412606, 56222, 280, 4, 338, "Print"],
Cell[2412889, 56228, 152, 3, 103, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2413090, 56237, 683, 20, 128, "Input"],
Cell[2413776, 56259, 172447, 4076, 370, "Output"]
}, Open  ]],
Cell[2586238, 60338, 516, 19, 75, "Input"],
Cell[CellGroupData[{
Cell[2586779, 60361, 766, 22, 128, "Input"],
Cell[2587548, 60385, 13928, 383, 585, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2601525, 60774, 34, 0, 110, "Subsection"],
Cell[CellGroupData[{
Cell[2601584, 60778, 381, 11, 75, "Input"],
Cell[2601968, 60791, 5659, 146, 550, "Output"]
}, Open  ]],
Cell[2607642, 60940, 132, 4, 75, "Input"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2607823, 60950, 90, 2, 64, "Section"],
Cell[2607916, 60954, 101, 2, 71, "Subsection"],
Cell[CellGroupData[{
Cell[2608042, 60960, 62, 0, 57, "Subsection"],
Cell[2608107, 60962, 839, 25, 151, "Input"],
Cell[CellGroupData[{
Cell[2608971, 60991, 631, 22, 83, "Input"],
Cell[2609605, 61015, 1978, 63, 338, "Output"],
Cell[2611586, 61080, 2318, 74, 336, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2613953, 61160, 56, 0, 71, "Subsection"],
Cell[2614012, 61162, 2886, 78, 421, "Input"],
Cell[CellGroupData[{
Cell[2616923, 61244, 364, 6, 117, "Input"],
Cell[2617290, 61252, 742, 20, 203, "Output"],
Cell[2618035, 61274, 2032, 60, 261, "Output"],
Cell[2620070, 61336, 4900, 147, 261, "Output"]
}, Open  ]],
Cell[2624985, 61486, 4193, 115, 556, "Input"],
Cell[2629181, 61603, 1977, 55, 185, "Input"],
Cell[CellGroupData[{
Cell[2631183, 61662, 117, 2, 48, "Input"],
Cell[2631303, 61666, 19799, 531, 301, "Output"]
}, Open  ]],
Cell[2651117, 62200, 1415, 38, 117, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2652569, 62243, 35, 0, 71, "Subsection"],
Cell[2652607, 62245, 1517, 38, 218, "Input"],
Cell[CellGroupData[{
Cell[2654149, 62287, 231, 7, 48, "Input"],
Cell[2654383, 62296, 5186, 148, 142, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2659606, 62449, 231, 7, 48, "Input"],
Cell[2659840, 62458, 5186, 148, 142, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2665075, 62612, 37, 0, 71, "Subsection"],
Cell[CellGroupData[{
Cell[2665137, 62616, 381, 11, 48, "Input"],
Cell[2665521, 62629, 1079, 33, 58, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2666637, 62667, 1009, 32, 83, "Input"],
Cell[2667649, 62701, 2067, 65, 80, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2669753, 62771, 467, 14, 83, "Input"],
Cell[2670223, 62787, 1079, 33, 58, "Output"],
Cell[2671305, 62822, 1079, 33, 58, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2672421, 62860, 540, 16, 83, "Input"],
Cell[2672964, 62878, 1079, 33, 58, "Output"],
Cell[2674046, 62913, 1079, 33, 58, "Output"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}
]
*)

