(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 10.4' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     22157,        689]
NotebookOptionsPosition[     19771,        612]
NotebookOutlinePosition[     20888,        649]
CellTagsIndexPosition[     20629,        640]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[BoxData[
 RowBox[{
  RowBox[{"SetDirectory", "[", 
   RowBox[{"NotebookDirectory", "[", "]"}], "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "LiteRed`"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"**************** \"\>", "\[InvisibleSpace]", 
   StyleBox["\<\"LiteRed v1.82\"\>",
    StripOnInput->False,
    LineColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    FrontFaceColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    BackFaceColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    GraphicsColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]],
    FontWeight->Bold,
    FontColor->RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]]], 
   "\[InvisibleSpace]", "\<\" ********************\\nAuthor: Roman N. Lee, \
Budker Institute of Nuclear Physics, Novosibirsk.\\nRelease Date: \
01.06.2015\\nLiteRed stands for Loop InTEgrals REDuction.\\nThe package is \
designed for the search and application of the Integration-By-Part reduction \
rules. It also contains some other useful tools.\\nSee ?LiteRed`* for a list \
of functions. \"\>"}],
  SequenceForm["**************** ", 
   Style["LiteRed v1.82", {Bold, 
     RGBColor[1, 
      Rational[1, 3], 
      Rational[1, 3]]}], 
   " ********************\nAuthor: Roman N. Lee, Budker Institute of Nuclear \
Physics, Novosibirsk.\nRelease Date: 01.06.2015\nLiteRed stands for Loop \
InTEgrals REDuction.\nThe package is designed for the search and application \
of the Integration-By-Part reduction rules. It also contains some other \
useful tools.\nSee ?LiteRed`* for a list of functions. "],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Get", "[", "\"\<litered.ps3l1t1.dir/ps3l1t1\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Get", "[", "\"\<litered.ps3l1t2.dir/ps3l1t2\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Get", "[", "\"\<litered.ps3l1t3.dir/ps3l1t3\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Get", "[", "\"\<litered.ps3l1t4.dir/ps3l1t4\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Get", "[", "\"\<litered.ps3l1t5.dir/ps3l1t5\>\"", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ExecuteDefinitions", "[", "ps3l1t1", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ExecuteDefinitions", "[", "ps3l1t2", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ExecuteDefinitions", "[", "ps3l1t3", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ExecuteDefinitions", "[", "ps3l1t4", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ExecuteDefinitions", "[", "ps3l1t5", "]"}], ";"}]}], "Input"],

Cell[BoxData[
 StyleBox["\<\"1-loop basis with 3 external momenta.\\nUse \
Information[ps3l1t1] to get basis summary.\\nThis string can be redefined by \
setting ps3l1t1::usage=\\\"\[Ellipsis]\\\"\"\>", "MSG"]], "Print", \
"PrintUsage",
 CellTags->"Info523762677513-5589579"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"CheckDefinitions", "::", "warn"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Some external definitions used in \
'\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t1\\\"}]\\)\[NoBreak]' basis are \
missing. Probably, you should run \
ExecuteDefinitions[\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t1\\\"}]\\)\[NoBreak]].\
\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 StyleBox["\<\"1-loop basis with 3 external momenta.\\nUse \
Information[ps3l1t2] to get basis summary.\\nThis string can be redefined by \
setting ps3l1t2::usage=\\\"\[Ellipsis]\\\"\"\>", "MSG"]], "Print", \
"PrintUsage",
 CellTags->"Info533762677513-5589579"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"CheckDefinitions", "::", "warn"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Some external definitions used in \
'\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t2\\\"}]\\)\[NoBreak]' basis are \
missing. Probably, you should run \
ExecuteDefinitions[\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t2\\\"}]\\)\[NoBreak]].\
\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 StyleBox["\<\"1-loop basis with 3 external momenta.\\nUse \
Information[ps3l1t3] to get basis summary.\\nThis string can be redefined by \
setting ps3l1t3::usage=\\\"\[Ellipsis]\\\"\"\>", "MSG"]], "Print", \
"PrintUsage",
 CellTags->"Info543762677513-5589579"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"CheckDefinitions", "::", "warn"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Some external definitions used in \
'\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t3\\\"}]\\)\[NoBreak]' basis are \
missing. Probably, you should run \
ExecuteDefinitions[\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t3\\\"}]\\)\[NoBreak]].\
\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 StyleBox["\<\"1-loop basis with 3 external momenta.\\nUse \
Information[ps3l1t4] to get basis summary.\\nThis string can be redefined by \
setting ps3l1t4::usage=\\\"\[Ellipsis]\\\"\"\>", "MSG"]], "Print", \
"PrintUsage",
 CellTags->"Info553762677513-5589579"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"CheckDefinitions", "::", "warn"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Some external definitions used in \
'\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t4\\\"}]\\)\[NoBreak]' basis are \
missing. Probably, you should run \
ExecuteDefinitions[\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t4\\\"}]\\)\[NoBreak]].\
\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 StyleBox["\<\"1-loop basis with 3 external momenta.\\nUse \
Information[ps3l1t5] to get basis summary.\\nThis string can be redefined by \
setting ps3l1t5::usage=\\\"\[Ellipsis]\\\"\"\>", "MSG"]], "Print", \
"PrintUsage",
 CellTags->"Info563762677513-5589579"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"CheckDefinitions", "::", "warn"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Some external definitions used in \
'\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t5\\\"}]\\)\[NoBreak]' basis are \
missing. Probably, you should run \
ExecuteDefinitions[\[NoBreak]\\!\\(\\*RowBox[{\\\"ps3l1t5\\\"}]\\)\[NoBreak]].\
\"\>"}]], "Message", "MSG"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"masters", "=", 
  RowBox[{"Join", "[", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"MIs", "[", "ps3l1t1", "]"}], ",", "\[IndentingNewLine]", 
    RowBox[{"MIs", "[", "ps3l1t2", "]"}], ",", "\[IndentingNewLine]", 
    RowBox[{"MIs", "[", "ps3l1t3", "]"}], ",", "\[IndentingNewLine]", 
    RowBox[{"MIs", "[", "ps3l1t4", "]"}], ",", "\[IndentingNewLine]", 
    RowBox[{"MIs", "[", "ps3l1t5", "]"}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"j", "[", 
    RowBox[{"ps3l1t1", ",", "1", ",", "1", ",", "0", ",", "0"}], "]"}], ",", 
   RowBox[{"j", "[", 
    RowBox[{"ps3l1t2", ",", "1", ",", "0", ",", "1", ",", "0"}], "]"}], ",", 
   RowBox[{"j", "[", 
    RowBox[{"ps3l1t4", ",", "0", ",", "0", ",", "1", ",", "1"}], "]"}], ",", 
   RowBox[{"j", "[", 
    RowBox[{"ps3l1t4", ",", "1", ",", "1", ",", "1", ",", "1"}], "]"}], ",", 
   RowBox[{"j", "[", 
    RowBox[{"ps3l1t5", ",", "1", ",", "0", ",", "1", ",", "0"}], "]"}], ",", 
   RowBox[{"j", "[", 
    RowBox[{"ps3l1t5", ",", "1", ",", "1", ",", "1", ",", "1"}], "]"}]}], 
  "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ds13", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"masters", "//", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{"Dinv", "[", 
         RowBox[{"#", ",", "s13"}], "]"}], "&"}], "]"}]}], "//", 
     "IBPReduce"}], "//", 
    RowBox[{"ReplaceAll", "[", 
     RowBox[{"d", "\[Rule]", 
      RowBox[{"4", "-", 
       RowBox[{"2", "ep"}]}]}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ds23", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"masters", "//", 
      RowBox[{"Map", "[", 
       RowBox[{
        RowBox[{"Dinv", "[", 
         RowBox[{"#", ",", "s23"}], "]"}], "&"}], "]"}]}], "//", 
     "IBPReduce"}], "//", 
    RowBox[{"ReplaceAll", "[", 
     RowBox[{"d", "\[Rule]", 
      RowBox[{"4", "-", 
       RowBox[{"2", "ep"}]}]}], "]"}]}]}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ms13", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"Coefficient", "[", 
      RowBox[{"dm", ",", "m"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"dm", ",", "ds13"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"m", ",", "masters"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ms13", ".", "masters"}], "\[Equal]", "ds13"}]}], "Input"],

Cell[BoxData["True"], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"ms23", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"Coefficient", "[", 
      RowBox[{"dm", ",", "m"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"dm", ",", "ds23"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"m", ",", "masters"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ms23", ".", "masters"}], "\[Equal]", "ds23"}]}], "Input"],

Cell[BoxData["True"], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"ms13", "//", "MatrixForm"}], "\[IndentingNewLine]", 
 RowBox[{"ms23", "//", "MatrixForm"}]}], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["ep", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "s13", "+", "s23"}]]}], "0", "0", "0", 
      "0"},
     {"0", "0", "0", "0", "0", "0"},
     {
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}]}], 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", "s13", " ", "s23", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}]}], 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", "s13", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"s13", "+", "s23"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}], ")"}]}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "s13", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"s13", "+", "s23"}], ")"}]}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"2", " ", "ep"}], "+", 
         RowBox[{"2", " ", "s13"}], "-", 
         RowBox[{"4", " ", "s23"}], "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"4", "-", 
            RowBox[{"2", " ", "ep"}]}], ")"}], " ", "s23"}]}], 
        RowBox[{"2", " ", "s13", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]]}], "0", "0"},
     {"0", "0", "0", "0", 
      RowBox[{"-", 
       FractionBox["ep", "s13"]}], "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}], ")"}]}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "s13", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"s13", "+", "s23"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"2", " ", "s13"}], "+", "s23"}], ")"}]}], 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13"}], ")"}], " ", "s13", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"s13", "+", "s23"}], ")"}]}]], "0", 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}], ")"}]}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13"}], ")"}], " ", "s13", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", 
          RowBox[{"2", " ", "s13"}], "+", "s23"}], ")"}]}], 
       RowBox[{"2", " ", "s13", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", 
       FractionBox["ep", "s23"]}], "0", "0", "0", "0", "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["ep", 
        RowBox[{
         RowBox[{"-", "1"}], "+", "s13", "+", "s23"}]]}], "0", "0", "0", 
      "0"},
     {"0", "0", "0", "0", "0", "0"},
     {
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}], ")"}]}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s23"}], ")"}], " ", "s23", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]]}], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}], ")"}]}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "s23", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"s13", "+", "s23"}], ")"}]}]]}], 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", 
          RowBox[{"2", " ", "s23"}]}], ")"}]}], 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s23"}], ")"}], " ", "s23", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"s13", "+", "s23"}], ")"}]}]], 
      FractionBox[
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "2"}], "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", 
          RowBox[{"2", " ", "s23"}]}], ")"}]}], 
       RowBox[{"2", " ", "s23", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]], "0", "0"},
     {"0", "0", "0", "0", "0", "0"},
     {"0", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}]}], 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", "s23", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{"s13", "+", "s23"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", 
         RowBox[{"(", 
          RowBox[{"1", "-", 
           RowBox[{"2", " ", "ep"}]}], ")"}]}], 
        RowBox[{
         SuperscriptBox["Q2", "2"], " ", "s23", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}], " ", 
         RowBox[{"(", 
          RowBox[{"s13", "+", "s23"}], ")"}]}]]}], "0", 
      FractionBox[
       RowBox[{"2", " ", 
        RowBox[{"(", 
         RowBox[{"1", "-", 
          RowBox[{"2", " ", "ep"}]}], ")"}]}], 
       RowBox[{
        SuperscriptBox["Q2", "2"], " ", "s13", " ", "s23", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]], 
      RowBox[{"-", 
       FractionBox[
        RowBox[{
         RowBox[{"2", " ", "ep"}], "-", 
         RowBox[{"4", " ", "s13"}], "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{"4", "-", 
            RowBox[{"2", " ", "ep"}]}], ")"}], " ", "s13"}], "+", 
         RowBox[{"2", " ", "s23"}]}], 
        RowBox[{"2", " ", "s23", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", "s13", "+", "s23"}], ")"}]}]]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"masters", ",", "\"\<ps3l1.masters\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"ms13", ",", "\"\<ps3l1.ms13\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Put", "[", 
   RowBox[{"ms23", ",", "\"\<ps3l1.ms23\>\""}], "]"}], ";"}]}], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"To", " ", "find", " ", "the", " ", "epsilon"}], "-", 
     RowBox[{"form", " ", "transformation"}]}], ",", " ", 
    RowBox[{"run", ":", " ", 
     RowBox[{
      RowBox[{"fuchsia", " ", "reduce", " ", 
       RowBox[{"ps3l1", ".", "ms13"}]}], "-", 
      RowBox[{"x", " ", "s13", " ", 
       RowBox[{"ps3l1", ".", "ms23"}]}], "-", 
      RowBox[{"x", " ", "s23"}], "-", 
      RowBox[{"e", " ", "ep"}], "-", 
      RowBox[{"t", " ", 
       RowBox[{"ps3l1", ".", "ep", ".", "t"}]}]}]}]}], " ", "*)"}], 
  " "}]], "Input"]
},
WindowSize->{1920, 1046},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
FrontEndVersion->"10.4 for Linux x86 (64-bit) (April 11, 2016)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info523762677513-5589579"->{
  Cell[3354, 103, 275, 5, 81, "Print",
   CellTags->"Info523762677513-5589579"]},
 "Info533762677513-5589579"->{
  Cell[4016, 121, 275, 5, 81, "Print",
   CellTags->"Info533762677513-5589579"]},
 "Info543762677513-5589579"->{
  Cell[4678, 139, 275, 5, 81, "Print",
   CellTags->"Info543762677513-5589579"]},
 "Info553762677513-5589579"->{
  Cell[5340, 157, 275, 5, 81, "Print",
   CellTags->"Info553762677513-5589579"]},
 "Info563762677513-5589579"->{
  Cell[6002, 175, 275, 5, 81, "Print",
   CellTags->"Info563762677513-5589579"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info523762677513-5589579", 20072, 622},
 {"Info533762677513-5589579", 20185, 625},
 {"Info543762677513-5589579", 20298, 628},
 {"Info553762677513-5589579", 20411, 631},
 {"Info563762677513-5589579", 20524, 634}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[558, 20, 123, 3, 32, "Input"],
Cell[CellGroupData[{
Cell[706, 27, 52, 1, 32, "Input"],
Cell[761, 30, 1500, 37, 119, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2298, 72, 1053, 29, 231, "Input"],
Cell[3354, 103, 275, 5, 81, "Print",
 CellTags->"Info523762677513-5589579"],
Cell[3632, 110, 381, 9, 24, "Message"],
Cell[4016, 121, 275, 5, 81, "Print",
 CellTags->"Info533762677513-5589579"],
Cell[4294, 128, 381, 9, 24, "Message"],
Cell[4678, 139, 275, 5, 81, "Print",
 CellTags->"Info543762677513-5589579"],
Cell[4956, 146, 381, 9, 24, "Message"],
Cell[5340, 157, 275, 5, 81, "Print",
 CellTags->"Info553762677513-5589579"],
Cell[5618, 164, 381, 9, 24, "Message"],
Cell[6002, 175, 275, 5, 81, "Print",
 CellTags->"Info563762677513-5589579"],
Cell[6280, 182, 381, 9, 24, "Message"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6698, 196, 472, 9, 165, "Input"],
Cell[7173, 207, 656, 15, 32, "Output"]
}, Open  ]],
Cell[7844, 225, 835, 28, 55, "Input"],
Cell[CellGroupData[{
Cell[8704, 257, 420, 13, 55, "Input"],
Cell[9127, 272, 31, 0, 32, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[9195, 277, 420, 13, 55, "Input"],
Cell[9618, 292, 31, 0, 32, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[9686, 297, 127, 2, 55, "Input"],
Cell[9816, 301, 4476, 137, 174, "Output"],
Cell[14295, 440, 4476, 137, 174, "Output"]
}, Open  ]],
Cell[18786, 580, 370, 11, 77, "Input"],
Cell[19159, 593, 608, 17, 32, "Input"]
}
]
*)

